/***************************************************************************
                          add.h  -  description
                             -------------------
    begin                : Wed Apr 14 2004
    copyright            : (C) 2004 by Joern Ossowski
    email                : ossowski@cs.uni-bonn.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ADD_H
#define ADD_H

#include <string>

class ADDNode;
class ADDGroup;
class ADDVarOrder;
class ADDWriter;
class PtrHashMap2;
class PtrHashMap3;

template <class T1, class T2, class T3> class Reordering;

#include <add/addfunction.h>
#include <add/addfunctioncontainer.h>
#include <add/addvarorder.h>
#include <add/addgroup.h>
#include <common/constants.h>
#include <common/assignment.h>

class ADD{
friend class ADDNode;
friend class ADDFunction;
friend class ADDGroup;
friend class ADDWriter;
public:
  ADD(void);
  ~ADD(void);

  static bool addVariable(const std::string&);
  static bool clear(void);
  static void reset();
  static unsigned long size(void);
  static unsigned long countDeadNodes(void);
  static unsigned long PEAK(void);
	static void printStat(void);

  static ADDFunction loadFunctionBLIF(const std::string&, const std::string&);
  static ADDFunctionContainer loadFunctionsBLIF(const std::string&);
  static ADDFunction loadMatrix(const std::string&, const unsigned short=MATRIX_INPUT_DEFAULT);
  static bool storeFunction(const std::string&, const ADDFunction&);
  static bool storeFunctions(const std::string&, const ADDFunctionContainer&);

  static void sifting(const float=MAX_GROWTH_DEFAULT);
  static void variableSifting(const float=MAX_GROWTH_DEFAULT);
  static void groupSifting(const float=MAX_GROWTH_DEFAULT);
  static void windowPermutation2(void);
  static void windowPermutation3(void);
  static void windowPermutation4(void);
  static void geneticMinimize(void);

  struct CallBack {
    virtual ~CallBack() {}
    virtual bool operator()() {return false;}
  };
  static void groupExhaustive(CallBack& call_back);

  static void enableDynamicReordering(void);
  static void disableDynamicReordering(void);
  static void setDynamicReorderingMaxGrowth(float);
  static void setDynamicReorderingTimeout(unsigned int); // in ms
  static void setUniqueTableMaxLoadFactor(float);
  static void reorderingHook(bool forceReordering=false);
  

  static void garbageCollect(void);
  static unsigned long setGarbageCollectionDelay(void);
  static unsigned long setGarbageCollectionDelay(const unsigned long);

  ADDFunction projection(const std::string&);
  ADDFunction projection(const unsigned long);
  ADDFunction negate(const ADDFunction&);
  ADDFunction addition(const ADDFunction&, const ADDFunction&);
  ADDFunction sum(const ADDFunction&, const ADDFunction&);
  ADDFunction subtraction(const ADDFunction&, const ADDFunction&);
  ADDFunction multiplication(const ADDFunction&, const ADDFunction&);
  ADDFunction product(const ADDFunction&, const ADDFunction&);
  ADDFunction equivalenceClasses(const ADDFunction&, const ADDFunction&, const ADDFunction&);
	ADDFunction constrain(const ADDFunction&, const ADDFunction&);
  ADDFunction MM(const ADDFunction&, const ADDFunction&, const ADDFunction&);
  ADDFunction MV(const ADDFunction&, const ADDFunction&, const ADDFunction&);
  ADDFunction VM(const ADDFunction&, const ADDFunction&, const ADDFunction&);
  ADDFunction ITE(const ADDFunction&, const ADDFunction&, const ADDFunction&);
  ADDFunction NOT(const ADDFunction&);
  ADDFunction AND(const ADDFunction&, const ADDFunction&);
  ADDFunction OR(const ADDFunction&, const ADDFunction&);
  ADDFunction XOR(const ADDFunction&, const ADDFunction&);
  ADDFunction EQUIV(const ADDFunction&, const ADDFunction&);
  ADDFunction cofactor(const ADDFunction&, const unsigned long, const bool);
  ADDFunction cofactor(const ADDFunction&, const ADDFunction&);
  ADDFunction change(const ADDFunction&);
  ADDFunction change(const ADDFunction&, const unsigned long);
  ADDFunction change(const ADDFunction&, const ADDFunction&);
  ADDFunction exists(const ADDFunction&, const ADDFunction&);
  ADDFunction exists(const ADDFunction&, const unsigned long);
  ADDFunction forall(const ADDFunction&, const ADDFunction&);
  ADDFunction forall(const ADDFunction&, const unsigned long);
  ADDFunction less(const ADDFunction&, const ADDFunction&);
  ADDFunction lessEqual(const ADDFunction&, const ADDFunction&);
  ADDFunction equal(const ADDFunction&, const ADDFunction&);
  ADDFunction maximum(const ADDFunction&, const ADDFunction&);
  ADDFunction max(const ADDFunction&, const ADDFunction&);
  double max(const ADDFunction&);
  ADDFunction minimum(const ADDFunction&, const ADDFunction&);
  ADDFunction min(const ADDFunction&, const ADDFunction&);
  double min(const ADDFunction&);
  ADDFunction oneOver(const ADDFunction&);
  ADDFunction moveUp(const ADDFunction&, const unsigned int=1);
  ADDFunction moveUp(const ADDFunction&, const ADDFunction&, const unsigned int=1);
  ADDFunction moveDown(const ADDFunction&, const unsigned int=1);
  ADDFunction moveDown(const ADDFunction&, const ADDFunction&, const unsigned int=1);
  ADDFunction getAssignment(const ADDFunction&);
  void getAssignment(const ADDFunction&, char*) const;
  double countAssignments(const ADDFunction&) const;
  double countAssignments(const ADDFunction& function, const ADDFunction& cubeSet) const;
  ADDFunction getRandomAssignment(const ADDFunction& function, const ADDFunction& cubeSet);
  ADDFunction getPath(const ADDFunction&);
  void getPath(const ADDFunction&, char*) const;
protected:
  //basic reordering function
  bool swap(const unsigned long);

  //function for ADDNode
  void deleteSubTree(ADDNode*);
  void deleteNode(ADDNode*);
private:
	static unsigned long deadNodes;
  static ADDVariableOrdering variableOrdering;
  static PtrHashMap2 ArithmeticComputedTable;
  static PtrHashMap2 ExtremaComputedTable;
  static PtrHashMap2 AbstractComputedTable;
  static PtrHashMap2 CompareComputedTable;
  static PtrHashMap3 MatrixComputedTable;
  static PtrHashMap3 BoolComputedTable;
  static PtrHashMap3 MoveComputedTable;
  static PtrHashMap3 SpecialComputedTable;
  static Reordering<ADDVarOrder,ADDGroup,ADD>* reordering;
  static ADD* BDD;
  static unsigned long garbageCollectionDelay;
  static unsigned long reorderingThreshold;
  static unsigned long operatorCounter;
  static unsigned long lastReorderSize;
  static bool dynamicReordering;
  static float dynamicReorderingMaxGrowth;
  // timeout in milliseconds for sifting up/down per variable
  static unsigned int dynamicReorderingTimeout;
  static ADDNode* zeroDrain;
  static ADDNode* oneDrain;
  static ADDNode* (ADD::*currentAbstractFunction)(ADDNode*, ADDNode*);
  static short currentAbstractFunctionID;
  static unsigned long swapcount;

  ADDNode* abstract(ADDNode*, ADDNode*);
  ADDNode* idempotentAbstract(ADDNode*, ADDNode*);
  ADDNode* exists(ADDNode*, ADDNode*);
  ADDNode* negate(const ADDNode*);
  ADDNode* addition(ADDNode*, ADDNode*);
	ADDNode* constAddition(ADDNode*, ADDNode*);
  ADDNode* subtraction(ADDNode*, ADDNode*);
  ADDNode* multiplication(ADDNode*, ADDNode*);
	ADDNode* constMultiplication(ADDNode*, ADDNode*);
  ADDNode* MM(ADDNode*, ADDNode*, ADDNode*, unsigned long);
  ADDNode* MV(ADDNode*, ADDNode*, ADDNode*, unsigned long);
  ADDNode* VM(ADDNode*, ADDNode*, ADDNode*, unsigned long);
  ADDNode* equivalenceClasses(ADDNode*, ADDNode*, ADDNode*);
	ADDNode* constrain(ADDNode*, ADDNode*);
  ADDNode* ITE(ADDNode*, ADDNode*, ADDNode*);
  ADDNode* AND(ADDNode*, ADDNode*);
  ADDNode* OR(ADDNode*, ADDNode*);
  ADDNode* NOT(ADDNode*);
  ADDNode* cofactor(ADDNode*, ADDNode*);
  ADDNode* change(ADDNode*);
  ADDNode* change(ADDNode*, ADDNode*);
  ADDNode* less(ADDNode*, ADDNode*);
  ADDNode* lessEqual(ADDNode*, ADDNode*);
  ADDNode* equal(ADDNode*, ADDNode*);
  ADDNode* max(ADDNode*, ADDNode*);
  double max(ADDNode*, map<ADDNode*,double>&);
  ADDNode* min(ADDNode*, ADDNode*);
  double min(ADDNode*, map<ADDNode*,double>&);
  ADDNode* oneOver(ADDNode*);
  ADDNode* moveUp(ADDNode*);
  ADDNode* moveUp(ADDNode*, const unsigned int);
  ADDNode* moveUp(ADDNode*, ADDNode*);
  ADDNode* moveUp(ADDNode*, ADDNode*, const unsigned int);
  ADDNode* moveDown(ADDNode*);
  ADDNode* moveDown(ADDNode*, const unsigned int);
  ADDNode* moveDown(ADDNode*, ADDNode*);
  ADDNode* moveDown(ADDNode*, ADDNode*, const unsigned int);
  ADDNode* getAssignment(const unsigned long, const ADDNode*);
  bool getAssignment(const ADDNode*, Assignment&) const;
  bool getAssignment(const ADDNode*, char*) const;
  double countAssignments(const ADDNode*, map<ADDNode*,double>&) const;
  double countAssignments(const ADDNode*, const ADDNode*, map<pair<const ADDNode*,const ADDNode*>,double>&) const;
  ADDNode* getPath(const ADDNode*);
  bool getPath(const ADDNode*, char*) const;

  ADDNode* createConstantFunction(const double);
  inline ADDFunction createFunction(ADDNode*);
  ADDNode* findOrAdd(const unsigned long, ADDNode*, ADDNode*);
  inline ADDNode* findOrAdd(ADDUniqueTable*, ADDNode*, ADDNode*);
  ADDNode* findOrAddDrain(const double);
  inline ADDNode* findDrain(const double);
};
#endif
