/***************************************************************************
                          assignment.cpp  -  description
                             -------------------
    begin                : Tue Feb 10 2004
    copyright            : (C) 2004 by Joern Ossowski
    email                : ossowski@cs.uni-bonn.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "assignment.h"
#include <algorithm>

using namespace std;

/**
 * \brief Constructor.
 */
Assignment::Assignment(void){
}

/**
 * \brief Copies the given Assignment.
 */
Assignment::Assignment(const Assignment& assign){
  assignment=assign.assignment;
  stringVector=assign.stringVector;
}

/**
 * \brief Destructor.
 */
Assignment::~Assignment(void){
}

/**
 * \brief Clears the Assignment.
 */
void Assignment::clear(void){
  assignment.clear();
  stringVector.clear();
}

/**
 * \brief Returns if the given variable can be found inside the Assignment.
 */
bool Assignment::find(const string& variable) const {
  return (assignment.find(variable)!=assignment.end());
}

/**
* \brief Returns the reversed assignment (a copy).
 */
Assignment Assignment::reverse(void) const {
  Assignment ret;
	ret.assignment=assignment;
	ret.stringVector=vector<string>(stringVector.size());
	std::reverse_copy(stringVector.begin(), stringVector.end(), ret.stringVector.begin());	
  return ret;
}

/**
 * \brief Multiplies two Assignments.
 *
 * This function can be used for multiplying two Assignments. The result is
 * the multiplication of every element.
 * Every not existing element will be interpreted as zero.
 *
 * \code
 * Assignment a;
 * a["x0"]=1;
 * a["x1"]=1;
 * a["x2"]=1;
 *
 * Assignment b;
 * b["x0"]=0;
 * b["x1"]=1;
 * b["y0"]=0;
 *
 * Assignment c=a*b;
 * //c["x0"]=1*0=0
 * //c["x1"]=1*1=1
 * //c["x2"]=1*0=0 (x2 does not exist in b)
 * //c["y0"]=0*0=0 (y0 does not exist in a)
 * \endcode
 */
Assignment Assignment::operator*(const Assignment& assign) const {
  Assignment a;
  AssignmentPair aP;

  for(unsigned long i=0;i<size();++i){
    aP=(*this)[i];
    a[aP.variable]=aP.value;
  }
  for(unsigned long i=0;i<assign.size();++i){
    aP=assign[i];
    a[aP.variable]&=aP.value;
  }

  return a;
}

/**
 * \brief Adds two Assignments.
 *
 * This function can be used for adding two Assignments. The result is
 * the addition of every element. Note that in a boolean assignment 1+1=1.
 * Every not existing element will be interpreted as zero.
 *
 * \code
 * Assignment a;
 * a["x0"]=0;
 * a["x1"]=1;
 * a["x2"]=1;
 *
 * Assignment b;
 * b["x0"]=0;
 * b["x1"]=0;
 * b["x2"]=1;
 * b["y0"]=1;
 *
 * Assignment c=a+b;
 * //c["x0"]=0+0=0
 * //c["x1"]=1+0=1
 * //c["x2"]=1+1=1
 * //c["y0"]=0+1=1  (y0 does not exist in a)
 * \endcode
 */
Assignment Assignment::operator+(const Assignment& assign) const {
  Assignment a;
  AssignmentPair aP;

  for(unsigned long i=0;i<size();++i){
    aP=(*this)[i];
    a[aP.variable]=aP.value;
  }
  for(unsigned long i=0;i<assign.size();++i){
    aP=assign[i];
    a[aP.variable]|=aP.value;
  }

  return a;
}

/**
 * \brief Copies an Assignment.
 */
Assignment& Assignment::operator=(const Assignment& assign){
  if(&assign!=this){
    assignment=assign.assignment;
    stringVector=assign.stringVector;
  }

  return *this;
}

/**
 * \brief Encodes a number into a boolean assignment.
 *
 * Note that the size of the current Assignment defines the length of the boolean encoding.
 *
 * \code
 * //init Assignment a
 * Assignment a;
 * a["x0"];
 * a["x1"];
 * a["x2"];
 *
 * a=14;
 * //the boolean encoding of 14 equals 1110
 * //a["x2"]=1
 * //a["x1"]=1
 * //a["x0"]=0
 * \endcode
 */
Assignment& Assignment::operator=(const unsigned long number){
  for(unsigned long i=0;i<size();++i) assignment[stringVector[i]]=number&(1<<i);

  return *this;
}

/**
 * \brief Calculates the boolean negation of the Assignment.
 */
Assignment& Assignment::operator!(void){
  for(unsigned long i=0;i<size();++i) assignment[stringVector[i]]=!assignment[stringVector[i]];

  return *this;
}

/**
 * \brief Accesses the i-th element of an Assignment.
 *
 * Returns the AssignmentPair from the given position.
 * Note that changing the values inside the AssignmentPair does not change the values inside
 * the Assignment. This pair is just a copy.
 * \todo Replace [ERROR] with =ERROR=.
 */
AssignmentPair Assignment::operator[](const unsigned long index) const {
  AssignmentPair ret;
  if(index>=assignment.size()){
    ret.variable="[ERROR]";
    ret.value=false;
  } else {
    ret.variable=stringVector.at(index);
    ret.value=at(ret.variable);
  }

  return ret;
}

/**
 * \brief Returns the boolean assignment of an Assignment.
 *
 * Returns the assignment of the given variable.
 * Note that changing the returned values changes the value inside
 * the Assignment.
 */
bool& Assignment::operator[](const string& variable){
  if(!find(variable)) stringVector.push_back(variable);

  return assignment[variable];
}

/**
 * \brief Returns the boolean assignment of an Assignment.
 *
 * Returns the assignment of the given variable.
 * Note that changing the returned values does not change the value inside
 * the Assignment.
 */
bool Assignment::at(const string& variable) const {
  return assignment.find(variable)->second;
}

/**
 * \brief Removes a variable from the Assignment.
 */
bool Assignment::remove(const string& variable){
  map<string,bool>::iterator it1=assignment.find(variable);
  if(it1==assignment.end()) return false;

  assignment.erase(it1);
  vector<string>::iterator it2=stringVector.begin();
  while(*it2!=variable) ++it2;
  stringVector.erase(it2);

  return true;
}

/**
 * \brief Returns the number of variables inside the Assignment.
 */
unsigned long Assignment::size(void) const {
  return assignment.size();
}

//*****************non class functions******************
/**
 * \brief Calculates the boolean encoding of the increased Assignment.
 */
Assignment& operator++(Assignment& assignment){
  unsigned long i=0;
  while((i<assignment.size()) && (assignment[i].value)){
    assignment[assignment[i].variable]=0;
    ++i;
  }
  if(i<assignment.size()) assignment[assignment[i].variable]=1;

  return assignment;
}

/**
 * \brief Calculates the boolean encoding of the decreased Assignment.
 */
Assignment& operator--(Assignment& assignment){
  unsigned long i=0;
  while((i<assignment.size()) && (!assignment[i].value)){
    assignment[assignment[i].variable]=1;
    ++i;
  }
  if(i<assignment.size()) assignment[assignment[i].variable]=0;

  return assignment;
}
