/***************************************************************************
                          hashdrainuniquetable.h  -  description
                             -------------------
    begin                : Sat Mar 29 2008
    copyright            : (C) 2008 by Joern Ossowski
    email                : mail@jossowski.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef HASHDRAINUNIQUETABLE_H
#define HASHDRAINUNIQUETABLE_H

#include <climits>

#include <common/helper.h>

#include <vector>
#include <boost/dynamic_bitset.hpp>

inline unsigned long getHash(double value){
  unsigned long iValue=static_cast<unsigned long>(value*1024.0);
  return Common::hashFunction<unsigned long>(iValue/1024,iValue%1024);
}

//******************HashDrainUniqueTable********************
template <class T>
class HashDrainUniqueTable {
public:
  HashDrainUniqueTable(void);
  unsigned long size(void) const {return count;}
  inline void resetDeadNodeCount(void){deadNodes=0;}
  inline void incDeadNodeCount(void){++deadNodes;}
  inline void decDeadNodeCount(void){--deadNodes;}
  inline unsigned long getDeadNodes(void) const {return deadNodes;}
  inline bool remove(const T*);
  inline T* findOrAddValue(double);
  T* getFirst(void){
    if(count==0) return 0;

    currentPos=bitSet.find_first();
    if(currentPos==boost::dynamic_bitset<>::npos) return 0;
    return uniqueTable[currentPos];
  }
  T* next(void){
    currentPos=bitSet.find_next(currentPos);
    if(currentPos==boost::dynamic_bitset<>::npos) return 0;
    return uniqueTable[currentPos];
  }
  inline unsigned long getLevel(void) const {return ULONG_MAX;}
  inline unsigned long getRealLevel(void) const {return ULONG_MAX;}
  inline static void setInitHashSize(unsigned int newInit){initHashSize=newInit;}
  inline static void setMaxHashSize(unsigned int newMax){maximumHashSize=newMax;}
private:
  static unsigned int initHashSize;
  static unsigned int maximumHashSize;
  unsigned long count;
  unsigned long maxCollisionListSize;
  unsigned int hashSize;
  unsigned int maxHashSize;
  unsigned short shift;
  unsigned short collisionListSizeLimit;
  unsigned long deadNodes;

  boost::dynamic_bitset<> bitSet;
  boost::dynamic_bitset<>::size_type currentPos;

  std::vector<T*> uniqueTable;

  void resize(void);
};
template <class T> unsigned int HashDrainUniqueTable<T>::initHashSize=131072;
template <class T> unsigned int HashDrainUniqueTable<T>::maximumHashSize=262144;

template <class T>
HashDrainUniqueTable<T>::HashDrainUniqueTable(void) : count(0), maxCollisionListSize(0), deadNodes(0) {
  collisionListSizeLimit=12;

  hashSize=2;
  //shift for smallest hashmap
  shift=sizeof(unsigned long)*8-1;
  while((hashSize<<1)<=initHashSize){
    hashSize<<=1;
    --shift;
  }
  maxHashSize=maximumHashSize;
  bitSet.resize(hashSize,false);
  uniqueTable.resize(hashSize,0);

  currentPos=boost::dynamic_bitset<>::npos;
}

template <class T>
bool HashDrainUniqueTable<T>::remove(const T* element){
  unsigned long index=getHash(element->getValue())>>shift;

  T* tempNode=uniqueTable[index];
  T* lastNode=0;
  while(tempNode && tempNode!=element){
    lastNode=tempNode;
    tempNode=tempNode->getNext();
  }
  if(tempNode!=element) return false;

  if(lastNode==0){
    //change entry in the unique table and check if the value is 0
    if((uniqueTable[index]=tempNode->getNext())==0) bitSet[index]=false;
  } else lastNode->setNext(tempNode->getNext());
  tempNode->setNext(0);
  --count;

  return true;
}

template <class T>
T* HashDrainUniqueTable<T>::findOrAddValue(double value){
  unsigned long index=getHash(value)>>shift;

  unsigned long collisionListSize=0;
  T* tempNode=uniqueTable[index];
  T* lastNode=0;
  while(tempNode && tempNode->getValue()<value){
    ++collisionListSize;
    lastNode=tempNode;
    tempNode=tempNode->getNext();
  }
  if(tempNode && !(value<tempNode->getValue())) return tempNode;
  T* result=T::getNewNode();
  result->setValue(value);

  if(lastNode==0){
    result->setNext(uniqueTable[index]);
    uniqueTable[index]=(T*) result;
    bitSet[index]=true;
  } else {
    result->setNext(lastNode->getNext());
    lastNode->setNext(result);
  }
  ++count;
  if(collisionListSize>maxCollisionListSize){
    maxCollisionListSize=collisionListSize;
    if(maxCollisionListSize>collisionListSizeLimit){
      if((count>=(hashSize*collisionListSizeLimit)/4)&&(hashSize<maxHashSize)) resize();
    }
  }

  return result;
}

template <class T>
void HashDrainUniqueTable<T>::resize(void){
  std::vector<T*> tempUniqueTable(2*hashSize,0);
  boost::dynamic_bitset<> tempBitSet(2*hashSize,false);

  --shift;
  maxCollisionListSize=0;

  //transfer all nodes from old unique table to new one
  T* currentTable=(T*) getFirst();
  T* currentNode;
  T* tempNode;
  T* evenTable;
  T* oddTable;
  unsigned long index;
  unsigned long evenCollisionListSize;
  unsigned long oddCollisionListSize;
  while(currentTable){
    currentNode=currentTable;
    tempNode=currentNode->getNext();
    //shift is already updated so that the shift has to be reverted
    index=getHash(currentNode->getValue())>>(shift+1);

    evenTable=0;
    oddTable=0;

    evenCollisionListSize=0;
    oddCollisionListSize=0;
    while(currentNode){
      tempNode=currentNode->getNext();
      //shift is already updated
      if((getHash(currentNode->getValue())>>shift)&1){
        if(oddTable!=0) oddTable->setNext(currentNode);
        else tempUniqueTable[2*index+1]=currentNode;
        oddTable=currentNode;
        ++oddCollisionListSize;
      } else {
        if(evenTable!=0) evenTable->setNext(currentNode);
        else tempUniqueTable[2*index]=currentNode;
        evenTable=currentNode;
        ++evenCollisionListSize;
      }

      //--------------------------------------
      currentNode=tempNode;
    }

    if(oddTable!=0){
      oddTable->setNext(0);
      tempBitSet[2*index+1]=true;
    } else tempUniqueTable[2*index+1]=0;
    if(evenTable!=0){
      evenTable->setNext(0);
      tempBitSet[2*index]=true;
    } else tempUniqueTable[2*index]=0;

    if(oddCollisionListSize>maxCollisionListSize) maxCollisionListSize=oddCollisionListSize;
    if(evenCollisionListSize>maxCollisionListSize) maxCollisionListSize=evenCollisionListSize;

    currentTable=(T*) next();
  }

  uniqueTable.swap(tempUniqueTable);
  bitSet.swap(tempBitSet);

  hashSize*=2;
  ++collisionListSizeLimit;
  currentPos=boost::dynamic_bitset<>::npos;
}

#endif
