/***************************************************************************
                          hashuniquetable.h  -  description
                             -------------------
    begin                : Mon Nov 28 2005
    copyright            : (C) 2005 by Joern Ossowski
    email                : ossowski@cs.uni-bonn.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef HASHUNIQUETABLE_H
#define HASHUNIQUETABLE_H

#include <cstring>
#include "helper.h"

const unsigned short nBits=sizeof(unsigned long)*8;

//******************HashUniqueTable********************
template <class T>
class HashUniqueTable {
public:
  HashUniqueTable(const unsigned long=0);
  ~HashUniqueTable(void);
  unsigned long size(void) const {
    return count;
  }
  inline void resetDeadNodeCount(void){
    deadNodes=0;
  }
  inline void incDeadNodeCount(void){
    ++deadNodes;
  }
  inline void decDeadNodeCount(void){
    --deadNodes;
  }
  inline unsigned long getDeadNodes(void) const {
    return deadNodes;
  }
  float get_load_factor() const;
  inline void add(T*);
  inline void addValue(T*);
  inline bool remove(const T*);
  inline T* get(const T*, const T*) const;
  inline T* findOrAdd(T*,T*);
  inline T* getValue(const T*) const;
  inline T* findOrAddValue(T*);
  inline bool contains(const T* node1, const T* node2) const {
    return (get(node1,node2)!=0);
  }
  T* getFirst(void){
    if(count==0) return (T*) 0;
    
    //set bit pointer
    unsigned long* bitPointer=bitMask;
    while(bitPointer!=EOM){
      if(*bitPointer!=0){
	unsigned long value=*bitPointer;
	unsigned short i=0;
	while((value&1)==0){
	  ++i;
	  value>>=1;
	}
	currentSlot=uniqueTable+(bitPointer-bitMask)*nBits+i;
	return *currentSlot;
      }
      ++bitPointer;
    }
    
    currentSlot=EOT;
    return (T*) 0;
  }
  T* next(void){
    //calculate position of next slot
    unsigned long index=(currentSlot-uniqueTable)+1;
    unsigned short i=(unsigned short)(index%nBits);
	
    unsigned long* bitPointer;
    if(i!=0){
      bitPointer=bitMask+index/nBits;
      if((*bitPointer)>>i){
	unsigned long value=(*bitPointer)>>i;
	while((value&1)==0){
	  ++i;
	  value>>=1;
	}
	currentSlot=uniqueTable+nBits*(index/nBits)+i;
	return *currentSlot;
      }
      //no element in current bitfield, change to next
      index+=nBits;
    }
    index/=nBits;
    bitPointer=bitMask+index;
    
    while(bitPointer!=EOM){
      if(*bitPointer!=0){
	unsigned long value=*bitPointer;
	i=0;
	while((value&1)==0){
	  ++i;
	  value>>=1;
	}
	currentSlot=uniqueTable+(bitPointer-bitMask)*nBits+i;
	return *currentSlot;
      }
      ++bitPointer;
    }
    
    currentSlot=EOT;
    return (T*) 0;
  }
  void setLevel(const unsigned long newLevel){
    level=newLevel;
  }
  inline unsigned long getLevel(void) const {
    return level-skip;
  }
  inline unsigned long getRealLevel(void) const {
    return level;
  }
  inline bool deleteable(void) const {
    return (count==0);
  }
  inline unsigned long hashFunction(const T*, const T*) const;
  inline static void setSkip(unsigned long newSkip){
    skip=newSkip;
  }
  inline static void setInitHashSize(unsigned int newInit){
    initHashSize=newInit;
  }
  inline static void setMaxHashSize(unsigned int newMax){
    maximumHashSize=newMax;
  }
  inline static unsigned long getSkip(void){
    return skip;
  }
  inline void setDrainLevel(void){
    drainLevel=true;
  }
  inline bool isDrainLevel(void) const {
    return drainLevel;
  }

public:
  static float max_load_factor;
private:
  static unsigned int initHashSize;
  static unsigned int maximumHashSize;
  static unsigned long skip;
  unsigned long level;
  unsigned long count;
  unsigned long maxCollisionListSize;
  unsigned int hashSize;
  unsigned int maxHashSize;
  unsigned short shift;
  unsigned short collisionListSizeLimit;
  unsigned long deadNodes;

  T** uniqueTable;
  T** EOT;
  T** currentSlot;
  unsigned long* bitMask;
  unsigned long* EOM;
  bool drainLevel;

  inline void check_for_resize(unsigned long collisions);
  void resize(void);
};

template <class T>
unsigned long HashUniqueTable<T>::skip=0;

template <class T>
unsigned int HashUniqueTable<T>::initHashSize=2048;

template <class T>
unsigned int HashUniqueTable<T>::maximumHashSize= 1<<30;

template <class T> 
float HashUniqueTable<T>::max_load_factor=1.0f;


template <class T>
HashUniqueTable<T>::HashUniqueTable(const unsigned long newLevel){
  level=newLevel;
  count=0;
  maxCollisionListSize=0;
  collisionListSizeLimit=12;

  hashSize=2;
  //shift for smallest hashmap
  shift=sizeof(unsigned long)*8-1;
  while((hashSize<<1)<=initHashSize){
    hashSize<<=1;
    --shift;
  }
  maxHashSize=maximumHashSize;
  
  uniqueTable=new T*[hashSize];
  memset(uniqueTable,0,hashSize*sizeof(T*));
  EOT=uniqueTable+hashSize;
  bitMask=new unsigned long[hashSize/nBits];
  memset(bitMask,0,hashSize/8);
  EOM=bitMask+hashSize/nBits;
  currentSlot=EOT;
  deadNodes=0;
  drainLevel=false;
}

template <class T>
HashUniqueTable<T>::~HashUniqueTable(void){
  delete[] uniqueTable;
  delete[] bitMask;
}

template <class T>
void HashUniqueTable<T>::add(T* result){
  const T* node1=result->getSucc0();
  const T* node2=result->getSucc1();

  const unsigned long index=hashFunction(node1,node2);
  unsigned long collisionListSize=0;

  T* tempNode=uniqueTable[index];
  T* lastNode=0;
  if(tempNode!=0){
    while(tempNode->getSucc0()<node1){
      lastNode=tempNode;
      tempNode=tempNode->getNext();
      ++collisionListSize;
      if(tempNode==0) break;
    }
    if(tempNode!=0){
      while((tempNode->getSucc0()==node1)&&(tempNode->getSucc1()<node2)){
        lastNode=tempNode;
        tempNode=tempNode->getNext();
        ++collisionListSize;
        if(tempNode==0) break;
      }
    }
  }
  //no check if entry is already there
  if(lastNode==0){
    result->setNext(uniqueTable[index]);
    uniqueTable[index]=(T*) result;
    bitMask[index/nBits]|=1<<(index%nBits);
  } else {
    result->setNext(lastNode->getNext());
    lastNode->setNext(result);
  }
  ++count;

  check_for_resize(collisionListSize);
}

template <class T>
void HashUniqueTable<T>::addValue(T* result){
  const unsigned long index=hashFunction(result->getSucc0(),result->getSucc1());
  unsigned long collisionListSize=0;

  T* tempNode=uniqueTable[index];
  T* lastNode=0;
  if(tempNode!=0){
    while((*tempNode)<(*result)){
      lastNode=tempNode;
      tempNode=tempNode->getNext();
      ++collisionListSize;
      if(tempNode==0) break;
    }
  }
  //no check if entry is already there
  if(lastNode==0){
    result->setNext(uniqueTable[index]);
    uniqueTable[index]=(T*) result;
    bitMask[index/nBits]|=1<<(index%nBits);
  } else {
    result->setNext(lastNode->getNext());
    lastNode->setNext(result);
  }
  ++count;

  check_for_resize(collisionListSize);
}

template <class T>
bool HashUniqueTable<T>::remove(const T* value){
  const unsigned long index=hashFunction(value->getSucc0(),value->getSucc1());
	
  T* tempNode=uniqueTable[index];
  T* lastNode=0;
  if(tempNode!=0){
    while(tempNode!=value){
      lastNode=tempNode;
      tempNode=tempNode->getNext();
      if(tempNode==0) break;
    }
  }
  if(tempNode!=value) return false;
	
  if(lastNode==0){
    //change entry in the unique table and check if the value is 0
    if((uniqueTable[index]=tempNode->getNext())==0) bitMask[index/nBits]&=~(1<<(index%nBits));
  } else lastNode->setNext(tempNode->getNext());
  
  tempNode->setNext(0);
  --count;
  return true;
}

template <class T>
T* HashUniqueTable<T>::get(const T* node1, const T* node2) const {
  T* tempNode=uniqueTable[hashFunction(node1,node2)];
  T* lastNode=0;
  if(tempNode!=0){
    while(tempNode->getSucc0()<node1){
      lastNode=tempNode;
      tempNode=tempNode->getNext();
      if(tempNode==0) break;
    }
    if(tempNode!=0){
      while((tempNode->getSucc0()==node1)&&(tempNode->getSucc1()<node2)){
        lastNode=tempNode;
        tempNode=tempNode->getNext();
        if(tempNode==0) break;
      }
    }
  }
  if(tempNode!=0){
    if((tempNode->getSucc0()==node1)&&(tempNode->getSucc1()==node2)) return tempNode;
  }
  
  return 0;
}

template <class T>
T* HashUniqueTable<T>::findOrAdd(T* node1, T* node2){
  const unsigned long index=hashFunction(node1,node2);
	
  unsigned long collisionListSize=0;
  T* tempNode=uniqueTable[index];
  T* lastNode=0;
  while((tempNode!=0) && (tempNode->getSucc0()<node1)){
    ++collisionListSize;
    lastNode=tempNode;
    tempNode=tempNode->getNext();
  }
  while((tempNode!=0) && (tempNode->getSucc0()==node1)&&(tempNode->getSucc1()<node2)){
    ++collisionListSize;
    lastNode=tempNode;
    tempNode=tempNode->getNext();
  }
  if((tempNode!=0)&&(tempNode->getSucc0()==node1)&&(tempNode->getSucc1()==node2)) return tempNode;
  T* result=T::getNewNode();
  result->initNode(this,node1,node2);
  if(lastNode==0){
    result->setNext(uniqueTable[index]);
    uniqueTable[index]=(T*) result;
    bitMask[index/nBits]|=1<<(index%nBits);
  } else {
    result->setNext(lastNode->getNext());
    lastNode->setNext(result);
  }
  ++count;

  check_for_resize(collisionListSize);
  
  return result;
}

template <class T>
T* HashUniqueTable<T>::getValue(const T* value) const {
  const unsigned long index=hashFunction(value->getSucc0(),value->getSucc1());
  
  T* tempNode=uniqueTable[index];
  T* lastNode=0;
  while((tempNode!=0)&&((*tempNode)<(*value))){
    lastNode=tempNode;
    tempNode=tempNode->getNext();
  }
  if(tempNode!=0){
    if((*value)<(*tempNode)) return 0;
    else return tempNode;
  }
  
  return 0;
}

template <class T>
T* HashUniqueTable<T>::findOrAddValue(T* value){
  const unsigned long index=hashFunction(value->getSucc0(),value->getSucc1());
	
  unsigned long collisionListSize=0;
  T* tempNode=uniqueTable[index];
  T* lastNode=0;
  while((tempNode!=0)&&((*tempNode)<(*value))){
    ++collisionListSize;
    lastNode=tempNode;
    tempNode=tempNode->getNext();
  }
  if(tempNode!=0){
    if(!((*value)<(*tempNode))) return tempNode;
  }
  T* result=T::getNewNode();
  result->setValue(value->getValue());
  result->setDrainTable(this);
	
  if(lastNode==0){
    result->setNext(uniqueTable[index]);
    uniqueTable[index]=(T*) result;
    bitMask[index/nBits]|=1<<(index%nBits);
  } else {
    result->setNext(lastNode->getNext());
    lastNode->setNext(result);
  }
  ++count;

  check_for_resize(collisionListSize);
  
  return result;
}

template <class T>
unsigned long HashUniqueTable<T>::hashFunction(const T* node1, const T* node2) const {
  return ((((unsigned long) node1*12582917+(unsigned long) node2)*4256249)>>shift);
}

template <class T> 
float HashUniqueTable<T>::get_load_factor() const {
  return ((float)count) / hashSize;
}

template <class T>
void HashUniqueTable<T>::check_for_resize(unsigned long collisions) {
  if(collisions>maxCollisionListSize){
    maxCollisionListSize=collisions;
  }

  if (UNLIKELY(get_load_factor() >= max_load_factor) && 
      hashSize<maxHashSize) {
    resize();
  }
}

template <class T>
void HashUniqueTable<T>::resize(void){
  T** tempUniqueTable=new T*[2*hashSize];
  //out of memory
  if(tempUniqueTable==0) return;
  unsigned long* tempBitMask=new unsigned long[(2*hashSize)/nBits];
  //out of memory
  if(tempBitMask==0){
    delete[] tempUniqueTable;
    return;
  }
  
  --shift;
  memset(tempUniqueTable,0,2*hashSize*sizeof(T*));
  memset(tempBitMask,0,(2*hashSize)/8);
  maxCollisionListSize=0;
  
  //transfer all nodes from old unique table to new one
  T* currentTable=(T*) getFirst();
  T* currentNode;
  T* tempNode;
  T* evenTable;
  T* oddTable;
  unsigned long index;
  unsigned long evenCollisionListSize;
  unsigned long oddCollisionListSize;
  while(currentTable){
    currentNode=currentTable;
    tempNode=currentNode->getNext();
    //shift is already updated so that the shift has to be reverted
    index=hashFunction(currentNode->getSucc0(),currentNode->getSucc1())>>1;
    
    evenTable=0;
    oddTable=0;
    
    evenCollisionListSize=0;
    oddCollisionListSize=0;
    while(currentNode){
      tempNode=currentNode->getNext();
      //shift is already updated
      if(hashFunction(currentNode->getSucc0(),currentNode->getSucc1())&1){
	if(oddTable!=0) oddTable->setNext(currentNode);
	else tempUniqueTable[2*index+1]=currentNode;
        oddTable=currentNode;
	++oddCollisionListSize;
      } else {
	if(evenTable!=0) evenTable->setNext(currentNode);
	else tempUniqueTable[2*index]=currentNode;
        evenTable=currentNode;
	++evenCollisionListSize;
      }
      
      //--------------------------------------
      currentNode=tempNode;
    }
    
    if(oddTable!=0){
      oddTable->setNext(0);
      tempBitMask[(2*index+1)/nBits]|=1<<((2*index+1)%nBits);
    } else tempUniqueTable[2*index+1]=0;
    if(evenTable!=0){
      evenTable->setNext(0);
      tempBitMask[2*index/nBits]|=1<<((2*index)%nBits);
    } else tempUniqueTable[2*index]=0;
    
    if(oddCollisionListSize>maxCollisionListSize) maxCollisionListSize=oddCollisionListSize;
    if(evenCollisionListSize>maxCollisionListSize) maxCollisionListSize=evenCollisionListSize;

    currentTable=(T*) next();
  }
  
  //updating informations
  delete[] uniqueTable;
  delete[] bitMask;
  
  uniqueTable=tempUniqueTable;
  bitMask=tempBitMask;  
  
  hashSize*=2;
  ++collisionListSizeLimit;
  EOT=uniqueTable+hashSize;
  EOM=bitMask+hashSize/nBits;
  currentSlot=EOT;
}

//******************HashUniqueTable2********************
template <class T1, class T2>
class HashUniqueTable2 {
public:
  HashUniqueTable2(const unsigned long=0);
  ~HashUniqueTable2(void);
  unsigned long size(void) const {
    return count;
  }
  inline void resetDeadNodeCount(void){
    deadNodes=0;
  }
  inline void incDeadNodeCount(void){
    ++deadNodes;
  }
  inline void decDeadNodeCount(void){
    --deadNodes;
  }
  inline unsigned long getDeadNodes(void){
    return deadNodes;
  }
  inline void add(T1*);
  inline bool remove(const T1*);
  inline T1* get(const T2&, const T2&) const;
  inline T1* findOrAdd(const T2&, const T2&);
  inline bool contains(const T2& nodeInfo1, const T2& nodeInfo2) const {
    return (get(nodeInfo1,nodeInfo2)!=0);
  }
  T1* getFirst(void){
    if(count==0) return (T1*) 0;
    
    //set bit pointer
    unsigned long* bitPointer=bitMask;
    while(bitPointer!=EOM){
      if(*bitPointer!=0){
	unsigned long value=*bitPointer;
	unsigned short i=0;
	while((value&1)==0){
	  ++i;
	  value>>=1;
	}
	currentSlot=uniqueTable+(bitPointer-bitMask)*nBits+i;
	return *currentSlot;
      }
      ++bitPointer;
    }
    
    currentSlot=EOT;
    return (T1*) 0;
  }
  T1* next(void){
    //calculate position of next slot
    unsigned long index=(currentSlot-uniqueTable)+1;
    unsigned short i=index%nBits;
    
    unsigned long* bitPointer;
    if(i!=0){
      bitPointer=bitMask+index/nBits;
      if((*bitPointer)>>i){
	unsigned long value=(*bitPointer)>>i;
	while((value&1)==0){
	  ++i;
	  value>>=1;
	}
	currentSlot=uniqueTable+nBits*(index/nBits)+i;
	return *currentSlot;
      }
      //no element in current bitfield, change to next
      index+=nBits;
    }
    index/=nBits;
    bitPointer=bitMask+index;
    
    while(bitPointer!=EOM){
      if(*bitPointer!=0){
	unsigned long value=*bitPointer;
	i=0;
	while((value&1)==0){
	  ++i;
	  value>>=1;
	}
	currentSlot=uniqueTable+(bitPointer-bitMask)*nBits+i;
	return *currentSlot;
      }
      ++bitPointer;
    }
    
    currentSlot=EOT;
    return (T1*) 0;
  }
  void setLevel(const unsigned long newLevel){
    level=newLevel;
  }
  inline unsigned long getLevel(void) const {
    return level-skip;
  }
  inline unsigned long getRealLevel(void) const {
    return level;
  }
  inline bool deleteable(void) const {
    return (count==0);
  }
  inline unsigned long hashFunction(const T2&, const T2&) const;
  inline static void setSkip(unsigned long newSkip){
    skip=newSkip;
  }
  inline static void setInitHashSize(unsigned int newInit){
    initHashSize=newInit;
  }
  inline static void setMaxHashSize(unsigned int newMax){
    maximumHashSize=newMax;
  }
  inline static unsigned long getSkip(void){
    return skip;
  }
  inline void setDrainLevel(void){
    drainLevel=true;
  }
  inline bool isDrainLevel(void) const {
    return drainLevel;
  }
private:
  static unsigned int initHashSize;
  static unsigned int maximumHashSize;
  static unsigned long skip;
  unsigned long level;
  unsigned long count;
  unsigned long maxCollisionListSize;
  unsigned int hashSize;
  unsigned int maxHashSize;
  unsigned short shift;
  unsigned short collisionListSizeLimit;
  unsigned long deadNodes;

  T1** uniqueTable;
  T1** EOT;
  T1** currentSlot;
  unsigned long* bitMask;
  unsigned long* EOM;
  bool drainLevel;
  
  void resize(void);
};

template <class T1, class T2>
unsigned long HashUniqueTable2<T1,T2>::skip=0;

template <class T1, class T2>
unsigned int HashUniqueTable2<T1,T2>::initHashSize=2048;

template <class T1, class T2>
unsigned int HashUniqueTable2<T1,T2>::maximumHashSize=262144;

template <class T1, class T2>
HashUniqueTable2<T1,T2>::HashUniqueTable2(const unsigned long newLevel){
  level=newLevel;
  count=0;
  maxCollisionListSize=0;
  collisionListSizeLimit=12;

  hashSize=2;
  //shift for smallest hashmap
  shift=sizeof(unsigned long)*8-1;
  while((hashSize<<1)<=initHashSize){
    hashSize<<=1;
    --shift;
  }
  maxHashSize=maximumHashSize;
  
  uniqueTable=new T1*[hashSize];
  memset(uniqueTable,0,hashSize*sizeof(T1*));
  EOT=uniqueTable+hashSize;
  bitMask=new unsigned long[hashSize/nBits];
  memset(bitMask,0,hashSize/8);
  EOM=bitMask+hashSize/nBits;
  currentSlot=EOT;
	deadNodes=0;
	drainLevel=false;
}

template <class T1, class T2>
HashUniqueTable2<T1,T2>::~HashUniqueTable2(void){
  delete[] uniqueTable;
  delete[] bitMask;
}

template <class T1, class T2>
void HashUniqueTable2<T1,T2>::add(T1* result){
  T2 nodeInfo1=result->getSucc0NodeInfo();
  T2 nodeInfo2=result->getSucc1NodeInfo();

  unsigned long index=hashFunction(nodeInfo1,nodeInfo2);
  unsigned long collisionListSize=0;

  T1* tempNode=uniqueTable[index];
  T1* lastNode=0;
  if(tempNode!=0){
    while(tempNode->getSucc0NodeInfo()<nodeInfo1){
      lastNode=tempNode;
      tempNode=tempNode->getNext();
      ++collisionListSize;
      if(tempNode==0) break;
    }
    if(tempNode!=0){
      while((tempNode->getSucc0NodeInfo()==nodeInfo1)&&(tempNode->getSucc1NodeInfo()<nodeInfo2)){
        lastNode=tempNode;
        tempNode=tempNode->getNext();
        ++collisionListSize;
        if(tempNode==0) break;
      }
    }
  }
  //no check if entry is already there
  if(lastNode==0){
    result->setNext(uniqueTable[index]);
    uniqueTable[index]=(T1*) result;
		bitMask[index/nBits]|=1<<(index%nBits);
  } else {
    result->setNext(lastNode->getNext());
    lastNode->setNext(result);
  }
  ++count;
  if(collisionListSize>maxCollisionListSize){
    maxCollisionListSize=collisionListSize;
	  if(maxCollisionListSize>collisionListSizeLimit){
	    if((count>=(hashSize*collisionListSizeLimit)/4)&&(hashSize<maxHashSize)) resize();
	  }
  }
}

template <class T1, class T2>
bool HashUniqueTable2<T1,T2>::remove(const T1* value){
  T2 nodeInfo1=value->getSucc0NodeInfo();
  T2 nodeInfo2=value->getSucc1NodeInfo();
  unsigned long index=hashFunction(nodeInfo1,nodeInfo2);
	
  T1* tempNode=uniqueTable[index];
  T1* lastNode=0;
  if(tempNode!=0){
    while(tempNode!=value){
      lastNode=tempNode;
      tempNode=tempNode->getNext();
      if(tempNode==0) break;
    }
  }
	if(tempNode!=value) return false;
	if(lastNode==0){
		//change entry in the unique table and check if the value is 0
		if((uniqueTable[index]=tempNode->getNext())==0) bitMask[index/nBits]&=~(1<<(index%nBits));
	} else lastNode->setNext(tempNode->getNext());

	tempNode->setNext(0);
	--count;
	return true;
}

template <class T1, class T2>
T1* HashUniqueTable2<T1,T2>::get(const T2& nodeInfo1, const T2& nodeInfo2) const {
  unsigned long index=hashFunction(nodeInfo1,nodeInfo2);

  T1* tempNode=uniqueTable[index];
  T1* lastNode=0;
  if(tempNode!=0){
    while(tempNode->getSucc0NodeInfo()<nodeInfo1){
      lastNode=tempNode;
      tempNode=tempNode->getNext();
      if(tempNode==0) break;
    }
    if(tempNode!=0){
      while((tempNode->getSucc0NodeInfo()==nodeInfo1)&&(tempNode->getSucc1NodeInfo()<nodeInfo2)){
        lastNode=tempNode;
        tempNode=tempNode->getNext();
        if(tempNode==0) break;
      }
    }
  }
  if(tempNode!=0){
    if((tempNode->getSucc0NodeInfo()==nodeInfo1)&&(tempNode->getSucc1NodeInfo()==nodeInfo2)) return tempNode;
  }
  
  return 0;
}

template <class T1, class T2>
T1* HashUniqueTable2<T1,T2>::findOrAdd(const T2& nodeInfo1, const T2& nodeInfo2){
  unsigned long index=hashFunction(nodeInfo1,nodeInfo2);
	
  unsigned long collisionListSize=0;
  T1* tempNode=uniqueTable[index];
  T1* lastNode=0;
	
  if(tempNode!=0){
    while(tempNode->getSucc0NodeInfo()<nodeInfo1){
			++collisionListSize;
      lastNode=tempNode;
      tempNode=tempNode->getNext();
      if(tempNode==0) break;
    }
    if(tempNode!=0){
      while((tempNode->getSucc0NodeInfo()==nodeInfo1)&&(tempNode->getSucc1NodeInfo()<nodeInfo2)){
				++collisionListSize;
        lastNode=tempNode;
        tempNode=tempNode->getNext();
        if(tempNode==0) break;
      }
    }
  }
  if(tempNode!=0){
    if((tempNode->getSucc0NodeInfo()==nodeInfo1)&&(tempNode->getSucc1NodeInfo()==nodeInfo2)) return tempNode;
  }

	T1* result=T1::getNewNode();

	result->initNode(this,nodeInfo1,nodeInfo2);
  if(lastNode==0){
    result->setNext(uniqueTable[index]);
    uniqueTable[index]=(T1*) result;
		bitMask[index/nBits]|=1<<(index%nBits);
  } else {
    result->setNext(lastNode->getNext());
    lastNode->setNext(result);
  }
  ++count;
  if(collisionListSize>maxCollisionListSize){
    maxCollisionListSize=collisionListSize;
	  if(maxCollisionListSize>collisionListSizeLimit){
	    if((count>=(hashSize*collisionListSizeLimit)/4)&&(hashSize<maxHashSize)) resize();
	  }
  }
  
  return result;
}

template <class T1, class T2>
unsigned long HashUniqueTable2<T1,T2>::hashFunction(const T2& nodeInfo1, const T2& nodeInfo2) const {
  return ((((unsigned long) nodeInfo1.getNode()*12582917+(unsigned long) nodeInfo2.getNode())*4256249)>>shift);
}

template <class T1, class T2>
void HashUniqueTable2<T1,T2>::resize(void){
  T1** tempHashUniqueTable2=new T1*[2*hashSize];
  //out of memory
  if(tempHashUniqueTable2==0) return;
  unsigned long* tempBitMask=new unsigned long[(2*hashSize)/nBits];
  //out of memory
  if(tempBitMask==0){
    delete[] tempHashUniqueTable2;
    return;
  }
  
  --shift;
  memset(tempHashUniqueTable2,0,2*hashSize*sizeof(T1*));
  memset(tempBitMask,0,(2*hashSize)/8);
  maxCollisionListSize=0;
  
  //transfer all nodes from old unique table to new one
  T1* currentTable=(T1*) getFirst();
  T1* currentNode;
  T1* tempNode;
	T1* evenTable;
	T1* oddTable;
  unsigned long index;
  unsigned long evenCollisionListSize;
  unsigned long oddCollisionListSize;
  while(currentTable){
    currentNode=currentTable;
		tempNode=currentNode->getNext();
		//shift is already updated so that the shift has to be reverted
		index=hashFunction(currentNode->getSucc0NodeInfo(),currentNode->getSucc1NodeInfo())>>1;
		
		evenTable=0;
		oddTable=0;
		
		evenCollisionListSize=0;
		oddCollisionListSize=0;
	  while(currentNode){
      tempNode=currentNode->getNext();
			//shift is already updated
			if(hashFunction(currentNode->getSucc0NodeInfo(),currentNode->getSucc1NodeInfo())&1){
				if(oddTable!=0) oddTable->setNext(currentNode);
				else tempHashUniqueTable2[2*index+1]=currentNode;
        oddTable=currentNode;
				++oddCollisionListSize;
			} else {
				if(evenTable!=0) evenTable->setNext(currentNode);
				else tempHashUniqueTable2[2*index]=currentNode;
        evenTable=currentNode;
				++evenCollisionListSize;
			}
						
	    //--------------------------------------
	    currentNode=tempNode;
    }
		
		if(oddTable!=0){
			oddTable->setNext(0);
			tempBitMask[(2*index+1)/nBits]|=1<<((2*index+1)%nBits);
		} else tempHashUniqueTable2[2*index+1]=0;
		if(evenTable!=0){
			evenTable->setNext(0);
			tempBitMask[2*index/nBits]|=1<<((2*index)%nBits);
		} else tempHashUniqueTable2[2*index]=0;

		if(oddCollisionListSize>maxCollisionListSize) maxCollisionListSize=oddCollisionListSize;
		if(evenCollisionListSize>maxCollisionListSize) maxCollisionListSize=evenCollisionListSize;

    currentTable=(T1*) next();
  }
  
  //updating informations
  delete[] uniqueTable;
  delete[] bitMask;
  
  uniqueTable=tempHashUniqueTable2;
  bitMask=tempBitMask;  
  
  hashSize*=2;
  ++collisionListSizeLimit;
  EOT=uniqueTable+hashSize;
  EOM=bitMask+hashSize/nBits;
  currentSlot=EOT;
}

#endif
