/***************************************************************************
                          reordering.h  -  description
                             -------------------
    begin                : Tue Apr 13 2004
    copyright            : (C) 2004 by Wolfgang Lenders, Joern Ossowski
    email                : {lenders,ossowski}@cs.uni-bonn.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software;you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation;either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef REORDERING_H
#define REORDERING_H

#include <common/constants.h>
#include <common/ulongarray.h>
#include <common/randomhelper.h>
#include <common/vsizearray.h>

#include <iostream>
#include <vector>
#include <iostream>

#include "vereofy/common/tracing.hpp"
#include "vereofy/common/interruptible.hpp"
#include "vereofy/common/countdown_clock.hpp"
#include "vereofy/common/stopwatch.hpp"
#include "vereofy/common/PermutationBySwap.hpp"

using namespace std;

template <class T1, class T2, class T3>
class Reordering{
public:
  Reordering(bool (T3::*fptr)(const unsigned long), T3*);
  void windowPermutation2(void);
  void windowPermutation3(void);
  void windowPermutation4(void);
  void variableSifting(const float=MAX_GROWTH_DEFAULT, unsigned int timeout_ms=0);
  void groupSifting(const float=MAX_GROWTH_DEFAULT, unsigned int timeout_ms=0);
  bool shift(unsigned long, const unsigned long) const;
  void geneticMinimize(void);
  void groupGeneticMinimize(void);
  template <class CallBack> void groupExhaustive(CallBack& call_back);
private:
  void windowPermutation2(const unsigned long, const unsigned long) const;
  void windowPermutation3(const unsigned long, const unsigned long) const;
  void windowPermutation4(const unsigned long, const unsigned long) const;
  void sifting(const unsigned long, const unsigned long, const float, const bool, unsigned int timeout_ms=0) const;
  bool groupSwap(const T2&) const;
  bool groupShift(unsigned long, const unsigned long) const;
  void geneticMinimize(unsigned long, const unsigned long) const;
  void minimumSifting(ULongArray&, const unsigned long) const;
  void crossover(const ULongArray&, const ULongArray&, ULongArray&, const unsigned long) const;
  void adoptCurrentGroupOrderToArray(ULongArray&, const ULongArray&) const;
  void adoptCurrentOrderToArray(ULongArray&, const ULongArray&, const unsigned long) const;
  void chooseParents(ULongArray&, unsigned long&, unsigned long&) const;
  void prepareReordering(void);
  void finalizeReordering(void);

  bool (T3::*swapfct)(const unsigned long);
  T3* object;
  unsigned long tempDelay;
  vector<T2> ungrouped;
  unsigned long GENETIC_DEBUG;

  struct ReorderingSwapOp {
  ReorderingSwapOp(Reordering<T1,T2,T3>& reorder_) :
    reorder(reorder_) {}
    
    void operator()(unsigned int index_a, unsigned int index_b) {
      unsigned int to_swap = (index_a < index_b) ? index_a : index_b;
      
      reorder.groupSwap(T1::getGroup(to_swap));
    }
    
    Reordering<T1,T2,T3>& reorder;
  };
};

template <class T1, class T2, class T3>
Reordering<T1,T2,T3>::Reordering(bool (T3::*fptr)(const unsigned long), T3* object){
  swapfct=fptr;
  this->object=object;
  GENETIC_DEBUG=1;
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::windowPermutation2(void){
  prepareReordering();

  T2 tempGroup=T1::getGroup(0);
  for(unsigned long i=0;i<T1::groupCount();++i){
    windowPermutation2(tempGroup.getIndexOfFirstElement(),tempGroup.size());
    tempGroup=T1::getNextGroup(tempGroup);
  }

  finalizeReordering();
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::windowPermutation3(void){
  prepareReordering();

  T2 tempGroup=T1::getGroup(0);
  for(unsigned long i=0;i<T1::groupCount();++i){
    windowPermutation3(tempGroup.getIndexOfFirstElement(),tempGroup.size());
    tempGroup=T1::getNextGroup(tempGroup);
  }

  finalizeReordering();
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::windowPermutation4(void){
  prepareReordering();

  T2 tempGroup=T1::getGroup(0);
  for(unsigned long i=0;i<T1::groupCount();++i){
    windowPermutation4(tempGroup.getIndexOfFirstElement(),tempGroup.size());
    tempGroup=T1::getNextGroup(tempGroup);
  }

  finalizeReordering();
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::variableSifting(const float maxGrowth, unsigned int timeout_ms){
  prepareReordering();

  T2 tempGroup=T1::getGroup(0);
  for(unsigned long i=0;i<T1::groupCount();++i){
    sifting(tempGroup.getIndexOfFirstElement(),tempGroup.size(),maxGrowth,false, timeout_ms);
    tempGroup=T1::getNextGroup(tempGroup);
  }

  finalizeReordering();
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::groupSifting(const float maxGrowth, unsigned int timeout_ms){
  prepareReordering();

  sifting(0,T1::groupCount(),maxGrowth,true, timeout_ms);

  finalizeReordering();
}



template <class T1, class T2, class T3>
template <class CallBack>
void Reordering<T1,T2,T3>::groupExhaustive(CallBack& call_back){
  prepareReordering();

  std::string saved_order="";

  ReorderingSwapOp swap_op(*this);
  PermutationBySwap permutation(T1::groupCount()); // count the groups

  std::cerr << "Starting exhaustive reordering for " << T1::groupCount() << " BDD variable groups (" << permutation.countIterations() << " iterations)...\n";

  vereofy::StopWatch stopwatch;
  stopwatch.start();
  CountdownClock count_down(1000);
  unsigned long counter = 0;

  bool store_order = call_back();
  if (store_order) {
    saved_order = T1::toString();
  }
  while (permutation.hasNext()) {
    permutation.next<ReorderingSwapOp>(swap_op);
    store_order = call_back();
    if (store_order) {
      saved_order = T1::toString();
    }
    ++counter;

    if (count_down.isFinished()) {
      std::cerr << "\rIteration " << counter << ", size = " << T3::size() << "       ";
      count_down.restart();
    }
  }

  stopwatch.stop();
  std::cerr << "\nExhaustive reordering done...";
  std::cerr << " (" << stopwatch.getElapsedTimeString() << "s)\n";

  std::cerr << "Best order:\n";
  std::cerr << saved_order;
  
  finalizeReordering();
}



template <class T1, class T2, class T3>
bool Reordering<T1,T2,T3>::shift(unsigned long index, const unsigned long position) const {
  //check if shift is possible
  if(index>=T1::size()) return false;
  if(position>=T1::size()) return false;

  //if index==position we are ready
  if(index==position) return true;

  if(index<position){ //jump down
    while(index<position){
      (object->*swapfct)(index);
      ++index;
    }
  } else { //jump up
    while(index>position){
      (object->*swapfct)(index-1);
      --index;
    }
  }

  return true;
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::geneticMinimize(void){
  prepareReordering();

  T2 tempGroup=T1::getGroup(0);
  for(unsigned long i=0;i<T1::groupCount();++i){
    geneticMinimize(tempGroup.getIndexOfFirstElement(),tempGroup.size());
    tempGroup=T1::getNextGroup(tempGroup);
  }

  finalizeReordering();
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::groupGeneticMinimize(){
  return;
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::windowPermutation2(const unsigned long begin, const unsigned long length) const {
  unsigned long optSize;
  unsigned long i;

  for(i=begin,optSize=T3::size();i+1<begin+length;++i){
    (object->*swapfct)(i);
    if(T3::size()>=optSize) (object->*swapfct)(i);
  }
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::windowPermutation3(const unsigned long begin, const unsigned long length) const {
  enum{ABC,BAC,BCA,CBA,CAB,ACB};
  unsigned long optSize;
  short optPos;
  unsigned long i;

  if(length<3) windowPermutation2(begin,length);
  else for(i=begin,optSize=T3::size(),optPos=ABC;i+2<begin+length;++i){
    optPos=ABC;
    (object->*swapfct)(i);  //BAC
    if(T3::size()<optSize){optSize=T3::size();optPos=BAC;}
    (object->*swapfct)(i+1);//BCA
    if(T3::size()<optSize){optSize=T3::size();optPos=BCA;}
    (object->*swapfct)(i);  //CBA
    if(T3::size()<optSize){optSize=T3::size();optPos=CBA;}
    (object->*swapfct)(i+1);//CAB
    if(T3::size()<optSize){optSize=T3::size();optPos=CAB;}
    (object->*swapfct)(i);  //ACB
    if(T3::size()<optSize){optSize=T3::size();optPos=ACB;}

    switch(optPos){
      case CBA: (object->*swapfct)(i);  //ACB=>CAB=>CBA
      case ABC: (object->*swapfct)(i+1);//ACB=>ABC
                break;
      case BCA: (object->*swapfct)(i);  //ACB=>CAB=>CBA=>BCA
      case BAC: (object->*swapfct)(i+1);//ACB=>ABC=>BAC
      case CAB: (object->*swapfct)(i);  //ACB=>CAB
      default:  break;
    }
  }
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::windowPermutation4(const unsigned long begin, const unsigned long length) const {
  enum{ABCD,ABDC,ADBC,DABC,DACB,ADCB,ACDB,ACBD,CABD,CADB,CDAB,DCAB,DCBA,CDBA,CBDA,CBAD,BCAD,BCDA,BDCA,DBCA,DBAC,BDAC,BADC,BACD};
  unsigned long optSize;
  short optPos;
  unsigned long i;

  if(length<4) windowPermutation3(begin,length);
  else for(i=begin,optSize=T3::size(),optPos=ABCD;i+3<begin+length;++i){
    optPos=ABCD;
    (object->*swapfct)(i+2);//ABDC
    if(T3::size()<optSize){optSize=T3::size();optPos=ABDC;}
    (object->*swapfct)(i+1);//ADBC
    if(T3::size()<optSize){optSize=T3::size();optPos=ADBC;}
    (object->*swapfct)(i);  //DABC
    if(T3::size()<optSize){optSize=T3::size();optPos=DABC;}
    (object->*swapfct)(i+2);//DACB
    if(T3::size()<optSize){optSize=T3::size();optPos=DACB;}
    (object->*swapfct)(i);  //ADCB
    if(T3::size()<optSize){optSize=T3::size();optPos=ADCB;}
    (object->*swapfct)(i+1);//ACDB
    if(T3::size()<optSize){optSize=T3::size();optPos=ACDB;}
    (object->*swapfct)(i+2);//ACBD
    if(T3::size()<optSize){optSize=T3::size();optPos=ACBD;}
    (object->*swapfct)(i);  //CABD
    if(T3::size()<optSize){optSize=T3::size();optPos=CABD;}
    (object->*swapfct)(i+2);//CADB
    if(T3::size()<optSize){optSize=T3::size();optPos=CADB;}
    (object->*swapfct)(i+1);//CDAB
    if(T3::size()<optSize){optSize=T3::size();optPos=CDAB;}
    (object->*swapfct)(i);  //DCAB
    if(T3::size()<optSize){optSize=T3::size();optPos=DCAB;}
    (object->*swapfct)(i+2);//DCBA
    if(T3::size()<optSize){optSize=T3::size();optPos=DCBA;}
    (object->*swapfct)(i);  //CDBA
    if(T3::size()<optSize){optSize=T3::size();optPos=CDBA;}
    (object->*swapfct)(i+1);//CBDA
    if(T3::size()<optSize){optSize=T3::size();optPos=CBDA;}
    (object->*swapfct)(i+2);//CBAD
    if(T3::size()<optSize){optSize=T3::size();optPos=CBAD;}
    (object->*swapfct)(i);  //BCAD
    if(T3::size()<optSize){optSize=T3::size();optPos=BCAD;}
    (object->*swapfct)(i+2);//BCDA
    if(T3::size()<optSize){optSize=T3::size();optPos=BCDA;}
    (object->*swapfct)(i+1);//BDCA
    if(T3::size()<optSize){optSize=T3::size();optPos=BDCA;}
    (object->*swapfct)(i);  //DBCA
    if(T3::size()<optSize){optSize=T3::size();optPos=DBCA;}
    (object->*swapfct)(i+2);//DBAC
    if(T3::size()<optSize){optSize=T3::size();optPos=DBAC;}
    (object->*swapfct)(i);  //BDAC
    if(T3::size()<optSize){optSize=T3::size();optPos=BDAC;}
    (object->*swapfct)(i+1);//BADC
    if(T3::size()<optSize){optSize=T3::size();optPos=BADC;}
    (object->*swapfct)(i+2);//BACD
    if(T3::size()<optSize){optSize=T3::size();optPos=BACD;}

    switch(optPos){
      case CABD: (object->*swapfct)(i);
      case CBAD: (object->*swapfct)(i+1);
      case ABCD: (object->*swapfct)(i);
                 break;
      case ACBD: (object->*swapfct)(i);
      case BCAD: (object->*swapfct)(i+1);
                 break;
      case ACDB: (object->*swapfct)(i);
      case BCDA: (object->*swapfct)(i+1);
      case BADC: (object->*swapfct)(i+2);
                 break;
      case CDAB: (object->*swapfct)(i);
      case CDBA: (object->*swapfct)(i+1);
      case ADBC: (object->*swapfct)(i);
      case BDAC: (object->*swapfct)(i+2);
                 (object->*swapfct)(i+1);
                 break;
      case CADB: (object->*swapfct)(i);
      case CBDA: (object->*swapfct)(i+1);
      case ABDC: (object->*swapfct)(i);
                 (object->*swapfct)(i+2);
                 break;
      case ADCB: (object->*swapfct)(i);
      case BDCA: (object->*swapfct)(i+1);
                 (object->*swapfct)(i+2);
                 (object->*swapfct)(i+1);
                 break;
      case DCAB: (object->*swapfct)(i);
      case DCBA: (object->*swapfct)(i+1);
      case DABC: (object->*swapfct)(i);
      case DBAC: (object->*swapfct)(i+2);
                 (object->*swapfct)(i+1);
                 (object->*swapfct)(i);
                 break;
      case DACB: (object->*swapfct)(i);
      case DBCA: (object->*swapfct)(i+1);
                 (object->*swapfct)(i+2);
                 (object->*swapfct)(i+1);
                 (object->*swapfct)(i);
                 break;
      default:  break;
    }
  }
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::sifting(const unsigned long begin, const unsigned long length, const float maxGrowth, const bool group, 
                                   unsigned int timeout_ms) const {
  unsigned long optSize=T3::size();
  unsigned long optPos;
  unsigned long l,r;
  VSIZEARRAY(unsigned long, elemPos, length);
  T2 currentGroup;

  TRACE("JINC", "reorder_details",     \
        std::cout << "\nStarting reordering, start size = " << optSize << "\n";);

  CountdownClock countdown(timeout_ms);

  for(l=0;l<length;++l) elemPos[l]=begin+l;

  for(l=0;l<length;++l){
    if((elemPos[l]-begin)<(begin+length-elemPos[l]-1)){
      //going left
      countdown.restart();
      r=optPos=elemPos[l];
      if(group) currentGroup=T1::getGroup(r);
      while((r>begin) && (T3::size()<=(unsigned long) (maxGrowth*(float)optSize))){
        //swap element
        INTERRUPTIBLE();
        if(group) groupSwap(T1::getPrevGroup(currentGroup));
        else (object->*swapfct)(r-1);
        if(T3::size()<optSize){
          countdown.restart();
          optSize=T3::size();
          optPos=r-1;
        }
        --r;
      }
      //going to initial position
      if(group) groupShift(r,elemPos[l]);
      else shift(r,elemPos[l]);

      //going right
      countdown.restart();
      r=elemPos[l];
      while((r+1<begin+length) && (T3::size()<=(unsigned long) (maxGrowth*(float)optSize))){
        INTERRUPTIBLE();
        if(group) groupSwap(currentGroup);
        else (object->*swapfct)(r);
        if(T3::size()<optSize){
          countdown.restart();
          optSize=T3::size();
          optPos=r+1;
        }
        ++r;
      }
    } else {
      //going right
      countdown.restart();
      r=optPos=elemPos[l];
      if(group) currentGroup=T1::getGroup(r);
      while((r+1<begin+length) && (T3::size()<=(unsigned long) (maxGrowth*(float)optSize)) && !countdown.isFinished()){
        INTERRUPTIBLE();
        if(group) groupSwap(currentGroup);
        else (object->*swapfct)(r);

        if(T3::size()<optSize){
          countdown.restart();
          optSize=T3::size();
          optPos=r+1;
        }
        ++r;
      }
      //going to initial position
      if(group) groupShift(r,elemPos[l]);
      else shift(r,elemPos[l]);

      //going left
      countdown.restart();
      r=elemPos[l];
      if(group) currentGroup=T1::getGroup(r);
      while((r>begin) && (T3::size()<=(unsigned long) (maxGrowth*(float)optSize)) && !countdown.isFinished()){
        INTERRUPTIBLE();
        if(group) groupSwap(T1::getPrevGroup(currentGroup));
        else (object->*swapfct)(r-1);

        if(T3::size()<optSize){
          countdown.restart();
          optSize=T3::size();
          optPos=r-1;
        }
        --r;
      }
    }
    //move variable to best position
    if(group) groupShift(r,optPos);
    else shift(r,optPos);


    TRACE("JINC", "reorder_details",   \
          std::cout << "\r" <<                                          \
          "Reorder " << l+1 << " of " << length << ", current size=" << T3::size() << "             "; \
          std::cout.flush(););
    
    //search for next variable
    if(optPos<elemPos[l]){
      for(r=l+1;r<length;++r){
        if((elemPos[r]>=optPos)&&(elemPos[r]<elemPos[l])) ++elemPos[r];
      }
    } else if(optPos>elemPos[l]){
      for(r=l+1;r<length;++r){
        if((elemPos[r]>elemPos[l])&&(elemPos[r]<=optPos)) --elemPos[r];
      }
    }
  }

  TRACE("JINC", "reorder_details", std::cout << "\n");

}

template <class T1, class T2, class T3>
bool Reordering<T1,T2,T3>::groupSwap(const T2& group) const {
  T2 tempGroup=T1::getNextGroup(group);
  if(tempGroup.size()==0) return false;

  unsigned long startIndex=group.getIndexOfFirstElement();
  unsigned long sizeOfGroup1=group.size();
  unsigned long sizeOfGroup2=tempGroup.size();
  unsigned long i,j;

  if(RandomHelper::getULong(2)){
    for(i=0;i<sizeOfGroup1;++i){
      for(j=0;j<sizeOfGroup2;++j){
        (object->*swapfct)(startIndex+sizeOfGroup1-1+j-i);
      }
    }
  } else {
    for(i=0;i<sizeOfGroup2;++i){
      for(j=0;j<sizeOfGroup1;++j){
        (object->*swapfct)(startIndex+sizeOfGroup1-1-j+i);
      }
    }
  }

  return true;
}

template <class T1, class T2, class T3>
bool Reordering<T1,T2,T3>::groupShift(unsigned long index, const unsigned long position) const {
  //check if shift is possible
  if(index>=T1::groupCount()) return false;
  if(position>=T1::groupCount()) return false;

  //if index==position we are done
  if(index==position) return true;

  if(index<position){ //jump down
    T2 currentGroup=T1::getGroup(index);
    while(index<position){
      groupSwap(currentGroup);
      ++index;
    }
  } else { //jump up
    T2 currentGroup=T1::getGroup(index);
    while(index>position){
      groupSwap(T1::getPrevGroup(currentGroup));
      --index;
    }
  }

  return true;
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::geneticMinimize(const unsigned long begin, const unsigned long length)  const {
  if(length<2) return;
  if(GENETIC_DEBUG) cout << endl << "----- Entering geneticMinimize() -----" << endl << endl;
  if(GENETIC_DEBUG) cout << "   using ordering: " << T1::getVarOrder() << endl;

  /** the population stores <POPULATION_SIZE> elements, here called "individuals"
   *  each individual is an array of unsigned longs, which stand for the current position of
   *  the variable in relation to the position when geneticMinimize() was called
   *  Example:
   *  Let individual population[0] be {0, 1, 4, 2, 3} means that calling
   *    swap(2);
   *    swap(3);
   *  would bring you back to the original order
   */
  unsigned long orderLength=length;
  ULongArray population[POPULATION_SIZE];
  ULongArray fitness(POPULATION_SIZE);
  ULongArray sizeOfBDD(POPULATION_SIZE);
  ULongArray distances(POPULATION_SIZE);
  unsigned long bestBDDSize=ULONG_MAX;
  unsigned long oldBDDSize=ULONG_MAX;
  unsigned long i,j,worstBDDSize,averageDistance,indexOfNearestOrder;
  i=j=worstBDDSize=averageDistance=indexOfNearestOrder=0;
  unsigned long maxDistance=orderLength*orderLength/2;
  double averageFitness=0.0;
  ULongArray matingPool(MATING_POOL_SIZE);

  // this array's order corresponds to the SOBDD's variableOrdering
  ULongArray currentOrder(orderLength);
  for(i=0;i<orderLength;++i) currentOrder[i]=i;
  ULongArray fittestIndividual=currentOrder;

  // create initial population
  population[0]=currentOrder;
  // randomize the rest of the initial population
  for(i=1;i<POPULATION_SIZE;++i){
    population[i]=currentOrder;
    for(j=0;j<orderLength;++j) population[i].shift(j+RandomHelper::getULong(orderLength-j), j);
    distances[i]=currentOrder.distance(population[i]);
    cout << "population[" << i << "]: " << population[i].print() << endl;
  }
  // check the bdd sizes
  sizeOfBDD[0]=T3::size();
  bestBDDSize=worstBDDSize=sizeOfBDD[0];
  if(GENETIC_DEBUG) cout << endl << "----- initial population: -----" << endl << endl;
  if(GENETIC_DEBUG) cout << "   population[0]: " << T1::getVarOrder() << " \tbdd size: " << sizeOfBDD[0] << endl;
  for(i=1;i<POPULATION_SIZE;++i){
    adoptCurrentOrderToArray(currentOrder, population[i], begin);
    sizeOfBDD[i]=T3::size();
    if(GENETIC_DEBUG) cout << "   population[" << i << "]: " << T1::getVarOrder() << " \tbdd size: " << sizeOfBDD[i] << endl;
    if(sizeOfBDD[i]<bestBDDSize){
      bestBDDSize=sizeOfBDD[i];
      fittestIndividual=population[i];
    } else if(sizeOfBDD[i]>worstBDDSize) worstBDDSize=sizeOfBDD[i];
  }
  for(i=0;i<POPULATION_SIZE;++i) fitness[i]=worstBDDSize-sizeOfBDD[i]+1;

  // initialize mating pool
  if(DEFAULT_SELECTION==STOCHASTIC_REMAINDER_SELECTION){
    unsigned long popLoop=0;
    for(i=0;i<MATING_POOL_SIZE;++i){
      matingPool[i]=popLoop;
      ++popLoop%=POPULATION_SIZE;
    }
  }

  // now loop until no more improvement is obtained for several continuous iterations
  unsigned long iterationCount,iterationsWithoutImprovement,nextGenerationWorstBDDSize,nextGenerationMembersFound;
  iterationCount=iterationsWithoutImprovement=0;
  while ((bestBDDSize<oldBDDSize) || (++iterationsWithoutImprovement <= (iterationCount/8)+5)){
    // handle general loop conditions
    if(GENETIC_DEBUG) cout << "----- creating generation " << ++iterationCount << " -----" << endl;
    if(bestBDDSize<oldBDDSize) iterationsWithoutImprovement=0;
    oldBDDSize=bestBDDSize;
    nextGenerationWorstBDDSize=nextGenerationMembersFound=0;

    // initialize new generation
    ULongArray nextGeneration[POPULATION_SIZE];
    for(i=0;i<POPULATION_SIZE;++i) nextGeneration[i]=currentOrder;
    ULongArray nextGenerationBDDSize(POPULATION_SIZE);
    ULongArray nextGenerationFitness(POPULATION_SIZE);

    // handle elitarism
    ULongArray indicesOrderedByFitness(POPULATION_SIZE);
    indicesOrderedByFitness=fitness.getIndicesOrderedByValues();
    for(i=0;i<ELITARISM_WINNERS;++i){
      if(GENETIC_DEBUG>=2) cout << "   population[" << indicesOrderedByFitness[i] << "] is an elitarism-winner" << endl;
      nextGeneration[nextGenerationMembersFound]=population[indicesOrderedByFitness[i]];
      nextGenerationBDDSize[nextGenerationMembersFound]=sizeOfBDD[indicesOrderedByFitness[i]];
      if(nextGenerationBDDSize[nextGenerationMembersFound]>nextGenerationWorstBDDSize) nextGenerationWorstBDDSize=nextGenerationBDDSize[nextGenerationMembersFound];
      ++nextGenerationMembersFound;
    }

    // do crossover
    if(GENETIC_DEBUG>=2) cout << "   doing " << CROSSOVER_COUNT << " crossovers" << endl;
    unsigned long firstParentIndex, secondParentIndex, crossoverType;
    crossoverType=ORDERED_CROSSOVER;
    for(i=0;i<CROSSOVER_COUNT;++i){
      //choose parents
      if(DEFAULT_SELECTION==STOCHASTIC_REMAINDER_SELECTION) chooseParents(matingPool, firstParentIndex, secondParentIndex);
      else chooseParents(fitness, firstParentIndex, secondParentIndex);
      if(i==ORDERED_CROSSOVER_COUNT) crossoverType=INVERTED_CROSSOVER_COUNT;
      if(i==ORDERED_CROSSOVER_COUNT+INVERTED_CROSSOVER_COUNT) crossoverType=PARTIALLY_MATCHED_CROSSOVER;
      if(i==ORDERED_CROSSOVER_COUNT+INVERTED_CROSSOVER_COUNT+PARTIALLY_MATCHED_CROSSOVER_COUNT) crossoverType=CYCLE_CROSSOVER;
      if(i==ORDERED_CROSSOVER_COUNT+INVERTED_CROSSOVER_COUNT+PARTIALLY_MATCHED_CROSSOVER_COUNT+ALTERNATING_CROSSOVER_COUNT) crossoverType=ALTERNATING_CROSSOVER;
      crossover(population[firstParentIndex], population[secondParentIndex], nextGeneration[nextGenerationMembersFound], crossoverType);
      // do mutation with a chance of 5% on each newly generated order in nextGeneration
      if(GENETIC_DEBUG>=2) cout << endl << "   ----- doing a mutation?";
      if((population[firstParentIndex]==nextGeneration[nextGenerationMembersFound]) || (population[secondParentIndex]==nextGeneration[nextGenerationMembersFound]) || (RandomHelper::getULong(10)<1)){
        nextGeneration[nextGenerationMembersFound].randomShuffle(RandomHelper::getULong(2)+1);
        if(GENETIC_DEBUG>=2) cout << " yes -----" << endl << endl << "   resulting offspring is population[" << nextGenerationMembersFound << "] = " <<  nextGeneration[nextGenerationMembersFound].print() << endl;
      } else if(GENETIC_DEBUG>=2) cout << " no -----" << endl << endl;
      distances[nextGenerationMembersFound]=currentOrder.distance(nextGeneration[nextGenerationMembersFound]);
      ++nextGenerationMembersFound;
    }

    // handle incest-prevention by random replacement
    averageFitness=0;
    for(i=0;i<nextGenerationMembersFound;++i) averageDistance+=distances[i];
    averageDistance/=nextGenerationMembersFound;
    for(i=0;i<RANDOM_REPLACEMENTS;++i){
      nextGeneration[nextGenerationMembersFound]=currentOrder;
      if(averageDistance<maxDistance/4){
        // randomize
        for(j=0;j<orderLength;++j) nextGeneration[nextGenerationMembersFound].shift(j+RandomHelper::getULong(orderLength-j), j);
        distances[nextGenerationMembersFound]=currentOrder.distance(nextGeneration[nextGenerationMembersFound]);
        if(GENETIC_DEBUG>=2) cout << "   incest prevention adds nextGeneration[" << nextGenerationMembersFound << "] = " <<  nextGeneration[nextGenerationMembersFound].print() << endl;
      } else {
        if(DEFAULT_SELECTION==STOCHASTIC_REMAINDER_SELECTION) chooseParents(matingPool, firstParentIndex, secondParentIndex);
        else chooseParents(fitness, firstParentIndex, secondParentIndex);
        crossover(population[firstParentIndex], population[secondParentIndex], nextGeneration[nextGenerationMembersFound], ORDERED_CROSSOVER);
      }
      ++nextGenerationMembersFound;
    }
    // sanity check
    if(nextGenerationMembersFound!=POPULATION_SIZE) cerr << "SANITY CHECK FAILED: nextGenerationMembersFound = " << nextGenerationMembersFound << "POPULATION_SIZE = " << POPULATION_SIZE;

    // update nextGenerationBDDSize (already done for elitarism winners)
    for(i=ELITARISM_WINNERS;i<POPULATION_SIZE;++i){
      indexOfNearestOrder=i;
      for(j=i+1;j<POPULATION_SIZE;++j) if(distances[j]<distances[indexOfNearestOrder]) indexOfNearestOrder=j;
      // swap the found element to the current position in population
      if(indexOfNearestOrder!=i){
        ULongArray tempOrder=nextGeneration[i];
        nextGeneration[i]=nextGeneration[indexOfNearestOrder];
        nextGeneration[indexOfNearestOrder]=tempOrder;
      }
      adoptCurrentOrderToArray(currentOrder, nextGeneration[i], begin);
      // hybrid GA with minimumSifting
      minimumSifting(currentOrder, begin);
      nextGeneration[i]=currentOrder;
      nextGenerationBDDSize[i]=T3::size();

      if(nextGenerationBDDSize[i]<bestBDDSize){
        bestBDDSize=nextGenerationBDDSize[i];
        fittestIndividual=nextGeneration[i];
      }
      if(nextGenerationBDDSize[i]>nextGenerationWorstBDDSize) nextGenerationWorstBDDSize=nextGenerationBDDSize[i];
      // update distances
      for(j=i+1;j<POPULATION_SIZE;++j) distances[j]=currentOrder.distance(nextGeneration[j]);
    }
    // transform bdd sizes in preliminary fitness values
    averageFitness=0;
    for(i=0;i<POPULATION_SIZE;++i){
      nextGenerationFitness[i]=nextGenerationWorstBDDSize-nextGenerationBDDSize[i]+1;
      averageFitness+=nextGenerationFitness[i];
    }
    averageFitness/=POPULATION_SIZE;
    // scale bdd sizes to obtain final fitness values
    if(averageFitness<nextGenerationWorstBDDSize-bestBDDSize+1){
      double growthFactor=2*averageFitness/(nextGenerationWorstBDDSize-bestBDDSize+1-averageFitness);
      double addend=averageFitness-2*averageFitness*averageFitness/(nextGenerationWorstBDDSize-bestBDDSize+1-averageFitness);
      if(GENETIC_DEBUG>=2) cout << endl << "   scaling function:   scaledFitness = " << growthFactor << " * oldFitness " << (addend<0 ?"":"+ ") << addend << endl;
      for(i=0;i<POPULATION_SIZE;++i) nextGenerationFitness[i]=(unsigned long)(growthFactor*(double)nextGenerationFitness[i]+addend>=0 ? (growthFactor*(double)nextGenerationFitness[i]+addend):0);
    }

    //update population
    fitness=nextGenerationFitness;
    if(GENETIC_DEBUG>=2) cout << endl << "   ----- fitness after iteration " << iterationCount << ": -----" << endl;
    for(i=0;i<POPULATION_SIZE;++i){
      population[i]=nextGeneration[i];
      if(GENETIC_DEBUG>=2) cout << "   population[" << i << "]:\t" << fitness[i] << endl;
    }
    if(GENETIC_DEBUG) cout << "   best BDD size until now:     " << bestBDDSize << endl << endl;
    // update mating pool
    if(DEFAULT_SELECTION==STOCHASTIC_REMAINDER_SELECTION){
      unsigned long matesFound,popLoop,integerPart;
      matesFound=popLoop=integerPart=0;
      double remainder[POPULATION_SIZE];
      double relativeFitness;
      for(i=0;i<POPULATION_SIZE;++i){
        relativeFitness=fitness.at(i)/averageFitness;
        integerPart=(unsigned long)relativeFitness;
        remainder[i]=relativeFitness-integerPart;
        while((integerPart>0)&&(matesFound<MATING_POOL_SIZE)) {
          matingPool[matesFound]=i;
          ++matesFound;
          --integerPart;
        }
      }
      while(matesFound<MATING_POOL_SIZE){
        if(remainder[popLoop]>0.0){
          if(RandomHelper::getUDouble()<remainder[popLoop]) {
            matingPool[matesFound]=popLoop;
            ++matesFound;
            remainder[popLoop]-=0.1;
          }
        } else remainder[popLoop]+=0.1;
        ++popLoop%=POPULATION_SIZE;
      }
    }
  }
  // update current order
  adoptCurrentOrderToArray(currentOrder, fittestIndividual, begin);
  if(GENETIC_DEBUG>=2){
    cout << endl << "leaving geneticMinimize(), opt size = " << T3::size() << endl;
    cout << "in individual:  " << fittestIndividual.print() << endl;
    cout << "using ordering: " << T1::getVarOrder() << endl << endl;
  }
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::minimumSifting(ULongArray& currentOrder, const unsigned long groupOffset) const {
  unsigned long optSize=T3::size();
  unsigned long optPos;
  unsigned long l,r;
  unsigned long count=currentOrder.size();
  VSIZEARRAY(unsigned long, varPos, count);

  for(l=0;l<count;++l) varPos[l]=l;

  for(l=0;l<count;++l){
    //going left
    r=optPos=varPos[l];
    while((r>0) && (T3::size()<=HYBRID_SIFTING_MAX_GROWTH*optSize)){
      (object->*swapfct)(r-1+groupOffset);
      currentOrder.swap(r-1);
      if(T3::size()<optSize){
        optSize=T3::size();
        optPos=r-1;
      }
      --r;
    }

    //going to init position
    shift(r+groupOffset,varPos[l]+groupOffset);
    currentOrder.shift(r, varPos[l]);

    //going right
    r=varPos[l];
    while((r+1<count) && (T3::size()<=HYBRID_SIFTING_MAX_GROWTH*optSize)){
      (object->*swapfct)(r+groupOffset);
      currentOrder.swap(r);
      if(T3::size()<optSize){
        optSize=T3::size();
        optPos=r+1;
      }
      ++r;
    }

    //move variables to their best position
    shift(r+groupOffset,optPos+groupOffset);
    currentOrder.shift(r, optPos);

    if(optPos<varPos[l]){
      for(r=l+1;r<count;++r) if((varPos[r]>=optPos)&&(varPos[r]<varPos[l])) ++varPos[r];
    } else {
      for(r=l+1;r<count;++r) if((varPos[r]>varPos[l])&&(varPos[r]<=optPos)) --varPos[r];
    }
  }
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::crossover(const ULongArray& firstParent, const ULongArray& secondParent, ULongArray& offspring, const unsigned long type) const {
  if(GENETIC_DEBUG>=2){
    cout << endl << "   ------ doing crossover: -----" << endl << endl;
    cout << "      first  parent: " << firstParent.print() << endl;
    cout << "      second parent: " << secondParent.print() << endl;
  }
  switch(type){
    case ORDERED_CROSSOVER:{
      // select length/2 positions to be taken from firstParent
      unsigned long choicesFound=0;
      VSIZEARRAY(bool, selected, firstParent.size());
      for(unsigned long i=0;i<firstParent.size();++i) selected[i]=false;
      if(GENETIC_DEBUG>=2) cout << "      chosing elements at indices [";
      while(choicesFound!=firstParent.size()/2){
        unsigned long choice=(RandomHelper::getULong(firstParent.size()));
        if(!selected[choice]){
          selected[choice]=true;
          ++choicesFound;
          if(GENETIC_DEBUG>=2) cout << choice;
          if(choicesFound!=firstParent.size()/2 && GENETIC_DEBUG>=2) cout << ", ";
        }
      }
      if(GENETIC_DEBUG>=2) cout << "] to be copied from first parent" << endl;
      // construct offspring
      unsigned long secondParentIndex=0;
      for(unsigned long i=0;i<firstParent.size();++i){
        // copy selections to offspring
        if(selected[i]) offspring[i]=firstParent.at(i);
        else{
          // choose rest from secondParent, preserving order
          while(selected[firstParent.find(secondParent.at(secondParentIndex))]) ++secondParentIndex;
          offspring[i]=secondParent.at(secondParentIndex);
          ++secondParentIndex;
        }
      }
      if(GENETIC_DEBUG>=2) cout << "      ordered crossover returns " << offspring.print() << endl;
    }
    break;
    case INVERTED_CROSSOVER:{
      unsigned long firstCutPoint, secondCutPoint, temp;
      RandomHelper::getULongs(firstCutPoint, secondCutPoint, firstParent.size());
      if(firstCutPoint>secondCutPoint){
        temp=secondCutPoint;
        secondCutPoint=firstCutPoint;
        firstCutPoint=temp;
      }
      if(GENETIC_DEBUG>=2) cout << "inverting from " << firstCutPoint << " to " << secondCutPoint << endl;
      for(unsigned long i=0;i<firstParent.size();++i){
        if(i>=firstCutPoint && i<=secondCutPoint) offspring[i]=firstParent.at(secondCutPoint-(i-firstCutPoint));
        else offspring[i]=firstParent.at(i);
      }
    }
    break;
    case PARTIALLY_MATCHED_CROSSOVER: {
      unsigned long firstCutpoint, secondCutpoint, i;
      RandomHelper::getULongs(firstCutpoint, secondCutpoint, offspring.size());
      if(firstCutpoint>secondCutpoint){
        unsigned long temp=firstCutpoint;
        firstCutpoint=secondCutpoint;
        secondCutpoint=temp;
      }
      if(GENETIC_DEBUG>=2) cout << "PMX cutpoints: " << firstCutpoint << " " << secondCutpoint << endl;
      if(GENETIC_DEBUG>=2) cout << "PMX before: fp: " << firstParent.print() << " sp: " << secondParent.print() << endl;
      offspring=firstParent;
      for(i=firstCutpoint;i<secondCutpoint;++i) offspring.shift(i, firstParent.find(secondParent.at(i)));
      if(GENETIC_DEBUG>=2) cout << "PMX after: os: " << offspring.print() << endl;
    }
    break;
    case CYCLE_CROSSOVER: {
      VSIZEARRAY(bool, selected, offspring.size());
      unsigned long i,selectedCounter;
      const ULongArray* first=&firstParent;
      const ULongArray* second=&secondParent;
      const ULongArray* temp;
      for(i=0;i<offspring.size();++i) selected[i]=false;
      selectedCounter=i=0;
      if(GENETIC_DEBUG>=2) cout << "CX before: fp: " << firstParent.print() << " sp: " << secondParent.print() << endl;
      while(selectedCounter<offspring.size()){
        for(i=0;selected[first->at(i)];++i);
        do{
          offspring[i]=first->at(i);
          selected[first->at(i)]=true;
          i=first->find(second->at(i));
          ++selectedCounter;
        } while(!selected[first->at(i)]);
        temp=first;
        first=second;
        second=temp;
      }
      if(GENETIC_DEBUG>=2) cout << "CX after: os: " << offspring.print() << endl;
    }
    break;
    case ALTERNATING_CROSSOVER: {
      unsigned long first, second, i, candidate=firstParent.at(0);
      first=second=0;
      VSIZEARRAY(bool, selected, offspring.size());
      for(i=0;i<offspring.size();++i) selected[i]=false;
      for(i=0;i<offspring.size();++i){
        while (selected[candidate]) candidate=(i%2==0?firstParent.at(first++):secondParent.at(second++));
        selected[candidate]=true;
        offspring[i]=candidate;
      }
      if(GENETIC_DEBUG>=2) cout << "AX after: os: " << offspring.print() << endl;
    }
    break;
  }
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::adoptCurrentGroupOrderToArray(ULongArray& currentOrder, const ULongArray& targetOrder) const {
  // shifts currentOrder and corresponding variableOrdering to resemble targetOrder[]
  for(unsigned long i=0;i<currentOrder.size();++i){
    if(currentOrder[i]!=targetOrder.at(i)){
      unsigned long indexInCurrentOrder=currentOrder.find(targetOrder.at(i));
      currentOrder.shift(indexInCurrentOrder, i);
      groupShift(indexInCurrentOrder, i);
    }
  }
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::adoptCurrentOrderToArray(ULongArray& currentOrder, const ULongArray& targetOrder, const unsigned long groupOffset) const {
  // shifts currentOrder and corresponding variableOrdering to resemble targetOrder[]
  for(unsigned long i=0;i<currentOrder.size();++i){
    if(currentOrder[i]!=targetOrder.at(i)){
      unsigned long indexInCurrentOrder=currentOrder.find(targetOrder.at(i));
      currentOrder.shift(indexInCurrentOrder, i);
      shift(indexInCurrentOrder, i + groupOffset);
    }
  }
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::chooseParents(ULongArray& fitnessOrMatingPool, unsigned long& firstParentIndex, unsigned long& secondParentIndex) const {
  switch(DEFAULT_SELECTION){
    case ROULETTE_WHEEL_SELECTION: {
      unsigned long probabilitySum=0;
      unsigned long i;
      for(i=0;i<POPULATION_SIZE;++i) probabilitySum+=fitnessOrMatingPool.at(i);
      unsigned long firstChoice=RandomHelper::getULong(probabilitySum);
      unsigned long secondChoice=RandomHelper::getULong(probabilitySum);
      bool firstSelected=false;
      bool secondSelected=false;
      unsigned long sum=0;
      for(i=0;i<POPULATION_SIZE;++i){
        if(firstSelected && secondSelected) return;
        sum+=fitnessOrMatingPool.at(i);
        if(firstChoice<=sum && !firstSelected){
          firstParentIndex=i;
          firstSelected=true;
        }
        if(secondChoice<=sum && !secondSelected){
          secondParentIndex=i;
          secondSelected=true;
        }
      }
      break;
    }
    case STOCHASTIC_REMAINDER_SELECTION: {
      unsigned long length=fitnessOrMatingPool.find(ULONG_MAX);
      if(length==ULONG_MAX) length=fitnessOrMatingPool.size();
      if(length<2){
        cerr << "INSUFFICIENT MATING POOL SIZE: " << length << endl;
        return;
      }
      unsigned long firstIndex, secondIndex;
      RandomHelper::getULongs(firstIndex, secondIndex, length);
      firstParentIndex=fitnessOrMatingPool[firstIndex];
      secondParentIndex=fitnessOrMatingPool[secondIndex];
      fitnessOrMatingPool[firstIndex]=fitnessOrMatingPool[length-1];
      fitnessOrMatingPool[secondIndex]=fitnessOrMatingPool[length-2];
      fitnessOrMatingPool[length-1]=ULONG_MAX;
      fitnessOrMatingPool[length-2]=ULONG_MAX;
      break;
    }
  }
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::prepareReordering(void){
  unsigned long lastGroupVariable=0;
  for(unsigned long i=0;i<T1::size();++i){
    if(!T1::isGroupMember(i)){
      if((ungrouped.size()==0)||(i-lastGroupVariable>1)) ungrouped.push_back(T2());
      ungrouped[ungrouped.size()-1]<<T1::getVariable(i);
      lastGroupVariable=i;
    }
  }
  T3::garbageCollect();
}

template <class T1, class T2, class T3>
void Reordering<T1,T2,T3>::finalizeReordering(void){
  for(unsigned long i=0;i<ungrouped.size();++i) ungrouped[i].ungroup();
  ungrouped.clear();
  T1::clearGroups();
}

#endif
