/***************************************************************************
                          varorder.h  -  description
                             -------------------
    begin                : Tue Nov 29 2005
    copyright            : (C) 2005 by Joern Ossowski
    email                : ossowski@cs.uni-bonn.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef VARORDER_H
#define VARORDER_H

#include <map>
#include <string>
#include <ctime>
#include <cstdlib>
#include <climits>
#include <sstream>

using namespace std;

template <class T> class Variable;
template <class T> class Group;
template <class T> class VarOrder;

//******************Variable*****************
//*this template stores the information for *
//*one variable                             *
//*******************************************
//*NOTE: the object which is used with this *
//*      template must implement:           *
//*       -T(unsigned long)                 *
//*       -void setLevel(unsigned long)     *
//*       -unsigned long getLevel(void)     *
//*       -bool deleteable(void)            *
//*******************************************
template <class T>
class Variable{
public:
  Variable(const string&, const unsigned long=0);
  ~Variable(void);
  string getName(void) const;
  T* getData(void) const;
  bool changeData(T*);
  void setLevel(const unsigned long);
  unsigned long getLevel(void) const;
  bool deleteable(void) const;
  void setGroup(Group<T>*);
  Group<T>* getGroup(void) const;
  Variable& operator=(const string&);
  Variable& operator=(const unsigned long);
private:
  string name;
  T* data;
  Group<T>* group;
};

template <class T>
Variable<T>::Variable(const string& name, const unsigned long level){
  this->name=name;
  data=new T(level);
  group=0;
}

template <class T>
Variable<T>::~Variable(void){
  //the destructor is called when variable (and its data) can savely be removed or in the case of quitting application
  //(in this case there is no check needed)
  delete data;
}

template <class T>
string Variable<T>::getName(void) const {
  return name;
}

template <class T>
T* Variable<T>::getData(void) const {
  return data;
}

template <class T>
bool Variable<T>::changeData(T* newData){
  //check if change is possible
  if(data==newData) return false;
  if(!data->deleteable()) return false;

  //set level for new data
  newData->setLevel(T::getSkip()+data->getLevel());

  //set new data
  data=newData;

  return true;
}

template <class T>
void Variable<T>::setLevel(const unsigned long level){
  data->setLevel(level);
}

template <class T>
unsigned long Variable<T>::getLevel(void) const {
  return data->getLevel();
}

template <class T>
bool Variable<T>::deleteable(void) const {
  return data->deleteable();
}

template <class T>
void Variable<T>::setGroup(Group<T>* newGroup){
  group=newGroup;
}

template <class T>
Group<T>* Variable<T>::getGroup(void) const {
  return group;
}

template <class T>
Variable<T>& Variable<T>::operator=(const string& newName){
  name=newName;

  return *this;
}

template <class T>
Variable<T>& Variable<T>::operator=(const unsigned long level){
  data->setLevel(level);

  return *this;
}

template <class T>
class Group{
friend class VarOrder<T>;
public:
  Group(VarOrder<T>&);
  ~Group(void);
  string getName(void) const;
  bool setName(const string&);
  bool group(const string&);
  void ungroup(void);
  bool ungroup(const string&);
  bool addVariable(const string&);
  bool removeVariable(const string&);
  string getVariableOrdering(void) const;
  void incRefCount(void);
  void decRefCount(void);
  unsigned long getRefCount(void) const;
  Variable<T>* getFirst(void) const;
  Variable<T>* getLast(void) const;
  void setFirst(Variable<T>*);
  void setLast(Variable<T>*);
private:
  string name;
  Variable<T>* firstElement;
  Variable<T>* lastElement;
  VarOrder<T>* varOrder;
  unsigned long refCount;
};

template <class T>
Group<T>::Group(VarOrder<T>& newVarOrder){
  //create an empty group
  firstElement=lastElement=0;

  varOrder=&newVarOrder;
  //create an unique group name and add group to varorder
  name=varOrder->getUniqueGroupName(this);
  varOrder->addGroup(*this);

  //no BDDgroup is pointing on this group
  refCount=0;
}

template <class T>
Group<T>::~Group(void){
}

template <class T>
string Group<T>::getName(void) const {
  return name;
}

template <class T>
bool Group<T>::setName(const string& name){
  return varOrder->renameGroup(this->name,name);
}

template <class T>
bool Group<T>::group(const string& name){
  //you can only group a variable if
  //1. the variable exists
  //2. the variable belongs to no other group
  //3. the group is empty or the variable is next to the group
  //    (a,[b,c])->([a,b,c]) or ([a,b],c)->([a,b,c]) but not (a,b,[c,d])->(b,[a,c,d])

  Variable<T>* tempVariable=varOrder->getVariable(name);
  //check 1
  if(!tempVariable) return false;
  //check 2
  if(tempVariable->getGroup()) return tempVariable->getGroup()==this;
  //check 3 (group is empty)
  if(!firstElement){
    firstElement=lastElement=tempVariable;
    //group isn't empty anymore
    varOrder->incGroupCount();
  } else {
    //check 3 (variable is next to group)
    if(tempVariable->getLevel()+1==firstElement->getLevel()) firstElement=tempVariable;
    else if(tempVariable->getLevel()==lastElement->getLevel()+1) lastElement=tempVariable;
    else return false;
  }
  //set group
  tempVariable->setGroup(this);

  return true;
}

template <class T>
void Group<T>::ungroup(void){
  //ungroups all variables
  if(firstElement){
    for(unsigned long i=firstElement->getLevel();i<=lastElement->getLevel();++i) varOrder->getVariable(i)->setGroup(0);
    firstElement=lastElement=0;
    //group is empty now
    varOrder->decGroupCount();
  }
}

template <class T>
bool Group<T>::ungroup(const string& name){
  //you can only ungroup a variable if
  //1. the variable exists
  //2. the variable belongs to no other group
  //3. the variable is the first or last variable in the group
  //    ([a,b,c])->(a,[b,c]) or ([a,b,c])->([a,b],c)

  //check 1
  Variable<T>* tempVariable=varOrder->getVariable(name);
  if(!tempVariable) return false;

  //check 3 (check 2 is done by check 3)
  if(firstElement==tempVariable){
    //this variable is the only one in the group
    if(lastElement==tempVariable){
      //group is empty now
      firstElement=lastElement=0;
      varOrder->decGroupCount();
    } else firstElement=varOrder->getVariable(tempVariable->getLevel()+1);
  } else if(lastElement==tempVariable) lastElement=varOrder->getVariable(tempVariable->getLevel()-1);
  else return false;

  //ungroup variable
  tempVariable->setGroup(0);

  return true;
}

template <class T>
bool Group<T>::addVariable(const string& name){
  //addVariable acts like group() but if check 1 fails a variable will be created at the end of the group
  //(or at the end of varorder if group is empty)
  varOrder->insertVariable(name,lastElement?lastElement->getLevel()+1:varOrder->size());

  return group(name);
}

template <class T>
bool Group<T>::removeVariable(const string& name){
  //check if variable exists
  Variable<T>* tempVariable=varOrder->getVariable(name);
  if(!tempVariable) return false;
  //check if variable is member of "this" group
  if(tempVariable->getGroup()!=this) return false;

  return varOrder->removeVariable(name);
}

template <class T>
string Group<T>::getVariableOrdering(void) const {
  string ret="[";
  if(firstElement){
    for(unsigned long i=firstElement->getLevel();i<=lastElement->getLevel();++i){
      ret+=varOrder->getVariable(i)->getName();
      if(i+1<=lastElement->getLevel()) ret+=",";
    }
  }
  ret+="]";

  return ret;
}

template <class T>
void Group<T>::incRefCount(void){
  ++refCount;
}

template <class T>
void Group<T>::decRefCount(void){
  --refCount;
}

template <class T>
unsigned long Group<T>::getRefCount(void) const {
  return refCount;
}

template <class T>
Variable<T>* Group<T>::getFirst(void) const {
  return firstElement;
}

template <class T>
Variable<T>* Group<T>::getLast(void) const {
  return lastElement;
}

template <class T>
void Group<T>::setFirst(Variable<T>* first){
  firstElement=first;
}

template <class T>
void Group<T>::setLast(Variable<T>* last){
  lastElement=last;
}

//********************VarOrder**********************
//*this template implements a common variable      *
//*ordering                                        *
//**************************************************
template <class T>
class VarOrder{
  typedef typename map<string,Group<T>*>::iterator groupIterator;
public:
  VarOrder(void);
  ~VarOrder(void);
  unsigned long size(void) const;
  unsigned long groupCount(void) const;
  bool prependVariable(const string&);
  bool appendVariable(const string&);
  bool insertVariable(const string&, const unsigned long=0);
  bool removeVariable(const string&);
  bool removeVariable(const unsigned long);
  bool renameVariable(const string&, const string&);
  bool renameVariable(const unsigned long, const string&);
  Variable<T>* getVariable(const string&) const;
  Variable<T>* getVariable(const unsigned long) const;
  bool findVariable(const string&) const;
  bool swapVariables(const unsigned long);
  bool changeData(const string&, T*);
  bool changeData(const unsigned long, T*);
  T* getDrainData(void) const;
  T* getData(const string&) const;
  T* getData(const unsigned long) const;
  bool addGroup(Group<T>&);
  bool removeGroup(const string&);
  bool renameGroup(const string&, const string&);
  Group<T>* getGroup(const string&) const;
  Group<T>* getGroup(const unsigned long) const;
  Group<T>* getNextGroup(const Group<T>*) const;
  Group<T>* getPrevGroup(const Group<T>*) const;
  void incGroupCount(void);
  void decGroupCount(void);
  string getUniqueGroupName(Group<T>* =0) const;
  string getVariableOrdering(void) const;
  string toString() const;
  void clear(void);
  void clearGroups(void);
  string operator[](const unsigned long) const;
  unsigned long operator[](const string&) const;
private:
  unsigned long count;
  unsigned long groupCounter;
  unsigned long reservedVariables;
  unsigned long skip;
  T* drains;
  Variable<T>** varArray;
  map<string,Variable<T>*> nameMap;
  map<string,Group<T>*> groupMap;
};

template <class T>
VarOrder<T>::VarOrder(void){
  //randomize
  srand(static_cast<unsigned>(time(0)));

  //variable count
  count=0;
  //variable count (count just non empty groups)
  groupCounter=0;

  drains=new T;
  drains->setLevel(ULONG_MAX);
  drains->setDrainLevel();
	
  reservedVariables=1000;
  varArray=new Variable<T>*[reservedVariables];
  skip=reservedVariables/2;	
  T::setSkip(skip);		
}

template <class T>
VarOrder<T>::~VarOrder(void){
  if(count>0){
    //delete all variables
    for(unsigned long i=skip;i<skip+count;++i) delete varArray[i];
  }
  //delete array
  delete[] varArray;

  //delete all groups
  //(be carefull if there are BDDgroups still pointing to a group)
  groupIterator it=groupMap.begin();
  while(it!=groupMap.end()) delete (it++)->second;
  
  delete drains;
}

template <class T>
unsigned long VarOrder<T>::size(void) const {
  return count;
}

template <class T>
unsigned long VarOrder<T>::groupCount(void) const {
  return groupCounter;
}

template <class T>
bool VarOrder<T>::prependVariable(const string& name){
  return insertVariable(name,0);
}

template <class T>
bool VarOrder<T>::appendVariable(const string& name){
  return insertVariable(name,count);
}

template <class T>
bool VarOrder<T>::insertVariable(const string& name, const unsigned long pos){
  Variable<T>** tempVarArray;
  Variable<T>* tempVariable;
  unsigned long i;

  //check if position of the new variable is inside [0,count]
  if(pos>count) return false;
  //empty variable name is not allowed
  if(name=="") return false;
  //check if name is already used
  if(nameMap.find(name)!=nameMap.end()) return false;

  //create new variable
  tempVariable=new Variable<T>(name,skip+pos);

  //check if resize is needed
  bool resizeNeeded=false;
  bool moveUp;
  if(pos<=count/2){
    moveUp=true;
    if(skip==0) resizeNeeded=true;
  } else {
    moveUp=false;
    if(skip+count>=reservedVariables) resizeNeeded=true;
  }
  
  if(resizeNeeded){
    unsigned int resizeValue=1000;
    //resize array
    reservedVariables+=resizeValue;
    //calc skip
    unsigned int newSkip=(reservedVariables-count)/2;
    //create temporary varArray
    tempVarArray=new Variable<T>*[reservedVariables];
    
    //copy old values (before pos)
    if(pos>0) memcpy(tempVarArray+newSkip,varArray+skip,sizeof(Variable<T>*)*pos);
    for(i=0;i<pos;++i){
      //update level information
      *(tempVarArray[newSkip+i])=newSkip+i;
    }
    
    tempVarArray[newSkip+pos]=tempVariable;
    //update level information
    *(tempVarArray[newSkip+pos])=newSkip+pos;
    
    //copy old values (after pos)
    if(pos<count) memcpy(tempVarArray+(newSkip+pos+1),varArray+(skip+pos),sizeof(Variable<T>*)*(count-pos));
    for(i=pos;i<count;++i){
      //update level information
      *(tempVarArray[newSkip+i+1])=newSkip+i+1;
    }
    
    skip=newSkip;
    T::setSkip(skip);
    
    //delete old varArray
    delete[] varArray;
    
    //update varArray
    varArray=tempVarArray;
  } else {
    if(moveUp){
      //copy old values (before pos)
      for(i=0;i<pos;++i){
	//copy value
	varArray[skip+i-1]=varArray[skip+i];
	//update level information
	*(varArray[skip+i-1])=skip+i-1;
      }
      varArray[skip+pos-1]=tempVariable;
      //update level information
      *(varArray[skip+pos-1])=skip+pos-1;
      --skip;
      T::setSkip(skip);
    } else {
      //copy old values (after pos)
      for(i=count;i>pos;--i){
	//copy value
	varArray[skip+i]=varArray[skip+i-1];
	//update level information
	*(varArray[skip+i])=skip+i;
      }
      varArray[skip+pos]=tempVariable;
    }
  }
  
  //check if position is inside a group
  if((pos>0) && (pos<count)){
    if(varArray[skip+pos-1]->getGroup()){
      if(varArray[skip+pos-1]->getGroup()==varArray[skip+pos+1]->getGroup()) varArray[skip+pos]->setGroup(varArray[skip+pos-1]->getGroup());
    }
  }

  //update count
  ++count;

  //insert name into nameMap
  nameMap[name]=tempVariable;

  return true;
}

template <class T>
bool VarOrder<T>::removeVariable(const string& name){
  //check if name doesn't exist
  if(nameMap.find(name)==nameMap.end()) return false;

  return removeVariable(nameMap.find(name)->second->getLevel());
}

template <class T>
bool VarOrder<T>::removeVariable(const unsigned long index){
  unsigned long i;

  //check if index has a valid value
  if(index>=count) return false;

  //check if variable isn't deletable
  if(!varArray[skip+index]->deleteable()) return false;
  
  //ungroup variable if necessary
  if(varArray[skip+index]->getGroup()) varArray[skip+index]->getGroup()->ungroup(varArray[skip+index]->getName());

  //remove name from nameMap if neccessary
  nameMap.erase(varArray[skip+index]->getName());

  //delete variable
  delete varArray[skip+index];
  
  if(count==1){
    skip=reservedVariables/2;
    T::setSkip(skip);
  } else {
    //check if first variable is removed (in this case it is only necessary to increase skip)
    if(index==0){
      ++skip;
      T::setSkip(skip);
    } else {
      //check if move down or move up
      if(index<=count/2){
	//copy old variables (before index)
	Variable<T>** target=&varArray[skip+index];
	Variable<T>** source=target-1;
	Variable<T>** end=&varArray[skip-1];
	i=skip+index;
	while(source!=end){
	  *((*target)=(*source))=i;
	  target=(source--);
	  --i;
	}
	++skip;
	T::setSkip(skip);				
      } else {
        //copy old variables (after index)
	if(index+1<count){
	  Variable<T>** target=&varArray[skip+index];
	  Variable<T>** source=target+1;
	  Variable<T>** end=&varArray[skip+count];
	  i=skip+index;
	  while(source!=end){
	    *((*target)=(*source))=i;
	    target=(source++);
	    ++i;
	  }
	}
      }
    }
  }
  
  //update count
  --count;

  return true;
}

template <class T>
bool VarOrder<T>::renameVariable(const string& oldName, const string& newName){
  //find index and rename variable
  return renameVariable((*this)[oldName],newName);
}

template <class T>
bool VarOrder<T>::renameVariable(const unsigned long index, const string& newName){
  //check if index is possible
  if(index>=count) return false;

  if(newName=="") return false;

  //check if newName is already in use
  if(nameMap.find(newName)!=nameMap.end()) return false;

  //remove old name
  nameMap.erase(varArray[skip+index]->getName());

  //insert newName into nameMap
  nameMap[newName]=varArray[skip+index];

  //set new name
  (*varArray[skip+index])=newName;

  return true;
}

template <class T>
Variable<T>* VarOrder<T>::getVariable(const string& name) const {
  return (findVariable(name))?getVariable(nameMap.find(name)->second->getLevel()):0;
}

template <class T>
Variable<T>* VarOrder<T>::getVariable(const unsigned long index) const {
  return (index<count)?varArray[skip+index]:0;
}

template <class T>
bool VarOrder<T>::findVariable(const string& name) const {
  return (nameMap.find(name)!=nameMap.end());
}

template <class T>
bool VarOrder<T>::swapVariables(const unsigned long index){
  if(index+1>=count) return false;

  //swap variables
  Variable<T>* tempVariable=varArray[skip+index];
  varArray[skip+index]=varArray[skip+index+1];
  varArray[skip+index+1]=tempVariable;

  //set new indices
  varArray[skip+index]->setLevel(skip+index);
  varArray[skip+index+1]->setLevel(skip+index+1);

  //be carefull with this function
  //following situations could occure
  //([a,b],{c,d})->([a,{c,b],d})
  //(a,[b,c])->([b,a,c])
  //this behaviour is needed for reordering algorithmns on groups
  Group<T>* tempGroup=varArray[skip+index]->getGroup();
  if(tempGroup){
    if(tempGroup==varArray[skip+index+1]->getGroup()){
      //check for ([a,b,c])->([b,a,c]), ([a,b,c])->([a,c,b]) or ([a,b])->([b,a])
      if(varArray[skip+index+1]==tempGroup->getFirst()) tempGroup->setFirst(varArray[skip+index]);
      if(varArray[skip+index]==tempGroup->getLast()) tempGroup->setLast(varArray[skip+index+1]);
    }
  }

  return true;
}

template <class T>
bool VarOrder<T>::changeData(const string& name, T* data){
  if(!findVariable(name)) return false;

  //get index and return changeData(index,data)
  return changeData(nameMap.find(name)->second->getLevel(),data);
}

template <class T>
bool VarOrder<T>::changeData(unsigned long index, T* data){
  //check if index is possible
  if(index>=count) return false;

  return varArray[skip+index]->changeData(data);
}

template <class T>
T* VarOrder<T>::getDrainData(void) const {
  return drains;
}

template <class T>
T* VarOrder<T>::getData(const string& name) const {
  if(!findVariable(name)) return 0;

  //get index and return data
  return getData(nameMap.find(name)->second->getLevel());
}

template <class T>
T* VarOrder<T>::getData(const unsigned long index) const {
  //check if index is possible
  if(index>=count) return 0;

  return varArray[skip+index]->getData();
}

template <class T>
bool VarOrder<T>::addGroup(Group<T>& group){
  if(groupMap.find(group.getName())!=groupMap.end()) return false;
  groupMap[group.getName()]=&group;

  return true;
}

template <class T>
bool VarOrder<T>::removeGroup(const string& groupName){
  //check if group exists
  Group<T>* tempGroup=getGroup(groupName);
  if(!tempGroup) return false;
  //check if BDDgroups are still pointing to the group
  if(tempGroup->refCount) return false;

  tempGroup->ungroup();
  groupMap.erase(groupName);
  delete tempGroup;

  return true;
}

template <class T>
bool VarOrder<T>::renameGroup(const string& oldGroupName, const string& newGroupName){
  if(groupMap.find(oldGroupName)==groupMap.end()) return false;
  if(groupMap.find(newGroupName)!=groupMap.end()) return false;

  //set new group
  groupMap[newGroupName]=groupMap[oldGroupName];
  //erase old entry (!!before!! setting newGroupName otherwise oldGroupName would equal newGroupName(reference))
  groupMap.erase(oldGroupName);
  //set newGroupName
  groupMap[newGroupName]->name=newGroupName;

  return true;
}

template <class T>
Group<T>* VarOrder<T>::getGroup(const string& name) const {
  if(groupMap.find(name)==groupMap.end()) return 0;

  return groupMap.find(name)->second;
}

template <class T>
Group<T>* VarOrder<T>::getGroup(const unsigned long index) const {
  //empty groups are not accessable through this function

  if(index>=groupCount()) return 0;

  unsigned long i=0;
  unsigned long groupIndex=0;
  while(i<count){
    if(varArray[skip+i]->getGroup()){
      if(index==groupIndex) return varArray[skip+i]->getGroup();
      ++groupIndex;
      //overjump the whole group
      i=varArray[skip+i]->getGroup()->getLast()->getLevel();
    }
    ++i;
  }

  return 0;
}

template <class T>
Group<T>* VarOrder<T>::getNextGroup(const Group<T>* group) const {
  //an empty group has no successor
  if(!group->getFirst()) return 0;

  for(unsigned long i=group->getLast()->getLevel()+1;i<count;++i){
    if(varArray[skip+i]->getGroup()) return varArray[skip+i]->getGroup();
  }

  return 0;
}

template <class T>
Group<T>* VarOrder<T>::getPrevGroup(const Group<T>* group) const {
  //an empty group has no predeccessor
  if(!group->getFirst()) return 0;

  for(unsigned long i=group->getFirst()->getLevel();i>0;--i){
    if(varArray[skip+i-1]->getGroup()) return varArray[skip+i-1]->getGroup();
  }

  return 0;
}

template <class T>
void VarOrder<T>::incGroupCount(void){
  ++groupCounter;
}

template <class T>
void VarOrder<T>::decGroupCount(void){
  --groupCounter;
}

template <class T>
string VarOrder<T>::getUniqueGroupName(Group<T>* group) const {
  //get name from address in memory
  ostringstream tempStream;
  tempStream<<group;
  string tempString=tempStream.str();

  //if name is already used try to "guess" a new name
  while(groupMap.find(tempString)!=groupMap.end()){
    switch(rand()%3){
      case 0: tempString+=char(int('0')+rand()%10);
        break;
      case 1: tempString+=char(int('A')+rand()%26);
        break;
      case 2: tempString+=char(int('a')+rand()%26);
        break;
    }
  }

  return tempString;
}

template <class T>
string VarOrder<T>::getVariableOrdering(void) const {
  string ret="(";

  unsigned long i=0;
  while(i<count){
    if(!varArray[skip+i]->getGroup()){
      ret+=varArray[skip+i]->getName();
      //check if "," is needed
      if(i+1<count) ret+=",";
    } else {
      ret+=varArray[skip+i]->getGroup()->getVariableOrdering();
      //overjump the whole group
      i=varArray[skip+i]->getGroup()->getLast()->getLevel();
      //check if "," is needed
      if(i+1<count) ret+=",";
    }
    ++i;
  }
  ret+=")";

  return ret;
}

template <class T>
string VarOrder<T>::toString() const {
  std::stringstream out;
  for (unsigned int i=0; i<size(); i++) {
    out << (*this)[i] << "\n";
  }

  return out.str();
}

template <class T>
void VarOrder<T>::clear(void){
  groupIterator it=groupMap.begin();

  //remove all unused variables
  unsigned long newIndex=0;
  unsigned long newCount=count;
  unsigned long newSkip=skip;
  for(unsigned long index=0;index<count;++index){
    //check if variable isn't deletable
    if(!varArray[skip+index]->deleteable()){
      if(index!=newIndex){
	//first variables have been removed (otherwise index would be zero) so skip must be updated
	if(newIndex==0) newSkip=skip+index;
	*(varArray[newSkip+newIndex]=varArray[skip+index])=newSkip+newIndex;
      }
      ++newIndex;
    } else {
      //ungroup variable if necessary
      if(varArray[skip+index]->getGroup()) varArray[skip+index]->getGroup()->ungroup(varArray[skip+index]->getName());
      
      //remove name from nameMap if neccessary
      nameMap.erase(varArray[skip+index]->getName());
      
      //delete variable
      delete varArray[skip+index];
      
      --newCount;
    }
  }
  
  count=newCount;
  skip=newSkip;
  T::setSkip(skip);
  
  //remove all "unused" groups
  while(it!=groupMap.end()) removeGroup((it++)->second->getName());
}

template <class T>
void VarOrder<T>::clearGroups(void){
  groupIterator it=groupMap.begin();

  //remove all "unused" groups
  while(it!=groupMap.end()) removeGroup((it++)->second->getName());
}

template <class T>
string VarOrder<T>::operator[](const unsigned long index) const {
  if(index>=count) return "[ERROR]";
  else return varArray[skip+index]->getName();
}

template <class T>
unsigned long VarOrder<T>::operator[](const string& name) const {
  if(!findVariable(name)) return ULONG_MAX;

  //return index
  return nameMap.find(name)->second->getLevel();
}
#endif
