/***************************************************************************
                          varset.cpp  -  description
                             -------------------
    begin                : Tue Apr 28 2004
    copyright            : (C) 2004 by Joern Ossowski
    email                : ossowski@cs.uni-bonn.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "varset.h"
#include "constants.h"
#include <algorithm>

using namespace std;

VarSet::VarSet(void){
}

/**
 * \brief Copies a VarSet
 */
VarSet::VarSet(const VarSet& varSet){
  this->varSet=varSet.varSet;
  this->stringVector=varSet.stringVector;
}

VarSet::~VarSet(void){
}

/**
 * \brief Clears all variables inside the VarSet.
 */
void VarSet::clear(void){
  varSet.clear();
  stringVector.clear();
}

/**
 * \brief Returns if a variable is inside the VarSet.
 */
bool VarSet::find(const string& variable) const {
  return (varSet.find(variable)!=varSet.end());
}

/**
 * \brief Inserts a variable at the end of the VarSet.
 *
 * The variable is inserted into the set and is positioned at the end.
 * You have to be carefull with inserting variable because there is no way to reorder a
 * VarSet.
 */
bool VarSet::insert(const string& variable){
  if(find(variable)) return false;
  if(variable.find_first_of(badChars)!=string::npos) return false;

  varSet.insert(variable);
  stringVector.push_back(variable);

  return true;
}

/**
 * \brief Removes a variable from the VarSet.
 *
 * This operation needs O(n) steps to perform the removemend of an element because
 * rearranging the vector needs this time.
 */
bool VarSet::remove(const string& variable){
  set<string>::iterator it1=varSet.find(variable);
  if(it1==varSet.end()) return false;

  varSet.erase(it1);
  vector<string>::iterator it2=stringVector.begin();
  while(*it2!=variable) ++it2;
  stringVector.erase(it2);

  return true;
}

/**
* \brief Returns the reversed VarSet (a copy).
 */
VarSet VarSet::reverse(void) const {
  VarSet ret;
	ret.varSet=varSet;
	ret.stringVector=vector<string>(stringVector.size());
	std::reverse_copy(stringVector.begin(), stringVector.end(), ret.stringVector.begin());	
  return ret;
}

/**
 * \brief Copies a VarSet.
 */
VarSet& VarSet::operator=(const VarSet& varSet){
  if(this!=&varSet){
    this->varSet=varSet.varSet;
    this->stringVector=varSet.stringVector;
  }

  return *this;
}

/**
 * \brief Inserts a variable at the end of the VarSet.
 *
 * Another way to insert a variable.
 * \sa insert(const string&)
 */
VarSet& VarSet::operator<<(const string& variable){
  insert(variable);

  return *this;
}

/**
 * \brief Removes a variable from the VarSet.
 *
 * \sa remove(const string&)
 */
VarSet& VarSet::operator>>(const string& variable){
  remove(variable);

  return *this;
}

/**
 * \brief Accesses an element in the VarSet.
 *
 * If the requestes element is out of range "[ERROR]" will be returned otherwise
 * the element. The elements are ordering in the same way as the were inserted into the
 * VarSet.
 * \todo [ERROR] in =ERROR=?
 * \sa size(), insert(const string&)
 */
string VarSet::operator[](const unsigned long index) const {
  return (index>=size())?"[ERROR]":stringVector[index];
}

/**
 * \brief Returns the number of elements in the VarSet.
 */
unsigned long VarSet::size(void) const {
  return varSet.size();
}
