/***************************************************************************
                          zaddfunction.cpp  -  description
                             -------------------
    begin                : Thu Apr 15 2004
    copyright            : (C) 2004 by Joern Ossowski
    email                : ossowski@cs.uni-bonn.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "zaddfunction.h"
#include "zadd.h"
#include "zaddnode.h"
#include "zaddvarorder.h"
#include "../common/assignment.h"
#include "../common/varset.h"
#include "../common/stringhelper.h"
#include "../common/varorder.h"

#include <stack>
#include <iostream>

inline unsigned long pow2(const unsigned long n){
  return (1 << n);
}

inline void addFunction(ZADDNode* node, ZADDNode* varSet){
  if(node) node->incNodeCount();
  if(varSet) varSet->incNodeCount();
}

inline void removeFunction(ZADDNode* node, ZADDNode* varSet){
  if(node) node->deleteNode();
  if(varSet) varSet->deleteNode();
}

//************************ZADDFunction************************
ZADD* ZADDFunction::BDD=0;

ZADDFunction::ZADDFunction(const ZADDFunction& function){
  rootNode=function.rootNode;
  varSet=function.varSet;
  addFunction(rootNode,varSet);
}

ZADDFunction::ZADDFunction(const Assignment& assignment, const double value){
  if(assignment.size()==0){
    rootNode=BDD->zeroDrain;
    varSet=BDD->oneDrain;
    addFunction(rootNode,varSet);

    return;
  } else {
    rootNode=BDD->findOrAddDrain(value);
    varSet=BDD->oneDrain;
    if(rootNode!=BDD->zeroDrain){
      for(unsigned long i=0;i<assignment.size();++i) ZADDVarOrder::addVariable(assignment[i].variable);

      for(unsigned long i=BDD->variableOrdering.size();i>0;--i){
        if(assignment.find(BDD->variableOrdering[i-1])){
          if(assignment.at(BDD->variableOrdering[i-1])){
            rootNode=BDD->findOrAdd(i-1,rootNode,BDD->zeroDrain);
          }
          varSet=BDD->findOrAdd(i-1,varSet,BDD->zeroDrain);
        }
      }
    }
  }

  addFunction(rootNode,varSet);
}

ZADDFunction::ZADDFunction(char* buffer, const double value){
  rootNode=BDD->findOrAddDrain(value);
  varSet=BDD->oneDrain;
  if(rootNode!=BDD->zeroDrain){
    for(unsigned long i=BDD->variableOrdering.size();i>0;--i){
      switch(buffer[i-1]){
        case '0': varSet=BDD->findOrAdd(i-1,varSet,BDD->zeroDrain);
                  break;
        case '1': rootNode=BDD->findOrAdd(i-1,rootNode,BDD->zeroDrain);
                  varSet=BDD->findOrAdd(i-1,varSet,BDD->zeroDrain);
                  break;
        default: break;
      }
    }
  }

  addFunction(rootNode,varSet);
}

ZADDFunction::ZADDFunction(const VarSet& varSet){
  if(varSet.size()==0){
    rootNode=BDD->zeroDrain;
    this->varSet=BDD->oneDrain;
    addFunction(rootNode,this->varSet);

    return;
  } else {
    rootNode=BDD->oneDrain;
    for(unsigned long i=0;i<varSet.size();++i) ZADDVarOrder::addVariable(varSet[i]);

    for(unsigned long i=BDD->variableOrdering.size();i>0;--i){
      if(varSet.find(BDD->variableOrdering[i-1])){
        rootNode=BDD->findOrAdd(i-1,rootNode,BDD->zeroDrain);
      }
    }
    this->varSet=rootNode;
  }

  addFunction(rootNode,this->varSet);
}

ZADDFunction::ZADDFunction(const string& variable){
  ZADDFunction tempFunction=ZADDFunction::BDD->projection(variable);

  rootNode=tempFunction.rootNode;
  varSet=tempFunction.varSet;

  addFunction(rootNode,varSet);
}

ZADDFunction::ZADDFunction(const double value){
  rootNode=ZADD::BDD->createConstantFunction(value);
  varSet=ZADD::BDD->oneDrain;

  addFunction(rootNode,varSet);
}

ZADDFunction::ZADDFunction(void){
  rootNode=0;
  varSet=0;
}

ZADDFunction::~ZADDFunction(void){
  if(rootNode) removeFunction(rootNode,varSet);
}

ZADDFunction ZADDFunction::ERROR_FUNCTION(void){
  return ZADDFunction((ZADDNode*) 0,(ZADDNode*) 0);
}

ZADDFunction ZADDFunction::ONE(void){
  return ZADDFunction(BDD->oneDrain,BDD->oneDrain);
}

ZADDFunction ZADDFunction::ZERO(void){
  return ZADDFunction(BDD->zeroDrain,BDD->oneDrain);
}

ZADDFunction ZADDFunction::createConstantFunction(const double value){
  return ZADDFunction(ZADDFunction::BDD->createConstantFunction(value),BDD->oneDrain);
}

ZADDFunction ZADDFunction::createStabilityFunction(const VarSet& set1, const VarSet& set2){
  if(set1.size()!=set2.size()) return ZADDFunction::ERROR_FUNCTION();

  ZADDFunction temp=ZADDFunction::ONE();
  for(unsigned long i=0;i<set1.size();++i){
    temp&=EQUIV(ZADDFunction(set1[i]),ZADDFunction(set2[i]));
  }

  return temp;
}

ZADDFunction ZADDFunction::projection(const string& variable){
  return ZADDFunction::BDD->projection(variable);
}

ZADDFunction ZADDFunction::projection(const unsigned long index){
  return ZADDFunction::BDD->projection(index);
}

unsigned long ZADDFunction::size(void) const {
  set<ZADDNode*> visited;
  stack<ZADDNode*> nodeStack;
  ZADDNode* node;

  if(!rootNode) return 0;

  visited.insert(rootNode);
  visited.insert(varSet);
  nodeStack.push(rootNode);
  nodeStack.push(varSet);
  while(nodeStack.size()>0){
    node=nodeStack.top();
    nodeStack.pop();
    if(!node->isDrain()){
      if(visited.find(node->getSucc0())==visited.end()){
        visited.insert(node->getSucc0());
        nodeStack.push(node->getSucc0());
      }
      if(visited.find(node->getSucc1())==visited.end()){
        visited.insert(node->getSucc1());
        nodeStack.push(node->getSucc1());
      }
    }
  }

  return visited.size();
}

std::set<ZADDNode*> ZADDFunction::getNodes() const {
  set<ZADDNode*> visited;
  stack<ZADDNode*> nodeStack;
  ZADDNode* node;

  if(!rootNode) return visited;

  visited.insert(rootNode);
  nodeStack.push(rootNode);
  while (nodeStack.size()>0){
    node=nodeStack.top();
    nodeStack.pop();
    if(!node->isDrain()){
      if(visited.find(node->getSucc0())==visited.end()){
        visited.insert(node->getSucc0());
        nodeStack.push(node->getSucc0());
      }
      if(visited.find(node->getSucc1())==visited.end()){
        visited.insert(node->getSucc1());
        nodeStack.push(node->getSucc1());
      }
    }
  }

  return visited;
}


ZADDFunction& ZADDFunction::operator=(const ZADDFunction& function){
  //insert function
  addFunction(function.rootNode,function.varSet);
  //remove function
  removeFunction(rootNode,varSet);
  //copy values
  rootNode=function.rootNode;
  varSet=function.varSet;

  return *this;
}

ZADDFunction ZADDFunction::operator+(const ZADDFunction& function) const {
  return ZADDFunction::BDD->addition(*this,function);
}

ZADDFunction& ZADDFunction::operator+=(const ZADDFunction& function){
  return *this=ZADDFunction::BDD->addition(*this,function);
}

ZADDFunction ZADDFunction::operator-(const ZADDFunction& function) const {
  return ZADDFunction::BDD->subtraction(*this,function);
}

ZADDFunction& ZADDFunction::operator-=(const ZADDFunction& function){
  return *this=ZADDFunction::BDD->subtraction(*this,function);
}

ZADDFunction ZADDFunction::operator*(const ZADDFunction& function) const {
  return ZADDFunction::BDD->multiplication(*this,function);
}

ZADDFunction& ZADDFunction::operator*=(const ZADDFunction& function){
  return *this=ZADDFunction::BDD->multiplication(*this,function);
}

ZADDFunction ZADDFunction::operator|(const ZADDFunction& function) const {
  return ZADDFunction::BDD->OR(*this,function);
}

ZADDFunction& ZADDFunction::operator|=(const ZADDFunction& function){
  return *this=ZADDFunction::BDD->OR(*this,function);
}

ZADDFunction ZADDFunction::operator&(const ZADDFunction& function) const {
  return ZADDFunction::BDD->AND(*this,function);
}

ZADDFunction& ZADDFunction::operator&=(const ZADDFunction& function){
  return *this=ZADDFunction::BDD->AND(*this,function);
}

ZADDFunction ZADDFunction::operator^(const ZADDFunction& function) const {
  return !ZADDFunction::BDD->equal(*this,function);
}

ZADDFunction& ZADDFunction::operator^=(const ZADDFunction& function){
  return *this=!ZADDFunction::BDD->equal(*this,function);
}

ZADDFunction ZADDFunction::operator!(void) const {
  ZADD zadd;
  return zadd.NOT(*this);
}

bool ZADDFunction::operator==(const ZADDFunction& function) const {
  if(rootNode==0) return (function.rootNode==0);
  if(varSet==function.varSet) return (rootNode==function.rootNode);
  return (ZADDFunction::BDD->equiv(rootNode,varSet,function.rootNode,function.varSet));
}

bool ZADDFunction::operator!=(const ZADDFunction& function) const {
  return !((*this)==function);
}

bool ZADDFunction::operator<(const ZADDFunction& function) const {
  return LESS(*this,function)==1.0;
}

bool ZADDFunction::operator<=(const ZADDFunction& function) const {
  return LESS_EQUAL(*this,function)==1.0;
}

bool ZADDFunction::operator>(const ZADDFunction& function) const {
  return LESS(function,*this)==1.0;
}

bool ZADDFunction::operator>=(const ZADDFunction& function) const {
  return LESS_EQUAL(function,*this)==1.0;
}

bool ZADDFunction::isConstant(void) const {
  if(rootNode){
    if(rootNode->isDrain()){
      if((rootNode==BDD->oneDrain)&&(varSet!=BDD->oneDrain)) return false;
      else return true;
    } else return false;
  }

  return false;
}

double ZADDFunction::getValue(void) const {
  return isConstant()?rootNode->getValue():0.0;
}

string ZADDFunction::getFirstVariable(void) const {
  if(!rootNode) return "[ERROR]";
  else {
    ZADDNode* node=rootNode;
    ZADDNode* nodeVarSet=varSet;

    while(!nodeVarSet->isDrain()){
      if(node->getUniqueTable()==nodeVarSet->getUniqueTable()){
        if(node->getSucc0()!=node->getSucc1()) return ZADDVarOrder::getVariable(nodeVarSet->getLevel());
        else {
          node=node->getSucc0();
          nodeVarSet=nodeVarSet->getSucc1();
        }
      } else return ZADDVarOrder::getVariable(nodeVarSet->getLevel());
    }
    return "[ERROR]";
  }
}

unsigned long ZADDFunction::getFirstVariableIndex(void) const {
  if(!rootNode) return ULONG_MAX;
  else {
    ZADDNode* node=rootNode;
    ZADDNode* nodeVarSet=varSet;

    while(!nodeVarSet->isDrain()){
      if(node->getUniqueTable()==nodeVarSet->getUniqueTable()){
        if(node->getSucc0()!=node->getSucc1()) return nodeVarSet->getLevel();
        else {
          node=node->getSucc0();
          nodeVarSet=nodeVarSet->getSucc1();
        }
      } else return nodeVarSet->getLevel();
    }
    return ULONG_MAX;
  }
}

ZADDFunction ZADDFunction::cofactor(const string& variable, const bool value) const {
  return cofactor(ZADDVarOrder::getVariable(variable),value);
}

ZADDFunction ZADDFunction::cofactor(const unsigned long index, const bool value) const {
  return ZADDFunction::BDD->cofactor(*this,index,value);
}

ZADDFunction ZADDFunction::cofactor(const ZADDFunction& assignment) const {
  return ZADDFunction::BDD->cofactor(*this,assignment);
}

ZADDFunction ZADDFunction::subSet(const string& variable, const bool value) const {
  if(!ZADDVarOrder::findVariable(variable)) return (*this);

  ZADDFunction tempVar=ZADDFunction::projection(variable);
  if(!value) tempVar=!tempVar;

  return ((*this)*tempVar);
}

ZADDFunction ZADDFunction::subSet(const unsigned long index, const bool value) const {
  if((index>=ZADDVarOrder::size())||(!rootNode)) return (*this);

  ZADDFunction tempVar=ZADDFunction::projection(index);
  if(!value) tempVar=!tempVar;

  return ((*this)*tempVar);
}

ZADDFunction ZADDFunction::subSet(const ZADDFunction& assignment) const {
  return ((*this)*assignment);
}

ZADDFunction ZADDFunction::change(void) const {
  return ZADDFunction::BDD->change(*this);
}

ZADDFunction ZADDFunction::change(const string& variable) const {
  return change(ZADDVarOrder::getVariable(variable));
}

ZADDFunction ZADDFunction::change(const unsigned long index) const {
  return ZADDFunction::BDD->change(*this,index);
}

ZADDFunction ZADDFunction::change(const ZADDFunction& cubeSet) const {
  return ZADDFunction::BDD->change(*this,cubeSet);
}

ZADDFunction ZADDFunction::exists(const ZADDFunction& varSet) const {
  return ZADDFunction::BDD->exists(*this,varSet);
}

ZADDFunction ZADDFunction::exists(const string& variable) const {
  return exists(ZADDVarOrder::getVariable(variable));
}

ZADDFunction ZADDFunction::exists(const unsigned long index) const {
  return ZADDFunction::BDD->exists(*this,index);
}

ZADDFunction ZADDFunction::forall(const ZADDFunction& varSet) const {
  return ZADDFunction::BDD->forall(*this,varSet);
}

ZADDFunction ZADDFunction::forall(const string& variable) const {
  return forall(ZADDVarOrder::getVariable(variable));
}

ZADDFunction ZADDFunction::forall(const unsigned long index) const {
  return ZADDFunction::BDD->forall(*this,index);
}

ZADDFunction ZADDFunction::equivalenceClasses(const ZADDFunction& pattern, const ZADDFunction& cubeSet) const {
  return ZADDFunction::BDD->equivalenceClasses(*this,pattern,cubeSet);
}

ZADDFunction ZADDFunction::constrain(const ZADDFunction& g) const {
  return ZADDFunction::BDD->constrain(*this,g);
}

ZADDFunction ZADDFunction::sum(const ZADDFunction& varset) const {
  return ZADDFunction::BDD->sum(*this,varset);
}

ZADDFunction ZADDFunction::product(const ZADDFunction& varset) const {
  return ZADDFunction::BDD->product(*this,varset);
}

ZADDFunction ZADDFunction::maximum(const ZADDFunction& varset) const {
  return ZADDFunction::BDD->maximum(*this,varset);
}

double ZADDFunction::max(void) const {
  return ZADDFunction::BDD->max(*this);
}

ZADDFunction ZADDFunction::minimum(const ZADDFunction& varset) const {
  return ZADDFunction::BDD->minimum(*this,varset);
}

double ZADDFunction::min(void) const {
  return ZADDFunction::BDD->min(*this);
}

ZADDFunction ZADDFunction::oneOver(void) const {
  return ZADDFunction::BDD->oneOver(*this);
}

ZADDFunction ZADDFunction::compose(const string& variable, const ZADDFunction& function) const {
  return compose(ZADDVarOrder::getVariable(variable),function);
}

ZADDFunction ZADDFunction::compose(const unsigned long index, const ZADDFunction& function) const {
  if(index>=ZADDVarOrder::size()) return (*this);
  else return ITE(function,cofactor(index,1),cofactor(index,0));
}

ZADDFunction ZADDFunction::moveUp(const unsigned int skip) const {
  return ZADDFunction::BDD->moveUp(*this,skip);
}

ZADDFunction ZADDFunction::moveUp(const ZADDFunction& varSet, const unsigned int skip) const {
  return ZADDFunction::BDD->moveUp(*this,varSet,skip);
}

ZADDFunction ZADDFunction::moveDown(const unsigned int skip) const {
  return ZADDFunction::BDD->moveDown(*this,skip);
}

ZADDFunction ZADDFunction::moveDown(const ZADDFunction& varSet, const unsigned int skip) const {
  return ZADDFunction::BDD->moveDown(*this,varSet,skip);
}

void ZADDFunction::printEssentialVariables(void) const {
  string ret;
  set<string> variables;
  set<ZADDNode*> visited;
  stack<ZADDNode*> nodeStack;
  ZADDNode* node;

  if(!varSet) return;

  visited.insert(varSet);
  nodeStack.push(varSet);
  while(nodeStack.size()>0){
    node=nodeStack.top();
    nodeStack.pop();
    if(!node->isDrain()){
      if(variables.find(ZADDVarOrder::getVariable(node->getLevel()))==variables.end()) variables.insert(ZADDVarOrder::getVariable(node->getLevel()));
      if(visited.find(node->getSucc0())==visited.end()){
        visited.insert(node->getSucc0());
        nodeStack.push(node->getSucc0());
      }
      if(visited.find(node->getSucc1())==visited.end()){
        visited.insert(node->getSucc1());
        nodeStack.push(node->getSucc1());
      }
    }
  }

  for(unsigned long i=0;i<ZADDVarOrder::size();++i){
    if(variables.find(ZADDVarOrder::getVariable(i))!=variables.end()) ret+=ZADDVarOrder::getVariable(i)+",";
  }

  cout << ret.substr(0,ret.length()-1) << endl;
}

set<string> ZADDFunction::getEssentialVariables(void) const {
  set<string> variables;
  set<ZADDNode*> visited;
  stack<ZADDNode*> nodeStack;
  ZADDNode* node;

  if(!varSet) return variables;

  visited.insert(varSet);
  nodeStack.push(varSet);
  while(nodeStack.size()>0){
    node=nodeStack.top();
    nodeStack.pop();
    if(!node->isDrain()){
      if(variables.find(ZADDVarOrder::getVariable(node->getLevel()))==variables.end()) variables.insert(ZADDVarOrder::getVariable(node->getLevel()));
      if(visited.find(node->getSucc0())==visited.end()){
        visited.insert(node->getSucc0());
        nodeStack.push(node->getSucc0());
      }
      if(visited.find(node->getSucc1())==visited.end()){
        visited.insert(node->getSucc1());
        nodeStack.push(node->getSucc1());
      }
    }
  }

  return variables;
}

ZADDFunction::operator Assignment() const {
  Assignment assignment;

  if(!rootNode) return assignment;

  //initially, every variable gets the value false
  for(unsigned long i=0;i<ZADDVarOrder::size();++i) assignment[ZADDVarOrder::getVariable(i)]=false;

  ZADDFunction::BDD->getAssignment(rootNode,assignment);

  return assignment;
}

ZADDFunction ZADDFunction::getAssignment(void) const {
  return BDD->getAssignment(*this);
}

void ZADDFunction::getAssignment(char* buffer) const {
  ZADDFunction::BDD->getAssignment(*this,buffer);
}

double ZADDFunction::countAssignments(void) const {
  return ZADDFunction::BDD->countAssignments(*this);
}

double ZADDFunction::countAssignments(const ZADDFunction& cubeSet) const {
  throw "UNIMPLEMENTED";
}
 
ZADDFunction ZADDFunction::getRandomAssignment(const ZADDFunction& cube_set) const {
  throw "UNIMPLEMENTED";
}


void ZADDFunction::printPaths(void) const {
  char minterm[ZADDVarOrder::size()];

  if(!rootNode) return;
  printPaths(0,rootNode,varSet,minterm);
}

double ZADDFunction::countPaths(void) const {
  if(!rootNode) return 0.0;

  map<ZADDNode*,double> cache;

  return countPaths(rootNode,varSet,&cache);
}

ZADDFunction ZADDFunction::getPath(void) const {
  return ZADDFunction::BDD->getPath(*this);
}

void ZADDFunction::getPath(char* buffer) const {
  ZADDFunction::BDD->getPath(*this,buffer);
}

//****************************private**************************
ZADDFunction::ZADDFunction(const ZADDNode* rootNode, const ZADDNode* varSet){
  this->rootNode=(ZADDNode*) rootNode;
  this->varSet=(ZADDNode*) varSet;

  addFunction(this->rootNode,this->varSet);
}

void ZADDFunction::printPaths(const unsigned long lastLevel, const ZADDNode* node, const ZADDNode* varSet, char* minterm) const {
  unsigned long i;

  if(!varSet->isDrain()){
    for(i=lastLevel;i<varSet->getLevel();++i) minterm[i]='-';
    if(varSet->getLevel()==node->getLevel()){
      if(node->getSucc0()==node->getSucc1()){
        minterm[varSet->getLevel()]='-';
        printPaths(varSet->getLevel()+1,node->getSucc0(),varSet->getSucc1(),minterm);
      } else {
        minterm[varSet->getLevel()]='0';
        printPaths(varSet->getLevel()+1,node->getSucc0(),varSet->getSucc1(),minterm);
        minterm[varSet->getLevel()]='1';
        printPaths(varSet->getLevel()+1,node->getSucc1(),varSet->getSucc1(),minterm);
      }
    } else {
      minterm[varSet->getLevel()]='0';
      printPaths(varSet->getLevel()+1,node,varSet->getSucc1(),minterm);
    }
  } else {
    if(node->getValue()!=0.0){
      for(i=lastLevel;i<ZADDVarOrder::size();++i) minterm[i]='-';
      for(i=0;i<ZADDVarOrder::size();++i) cout << minterm[i];
      cout << endl;
    }
  }
}

double ZADDFunction::countPaths(const ZADDNode* node, const ZADDNode* varSet, map<ZADDNode*,double>* cache) const {
  if(!varSet->isDrain()){
    if(cache->find((ZADDNode*) node)!=cache->end()) return (*cache)[(ZADDNode*) node];

    double ret=0.0;
    if(varSet->getLevel()==node->getLevel()){
      if(node->getSucc0()==node->getSucc1()){
        ret=countPaths(node->getSucc0(),varSet->getSucc1(),cache);
      } else {
        ret=countPaths(node->getSucc0(),varSet->getSucc1(),cache);
        ret+=countPaths(node->getSucc1(),varSet->getSucc1(),cache);
      }
    } else {
      ret=countPaths(node,varSet->getSucc1(),cache);
    }
    (*cache)[(ZADDNode*) node]=ret;

    return ret;
  } else {
    if(node->getValue()!=0.0) return 1.0;
    else return 0.0;
  }
}

//********************************non class functions********************
ZADDFunction operator+(const double value, const ZADDFunction& function){
  return function+value;
}

ZADDFunction operator-(const double value, const ZADDFunction& function){
  return ZADDFunction(value)-function;
}

ZADDFunction operator*(const double value, const ZADDFunction& function){
  return function*value;
}

ZADDFunction operator-(const ZADDFunction& function){
  ZADD zadd;
  return zadd.negate(function);
}

ZADDFunction EQUIV(const ZADDFunction& function1, const ZADDFunction& function2){
  return EQUAL(function1,function2);
}

ZADDFunction IMPLIES(const ZADDFunction& function1, const ZADDFunction& function2){
  return !function1|function2;
}

ZADDFunction ITE(const ZADDFunction& function1, const ZADDFunction& function2, const ZADDFunction& function3){
  ZADD zadd;
  return zadd.ITE(function1,function2,function3);
}

ZADDFunction LESS(const ZADDFunction& function1, const ZADDFunction& function2){
  ZADD zadd;
  return zadd.less(function1,function2);
}

ZADDFunction LESS_EQUAL(const ZADDFunction& function1, const ZADDFunction& function2){
  ZADD zadd;
  return zadd.lessEqual(function1,function2);
}

ZADDFunction EQUAL(const ZADDFunction& function1, const ZADDFunction& function2){
  ZADD zadd;
  return zadd.equal(function1,function2);
}

ZADDFunction GREATER(const ZADDFunction& function1, const ZADDFunction& function2){
  ZADD zadd;
  return zadd.less(function2,function1);
}

ZADDFunction GREATER_EQUAL(const ZADDFunction& function1, const ZADDFunction& function2){
  ZADD zadd;
  return zadd.lessEqual(function2,function1);
}

ZADDFunction MAXIMUM(const ZADDFunction& function1, const ZADDFunction& function2){
  ZADD zadd;
  return zadd.max(function1,function2);
}

ZADDFunction MINIMUM(const ZADDFunction& function1, const ZADDFunction& function2){
  ZADD zadd;
  return zadd.min(function1,function2);
}

ZADDFunction MM(const ZADDFunction& A, const ZADDFunction& b, const ZADDFunction& varSet){
  ZADD zadd;
  return zadd.MM(A,b,varSet);
}

ZADDFunction MV(const ZADDFunction& A, const ZADDFunction& b, const ZADDFunction& varSet){
  ZADD zadd;
  return zadd.MV(A,b,varSet);
}

ZADDFunction VM(const ZADDFunction& b, const ZADDFunction& A, const ZADDFunction& varSet){
  ZADD zadd;
  return zadd.VM(b,A,varSet);
}
