/***************************************************************************
                          zaddfunction.h  -  description
                             -------------------
    begin                : Thu Apr 15 2004
    copyright            : (C) 2004 by Joern Ossowski
    email                : ossowski@cs.uni-bonn.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ZADDFUNCTION_H
#define ZADDFUNCTION_H

class ZADD;
class ZADDNode;
class Assignment;
class VarSet;

#include <string>
#include <set>
#include <map>


class ZADDFunction{
friend class ZADD;
friend class ZADDOutput;
public:
  ZADDFunction(const ZADDFunction&);
  ZADDFunction(const Assignment&, const double=1.0);
  ZADDFunction(char*, const double=1.0);
  ZADDFunction(const VarSet& varSet);
  ZADDFunction(const std::string&);
  ZADDFunction(const double);
  ZADDFunction(void);
  ~ZADDFunction(void);
  static ZADDFunction ERROR_FUNCTION(void);
  static ZADDFunction ONE(void);
  static ZADDFunction ZERO(void);
  static ZADDFunction createConstantFunction(const double);
  static ZADDFunction createStabilityFunction(const VarSet&, const VarSet&);
  static ZADDFunction projection(const std::string&);
  static ZADDFunction projection(const unsigned long);
  unsigned long size(void) const;
  ZADDFunction& operator=(const ZADDFunction&);
  ZADDFunction operator+(const ZADDFunction&) const; //+
  ZADDFunction& operator+=(const ZADDFunction&);
  ZADDFunction operator-(const ZADDFunction&) const; //-
  ZADDFunction& operator-=(const ZADDFunction&);
  ZADDFunction operator*(const ZADDFunction&) const; //*
  ZADDFunction& operator*=(const ZADDFunction&);
  ZADDFunction operator|(const ZADDFunction&) const; //or
  ZADDFunction& operator|=(const ZADDFunction&);
  ZADDFunction operator&(const ZADDFunction&) const; //and
  ZADDFunction& operator&=(const ZADDFunction&);
  ZADDFunction operator^(const ZADDFunction&) const; //xor
  ZADDFunction& operator^=(const ZADDFunction&);
  ZADDFunction operator!(void) const;
  bool operator==(const ZADDFunction&) const;
  bool operator!=(const ZADDFunction&) const;
  bool operator<(const ZADDFunction&) const;
  bool operator<=(const ZADDFunction&) const;
  bool operator>(const ZADDFunction&) const;
  bool operator>=(const ZADDFunction&) const;
  bool isConstant(void) const;
  double getValue(void) const;
  std::string getFirstVariable(void) const;
	unsigned long getFirstVariableIndex(void) const;
  ZADDFunction cofactor(const std::string&, const bool) const;
  ZADDFunction cofactor(const unsigned long, const bool) const;
  ZADDFunction cofactor(const ZADDFunction&) const;
  ZADDFunction subSet(const std::string&, const bool) const;
  ZADDFunction subSet(const unsigned long, const bool) const;
	ZADDFunction subSet(const ZADDFunction&) const;
	ZADDFunction change(void) const;
  ZADDFunction change(const std::string&) const;
  ZADDFunction change(const unsigned long) const;
	ZADDFunction change(const ZADDFunction&) const;
  ZADDFunction exists(const ZADDFunction&) const;
  ZADDFunction exists(const std::string&) const;
  ZADDFunction exists(const unsigned long) const;
  ZADDFunction forall(const ZADDFunction&) const;
  ZADDFunction forall(const std::string&) const;
  ZADDFunction forall(const unsigned long) const;
	ZADDFunction equivalenceClasses(const ZADDFunction&, const ZADDFunction&) const;
	ZADDFunction constrain(const ZADDFunction&) const;
  ZADDFunction sum(const ZADDFunction&) const;
  ZADDFunction product(const ZADDFunction&) const;
  ZADDFunction maximum(const ZADDFunction&) const;
  double max(void) const;
	ZADDFunction minimum(const  ZADDFunction&) const;
  double min(void) const;
  ZADDFunction oneOver(void) const;
  ZADDFunction compose(const std::string&, const ZADDFunction&) const;
  ZADDFunction compose(const unsigned long, const ZADDFunction&) const;
  ZADDFunction moveUp(const unsigned int=1) const;
  ZADDFunction moveUp(const ZADDFunction&, const unsigned int=1) const;  
  ZADDFunction moveDown(const unsigned int=1) const;
  ZADDFunction moveDown(const ZADDFunction&, const unsigned int=1) const;  
  void printEssentialVariables(void) const;
  std::set<std::string> getEssentialVariables(void) const;
	operator Assignment() const;
  ZADDFunction getAssignment(void) const;
	void getAssignment(char*) const;
  double countAssignments(void) const;
  double countAssignments(const ZADDFunction& cubeSet) const;
  ZADDFunction getRandomAssignment(const ZADDFunction& cube_set) const;
  void printPaths(void) const;
  double countPaths(void) const;
  ZADDFunction getPath(void) const;
	void getPath(char*) const;

  std::set<ZADDNode*> getNodes() const;


  struct less : public std::binary_function<ZADDFunction, ZADDFunction, bool>  {
    bool operator()(const ZADDFunction& x, const ZADDFunction& y) const {
      std::less<ZADDNode*> l;

      return 
        (x.rootNode==y.rootNode) 
        ? l(x.varSet,y.varSet) 
        : l(x.rootNode,y.rootNode);
    }
  };

private:
  static ZADD* BDD;
  ZADDNode* rootNode;
	ZADDNode* varSet;

  ZADDFunction(const ZADDNode* rootNode, const ZADDNode* varSet);
  void printPaths(const unsigned long, const ZADDNode*, const ZADDNode*, char*) const;
  double countPaths(const ZADDNode*, const ZADDNode*, std::map<ZADDNode*,double>*) const;
};

ZADDFunction operator+(const double, const ZADDFunction&);
ZADDFunction operator-(const double, const ZADDFunction&);
ZADDFunction operator*(const double, const ZADDFunction&);
ZADDFunction operator-(const ZADDFunction&);
ZADDFunction EQUIV(const ZADDFunction&, const ZADDFunction&);
ZADDFunction IMPLIES(const ZADDFunction&, const ZADDFunction&);
ZADDFunction ITE(const ZADDFunction&, const ZADDFunction&, const ZADDFunction&);
ZADDFunction LESS(const ZADDFunction&, const ZADDFunction&);
ZADDFunction LESS_EQUAL(const ZADDFunction&, const ZADDFunction&);
ZADDFunction EQUAL(const ZADDFunction&, const ZADDFunction&);
ZADDFunction GREATER(const ZADDFunction&, const ZADDFunction&);
ZADDFunction GREATER_EQUAL(const ZADDFunction&, const ZADDFunction&);
ZADDFunction MAXIMUM(const ZADDFunction&, const ZADDFunction&);
ZADDFunction MINIMUM(const ZADDFunction&, const ZADDFunction&);

ZADDFunction MM(const ZADDFunction&, const ZADDFunction&, const ZADDFunction&);
ZADDFunction MV(const ZADDFunction&, const ZADDFunction&, const ZADDFunction&);
ZADDFunction VM(const ZADDFunction&, const ZADDFunction&, const ZADDFunction&);

/*
ZADDFunction MM(const ZADDFunction&, const ZADDFunction&, const unsigned long, const short=2);
ZADDFunction MV(const ZADDFunction&, const ZADDFunction&, const unsigned long);

ZADDFunction JACOBI(const ZADDFunction&, const ZADDFunction&, const unsigned long, const unsigned long, const double=0.0);
ZADDFunction BF(const ZADDFunction&, const unsigned long, const unsigned long);
ZADDFunction BF(const ZADDFunction&, const unsigned long);
ZADDFunction FLOYD(const ZADDFunction&, const unsigned long);

void printMatrix(const ZADDFunction&, const unsigned long, const unsigned long);
void printMatrix2PNM(const ZADDFunction&, const unsigned long, const unsigned long);
void printLeftMatrix(const ZADDFunction&, const unsigned long, const unsigned long);
void printRightMatrix(const ZADDFunction&, const unsigned long, const unsigned long);
*/

#endif
