/***************************************************************************
                          addfunction.cpp  -  description
                             -------------------
    begin                : Thu Apr 15 2004
    copyright            : (C) 2004-2008 by Joern Ossowski
    email                : mail@jossowski.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <add/addfunction.h>
#include <add/addbasenode.h>
#include <add/addinnernode.h>
#include <add/addterminalnode.h>
#include <add/addvarorder.h>
#include <add/add.h>
#include <common/assignment.h>
#include <common/randomhelper.h>
#include <common/stringhelper.h>
#include <common/helper.h>
#include <common/ptrhashmap.h>
#include <common/vsizearray.h>

#include <stack>
#include <set>
#include <iostream>

#include "vereofy/common/interruptible.hpp"

using namespace std;

inline ADDTerminalNode* findOrAddDrain(const double value){
  if UNLIKELY(ABS(value)<0.0000001) return ADD::zeroDrain;

  return ADD::drainUniqueTable.findOrAddValue(value);
}
inline ADDInnerNode* findOrAdd(ADDUniqueTable* uTable, ADDBaseNode* succ0, ADDBaseNode* succ1){
  return uTable->findOrAdd(succ0,succ1);
}

inline unsigned long pow2(const unsigned long n){
  return (1 << n);
}

//boolean constants
#define AND_OP              0x01
#define OR_OP               0x02
#define DIF_OP              0x03
#define ITE_OP              0x02
#define EXI_OP              0x04
#define COF_OP              0x06
#define CHA_OP              0x08
//artihmetic constants
#define ADD_OP              0x02
#define SUB_OP              0x04
#define MUL_OP              0x06
#define OVE_OP              0x08
//move constants
#define MUP_OP              0x02
#define MDO_OP              0x04
//compare constants
#define LES_OP              0x02
#define LEE_OP              0x04
#define EQU_OP              0x06
//abstract constants
#define SUM_OP              0x02
#define PRO_OP              0x04
#define MAX_OP              0x06
#define MIN_OP              0x08
//matrix constants
#define MAV_OP              0x02
#define VMA_OP              0x04
#define MMA_OP              0x06
//special constants
#define EQI_OP              0x02
#define CON_OP              0x04

inline ADDBaseNode* calcPtr(const ADDBaseNode* node, unsigned long op){
  return reinterpret_cast<ADDBaseNode*>(reinterpret_cast<unsigned long>(node)+op);
}

inline void swapPointer(const ADDBaseNode*& u, const ADDBaseNode*& v){
  //swap nodes for more computed table hits
  if(u>v){
    std::swap(u,v);
  }
}

inline void addFunction(ADDBaseNode* node){node->incNodeCount();}

inline void removeFunction(ADDBaseNode* node){node->deleteNode();}

//************************ADDFunction************************
ADDFunction::ADDFunction(const ADDFunction& function) : rootNode(function.rootNode){
  addFunction(rootNode);
  if (ADD::dynamicReordering) ADD::reorderingHook();
}

ADDFunction::ADDFunction(const Assignment& assignment, const double value){
  if UNLIKELY(assignment.size()==0){
    rootNode=ADD::zeroDrain;
  } else {
    rootNode=findOrAddDrain(value);
    if(rootNode!=ADD::zeroDrain){
      for(unsigned long i=0;i<assignment.size();++i) ADDVarOrder::addVariable(assignment[i].variable);

      for(unsigned long i=ADD::variableOrdering.size();i>0;--i){
	std::string var=ADDVarOrder::getVariableName (i-1);
        if(assignment.find(var)){
          rootNode=(assignment.at(var))?findOrAdd(ADD::variableOrdering.getData(i-1),ADD::zeroDrain,rootNode):findOrAdd(ADD::variableOrdering.getData(i-1),rootNode,ADD::zeroDrain);
        }
      }
    }
  }

  addFunction(rootNode);
  if (ADD::dynamicReordering) ADD::reorderingHook();
}

ADDFunction::ADDFunction(char* buffer, const double value) : rootNode(findOrAddDrain(value)){
  if(rootNode!=ADD::zeroDrain){
    for(unsigned long i=ADD::variableOrdering.size();i>0;--i){
      switch(buffer[i-1]){
        case '0': rootNode=findOrAdd(ADD::variableOrdering.getData(i-1),rootNode,ADD::zeroDrain);
                  break;
        case '1': rootNode=findOrAdd(ADD::variableOrdering.getData(i-1),ADD::zeroDrain,rootNode);
                  break;
        default:  break;
      }
    }
  }

  addFunction(rootNode);
  if (ADD::dynamicReordering) ADD::reorderingHook();
}

ADDFunction::ADDFunction(const std::string& variable){
  //append label if variableOrdering has no such variable
  VariableIndexType index=ADD::variableOrdering[variable];
  if(index==VARIABLE_INDEX_MAX){
//    if(variable.find_first_of(badChars)!=string::npos) throw "Wrong characters in variable name";
    if UNLIKELY(!ADD::variableOrdering.appendVariable(variable)) throw "Could not append variable";
    index=ADD::variableOrdering.size()-1;
  }

  rootNode=findOrAdd(ADD::variableOrdering.getData(index),ADD::zeroDrain,ADD::oneDrain);
  addFunction(rootNode);
  if (ADD::dynamicReordering) ADD::reorderingHook();
}

ADDFunction::ADDFunction(const double value) : rootNode(findOrAddDrain(value)){
  addFunction(rootNode);
  if (ADD::dynamicReordering) ADD::reorderingHook();
}

ADDFunction::ADDFunction(const ADDBaseNode* node) : rootNode(const_cast<ADDBaseNode*>(node)){
  addFunction(rootNode);
  if (ADD::dynamicReordering) ADD::reorderingHook();
}

ADDFunction::ADDFunction() : rootNode(ADD::zeroDrain) {
  addFunction(rootNode);
  if (ADD::dynamicReordering) ADD::reorderingHook();
}

ADDFunction::~ADDFunction(void){removeFunction(rootNode);}

ADDFunction ADDFunction::ONE(void){return ADDFunction(ADD::oneDrain);}

ADDFunction ADDFunction::ZERO(void){return ADDFunction(ADD::zeroDrain);}

ADDFunction ADDFunction::projection(const std::string& variable){return ADDFunction(variable);}

ADDFunction ADDFunction::projection(const unsigned long index){
  if(index>=ADD::variableOrdering.size()) throw "Index out of range";
  else {
    ADDBaseNode* ret=findOrAdd(ADD::variableOrdering.getData(index),ADD::zeroDrain,ADD::oneDrain);
    return ADDFunction(ret);
  }
}

unsigned long ADDFunction::size(void) const {
  std::set<const ADDBaseNode*> visited;
  std::stack<const ADDBaseNode*> nodeStack;

  visited.insert(rootNode);
  nodeStack.push(rootNode);
  while(nodeStack.size()>0){
    const ADDBaseNode* node=nodeStack.top();
    nodeStack.pop();
    if LIKELY(!node->isDrain()){
      {
        ADDBaseNode* tempNode=static_cast<const ADDInnerNode*>(node)->getSucc0();
        if(visited.find(tempNode)==visited.end()){
          visited.insert(tempNode);
          nodeStack.push(tempNode);
        }
      }
      {
        ADDBaseNode* tempNode=static_cast<const ADDInnerNode*>(node)->getSucc1();
        if(visited.find(tempNode)==visited.end()){
          visited.insert(tempNode);
          nodeStack.push(tempNode);
        }
      }
    }
  }

  return visited.size();
}

ADDFunction& ADDFunction::operator=(const ADDFunction& function){
  //insert function
  addFunction(function.rootNode);
  //remove function
  removeFunction(rootNode);
  //copy values
  rootNode=function.rootNode;

  return *this;
}

struct abstract{
  const short OPERATOR_;
  PtrHashMap2<ADDBaseNode>& table_;
  ADDBaseNode* (*function_)(const ADDBaseNode*,const ADDBaseNode*);

  abstract(const short op, PtrHashMap2<ADDBaseNode>& table, ADDBaseNode* (*function)(const ADDBaseNode*,const ADDBaseNode*)): OPERATOR_(op),table_(table), function_(function){}
  ADDBaseNode* calc(const ADDBaseNode* node, const ADDBaseNode* varSet){
    if(varSet->isDrain()) return const_cast<ADDBaseNode*>(node);
    if(ADDBaseNode* comp=table_.find(calcPtr(node,OPERATOR_),varSet)) return comp;

    if(varSet->getUniqueTable()==node->getUniqueTable()){
      ADDBaseNode* newVarSet=static_cast<const ADDInnerNode*>(varSet)->getSucc1();
      ADDBaseNode* w0=calc(static_cast<const ADDInnerNode*>(node)->getSucc0(),newVarSet);
      ADDBaseNode* w1=calc(static_cast<const ADDInnerNode*>(node)->getSucc1(),newVarSet);
      ADDBaseNode* result=(*function_)(w0,w1);
      if(result!=w0 && w0->preNodeCount==0) w0->garbageCollect();
      if(result!=w1 && w1->preNodeCount==0) w1->garbageCollect();
      table_.insert(calcPtr(node,OPERATOR_),varSet,result);

      return result;
    } else {
      ADDBaseNode* result;
      if(varSet->getRealLevel()<node->getRealLevel()){
        ADDBaseNode* w=calc(node,static_cast<const ADDInnerNode*>(varSet)->getSucc1());
        result=(*function_)(w,w);
        if(result!=w && w->preNodeCount==0) w->garbageCollect();
      } else {
        ADDBaseNode* w0=calc(static_cast<const ADDInnerNode*>(node)->getSucc0(),varSet);
        ADDBaseNode* w1=calc(static_cast<const ADDInnerNode*>(node)->getSucc1(),varSet);
        result=(w0==w1)?w0:findOrAdd(node->getUniqueTable(),w0,w1);
      }
      table_.insert(calcPtr(node,OPERATOR_),varSet,result);

      return result;
    }
  }
};

struct idempotentAbstract{
  const short OPERATOR_;
  PtrHashMap2<ADDBaseNode>& table_;
  ADDBaseNode* (*function_)(const ADDBaseNode*,const ADDBaseNode*);

  idempotentAbstract(const short op, PtrHashMap2<ADDBaseNode>& table, ADDBaseNode* (*function)(const ADDBaseNode*,const ADDBaseNode*)): OPERATOR_(op),table_(table), function_(function){}
  ADDBaseNode* calc(const ADDBaseNode* node, const ADDBaseNode* varSet){
    if(node->isDrain()) return const_cast<ADDBaseNode*>(node);

    ADDBaseNode* newVarSet=const_cast<ADDBaseNode*>(varSet);
    while(newVarSet->getRealLevel()<node->getRealLevel()) newVarSet=static_cast<ADDInnerNode*>(newVarSet)->getSucc1();
    if(newVarSet->isDrain()) return const_cast<ADDBaseNode*>(node);
    
    if(ADDBaseNode* comp=table_.find(calcPtr(node,OPERATOR_),newVarSet)) return comp;

    if(varSet->getUniqueTable()==node->getUniqueTable()){
      ADDBaseNode* w0=calc(static_cast<const ADDInnerNode*>(node)->getSucc0(),static_cast<ADDInnerNode*>(newVarSet)->getSucc1());
      ADDBaseNode* w1=calc(static_cast<const ADDInnerNode*>(node)->getSucc1(),static_cast<ADDInnerNode*>(newVarSet)->getSucc1());
      ADDBaseNode* result=(*function_)(w0,w1);
      if(result!=w0 && w0->preNodeCount==0) w0->garbageCollect();
      if(result!=w1 && w1->preNodeCount==0) w1->garbageCollect();
      table_.insert(calcPtr(node,OPERATOR_),varSet,result);

      return result;
    } else {
      ADDBaseNode* w0=calc(static_cast<const ADDInnerNode*>(node)->getSucc0(),newVarSet);
      ADDBaseNode* w1=calc(static_cast<const ADDInnerNode*>(node)->getSucc1(),newVarSet);
      ADDBaseNode* result=(w0==w1)?w0:findOrAdd(node->getUniqueTable(),w0,w1);
      table_.insert(calcPtr(node,OPERATOR_),newVarSet,result);

      return result;
    }
  }
};

ADDBaseNode* addition(const ADDBaseNode* v){
  if(v->isDrain()) return (v==ADD::zeroDrain)?ADD::zeroDrain:findOrAddDrain(2.0*static_cast<const ADDTerminalNode*>(v)->getValue());

  //try to get result from computed table
  if(ADDBaseNode* comp=ADD::ArithmeticComputedTable.find(calcPtr(v,ADD_OP),v)) return comp;

  ADDUniqueTable* minTable=v->getUniqueTable();
  ADDBaseNode* w0=addition(static_cast<const ADDInnerNode*>(v)->getSucc0());
  ADDBaseNode* w1=addition(static_cast<const ADDInnerNode*>(v)->getSucc1());

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);

  //insert result into computed table
  ADD::ArithmeticComputedTable.insert(calcPtr(v,ADD_OP),v,node);

  return node;
}

ADDBaseNode* addition(const ADDBaseNode* v1, const ADDBaseNode* v2){
  if(v1==ADD::zeroDrain) return const_cast<ADDBaseNode*>(v2);
  if(v2==ADD::zeroDrain) return const_cast<ADDBaseNode*>(v1);
  if(v1==v2) return addition(v1);
  if(v1->isDrain() && v2->isDrain()) return findOrAddDrain(static_cast<const ADDTerminalNode*>(v1)->getValue()+static_cast<const ADDTerminalNode*>(v2)->getValue());

  swapPointer(v1,v2);

  //try to get result from computed table
  if(ADDBaseNode* comp=ADD::ArithmeticComputedTable.find(calcPtr(v1,ADD_OP),v2)) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;
  ADDUniqueTable* minTable;

  //calculate minTable
  if(v1->getRealLevel()<=v2->getRealLevel()){
    minTable=v1->getUniqueTable();
    if(v2->getUniqueTable()==minTable){
      w0=addition(static_cast<const ADDInnerNode*>(v1)->getSucc0(),static_cast<const ADDInnerNode*>(v2)->getSucc0());
      w1=addition(static_cast<const ADDInnerNode*>(v1)->getSucc1(),static_cast<const ADDInnerNode*>(v2)->getSucc1());
    } else {
      w0=addition(static_cast<const ADDInnerNode*>(v1)->getSucc0(),v2);
      w1=addition(static_cast<const ADDInnerNode*>(v1)->getSucc1(),v2);
    }
  } else {
    minTable=v2->getUniqueTable();
    w0=addition(v1,static_cast<const ADDInnerNode*>(v2)->getSucc0());
    w1=addition(v1,static_cast<const ADDInnerNode*>(v2)->getSucc1());
  }

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);

  //insert result into computed table
  ADD::ArithmeticComputedTable.insert(calcPtr(v1,ADD_OP),v2,node);

  return node;
}

ADDFunction& ADDFunction::operator+=(const ADDFunction& function){
  return *this=ADDFunction(::addition(rootNode,function.rootNode));
}

ADDBaseNode* negate(const ADDBaseNode* v){
  INTERRUPTIBLE();
  if(v==ADD::zeroDrain) return ADD::zeroDrain;
  if(v->isDrain()) return findOrAddDrain(-static_cast<const ADDTerminalNode*>(v)->getValue());

  if(ADDBaseNode* comp=ADD::ArithmeticComputedTable.find(calcPtr(ADD::zeroDrain,SUB_OP),v)) return comp;

  ADDUniqueTable* minTable=v->getUniqueTable();
  ADDBaseNode* w0=::negate(static_cast<const ADDInnerNode*>(v)->getSucc0());
  ADDBaseNode* w1=::negate(static_cast<const ADDInnerNode*>(v)->getSucc1());

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);

  ADD::ArithmeticComputedTable.insert(calcPtr(ADD::zeroDrain,SUB_OP),v,node);

  return node;
}

ADDBaseNode* change(const ADDBaseNode* v){
  INTERRUPTIBLE();
  if(v->isDrain()) return const_cast<ADDBaseNode*>(v);

  if(ADDBaseNode* comp=ADD::BoolComputedTable.find(calcPtr(v,CHA_OP),0,0)) return comp;

  ADDUniqueTable* minTable=v->getUniqueTable();
  ADDBaseNode* w0=::change(static_cast<const ADDInnerNode*>(v)->getSucc0());
  ADDBaseNode* w1=::change(static_cast<const ADDInnerNode*>(v)->getSucc1());

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w1,w0);

  ADD::BoolComputedTable.insert(calcPtr(v,CHA_OP),0,0,node);

  return node;
}

ADDBaseNode* change(const ADDBaseNode* v, ADDBaseNode* cubeSet){
  INTERRUPTIBLE();
  if(v->isDrain()) return const_cast<ADDBaseNode*>(v);

  while(cubeSet->getRealLevel()<v->getRealLevel()) cubeSet=static_cast<ADDInnerNode*>(cubeSet)->getSucc1();
  if(cubeSet->isDrain()) return const_cast<ADDBaseNode*>(v);

  if(ADDBaseNode* comp=ADD::BoolComputedTable.find(calcPtr(v,CHA_OP),cubeSet,0)) return comp;

  ADDBaseNode* result;
  ADDUniqueTable* minTable=v->getUniqueTable();
  if(cubeSet->getUniqueTable()==v->getUniqueTable()){
    ADDBaseNode* w0=::change(static_cast<const ADDInnerNode*>(v)->getSucc0(),static_cast<ADDInnerNode*>(cubeSet)->getSucc1());
    ADDBaseNode* w1=::change(static_cast<const ADDInnerNode*>(v)->getSucc1(),static_cast<ADDInnerNode*>(cubeSet)->getSucc1());

    result=(w0==w1)?w0:findOrAdd(minTable,w1,w0);
  } else {
    ADDBaseNode* w0=::change(static_cast<const ADDInnerNode*>(v)->getSucc0(),cubeSet);
    ADDBaseNode* w1=::change(static_cast<const ADDInnerNode*>(v)->getSucc1(),cubeSet);

    result=(w0==w1)?w0:findOrAdd(minTable,w0,w1);
  }

  ADD::BoolComputedTable.insert(calcPtr(v,CHA_OP),cubeSet,0,result);

  return result;
}

ADDBaseNode* subtraction(const ADDBaseNode* v1, const ADDBaseNode* v2){
  if(v1==ADD::zeroDrain) return ::negate(v2);
  if(v2==ADD::zeroDrain) return const_cast<ADDBaseNode*>(v1);
  if(v1==v2) return ADD::zeroDrain;
  if(v1->isDrain() && v2->isDrain()) return findOrAddDrain(static_cast<const ADDTerminalNode*>(v1)->getValue()-static_cast<const ADDTerminalNode*>(v2)->getValue());

  //try to get result from computed table
  if(ADDBaseNode* comp=ADD::ArithmeticComputedTable.find(calcPtr(v1,SUB_OP),v2)) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;
  ADDUniqueTable* minTable;

  //calculate minTable
  if(v1->getRealLevel()<=v2->getRealLevel()){
    minTable=v1->getUniqueTable();
    if(v2->getUniqueTable()==minTable){
      w0=subtraction(static_cast<const ADDInnerNode*>(v1)->getSucc0(),static_cast<const ADDInnerNode*>(v2)->getSucc0());
      w1=subtraction(static_cast<const ADDInnerNode*>(v1)->getSucc1(),static_cast<const ADDInnerNode*>(v2)->getSucc1());
    } else {
      w0=subtraction(static_cast<const ADDInnerNode*>(v1)->getSucc0(),v2);
      w1=subtraction(static_cast<const ADDInnerNode*>(v1)->getSucc1(),v2);
    }
  } else {
    minTable=v2->getUniqueTable();
    w0=subtraction(v1,static_cast<const ADDInnerNode*>(v2)->getSucc0());
    w1=subtraction(v1,static_cast<const ADDInnerNode*>(v2)->getSucc1());
  }

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);

  //insert result into computed table
  ADD::ArithmeticComputedTable.insert(calcPtr(v1,SUB_OP),v2,node);

  return node;
}

ADDFunction& ADDFunction::operator-=(const ADDFunction& function){
  return *this=ADDFunction(::subtraction(rootNode,function.rootNode));
}

ADDBaseNode* multiplication(const ADDBaseNode* v){
  if(v->isDrain()) return (v==ADD::zeroDrain)?ADD::zeroDrain:findOrAddDrain(static_cast<const ADDTerminalNode*>(v)->getValue()*static_cast<const ADDTerminalNode*>(v)->getValue());

  //try to get result from computed table
  if(ADDBaseNode* comp=ADD::ArithmeticComputedTable.find(calcPtr(v,MUL_OP),v)) return comp;

  ADDUniqueTable* minTable=v->getUniqueTable();
  ADDBaseNode* w0=multiplication(static_cast<const ADDInnerNode*>(v)->getSucc0());
  ADDBaseNode* w1=multiplication(static_cast<const ADDInnerNode*>(v)->getSucc1());

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);

  //insert result into computed table
  ADD::ArithmeticComputedTable.insert(calcPtr(v,MUL_OP),v,node);

  return node;
}

ADDBaseNode* multiplication(const ADDBaseNode* v1, const ADDBaseNode* v2){
  if(v1==ADD::zeroDrain || v2==ADD::zeroDrain) return ADD::zeroDrain;
  if(v1==v2) return multiplication(v1);
  if(v1->isDrain() && v2->isDrain()) return findOrAddDrain(static_cast<const ADDTerminalNode*>(v1)->getValue()*static_cast<const ADDTerminalNode*>(v2)->getValue());

  swapPointer(v1,v2);

  //try to get result from computed table
  if(ADDBaseNode* comp=ADD::ArithmeticComputedTable.find(calcPtr(v1,MUL_OP),v2)) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;
  ADDUniqueTable* minTable;

  //calculate minTable
  if(v1->getRealLevel()<=v2->getRealLevel()){
    minTable=v1->getUniqueTable();
    if(v2->getUniqueTable()==minTable){
      w0=multiplication(static_cast<const ADDInnerNode*>(v1)->getSucc0(),static_cast<const ADDInnerNode*>(v2)->getSucc0());
      w1=multiplication(static_cast<const ADDInnerNode*>(v1)->getSucc1(),static_cast<const ADDInnerNode*>(v2)->getSucc1());
    } else {
      w0=multiplication(static_cast<const ADDInnerNode*>(v1)->getSucc0(),v2);
      w1=multiplication(static_cast<const ADDInnerNode*>(v1)->getSucc1(),v2);
    }
  } else {
    minTable=v2->getUniqueTable();
    w0=multiplication(v1,static_cast<const ADDInnerNode*>(v2)->getSucc0());
    w1=multiplication(v1,static_cast<const ADDInnerNode*>(v2)->getSucc1());
  }

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);

  //insert result into computed table
  ADD::ArithmeticComputedTable.insert(calcPtr(v1,MUL_OP),v2,node);

  return node;
}

ADDFunction& ADDFunction::operator*=(const ADDFunction& function){
  return *this=ADDFunction(::multiplication(rootNode,function.rootNode));
}

ADDBaseNode* OR(const ADDBaseNode* v1, const ADDBaseNode* v2){
  INTERRUPTIBLE();

  if(v1==v2) return const_cast<ADDBaseNode*>(v1);
  if(v1->isDrain()) return (v1==ADD::zeroDrain)?const_cast<ADDBaseNode*>(v2):ADD::oneDrain;
  if(v2->isDrain()) return (v2==ADD::zeroDrain)?const_cast<ADDBaseNode*>(v1):ADD::oneDrain;

  swapPointer(v1,v2);
  if(ADDBaseNode* comp=ADD::BoolComputedTable.find(calcPtr(v1,ITE_OP),ADD::oneDrain,v2)) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;

  ADDUniqueTable* minTable;
  ADDUniqueTable* minTable2;
  if(v1->getRealLevel()<=v2->getRealLevel()){
    minTable=v1->getUniqueTable();
    minTable2=ADD::variableOrdering.getData(minTable->getLevel()+1);
    if(minTable2==0) minTable2=minTable;
    if(v2->getUniqueTable()==minTable){
      {
        ADDBaseNode* v1_00, *v1_01, *v2_00, *v2_01;
        {
          ADDBaseNode* temp0=static_cast<const ADDInnerNode*>(v1)->getSucc0();
	  if(temp0->getUniqueTable()==minTable2){
	    v1_00=static_cast<ADDInnerNode*>(temp0)->getSucc0();
            v1_01=static_cast<ADDInnerNode*>(temp0)->getSucc1();
	  } else v1_00=v1_01=temp0;
	}
	{
	  ADDBaseNode* temp1=static_cast<const ADDInnerNode*>(v2)->getSucc0();
          if(temp1->getUniqueTable()==minTable2){
            v2_00=static_cast<ADDInnerNode*>(temp1)->getSucc0();
            v2_01=static_cast<ADDInnerNode*>(temp1)->getSucc1();
          } else v2_00=v2_01=temp1;
	}
	ADDBaseNode* w00=OR(v1_00,v2_00);
	ADDBaseNode* w01=OR(v1_01,v2_01);
	w0=(w00==w01)?w00:findOrAdd(minTable2,w00,w01);
      }
      {
        ADDBaseNode* v1_10, *v1_11, *v2_10, *v2_11;
        {
          ADDBaseNode* temp0=static_cast<const ADDInnerNode*>(v1)->getSucc1();
          if(temp0->getUniqueTable()==minTable2){
            v1_10=static_cast<ADDInnerNode*>(temp0)->getSucc0();
            v1_11=static_cast<ADDInnerNode*>(temp0)->getSucc1();
          } else v1_10=v1_11=temp0;
        }
        {
          ADDBaseNode* temp1=static_cast<const ADDInnerNode*>(v2)->getSucc1();
          if(temp1->getUniqueTable()==minTable2){
            v2_10=static_cast<ADDInnerNode*>(temp1)->getSucc0();
            v2_11=static_cast<ADDInnerNode*>(temp1)->getSucc1();
          } else v2_10=v2_11=temp1;
        }
        ADDBaseNode* w10=OR(v1_10,v2_10);
        ADDBaseNode* w11=OR(v1_11,v2_11);
        w1=(w10==w11)?w10:findOrAdd(minTable2,w10,w11);
      }

 
//      w0=OR(static_cast<const ADDInnerNode*>(v1)->getSucc0(),static_cast<const ADDInnerNode*>(v2)->getSucc0());
//      w1=OR(static_cast<const ADDInnerNode*>(v1)->getSucc1(),static_cast<const ADDInnerNode*>(v2)->getSucc1());
    } else {
      w0=OR(static_cast<const ADDInnerNode*>(v1)->getSucc0(),v2);
      w1=OR(static_cast<const ADDInnerNode*>(v1)->getSucc1(),v2);
    }
  } else {
    minTable=v2->getUniqueTable();
    w0=OR(v1,static_cast<const ADDInnerNode*>(v2)->getSucc0());
    w1=OR(v1,static_cast<const ADDInnerNode*>(v2)->getSucc1());
  }

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);
  ADD::BoolComputedTable.insert(calcPtr(v1,ITE_OP),ADD::oneDrain,v2,node);

  return node;
}

ADDFunction& ADDFunction::operator|=(const ADDFunction& function){
  return *this=ADDFunction(OR(rootNode,function.rootNode));
}

ADDBaseNode* AND(const ADDBaseNode* v1, const ADDBaseNode* v2){
  INTERRUPTIBLE();

  if(v1==v2) return const_cast<ADDBaseNode*>(v1);
  if(v1->isDrain()) return (v1==ADD::zeroDrain)?ADD::zeroDrain:const_cast<ADDBaseNode*>(v2);
  if(v2->isDrain()) return (v2==ADD::zeroDrain)?ADD::zeroDrain:const_cast<ADDBaseNode*>(v1);

  swapPointer(v1,v2);
  if(ADDBaseNode* comp=ADD::BoolComputedTable.find(calcPtr(v1,ITE_OP),v2,ADD::zeroDrain)) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;

  ADDUniqueTable* minTable;
  ADDUniqueTable* minTable2;

  if(v1->getRealLevel()<=v2->getRealLevel()){
    minTable=v1->getUniqueTable();
    minTable2=ADD::variableOrdering.getData(minTable->getLevel()+1);
    if(minTable2==0) minTable2=minTable;
    if(v2->getUniqueTable()==minTable){
      {
        ADDBaseNode* v1_00, *v1_01, *v2_00, *v2_01;
        {
          ADDBaseNode* temp0=static_cast<const ADDInnerNode*>(v1)->getSucc0();
          if(temp0->getUniqueTable()==minTable2){
            v1_00=static_cast<ADDInnerNode*>(temp0)->getSucc0();
            v1_01=static_cast<ADDInnerNode*>(temp0)->getSucc1();
          } else v1_00=v1_01=temp0;
        }
        {
          ADDBaseNode* temp1=static_cast<const ADDInnerNode*>(v2)->getSucc0();
          if(temp1->getUniqueTable()==minTable2){
            v2_00=static_cast<ADDInnerNode*>(temp1)->getSucc0();
            v2_01=static_cast<ADDInnerNode*>(temp1)->getSucc1();
          } else v2_00=v2_01=temp1;
        }
        ADDBaseNode* w00=AND(v1_00,v2_00);
        ADDBaseNode* w01=AND(v1_01,v2_01);
        w0=(w00==w01)?w00:findOrAdd(minTable2,w00,w01);
      }
      {
        ADDBaseNode* v1_10, *v1_11, *v2_10, *v2_11;
        {
          ADDBaseNode* temp0=static_cast<const ADDInnerNode*>(v1)->getSucc1();
          if(temp0->getUniqueTable()==minTable2){
            v1_10=static_cast<ADDInnerNode*>(temp0)->getSucc0();
            v1_11=static_cast<ADDInnerNode*>(temp0)->getSucc1();
          } else v1_10=v1_11=temp0;
        }
        {
          ADDBaseNode* temp1=static_cast<const ADDInnerNode*>(v2)->getSucc1();
          if(temp1->getUniqueTable()==minTable2){
            v2_10=static_cast<ADDInnerNode*>(temp1)->getSucc0();
            v2_11=static_cast<ADDInnerNode*>(temp1)->getSucc1();
          } else v2_10=v2_11=temp1;
        }
        ADDBaseNode* w10=AND(v1_10,v2_10);
        ADDBaseNode* w11=AND(v1_11,v2_11);
        w1=(w10==w11)?w10:findOrAdd(minTable2,w10,w11);
      }
    } else {
      ADDBaseNode* v2_0, *v2_1;
      if(v2->getUniqueTable()==minTable2){
        v2_0=static_cast<const ADDInnerNode*>(v2)->getSucc0();
        v2_1=static_cast<const ADDInnerNode*>(v2)->getSucc1();
      } else v2_0=v2_1=const_cast<ADDBaseNode*>(v2);
      {
        ADDBaseNode* v1_00, *v1_01;
        {
          ADDBaseNode* temp0=static_cast<const ADDInnerNode*>(v1)->getSucc0();
          if(temp0->getUniqueTable()==minTable2){
            v1_00=static_cast<ADDInnerNode*>(temp0)->getSucc0();
            v1_01=static_cast<ADDInnerNode*>(temp0)->getSucc1();
          } else v1_00=v1_01=temp0;
        }
        ADDBaseNode* w00=AND(v1_00,v2_0);
        ADDBaseNode* w01=AND(v1_01,v2_1);
        w0=(w00==w01)?w00:findOrAdd(minTable2,w00,w01);
      }
      {
        ADDBaseNode* v1_10, *v1_11;
        {
          ADDBaseNode* temp0=static_cast<const ADDInnerNode*>(v1)->getSucc1();
          if(temp0->getUniqueTable()==minTable2){
            v1_10=static_cast<ADDInnerNode*>(temp0)->getSucc0();
            v1_11=static_cast<ADDInnerNode*>(temp0)->getSucc1();
          } else v1_10=v1_11=temp0;
        }
        ADDBaseNode* w10=AND(v1_10,v2_0);
        ADDBaseNode* w11=AND(v1_11,v2_1);
        w1=(w10==w11)?w10:findOrAdd(minTable2,w10,w11);
      }
    }
  } else {
    minTable=v2->getUniqueTable();
    minTable2=ADD::variableOrdering.getData(minTable->getLevel()+1);
    if(minTable2==0) minTable2=minTable;
      ADDBaseNode* v1_0, *v1_1;
      if(v1->getUniqueTable()==minTable2){
        v1_0=static_cast<const ADDInnerNode*>(v1)->getSucc0();
        v1_1=static_cast<const ADDInnerNode*>(v1)->getSucc1();
      } else v1_0=v1_1=const_cast<ADDBaseNode*>(v1);
      {
        ADDBaseNode* v2_00, *v2_01;
        {
          ADDBaseNode* temp0=static_cast<const ADDInnerNode*>(v2)->getSucc0();
          if(temp0->getUniqueTable()==minTable2){
            v2_00=static_cast<ADDInnerNode*>(temp0)->getSucc0();
            v2_01=static_cast<ADDInnerNode*>(temp0)->getSucc1();
          } else v2_00=v2_01=temp0;
        }
        ADDBaseNode* w00=AND(v1_0,v2_00);
        ADDBaseNode* w01=AND(v1_1,v2_01);
        w0=(w00==w01)?w00:findOrAdd(minTable2,w00,w01);
      }
      {
        ADDBaseNode* v2_10, *v2_11;
        {
          ADDBaseNode* temp0=static_cast<const ADDInnerNode*>(v2)->getSucc1();
          if(temp0->getUniqueTable()==minTable2){
            v2_10=static_cast<ADDInnerNode*>(temp0)->getSucc0();
            v2_11=static_cast<ADDInnerNode*>(temp0)->getSucc1();
          } else v2_10=v2_11=temp0;
        }
        ADDBaseNode* w10=AND(v1_0,v2_10);
        ADDBaseNode* w11=AND(v1_1,v2_11);
        w1=(w10==w11)?w10:findOrAdd(minTable2,w10,w11);
      }
  }

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);
  ADD::BoolComputedTable.insert(calcPtr(v1,ITE_OP),v2,ADD::zeroDrain,node);

  return node;
}

ADDFunction& ADDFunction::operator&=(const ADDFunction& function){
  return *this=ADDFunction(AND(rootNode,function.rootNode));
}

ADDBaseNode* exists(const ADDBaseNode* node, const ADDBaseNode* varSet){
  INTERRUPTIBLE();

  if(node->isDrain()) return const_cast<ADDBaseNode*>(node);

  unsigned long nodeLevel=node->getRealLevel();
  while(varSet->getRealLevel()<nodeLevel) varSet=static_cast<const ADDInnerNode*>(varSet)->getSucc1();

  if(varSet->isDrain()) return const_cast<ADDBaseNode*>(node);

  if(ADDBaseNode* comp=ADD::BoolComputedTable.find(calcPtr(node,EXI_OP),varSet,0)) return comp;

  if(varSet->getUniqueTable()==node->getUniqueTable()){
    ADDBaseNode* w0=exists(static_cast<const ADDInnerNode*>(node)->getSucc0(),static_cast<const ADDInnerNode*>(varSet)->getSucc1());
    ADDBaseNode* w1=exists(static_cast<const ADDInnerNode*>(node)->getSucc1(),static_cast<const ADDInnerNode*>(varSet)->getSucc1());
    ADDBaseNode* result=::OR(w0,w1);

    if(result!=w0 && w0->preNodeCount==0) w0->garbageCollect();
    if(result!=w1 && w1->preNodeCount==0) w1->garbageCollect();

    ADD::BoolComputedTable.insert(calcPtr(node,EXI_OP),varSet,0,result);

    return result;
  } else {
    ADDBaseNode* w0=exists(static_cast<const ADDInnerNode*>(node)->getSucc0(),varSet);
    ADDBaseNode* w1=exists(static_cast<const ADDInnerNode*>(node)->getSucc1(),varSet);

    ADDBaseNode* temp=(w0==w1)?w0:findOrAdd(node->getUniqueTable(),w0,w1);
    ADD::BoolComputedTable.insert(calcPtr(node,EXI_OP),varSet,0,temp);

    return temp;
  }
}

ADDFunction ADDFunction::exists(const ADDFunction& function) const {
  return ADDFunction(::exists(rootNode,function.rootNode));
}

ADDFunction& ADDFunction::operator^=(const ADDFunction& function){
  return *this=ITE(*this,!function,function);
}

ADDFunction& ADDFunction::operator++(){
  return *this=::addition(rootNode,ADD::oneDrain);
}

ADDFunction& ADDFunction::operator--(){
  return *this=::subtraction(rootNode,ADD::oneDrain);
}

ADDBaseNode* NOT(const ADDBaseNode* u){
  INTERRUPTIBLE();

  if(u->isDrain()) return (u==ADD::zeroDrain)?ADD::oneDrain:ADD::zeroDrain;

  if(ADDBaseNode* comp=ADD::BoolComputedTable.find(calcPtr(u,ITE_OP),ADD::zeroDrain,ADD::oneDrain)) return comp;

  ADDBaseNode* w0=NOT(static_cast<const ADDInnerNode*>(u)->getSucc0());
  ADDBaseNode* w1=NOT(static_cast<const ADDInnerNode*>(u)->getSucc1());

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(u->getUniqueTable(),w0,w1);

  //insert result into computed table
  ADD::BoolComputedTable.insert(calcPtr(u,ITE_OP),ADD::zeroDrain,ADD::oneDrain,node);

  return node;
}

ADDFunction ADDFunction::operator!(void) const {
  return ADDFunction(NOT(rootNode));
}

bool ADDFunction::operator==(const ADDFunction& function) const {
  return (function.rootNode==rootNode);
}

ADDBaseNode* less(const ADDBaseNode* u, const ADDBaseNode* v){
  INTERRUPTIBLE();

  if(u==v) return ADD::zeroDrain;
  if(u->isDrain() && v->isDrain()) return (static_cast<const ADDTerminalNode*>(u)->getValue()<static_cast<const ADDTerminalNode*>(v)->getValue())?ADD::oneDrain:ADD::zeroDrain;

  if(ADDBaseNode* comp=ADD::CompareComputedTable.find(calcPtr(u,LES_OP),v)) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;

  ADDUniqueTable* minTable;

  if(u->getRealLevel()<=v->getRealLevel()){
    minTable=u->getUniqueTable();
    if(v->getUniqueTable()==minTable){
      w0=::less(static_cast<const ADDInnerNode*>(u)->getSucc0(),static_cast<const ADDInnerNode*>(v)->getSucc0());
      w1=::less(static_cast<const ADDInnerNode*>(u)->getSucc1(),static_cast<const ADDInnerNode*>(v)->getSucc1());
    } else {
      w0=::less(static_cast<const ADDInnerNode*>(u)->getSucc0(),v);
      w1=::less(static_cast<const ADDInnerNode*>(u)->getSucc1(),v);
    }
  } else {
    minTable=v->getUniqueTable();
    w0=::less(u,static_cast<const ADDInnerNode*>(v)->getSucc0());
    w1=::less(u,static_cast<const ADDInnerNode*>(v)->getSucc1());
  }

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);
  ADD::CompareComputedTable.insert(calcPtr(u,LES_OP),v,node);

  return node;
}

ADDBaseNode* max(const ADDBaseNode* u, const ADDBaseNode* v){
  if(u==v) return const_cast<ADDBaseNode*>(u);
  if(u->isDrain() && v->isDrain()) return (static_cast<const ADDTerminalNode*>(u)->getValue()<static_cast<const ADDTerminalNode*>(v)->getValue())?const_cast<ADDBaseNode*>(v):const_cast<ADDBaseNode*>(u);

  swapPointer(u,v);
  if(ADDBaseNode* comp=ADD::ExtremaComputedTable.find(calcPtr(u,MAX_OP),v)) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;

  ADDUniqueTable* minTable;

  if(u->getRealLevel()<=v->getRealLevel()){
    minTable=u->getUniqueTable();
    if(v->getUniqueTable()==minTable){
      w0=::max(static_cast<const ADDInnerNode*>(u)->getSucc0(),static_cast<const ADDInnerNode*>(v)->getSucc0());
      w1=::max(static_cast<const ADDInnerNode*>(u)->getSucc1(),static_cast<const ADDInnerNode*>(v)->getSucc1());
    } else {
      w0=::max(static_cast<const ADDInnerNode*>(u)->getSucc0(),v);
      w1=::max(static_cast<const ADDInnerNode*>(u)->getSucc1(),v);
    }
  } else {
    minTable=v->getUniqueTable();
    w0=::max(u,static_cast<const ADDInnerNode*>(v)->getSucc0());
    w1=::max(u,static_cast<const ADDInnerNode*>(v)->getSucc1());
  }

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);
  ADD::ExtremaComputedTable.insert(calcPtr(u,MAX_OP),v,node);

  return node;
}

ADDBaseNode* max(const ADDBaseNode* u){
  if(u->isDrain()) return const_cast<ADDBaseNode*>(u);

  if(ADDBaseNode* comp=ADD::ExtremaComputedTable.find(calcPtr(u,MAX_OP),0)) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;

  w0=::max(static_cast<const ADDInnerNode*>(u)->getSucc0());
  w1=::max(static_cast<const ADDInnerNode*>(u)->getSucc1());

  ADDBaseNode* node=(static_cast<ADDTerminalNode*>(w0)->getValue()>=static_cast<ADDTerminalNode*>(w1)->getValue())?w0:w1;
  ADD::ExtremaComputedTable.insert(calcPtr(u,MAX_OP),0,node);

  return node;
}

ADDBaseNode* min(const ADDBaseNode* u, const ADDBaseNode* v){
  if(u==v) return const_cast<ADDBaseNode*>(u);
  if(u->isDrain() && v->isDrain()) return (static_cast<const ADDTerminalNode*>(u)->getValue()<=static_cast<const ADDTerminalNode*>(v)->getValue())?const_cast<ADDBaseNode*>(u):const_cast<ADDBaseNode*>(v);

  swapPointer(u,v);
  if(ADDBaseNode* comp=ADD::ExtremaComputedTable.find(calcPtr(u,MIN_OP),v)) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;

  ADDUniqueTable* minTable;

  if(u->getRealLevel()<=v->getRealLevel()){
    minTable=u->getUniqueTable();
    if(v->getUniqueTable()==minTable){
      w0=::min(static_cast<const ADDInnerNode*>(u)->getSucc0(),static_cast<const ADDInnerNode*>(v)->getSucc0());
      w1=::min(static_cast<const ADDInnerNode*>(u)->getSucc1(),static_cast<const ADDInnerNode*>(v)->getSucc1());
    } else {
      w0=::min(static_cast<const ADDInnerNode*>(u)->getSucc0(),v);
      w1=::min(static_cast<const ADDInnerNode*>(u)->getSucc1(),v);
    }
  } else {
    minTable=v->getUniqueTable();
    w0=::min(u,static_cast<const ADDInnerNode*>(v)->getSucc0());
    w1=::min(u,static_cast<const ADDInnerNode*>(v)->getSucc1());
  }

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);
  ADD::ExtremaComputedTable.insert(calcPtr(u,MIN_OP),v,node);

  return node;
}

ADDBaseNode* min(const ADDBaseNode* u){
  if(u->isDrain()) return const_cast<ADDBaseNode*>(u);

  if(ADDBaseNode* comp=ADD::ExtremaComputedTable.find(calcPtr(u,MIN_OP),0)) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;

  w0=::min(static_cast<const ADDInnerNode*>(u)->getSucc0());
  w1=::min(static_cast<const ADDInnerNode*>(u)->getSucc1());

  ADDBaseNode* node=(static_cast<ADDTerminalNode*>(w0)->getValue()<=static_cast<ADDTerminalNode*>(w1)->getValue())?w0:w1;
  ADD::ExtremaComputedTable.insert(calcPtr(u,MIN_OP),0,node);

  return node;
}


bool ADDFunction::operator<(const ADDFunction& function) const {
  return (LESS(*this,function)==ADDFunction::ONE());
}

bool ADDFunction::isConstant(void) const {
  return rootNode->isDrain();
}

double ADDFunction::getValue(void) const {
  return isConstant()?static_cast<const ADDTerminalNode*>(rootNode)->getValue():0.0;
}

unsigned long ADDFunction::getFirstVariableIndex(void) const {
  return rootNode->getLevel();
}

ADDBaseNode* cofactor(const ADDBaseNode* node, const ADDBaseNode* assignment){
  INTERRUPTIBLE();

  if(node->isDrain()) return const_cast<ADDBaseNode*>(node);

  while(assignment->getRealLevel()<node->getRealLevel()){
    assignment=(static_cast<const ADDInnerNode*>(assignment)->getSucc0()==ADD::zeroDrain)?static_cast<const ADDInnerNode*>(assignment)->getSucc1():static_cast<const ADDInnerNode*>(assignment)->getSucc0();
  }

  if(assignment->isDrain()) return const_cast<ADDBaseNode*>(node);


  //check for computed table
  if(ADDBaseNode* comp=ADD::BoolComputedTable.find(calcPtr(node,COF_OP),assignment,0)) return comp;

  //do it the hard way
  ADDBaseNode* tempNode;
  if(assignment->getRealLevel()==node->getRealLevel()){
    const bool value=(static_cast<const ADDInnerNode*>(assignment)->getSucc0()==ADD::zeroDrain)?true:false;
    tempNode=cofactor(static_cast<const ADDInnerNode*>(node)->getSucc(value),static_cast<const ADDInnerNode*>(assignment)->getSucc(value));
  } else {
    ADDBaseNode* w0=cofactor(static_cast<const ADDInnerNode*>(node)->getSucc0(),assignment);
    ADDBaseNode* w1=cofactor(static_cast<const ADDInnerNode*>(node)->getSucc1(),assignment);

    tempNode=(w0==w1)?w0:findOrAdd(node->getUniqueTable(),w0,w1);
  }

  //insert result into computed table
  ADD::BoolComputedTable.insert(calcPtr(node,COF_OP),assignment,0,tempNode);

  return tempNode;
}

ADDFunction ADDFunction::cofactor(const unsigned long index, const bool value) const {
  INTERRUPTIBLE();

  if(index>=ADD::variableOrdering.size()) return ADDFunction(rootNode);
  if(index==rootNode->getLevel()) return ADDFunction(static_cast<const ADDInnerNode*>(rootNode)->getSucc(value));

  //create temporary node
  ADDBaseNode* tempVar=value?findOrAdd(ADD::variableOrdering.getData(index),ADD::zeroDrain,ADD::oneDrain):findOrAdd(ADD::variableOrdering.getData(index),ADD::oneDrain,ADD::zeroDrain);
  ADDFunction nrv(::cofactor(rootNode,tempVar));

  //if node was really created above delete it
  if(tempVar->preNodeCount==0) tempVar->garbageCollect();

  return nrv;
}

ADDFunction ADDFunction::cofactor(const ADDFunction& assignment) const {
  return ADDFunction(::cofactor(rootNode,assignment.rootNode));
}

ADDFunction ADDFunction::subSet(const unsigned long index, const bool value) const {
  if(index>=ADDVarOrder::size()) return ADDFunction(rootNode);

  //create temporary node
  ADDBaseNode* tempVar=value?findOrAdd(ADD::variableOrdering.getData(index),ADD::zeroDrain,ADD::oneDrain):findOrAdd(ADD::variableOrdering.getData(index),ADD::oneDrain,ADD::zeroDrain);
  ADDFunction nrv(::multiplication(rootNode,tempVar));

  //if node was really created above delete it
  if(tempVar->preNodeCount==0) tempVar->garbageCollect();

  return nrv;
}

ADDFunction ADDFunction::subSet(const ADDFunction& assignment) const {
  return ((*this)*assignment);
}

ADDFunction ADDFunction::change(void) const {
  return ADDFunction(::change(rootNode));
}

ADDFunction ADDFunction::change(const unsigned long index) const {
  if(index>=ADDVarOrder::size()) return ADDFunction(rootNode);

  ADDBaseNode* tempVar=findOrAdd(ADD::variableOrdering.getData(index),ADD::zeroDrain,ADD::oneDrain);
  ADDFunction nrv(::change(rootNode,tempVar));

  if(tempVar->preNodeCount==0) tempVar->garbageCollect();

  return nrv;
}

ADDFunction ADDFunction::change(const ADDFunction& cubeSet) const {
  return ADDFunction(::change(rootNode,cubeSet.rootNode));
}

ADDFunction ADDFunction::exists(const unsigned long index) const {
  if(index>=ADDVarOrder::size()) return ADDFunction(rootNode);

  ADDBaseNode* tempVar=findOrAdd(ADD::variableOrdering.getData(index),ADD::zeroDrain,ADD::oneDrain);
  ADDFunction nrv(::exists(rootNode,tempVar));

  if(tempVar->preNodeCount==0) tempVar->garbageCollect();

  return nrv;
}

ADDFunction ADDFunction::forall(const ADDFunction& varSet) const {
  ADDFunction temp=!(*this);
  return !temp.exists(varSet);
}

ADDFunction ADDFunction::forall(const unsigned long index) const {
  ADDFunction temp=!(*this);
  return !temp.exists(index);
}

// ADDFunction ADDFunction::equivalenceClasses(const ADDFunction& pattern, const ADDFunction& cubeSet) const {
//   return ADDFunction::BDD->equivalenceClasses(*this,pattern,cubeSet);
// }
//

ADDBaseNode* constrain(const ADDBaseNode* v, const ADDBaseNode* w){
  INTERRUPTIBLE();

  // if g is a drain return constrain(f,0)=0 or constrain(f,1)=f
  if(w->isDrain()) return (w==ADD::zeroDrain)?ADD::zeroDrain:const_cast<ADDBaseNode*>(v);
  
  // if f is a drain and g is not, return constrain(1,g)=1 or constrain(0,g)=0
  if (v->isDrain()) return const_cast<ADDBaseNode*>(v);
  
  // if v=w constrain(v,w)=1
  if (v==w) return ADD::oneDrain;
  
  // computed table lookup
  if (ADDBaseNode* comp=ADD::SpecialComputedTable.find(calcPtr(v,CON_OP),w,0)) return comp;  
  
  ADDBaseNode* result;
  ADDUniqueTable* minTable;
  
  const ADDBaseNode* f0;
  const ADDBaseNode* f1;
  const ADDBaseNode* g0;
  const ADDBaseNode* g1;
  
  if (v->getRealLevel() >= w->getRealLevel()){
    minTable = w->getUniqueTable();
    g0=static_cast<const ADDInnerNode*>(w)->getSucc0();
    g1=static_cast<const ADDInnerNode*>(w)->getSucc1();
    
    if (v->getRealLevel() == w->getRealLevel()){
      f0=static_cast<const ADDInnerNode*>(v)->getSucc0();
      f1=static_cast<const ADDInnerNode*>(v)->getSucc1();
    } else{
      f0=v;
      f1=v;
    }
    
    if (g0==ADD::zeroDrain){
      result = ::constrain(f1,g1);
      ADD::SpecialComputedTable.insert(calcPtr(v,CON_OP),w,0,result);
      return result;
    } 
    
    else if (g1==ADD::zeroDrain){
        result = ::constrain(f0,g0);
        ADD::SpecialComputedTable.insert(calcPtr(v,CON_OP),w,0,result);
        return result;
      }
  } else{
    minTable = v->getUniqueTable();
    f0 = static_cast<const ADDInnerNode*>(v)->getSucc0();
    f1 = static_cast<const ADDInnerNode*>(v)->getSucc1();    
    g0 = w;
    g1 = w;
  }
  
  ADDBaseNode* w0 = ::constrain(f0,g0);
  ADDBaseNode* w1 = ::constrain(f1,g1);
  
  result = (w0==w1)?w0:findOrAdd(minTable,w0,w1);
  ADD::SpecialComputedTable.insert(calcPtr(v,CON_OP),w,0,result);

  return result;
}

ADDFunction ADDFunction::constrain(const ADDFunction& g) const {
  return ADDFunction(::constrain(rootNode,g.rootNode));
}

ADDFunction ADDFunction::sum(const ADDFunction& varSet) const {
  return ADDFunction(abstract(SUM_OP,ADD::AbstractComputedTable,addition).calc(rootNode,varSet.rootNode));
}

ADDFunction ADDFunction::product(const ADDFunction& varSet) const {
  return ADDFunction(abstract(PRO_OP,ADD::AbstractComputedTable,multiplication).calc(rootNode,varSet.rootNode));
}

ADDFunction ADDFunction::maximum(const ADDFunction& varSet) const {
  return ADDFunction(idempotentAbstract(MAX_OP,ADD::ExtremaComputedTable,::max).calc(rootNode,varSet.rootNode));
}

double ADDFunction::max(void) const {
  return static_cast<ADDTerminalNode*>(::max(rootNode))->getValue();
}

ADDFunction ADDFunction::minimum(const ADDFunction& varSet) const {
  return ADDFunction(idempotentAbstract(MIN_OP,ADD::ExtremaComputedTable,::min).calc(rootNode,varSet.rootNode));
}

double ADDFunction::min(void) const {
  return static_cast<ADDTerminalNode*>(::min(rootNode))->getValue();
}

ADDBaseNode* oneOver(const ADDBaseNode* v){
  INTERRUPTIBLE();

  if(v==ADD::zeroDrain) return ADD::zeroDrain;
  if(v->isDrain()) return findOrAddDrain(1.0/static_cast<const ADDTerminalNode*>(v)->getValue());

  //try to get result from computed table
  if(ADDBaseNode* comp=ADD::ArithmeticComputedTable.find(calcPtr(ADD::oneDrain,OVE_OP),v)) return comp;

  ADDUniqueTable* minTable=v->getUniqueTable();
  ADDBaseNode* w0=::oneOver(static_cast<const ADDInnerNode*>(v)->getSucc0());
  ADDBaseNode* w1=::oneOver(static_cast<const ADDInnerNode*>(v)->getSucc1());

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);

  //insert result into computed table
  ADD::ArithmeticComputedTable.insert(calcPtr(ADD::oneDrain,OVE_OP),v,node);

  return node;
}

ADDFunction ADDFunction::oneOver(void) const {
  return ADDFunction(::oneOver(rootNode));
}

ADDFunction ADDFunction::compose(const unsigned long index, const ADDFunction& function) const {
  if(index>=ADDVarOrder::size()) return (*this);
  else return ITE(function,cofactor(index,1),cofactor(index,0));
}

template <int N>
ADDBaseNode* moveUp(const ADDBaseNode* u){
  INTERRUPTIBLE();

  if(u->isDrain()) return const_cast<ADDBaseNode*>(u);

  if(ADDBaseNode* comp=ADD::MoveComputedTable.find(calcPtr(u,MUP_OP),0,reinterpret_cast<ADDBaseNode*>(N))) return comp;

  ADDBaseNode* node=findOrAdd(ADD::variableOrdering.getData(u->getLevel()-N),moveUp<N>(static_cast<const ADDInnerNode*>(u)->getSucc0()),moveUp<N>(static_cast<const ADDInnerNode*>(u)->getSucc1()));
  ADD::MoveComputedTable.insert(calcPtr(u,MUP_OP),0,reinterpret_cast<ADDBaseNode*>(N),node);

  return node;
}

ADDFunction ADDFunction::moveUp() const {
  if(isConstant()) return ADDFunction(rootNode);

  return ADDFunction(::moveUp<1>(rootNode));
}

template <int N>
ADDBaseNode* moveUp(const ADDBaseNode* u, const ADDBaseNode* v){
  INTERRUPTIBLE();

  if(u->isDrain()) return const_cast<ADDBaseNode*>(u);

  unsigned long uLevel=u->getLevel();
  while(v->getLevel()<uLevel) v=static_cast<const ADDInnerNode*>(v)->getSucc1();

  if(v->isDrain()) return const_cast<ADDBaseNode*>(u);
  if((uLevel<N)&&(uLevel==v->getLevel())) return const_cast<ADDBaseNode*>(u);

  if(ADDBaseNode* comp=ADD::MoveComputedTable.find(calcPtr(u,MUP_OP),v,reinterpret_cast<ADDBaseNode*>(N))) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;
  ADDBaseNode* node;
  if(uLevel==v->getLevel()){
    w0=moveUp<N>(static_cast<const ADDInnerNode*>(u)->getSucc0(),static_cast<const ADDInnerNode*>(v)->getSucc1());
    w1=moveUp<N>(static_cast<const ADDInnerNode*>(u)->getSucc1(),static_cast<const ADDInnerNode*>(v)->getSucc1());

    node=findOrAdd(ADD::variableOrdering.getData(uLevel-N),w0,w1);
  } else {
    w0=moveUp<N>(static_cast<const ADDInnerNode*>(u)->getSucc0(),v);
    w1=moveUp<N>(static_cast<const ADDInnerNode*>(u)->getSucc1(),v);

    if((w0->getLevel()<=uLevel)||(w1->getLevel()<=uLevel)){
      if(w0->preNodeCount==0) w0->garbageCollect();
      if(w1->preNodeCount==0) w1->garbageCollect();
      node=const_cast<ADDBaseNode*>(u);
    } else node=findOrAdd(ADD::variableOrdering.getData(uLevel),w0,w1);
  }

  ADD::MoveComputedTable.insert(calcPtr(u,MUP_OP),v,reinterpret_cast<ADDBaseNode*>(N),node);

  return node;
}

ADDFunction ADDFunction::moveUp(const ADDFunction& varSet) const {
  if(isConstant()) return ADDFunction(rootNode);

  return ADDFunction(::moveUp<1>(rootNode,const_cast<ADDBaseNode*>(varSet.rootNode)));
}

template <int N>
ADDBaseNode* moveDown(const ADDBaseNode* u){
  INTERRUPTIBLE();

  if(u->isDrain()) return const_cast<ADDBaseNode*>(u);

  if(ADDBaseNode* comp=ADD::MoveComputedTable.find(calcPtr(u,MDO_OP),0,reinterpret_cast<ADDBaseNode*>(N))) return comp;

  ADDBaseNode* w0=moveDown<N>(static_cast<const ADDInnerNode*>(u)->getSucc0());
  ADDBaseNode* w1=moveDown<N>(static_cast<const ADDInnerNode*>(u)->getSucc1());

  ADDBaseNode* node=((w0==w1) || (u->getLevel()+N>=ADD::variableOrdering.size()))?static_cast<ADDBaseNode*>(ADD::zeroDrain):findOrAdd(ADD::variableOrdering.getData(u->getLevel()+N),w0,w1);
  ADD::MoveComputedTable.insert(calcPtr(u,MDO_OP),0,reinterpret_cast<ADDBaseNode*>(N),node);

  return node;
}

ADDFunction ADDFunction::moveDown() const {
  if(isConstant()) return ADDFunction(rootNode);

  return ADDFunction(::moveDown<1>(rootNode));
}

template <int N>
ADDBaseNode* moveDown(const ADDBaseNode* u, const ADDBaseNode* v){
  INTERRUPTIBLE();

  if(u->isDrain()) return const_cast<ADDBaseNode*>(u);

  unsigned long uLevel=u->getLevel();
  while(v->getLevel()<uLevel) v=static_cast<const ADDInnerNode*>(v)->getSucc1();

  if(v->isDrain()) return const_cast<ADDBaseNode*>(u);

  if(ADDBaseNode* comp=ADD::MoveComputedTable.find(calcPtr(u,MDO_OP),v,reinterpret_cast<ADDBaseNode*>(N))) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;
  ADDBaseNode* node;
  if(uLevel==v->getLevel()){
    w0=moveDown<N>(static_cast<const ADDInnerNode*>(u)->getSucc0(),static_cast<const ADDInnerNode*>(v)->getSucc1());
    w1=moveDown<N>(static_cast<const ADDInnerNode*>(u)->getSucc1(),static_cast<const ADDInnerNode*>(v)->getSucc1());

    if((uLevel+N>=w0->getLevel())||(uLevel+N>=w1->getLevel())) node=findOrAdd(ADD::variableOrdering.getData(uLevel),w0,w1);
    else node=findOrAdd(ADD::variableOrdering.getData(uLevel+N),w0,w1);
  } else {
    w0=moveDown<N>(static_cast<const ADDInnerNode*>(u)->getSucc0(),v);
    w1=moveDown<N>(static_cast<const ADDInnerNode*>(u)->getSucc1(),v);

    node=findOrAdd(u->getUniqueTable(),w0,w1);
  }

  ADD::MoveComputedTable.insert(calcPtr(u,MDO_OP),v,reinterpret_cast<ADDBaseNode*>(N),node);

  return node;
}

ADDFunction ADDFunction::moveDown(const ADDFunction& varSet) const {
  if(isConstant()) return ADDFunction(rootNode);

  return ADDFunction(::moveDown<1>(rootNode,const_cast<ADDBaseNode*>(varSet.rootNode)));
}


void ADDFunction::printEssentialVariables(void) const {
  std::string ret;
  std::set<std::string> variables=ADDFunction::getEssentialVariables();

  if((ADD::variableOrdering.size()>0) && (variables.find(ADD::variableOrdering.getVariable(0)->getName())!=variables.end())) ret+=ADD::variableOrdering.getVariable(0)->getName();
  for(unsigned long i=1;i<ADD::variableOrdering.size();++i){
    if(variables.find(ADD::variableOrdering.getVariable(i)->getName())!=variables.end()) ret+=","+ADD::variableOrdering.getVariable(i)->getName();
  }

  std::cout << ret << std::endl;
}

std::set<std::string> ADDFunction::getEssentialVariables(void) const {
  std::set<std::string> variables;
  std::set<ADDBaseNode*> visited;
  std::stack<ADDBaseNode*> nodeStack;
  ADDBaseNode* node;

  visited.insert(rootNode);
  nodeStack.push(rootNode);
  while(nodeStack.size()>0){
    node=nodeStack.top();
    nodeStack.pop();
    if(node && !node->isDrain()){
      ADDBaseNode* succ0=((ADDInnerNode*)node)->getSucc0();
      if(variables.find(ADD::variableOrdering.getVariable(node->getLevel())->getName())==variables.end()) variables.insert(ADD::variableOrdering.getVariable(node->getLevel())->getName());
      if(visited.find(succ0)==visited.end()){
        visited.insert(succ0);
        nodeStack.push(succ0);
      }
      ADDBaseNode* succ1=((ADDInnerNode*)node)->getSucc1();
      if(visited.find(succ1)==visited.end()){
        visited.insert(succ1);
        nodeStack.push(succ1);
      }
    }
  }

  return variables;
}

ADDBaseNode* getAssignment(unsigned int lastLevel, const ADDBaseNode* node){
  if(node->isDrain()) return (node!=ADD::zeroDrain)?ADD::oneDrain:ADD::zeroDrain;

  ADDBaseNode* tempNode;
  unsigned long currentLevel=node->getLevel();

  if(static_cast<const ADDInnerNode*>(node)->getSucc0()!=ADD::zeroDrain){
    tempNode=findOrAdd(node->getUniqueTable(),getAssignment(currentLevel+1, static_cast<const ADDInnerNode*>(node)->getSucc0()),ADD::zeroDrain);
  } else {
    tempNode=findOrAdd(node->getUniqueTable(),ADD::zeroDrain,getAssignment(currentLevel+1, static_cast<const ADDInnerNode*>(node)->getSucc1()));
  }

  while(lastLevel<currentLevel){
    tempNode=findOrAdd(ADD::variableOrdering.getData(--currentLevel),tempNode,ADD::zeroDrain);
  }

  return tempNode;
}

ADDFunction ADDFunction::getAssignment(void) const {
  ADDBaseNode* node=::getAssignment(0,rootNode);
  unsigned long level=node->getLevel();
  while(level>0){
    node=findOrAdd(ADD::variableOrdering.getData(--level),node,ADD::zeroDrain);
  }
  return ADDFunction(node);
}

bool getAssignment(const ADDBaseNode* node, char* buffer){
  if(node->isDrain()) return (node!=ADD::zeroDrain);

  if(!getAssignment(static_cast<const ADDInnerNode*>(node)->getSucc0(),buffer)){
    getAssignment(static_cast<const ADDInnerNode*>(node)->getSucc1(),buffer);
    buffer[node->getLevel()]='1';
  }

  return true;
}

void ADDFunction::getAssignment(char* buffer) const {
  //initially, every variable gets the value false
  for(unsigned long i=0;i<ADD::variableOrdering.size();++i) buffer[i]='0';

  ::getAssignment(rootNode,buffer);
}

double countAssignments(const ADDBaseNode* node, map<const ADDBaseNode*,double>& CT){
  if(node->isDrain()) return (node==ADD::zeroDrain)?0.0:1.0;

  double ret1, ret2;

  if(CT.find(node)!=CT.end()) return CT[node];

  {
    ADDBaseNode* tempNode=static_cast<const ADDInnerNode*>(node)->getSucc0();
    ret1=countAssignments(tempNode,CT);
    unsigned long diff=((!tempNode || tempNode->isDrain())?ADD::variableOrdering.size():tempNode->getLevel())-node->getLevel()-1;
    while(diff){
      ret1*=2.0;
      --diff;
    }
  }

  {
    ADDBaseNode* tempNode=static_cast<const ADDInnerNode*>(node)->getSucc1();
    ret2=countAssignments(tempNode,CT);
    unsigned long diff=((!tempNode || tempNode->isDrain())?ADD::variableOrdering.size():tempNode->getLevel())-node->getLevel()-1;
    while(diff){
      ret2*=2.0;
      --diff;
    }
  }

  CT[node]=ret1+ret2;

  return ret1+ret2;
}

double ADDFunction::countAssignments(void) const {
  map<const ADDBaseNode*,double> CT;
  double ret=::countAssignments(rootNode,CT);

  unsigned long diff=rootNode->isDrain()?ADD::variableOrdering.size():rootNode->getLevel();
  while(diff){
    ret*=2.0;
    --diff;
  }

  return ret;
}

double countAssignments(const ADDBaseNode* node, const ADDBaseNode* cubeset_node, map<pair<const ADDBaseNode*, const ADDBaseNode*>,double>& CT) {
  double factor=1.0;
  double count, count0, count1;
  
  if(CT.find(make_pair(node, cubeset_node))!=CT.end()) return CT[make_pair(node, cubeset_node)];
  
  const ADDBaseNode* cubeset_cur=cubeset_node;
  while (cubeset_cur->getLevel() < node->getLevel()) {
    factor*=2.0;
    cubeset_cur=static_cast<const ADDInnerNode*>(cubeset_cur)->getSucc1();
  }

  if(node->isDrain()) {
    count=(node==ADD::zeroDrain)?0.0:1.0;
  } else {
    // check that node->getLevel() == cubeset_node->getLevel()
    // ....
    
    // at inner node
    count0=countAssignments(static_cast<const ADDInnerNode*>(node)->getSucc0(), static_cast<const ADDInnerNode*>(cubeset_cur)->getSucc1(), CT);
    count1=countAssignments(static_cast<const ADDInnerNode*>(node)->getSucc1(), static_cast<const ADDInnerNode*>(cubeset_cur)->getSucc1(), CT);
    
    count=(count0+count1);
  }
  
  CT[make_pair(node, cubeset_cur)]=count;
  
  double count_with_factor=count*factor;
  CT[make_pair(node, cubeset_node)]=count_with_factor;

  return count_with_factor;     
}


double ADDFunction::countAssignments(const ADDFunction& cubeSet) const {
  if(!rootNode ||
     !cubeSet.rootNode) return 0.0;

  map<pair<const ADDBaseNode*, const ADDBaseNode*>, double> CT;
  double count=::countAssignments(rootNode, cubeSet.rootNode, CT);

  return count;
}


ADDFunction ADDFunction::getRandomAssignment(const ADDFunction& cube_set) const {
  map<pair<const ADDBaseNode*, const ADDBaseNode*>, double> CT;
  ::countAssignments(rootNode, cube_set.rootNode, CT);
  
  ADDBaseNode* cube_current=cube_set.rootNode;
  ADDBaseNode* node_current=rootNode;

  std::vector<ADDBaseNode*> cube_nodes;
  std::vector<bool> choices;  // true = choose 1-successor, false = 0-successor

  while (cube_current!=ADD::oneDrain) {
    cube_nodes.push_back(cube_current);

    if (node_current->isDrain() ||
        cube_current->getLevel() < node_current->getLevel()) {
      // level is not relevant in function, 0 or 1 with 1/2 probability
      
      bool rand=RandomHelper::getBool();
      choices.push_back(rand);
    } else {
      ADDBaseNode* w0=static_cast<const ADDInnerNode*>(node_current)->getSucc0();
      ADDBaseNode* w1=static_cast<const ADDInnerNode*>(node_current)->getSucc1();

      if (w0->isDrain() && w0==ADD::zeroDrain) {
        choices.push_back(true);
        node_current=w1;
      } else if (w1->isDrain() && w1==ADD::zeroDrain) {
        choices.push_back(false);
        node_current=w0;
      } else {
        double count0=CT[make_pair(w0, static_cast<const ADDInnerNode*>(cube_current)->getSucc1())];
        double count1=CT[make_pair(w1, static_cast<const ADDInnerNode*>(cube_current)->getSucc1())];

        double sum=count0+count1;
        double rand=RandomHelper::getUDouble(sum);
        if (rand >= count0) {
          choices.push_back(true);
          node_current=w1;
        } else {
          choices.push_back(false);
          node_current=w0;
        }          
      }
    }

    cube_current=static_cast<const ADDInnerNode*>(cube_current)->getSucc1();
  }

  // construct the assignment
  ADDBaseNode* result=ADD::oneDrain;
  for (unsigned int i=cube_nodes.size(); i>0; i--) {
    bool choice=choices.at(i-1);
    
    if (choice) {
      // succ1 chosen
      result=findOrAdd(cube_nodes.at(i-1)->getUniqueTable(), ADD::zeroDrain, result);
    } else {
      // succ0 chosen
      result=findOrAdd(cube_nodes.at(i-1)->getUniqueTable(), result, ADD::zeroDrain);
    }
  }

  return ADDFunction(result);
}


void printPaths(const unsigned long lastLevel, const ADDBaseNode* node, char* minterm){
  unsigned long i;

  if(!node->isDrain()){
    for(i=lastLevel;i<node->getLevel();++i) minterm[i]='-';
    minterm[node->getLevel()]='0';
    printPaths(node->getLevel()+1,static_cast<const ADDInnerNode*>(node)->getSucc0(),minterm);
    minterm[node->getLevel()]='1';
    printPaths(node->getLevel()+1,static_cast<const ADDInnerNode*>(node)->getSucc1(),minterm);
  } else {
    if(node!=ADD::zeroDrain){
      for(i=lastLevel;i<ADD::variableOrdering.size();++i) minterm[i]='-';
      for(i=0;i<ADD::variableOrdering.size();++i) std::cout << minterm[i];
      std::cout << "|" << static_cast<const ADDTerminalNode*>(node)->getValue() << std::endl;
    }
  }
}

void ADDFunction::printPaths(void) const {
  VSIZEARRAY(char, minterm, ADD::variableOrdering.size());

  ::printPaths(0,rootNode,VSIZEARRAY_PTR(minterm));
}

bool getPath(const ADDBaseNode* node, char *minterm) {
  if (node->isDrain()) {
    return node!=ADD::zeroDrain;
  }

  if (getPath(static_cast<const ADDInnerNode*>(node)->getSucc0(), minterm)) {
    minterm[node->getLevel()]='0';
  } else if (getPath(static_cast<const ADDInnerNode*>(node)->getSucc1(), minterm)) {
    minterm[node->getLevel()]='1';
  } else {
    throw "No path!";
  }

  return true;
}

void ADDFunction::getPath(char* minterm) const {
  //initially, every variable is a don't care symbol
  for(unsigned long i=0;i<ADD::variableOrdering.size();++i) 
    minterm[i]='-';

  ::getPath(rootNode, minterm);
}

double countPaths(const ADDBaseNode* node, map<const ADDBaseNode*,double>& CT){
  if(!node->isDrain()){
    double ret=0.0;
    if(CT.find(node)!=CT.end()) return CT[node];
    else {
      ret=countPaths(static_cast<const ADDInnerNode*>(node)->getSucc0(),CT);
      ret+=countPaths(static_cast<const ADDInnerNode*>(node)->getSucc1(),CT);
    }

    CT[node]=ret;

    return ret;
  } else return (node!=ADD::zeroDrain)?1.0:0.0;
}

double ADDFunction::countPaths(void) const {
  map<const ADDBaseNode*,double> CT;
  return ::countPaths(rootNode,CT);
}

ADDFunction operator-(const ADDFunction& function){
  return ADDFunction(::negate(const_cast<ADDBaseNode*>(function.rootNode)));
}

ADDBaseNode* ITE(const ADDBaseNode* u, const ADDBaseNode* v1, const ADDBaseNode* v2){
  INTERRUPTIBLE();

  if(v1==v2) return const_cast<ADDBaseNode*>(v1);
  if(u->isDrain()) return const_cast<ADDBaseNode*>((u==ADD::zeroDrain)?v2:v1);
  if(v2==ADD::zeroDrain) return AND(u,v1);
  if(v1==ADD::oneDrain) return OR(u,v2);

  unsigned long index=v1->getRealLevel();
  if(v2->getRealLevel()<index) index=v2->getRealLevel();
  if(u->getRealLevel()<index){
    index=u->getRealLevel();
    const ADDInnerNode* innerNode=static_cast<const ADDInnerNode*>(u);
    if((innerNode->getSucc0()==ADD::zeroDrain) && (innerNode->getSucc1()==ADD::oneDrain)) return findOrAdd(u->getUniqueTable(),const_cast<ADDBaseNode*>(v2),const_cast<ADDBaseNode*>(v1));
    if((innerNode->getSucc1()==ADD::zeroDrain) && (innerNode->getSucc0()==ADD::oneDrain)) return findOrAdd(u->getUniqueTable(),const_cast<ADDBaseNode*>(v1),const_cast<ADDBaseNode*>(v2));
  }

  if(ADDBaseNode* comp=ADD::BoolComputedTable.find(calcPtr(u,ITE_OP),v1,v2)) return comp;

  ADDBaseNode* u0, *u1;
  ADDBaseNode* v10, *v11;
  ADDBaseNode* v20, *v21;

  ADDUniqueTable* minTable=0;
  if(u->getRealLevel()<=index){
    minTable=u->getUniqueTable();
    u0=static_cast<const ADDInnerNode*>(u)->getSucc0();
    u1=static_cast<const ADDInnerNode*>(u)->getSucc1();
  } else u0=u1=const_cast<ADDBaseNode*>(u);
  if(v1->getRealLevel()==index){
    minTable=v1->getUniqueTable();
    v10=static_cast<const ADDInnerNode*>(v1)->getSucc0();
    v11=static_cast<const ADDInnerNode*>(v1)->getSucc1();
  } else v10=v11=const_cast<ADDBaseNode*>(v1);
  if(v2->getRealLevel()==index) {
    minTable=v2->getUniqueTable();
    v20=static_cast<const ADDInnerNode*>(v2)->getSucc0();
    v21=static_cast<const ADDInnerNode*>(v2)->getSucc1();
  } else v20=v21=const_cast<ADDBaseNode*>(v2);

  u0=ITE(u0,v10,v20);
  u1=ITE(u1,v11,v21);

  ADDBaseNode* node=(u0==u1)?u0:findOrAdd(minTable,u0,u1);

  //insert result into computed table
  ADD::BoolComputedTable.insert(calcPtr(u,ITE_OP),v1,v2,node);

  return node;
}

ADDFunction ITE(const ADDFunction& i, const ADDFunction& t, const ADDFunction& e){
  return ADDFunction(ITE(i.rootNode,t.rootNode,e.rootNode));
}

ADDFunction LESS(const ADDFunction& u, const ADDFunction& v){
  return ADDFunction(::less(u.rootNode,v.rootNode));
}

ADDBaseNode* lessEqual(const ADDBaseNode* u, const ADDBaseNode* v){
  INTERRUPTIBLE();

  if(u==v) return ADD::oneDrain;
  if(u->isDrain() && v->isDrain()) return (static_cast<const ADDTerminalNode*>(u)->getValue()<=static_cast<const ADDTerminalNode*>(v)->getValue())?ADD::oneDrain:ADD::zeroDrain;

  if(ADDBaseNode* comp=ADD::CompareComputedTable.find(calcPtr(u,LEE_OP),v)) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;

  ADDUniqueTable* minTable;

  if(u->getRealLevel()<=v->getRealLevel()){
    minTable=u->getUniqueTable();
    if(v->getUniqueTable()==minTable){
      w0=::lessEqual(static_cast<const ADDInnerNode*>(u)->getSucc0(),static_cast<const ADDInnerNode*>(v)->getSucc0());
      w1=::lessEqual(static_cast<const ADDInnerNode*>(u)->getSucc1(),static_cast<const ADDInnerNode*>(v)->getSucc1());
    } else {
      w0=::lessEqual(static_cast<const ADDInnerNode*>(u)->getSucc0(),v);
      w1=::lessEqual(static_cast<const ADDInnerNode*>(u)->getSucc1(),v);
    }
  } else {
    minTable=v->getUniqueTable();
    w0=::lessEqual(u,static_cast<const ADDInnerNode*>(v)->getSucc0());
    w1=::lessEqual(u,static_cast<const ADDInnerNode*>(v)->getSucc1());
  }

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);
  ADD::CompareComputedTable.insert(calcPtr(u,LEE_OP),v,node);

  return node;
}

ADDFunction LESS_EQUAL(const ADDFunction& u, const ADDFunction& v){
  return ADDFunction(::lessEqual(u.rootNode,v.rootNode));
}

ADDBaseNode* equal(const ADDBaseNode* u, const ADDBaseNode* v){
  INTERRUPTIBLE();

  if(u==v) return ADD::oneDrain;
  if(u->isDrain() && v->isDrain()) return (static_cast<const ADDTerminalNode*>(u)->getValue()==static_cast<const ADDTerminalNode*>(v)->getValue())?ADD::oneDrain:ADD::zeroDrain;

  if(ADDBaseNode* comp=ADD::CompareComputedTable.find(calcPtr(u,EQU_OP),v)) return comp;

  ADDBaseNode* w0;
  ADDBaseNode* w1;

  ADDUniqueTable* minTable;

  if(u->getRealLevel()<=v->getRealLevel()){
    minTable=u->getUniqueTable();
    if(v->getUniqueTable()==minTable){
      w0=::equal(static_cast<const ADDInnerNode*>(u)->getSucc0(),static_cast<const ADDInnerNode*>(v)->getSucc0());
      w1=::equal(static_cast<const ADDInnerNode*>(u)->getSucc1(),static_cast<const ADDInnerNode*>(v)->getSucc1());
    } else {
      w0=::equal(static_cast<const ADDInnerNode*>(u)->getSucc0(),v);
      w1=::equal(static_cast<const ADDInnerNode*>(u)->getSucc1(),v);
    }
  } else {
    minTable=v->getUniqueTable();
    w0=::equal(u,static_cast<const ADDInnerNode*>(v)->getSucc0());
    w1=::equal(u,static_cast<const ADDInnerNode*>(v)->getSucc1());
  }

  ADDBaseNode* node=(w0==w1)?w0:findOrAdd(minTable,w0,w1);
  ADD::CompareComputedTable.insert(calcPtr(u,EQU_OP),v,node);

  return node;
}

ADDFunction EQUAL(const ADDFunction& u, const ADDFunction& v){
  return ADDFunction(::equal(u.rootNode,v.rootNode));
}

ADDFunction MAXIMUM(const ADDFunction& u, const ADDFunction& v){
  return ADDFunction(::max(u.rootNode,v.rootNode));
}

ADDFunction MINIMUM(const ADDFunction& u, const ADDFunction& v){
  return ADDFunction(::min(u.rootNode,v.rootNode));
}

ADDBaseNode* MM(const ADDBaseNode* A, const ADDBaseNode* B, const ADDBaseNode* varSet, const unsigned long m){
  if((A==ADD::zeroDrain) || (B==ADD::zeroDrain)) return ADD::zeroDrain;
  if((A->isDrain() && B->isDrain()) || (m==0)){
    ADDTerminalNode* tempDrain=findOrAddDrain(static_cast<double>(1<<m));
    ADDBaseNode* result=::multiplication(::multiplication(A,B),static_cast<ADDBaseNode*>(tempDrain));

    if(result!=tempDrain && tempDrain->preNodeCount==0) tempDrain->garbageCollect();

    return result;
  }

  if(ADDBaseNode* comp=ADD::MatrixComputedTable.find(calcPtr(A,MMA_OP),B,varSet)) return comp;

  ADDUniqueTable* uniqueTable=varSet->getUniqueTable();
  ADDUniqueTable* nextUniqueTable=ADD::variableOrdering.getData(uniqueTable->getLevel()+1);

  ADDBaseNode* A00, *A01, *A10, *A11;
  if(uniqueTable==A->getUniqueTable()){
    ADDBaseNode* A0=static_cast<const ADDInnerNode*>(A)->getSucc0();
    ADDBaseNode* A1=static_cast<const ADDInnerNode*>(A)->getSucc1();
    if(A0->getUniqueTable()==nextUniqueTable){
      A00=static_cast<ADDInnerNode*>(A0)->getSucc0();
      A01=static_cast<ADDInnerNode*>(A0)->getSucc1();
    } else {
      A00=A01=A0;
    }
    if(A1->getUniqueTable()==nextUniqueTable){
      A10=static_cast<ADDInnerNode*>(A1)->getSucc0();
      A11=static_cast<ADDInnerNode*>(A1)->getSucc1();
    } else {
      A10=A11=A1;
    }
  } else {
    if(A->getUniqueTable()==nextUniqueTable){
      A00=A10=static_cast<const ADDInnerNode*>(A)->getSucc0();
      A01=A11=static_cast<const ADDInnerNode*>(A)->getSucc1();
    } else {
      A00=A01=A10=A11=const_cast<ADDBaseNode*>(A);
    }
  }

  ADDBaseNode* B00, *B01, *B10, *B11;
  if(uniqueTable==B->getUniqueTable()){
    ADDBaseNode* B0=static_cast<const ADDInnerNode*>(B)->getSucc0();
    ADDBaseNode* B1=static_cast<const ADDInnerNode*>(B)->getSucc1();
    if(B0->getUniqueTable()==nextUniqueTable){
      B00=static_cast<ADDInnerNode*>(B0)->getSucc0();
      B01=static_cast<ADDInnerNode*>(B0)->getSucc1();
    } else {
      B00=B01=B0;
    }
    if(B1->getUniqueTable()==nextUniqueTable){
      B10=static_cast<ADDInnerNode*>(B1)->getSucc0();
      B11=static_cast<ADDInnerNode*>(B1)->getSucc1();
    } else {
      B10=B11=B1;
    }
  } else {
    if(B->getUniqueTable()==nextUniqueTable){
      B00=B10=static_cast<const ADDInnerNode*>(B)->getSucc0();
      B01=B11=static_cast<const ADDInnerNode*>(B)->getSucc1();
    } else {
      B00=B01=B10=B11=const_cast<ADDBaseNode*>(B);
    }
  }

  ADDBaseNode* newVarSet=static_cast<const ADDInnerNode*>(varSet)->getSucc1();

  ADDBaseNode* w0000=::MM(A00,B00,newVarSet,m-1);
  ADDBaseNode* w0110=::MM(A01,B10,newVarSet,m-1);
  ADDBaseNode* w0001=::MM(A00,B01,newVarSet,m-1);
  ADDBaseNode* w0111=::MM(A01,B11,newVarSet,m-1);
  ADDBaseNode* w1000=::MM(A10,B00,newVarSet,m-1);
  ADDBaseNode* w1110=::MM(A11,B10,newVarSet,m-1);
  ADDBaseNode* w1001=::MM(A10,B01,newVarSet,m-1);
  ADDBaseNode* w1111=::MM(A11,B11,newVarSet,m-1);

  ADDBaseNode* c00=addition(w0000,w0110);
  ADDBaseNode* c01=addition(w0001,w0111);
  ADDBaseNode* c10=addition(w1000,w1110);
  ADDBaseNode* c11=addition(w1001,w1111);

  ADDBaseNode* result0=(c00==c01)?c00:findOrAdd(nextUniqueTable,c00,c01);
  ADDBaseNode* result1=(c10==c11)?c11:findOrAdd(nextUniqueTable,c10,c11);
  ADDBaseNode* result=(result0==result1)?result1:findOrAdd(uniqueTable,result0,result1);

  if(result!=w0000 && w0000->preNodeCount==0) w0000->garbageCollect();
  if(result!=w0110 && w0110->preNodeCount==0) w0110->garbageCollect();
  if(result!=w0001 && w0001->preNodeCount==0) w0001->garbageCollect();
  if(result!=w0111 && w0111->preNodeCount==0) w0111->garbageCollect();
  if(result!=w1000 && w1000->preNodeCount==0) w1000->garbageCollect();
  if(result!=w1110 && w1110->preNodeCount==0) w1110->garbageCollect();
  if(result!=w1001 && w1001->preNodeCount==0) w1001->garbageCollect();
  if(result!=w1111 && w1111->preNodeCount==0) w1111->garbageCollect();
  if(result!=c00 && c00->preNodeCount==0) c00->garbageCollect();
  if(result!=c01 && c01->preNodeCount==0) c01->garbageCollect();
  if(result!=c10 && c10->preNodeCount==0) c10->garbageCollect();
  if(result!=c11 && c11->preNodeCount==0) c11->garbageCollect();

  ADD::MatrixComputedTable.insert(calcPtr(A,MMA_OP),B,varSet,result);

  return result;
}

ADDFunction MM(const ADDFunction& A, const ADDFunction& B, const ADDFunction& varSet){
  ADDBaseNode* node=varSet.rootNode;

  unsigned long m=0;
  while(!node->isDrain()){
    node=static_cast<ADDInnerNode*>(node)->getSucc1();
    ++m;
  }

  return ADDFunction(::MM(A.rootNode,B.rootNode,varSet.rootNode,m));
}

ADDBaseNode* MV(const ADDBaseNode* A, const ADDBaseNode* b, const ADDBaseNode* varSet, const unsigned long m){
  if((A==ADD::zeroDrain) || (b==ADD::zeroDrain)) return ADD::zeroDrain;
  if((A->isDrain() && b->isDrain()) || (m==0)){
    ADDTerminalNode* tempDrain=findOrAddDrain(static_cast<double>(1<<m));
    ADDBaseNode* result=::multiplication(::multiplication(A,b),static_cast<ADDBaseNode*>(tempDrain));

    if(result!=tempDrain && tempDrain->preNodeCount==0) tempDrain->garbageCollect();

    return result;
  }

  if(ADDBaseNode* comp=ADD::MatrixComputedTable.find(calcPtr(A,MAV_OP),b,varSet)) return comp;

  ADDUniqueTable* uniqueTable=varSet->getUniqueTable();
  ADDUniqueTable* nextUniqueTable=ADD::variableOrdering.getData(uniqueTable->getLevel()+1);

  ADDBaseNode* A00, *A01, *A10, *A11;
  if(uniqueTable==A->getUniqueTable()){
    ADDBaseNode* A0=static_cast<const ADDInnerNode*>(A)->getSucc0();
    ADDBaseNode* A1=static_cast<const ADDInnerNode*>(A)->getSucc1();
    if(A0->getUniqueTable()==nextUniqueTable){
      A00=static_cast<ADDInnerNode*>(A0)->getSucc0();
      A01=static_cast<ADDInnerNode*>(A0)->getSucc1();
    } else {
      A00=A01=A0;
    }
    if(A1->getUniqueTable()==nextUniqueTable){
      A10=static_cast<ADDInnerNode*>(A1)->getSucc0();
      A11=static_cast<ADDInnerNode*>(A1)->getSucc1();
    } else {
      A10=A11=A1;
    }
  } else {
    if(A->getUniqueTable()==nextUniqueTable){
      A00=A10=static_cast<const ADDInnerNode*>(A)->getSucc0();
      A01=A11=static_cast<const ADDInnerNode*>(A)->getSucc1();
    } else {
      A00=A01=A10=A11=const_cast<ADDBaseNode*>(A);
    }
  }

  ADDBaseNode* b0, *b1;
  if(uniqueTable==b->getUniqueTable()){
    b0=static_cast<const ADDInnerNode*>(b)->getSucc0();
    b1=static_cast<const ADDInnerNode*>(b)->getSucc1();
  } else {
    b0=b1=const_cast<ADDBaseNode*>(b);
  }

  ADDBaseNode* newVarSet=static_cast<const ADDInnerNode*>(varSet)->getSucc1();

  ADDBaseNode* w00=::MV(A00,b0,newVarSet,m-1);
  ADDBaseNode* w01=::MV(A01,b1,newVarSet,m-1);
  ADDBaseNode* w10=::MV(A10,b0,newVarSet,m-1);
  ADDBaseNode* w11=::MV(A11,b1,newVarSet,m-1);

  ADDBaseNode* c0=addition(w00,w01);
  ADDBaseNode* c1=addition(w10,w11);

  ADDBaseNode* result=(c0==c1)?c1:findOrAdd(uniqueTable,c0,c1);

  if(result!=w00 && w00->preNodeCount==0) w00->garbageCollect();
  if(result!=w01 && w01->preNodeCount==0) w01->garbageCollect();
  if(result!=w10 && w10->preNodeCount==0) w10->garbageCollect();
  if(result!=w11 && w11->preNodeCount==0) w11->garbageCollect();
  if(result!=c0 && c0->preNodeCount==0) c0->garbageCollect();
  if(result!=c1 && c1->preNodeCount==0) c1->garbageCollect();

  ADD::MatrixComputedTable.insert(calcPtr(A,MAV_OP),b,varSet,result);

  return result;
}

ADDFunction MV(const ADDFunction& A, const ADDFunction& b, const ADDFunction& varSet){
  ADDBaseNode* node=varSet.rootNode;

  unsigned long m=0;
  while(!node->isDrain()){
    node=static_cast<ADDInnerNode*>(node)->getSucc1();
    ++m;
  }

  return ADDFunction(::MV(A.rootNode,b.rootNode,varSet.rootNode,m));
}

ADDBaseNode* VM(const ADDBaseNode* b, const ADDBaseNode* A, const ADDBaseNode* varSet, const unsigned long m){
  if((A==ADD::zeroDrain) || (b==ADD::zeroDrain)) return ADD::zeroDrain;
  if((A->isDrain() && b->isDrain()) || (m==0)){
    ADDTerminalNode* tempDrain=findOrAddDrain(static_cast<double>(1<<m));
    ADDBaseNode* result=::multiplication(::multiplication(A,b),static_cast<ADDBaseNode*>(tempDrain));

    if(result!=tempDrain && tempDrain->preNodeCount==0) tempDrain->garbageCollect();

    return result;
  }

  if(ADDBaseNode* comp=ADD::MatrixComputedTable.find(calcPtr(b,VMA_OP),A,varSet)) return comp;

  ADDUniqueTable* uniqueTable=varSet->getUniqueTable();
  ADDUniqueTable* nextUniqueTable=ADD::variableOrdering.getData(uniqueTable->getLevel()+1);

  ADDBaseNode* A00, *A01, *A10, *A11;
  if(uniqueTable==A->getUniqueTable()){
    ADDBaseNode* A0=static_cast<const ADDInnerNode*>(A)->getSucc0();
    ADDBaseNode* A1=static_cast<const ADDInnerNode*>(A)->getSucc1();
    if(A0->getUniqueTable()==nextUniqueTable){
      A00=static_cast<ADDInnerNode*>(A0)->getSucc0();
      A01=static_cast<ADDInnerNode*>(A0)->getSucc1();
    } else {
      A00=A01=A0;
    }
    if(A1->getUniqueTable()==nextUniqueTable){
      A10=static_cast<ADDInnerNode*>(A1)->getSucc0();
      A11=static_cast<ADDInnerNode*>(A1)->getSucc1();
    } else {
      A10=A11=A1;
    }
  } else {
    if(A->getUniqueTable()==nextUniqueTable){
      A00=A10=static_cast<const ADDInnerNode*>(A)->getSucc0();
      A01=A11=static_cast<const ADDInnerNode*>(A)->getSucc1();
    } else {
      A00=A01=A10=A11=const_cast<ADDBaseNode*>(A);
    }
  }

  ADDBaseNode* b0, *b1;
  if(uniqueTable==b->getUniqueTable()){
    b0=static_cast<const ADDInnerNode*>(b)->getSucc0();
    b1=static_cast<const ADDInnerNode*>(b)->getSucc1();
  } else {
    b0=b1=const_cast<ADDBaseNode*>(b);
  }

  ADDBaseNode* newVarSet=static_cast<const ADDInnerNode*>(varSet)->getSucc1();

  ADDBaseNode* w00=::VM(b0,A00,newVarSet,m-1);
  ADDBaseNode* w01=::VM(b0,A01,newVarSet,m-1);
  ADDBaseNode* w10=::VM(b1,A10,newVarSet,m-1);
  ADDBaseNode* w11=::VM(b1,A11,newVarSet,m-1);

  ADDBaseNode* c0=addition(w00,w10);
  ADDBaseNode* c1=addition(w01,w11);

  ADDBaseNode* result=(c0==c1)?c1:findOrAdd(uniqueTable,c0,c1);

  if(result!=w00 && w00->preNodeCount==0) w00->garbageCollect();
  if(result!=w01 && w01->preNodeCount==0) w01->garbageCollect();
  if(result!=w10 && w10->preNodeCount==0) w10->garbageCollect();
  if(result!=w11 && w11->preNodeCount==0) w11->garbageCollect();
  if(result!=c0 && c0->preNodeCount==0) c0->garbageCollect();
  if(result!=c1 && c1->preNodeCount==0) c1->garbageCollect();

  ADD::MatrixComputedTable.insert(calcPtr(b,VMA_OP),A,varSet,result);

  return result;
}

ADDFunction VM(const ADDFunction& A, const ADDFunction& b, const ADDFunction& varSet){
  ADDBaseNode* node=varSet.rootNode;

  unsigned long m=0;
  while(!node->isDrain()){
    node=static_cast<ADDInnerNode*>(node)->getSucc1();
    ++m;
  }

  return ADDFunction(::VM(b.rootNode,A.rootNode,varSet.rootNode,m));
}




ADDBaseNode* equivalenceClasses(const ADDBaseNode* u, const ADDBaseNode* equiv, const ADDBaseNode* cubeSet){
  INTERRUPTIBLE();

  if (u==equiv) return ADD::oneDrain;

  if(u->isDrain()) return ADD::zeroDrain;

  unsigned long index=u->getRealLevel();
  
  if (equiv->getRealLevel()<index) return ADD::zeroDrain;

  while (cubeSet->getRealLevel()<index) {
    cubeSet=static_cast<const ADDInnerNode*>(cubeSet)->getSucc1();
  }

  if(ADDBaseNode* e_class=ADD::SpecialComputedTable.find(calcPtr(u,EQI_OP), equiv, cubeSet)) return e_class;

  ADDBaseNode* w0;
  ADDBaseNode* w1;
  ADDBaseNode* result;

  if (u->getUniqueTable()==cubeSet->getUniqueTable()) {
    const ADDBaseNode* equiv0;
    const ADDBaseNode* equiv1;
    
    if (equiv->getUniqueTable()==u->getUniqueTable()) {
      equiv0=static_cast<const ADDInnerNode*>(equiv)->getSucc0();
      equiv1=static_cast<const ADDInnerNode*>(equiv)->getSucc1();
    } else equiv0=equiv1=equiv;

    w0=equivalenceClasses(static_cast<const ADDInnerNode*>(u)->getSucc0(),
                          equiv0,
                          static_cast<const ADDInnerNode*>(cubeSet)->getSucc1());

    w1=equivalenceClasses(static_cast<const ADDInnerNode*>(u)->getSucc1(),
                          equiv1,
                          static_cast<const ADDInnerNode*>(cubeSet)->getSucc1());
    
    result=::AND(w0, w1);

    if(result!=w0 && w0->preNodeCount==0) w0->garbageCollect();
    if(result!=w1 && w1->preNodeCount==0) w1->garbageCollect();
  } else {
    w0=equivalenceClasses(static_cast<const ADDInnerNode*>(u)->getSucc0(),
                          equiv,
                          cubeSet);
    w1=equivalenceClasses(static_cast<const ADDInnerNode*>(u)->getSucc1(),
                          equiv,
                          cubeSet);

    result=(w0==w1)?w0:findOrAdd(u->getUniqueTable(),w0,w1);
  }

  ADD::SpecialComputedTable.insert(calcPtr(u,EQI_OP),equiv,cubeSet,result);

  return result;
}

ADDFunction ADDFunction::equivalenceClasses(const ADDFunction& equiv,
                                            const ADDFunction& cubeSet) const {
  return ADDFunction(::equivalenceClasses(rootNode, equiv.rootNode, cubeSet.rootNode));
}


std::set<ADDBaseNode*> ADDFunction::getNodes() const {
  std::set<ADDBaseNode*> visited;
  std::stack<const ADDBaseNode*> nodeStack;

  visited.insert(rootNode);
  nodeStack.push(rootNode);
  while(nodeStack.size()>0){
    const ADDBaseNode* node=nodeStack.top();
    nodeStack.pop();
    if LIKELY(!node->isDrain()){
      {
        ADDBaseNode* tempNode=static_cast<const ADDInnerNode*>(node)->getSucc0();
        if(visited.find(tempNode)==visited.end()){
          visited.insert(tempNode);
          nodeStack.push(tempNode);
        }
      }
      {
        ADDBaseNode* tempNode=static_cast<const ADDInnerNode*>(node)->getSucc1();
        if(visited.find(tempNode)==visited.end()){
          visited.insert(tempNode);
          nodeStack.push(tempNode);
        }
      }
    }
  }

  return visited;
}
