/***************************************************************************
                           blifinput.h  -  description
                             -------------------
    begin                : Fri May 28 2004
    copyright            :(C) 2002 by Joern Ossowski
    email                : ossowski@cs.uni-bonn.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software;you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation;either version 2 of the License, or      *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef BLIF_INPUT_H
#define BLIF_INPUT_H

#include <common/stringhelper.h>

#include <iostream>
#include <fstream>
#include <map>
#include <set>
#include <vector>

using namespace std;

struct BLIFProgressBar{
  unsigned long index;
  unsigned long size;
  unsigned long step;
  unsigned long maxStepSize;
  unsigned long lineSize;
};

struct BLIFType{
  enum{INPUT,POSSIBLE_INPUT,OUTPUT,POSSIBLE_OUTPUT,GATE};
  unsigned short type;
  unsigned long line;
  unsigned long gateCount;
};

struct BLIFNames{
  vector<string> lines;
  vector<string> inputs;
  string output;
  unsigned long line;
  bool onSetChosen;
};

inline void printBLIFProgressBar(BLIFProgressBar& bar){
  string temp;
  string print;
  unsigned int digits;

  temp=string("")+bar.size;
  digits=temp.size();

  temp=string("")+bar.index;
  while(temp.size()<digits) temp=" "+temp;

  print="statement: [";
  print+=temp;
  print+="/";
  print+=string("")+bar.size;
  print+=+"]     step:[";

  temp=string("")+bar.maxStepSize;
  digits=temp.size();

  temp=string("")+bar.step;
  while(temp.size()<digits) temp=" "+temp;

  print+=temp;
  print+="/";
  print+=string("")+bar.maxStepSize;
  print+="]";

  bar.lineSize=print.length();

  cerr << print;
  cerr.flush();
}

inline void clearBLIFProgressBar(BLIFProgressBar& bar){
  cerr << "\r";
  while(bar.lineSize){
    cerr << " ";
    --bar.lineSize;
  }
  cerr << "\r";
  cerr.flush();
}

//**************************ParseModule**************************
template <class T1, class T2, class T3>
class ParseModel{
enum{ERROR,UNRESOLVED,SUCCESS};
public:
  ParseModel(const string&);
  ParseModel(const string&, const string&);
  ~ParseModel(void);

  T2 getOutput(const string&);
  T2 getOutput(const unsigned long);

  unsigned long size(void) const;
  void clear(void);
private:
  void parseModel(void);
  void buildOutputs(void);
  bool declareInput(const string&);
  bool useInput(const string&);
  bool declareOutput(const string&);
  bool useOutput(const string&);
  unsigned short createNames(const string&);
  bool gotoModel(void) const;
  bool isCommand(string&, const string&) const;
  string readLine(void);

  ifstream file;
  vector<string> inputVector;
  vector<string> outputVector;
  map<string,BLIFType> typeMap;
  map<string,BLIFNames> blifNames;
  map<string,T2> gatesMap;
  string modelName;
  string filename;
  unsigned long lineNumber;
  unsigned long diff;
  bool inputsDeclared, outputsDeclared;
  bool modelNameSet;
  bool readNextLine;
  bool createOutputs;
  BLIFProgressBar progressBar;
};

template <class T1, class T2, class T3>
ParseModel<T1,T2,T3>::ParseModel(const string& filename){
  this->filename=filename;
  modelName=filename;
  modelNameSet=false;
  readNextLine=true;
  createOutputs=false;
  inputsDeclared=outputsDeclared=false;
  lineNumber=0;
  diff=0;
  file.open(filename.c_str(),ios::in);
  //TODO: check if file is a directory (strange sentence :-) )
  //TODO: whenever a submodel can be used the varorder must be created first (ZBDDs)

  parseModel();
}

template <class T1, class T2, class T3>
ParseModel<T1,T2,T3>::ParseModel(const string& filename, const string& modelname){
  this->filename=filename;
  modelName=modelname;
  modelNameSet=true;
  readNextLine=true;
  createOutputs=false;
  inputsDeclared=outputsDeclared=false;
  lineNumber=0;
  diff=0;
  file.open(filename.c_str(),ios::in);
  //TODO: check if file is a directory (strange sentence :-) )

  //TODO: check if model exists in file and search other files if not
  gotoModel(modelName);
  parseModel();
}

template <class T1, class T2, class T3>
ParseModel<T1,T2,T3>::~ParseModel(void){
  if(file) file.close();
}

template <class T1, class T2, class T3>
T2 ParseModel<T1,T2,T3>::getOutput(const string& outputname){
  //if outputname is not in typeMap something went wrong
  if(typeMap.find(outputname)==typeMap.end()){
    cerr << "[ERROR] requested function \"" << outputname << "\" not declared in " << filename << endl;
    return T2::ERROR();
  }

  //check if outputname is really an outpt
  if(typeMap[outputname].type!=BLIFType::OUTPUT) return T2::ERROR();

  //check if outputname has been defined
  if(gatesMap.find(outputname)!=gatesMap.end()) return gatesMap[outputname];
  cerr << "[ERROR] requested function \"" << outputname << "\" not defined in " << filename << endl;

  return T2::ERROR();
}

template <class T1, class T2, class T3>
T2 ParseModel<T1,T2,T3>::getOutput(const unsigned long index){
  return (index>=size())?T2::ERROR():getOutput(outputVector[index]);
}

template <class T1, class T2, class T3>
unsigned long ParseModel<T1,T2,T3>::size(void) const {
  return outputVector.size();
}

template <class T1, class T2, class T3>
void ParseModel<T1,T2,T3>::clear(void){
  typeMap.clear();
  blifNames.clear();
  outputVector.clear();
  gatesMap.clear();
}

template <class T1, class T2, class T3>
void ParseModel<T1,T2,T3>::parseModel(void){
  string line;
  string temp;

  //check if file could not be opened
  if(!file) return;

  //read next line till eof
  //(readNextLine can be set to ignore the last line in a command and read it again in this loop)
  while((!file.eof())||(!readNextLine)){
    //ignore reading a new line just once
    if(readNextLine) line=readLine();
    else readNextLine=true;

    if(isCommand(line,".model")){
    // handle .model
      //if model name is not set it is the main model
      if(!modelNameSet){
        modelName=line;
        modelNameSet=true;
      } else {
        //if name is set the model is completely read (outputs can be build)
        buildOutputs();
        return;
      }
    } else if(isCommand(line,".inputs")){
    // handle .input
      inputsDeclared=true;

      line=trim(line);
      while(!line.empty()){
        temp=popToken(line);
        //check if input can be declared
        if(!declareInput(temp)){
          clear();
          return;
        }
      }
    } else if(isCommand(line,".outputs")){
    // handle .output
      outputsDeclared=true;

      line=trim(line);
      while(!line.empty()){
        temp=popToken(line);
        //check if output can be declared
        if(!declareOutput(temp)){
          clear();
          return;
        }
      }
    } else if(isCommand(line,".latch")){
    // handle .latch
      line=trim(line);
      if((countTokens(line)<2) || (countTokens(line)>3)){
        cerr << "[ERROR:" << lineNumber << "] illegal latch declaration" << endl;
        clear();
        return;
      }
      //TODO: do not ignore initial state (this can be done with a circuit class)
      string inputName=popToken(line);
      string outputName=popToken(line);
      if(!declareInput(outputName)){
        clear();
        return;
      }
      if(!useInput(inputName)){
        clear();
        return;
      }
      outputName=outputName+"'";
      if(!declareOutput(outputName)){
        clear();
        return;
      }

      BLIFNames blifName;
      blifName.onSetChosen=true;
      blifName.inputs.push_back(inputName);
      blifName.output=outputName;
      blifName.lines.push_back("1");
      blifName.line=lineNumber;
      //insert '.names'
      blifNames[blifName.output]=blifName;
    } else if(isCommand(line,".names")){
    // handle .names
      unsigned long i;
      unsigned tokenCounter=countTokens(line);
      if(tokenCounter==0){
        cerr << "[ERROR:" << lineNumber << "] illegal single output cover" << endl;
        clear();
        return;
      }

      BLIFNames blifName;
      //set input names
      for(i=1;i<tokenCounter;++i) blifName.inputs.push_back(popToken(line));
      //set output name
      blifName.output=popToken(line);
      //set line number
      blifName.line=lineNumber;

      //store all .names lines in blifName
      do {
        line=readLine();
        line=trim(line);
        //if line is a new command or the last line give line to outer loop
        if((line.substr(0,1)==".")||(line.empty())) readNextLine=false;
        else {
          //check lines to build the outputs faster
          if(blifName.inputs.size()==0){
            //check if exactly one output exists
            if(line.length()!=1){
              cerr << "[ERROR:" << lineNumber << "] wrong number of outputs" << endl;
              clear();
              return;
            }
            //check if more than one line exists
            if(blifName.lines.size()>1){
              cerr << "[ERROR:" << lineNumber << "] more than one line in .names statement with no input variables" << endl;
              clear();
              return;
            }
            //check char
            if(line[0]=='0') blifName.onSetChosen=false;
            else if(line[0]=='1') blifName.onSetChosen=true;
            else {
              cerr << "[ERROR:" << lineNumber << "] unknown character '" << line[0] << "'"<< endl;
              clear();
              return;
            }
          } else {
            bool outputSpecified=false;
            //if a whitespace is missing try to add one
            if((countTokens(line)==1) && (line.length()==blifName.inputs.size()+1)) line=line.substr(0,line.length()-2)+" "+line[line.length()-1];
            //check number of inputs
            if(peekToken(line).length()!=blifName.inputs.size()){
              cerr << "[ERROR:" << lineNumber << "] wrong number of inputs" << endl;
              clear();
              return;
            }
            //check input chars
            for(i=0;i<blifName.inputs.size();++i){
              if((line[i]!='0')&&(line[i]!='1')&&(line[i]!='-')){
                cerr << "[ERROR:" << lineNumber << "] unknown character '" << line[i] << "'"<< endl;
                clear();
                return;
              }
            }
            //check output char
            if(line[line.length()-1]=='0'){
              if(outputSpecified){
                if(blifName.onSetChosen){
                  cerr << "[ERROR:" << lineNumber << "] mixing ON and OFF set definition is not allowed" << endl;
                  clear();
                  return;
                }
              } else {
                outputSpecified=true;
                blifName.onSetChosen=false;
              }
            } else if(line[line.length()-1]=='1'){
              if(outputSpecified){
                if(!blifName.onSetChosen){
                  cerr << "[ERROR:" << lineNumber << "] mixing ON and OFF set definition is not allowed" << endl;
                  clear();
                  return;
                }
              } else {
                outputSpecified=true;
                blifName.onSetChosen=true;
              }
            } else {
              cerr << "[ERROR:" << lineNumber << "] unknown character '" << line[line.length()-1] << "'"<< endl;
              clear();
              return;
            }
            line=line.substr(0,line.length()-1);
          }
          //if everything is right the line can be added to '.names'
          blifName.lines.push_back(line);
        }
      } while((line.substr(0,1)!=".")&&(!file.eof()));
      //check if output was already defined
      if(blifNames.find(blifName.output)!=blifNames.end()){
        cerr << "[ERROR:" << blifName.line << "] multiple definition of: "<< blifName.output << endl;
        clear();
        return;
      }
      //insert '.names'
      blifNames[blifName.output]=blifName;
      //save line number
      unsigned long temp=lineNumber;
      //set line number to the beginning of .names statement (for error messages)
      lineNumber=blifName.line;
      //check if inputs and outputs can be used
      if(createNames(blifName.output)==ParseModel::ERROR) return;
      //set old line number
      lineNumber=temp;
    } else if(isCommand(line,".end")){
      //end of model reached - outputs can be build
      buildOutputs();
      return;
    } else {
      //an unknown command was found if line is not empty
      if(!line.empty()){
        cerr << "[ERROR:" << lineNumber << "] command not known: " << line << endl;
        clear();
        return;
      }
    }
  }
  //build outputs after reading file
  buildOutputs();
}

template <class T1, class T2, class T3>
void ParseModel<T1,T2,T3>::buildOutputs(void){
  //this signals that the outputs can be build
  //(otherwise just a quick check would be done)
  createOutputs=true;

  //create inputs
  for(unsigned long i=0;i<inputVector.size();++i) gatesMap[inputVector[i]]=T2::projection(inputVector[i]);

  if(blifNames.size()==0) return;

  unsigned long resolved=0;
  unsigned long oldResolved=0;
  map<const string,BLIFNames>::iterator it;

  progressBar.size=blifNames.size();

  //try to resolve everything
  do {
    oldResolved=resolved;
    it=blifNames.begin();
    while(it!=blifNames.end()){
      progressBar.index=resolved+1;
      switch(createNames(it->first)){
        case ParseModel::ERROR: return;
        case ParseModel::UNRESOLVED:
          ++it;
          break;
        case ParseModel::SUCCESS:
        default:
          blifNames.erase(it++);
          ++resolved;
      }
    }
  } while(resolved!=oldResolved);

  //check if something could not be resolved
  if(blifNames.size()){
    cerr << "[ERROR] could not build BDD (cyclic dependencies)" << endl;
    clear();
  }
}

template <class T1, class T2, class T3>
bool ParseModel<T1,T2,T3>::declareInput(const string& input){
  //check if input was already used
  if(typeMap.find(input)==typeMap.end()){
    //create new input
    BLIFType tempType;
    tempType.line=lineNumber;
    tempType.type=BLIFType::INPUT;
    tempType.gateCount=0;
    typeMap[input]=tempType;
    T1::addVariable(input);
    inputVector.push_back(input);
  } else {
    //check type
    switch(typeMap[input].type){
      case BLIFType::INPUT:
        cerr << "[WARNING:" << lineNumber << "] input " << input << " already declared in line " << typeMap[input].line << endl;
        break;
      case BLIFType::POSSIBLE_INPUT:
        //change type
        typeMap[input].type=BLIFType::INPUT;
        T1::addVariable(input);
        inputVector.push_back(input);
        break;
      case BLIFType::OUTPUT: break;
      case BLIFType::POSSIBLE_OUTPUT:
        cerr << "[ERROR:" << typeMap[input].line << "] output " << input << " already declared as an input (line " << lineNumber << ")" << endl;
        return false;
      case BLIFType::GATE:
      default:
        cerr << "[ERROR:" << typeMap[input].line << "] gate " << input << " already declared as an input (line " << lineNumber << ")" << endl;
        return false;
    }
  }

  return true;
}

template <class T1, class T2, class T3>
bool ParseModel<T1,T2,T3>::useInput(const string& input){
  //check if input was already used
  if(typeMap.find(input)==typeMap.end()){
    //create new (possible) input type
    BLIFType tempType;
    tempType.line=lineNumber;
    tempType.type=BLIFType::POSSIBLE_INPUT;
    tempType.gateCount=1;
    typeMap[input]=tempType;
  } else {
    switch(typeMap[input].type){
      case BLIFType::POSSIBLE_OUTPUT:
        //change type
        typeMap[input].type=BLIFType::GATE;
        ++typeMap[input].gateCount;
        break;
      case BLIFType::POSSIBLE_INPUT:
      case BLIFType::GATE: ++typeMap[input].gateCount;
      case BLIFType::INPUT:
      case BLIFType::OUTPUT:
      default: break;
    }
  }

  return true;
}

template <class T1, class T2, class T3>
bool ParseModel<T1,T2,T3>::declareOutput(const string& output){
  //check if output was already used
  if(typeMap.find(output)==typeMap.end()){
    //create new output type
    BLIFType tempType;
    tempType.line=lineNumber;
    tempType.type=BLIFType::OUTPUT;
    tempType.gateCount=0;
    typeMap[output]=tempType;
    outputVector.push_back(output);
  } else {
    //check type
    switch(typeMap[output].type){
      case BLIFType::INPUT:
        cerr << "[ERROR:" << lineNumber << "] input " << output << " (line " << typeMap[output].line << ") can not be used as an output" << endl;
        return false;
      case BLIFType::POSSIBLE_INPUT:
        cerr << "[ERROR:" << typeMap[output].line << "] input " << output << " already declared as an output (line " << lineNumber << ")" << endl;
        return false;
        break;
      case BLIFType::OUTPUT:
        cerr << "[WARNING:" << lineNumber << "] output " << output << " already declared (line " << typeMap[output].line << ")" << endl;
        break;
      case BLIFType::POSSIBLE_OUTPUT:
        //change tye
        typeMap[output].type=BLIFType::OUTPUT;
        outputVector.push_back(output);
        break;
      case BLIFType::GATE:
      default: break;
    }
  }

  return true;
}

template <class T1, class T2, class T3>
bool ParseModel<T1,T2,T3>::useOutput(const string& output){
  if(typeMap.find(output)==typeMap.end()){
    BLIFType tempType;
    tempType.line=lineNumber;
    tempType.type=BLIFType::POSSIBLE_OUTPUT;
    tempType.gateCount=0;
    typeMap[output]=tempType;
  } else {
    switch(typeMap[output].type){
      case BLIFType::INPUT:
        cerr << "[ERROR:" << lineNumber << "] input " << output << " defined (line " << typeMap[output].line << ") can not be used as an output" << endl;
        return false;
      case BLIFType::POSSIBLE_INPUT:
        //change type
        typeMap[output].type=BLIFType::GATE;
        break;
      case BLIFType::GATE:
        cerr << "[ERROR:" << typeMap[output].line << "] gate " << output << " already declared as an output (line " << lineNumber << ")" << endl;
        return false;
      case BLIFType::OUTPUT:
      case BLIFType::POSSIBLE_OUTPUT:
      default: break;
    }
  }

  return true;
}

template <class T1, class T2, class T3>
unsigned short ParseModel<T1,T2,T3>::createNames(const string& output){
  BLIFNames names=blifNames[output];
  unsigned long i,j;

  //make a quick check if everything can be done
  if(!createOutputs){
    for(i=0;i<names.inputs.size();++i) if(!useInput(names.inputs[i])) return ParseModel::ERROR;
    if(!useOutput(names.output)) return ParseModel::ERROR;

    return ParseModel::UNRESOLVED;
  }

  //start building output
  T2 inputs[names.inputs.size()];

  //check if all inputs can be used
  for(i=0;i<names.inputs.size();++i){
    switch(typeMap[names.inputs[i]].type){
      case BLIFType::INPUT:
        inputs[i]=gatesMap[names.inputs[i]];
        break;
      case BLIFType::POSSIBLE_INPUT:
        if(!inputsDeclared){
          declareInput(names.inputs[i]);
          inputs[i]=gatesMap[names.inputs[i]];
        } else {
          cerr << "[ERROR:" << names.line << "] input " << names.inputs[i] << " not declared" << endl;
          clear();
          return ParseModel::ERROR;
        }
        break;
      case BLIFType::POSSIBLE_OUTPUT:
        cerr << "[ERROR:" << names.line << "] no definition for " << names.inputs[i] << endl;
        clear();
        return ParseModel::ERROR;
      case BLIFType::OUTPUT:
      case BLIFType::GATE:
      default:
        if(gatesMap.find(names.inputs[i])==gatesMap.end()) return ParseModel::UNRESOLVED;
        else inputs[i]=gatesMap[names.inputs[i]];
    }
  }

  //check if output can be used
  switch(typeMap[names.output].type){
    case BLIFType::INPUT:
      cerr << "[ERROR:" << names.line << "] input " << names.output << " can not be used as output" << endl;
      clear();
      return ParseModel::ERROR;
    case BLIFType::OUTPUT:
      //check if output was already defined
      if(gatesMap.find(names.output)!=gatesMap.end()){
        cerr << "[ERROR:" << names.line << "] multiple definitions for output " << names.output << endl;
        clear();
        return ParseModel::ERROR;
      }
      break;
    case BLIFType::POSSIBLE_OUTPUT:
        //declare output or warn that this output was never defined (and erase .names statement)
        if(!outputsDeclared) declareOutput(names.output);
        else {
          cerr << "[WARNING:" << typeMap[names.output].line << "] " << names.output << " defined but never used" << endl;

          return ParseModel::SUCCESS;
        }
      break;
    case BLIFType::GATE:
      //check if gate was already defined before
      if(gatesMap.find(names.output)!=gatesMap.end()){
        cerr << "[ERROR:" << typeMap[names.output].line << "] multiple definitions for gate " << names.output << endl;
        clear();
        return ParseModel::ERROR;
      }
      break;
    case BLIFType::POSSIBLE_INPUT: //this should not happen
    default: break;
  }

  if(names.inputs.size()==0) gatesMap[names.output]=names.onSetChosen?T2::ONE():T2::ZERO();
  else {
    progressBar.step=1;
    progressBar.maxStepSize=2*names.lines.size()-1;
    unsigned long k;
    T2 tempFunctions[names.lines.size()];
    T2 tempFunctions2[names.inputs.size()];
    for(i=0;i<names.lines.size();++i){
      printBLIFProgressBar(progressBar);
      ++progressBar.step;
      //create input assignment
      k=0;
      for(j=0;j<names.inputs.size();++j){
        if(names.lines[i][j]=='0'){
          tempFunctions2[k]=!inputs[j];
          ++k;
        } else if(names.lines[i][j]=='1'){
          tempFunctions2[k]=inputs[j];
          ++k;
        }
      }
      j=k;
      while(j>1){
        for(k=0;k<j;k+=2){
          if(k+1<j){
            tempFunctions2[k/2]=tempFunctions2[k]&tempFunctions2[k+1];
            tempFunctions2[k+1]=T2::ERROR();
            //if k/2==k it would not be a good idea to overwrite k/2
            if(k>0) tempFunctions2[k]=T2::ERROR();
          } else {
            if(k>0){
              tempFunctions2[(k/2)-1]&=tempFunctions2[k];
              tempFunctions2[k]=T2::ERROR();
            }
          }
        }
        j/=2;
      }
      if(j==0) tempFunctions[i]=T2::ONE();
      else tempFunctions[i]=names.inputs.size()?tempFunctions2[0]:T2::ONE();
      //set temp function
      if(!names.onSetChosen) tempFunctions[i]=!tempFunctions[i];
      clearBLIFProgressBar(progressBar);
    }
    j=names.lines.size();
    while(j>1){
      for(i=0;i<j;i+=2){
        printBLIFProgressBar(progressBar);
        ++progressBar.step;
        if(i+1<j){
          tempFunctions[i/2]=names.onSetChosen?tempFunctions[i]|tempFunctions[i+1]:tempFunctions[i]&tempFunctions[i+1];
          tempFunctions[i+1]=T2::ERROR();
          //if i/2==i it would not be a good idea to overwrite i/2
          if(i>0) tempFunctions[i]=T2::ERROR();
        } else {
          if(i>0){
            if(names.onSetChosen) tempFunctions[(i/2)-1]|=tempFunctions[i];
            else tempFunctions[(i/2)-1]&=tempFunctions[i];
            tempFunctions[i]=T2::ERROR();
          }
        }
        clearBLIFProgressBar(progressBar);
      }
      j/=2;
    }
    //set output
    gatesMap[names.output]=(names.lines.size())?tempFunctions[0]:T2::ZERO();
  }
  for(i=0;i<names.inputs.size();++i){
    switch(typeMap[names.inputs[i]].type){
      case BLIFType::GATE:
        //check if gate can be removed
        --typeMap[names.inputs[i]].gateCount;
        if(typeMap[names.inputs[i]].gateCount==0) gatesMap.erase(names.inputs[i]);
        break;
      case BLIFType::INPUT:
      case BLIFType::POSSIBLE_INPUT:
      case BLIFType::POSSIBLE_OUTPUT:
      case BLIFType::OUTPUT:
      default: break;
    }
  }

  return ParseModel::SUCCESS;
}

template <class T1, class T2, class T3>
bool ParseModel<T1,T2,T3>::gotoModel(void) const {
  string line;

  while(!file.eof()){
    line=readLine();
    //if mode was reached stop reading
    if(line.compare(0,string(".model").length(),".model")==0){
      line.erase(0,string(".model").length());
      line=trim(line);
      if(line==modelName) return true;
    }
  }

  return false;
}

template <class T1, class T2, class T3>
bool ParseModel<T1,T2,T3>::isCommand(string& line, const string& command) const {
  //check if line starts with given command
  if(line.compare(0,command.length(),command)==0){
    //erase command from line and trim it
    line.erase(0,command.length());
    line=trim(line);

    return true;
  }

  return false;
}

template <class T1, class T2, class T3>
string ParseModel<T1,T2,T3>::readLine(void){
  string ret="";
  string line;
  bool concatenated=false;

  //the line number is delayed if a line was concatenated (for error messages)
  //set right line number
  lineNumber+=diff;
  diff=0;
  while(!file.eof()){
    getline(file,line);
    //if concatenated do not increase the line number
    if(concatenated) ++diff;
    else ++lineNumber;
    //is line a comment?
    if(line.find("#")!=string::npos) line=line.substr(0,line.find("#"));
    //do not trim a concatenated string on the left side
    line=concatenated?trimRight(line):trim(line);

    if(line.length()>0){
      // concatenate lines if line finishes with "\"
      if(line.at(line.length()-1)=='\\'){
        ret+=line.substr(0,line.length()-1);
        concatenated=true;
      } else return ret+line;
    } else if(concatenated) return ret;
  }

  return ret;
}

//**************************BLIF Input***************************
template <class T1, class T2, class T3>
class BLIFInput{
public:
  static T2 getOutput(const string&, const string&);
  static T3 getAllOutputs(const string&);
};

template <class T1, class T2, class T3>
T2 BLIFInput<T1,T2,T3>::getOutput(const string& filename, const string& outputname){
  ParseModel<T1,T2,T3> model(filename);
  T2 ret=model.getOutput(outputname);
  model.clear();
  T1::clear();

  return ret;
}

template <class T1, class T2, class T3>
T3 BLIFInput<T1,T2,T3>::getAllOutputs(const string& filename){
  ParseModel<T1,T2,T3> model(filename);
  T3 ret;
  bool error=false;

  for(unsigned long i=0;(i<model.size())&&(!error);++i){
    ret.push(model.getOutput(i));
    error|=(ret[i]==T2::ERROR());
  }
  //return empty function container if something went wrong
  if(error) ret.clear();
  model.clear();
  T1::clear();

  return ret;
}

#endif
