/***************************************************************************
                          hashuniquetable.h  -  description
                             -------------------
    begin                : Mon Nov 28 2005
    copyright            : (C) 2005-2008 by Joern Ossowski
    email                : mail@jossowski.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef HASHUNIQUETABLE_H
#define HASHUNIQUETABLE_H

#include <common/helper.h>

#include <vector>
#include <boost/dynamic_bitset.hpp>

//******************HashUniqueTable********************
template <class T, class B>
class HashUniqueTable {
public:
  HashUniqueTable(const unsigned long);
  unsigned long size(void) const {return count;}
  inline void resetDeadNodeCount(void){deadNodes=0;}
  inline void incDeadNodeCount(void){++deadNodes;}
  inline void decDeadNodeCount(void){--deadNodes;}
  inline unsigned long getDeadNodes(void) const {return deadNodes;}
  inline void add(T*);
  inline bool remove(const T*);
  inline T* findOrAdd(B*,B*);
  T* getFirst(void){
    if(count==0) return 0;

    currentPos=bitSet.find_first();
    return uniqueTable[currentPos];
  }
  T* next(void){
    currentPos=bitSet.find_next(currentPos);
    if(currentPos==boost::dynamic_bitset<>::npos) return 0;
    return uniqueTable[currentPos];
  }
  void setLevel(const unsigned long newLevel){level=newLevel;}
  inline unsigned long getLevel(void) const {return level-skip;}
  inline unsigned long getRealLevel(void) const {return level;}
  inline bool deleteable(void) const {return (count==0);}
  inline static void setSkip(unsigned long newSkip){skip=newSkip;}
  inline static void setInitHashSize(unsigned int newInit){initHashSize=newInit;}
  inline static void setMaxHashSize(unsigned int newMax){maximumHashSize=newMax;}
  inline static unsigned long getSkip(void){return skip;}
  float get_load_factor() const;

public:
  static float max_load_factor;
private:
  static unsigned int initHashSize;
  static unsigned int maximumHashSize;
  static unsigned long skip;
  unsigned long level;
  unsigned long count;
  unsigned long maxCollisionListSize;
  unsigned int hashSize;
  unsigned int maxHashSize;
  unsigned short shift;
  unsigned short collisionListSizeLimit;
  unsigned long deadNodes;

  boost::dynamic_bitset<> bitSet;
  boost::dynamic_bitset<>::size_type currentPos;

  std::vector<T*> uniqueTable;

  inline void check_for_resize(unsigned long collisions);
  void resize(void);
};
template <class T, class B> unsigned long HashUniqueTable<T,B>::skip=0;
template <class T, class B> unsigned int HashUniqueTable<T,B>::initHashSize=256;
//template <class T, class B> unsigned int HashUniqueTable<T,B>::initHashSize=131072;
template <class T, class B> unsigned int HashUniqueTable<T,B>::maximumHashSize= 1<<30;
template <class T, class B> float HashUniqueTable<T,B>::max_load_factor=1.0f;

template <class T, class B>
HashUniqueTable<T,B>::HashUniqueTable(const unsigned long newLevel) : level(newLevel), count(0), maxCollisionListSize(0), deadNodes(0) {
  collisionListSizeLimit=12;

  hashSize=2;
  //shift for smallest hashmap
  shift=sizeof(unsigned long)*8-1;
  while((hashSize<<1)<=initHashSize){
    hashSize<<=1;
    --shift;
  }
  maxHashSize=maximumHashSize;

  bitSet.resize(hashSize,false);
  uniqueTable.resize(hashSize,0);

  currentPos=boost::dynamic_bitset<>::npos;
}

template <class T, class B>
void HashUniqueTable<T,B>::add(T* result){
  T* node1=static_cast<T*>(result->getSucc0());
  T* node2=static_cast<T*>(result->getSucc1());
  const unsigned long index=Common::hashFunction<T*>(node1,node2)>>shift;
  unsigned long collisionListSize=0;

  T* tempNode=uniqueTable[index];
  T* lastNode=0;
  while(tempNode &&tempNode->getSucc0()<node1){
    lastNode=tempNode;
    tempNode=tempNode->getNext();
    ++collisionListSize;
  }
  while(tempNode&&(tempNode->getSucc0()==node1)&&(tempNode->getSucc1()<node2)){
    lastNode=tempNode;
    tempNode=tempNode->getNext();
    ++collisionListSize;
  }
  //no check if entry is already there
  if(lastNode==0){
    result->setNext(uniqueTable[index]);
    uniqueTable[index]=result;
    bitSet[index]=true;
  } else {
    result->setNext(lastNode->getNext());
    lastNode->setNext(result);
  }
  ++count;

  check_for_resize(collisionListSize);
}

template <class T, class B>
bool HashUniqueTable<T,B>::remove(const T* value){
  T* node1=static_cast<T*>(value->getSucc0());
  T* node2=static_cast<T*>(value->getSucc1());
  const unsigned long index=Common::hashFunction<T*>(node1,node2)>>shift;

  T* tempNode=uniqueTable[index];
  T* lastNode=0;
  while(tempNode && tempNode!=value){
    lastNode=tempNode;
    tempNode=tempNode->getNext();
  }
  if(tempNode!=value) return false;

  if(lastNode==0){
    //change entry in the unique table and check if the value is 0
    if((uniqueTable[index]=tempNode->getNext())==0) bitSet[index]=false;
  } else lastNode->setNext(tempNode->getNext());

  tempNode->setNext(0);
  --count;
  return true;
}

template <class T, class B>
T* HashUniqueTable<T,B>::findOrAdd(B* node1, B* node2){
  const unsigned long index=Common::hashFunction<B*>(node1,node2)>>shift;

  unsigned long collisionListSize=0;
  T* tempNode=uniqueTable[index];
  T* lastNode=0;
  while(tempNode && tempNode->getSucc0()<node1){
    ++collisionListSize;
    lastNode=tempNode;
    tempNode=tempNode->getNext();
  }
  while(tempNode&&(tempNode->getSucc0()==node1)&&(tempNode->getSucc1()<node2)){
    ++collisionListSize;
    lastNode=tempNode;
    tempNode=tempNode->getNext();
  }
  if(tempNode&&(tempNode->getSucc0()==node1)&&(tempNode->getSucc1()==node2)) return tempNode;
  T* result=T::getNewNode();
  result->initNode(this,node1,node2);
  if(lastNode==0){
    result->setNext(uniqueTable[index]);
    uniqueTable[index]=result;
    bitSet[index]=true;
  } else {
    result->setNext(lastNode->getNext());
    lastNode->setNext(result);
  }
  ++count;

  check_for_resize(collisionListSize);

  return result;
}

template <class T, class B> 
float HashUniqueTable<T,B>::get_load_factor() const {
  return ((float)count) / hashSize;
}

template <class T, class B>
void HashUniqueTable<T,B>::check_for_resize(unsigned long collisions) {
  if(collisions>maxCollisionListSize){
    maxCollisionListSize=collisions;
  }

  if (UNLIKELY(get_load_factor() >= max_load_factor) && 
      hashSize<maxHashSize) {
    resize();
  }
}

template <class T, class B>
void HashUniqueTable<T,B>::resize(void){
  std::vector<T*> tempUniqueTable(2*hashSize,0);
  boost::dynamic_bitset<> tempBitSet(2*hashSize,false);

  --shift;
  maxCollisionListSize=0;

  //transfer all nodes from old unique table to new one
  T* currentTable=getFirst();
  T* currentNode;
  T* tempNode;
  T* evenTable;
  T* oddTable;
  unsigned long index;
  unsigned long evenCollisionListSize;
  unsigned long oddCollisionListSize;
  while(currentTable){
    currentNode=currentTable;
    tempNode=currentNode->getNext();
    //shift is already updated so that the shift has to be reverted
    index=Common::hashFunction<B*>(currentNode->getSucc0(),currentNode->getSucc1())>>(shift+1);

    evenTable=0;
    oddTable=0;

    evenCollisionListSize=0;
    oddCollisionListSize=0;
    while(currentNode){
      tempNode=currentNode->getNext();
      //shift is already updated
      if((Common::hashFunction<B*>(currentNode->getSucc0(),currentNode->getSucc1())>>shift)&1){
        if(oddTable!=0) oddTable->setNext(currentNode);
        else tempUniqueTable[2*index+1]=currentNode;
        oddTable=currentNode;
        ++oddCollisionListSize;
      } else {
        if(evenTable!=0) evenTable->setNext(currentNode);
        else tempUniqueTable[2*index]=currentNode;
        evenTable=currentNode;
        ++evenCollisionListSize;
      }

      //--------------------------------------
      currentNode=tempNode;
    }

    if(oddTable!=0){
      oddTable->setNext(0);
      tempBitSet[2*index+1]=true;
    } else tempUniqueTable[2*index+1]=0;
    if(evenTable!=0){
      evenTable->setNext(0);
      tempBitSet[2*index]=true;
    } else tempUniqueTable[2*index]=0;

    if(oddCollisionListSize>maxCollisionListSize) maxCollisionListSize=oddCollisionListSize;
    if(evenCollisionListSize>maxCollisionListSize) maxCollisionListSize=evenCollisionListSize;

    currentTable=next();
  }

  uniqueTable.swap(tempUniqueTable);
  bitSet.swap(tempBitSet);

  hashSize*=2;
  ++collisionListSizeLimit;
  currentPos=boost::dynamic_bitset<>::npos;
}

//******************HashUniqueTable2********************
template <class T1, class T2>
class HashUniqueTable2 {
public:
  HashUniqueTable2(const unsigned long=0);
  ~HashUniqueTable2(void);
  unsigned long size(void) const {return count;}
  inline void resetDeadNodeCount(void){deadNodes=0;}
  inline void incDeadNodeCount(void){++deadNodes;}
  inline void decDeadNodeCount(void){--deadNodes;}
  inline unsigned long getDeadNodes(void){return deadNodes;}
  inline void add(T1*);
  inline bool remove(const T1*);
  inline T1* get(const T2&, const T2&) const;
  inline T1* findOrAdd(const T2&, const T2&);
  T1* getFirst(void){
    if(count==0) return 0;

    boost::dynamic_bitset<>::size_type pos=bitSet->find_first();
    currentSlot=uniqueTable+pos;
    return *currentSlot;
  }
  T1* next(void){
    //calculate position of next slot
    unsigned long index=(currentSlot-uniqueTable);

    boost::dynamic_bitset<>::size_type pos=bitSet->find_next(index);
    if(pos==boost::dynamic_bitset<>::npos) return 0;
    currentSlot=uniqueTable+pos;
    return *currentSlot;
  }
  void setLevel(const unsigned long newLevel){level=newLevel;}
  inline unsigned long getLevel(void) const {return level-skip;}
  inline unsigned long getRealLevel(void) const {return level;}
  inline bool deleteable(void) const {return (count==0);}
  inline unsigned long hashFunction(const T2&, const T2&) const;
  inline static void setSkip(unsigned long newSkip){skip=newSkip;}
  inline static void setInitHashSize(unsigned int newInit){initHashSize=newInit;}
  inline static void setMaxHashSize(unsigned int newMax){maximumHashSize=newMax;}
  inline static unsigned long getSkip(void){return skip;}
private:
  static unsigned int initHashSize;
  static unsigned int maximumHashSize;
  static unsigned long skip;
  unsigned long level;
  unsigned long count;
  unsigned long maxCollisionListSize;
  unsigned int hashSize;
  unsigned int maxHashSize;
  unsigned short shift;
  unsigned short collisionListSizeLimit;
  unsigned long deadNodes;

  boost::dynamic_bitset<>* bitSet;
  T1** uniqueTable;
  T1** EOT;
  T1** currentSlot;

  void resize(void);
};

template <class T1, class T2>
unsigned long HashUniqueTable2<T1,T2>::skip=0;

template <class T1, class T2>
unsigned int HashUniqueTable2<T1,T2>::initHashSize=131072;

template <class T1, class T2>
unsigned int HashUniqueTable2<T1,T2>::maximumHashSize=262144;

template <class T1, class T2>
HashUniqueTable2<T1,T2>::HashUniqueTable2(const unsigned long newLevel){
  level=newLevel;
  count=0;
  maxCollisionListSize=0;
  collisionListSizeLimit=12;

  hashSize=2;
  //shift for smallest hashmap
  shift=sizeof(unsigned long)*8-1;
  while((hashSize<<1)<=initHashSize){
    hashSize<<=1;
    --shift;
  }
  maxHashSize=maximumHashSize;
  bitSet=new boost::dynamic_bitset<>(hashSize);

  uniqueTable=new T1*[hashSize];
  memset(uniqueTable,0,hashSize*sizeof(T1*));
  EOT=uniqueTable+hashSize;
  currentSlot=EOT;
  deadNodes=0;
}

template <class T1, class T2>
HashUniqueTable2<T1,T2>::~HashUniqueTable2(void){
  delete[] uniqueTable;
  delete bitSet;
}

template <class T1, class T2>
void HashUniqueTable2<T1,T2>::add(T1* result){
  T2 nodeInfo1=result->getSucc0NodeInfo();
  T2 nodeInfo2=result->getSucc1NodeInfo();

  unsigned long index=hashFunction(nodeInfo1,nodeInfo2);
  unsigned long collisionListSize=0;

  T1* tempNode=uniqueTable[index];
  T1* lastNode=0;
  if(tempNode!=0){
    while(tempNode->getSucc0NodeInfo()<nodeInfo1){
      lastNode=tempNode;
      tempNode=tempNode->getNext();
      ++collisionListSize;
      if(tempNode==0) break;
    }
    if(tempNode!=0){
      while((tempNode->getSucc0NodeInfo()==nodeInfo1)&&(tempNode->getSucc1NodeInfo()<nodeInfo2)){
        lastNode=tempNode;
        tempNode=tempNode->getNext();
        ++collisionListSize;
        if(tempNode==0) break;
      }
    }
  }
  //no check if entry is already there
  if(lastNode==0){
    result->setNext(uniqueTable[index]);
    uniqueTable[index]=(T1*) result;
    (*bitSet)[index]=true;
  } else {
    result->setNext(lastNode->getNext());
    lastNode->setNext(result);
  }
  ++count;
  if(collisionListSize>maxCollisionListSize){
    maxCollisionListSize=collisionListSize;
    if(maxCollisionListSize>collisionListSizeLimit){
      if((count>=(hashSize*collisionListSizeLimit)/4)&&(hashSize<maxHashSize)) resize();
    }
  }
}

template <class T1, class T2>
bool HashUniqueTable2<T1,T2>::remove(const T1* value){
  T2 nodeInfo1=value->getSucc0NodeInfo();
  T2 nodeInfo2=value->getSucc1NodeInfo();
  unsigned long index=hashFunction(nodeInfo1,nodeInfo2);

  T1* tempNode=uniqueTable[index];
  T1* lastNode=0;
  if(tempNode!=0){
    while(tempNode!=value){
      lastNode=tempNode;
      tempNode=tempNode->getNext();
      if(tempNode==0) break;
    }
  }
  if(tempNode!=value) return false;
  if(lastNode==0){
    //change entry in the unique table and check if the value is 0
    if((uniqueTable[index]=tempNode->getNext())==0) (*bitSet)[index]=false;
  } else lastNode->setNext(tempNode->getNext());

  tempNode->setNext(0);
  --count;
  return true;
}

template <class T1, class T2>
T1* HashUniqueTable2<T1,T2>::get(const T2& nodeInfo1, const T2& nodeInfo2) const {
  unsigned long index=hashFunction(nodeInfo1,nodeInfo2);

  T1* tempNode=uniqueTable[index];
  T1* lastNode=0;
  if(tempNode!=0){
    while(tempNode->getSucc0NodeInfo()<nodeInfo1){
      lastNode=tempNode;
      tempNode=tempNode->getNext();
      if(tempNode==0) break;
    }
    if(tempNode!=0){
      while((tempNode->getSucc0NodeInfo()==nodeInfo1)&&(tempNode->getSucc1NodeInfo()<nodeInfo2)){
        lastNode=tempNode;
        tempNode=tempNode->getNext();
        if(tempNode==0) break;
      }
    }
  }
  if(tempNode!=0){
    if((tempNode->getSucc0NodeInfo()==nodeInfo1)&&(tempNode->getSucc1NodeInfo()==nodeInfo2)) return tempNode;
  }

  return 0;
}

template <class T1, class T2>
T1* HashUniqueTable2<T1,T2>::findOrAdd(const T2& nodeInfo1, const T2& nodeInfo2){
  unsigned long index=hashFunction(nodeInfo1,nodeInfo2);

  unsigned long collisionListSize=0;
  T1* tempNode=uniqueTable[index];
  T1* lastNode=0;

  if(tempNode!=0){
    while(tempNode->getSucc0NodeInfo()<nodeInfo1){
      ++collisionListSize;
      lastNode=tempNode;
      tempNode=tempNode->getNext();
      if(tempNode==0) break;
    }
    if(tempNode!=0){
      while((tempNode->getSucc0NodeInfo()==nodeInfo1)&&(tempNode->getSucc1NodeInfo()<nodeInfo2)){
        ++collisionListSize;
        lastNode=tempNode;
        tempNode=tempNode->getNext();
        if(tempNode==0) break;
      }
    }
  }
  if(tempNode!=0){
    if((tempNode->getSucc0NodeInfo()==nodeInfo1)&&(tempNode->getSucc1NodeInfo()==nodeInfo2)) return tempNode;
  }

  T1* result=T1::getNewNode();

  result->initNode(this,nodeInfo1,nodeInfo2);
  if(lastNode==0){
    result->setNext(uniqueTable[index]);
    uniqueTable[index]=(T1*) result;
    (*bitSet)[index]=true;
  } else {
    result->setNext(lastNode->getNext());
    lastNode->setNext(result);
  }
  ++count;
  if(collisionListSize>maxCollisionListSize){
    maxCollisionListSize=collisionListSize;
    if(maxCollisionListSize>collisionListSizeLimit){
      if((count>=(hashSize*collisionListSizeLimit)/4)&&(hashSize<maxHashSize)) resize();
    }
  }

  return result;
}

template <class T1, class T2>
unsigned long HashUniqueTable2<T1,T2>::hashFunction(const T2& nodeInfo1, const T2& nodeInfo2) const {
  return ((((unsigned long) nodeInfo1.getNode()*12582917+(unsigned long) nodeInfo2.getNode())*4256249)>>shift);
}

template <class T1, class T2>
void HashUniqueTable2<T1,T2>::resize(void){
  T1** tempHashUniqueTable2=new T1*[2*hashSize];
  //out of memory
  if(tempHashUniqueTable2==0) return;
  boost::dynamic_bitset<>* tempBitSet=new boost::dynamic_bitset<>(2*hashSize);
  //out of memory
  if(tempBitSet==0){
    delete[] tempHashUniqueTable2;
    return;
  }

  --shift;
  memset(tempHashUniqueTable2,0,2*hashSize*sizeof(T1*));
  maxCollisionListSize=0;

  //transfer all nodes from old unique table to new one
  T1* currentTable=(T1*) getFirst();
  T1* currentNode;
  T1* tempNode;
  T1* evenTable;
  T1* oddTable;
  unsigned long index;
  unsigned long evenCollisionListSize;
  unsigned long oddCollisionListSize;
  while(currentTable){
    currentNode=currentTable;
    tempNode=currentNode->getNext();
    //shift is already updated so that the shift has to be reverted
    index=hashFunction(currentNode->getSucc0NodeInfo(),currentNode->getSucc1NodeInfo())>>1;

    evenTable=0;
    oddTable=0;

    evenCollisionListSize=0;
    oddCollisionListSize=0;
    while(currentNode){
      tempNode=currentNode->getNext();
      //shift is already updated
      if(hashFunction(currentNode->getSucc0NodeInfo(),currentNode->getSucc1NodeInfo())&1){
        if(oddTable!=0) oddTable->setNext(currentNode);
        else tempHashUniqueTable2[2*index+1]=currentNode;
        oddTable=currentNode;
        ++oddCollisionListSize;
      } else {
        if(evenTable!=0) evenTable->setNext(currentNode);
        else tempHashUniqueTable2[2*index]=currentNode;
        evenTable=currentNode;
        ++evenCollisionListSize;
      }

      //--------------------------------------
      currentNode=tempNode;
    }

    if(oddTable!=0){
      oddTable->setNext(0);
      (*tempBitSet)[2*index+1]=true;
    } else tempHashUniqueTable2[2*index+1]=0;
    if(evenTable!=0){
      evenTable->setNext(0);
      (*tempBitSet)[2*index]=true;
    } else tempHashUniqueTable2[2*index]=0;

    if(oddCollisionListSize>maxCollisionListSize) maxCollisionListSize=oddCollisionListSize;
    if(evenCollisionListSize>maxCollisionListSize) maxCollisionListSize=evenCollisionListSize;

    currentTable=(T1*) next();
  }

  //updating informations
  delete[] uniqueTable;
  delete bitSet;

  uniqueTable=tempHashUniqueTable2;
  bitSet=tempBitSet;

  hashSize*=2;
  ++collisionListSizeLimit;
  EOT=uniqueTable+hashSize;
  currentSlot=EOT;
}

#endif
