/***************************************************************************
                          zaddfunction.cpp  -  description
                             -------------------
    begin                : Thu Apr 15 2004
    copyright            : (C) 2004-2008 by Joern Ossowski
    email                : mail@jossowski.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <zadd/zaddfunction.h>
#include <zadd/zaddbasenode.h>
#include <zadd/zaddinnernode.h>
#include <zadd/zaddterminalnode.h>
#include <zadd/zaddvarorder.h>
#include <zadd/zadd.h>
#include <common/assignment.h>
#include <common/stringhelper.h>
#include <common/helper.h>
#include <common/ptrhashmap.h>

#include <stack>
#include <set>
#include <iostream>

inline ZADDTerminalNode* findOrAddDrain(const double value){
  if UNLIKELY(ABS(value)<0.0000001) return ZADD::zeroDrain;

  return ZADD::drainUniqueTable.findOrAddValue(value);
}
inline ZADDInnerNode* findOrAdd(ZADDUniqueTable* uTable, ZADDBaseNode* succ0, ZADDBaseNode* succ1){
  return uTable->findOrAdd(succ0,succ1);
}

inline unsigned long pow2(const unsigned long n){
  return (1 << n);
}

//boolean constants
#define AND_OP              0x01
#define OR_OP               0x02
#define DIF_OP              0x03
#define ITE_OP              0x02
#define EXI_OP              0x04
#define COF_OP              0x06
#define CHA_OP              0x08
//artihmetic constants
#define ADD_OP              0x02
#define SUB_OP              0x04
#define MUL_OP              0x06
#define OVE_OP              0x08
//move constants
#define MUP_OP              0x02
#define MDO_OP              0x04
//compare constants
#define LES_OP              0x02
#define LEE_OP              0x04
#define EQU_OP              0x06
//abstract constants
#define SUM_OP              0x02
#define PRO_OP              0x04
#define MAX_OP              0x06
#define MIN_OP              0x08
//matrix constants
#define MAV_OP              0x02
#define VMA_OP              0x04
#define MMA_OP              0x06

inline ZADDBaseNode* calcPtr(const ZADDBaseNode* node, unsigned long op){
  return reinterpret_cast<ZADDBaseNode*>(reinterpret_cast<unsigned long>(node)+op);
}

inline void swapPointer(const ZADDBaseNode*& u, const ZADDBaseNode*& v){
  if(u>v){
    std::swap(u,v);
  }
}

inline void swapPointer(const ZADDBaseNode*& u, const ZADDBaseNode*& uVS, const ZADDBaseNode*& v, const ZADDBaseNode*& vVS){
  if(u>v){
    std::swap(u,v);
    std::swap(uVS,vVS);
  }
}

inline void swapPointer(const ZADDBaseNode*& u, ZADDBaseNode*& uVS, const ZADDBaseNode*& v, ZADDBaseNode*& vVS){
  if(u>v){
    std::swap(u,v);
    std::swap(uVS,vVS);
  }
}

inline void addFunction(ZADDBaseNode* node, ZADDBaseNode* nodeVarSet){
  node->incNodeCount();
  nodeVarSet->incNodeCount();
}

inline void removeFunction(ZADDBaseNode* node, ZADDBaseNode* nodeVarSet){
  node->deleteNode();
  nodeVarSet->deleteNode();
}

//************************ZADDFunction************************
ZADDFunction::ZADDFunction(const ZADDFunction& function) : rootNode(function.rootNode), varSet(function.varSet){
  addFunction(rootNode,varSet);
}

ZADDFunction::ZADDFunction(const Assignment& assignment, const double value) : varSet(ZADD::oneDrain){
  if UNLIKELY(assignment.size()==0){
    rootNode=ZADD::zeroDrain;
  } else {
    rootNode=findOrAddDrain(value);
    if(rootNode!=ZADD::zeroDrain){
      for(unsigned long i=0;i<assignment.size();++i) ZADDVarOrder::addVariable(assignment[i].variable);

      for(unsigned long i=ZADD::variableOrdering.size();i>0;--i){
	std::string var=ZADDVarOrder::getVariableName (i-1);
        if(assignment.find(var)){
	  if(assignment.at(var)) rootNode=findOrAdd(ZADD::variableOrdering.getData(i-1),ZADD::zeroDrain,rootNode);
	  varSet=findOrAdd(ZADD::variableOrdering.getData(i-1),ZADD::zeroDrain,varSet);
        }
      }
    }
  }

  addFunction(rootNode,varSet);
}

ZADDFunction::ZADDFunction(char* buffer, const double value) : rootNode(findOrAddDrain(value)), varSet(ZADD::oneDrain){
  if(rootNode!=ZADD::zeroDrain){
    for(unsigned long i=ZADD::variableOrdering.size();i>0;--i){
      switch(buffer[i-1]){
        case '0': varSet=findOrAdd(ZADD::variableOrdering.getData(i-1),ZADD::zeroDrain,varSet);
                  break;
        case '1': rootNode=findOrAdd(ZADD::variableOrdering.getData(i-1),ZADD::zeroDrain,rootNode);
	          varSet=findOrAdd(ZADD::variableOrdering.getData(i-1),ZADD::zeroDrain,varSet);
                  break;
        default:  break;
      }
    }
  }

  addFunction(rootNode,varSet);
}

ZADDFunction::ZADDFunction(const std::string& variable){
  unsigned long index=ZADD::variableOrdering[variable];
  if(index==ULONG_MAX){
    if UNLIKELY(!ZADD::variableOrdering.appendVariable(variable)) throw "Could not append variable";
    index=ZADD::variableOrdering.size()-1;
  }

  varSet=rootNode=findOrAdd(ZADD::variableOrdering.getData(index),ZADD::zeroDrain,ZADD::oneDrain);
  addFunction(rootNode,varSet);
}

ZADDFunction::ZADDFunction(const double value) : rootNode(findOrAddDrain(value)), varSet(ZADD::oneDrain){
  addFunction(rootNode,varSet);
}

ZADDFunction::ZADDFunction(const ZADDBaseNode* node, const ZADDBaseNode* nodeVarSet) : rootNode(const_cast<ZADDBaseNode*>(node)), varSet(const_cast<ZADDBaseNode*>(nodeVarSet)){
  addFunction(rootNode,varSet);
}

ZADDFunction::ZADDFunction() : rootNode(ZADD::zeroDrain), varSet(ZADD::oneDrain) {
  addFunction(rootNode,varSet);
}

ZADDFunction::~ZADDFunction(void){removeFunction(rootNode,varSet);}

ZADDFunction ZADDFunction::ONE(void){return ZADDFunction(ZADD::oneDrain,ZADD::oneDrain);}

ZADDFunction ZADDFunction::ZERO(void){return ZADDFunction(ZADD::zeroDrain,ZADD::oneDrain);}

ZADDFunction ZADDFunction::projection(const std::string& variable){return ZADDFunction(variable);}

ZADDFunction ZADDFunction::projection(const unsigned long index){
  if(index>=ZADD::variableOrdering.size()) throw "Index out of range";
  else {
    ZADDBaseNode* ret=findOrAdd(ZADD::variableOrdering.getData(index),ZADD::zeroDrain,ZADD::oneDrain);
    return ZADDFunction(ret,ret);
  }
}

ZADDBaseNode* calcUnion(const ZADDBaseNode* node1, const ZADDBaseNode* node2){
  if(node1==node2) return const_cast<ZADDBaseNode*>(node1);
  if(node1->isDrain()) return const_cast<ZADDBaseNode*>(node2);
  if(node2->isDrain()) return const_cast<ZADDBaseNode*>(node1);

  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(node1,AND_OP),node1,node2,node2)) return comp;

  ZADDUniqueTable* uniqueTable;
  ZADDBaseNode* n1;
  ZADDBaseNode* n2;
  if(node1->getRealLevel()<=node2->getRealLevel()){
    uniqueTable=node1->getUniqueTable();
    if(node2->getUniqueTable()==uniqueTable){
      n1=static_cast<const ZADDInnerNode*>(node1)->getSucc1();
      n2=static_cast<const ZADDInnerNode*>(node2)->getSucc1();
    } else {
      n1=static_cast<const ZADDInnerNode*>(node1)->getSucc1();
      n2=const_cast<ZADDBaseNode*>(node2);
    }
  } else {
    uniqueTable=node2->getUniqueTable();
    n1=const_cast<ZADDBaseNode*>(node1);
    n2=static_cast<const ZADDInnerNode*>(node2)->getSucc1();
  }

  ZADDBaseNode* node=findOrAdd(uniqueTable,ZADD::zeroDrain,calcUnion(n1,n2));

  ZADD::BoolComputedTable.insert(calcPtr(node1,AND_OP),node1,node2,node2,node);

  return node;
}

ZADDBaseNode* calcSetMinus(const ZADDBaseNode* node1, ZADDBaseNode* node2){
  if(node1==node2) return ZADD::oneDrain;
  if(node1->isDrain()) return ZADD::oneDrain;
  if(node2->isDrain()) return const_cast<ZADDBaseNode*>(node1);

  ZADDBaseNode* newNode2=node2;
  while(node1->getRealLevel()>newNode2->getRealLevel()) newNode2=static_cast<ZADDInnerNode*>(newNode2)->getSucc1();

  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(node1,EXI_OP),node1,newNode2,newNode2)) return comp;

  ZADDBaseNode* node;
  if(node1->getRealLevel()==newNode2->getRealLevel()) node=calcSetMinus(static_cast<const ZADDInnerNode*>(node1)->getSucc1(),static_cast<ZADDInnerNode*>(newNode2)->getSucc1());
  else node=findOrAdd(node1->getUniqueTable(),ZADD::zeroDrain,calcSetMinus(static_cast<const ZADDInnerNode*>(node1)->getSucc1(),newNode2));

  ZADD::BoolComputedTable.insert(calcPtr(node1,EXI_OP),node1,node2,node2,node);

  return node;
}

ZADDBaseNode* calcAllVars(void){
  ZADDBaseNode* tempVar=ZADD::oneDrain;
  for(unsigned long i=ZADD::variableOrdering.size();i>0;--i) tempVar=findOrAdd(ZADD::variableOrdering.getData(i-1),ZADD::zeroDrain,tempVar);

  return tempVar;
}

ZADDBaseNode* calcLiftedBDD(const ZADDBaseNode* function, const ZADDBaseNode* functionVS, const ZADDBaseNode* varSet){
  if(varSet==ZADD::oneDrain) return const_cast<ZADDBaseNode*>(function);
  if(function==ZADD::zeroDrain) return ZADD::zeroDrain;

  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(function,OR_OP),functionVS,ZADD::zeroDrain,varSet)) return comp;

  ZADDBaseNode* node;
  if(varSet->getRealLevel()<=functionVS->getRealLevel()){
    if(varSet->getUniqueTable()==functionVS->getUniqueTable()){
      if(function->getUniqueTable()==functionVS->getUniqueTable()){
        ZADDBaseNode* w0=calcLiftedBDD(static_cast<const ZADDInnerNode*>(function)->getSucc0(),static_cast<const ZADDInnerNode*>(functionVS)->getSucc1(),static_cast<const ZADDInnerNode*>(varSet)->getSucc1());
        ZADDBaseNode* w1=calcLiftedBDD(static_cast<const ZADDInnerNode*>(function)->getSucc1(),static_cast<const ZADDInnerNode*>(functionVS)->getSucc1(),static_cast<const ZADDInnerNode*>(varSet)->getSucc1());

        node=(w1==ZADD::zeroDrain)?w0:findOrAdd(function->getUniqueTable(),w0,w1);
      } else {
        node=calcLiftedBDD(function,static_cast<const ZADDInnerNode*>(functionVS)->getSucc1(),static_cast<const ZADDInnerNode*>(varSet)->getSucc1());
      }
    } else {
      node=calcLiftedBDD(function,functionVS,static_cast<const ZADDInnerNode*>(varSet)->getSucc1());
      node=findOrAdd(varSet->getUniqueTable(),node,node);
    }
  } else {
    if(function->getUniqueTable()==functionVS->getUniqueTable()){
      ZADDBaseNode* w0=calcLiftedBDD(static_cast<const ZADDInnerNode*>(function)->getSucc0(),static_cast<const ZADDInnerNode*>(functionVS)->getSucc1(),varSet);
      ZADDBaseNode* w1=calcLiftedBDD(static_cast<const ZADDInnerNode*>(function)->getSucc1(),static_cast<const ZADDInnerNode*>(functionVS)->getSucc1(),varSet);

      node=(w1==ZADD::zeroDrain)?w0:findOrAdd(function->getUniqueTable(),w0,w1);
    } else {
      node=calcLiftedBDD(function,static_cast<const ZADDInnerNode*>(functionVS)->getSucc1(),varSet);
    }
  }

  ZADD::BoolComputedTable.insert(calcPtr(function,OR_OP),functionVS,ZADD::zeroDrain,varSet,node);

  return node;
}

unsigned long ZADDFunction::size(void) const {
  std::set<const ZADDBaseNode*> visited;
  std::stack<const ZADDBaseNode*> nodeStack;

  visited.insert(rootNode);
  if(rootNode!=varSet) visited.insert(varSet);
  nodeStack.push(rootNode);
  if(rootNode!=varSet) nodeStack.push(varSet);
  while(nodeStack.size()>0){
    const ZADDBaseNode* node=nodeStack.top();
    nodeStack.pop();
    if LIKELY(!node->isDrain()){
      {
        ZADDBaseNode* tempNode=static_cast<const ZADDInnerNode*>(node)->getSucc0();
        if(visited.find(tempNode)==visited.end()){
          visited.insert(tempNode);
          nodeStack.push(tempNode);
        }
      }
      {
        ZADDBaseNode* tempNode=static_cast<const ZADDInnerNode*>(node)->getSucc1();
        if(visited.find(tempNode)==visited.end()){
          visited.insert(tempNode);
          nodeStack.push(tempNode);
        }
      }
    }
  }

  return visited.size();
}

ZADDFunction& ZADDFunction::operator=(const ZADDFunction& function){
  addFunction(function.rootNode,function.varSet);
  removeFunction(rootNode,varSet);
  rootNode=function.rootNode;
  varSet=function.varSet;

  return *this;
}

struct abstract{
  const short OPERATOR_;
  PtrHashMap2<ZADDBaseNode>& table_;
  ZADDBaseNode* (*function_)(const ZADDBaseNode*,const ZADDBaseNode*);

  abstract(const short op, PtrHashMap2<ZADDBaseNode>& table, ZADDBaseNode* (*function)(const ZADDBaseNode*,const ZADDBaseNode*)): OPERATOR_(op),table_(table), function_(function){}
  ZADDBaseNode* calc(const ZADDBaseNode* node, const ZADDBaseNode* varSet){
    if(varSet->isDrain()) return const_cast<ZADDBaseNode*>(node);
    if(ZADDBaseNode* comp=table_.find(calcPtr(node,OPERATOR_),varSet)) return comp;

    if(varSet->getUniqueTable()==node->getUniqueTable()){
      ZADDBaseNode* newVarSet=static_cast<const ZADDInnerNode*>(varSet)->getSucc1();
      ZADDBaseNode* w0=calc(static_cast<const ZADDInnerNode*>(node)->getSucc0(),newVarSet);
      ZADDBaseNode* w1=calc(static_cast<const ZADDInnerNode*>(node)->getSucc1(),newVarSet);
      ZADDBaseNode* result=(*function_)(w0,w1);
      if(result!=w0 && w0->preNodeCount==0) w0->garbageCollect();
      if(result!=w1 && w1->preNodeCount==0) w1->garbageCollect();
      table_.insert(calcPtr(node,OPERATOR_),varSet,result);

      return result;
    } else {
      ZADDBaseNode* result;
      if(varSet->getRealLevel()<node->getRealLevel()){
        ZADDBaseNode* w=calc(node,static_cast<const ZADDInnerNode*>(varSet)->getSucc1());
        result=(*function_)(w,w);
        if(result!=w && w->preNodeCount==0) w->garbageCollect();
      } else {
        ZADDBaseNode* w0=calc(static_cast<const ZADDInnerNode*>(node)->getSucc0(),varSet);
        ZADDBaseNode* w1=calc(static_cast<const ZADDInnerNode*>(node)->getSucc1(),varSet);
        result=(w1==ZADD::zeroDrain)?w0:findOrAdd(node->getUniqueTable(),w0,w1);
      }
      table_.insert(calcPtr(node,OPERATOR_),varSet,result);

      return result;
    }
  }
};

struct idempotentAbstract{
  const short OPERATOR_;
  PtrHashMap2<ZADDBaseNode>& table_;
  ZADDBaseNode* (*function_)(const ZADDBaseNode*,const ZADDBaseNode*);

  idempotentAbstract(const short op, PtrHashMap2<ZADDBaseNode>& table, ZADDBaseNode* (*function)(const ZADDBaseNode*,const ZADDBaseNode*)): OPERATOR_(op),table_(table), function_(function){}
  ZADDBaseNode* calc(const ZADDBaseNode* node, const ZADDBaseNode* varSet){
    if(varSet->isDrain()) return const_cast<ZADDBaseNode*>(node);
    if(ZADDBaseNode* comp=table_.find(calcPtr(node,OPERATOR_),varSet)) return comp;

    if(varSet->getUniqueTable()==node->getUniqueTable()){
      ZADDBaseNode* newVarSet=static_cast<const ZADDInnerNode*>(varSet)->getSucc1();
      ZADDBaseNode* w0=calc(static_cast<const ZADDInnerNode*>(node)->getSucc0(),newVarSet);
      ZADDBaseNode* w1=calc(static_cast<const ZADDInnerNode*>(node)->getSucc1(),newVarSet);
      ZADDBaseNode* result=(*function_)(w0,w1);
      if(result!=w0 && w0->preNodeCount==0) w0->garbageCollect();
      if(result!=w1 && w1->preNodeCount==0) w1->garbageCollect();
      table_.insert(calcPtr(node,OPERATOR_),varSet,result);

      return result;
    } else {
      ZADDBaseNode* w0=calc(static_cast<const ZADDInnerNode*>(node)->getSucc0(),varSet);
      ZADDBaseNode* w1=calc(static_cast<const ZADDInnerNode*>(node)->getSucc1(),varSet);
      ZADDBaseNode* result=(w1==ZADD::zeroDrain)?w0:findOrAdd(node->getUniqueTable(),w0,w1);
      table_.insert(calcPtr(node,OPERATOR_),varSet,result);

      return result;
    }
  }
};

ZADDBaseNode* addition(const ZADDBaseNode* v){
  if(v->isDrain()) return (v==ZADD::zeroDrain)?ZADD::zeroDrain:findOrAddDrain(2.0*static_cast<const ZADDTerminalNode*>(v)->getValue());

  //try to get result from computed table
  if(ZADDBaseNode* comp=ZADD::ArithmeticComputedTable.find(calcPtr(v,ADD_OP),0,v,0)) return comp;

  ZADDUniqueTable* minTable=v->getUniqueTable();
  ZADDBaseNode* w0=addition(static_cast<const ZADDInnerNode*>(v)->getSucc0());
  ZADDBaseNode* w1=addition(static_cast<const ZADDInnerNode*>(v)->getSucc1());

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);

  //insert result into computed table
  ZADD::ArithmeticComputedTable.insert(calcPtr(v,ADD_OP),0,v,0,node);

  return node;
}

ZADDBaseNode* addition(const ZADDBaseNode* v1, const ZADDBaseNode* v2){
  if(v1==ZADD::zeroDrain) return const_cast<ZADDBaseNode*>(v2);
  if(v2==ZADD::zeroDrain) return const_cast<ZADDBaseNode*>(v1);
  if(v1==v2) return addition(v1);
  if(v1->isDrain() && v2->isDrain()) return findOrAddDrain(static_cast<const ZADDTerminalNode*>(v1)->getValue()+static_cast<const ZADDTerminalNode*>(v2)->getValue());

  swapPointer(v1,v2);

  if(ZADDBaseNode* comp=ZADD::ArithmeticComputedTable.find(calcPtr(v1,ADD_OP),0,v2,0)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;
  ZADDUniqueTable* minTable;

  //calculate minTable
  if(v1->getRealLevel()<=v2->getRealLevel()){
    minTable=v1->getUniqueTable();
    if(v2->getUniqueTable()==minTable){
      w0=addition(static_cast<const ZADDInnerNode*>(v1)->getSucc0(),static_cast<const ZADDInnerNode*>(v2)->getSucc0());
      w1=addition(static_cast<const ZADDInnerNode*>(v1)->getSucc1(),static_cast<const ZADDInnerNode*>(v2)->getSucc1());
    } else {
      w0=addition(static_cast<const ZADDInnerNode*>(v1)->getSucc0(),v2);
      w1=addition(static_cast<const ZADDInnerNode*>(v1)->getSucc1(),v2);
    }
  } else {
    minTable=v2->getUniqueTable();
    w0=addition(v1,static_cast<const ZADDInnerNode*>(v2)->getSucc0());
    w1=addition(v1,static_cast<const ZADDInnerNode*>(v2)->getSucc1());
  }

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);

  //insert result into computed table
  ZADD::ArithmeticComputedTable.insert(calcPtr(v1,ADD_OP),0,v2,0,node);

  return node;
}

ZADDBaseNode* addition(const ZADDBaseNode* v1, const ZADDBaseNode* v1VS, const ZADDBaseNode* v2, const ZADDBaseNode* v2VS){
  if(v1VS==v2VS) return ::addition(v1,v2);
  if(v1==ZADD::zeroDrain) return calcLiftedBDD(v2,v2VS,v1VS);
  if(v2==ZADD::zeroDrain) return calcLiftedBDD(v1,v1VS,v2VS);

  swapPointer(v1,v1VS,v2,v2VS);

  if(ZADDBaseNode* comp=ZADD::ArithmeticComputedTable.find(calcPtr(v1,ADD_OP),v1VS,v2,v2VS)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;
  ZADDBaseNode* v10;
  ZADDBaseNode* v11;
  ZADDBaseNode* v20;
  ZADDBaseNode* v21;
  ZADDBaseNode* newV1VS;
  ZADDBaseNode* newV2VS;

  ZADDUniqueTable* minTable;

  if(v1VS->getRealLevel()<=v2VS->getRealLevel()){
    minTable=v1VS->getUniqueTable();
    if(v1->getUniqueTable()==minTable){
      v10=static_cast<const ZADDInnerNode*>(v1)->getSucc0();
      v11=static_cast<const ZADDInnerNode*>(v1)->getSucc1();
    } else {
      v10=const_cast<ZADDBaseNode*>(v1);
      v11=ZADD::zeroDrain;
    }
    newV1VS=static_cast<const ZADDInnerNode*>(v1VS)->getSucc1();
    if(v2VS->getUniqueTable()==minTable){
      if(v2->getUniqueTable()==minTable){
        v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
        v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
      } else {
        v20=const_cast<ZADDBaseNode*>(v2);
        v21=ZADD::zeroDrain;
      }
      newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
    } else {
      newV2VS=const_cast<ZADDBaseNode*>(v2VS);
      v20=v21=const_cast<ZADDBaseNode*>(v2);
    }
  } else {
    minTable=v2VS->getUniqueTable();
    v10=v11=const_cast<ZADDBaseNode*>(v1);
    if(v2->getUniqueTable()==minTable){
      v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
      v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
    } else {
      v20=const_cast<ZADDBaseNode*>(v2);
      v21=ZADD::zeroDrain;
    }
    newV1VS=const_cast<ZADDBaseNode*>(v1VS);
    newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
  }

  w0=addition(v10,newV1VS,v20,newV2VS);
  w1=addition(v11,newV1VS,v21,newV2VS);

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);

  ZADD::ArithmeticComputedTable.insert(calcPtr(v1,ADD_OP),v1VS,v2,v2VS,node);

  return node;
}

ZADDFunction& ZADDFunction::operator+=(const ZADDFunction& function){
  return *this=ZADDFunction(::addition(rootNode,varSet,function.rootNode,function.varSet),calcUnion(varSet,function.varSet));
}

ZADDBaseNode* change(const ZADDBaseNode* v, const ZADDBaseNode* vVS){
  if(vVS->isDrain()) return const_cast<ZADDBaseNode*>(v);

  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(v,CHA_OP),vVS,0,0)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;
  if(v->getUniqueTable()==vVS->getUniqueTable()){
    w0=change(static_cast<const ZADDInnerNode*>(v)->getSucc0(),static_cast<const ZADDInnerNode*>(vVS)->getSucc1());
    w1=change(static_cast<const ZADDInnerNode*>(v)->getSucc1(),static_cast<const ZADDInnerNode*>(vVS)->getSucc1());
  } else {
    w0=change(v,static_cast<const ZADDInnerNode*>(vVS)->getSucc1());
    w1=ZADD::zeroDrain;
  }

  ZADDBaseNode* node=(w0==ZADD::zeroDrain)?w1:findOrAdd(vVS->getUniqueTable(),w1,w0);

  ZADD::BoolComputedTable.insert(calcPtr(v,CHA_OP),vVS,0,0,node);

  return node;
}

ZADDBaseNode* change(const ZADDBaseNode* v, const ZADDBaseNode* vVS, ZADDBaseNode* cubeSet){
  if(vVS->isDrain()) return const_cast<ZADDBaseNode*>(v);

  while(cubeSet->getRealLevel()<vVS->getRealLevel()) cubeSet=static_cast<ZADDInnerNode*>(cubeSet)->getSucc1();
  if(cubeSet->isDrain()) return const_cast<ZADDBaseNode*>(v);

  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(v,CHA_OP),vVS,cubeSet,0)) return comp;

  ZADDBaseNode* v0;
  ZADDBaseNode* v1;
  if(v->getUniqueTable()==vVS->getUniqueTable()){
    v0=static_cast<const ZADDInnerNode*>(v)->getSucc0();
    v1=static_cast<const ZADDInnerNode*>(v)->getSucc1();
  } else {
    v0=const_cast<ZADDBaseNode*>(v);
    v1=ZADD::zeroDrain;
  }

  ZADDBaseNode* result;
  ZADDBaseNode* newVS=static_cast<const ZADDInnerNode*>(vVS)->getSucc1();
  if(vVS->getUniqueTable()==cubeSet->getUniqueTable()){
    ZADDBaseNode* w0=change(v0,newVS,static_cast<ZADDInnerNode*>(cubeSet)->getSucc1());
    ZADDBaseNode* w1=change(v1,newVS,static_cast<ZADDInnerNode*>(cubeSet)->getSucc1());

    result=(w0==ZADD::zeroDrain)?w1:findOrAdd(vVS->getUniqueTable(),w1,w0);
  } else {
    ZADDBaseNode* w0=change(v0,newVS,cubeSet);
    ZADDBaseNode* w1=change(v1,newVS,cubeSet);

    result=(w1==ZADD::zeroDrain)?w0:findOrAdd(vVS->getUniqueTable(),w0,w1);
  }

  ZADD::BoolComputedTable.insert(calcPtr(v,CHA_OP),vVS,cubeSet,0,result);

  return result;
}

ZADDBaseNode* subtraction(const ZADDBaseNode* v1,const ZADDBaseNode* v1VS,const ZADDBaseNode* v2,const ZADDBaseNode* v2VS){
  if((v1==v2)&&(v1VS==v2VS)) return ZADD::zeroDrain;
  if((v1VS->isDrain())&&(v2VS->isDrain())) return findOrAddDrain(static_cast<const ZADDTerminalNode*>(v1)->getValue()-static_cast<const ZADDTerminalNode*>(v2)->getValue());

  if(ZADDBaseNode* comp=ZADD::ArithmeticComputedTable.find(calcPtr(v1,SUB_OP),v1VS,v2,v2VS)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;
  ZADDBaseNode* v10;
  ZADDBaseNode* v11;
  ZADDBaseNode* v20;
  ZADDBaseNode* v21;
  ZADDBaseNode* newV1VS;
  ZADDBaseNode* newV2VS;

  ZADDUniqueTable* minTable;

  if(v1VS->getRealLevel()<=v2VS->getRealLevel()){
    minTable=v1VS->getUniqueTable();
    if(v1->getUniqueTable()==minTable){
      v10=static_cast<const ZADDInnerNode*>(v1)->getSucc0();
      v11=static_cast<const ZADDInnerNode*>(v1)->getSucc1();
    } else {
      v10=const_cast<ZADDBaseNode*>(v1);
      v11=ZADD::zeroDrain;
    }
    newV1VS=static_cast<const ZADDInnerNode*>(v1VS)->getSucc1();
    if(v2VS->getUniqueTable()==minTable){
      if(v2->getUniqueTable()==minTable){
        v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
        v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
      } else {
        v20=const_cast<ZADDBaseNode*>(v2);
        v21=ZADD::zeroDrain;
      }
      newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
    } else {
      newV2VS=const_cast<ZADDBaseNode*>(v2VS);
      v20=v21=const_cast<ZADDBaseNode*>(v2);
    }
  } else {
    minTable=v2VS->getUniqueTable();
    v10=v11=const_cast<ZADDBaseNode*>(v1);
    if(v2->getUniqueTable()==minTable){
      v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
      v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
    } else {
      v20=const_cast<ZADDBaseNode*>(v2);
      v21=ZADD::zeroDrain;
    }
    newV1VS=const_cast<ZADDBaseNode*>(v1VS);
    newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
  }

  w0=subtraction(v10,newV1VS,v20,newV2VS);
  w1=subtraction(v11,newV1VS,v21,newV2VS);

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);

  ZADD::ArithmeticComputedTable.insert(calcPtr(v1,SUB_OP),v1VS,v2,v2VS,node);

  return node;
}

ZADDFunction& ZADDFunction::operator-=(const ZADDFunction& function){
  return *this=ZADDFunction(::subtraction(rootNode,varSet,function.rootNode,function.varSet),calcUnion(varSet,function.varSet));
}

ZADDBaseNode* multiplication(const ZADDBaseNode* v){
  if(v->isDrain()) return (v==ZADD::zeroDrain)?ZADD::zeroDrain:findOrAddDrain(static_cast<const ZADDTerminalNode*>(v)->getValue()*static_cast<const ZADDTerminalNode*>(v)->getValue());

  //try to get result from computed table
  if(ZADDBaseNode* comp=ZADD::ArithmeticComputedTable.find(calcPtr(v,MUL_OP),0,v,0)) return comp;

  ZADDUniqueTable* minTable=v->getUniqueTable();
  ZADDBaseNode* w0=multiplication(static_cast<const ZADDInnerNode*>(v)->getSucc0());
  ZADDBaseNode* w1=multiplication(static_cast<const ZADDInnerNode*>(v)->getSucc1());

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);

  //insert result into computed table
  ZADD::ArithmeticComputedTable.insert(calcPtr(v,MUL_OP),0,v,0,node);

  return node;
}

ZADDBaseNode* multiplication(const ZADDBaseNode* v1, const ZADDBaseNode* v2){
  if(v1==ZADD::zeroDrain || v2==ZADD::zeroDrain) return ZADD::zeroDrain;
  if(v1==v2) return multiplication(v1);
  if(v1->isDrain() && v2->isDrain()) return findOrAddDrain(static_cast<const ZADDTerminalNode*>(v1)->getValue()*static_cast<const ZADDTerminalNode*>(v2)->getValue());

  swapPointer(v1,v2);

  //try to get result from computed table
  if(ZADDBaseNode* comp=ZADD::ArithmeticComputedTable.find(calcPtr(v1,MUL_OP),0,v2,0)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;
  ZADDUniqueTable* minTable;

  //calculate minTable
  if(v1->getRealLevel()<=v2->getRealLevel()){
    minTable=v1->getUniqueTable();
    if(v2->getUniqueTable()==minTable){
      w0=multiplication(static_cast<const ZADDInnerNode*>(v1)->getSucc0(),static_cast<const ZADDInnerNode*>(v2)->getSucc0());
      w1=multiplication(static_cast<const ZADDInnerNode*>(v1)->getSucc1(),static_cast<const ZADDInnerNode*>(v2)->getSucc1());
    } else {
      w0=multiplication(static_cast<const ZADDInnerNode*>(v1)->getSucc0(),v2);
      w1=multiplication(static_cast<const ZADDInnerNode*>(v1)->getSucc1(),v2);
    }
  } else {
    minTable=v2->getUniqueTable();
    w0=multiplication(v1,static_cast<const ZADDInnerNode*>(v2)->getSucc0());
    w1=multiplication(v1,static_cast<const ZADDInnerNode*>(v2)->getSucc1());
  }

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);

  //insert result into computed table
  ZADD::ArithmeticComputedTable.insert(calcPtr(v1,MUL_OP),0,v2,0,node);

  return node;
}

ZADDBaseNode* multiplication(const ZADDBaseNode* v1, ZADDBaseNode* v1VS, const ZADDBaseNode* v2, ZADDBaseNode* v2VS){
  if((v1==ZADD::zeroDrain)||(v2==ZADD::zeroDrain)) return ZADD::zeroDrain;

  ZADDUniqueTable* minTable;
  if(v1->getRealLevel()<=v2->getRealLevel()) minTable=v1->getUniqueTable();
  else minTable=v2->getUniqueTable();
	
  while(v1VS->getRealLevel()<minTable->getRealLevel()) v1VS=static_cast<ZADDInnerNode*>(v1VS)->getSucc1();
  while(v2VS->getRealLevel()<minTable->getRealLevel()) v2VS=static_cast<ZADDInnerNode*>(v1VS)->getSucc1();
	  
  if(v1VS->isDrain() && v2VS->isDrain()) return findOrAddDrain(static_cast<const ZADDTerminalNode*>(v1)->getValue()*static_cast<const ZADDTerminalNode*>(v2)->getValue());

  if(v1VS==v2VS) return ::multiplication(v1,v2);

  swapPointer(v1,v1VS,v2,v2VS);

  if(ZADDBaseNode* comp=ZADD::ArithmeticComputedTable.find(calcPtr(v1,MUL_OP),v1VS,v2,v2VS)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;
  ZADDBaseNode* v10;
  ZADDBaseNode* v11;
  ZADDBaseNode* v20;
  ZADDBaseNode* v21;
  ZADDBaseNode* newV1VS;
  ZADDBaseNode* newV2VS;

  if(v1VS->getRealLevel()<=v2VS->getRealLevel()){
    minTable=v1VS->getUniqueTable();
    if(v1->getUniqueTable()==minTable){
      v10=static_cast<const ZADDInnerNode*>(v1)->getSucc0();
      v11=static_cast<const ZADDInnerNode*>(v1)->getSucc1();
    } else {
      v10=const_cast<ZADDBaseNode*>(v1);
      v11=ZADD::zeroDrain;
    }
    newV1VS=static_cast<const ZADDInnerNode*>(v1VS)->getSucc1();
    if(v2VS->getUniqueTable()==minTable){
      if(v2->getUniqueTable()==minTable){
        v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
        v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
      } else {
        v20=const_cast<ZADDBaseNode*>(v2);
        v21=ZADD::zeroDrain;
      }
      newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
    } else {
      newV2VS=const_cast<ZADDBaseNode*>(v2VS);
      v20=v21=const_cast<ZADDBaseNode*>(v2);
    }
  } else {
    minTable=v2VS->getUniqueTable();
    v10=v11=const_cast<ZADDBaseNode*>(v1);
    if(v2->getUniqueTable()==minTable){
      v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
      v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
    } else {
      v20=const_cast<ZADDBaseNode*>(v2);
      v21=ZADD::zeroDrain;
    }
    newV1VS=const_cast<ZADDBaseNode*>(v1VS);
    newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
  }

  w0=multiplication(v10,newV1VS,v20,newV2VS);
  w1=multiplication(v11,newV1VS,v21,newV2VS);

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);

  ZADD::ArithmeticComputedTable.insert(calcPtr(v1,MUL_OP),v1VS,v2,v2VS,node);

  return node;
}

ZADDFunction& ZADDFunction::operator*=(const ZADDFunction& function){
  return *this=ZADDFunction(::multiplication(rootNode,function.rootNode),calcUnion(varSet,function.varSet));
}

ZADDBaseNode* OR(const ZADDBaseNode* v1, const ZADDBaseNode* v2){
  if(v1==v2) return const_cast<ZADDBaseNode*>(v1);
  if(v1->isDrain()) return (v1==ZADD::zeroDrain)?const_cast<ZADDBaseNode*>(v2):ZADD::oneDrain;
  if(v2->isDrain()) return (v2==ZADD::zeroDrain)?const_cast<ZADDBaseNode*>(v1):ZADD::oneDrain;

  swapPointer(v1,v2);
  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(v1,OR_OP),0,v2,0)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;

  ZADDUniqueTable* minTable;

  if(v1->getRealLevel()<=v2->getRealLevel()){
    minTable=v1->getUniqueTable();
    if(v2->getUniqueTable()==minTable){
      w0=OR(static_cast<const ZADDInnerNode*>(v1)->getSucc0(),static_cast<const ZADDInnerNode*>(v2)->getSucc0());
      w1=OR(static_cast<const ZADDInnerNode*>(v1)->getSucc1(),static_cast<const ZADDInnerNode*>(v2)->getSucc1());
    } else {
      w0=OR(static_cast<const ZADDInnerNode*>(v1)->getSucc0(),v2);
      w1=OR(static_cast<const ZADDInnerNode*>(v1)->getSucc1(),v2);
    }
  } else {
    minTable=v2->getUniqueTable();
    w0=OR(v1,static_cast<const ZADDInnerNode*>(v2)->getSucc0());
    w1=OR(v1,static_cast<const ZADDInnerNode*>(v2)->getSucc1());
  }

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);
  ZADD::BoolComputedTable.insert(calcPtr(v1,OR_OP),0,v2,0,node);

  return node;
}

ZADDBaseNode* OR(const ZADDBaseNode* v1, const ZADDBaseNode* v1VS, const ZADDBaseNode* v2, const ZADDBaseNode* v2VS){
  if(v1VS==v2VS) return ::OR(v1,v2);
  if(v1==ZADD::zeroDrain) return calcLiftedBDD(v2,v2VS,v1VS);
  if(v2==ZADD::zeroDrain) return calcLiftedBDD(v1,v1VS,v2VS);

  swapPointer(v1,v1VS,v2,v2VS);

  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(v1,OR_OP),v1VS,v2,v2VS)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;
  ZADDBaseNode* v10;
  ZADDBaseNode* v11;
  ZADDBaseNode* v20;
  ZADDBaseNode* v21;
  ZADDBaseNode* newV1VS;
  ZADDBaseNode* newV2VS;

  ZADDUniqueTable* minTable;

  if(v1VS->getRealLevel()<=v2VS->getRealLevel()){
    minTable=v1VS->getUniqueTable();
    if(v1->getUniqueTable()==minTable){
      v10=static_cast<const ZADDInnerNode*>(v1)->getSucc0();
      v11=static_cast<const ZADDInnerNode*>(v1)->getSucc1();
    } else {
      v10=const_cast<ZADDBaseNode*>(v1);
      v11=ZADD::zeroDrain;
    }
    newV1VS=static_cast<const ZADDInnerNode*>(v1VS)->getSucc1();
    if(v2VS->getUniqueTable()==minTable){
      if(v2->getUniqueTable()==minTable){
        v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
        v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
      } else {
        v20=const_cast<ZADDBaseNode*>(v2);
        v21=ZADD::zeroDrain;
      }
      newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
    } else {
      newV2VS=const_cast<ZADDBaseNode*>(v2VS);
      v20=v21=const_cast<ZADDBaseNode*>(v2);
    }
  } else {
    minTable=v2VS->getUniqueTable();
    v10=v11=const_cast<ZADDBaseNode*>(v1);
    if(v2->getUniqueTable()==minTable){
      v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
      v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
    } else {
      v20=const_cast<ZADDBaseNode*>(v2);
      v21=ZADD::zeroDrain;
    }
    newV1VS=const_cast<ZADDBaseNode*>(v1VS);
    newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
  }

  w0=OR(v10,newV1VS,v20,newV2VS);
  w1=OR(v11,newV1VS,v21,newV2VS);

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);

  ZADD::BoolComputedTable.insert(calcPtr(v1,OR_OP),v1VS,v2,v2VS,node);

  return node;
}

ZADDFunction& ZADDFunction::operator|=(const ZADDFunction& function){
  return *this=ZADDFunction(OR(rootNode,varSet,function.rootNode,function.varSet),calcUnion(varSet,function.varSet));
}

ZADDBaseNode* AND(const ZADDBaseNode* v1, const ZADDBaseNode* v2){
  if(v1==v2) return const_cast<ZADDBaseNode*>(v1);
  if(v1->isDrain()) return (v1==ZADD::zeroDrain)?ZADD::zeroDrain:const_cast<ZADDBaseNode*>(v2);
  if(v2->isDrain()) return (v2==ZADD::zeroDrain)?ZADD::zeroDrain:const_cast<ZADDBaseNode*>(v1);

  swapPointer(v1,v2);
  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(v1,AND_OP),0,v2,0)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;

  ZADDUniqueTable* minTable;

  if(v1->getRealLevel()<=v2->getRealLevel()){
    minTable=v1->getUniqueTable();
    if(v2->getUniqueTable()==minTable){
      w0=AND(static_cast<const ZADDInnerNode*>(v1)->getSucc0(),static_cast<const ZADDInnerNode*>(v2)->getSucc0());
      w1=AND(static_cast<const ZADDInnerNode*>(v1)->getSucc1(),static_cast<const ZADDInnerNode*>(v2)->getSucc1());
    } else {
      w0=AND(static_cast<const ZADDInnerNode*>(v1)->getSucc0(),v2);
      w1=AND(static_cast<const ZADDInnerNode*>(v1)->getSucc1(),v2);
    }
  } else {
    minTable=v2->getUniqueTable();
    w0=AND(v1,static_cast<const ZADDInnerNode*>(v2)->getSucc0());
    w1=AND(v1,static_cast<const ZADDInnerNode*>(v2)->getSucc1());
  }

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);
  ZADD::BoolComputedTable.insert(calcPtr(v1,AND_OP),0,v2,0,node);

  return node;
}

ZADDBaseNode* AND(const ZADDBaseNode* v1, ZADDBaseNode* v1VS, const ZADDBaseNode* v2, ZADDBaseNode* v2VS){
  if((v1==ZADD::zeroDrain)||(v2==ZADD::zeroDrain)) return ZADD::zeroDrain;

  ZADDUniqueTable* minTable;
  if(v1->getRealLevel()<=v2->getRealLevel()) minTable=v1->getUniqueTable();
  else minTable=v2->getUniqueTable();
	
  while(v1VS->getRealLevel()<minTable->getRealLevel()) v1VS=static_cast<ZADDInnerNode*>(v1VS)->getSucc1();
  while(v2VS->getRealLevel()<minTable->getRealLevel()) v2VS=static_cast<ZADDInnerNode*>(v1VS)->getSucc1();
	  
  if(v1VS->isDrain() && v2VS->isDrain()) return findOrAddDrain(static_cast<const ZADDTerminalNode*>(v1)->getValue()*static_cast<const ZADDTerminalNode*>(v2)->getValue());

  if(v1VS==v2VS) return ::multiplication(v1,v2);
  if(v1==v2) return const_cast<ZADDBaseNode*>(v1);

  swapPointer(v1,v1VS,v2,v2VS);

  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(v1,AND_OP),v1VS,v2,v2VS)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;
  ZADDBaseNode* v10;
  ZADDBaseNode* v11;
  ZADDBaseNode* v20;
  ZADDBaseNode* v21;
  ZADDBaseNode* newV1VS;
  ZADDBaseNode* newV2VS;

  if(v1VS->getRealLevel()<=v2VS->getRealLevel()){
    minTable=v1VS->getUniqueTable();
    if(v1->getUniqueTable()==minTable){
      v10=static_cast<const ZADDInnerNode*>(v1)->getSucc0();
      v11=static_cast<const ZADDInnerNode*>(v1)->getSucc1();
    } else {
      v10=const_cast<ZADDBaseNode*>(v1);
      v11=ZADD::zeroDrain;
    }
    newV1VS=static_cast<const ZADDInnerNode*>(v1VS)->getSucc1();
    if(v2VS->getUniqueTable()==minTable){
      if(v2->getUniqueTable()==minTable){
        v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
        v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
      } else {
        v20=const_cast<ZADDBaseNode*>(v2);
        v21=ZADD::zeroDrain;
      }
      newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
    } else {
      newV2VS=const_cast<ZADDBaseNode*>(v2VS);
      v20=v21=const_cast<ZADDBaseNode*>(v2);
    }
  } else {
    minTable=v2VS->getUniqueTable();
    v10=v11=const_cast<ZADDBaseNode*>(v1);
    if(v2->getUniqueTable()==minTable){
      v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
      v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
    } else {
      v20=const_cast<ZADDBaseNode*>(v2);
      v21=ZADD::zeroDrain;
    }
    newV1VS=const_cast<ZADDBaseNode*>(v1VS);
    newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
  }

  w0=AND(v10,newV1VS,v20,newV2VS);
  w1=AND(v11,newV1VS,v21,newV2VS);

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);

  ZADD::BoolComputedTable.insert(calcPtr(v1,AND_OP),v1VS,v2,v2VS,node);

  return node;
}

ZADDFunction& ZADDFunction::operator&=(const ZADDFunction& function){
  return *this=ZADDFunction(AND(rootNode,function.rootNode),calcUnion(varSet,function.varSet));
}

ZADDBaseNode* exists(const ZADDBaseNode* node, const ZADDBaseNode* varSet){
  if(node->isDrain()) return const_cast<ZADDBaseNode*>(node);

  unsigned long nodeLevel=node->getRealLevel();
  while(varSet->getRealLevel()<nodeLevel) varSet=static_cast<const ZADDInnerNode*>(varSet)->getSucc1();

  if(varSet->isDrain()) return const_cast<ZADDBaseNode*>(node);

  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(node,EXI_OP),varSet,0)) return comp;

  if(varSet->getUniqueTable()==node->getUniqueTable()){
    ZADDBaseNode* w0=exists(static_cast<const ZADDInnerNode*>(node)->getSucc0(),static_cast<const ZADDInnerNode*>(varSet)->getSucc1());
    ZADDBaseNode* w1=exists(static_cast<const ZADDInnerNode*>(node)->getSucc1(),static_cast<const ZADDInnerNode*>(varSet)->getSucc1());
    ZADDBaseNode* result=::OR(w0,w1);

    if(result!=w0 && w0->preNodeCount==0) w0->garbageCollect();
    if(result!=w1 && w1->preNodeCount==0) w1->garbageCollect();

    ZADD::BoolComputedTable.insert(calcPtr(node,EXI_OP),varSet,0,result);

    return result;
  } else {
    ZADDBaseNode* w0=exists(static_cast<const ZADDInnerNode*>(node)->getSucc0(),varSet);
    ZADDBaseNode* w1=exists(static_cast<const ZADDInnerNode*>(node)->getSucc1(),varSet);

    ZADDBaseNode* temp=(w1==ZADD::zeroDrain)?w0:findOrAdd(node->getUniqueTable(),w0,w1);
    ZADD::BoolComputedTable.insert(calcPtr(node,EXI_OP),varSet,0,temp);

    return temp;
  }
}

ZADDFunction ZADDFunction::exists(const ZADDFunction& function) const {
  return ZADDFunction(::exists(rootNode,function.rootNode));
}

ZADDFunction& ZADDFunction::operator^=(const ZADDFunction& function){
  return *this=ITE(*this,!function,function);
}

ZADDFunction& ZADDFunction::operator++(){
  return *this=ZADDFunction(::addition(rootNode,varSet,ZADD::oneDrain,ZADD::oneDrain),varSet);
}

ZADDFunction& ZADDFunction::operator--(){
  return *this=ZADDFunction(::subtraction(rootNode,varSet,ZADD::oneDrain,ZADD::oneDrain),varSet);
}

ZADDBaseNode* NOT(const ZADDBaseNode* u, const ZADDBaseNode* uVS){
  if(u==ZADD::zeroDrain) return calcLiftedBDD(ZADD::oneDrain,ZADD::oneDrain,uVS);
  if(uVS->isDrain()) return (u==ZADD::zeroDrain)?ZADD::oneDrain:ZADD::zeroDrain;

  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(ZADD::oneDrain,DIF_OP),ZADD::oneDrain,u,uVS)) return comp;

  ZADDBaseNode* node;
  ZADDBaseNode* newUVS=static_cast<const ZADDInnerNode*>(uVS)->getSucc1();
  if(u->getUniqueTable()==uVS->getUniqueTable()){
    ZADDBaseNode* w0=NOT(static_cast<const ZADDInnerNode*>(u)->getSucc0(),newUVS);
    ZADDBaseNode* w1=NOT(static_cast<const ZADDInnerNode*>(u)->getSucc1(),newUVS);

    node=(w1==ZADD::zeroDrain)?w0:findOrAdd(uVS->getUniqueTable(),w0,w1);
  } else {
    ZADDBaseNode* w0=NOT(u,newUVS);
    ZADDBaseNode* w1=calcLiftedBDD(ZADD::oneDrain,ZADD::oneDrain,newUVS);

    node=findOrAdd(uVS->getUniqueTable(),w0,w1);
  }

  //insert result into computed table
  ZADD::BoolComputedTable.insert(calcPtr(ZADD::oneDrain,DIF_OP),ZADD::oneDrain,u,uVS,node);

  return node;
}

ZADDFunction ZADDFunction::operator!(void) const {
  return ZADDFunction(NOT(rootNode,varSet),varSet);
}

bool equiv(const ZADDBaseNode* u, const ZADDBaseNode* uVS, const ZADDBaseNode* v, const ZADDBaseNode* vVS){
  if((uVS->isDrain())&&(vVS->isDrain())) return (u==v);

  if(uVS->getRealLevel()<vVS->getRealLevel()){
    if(u->getUniqueTable()!=uVS->getUniqueTable()) return false; //zero suppressed vs don't care
    if(static_cast<const ZADDInnerNode*>(u)->getSucc0()!=static_cast<const ZADDInnerNode*>(u)->getSucc1()) return false;
    return equiv(static_cast<const ZADDInnerNode*>(u)->getSucc0(),static_cast<const ZADDInnerNode*>(uVS)->getSucc1(),v,vVS);
  } else if(vVS->getRealLevel()<uVS->getRealLevel()){
    if(v->getUniqueTable()!=vVS->getUniqueTable()) return false; //zero suppressed vs don't care
    if(static_cast<const ZADDInnerNode*>(v)->getSucc0()!=static_cast<const ZADDInnerNode*>(v)->getSucc1()) return false;
    return equiv(u,uVS,static_cast<const ZADDInnerNode*>(v)->getSucc0(),static_cast<const ZADDInnerNode*>(vVS)->getSucc1());
  }

  if(u->getUniqueTable()!=v->getUniqueTable()) return false;

  if(u->getUniqueTable()==uVS->getUniqueTable()){
    if(!equiv(static_cast<const ZADDInnerNode*>(u)->getSucc0(),static_cast<const ZADDInnerNode*>(uVS)->getSucc1(),static_cast<const ZADDInnerNode*>(v)->getSucc0(),static_cast<const ZADDInnerNode*>(vVS)->getSucc1())) return false;
    return equiv(static_cast<const ZADDInnerNode*>(u)->getSucc1(),static_cast<const ZADDInnerNode*>(uVS)->getSucc1(),static_cast<const ZADDInnerNode*>(v)->getSucc1(),static_cast<const ZADDInnerNode*>(vVS)->getSucc1());
  } else {
    return equiv(u,static_cast<const ZADDInnerNode*>(uVS)->getSucc1(),v,static_cast<const ZADDInnerNode*>(vVS)->getSucc1());
  }
}

bool ZADDFunction::operator==(const ZADDFunction& function) const {
  if(varSet==function.varSet) return (rootNode==function.rootNode);
  return (equiv(rootNode,varSet,function.rootNode,function.varSet));
}

ZADDBaseNode* less(const ZADDBaseNode* u, const ZADDBaseNode* v){
  if(u==v) return ZADD::zeroDrain;
  if(u->isDrain() && v->isDrain()) return (static_cast<const ZADDTerminalNode*>(u)->getValue()<static_cast<const ZADDTerminalNode*>(v)->getValue())?ZADD::oneDrain:ZADD::zeroDrain;

  if(ZADDBaseNode* comp=ZADD::CompareComputedTable.find(calcPtr(u,LES_OP),v)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;

  ZADDUniqueTable* minTable;

  if(u->getRealLevel()<=v->getRealLevel()){
    minTable=u->getUniqueTable();
    if(v->getUniqueTable()==minTable){
      w0=::less(static_cast<const ZADDInnerNode*>(u)->getSucc0(),static_cast<const ZADDInnerNode*>(v)->getSucc0());
      w1=::less(static_cast<const ZADDInnerNode*>(u)->getSucc1(),static_cast<const ZADDInnerNode*>(v)->getSucc1());
    } else {
      w0=::less(static_cast<const ZADDInnerNode*>(u)->getSucc0(),v);
      w1=::less(static_cast<const ZADDInnerNode*>(u)->getSucc1(),v);
    }
  } else {
    minTable=v->getUniqueTable();
    w0=::less(u,static_cast<const ZADDInnerNode*>(v)->getSucc0());
    w1=::less(u,static_cast<const ZADDInnerNode*>(v)->getSucc1());
  }

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);
  ZADD::CompareComputedTable.insert(calcPtr(u,LES_OP),v,node);

  return node;
}

ZADDBaseNode* max(const ZADDBaseNode* v1, const ZADDBaseNode* v1VS, const ZADDBaseNode* v2, const ZADDBaseNode* v2VS){
  if(v1VS==v2VS){
    if(v1==v2) return const_cast<ZADDBaseNode*>(v1);
    if(v1VS->isDrain()) return const_cast<ZADDBaseNode*>(static_cast<const ZADDTerminalNode*>(v1)->getValue()>=static_cast<const ZADDTerminalNode*>(v2)->getValue()?v1:v2);
  }

  swapPointer(v1,v1VS,v2,v2VS);

  if(ZADDBaseNode* comp=ZADD::ExtremaComputedTable.find(calcPtr(v1,MAX_OP),v1VS,v2,v2VS)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;
  ZADDBaseNode* v10;
  ZADDBaseNode* v11;
  ZADDBaseNode* v20;
  ZADDBaseNode* v21;
  ZADDBaseNode* newV1VS;
  ZADDBaseNode* newV2VS;

  ZADDUniqueTable* minTable;

  if(v1VS->getRealLevel()<=v2VS->getRealLevel()){
    minTable=v1VS->getUniqueTable();
    if(v1->getUniqueTable()==minTable){
      v10=static_cast<const ZADDInnerNode*>(v1)->getSucc0();
      v11=static_cast<const ZADDInnerNode*>(v1)->getSucc1();
    } else {
      v10=const_cast<ZADDBaseNode*>(v1);
      v11=ZADD::zeroDrain;
    }
    newV1VS=static_cast<const ZADDInnerNode*>(v1VS)->getSucc1();
    if(v2VS->getUniqueTable()==minTable){
      if(v2->getUniqueTable()==minTable){
        v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
        v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
      } else {
        v20=const_cast<ZADDBaseNode*>(v2);
        v21=ZADD::zeroDrain;
      }
      newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
    } else {
      newV2VS=const_cast<ZADDBaseNode*>(v2VS);
      v20=v21=const_cast<ZADDBaseNode*>(v2);
    }
  } else {
    minTable=v2VS->getUniqueTable();
    v10=v11=const_cast<ZADDBaseNode*>(v1);
    if(v2->getUniqueTable()==minTable){
      v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
      v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
    } else {
      v20=const_cast<ZADDBaseNode*>(v2);
      v21=ZADD::zeroDrain;
    }
    newV1VS=const_cast<ZADDBaseNode*>(v1VS);
    newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
  }

  w0=max(v10,newV1VS,v20,newV2VS);
  w1=max(v11,newV1VS,v21,newV2VS);

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);

  ZADD::ExtremaComputedTable.insert(calcPtr(v1,MAX_OP),v1VS,v2,v2VS,node);

  return node;
}

ZADDBaseNode* max(const ZADDBaseNode* u){
  if(u->isDrain()) return const_cast<ZADDBaseNode*>(u);

  if(ZADDBaseNode* comp=ZADD::ExtremaComputedTable.find(calcPtr(u,MAX_OP),0,0,0)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;

  w0=::max(static_cast<const ZADDInnerNode*>(u)->getSucc0());
  w1=::max(static_cast<const ZADDInnerNode*>(u)->getSucc1());

  ZADDBaseNode* node=(static_cast<ZADDTerminalNode*>(w0)->getValue()>=static_cast<ZADDTerminalNode*>(w1)->getValue())?w0:w1;
  ZADD::ExtremaComputedTable.insert(calcPtr(u,MAX_OP),0,0,0,node);

  return node;
}

ZADDBaseNode* min(const ZADDBaseNode* v1, const ZADDBaseNode* v1VS, const ZADDBaseNode* v2, const ZADDBaseNode* v2VS){
  if(v1VS==v2VS){
    if(v1==v2) return const_cast<ZADDBaseNode*>(v1);
    if(v1VS->isDrain()) return const_cast<ZADDBaseNode*>(static_cast<const ZADDTerminalNode*>(v1)->getValue()<=static_cast<const ZADDTerminalNode*>(v2)->getValue()?v1:v2);
  }

  swapPointer(v1,v1VS,v2,v2VS);

  if(ZADDBaseNode* comp=ZADD::ExtremaComputedTable.find(calcPtr(v1,MIN_OP),v1VS,v2,v2VS)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;
  ZADDBaseNode* v10;
  ZADDBaseNode* v11;
  ZADDBaseNode* v20;
  ZADDBaseNode* v21;
  ZADDBaseNode* newV1VS;
  ZADDBaseNode* newV2VS;

  ZADDUniqueTable* minTable;

  if(v1VS->getRealLevel()<=v2VS->getRealLevel()){
    minTable=v1VS->getUniqueTable();
    if(v1->getUniqueTable()==minTable){
      v10=static_cast<const ZADDInnerNode*>(v1)->getSucc0();
      v11=static_cast<const ZADDInnerNode*>(v1)->getSucc1();
    } else {
      v10=const_cast<ZADDBaseNode*>(v1);
      v11=ZADD::zeroDrain;
    }
    newV1VS=static_cast<const ZADDInnerNode*>(v1VS)->getSucc1();
    if(v2VS->getUniqueTable()==minTable){
      if(v2->getUniqueTable()==minTable){
        v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
        v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
      } else {
        v20=const_cast<ZADDBaseNode*>(v2);
        v21=ZADD::zeroDrain;
      }
      newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
    } else {
      newV2VS=const_cast<ZADDBaseNode*>(v2VS);
      v20=v21=const_cast<ZADDBaseNode*>(v2);
    }
  } else {
    minTable=v2VS->getUniqueTable();
    v10=v11=const_cast<ZADDBaseNode*>(v1);
    if(v2->getUniqueTable()==minTable){
      v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
      v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
    } else {
      v20=const_cast<ZADDBaseNode*>(v2);
      v21=ZADD::zeroDrain;
    }
    newV1VS=const_cast<ZADDBaseNode*>(v1VS);
    newV2VS=static_cast<const ZADDInnerNode*>(v2VS)->getSucc1();
  }

  w0=min(v10,newV1VS,v20,newV2VS);
  w1=min(v11,newV1VS,v21,newV2VS);

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);

  ZADD::ExtremaComputedTable.insert(calcPtr(v1,MIN_OP),v1VS,v2,v2VS,node);

  return node;
}

ZADDBaseNode* min(const ZADDBaseNode* u){
  if(u->isDrain()) return const_cast<ZADDBaseNode*>(u);

  if(ZADDBaseNode* comp=ZADD::ExtremaComputedTable.find(calcPtr(u,MIN_OP),0,0,0)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;

  w0=::min(static_cast<const ZADDInnerNode*>(u)->getSucc0());
  w1=::min(static_cast<const ZADDInnerNode*>(u)->getSucc1());

  ZADDBaseNode* node=(static_cast<ZADDTerminalNode*>(w0)->getValue()<=static_cast<ZADDTerminalNode*>(w1)->getValue())?w0:w1;
  ZADD::ExtremaComputedTable.insert(calcPtr(u,MIN_OP),0,0,0,node);

  return node;
}


bool ZADDFunction::operator<(const ZADDFunction& function) const {
  return (LESS(*this,function)==ZADDFunction::ONE());
}

bool ZADDFunction::isConstant(void) const {
  if(varSet->isDrain()) return true;
  
  ZADDBaseNode* node=rootNode;
  ZADDBaseNode* nodeVarSet=varSet;

  while(!nodeVarSet->isDrain()){
    if(node->getUniqueTable()==nodeVarSet->getUniqueTable()){
      if(static_cast<ZADDInnerNode*>(node)->getSucc0()!=static_cast<ZADDInnerNode*>(node)->getSucc1()) return false;
      else {
	node=static_cast<ZADDInnerNode*>(node)->getSucc0();
	nodeVarSet=static_cast<ZADDInnerNode*>(nodeVarSet)->getSucc1();
      }
    } else return false;
  }
  return ULONG_MAX;
}

double ZADDFunction::getValue(void) const {
  return isConstant()?static_cast<const ZADDTerminalNode*>(rootNode)->getValue():0.0;
}

unsigned long ZADDFunction::getFirstVariableIndex(void) const {
  ZADDBaseNode* node=rootNode;
  ZADDBaseNode* nodeVarSet=varSet;

  while(!nodeVarSet->isDrain()){
    if(node->getUniqueTable()==nodeVarSet->getUniqueTable()){
      if(static_cast<ZADDInnerNode*>(node)->getSucc0()!=static_cast<ZADDInnerNode*>(node)->getSucc1()) return nodeVarSet->getLevel();
      else {
	node=static_cast<ZADDInnerNode*>(node)->getSucc0();
	nodeVarSet=static_cast<ZADDInnerNode*>(nodeVarSet)->getSucc1();
      }
    } else return nodeVarSet->getLevel();
  }
  return ULONG_MAX;
}

ZADDBaseNode* cofactor(const ZADDBaseNode* node, const ZADDBaseNode* assignment){
  if(node->isDrain()) return const_cast<ZADDBaseNode*>(node);

  while(assignment->getRealLevel()<node->getRealLevel()){
    assignment=(static_cast<const ZADDInnerNode*>(assignment)->getSucc0()==ZADD::zeroDrain)?static_cast<const ZADDInnerNode*>(assignment)->getSucc1():static_cast<const ZADDInnerNode*>(assignment)->getSucc0();
  }

  if(assignment->isDrain()) return const_cast<ZADDBaseNode*>(node);


  //check for computed table
  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(node,COF_OP),assignment,0)) return comp;

  //do it the hard way
  ZADDBaseNode* tempNode;
  if(assignment->getRealLevel()==node->getRealLevel()){
    const bool value=(static_cast<const ZADDInnerNode*>(assignment)->getSucc0()==ZADD::zeroDrain)?true:false;
    tempNode=cofactor(static_cast<const ZADDInnerNode*>(node)->getSucc(value),static_cast<const ZADDInnerNode*>(assignment)->getSucc(value));
  } else {
    ZADDBaseNode* w0=cofactor(static_cast<const ZADDInnerNode*>(node)->getSucc0(),assignment);
    ZADDBaseNode* w1=cofactor(static_cast<const ZADDInnerNode*>(node)->getSucc1(),assignment);

    tempNode=(w1==ZADD::zeroDrain)?w0:findOrAdd(node->getUniqueTable(),w0,w1);
  }

  //insert result into computed table
  ZADD::BoolComputedTable.insert(calcPtr(node,COF_OP),assignment,0,tempNode);

  return tempNode;
}

ZADDFunction ZADDFunction::cofactor(const unsigned long index, const bool value) const {
  if(index>=ZADD::variableOrdering.size()) return ZADDFunction(rootNode);
  if(index==rootNode->getLevel()) return ZADDFunction(static_cast<const ZADDInnerNode*>(rootNode)->getSucc(value));

  //create temporary node
  ZADDBaseNode* tempVar=value?findOrAdd(ZADD::variableOrdering.getData(index),ZADD::zeroDrain,ZADD::oneDrain):findOrAdd(ZADD::variableOrdering.getData(index),ZADD::oneDrain,ZADD::zeroDrain);
  ZADDFunction nrv(::cofactor(rootNode,tempVar));

  //if node was really created above delete it
  if(tempVar->preNodeCount==0) tempVar->garbageCollect();

  return nrv;
}

ZADDFunction ZADDFunction::cofactor(const ZADDFunction& assignment) const {
  return ZADDFunction(::cofactor(rootNode,assignment.rootNode));
}

ZADDFunction ZADDFunction::subSet(const unsigned long index, const bool value) const {
  if(index>=ZADDVarOrder::size()) return ZADDFunction(rootNode);

  //create temporary node
  ZADDBaseNode* tempVar=value?findOrAdd(ZADD::variableOrdering.getData(index),ZADD::zeroDrain,ZADD::oneDrain):findOrAdd(ZADD::variableOrdering.getData(index),ZADD::oneDrain,ZADD::zeroDrain);
  ZADDFunction nrv(::multiplication(rootNode,tempVar));

  //if node was really created above delete it
  if(tempVar->preNodeCount==0) tempVar->garbageCollect();

  return nrv;
}

ZADDFunction ZADDFunction::subSet(const ZADDFunction& assignment) const {
  return ((*this)*assignment);
}

ZADDFunction ZADDFunction::change(void) const {
  return ZADDFunction(::change(rootNode,varSet),varSet);
}

ZADDFunction ZADDFunction::change(const unsigned long index) const {
  if(index>=ZADDVarOrder::size()) return ZADDFunction(rootNode,varSet);

  ZADDBaseNode* var=findOrAdd(ZADD::variableOrdering.getData(index),ZADD::zeroDrain,ZADD::oneDrain);
  return ZADDFunction(::change(rootNode,var),var);
}

ZADDFunction ZADDFunction::change(const ZADDFunction& cubeSet) const {
  return ZADDFunction(::change(rootNode,varSet,cubeSet.rootNode),varSet);
}

ZADDFunction ZADDFunction::exists(const unsigned long index) const {
  if(index>=ZADDVarOrder::size()) return ZADDFunction(rootNode);

  ZADDBaseNode* tempVar=findOrAdd(ZADD::variableOrdering.getData(index),ZADD::zeroDrain,ZADD::oneDrain);
  ZADDFunction nrv(::exists(rootNode,tempVar));

  if(tempVar->preNodeCount==0) tempVar->garbageCollect();

  return nrv;
}

ZADDFunction ZADDFunction::forall(const ZADDFunction& varSet) const {
  ZADDFunction temp=!(*this);
  return !temp.exists(varSet);
}

ZADDFunction ZADDFunction::forall(const unsigned long index) const {
  ZADDFunction temp=!(*this);
  return !temp.exists(index);
}

// ZADDFunction ZADDFunction::equivalenceClasses(const ZADDFunction& pattern, const ZADDFunction& cubeSet) const {
//   return ZADDFunction::BDD->equivalenceClasses(*this,pattern,cubeSet);
// }
//
// ZADDFunction ZADDFunction::constrain(const ZADDFunction& g) const {
//   return ZADDFunction::BDD->constrain(*this,g);
// }

ZADDFunction ZADDFunction::sum(const ZADDFunction& varSet) const {
  return ZADDFunction(abstract(SUM_OP,ZADD::AbstractComputedTable,addition).calc(rootNode,varSet.rootNode));
}

ZADDFunction ZADDFunction::product(const ZADDFunction& varSet) const {
  return ZADDFunction(abstract(PRO_OP,ZADD::AbstractComputedTable,multiplication).calc(rootNode,varSet.rootNode));
}

ZADDFunction ZADDFunction::maximum(const ZADDFunction& varSet) const {
  return ZADDFunction(idempotentAbstract(MAX_OP,ZADD::ExtremaComputedTable,::max).calc(rootNode,varSet.rootNode));
}

double ZADDFunction::max(void) const {
  return static_cast<ZADDTerminalNode*>(::max(rootNode))->getValue();
}

ZADDFunction ZADDFunction::minimum(const ZADDFunction& varSet) const {
  return ZADDFunction(idempotentAbstract(MIN_OP,ZADD::ExtremaComputedTable,::min).calc(rootNode,varSet.rootNode));
}

double ZADDFunction::min(void) const {
  return static_cast<ZADDTerminalNode*>(::min(rootNode))->getValue();
}

ZADDBaseNode* oneOver(const ZADDBaseNode* v){
  if(v==ZADD::zeroDrain) return ZADD::zeroDrain;
  if(v->isDrain()) return findOrAddDrain(1.0/static_cast<const ZADDTerminalNode*>(v)->getValue());

  //try to get result from computed table
  if(ZADDBaseNode* comp=ZADD::ArithmeticComputedTable.find(calcPtr(ZADD::oneDrain,OVE_OP),v)) return comp;

  ZADDUniqueTable* minTable=v->getUniqueTable();
  ZADDBaseNode* w0=::oneOver(static_cast<const ZADDInnerNode*>(v)->getSucc0());
  ZADDBaseNode* w1=::oneOver(static_cast<const ZADDInnerNode*>(v)->getSucc1());

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);

  //insert result into computed table
  ZADD::ArithmeticComputedTable.insert(calcPtr(ZADD::oneDrain,OVE_OP),v,node);

  return node;
}

ZADDFunction ZADDFunction::oneOver(void) const {
  return ZADDFunction(::oneOver(rootNode));
}

ZADDFunction ZADDFunction::compose(const unsigned long index, const ZADDFunction& function) const {
  if(index>=ZADDVarOrder::size()) return (*this);
  else return ITE(function,cofactor(index,1),cofactor(index,0));
}

template <int N>
ZADDBaseNode* moveUp(const ZADDBaseNode* u){
  if(u->isDrain()) return const_cast<ZADDBaseNode*>(u);

  if(ZADDBaseNode* comp=ZADD::MoveComputedTable.find(calcPtr(u,MUP_OP),0,reinterpret_cast<ZADDBaseNode*>(N))) return comp;

  ZADDBaseNode* node=findOrAdd(ZADD::variableOrdering.getData(u->getLevel()-N),moveUp<N>(static_cast<const ZADDInnerNode*>(u)->getSucc0()),moveUp<N>(static_cast<const ZADDInnerNode*>(u)->getSucc1()));
  ZADD::MoveComputedTable.insert(calcPtr(u,MUP_OP),0,reinterpret_cast<ZADDBaseNode*>(N),node);

  return node;
}

ZADDFunction ZADDFunction::moveUp() const {
  if(rootNode->isDrain) return ZADDFunction(rootNode,varSet);

  return ZADDFunction(::moveUp<1>(rootNode),::moveUp<1>(varSet));
}

template <int N>
ZADDBaseNode* moveUp(const ZADDBaseNode* u, const ZADDBaseNode* v){
  if(u->isDrain()) return const_cast<ZADDBaseNode*>(u);

  unsigned long uLevel=u->getLevel();
  while(v->getLevel()<uLevel) v=static_cast<const ZADDInnerNode*>(v)->getSucc1();

  if(v->isDrain()) return const_cast<ZADDBaseNode*>(u);
  if((uLevel<N)&&(uLevel==v->getLevel())) return const_cast<ZADDBaseNode*>(u);

  if(ZADDBaseNode* comp=ZADD::MoveComputedTable.find(calcPtr(u,MUP_OP),v,reinterpret_cast<ZADDBaseNode*>(N))) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;
  ZADDBaseNode* node;
  if(uLevel==v->getLevel()){
    w0=moveUp<N>(static_cast<const ZADDInnerNode*>(u)->getSucc0(),static_cast<const ZADDInnerNode*>(v)->getSucc1());
    w1=moveUp<N>(static_cast<const ZADDInnerNode*>(u)->getSucc1(),static_cast<const ZADDInnerNode*>(v)->getSucc1());

    node=findOrAdd(ZADD::variableOrdering.getData(uLevel-N),w0,w1);
  } else {
    w0=moveUp<N>(static_cast<const ZADDInnerNode*>(u)->getSucc0(),v);
    w1=moveUp<N>(static_cast<const ZADDInnerNode*>(u)->getSucc1(),v);

    if((w0->getLevel()<=uLevel)||(w1->getLevel()<=uLevel)){
      if(w0->preNodeCount==0) w0->garbageCollect();
      if(w1->preNodeCount==0) w1->garbageCollect();
      node=const_cast<ZADDBaseNode*>(u);
    } else node=findOrAdd(ZADD::variableOrdering.getData(uLevel),w0,w1);
  }

  ZADD::MoveComputedTable.insert(calcPtr(u,MUP_OP),v,reinterpret_cast<ZADDBaseNode*>(N),node);

  return node;
}

ZADDFunction ZADDFunction::moveUp(const ZADDFunction& varSet) const {
  if(rootNode->isDrain()) return ZADDFunction(rootNode,varSet);

  return ZADDFunction(::moveUp<1>(rootNode,const_cast<ZADDBaseNode*>(varSet.varSet)),::moveUp<1>(varSet,const_cast<ZADDBaseNode*>(varSet.varSet)));
}

template <int N>
ZADDBaseNode* moveDown(const ZADDBaseNode* u){
  if(u->isDrain()) return const_cast<ZADDBaseNode*>(u);

  if(ZADDBaseNode* comp=ZADD::MoveComputedTable.find(calcPtr(u,MDO_OP),0,reinterpret_cast<ZADDBaseNode*>(N))) return comp;

  ZADDBaseNode* w0=moveDown<N>(static_cast<const ZADDInnerNode*>(u)->getSucc0());
  ZADDBaseNode* w1=moveDown<N>(static_cast<const ZADDInnerNode*>(u)->getSucc1());

  ZADDBaseNode* node=((w1==ZADD::zeroDrain) || (u->getLevel()+N>=ZADD::variableOrdering.size()))?static_cast<ZADDBaseNode*>(ZADD::zeroDrain):findOrAdd(ZADD::variableOrdering.getData(u->getLevel()+N),w0,w1);
  ZADD::MoveComputedTable.insert(calcPtr(u,MDO_OP),0,reinterpret_cast<ZADDBaseNode*>(N),node);

  return node;
}

ZADDFunction ZADDFunction::moveDown() const {
  if(rootNode->isDrain()) return ZADDFunction(rootNode,varSet);

  return ZADDFunction(::moveDown<1>(rootNode),::moveDown<1>(varSet));
}

template <int N>
ZADDBaseNode* moveDown(const ZADDBaseNode* u, const ZADDBaseNode* v){
  if(u->isDrain()) return const_cast<ZADDBaseNode*>(u);

  unsigned long uLevel=u->getLevel();
  while(v->getLevel()<uLevel) v=static_cast<const ZADDInnerNode*>(v)->getSucc1();

  if(v->isDrain()) return const_cast<ZADDBaseNode*>(u);

  if(ZADDBaseNode* comp=ZADD::MoveComputedTable.find(calcPtr(u,MDO_OP),v,reinterpret_cast<ZADDBaseNode*>(N))) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;
  ZADDBaseNode* node;
  if(uLevel==v->getLevel()){
    w0=moveDown<N>(static_cast<const ZADDInnerNode*>(u)->getSucc0(),static_cast<const ZADDInnerNode*>(v)->getSucc1());
    w1=moveDown<N>(static_cast<const ZADDInnerNode*>(u)->getSucc1(),static_cast<const ZADDInnerNode*>(v)->getSucc1());

    if((uLevel+N>=w0->getLevel())||(uLevel+N>=w1->getLevel())) node=findOrAdd(ZADD::variableOrdering.getData(uLevel),w0,w1);
    else node=findOrAdd(ZADD::variableOrdering.getData(uLevel+N),w0,w1);
  } else {
    w0=moveDown<N>(static_cast<const ZADDInnerNode*>(u)->getSucc0(),v);
    w1=moveDown<N>(static_cast<const ZADDInnerNode*>(u)->getSucc1(),v);

    node=findOrAdd(u->getUniqueTable(),w0,w1);
  }

  ZADD::MoveComputedTable.insert(calcPtr(u,MDO_OP),v,reinterpret_cast<ZADDBaseNode*>(N),node);

  return node;
}

ZADDFunction ZADDFunction::moveDown(const ZADDFunction& varSet) const {
  if(rootNode->isDrain()) return ZADDFunction(rootNode,varSet);

  return ZADDFunction(::moveDown<1>(rootNode,const_cast<ZADDBaseNode*>(varSet.varSet)),::moveDown<1>(varSet,const_cast<ZADDBaseNode*>(varSet.varSet)));
}


void ZADDFunction::printEssentialVariables(void) const {
  std::string ret;
  std::set<std::string> variables=ZADDFunction::getEssentialVariables();

  if((ZADD::variableOrdering.size()>0) && (variables.find(ZADD::variableOrdering.getVariable(0)->getName())!=variables.end())) ret+=ZADD::variableOrdering.getVariable(0)->getName();
  for(unsigned long i=1;i<ZADD::variableOrdering.size();++i){
    if(variables.find(ZADD::variableOrdering.getVariable(i)->getName())!=variables.end()) ret+=","+ZADD::variableOrdering.getVariable(i)->getName();
  }

  std::cout << ret << std::endl;
}

std::set<std::string> ZADDFunction::getEssentialVariables(void) const {
  std::set<std::string> variables;
  std::set<ZADDBaseNode*> visited;
  std::stack<ZADDBaseNode*> nodeStack;
  ZADDBaseNode* node;

  visited.insert(rootNode);
  nodeStack.push(rootNode);
  while(nodeStack.size()>0){
    node=nodeStack.top();
    nodeStack.pop();
    if(node && !node->isDrain()){
      ZADDBaseNode* succ0=((ZADDInnerNode*)node)->getSucc0();
      if(variables.find(ZADD::variableOrdering.getVariable(node->getLevel())->getName())==variables.end()) variables.insert(ZADD::variableOrdering.getVariable(node->getLevel())->getName());
      if(visited.find(succ0)==visited.end()){
        visited.insert(succ0);
        nodeStack.push(succ0);
      }
      ZADDBaseNode* succ1=((ZADDInnerNode*)node)->getSucc1();
      if(visited.find(succ1)==visited.end()){
        visited.insert(succ1);
        nodeStack.push(succ1);
      }
    }
  }

  return variables;
}

ZADDBaseNode* getAssignment(const ZADDBaseNode* node){
  if(node->isDrain()) return (node!=ZADD::zeroDrain)?ZADD::oneDrain:ZADD::zeroDrain;

  ZADDBaseNode* tempNode;
  unsigned long currentLevel=node->getLevel();

  if(static_cast<const ZADDInnerNode*>(node)->getSucc0()!=ZADD::zeroDrain){
    tempNode=findOrAdd(node->getUniqueTable(),getAssignment(static_cast<const ZADDInnerNode*>(node)->getSucc0()),ZADD::zeroDrain);
  } else {
    tempNode=findOrAdd(node->getUniqueTable(),ZADD::zeroDrain,getAssignment(static_cast<const ZADDInnerNode*>(node)->getSucc1()));
  }

  unsigned long tempNodeLevel=tempNode->getLevel();
  while(currentLevel>tempNodeLevel){
    tempNode=findOrAdd(ZADD::variableOrdering.getData(--tempNodeLevel),tempNode,ZADD::zeroDrain);
  }

  return tempNode;
}

ZADDFunction ZADDFunction::getAssignment(void) const {
  ZADDBaseNode* node=::getAssignment(rootNode);
  unsigned long level=node->getLevel();
  while(level>0){
    node=findOrAdd(ZADD::variableOrdering.getData(--level),node,ZADD::zeroDrain);
  }
  return ZADDFunction(node);
}

bool getAssignment(const ZADDBaseNode* node, char* buffer){
  if(node->isDrain()) return (node!=ZADD::zeroDrain);

  if(!getAssignment(static_cast<const ZADDInnerNode*>(node)->getSucc0(),buffer)){
    getAssignment(static_cast<const ZADDInnerNode*>(node)->getSucc1(),buffer);
    buffer[node->getLevel()]='1';
  }

  return true;
}

void ZADDFunction::getAssignment(char* buffer) const {
  //initially, every variable gets the value false
  for(unsigned long i=0;i<ZADD::variableOrdering.size();++i) buffer[i]='0';

  ::getAssignment(rootNode,buffer);
}

double countAssignments(const ZADDBaseNode* node, map<const ZADDBaseNode*,double>& CT){
  if(node->isDrain()) return (node==ZADD::zeroDrain)?0.0:1.0;

  unsigned long diff;
  double ret1, ret2;

  if(CT.find(node)!=CT.end()) return CT[node];

  {
    ZADDBaseNode* tempNode=static_cast<const ZADDInnerNode*>(node)->getSucc0();
    ret1=countAssignments(tempNode,CT);
    diff=((!tempNode || tempNode->isDrain())?ZADD::variableOrdering.size():tempNode->getLevel())-node->getLevel()-1;
  }
  ret1*=(1<<diff);

  {
    ZADDBaseNode* tempNode=static_cast<const ZADDInnerNode*>(node)->getSucc1();
    ret2=countAssignments(tempNode,CT);
    diff=((!tempNode || tempNode->isDrain())?ZADD::variableOrdering.size():tempNode->getLevel())-node->getLevel()-1;
  }
  ret2*=(1<<diff);

  CT[node]=ret1+ret2;

  return ret1+ret2;
}

double ZADDFunction::countAssignments(void) const {
  map<const ZADDBaseNode*,double> CT;
  double ret=::countAssignments(rootNode,CT);

  unsigned long diff=rootNode->isDrain()?ZADD::variableOrdering.size():rootNode->getLevel();
  while(diff){
    ret*=2.0;
    --diff;
  }

  return ret;
}

void printPaths(const unsigned long lastLevel, const ZADDBaseNode* node, char* minterm){
  unsigned long i;

  if(!node->isDrain()){
    for(i=lastLevel;i<node->getLevel();++i) minterm[i]='-';
    minterm[node->getLevel()]='0';
    printPaths(node->getLevel()+1,static_cast<const ZADDInnerNode*>(node)->getSucc0(),minterm);
    minterm[node->getLevel()]='1';
    printPaths(node->getLevel()+1,static_cast<const ZADDInnerNode*>(node)->getSucc1(),minterm);
  } else {
    if(node!=ZADD::zeroDrain){
      for(i=lastLevel;i<ZADD::variableOrdering.size();++i) minterm[i]='-';
      for(i=0;i<ZADD::variableOrdering.size();++i) std::cout << minterm[i];
      std::cout << "|" << static_cast<const ZADDTerminalNode*>(node)->getValue() << std::endl;
    }
  }
}

void ZADDFunction::printPaths(void) const {
  char minterm[ZADD::variableOrdering.size()];

  ::printPaths(0,rootNode,minterm);
}

double countPaths(const ZADDBaseNode* node, map<const ZADDBaseNode*,double>& CT){
  if(!node->isDrain()){
    double ret=0.0;
    if(CT.find(node)!=CT.end()) return CT[node];
    else {
      ret=countPaths(static_cast<const ZADDInnerNode*>(node)->getSucc0(),CT);
      ret+=countPaths(static_cast<const ZADDInnerNode*>(node)->getSucc1(),CT);
    }

    CT[node]=ret;

    return ret;
  } else return (node!=ZADD::zeroDrain)?1.0:0.0;
}

double ZADDFunction::countPaths(void) const {
  map<const ZADDBaseNode*,double> CT;
  return ::countPaths(rootNode,CT);
}

ZADDFunction operator-(const ZADDFunction& function){
  return ZADDFunction(::subtraction(ZADD::zeroDrain,ZADD::oneDrain,function.rootNode,function.varSet),varSet);
}

ZADDBaseNode* ITE(const ZADDBaseNode* u, const ZADDBaseNode* v1, const ZADDBaseNode* v2){
  if(v1==v2) return const_cast<ZADDBaseNode*>(v1);
  if(u->isDrain()) return const_cast<ZADDBaseNode*>((u==ZADD::zeroDrain)?v2:v1);
  if(v2==ZADD::zeroDrain) return AND(u,v1);
  if(v1==ZADD::oneDrain) return OR(u,v2);

  unsigned long index=v1->getRealLevel();
  if(v2->getRealLevel()<index) index=v2->getRealLevel();
  if(u->getRealLevel()<index){
    const ZADDInnerNode* innerNode=static_cast<const ZADDInnerNode*>(u);
    if((innerNode->getSucc0()==ZADD::zeroDrain) && (innerNode->getSucc1()==ZADD::oneDrain)) return findOrAdd(u->getUniqueTable(),const_cast<ZADDBaseNode*>(v1),const_cast<ZADDBaseNode*>(v2));
    if((innerNode->getSucc1()==ZADD::zeroDrain) && (innerNode->getSucc0()==ZADD::oneDrain)) return findOrAdd(u->getUniqueTable(),const_cast<ZADDBaseNode*>(v2),const_cast<ZADDBaseNode*>(v1));
  }

  if(ZADDBaseNode* comp=ZADD::BoolComputedTable.find(calcPtr(u,ITE_OP),v1,v2)) return comp;

  ZADDBaseNode* u0, *u1;
  ZADDBaseNode* v10, *v11;
  ZADDBaseNode* v20, *v21;

  ZADDUniqueTable* minTable=0;
  if(u->getRealLevel()<=index){
    minTable=u->getUniqueTable();
    u0=static_cast<const ZADDInnerNode*>(u)->getSucc0();
    u1=static_cast<const ZADDInnerNode*>(u)->getSucc1();
  } else u0=u1=const_cast<ZADDBaseNode*>(u);
  if(v1->getRealLevel()==index){
    minTable=v1->getUniqueTable();
    v10=static_cast<const ZADDInnerNode*>(v1)->getSucc0();
    v11=static_cast<const ZADDInnerNode*>(v1)->getSucc1();
  } else v10=v11=const_cast<ZADDBaseNode*>(v1);
  if(v2->getRealLevel()==index) {
    minTable=v2->getUniqueTable();
    v20=static_cast<const ZADDInnerNode*>(v2)->getSucc0();
    v21=static_cast<const ZADDInnerNode*>(v2)->getSucc1();
  } else v20=v21=const_cast<ZADDBaseNode*>(v2);

  u0=ITE(u0,v10,v20);
  u1=ITE(u1,v11,v21);

  ZADDBaseNode* node=(u0==u1)?u0:findOrAdd(minTable,u0,u1);

  //insert result into computed table
  ZADD::BoolComputedTable.insert(calcPtr(u,ITE_OP),v1,v2,node);

  return node;
}

ZADDFunction ITE(const ZADDFunction& i, const ZADDFunction& t, const ZADDFunction& e){
  return ((i&t)|(!i&e));
}

ZADDFunction LESS(const ZADDFunction& u, const ZADDFunction& v){
  return ZADDFunction(::less(u.rootNode,v.rootNode));
}

ZADDBaseNode* lessEqual(const ZADDBaseNode* u, const ZADDBaseNode* v){
  if(u==v) return ZADD::oneDrain;
  if(u->isDrain() && v->isDrain()) return (static_cast<const ZADDTerminalNode*>(u)->getValue()<=static_cast<const ZADDTerminalNode*>(v)->getValue())?ZADD::oneDrain:ZADD::zeroDrain;

  if(ZADDBaseNode* comp=ZADD::CompareComputedTable.find(calcPtr(u,LEE_OP),v)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;

  ZADDUniqueTable* minTable;

  if(u->getRealLevel()<=v->getRealLevel()){
    minTable=u->getUniqueTable();
    if(v->getUniqueTable()==minTable){
      w0=::lessEqual(static_cast<const ZADDInnerNode*>(u)->getSucc0(),static_cast<const ZADDInnerNode*>(v)->getSucc0());
      w1=::lessEqual(static_cast<const ZADDInnerNode*>(u)->getSucc1(),static_cast<const ZADDInnerNode*>(v)->getSucc1());
    } else {
      w0=::lessEqual(static_cast<const ZADDInnerNode*>(u)->getSucc0(),v);
      w1=::lessEqual(static_cast<const ZADDInnerNode*>(u)->getSucc1(),v);
    }
  } else {
    minTable=v->getUniqueTable();
    w0=::lessEqual(u,static_cast<const ZADDInnerNode*>(v)->getSucc0());
    w1=::lessEqual(u,static_cast<const ZADDInnerNode*>(v)->getSucc1());
  }

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);
  ZADD::CompareComputedTable.insert(calcPtr(u,LEE_OP),v,node);

  return node;
}

ZADDFunction LESS_EQUAL(const ZADDFunction& u, const ZADDFunction& v){
  return ZADDFunction(::lessEqual(u.rootNode,v.rootNode));
}

ZADDBaseNode* equal(const ZADDBaseNode* u, const ZADDBaseNode* v){
  if(u==v) return ZADD::oneDrain;
  if(u->isDrain() && v->isDrain()) return (static_cast<const ZADDTerminalNode*>(u)->getValue()==static_cast<const ZADDTerminalNode*>(v)->getValue())?ZADD::oneDrain:ZADD::zeroDrain;

  if(ZADDBaseNode* comp=ZADD::CompareComputedTable.find(calcPtr(u,EQU_OP),v)) return comp;

  ZADDBaseNode* w0;
  ZADDBaseNode* w1;

  ZADDUniqueTable* minTable;

  if(u->getRealLevel()<=v->getRealLevel()){
    minTable=u->getUniqueTable();
    if(v->getUniqueTable()==minTable){
      w0=::equal(static_cast<const ZADDInnerNode*>(u)->getSucc0(),static_cast<const ZADDInnerNode*>(v)->getSucc0());
      w1=::equal(static_cast<const ZADDInnerNode*>(u)->getSucc1(),static_cast<const ZADDInnerNode*>(v)->getSucc1());
    } else {
      w0=::equal(static_cast<const ZADDInnerNode*>(u)->getSucc0(),v);
      w1=::equal(static_cast<const ZADDInnerNode*>(u)->getSucc1(),v);
    }
  } else {
    minTable=v->getUniqueTable();
    w0=::equal(u,static_cast<const ZADDInnerNode*>(v)->getSucc0());
    w1=::equal(u,static_cast<const ZADDInnerNode*>(v)->getSucc1());
  }

  ZADDBaseNode* node=(w1==ZADD::zeroDrain)?w0:findOrAdd(minTable,w0,w1);
  ZADD::CompareComputedTable.insert(calcPtr(u,EQU_OP),v,node);

  return node;
}

ZADDFunction EQUAL(const ZADDFunction& u, const ZADDFunction& v){
  return ZADDFunction(::equal(u.rootNode,v.rootNode));
}

ZADDFunction MAXIMUM(const ZADDFunction& u, const ZADDFunction& v){
  return ZADDFunction(::max(u.rootNode,u.varSet,v.rootNode,v.varSet));
}

ZADDFunction MINIMUM(const ZADDFunction& u, const ZADDFunction& v){
  return ZADDFunction(::min(u.rootNode,u.varSet,v.rootNode,v.varSet));
}

ZADDBaseNode* MM(const ZADDBaseNode* A, const ZADDBaseNode* B, const ZADDBaseNode* varSet, const unsigned long m){
  if((A==ZADD::zeroDrain) || (B==ZADD::zeroDrain)) return ZADD::zeroDrain;
  if((A->isDrain() && B->isDrain()) || (m==0)){
    ZADDTerminalNode* tempDrain=findOrAddDrain(static_cast<double>(1<<m));
    ZADDBaseNode* result=::multiplication(::multiplication(A,B),static_cast<ZADDBaseNode*>(tempDrain));

    if(result!=tempDrain && tempDrain->preNodeCount==0) tempDrain->garbageCollect();

    return result;
  }

  if(ZADDBaseNode* comp=ZADD::MatrixComputedTable.find(calcPtr(A,MMA_OP),B,varSet)) return comp;

  ZADDUniqueTable* uniqueTable=varSet->getUniqueTable();
  ZADDUniqueTable* nextUniqueTable=ZADD::variableOrdering.getData(uniqueTable->getLevel()+1);

  ZADDBaseNode* A00, *A01, *A10, *A11;
  if(uniqueTable==A->getUniqueTable()){
    ZADDBaseNode* A0=static_cast<const ZADDInnerNode*>(A)->getSucc0();
    ZADDBaseNode* A1=static_cast<const ZADDInnerNode*>(A)->getSucc1();
    if(A0->getUniqueTable()==nextUniqueTable){
      A00=static_cast<ZADDInnerNode*>(A0)->getSucc0();
      A01=static_cast<ZADDInnerNode*>(A0)->getSucc1();
    } else {
      A00=A01=A0;
    }
    if(A1->getUniqueTable()==nextUniqueTable){
      A10=static_cast<ZADDInnerNode*>(A1)->getSucc0();
      A11=static_cast<ZADDInnerNode*>(A1)->getSucc1();
    } else {
      A10=A11=A1;
    }
  } else {
    if(A->getUniqueTable()==nextUniqueTable){
      A00=A10=static_cast<const ZADDInnerNode*>(A)->getSucc0();
      A01=A11=static_cast<const ZADDInnerNode*>(A)->getSucc1();
    } else {
      A00=A01=A10=A11=const_cast<ZADDBaseNode*>(A);
    }
  }

  ZADDBaseNode* B00, *B01, *B10, *B11;
  if(uniqueTable==B->getUniqueTable()){
    ZADDBaseNode* B0=static_cast<const ZADDInnerNode*>(B)->getSucc0();
    ZADDBaseNode* B1=static_cast<const ZADDInnerNode*>(B)->getSucc1();
    if(B0->getUniqueTable()==nextUniqueTable){
      B00=static_cast<ZADDInnerNode*>(B0)->getSucc0();
      B01=static_cast<ZADDInnerNode*>(B0)->getSucc1();
    } else {
      B00=B01=B0;
    }
    if(B1->getUniqueTable()==nextUniqueTable){
      B10=static_cast<ZADDInnerNode*>(B1)->getSucc0();
      B11=static_cast<ZADDInnerNode*>(B1)->getSucc1();
    } else {
      B10=B11=B1;
    }
  } else {
    if(B->getUniqueTable()==nextUniqueTable){
      B00=B10=static_cast<const ZADDInnerNode*>(B)->getSucc0();
      B01=B11=static_cast<const ZADDInnerNode*>(B)->getSucc1();
    } else {
      B00=B01=B10=B11=const_cast<ZADDBaseNode*>(B);
    }
  }

  ZADDBaseNode* newVarSet=static_cast<const ZADDInnerNode*>(varSet)->getSucc1();

  ZADDBaseNode* w0000=::MM(A00,B00,newVarSet,m-1);
  ZADDBaseNode* w0110=::MM(A01,B10,newVarSet,m-1);
  ZADDBaseNode* w0001=::MM(A00,B01,newVarSet,m-1);
  ZADDBaseNode* w0111=::MM(A01,B11,newVarSet,m-1);
  ZADDBaseNode* w1000=::MM(A10,B00,newVarSet,m-1);
  ZADDBaseNode* w1110=::MM(A11,B10,newVarSet,m-1);
  ZADDBaseNode* w1001=::MM(A10,B01,newVarSet,m-1);
  ZADDBaseNode* w1111=::MM(A11,B11,newVarSet,m-1);

  ZADDBaseNode* c00=addition(w0000,w0110);
  ZADDBaseNode* c01=addition(w0001,w0111);
  ZADDBaseNode* c10=addition(w1000,w1110);
  ZADDBaseNode* c11=addition(w1001,w1111);

  ZADDBaseNode* result0=(c00==c01)?c00:findOrAdd(nextUniqueTable,c00,c01);
  ZADDBaseNode* result1=(c10==c11)?c11:findOrAdd(nextUniqueTable,c10,c11);
  ZADDBaseNode* result=(result0==result1)?result1:findOrAdd(uniqueTable,result0,result1);

  if(result!=w0000 && w0000->preNodeCount==0) w0000->garbageCollect();
  if(result!=w0110 && w0110->preNodeCount==0) w0110->garbageCollect();
  if(result!=w0001 && w0001->preNodeCount==0) w0001->garbageCollect();
  if(result!=w0111 && w0111->preNodeCount==0) w0111->garbageCollect();
  if(result!=w1000 && w1000->preNodeCount==0) w1000->garbageCollect();
  if(result!=w1110 && w1110->preNodeCount==0) w1110->garbageCollect();
  if(result!=w1001 && w1001->preNodeCount==0) w1001->garbageCollect();
  if(result!=w1111 && w1111->preNodeCount==0) w1111->garbageCollect();
  if(result!=c00 && c00->preNodeCount==0) c00->garbageCollect();
  if(result!=c01 && c01->preNodeCount==0) c01->garbageCollect();
  if(result!=c10 && c10->preNodeCount==0) c10->garbageCollect();
  if(result!=c11 && c11->preNodeCount==0) c11->garbageCollect();

  ZADD::MatrixComputedTable.insert(calcPtr(A,MMA_OP),B,varSet,result);

  return result;
}

ZADDFunction MM(const ZADDFunction& A, const ZADDFunction& B, const ZADDFunction& varSet){
  ZADDBaseNode* node=varSet.rootNode;

  unsigned long m=0;
  while(!node->isDrain()){
    node=static_cast<ZADDInnerNode*>(node)->getSucc1();
    ++m;
  }

  return ZADDFunction(::MM(A.rootNode,B.rootNode,varSet.rootNode,m));
}

ZADDBaseNode* MV(const ZADDBaseNode* A, const ZADDBaseNode* b, const ZADDBaseNode* varSet, const unsigned long m){
  if((A==ZADD::zeroDrain) || (b==ZADD::zeroDrain)) return ZADD::zeroDrain;
  if((A->isDrain() && b->isDrain()) || (m==0)){
    ZADDTerminalNode* tempDrain=findOrAddDrain(static_cast<double>(1<<m));
    ZADDBaseNode* result=::multiplication(::multiplication(A,b),static_cast<ZADDBaseNode*>(tempDrain));

    if(result!=tempDrain && tempDrain->preNodeCount==0) tempDrain->garbageCollect();

    return result;
  }

  if(ZADDBaseNode* comp=ZADD::MatrixComputedTable.find(calcPtr(A,MAV_OP),b,varSet)) return comp;

  ZADDUniqueTable* uniqueTable=varSet->getUniqueTable();
  ZADDUniqueTable* nextUniqueTable=ZADD::variableOrdering.getData(uniqueTable->getLevel()+1);

  ZADDBaseNode* A00, *A01, *A10, *A11;
  if(uniqueTable==A->getUniqueTable()){
    ZADDBaseNode* A0=static_cast<const ZADDInnerNode*>(A)->getSucc0();
    ZADDBaseNode* A1=static_cast<const ZADDInnerNode*>(A)->getSucc1();
    if(A0->getUniqueTable()==nextUniqueTable){
      A00=static_cast<ZADDInnerNode*>(A0)->getSucc0();
      A01=static_cast<ZADDInnerNode*>(A0)->getSucc1();
    } else {
      A00=A01=A0;
    }
    if(A1->getUniqueTable()==nextUniqueTable){
      A10=static_cast<ZADDInnerNode*>(A1)->getSucc0();
      A11=static_cast<ZADDInnerNode*>(A1)->getSucc1();
    } else {
      A10=A11=A1;
    }
  } else {
    if(A->getUniqueTable()==nextUniqueTable){
      A00=A10=static_cast<const ZADDInnerNode*>(A)->getSucc0();
      A01=A11=static_cast<const ZADDInnerNode*>(A)->getSucc1();
    } else {
      A00=A01=A10=A11=const_cast<ZADDBaseNode*>(A);
    }
  }

  ZADDBaseNode* b0, *b1;
  if(uniqueTable==b->getUniqueTable()){
    b0=static_cast<const ZADDInnerNode*>(b)->getSucc0();
    b1=static_cast<const ZADDInnerNode*>(b)->getSucc1();
  } else {
    b0=b1=const_cast<ZADDBaseNode*>(b);
  }

  ZADDBaseNode* newVarSet=static_cast<const ZADDInnerNode*>(varSet)->getSucc1();

  ZADDBaseNode* w00=::MV(A00,b0,newVarSet,m-1);
  ZADDBaseNode* w01=::MV(A01,b1,newVarSet,m-1);
  ZADDBaseNode* w10=::MV(A10,b0,newVarSet,m-1);
  ZADDBaseNode* w11=::MV(A11,b1,newVarSet,m-1);

  ZADDBaseNode* c0=addition(w00,w01);
  ZADDBaseNode* c1=addition(w10,w11);

  ZADDBaseNode* result=(c0==c1)?c1:findOrAdd(uniqueTable,c0,c1);

  if(result!=w00 && w00->preNodeCount==0) w00->garbageCollect();
  if(result!=w01 && w01->preNodeCount==0) w01->garbageCollect();
  if(result!=w10 && w10->preNodeCount==0) w10->garbageCollect();
  if(result!=w11 && w11->preNodeCount==0) w11->garbageCollect();
  if(result!=c0 && c0->preNodeCount==0) c0->garbageCollect();
  if(result!=c1 && c1->preNodeCount==0) c1->garbageCollect();

  ZADD::MatrixComputedTable.insert(calcPtr(A,MAV_OP),b,varSet,result);

  return result;
}

ZADDFunction MV(const ZADDFunction& A, const ZADDFunction& b, const ZADDFunction& varSet){
  ZADDBaseNode* node=varSet.rootNode;

  unsigned long m=0;
  while(!node->isDrain()){
    node=static_cast<ZADDInnerNode*>(node)->getSucc1();
    ++m;
  }

  return ZADDFunction(::MV(A.rootNode,b.rootNode,varSet.rootNode,m));
}

ZADDBaseNode* VM(const ZADDBaseNode* b, const ZADDBaseNode* A, const ZADDBaseNode* varSet, const unsigned long m){
  if((A==ZADD::zeroDrain) || (b==ZADD::zeroDrain)) return ZADD::zeroDrain;
  if((A->isDrain() && b->isDrain()) || (m==0)){
    ZADDTerminalNode* tempDrain=findOrAddDrain(static_cast<double>(1<<m));
    ZADDBaseNode* result=::multiplication(::multiplication(A,b),static_cast<ZADDBaseNode*>(tempDrain));

    if(result!=tempDrain && tempDrain->preNodeCount==0) tempDrain->garbageCollect();

    return result;
  }

  if(ZADDBaseNode* comp=ZADD::MatrixComputedTable.find(calcPtr(b,VMA_OP),A,varSet)) return comp;

  ZADDUniqueTable* uniqueTable=varSet->getUniqueTable();
  ZADDUniqueTable* nextUniqueTable=ZADD::variableOrdering.getData(uniqueTable->getLevel()+1);

  ZADDBaseNode* A00, *A01, *A10, *A11;
  if(uniqueTable==A->getUniqueTable()){
    ZADDBaseNode* A0=static_cast<const ZADDInnerNode*>(A)->getSucc0();
    ZADDBaseNode* A1=static_cast<const ZADDInnerNode*>(A)->getSucc1();
    if(A0->getUniqueTable()==nextUniqueTable){
      A00=static_cast<ZADDInnerNode*>(A0)->getSucc0();
      A01=static_cast<ZADDInnerNode*>(A0)->getSucc1();
    } else {
      A00=A01=A0;
    }
    if(A1->getUniqueTable()==nextUniqueTable){
      A10=static_cast<ZADDInnerNode*>(A1)->getSucc0();
      A11=static_cast<ZADDInnerNode*>(A1)->getSucc1();
    } else {
      A10=A11=A1;
    }
  } else {
    if(A->getUniqueTable()==nextUniqueTable){
      A00=A10=static_cast<const ZADDInnerNode*>(A)->getSucc0();
      A01=A11=static_cast<const ZADDInnerNode*>(A)->getSucc1();
    } else {
      A00=A01=A10=A11=const_cast<ZADDBaseNode*>(A);
    }
  }

  ZADDBaseNode* b0, *b1;
  if(uniqueTable==b->getUniqueTable()){
    b0=static_cast<const ZADDInnerNode*>(b)->getSucc0();
    b1=static_cast<const ZADDInnerNode*>(b)->getSucc1();
  } else {
    b0=b1=const_cast<ZADDBaseNode*>(b);
  }

  ZADDBaseNode* newVarSet=static_cast<const ZADDInnerNode*>(varSet)->getSucc1();

  ZADDBaseNode* w00=::VM(b0,A00,newVarSet,m-1);
  ZADDBaseNode* w01=::VM(b0,A01,newVarSet,m-1);
  ZADDBaseNode* w10=::VM(b1,A10,newVarSet,m-1);
  ZADDBaseNode* w11=::VM(b1,A11,newVarSet,m-1);

  ZADDBaseNode* c0=addition(w00,w10);
  ZADDBaseNode* c1=addition(w01,w11);

  ZADDBaseNode* result=(c0==c1)?c1:findOrAdd(uniqueTable,c0,c1);

  if(result!=w00 && w00->preNodeCount==0) w00->garbageCollect();
  if(result!=w01 && w01->preNodeCount==0) w01->garbageCollect();
  if(result!=w10 && w10->preNodeCount==0) w10->garbageCollect();
  if(result!=w11 && w11->preNodeCount==0) w11->garbageCollect();
  if(result!=c0 && c0->preNodeCount==0) c0->garbageCollect();
  if(result!=c1 && c1->preNodeCount==0) c1->garbageCollect();

  ZADD::MatrixComputedTable.insert(calcPtr(b,VMA_OP),A,varSet,result);

  return result;
}

ZADDFunction VM(const ZADDFunction& A, const ZADDFunction& b, const ZADDFunction& varSet){
  ZADDBaseNode* node=varSet.rootNode;

  unsigned long m=0;
  while(!node->isDrain()){
    node=static_cast<ZADDInnerNode*>(node)->getSucc1();
    ++m;
  }

  return ZADDFunction(::VM(b.rootNode,A.rootNode,varSet.rootNode,m));
}



std::set<ZADDBaseNode*> ZADDFunction::getNodes() const {
  std::set<ZADDBaseNode*> visited;
  std::stack<const ZADDBaseNode*> nodeStack;

  visited.insert(rootNode);
  nodeStack.push(rootNode);
  while(nodeStack.size()>0){
    const ZADDBaseNode* node=nodeStack.top();
    nodeStack.pop();
    if LIKELY(!node->isDrain()){
      {
        ZADDBaseNode* tempNode=static_cast<const ZADDInnerNode*>(node)->getSucc0();
        if(visited.find(tempNode)==visited.end()){
          visited.insert(tempNode);
          nodeStack.push(tempNode);
        }
      }
      {
        ZADDBaseNode* tempNode=static_cast<const ZADDInnerNode*>(node)->getSucc1();
        if(visited.find(tempNode)==visited.end()){
          visited.insert(tempNode);
          nodeStack.push(tempNode);
        }
      }
    }
  }

  return visited;
}
