/*
 * This file is part of the program ltl2dstar (http://www.ltl2dstar.de/).
 * Copyright (C) 2005-2007 Joachim Klein <j.klein@ltl2dstar.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 *  published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef DPA_H
#define DPA_H

/** @file
 * Provide class DPA<...> which can store a deterministic Parity automaton
 */

#include "DA.hpp"
#include "DA_State.hpp"
#include "ParityAcceptance.hpp"
#include "APSet.hpp"
#include "APElement.hpp"

#include "ext/ltl2dstar/common/Index.hpp"
#include <iterator>
#include <iostream>
#include <map>
#include <string>

#include "boost/shared_ptr.hpp"

namespace ltl2dstar {

/**
 * A class representing a deterministic Parity automaton.
 * <p>
 * For details on the template parameters, see class DA.
 * </p>
 */
template <typename Label, template <typename N> class EdgeContainer >
class DPA : public DA<Label, EdgeContainer, ParityAcceptance> {
 public:
  DPA(APSet_cp ap_set, const ParityAcceptance::Initializer& init=ParityAcceptance::getStandardInitializer());
  virtual ~DPA();
  
  /** The type of the states of the DPA. */
  typedef typename DA<Label,EdgeContainer,ParityAcceptance>::state_type state_type;

  /** The type of the label on the edges. */
  typedef typename DA<Label,EdgeContainer,ParityAcceptance>::label_type label_type;

  /** The type of an iterator over the states. */
  typedef typename DA<Label,EdgeContainer,ParityAcceptance>::iterator iterator;

  /** The type of an iterator over the edges of a state. */
  typedef typename DA<Label,EdgeContainer,ParityAcceptance>::edge_iterator edge_iterator;

  typedef ParityAcceptance acceptance_condition_type;

  /** Type of a reference counted pointer to the DPA (boost::shared_ptr) */
  typedef boost::shared_ptr< DPA<Label,EdgeContainer> > shared_ptr;

  /** Create a new instance of the automaton. */
  virtual DA<Label,EdgeContainer,ParityAcceptance> *createInstance(APSet_cp ap_set) {
    return new DPA<Label,EdgeContainer>(ap_set);
  }

  virtual DA<Label,EdgeContainer,ParityAcceptance> *createInstance(APSet_cp ap_set,
								   const ParityAcceptance::Initializer& init) {
    return new DPA<Label,EdgeContainer>(ap_set,init);
  }

  /** Make this automaton into an never accepting automaton */
  void constructEmpty() {
    state_type* n=DA<Label,EdgeContainer,ParityAcceptance>::newState();
    n->acceptance().setRank(1);
    setStartState(n);
    
    for (APSet::element_iterator it_elem=
	   DA<Label,EdgeContainer,ParityAcceptance>::getAPSet().all_elements_begin();
	 it_elem!=DA<Label,EdgeContainer,ParityAcceptance>::getAPSet().all_elements_end();
	 ++it_elem) {
      APElement label=(*it_elem);
      n->edges().addEdge(label, n);
    }
  }

 
  /**
   * Print the DPA in v2 format to the output stream.
   * This function can compact the automaton, which may invalidate iterators!
   */
  void print_explicit_v2(std::ostream& out) {
    if (!this->isCompact()) {
      this->makeCompact();
    }

    this->print_da_explicit_v2(typeID(),
			       out);
  }

  /**
   * Print the DPA in DOT format to the output stream.
   * This function can compact the automaton, which may invalidate iterators!
   */
  void print_dot(std::ostream& out) {
    if (!this->isCompact()) {
      this->makeCompact();
    }

    this->print_da_dot(typeID(),
		       out);
  }

  /**
   * Output operator for the DPA.
   */
  friend std::ostream& operator<<(std::ostream& out, DPA& dpa) {
    dpa.print_explicit_v2(out);
    return out;
  }


  virtual void formatStateForDOT(std::ostream& out, unsigned int state_index) {
    state_type* cur_state=this->get(state_index);

    bool is_html=false;

    ParityAcceptance::parity_rank_t rank=cur_state->acceptance().getRank();
    bool is_odd= (rank % 2) == 1;
    std::string rank_description=boost::lexical_cast<std::string>(rank);
    
    std::string label;

    if (cur_state->hasDescription()) {
      std::string description=cur_state->getDescription();
      is_html=true;
      
      StringAlgorithms::replace_all(description, 
				    "<TABLE>", "<TABLE BORDER=\"0\" CELLBORDER=\"1\">");

      out << "label= < ";
      out << "<TABLE BORDER=\"0\"><TR><TD BORDER=\"1\">";
      out << state_index;
      out << "</TD><TD " << (is_odd ? "BGCOLOR=\"red\"" : "BGCOLOR=\"green\"" ) << ">";
      out << rank_description;
      out << "</TD></TR><TR><TD COLSPAN=\"2\">";
      out << description;
      out << "</TD></TR></TABLE>";
      
      out << " >, shape=box";
      
    } else {
      out << "label= \"" << state_index;
      out  << " r=" << rank_description;
      out << "\"";
    }
    
    out << ", shape=box";
      
    if (this->getStartState() == cur_state) {
      out << ", style=filled, color=black, fillcolor=grey";
    }
  }

  /**
   * Optimizes the acceptance condition.
   */
  void optimizeAcceptanceCondition() {
    DA<Label,EdgeContainer,ParityAcceptance>::acceptance().optimize(*this);
  }

  void complement() {
    DA<Label,EdgeContainer,ParityAcceptance>::acceptance().complement();
  }
  
 private:
  /** Return a string giving the type of the automaton. */
  std::string typeID() const {
    return std::string("DPA");
  }
  
};


/**
 * Constructor.
 * @param ap_set the underlying APSet
 */
template <typename Label, template <typename N> class EdgeContainer >
DPA<Label, EdgeContainer>::DPA(APSet_cp ap_set, const ParityAcceptance::Initializer& init)
  : DA<Label,EdgeContainer,ParityAcceptance>(ap_set, init) {
}


/**
 * Destructor.
 */
template <typename Label, template <typename N> class EdgeContainer >
DPA<Label, EdgeContainer>::~DPA() {

}

}
#endif
