/*
 * This file is part of the program ltl2dstar (http://www.ltl2dstar.de/).
 * Copyright (C) 2005-2007 Joachim Klein <j.klein@ltl2dstar.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 *  published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef PARITYACCEPTANCE_H
#define PARITYACCEPTANCE_H

#include "vereofy/common/BitSet.hpp"

#include "RabinAcceptance.hpp"
#include "GraphAlgorithms.hpp"

#include <iostream>
#include <string>
#include <vector>
#include <limits>
#include <cassert>

namespace ltl2dstar {

class ParityAcceptance {
public:
  typedef unsigned int parity_rank_t;
  
  const static unsigned int INVALID_RANK;

  ParityAcceptance(parity_rank_t largest_rank_=0, 
		   bool minimum_parity_type=true) :
    _largest_rank(largest_rank_),
    _minimum_parity_type(minimum_parity_type),
    _is_compact(true) {
  }
  
  ~ParityAcceptance() {
  }


  // ---- The Acceptance Interface

  class Initializer {
  public:
    Initializer(parity_rank_t largest_rank=0, bool minimum_parity_type=true) 
      : _largest_rank(largest_rank), _minimum_parity_type(minimum_parity_type)  { }
    parity_rank_t getLargestRank() const {return _largest_rank;}
    bool isMinimumType() const {return _minimum_parity_type;}
  private:
    parity_rank_t _largest_rank;
    bool _minimum_parity_type;
  };

  static Initializer getStandardInitializer() {
    return Initializer();
  };

  static Initializer getSameSizeInitializer(const ParityAcceptance& other) {
    return Initializer(other.getMaximalUsedRank(), other.isMinimumParityType());
  }


  ParityAcceptance(const Initializer& init) :
    _largest_rank(init.getLargestRank()),
    _minimum_parity_type(init.isMinimumType()),
    _is_compact(true) {
  }

  
  bool isCompact() const {
    return _is_compact;
  }

  void makeCompact() {
    if (isCompact()) {
      return;
    }

    // do nothing
    _is_compact=true;
  }

  void complement() {
    for (unsigned int state=0;state<_ranks.size();state++) {
      setRankForState(state, getRankForState(state)+1);
    }
  }


  /** Update the acceptance condition upon renaming of states according to the mapping.
   *  Assumes that states can only get a lower name
   */
  void moveStates(std::vector<unsigned int> mapping) {
    // TODO: not-THROW
    throw std::string("moveStates not yet implemented");
    if (!isCompact()) {
      makeCompact();
    }
  }

  void outputAcceptanceHeader(std::ostream& out) const {
    out << "Ranks: "<< getMaximalUsedRank() << std::endl;
  }

  void outputAcceptanceForState(std::ostream& out, unsigned int state_index) const {
    out << "Rank: " << getRankForState(state_index) << "\n";
  }


  void addState(unsigned int state_index) {
    if (state_index >= _ranks.size()) {
      _ranks.resize(state_index+1, INVALID_RANK);
    }
  }

  class ParitySignature {
  public:
    ParitySignature() : rank(INVALID_RANK) {}
    ParitySignature(const parity_rank_t& rank_) : rank(rank_) {}

    void operator=(const parity_rank_t& rank_) {
      rank=rank_;
    }

    operator parity_rank_t() const {
      return rank;
    }

    bool operator==(const ParitySignature& other) const {
      return rank==other.rank;
    }

    bool operator<(const ParitySignature& other) const {
      return rank<other.rank;
    }

    /**
     * Calculate a hash value using HashFunction
     * @param hashfunction the HashFunction
     * @only_structure ignore the nameing of the nodes
     */
    template <class HashFunction>
    void hashCode(HashFunction& hashfunction) {
      hashfunction.hash(rank);
    }

  private:
    parity_rank_t rank;
  };

  class AcceptanceForState {
  private:
    ParityAcceptance& _acceptance;
    unsigned int _state_index;

  public:
    AcceptanceForState(ParityAcceptance& acceptance,
		       unsigned int state_index) :
      _acceptance(acceptance), _state_index(state_index) {}

    parity_rank_t getRank() const {
      return _acceptance.getRankForState(_state_index);
    }

    void setRank(parity_rank_t rank) {
      _acceptance.setRankForState(_state_index, rank);
    }

    ParitySignature getSignature() const {
      return ParitySignature(getRank());
    }

    void setSignature(const ParitySignature& signature) {
      setRank(signature);
    }
    
    RabinAcceptance::signature_type getRabinSignature() const {
      return getRabinForRank(getRank(), _acceptance.getMaximalUsedRank());
    }

    ParityAcceptance& getParityAcceptance() {
      return _acceptance;
    }
  };
  

  // ---- Parity specific

  const parity_rank_t getRankForState(unsigned int state_index) const {
    return _ranks[state_index];
  }

  void setRankForState(unsigned int state_index, parity_rank_t rank) {
    _ranks[state_index]=rank;
    if (rank>_largest_rank) {
      _largest_rank=rank;
    }
  }

  unsigned int getMaximalUsedRank() const {
    return _largest_rank;
  }

  bool isMinimumParityType() const {return _minimum_parity_type;}
  
  
  
  typedef ParitySignature signature_type;
  
  const ParitySignature getEmptySignature() {
    return ParitySignature(INVALID_RANK);
  }

  unsigned int getRabinSize() const {
    return (_largest_rank+1)/2;
  }

  static
  RabinAcceptance::RabinSignature getRabinForRank(parity_rank_t rank,
						  parity_rank_t max_rank) {
    unsigned int start_index=(rank+1)/2;
    unsigned int size=(max_rank+1)/2;

    RabinAcceptance::RabinSignature rabin(size);
    
    rabin.setColor(start_index, rank%2==1 ? RabinAcceptance::RABIN_RED : RabinAcceptance::RABIN_GREEN);
    for (unsigned int i=start_index+1;i<size;i++) {
      rabin.setColor(i, RabinAcceptance::RABIN_RED);
    }
    return rabin;
  }

  static
  RabinAcceptance::RabinSignature getNegatedRabinForRank(parity_rank_t rank,
							 parity_rank_t max_rank) {
    return getRabinForRank(rank+1, max_rank+1);
  }

  template <typename Graph>
  void optimize(Graph& graph) {
    AcceptanceOptimizer<Graph> opt(graph, *this);
    opt.optimize();    
  }

private: // members
  unsigned int _largest_rank;

  bool _minimum_parity_type;
			    
  /** A vector of parity_ranks */
  std::vector<parity_rank_t> _ranks;

  bool _is_compact;

  template <typename Graph>
  class AcceptanceOptimizer {
  public:
    AcceptanceOptimizer(Graph& graph, ParityAcceptance& acceptance) :
      _graph(graph), _acceptance(acceptance) {
      switch_min_max(acceptance._ranks, pi, acceptance._largest_rank);
      pi_.resize(pi.size(), 0);
    }

    void optimize() {
      BitSet Q;
      for (unsigned int i=0;i<pi.size();i++) {
	Q.set(i);
      }
      unsigned int m=M(Q);
      _acceptance._largest_rank=switch_min_max(pi_, _acceptance._ranks, m);
    }

    parity_rank_t M(const BitSet& P) {
      parity_rank_t max=0;

      std::cerr << "Calculating M for P = " << P <<std::endl;
      SCCs sccs;
      BitSetFilter filter(P);
      GraphAlgorithms<Graph, StdSuccessorAccess<Graph>, BitSetFilter>::calculateSCCs(_graph, sccs, true, StdSuccessorAccess<Graph>(), filter);
      
      for (unsigned int scc=0;
	   scc<sccs.countSCCs();
	   ++scc) {
	const BitSet& R=sccs[scc];
	if (R.cardinality()==1) {
	  unsigned int s=R.nextSetBit(0);
	  if (sccs.stateIsReachable(s,s)==false) {
	    // is trivial SCC (one state, no self-loop)
	  
	    continue; // ?
	  }
	}
	
	std::cerr << "R_"<< scc << " = " << R << std::endl;
	// calculate pi(R)
	parity_rank_t piR=calcPiR(R);

	std::cerr << "pi(R) = "<< piR << std::endl;

	// calculate R'
	BitSet Rprime;
	calcRprime(R, Rprime, piR);
	
	std::cerr << "R' = "<< Rprime << std::endl;

	parity_rank_t m;
	if (piR == 0) {
	  m=0;
	} else {
	  m=M(Rprime);
	  if ((piR - m) % 2) {
	    // odd
	    m=m+1;
	  }
	}

	for (BitSetIterator it=BitSetIterator(R);
	     it!=BitSetIterator::end(R);
	     ++it) {
	  unsigned int state=*it;
	  
	  // foreach R-Rprime
	  if (!Rprime.get(state)) {
	    std::cerr << "Setting pi'[" << state << "] = " << m << std::endl;
	    pi_[state]=m;
	  }
	}
	max=std::max(max, m);
      }
      
      std::cerr << "< " << max << std::endl;
      return max;
    }

    parity_rank_t calcPiR(const BitSet& R) {
      parity_rank_t piR=0;
      
      for (BitSetIterator it=BitSetIterator(R);
	   it!=BitSetIterator::end(R);
	   ++it) {
	parity_rank_t r=pi[*it];
	piR=std::max(piR, r);
      }
      return piR;
    }

    void calcRprime(const BitSet& R, BitSet& R_, parity_rank_t piR) {
      for (BitSetIterator it=BitSetIterator(R);
	   it!=BitSetIterator::end(R);
	   ++it) {
	parity_rank_t r=pi[*it];
	if (r != piR) {
	  assert(r < piR);
	  R_.set(*it);
	}
      }
    }
    
    parity_rank_t switch_min_max(const std::vector<parity_rank_t>& from,
				 std::vector<parity_rank_t>& to,
				 unsigned int largest_rank) {
      if (largest_rank % 2 == 1) {
	largest_rank++;
      }

      std::cerr << "Largest rank = " << largest_rank << std::endl;

      to.resize(from.size());
      for (unsigned int i=0;i<from.size();i++) {
	to[i]=largest_rank-from[i];
	std::cerr << "State " << i << ": " << from[i] << " -> " << to[i] << std::endl;
      }

      return largest_rank;
    }

  private:
    Graph& _graph;
    ParityAcceptance& _acceptance;
    
    std::vector<parity_rank_t> pi;
    std::vector<parity_rank_t> pi_;
  };
};

}
#endif

