/*
 * This file is part of the program ltl2dstar (http://www.ltl2dstar.de/).
 * Copyright (C) 2005-2007 Joachim Klein <j.klein@ltl2dstar.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 *  published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef RABINACCEPTANCE_H
#define RABINACCEPTANCE_H

/** @file
 * Provides class RabinAcceptance.
 */

#include "vereofy/common/BitSet.hpp"
#include <boost/iterator/filter_iterator.hpp>
#include <boost/iterator/counting_iterator.hpp>
#include <boost/lexical_cast.hpp>

#include <iostream>
#include <string>
#include <vector>
#include <cassert>

using namespace vereofy;

namespace ltl2dstar {

/**
 * Class storing a Rabin acceptance condition.
 * Contains a number k of pairs of BitSets (L_i,U_i), where the states
 * in the acceptance sets L_i and U_i are set.
 */
class RabinAcceptance {
public:
  /**
   * Constructor
   * @param number_of_initial_pairs The initial numbers of pairs to allocate
   */
  RabinAcceptance(unsigned int number_of_initial_pairs=0,
                  bool allow_automatic_growth=false) :
    _acceptance_count(0),
    _max_state_id(0),
    _is_compact(true), 
    _allow_automatic_growth(allow_automatic_growth), 
    _empty_signature(0) {
    if (number_of_initial_pairs>0) {
      newAcceptancePairs(number_of_initial_pairs);
    }
  }

  /** Destructor */
  ~RabinAcceptance() {
    // delete BitSets in the acceptance pairs
    for (std::vector<BitSet*>::iterator it=_acceptance_L.begin();
         it!=_acceptance_L.end();
         ++it) {
      delete *it;
    }
    
    for (std::vector<BitSet*>::iterator it=_acceptance_U.begin();
         it!=_acceptance_U.end();
         ++it) {
      delete *it;
    }
  }

private:
  RabinAcceptance(const RabinAcceptance& other); // not yet implemented

public:

  
  // ---- The AcceptanceCondition Interface

  class Initializer {
  public:
    Initializer(unsigned int nr_of_pairs, 
                bool allow_automatic_growth) : 
      _nr_of_pairs(nr_of_pairs),
      _allow_automatic_growth(allow_automatic_growth) { }
    unsigned int getNumberOfPairs() const {return _nr_of_pairs;}
    bool allowAutomaticGrowth() const {return _allow_automatic_growth;}
  private:
    unsigned int _nr_of_pairs;
    bool _allow_automatic_growth;
  };

  static Initializer getStandardInitializer() {
    return Initializer(0, false);
  };

  static Initializer getSameSizeInitializer(const RabinAcceptance& other) {
    return Initializer(other.size(), other.allowAutomaticGrowth());
  }

  RabinAcceptance(const Initializer& init) : 
    _acceptance_count(0),
    _max_state_id(0),
    _is_compact(true), 
    _allow_automatic_growth(init.allowAutomaticGrowth()),
    _empty_signature(0) {
    unsigned int number_of_initial_pairs=init.getNumberOfPairs();
    
    if (number_of_initial_pairs>0) {
      newAcceptancePairs(number_of_initial_pairs);
    }
  }
  
  /**
   * Check if this RabinAcceptance is compact (part of interface AcceptanceCondition).
   * @return true iff compact
   */
  bool isCompact() const {
    return _is_compact;
  }


  /**
   * Make this RabinAcceptance compact (part of interface AcceptanceCondition).
   */
  void makeCompact() {
    if (isCompact()) {
      return;
    }

    // Compress Acceptance-Pairs 
    unsigned pair_to=0;
    for (unsigned int pair_from=0;
         pair_from<_acceptance_L.size();
         pair_from++) {
      if (_acceptance_L.at(pair_from)!=0) {
        if (pair_from==pair_to) {
          // nothing to do
        } else {
          _acceptance_L.at(pair_to)=_acceptance_L.at(pair_from);
          _acceptance_U.at(pair_to)=_acceptance_U.at(pair_from);
        }
        
        ++pair_to;
      }
    }
   
    unsigned int new_acceptance_count=pair_to;
    
    _acceptance_L.resize(new_acceptance_count);
    _acceptance_U.resize(new_acceptance_count);

    _acceptance_count=new_acceptance_count;

    _is_compact=true;
  }


  /** Update the acceptance condition upon renaming of states acording
   *  to the mapping (part of AcceptanceCondition interface).
   *  Assumes that states can only get a lower name.
   * @param mapping vector with mapping a[i] -> j
   */
  void moveStates(std::vector<unsigned int> mapping) {
    if (!isCompact()) {
      makeCompact();
    }

    for (unsigned int i=0;i<size();++i) {
      move_acceptance_bits(*_acceptance_L.at(i), mapping);
      move_acceptance_bits(*_acceptance_U.at(i), mapping);
    }
  }

  /**
   * Print the Acceptance-Pairs: header (part of interface AcceptanceCondition).
   * @param out the output stream.
   */
  void outputAcceptanceHeader(std::ostream& out) const {
    out << "Acceptance-Pairs: "<< size() << std::endl;
  }

  /**
   * Print the Acc-Sig: line for a state (part of interface AcceptanceCondition).
   * @param out the output stream.
   * @param state_index the state
   */
  void outputAcceptanceForState(std::ostream& out, unsigned int state_index) const {
    out << "Acc-Sig:";
    for (unsigned int pair_index=0;pair_index<size();pair_index++) {
      if (isStateInAcceptance_L(pair_index, state_index)) {
        out << " +" << pair_index;
      }

      if (isStateInAcceptance_U(pair_index, state_index)) {
        out << " -" << pair_index;
      }
    }
    out << std::endl;
  }


  /**
   * Add a state (part of interface AcceptanceCondition).
   * @param state_index the index of the added state.
   */
  void addState(unsigned int state_index) {
    // TODO: Assert that state_index > highest set bit
    _max_state_id=std::max(state_index, _max_state_id);
  }

  /** The 3 different colors for RabinAcceptance */
  enum RabinColor {RABIN_WHITE=0, RABIN_GREEN=1, RABIN_RED=2};

  /** A class storing the acceptance signature for a state
   * (for every acceptance pair one color).
   * The default color is RED.
   */
  class RabinSignature {
  public:
    /** Constructor 
     * @param size the number of acceptance pairs 
     */
    RabinSignature(unsigned int size) : _size(size) {}

    /** Constructor
     * @param size the number of acceptance pairs
     * @color the color with which to initialize 
     */
    RabinSignature(unsigned int size, RabinColor color) : _size(size) {
      for (unsigned int i=0;i<size;i++) {
        setColor(i, color);
      }
    }

    /** Constructor
     * @param other another RabinSignature
     */
    RabinSignature(const RabinSignature& other) : 
      _WHITE_bits(other._WHITE_bits), _GREEN_bits(other._GREEN_bits), _size(other._size) {}

    /** Constructor
     * @param L the L part of the acceptance signature.
     * @param U the U part of the acceptance signature.
     * @param size the number of acceptance pairs
     */
    RabinSignature(const BitSet& L_bits, const BitSet U_bits, unsigned int size) 
      : _size(size) {
      if (size>0) {
        // the green bits are the L bits, minus the U_bits
        _GREEN_bits = L_bits;
        _GREEN_bits.Minus(U_bits);
        
        // the white bits are all bits minus the U_bits and the L_bits
        _WHITE_bits.set(0, size-1, true);
        _WHITE_bits.Minus(U_bits);
        _WHITE_bits.Minus(L_bits);
      }
    }


    /** Set the RabinColor c. 
     * @param i The pair index
     * @param c the RabinColor
     */
    void setColor(unsigned int i, RabinColor c) {
      if (i>=_size) setSize(i+1);
      switch (c) {
      case RABIN_RED:
	_WHITE_bits.clear(i);
	_GREEN_bits.clear(i);
	break;

      case RABIN_GREEN:
	_WHITE_bits.clear(i);
	_GREEN_bits.set(i);
	break;

      case RABIN_WHITE:
	_WHITE_bits.set(i);
	_GREEN_bits.clear(i);
	break;
      }
    };

    /** Get the RabinColor for a pair i */
    RabinColor getColor(unsigned int i) const {
      if (i>=_size) return RABIN_RED;
      if (_WHITE_bits.get(i)) {
        return RABIN_WHITE;
      } else if (_GREEN_bits.get(i)) {
        return RABIN_GREEN;
      } else {
        return RABIN_RED;
      }
    }

    /** Get string representation of this signature. */
    std::string toString() const {
      std::string a;
      a="{";
      for (unsigned int i=0;i<size();i++) {
	switch (getColor(i)) {
	case RABIN_RED:
	  a+='-'+boost::lexical_cast<std::string>(i);
	  break;
	case RABIN_GREEN:
	  a+='+'+boost::lexical_cast<std::string>(i);
	  break;
	case RABIN_WHITE:
	  break;
	}
      }
      a+="}";
      
      return a;
    }

    /** Compare to other signature for equality. */
    bool operator==(const RabinSignature& other) const {
      return _GREEN_bits == other._GREEN_bits &&
        _WHITE_bits == other._WHITE_bits;
    }

    /** Compare less_than to other signature */
    bool operator<(const RabinSignature& other) const {
      if (_WHITE_bits<other._WHITE_bits) {
	return true;
      } else if (_WHITE_bits==other._WHITE_bits) {
      return _GREEN_bits < other._GREEN_bits;
      }
      return false;
    }

    /** Get the number of acceptance pairs */
    unsigned int getSize() const {return _size;}
    /** Get the number of acceptance pairs */
    unsigned int size() const {return _size;}

    /** Set the number of acceptance pairs */
    void setSize(unsigned int size) {_size=size;}

    /** Merge this acceptance signature with other signature,
     *  for each tuple element calculate the maximum of the
     *  colors according to the order 
     * RABIN_WHITE < RABIN_GREEN < RABIN_RED */
    void maxMerge(const RabinSignature& other) {
      setSize(std::max(size(), other.size()));
              
      for (unsigned int i=0; i<size(); i++) {
	if (getColor(i) < other.getColor(i)) {
	  setColor(i, other.getColor(i));
	}
      }
    }

    /**
     * Calculate a hash value using HashFunction
     * @param hashfunction the HashFunction
     */
    template <class HashFunction>
    void hashCode(HashFunction& hashfunction) {
      _WHITE_bits.hashCode(hashfunction);
      _GREEN_bits.hashCode(hashfunction);
    }
    
    
  private:
    /** The WHITE bits */
    BitSet _WHITE_bits;
    /** The GREEN bits */
    BitSet _GREEN_bits;    
    /** The number of acceptance pairs */
    unsigned int _size;
  };

  /** The signature_type (part of AcceptanceCondition interface) */
  typedef RabinSignature signature_type;

  /** Accessor for the acceptance signature for a state 
   *  (part of AcceptanceCondition interface)
   */
  class AcceptanceForState {
  private:
    /** Reference to the underlying RabinAcceptance */
    RabinAcceptance& _acceptance;
    /** The state index for this accessor */
    unsigned int _state_index;

  public:
    /** Constructor */
    AcceptanceForState(RabinAcceptance& acceptance,
		       unsigned int state_index) :
      _acceptance(acceptance), _state_index(state_index) {}

    void setColor(unsigned int pair_index, RabinColor color) {
      _acceptance.setRabinColorForState(pair_index, _state_index, color);
    }

    RabinColor getColor(unsigned int pair_index) const {
      return _acceptance.getRabinColorForState(pair_index, _state_index);
    }
    
    /** Set the acceptance for this state according to RabinSignature */
    void setSignature(const RabinSignature& signature) {
      for (unsigned int i=0;i<signature.size();i++) {
        _acceptance.setRabinColorForState(i, _state_index, signature.getColor(i));
      }
    }

    /** Get number of acceptance pairs */
    unsigned int size() const {return _acceptance.size();}

    /** Get the signature for this state */
    RabinSignature getSignature() const {
      return RabinSignature(_acceptance.getAcceptance_L_forState(_state_index),
			    _acceptance.getAcceptance_U_forState(_state_index),
			    _acceptance.size());
    }

    RabinAcceptance& getRabinAcceptance() {
      return _acceptance;
    }
    
    unsigned int getStateIndex() {
      return _state_index;
    }
  };


  const RabinSignature& getEmptySignature() {
    return _empty_signature;
  }
  

  // ---- Rabin/Streett acceptance specific

  /**
   * Creates a new acceptance pair.
   * @return the index of the new acceptance pair.
   */
  unsigned int newAcceptancePair() {
    BitSet *l=new BitSet();
    BitSet *u=new BitSet();
    
    _acceptance_L.push_back(l);
    _acceptance_U.push_back(u);
    
    _acceptance_count++;  

    _empty_signature.setSize(_acceptance_L.size());
    return _acceptance_L.size()-1;
  }


  /**
   * Creates count new acceptance pairs.
   * @return the index of the first new acceptance pair.
   */
  unsigned int newAcceptancePairs(unsigned int count) {
    unsigned int rv=_acceptance_L.size();
    
    for (unsigned int i=0;i<count;i++) {
      newAcceptancePair();
    }
    
    return rv;
  }
  

  /**
   * Delete an acceptance pair.
   */
  void removeAcceptancePair(unsigned int pair_index) {
    if (_acceptance_L.at(pair_index)!=0) {
      _acceptance_count--;
    }
    
    delete _acceptance_L.at(pair_index);
    delete _acceptance_U.at(pair_index);
    
    _acceptance_L.at(pair_index)=0;
    _acceptance_U.at(pair_index)=0;
    
    _is_compact=false;
  }

  bool allowAutomaticGrowth() const {return _allow_automatic_growth;}
  void setAllowAutomaticGrowth(bool value=true) {
    _allow_automatic_growth=value;
  }

  /**
   * Get a reference to the BitSet representing L[pair_index], 
   * allowing changes to this set.
   */
  BitSet& getAcceptance_L(unsigned int pair_index) {
    accessPair(pair_index);
    return *_acceptance_L.at(pair_index);
  }

  /**
   * Get a reference to the BitSet representing U[pair_index], 
   * allowing changes to this set.
   */
  BitSet& getAcceptance_U(unsigned int pair_index) {
    accessPair(pair_index);
    return *_acceptance_U.at(pair_index);
  }
  

  /**
   * Get the L part of the acceptance signature for a state (changes to the
   * BitSet do not affect the automaton).
   */
  BitSet getAcceptance_L_forState(unsigned int state_index) const {
    BitSet result;
    getBitSetForState(state_index, _acceptance_L, &result);

    return result;
  }

  /**
   * Get the U part of the acceptance signature for a state (changes to the
   * BitSet do not affect the automaton).
   */
  BitSet getAcceptance_U_forState(unsigned int state_index) const {
    BitSet result;
    getBitSetForState(state_index, _acceptance_U, &result);

    return result;
  }

  /** Is a certain state in L[pair_index]? */
  bool isStateInAcceptance_L(unsigned int pair_index,
			     unsigned int state_index) const {
    return _acceptance_L.at(pair_index)->get(state_index);
  }

  /** Is a certain state in U[pair_index]? */
  bool isStateInAcceptance_U(unsigned int pair_index,
			     unsigned int state_index) const {
    return _acceptance_U.at(pair_index)->get(state_index);
  }

  RabinColor getRabinColorForState(unsigned int pair_index,
                                   unsigned int state_index) const {
    bool inL = isStateInAcceptance_L(pair_index, state_index);
    bool inU = isStateInAcceptance_U(pair_index, state_index);
    if (inU) return RABIN_RED;
    if (inL) return RABIN_GREEN;
    return RABIN_WHITE;
  }

  void setRabinColorForState(unsigned int pair_index,
                             unsigned state_index,
                             RabinColor color) {
    switch (color) {
    case RABIN_WHITE:
      stateIn_L(pair_index, state_index, false);
      stateIn_U(pair_index, state_index, false);
      break;
    case RABIN_RED:
      stateIn_L(pair_index, state_index, false);
      stateIn_U(pair_index, state_index, true);
      break;
    case RABIN_GREEN:
      stateIn_L(pair_index, state_index, true);
      stateIn_U(pair_index, state_index, false);
      break;
    }
  }

  /** Set L[pair_index] for this state to value. */
  void stateIn_L(unsigned int pair_index,
		 unsigned int state_index,
		 bool value=true) {
    getAcceptance_L(pair_index).set(state_index,value);
  }

  /** Set U[pair_index] for this state to value. */
  void stateIn_U(unsigned int pair_index,
		 unsigned int state_index,
		 bool value=true) {
    getAcceptance_U(pair_index).set(state_index,value);
  }
  
  /** Get the number of acceptance pairs. 
   *  Requires the acceptance pairs to be compact. */
  unsigned int size() const {
    if (!isCompact()) {
      throw std::string("Can't give acceptance pair count for uncompacted condition.");
    }
    return _acceptance_L.size();
  }

  void optimize() {
    if (!isCompact()) 
      makeCompact();

    assert(_acceptance_L.size() == _acceptance_U.size());

    for (unsigned int pair_index=0;
         pair_index < _acceptance_L.size();
         ++pair_index) {
      
      // L = L \ U
      if (_acceptance_L.at(pair_index)->intersects(*_acceptance_U.at(pair_index))) {
        BitSet L_minus_U=*_acceptance_L.at(pair_index);
        L_minus_U.Minus(*_acceptance_U.at(pair_index));
        (*_acceptance_L.at(pair_index))=L_minus_U;
      }
      
      // remove if L is empty
      if (_acceptance_L.at(pair_index)->isEmpty()) {
        // no state is in L(pair_index) -> remove
        removeAcceptancePair(pair_index);
      }
    }

    // make compact again
    if (!isCompact()) 
      makeCompact();
  }



private:
  /** A vector of BitSet* */
  typedef std::vector<BitSet*> BitSetVector;

  /** Calculate the BitSet for a state from the acceptance pairs, store
   *  result in result.
   *  @param state_index the state
   *  @param acc the BitSetVector (either _L or _U)
   *  @param result the Bitset where the results are stored, has to be clear 
   *                at the beginning!
   */
  void getBitSetForState(unsigned int state_index,
			 const BitSetVector& acc,
			 BitSet* result) const {
    
    for (unsigned int i=0;i<acc.size();i++) {
      if (acc.at(i)!=0) {
	if (acc.at(i)->get(state_index)) {
	  result->set(i);
	}
      }
    }
  }


  /** 
   * Move the bits set in acc to the places specified by mapping.
   */
  void move_acceptance_bits(BitSet& acc,
			    std::vector<unsigned int> mapping) {
    int i=acc.nextSetBit(0);
    while (i!=-1) {
      unsigned int j=mapping.at(i);
      // :: j is always <= i
      if (j>(unsigned int)i) {
	throw std::string("Wrong mapping in move_acceptance_bits");
      }

      if (((unsigned int)i) == j) {
	// do nothing
      } else {
	// move bit from i->j
	acc.set(j);
	acc.clear(i);
      }
      i=acc.nextSetBit(i+1);
    }
  }


private:
  void accessPair(unsigned int pair_index) {
    if (pair_index >= _acceptance_count) {
      if (_allow_automatic_growth) {
        // grow
        grow(pair_index+1);
      } else {
        throw std::string("Trying to access Rabin pair ")+boost::lexical_cast<std::string>(pair_index)+
                        ", but there are only "+boost::lexical_cast<std::string>(_acceptance_count)+" pairs.";
      }
    }    
  }

  /** 
   * Grow the number acceptance pair to to_pairs. The newly created pairs contain
   * all registered states in the color RED.
   */
  void grow(unsigned int to_pairs) {
    assert(to_pairs > _acceptance_count);
    assert(_is_compact);
    assert(_acceptance_count == _acceptance_L.size());
    assert(_acceptance_count == _acceptance_U.size());

    unsigned int new_pairs = to_pairs - _acceptance_count;
    
    // we have to add all states added up to now to the U set of
    // the newly generated pairs
    //  -> colored red
    BitSet U_states;
    U_states.set(0, _max_state_id, true);

    for (unsigned int i=0;i<new_pairs;i++) {
      unsigned int pair_index = newAcceptancePair();
      
      // initialize the pair by setting U to U_states
      // L is clear
      *(_acceptance_U.at(pair_index)) = U_states;
      _acceptance_L.at(pair_index)->clear();
    }
  }
  
private: // members
  /** The number of acceptance pairs */
  unsigned int _acceptance_count;

  /** The maximal state id */
  unsigned int _max_state_id;

  /** A vector of BitSet* representing the L part of the acceptance pairs. */
  BitSetVector _acceptance_L;
  /** A vector of BitSet* representing the U part of the acceptance pairs. */
  BitSetVector _acceptance_U;

  bool _is_compact;

  bool _allow_automatic_growth;

  RabinSignature _empty_signature;
};



}



#endif

