#include "DAParser.hpp"
#include <boost/algorithm/string.hpp>
#include <set>

namespace ltl2dstar {

typedef std::vector<ap_index_t> ap_permutation_t;

static std::string next_line(std::istream& in) {
  std::string result;
  while (in.good()) {
    std::getline(in, result);
    std::size_t comment_pos = result.find("//");
    if (comment_pos!=std::string::npos) {
      result.erase(comment_pos);
    }
    std::size_t n = 0;
    while ((n = result.find("  "))!=std::string::npos) {
      result.replace(n, 2, " ");
    }

    boost::algorithm::trim(result);

    if (result!="") {
      return result;
    }

    if (in.eof()) {
      return "";
    } else {
      throw std::string("Input error trying to parse deterministic automaton.");
    }
  }

  if (in.eof()) {
    return "";
  } else {
    throw std::string("Input error trying to parse deterministic automaton.");
  }
  return "";
}

static std::string get_part(const std::vector<std::string>& v,
                            unsigned int index,
                            const std::string& where,
                            const std::string& what) {
  if (index>=v.size()) {
    throw std::string("Error trying to parse deterministic automaton: In ")
      +where+", missing "+what;
  }
  return v.at(index);
}

static void detect_garbage(const std::vector<std::string>& v, unsigned int max_index) {
  if (v.size()!=max_index+1) {
    throw std::string("Error trying to parse deterministic automaton: Garbage at end of line:\n")+v.at(max_index+1);
  }
}

static unsigned int get_int(const std::string& s) {
  try {
    return boost::lexical_cast<unsigned int>(s);
  } catch (boost::bad_lexical_cast& e) {
    throw std::string("Error trying to parse deterministic automaton: ")+s+" is not a valid integer.";
  }
}

static void expect(const std::string s, const std::string e) {
  if (s!=e) {
    throw std::string("Error trying to parse deterministic automaton: Expected '")+e+"', found '"+s+"'";
  }
}

template <class AcceptanceForState>
void parse_acceptance_signature(AcceptanceForState& acceptance, const std::string& s);

template <>
void parse_acceptance_signature(RabinAcceptance::AcceptanceForState& acceptance, const std::string& s_) {
  std::string s(s_);

  if (!boost::algorithm::starts_with(s, "Acc-Sig:")) {
    throw std::string("Error trying to parse deterministic automaton: Expected Acc-Sig:");
  }
  s.erase(0, std::string("Acc-Sig:").length());
  boost::algorithm::trim(s);

  unsigned int pairs = acceptance.getRabinAcceptance().size();
  // initialize to WHITE
  for (unsigned int i=0;i<pairs;i++) {
    acceptance.setColor(i, RabinAcceptance::RABIN_WHITE);
  }

  if (s=="") {return;}

  std::vector<std::string> v;
  boost::algorithm::split(v, s, boost::algorithm::is_any_of(" "));
  for (unsigned int i=0;i<v.size();i++) {
    std::string acc = v.at(i);
    bool green=false;
    if (boost::algorithm::starts_with(acc, "+")) {
      green=true;
    } else if (boost::algorithm::starts_with(acc, "-")) {
      green=false;
    } else {
      throw std::string("Error trying to parse deterministic automaton: Invalid acceptance signature: ")+acc;
    }
    
    acc.erase(0,1);
    unsigned int pair = get_int(acc);
    if (pair >= pairs) {
      throw std::string("Error trying to parse deterministic automaton: Unknown Rabin/Streett pair "+boost::lexical_cast<std::string>(pair));
    }
    if (green) {
      acceptance.setColor(pair, RabinAcceptance::RABIN_GREEN);
    } else {
      acceptance.setColor(pair, RabinAcceptance::RABIN_RED);
    }
  }  
}

template <>
void parse_acceptance_signature(ParityAcceptance::AcceptanceForState& acceptance, const std::string& s) {
  std::vector<std::string> v;
  boost::algorithm::split(v, s, boost::algorithm::is_any_of(" "));
  expect(get_part(v, 0, "state definition", "Rank:"), "Rank:");
  unsigned int rank = get_int(get_part(v,1, "state definition", "rank"));
  if (rank>acceptance.getParityAcceptance().getMaximalUsedRank()) {
    throw std::string(
                    "Error trying to parse deterministic automaton: Rank "+boost::lexical_cast<std::string>(rank)+
                    " out of range!");
  }
  acceptance.setRank(rank);
}


template <class Automaton>
void parse_state(Automaton& automaton, std::istream& in, 
                 unsigned int index, 
                 unsigned int state_count,
                 unsigned int transitions,
                 ap_permutation_t* ap_permutation) {
  std::vector<std::string> v;
  std::string state = next_line(in);
  boost::algorithm::split(v, state, boost::algorithm::is_any_of(" "));
  
  expect(get_part(v, 0, "state", "\"State:\""), "State:");
  unsigned int state_index=get_int(get_part(v, 1, "state", "state index"));
  // don't detect garbage, possible comment
  
  if (index!=state_index) {
    throw std::string(
                    "Error trying to parse deterministic automaton: Expected state definition "+
                    boost::lexical_cast<std::string>(index)+
                    ", found "+
                    boost::lexical_cast<std::string>(state_index));
  }

  typename Automaton::state_type* state_p = automaton.get(index);

  std::string acc_sig = next_line(in);
  typename Automaton::acceptance_condition_type::AcceptanceForState
    acceptance_for_state(automaton.acceptance(), index);
  parse_acceptance_signature(acceptance_for_state, acc_sig);

  for (unsigned int i=0;i<transitions;i++) {
    std::string transition = next_line(in);
    if (transition=="") {
      throw std::string("Error trying to parse deterministic automaton: Missing transition!");
    }
    unsigned int next_state = get_int(transition);
    if (next_state >= state_count) {
      throw std::string("Error trying to parse deterministic automaton: Transition to non-existent state "+
                      boost::lexical_cast<std::string>(next_state));
    }

    APElement label(i);
    if (ap_permutation) {
      label.permutate(*ap_permutation);
    }
    state_p->edges().addEdge(label, automaton.get(next_state));
  }  
}

Automaton::ptr DAParser::parse(std::istream& in, APSet_cp for_ap_set) {
  std::vector<std::string> v;

  std::string header = next_line(in);
  boost::algorithm::split(v, header, boost::algorithm::is_any_of(" "));
  
  std::string automaton_type_name=get_part(v, 0, "header", "automaton type");
  std::string version=get_part(v, 1, "header", "format version");
  std::string transition_type=get_part(v, 2, "header", "transition type");
  detect_garbage(v,2);

  Automaton::automaton_t automaton_type;

  if (automaton_type_name=="DRA") {
    automaton_type=Automaton::DETERMINISTIC_RABIN_AUTOMATON;
  } else if (automaton_type_name=="DSA") {
    automaton_type=Automaton::DETERMINISTIC_STREETT_AUTOMATON;
  } else if (automaton_type_name=="DPA") {
    automaton_type=Automaton::DETERMINISTIC_PARITY_AUTOMATON;
  } else {
    throw std::string("Error trying to parse deterministic automaton: Automaton type "+automaton_type_name+" is not supported, only DRA, DSA and DPA");
  }

  if (version!="v2") {
    throw std::string("Error trying to parse deterministic automaton: Version "+version+" is not supported, only v2");
  }

  if (transition_type!="explicit") {
    throw std::string("Error trying to parse deterministic automaton: Transition format "+transition_type+" is not supported, only explicit");
  }

  
  std::string comment_line, state_line;
  comment_line = next_line(in);
  if (boost::algorithm::starts_with(comment_line, "Comment:")) {
    // ignore line;
    state_line = next_line(in);
  } else {
    state_line = comment_line;
    comment_line ="";
  }

  v.clear();
  boost::algorithm::split(v, state_line, boost::algorithm::is_any_of(" "));
  expect(get_part(v, 0, "states header", "\"States:\""), "States:");
  unsigned int state_count = get_int(get_part(v, 1, "states header", "state count"));
  detect_garbage(v, 1);
  
  std::string acceptance = next_line(in);
  v.clear();
  boost::algorithm::split(v, acceptance, boost::algorithm::is_any_of(" "));
  unsigned int acceptance_number = 0;
  switch (automaton_type) {
  case Automaton::DETERMINISTIC_RABIN_AUTOMATON:
  case Automaton::DETERMINISTIC_STREETT_AUTOMATON: {
    expect(get_part(v, 0, "acceptance definition", "Acceptance-Pairs:"), "Acceptance-Pairs:");
    acceptance_number = get_int(get_part(v, 1, "acceptance definition", "acceptance pair count"));
    detect_garbage(v,1);
    break;
  }
  case Automaton::DETERMINISTIC_PARITY_AUTOMATON:
    expect(get_part(v, 0, "acceptance definition", "Ranks:"), "Ranks:");
    acceptance_number = get_int(get_part(v, 1, "acceptance definition", "ranks"));
    detect_garbage(v,1);
    break;
  }
  
  std::string start_state = next_line(in);
  v.clear();
  boost::algorithm::split(v, start_state, boost::algorithm::is_any_of(" "));
  expect(get_part(v, 0, "state state header", "\"Start:\""), "Start:");
  unsigned int start_state_index = get_int(get_part(v, 1, "start state header", "start state index"));
  detect_garbage(v, 1);

  
  // AP

  std::string ap = next_line(in);
  APSet_p aps(new APSet());
  std::set<std::string> ap_set;

  v.clear();
  boost::algorithm::split(v, ap, boost::algorithm::is_any_of(" "));
  expect(get_part(v, 0, "AP header", "\"AP:\""), "AP:");
  unsigned int ap_count = get_int(get_part(v, 1, "AP header", "AP count"));

  if (ap_count > APSet::getMaxAP()) {
    throw std::string("Error trying to parse deterministic automaton: More than "+
                    boost::lexical_cast<std::string>(APSet::getMaxAP())+" APs are not supported.");
  }

  // reconstruct rest of line
  std::string ap_defs = boost::algorithm::join(std::vector<std::string>(v.begin()+2, v.end()), " ");
  for (unsigned int i=0;i<ap_count;i++) {
    boost::algorithm::trim(ap_defs);
    if (ap_defs=="") {
      throw std::string("Error trying to parse deterministic automaton: Expected "+
                      boost::lexical_cast<std::string>(ap_count)+" APs, found only "+
                      boost::lexical_cast<std::string>(i));
    }
    if (!boost::algorithm::starts_with(ap_defs, "\"")) {
      throw std::string("Error trying to parse deterministic automaton: Malformed AP, expected '\"'");
    }
    ap_defs=ap_defs.substr(1);
    std::size_t closing_quote = ap_defs.find("\"");
    if (closing_quote == std::string::npos) {
      throw std::string("Error trying to parse deterministic automaton: Malformed AP, could not find closing '\"'");
    }
    if (closing_quote == 0) {
      throw std::string("Error trying to parse deterministic automaton: Empty AP");
    }
    std::string cur_ap=ap_defs.substr(0, closing_quote);
    if (ap_set.count(cur_ap)>0) {
      throw std::string("Error trying to parse deterministic automaton: Duplicate AP \""+cur_ap+"\"");
    }

    aps->addAP(cur_ap);
    ap_set.insert(cur_ap);

    ap_defs.erase(0, closing_quote+1);
  }
  if (ap_defs!="") {
    throw std::string("Error trying to parse deterministic automaton: Garbage at end of AP header");
  }

  boost::shared_ptr< ap_permutation_t > ap_permutation;
  if (for_ap_set) {
    ap_permutation.reset(new ap_permutation_t(aps->generate_permutation_to(*for_ap_set)));
  }

  Automaton::ptr automaton;
  switch (automaton_type) {
  case Automaton::DETERMINISTIC_RABIN_AUTOMATON: {
    Automaton::DRA_ptr dra(new Automaton::DRA_t(aps));
    dra->acceptance().newAcceptancePairs(acceptance_number);
    automaton.reset(new Automaton(dra));
    break;
  }
  case Automaton::DETERMINISTIC_STREETT_AUTOMATON: {
    Automaton::DRA_ptr dsa(new Automaton::DRA_t(aps));
    dsa->considerAsStreett();
    dsa->acceptance().newAcceptancePairs(acceptance_number);
    automaton.reset(new Automaton(dsa));
    break;
  }
  case Automaton::DETERMINISTIC_PARITY_AUTOMATON: {
    Automaton::DPA_ptr dpa(new Automaton::DPA_t(aps));
    dpa->acceptance()=ParityAcceptance(acceptance_number);
    automaton.reset(new Automaton(dpa));
    break;
  }
  default:
    throw std::string("Unsupported automaton type");
  }

  // Create all the states
  for (unsigned int i=0;i<state_count;i++) {
    switch (automaton_type) {
    case Automaton::DETERMINISTIC_RABIN_AUTOMATON: 
    case Automaton::DETERMINISTIC_STREETT_AUTOMATON: 
      automaton->getDRA()->newState();
      break;
    case Automaton::DETERMINISTIC_PARITY_AUTOMATON: 
      automaton->getDPA()->newState();
      break;
    }
  }

  // set start state
  automaton->setStartState(start_state_index);

  unsigned int transitions_per_state = aps->powersetSize();

  std::string separator = next_line(in);
  expect(separator, "---");  

  // actually parse state definitions
  for (unsigned int i=0;i<state_count;i++) {
    switch (automaton_type) {
    case Automaton::DETERMINISTIC_RABIN_AUTOMATON: 
    case Automaton::DETERMINISTIC_STREETT_AUTOMATON: 
      parse_state(*automaton->getDRA(), in, i, state_count, transitions_per_state,
                  ap_permutation.get());
      break;
    case Automaton::DETERMINISTIC_PARITY_AUTOMATON: 
      parse_state(*automaton->getDPA(), in, i, state_count, transitions_per_state,
                  ap_permutation.get());
      break;
    }
  }

  return automaton;
}

}
