/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef BDD_ASSIGNMENT_HELPERS_HPP
#define BDD_ASSIGNMENT_HELPERS_HPP

// don't include individually, include bdd_helpers.hpp instead

namespace vereofy {

  /*
   * This function checks whether the bit corresponding to the projection
   * function var is set in the assignment function (having exactly one path).
   */
  inline bool bdd_is_bit_set(const BDDFunction& assignment, const BDDFunction& var) {
    BDD_CHECK(3, BDDChecks::isProjection(var), "", var);
    BDD_CHECK2(3, BDDChecks::isAssignment(assignment, var), "", assignment, var);

    return (assignment & var) != BDDFunction::ZERO();
  }


  /**
   * Returns a bit string corresponding to the assignment, vars specifies the order of bits
   */
  inline std::string projectionAssignment(const BDDFunction& assignment, const BDDFunctionContainer& vars) {
    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    std::string result(vars.size(), '0');
    for (unsigned int i=0;i<vars.size();i++) {
      if (bdd_is_bit_set(assignment, vars.at(i))) {
        result[i]='1';
      }
    }
    return result;
  }

  /**
   * Generates a bitstream from the assignment (ordered by vars) and calls
   * void BitInserter::operator()(bool bit) for each bit
   */
  template <class BitInserter>
  inline void projectionAssignment(const BDDFunction& assignment, const BDDFunctionContainer& vars,
                                   BitInserter& inserter) {
    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    for (unsigned int i=0;i<vars.size();i++) {
      bool bit = bdd_is_bit_set(assignment, vars.at(i));
      inserter(bit);
    }
  }

  /**
   * Returns a bit string corresponding to the assignment, 
   * csVars specifies the significant variables. The bits are ordered
   * by the current variable order.
   */
  inline std::string assignmentToBin(const BDDFunction& assignment, 
                                const BDDFunction& csVars_) {  
    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    BDDFunction csVars(csVars_);
    BDD_CHECK2(3, BDDChecks::isAssignment(assignment, csVars), "", assignment, csVars);
    
    std::stringstream out;

    while (csVars!=BDDFunction::ONE()) {
      BDDFunction cur_var=popCubeSet(csVars, csVars);

      out << (bdd_is_bit_set(assignment, cur_var) ? "1" : "0");
    }

    return out.str();
  }

  /**
   * Returns a hexadecimal encoding of the bits of the assignment, 
   * csVars specifies the significant variables. The bits are ordered
   * by the current variable order.
   */
  inline std::string assignmentToHex(const BDDFunction& assignment, 
                                const BDDFunction& csVars) {
    BDD_CHECK2(3, BDDChecks::isAssignment(assignment, csVars), "", assignment, csVars);

    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    std::string binString = "000" + assignmentToBin(assignment, csVars);
    int pos = binString.length() / 4;  
    std::string result(pos, '0');
    for (int i=binString.length()-1; i>=0; i-=4) {
      char j = (binString[i]-'0') + (binString[i-1]-'0')*2 + (binString[i-2]-'0')*4 + (binString[i-3]-'0')*8;
      result[--pos] = j<10?'0'+j:'A' + (j -10);
    }
    return "0x"+result;
  }
  

  inline BDDFunction popAssignment(BDDFunction& target, 
                                   const BDDFunction& csVariables) {
    if (target==BDDFunction::ZERO()) {
      throw "popAssignment called with empty set!";
    }

    BDD_CHECK(3, BDDChecks::isCubeSet(csVariables), "", csVariables);

    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    unsigned long size = BDDVarOrder::size(); 
    boost::scoped_array<char> bufPath(new char[size]);
    target.getPath(bufPath.get());
    
    boost::scoped_array<char> bufCsVariables(new char[size]);
    csVariables.getPath(bufCsVariables.get());
    
    for (unsigned long i=0; i<size; ++i) {
      if (bufCsVariables[i] == '-') bufPath[i] = '-';
        else if (bufPath[i] == '-') bufPath[i] = '0';
    }
    BDDFunction result = BDDFunction(bufPath.get());
    target &= !result;

    BDD_CHECK2(3, BDDChecks::isAssignment(result, csVariables), "", result, csVariables);
    return result;
  }


  // converts a hexadecimal representation (lowest byte first) into an
  // BDDFunction assignment, with the variables taken from csVars (lowest bit first)
  inline BDDFunction hexToAssignment(const std::string& hex_string,
                                     const BDDFunctionContainer& csVars) {
    BDDFunction result = BDDFunction::ONE();

    std::string::const_iterator it = hex_string.begin();
    unsigned short byte_value = 0;

    for (unsigned int index = 0; index < csVars.size(); index++) {
      if (index % 8 == 0) {
        // get most significant nibble
        unsigned short nibble;
        if (it == hex_string.end()) {
          nibble = 0;
        } else {
          char ch = *it++;
          if (ch>='0' && ch<='9') {
            nibble = ch-'0';
          } else if (ch>='a' && ch<='f') {
            nibble = 10+(ch-'a');
          } else if (ch>='A' && ch<='F') {
            nibble = 10+(ch-'A');
          } else {
            throw std::string("Illegal character ")+ch+" in hex string";
          }

          byte_value = nibble << 4;

          if (it == hex_string.end()) {
            throw "Premature end of hex string!";
          }
          ch = *it++;
          if (ch>='0' && ch<='9') {
            nibble = ch-'0';
          } else if (ch>='a' && ch<='f') {
            nibble = 10+(ch-'a');
          } else if (ch>='A' && ch<='F') {
            nibble = 10+(ch-'A');
          } else {
            throw std::string("Illegal character ")+ch+" in hex string";
          }
          byte_value+=nibble;
        }
      }

      bool bit = byte_value & 0x01;
      byte_value = byte_value >> 1;
      if (bit) {
        result &= csVars.at(index);
      } else {
        result &= !csVars.at(index);
      }
    }

    return result;
  }
}

#endif
