/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/



#ifndef BDD_CHECKS_HPP
#define BDD_CHECKS_HPP

namespace vereofy {

struct BDDChecks {
  /**
   * Returns true, if f is a projection function (exactly one essential variable, with if==ONE, else==ZERO
   */
  static bool isProjection(const BDDFunction& f) {
    if (f.isConstant()) return false;
    
    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope
    return f==BDDFunction::projection(f.getFirstVariableIndex());
  }

  /**
   * Returns true, if f is an assignment (exactly one path, only 0 and 1 for every variable)
   */
  static bool isAssignment(const BDDFunction& f) {
    if (f.isConstant()) return false;

    return hasOnePath(f, true, true, false);
  }

  /**
   * Returns true, if f is an assignment (exactly one path, and only 0 and 1 for 
   *  the variables given as the cubeset)
   */
  static bool isAssignment(const BDDFunction& f, const BDDFunction& csVars) {
    if (csVars==BDDFunction::ONE()) return true;
    if (f.isConstant()) return false;

    if (!hasOnePath(f, true, true, false, csVars)) return false;

    return true;
  }

  static bool isCubeSet(const BDDFunction& f) {
    if (f==BDDFunction::ZERO()) {
      return false;
    }

    if (f==BDDFunction::ONE()) {
      // empty cubeset
      return true;
    }
    
    return hasOnePath(f, false, true, true);
  }

  static bool hasOnePath(const BDDFunction& f, 
                         bool allow_zero=true, 
                         bool allow_one=true, 
                         bool allow_dont_care=true,
                         const BDDFunction& essential_vars=BDDFunction::ONE()) {

#ifdef USE_ZADD
    return true; // we don't know
#endif

    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    unsigned long size = BDDVarOrder::size(); 
    boost::scoped_array<char> bufPath(new char[size]);
    f.getPath(bufPath.get());

    boost::scoped_array<char> bufCsVariables;
    bool check_essential=(essential_vars!=BDDFunction::ONE());
    
    if (check_essential) {
      bufCsVariables.reset(new char[size]);
      essential_vars.getPath(bufCsVariables.get());
    }

    for (unsigned long i=0; i<size; ++i) {
      if (check_essential) {
        if (bufCsVariables[i]=='-') {
          // ignore non-essential value
          continue;
        }
      }

      switch (bufPath[i]) {
      case '0':
        if (!allow_zero) return false;
        break;
      case '1':
        if (!allow_one) return false;
        break;
      case '-':
        if (!allow_dont_care) return false;
        break;
      default:
        throw "Error in getPath!";
      }
    }
    BDDFunction path = BDDFunction(bufPath.get());

    // this should be the only path
    return ((f&!path)==BDDFunction::ZERO());    
  }
  
  static bool essentialVarsInCubeSet(const BDDFunction& f, const BDDFunction& cubeSet) {
    if (f.isConstant()) {
      return true;
    }

    return f.exists(cubeSet)==BDDFunction::ONE();
  }
};

}
#endif
