/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef BDD_CONTAINER_HELPERS_HPP
#define BDD_CONTAINER_HELPERS_HPP

// don't include individually, include bdd_helpers.hpp instead

namespace vereofy {

  inline BDDFunction container2cs(const BDDFunctionContainer& container) {
    BDDFunction result = BDDFunction::ONE();
    for (unsigned int i=0; i<container.size(); ++i) 
      result &= container.at(i);
    
    BDD_CHECK(3, BDDChecks::isCubeSet(result), "", result);
    return result;
  }

  inline BDDFunction containerAND(const BDDFunctionContainer& container) {
    BDDFunction result = BDDFunction::ONE();
    for (unsigned int i=0; i<container.size(); ++i) {
      result &= container.at(i);
    }

    return result;
  }


  inline BDDFunction pairwiseEQUIV(const BDDFunctionContainer& a, const BDDFunctionContainer& b) {
    if (a.size() != b.size()) throw "pairwiseEQUIV FunctionContainer of different size";
    BDDFunction result = BDDFunction::ONE();
    for (unsigned int i=0; i < a.size(); ++i) {
      result &= EQUIV(a.at(i), b.at(i));
    }
    return result;
  }

  inline BDDFunction pairwiseEQUIV_with_fill(const BDDFunctionContainer& a, const BDDFunctionContainer& b) {
    BDDFunction result = BDDFunction::ONE();
    for (unsigned int i=0; i < a.size() || i<b.size(); ++i) {
      if (i>=a.size()) {
        result&=EQUIV(b.at(i), BDDFunction::ZERO());
      } else if (i>=b.size()) {
        result&=EQUIV(a.at(i), BDDFunction::ZERO());
      } else {
        result &= EQUIV(a.at(i), b.at(i));
      }
    }
    return result;
  }

  inline BDDFunction pairwiseIMPLIES(const BDDFunctionContainer& premise, const BDDFunctionContainer& conclusion) {
    if (premise.size() != conclusion.size()) throw "pairwiseIMPLIES FunctionContainer of different size";
    BDDFunction result = BDDFunction::ONE();
    for (unsigned int i=0; i < premise.size() || i < conclusion.size(); ++i) {
        result &= IMPLIES(premise.at(i), conclusion.at(i));
    }
    return result;
  }

  inline BDDFunctionContainer containerITE(const BDDFunction& bdd_if, 
                                           const BDDFunctionContainer& bdd_then, 
                                           const BDDFunctionContainer& bdd_else) {
    CHECK_SUCCESS(bdd_then.size() == bdd_else.size());
    
    BDDFunctionContainer result;
    for (unsigned int i=0;i<bdd_then.size();i++) {
      result.push(ITE(bdd_if, bdd_then[i], bdd_else[i]));
    }

    return result;
  }

  inline BDDFunctionContainer container_moveDown(const BDDFunctionContainer& functions) {
    BDDFunctionContainer result(functions);

    for (unsigned int i=0;i<result.size();i++) {
      result[i]=result[i].moveDown();
    }

    return result;
  }

  inline BDDFunctionContainer containerToSize(const BDDFunctionContainer& a, 
                                              unsigned int bits,
                                              const BDDFunction& filler=BDDFunction::ZERO()) {
    BDDFunctionContainer result;
    for (unsigned int i=0;i<bits;i++) {
      if (i<a.size()) {
        result.push(a[i]);
      } else {
        result.push(filler);
      }
    }
    return result;
  }


  /**
   * containerLEQ returns a BDDFunction that is true if the value represented by the container left is
   * less or equal to the value of the container right (assuming least-significant-bit at index 0).
   */
  inline BDDFunction containerLEQ(const BDDFunctionContainer& left, const BDDFunctionContainer& right) {
    unsigned int i;
    unsigned int depth_left=left.size();
    unsigned int depth_right=right.size();
    
    unsigned int depth=(depth_left < depth_right) ? depth_left : depth_right;
    BDDFunction result=BDDFunction::ONE();
    
    for(i=0;i<depth;++i){
      BDDFunction left_bit=(i<depth_left) ? left.at(i) : BDDFunction::ZERO();
      BDDFunction right_bit=(i<depth_right) ? right.at(i) : BDDFunction::ZERO();

      // leq_{i}(x,y) := (x_i==y_i & leq_{i-1}(x,y) | (x_i < y_i)
      // leq_{0}(x,y) := (x_0 <= y_0) 
      
      if (i==0) {
        result=!(left_bit & !right_bit);   // left_bit <= right_bit
      } else {
        result&=!(left_bit ^ right_bit);   // left_bit == right_bit
        result|=(!left_bit & right_bit);   // left_bit <  right_bit
      }
    }
    return result;
  }

  /**
   * containerLE returns a BDDFunction that is true if the value represented by the container left is
   * less than the value of the container right (assuming least-significant-bit at index 0).
   */
  inline BDDFunction containerLE(const BDDFunctionContainer& left, const BDDFunctionContainer& right) {
    unsigned int i;
    unsigned int depth_left=left.size();
    unsigned int depth_right=right.size();
    
    unsigned int depth=(depth_left < depth_right) ? depth_left : depth_right;
    BDDFunction result=BDDFunction::ONE();
    
    for(i=0;i<depth;++i){
      BDDFunction left_bit=(i<depth_left) ? left.at(i) : BDDFunction::ZERO();
      BDDFunction right_bit=(i<depth_right) ? right.at(i) : BDDFunction::ZERO();

      // le_{i}(x,y) := (x_i==y_i & le_{i-1}(x,y) | (x_i < y_i)
      // le_{0}(x,y) := (x_0 < y_0) 
      
      if (i==0) {
        result=(!left_bit & right_bit);   // left_bit < right_bit
      } else {
        result&=!(left_bit ^ right_bit);   // left_bit == right_bit
        result|=(!left_bit & right_bit);   // left_bit <  right_bit
      }
    }
    return result;
  }

  inline BDDFunctionContainer containerShiftRight(const BDDFunctionContainer& x, unsigned int v) {
    BDDFunctionContainer result;
    for (unsigned int i=0;i<x.size();i++) {
      if (i<v) {
        continue;
      } else {
        result.push(x[i]);      
      }
    }
    return result;
  }

  inline BDDFunctionContainer containerShiftLeft(const BDDFunctionContainer& x, unsigned int v) {
    BDDFunctionContainer result;
    for (unsigned int i=0;i<v;i++) {
      result.push(BDDFunction::ZERO());
    }
    for (unsigned int i=0;i<x.size();i++) {
      result.push(x[i]);      
    }

    return result;
  }

  inline BDDFunctionContainer containerSUM(const BDDFunctionContainer& x,
                                           const BDDFunctionContainer& y) {
    if (x.size()!=y.size()) throw "SUM for Container of different sizes.";
    BDDFunctionContainer sum;

    BDDFunction carry = BDDFunction::ZERO();
    for(unsigned int i=0;i<x.size();++i){
      sum.push(x[i]^y[i]^carry);
      carry=((x[i]|y[i])&carry)|(x[i]&y[i]);
    }
    sum.push(carry);
    return sum;
  }   

  /**
   * Perform unsigned subtraction x-y, both having an equal number of
   * bits. Returns the resulting bits, as well as a function
   * representing whether underflow occured.
   */
  inline std::pair<BDDFunctionContainer,
                   BDDFunction> 
  containerSUB(const BDDFunctionContainer& x,
               const BDDFunctionContainer& y) {
    if (x.size()!=y.size()) throw "SUB for Container of different sizes.";
    BDDFunctionContainer result;
    
    BDDFunction borrow = BDDFunction::ZERO();
    for(unsigned int i=0;i<x.size();++i){
      BDDFunction value=x[i]^y[i]^borrow;
      borrow=(!x[i] & (y[i] | borrow)) | (x[i] & (y[i]&borrow));
      result.push(value);
      
      
      TRACE_BDD_FUNCTION("containerSUB", x[i], i);
      TRACE_BDD_FUNCTION("containerSUB", y[i], i);
      TRACE_BDD_FUNCTION("containerSUB", value, i);
      TRACE_BDD_FUNCTION("containerSUB", borrow, i);
    }
    return std::make_pair(result, borrow);
  }   


  inline BDDFunctionContainer 
  containerOnesComplement(const BDDFunctionContainer& x) {
    BDDFunctionContainer result;
    
    for (unsigned int i=0;i<x.size();++i) {
      result.push(!x[i]);
    }
    
    return result;
  }
  
  inline BDDFunctionContainer 
  containerTwosComplement(const BDDFunctionContainer& x) {
    // onesComplement
    BDDFunctionContainer result=containerOnesComplement(x);

    // and then add one
    BDDFunction carry = BDDFunction::ONE();
    for(unsigned int i=0;i<x.size();++i){
      BDDFunction cur=result[i];
      result[i]=cur ^ carry;
      carry=cur & carry;
    }

    // ignore carry if any

    return result;
  }   

  inline BDDFunctionContainer containerMULT(const BDDFunctionContainer& x,
                                            const BDDFunctionContainer& y) {
    if (x.size()!=y.size()) throw "MULT for Container of different sizes.";

    unsigned int size=x.size();
    BDDFunctionContainer result;
    result=containerToSize(BDDFunctionContainer(), 2*size, BDDFunction::ZERO());

    BDDFunctionContainer zero=containerToSize(BDDFunctionContainer(), size, BDDFunction::ZERO());
    
    for (unsigned int i=0;i<size;i++) {
      BDDFunctionContainer partial=containerITE(y[i], x, zero);
      partial=containerShiftLeft(partial, i);
      partial=containerToSize(partial, 2*size);

      result=containerSUM(result, partial);
      result=containerToSize(result, 2*size);
    }

    return result;
  }   


  inline BDDFunctionContainer containerDIV(const BDDFunctionContainer& x,
                                           const BDDFunctionContainer& y,
                                           bool get_remainder=false) {
    unsigned int x_size=x.size();

    if (x_size==0) {
      return x;
    }

    BDDFunctionContainer result;
    result=containerToSize(BDDFunctionContainer(), x_size, BDDFunction::ZERO());

    BDDFunctionContainer remainder=x;
    for (unsigned int i=0;i<x_size;i++) {
      BDDFunctionContainer subtrahend=containerShiftLeft(y, x_size-i-1);
      
      std::pair<BDDFunctionContainer,
        BDDFunction> tmp=containerSUB(containerToSize(remainder, subtrahend.size()), subtrahend);
      
      BDDFunction sign_bit=tmp.second;
      remainder=containerITE(!sign_bit, tmp.first, containerToSize(remainder, tmp.first.size()));

      result[x_size-i-1]=!sign_bit;
    }

    if (get_remainder) {
      return containerToSize(remainder, x_size);
    }

    return result;
  }   
  

  inline BDDFunctionContainer containerChoose(const BDDFunction& choice,
                              const BDDFunctionContainer& choice_true,
                              const BDDFunctionContainer& choice_false) {
    CHECK_SUCCESS(choice_true.size() == choice_false.size());

    BDDFunctionContainer result;
    for (unsigned int i=0;i<choice_true.size();i++) {
      result.push( (choice & choice_true[i]) | (!choice & choice_false[i]) );
    }

    return result;
  }

  inline void containerFillToEqualSize(BDDFunctionContainer& a,
                                       BDDFunctionContainer& b,
                                       const BDDFunction& filler=BDDFunction::ZERO()) {
    if (b.size()<a.size()) {
      containerFillToEqualSize(b,a, filler);
      return;
    }

    if (a.size()==b.size()) {
      return;
    }

    for (unsigned int i=a.size();i<b.size();i++) {
      a.push(filler);
    }

    CHECK_SUCCESS(a.size()==b.size());
  }


  inline BDDFunctionContainer containerTruncate(const BDDFunctionContainer& container,
                                                unsigned int bits) {
    CHECK_SUCCESS(bits<=container.size());
    return containerToSize(container, bits);
  }


  inline std::pair<BDDFunctionContainer, BDDFunction> containerCastDown(const BDDFunctionContainer& container,
                                                                        unsigned int bits) {
    CHECK_SUCCESS(bits<=container.size());

    BDDFunctionContainer result = containerTruncate(container, bits);

    BDDFunction restriction=BDDFunction::ONE();

    for (unsigned int i=bits;i<container.size();i++) {
      restriction&=EQUIV(container.at(i), BDDFunction::ZERO());
    }

    return std::make_pair(result, restriction);
  }


  inline void containerPrint(const BDDFunctionContainer& function_container) {
    for (unsigned int i=0;i<function_container.size();i++) {
      std::cout << "[" << i << "]:\n";
      function_container.at(i).printEssentialVariables();
      function_container.at(i).printPaths();
    }
  }


  /** 
   * Compares the variable level of the top-most variable for all functions in the
   * container, identifies the function where this variable level is the lowest
   * in the BDD, pops this function from the container and returns this function.
   */
  inline BDDFunction containerPopLowestFunction(BDDFunctionContainer& container) {
    if (container.size()==0) {
      THROW_UNIMPLEMENTED();
    }

    unsigned long max_level=container.at(0).getFirstVariableIndex();
    unsigned int max_index=0;
    for (unsigned int i=1;i<container.size();i++) {
      unsigned long level = container.at(i).getFirstVariableIndex();
      if (level > max_level) {
        max_level=level;
        max_index=i;
      }
    }

    BDDFunction result = container.at(max_index);
    container.remove(max_index);
    return result;
  }


  /**
   * Returns function & container.first & ... & container.last.
   * The application of the AND for the intermediate results is
   * performed in the order of the top-mode variable of the 
   * functions (see containerPopLowestFunction), to achieve
   * a bottom-up building of the whole result.
   */
  inline BDDFunction containerApplyAND(const BDDFunction& function,
                                       const BDDFunctionContainer& container) {
    BDDFunctionContainer c(container);

    BDDFunction result=function;
    while (c.size()>0) {
      // no reorder lock, the lowest function is determined 
      // in every iteration according to the current varorder
      result &= containerPopLowestFunction(c);
    }
    
    return result;
  }
  
}

#endif
