/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef BDD_CUBESET_HELPERS_HPP
#define BDD_CUBESET_HELPERS_HPP

// don't include individually, include bdd_helpers.hpp instead

namespace vereofy {

  /* Returns the projection function of one variable from the cubeSet and returns in target
   * a cubeset without this variable.
   */
  inline BDDFunction popCubeSet(BDDFunction& target, const BDDFunction& cubeSet) {
    BDD_CHECK(3, BDDChecks::isCubeSet(cubeSet), "", cubeSet);

    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    unsigned long v=cubeSet.getFirstVariableIndex();
    target=cubeSet.cofactor(v, 1);
    BDD_CHECK(3, BDDChecks::isCubeSet(target), "", target);
    
    BDDFunction v_projection=BDDFunction::projection(v);
    return v_projection;
  }

  inline std::string cubeSetToString(const BDDFunction& cube_set_) {
    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    std::stringstream out;
    bool first=true;

    BDDFunction cube_set(cube_set_);   // make non-const copy
    BDD_CHECK(3, BDDChecks::isCubeSet(cube_set), "", cube_set);

    while (cube_set!=BDDFunction::ONE()) {
      BDDFunction cs_var=popCubeSet(cube_set, cube_set);

      if (!first) {out << ",";} first=false;
      out << bdd_var_name(cs_var);
    }

    return out.str();
  }

  inline bool cubeSetIsEmpty(const BDDFunction& cube_set) {
    BDD_CHECK(3, BDDChecks::isCubeSet(cube_set), "", cube_set);
    return cube_set == BDDFunction::ONE();
  }

  /**
   * Returns a BDDFunction where none of the bits in the cube_set are active
   */
  inline BDDFunction cubeSetNone(const BDDFunction& cube_set) {
    BDD_CHECK(3, BDDChecks::isCubeSet(cube_set), "", cube_set);
    return cube_set.change();    
  }


  /**
   * Returns a BDDFunction that is the negation of the cube_set, i.e.,
   * where all variables NOT in the cube_set are 1
   */
  inline BDDFunction cubeSetNegate(const BDDFunction& cube_set) {
    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    BDD_CHECK(3, BDDChecks::isCubeSet(cube_set), "", cube_set);

    BDDFunction all_vars=BDDFunction::ONE();
    for(unsigned long i=BDDVarOrder::size();i>0;--i){
      all_vars&=BDDFunction::projection(i-1);
    }

    return all_vars.cofactor(cube_set);
  }

  /**
   * Returns a BDDFunction where at least one of the bits in the cube_set is active
   */
  inline BDDFunction cubeSetAtLeastOne(const BDDFunction& cube_set) {
    BDD_CHECK(3, BDDChecks::isCubeSet(cube_set), "", cube_set);
    return !cubeSetNone(cube_set);
  }

  inline BDDFunctionContainer cubeSetToFunctionContainer(const BDDFunction& cube_set) {
    BDD_CHECK(3, BDDChecks::isCubeSet(cube_set), "", cube_set);

    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    BDDFunction cur = cube_set;
    TRACE_BDD_FUNCTION("cubeSetToFunctionContainer", cur, "");
    BDDFunctionContainer result;
    while (cur!=BDDFunction::ONE()) {
      unsigned long first_var = cur.getFirstVariableIndex();
      BDDFunction v = BDDFunction::projection(first_var);
      TRACE_BDD_FUNCTION("cubeSetToFunctionContainer", v, "");      
    
      result.push(v);
      cur = cur.exists(v);
      TRACE_BDD_FUNCTION("cubeSetToFunctionContainer", cur, "");
    }

    return result;
  }
}

#endif
