/***************************************************************************
 *   Copyright (C) 2009 by Joachim Klein                                   *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "setup.hpp"
#include "bdd_op_trace.hpp"

#include <iostream>
#include <iomanip>
#include <cstdio>
#include <algorithm>
#include <sstream>
#include <boost/algorithm/string.hpp>

using namespace std;

static unsigned int bdd_op_trace_counter = 0;

#if defined(JINC_NEW) && defined(USE_ADD)
#include "add/addinnernode.h"
typedef set<ADDBaseNode*> node_set_t;
#endif

#if defined(JINC_NEW) && defined(USE_ZADD) 
#include "zadd/zaddinnernode.h"
typedef set<ZADDBaseNode*> node_set_t;
#endif

#if !defined(JINC_NEW) && defined(USE_ADD) 
#include "add/addnode.h"
typedef set<ADDNode*> node_set_t;
#endif

#if !defined(JINC_NEW) && defined(USE_ZADD) 
#include "zadd/zaddnode.h"
typedef set<ZADDNode*> node_set_t;
#endif

using namespace vereofy;

typedef std::map<unsigned int, unsigned int> size_per_level_t;

size_per_level_t calculate_size_per_level(const BDDFunction& f) {
  size_per_level_t result;

  node_set_t nodes=f.getNodes();

  for (node_set_t::iterator it=nodes.begin();
       it!=nodes.end();
       ++it) {
    node_set_t::key_type node = *it;

    if (!node->isDrain()) {
      unsigned int level = node->getLevel();
      if (result.count(level)==0) {
        result[level]=1;
      } else {
        result[level]++;
      }
    }
  }

  return result;
}

std::string calculate_sharing(const BDDFunction& result,
                              const std::vector<BDDFunction>& arguments) {
  std::stringstream ss;

  node_set_t result_nodes=result.getNodes();
  std::size_t result_size=result_nodes.size();

  ss << "Sizes: Result=" << result_size;

  for (unsigned int i=0;i<arguments.size();i++) {
    node_set_t arg_nodes=arguments.at(i).getNodes();
    ss << ", A" << (i+1) << "=" << arg_nodes.size();

    node_set_t tmp;    
    std::set_difference(result_nodes.begin(), result_nodes.end(),
                        arg_nodes.begin(), arg_nodes.end(),
                        std::insert_iterator<node_set_t>(tmp, tmp.end()));
    result_nodes=tmp;
  }

  std::size_t unshared=result_nodes.size();
  std::size_t shared=(result_size-unshared);
  
  ss << " -- Shared: " << shared << " (";
  ss << std::fixed << std::setprecision(2) << (((double)shared)/result_size*100) << "%)";

  return ss.str();
}

std::string get_variable_type(unsigned int level) {
  string var_name;
  var_name=bdd_var_name(level);

  if (boost::algorithm::starts_with(var_name, "X")) {
    return "1";
  } else if (boost::algorithm::starts_with(var_name, "Y")) {
    return "2";
  } else if (boost::algorithm::starts_with(var_name, "Z")) {
    return "3";
  } else if (boost::algorithm::starts_with(var_name, "*")) {
    return "4";
  } else if (boost::algorithm::starts_with(var_name, "#")) {
    return "5";
  }

  return "0";
}


void output_level_size(std::ostream& out, const std::vector<size_per_level_t>& levels, bool only_relevant=false) {
  std::vector<size_per_level_t::const_iterator> iterators;
  for (unsigned int i=0;i<levels.size();i++) {
    iterators.push_back(levels.at(i).begin());
  }

  for (unsigned level=0;level<bdd_var_count();level++) {
    if (only_relevant) {
      // check whether there is at least one iterator with data for this level
      bool have_data=false;
      for (unsigned int i=0;i<levels.size();i++) {
        if (iterators.at(i) != levels.at(i).end() &&
            iterators.at(i)->first == level) {
          have_data=true;
          break;
        }
      }

      if (!have_data) {
        // skip to next level
        continue;
      }
    }

    // output data

    // first column: level
    out << level;  

    // second column: level type
    out << ";" << get_variable_type(level);
        
    for (unsigned int j=0;j<levels.size();j++) {
      unsigned int value=0;  // the default value

      // if there is data for this level for iterator j, store it in value
      if (iterators.at(j) != levels.at(j).end() &&
          iterators.at(j)->first == level) {
        value=iterators.at(j)->second;
        ++iterators.at(j);  // increment iterator
      }

      out << ";";
      out << value;
    }
    out << "\n";
  }
}


std::string get_number(unsigned int i) {
  char buffer[20];
  sprintf(buffer, "%05d", i);
  return std::string(buffer);
}


void BDD_OpTraceHelper::trace(const std::string &name, 
                              unsigned int counter,
                              const BDDFunction& result,
                              const std::vector<BDDFunction>& arguments) {

  string filename=get_number(bdd_op_trace_counter)+"-"+get_number(counter)+"_"+name+".sizes";
  bdd_op_trace_counter++;

  std::cout << "    ## Tracing size per level for " << name << ": " << filename << "\n";


  std::vector<size_per_level_t> levels;
  levels.push_back(calculate_size_per_level(result));

  for (unsigned int i=0;i<arguments.size();i++) {
    levels.push_back(calculate_size_per_level(arguments.at(i)));
  }

  std::cout << "    ## " << calculate_sharing(result, arguments) << "\n";

  ofstream out(filename.c_str());
  if (out.good()) {
    output_level_size(out, levels);
  }
  out.close();
}
