/***************************************************************************
 *   Copyright (C) 2011 by Joachim Klein                                   *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef VALUE_OR_UNDEFINED_HPP
#define VALUE_OR_UNDEFINED_HPP

#include "vereofy/common/errors.hpp"

namespace vereofy {

/**
 * ValueOrUndefined<T>:
 * Stores either a value of type T or 'undefined'
 */
template <class T>
class ValueOrUndefined {
public:
  ValueOrUndefined() : is_defined(false), value(T()) {}
  ValueOrUndefined(const T& value_) : is_defined(true), value(value_) {}
  ValueOrUndefined(const ValueOrUndefined<T>& other) :
    is_defined(other.is_defined), value(other.value) {}
  
  template <class Y>
  explicit ValueOrUndefined(const ValueOrUndefined<Y>& other) {
    is_defined=other.isDefined();
    if (is_defined) {
      value=other.getValue();
    } else {
      value=Y(); // default constructor
    }
  }

  static ValueOrUndefined<T> UNDEFINED() {return ValueOrUndefined<T>();}

  bool isDefined() const {return is_defined;}
  
  const T& getValue() const {
    if (!is_defined) {
      throw "Value is undefined!";
    };
    return value;
  }

  ValueOrUndefined<T> operator+(const ValueOrUndefined<T>& other) const {
    if (isDefined() && other.isDefined()) {
      return getValue() + other.getValue();
    }
    return ValueOrUndefined<T>();
  }
  
  ValueOrUndefined<T> operator-(const ValueOrUndefined<T>& other) const {
    if (isDefined() && other.isDefined()) {
      return getValue() - other.getValue();
    }
    return ValueOrUndefined<T>();
  }
  
  ValueOrUndefined<T> operator-() const {
    if (isDefined()) {
      return -getValue();
    }
    return ValueOrUndefined<T>();
  }
 

  ValueOrUndefined<T> operator*(const ValueOrUndefined<T>& other) const {
    if (isDefined() && other.isDefined()) {
      return getValue() * other.getValue();
    }
    return ValueOrUndefined<T>();
  }

  ValueOrUndefined<T> operator/(const ValueOrUndefined<T>& other) const {
    if (isDefined() && other.isDefined()) {
      return getValue() / other.getValue();
    }
    return ValueOrUndefined<T>();
  }

  bool operator<(const ValueOrUndefined<T>& other) const {
    CHECK_SUCCESS(isDefined() && other.isDefined());
    return getValue() < other.getValue();
  }

  bool operator>(const ValueOrUndefined<T>& other) const {
    CHECK_SUCCESS(isDefined() && other.isDefined());
    return getValue() > other.getValue();
  }

  bool operator<=(const ValueOrUndefined<T>& other) const {
    CHECK_SUCCESS(isDefined() && other.isDefined());
    return getValue() <= other.getValue();
  }

  bool operator>=(const ValueOrUndefined<T>& other) const {
    CHECK_SUCCESS(isDefined() && other.isDefined());
    return getValue() >= other.getValue();
  }

  bool operator==(const ValueOrUndefined<T>& other) const {
    CHECK_SUCCESS(isDefined() && other.isDefined());
    return getValue() == other.getValue();
  }

  bool operator!=(const ValueOrUndefined<T>& other) const {
    CHECK_SUCCESS(isDefined() && other.isDefined());
    return getValue() != other.getValue();
  }

  bool operator==(const T& other) const {
    CHECK_SUCCESS(isDefined());
    return getValue() == other;
  }

  bool operator!=(const T& other) const {
    CHECK_SUCCESS(isDefined());
    return getValue() != other;
  }

  bool operator<(const T& other) const {
    CHECK_SUCCESS(isDefined());
    return getValue() < other;
  }

  bool operator>(const T& other) const {
    CHECK_SUCCESS(isDefined());
    return getValue() > other;
  }

  bool operator>=(const T& other) const {
    CHECK_SUCCESS(isDefined());
    return getValue() >= other;
  }

  bool operator<=(const T& other) const {
    CHECK_SUCCESS(isDefined());
    return getValue() <= other;
  }

  friend std::ostream& operator<< (std::ostream& o, const ValueOrUndefined<T>& v) {
    if (v.isDefined()) {
      o << v.getValue();
    } else {
      o << "undefined";
    }
    return o;
  }


private:
  bool is_defined;
  T value;
};

typedef ValueOrUndefined<unsigned int> uint_value_t;
typedef ValueOrUndefined<int> int_value_t;

}

#endif
