/***************************************************************************
 *   Copyright (C) 2009 by Joachim Klein                                   *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef XML_HELPERS_HPP
#define XML_HELPERS_HPP

#include <iosfwd>
#include <list>
#include <boost/algorithm/string/replace.hpp>
#include <boost/lexical_cast.hpp>

namespace vereofy {

  inline void xml_print_quoted(std::ostream& out, const std::string& text) {
    std::string::size_type first_problematic_char=text.find_first_of("<>'\"&");
    if (first_problematic_char==std::string::npos) {
      // this string does not need quoting
      out << text;
      return;
    }

    std::string quoted_text(text);
    boost::algorithm::replace_all(quoted_text, "&", "&amp;");
    boost::algorithm::replace_all(quoted_text, "\"", "&quot;");
    boost::algorithm::replace_all(quoted_text, "'", "&apos;");
    boost::algorithm::replace_all(quoted_text, "<", "&lt;");
    boost::algorithm::replace_all(quoted_text, ">", "&gt;");

    out << quoted_text;
  }

  class xml_parameter {
  public:
    xml_parameter(const std::string& name_, const std::string& value_) :
      name(name_), value(value_), truth_value(true) {};
    
    template <typename Value>
    xml_parameter(const std::string& name_, const Value& value_) :
      name(name_), value(boost::lexical_cast<std::string>(value_)), truth_value(true) {};

    void print(std::ostream& out) const {
      if (!truth_value) return; 
      out << name << "=\"";
      xml_print_quoted(out, value);
      out << "\"";
    }

    xml_parameter& only_if(bool truth_value_) {
      truth_value=truth_value_;
      return *this;
    }

  private:
    std::string name;
    std::string value;

    bool truth_value;
  };

  class xml_parameters {
  public:
    xml_parameters(const xml_parameter& param) {
      params.push_back(param);
    }

    xml_parameters& operator+(const xml_parameter& other) {
      params.push_back(other);
      return *this;
    }

    void print(std::ostream& out) const {
      for (std::list<xml_parameter>::const_iterator it=params.begin();
           it!=params.end();
           it++) {
        out << " ";
          
        it->print(out);
      }
    }

  private:
    std::list<xml_parameter> params;
  };

  inline xml_parameters operator+(const xml_parameter& a, const xml_parameter& b) {
    return xml_parameters(a)+b;
  }


  inline std::ostream& xml_tag_open(std::ostream &out, const std::string& name, bool no_crlf=false) {
    out << "<" << name << ">";
    if (!no_crlf) out << "\n";
    return out;
  }

  inline std::ostream& xml_tag_open(std::ostream &out, const std::string& name, const xml_parameters& parameters, bool no_crlf=false) {
    out << "<" << name;
    parameters.print(out);
    out << ">";
    if (!no_crlf) out << "\n";
    return out;
  }

  inline void xml_tag_close(std::ostream &out, const std::string& name) {
    out << "</"<< name << ">\n";
  }

  inline void xml_tag_simple(std::ostream &out, const std::string& name) {
    out << "<" << name << "/>\n";
  }

  inline void xml_tag_simple(std::ostream &out, const std::string& name, const xml_parameters& parameters) {
    out << "<" << name;
    parameters.print(out);
    out << "/>\n";
  }

  inline void xml_text(std::ostream& out, const std::string& text) {
    xml_print_quoted(out, text);
  }

  inline void xml_header(std::ostream &out) {
    out << "<?xml version=\"1.0\" encoding='ISO-8859-1'?>\n";
  }

};


#endif
