#include "BDD_G4G_NRA.hpp"
#include "SCCCounter.hpp"
#include "common/varset.h"
#include <boost/lexical_cast.hpp>

namespace vereofy { namespace ltl_check {


void BDD_G4G_NRA::debug_bdd(const std::string& dsc, const BDDFunction& function) {
    std::cout << dsc << std::endl << std::endl;
    std::cout << "Essential variables: ";
    function.printEssentialVariables();
    std::cout << "Paths: " << std::endl;
    function.printPaths();
    std::cout << std::endl;
}

BDDFunction BDD_G4G_NRA::post(const BDDFunction& Q) {
    BDDFunction result = Q & delta;
    TRACE_BDD_FUNCTION("g4g_post", Q & delta, "q_delta");
    result = result.exists(csTransitionVariables);
    result = result.exists(get_state_vars(false));
    TRACE_BDD_FUNCTION("g4g_post", result, "result");
    result = result.moveUp();
    TRACE_BDD_FUNCTION("g4g_post", result, "result");
    return result;
}

BDDFunction BDD_G4G_NRA::calc_reachable() {
    BDDFunctionContainer dependencies;
    dependencies.push(delta);
    dependencies.push(getInitialStates());
    dependencies.push(csTransitionVariables & get_state_vars(false));

    if (cachedReachableStates.isValid() &&
        !cachedReachableStates.haveDependenciesChanged(dependencies)) {
      // valid result was cached
      return cachedReachableStates.getValue();
    }

    stopwatch_reachable.start();
    TRACE_BDD_FUNCTION("g4g_calc_reachable", delta, "");
    BDDFunction oldReachable=BDDFunction::ZERO();
    BDDFunction reachable=getInitialStates();
    int steps = 0;
    while (reachable!=oldReachable) {
      TRACE_BDD_FUNCTION("g4g_calc_reachable", reachable, "step " << steps);
      oldReachable=reachable;
      ++steps;
      reachable |= post(reachable);
    }
    stopwatch_reachable.stop();
    
    TRACE_BDD_FUNCTION("g4g_calc_reachable", reachable, "finished");

    // cache result
    cachedReachableStates=BDDCachedFunction(reachable, dependencies);
    return reachable;
}


double BDD_G4G_NRA::count_states(const BDDFunction& Q) {
    return Q.countAssignments(get_state_vars(false));
}

double BDD_G4G_NRA::count_reachable() {
    return count_states(calc_reachable());
}

BDDFunction BDD_G4G_NRA::get_state_vars(const bool& primed) const {
  return primed ? 
    csStateVariables.moveDown() :
    csStateVariables;
}

BDDFunctionContainer BDD_G4G_NRA::get_state_var_container(const bool& primed) {
  if (primed) {
    return container_moveDown(state_variables);
  } else {
    return state_variables;
  }
}


void BDD_G4G_NRA::toDot(std::ostream& out) {
  VarManagerG4G::ReorderLock reorder_lock; // inhibit reordering for the rest of this scope
  
  out << "digraph G {\n";

  BDDFunction reachableCopy = calc_reachable();
  TRACE_BDD_FUNCTION("toDot", reachableCopy, "");
  while (reachableCopy != BDDFunction::ZERO()) {
    BDDFunction from=popAssignment(reachableCopy, get_state_vars(false));
    out << "  \"" << assignmentToBin(from, get_state_vars(false)) << "\" [";
    
    bool is_initial= (from == getInitialStates());
    bool is_accepting = false; // TODO

    out << "shape=box, color=";
    if (is_initial && is_accepting) {
      out << "violet";
    } else if (is_initial) {
      out << "blue";
    } else if (is_accepting) {
      out << "red";
    } else {
      out << "black";
    }

    // add label
    out << ", label=\"";
    out << print_single_state(from, true);
    out << "\"]\n";
    
    out << std::endl;

    BDDFunction deltaFrom = (delta & from).exists(get_state_vars(false));
    while (deltaFrom != BDDFunction::ZERO()) {
      BDDFunction to_pair = popAssignment(deltaFrom, 
                                          get_state_vars(true) & 
                                          csTransitionVariables);
      
      out << " \"" << assignmentToBin(from, get_state_vars(false)) << "\" -> \""
          << assignmentToBin(to_pair, get_state_vars(true))
          << "\" [label=\"" << assignmentToBin(to_pair, csTransitionVariables)
          << "\" ];\n";
    }
  }

  out << "}\n";
}


std::string BDD_G4G_NRA::print_single_state(const BDDFunction& q, bool with_acceptance) {
    std::string result;

    result = assignmentToBin(q, get_state_vars(false));

    if (with_acceptance) {
      if (acceptance_parity) {
        for (unsigned int i=0;i<acceptance_parity->countColors();i++) {
          if (bdd_is_element_of_set(q, acceptance_parity->getColor(i))) {
            result+=" | "+boost::lexical_cast<std::string>(i);
          }
        }
      }
      result+=" [";
      for (unsigned int i=0;i<acceptance_rabin.countPairs();i++) {
        bool in_R = bdd_is_element_of_set(q, acceptance_rabin.getPair(i).R);
        bool in_G = bdd_is_element_of_set(q, acceptance_rabin.getPair(i).G);
        
        if (in_R) {
          result += "-" + boost::lexical_cast<std::string>(i);
        } else if (in_G) {
          result += "+" + boost::lexical_cast<std::string>(i);
        }
      }
      result+="]";
    }
      
    return result;
}

unsigned long long int BDD_G4G_NRA::state2int(const BDDFunction& q) {
    return reachables_index[q];
}

std::string BDD_G4G_NRA::ap2string(const BDDFunction& ap) {
    const BDDFunctionContainer& ap_container = varmanager.getAPVariables();
    if(ap_container.size() == 0) {
        return "t";
    } else {
        std::string result("");
        for(unsigned int i = 0; i < ap_container.size() - 1; i++) {
            result += "& ";
        }
        for(unsigned int i = 0; i < ap_container.size(); i++) {
            if(bdd_is_bit_set(ap, ap_container.at(i))) {
                result +=  apset->getAP(i)+ " ";
            } else {
                result += "!" + apset->getAP(i) + " ";
            }
        }
        return result;
    }
}

void BDD_G4G_NRA::create_reachable_index() {
    BDDFunction reachable_copy = calc_reachable();
    unsigned long long int index = 0;
    while(reachable_copy != BDDFunction::ZERO()) {
        BDDFunction state = popAssignment(reachable_copy, get_state_vars(false));
        reachables_index[state] = index;
        index++;
    }
}


//Coding of the states
//every state has an index i
//Assume the G4G NRA has n Rabin pairs
//The uncopied state has number i * n
//state i of copy j has number i * n + j
void BDD_G4G_NRA::to_lbtt(std::ostream& out, bool detailed, bool parity) {
    VarManagerG4G::ReorderLock reorder_lock; // inhibit reordering for the rest of this scope
    create_reachable_index();
    unsigned int reachable_nba_states;
    if(parity) {
        reachable_nba_states = (acceptance_parity->countColors()/2 + 1) * (reachables_index.size());
    } else {
        reachable_nba_states = (acceptance_rabin.countPairs() + 1) * (reachables_index.size());
    }

    out << reachable_nba_states << " " << "1s" << std::endl;

    BDDFunction reachable_copy = calc_reachable();
    while(reachable_copy != BDDFunction::ZERO()) {
        BDDFunction from=popAssignment(reachable_copy, get_state_vars(false));
        unsigned int from_index;
        if(parity) {
            from_index = (state2int(from) * (acceptance_parity->countColors() / 2 + 1));
        } else {
            from_index = (state2int(from) * (acceptance_rabin.countPairs() + 1));
        }
        out << from_index << " ";
        if(bdd_is_element_of_set(from, getInitialStates())) {
            out << "1 -1";
        } else {
            out << "0 -1";
        }
        if (detailed) {
          out << " // " << print_single_state(from, true);
        }
        out << std::endl;

        //calculate the successors states
        BDDFunction deltaFrom = (delta & from).exists(get_state_vars(false));
        while (deltaFrom != BDDFunction::ZERO()) {
            BDDFunction to_pair = popAssignment(deltaFrom, 
                                              get_state_vars(true) & 
                                              csTransitionVariables);
          
            BDDFunction to_state = to_pair.exists(csTransitionVariables).moveUp();
            unsigned long long int state_index;
            if(parity) {
                state_index = state2int(to_state) * (acceptance_parity->countColors()/2 + 1);
            } else {
                state_index = state2int(to_state) * (acceptance_rabin.countPairs() + 1);
            }
            out << state_index << " ";
            BDDFunction ap = to_pair.exists(get_state_vars(true));
            out << ap2string(ap) << std::endl;
            if(parity) {
                for(unsigned int i = 1; i <= (acceptance_parity->countColors()/2); i++) {
                    out << (state_index + i) << " " << ap2string(ap) << std::endl;
                }
            } else {
                for(unsigned int i = 1; i <= acceptance_rabin.countPairs(); i++) {
                    out << (state_index + i) << " " << ap2string(ap) << std::endl;
                }
            }
        }
        out << "-1" << std::endl;

        if(parity) {
            //calculate for parity condition 
            BDDFunction lower_color = BDDFunction::ZERO();
            for(unsigned int i = 0; i < acceptance_parity->countColors(); i += 2) {
                if(i > 0) {
                    lower_color |= acceptance_parity->getColor(i-1);
                }
                out << (state2int(from) * (acceptance_parity->countColors()/2 + 1) + i/2 + 1) << " ";
                if(bdd_is_element_of_set(from, acceptance_parity->getColor(i)) && !bdd_is_element_of_set(from, lower_color)) {
                    out << "0 0 -1";
                } else {
                    out << "0 -1";
                }
                if (detailed) {
                  out << " // copy " << i/2 << " of " << print_single_state(from, true);
                }
                out << std::endl;
            

                if(!bdd_is_element_of_set(from, lower_color)) {
                    deltaFrom = (delta & from).exists(get_state_vars(false));
                    while (deltaFrom != BDDFunction::ZERO()) {
                        BDDFunction to_pair = popAssignment(deltaFrom, 
                                                          get_state_vars(true) & 
                                                          csTransitionVariables);
                      
                        BDDFunction to_state = to_pair.exists(csTransitionVariables).moveUp();
                        unsigned long long int state_index = state2int(to_state) * (acceptance_parity->countColors() / 2 + 1) + i/2 + 1;
                        out << state_index << " ";
                        BDDFunction ap = to_pair.exists(get_state_vars(true));
                        out << ap2string(ap) << std::endl;
                    }
                }
                
                lower_color |= acceptance_parity->getColor(i);
                out << "-1" << std::endl;
            }
        } else {
            //calculate for rabin pairs 
            for(unsigned int i = 0; i < acceptance_rabin.countPairs(); i++) {
                BDD_RabinPair rabin_pair = acceptance_rabin.getPair(i);
                out << (state2int(from) * (acceptance_rabin.countPairs() + 1) + i + 1) << " ";
                if(bdd_is_element_of_set(from, rabin_pair.G)) {
                    out << "0 0 -1";
                } else {
                    out << "0 -1";
                }
                if (detailed) {
                  out << " // copy " << i << " of " << print_single_state(from, true);
                }
                out << std::endl;
            

                if(!bdd_is_element_of_set(from, rabin_pair.R)) {
                    deltaFrom = (delta & from).exists(get_state_vars(false));
                    while (deltaFrom != BDDFunction::ZERO()) {
                        BDDFunction to_pair = popAssignment(deltaFrom, 
                                                          get_state_vars(true) & 
                                                          csTransitionVariables);
                      
                        BDDFunction to_state = to_pair.exists(csTransitionVariables).moveUp();
                        unsigned long long int state_index = state2int(to_state) * (acceptance_rabin.countPairs() + 1) + i + 1;
                        out << state_index << " ";
                        BDDFunction ap = to_pair.exists(get_state_vars(true));
                        out << ap2string(ap) << std::endl;
                    }
                }
                out << "-1" << std::endl;
            }
        }
    }
}

unsigned int BDD_G4G_NRA::var_count() {
    return state_variables.size();
}
    
BDDFunction BDD_G4G_NRA::no_change() {
  return pairwiseEQUIV(get_state_var_container(false), get_state_var_container(true));
}

bool BDD_G4G_NRA::check_isTotal() {
  BDDFunction reachable = calc_reachable();
  BDDFunction total_nra = 
    getDelta()
    .exists(get_state_vars(true))
    .forall(csTransitionVariables);

  if (!bdd_set_is_contained(reachable, total_nra)) {
    TRACE_BDD_FUNCTION("check_isTotal()", reachable & !total_nra, "");
    return false;
  }
  
  return true;
}


unsigned long BDD_G4G_NRA::countSCCs() {
  SCCCounter counter;
  
  std::cerr << "Counting SCCs...\n";
  
  BDDFunction reachable = calc_reachable();
  BDDFunction delta_xy = delta.exists(csTransitionVariables);
  return SCCCounter::count_sccs(reachable, delta_xy, get_state_vars(false));
}


}}
