#ifndef BDD_G4G_NRA_HPP
#define BDD_G4G_NRA_HPP

#include "VarManagerG4G.hpp"

#include "vereofy/setup.hpp"
#include "vereofy/common/errors.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/modelcheck/checkers/ltl_check/NBA_I.hpp"
#include "BDD_NBA_Unary.hpp"
#include "vereofy/common/stopwatch.hpp"
#include "vereofy/common/bdd_cached_function.hpp"
#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>
#include <string>
#include <iostream>

#include "BDD_RabinAcceptance.hpp"
#include "BDD_ParityAcceptance.hpp"

namespace vereofy {namespace ltl_check {

class BDD_G4G_NRA {
public:
  typedef boost::shared_ptr<BDD_G4G_NRA> ptr;

  BDD_G4G_NRA(APSet_cp apset_,
              VarManagerG4G& varmanager_) :
    apset(apset_),
    varmanager(varmanager_),
    
    initial(BDDFunction::ZERO()),
    delta(BDDFunction::ZERO()),
    csTransitionVariables(BDDFunction::ONE()),
    csStateVariables(BDDFunction::ONE()) {
  }

  virtual ~BDD_G4G_NRA() {
    initial = BDDFunction::ZERO();
    delta = BDDFunction::ZERO();
    csTransitionVariables = BDDFunction::ZERO();
    csStateVariables = BDDFunction::ZERO();

    state_variables.clear();
    
    acceptance_parity.reset();
    acceptance_rabin = BDD_RabinAcceptance();  // clear
    reachables_index.clear();
    cachedReachableStates = BDDCachedFunction();

    // last: remove registered State Vars
    registered_vars.clear();
  }

  virtual void slim() {}
  
  const BDDFunction& getInitialStates() const {
    return initial;
  }

  const BDDFunction& getDelta() const {
    return delta;
  } 

  virtual void print_stats(std::ostream& out, bool count_sccs, const std::string &prefix) {
    out << prefix << "G4G-number of reachable states: " << count_reachable() << std::endl;
    if (acceptance_parity) {
      out << prefix << "G4G-parity colors: " << acceptance_parity->countColors() << std::endl;
    }
    out << prefix << "G4G-Rabin pairs: " << acceptance_rabin.countPairs() << std::endl;
    out << prefix << "G4G-time for constructing: " << stopwatch.getElapsedTimeString() << std::endl;
    out << prefix << "G4G-time for calculating the reachable space: " << stopwatch_reachable.getElapsedTimeString() << std::endl << std::endl;

    out << prefix << "G4G-BDD Encoding" << std::endl;
    out << prefix << "G4G-variables: " << var_count() << std::endl;
    out << prefix << "G4G-initial: " << getInitialStates().size() << std::endl;
    out << prefix << "G4G-delta: " << getDelta().size() << std::endl;
    if (acceptance_parity) {
      for(unsigned int i = 0; i < acceptance_parity->countColors(); i++) {
        out << prefix << "G4G-parity c(" << i << "): " << acceptance_parity->getColor(i).size() << std::endl;
      }
    } else {
      for(unsigned int i = 0; i < acceptance_rabin.countPairs(); i++) {
        out << prefix << "G4G-rabin R(" << i << "): " << acceptance_rabin.getPair(i).R.size() << std::endl;
        out << prefix << "G4G-rabin G(" << i << "): " << acceptance_rabin.getPair(i).G.size() << std::endl;
      }
    }

    if (count_sccs) {
      out << prefix << "SCCs: " << countSCCs() << std::endl;
    }
  }

  unsigned long countSCCs();
  BDDFunction post(const BDDFunction& Q);
  BDDFunction calc_reachable();
  double count_reachable();

  // return a delta where no change of state happens
  BDDFunction no_change();

  BDDFunction get_state_vars(const bool& primed) const;
  BDDFunctionContainer get_state_var_container(const bool& primed);
  BDDFunction get_transition_vars() const {return csTransitionVariables;}

  void toDot(std::ostream& out);
  void to_lbtt(std::ostream& out, bool detailed, bool parity = false); 

  unsigned int countRabinPairs() const {
    return acceptance_rabin.countPairs();
  }
  
  const BDD_RabinPair& getRabinPair(unsigned int i) const {
    return acceptance_rabin.getPair(i);
  }

  virtual std::string print_single_state(const BDDFunction& q, bool with_acceptance=false);

  APSet_cp getAPSet() const {return apset;}
  VarManagerG4G& getVarManager() {return varmanager;}

  bool check_isTotal();

  
  const std::vector<StateVars_G4G::ptr>& getRegisteredStateVars_G4G() {
    return registered_vars;
  }

protected:
  APSet_cp apset;
  VarManagerG4G& varmanager;

  std::vector<StateVars_G4G::ptr> registered_vars;

  StopWatch stopwatch;
  StopWatch stopwatch_reachable;

  BDDFunction initial;
  BDDFunction delta;

  BDDFunction csTransitionVariables;
  BDDFunction csStateVariables;
  BDDFunctionContainer state_variables;
  
  BDD_ParityAcceptance::ptr acceptance_parity;
  BDD_RabinAcceptance acceptance_rabin;
  std::map<BDDFunction,unsigned long long int,BDDFunction::less> reachables_index;

  BDDCachedFunction cachedReachableStates;

  void registerStateVars_G4G(StateVars_G4G::ptr vars) {
    registered_vars.push_back(vars);
  }

  void registerStateVars_G4G(const std::vector<StateVars_G4G::ptr> &vars) {
    registered_vars.insert(registered_vars.end(),
                           vars.begin(), vars.end());
  }

  void unregisterStateVars_G4G(StateVars_G4G::ptr vars) {
    for (std::vector<StateVars_G4G::ptr>::iterator it = registered_vars.begin();
         it != registered_vars.end();
         ++it) {
      if (*it == vars) {
        registered_vars.erase(it);
        return;
      }
    }
    throw "Unregistering StateVars_G4G that was not registered!";
  }
  
  void create_reachable_index();
  unsigned long long int state2int(const BDDFunction& q); 
  void debug_bdd(const std::string& dsc, const BDDFunction& function); 
  std::string ap2string(const BDDFunction& ap);
  void print_states(BDDFunction Q);
  double count_states(const BDDFunction& Q);
  unsigned int var_count();
};

}}

#endif
