#include "BDD_G4G_Transformation.hpp"
#include "SCCCounter.hpp"
#include "common/varset.h"
#include <boost/lexical_cast.hpp>

namespace vereofy { namespace ltl_check {

BDD_G4G_Transformation::~BDD_G4G_Transformation() {
  nba.reset();
  vars.reset();

  state_pattern = BDDFunction::ZERO();
  state_pattern_primed = BDDFunction::ZERO();

  // registered state vars will be removed
  // in ~BDD_G4G_NRA
}



//apply transformation from Henzinger and Piterman "Solving games without Determinization"
//the result is saved internally, use getters for taking the result
//@param nba_: the nba which should be made g4g
//@param max_set_: for iterative approach, only max_set-tuples of marked sets are considered
//Ansatz mit Max_set stimmt noch nicht, die marked state mit Index > max_set müssen noch genullt werden
void BDD_G4G_Transformation::from_nba(const BDD_NBA_Unary::ptr nba_, 
                                      StateVars_G4G::ptr vars_,
                                      const unsigned int max_set_) {

    stopwatch.start();
    nba = nba_;

    if (!vars) {
      vars = vars_;
      registerStateVars_G4G(vars);
    } else if (vars != vars_) {
      unregisterStateVars_G4G(vars);
      vars = vars_;
      registerStateVars_G4G(vars);
    } else {
      // vars == vars_, nothing to do
    }

    max_set = max_set_;

    if(max_set > nba->getStateCount()) {
        throw "Parameter max-set cannot be larger than the number of states used in the NBA";
    }

    csTransitionVariables = container2cs(varmanager.getAPVariables());

    // initialize csStateVariables and state_variables from A, B until max_set
    csStateVariables = BDDFunction::ONE();
    for(unsigned int i = 0; i < max_set; i++) {
      csStateVariables &= container2cs(vars->getA(i)) & container2cs(vars->getB(i));

      state_variables.push(vars->getA(i));
      state_variables.push(vars->getB(i));
    }
    

    initial = (pairwiseEQUIV(vars->getA(0), nba->getStateVariables()) & nba->getInitialStates()).exists(nba->getStateVars());
    
    
    if(nba->getAcceptingContainer().at(nba->getInitialStateIndex()) == BDDFunction::ONE()) {
        initial &= (pairwiseEQUIV(vars->getB(0), nba->getStateVariables()) & nba->getInitialStates()).exists(nba->getStateVars());    
    } else {
        initial &= empty_set(vars->getB(0));
    } 

    for(unsigned int i = 1; i < max_set; i++) {
        initial &= empty_set(vars->getA(i)) & empty_set(vars->getB(i));
    }
    TRACE_BDD_FUNCTION("g4g_initial", initial, "");

    state_pattern = BDDFunction::ONE();
    if(twice_state_pattern) {
        state_pattern = create_state_pattern(false);
    }
    
    state_pattern_primed = create_state_pattern(true);

    BDDFunction nba_delta = nba->getDelta();

    //Creating Delta of G4G
    
    delta = BDDFunction::ONE();
    for(unsigned int i = 0; i < max_set; i++) {
        const BDDFunctionContainer A0 = vars->getA(0);
        const BDDFunctionContainer B0 = vars->getB(0);
        const BDDFunctionContainer Ai = vars->getA(i);
        const BDDFunctionContainer Bi = vars->getB(i);

        //first case Ai is not an emptyset and does not equal Bi
        BDDFunction Ai_nonempty_inequal_Bi = (!empty_set(Ai)) & (!pairwiseEQUIV(Ai, Bi));

        const BDDFunctionContainer nba_states_unprimed = nba->getStateVariables();
        const BDDFunctionContainer nba_states_primed = container_moveDown(nba->getStateVariables());

        //Defining Ai for the first case
        BDDFunction delta1_Ai = nba_delta &    
                               pairwiseEQUIV(Ai, nba_states_unprimed) & pairwiseIMPLIES(container_moveDown(Ai), nba_states_primed);
        delta1_Ai = delta1_Ai.exists(nba->getStateVars()).exists(nba->getStateVars().moveDown());
        TRACE_BDD_FUNCTION("g4g_delta", delta1_Ai, "cond1_A" + boost::lexical_cast<std::string>(i));

        //Defining Bi for the first case
       
        BDDFunction delta1_Bi = nba_delta & 
                               pairwiseEQUIV(Bi, nba_states_unprimed) & pairwiseIMPLIES(container_moveDown(Bi), nba_states_primed);
        delta1_Bi = delta1_Bi.exists(nba->getStateVars()).exists(nba->getStateVars().moveDown());
        
        //union with A' cap accepting sets
        BDDFunction Bi_prime_cap_alpha = pairwiseIMPLIES(container_moveDown(Bi), nba->getAcceptingContainer());
        //BDDFunction Bi_prime_cap_alpha = pairwiseIMPLIES(container_moveDown(Bi), nba_states_unprimed) & nba->getAccepting();
        delta1_Bi |= Bi_prime_cap_alpha.exists(nba->getStateVars());
        TRACE_BDD_FUNCTION("g4g_delta", delta1_Bi, "cond1_B" + boost::lexical_cast<std::string>(i));

        delta &= (!Ai_nonempty_inequal_Bi) | (delta1_Ai & delta1_Bi);
 
        //second case Ai and Bi are equal and nonempty
        BDDFunction Ai_equal_Bi_nonempty = (!empty_set(Ai)) & pairwiseEQUIV(Ai, Bi);
        //
        //Defining Ai for the second case
        BDDFunction delta2_Ai = nba_delta &
                               pairwiseEQUIV(Ai, nba_states_unprimed) & pairwiseIMPLIES(container_moveDown(Ai), nba_states_primed);
        delta2_Ai = delta2_Ai.exists(nba->getStateVars()).exists(nba->getStateVars().moveDown());
        TRACE_BDD_FUNCTION("g4g_delta", delta2_Ai, "cond2_A" + boost::lexical_cast<std::string>(i));

        //Defining Bi for the second case
        Bi_prime_cap_alpha = pairwiseIMPLIES(container_moveDown(Bi), nba->getAcceptingContainer());
        //Bi_prime_cap_alpha = pairwiseIMPLIES(container_moveDown(Bi), nba_states_unprimed) & nba->getAccepting();
        BDDFunction delta2_Bi = Bi_prime_cap_alpha.exists(nba->getStateVars());
        TRACE_BDD_FUNCTION("g4g_delta", delta2_Bi, "cond2_B" + boost::lexical_cast<std::string>(i));


        delta &= (!Ai_equal_Bi_nonempty) | (delta2_Ai & delta2_Bi); 
 
        if((initial & delta) == BDDFunction::ZERO()) {
            std::cout << "Delta Error" << std::endl;
        }       

        //third case Ai is empty
        BDDFunction Ai_empty = empty_set(Ai);
        //Defining Ai for the third case
        BDDFunction delta3_Ai = nba_delta & 
                               pairwiseEQUIV(A0, nba_states_unprimed) & pairwiseIMPLIES(container_moveDown(Ai), nba_states_primed);
        delta3_Ai = delta3_Ai.exists(nba->getStateVars()).exists(nba->getStateVars().moveDown());
        TRACE_BDD_FUNCTION("g4g_delta", delta2_Bi, "cond3_A" + boost::lexical_cast<std::string>(i));

        //Defining Bi for the third case is not necessary, since Bi subseteq A
        delta &= (!Ai_empty) | delta3_Ai;
 
        if((initial & delta) == BDDFunction::ZERO()) {
            std::cout << "Delta Error" << std::endl;
        }       
    }

    if(twice_state_pattern) {
        delta &= state_pattern & state_pattern_primed;
    } else {
        delta &= state_pattern_primed;
    }

    TRACE_BDD_FUNCTION("g4g_delta", delta, "");
    
    //create accepting condition

    acceptance_parity = BDD_ParityAcceptance::ptr(new BDD_ParityAcceptance());
    acceptance_rabin = BDD_RabinAcceptance();

    //first parity set

    acceptance_parity->addColor(pairwiseEQUIV(vars->getA(0), vars->getB(0)) & !empty_set(vars->getA(0)));
    TRACE_BDD_FUNCTION("g4g_accepting_parity", pairwiseEQUIV(vars->getA(0), vars->getB(0)) & !empty_set(vars->getA(0)), "F0");
    
    BDDFunction Ei = BDDFunction::ZERO();
    BDD_RabinPair pair_0(Ei,
                         pairwiseEQUIV(vars->getA(0), vars->getB(0)) 
                         & !empty_set(vars->getA(0)));
                         
    acceptance_rabin.addPair(pair_0);
    TRACE_BDD_FUNCTION("g4g_accepting_rabin", pair_0.R, "E0");
    TRACE_BDD_FUNCTION("g4g_accepting_rabin", pair_0.G, "F0");
    BDDFunction trap_state = BDDFunction::ONE();
    for(unsigned int i = 0; i < max_set; i++) {
        trap_state &= empty_set(vars->getA(i)) & empty_set(vars->getB(i));
    }
    
    BDDFunction second_parity_set = get_odd_parity_set(0);
    second_parity_set |= trap_state;
    acceptance_parity->addColor(second_parity_set);
    Ei |= second_parity_set;

    TRACE_BDD_FUNCTION("g4g_accepting_parity", second_parity_set, "F1");

    if(max_set > 1) {
        acceptance_parity->addColor(get_even_parity_set(0));
        TRACE_BDD_FUNCTION("g4g_accepting_parity", get_even_parity_set(0), "F2");
        acceptance_rabin.addPair(BDD_RabinPair(Ei, get_even_parity_set(0)));
        TRACE_BDD_FUNCTION("g4g_accepting_rabin", Ei, "E1");
        TRACE_BDD_FUNCTION("g4g_accepting_rabin", get_even_parity_set(0), "F1");
       
        Ei |= get_even_parity_set(0);
    }

    for(int i = 1; i < max_set - 1; i++) {
        acceptance_parity->addColor(get_odd_parity_set(i));
        Ei |= get_odd_parity_set(i);
        TRACE_BDD_FUNCTION("g4g_accepting_parity", get_odd_parity_set(i), "F" + boost::lexical_cast<std::string>(2*i + 1));
        acceptance_parity->addColor(get_even_parity_set(i));
        TRACE_BDD_FUNCTION("g4g_accepting_parity", get_even_parity_set(i), "F" + boost::lexical_cast<std::string>(2*i + 2));
        acceptance_rabin.addPair(BDD_RabinPair(Ei, get_even_parity_set(i)));
        TRACE_BDD_FUNCTION("g4g_accepting_rabin", Ei, "E" + boost::lexical_cast<std::string>(i));
        TRACE_BDD_FUNCTION("g4g_accepting_rabin", get_even_parity_set(i), "F" + boost::lexical_cast<std::string>(i));
        Ei |= get_even_parity_set(i);
    }
    if(max_set > 2) {
        acceptance_parity->addColor(get_odd_parity_set(max_set - 1));
        TRACE_BDD_FUNCTION("g4g_accepting_parity", get_odd_parity_set(max_set-1), "F" + boost::lexical_cast<std::string>(2*max_set - 1));
    }
    stopwatch.stop();

    calc_reachable();
}

BDDFunction BDD_G4G_Transformation::get_odd_parity_set(unsigned int i) {
    BDDFunction result = BDDFunction::ZERO();
    for(unsigned j = i + 1; j <= max_set; j++) {
        result |= ind_F(j);
    }
    result &= ind_E(i+1);
    return result;
}

BDDFunction BDD_G4G_Transformation::get_even_parity_set(unsigned int i) {
    BDDFunction result = BDDFunction::ZERO();
    for(unsigned j = i + 2; j <= max_set; j++) {
        result |= ind_E(j);
    }
    result &= ind_F(i+1);
    return result;
}

BDDFunction BDD_G4G_Transformation::ind_E(unsigned int min) {
    BDDFunction result = state_pattern;
    for(unsigned int i = 0; i < min; i++) {
        result &= !empty_set(vars->getA(i));
    }
    if(min < max_set) {
        result &= empty_set(vars->getA(min));
    }
    return result;
}

BDDFunction BDD_G4G_Transformation::ind_F(unsigned int min) {
    BDDFunction result = state_pattern;
    for(unsigned int i = 0; i < min; i++) {
        result &= (!pairwiseEQUIV(vars->getA(i), vars->getB(i))) | empty_set(vars->getB(i));
    }
    if(min < max_set) {
        result &= pairwiseEQUIV(vars->getA(min), vars->getB(min)) & !empty_set(vars->getA(min));
    }
    return result;
}

BDDFunction BDD_G4G_Transformation::empty_set(const BDDFunctionContainer& container) {
  BDDFunction result = BDDFunction::ONE();
  for(unsigned int i = 0; i < container.size(); i++) {
    result &= !(container.at(i));
  }
  return result;
}


void BDD_G4G_Transformation::print_states(BDDFunction Q) {
    unsigned int step = 0;
    while (Q != BDDFunction::ZERO()) {
        BDDFunction single_state = popAssignment(Q, get_state_vars(false));
        std::cout << step << ".state:" << std::endl;
        for(unsigned int j = 0; j < max_set; j++) {
            BDDFunctionContainer Aj = vars->getA(j);
            BDDFunctionContainer Bj = vars->getB(j);

            std::cout << "A" << j << ": ";
            for(unsigned int i = 0; i < nba->getStateCount(); i++) {
                BDDFunction si_in_Aj = single_state & Aj.at(i);
                if(si_in_Aj != BDDFunction::ZERO()) {
                    std::cout << "s" << i << " ";
                }
            }
            std::cout << std::endl;

            std::cout << "B" << j << ": ";
            for(unsigned int i = 0; i < nba->getStateCount(); i++) {
                BDDFunction si_in_Bj = single_state & Bj.at(i);
                if(si_in_Bj != BDDFunction::ZERO()) {
                    std::cout << "s" << i << " ";
                }
            }
            std::cout << std::endl;
        }
        std::cout << std::endl;
        step++;
    }
}

std::string BDD_G4G_Transformation::print_single_state(const BDDFunction& q, bool with_acceptance) {
    std::string result("");
    for(unsigned int j = 0; j < max_set; j++) {
        result += "{";
        
        bool first = true;
        for(unsigned int i = 0; i < nba->getStateCount(); i++) {
          bool si_in_Aj = (q & vars->getA(j).at(i)) != BDDFunction::ZERO();
          bool si_in_Bj = (q & vars->getB(j).at(i)) != BDDFunction::ZERO();
          
          // sanity check: verify that a state in 
          // Bj appears in Aj as well, Bj -> Aj
          CHECK_SUCCESS( (!si_in_Bj) | si_in_Aj );

          if(si_in_Aj) {
            if (!first) {result += ",";}
            first = false;

            result += "s" + boost::lexical_cast<std::string>(i);
            if(si_in_Bj) {
              result+="!";
            }
          }
        }
        result += "}";
    }

    if (with_acceptance) {
      if (acceptance_parity) {
        for (unsigned int i=0;i<acceptance_parity->countColors();i++) {
          if (bdd_is_element_of_set(q, acceptance_parity->getColor(i))) {
            result+=" | "+boost::lexical_cast<std::string>(i);
          }
        }
      }
      result+=" [";
      for (unsigned int i=0;i<acceptance_rabin.countPairs();i++) {
        bool in_R = bdd_is_element_of_set(q, acceptance_rabin.getPair(i).R);
        bool in_G = bdd_is_element_of_set(q, acceptance_rabin.getPair(i).G);
        
        if (in_R) {
          result += "-" + boost::lexical_cast<std::string>(i);
        } else if (in_G) {
          result += "+" + boost::lexical_cast<std::string>(i);
        }
      }
      result+="]";
    }
      
    return result;
}



BDDFunction BDD_G4G_Transformation::create_state_pattern(const bool& primed) {
    //for all i Bi must be a subseteq of Ai
    BDDFunction B_subseteq_of_A = BDDFunction::ONE();
    for(unsigned int i = 0; i < max_set; i++) {
        if(primed) {
            B_subseteq_of_A &= pairwiseIMPLIES(container_moveDown(vars->getB(i)), container_moveDown(vars->getA(i)));
        } else {
            B_subseteq_of_A &= pairwiseIMPLIES(vars->getB(i), vars->getA(i));
        }
    }
    TRACE_BDD_FUNCTION("g4g_state_pattern", B_subseteq_of_A, "B_subseteq_A");

    //for all i Ai == emptyset implies Ai+1==emptyset
    BDDFunction empty_set_implies_empty_successor = BDDFunction::ONE();
    for(unsigned int i = 0; i < max_set - 1; i++) {
        if(primed) {
            empty_set_implies_empty_successor &= (!empty_set(container_moveDown(vars->getA(i)))) |
                empty_set(container_moveDown(vars->getA(i+1)));
        } else {
            empty_set_implies_empty_successor &= (!empty_set(vars->getA(i))) | empty_set(vars->getA(i+1));

        }
    }
    TRACE_BDD_FUNCTION("g4g_state_pattern", empty_set_implies_empty_successor, "empty_set_implies_empty_successor");

    BDDFunction no_duplication = BDDFunction::ONE();
    if(!duplicate_sets) {
        for(unsigned int i = 1; i + 1 < max_set; i++) {
            BDDFunctionContainer Ai;
            if(primed) {
                Ai = container_moveDown(vars->getA(i));
            } else {
                Ai = vars->getA(i);
            }
            BDDFunctionContainer Bi;
            if(primed) {
                Bi = container_moveDown(vars->getB(i));
            } else {
                Bi = vars->getB(i);
            }
            for(unsigned int j = i + 1; j < max_set; j++) {
                BDDFunctionContainer Aj;
                if(primed) {
                    Aj = container_moveDown(vars->getA(j));
                } else {
                    Aj = vars->getA(j);
                }
                BDDFunctionContainer Bj;
                if(primed) {
                    Bj = container_moveDown(vars->getB(j));
                } else {
                    Bj = vars->getB(j);
                }

                no_duplication &= (!(pairwiseEQUIV(Ai,Aj) & pairwiseEQUIV(Bi,Bj))) |
                    empty_set(Ai);
            }
        }
        //for(unsigned int i = 1; i + 2 < max_set; i += 2) {
        //    BDDFunctionContainer Ai;
        //    if(primed) {
        //        Ai = container_moveDown(vars->getA(i));
        //    } else {
        //        Ai = vars->getA(i);
        //    }
        //    BDDFunctionContainer Bi;
        //    if(primed) {
        //        Bi = container_moveDown(vars->getB(i));
        //    } else {
        //        Bi = vars->getB(i);
        //    }
        //    for(int j = i + 2; j < max_set; j += 2) {
        //        BDDFunctionContainer Aj;
        //        if(primed) {
        //            Aj = container_moveDown(vars->getA(j));
        //        } else {
        //            Aj = vars->getA(j);
        //        }
        //        BDDFunctionContainer Bj;
        //        if(primed) {
        //            Bj = container_moveDown(vars->getB(j));
        //        } else {
        //            Bj = vars->getB(j);
        //        }

        //        no_duplication &= (!(pairwiseEQUIV(Ai,Aj) & pairwiseEQUIV(Bi,Bj))) |
        //            empty_set(Ai);
        //    }
        //}
    }

    if(complete_state_pattern) {
        //for all i < j: either (A_i cap A_j = emptyset) or (Aj subseteq Bi)
        BDDFunction second_condition = BDDFunction::ONE();
        for(unsigned int i = 0; i < max_set - 1; i++) {
            for(unsigned int j = i + 1; j < max_set; j++) {
                BDDFunction Ai_Aj_disjunct = BDDFunction::ONE();
                for(unsigned int k = 0; k < nba->getStateCount(); k++) {
                    //s_k should not be in Ai and Aj at the same time
                    if(primed) {
                        Ai_Aj_disjunct &= !(container_moveDown(vars->getA(i)).at(k) &
                                container_moveDown(vars->getA(j)).at(k));
                    } else {
                        Ai_Aj_disjunct &= !(vars->getA(i).at(k) & vars->getA(j).at(k));
                    }
                }

                if(primed) {
                    BDDFunction Aj_subseteq_of_Bi = pairwiseIMPLIES(container_moveDown(vars->getA(j)), container_moveDown(vars->getB(i)));
                    second_condition &= Ai_Aj_disjunct | Aj_subseteq_of_Bi; 
                } else {
                    BDDFunction Aj_subseteq_of_Bi = pairwiseIMPLIES(vars->getA(j), vars->getB(i));
                    second_condition &= Ai_Aj_disjunct | Aj_subseteq_of_Bi; 
                }
            }
        }
        TRACE_BDD_FUNCTION("g4g_state_pattern", second_condition, "second_condition");

        TRACE_BDD_FUNCTION("g4g_state_pattern", B_subseteq_of_A & empty_set_implies_empty_successor & second_condition & no_duplication, (primed ? "primed" : "unprimed"));
        return B_subseteq_of_A & empty_set_implies_empty_successor & second_condition & no_duplication;
    } else {
        TRACE_BDD_FUNCTION("g4g_state_pattern", B_subseteq_of_A & empty_set_implies_empty_successor & no_duplication, (primed ? "primed" : "unprimed"));
        return B_subseteq_of_A & empty_set_implies_empty_successor & no_duplication;
    }
}

}}
