
#ifndef BDD_NBA_UNARY_HPP
#define BDD_NBA_UNARY_HPP

#include "VarManagerG4G.hpp"
#include "vereofy/setup.hpp"

#include "vereofy/common/errors.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/modelcheck/checkers/ltl_check/NBA_I.hpp"
#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>
#include <string>
#include <iostream>



namespace vereofy {namespace ltl_check {

class BDD_NBA_Unary : public NBA_I {
public:
  typedef boost::shared_ptr<BDD_NBA_Unary> ptr;

  BDD_NBA_Unary(APSet_cp apset_) :
  varmanager(*(VarManagerG4G*)NULL) {
    throw "Missing VarManagerG4G!";
  }

  BDD_NBA_Unary(APSet_cp apset_,
                VarManagerG4G& varmanager_,
                std::ostream* dot_stream_=(std::ostream*)(NULL)) :
    apset(apset_),
    varmanager(varmanager_),
    vars(new StateVars_NBA(varmanager)),
    
    states(0),
    var_count(0),
    initial(BDDFunction::ZERO()),
    delta(BDDFunction::ZERO()),
    all_states(BDDFunction::ZERO()),
    state_vars(BDDFunction::ONE()),
    apset_vars(container2cs(varmanager.getAPVariables())),
    dot_stream(dot_stream_) {

    accepting_container.clear();

    if (dot_stream) {
      (*dot_stream) << "digraph { \n";
    }
  }
  
  virtual ~BDD_NBA_Unary() {
    initial = BDDFunction::ZERO();
    delta = BDDFunction::ZERO();
    all_states = BDDFunction::ZERO();
    state_vars = BDDFunction::ZERO();
    apset_vars = BDDFunction::ZERO();
    accepting_container.clear();
    state_variables.clear();
    to_implications.clear();

    // last
    vars.reset();
  }

  /** 
   * Create a new state.
   * @return the index of the new state
   */
  virtual unsigned int nba_i_newState();

  /**
   * Add an edge from state <i>from</i> to state <i>to</i>
   * for the edges covered by the APMonom.
   * @param from the index of the 'from' state
   * @param m the APMonom
   * @param to the index of the 'to' state
   */
  virtual void nba_i_addEdge(unsigned int from,
			     const APMonom& m,
			     unsigned int to);

  /**
   * Get the underlying APSet 
   * @return a const pointer to the APSet
   */
  virtual APSet_cp nba_i_getAPSet();

  /** 
   * Set the final flag (accepting) for a state.
   * @param state the state index
   * @param final the flag
   */
  virtual void nba_i_setFinal(unsigned int state, bool final);
  
  /**
   * Set the state as the start state.
   * @param state the state index
   */
  virtual void nba_i_setStartState(unsigned int state);

  /** Constructing the NBA is finished. */
  virtual void nba_i_finished();

  /** Switch the APSet to another with the same number of APs. */
  void switchAPSet(APSet_cp new_apset) {
    if (new_apset->size()!=apset->size()) {
      throw std::string("New APSet has to have the same size as the old APSet!");
    }
    
    apset=new_apset;
  }

  APSet_cp getAPSet() const {return apset;}


  const BDDFunction& getInitialStates() const {
    return initial;
  }

  //const BDDFunction& getAccepting() const {
  //  return accepting;
  //}

  const BDDFunctionContainer& getAcceptingContainer() const {
    return accepting_container;
  }

  const BDDFunction& getDelta() const {
    return delta;
  } 

  unsigned int getStateCount() const {
    return states;
  }

  BDDFunction getAll() {
    return all_states;
  }

  unsigned int getInitialStateIndex() {
    return initial_state_index;
  }

  const StateVars_NBA& getStateVariables() const {return *vars;}

  const BDDFunction& getStateVars() const {return state_vars;}
  const BDDFunction& getAPSetVars() const {return apset_vars;}
  
  BDDFunction post(BDDFunction Q) {
    return (Q&delta).exists(state_vars & apset_vars).moveUp();
  }

  BDDFunction pre(BDDFunction Q) {
    BDDFunction deltaQ=delta.exists(apset_vars);
    return pre(Q, deltaQ);
  }

  BDDFunction pre(BDDFunction Q, BDDFunction deltaQ) {
    return (Q.moveDown()&deltaQ).exists(state_vars.moveDown());
  }
  
  BDDFunction cycle_states(BDDFunction Q) {
    BDDFunction Q_old=Q;
    BDDFunction Q_new;
    while (true) {
      if (debugLevel>=4) {
	std::cout << "**********************\n";
	Q_old.printPaths();
	std::cout << "Pre:\n";
      }
      BDDFunction preimg=(Q_old.moveDown()&delta).exists(state_vars.moveDown());
      if (debugLevel>=4) {
	preimg.printPaths();
      }
      Q_new=Q_old&preimg;
      if (Q_new == Q_old) {
	break;
      }
      
      Q_old=Q_new;
    }
    
    return Q_new;
  }

  typedef BDDFunction (*fn_precalc)(BDDFunction);
  
  //BDDFunction fair_cycle(fn_precalc precalc) {
  //  BDDFunction Z, Z_old;
  //  
  //  Z=Z_old=all_states;
  //  while (true) {
  //    BDDFunction Q, Q_old;
  //    Q=Q_old=Z & getAccepting();
  //    while (true) {
  //  Q|=(*precalc)(Q);
  //  if (Q==Q_old) {break;}
  //  Q_old=Q;
  //    }

  //    Z&=(*precalc)(Q);

  //    if (Z==Z_old) {break;}
  //    Z_old=Z;
  //  }    
  //  return Z;
  //}

  BDDFunction calcReachable();

  void print() {
    std::cout << "NBA:\n";
    std::cout << "States: \n";
    all_states.printPaths();

    std::cout << "\nInitial: \n";
    initial.printPaths();

    //std::cout << "\nAccepting" << std::endl;
    //accepting.printPaths();

    std::cout << "\nDelta" << std::endl;
    delta.printPaths();

    std::cout << "\nState Variables" << std::endl;
    state_vars.printPaths();
  }

  void toDot(std::ostream& out);

protected:
  APSet_cp apset;
  VarManagerG4G& varmanager;
  StateVars_NBA::ptr vars;

  unsigned int states;

  unsigned int var_count;

  BDDFunction initial;
  unsigned int initial_state_index;
  //BDDFunction accepting;
  BDDFunctionContainer accepting_container;
  BDDFunction delta;
  BDDFunction all_states;

  BDDFunction state_vars;
  BDDFunction apset_vars;

  BDDFunctionContainer state_variables;
  BDDFunctionContainer to_implications;

  std::ostream* dot_stream;

  BDDFunction state2bdd(unsigned int state, bool copy) {
    CHECK_SUCCESS(state <= states);

    BDDFunction result = BDDFunction::ONE();
    for (unsigned int i=0; i<states; i++) {
      if (i == state) {
        result &= state_variables.at(i);
      } else {
        result &= !state_variables.at(i);
      }
    }

    if (copy) {
      return result.moveDown();
    } else {
      return result;
    }
  }
  
  BDDFunction monom2bdd(const APMonom& m) const {
    BDDFunction result=BDDFunction::ONE();

    if (m.isTrue()) {return result;}
    
    for (unsigned int i=0;i<apset->size();i++) {
      if (m.isSet(i)) {
        BDDFunction x=varmanager.getAPVariables().at(i);
        result&=(m.getValue(i) ? x : !x);
      }
    }
    
    return result;
  }
};

}}

#endif
