
#include "CheckLanguage.hpp"
#include "LTL2DSTAR.hpp"
#include "DRA.hpp"
#include "DRA_BDD.hpp"
#include "RabinGame.hpp"
#include "APTransformation.hpp"
#include "APSet.hpp"
#include <iostream>

using namespace vereofy;
using namespace vereofy::ltl_check;
using namespace ltl2dstar;



void CheckLanguage::init_dra() {
  LTLFormula_ptr ltl_neg = ltl->negate();
  
  LTL2DSTAR::DA_ptr dra = LTL2DSTAR::ltl2da(ltl_neg,
                                            "",
                                            std::vector<std::string>());
  
  dra_bdd = DRA_BDD::dra_to_bdd(varmanager,
                                dra->getDRA(), ltl->getAPSet());
  CHECK_SUCCESS(dra_bdd);

  // Check structure of  dra_bdd
  CHECK(1, dra_bdd->check_isTotal(), "");  
}



bool CheckLanguage::is_equivalent(BDD_G4G_NRA& g4g_nra) {
  // Check structure of g4g_nra
  CHECK(1, g4g_nra.check_isTotal(), "");
 
  std::cerr << "Checking equivalence via Rabin game and det. complement automaton..." << std::flush;
  
  StopWatch stopwatch_language_check;
  stopwatch_language_check.start();

  BDDFunction game_bit = varmanager.getGameVariable();

  BDDFunction V0 = !game_bit;
  BDDFunction V1 =  game_bit;

  BDDFunction csAP = container2cs(varmanager.getAPVariables());  
  BDDFunction csVertices = dra_bdd->getStateVars() & g4g_nra.get_state_vars(false) & csAP & game_bit;


  TRACE_BDD_FUNCTION("check_language", g4g_nra.getDelta(), "");
  TRACE_BDD_FUNCTION("check_language", dra_bdd->getDelta(), "");

  BDDFunction delta_v1_v0 = 
    // choose non-deterministiclly some AP...
    V1 & V0.moveDown() 
    & dra_bdd->no_change()
    & g4g_nra.no_change();


  BDDFunction delta_v0_v1 = 
    // can only move if AP is correct
    V0 & V1.moveDown()
    & g4g_nra.getDelta()
    & dra_bdd->getDelta();

  if (tight_game) {
    // move to defined AP state
    delta_v0_v1 &= csAP.moveDown();
  }

  TRACE_BDD_FUNCTION("check_language", delta_v1_v0, "");
  TRACE_BDD_FUNCTION("check_language", delta_v0_v1, "");
  
      
  BDDFunction E = delta_v1_v0 | delta_v0_v1;
  TRACE_BDD_FUNCTION("check_language", E, "");

  if (tight_game) {
    BDDFunction E_restriction = IMPLIES(V1.moveDown(), csAP.moveDown());
    E &= E_restriction;
  }
  
  BDDFunction initial_vertices = 
    V1 & g4g_nra.getInitialStates() & dra_bdd->getInitial();
  if (tight_game) {
    initial_vertices &= csAP;
  }
  TRACE_BDD_FUNCTION("check_language", initial_vertices, "");


  RabinGame game(initial_vertices, V0, V1, csVertices, E);
  for (unsigned int i=0;i<g4g_nra.countRabinPairs();i++) {
    BDD_RabinPair rg(V1 & g4g_nra.getRabinPair(i).R,
                     V1 & g4g_nra.getRabinPair(i).G);
    game.addRabinPair(rg);
  }

  for (unsigned int i=0;i<dra_bdd->countRabinPairs();i++) {
    BDD_RabinPair rg(V1 & dra_bdd->getRabinPair(i).R,
                     V1 & dra_bdd->getRabinPair(i).G);
    game.addRabinPair(rg);
  }
  //  game.addRabinPair(RabinGame::RabinPair(BDDFunction::ONE(), BDDFunction::ZERO()));

  class DotVertexWriter : public DotStateWriter {
  public:
    DotVertexWriter(BDDFunction V0_,
                    BDDFunction V1_,
                    DRA_BDD::ptr dra_bdd_,
                    BDD_G4G_NRA& g4g_nra_) : 
      V0(V0_),
      V1(V1_),
      dra_bdd(dra_bdd_),
      g4g_nra(g4g_nra_) {

      csG4G_NRA = g4g_nra.get_state_vars(false);
      csDA = dra_bdd->getStateVars();
      csAP = container2cs(g4g_nra.getVarManager().getAPVariables());
    }
    
    virtual void print(std::ostream& out, BDDFunction assignment) {
      out << g4g_nra.print_single_state(assignment,false);
      out << " | ";
      out << assignmentToBin(assignment, csDA);
      out << " | ";
      out << assignmentToBin(assignment, csAP);
    }
    
    BDDFunction V0;
    BDDFunction V1;
    DRA_BDD::ptr dra_bdd;    
    BDD_G4G_NRA& g4g_nra;

    BDDFunction csG4G_NRA;
    BDDFunction csAP;
    BDDFunction csDA;
  } dot_vertex_writer(V0, V1, dra_bdd, g4g_nra);


  if (print_stats) {
    game.print_stats(std::cerr);
  }



  TRACE("CheckLanguage", "game.dot", \
        std::ofstream game_dot("game.dot");     \
        game.toDot(game_dot, BDDFunction::ZERO(), &dot_vertex_writer); \
        game_dot.close();             );

  BDDFunction W = game.calculateW();
  TRACE_BDD_FUNCTION("check_language", W, "");


  TRACE("CheckLanguage", "game.dot", \
        std::ofstream game_dot("game.dot");     \
        game.toDot(game_dot, W, &dot_vertex_writer); \
        game_dot.close();             );



  // are all initial vertices winning?
  bool rv = bdd_set_is_contained(initial_vertices, W);


  if (rv && check_strategy) {
    RabinGame winning = game.restrictedToVertices(W);
    BDDFunction strategy = winning.find_strategy();
    if (strategy != BDDFunction::ZERO()) {
      std::cout << "Found strategy!\n";
      
      TRACE("CheckLanguage", "game.dot",                \
            std::ofstream strategy_dot("strat.dot");                 \
            winning.accordingToV0Strategy(strategy).toDot(strategy_dot,   \
                                                          W, &dot_vertex_writer));
    }
  }

  stopwatch_language_check.stop();
  std::cerr << "(" << stopwatch_language_check.getElapsedTimeString() << "s)\n";

  return rv;
}
