#ifndef DRA_BDD_HPP
#define DRA_BDD_HPP

#include "vereofy/setup.hpp"
#include "APSet.hpp"
#include "LTL2DSTAR.hpp"
#include "DRA.hpp"
#include "Reachability.hpp"


class DRA_BDD {
 public:
  typedef boost::shared_ptr<DRA_BDD> ptr;

  struct RabinPair {
    RabinPair(BDDFunction R_,
              BDDFunction G_) : R(R_), G(G_) {}

    BDDFunction R;    
    BDDFunction G;
  };

  DRA_BDD(ltl2dstar::APSet_cp ap_set_,
          BDDFunction dra_delta_,
          BDDFunction dra_initial_,
          BDDFunction csStates_,
          BDDFunction csAP_,
          StateVars_DA::ptr state_vars_da_) :
    ap_set(ap_set_),
    dra_delta(dra_delta_),
    dra_initial(dra_initial_),
    csStates(csStates_),
    csAP(csAP_),
    state_vars_da(state_vars_da_)
  {}

  ~DRA_BDD() {
    dra_delta = BDDFunction::ZERO();
    dra_initial = BDDFunction::ZERO();
    csStates = BDDFunction::ZERO();
    csAP = BDDFunction::ZERO();
    pairs.clear();
    state_vars_da.reset();
  }
  
  ltl2dstar::APSet_cp getAPSet() {return ap_set;}
  BDDFunction getDelta() const {return dra_delta;}
  BDDFunction getInitial() const {return dra_initial;}
  
  BDDFunction getStateVars() const {return csStates;}
  BDDFunction getAPVars() const {return csAP;}

  unsigned int countStateVars() const {return state_vars_da->getDAStateCount();}

  unsigned int countRabinPairs() {return pairs.size();}
  const RabinPair& getRabinPair(unsigned int i) {return pairs.at(i);}

  void addRabinPair(RabinPair rg) {pairs.push_back(rg);}

  BDDFunction no_change() {
    BDDFunctionContainer projStates(cubeSetToFunctionContainer(csStates));
    return pairwiseEQUIV(projStates, container_moveDown(projStates));
  }
  
  BDDFunction calcReachable() const {
    return Reachability::reachable(getInitial(),
                                   getDelta(),
                                   getStateVars(),
                                   getAPVars());
  }

  bool check_isTotal() const {
    BDDFunction reachable = calcReachable();
    BDDFunction total_dra = getDelta()
      .exists(getStateVars().moveDown())
      .forall(getAPVars());

    if (!bdd_set_is_contained(reachable, total_dra)) {
      TRACE_BDD_FUNCTION("check", reachable & !total_dra, "");
      return false;
    }
    return true;
  }

  void toDot(std::ostream& out) const {
    VarManagerG4G::ReorderLock reorder_lock; // inhibit reordering for the rest of this scope
    
    out << "digraph G {\n";
    BDDFunction reachable = calcReachable();
    TRACE_BDD_FUNCTION("toDot", reachable, "");
    
    BDDFunction reachableCopy = reachable;
    while (reachableCopy != BDDFunction::ZERO()) {
      BDDFunction from=popAssignment(reachableCopy, csStates);
      out << "  \"" << assignmentToBin(from, csStates) << "\" [";
      
      out << "shape=box";
      
      // add label
      out << ", label=\"";
      out << assignmentToBin(from, csStates);
      out << " | ";
      
      for (unsigned int i=0;i<pairs.size();i++) {
        if ((from & pairs.at(i).R) != BDDFunction::ZERO()) {
          out << "-" << i;
        } else if ((from & pairs.at(i).G) != BDDFunction::ZERO()) {
          out << "+" << i;
        }
      }
      out << "\"]\n";
      
      out << std::endl;
      
      BDDFunction delta_from = (dra_delta & from).exists(csStates);
      while (delta_from != BDDFunction::ZERO()) {
        BDDFunction toPair = popAssignment(delta_from, 
                                           csAP & 
                                           csStates.moveDown());
        
        out << " \"" << assignmentToBin(from, csStates) << "\" -> \""
          << assignmentToBin(toPair, csStates.moveDown())
            << "\" [label = \"" <<assignmentToBin(toPair, csAP) <<  "\"];\n";
      }
    }
    
    out << "}\n";
  }
  
  static DRA_BDD::ptr dra_to_bdd(VarManagerG4G& varmanager,
                                 ltl2dstar::Automaton::DRA_ptr dra,
                                 ltl2dstar::APSet_cp target_apset) {
    typedef ltl2dstar::Automaton::DRA_t DRA_t;
    
    ltl2dstar::APSet_cp ap_set = dra->getAPSet_cp();
    APTransformation transform(ap_set, target_apset);
    
    unsigned int state_count = dra->size();
    
    std::cout << "Building BDD representation for DRA with " << state_count << " states..." << std::flush;
    StopWatch stopwatch_bdd;
    stopwatch_bdd.start();
    
    StateVars_DA::ptr da_vars = varmanager.create_da_state_variables(state_count);
    
    BDDFunction da_delta = BDDFunction::ZERO();
    BDDFunction da_initial = BDDFunction::ZERO();
    
    for (std::size_t state=0; state<state_count; state++) {
      BDDFunction from_bdd = da_vars->da_state_2_bdd(state, false);
      
      DRA_t::state_type* cur_state=dra->get(state);      
      if (cur_state == dra->getStartState()) {
        da_initial |= da_vars->da_state_2_bdd(state, false);
      }
      
      // transitions
      for (ltl2dstar::APSet::element_iterator el_it=ap_set->all_elements_begin();
           el_it!=ap_set->all_elements_end();
           ++el_it) {
        ltl2dstar::APElement label=*el_it;
        DRA_t::state_type *to_state=cur_state->edges().get(label);
        std::size_t to_state_index=to_state->getName();
        
        BDDFunction to_bdd =
          da_vars->da_state_2_bdd(to_state_index, true);
        
        BDDFunction ap_bdd = varmanager.apelement_2_bdd(label,
                                                        transform);
        
        BDDFunction transition = from_bdd & to_bdd & ap_bdd;
        da_delta |= transition;
      }
    }
    
    DRA_BDD::ptr dra_bdd(new DRA_BDD(ap_set,
                                     da_delta,
                                     da_initial,
                                     container2cs(*da_vars),
                                     container2cs(varmanager.getAPVariables()),
                                     da_vars));
    
    
    for (unsigned int i = 0; i<dra->acceptance().size(); i++) {
      BDDFunction R = BDDFunction::ZERO();
      BDDFunction G = BDDFunction::ZERO();
      
      for (unsigned int s=0; s<dra->size(); s++) {
        switch (dra->acceptance().getRabinColorForState(i, s)) {
        case ltl2dstar::RabinAcceptance::RABIN_RED:
          R |= da_vars->da_state_2_bdd(s, false);
          break;
        case ltl2dstar::RabinAcceptance::RABIN_GREEN:
          G |= da_vars->da_state_2_bdd(s, false);
          break;
        case ltl2dstar::RabinAcceptance::RABIN_WHITE:
          // nop
          break;        
        }
      }
      
      DRA_BDD::RabinPair rg(R, G);
      dra_bdd->addRabinPair(rg);
    }

    stopwatch_bdd.stop();
    std::cout << " (" << stopwatch_bdd.getElapsedTimeString() << "s)\n";

    /*
    std::ofstream dra_dot("dra.dot");
    dra_bdd->toDot(dra_dot);
    dra_dot.close();
    */
    
    return dra_bdd;
  }

  
 private:
  ltl2dstar::APSet_cp ap_set;
  
  BDDFunction dra_delta;
  BDDFunction dra_initial;
  BDDFunction csStates;
  BDDFunction csAP;

  StateVars_DA::ptr state_vars_da;
  std::vector<RabinPair> pairs;
};



#endif
