#include "VarManagerG4G.hpp"
#include "RabinGame.hpp"
#include "Reachability.hpp"
#include "SCCFinder.hpp"
#include "vereofy/common/BitSetIterator.hpp"
#include "vereofy/common/errors.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/common/bdd_checks.hpp"
#include "vereofy/common/stopwatch.hpp"
#include "vereofy/common/return_value.hpp"

using namespace vereofy;

BDDFunction RabinGame::cpred(BDDFunction Q) {
  TRACE_BDD_FUNCTION("cpred", Q, "");
  BDDFunction edge_to_Q = (E & Q.moveDown()).exists(csVertices.moveDown());
  TRACE_BDD_FUNCTION("cpred", edge_to_Q, "");
  BDDFunction edge_to_notQ = (E & (!Q).moveDown()).exists(csVertices.moveDown());
  TRACE_BDD_FUNCTION("cpred", edge_to_notQ, "");

  BDDFunction v0_cpred = V_0 & edge_to_Q;
  TRACE_BDD_FUNCTION("cpred", v0_cpred, "");
  BDDFunction v1_cpred = V_1 & edge_to_Q & !edge_to_notQ;
  TRACE_BDD_FUNCTION("cpred", v1_cpred, "");

  BDDFunction result = v0_cpred | v1_cpred;
  TRACE_BDD_FUNCTION("cpred", result, "");
  return result;
}

BDDFunction RabinGame::post(BDDFunction Q) const {
  return ((E & Q).exists(csVertices)).moveUp();
}

BDDFunction RabinGame::pre(BDDFunction Q) const {
  return (E & Q.moveDown()).exists(csVertices.moveDown());
}

BDDFunction RabinGame::calculateW() {
  // LeastFixPoint(Z)
  BDDFunction Z = BDDFunction::ZERO();
  BDDFunction Zold = BDDFunction::ONE();

  BitSet rabin_pairs;
  if (pairs.countPairs()>0) {
    rabin_pairs.set(0, pairs.countPairs()-1, true);
  }

  while (Z!=Zold) {
    Zold = Z;

    TRACE_BDD_FUNCTION("calculateW", Z, "before");
    BDDFunction p1 = cpred(Z);
    TRACE_BDD_FUNCTION("calculateW", p1, "");
    Z = Rabin(rabin_pairs, BDDFunction::ONE(), p1);
    TRACE_BDD_FUNCTION("calculateW", Z, "after");
  }

  return Z;
}

BDDFunction RabinGame::Rabin(const BitSet &rabin_pairs,
                             BDDFunction seqnr,
                             BDDFunction right) {
  BDDFunction U = BDDFunction::ZERO();
  CHECK_SUCCESS(U == BDDFunction::ZERO());

  TRACE_BDD_FUNCTION("Rabin", seqnr, "");
  TRACE_BDD_FUNCTION("Rabin", right, "");

  for (BitSetIterator it = BitSetIterator(rabin_pairs);
       it != BitSetIterator::end(rabin_pairs);
       ++it) {
    unsigned int rabin_index = *it;
    const BDDFunction& r = pairs.getPair(rabin_index).R;
    const BDDFunction& g = pairs.getPair(rabin_index).G;

    TRACE_BDD_FUNCTION("Rabin", r, rabin_index);
    TRACE_BDD_FUNCTION("Rabin", g, rabin_index);


    BitSet n_rabin_pairs(rabin_pairs);
    n_rabin_pairs.clear(rabin_index);
    
    // GreatestFixPoint(Y)
    BDDFunction Y = BDDFunction::ONE();
    BDDFunction Yold = BDDFunction::ZERO();
    while (Y!=Yold) {
      Yold = Y;
      
      TRACE_BDD_FUNCTION("RabinGame", Y, "before");

      BDDFunction p2 = right | (seqnr & g & !r & cpred(Y));
      TRACE_BDD_FUNCTION("RabinGame", p2, "");      

      // LeastFixPoint(X)
      BDDFunction X = BDDFunction::ZERO();
      BDDFunction Xold = BDDFunction::ONE();
      while (X!=Xold) {
        Xold = X;

      TRACE_BDD_FUNCTION("RabinGame", X, "before");        

        BDDFunction p3 = p2 |
          (seqnr & !r & cpred(X));

      TRACE_BDD_FUNCTION("RabinGame", p3, "");      

        if (n_rabin_pairs.isEmpty()) {
          X = p3;
        } else {
          X = Rabin(n_rabin_pairs,
                    seqnr & !r,
                    p3);
        }
        TRACE_BDD_FUNCTION("RabinGame", X, "after");      
      }
      Y = X;
      TRACE_BDD_FUNCTION("RabinGame", Y, "after");
    }
    
    TRACE_BDD_FUNCTION("RabinGame", U, "before");      
    U |= Y;
    TRACE_BDD_FUNCTION("RabinGame", U, "after");      
  }

  return U;
}

void RabinGame::toDot(std::ostream& out, 
                      BDDFunction winning,
                      DotStateWriter* state_writer) const {
  VarManagerG4G::ReorderLock reorder_lock; // inhibit reordering for the rest of this scope
  

  out << "digraph G {\n";
  BDDFunction reachable = calculateReachable();
  TRACE_BDD_FUNCTION("toDot", reachable, "");
  
  BDDFunction reachableCopy = reachable;
  while (reachableCopy != BDDFunction::ZERO()) {
    BDDFunction from=popAssignment(reachableCopy, csVertices);
    out << "  \"" << assignmentToBin(from, csVertices) << "\" [";

    if ((from & V_0) != BDDFunction::ZERO()) {
      out << "shape=ellipse";
    } else {
      out << "shape=box";
    }

    if ((from & initial_vertices) != BDDFunction::ZERO()) {
      out << ", style=filled, fillcolor=lightgrey";
    }

    if ((winning & from) != BDDFunction::ZERO()) {
      out << ", color=green";
    }
    
    // add label
    out << ", label=\"";
    if (state_writer) {
      state_writer->print(out, from);
    } else {
      out << assignmentToBin(from, csVertices);
    }
    out << " [ ";

    for (unsigned int i=0;i<pairs.countPairs();i++) {
      TRACE_BDD_FUNCTION("RabinGame.toDot", from, "");
      TRACE_BDD_FUNCTION("RabinGame.toDot", pairs.getPair(i).R, i);
      TRACE_BDD_FUNCTION("RabinGame.toDot", from & pairs.getPair(i).R, i);
      TRACE_BDD_FUNCTION("RabinGame.toDot", pairs.getPair(i).G, i);
      TRACE_BDD_FUNCTION("RabinGame.toDot", from & pairs.getPair(i).G, i);
      if ((from & pairs.getPair(i).R) != BDDFunction::ZERO()) {
        out << "-" << i;
      } else if ((from & pairs.getPair(i).G) != BDDFunction::ZERO()) {
        out << "+" << i;
      }
    }
    out << "]\"]\n";
    
    out << std::endl;

    BDDFunction E_From = (E & from).exists(csVertices);
    while (E_From != BDDFunction::ZERO()) {
      BDDFunction to = popAssignment(E_From, 
                                     csVertices.moveDown());
      
      out << " \"" << assignmentToBin(from, csVertices) << "\" -> \""
          << assignmentToBin(to, csVertices.moveDown())
          << "\";\n";
    }
  }
  
  out << "}\n";
}

void RabinGame::print_stats(std::ostream& out) const {
  out << "Rabin game with " << pairs.countPairs() << " acceptance pairs.\n";
  out << "V0 BDD nodes = " << V_0.size() << "\n";
  out << "V1 BDD nodes = " << V_1.size() << "\n";
  out << "E BDD nodes = " << E.size() << "\n";
  BDDFunction reachable = calculateReachable();
  out << "reach. V0 vertices = " << (V_0 & reachable).countAssignments(csVertices) << "\n";
  out << "reach. V1 vertices = " << (V_1 & reachable).countAssignments(csVertices) << "\n\n";
}


BDDFunction RabinGame::calculateReachable(BDDFunction from) const {
  StopWatch stopwatch;
  stopwatch.start();
  TRACE_BDD_FUNCTION("RabinGame.calculateReachable", E, "");
  BDDFunction oldReachable=BDDFunction::ZERO();
  BDDFunction reachable=from;
  int steps = 0;
  while (reachable!=oldReachable) {
    TRACE_BDD_FUNCTION("RabinGame.calcReachable", reachable, "step " << steps);
    oldReachable=reachable;
    ++steps;
    reachable |= post(reachable);
  }
  stopwatch.stop();


  TRACE_BDD_FUNCTION("RabinGame.calcReachable", reachable, "finished");
  return reachable;
}


RabinGame RabinGame::accordingToV0Strategy(BDDFunction strategy) const {
  BDDFunction restricted_E = E & (IMPLIES(V_0, strategy));
  RabinGame result(initial_vertices, V_0, V_1, csVertices, restricted_E);

  // copy acceptance
  result.pairs = pairs;

  return result;
}

RabinGame RabinGame::restrictedToVertices(BDDFunction V) const {
  BDDFunction restricted_E = E & V & V.moveDown();
  RabinGame result(initial_vertices & V,
                   V_0 & V, 
                   V_1 & V, 
                   csVertices, 
                   restricted_E);
  // copy acceptance
  result.pairs = pairs;
  
  return result;
}

struct StrategyIterator {
  typedef std::vector<unsigned int> strategy_t;

  StrategyIterator(RabinGame game) {
    BDDFunction reachable = game.calculateReachable();
    BDDFunction csVertices = game.getCsVertices();

    while (reachable != BDDFunction::ZERO()) {
      BDDFunction v = popAssignment(reachable, csVertices);

      unsigned int index = vertices.size();
      vertices.push_back(v);

      successors.push_back(std::vector<BDDFunction>());

      BDDFunction all_successors = (game.getE() & v).exists(csVertices);
      while (all_successors != BDDFunction::ZERO()) {
        BDDFunction successor = popAssignment(all_successors, 
                                              csVertices.moveDown());
        successors.at(index).push_back(successor);
      }
    }
  }

  strategy_t get_first_strategy() const {
    strategy_t result;
    
    for (unsigned int i=0;i<vertices.size();i++) {
      CHECK_SUCCESS(successors.at(i).size() >= 1);
      result.push_back(0);
    }
    
    return result;
  }
  
  ReturnValue<strategy_t> get_next_strategy(strategy_t& cur) const {
    strategy_t next = cur;
    for (unsigned int i=0;i<vertices.size();i++) {
      if (cur.at(i)+1 == successors.at(i).size()) {
        // overflow
        next.at(i) = 0;
      } else {
        next.at(i) = cur.at(i) + 1;
        return next;
      }
    }
    
    return ReturnValue<strategy_t>(); // overflow
  }

  void print(const strategy_t& strategy, std::ostream& out) const {
    for (unsigned int i=0;i<strategy.size();i++) {
      out << i << " -> " << strategy.at(i) << " / " << successors.at(i).size() << "\n";
    }
  }

  BDDFunction bdd_strategy(const strategy_t& strategy) const {
    BDDFunction result = BDDFunction::ZERO();
    for (unsigned int i=0;i<strategy.size();i++) {
      result |= vertices.at(i) & successors.at(i).at(strategy.at(i));
    }
    return result;
  }
  
  std::vector<BDDFunction> vertices;
  std::vector< std::vector<BDDFunction> > successors;
};


BDDFunction RabinGame::find_strategy() const {
  StrategyIterator strategies(*this);
  StrategyIterator::strategy_t strategy = strategies.get_first_strategy();

  bool done = false;
  unsigned int counter = 0;
  while (!done) {
    ++counter;

    TRACE("find_strategy", "strategy", \
          std::cerr << "Checking strategy " << counter << ": \n";    \
          strategies.print(strategy, std::cerr));

    BDDFunction strategy_bdd = strategies.bdd_strategy(strategy);

    TRACE("find_strategy", "strategy.dot",                             \
          RabinGame game_with_strategy = accordingToV0Strategy(strategy_bdd); \
          std::ofstream strategy_dot((std::string("strat_")+            \
                                     boost::lexical_cast<std::string>(counter)+\
                                      ".dot").c_str());                 \
          game_with_strategy.toDot(strategy_dot));    

    bool rv = check_strategy(strategy_bdd);
    
    if (rv) {
      std::cerr << "Found strategy after " << counter << " tries.\n";
      return strategy_bdd;
    }

    ReturnValue<StrategyIterator::strategy_t> next = 
      strategies.get_next_strategy(strategy);
    if (next.valid) {
      strategy = next;
    } else {
      done = true;
    }
  }
  
  return BDDFunction::ZERO();
}

class StreettChecker {
public:
  StreettChecker(SCCFinder& scc_finder_,
                 BDD_StreettAcceptance &pairs_) 
    : scc_finder(scc_finder_),
      pairs(pairs_),
      non_empty(false) {}

  bool isNonEmpty() const {return non_empty;}

  bool operator()(BDDFunction C) {
    // if C is trivial...
    if (BDDChecks::isAssignment(C, scc_finder.getCsStateVars())) {
      // only single state
      if ((scc_finder.post(C) & C) == BDDFunction::ZERO()) {
        // no transition back to C
        // is trivial SCC
        return true;
      }
    }
    
    bool fair = true;
    for (unsigned int i=0;i<pairs.countPairs();i++) {
      if ((!bdd_sets_have_intersection(C,pairs.getPair(i).R)) ||
          bdd_sets_have_intersection(C, pairs.getPair(i).G)) {
        // good
        continue;
      } else {
        fair = false;
        break;
      }          
    }

    if (fair) {
      non_empty = true;
      return false;
    }

    BDDFunction C_prime = C;
    for (unsigned int i=0;i<pairs.countPairs();i++) {
      if (bdd_sets_have_intersection(C,pairs.getPair(i).R) &&
          (!bdd_sets_have_intersection(C,pairs.getPair(i).G))) {
          C_prime = C_prime & ! pairs.getPair(i).R;
        }
    }
    
    if (C_prime != BDDFunction::ZERO()) {
      return scc_finder.Lockstep(C_prime, *this);
    }
    return true;
  }

  SCCFinder scc_finder;
  BDD_StreettAcceptance &pairs;
  bool non_empty;
};


bool RabinGame::check_strategy(BDDFunction v0_strategy) const {
  RabinGame game_with_strategy = accordingToV0Strategy(v0_strategy);
  BDDFunction reachable = game_with_strategy.calculateReachable();

  TRACE("check_strategy", "strategy.dot",                               \
        std::ofstream strategy_dot("strat.dot");                        \
        game_with_strategy.toDot(strategy_dot));
  
  BDD_StreettAcceptance streett_condition;
  for (unsigned int i=0;i<pairs.countPairs();i++) {
    // switch R and G
    streett_condition.addPair(BDD_StreettPair(pairs.getPair(i).G,
                                              pairs.getPair(i).R));
  }
  
  
  SCCFinder scc_finder(game_with_strategy.getE() & reachable, csVertices);
  StreettChecker streett_check(scc_finder, streett_condition);

  scc_finder.Lockstep(reachable, streett_check);

  if (streett_check.isNonEmpty()) {
    return false;
  }

  return true;
}

bool RabinGame::check_game_structure() const {
  BDDFunction reachable = calculateReachable();
  BDDFunction terminal = reachable & !pre(reachable);
  if (terminal != BDDFunction::ZERO()) {
    TRACE_BDD_FUNCTION("check_game_structure", terminal, "");
    return false;
  } else {
    return true;
  }
}
