#include "VarManagerG4G.hpp"
#include "vereofy/common/errors.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/system/symb_automaton/symb_flat_datadomain.hpp"
#include <boost/lexical_cast.hpp>

using namespace vereofy;

bool VarManagerG4G::dynamic_reorder_enabled = 0;
unsigned int VarManagerG4G::reorder_lock = 0;

// create ap variables. If there are already some, check that size is
// is the same and don't create new ones
void VarManagerG4G::create_ap_variables(unsigned int size) {
  if (have_ap_variables) {
    CHECK_SUCCESS(size == ap_variables.size());
    return ;
  }

  for (unsigned int i=0;i<size;i++) {
    BDDFunction f = createVariablePair("ap_"+
                                       boost::lexical_cast<std::string>(i),
                                       at_end());
    ap_variables.push(f);
  }
  
  have_ap_variables = true;
}

BDDFunction VarManagerG4G::create_nba_state_variable_in(StateVars_NBA& nba_variables) {
  unsigned int k = nba_variables.size();

  if(debugLevel >= 4) {
    std::cout << " ------- " << k << "---------------\n";

    print_varorder(std::cout);
  }

  std::string name = "n_"+boost::lexical_cast<std::string>(k);
  if (nba_variables.nba_index != 0) {
    name = boost::lexical_cast<std::string>(nba_variables.nba_index)+"-"+name;
  }
  
  // create n_j variable pair at the end
  BDDFunction f = createVariablePair(name,
                                     at_end());

  nba_variables.push(f);

  if(debugLevel >= 4) {
    print_varorder(std::cout);
  }

  return f;
}

StateVars_G4G::ptr VarManagerG4G::createG4GVariablesFor(const StateVars_NBA& nba_variables) {
  ReorderLock lock;

  static unsigned int counter = 0;
  std::string prefix = counter == 0 ?
    "" :
    boost::lexical_cast<std::string>(counter)+"-";
  ++counter;

  unsigned int k = nba_variables.size();

  StateVars_G4G::ptr g4g_variables(new StateVars_G4G(*this));
  g4g_variables->A.resize(k);
  g4g_variables->B.resize(k);

  // i = state
  // j = set


  unsigned int pos;  // position for newly created vars
  for (unsigned int i=0;i<k;i++) {
    // interleave with the NBA variables
    pos = after_prime_var(nba_variables.at(i));

    for (unsigned int j=0;j<k;j++) {
      // insert s_i-in-A_j
      BDDFunction a_i_j = createVariablePair(prefix+"a_"+
                                             boost::lexical_cast<std::string>(i)+
                                             "_"+
                                             boost::lexical_cast<std::string>(j),
                                             pos
                                             );
      
      g4g_variables->A.at(j).push(a_i_j);
      pos = after_prime_var(a_i_j);
      
      // insert s_i-in-B_j
      BDDFunction b_i_j = createVariablePair(prefix+"b_"+
                                             boost::lexical_cast<std::string>(i)+
                                             "_"+
                                             boost::lexical_cast<std::string>(j),
                                             pos
                                             );
      
      g4g_variables->B.at(j).push(b_i_j);

      pos = after_prime_var(b_i_j);
    }
    
    if(debugLevel >= 4) {
      print_varorder(std::cout);
    }
  }

  return g4g_variables;
}  

BDDFunction VarManagerG4G::create_game_variable() {
  game_variable = createVariablePair("game", 0);
  return game_variable;
}

void VarManagerG4G::free_game_variable() {
  if (game_variable != BDDFunction::ZERO()) {
    CHECK_SUCCESS(free(game_variable));
  }
}


BDDFunctionContainer VarManagerG4G::create_action_variables(BDDFunctionContainer state_vars_prime,
                                             bool action_vars_first) {
  BDDFunctionContainer result;

  for (unsigned int i=0;i<state_vars_prime.size();i++) {
    BDDFunction var;
    if (action_vars_first) {
      if (i == 0) {
        var = createVariable("act0", at_start());
      } else {
        var = createVariable("act"+boost::lexical_cast<std::string>(i), after_var(result.at(i-1)));
      }
    } else {
      var = createVariable("act"+boost::lexical_cast<std::string>(i), after_var(state_vars_prime.at(i)));
    }

    result.push(var);
  }

  return result;
}



BDDFunction VarManagerG4G::createVariable(const std::string& name, 
                                          unsigned int position) {
  std::vector<std::string> names(1, name);
  return createVarsInGroup(names, position).at(0);
}

BDDFunction VarManagerG4G::createVariablePair(const std::string& name, 
                                              unsigned int position) {
  std::vector<std::string> names;
  names.push_back(name);
  names.push_back(name+"'");
  return createVarsInGroup(names, position).at(0);
}

BDDFunctionContainer VarManagerG4G::createVarsInGroup(const std::vector<std::string>& names,
                                                      unsigned int position) {
  BDDGroup_ptr varGroup(new BDDGroup);
  std::list<BDDFunction> projection_functions;
  
  bool at_end = (position == bdd_var_count());

  if (at_end) {
    for (std::vector<std::string>::const_iterator it=names.begin();
       it!=names.end();
       it++) {
    
      CHECK_SUCCESS(!bdd_var_exists(*it));
      
      BDDVarOrder::insertVariable(*it, position);
      projection_functions.push_back(BDDFunction::projection(position));
#ifdef JINC_NEW 
      varGroup->groupVariable(position);
#else  // JINC_OLD
      varGroup->addVariable(*it);
#endif
      
      ++position;
    }
  } else {
    // not at end
    for (std::vector<std::string>::const_reverse_iterator it=names.rbegin();
         it!=names.rend();
         it++) {
      
      CHECK_SUCCESS(!bdd_var_exists(*it));
      
      BDDVarOrder::insertVariable(*it, position);
      projection_functions.push_front(BDDFunction::projection(position));
#ifdef JINC_NEW 
      varGroup->groupVariable(position);
#else  // JINC_OLD
      varGroup->addVariable(*it);
#endif
    }
  }
    
  group_list.push_back(varGroup);
  
  BDDFunctionContainer result;
  for (std::list<BDDFunction>::iterator it=projection_functions.begin();
       it!=projection_functions.end();
       ++it) {
    result.push(*it);
  }
  return result;
}


unsigned int VarManagerG4G::after_var(BDDFunction projection) const {
  return projection.getFirstVariableIndex()+1;
}

unsigned int VarManagerG4G::after_prime_var(BDDFunction projection) const {
  return after_var(projection)+primed_count();
}


unsigned int VarManagerG4G::after_last_var(const BDDFunctionContainer& vars) const {
  CHECK_SUCCESS(vars.size()>0);
  unsigned int index = vars.size()-1;
  
  return after_var(vars.at(index));
}

unsigned int VarManagerG4G::after_last_prime_var(const BDDFunctionContainer& vars) const {
  CHECK_SUCCESS(vars.size()>0);
  unsigned int index = vars.size()-1;
  
  return after_var(vars.at(index))+primed_count();
}

unsigned int VarManagerG4G::at_start() const {
  return 0;
}

unsigned int VarManagerG4G::at_end() const {
  return bdd_var_count();
}


void VarManagerG4G::print_varorder(std::ostream& out) const {
  for (unsigned int i=0;i<BDDVarOrder::size();i++) {
#ifdef JINC_NEW
      std::string name=BDDVarOrder::getVariableName(i);
#else
      std::string name=BDDVarOrder::getVariable(i);
#endif
    
    out << i << " : " << name << "\n";
  }
  out << "\n";
}
  
void VarManagerG4G::print_vars(std::ostream& out,
                               const BDDFunctionContainer& vars,
                               bool compact) const {
  for (unsigned int i=0;i<vars.size();i++) {
    if (compact && i>0) {
      out << ", ";
    }
    out << bdd_var_name(vars.at(i));
    if (!compact) out << "\n";
  }
  out << "\n";
}  


StateVars_DA::ptr VarManagerG4G::create_da_state_variables(unsigned int state_count) {
  ReorderLock lock;

  static unsigned int counter = 0;
  std::string prefix = counter == 0 ?
    "" :
    boost::lexical_cast<std::string>(counter)+"-";
  ++counter;

  StateVars_DA::ptr da_variables(new StateVars_DA(*this));
  da_variables->da_state_count = state_count;
  
  DataDomain::ptr dd(new IntRangeDataDomain(0, state_count-1));
  da_variables->da_state_encoding = boost::dynamic_pointer_cast<Symb_FlatDataDomain>(Symb_DataDomain::getFor(dd));
  unsigned int bits = da_variables->da_state_encoding->getLogSize();
  
  for (unsigned int i=0;i<bits;i++) {
    BDDFunction var = createVariablePair(prefix+"d_"+boost::lexical_cast<std::string>(i),
					 at_end());
    da_variables->push(var);
  }

  return da_variables;
}



BDDFunction VarManagerG4G::apelement_2_bdd(ltl2dstar::APElement element) const {
  BDDFunction result = BDDFunction::ONE();
  for (unsigned int i=0;i<ap_variables.size();i++) {
    if (element.get(i)) {
      result &= ap_variables.at(i);
    } else {
      result &= !ap_variables.at(i);
    }
  } 
  return result;
}

BDDFunction VarManagerG4G::apelement_2_bdd(ltl2dstar::APElement element,
                                           APTransformation& transformation) const {
  BDDFunction result = BDDFunction::ONE();
  for (unsigned int i=0;i<ap_variables.size();i++) {
    if (transformation.hasSource(i)) {
      if (element.get(transformation.getSourceFor(i))) {
        result &= ap_variables.at(i);
      } else {
        result &= !ap_variables.at(i);
      }
    } else {
      // nothing to do, don't care
    }
  } 
  return result;
}




void VarManagerG4G::enableDynamicReordering() {
  dynamic_reorder_enabled=true;
  setReorderingState();
}

void VarManagerG4G::disableDynamicReordering() {
  dynamic_reorder_enabled=false;
  setReorderingState();
}

void VarManagerG4G::setDynamicReorderingMaxGrowth(float maxGrowth) {
  BDD::setDynamicReorderingMaxGrowth(maxGrowth);
}

void VarManagerG4G::setDynamicReorderingTimeout(unsigned int timeout_ms) {
  BDD::setDynamicReorderingTimeout(timeout_ms);
}
  
void VarManagerG4G::acquireReorderingLock() {
  ++reorder_lock;
  if (reorder_lock==1) {
    // outmost lock acquired, change in reordering state
    setReorderingState();
  }
}

void VarManagerG4G::releaseReorderingLock() {
  CHECK_SUCCESS(reorder_lock>0);
  if (--reorder_lock == 0) {
    // outmost lock released, change in reordering state
    setReorderingState();
  }
}

void VarManagerG4G::setReorderingState() {
  if (dynamic_reorder_enabled && reorder_lock==0) {
    // enable reordering
    BDD::enableDynamicReordering();
  } else {
    // disable reordering
    BDD::disableDynamicReordering();
  }
}

void VarManagerG4G::triggerReordering() {
  if (reorder_lock==0) {
    BDD::reorderingHook(true);  // force reordering
  }
}


VarManagerG4G::ReorderLock::ReorderLock() {
  VarManagerG4G::acquireReorderingLock();
}

VarManagerG4G::ReorderLock::~ReorderLock() {
  VarManagerG4G::releaseReorderingLock();
}


// proj will be set to BDDFunction::ZERO() to enable
// freeing of the underlying variable
// returns 
//   true if the variable was successfully freed,
//   false if there were still BDD nodes for this variable
bool VarManagerG4G::free(BDDFunction& proj) {
  ReorderLock lock;

  CHECK_SUCCESS(BDDChecks::isProjection(proj));

  unsigned long index = proj.getFirstVariableIndex();
  proj = BDDFunction::ZERO();

  BDDGroup_ptr group;
  for (BDDGroupList::iterator it=group_list.begin();
       it!=group_list.end();
       ++it) {
#ifdef JINC_NEW
    unsigned int first_group_var = (*it)->getFirst()->getLevel();
#else // JINC_OLD
    unsigned int first_group_var = (*it)->getIndexOfFirstElement();
#endif
    if (first_group_var == index) {
      group = *it;
      group_list.erase(it);
      break;
    }
  }

  if (!group) {
    return false;
  }

  unsigned int size = group->size();
#ifdef JINC_NEW
  group->clear();
#else
  group->ungroup();
#endif
  group.reset();

  // store names to have persistent mean of talking about
  // vars
  std::vector<std::string> names;
  for (unsigned long i=0;i<size;i++) {
    names.push_back(bdd_var_name(index + i));
  }
  
  bool rv = true;
  for (unsigned long i=0;i<names.size();i++) {
    rv &= BDDVarOrder::removeVariable(names[i]);
  }
  return rv;
}

