#include "VarManagerG4G.hpp"
#include "vereofy/setup.hpp"

#include "vereofy/common/tracing.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/modelcheck/specifications/ltl/LTLInfixParser.hpp"
#include "vereofy/modelcheck/specifications/ltl/LTLPrefixParser.hpp"
#include "vereofy/modelcheck/checkers/ltl_check/LTL2NBA.hpp"

#include "BDD_NBA_Unary.hpp"
#include "BDD_G4G_Transformation.hpp"
#include "BDD_G4G_Union.hpp"
#include "CheckLanguage.hpp"


#include <boost/program_options.hpp>

namespace po=boost::program_options;
using namespace vereofy::ltl_check;
using namespace vereofy;
using namespace std;

class LTL_To_G4G {
private:
  VarManagerG4G varmanager;
  std::ofstream* dot_stream;

  std::string param_ltl_formula;
  std::string param_ltl_file;
  std::string param_write_varorder;
  std::string g4g_dot_file;
  std::string to_lbtt;
  std::string to_lbtt_parity;
  std::string param_dump_bdd;
  bool param_iterative;
  bool param_ltl_is_prefix_format;
  bool param_lbtt_detailed;
  bool param_complete_state_pattern;
  bool param_g4g_stats;
  bool param_g4g_count_sccs;
  bool param_twice_state_pattern;
  bool param_duplicate_sets;
  bool param_check_language;
  bool param_game_tight;  
  bool param_game_check_strategy;
  bool param_exhaustive;
  bool param_actions_on_top;
  unsigned int param_max_set;
  bool param_allow_union;

  bool dynamic_reordering;
  float dynamic_reordering_max_growth;
  unsigned int dynamic_reordering_timeout;

public:
  
  LTL_To_G4G() : 
    dot_stream(new std::ofstream("nba.dot")) {}

  ~LTL_To_G4G() {
    delete dot_stream;
  }

  LTLFormula_ptr parse_formula(const std::string& formula,
                                     APSet_cp ap_set=APSet_cp()) {
    if (param_ltl_is_prefix_format) {
      return parse_formula_prefix(formula);
    } else {
      return parse_formula_infix(formula);
    }
  }

  LTLFormula_ptr parse_formula_infix(const std::string& formula,
                                     APSet_cp ap_set=APSet_cp()) {
    return LTLInfixParser::parse(formula, ap_set);
  }
  
  LTLFormula_ptr parse_formula_prefix(const std::string& formula,
                                     APSet_cp ap_set=APSet_cp()) {
    return LTLPrefixParser::parse(formula, ap_set);
  }
  
  
  int usage(const std::string& message="") {
    if (message!="") {
      std::cout << "Error: " << message << "\n";    
    }
    std::cout << "\n\n nba-to-g4g formula\n";
    return 2;
  }
  
  
  BDD_NBA_Unary::ptr ltl_to_nba(LTLFormula_ptr ltl_formula) {

    VarManagerG4G::ReorderLock lock;  // don't allow reordering during
                                      // NBA parsing
   
    unsigned int ap_size = ltl_formula->getAPSet()->size();
    
    // create AP variables
    varmanager.create_ap_variables(ap_size);
        

    LTL2NBA_SPIN<BDD_NBA_Unary> ltl2nba("./ltl2ba");
    BDD_NBA_Unary::ptr nba(new BDD_NBA_Unary(ltl_formula->getAPSet(), 
                                             varmanager,
                                             dot_stream));
    
    if (!ltl2nba.ltl2nba(*ltl_formula, nba.get())) {
      throw "Could not generate NBA!";
    }
    
    return nba;
  }

  
  bool apply_options(int argc, char** argv) {
    po::options_description opt_general("General options");
    opt_general.add_options()
      ("help", "produce a help message")
      ("ltl", 
       po::value<string>(&param_ltl_formula)->default_value(""),
       "LTL formula to generate")
      ("ltl-file",
       po::value<string>(&param_ltl_file)->default_value(""),
       "Specifies the file from which the formula should be read")
      ("prefix-format",
       po::value<bool>(&param_ltl_is_prefix_format)->default_value(true),
       "LTL formula is in prefix format (like input for ltl2dstar)")
      ("g4g-to-dot",
       po::value<std::string>(&g4g_dot_file)->default_value(""),
       "The dot file to which the G4G automata is written")
      ("show-g4g-stats",
       po::value<bool>(&param_g4g_stats)->default_value(true),
       "Write statistics about the G4G automaton to std::cerr")
      ("show-g4g-count-sccs",
       po::value<bool>(&param_g4g_count_sccs)->default_value(true),
       "Determine number of SCCs in the G4G automaton std::cerr")
      ("to-lbtt",
       po::value<std::string>(&to_lbtt)->default_value(""),
       "The file to which the NBA for LBTT should be written (- stands for writing to stdout)")
      ("to-lbtt-parity",
       po::value<std::string>(&to_lbtt_parity)->default_value(""),
       "The file to which the NBA obtained by the G4G NPA for LBTT should be written (- stands for writing to stdout)")
      ("lbtt-detailed",
       po::value<bool>(&param_lbtt_detailed)->default_value(false),
       "Print detailed state information in LBTT NBA")
      ("write-varorder",
       po::value<std::string>(&param_write_varorder)->default_value(""),
       "Write the varorder at the end to the file")
      ("check-language",
       po::value<bool>(&param_check_language)->default_value(false),
       "Check whether the generated G4G-NRA corresponds to L(phi)")
      ("iterative",
       po::value<bool>(&param_iterative)->default_value(false),
       "Iterative construction of G4G-NRA, abort when language is sufficient")
      ("dump-bdd",
       po::value<string>(&param_dump_bdd),
       "Dump the BDD to the specified file")
      ("actions-on-top",
       po::value<bool>(&param_actions_on_top)->default_value(false),
       "When dumping the BDD, place the (non-deterministic choice) actions at the top of the BDD")
      ;

    po::options_description opt_bdd("BDD options");
    opt_bdd.add_options()
      ("dynamic-reorder", 
       po::value<bool>(&dynamic_reordering)->default_value(0),
       "Dynamic reordering of the variables")
      ("dynamic-reorder-max-growth",
       po::value<float>(&dynamic_reordering_max_growth)->default_value(MAX_GROWTH_DEFAULT),
       "The maximum intermediate growth of the BDD size during dynamic reordering (1.0 = no growth, 2.0 = may double)")
      ("dynamic-reorder-timeout",
       po::value<unsigned int>(&dynamic_reordering_timeout)->default_value(0),
       "A timeout (in milliseconds) for unsuccessful sifting for each variable during reordering. 0 = no timeout");

    po::options_description opt_game("Game options");
    opt_game.add_options()
      ("game-tight", 
       po::value<bool>(&param_game_tight)->default_value(0),
       "Tight vertices / edge relation for game")
      ("game-check-strategy", 
       po::value<bool>(&param_game_check_strategy)->default_value(0),
       "Find and check a strategy in the game")
      ;
    
    po::options_description opt_debug("Debug options");
    opt_debug.add_options()
      ("debug",
       po::value<int>(&debugLevel),  // global variable
       "Debug Level (0=no info, 5=max info)")  
      ("trace,T", 
       po::value<vector<string> >(),
       "trace points for debugging")
      ("check-level",
       po::value<unsigned int>(&check_level),  // global variable
       "level of internal sanity checks (0=off)");

    po::options_description opt_optimisation("Optimisation options");
    opt_optimisation.add_options()
      ("complete-state-pattern",
       po::value<bool>(&param_complete_state_pattern)->default_value(true),
       "Specifies whether the complete state pattern is used or the disjunctness/subset check is dropped")
      ("twice-state-pattern",
       po::value<bool>(&param_twice_state_pattern)->default_value(true),
       "Specifies whether the state pattern should be applied unprimed and primed")
      ("max-set",
       po::value<unsigned int>(&param_max_set)->default_value(0),
       "Specifies how many sets are used in the state tuples of the G4G (0 means all)")
      ("duplicate-sets",
       po::value<bool>(&param_duplicate_sets)->default_value(true),
       "Specifies whether duplicates in the sets of a tuple are allowed")
      ("union",
       po::value<bool>(&param_allow_union)->default_value(false),
       "Allow union construction")
      ("exhaustive",
       po::value<bool>(&param_exhaustive)->default_value(false),
       "Exhaustive search for best variable ordering")
      ;
        

    po::options_description opt_visible("Options");
    opt_visible.add(opt_general).add(opt_bdd).add(opt_game).add(opt_debug).add(opt_optimisation);
    
    po::options_description opt_all("All Options");
    opt_all.add(opt_visible);


    po::positional_options_description p;
    p.add("ltl-file", 1);
    p.add("to-lbtt-parity", 2);
    
    po::variables_map vm;
    std::vector<std::string> cmdline_arguments;
    for (int i=1; i<argc; i++) {
      cmdline_arguments.push_back(argv[i]);
    }
    po::store(po::command_line_parser(cmdline_arguments).options(opt_all).positional(p).run(), vm);
    po::notify(vm);    
    
    if (vm.count("help")) {
      cout << opt_visible;
      return false;
    }
    
    if (vm.count("trace")) {
      std::vector<std::string> trace_points=vm["trace"].as< vector<string> >();
      
      tracingEnabled=true;
      for (unsigned int i=0;i<trace_points.size();i++) {
        Tracing::activateTracePoint(trace_points.at(i));
        cerr << "Activate trace point : " << trace_points.at(i) << endl;
      }
    }

    if (dynamic_reordering) {
      varmanager.enableDynamicReordering();
    } else {
      varmanager.disableDynamicReordering();
    }
    varmanager.setDynamicReorderingMaxGrowth(dynamic_reordering_max_growth);
    varmanager.setDynamicReorderingTimeout(dynamic_reordering_timeout);


    return true;
  }

  std::string read_ltl_formula(const std::string& file) {
    std::string ltl_formula;
    if (file=="-") {
      std::getline(std::cin, ltl_formula);
    } else {
      std::ifstream ltl_input(file.c_str());
      std::getline(ltl_input, ltl_formula);
    }

    return ltl_formula;
  }

  struct G4G_Result {
    G4G_Result() : is_sufficient(false), is_not_sufficient(false) {}
    
    BDD_G4G_NRA::ptr g4g_automaton;
    bool is_sufficient;
    bool is_not_sufficient;
  };


  G4G_Result ltl_to_g4g(LTLFormula_ptr ltl_formula,
                        std::string prefix="") {
    G4G_Result result;

    BDD_NBA_Unary::ptr nba=ltl_to_nba(ltl_formula);
    
    if(param_g4g_stats) {
      std::cerr << prefix << "----------------------------------------------------\n";
      std::cerr << prefix << "LTL-Formula: ";
      ltl_formula->printPrefix(std::cerr);
      std::cerr << std::endl;
    }
    
    TRACE("NBA", "encoded.dot", \
          std::ofstream dot_nba("nba-encoded.dot"); \
          nba->toDot(dot_nba);                      \
          dot_nba.close();                          \
          );

    StateVars_G4G::ptr g4g_vars = varmanager.createG4GVariablesFor(nba->getStateVariables());

    if(debugLevel >= 3) {
        std::cout << "\n\n =================================== \n\n";

        // debug:
        std::cout << "\nNBA: " << nba->getStateCount() << " states\n";

        // debug: APSet
        std::cout << "\nAPs:\n";
        for (unsigned int i=0; i< nba->getAPSet()->size(); i++) {
          std::cout << i << " : " << nba->getAPSet()->getAP(i) << "\n";
        }
        std::cout << "\n\n";

        // debug: variable ordering
        std::cout << "Variable ordering:\n\n";
        varmanager.print_varorder(std::cout);

        // debug: A_j, B_j
        for (unsigned int j=0; j<g4g_vars->getSubsetCount(); j++) {
          std::cout << "A_"<<j<< ": ";
          varmanager.print_vars(std::cout, g4g_vars->getA(j));
          std::cout << "B_"<<j<< ": ";
          varmanager.print_vars(std::cout, g4g_vars->getB(j));
        }
        std::cout << "\n";
    }

    
 
    // construction
    BDD_G4G_Transformation::ptr g4g(new BDD_G4G_Transformation(nba->getAPSet(), varmanager));
    g4g->set_twice_state_pattern(param_twice_state_pattern);
    g4g->set_complete_state_pattern(param_complete_state_pattern);
    g4g->set_duplicate_sets(param_duplicate_sets);
    
    unsigned int max_set;
    if (param_max_set == 0) {
      max_set = nba->getStateCount();
    } else if (param_max_set > nba->getStateCount()) {
      max_set = nba->getStateCount();
      std::cerr << prefix << "Note: --max-set=" << param_max_set << 
        " lowered to " << max_set << ", as NBA has only " << max_set << "\n";
    } else {
      max_set = param_max_set;
    }


    bool is_sufficient = false; 
    bool is_not_sufficient = false;

    if (param_iterative) {
      CheckLanguage check(varmanager, 
                          ltl_formula,
                          param_g4g_stats,
                          param_game_tight,
                          param_game_check_strategy);

      for (unsigned int i=1; i <= max_set; i++) {
        std::cerr << prefix << "Iteration: " << i << " / " << max_set << "\n";
        g4g->from_nba(nba, g4g_vars, i);
        
        if (i == max_set ||
            // the last one is sufficient
            check.is_equivalent(*g4g)) {
          is_sufficient = true;
          std::cerr << prefix << "Max-Set for G4G-NRA (iterative): " << i << "\n";
          break;
        }
      }
      CHECK_SUCCESS(is_sufficient);
    } else {
      g4g->from_nba(nba, g4g_vars, max_set);
      is_sufficient = (max_set == nba->getStateCount() );
    }
    
    if (dynamic_reordering) {
      varmanager.triggerReordering();
    }

    if (param_write_varorder!="") {
      std::ofstream varorder(param_write_varorder.c_str());
      varmanager.print_varorder(varorder);
      varorder.close();
    }

    if(param_g4g_stats) {
      g4g->print_stats(std::cerr, param_g4g_count_sccs, prefix);
    }

    if (param_check_language) {
      CheckLanguage check(varmanager, 
                          ltl_formula,
                          param_g4g_stats,
                          param_game_tight,
                          param_game_check_strategy);
      if (check.is_equivalent(*g4g)) {
        std::cerr << prefix << "Language of G4G-NRA with " << max_set << " subsets is sufficient.\n";
        is_sufficient = true;
      } else {
        std::cerr << prefix << "Language of G4G-NRA with " << max_set << " subsets is not sufficient.\n";
        is_not_sufficient = true;
        is_sufficient = false;
      }
    }

    result.g4g_automaton = g4g;
    result.is_sufficient = is_sufficient;
    result.is_not_sufficient = is_not_sufficient;

    return result;
  }

  G4G_Result union_g4g(LTLFormula_ptr ltl_formula,
                       G4G_Result& result_left, 
                       G4G_Result& result_right,
                       std::string prefix="") {
    G4G_Result result;

    result.g4g_automaton = 
      BDD_G4G_Union::ptr(new BDD_G4G_Union(result_left.g4g_automaton,
                                           result_right.g4g_automaton));

    result.is_sufficient = result_left.is_sufficient && result_right.is_sufficient;
    result.is_not_sufficient = result_left.is_not_sufficient || result_right.is_not_sufficient;


    if (dynamic_reordering) {
      varmanager.triggerReordering();
    }

    if(param_g4g_stats) {
      std::cerr << prefix << "----------------------------------------------------\n";
      std::cerr << prefix << "LTL-Formula: ";
      ltl_formula->printPrefix(std::cerr);
      std::cerr << std::endl;

      result.g4g_automaton->print_stats(std::cerr, param_g4g_count_sccs, prefix);
    }

    return result;
  }

  struct MinDeltaRecorder : public BDD::CallBack {
    MinDeltaRecorder(BDD_G4G_NRA::ptr g4g_nra_) :
      g4g_nra(g4g_nra_),
      have_value(false) {}

    // returns true if the varorder should by stored, i.e., is better
    virtual bool operator()() {
      unsigned int cur_size = g4g_nra->getDelta().size();
      
      bool better = false;

      if (!have_value) {
        min_size = cur_size;
        have_value = true;
        better = true;
      } else {
        if (min_size > cur_size) {
          min_size = cur_size;
          better = true;
        }
      }

      return better;
    }

    BDD_G4G_NRA::ptr g4g_nra;
    bool have_value;
    unsigned int min_size;
  };
  
  void exhaustive_reordering(BDD_G4G_NRA::ptr g4g_nra) {
    g4g_nra->slim();
    varmanager.free_game_variable();

    std::cerr << "Variable ordering:\n\n";
    varmanager.print_varorder(std::cerr);

    MinDeltaRecorder recorder(g4g_nra);
    BDD::groupExhaustive(recorder);

    std::cerr << "Minimal Delta: " << recorder.min_size << "\n";
  }
  

  void dump_g4g(BDD_G4G_NRA::ptr g4g_nra, bool action_vars_first=false) {
    VarManagerG4G::ReorderLock lock;

    varmanager.free_game_variable();
    g4g_nra->slim();

    BDDFunctionContainer action_variables = 
      varmanager.create_action_variables(g4g_nra->get_state_var_container(true),
                                         action_vars_first);

    StopWatch stopwatch_actions;
    stopwatch_actions.start();
    BDDFunction delta_with_actions = g4g_nra->getDelta();
    // action <-> q'
    delta_with_actions &= pairwiseEQUIV(action_variables, g4g_nra->get_state_var_container(true));

    stopwatch_actions.stop();
    std::cerr << "Dump, adding actions: " << stopwatch_actions.getElapsedTimeString() << "s\n";

    StopWatch stopwatch_dump;
    stopwatch_dump.start();

    std::ofstream dump_stream(param_dump_bdd.c_str());
    dump_stream << "BDD-G4G-NRA Pairs: " << g4g_nra->countRabinPairs() << " ";
    dump_stream << "Vars: " << bdd_var_count() << " ";      
    dump_stream << "AP: ";
    for (unsigned int i=0;i<g4g_nra->getAPSet()->size();i++) {
      if (i>0) dump_stream << " ";
      dump_stream << g4g_nra->getAPSet()->getAP(i);
    }
    dump_stream << "\n";
        
    BDDFunctionContainer to_dump;
    to_dump.push(delta_with_actions & g4g_nra->calc_reachable());
    to_dump.push(g4g_nra->getInitialStates());
    for (unsigned int i=0;i<g4g_nra->countRabinPairs();i++) {
      to_dump.push(g4g_nra->getRabinPair(i).R);
      to_dump.push(g4g_nra->getRabinPair(i).G);
    }
    
    BDDWriter::storeFunctions(dump_stream, to_dump);

    stopwatch_dump.stop();
    std::cerr << "Dumping BDD: " << stopwatch_dump.getElapsedTimeString() << "s\n";
  }



  int run(int argc, char **argv) {
    if (!apply_options(argc, argv)) {
      return 2;
    }


    StopWatch stopwatch_run;
    stopwatch_run.start();
    


    if (param_ltl_formula!="" && param_ltl_file!="") {
      std::cerr << "Multiple options for the LTL formula!\n";
      return 2;
    }



    if (param_ltl_file!="") {
      param_ltl_formula = read_ltl_formula(param_ltl_file);
    } 
    if (param_ltl_formula=="") {
      std::cerr << "No LTL formula. Abort...\n";
      return 1;
    }

    G4G_Result result;

    LTLFormula_ptr ltl_formula;
    try {
      ltl_formula = parse_formula(param_ltl_formula);
    } catch (...) {
      std::cerr << "Exception parsing formula: " << param_ltl_formula << "\n";
      return 3;
    }

    if(param_g4g_stats) {
      std::cerr << "============================================================\n";
      std::cerr << "LTL-Input: ";
      ltl_formula->printPrefix(std::cerr);
      std::cerr << std::endl;
    }


    LTLFormula_ptr ltl_pnf = ltl_formula->toPNF();
    if (param_allow_union && ltl_pnf->getRootNode()->getType() == LTLNode::T_OR) {
      // union construction 
      LTLFormula_ptr ltl_left =  ltl_pnf->getSubFormula(ltl_pnf->getRootNode()->getLeft());
      LTLFormula_ptr ltl_right = ltl_pnf->getSubFormula(ltl_pnf->getRootNode()->getRight());
      
      G4G_Result result_left  = ltl_to_g4g(ltl_left, "LEFT: ");
      G4G_Result result_right = ltl_to_g4g(ltl_right, "RIGHT: ");

      result = union_g4g(ltl_formula,
                         result_left, result_right);

    } else {
      // the normal case
      
      result = ltl_to_g4g(ltl_formula);
    }





    if(g4g_dot_file!="") {
        std::ofstream dot_g4g(g4g_dot_file.c_str());
        result.g4g_automaton->toDot(dot_g4g);
        dot_g4g.close();
    }

    if(to_lbtt!="") {
      if(to_lbtt=="-") {
        result.g4g_automaton->to_lbtt(std::cout, param_lbtt_detailed);
      } else {
        std::ofstream lbtt_stream(to_lbtt.c_str());
        result.g4g_automaton->to_lbtt(lbtt_stream, param_lbtt_detailed);
        lbtt_stream.close();
      }
    }

    if(to_lbtt_parity!="") {
      if(to_lbtt_parity=="-") {
        result.g4g_automaton->to_lbtt(std::cout, param_lbtt_detailed,true);
      } else {
        std::ofstream lbtt_stream(to_lbtt.c_str());
        result.g4g_automaton->to_lbtt(lbtt_stream, param_lbtt_detailed,true);
        lbtt_stream.close();
      }
    }

    if (param_exhaustive) {
      exhaustive_reordering(result.g4g_automaton);
    }

    stopwatch_run.stop();
    std::cerr << "Total runtime: " << stopwatch_run.getElapsedTimeString() << "s\n";

    if (param_dump_bdd != "") {
      dump_g4g(result.g4g_automaton, param_actions_on_top);
    }

    if (result.is_sufficient) {
      return 0;
    } else if (result.is_not_sufficient) {
      return 1;
    } else {
      return 2;
    }
  }

};



int main(int argc, char **argv) {
  try {
    LTL_To_G4G g4g;
    
    return g4g.run(argc, argv);
  } catch (std::string& s) {
    std::cerr << "Exception: " << s << "\n";
    return 3;
  } catch (const char* s) {
    std::cerr << "Exception: " << s << "\n";
    return 3;
  }
}


