
#include "VarManagerG4G.hpp"
#include "vereofy/setup.hpp"

#include "vereofy/common/tracing.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/common/stopwatch.hpp"
#include "vereofy/modelcheck/specifications/ltl/LTLInfixParser.hpp"
#include "vereofy/modelcheck/specifications/ltl/LTLPrefixParser.hpp"
#include "vereofy/modelcheck/checkers/ltl_check/LTL2NBA.hpp"

#include "BDD_NBA_Unary.hpp"
#include "DRA_BDD.hpp"
#include "LTL2DSTAR.hpp"

#include <boost/program_options.hpp>

namespace po=boost::program_options;
using namespace vereofy::ltl_check;
using namespace vereofy;
using namespace std;

class LTL2DSTAR_to_BDD {
private:
  VarManagerG4G varmanager;
  std::ofstream* dot_stream;

  DRA_BDD::ptr bdd_dra;

  APSet_cp ap_set;
  
  std::string da_type;

  std::string param_ltl_formula;
  std::string param_dump_bdd;
  bool param_ltl_is_prefix_format;

  std::vector<std::string> ltl2dstar_options;

  bool dynamic_reordering;
  float dynamic_reordering_max_growth;
  unsigned int dynamic_reordering_timeout;

public:
  
  LTL2DSTAR_to_BDD() : 
    dot_stream(new std::ofstream("nba.dot")) {}

  ~LTL2DSTAR_to_BDD() {
    delete dot_stream;
  }

  LTLFormula_ptr parse_formula_infix(const std::string& formula) {
    return LTLInfixParser::parse(formula);
  }
  
  LTLFormula_ptr parse_formula_prefix(const std::string& formula) {
    return LTLPrefixParser::parse(formula);
  }
  
  
  
  int usage(const std::string& message="") {
    if (message!="") {
      std::cout << "Error: " << message << "\n";    
    }
    std::cout << "\n\n ltl2dstar formula\n";
    return 2;
  }
  
  
  LTL2DSTAR::DA_ptr ltl_to_da(const std::string& formula) {
    LTLFormula_ptr ltl_formula;
    if (param_ltl_is_prefix_format) {
      ltl_formula = parse_formula_prefix(formula);
    } else {
      ltl_formula = parse_formula_infix(formula);
    }
    
    unsigned int ap_size = ltl_formula->getAPSet()->size();
    
    ap_set = ltl_formula->getAPSet();

    // create AP variables
    varmanager.create_ap_variables(ap_size);
        

    LTL2DSTAR::DA_ptr da = LTL2DSTAR::ltl2da(ltl_formula,
                                             "",
                                             ltl2dstar_options);
                                             
    return da;
  }

  
  
  
  bool apply_options(int argc, char** argv) {
    po::options_description opt_general("General options");
    opt_general.add_options()
      ("help", "produce a help message")
      ("ltl", 
       po::value<string>(&param_ltl_formula),
       "LTL formula to generate")
      ("prefix-format",
       po::value<bool>(&param_ltl_is_prefix_format)->default_value(true),
       "LTL formula is in prefix format (like input for ltl2dstar)")
      ("dump-bdd",
       po::value<string>(&param_dump_bdd),
       "Dump the BDD to the specified file")
      ("option,O",
       po::value<vector<string> >(),
       "Options for ltl2dstar")
      ;
    
    po::options_description opt_bdd("BDD options");
    opt_bdd.add_options()
      ("dynamic-reorder", 
       po::value<bool>(&dynamic_reordering)->default_value(0),
       "Dynamic reordering of the variables")
      ("dynamic-reorder-max-growth",
       po::value<float>(&dynamic_reordering_max_growth)->default_value(MAX_GROWTH_DEFAULT),
       "The maximum intermediate growth of the BDD size during dynamic reordering (1.0 = no growth, 2.0 = may double)")
      ("dynamic-reorder-timeout",
       po::value<unsigned int>(&dynamic_reordering_timeout)->default_value(0),
       "A timeout (in milliseconds) for unsuccessful sifting for each variable during reordering. 0 = no timeout");

    po::options_description opt_debug("Debug options");
    opt_debug.add_options()
      ("debug",
       po::value<int>(&debugLevel),  // global variable
       "Debug Level (0=no info, 5=max info)")  
      ("trace,T", 
       po::value<vector<string> >(),
       "trace points for debugging")
      ("check-level",
       po::value<unsigned int>(&check_level),  // global variable
       "level of internal sanity checks (0=off)");
    
    po::options_description opt_visible("Options");
    opt_visible.add(opt_general).add(opt_bdd).add(opt_debug);
    
    po::options_description opt_all("All Options");
    opt_all.add(opt_visible);
    
    po::variables_map vm;
    std::vector<std::string> cmdline_arguments;
    for (int i=1; i<argc; i++) {
      cmdline_arguments.push_back(argv[i]);
    }
    po::store(po::command_line_parser(cmdline_arguments).options(opt_all).run(), vm);
    po::notify(vm);    
    
    if (vm.count("help")) {
      cout << opt_visible;
      return false;
    }
    
    if (vm.count("trace")) {
      std::vector<std::string> trace_points=vm["trace"].as< vector<string> >();
      
      tracingEnabled=true;
      for (unsigned int i=0;i<trace_points.size();i++) {
        Tracing::activateTracePoint(trace_points.at(i));
        cerr << "Activate trace point : " << trace_points.at(i) << endl;
      }
    }

    if (dynamic_reordering) {
      varmanager.enableDynamicReordering();
    } else {
      varmanager.disableDynamicReordering();
    }
    varmanager.setDynamicReorderingMaxGrowth(dynamic_reordering_max_growth);
    varmanager.setDynamicReorderingTimeout(dynamic_reordering_timeout);

    if (vm.count("option")) {
      ltl2dstar_options=vm["option"].as< vector<string> >();
    }


    return true;
  }



  void dra_to_bdd(ltl2dstar::Automaton::DRA_ptr dra) {
    StopWatch stopwatch_bdd;
    stopwatch_bdd.start();
    
    bdd_dra = DRA_BDD::dra_to_bdd(varmanager,
                                  dra,
                                  dra->getAPSet_cp());
    
    stopwatch_bdd.stop();
  }

  void dump_dra() {
    VarManagerG4G::ReorderLock lock;

    std::ofstream dump_stream(param_dump_bdd.c_str());
    dump_stream << "BDD-DRA Pairs: " << bdd_dra->countRabinPairs() << " ";
    dump_stream << "Vars: " << bdd_var_count() << " ";      
    dump_stream << "AP: ";
    for (unsigned int i=0;i<bdd_dra->getAPSet()->size();i++) {
      if (i>0) dump_stream << " ";
      dump_stream << bdd_dra->getAPSet()->getAP(i);
    }
    dump_stream << "\n";
        
    BDDFunctionContainer to_dump;
    to_dump.push(bdd_dra->getDelta() & bdd_dra->calcReachable());
    to_dump.push(bdd_dra->getInitial());
    for (unsigned int i=0;i<bdd_dra->countRabinPairs();i++) {
      to_dump.push(bdd_dra->getRabinPair(i).R);
      to_dump.push(bdd_dra->getRabinPair(i).G);
    }
    
    BDDWriter::storeFunctions(dump_stream, to_dump);
  }

  int run(int argc, char **argv) {
    if (!apply_options(argc, argv)) {
      return 2;
    }

    // We expect write failures to occur but we want ignore them and
    // don't want SIGPIPE to be raised.
    signal(SIGPIPE, SIG_IGN);

    std::cout << "----------------------------------------------------\n";
    std::cout << "LTL-Input: " << param_ltl_formula << std::endl;

    StopWatch stopwatch_total;
    stopwatch_total.start();

    LTL2DSTAR::DA_ptr da = ltl_to_da(param_ltl_formula);

    switch (da->getAutomatonType()) {
    case Automaton::DETERMINISTIC_RABIN_AUTOMATON:
      dra_to_bdd(da->getDRA());
      da_type="Rabin";
      break;
    case Automaton::DETERMINISTIC_STREETT_AUTOMATON:
      //      da_to_bdd(da->getDSA());
      da_type="Streett";
      break;
    case Automaton::DETERMINISTIC_PARITY_AUTOMATON:
      //      da_to_bdd(da->getDPA());
      da_type="parity";
      break;
    default:
      THROW_UNIMPLEMENTED();
    }

    if (dynamic_reordering) {
      print_stats("PRE-REORDER ");
      
      // do reorder at the end
      std::cout << "\n\nDo one (last) iteration of dynamic reordering..." << std::flush;
      varmanager.triggerReordering();
      std::cout << "\nAfter reordering:\n";
    }

    print_stats();
    
    stopwatch_total.stop();
    std::cout << "Total-time: " << stopwatch_total.getElapsedTimeString() << "s\n";

    if (param_dump_bdd != "") {
      dump_dra();
    }

    return 0;
  }

  void print_stats(std::string prefix ="") {
    std::cout << prefix << "Deterministic " << da_type << " automaton:\n";
    std::cout << prefix << "States    : " << bdd_dra->countStateVars() << "\n";
    std::cout << "\n";
    std::cout << prefix << "BDD encoding:\n";
    std::cout << prefix << "Variables : " << bdd_var_count() << "\n";
    std::cout << prefix << "Delta     : " << bdd_dra->getDelta().size() << " nodes\n";
    std::cout << prefix << "Initial   : " << bdd_dra->getInitial().size() << " nodes\n"; 
  }    

};



  
int main(int argc, char **argv) {
  LTL2DSTAR_to_BDD ltl2da2bdd;

  try {
    return ltl2da2bdd.run(argc, argv);
  } catch (std::string& s) {
    std::cerr << "Exception!\n" << s << "\n";
    return -2;
  } catch (const char * s) {
    std::cerr << "Exception!\n" << s << "\n";
    return -2;
  }
}


