/***************************************************************************
 *   Copyright (C) 2008-2009 by Tobias Blechmann                           *
 *   blechman@tcs.inf.tu-dresden.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 @author Tobias Blechmann
 */

#include "bisimulation.hpp"
#include <iostream>
#include <sstream>
#include "vereofy/common/stopwatch.hpp"

using namespace std;

namespace vereofy {
  
  Bisimulation::Bisimulation(SymbComponentInstance::ptr _symbCompA, 
                             SymbComponentInstance::ptr _symbCompB) {
    symbCompA = _symbCompA;
    symbCompB = _symbCompB;

    TRACE_SYMB_COMPONENT_DOT("Bisimulation.init", "bisimulation_dot", *symbCompA, false);
    TRACE_SYMB_COMPONENT_PATHS("Bisimulation.init", "bisimulation_dot_path", *symbCompA);
    TRACE_SYMB_COMPONENT_DOT("Bisimulation.init", "bisimulation_dot", *symbCompB, false);
    TRACE_SYMB_COMPONENT_PATHS("Bisimulation.init", "bisimulation_dot_path", *symbCompB);

    StopWatch stopwatch_preparation;
    stopwatch_preparation.start();
    std::cout << "Bisimulation: Unite automata\n";

    if (debugLevel>=2) {
      cout << "CA(1) initialStates() = " << symbCompA->getInitialStates().size() << " BDD Nodes\n";
      cout << "CA(1) delta() = " << symbCompA->getDelta().size() << " BDD Nodes\n";
      cout << "CA(2) initialStates() = " << symbCompB->getInitialStates().size() << " BDD Nodes\n";
      cout << "CA(2) delta() = " << symbCompB->getDelta().size() << " BDD Nodes\n";
    }

    automaton = boost::dynamic_pointer_cast<SymbAutomaton>(symbCompA->copyUnite(*symbCompB));
    
    TRACE_SYMB_AUTOMATON_DOT("Bisimulation.init", "bisimulation_dot", *automaton);
    TRACE_SYMB_AUTOMATON_PATHS("Bisimulation.init", "bisimulation_dot_path", *automaton);
    TRACE_BDD_FUNCTION("Bisimulation.init", automaton->getInitialStates(), "");
    
    std::cout << "Bisimulation: Combine ports\n";
    
    symb_ce_patch(*symbCompA->info->c_interface, *symbCompB->info->c_interface);

    if (debugLevel>=2) {
      cout << "CA(combined) initialStates() = " << automaton->getInitialStates().size() << " BDD Nodes\n";
      cout << "CA(combined) delta() = " << automaton->getDelta().size() << " BDD Nodes\n";
    }

    stopwatch_preparation.stop();
    std::cout << "Bisimulation preparation: " << stopwatch_preparation.toString() << endl;
  }


  Bisimulation::~Bisimulation() {
  }

  
  void Bisimulation::symb_ce_patch(const ComponentInterface& interface_a, const ComponentInterface& interface_b) {
    if (interface_a.getInPortCount()!=interface_b.getInPortCount()) {
      throw "Bisimulation: Mismatch in input port number, A has "+boost::lexical_cast<string>(interface_a.getInPortCount()) + ", B has "+boost::lexical_cast<string>(interface_b.getInPortCount());
    }

    if (interface_a.getOutPortCount()!=interface_b.getOutPortCount()) {
      throw "Bisimulation: Mismatch in output port number, A has "+boost::lexical_cast<string>(interface_a.getOutPortCount()) + ", B has "+boost::lexical_cast<string>(interface_b.getOutPortCount());
    }

    /* Don't check for internal ports, they may still be in interface but have been hidden in automaton
    if (interface_a.getInternalPortCount()!=0) {
      throw "Bisimulation: A has internal ports!";
    }

    if (interface_b.getInternalPortCount()!=0) {
      throw "Bisimulation: B has internal ports!";
      }*/

    BDDFunction csPortVars = automaton->getPortVars();
    BDDFunction csDCVars = automaton->getDataConstraintVars();

    BDDFunction csExpectedPortVars = BDDFunction::ONE();
    BDDFunction csExpectedDCVars = BDDFunction::ONE();

    for (unsigned int i=0; i<interface_a.getInPortCount(); i++) {
      Symb_Channel_End::ptr symb_ce_a=symbCompA->get_symb_ce_for_port(interface_a.getInPort(i));
      Symb_Channel_End::ptr symb_ce_b=symbCompB->get_symb_ce_for_port(interface_b.getInPort(i));
      
      csExpectedPortVars&=symb_ce_a->getPortActive() & symb_ce_b->getPortActive();
      csExpectedDCVars&=container2cs(symb_ce_a->getDCVars()) & container2cs(symb_ce_b->getDCVars());

      automaton->rename_ce(symb_ce_a, symb_ce_b);

      if (debugLevel>=2) {
        cout << "Combined input port " << (i+1) << "/" << interface_a.getInPortCount() << ", delta() = " << automaton->getDelta().size() << " BDD Nodes\n";
      }
    }
    
    for (unsigned int i=0;i<interface_a.getOutPortCount();i++) {
      Symb_Channel_End::ptr symb_ce_a=symbCompA->get_symb_ce_for_port(interface_a.getOutPort(i));
      Symb_Channel_End::ptr symb_ce_b=symbCompB->get_symb_ce_for_port(interface_b.getOutPort(i));
      
      csExpectedPortVars&=symb_ce_a->getPortActive() & symb_ce_b->getPortActive();
      csExpectedDCVars&=container2cs(symb_ce_a->getDCVars()) & container2cs(symb_ce_b->getDCVars());
      
      automaton->rename_ce(symb_ce_a, symb_ce_b);

      if (debugLevel>=2) {
        cout << "Combined output port " << (i+1) << "/" << interface_a.getOutPortCount() << ", delta() = " << automaton->getDelta().size() << " BDD Nodes\n";
      }
    }

    CHECK_SUCCESS(csExpectedPortVars==csPortVars);
    CHECK_SUCCESS(csExpectedDCVars==csDCVars);
  }
  
  
  
  void Bisimulation::calculateSplitter(BDDFunction states, BDDFunction splitterStates) {
    BDDFunction ports = (delta&states).exists(csNotPorts) & csNotPorts;
    while (ports != 0.0) {
      BDDFunction currentPort = ports.getAssignment().cofactor(csNotPorts);
      BDDFunction splitterElement=currentPort & splitterStates;
      splitterContainer.push_back(splitterElement);
      ++nrOfSplitter;
      ++totalNrOfSplitters;
      ports &= !(currentPort&csNotPorts);
    }
  }
  
  
  void Bisimulation::insertStableBlock(BDDFunction state) {
    BDDFunction oldBlock = blockContainer[firstNonStableBlock];
    
    blockContainer[firstNonStableBlock]=state;
    blockContainer.push(oldBlock);
    ++nrOfPartitions;
    ++firstNonStableBlock;
  }

  
  void Bisimulation::markBlockAsStable(int blockNumber){
    BDDFunction oldBlock = blockContainer[firstNonStableBlock];
    
    blockContainer[firstNonStableBlock]=blockContainer[blockNumber];
    blockContainer[blockNumber]=oldBlock;
    
    ++firstNonStableBlock;
  }
  
  
  ModelCheckingResult::ptr Bisimulation::compute() {
    ModelCheckingResult::ptr result(new ModelCheckingResult(SymbComponentInstance::ptr()));
    std::stringstream result_stream;

    if (!cubeSetIsEmpty(automaton->getTransitionLabelVars())) {
      THROW_UNIMPLEMENTED();
    }

    cout << "###########Bisimulation started############" << endl;
    StopWatch stopwatch_bisimulation;
    stopwatch_bisimulation.start();

    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    totalNrOfSplitters=0;
    allVars=BDDFunction::ONE();
    for(unsigned long i=BDDVarOrder::size();i>0;--i){
      allVars&=BDDFunction::projection(i-1);
    }

    setOfReachable = automaton->calcReachableStates();
    delta = automaton->getDelta() & setOfReachable;
    csPorts = automaton->getPortVars();    
    csConstraints = automaton->getDataConstraintVars();
    csOriginalX = automaton->getOriginalStateVars();
    csCopyX = automaton->getCopyStateVars();
    csCopyCopyX = automaton->getCopyCopyStateVars();
    csNotPorts = allVars.cofactor(csPorts);
    csNotOriginalX= allVars.cofactor(csOriginalX);
    
    csSplitterVars = csPorts & csOriginalX;
    
    //initial splitters
    nrOfSplitter = 0;
    calculateSplitter(setOfReachable,setOfReachable);
    
    //inital partitions
    nrOfPartitions = 0;
    firstNonStableBlock = 0;
    blockContainer.push(setOfReachable);
    ++nrOfPartitions;
    
    int checkBlocks=1000;
    
    while((!splitterContainer.empty()) && (firstNonStableBlock!=nrOfPartitions)){
      //     cout << nrOfSplitter << " " << firstNonStableBlock << " " << nrOfPartitions << endl;
      
      //     int splitterNr = rand()%nrOfSplitter;
      //     BDDFunction temp = splitterContainer[splitterNr];
      //     splitterContainer[splitterNr] = splitterContainer[nrOfSplitter-1];
      //     splitterContainer[nrOfSplitter-1] = temp;
      
      --nrOfSplitter;
      //     BDDFunction currentSplitter = splitterContainer[nrOfSplitter];
      BDDFunction currentSplitter = splitterContainer.front();
      
      //     splitterContainer.pop();
      splitterContainer.pop_front();
      
      //loop over all partitions
      int oldNrOfPartitions=nrOfPartitions;
      for(int i=firstNonStableBlock; i < oldNrOfPartitions; ++i) {
        //get current partition
        BDDFunction currentPartition=blockContainer[i];
        if(checkBlocks==0){
          if((currentPartition&csNotOriginalX)==(currentPartition&csNotOriginalX).getAssignment()){
            markBlockAsStable(i);
            continue;
          }
        }
        //get equivalence class candidates for current partition
        BDDFunction currentPartitionCandidates=(delta & currentPartition & currentSplitter.moveDown(csOriginalX)).exists(csCopyX);
        //old       BDDFunction currentPartitionCandidates=(delta & blockContainer[i] & currentSplitter.moveDown(csOriginalX)).exists(csCopyX);
        //remove states which cannot be refined by this splitter (dc(q,N,P)=0 with q in currentPartition)
        BDDFunction clearedCurrentPartition=currentPartitionCandidates.exists(csNotOriginalX);
        //old       BDDFunction clearedCurrentPartition=(currentPartition&currentPartitionCandidates).exists(csNotOriginalX);
        
        //if not all states are removed they build their own equivalence class A
        bool notAllStatesRemoved=(clearedCurrentPartition!=currentPartition);
        //if the cleared current partition is not empty and there are states in equivalence class A we have to create new splitters for A
        if((clearedCurrentPartition!=BDDFunction::ZERO())&&(notAllStatesRemoved)){
          calculateSplitter((currentPartition&!clearedCurrentPartition).moveDown(), currentPartition&!clearedCurrentPartition);
        }
        
        //we have to calculate new splitter elements if A and clearedCurrentPartition are not empty
        bool calcSplitter=notAllStatesRemoved;
        while(clearedCurrentPartition!=BDDFunction::ZERO()){
          //get one state out of cleared current partition
          BDDFunction currentState=(clearedCurrentPartition&csNotOriginalX).getAssignment().cofactor(csNotOriginalX);
          
          //check if current partition has only one element
          if(currentState==clearedCurrentPartition){
            if(notAllStatesRemoved){
              blockContainer[i]&=!currentState;
              calculateSplitter(currentState.moveDown(),currentState);
              insertStableBlock(currentState);
            }
            break;
          }
          
          //calc pattern
          BDDFunction pattern=currentPartitionCandidates.cofactor(currentState);
          //get all states out of current equivalence class and calculate states that are equivalent in respect to the currentState and the pattern
          BDDFunction equivalentStates = currentPartitionCandidates.equivalenceClasses(pattern,csPorts&csConstraints);
          
          //remove equivalent states from equivalencePartitionCandidates
          currentPartitionCandidates&=!(equivalentStates&pattern);
          
          //if A is not empty or there are states in clearedCurrentPartition and not in equivalentStates we have to create a new partition
          if((clearedCurrentPartition!=equivalentStates) || notAllStatesRemoved){
            blockContainer[i]&=!equivalentStates;
            blockContainer.push(equivalentStates);
            ++nrOfPartitions;
            cout << nrOfPartitions << "  " << flush;
          }
          
          //if equivalentStates!=clearedCurrentPartition we have to create new splitters (because there are some states remaining)
          if(equivalentStates!=clearedCurrentPartition) calcSplitter=true;
          
          //calc splitter if necessary
          if(calcSplitter) calculateSplitter(equivalentStates.moveDown(),equivalentStates);
          
          //remove equivalent states
          clearedCurrentPartition&=!equivalentStates;
        }
      }
      if(checkBlocks==0) checkBlocks=1000;
      else --checkBlocks;
    }
    stopwatch_bisimulation.stop();
    result_stream << "Bisimulation took " << stopwatch_bisimulation.toString() << endl;
    result_stream << "totalNrOfSplitters: " << totalNrOfSplitters << endl;
    result_stream << "classes: " << nrOfPartitions << endl;

    bool bisimilar = true;
    result_stream << "start states in: " << flush;
    int classesWithStartStates=0;
    
    BDDFunction switch_var = automaton->getOriginalStateVars();         //all vars for the new automaton
    switch_var = switch_var.exists(symbCompA->getOriginalStateVars());  //substract the 
    switch_var = switch_var.exists(symbCompB->getOriginalStateVars());  //old variables
    BDDFunction var_first = !switch_var;
    BDDFunction var_second = switch_var;
    
    for (unsigned int i=0; bisimilar && (i<blockContainer.size()); ++i) {
      if ((blockContainer[i] & automaton->getInitialStates()) != 0.0) {
        //we found a block which contains starting states
        result_stream << i << " " << flush;
        ++classesWithStartStates;
        //check if start states of both original automata are present in this block
        if ( ((symbCompA->getInitialStates() & blockContainer[i] & var_first) == 0.0) || 
             ((symbCompB->getInitialStates() & blockContainer[i] & var_second) == 0.0) ) bisimilar=false;
      }
    }
    result_stream << endl;

    result->setCheckPassed(bisimilar);
    result->setAdditionalInfo(result_stream.str());

    return result;
  }
  
}

