/***************************************************************************
 *   Copyright (C) 2008-2009 by Sascha Klueppelholz                        *
 *   klueppel@tcs.inf.tu-dresden.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "SymbAutomatonLanguageAcceptor.hpp"
#include "mcreo.hpp"
#include "mcreo_functions.hpp"
#include "vereofy/system/datadomain.hpp"
#include "vereofy/system/symb_automaton/symb_component_instance.hpp"
#include "vereofy/common/tracing.hpp"

#include <iostream>
#include <boost/lexical_cast.hpp>

/**
 @author Sascha Klueppelholz
 */

using namespace std;
using namespace vereofy;

//********************************** public **********************************


SymbAutomatonLanguageAcceptor::SymbAutomatonLanguageAcceptor(const SymbAutomaton::ptr system) : SymbAutomaton(Symb_FlatDataDomain::ptr()){
  ca=system;
    
  acceptingStates = BDDFunction::ZERO();
  initialStates   = BDDFunction::ZERO();
  reachableStates = BDDFunction::ONE();

  csPorts=system->getPortVars();
  csDCVariables=system->getDataConstraintVars();
  csTransitionLabelVariables=system->getTransitionLabelVars();
  copy_symb_ces(*system);
  copy_symb_tls(*system);
}

/**
 * \brief Constructor of SymbAutomatonLanguageAcceptor
 *
 * Creates a finite automaton and parses the BTSL formula specified by the string passed to the constructor 
 * \sa mcreo.hpp
 *
 */
SymbAutomatonLanguageAcceptor::SymbAutomatonLanguageAcceptor(const string& expr) : SymbAutomaton(Symb_FlatDataDomain::ptr()){
  // remember expression
  stream_expr=expr;
  
  // initialize variable level
  varlevel = 0;
}

void SymbAutomatonLanguageAcceptor::parse(const SymbAutomaton::ptr p){
  ca=p;
  
  csPorts=p->getPortVars();
  csDCVariables=p->getDataConstraintVars();
  csTransitionLabelVariables=p->getTransitionLabelVars();
  copy_symb_ces(*p);
  
  parse();
}

void SymbAutomatonLanguageAcceptor::makeDeterministic(){
  SymbAutomaton::makeDeterministic(acceptingStates);
}

void SymbAutomatonLanguageAcceptor::SALA_State_Renderer::render(ostream& out, const BDDFunction& assignment, bool use_x) const {
  State_Renderer::render(out, assignment, use_x);
  
  BDDFunction accepting_states=((SymbAutomatonLanguageAcceptor&)automaton).getAccept();
  if (bdd_sets_have_intersection(assignment, use_x ? accepting_states : accepting_states.moveDown())) {
    out << " *";
  }
}

bool SymbAutomatonLanguageAcceptor::SALA_State_Renderer::render_dot_style(ostream& out, const BDDFunction& assignment, bool is_initial, bool is_terminal) const {
  return false;
}
      
SymbAutomatonLanguageAcceptor::State_Renderer::ptr SymbAutomatonLanguageAcceptor::getDefaultStateRenderer() const {
  return SALA_State_Renderer::ptr(new SALA_State_Renderer(*this));
}

void SymbAutomatonLanguageAcceptor::toDot(ostream& out) const {
  SymbAutomaton::toDot(out, SALA_State_Renderer(*this), SymbComponentInstance::Component_IO_Renderer(*boost::dynamic_pointer_cast<SymbComponentInstance>(ca)));
}

//********************************** private **********************************

/**
 * \brief This method parses the given expression
 *
 * This method parses the expression stored in stream_expr
 *
 */
void SymbAutomatonLanguageAcceptor::parse(){	
  debugOut(2," building automaton for expression: " + stream_expr);
	
  // expression not given?
  if (!stream_expr.size()) throw "Parsing an empty expression!";
  
  // star operator (alpha*)
  if (stream_expr[stream_expr.length()-1]=='*'){
    // create new automaton for alpha
    SymbAutomatonLanguageAcceptor::ptr tmp=SymbAutomatonLanguageAcceptor::ptr(new SymbAutomatonLanguageAcceptor(stream_expr.substr(0,stream_expr.length()-1))); tmp->parse(ca);
    
    // common operations:
    eat(tmp);
    
    initialStates = tmp->initialStates;
    delta = (tmp->delta) | (tmp->acceptingStates & (tmp->initialStates & tmp->delta).exists(tmp->csOriginalX));
    acceptingStates = tmp->acceptingStates|tmp->initialStates;
    return;
  }
  
  // star operator (alpha+)
  if (stream_expr[stream_expr.length()-1]=='+'){
    // create new automaton for alpha
    SymbAutomatonLanguageAcceptor::ptr tmp=SymbAutomatonLanguageAcceptor::ptr(new SymbAutomatonLanguageAcceptor(stream_expr.substr(0,stream_expr.length()-1))); tmp->parse(ca);
	  
    // common operations:
    eat(tmp);
    
    initialStates = tmp->initialStates;
    delta = (tmp->delta) | (tmp->acceptingStates & (tmp->initialStates & tmp->delta).exists(tmp->csOriginalX));
    acceptingStates = tmp->acceptingStates;
    return;
  }
  
  // negation of the automaton neg alpha
  if (stream_expr[0]=='!'){
    cerr << "Warning. Complement of alpha used. Are you sure?" << endl;
    
    // create new automaton for alpha
    stream_expr=stream_expr.erase(0,1);
    parse();
    
    // for complement the automaton needs to be a) complete 
    // @TODO: this could also be part of the bdd_helpers
    // introduce a new state variable for a capturing state
    StateVars states = SymbAutomaton::findOrAddGlobalState(string("EA::Q")+boost::lexical_cast<string>(++varlevel), 1);
    
    BDDFunction term = !delta.exists(getTransitionVars() & getCopyStateVars());
    
    delta = ( delta                           & !states.X->at(0) & !states.Y->at(0))
           | (!delta & !term                   & !states.X->at(0) &  states.Y->at(0))
           | (           term & term.moveDown()                    &  states.Y->at(0));
    
    // modify the starting and accepting state relation
    initialStates   &= !states.X->at(0);
    acceptingStates &= !states.X->at(0);
    
    // the automaton must be b) deterministic
    makeDeterministic();
    
    // we can set the accepting states to its complement
    acceptingStates=!acceptingStates;
    
    stream_expr="!"+stream_expr;
    return;
  }
  
  // watch out for binary operators (REOPS) on level zero
  unsigned int OPPos=findOperatorPosition(stream_expr,REOPS);
  
  // and now all binary cases...
  if (OPPos){
    // create new automata for alpha1 and alpha2
    SymbAutomatonLanguageAcceptor::ptr tmp1=SymbAutomatonLanguageAcceptor::ptr(new SymbAutomatonLanguageAcceptor(stream_expr.substr(0,OPPos))); tmp1->parse(ca);
    SymbAutomatonLanguageAcceptor::ptr tmp2=SymbAutomatonLanguageAcceptor::ptr(new SymbAutomatonLanguageAcceptor(stream_expr.substr(OPPos+1,stream_expr.length()-OPPos-1))); tmp2->parse(ca);
    
    // fill up to have the same variables
    fill(tmp1, tmp2->getOriginalStateVars().exists(tmp1->getOriginalStateVars()));
    fill(tmp2, tmp1->getOriginalStateVars().exists(tmp2->getOriginalStateVars()));
    
    // and eat them...
    eat(tmp1);
    eat(tmp2);
    
    // add a new variable
    StateVars states = SymbAutomaton::findOrAddGlobalState(string("EA::Q")+boost::lexical_cast<string>(++varlevel),1);
    
    // and now the operator specific part:
    // concatenation of SymbAutomatonLanguageAcceptor
    if (stream_expr[OPPos]==';'){
      initialStates = (tmp1->initialStates & states.X->at(0));
      delta = (tmp1->delta & states.X->at(0) & states.Y->at(0))
      | (tmp1->acceptingStates & (tmp2->initialStates & tmp2->delta).exists(tmp2->getOriginalStateVars()) & states.X->at(0) & !states.Y->at(0))
      | (tmp2->delta & !states.X->at(0) & !states.Y->at(0));
      acceptingStates = tmp2->acceptingStates & !states.X->at(0);
    }
	  
    // union of SymbAutomatonLanguageAcceptors (a|b)
    else if (stream_expr[OPPos]=='|'){
      initialStates = (tmp1->initialStates & states.X->at(0)) | (tmp2->initialStates & !states.X->at(0));
      delta = (tmp1->delta & states.X->at(0) & states.Y->at(0)) | (tmp2->delta & !states.X->at(0) & !states.Y->at(0));
      acceptingStates = (tmp1->acceptingStates & states.X->at(0)) | (tmp2->acceptingStates & !states.X->at(0));
    }
    
    return;
  }
  
  // remove brackets
  if ((stream_expr[0]=='(') && (stream_expr[stream_expr.length()-1]==')')){
    stream_expr=stream_expr.substr(1,stream_expr.length()-2);
    parse();
    
    return;
  }
  
  // and now the terminal case finally...
  
  // get a new state variable
  StateVars states = SymbAutomaton::findOrAddGlobalState("EA::Q0",1);
	
  // set initial relation	
  initialStates = !states.X->at(0);
	
  // set acceptingStatesing state
  acceptingStates = states.X->at(0);
	
  // get function for the data constraint
  BDDFunction dc=BDDFunction::ONE();
  try{
    vereofy::SymbComponentInstance::ptr ci_ca=boost::dynamic_pointer_cast<vereofy::SymbComponentInstance>(ca);
    if (!ci_ca) throw "Don't know how to get BDDFunction for this data constraint.";
    
    OPPos=findOperatorPosition(stream_expr,",");
    string working_copy=stream_expr; 
    
    while (OPPos){
      dc&=ci_ca->parse_io_guard(working_copy.substr(OPPos+1,working_copy.length()-OPPos), true, false);
      working_copy=working_copy.erase(OPPos,working_copy.length()-OPPos);
      OPPos=findOperatorPosition(working_copy,",");
    }
    
    dc&=ci_ca->parse_io_guard(working_copy, true, false);
  } catch(exception& e){ throw "Can not parse data constraint: " + stream_expr;}
  
  // set transition relation
  delta = !states.X->at(0) & dc & states.Y->at(0);
}

void SymbAutomatonLanguageAcceptor::fill(SymbAutomatonLanguageAcceptor::ptr tmp, BDDFunction diff){
  tmp->csOriginalX &= diff;
  tmp->csCopyY &= diff.moveDown();
  tmp->csCopyCopyZ &= diff.moveDown().moveDown();
  tmp->haltSystem &= !(diff ^ (diff.moveDown()));
  
  diff = diff.change();
  tmp->initialStates &= diff;
  tmp->delta &= diff & diff.moveDown();
  tmp->acceptingStates &= diff;
}

void SymbAutomatonLanguageAcceptor::eat(const SymbAutomatonLanguageAcceptor::ptr tmp){
  haltSystem&=tmp->haltSystem;
  
  csOriginalX&=tmp->csOriginalX;
  csCopyY&=tmp->csCopyY;
  csCopyCopyZ&=tmp->csCopyCopyZ;
  
  csPorts&=tmp->csPorts;
  csDCVariables&=tmp->csDCVariables;
  
  // take the maximum of variable levels as new value
  if (varlevel<tmp->varlevel) varlevel=tmp->varlevel;
}

