/***************************************************************************
 *   Copyright (C) 2008-2009 by Sascha Klueppelholz                        *
 *   klueppel@tcs.inf.tu-dresden.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "hybridFiniteAutomaton.hpp"

#include "vereofy/system/symb_automaton/symb_component_instance.hpp"
#include "vereofy/system/symb_automaton/symb_automaton.hpp"
#include "vereofy/modelcheck/specifications/asl/branching_time_logics_parser.hpp"
#include "vereofy/common/math_helpers.hpp"

#include <queue>
#include <list>
#include <boost/foreach.hpp>
#define  foreach BOOST_FOREACH

/**
 @author Sascha Klueppelholz
 @author Karsten Lehmann
 */

using namespace std;
using namespace vereofy;

unsigned long automatonState::last_id=0;

void automatonState::addTransition(const transition_t& transition){
  addTransition(transition.first,transition.second);
}

void automatonState::addTransition(automatonState* succ, const BDDFunction& label){
  if (label==BDDFunction::ZERO()){
    cerr << "Warning: false transition in automaton for the regular expression. (ignored)" << endl;
    return;
  }
  
  if (shouldBeTerminal){      
    throw "Error. Cannot add transition from a stopp state!";
  }
  
  successors_t::const_iterator it = successors.find(succ);
  successors[succ]=(it==successors.end())?(label):(successors[succ]|label);
}

///////////////////////////////////////////////////////////////////////////////////////////////////

hybridFiniteAutomaton hybridFiniteAutomaton::operator& (const hybridFiniteAutomaton& _second) const{
  hybridFiniteAutomaton result;
  
  hybridFiniteAutomaton first(*this);
  hybridFiniteAutomaton second(_second);
  
  // treat stopped
  if (first.containsStopp&&!second.containsStopp){
    second.addStopped();
  }else if (second.containsStopp&!containsStopp){
    first.addStopped();
  }
  result.containsStopp=first.containsStopp||second.containsStopp;
  
  // reached does not need id_less as comparator to lead to deterministic execution, 
  // as there is no iterator on it
  typedef pair<automatonState*, automatonState*> ststPair;
  map<ststPair, automatonState::ptr> reached;  
  
  queue<ststPair> toDo;
  ststPair active;
  foreach(automatonState* init_1, first.startingStates){
    foreach(automatonState* init_2, second.startingStates){
      active = ststPair(init_1, init_2);
      automatonState::ptr newstate(new automatonState);
      reached[active] = newstate;
      result.startingStates.insert(newstate.get());
      toDo.push(active);
    }
  }
  
  while (!toDo.empty()){
    active = toDo.front();
    toDo.pop();
    result.allStates.insert(reached[active]);
    if ((first.acceptingStates.find(active.first) != first.acceptingStates.end()) && (second.acceptingStates.find(active.second) != second.acceptingStates.end()))
      result.acceptingStates.insert(reached[active].get());
    
    foreach(automatonState::transition_t transfirst, active.first->successors){
      foreach(automatonState::transition_t transsecond, active.second->successors){
        BDDFunction intersection = transfirst.second & transsecond.second;
        if (intersection != BDDFunction::ZERO()){
          automatonState::ptr& dest = reached[ ststPair(transfirst.first, transsecond.first) ];
          if (dest.get() == NULL){
            dest.reset(new automatonState);
            toDo.push(ststPair(transfirst.first, transsecond.first));
            reached[ ststPair(transfirst.first, transsecond.first) ] = dest;
          }
          reached[active]->addTransition(dest.get(), intersection);     
        }
      }
    }
  }
  
  return result;
}

hybridFiniteAutomaton hybridFiniteAutomaton::operator| (const hybridFiniteAutomaton& second) const{
  hybridFiniteAutomaton first(*this);
  hybridFiniteAutomaton second_c(second);
  return first.cup(second_c);
}

hybridFiniteAutomaton hybridFiniteAutomaton::operator% (const hybridFiniteAutomaton& second) const{
  hybridFiniteAutomaton first(*this);
  hybridFiniteAutomaton second_c(second);
  return first.concat(second_c);
}

hybridFiniteAutomaton hybridFiniteAutomaton::operator/ (const hybridFiniteAutomaton& second) const{
  hybridFiniteAutomaton second_ptr(second);
  return (*this & !second_ptr);
}

hybridFiniteAutomaton hybridFiniteAutomaton::operator! (void) const{
  hybridFiniteAutomaton copy(*this);
  return copy.negate();
}

hybridFiniteAutomaton hybridFiniteAutomaton::operator* (void) const{
  hybridFiniteAutomaton copy(*this);
  return copy.star();
}

hybridFiniteAutomaton hybridFiniteAutomaton::operator+ (void) const{
  hybridFiniteAutomaton copy(*this);
  return copy.plus();
}

bool hybridFiniteAutomaton::operator== (const hybridFiniteAutomaton& second) const{
  if ((*this / second).acceptingStates.size() == 0){
    return ((second / *this).acceptingStates.size() == 0);
  } else return false;
}

bool hybridFiniteAutomaton::operator!= (const hybridFiniteAutomaton& second) const{
  return !(*this == second);
}

///////////////////////////////////////////////////////////////////////////////////////////////////

hybridFiniteAutomaton::hybridFiniteAutomaton(const hybridFiniteAutomaton& aut){
  map<automatonState*, automatonState::ptr, ptr_id_less<automatonState*> > statecopy;
  foreach(automatonState::ptr state, aut.allStates){
    automatonState::ptr newstate(new automatonState);
    statecopy[state.get()] = newstate;
    allStates.insert(newstate);
    if (aut.startingStates.find(state.get()) != aut.startingStates.end())
      startingStates.insert(newstate.get());
    if (aut.acceptingStates.find(state.get()) != aut.acceptingStates.end())
      acceptingStates.insert(newstate.get());
  }
  
  foreach(automatonState::ptr state, aut.allStates){
    foreach(automatonState::transition_t trans, state->successors){
      statecopy[state.get()]->addTransition(statecopy[trans.first].get(), trans.second);
    }
  }
  
  containsStopp = aut.containsStopp;
}

hybridFiniteAutomaton::ptr hybridFiniteAutomaton::toHybridFiniteAutomaton(const vparser::pt_node& node, const SymbAutomaton::ptr system){  
  hybridFiniteAutomaton::ptr hfa;
  
  switch (node.value.id().to_long()){
    case branching_time_logics_parser::ID_IO_GUARD_ATOM: case 0:{
      if (node.children.size()==1) {
        hfa = toHybridFiniteAutomaton(node.children.at(0),system);
        ERROR_CHECK_ASSERT(hfa->startingStates.size());
        return hfa;
      }
      if (node.children.size()==2) hfa = toHybridFiniteAutomaton(node.children.at(1),system);
      else {
        vereofy::SymbComponentInstance::ptr ci_ca=boost::dynamic_pointer_cast<vereofy::SymbComponentInstance>(system);
        if (!ci_ca) throw "Not a valid system.";
        
        const vparser::pt_node& aNode = (node.children.size()==0)?(node):(node.children.at(1));
        string result = string(aNode.value.begin(),aNode.value.end());
        
        if (result!="stopped"){  
          try{ 
            hfa = hybridFiniteAutomaton::ptr(new hybridFiniteAutomaton(ci_ca->parse_io_guard(result, true, false),false));
          }
          catch(std::exception& e){
            cout << e.what() << "\n";
            throw "Can not resolve I/O-guard: " + result;
          }
        }else{
          VarManager::getInstance()->findOrAddGlobalBits("stopped_a", 1).at(0);
          VarManager::getInstance()->findOrAddGlobalBits("stopped_o", 1).at(0);
          VarManager::getInstance()->findOrAddGlobalBits("stopped_t", 1).at(0);
          
          BDDFunction stopped_io_guard = !ci_ca->parse_io_guard("dataflow", true, false);
          stopped_io_guard &= BDDFunction::projection("stopped_t") | BDDFunction::projection("stopped_a") | BDDFunction::projection("stopped_o");
          
          hfa = hybridFiniteAutomaton::ptr(new hybridFiniteAutomaton(stopped_io_guard,true));
        }
      }
      
      if (node.children.size()==2) hfa->negate();
      
      ERROR_CHECK_ASSERT(hfa->startingStates.size());
      return hfa;
    }
      
    case branching_time_logics_parser::ID_IO_GUARD_STARED:{
      if (node.children.size()==1){ 
        hfa = toHybridFiniteAutomaton(node.children.at(0), system);
        ERROR_CHECK_ASSERT(hfa->startingStates.size());
        return hfa;
      }
      hfa = toHybridFiniteAutomaton(node.children.at(0), system);
      string op = string(node.children.at(1).value.begin(),node.children.at(1).value.end());
      
      if (op == "*") hfa->star();
      else if (op == "+") hfa->plus();
      
      ERROR_CHECK_ASSERT(hfa->startingStates.size());
      return hfa;
    }
      
    case branching_time_logics_parser::ID_IO_SEQUENCE:{
      hfa = toHybridFiniteAutomaton(node.children.at(0), system);
      for (unsigned int i=1;i<node.children.size();++i){
        hfa->concat(*toHybridFiniteAutomaton(node.children.at(i), system));
      }
      
      ERROR_CHECK_ASSERT(hfa->startingStates.size());
      return hfa;
    }
      
    case branching_time_logics_parser::ID_IO_TERM_AND:{
      hfa = toHybridFiniteAutomaton(node.children.at(0), system);
      for (unsigned int i=1;i<node.children.size();++i){
        hfa->cap(*toHybridFiniteAutomaton(node.children.at(i), system));
      }
      
      ERROR_CHECK_ASSERT(hfa->startingStates.size());
      return hfa;
    }
      
    case branching_time_logics_parser::ID_IO_REGULAR_IO_EXPRESSION:{
      hfa = toHybridFiniteAutomaton(node.children.at(0), system);
      for (unsigned int i=1;i<node.children.size();++i){
        hfa->cup(*toHybridFiniteAutomaton(node.children.at(i), system));
      }
      
      ERROR_CHECK_ASSERT(hfa->startingStates.size());
      return hfa;
    }
      
    default:{
      THROW_UNIMPLEMENTED();
    }
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////

hybridFiniteAutomaton::hybridFiniteAutomaton(const BDDFunction& transitionLabel, const bool stopped){
  automatonState* start = new automatonState();
  automatonState* accept = new automatonState();
  accept->shouldBeTerminal = stopped;
  
  startingStates.insert(start);
  allStates.insert(automatonState_ptr(start));
  
  acceptingStates.insert(accept);
  allStates.insert(automatonState_ptr(accept));
  
  start->addTransition(accept,transitionLabel);
  containsStopp = stopped;
}  

hybridFiniteAutomaton::~hybridFiniteAutomaton(){
  startingStates.clear(); 
  acceptingStates.clear();
  allStates.clear();
}

void automatonState::addSuccessors(ptr_state_set_t& theSet){
  successors_t::const_iterator it = successors.begin();
  while (it != successors.end()){
    if ((theSet.insert((*it).first)).second) (*it).first->addSuccessors(theSet);
    ++it;
  }
}

void hybridFiniteAutomaton::makeDeterministic(){
  ptr_state_set_t new_initstates, new_finalstates;
  shared_ptr_state_set_t allStates_copy = allStates;
  allStates.clear();
  
  typedef ptr_state_set_t StateSet;
  queue<StateSet> toDo;
  set<StateSet> reached;
  map<StateSet, automatonState::ptr> newstates;
  
  StateSet active(startingStates);
  newstates[active] = automatonState::ptr(new automatonState);
  allStates.insert(newstates[active]);
  new_initstates.insert(newstates[active].get());
  toDo.push(active);
  reached.insert(active);
  
  while (!toDo.empty()){
    active = toDo.front();
    toDo.pop();
    
    // set final status
    foreach(automatonState* state, active){
      if (acceptingStates.find(state) != acceptingStates.end()){
        new_finalstates.insert(newstates[active].get());
        break;
      }
    }
    
    // calulate then nonddeterministic transition_function for the new state -> store in nondet_transitons
    typedef map<automatonState*, BDDFunction, ptr_id_less<automatonState*> > nondet_transitions_map;
    nondet_transitions_map nondet_transitions;
    
    foreach(automatonState* state, active){
      foreach(automatonState::transition_t trans, state->successors){
        pair<nondet_transitions_map::iterator, bool> it = nondet_transitions.insert(trans);
        if (!it.second)
          it.first->second = it.first->second | trans.second;
      }
    }
    
    // determinise nondet_transition
    while(!nondet_transitions.empty()){
      
      // use the first state
      nondet_transitions_map::iterator it_trans = nondet_transitions.begin();
      StateSet newstateset;
      newstateset.insert(it_trans->first);
      BDDFunction tmp = it_trans->second;
      
      // calculate the biggest intersection with the rest states (tmp)
      foreach(automatonState::transition_t tr, nondet_transitions){
        BDDFunction intersection = tmp & tr.second;
        if (intersection != BDDFunction::ZERO()){
          tmp = intersection;
          newstateset.insert(tr.first);
        }
      }
      
      // now we have a new transition, lets find out if the destination was reached before
      pair<set<StateSet>::iterator, bool> it_reached = reached.insert(newstateset);
      if (!it_reached.second){
        newstateset = *(it_reached.first);
      }else {
        toDo.push(newstateset);
        newstates[newstateset] = automatonState::ptr(new automatonState);
        allStates.insert(newstates[newstateset]);
      }
      newstates[active]->addTransition(newstates[newstateset].get() ,tmp);
      
      // update the other state, transitionset pairs
      foreach(automatonState* state, newstateset){
        it_trans = nondet_transitions.find(state);
        it_trans->second = it_trans->second & !tmp;
        if (it_trans->second == BDDFunction::ZERO()) nondet_transitions.erase(it_trans);
      }
    }
  }
  
  startingStates = new_initstates;
  acceptingStates = new_finalstates;
}

void hybridFiniteAutomaton::powerSetConstruction(){  
  ptr_state_set_t newStartingStates;
  ptr_state_set_t newAcceptingStates;
  shared_ptr_state_set_t newAllStates;
  
  list<ptr_state_set_t > todoList;
  map<ptr_state_set_t,automatonState*> stateMapping;
  
  todoList.push_back(startingStates);
  newStartingStates.insert(makeDeterministic(todoList, stateMapping, newAcceptingStates));
  newAllStates.insert(automatonState_ptr(new automatonState(*(*(newStartingStates.begin())))));
  while (todoList.size()){
    newAllStates.insert(automatonState_ptr(makeDeterministic(todoList, stateMapping, newAcceptingStates)));
    todoList.pop_front();
  }
  
  // overwrite old automaton
  startingStates  = newStartingStates;
  acceptingStates = newAcceptingStates;
  allStates       = newAllStates;
}

void hybridFiniteAutomaton::makeComplete(){
  automatonState* capturingState = new automatonState();
  capturingState->addTransition(capturingState, BDDFunction::ONE());
  bool insertCapturingState = false;
  
  shared_ptr_state_set_t::const_iterator it = allStates.begin();
  while (it!=allStates.end()){
    if (!(*it)->shouldBeTerminal){
      BDDFunction remainder = BDDFunction::ONE();
      automatonState::successors_t::const_iterator transition_it = (*it)->successors.begin();
      
      while (transition_it!=(*it)->successors.end()){
        remainder &= !(*transition_it).second;
        ++transition_it;
      }
      
      if (remainder != BDDFunction::ZERO()){
        (*it)->addTransition(capturingState,remainder);
        insertCapturingState = true;
      }
    }
    ++it;
  }
  
  if (insertCapturingState) allStates.insert(automatonState_ptr(capturingState));
  else delete capturingState;
}

void hybridFiniteAutomaton::removeUselessStates(){
	reverseAutomaton();
	ptr_state_set_t reachable = computeReachableStates();
  
  shared_ptr_state_set_t::iterator it = allStates.begin();
  while (it!=allStates.end()){
    automatonState::ptr state = (*it);
    ++it;
    
    if (reachable.find(state.get()) == reachable.end()){
      if (allStates.size()>1) allStates.erase(state);
    }
  }
  
	reverseAutomaton();	
}

void hybridFiniteAutomaton::reverseAutomaton(){
  map<automatonState::ptr, automatonState::successors_t > transitions_copy;
  typedef pair<automatonState::ptr, automatonState::successors_t > trans_pair;
  
  foreach(automatonState::ptr state, allStates){
    transitions_copy.insert(trans_pair(state, state->successors));
    state->successors.clear();    
  }
  
  foreach(trans_pair trp, transitions_copy){
    foreach(automatonState::transition_t trans, trp.second){
      bool wasTerminal = trans.first->shouldBeTerminal;
      trans.first->shouldBeTerminal=false;  
      trans.first->addTransition(automatonState::transition_t(trp.first.get(), trans.second));
      trans.first->shouldBeTerminal=wasTerminal;
    }
  }
  
  ptr_state_set_t helper = startingStates;
  startingStates = acceptingStates;
  acceptingStates = helper;
  
  helper = computeReachableStates(); 
  shared_ptr_state_set_t::iterator it = allStates.begin();
  while(it != allStates.end()){
    automatonState::ptr state = (*it);
    ++it;
    
    if (helper.find(state.get()) == helper.end()){
      acceptingStates.erase(state.get());
      allStates.erase(state);
    }
  }  
}

hybridFiniteAutomaton::ptr_state_set_t hybridFiniteAutomaton::computeReachableStates() const{
  ptr_state_set_t result(startingStates);
  ptr_state_set_t::const_iterator it = startingStates.begin();
  while (it!=startingStates.end()){
    (*it)->addSuccessors(result);
    ++it;
  }
  
  return result;
}

SymbAutomatonLanguageAcceptor_ptr hybridFiniteAutomaton::encode(const SymbAutomaton::ptr system) const{
  SymbAutomatonLanguageAcceptor_ptr result=SymbAutomatonLanguageAcceptor::ptr(new SymbAutomatonLanguageAcceptor(system));
  
  ptr_state_set_t reachableStates = computeReachableStates();
  
  StateVars states = result->findOrAddGlobalState("HFA_Q" + boost::lexical_cast<string>(automatonState::last_id), logSize(reachableStates.size()));
  
  unsigned long id = 0;
  ptr_state_set_t::const_iterator it = reachableStates.begin();
  map<automatonState*,unsigned long, ptr_id_less<automatonState*> > stateNumber;
  while (it!=reachableStates.end()){
    stateNumber[(*it)] = id;
    ++it;
    ++id;
  }
  
  it = reachableStates.begin();
  result->reachableStates=BDDFunction::ZERO();
  
  while (it!=reachableStates.end()){
    BDDFunction startState = encode(stateNumber[(*it)],*(states.X));
    result->reachableStates|=startState;
    
    if (startingStates.find(*it)!=startingStates.end()){
      result->initialStates|=startState;
    }
    
    if (acceptingStates.find(*it)!=acceptingStates.end()){
      result->acceptingStates|=startState;
    }
    
    automatonState::successors_t::const_iterator transition_it=(*it)->successors.begin();
    while (transition_it!=(*it)->successors.end()){
      BDDFunction targetState = encode(stateNumber[(*transition_it).first],*(states.Y));
      result->delta|=startState & ((*transition_it).second) & targetState;
      ++transition_it;
    }
    ++it;
  }
  
  result->containsStopp = containsStopp;
  
  return result;
}

void hybridFiniteAutomaton::addStopped(void){
  BDDFunction stopped_a = BDDFunction::projection("stopped_a");
  BDDFunction stopped_o = BDDFunction::projection("stopped_o");
  BDDFunction stopped_t = BDDFunction::projection("stopped_t");
  BDDFunction stopped = stopped_a | stopped_o | stopped_t;
  
  foreach(automatonState::ptr state, allStates){
    automatonState::successors_t::iterator it = state->successors.begin();
    while (it != state->successors.end()){
      (*it).second&=!stopped;
      ++it;
    }
  }
}

hybridFiniteAutomaton& hybridFiniteAutomaton::cup(const hybridFiniteAutomaton& _second){
  hybridFiniteAutomaton second(_second);
  
  // treat stopped
  if (containsStopp&&!second.containsStopp){
    second.addStopped();
  }else if (second.containsStopp&!containsStopp){
    addStopped();
  }
  containsStopp|=second.containsStopp;
  
  foreach(automatonState::ptr state, second.allStates){
    allStates.insert(state);
  }
  
  foreach(automatonState* state, second.startingStates){
    startingStates.insert(state);
  }
  
  foreach(automatonState* state, second.acceptingStates){
    acceptingStates.insert(state);
  }
  
  return *this;
}

hybridFiniteAutomaton& hybridFiniteAutomaton::cap(const hybridFiniteAutomaton& second){
  hybridFiniteAutomaton result = *this & second;
  allStates = result.allStates;
  acceptingStates = result.acceptingStates;
  startingStates = result.startingStates;
  containsStopp =result.containsStopp;
  return *this;
}

hybridFiniteAutomaton& hybridFiniteAutomaton::concat(const hybridFiniteAutomaton& _second){
  bool firstContainsEpsilon = false;
  bool secondContainsEpsilon = false;
  
  hybridFiniteAutomaton second(_second);
  
  // treat stopped
  if (containsStopp&&!second.containsStopp){
    second.addStopped();
  }else if (second.containsStopp&!containsStopp){
    addStopped();
  }
  containsStopp|=second.containsStopp;
  
  foreach(automatonState* accept, acceptingStates){
    firstContainsEpsilon |= (startingStates.find(accept) != startingStates.end());
    foreach(automatonState* starting, second.startingStates){
      secondContainsEpsilon |= (second.acceptingStates.find(starting) != second.acceptingStates.end());
      foreach(automatonState::transition_t trans, starting->successors)
      accept->addTransition(trans);
    }
  }
  
  foreach(automatonState::ptr state, second.allStates)
  allStates.insert(state);
  if (firstContainsEpsilon)
    foreach(automatonState* state, second.startingStates)
    startingStates.insert(state);
  
  if (secondContainsEpsilon){
    foreach(automatonState* state, second.acceptingStates)
    acceptingStates.insert(state);
  } else { acceptingStates = second.acceptingStates;}
  
  return *this;
}

hybridFiniteAutomaton& hybridFiniteAutomaton::star(void){
  this->plus();
  
  automatonState* newState = new automatonState();
  startingStates.insert(newState);
  acceptingStates.insert(newState);
  
  automatonState::ptr newStatePtr = automatonState::ptr(newState);
  allStates.insert(newStatePtr);
  
  return *this;
}

hybridFiniteAutomaton& hybridFiniteAutomaton::plus(void){
  foreach(automatonState* accept, acceptingStates){
    foreach(automatonState* starting, startingStates){
      foreach(automatonState::transition_t trans, starting->successors){
        accept->addTransition(trans);}
    }
  }
  
  return *this;
}

hybridFiniteAutomaton& hybridFiniteAutomaton::negate(){
  makeDeterministic();
  makeComplete();  
  
  ptr_state_set_t newAcceptingStates;
  foreach(automatonState::ptr state, allStates){
    if (acceptingStates.find(state.get()) == acceptingStates.end()){
      newAcceptingStates.insert(state.get());
    }
  }
  acceptingStates = newAcceptingStates;
  
  return *this;
}

bool hybridFiniteAutomaton::isomorphic(const hybridFiniteAutomaton& aut1, const hybridFiniteAutomaton& aut2){
  if ((aut1.startingStates.size() != aut2.startingStates.size())) return false;
  state_to_state_map equal_states;
  ptr_state_set_t initstates_2 = aut2.startingStates;
  bool exit;
  foreach(automatonState* init1, aut1.startingStates){
    ptr_state_set_t::iterator it = initstates_2.begin();
    exit = false;
    while (it!=initstates_2.end()){
      automatonState* state = (*it);
      ++it;
      
      if (hybridFiniteAutomaton::checkStateEqual(equal_states, init1, state, aut1, aut2)){
        initstates_2.erase(state);
        exit = true;
        break;
      } else {
        equal_states.erase(init1);
      }   
    }
    if (!exit) return false;
  }
  return true;
}

void hybridFiniteAutomaton::minimizeBrzozowski(){
  reverseAutomaton();
  makeDeterministic();
  reverseAutomaton();
  makeDeterministic();
}

void hybridFiniteAutomaton::minimizeHopcroft(){
  makeComplete();
  map<automatonState*,int, ptr_id_less<automatonState*> > state_in_ref;
  map<automatonState*,int, ptr_id_less<automatonState*> > new_state_in_ref;
  ptr_state_set_t reachable_states = computeReachableStates();
  ptr_state_set_t not_finalstates;
  foreach(automatonState*  state, reachable_states){
    if (acceptingStates.find(state) == acceptingStates.end()){
      not_finalstates.insert(state);
      state_in_ref[state] = 0;
    } else {
      state_in_ref[state] = 1;
    }
  }
  
  set<ptr_state_set_t > refinements; 
  set<ptr_state_set_t > newrefinements;
  refinements.insert(acceptingStates);
  refinements.insert(not_finalstates);
  
  bool stchanged = true;
  while (stchanged){
    stchanged = false;
    
    set<ptr_state_set_t >::iterator it = refinements.begin();
    while (it!=refinements.end()){
      ptr_state_set_t aeqset = (*it);
      ++it;
      
      int ref = 0;
      while (!aeqset.empty()){
        ptr_state_set_t newset;
        ptr_state_set_t::iterator stateit = aeqset.begin();
        ptr_state_set_t::iterator helper;
        automatonState* acctual_state = *stateit;
        newset.insert(acctual_state);
        new_state_in_ref[acctual_state] = ref;
        aeqset.erase(stateit);
        stateit = aeqset.begin();
        while (stateit != aeqset.end()){
          helper = stateit;
          helper++;
          foreach(automatonState::transition_t trans1, acctual_state->successors)
          foreach(automatonState::transition_t trans2, (*stateit)->successors)
          if ((trans1.second & trans2.second) != BDDFunction::ZERO() && state_in_ref[trans1.first] != state_in_ref[trans2.first]){
            stchanged = true;
            goto NotEqual;
          }
          newset.insert(*stateit);
          new_state_in_ref[*stateit] = ref;
          aeqset.erase(stateit);
        NotEqual:
          stateit = helper; 
        }
        newrefinements.insert(newset);
        ref++;
      }
    }
    
    refinements  = newrefinements;
    state_in_ref = new_state_in_ref;
    
    newrefinements.clear();
    new_state_in_ref.clear();
  }
  
  map<ptr_state_set_t ,automatonState*> newmap;
  ptr_state_set_t finalstates = acceptingStates;
  acceptingStates.clear();
  
  shared_ptr_state_set_t allstates_copy = allStates;
  allStates.clear(); 
  
  foreach(ptr_state_set_t newset, refinements){
    newmap[newset] = new automatonState;
    allStates.insert(automatonState::ptr(newmap[newset]));
    ptr_state_set_t::iterator state = newset.begin();
    
    if (newset.find(*startingStates.begin()) != newset.end()){
      startingStates.clear();
      startingStates.insert(newmap[newset]);
    }
    
    if (finalstates.find(*state) != finalstates.end()) acceptingStates.insert(newmap[newset]);
  }
  
  foreach(ptr_state_set_t newset, refinements){
    ptr_state_set_t::iterator state = newset.begin();
    foreach(automatonState::transition_t trans, (*state)->successors){
      foreach(ptr_state_set_t destset, refinements)
      if (destset.find(trans.first) != destset.end()){
        newmap[newset]->addTransition(newmap[destset], trans.second);
        break;
      }
    }
  }
}

/////////

automatonState* hybridFiniteAutomaton::makeDeterministic(list<ptr_state_set_t >& todoList, map<ptr_state_set_t, automatonState*>& stateMapping, ptr_state_set_t& newAcceptingStates){  
  automatonState* newStartingState;
  ptr_state_set_t newStartingStateSet;
  bool languageContainsEpsilon = false;
  
  map<automatonState*,BDDFunction, ptr_id_less<automatonState*> > possibleTransitions;
  ptr_state_set_t::const_iterator it = todoList.front().begin();
  while (it!=todoList.front().end()){
    languageContainsEpsilon|=(acceptingStates.find(*it)!=acceptingStates.end());
    newStartingStateSet.insert(*it);
    
    automatonState::successors_t::const_iterator transition_it = (*it)->successors.begin();
    while (transition_it != (*it)->successors.end()){
      if (possibleTransitions.find((*transition_it).first) == possibleTransitions.end()){
        possibleTransitions[(*transition_it).first] = (*transition_it).second;
      } else possibleTransitions[(*transition_it).first] |= (*transition_it).second;
      
      ++transition_it;
    }
    ++it;
  }
  
  // find or add new starting state
  if (stateMapping.find(newStartingStateSet) != stateMapping.end()){
    newStartingState = stateMapping[newStartingStateSet];
  } else{
    newStartingState = new automatonState();
    stateMapping[newStartingStateSet] = newStartingState;
    if (languageContainsEpsilon) newAcceptingStates.insert(newStartingState);
  }
  
  for (unsigned long allCombinations = 1; allCombinations <(unsigned long)(1 <<possibleTransitions.size()); ++allCombinations){
    automatonState* newTargetState;
    bool targetIsAccepting = false;
    BDDFunction transitionLabel = BDDFunction::ONE();
    
    unsigned long combination = allCombinations;
    ptr_state_set_t targetStateSet;
    
    map<automatonState*,BDDFunction, ptr_id_less<automatonState*> >::const_iterator transition_it = possibleTransitions.begin();
    while (transition_it != possibleTransitions.end() && transitionLabel != BDDFunction::ZERO()){
      if (combination%2){
        combination=(combination-1)/2;
        targetStateSet.insert((*transition_it).first);
        transitionLabel &= (*transition_it).second;
        targetIsAccepting |= (acceptingStates.find((*transition_it).first) != acceptingStates.end());
      }else{
        combination=(combination)/2;
        transitionLabel &= !(*transition_it).second;
      }
      
      ++transition_it;
    }
    
    // find or add new target state
    if (transitionLabel != BDDFunction::ZERO()){
      if (stateMapping.find(targetStateSet) != stateMapping.end()){
        newTargetState = stateMapping[targetStateSet];
      } else{
        newTargetState = new automatonState();
        stateMapping[targetStateSet] = newTargetState;
        if (targetIsAccepting) newAcceptingStates.insert(newTargetState);
        todoList.push_back(targetStateSet);
      }
      
      // add transition
      newStartingState->addTransition(newTargetState,transitionLabel);
    }
  }
  
  return newStartingState;
}

bool hybridFiniteAutomaton::checkStateEqual(state_to_state_map& equal_states, automatonState* act1, automatonState* act2, const hybridFiniteAutomaton& aut1, const hybridFiniteAutomaton& aut2){
  if (equal_states[act1] == act2) return true;
  if (act1->successors.size() != act2->successors.size()) return false;
  if ((aut1.acceptingStates.find(act1) == aut1.acceptingStates.end()) != (aut2.acceptingStates.find(act2) == aut2.acceptingStates.end())) return false;
  
  equal_states[act1] = act2;
  automatonState::successors_t transitions_2 = act2->successors;  // make copy of act2 successors
  bool equal_transition_found;
  
  foreach(automatonState::transition_t trans1, act1->successors){
    equal_transition_found = false;
    
    for (automatonState::successors_t::iterator it = transitions_2.begin();
         it != transitions_2.end();
         ++it) {
      if (trans1.second == it->second) {
        if (hybridFiniteAutomaton::checkStateEqual(equal_states, trans1.first, it->first, aut1, aut2)){
          // remove transition from transitions_2 so it is not considered in subsequent aut1 transition checks
          transitions_2.erase(it);  // erase(it) is ok, because we don't touch it afterwards
          
          equal_transition_found = true;
          break;
        }
      }
    }
    
    if (!equal_transition_found){
      // both states are not equal
      equal_states.erase(act1);
      return false;
    }
  } 
  return true;
}

BDDFunction hybridFiniteAutomaton::encode(unsigned long id, const BDDFunctionContainer& variables) const{
  BDDFunction result = BDDFunction::ONE();
  for (unsigned long i=0;i<variables.size();++i){
    result&=(id%2)?(variables[i]):(!(variables[i]));
    id=id>>1;
  }
  return result;
}
