/***************************************************************************
 *   Copyright (C) 2008-2009 by Sascha Klueppelholz                        *
 *   klueppel@tcs.inf.tu-dresden.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef HYBRID_FINITE_AUTOMATON_HPP
#define HYBRID_FINITE_AUTOMATON_HPP

#include <boost/shared_ptr.hpp>
#include <boost/operators.hpp>
#include <set>
#include <map>
#include <iostream>
#include "vereofy/common/shared_ptr_map.hpp"
#include "vereofy/system/symb_automaton/symb_automaton.hpp"
#include "vereofy/reo_parser/reo_parser_types.hpp"
#include "vereofy/modelcheck/checkers/branching_time_check/SymbAutomatonLanguageAcceptor.hpp"

/**
 @author Sascha Klueppelholz
 @author Karsten Lehmann
 */

namespace vereofy{
  
  class automatonState{
    friend class hybridFiniteAutomaton;
  public:
    typedef boost::shared_ptr<automatonState> ptr;
    typedef pair<automatonState*,BDDFunction> transition_t;
    typedef std::map<automatonState*,BDDFunction, ptr_id_less<automatonState*> > successors_t;
    
    automatonState(){id=last_id++; shouldBeTerminal = false;}
    ~automatonState(){ successors.clear(); }
    void addTransition(const transition_t&);
    
    unsigned long getID() {return id;}
  private:
    typedef std::set<automatonState*, ptr_id_less<automatonState*> > ptr_state_set_t;
    
    void addTransition(automatonState*, const BDDFunction&);
    void addSuccessors(ptr_state_set_t&);
    void doAddTransition(const transition_t&);
    
    static unsigned long last_id;
    
    unsigned long id;
    successors_t successors;
    bool shouldBeTerminal;
  };
  typedef automatonState::ptr automatonState_ptr;
  
  ///////////////////////////////////////////////////////////////////////////////////////////////////
  
  class hybridFiniteAutomaton{ 
  public:
    typedef boost::shared_ptr<hybridFiniteAutomaton> ptr;
    typedef std::set<automatonState*, ptr_id_less<automatonState*> > ptr_state_set_t;
    
    hybridFiniteAutomaton(const hybridFiniteAutomaton&);
    hybridFiniteAutomaton(){};
    hybridFiniteAutomaton(const BDDFunction&, const bool);
    ~hybridFiniteAutomaton();
    
    void makeDeterministic(void);
    void powerSetConstruction(void);
    void makeComplete(void);
    void removeUselessStates(void);
    void reverseAutomaton(void);
    
    ptr_state_set_t computeReachableStates() const;
    
    SymbAutomatonLanguageAcceptor_ptr encode(const SymbAutomaton::ptr) const;
    static hybridFiniteAutomaton::ptr toHybridFiniteAutomaton(const vparser::pt_node&, const SymbAutomaton::ptr);    
    
    hybridFiniteAutomaton& negate(void);
    hybridFiniteAutomaton& plus(void);
    hybridFiniteAutomaton& star(void);
    
    hybridFiniteAutomaton operator& (const hybridFiniteAutomaton&) const;
    hybridFiniteAutomaton operator| (const hybridFiniteAutomaton&) const;
    hybridFiniteAutomaton operator/ (const hybridFiniteAutomaton&) const;
    hybridFiniteAutomaton operator% (const hybridFiniteAutomaton&) const;
    hybridFiniteAutomaton operator! (void) const;
    hybridFiniteAutomaton operator* (void) const;
    hybridFiniteAutomaton operator+ (void) const;
    bool operator== (const hybridFiniteAutomaton&) const;
    bool operator!= (const hybridFiniteAutomaton&) const;
    
    static bool isomorphic(const hybridFiniteAutomaton&, const hybridFiniteAutomaton&);
    
    void minimizeBrzozowski(void);
    void minimizeHopcroft(void);
    
  private:        
    typedef map<automatonState*, automatonState*, ptr_id_less<automatonState*> > state_to_state_map;
    
    automatonState* makeDeterministic(list<ptr_state_set_t >&, map< ptr_state_set_t, automatonState*>&, ptr_state_set_t&);
    void addStopped();
    BDDFunction encode(unsigned long, const BDDFunctionContainer&) const;
    
    static bool checkStateEqual(state_to_state_map& equal_states, automatonState* act1, automatonState* act2, const hybridFiniteAutomaton& aut1, const hybridFiniteAutomaton& aut2);    
    
    hybridFiniteAutomaton& cup(const hybridFiniteAutomaton&);
    hybridFiniteAutomaton& cap(const hybridFiniteAutomaton&);
    hybridFiniteAutomaton& concat(const hybridFiniteAutomaton&);
    
    typedef shared_ptr_id_set<automatonState::ptr>::type shared_ptr_state_set_t;
    shared_ptr_state_set_t allStates;
    
    ptr_state_set_t startingStates;
    ptr_state_set_t acceptingStates;
    bool containsStopp;
  };
  typedef hybridFiniteAutomaton::ptr hybridFiniteAutomaton_ptr;
}

#endif
