/***************************************************************************
 *   Copyright (C) 2008-2009 by Sascha Klueppelholz                        *
 *   klueppel@tcs.inf.tu-dresden.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#define USE_NEW_ASL_PARSER

#include "mcreo.hpp"
#include "vereofy/modelcheck/specifications/asl/ASL_Formula.hpp"

#include "vereofy/common/streamlistener.hpp"
#include "vereofy/common/stopwatch.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/system/symb_automaton/symb_automaton.hpp"
#include "vereofy/system/symb_automaton/symb_component_instance.hpp"
#include "vereofy/modelcheck/checkers/branching_time_check/ASLStrategyAutomaton.hpp"

#include "mcreo_functions.hpp"

#include <boost/lexical_cast.hpp>
#include <iostream>
#include <stack>

/**
 @author Sascha Klueppelholz
 */

using namespace std;
using namespace vereofy;

// using namespace vereofy;

//********************************** public **********************************

/**
 * \brief Standard constructor
 *
 * It creates a new reo model checker.
 * If no progresslistener is passed the progress will be writtenn on cout using a new streamListener instance.
 *
 * \sa clear() progresslistener.hpp streamlistener.hpp
 */
mcreo::mcreo(const OptionContainer& options, const streamListener::ptr l){
  debugOut(1,"MCREO is alive...\n");
  clear();
  
  listener=l;	
  
  opt_witness_level=options.getOption("btsl-witness-level").getUintValue();
  opt_asl_fairness=false;   // options.getOption("asl-fairness").getBoolValue();
  opt_precompute_reachable=options.getOption("precompute-reachable").getBoolValue();
  opt_keep_sat=options.getOption("keep-sat").getBoolValue();
  opt_compute_strategy=options.getOption("compute-asl-strategy").getBoolValue();
}

/**
 * \brief Standard destructor
 *
 * Reset all members by calling clear
 * If a new listener was instanciated by by this model checker it is going to be deleted here.
 *
 * \sa clear() progresslistener.hpp, streamlistener.hpp
 */
mcreo::~mcreo(){
  clear();
}

/**
 * \brief The findOrAddRequirements method
 *
 * This method receives a string variable representing a ASL/BTSL formula.
 * It ether increases the reference counter for existing (sub)formulas or creates a new one.
 * Afterwords requirement and formula are going to be linked via an entry in the requirements table
 */
bool mcreo::findOrAddRequirement(const string& line){
  return (bool)(formulas[trim(line)]=formula::findOrAdd(line));
}

/**
 * \brief The analyse method
 *
 * This method receives and automaton which is model checked agains all
 * formulas added via findOrAddRequirement(). The result of the model checking
 * is a boolean and will be returned.
 * The second parameter determines if the reachable part of the automaton should
 * be precomputed before the model checking begins.
 * This method uses the private function sat(const formula::ptr) to compute 
 * the satisfaction sets for all (sub)formulas recursively.
 *
 * \sa clearComputation() sat(const formula::ptr)
 */
ASLModelCheckingResult::ptr mcreo::analyse(const SymbAutomaton::ptr automaton){
  // is the automaton ok?
  if (!automaton) throw "automaton seems to be faulty!";
  
  // clear previous computations
  clearComputation();
  
  // copy the automaton
  ca=automaton;
  
  // compute reachable subsystem if flag was set
  if (opt_precompute_reachable) reachable=ca->calcReachableStates();
  
  // initialize computation
  progress=0;
  bool ret=true;
  bool currentIsSatisfied = false;
  string spec;
  
  SymbComponentInstance::ptr system=boost::dynamic_pointer_cast<SymbComponentInstance>(automaton);
  CHECK_SUCCESS(system);
  
  ASLModelCheckingResult::ptr result(new ASLModelCheckingResult(system));
  BDDFunction localSatSet = BDDFunction::ONE();
  result->setSatSet(localSatSet);
  stringstream out;
  
  map<string, formula::ptr>::iterator fit=formulas.begin();
  // model checking automaton against each formula
  while (fit!=formulas.end()){
    spec=fit->second->getSpecification();
    StopWatch stopwatch;
    stopwatch.start();
    preSteps=0;
    
    // model checking problem: M \models Phi <=> Init\Sat(Phi) = \emptyset
    localSatSet = sat(fit->second);
    currentIsSatisfied = (ca->getInitialStates()&!localSatSet)==BDDFunction::ZERO();
    
    if (currentIsSatisfied){
      access_weak_ptr(listener)->updateProgress(0,string("system meets: ") + spec);
    } else{
      access_weak_ptr(listener)->updateProgress(0,string("system does NOT ") + "meet: " + spec);
    }
    
    // combine results
    result->setSatSet(result->getSatSet()&localSatSet); 
    ret &= currentIsSatisfied;
    
    // provide time information to the user
    stopwatch.stop();
    debugOut(1,"Time: " + stopwatch.toString()+"\n");
    
    // store witness/counterexample for btsl
    if (opt_witness_level>0 && fit->second) if (!fit->second->isASL){
      // -1 = no parent
      fit->second->generateWitness(ca->getInitialStates(), opt_witness_level, -1, *result);
      debugOut(2, out.str());
    }
    
    // and now for asl
    if (fit->second) if (fit->second->isASL) if (opt_compute_strategy){
      string objective1 = fit->second->sub1->getParsableSpecification();
      string objective2 = "";
      string objective  = "";
      bool haveStrategy = false;
      
      switch (fit->second->getOpType()){
          // E{N}Phi -> witness for: A Phi
        case EX:{
          objective = "AX[" + objective1 + "]";
          haveStrategy = currentIsSatisfied;
          break;
        }
          
        case EF:{
          objective = "AF[" + objective1 + "]";
          haveStrategy = currentIsSatisfied;
          break;
        }
          
        case EU:{
          objective2 = fit->second->sub2->getParsableSpecification();
          objective = "AU[" + objective1 + "," + objective2 + "]";
          haveStrategy = currentIsSatisfied;
          break;
        }
          
        case EG:{
          objective = "AG[" + objective1 + "]";
          haveStrategy = currentIsSatisfied;
          break;
        }
          
        case ER:{
          objective2 = fit->second->sub2->getParsableSpecification();
          objective = "AR[" + objective1 + "," + objective2 + "]";            
          haveStrategy = currentIsSatisfied;
          break;
        }
          
        case EES:{
          objective = "A<" + fit->second->getRegExpression() + ">" + objective1;
          haveStrategy = currentIsSatisfied;
          break;
        }
          
        case EAS:{
          objective = "A[" + fit->second->getRegExpression() + "]" + objective1;
          haveStrategy = currentIsSatisfied;
          break;
        }
          
          // A{N}Phi -> counterexample for !E Phi 
        case AX:{
          objective = "!EX[" + objective1 + "]";            
          haveStrategy = !currentIsSatisfied;
          break;
        }
          
        case AF:{
          objective = "!EF[" + objective1 + "]";            
          haveStrategy = !currentIsSatisfied;
          break;
        }
          
        case AU:{
          objective2 = fit->second->sub2->getParsableSpecification();
          objective = "!EU[" + objective1 + "," + objective2 + "]";
          haveStrategy = !currentIsSatisfied;
          break;
        }
          
        case AG:{
          objective = "!EG[" + objective1 + "]";            
          haveStrategy = !currentIsSatisfied;
          break;
        }
          
        case AR:{
          objective2 = fit->second->sub2->getParsableSpecification();
          objective = "!ER[" + objective1 + "," + objective2 + "]";
          haveStrategy = !currentIsSatisfied;
          break;
        }
          
        case AES:{
          objective = "!E<" + fit->second->getRegExpression() + ">" + objective1;
          haveStrategy = !currentIsSatisfied;
          break;
        }
          
        case AAS:{
          objective = "!E[" + fit->second->getRegExpression() + "]" + objective1;
          haveStrategy = !currentIsSatisfied;
          break;
        }
          
        default: haveStrategy = false;
      }
      
      if (haveStrategy){
        CHECK_SUCCESS(!result->hasStrategy());
        
        SymbAutomaton::ptr strategyAutomaton;
        if (fit->second->isBTSL){
          strategyAutomaton = SymbAutomaton::ptr(new ASLStrategyAutomaton(automaton, fit->second->regExpr, strategy.exists(copies)));
        }else{
          strategyAutomaton = SymbAutomaton::ptr(new ASLStrategyAutomaton(automaton, strategy.exists(copies)));
        }
        
        result->setStrategy(strategyAutomaton);
        result->setStrategyObjective(objective);          
      }
    }
    
    // check the next requirement
    ++fit;
  }
  
  // print status and return the overall result
  getStatus();
  
  // clear result
  if (!opt_keep_sat) clearComputation();
  
  // the overall result is the bool + counterexample/witness
  result->setCheckPassed(ret);
  result->setAdditionalInfo(out.str());
  
  return result;
}

/**
 * \brief The getStatus method
 *
 * Privides some information about the current status to the progress listener.
 */
void mcreo::getStatus() const{
  string status="status: ";
  status+=boost::lexical_cast<string>(formulas.size()) + " formula(s), ";
  status+=boost::lexical_cast<string>(formula::allFormulas->size()) + " subformula(s), ";
  status+=boost::lexical_cast<string>(BDD::size()) + " BDD nodes";
#ifdef JINC_NEW
  status+=".";
#else
  status+=" and peak number of ";
  status+=boost::lexical_cast<string>(BDD::PEAK()) + " BDD nodes.";
#endif
  access_weak_ptr(listener)->updateProgress(0,status);
}

/**
 * \brief clearRequirement removes all requirements from memory
 *
 * This method removes all requirements from mcreo.
 *
 * \sa clearComputation() clear()
 */
void mcreo::clearRequirements(){
  // clear container
  formulas.clear();
}

//********************************** private **********************************

/**
 * \brief The core functionality for model checking
 *
 * This is the core method for model checking.
 * It receives a pointer to a formula and computes the set of states satiesfying the given formula.
 *
 * \sa analyse(const Automata&, bool=false)
 */
BDDFunction mcreo::sat(formula::ptr f){
  // the result for constant formulas true, false is already known
  // this is also the case when the satisfaction set was computed before 
  // and stored to the member variable of the formula 
  if (f->satSet.valid){
    if (f->formulaType==ONE || f->formulaType==ZERO) 
      if (!f->done) f->done=access_weak_ptr(listener)->updateProgress(++progress/formula::allFormulas->size(), "checked: " + f->specAsString);
    return f->satSet.value;
  }
  
  // store variable sets
  BDDFunction _allTransitions = allTransitions;
  BDDFunction _initialStates  = initialStates;
  BDDFunction _copies         = copies;
  BDDFunction _reachable      = reachable;
  BDDFunction _controllable   = controllable;
  
  // update variable sets
  allTransitions = ca->getDelta();
  initialStates  = ca->getInitialStates();
  copies         = ca->getOriginalStateVars().moveDown();
  ports          = ca->getPortVars();
  label_vars     = ca->getTransitionVars();
  
  // this is the magic function resolving all cases and prepares the reachability analysis
  SymbAutomatonLanguageAcceptor_ptr languageAcceptor = doWhatYouHaveToDo(f);
  
  // initialize computation
  ReturnValue<BDDFunction> result;
  BDDFunction satPhi;
  
  // laststep determines if a top level formula is computetd and no other requirement needs to be checked
  // if this is the case early termination might improve the runtime
  bool laststep=(formulas.size()==1)&&(f==formulas.begin()->second);
  if (laststep) debugOut(1,"early termination is enabled for " + f->specAsString + "\n");
  
  // depending on the formula type switch to the right computation method
  switch (f->formulaType){
    case EX:{
      satPhi=sat(f->sub1);
      pair <BDDFunction,BDDFunction> image = preImage(satPhi,f);
      result.value = image.first;
      
      // compute strategy
      if (opt_compute_strategy){
        strategy = image.second;
      }
      
      // store information for a possible witness
      if (opt_witness_level>0){
        f->createNewWitnessFor(f,ca);
        f->appendToWitness(satPhi);
        f->appendToWitness(result.value);
      }
      
      result.valid=true;
      break;
    }
      
    case EF:{
      result.value=lfp(sat(f->sub1),BDDFunction::ONE(),f,false,laststep);
      result.valid=true;
      break;
    }
      
    case EG:{
      result.value = gfp(sat(f->sub1),f,false,laststep);
      result.valid = true;
      break;
    }
      
    case EU:{
      result.value=lfp(sat(f->sub2),sat(f->sub1),f,false,laststep);
      result.valid=true;
      break;
    }
      
    case ER:{
      // E[Phi1 R Phi2] = E[Phi2 U (Phi1&Phi2)] | EG[Phi2])
      result.value = lfp(sat(f->sub1)&sat(f->sub2),sat(f->sub2),f,false,laststep);
      
      if (laststep && (result.value&initialStates)!=BDDFunction::ZERO()){
        result.valid=true;
        break;
      } else {
        result.value|=gfp(sat(f->sub2), f, false, laststep);
        result.valid=true;
      }
      break;
    }
      
    case AX:{
      // AX[Phi]=!EX[!Phi]: compute one image step
      satPhi=sat(f->sub1);
      
      pair <BDDFunction,BDDFunction> image = preImage(!satPhi,f);
      result.value = !image.first;
      
      // compute strategy
      if (opt_compute_strategy){
        strategy = image.second;
      }
      
      // store information for a possible counterexample
      if (opt_witness_level>0){
        f->createNewWitnessFor(f,ca);
        f->appendToWitness(!satPhi);
        f->appendToWitness(!result.value);
      }
      
      result.valid=true;
      break;
    }
      
    case AF:{
      // AF[Phi] = !EG[!Phi]
      result.value = !gfp(!sat(f->sub1),f,true,laststep);
      result.valid = true;
      break;
    }
      
    case AG:{
      // AG[Phi]=!EF[!Phi]: calculate least fixpoint
      result.value=!lfp(!sat(f->sub1),BDDFunction::ONE(),f,true,laststep);
      result.valid=true;
      break;
    }
      
    case AU:{
      // A[Phi1 U Phi2] = (!E[!Phi2 U (!Phi1&!Phi2)] & !EG[!Phi2])
      result.value = !lfp(!sat(f->sub1)&!sat(f->sub2),!sat(f->sub2),f,true,laststep);
      
      if (laststep && (result.value&initialStates)==BDDFunction::ZERO()){
        result.valid=true;
        break;
      } else {
        result.value&=!gfp(!sat(f->sub2), f, true, laststep);
        result.valid=true;
      }
      break;
    }
      
    case AR:{
      // A[Phi1 R Phi2] = !E[!Phi1 U !Phi2]: compute least fixpoint
      result.value=!lfp(!sat(f->sub2),!sat(f->sub1),f,true,laststep);
      result.valid=true;
      break;
    }
      
    case CONJ:{
      // Phi1&Phi2 (heuristic: chosing the shorter formula representation)
      bool choice=(f->sub1->getSpecification().length()<f->sub2->getSpecification().length());
      result.value=choice?sat(f->sub1):sat(f->sub2);
      if (result.value!=BDDFunction::ZERO()) result.value&=choice?sat(f->sub2):sat(f->sub1);
      
      // store information for a possible witness
      if (opt_witness_level>0){
        f->createNewWitnessFor(f,ca);
        f->appendToWitness(result.value);
      }
      
      result.valid=true;
      break;
    }
      
    case DISJ:{
      // Phi1|Phi2
      // heuristic: chosing the shorter formula representation
      bool choice=(f->sub1->specAsString.length()<f->sub2->specAsString.length());
      result.value=choice?sat(f->sub1):sat(f->sub2);
      if (result.value!=BDDFunction::ONE()) result.value|=choice?sat(f->sub2):sat(f->sub1);
      
      // store information for a possible witness
      if (opt_witness_level>0){
        f->createNewWitnessFor(f,ca);
        f->appendToWitness(result.value);
      }
      
      result.valid=true;
      break;
    }
      
    case EXOR:{
      // Phi1^Phi2
      result.value=(sat(f->sub1)^sat(f->sub2)).constrain(reachable);
      
      // store information for a possible witness
      if (opt_witness_level>0){
        f->createNewWitnessFor(f,ca);
        f->appendToWitness(result.value);
      }
      
      result.valid=true;
      break;
    }
      
    case NEG:{
      // !Phi
      result.value=!sat(f->sub1);
      
      // store information for a possible witness
      if (opt_witness_level && !f->sub1->isASL){
        // adopt witness of subformula
        f->wit_least=f->sub1->wit_least;
        
        if (f->wit_least){
          f->wit_least->representedFormula = f;
          
          // swap WITNESS <-> COUNTEREXAMPLE
          if (f->wit_least->whatAmI.myType == WITNESS) {
            f->wit_least->whatAmI.myType = COUNTEREXAMPLE;
          } else if (f->wit_least->whatAmI.myType == COUNTEREXAMPLE){
            f->wit_least->whatAmI.myType = WITNESS;
          }
        }
      }
      
      result.valid=true;
      break;
    }
      
      // E<alpha>Phi
    case EES:{
      result.value = lfp((languageAcceptor->getAccept()&sat(f->sub1)),BDDFunction::ONE(),f,false,laststep);
      result.value = (result.value & languageAcceptor->getInitialStates()).exists(languageAcceptor->getOriginalStateVars());
      result.valid = true;
      break;
    }
      
      // E[alpha]Phi
    case EAS:{
      result.value = gfp((!languageAcceptor->getAccept()|sat(f->sub1)),f,false,laststep);
      result.value = (result.value & languageAcceptor->getInitialStates()).exists(languageAcceptor->getOriginalStateVars());	
      result.valid = true;
      break;
    }
      
      // A<alpha>Phi = !E[alpha]!Phi
    case AES:{
      result.value = gfp((!languageAcceptor->getAccept()|!sat(f->sub1)),f,true,laststep);
      result.value = (result.value & languageAcceptor->getInitialStates()).exists(languageAcceptor->getOriginalStateVars());	
      result.value = !result.value;
      result.valid = true;
      break;
    }
      
      // A[alpha]Phi = !E<alpha>!Phi
    case AAS:{
      result.value = lfp((languageAcceptor->getAccept()&!sat(f->sub1)),BDDFunction::ONE(),f,true,laststep);
      result.value = (result.value & languageAcceptor->getInitialStates()).exists(languageAcceptor->getOriginalStateVars());
      result.value = !result.value;
      result.valid = true;
      break;
    }
      
    case ATOM:{
      // try fetching the symbol from the automaton
      vereofy::SymbComponentInstance::ptr ci_ca=boost::dynamic_pointer_cast<vereofy::SymbComponentInstance>(ca);
      if (!ci_ca) throw "Don't know how to get BDDFunction for this atom!";
      
      try{result.value=ci_ca->parse_state_guard(f->specAsString);}
      catch(std::exception& e){
        cout << e.what() << "\n";
        throw "Can not resolve atomic proposition: " + f->specAsString;
      }
      
      // create new counterexample for current formula and system if needed	
      if (opt_witness_level>0){
        f->createNewWitnessFor(f,ca);
        f->appendToWitness(result.value);
      }
      
      result.valid=true;
      break;
    }
      
    case ONE: case ZERO:{
      // constant values were stored already
      result=f->satSet;
      
      // create new counterexample for current formula and system if needed
      if (opt_witness_level>0){
        f->createNewWitnessFor(f,ca);
        f->appendToWitness(result.value);
      }
      break;
    }
      
    case INVALID:{
      throw "Implementation error.";
    }      
  }
  
  // computed table stores the result of sat(formula)
  // the result is constrained to the reachable part of the system
  // if reachable was not precomputed this will not change anything
  // because of initializing reachable to the one BDDFunction
  f->satSet.value=result.value.constrain(reachable);
  f->satSet.valid=true;
  f->done=true;
  
  // reset variable sets from previous recursion step
  allTransitions = _allTransitions;
  initialStates  = _initialStates;
  copies         = _copies;
  reachable      = _reachable;
  
  // report the listener and continue work if the listener agrees
  return (access_weak_ptr(listener)->updateProgress(++progress/formula::allFormulas->size(), "\nchecked: " + f->specAsString))?(result.value):BDDFunction::ZERO();
}

/**
 * \brief preImage step method
 *
 * This inline method is used to perform one backward step in transition system. 
 * This is elementary for the model checking procedure (CTL and ATL). 
 * BTSL is handled in \sat(formula::ptr) directly.
 * The parameter stands for a set of states from which to start (and the current subformula (optional)).
 * The transition relation comes from the variable ca->delta.
 *
 * \sa \sat(formula::ptr)
*/
pair <BDDFunction, BDDFunction> mcreo::preImage(const BDDFunction& p, const formula::ptr f){
  // show the number of current step
  debugOut(1,"\r step " + boost::lexical_cast<string>(preSteps) + ", " + boost::lexical_cast<string>(BDD::size()) + " bdd nodes");

  pair <BDDFunction,BDDFunction> result;
  
  // increase step counter
  ++preSteps;
  
  // first the ASL case:
  if (f->isASL){
    BDDFunction local_strategy = (controllable & p.moveDown()).exists(copies) & !(controllable & !p.moveDown().exists(copies));
    BDDFunction result_states  = !((unpreventable & !p.moveDown()).exists(label_vars&copies)) & (local_strategy).exists(label_vars);
    
    // store result
    result.first = result_states.constrain(reachable);
    
    // add strategy
    if (opt_compute_strategy){
      result.second = allTransitions & result_states & local_strategy;
    }
  } else{ 
    // else btsl
    result.first = (p.moveDown()&allTransitions).exists(copies).constrain(reachable);
  }
  
  return result; 
}

/**
 * \brief Clear resets mcreo completely
 *
 * This inline method removes all components, requirements and symbols from mcreo.
 * It also clears all other member variables and the bdd.
 *
 * \sa clearComputation() clearRequirements()
 */
void mcreo::clear(){
  debugOut(1,"clearing...");
  if (formula::allFormulas==NULL) formula::allFormulas=new formula_map;
  
  initialStates = BDDFunction::ZERO();
  
  copies     = BDDFunction::ZERO();
  ports      = BDDFunction::ZERO();
  label_vars = BDDFunction::ZERO();
	
  controllable   = BDDFunction::ZERO();
  allTransitions = BDDFunction::ZERO();
  unpreventable  = BDDFunction::ZERO();
  reachable      = BDDFunction::ONE();  
  
  strategy = BDDFunction::ZERO();
	
  clearRequirements();
  debugOut(1,"done!\n");
}

/**
 * \brief clearComputation clears the the result for each formula and the reachable computation  
 *
 * After computing the satisfaction set of a formula the result is stored within a member variable of the formula class.
 * These results may be cleared to e.g. redo the computation. This method calles formula::clearComputation() for each formula.
 * The second thing this BDDFunction takes care of is the computation of the reachable part of a component.
 *
 * \sa formula::clearComputation() component::clearReachable() clear() clearRequirements()
 */
void mcreo::clearComputation(){
  // clear sat sets of all requirements
  debugOut(1,"clearing computation...");
  map<string, formula::ptr>::iterator it=formulas.begin();
  while (it!=formulas.end()){
    it->second->clearComputation();
    ++it;
  }
  
  // reset computation of reachable part
  reachable=BDDFunction::ONE();
  debugOut(1,"done!\n");
}

inline BDDFunction mcreo::BTSLProduct(const BDDFunction& sys, const SymbAutomatonLanguageAcceptor_ptr lA){
  if (!lA->hasStoppedTransitions()) return (sys & lA->getDelta());
  
  BDDFunction stopped_a = BDDFunction::projection("stopped_a");
  BDDFunction stopped_o = BDDFunction::projection("stopped_o");
  BDDFunction stopped_t = BDDFunction::projection("stopped_t");
  BDDFunction stopped = stopped_a | stopped_o | stopped_t;
  
  // first non - stopp
  BDDFunction result = (sys & (lA->getDelta() & !stopped));
  
  // now add terminal
  BDDFunction terminalStates=!(sys.exists(copies&label_vars));
  result|= terminalStates & ca->getHaltSystem() & ports.change() & (lA->getDelta() & stopped_t & !stopped_o & !stopped_a);
  
  // change port set afterwards
  ports      &= stopped_a & stopped_o & stopped_t;
  label_vars &= stopped_a & stopped_o & stopped_t;
  
  return result;
}

inline BDDFunction mcreo::ASLProduct(const BDDFunction& sys, const SymbAutomatonLanguageAcceptor_ptr lA, const BDDFunction& cP){
  if (!lA->hasStoppedTransitions()) return (sys & lA->getDelta());
  
  BDDFunction stopped_a = BDDFunction::projection("stopped_a");
  BDDFunction stopped_o = BDDFunction::projection("stopped_o");
  BDDFunction stopped_t = BDDFunction::projection("stopped_t");
  BDDFunction stopped = stopped_a | stopped_o | stopped_t;
  
  // first non - stopp
  BDDFunction result = (sys & (lA->getDelta() & !stopped));
  
  // now add terminal
  BDDFunction terminalStates=!(sys.exists(copies&label_vars));
  result|= terminalStates & ca->getHaltSystem() & ports.change() & (lA->getDelta() & stopped_t & !stopped_o & !stopped_a);
  
  // now quiescent
  BDDFunction quiescentStates = !(sys.exists(copies) & ports.change()).exists(label_vars);
  
  // q1: first product rule
  BDDFunction rule43States = quiescentStates & !(sys.exists(copies) & ports.cofactor(cP).change()).exists(label_vars);
  result|= rule43States & ca->getHaltSystem() & ports.change() & (lA->getDelta() & stopped_o & !stopped_a & !stopped_t);
  
  // q2: second product rule
  BDDFunction rule44States = quiescentStates & (sys.exists(copies) & (!cP.change())).exists(label_vars);
  result|= rule44States & ca->getHaltSystem() & ports.change() & (lA->getDelta() & stopped_a & !stopped_t & !stopped_o);
  
  // change port set afterwards
  ports      &= stopped_a & stopped_o & stopped_t;
  label_vars &= stopped_a & stopped_o & stopped_t;
  
  return result;
}

SymbAutomatonLanguageAcceptor::ptr mcreo::doWhatYouHaveToDo(const formula::ptr f){  
  SymbAutomatonLanguageAcceptor_ptr languageAcceptor;
  
  // for ASL: find out which ports are controllable and which are not
  BDDFunction my_ports;
  BDDFunction opponent_ports;
  
  if (f->isASL){
    vereofy::SymbComponentInstance::ptr sci=boost::dynamic_pointer_cast<vereofy::SymbComponentInstance>(ca);
    CHECK_SUCCESS(sci);
    
    vereofy::ComponentInstance::ptr ci=sci->info->component_instance;
    CHECK_SUCCESS(ci);
    
    set<string>::iterator cps=f->controllablePorts.begin();
    PortDescription::ptr_set controllable_pds;
    try{
      while (cps!=f->controllablePorts.end()){
        PortID port_id = ci->lookupPort(*cps, false);
        CHECK_SUCCESS(ci.get() == &port_id.getComponentInstance());
        
        controllable_pds.insert(port_id.getPortDescription().shared_from_this());
        ++cps;
      }
    } catch(std::exception& e){ throw "Unknown Port: " + *cps; }
    
    pair<BDDFunction, BDDFunction> activity = sci->toplevel_port_activity(controllable_pds);
    my_ports       = activity.first;
    opponent_ports = activity.second;
  }
  
  // for btsl formulas build the (nondeterministic) automaton with respect to the current system
  if (f->isBTSL){
    bool det = f->isASL || (f->formulaType == EAS || f->formulaType == AES);
    
    // building symbolic representation of automaton for the regular expression
#if defined(USE_NEW_ASL_PARSER)
    languageAcceptor = f->composeEA(ca,det);
#else
    languageAcceptor = f->buildEA(ca,det);
#endif
    
    // building product
    if (f->isASL){
      // build ASL product
      allTransitions=ASLProduct(allTransitions,languageAcceptor,my_ports.exists(opponent_ports));
      
      if (languageAcceptor->hasStoppedTransitions()){
        my_ports       &= BDDFunction::projection("stopped_a");
        opponent_ports &= BDDFunction::projection("stopped_o");
      }
      
    }else{
      // build BTSL product
      allTransitions=BTSLProduct(allTransitions,languageAcceptor);
    }
    
    // unite copy sets
    copies&=languageAcceptor->getOriginalStateVars().moveDown();
    
    // update initial states
    initialStates&=languageAcceptor->getInitialStates();
  }
  
  if (f->isASL){
    // split delta into controllable and unpreventable transitions for ASL preImage
    controllable  = allTransitions & cubeSetNone(opponent_ports);
    unpreventable = allTransitions & cubeSetNone(my_ports);
  }else{
    // after building the product there is no need to reason about data-flow in case of BTSL    
    allTransitions=allTransitions.exists(label_vars).constrain(reachable);
  }
  
  return languageAcceptor;
}

BDDFunction mcreo::gfp(const BDDFunction& start, const formula::ptr f, const bool negate, const bool laststep){
  BDDFunction satSet = start;
  BDDFunction visited=BDDFunction::ZERO();
  BDDFunction terminalStates = !(allTransitions.exists(copies&label_vars));
  BDDFunction exceptionalStates = start & terminalStates;
  
  while (visited!=satSet){
    visited=satSet;
    if (f->isASL) exceptionalStates = (start & (terminalStates | !(unpreventable & !satSet.moveDown()).exists(copies&label_vars)));
    satSet&=((preImage(satSet,f).first)|exceptionalStates).constrain(reachable);
    
    // early termination of image computation
    if (laststep){
      if (!negate){
        if ((satSet&initialStates).exists(copies)==BDDFunction::ZERO()) visited=satSet;
      } else if ((!satSet&initialStates).exists(copies)!=BDDFunction::ZERO()) visited=satSet;
    }
  }
  
  // store information for a possible witness
  if (opt_witness_level>0){
    f->createSecondWitnessFor(f,ca,GREATEST,satSet);
  }
  
  // store strategy to stay in satSet
  if (opt_compute_strategy){
    strategy = (unpreventable | allTransitions & satSet & (allTransitions & satSet.moveDown()).exists(copies) & !(allTransitions & !satSet.moveDown()));
  }
  
  return satSet.constrain(reachable);    
}

BDDFunction mcreo::lfp(const BDDFunction& start, const BDDFunction& invariance, const formula::ptr f, const bool negated, const bool laststep){
  BDDFunction satSet  = start;
  BDDFunction visited = BDDFunction::ZERO();
  pair <BDDFunction,BDDFunction> image;
  
  // initiate strategy
  if(opt_compute_strategy){
    strategy = unpreventable;
  }
  
  // create new witness
  if (opt_witness_level>0){
    f->createNewWitnessFor(f,ca);
    f->appendToWitness(start);
  }  
  
  while (visited!=satSet){
    visited=satSet;
    image = preImage(satSet,f);
    
    satSet   |= (image.first & invariance).constrain(reachable);
    
    // compute strategy
    if (opt_compute_strategy){
      strategy |= !(strategy.exists(label_vars&copies)) & image.second;
    }
    
    // store information for a possible witness
    if (opt_witness_level>0){
      BDDFunction shell=satSet&!visited;
      if (shell!=BDDFunction::ZERO()) f->appendToWitness(shell);
    }
    
    // early termination of image computation
    if (laststep){
      if (negated){
        if ((!satSet&initialStates)==BDDFunction::ZERO()) visited=satSet;
      } else if (!negated) if ((satSet&initialStates)!=BDDFunction::ZERO()) visited=satSet;
    }
  }
  
  return satSet.constrain(reachable);    
}
