/***************************************************************************
 *   Copyright (C) 2008-2009 by Ilham Kurnia, Sascha Klueppelholz          *  
 *   s4036591@mail.inf.tu-dresden.de  , klueppel@tcs.inf.tu-dresden.de     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "witness.hpp"
#include "mcreo_functions.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/system/symb_automaton/symb_component_instance.hpp"

#include <iostream>
#include <boost/lexical_cast.hpp>

using namespace std;
using namespace vereofy;

/**
 @author Ilham Kurnia
 @author Sascha Klueppelholz
 */

/**
 * \brief witness standard constructor
 * 
 * Creating a new instance
 */
witness::witness(formula::ptr f, SymbAutomaton::ptr a) : 
SymbAutomaton(*a) { // clone from automaton a
  // store refs to formula and automaton
  representedFormula = f;
  ca = a;
  
  // copy relevant information
  if (f->isBTSL){
    csOriginalX    &= f->regExpr->getOriginalStateVars();
    csCopyY        = csOriginalX.moveDown();
    csCopyCopyZ    = csCopyY.moveDown();
    initialStates  &= f->regExpr->getInitialStates();
    delta          &= f->regExpr->getDelta();
    terminalStates = calcTerminalStates();
    terminalLoop   = (terminalStates & access_weak_ptr(ca)->getDelta() & terminalStates.exists(access_weak_ptr(ca)->getOriginalStateVars()).moveDown());
  } else{
    terminalStates = calcTerminalStates();
    terminalLoop   = BDDFunction::ZERO();
  }
  
  // decide whether it is a counter example or witness
  switch (f->formulaType){
    case ATOM: case EX: case EF: case EU: case ER: case EG: case EES: case EAS:{
      whatAmI.myType = WITNESS;
      break;
    }
		  
    case AX: case AF: case AU: case AR: case AG: case AES: case AAS:{
      whatAmI.myType = COUNTEREXAMPLE;
      break;  
    }
      
    case NEG: case CONJ: case DISJ: case EXOR: case ONE: case ZERO:{
      whatAmI.myType = DEPENDS;
      break;
    }
      
    case INVALID:{
      throw "Implementation error.";
    }
  }
  
  // default: only allow transitions in the product automaton
  relevantStateVars  = csOriginalX;
  
  // decide whether to compute a least, greatest or both fixpoints
  switch (f->formulaType){
    case EU: case EF: case AR: case AG: case EES: case AAS:{
      whatAmI.myFixpointType = LEAST;
      break;
    }
		  
    case AF: case EG: case AES: case EAS:{
      whatAmI.myFixpointType = GREATEST;
      // allow transitions in the system automaton
      relevantStateVars = a->getOriginalStateVars();
      break;
    }
      
      // EG and AG can have greatest and least fixpoints - first try will be LEAST
      // the model checker resolves this later
    case ER: case AU:{
      whatAmI.myFixpointType = LEAST;
      break;  
    }
      
    case EX: case AX: case NEG: case ATOM: case CONJ: case DISJ: case EXOR: case ONE: case ZERO:{
      whatAmI.myFixpointType = NONE;
      break;  
    }
      
    case INVALID:{
      throw "Implementation error.";
    }      
  }
}

/**
 * \brief The append method for witnesses and counterexamples
 *
 * This method appends a set of states to the existing result.
 * The function has been added later
 */
void witness::append(const BDDFunction& f){
  result.push(f);
}

void witness::setFixpointType(const f_info& ft){
  whatAmI.myFixpointType = ft;
}

/**
 * \brief The costumized forwardSet method
 *
 * This method calculates the set of states that are reachable from any of the state in initStates.
 * The reachability analysis is done based on the transitions given in delta. 
 */
BDDFunction witness::forwardSet(const BDDFunction& initStates, const BDDFunction& restriction) const{
  BDDFunction reachable = initStates&restriction;
  BDDFunction visited=BDDFunction::ZERO();
  
  while (visited!=reachable){
    visited=reachable;
    reachable|=postImage(reachable);
  }
  return reachable;
}

/**
 * \brief The costumized sccFind method
 *
 * This method returns a subset of the original constraint automaton transition relation where 
 * a relation from a state to another state is present in the subset iff that those two states
 * are of the same strongly connected component w.r.t to the transition relation. 
 * The SCC is calculated using the linear symbolic step SCC finding algorithm described by 
 * Gentilini, Piazza and Policriti.
 */
BDDFunction witness::sccFind(const BDDFunction& states, const BDDFunction& spineSet, const BDDFunction& spineNode) const{
  // contract: sccStates and sccTrans are initialized before the first call!
  if (states == BDDFunction::ZERO()) return states;
  
  BDDFunction node=(spineSet==BDDFunction::ZERO())?(SymbState::pickState(states, csOriginalX)):(spineNode);
  
  // compute the skeleton 
  vector<BDDFunction> onionRings;
  BDDFunction level = node;
  BDDFunction skelForwardSet = BDDFunction::ZERO();
  
  // contract: level is never initially equal to BDDFunction::ZERO()
  while (level != BDDFunction::ZERO()){
    onionRings.push_back(level);
    skelForwardSet |= level;
    level = postImage(level) & states & !skelForwardSet;
  }
  
  ERROR_CHECK_ASSERT(onionRings.back() != BDDFunction::ZERO());
  SymbState skelSpineNode = SymbState::pickState(onionRings.back(),csOriginalX);
  SymbState origSkelSpineNode = skelSpineNode;
  BDDFunction skelSpineSet = skelSpineNode;
  
  for (int i=onionRings.size()-2;i>0;i--){
    CHECK(2,(preImage(skelSpineNode) & onionRings[i]) != BDDFunction::ZERO(),"");
    skelSpineNode = SymbState::pickState(preImage(skelSpineNode) & onionRings[i], csOriginalX);
    skelSpineSet |= skelSpineNode;
  }
  
  // determine the scc with the node spineNode.
  BDDFunction scc = node;
  BDDFunction tmp = preImage(scc) & skelForwardSet;
  while ((tmp & !scc) != BDDFunction::ZERO()){
    scc |= tmp;
    tmp = preImage(scc) & skelForwardSet;
  }
  
  return (sccFind(states & !(skelForwardSet), spineSet & !scc, preImage(scc & spineSet) & (spineSet & !scc))) |
  (sccFind(skelForwardSet & !scc, skelSpineSet & !scc, origSkelSpineNode)) | 
  (delta & scc & scc.moveDown());
}


std::pair<SymbTrace::ptr,ModelCheckingResultTrace::trace_type_t> witness::computePath(const BDDFunction& initStates) const{
  if (!result.size()) throw "Implementation error: Witness/ counterexample has not been computed during model checking.";
  if (whatAmI.myFixpointType==UNDETERMINED) throw "Implementation error: fixpoint type unresolved when computing witness/ couterexample path.";
  
  SymbTrace::ptr witnessPath(new SymbTrace());
  
  if (initStates==BDDFunction::ZERO()){
    debugOut(4, "there is none in this context.");
    return make_pair(witnessPath, ModelCheckingResultTrace::TRACE_NONE);
  }
  
  if (whatAmI.myFixpointType!=GREATEST){
    // working copy of result
    stack<BDDFunction> _result=result;
    
    // try to find inital states modelling the formula
    BDDFunction currentState = _result.top()&initStates;
    
    if (whatAmI.myFixpointType==LEAST){ 
      while (_result.size() && currentState==BDDFunction::ZERO()){
        currentState = _result.top()&initStates;
        _result.pop();
      }
    }
    
    if (currentState==BDDFunction::ZERO()){
      debugOut(4, "there is none in this context.");
      return make_pair(witnessPath, ModelCheckingResultTrace::TRACE_NONE);
    }
    
    // initial state has been found; its the first element of the witness path
    ERROR_CHECK_ASSERT(currentState != BDDFunction::ZERO());
    SymbState aState = SymbState::pickState(currentState, relevantStateVars);
    witnessPath->appendState(aState);
    if (_result.size()) _result.pop();
    
    // append the suffix of the path
    while (_result.size()){
      currentState=postImage(aState)&_result.top();
      
      // if not a stopp step (last step) - the next state should exist
      ERROR_CHECK_ASSERT(_result.size()==1 || currentState != BDDFunction::ZERO());
      
      if (currentState!=BDDFunction::ZERO()){
        SymbState nextState = SymbState::pickState(currentState, relevantStateVars);
        ERROR_CHECK_ASSERT(nextState.getX() != BDDFunction::ZERO());
        
        witnessPath->appendTransition(SymbTransition::pickTransition(delta, aState, nextState, getTransitionVars()));
        aState = nextState;
      }
      _result.pop();
    }
  }
  
  if (whatAmI.myFixpointType==GREATEST){
    if ((initStates&result.top())==BDDFunction::ZERO()){
      debugOut(4, "there is none in this context.");
      return make_pair(witnessPath, ModelCheckingResultTrace::TRACE_NONE);
    }
    
    BDDFunction reachable = forwardSet(initStates, result.top());
    BDDFunction v = reachable;
    BDDFunction trivialScc = BDDFunction::ZERO();
    BDDFunction tmp = preImage(v) & result.top();
    while (v != tmp){
      trivialScc |= v & !tmp;
      v = tmp;
      tmp = preImage(v) & result.top();
    }
		
    tmp = reachable & !trivialScc;
    
    BDDFunction sccTrans = sccFind(tmp, BDDFunction::ZERO(), BDDFunction::ZERO()) | terminalLoop;
    BDDFunction sccTransSteps = sccTrans.exists(getTransitionVars());
    BDDFunction sccStates = sccTransSteps.exists(csCopyY) | terminalStates;
    
    ERROR_CHECK_ASSERT(sccStates != BDDFunction::ZERO());
    
    vector<BDDFunction> rings;
    rings.push_back(sccStates & result.top());
    
    tmp = sccStates;
    while ((rings.back() & initialStates) == BDDFunction::ZERO()){
      rings.push_back(preImage(rings.back()) & result.top() &!tmp);
      tmp |= rings.back();
    }
    
    SymbState aState = SymbState::pickState(rings.back() & initialStates, relevantStateVars);
    witnessPath->appendState(aState);
    for (int i=rings.size()-2;i>=0;i--){
      SymbState nextState = SymbState::pickState(postImage(aState) & rings[i], relevantStateVars);
      witnessPath->appendTransition(SymbTransition::pickTransition(delta, aState, nextState, getTransitionVars()));
      aState = nextState;
    }
    
    tmp = aState;
    reachable = BDDFunction::ZERO();
    
    // try to either find a cycle or terminal state
    while ((tmp&reachable) == BDDFunction::ZERO()){
      reachable|=tmp;
      tmp = (tmp & sccTransSteps).exists(csOriginalX).moveUp();
      
      if (tmp != BDDFunction::ZERO()){
        SymbState nextState = SymbState::pickState(tmp, relevantStateVars);
        ERROR_CHECK_ASSERT(nextState.getX() != BDDFunction::ZERO());
        
        witnessPath->appendTransition(SymbTransition::pickTransition(sccTrans, aState, nextState, getTransitionVars()));
        aState = nextState;
      }else{tmp=reachable;}
    }	
  }	
  
  
  ModelCheckingResultTrace::trace_type_t trace_type;
  switch (whatAmI.myType) {
    case WITNESS:
      trace_type=ModelCheckingResultTrace::TRACE_WITNESS;
      break;
    case COUNTEREXAMPLE:
      trace_type=ModelCheckingResultTrace::TRACE_COUNTEREXAMPLE;
      break;
    default:
      trace_type=ModelCheckingResultTrace::TRACE_EXAMPLE;
      break;
  }
  
  return make_pair(witnessPath, trace_type);
}

void witness::generateWitness(const BDDFunction& f, const unsigned int rl, const int parent_id, ModelCheckingResult& mcresult) const{
  debugOut(4, "Calculating witness for " + access_weak_ptr(representedFormula)->specAsString+"\n");
  
  std::pair<SymbTrace::ptr,ModelCheckingResultTrace::trace_type_t> example=computePath(f);
  
  ModelCheckingResultTrace::ptr result_trace(new ModelCheckingResultTrace(
                                                                          example.second,
                                                                          access_weak_ptr(representedFormula)->specAsString,
                                                                          SymbTraceLasso::ptr(new SymbTraceLasso(example.first)),
                                                                          parent_id));
  
  unsigned int id=mcresult.addTrace(result_trace);
  
  BDDFunction lastState = (example.first->countStates())?example.first->getLastState():BDDFunction::ZERO();
  if ((rl-1)){
    if (access_weak_ptr(representedFormula)->sub1) access_weak_ptr(representedFormula)->sub1->generateWitness(lastState,rl-1,id,mcresult);
    if (access_weak_ptr(representedFormula)->sub2) access_weak_ptr(representedFormula)->sub2->generateWitness(lastState,rl-1,id,mcresult);
  }
}
