/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/



#ifndef BDD_NBA_HPP
#define BDD_NBA_HPP

#include "vereofy/setup.hpp"
#include "vereofy/common/errors.hpp"
#include "NBA_I.hpp"
#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>
#include <string>
#include <iostream>

namespace vereofy {namespace ltl_check {

class BDD_NBA : public NBA_I {
public:
  typedef boost::shared_ptr<BDD_NBA> ptr;

  typedef std::vector<BDDFunction> ap_to_bddfunction_t;

  BDD_NBA(APSet_cp apset_,
	  const ap_to_bddfunction_t& ap_to_bddfunction_=ap_to_bddfunction_t()) :
    apset(apset_),
    ap_to_bddfunction(ap_to_bddfunction_),

    states(0),
    var_count(0),
    initial(BDDFunction::ZERO()),
    accepting(BDDFunction::ZERO()),
    delta(BDDFunction::ZERO()),
    all_states(BDDFunction::ZERO()),
    state_vars(BDDFunction::ONE()),
    apset_vars(BDDFunction::ONE()) {
    
    // TODO: apset_vars
    //      apset_vars&=BDDFunction::projection(ap_var_index[ap]);

  }

  virtual ~BDD_NBA() {}

  /** 
   * Create a new state.
   * @return the index of the new state
   */
  virtual unsigned int nba_i_newState();

  /**
   * Add an edge from state <i>from</i> to state <i>to</i>
   * for the edges covered by the APMonom.
   * @param from the index of the 'from' state
   * @param m the APMonom
   * @param to the index of the 'to' state
   */
  virtual void nba_i_addEdge(unsigned int from,
			     const APMonom& m,
			     unsigned int to);

  /**
   * Get the underlying APSet 
   * @return a const pointer to the APSet
   */
  virtual APSet_cp nba_i_getAPSet();

  /** 
   * Set the final flag (accepting) for a state.
   * @param state the state index
   * @param final the flag
   */
  virtual void nba_i_setFinal(unsigned int state, bool final);
  
  /**
   * Set the state as the start state.
   * @param state the state index
   */
  virtual void nba_i_setStartState(unsigned int state);

  /** Constructing the NBA is finished. */
  virtual void nba_i_finished();



  /** Switch the APSet to another with the same number of APs. */
  void switchAPSet(APSet_cp new_apset) {
    if (new_apset->size()!=apset->size()) {
      throw std::string("New APSet has to have the same size as the old APSet!");
    }
    
    apset=new_apset;
  }


  const BDDFunction& getInitialStates() const {
    return initial;
  }

  const BDDFunction& getAccepting() const {
    return accepting;
  }

  const BDDFunction& getDelta() const {
    return delta;
  } 

  BDDFunction getAll() {
    return all_states;
  }

  const BDDFunction& getStateVars() const {return state_vars;}
  const BDDFunction& getAPSetVars() const {return apset_vars;}
  
  BDDFunction post(BDDFunction Q) {
    return (Q&delta).exists(state_vars & apset_vars).moveUp();
  }

  BDDFunction pre(BDDFunction Q) {
    BDDFunction deltaQ=delta.exists(apset_vars);
    return pre(Q, deltaQ);
  }

  BDDFunction pre(BDDFunction Q, BDDFunction deltaQ) {
    return (Q.moveDown()&deltaQ).exists(state_vars.moveDown());
  }
  
  BDDFunction cycle_states(BDDFunction Q) {
    BDDFunction Q_old=Q;
    BDDFunction Q_new;
    while (true) {
      if (debugLevel>=4) {
	std::cout << "**********************\n";
	Q_old.printPaths();
	std::cout << "Pre:\n";
      }
      BDDFunction preimg=(Q_old.moveDown()&delta).exists(state_vars.moveDown());
      if (debugLevel>=4) {
	preimg.printPaths();
      }
      Q_new=Q_old&preimg;
      if (Q_new == Q_old) {
	break;
      }
      
      Q_old=Q_new;
    }
    
    return Q_new;
  }

  typedef BDDFunction (*fn_precalc)(BDDFunction);
  
  BDDFunction fair_cycle(fn_precalc precalc) {
    BDDFunction Z, Z_old;
    
    Z=Z_old=all_states;
    while (true) {
      BDDFunction Q, Q_old;
      Q=Q_old=Z & getAccepting();
      while (true) {
	Q|=(*precalc)(Q);
	if (Q==Q_old) {break;}
	Q_old=Q;
      }

      Z&=(*precalc)(Q);

      if (Z==Z_old) {break;}
      Z_old=Z;
    }    
    return Z;
  }

  void print() {
    std::cout << "NBA:\n";
    std::cout << "States: \n";
    all_states.printPaths();

    std::cout << "\nInitial: \n";
    initial.printPaths();

    std::cout << "\nAccepting" << std::endl;
    accepting.printPaths();

    std::cout << "\nDelta" << std::endl;
    delta.printPaths();

    std::cout << "\nState Variables" << std::endl;
    state_vars.printPaths();
  }

protected:
  APSet_cp apset;
  ap_to_bddfunction_t ap_to_bddfunction;

  unsigned int states;

  unsigned int var_count;

  BDDFunction initial;
  BDDFunction accepting;
  BDDFunction delta;
  BDDFunction all_states;

  BDDFunction state_vars;
  BDDFunction apset_vars;


  BDDFunctionContainer state_variables;

  BDDFunction newStateVariable();

  BDDFunction state2bdd(unsigned int state, bool copy) {
    BDDFunction result=BDDFunction::ONE();

    CHECK_SUCCESS(var_count>0);

    for (unsigned int j=0;j<var_count;j++) {
      bool is_set=(state >> j) & 1;
      BDDFunction v=state_variables.at(j);
      if (copy) {
        v=v.moveDown();
      }
      if (!is_set) {
	v=!v;
      }
      
      result&=v;
    }

    return result;
  }
  
  BDDFunction monom2bdd(const APMonom& m) const {
    BDDFunction result=BDDFunction::ONE();

    if (m.isTrue()) {return result;}
    
    for (unsigned int i=0;i<apset->size();i++) {
      if (m.isSet(i)) {
	BDDFunction x=ap_to_bddfunction.at(i);
	result&=(m.getValue(i) ? x : !x);
      }
    }
    
    return result;
  }
};

}}

#endif
