/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef FAIR_CYCLE_DETECTION_HPP
#define FAIR_CYCLE_DETECTION_HPP


#include "vereofy/setup.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/system/symb_automaton/symb_trace.hpp"
#include "vereofy/system/symb_automaton/symb_trace_lasso.hpp"
#include "vereofy/system/symb_automaton/symb_bfs.hpp"
#include "vereofy/system/symb_automaton/symb_automaton.hpp"
//#include <pair>
#include <string>
#include <iostream>
#include <sstream>

namespace vereofy {namespace ltl_check {


class Fair_Cycle_Detection {
public:

  Fair_Cycle_Detection(BDDFunction delta_, 
		       BDDFunction state_vars_,
		       BDDFunction transition_vars_,
		       BDDFunction fair_states_,
		       BDDFunction initial_,
		       bool with_witness_=false) :
    delta(delta_), 
    state_vars(state_vars_), 
    transition_vars(transition_vars_), 
    fair_states(fair_states_),
    initial(initial_),
    constraint(BDDFunction::ONE()),
    with_witness(with_witness_) {
   
    TRACE("Fair_Cycle_Detection", "deltaQ", cout << "Generate deltaQ\n"); 
    deltaQ=delta.exists(transition_vars);

    if (debugLevel>=4) {
      std::cout << "NBA:\n";
      std::cout << "\n delta\n"; delta.printPaths();
      std::cout << "\n state_vars\n"; state_vars.printPaths();
      std::cout << "\n transition_vars\n"; transition_vars.printPaths();
      std::cout << "\n fair_states\n"; fair_states.printPaths();
      std::cout << "\n initial\n"; initial.printPaths();
      std::cout << "\n deltaQ\n"; deltaQ.printPaths();
    }
  }

  void setStateConstraint(BDDFunction constraint) {
    this->constraint=constraint;
  }

  BDDFunction preImageQ(BDDFunction Q) {
    return (Q.moveDown() & deltaQ).exists(state_vars.moveDown()).constrain(constraint);
  }

  BDDFunction postImageQ(BDDFunction Q) {
    return (Q & deltaQ).exists(state_vars).moveUp().constrain(constraint);
  }

  BDDFunction post(BDDFunction Q) {
    return (Q & delta).exists(state_vars).moveUp().constrain(constraint);
  }

  BDDFunction pre(BDDFunction Q) {
    return (Q.moveDown() & delta).exists(state_vars.moveDown()).constrain(constraint);
  }

  bool find_cycle() {
    BDDFunction Z_old;
    Z=Z_old=BDDFunction::ONE();//postImageQ(BDDFunction::ONE());

    unsigned int z=0;
    while (true) {
      TRACE_BDD_FUNCTION("Fair_Cycle_Detection_find_cycle", Z, "z = " << z);
      
      BDDFunction Q, Q_old;
      Q=Q_old=BDDFunction::ZERO();
      BDDFunction q=Z & fair_states;
      TRACE_BDD_FUNCTION("Fair_Cycle_Detection_find_cycle", q, "z = " << z);
      unsigned int i=0;
      while (true) {
        TRACE("Fair_Cycle_Detection", "find_cycle", cout << "find_cycle, z = " << z << ", i=" << i << " (" << BDD::size() << ")\n");
	Q|=preImageQ(q);
	if (Q==Q_old) {break;}

        TRACE_BDD_FUNCTION("Fair_Cycle_Detection_find_cycle", Q, "z = " << z << ", i=" << i);

	Q_old=q=Q;
        i++;
      }
      
      TRACE("Fair_Cycle_Detection", "find_cycle", cout << "find_cycle, z = " << z << ", preImage(Q) (" << BDD::size() << ")\n");
      Z&=preImageQ(Q);
      
      if (Z==Z_old) {break;}
      Z_old=Z;

      z++;
    }    
  
    TRACE_BDD_FUNCTION("Fair_Cycle_Detection_find_cycle", Z, "result");

    
    TRACE_BDD_FUNCTION("Fair_Cycle_Detection_find_cycle", Z & initial, "result");
    bool rv=bdd_sets_have_intersection(Z, initial);

    if (!with_witness) {
      Z=BDDFunction::ZERO();
      deltaQ=BDDFunction::ZERO();
    }

    return rv;
  }

  SymbTraceLasso::ptr find_witness() {
    ERROR_CHECK_ASSERT(with_witness);

    if (debugLevel>=4) {
      std::cout << "Find Witness\n";
      
      std::cout << "Fair States\n";
      fair_states.printPaths();
      std::cout << std::endl;
    }
    
    SymbTraceLasso::ptr lasso(new SymbTraceLasso());
    BDDFunction Z_and_fair=Z & fair_states;
    
    SymbBFS bfs(delta, state_vars, transition_vars, constraint);
    lasso->prefix=SymbTrace::ptr(new SymbTrace(bfs.findTraceBetweenStateSets(initial, Z_and_fair)));
    
    SymbState potential_cycle_state=lasso->prefix->getLastState();
    while (true) {
      ReturnValue<SymbTrace> trace=bfs.findTraceBetweenStateSets(potential_cycle_state.getX(), potential_cycle_state.getX());
      if (trace.valid) {
        lasso->cycle=SymbTrace::ptr(new SymbTrace(trace.value));
        break;
      }

      // potential_cycle_state can not close the cycle, remove from Z_and_fair
      Z_and_fair&=!potential_cycle_state.getX();
      if (Z_and_fair==BDDFunction::ZERO()) {
        throw "Can't find cycle!";
      }
      
      SymbTrace to_next_potential_cycle_state=bfs.findTraceBetweenStateSets(potential_cycle_state, Z_and_fair);
      potential_cycle_state=to_next_potential_cycle_state.getLastState();
      
      lasso->prefix->appendTrace(to_next_potential_cycle_state);
    }

    return lasso;
  }

private:
  BDDFunction delta;
  BDDFunction deltaQ;

  BDDFunction state_vars;
  BDDFunction transition_vars;

  BDDFunction fair_states;
  BDDFunction initial;

  BDDFunction constraint;

  BDDFunction Z;

  bool with_witness;
};


}}
#endif
