/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "LTLModelChecker.hpp"
#include "vereofy/modelcheck/specifications/LTL_Specification.hpp"
#include "vereofy/system/symb_automaton/symb_component_instance.hpp"
#include "vereofy/common/stopwatch.hpp"

#include "symb_nba_product.hpp"
#include "Fair_Cycle_Detection.hpp"
#include "BDD_NBA.hpp"
#include "Symb_LTL_NBA.hpp"
#include "LTL2NBA.hpp"

#include <sstream>
#include <boost/lexical_cast.hpp>

namespace vereofy {

using namespace ltl_check;

  ModelCheckingResult::ptr LTLModelChecker::analyze(Automaton::ptr automaton, Specification::ptr spec,
                                                    const OptionContainer& options)  {

  LTL_Specification::ptr ltl_spec=boost::dynamic_pointer_cast<LTL_Specification>(spec);
  if (!ltl_spec) {
    throw "Not an LTL specification!";
  }
  
  LTLFormula_ptr ltl_formula=ltl_spec->getLTLFormula();
  APSet_cp ap_set=ltl_formula->getAPSet();
    
  
  SymbComponentInstance::ptr system_automaton=boost::dynamic_pointer_cast<SymbComponentInstance>(automaton);
  if (!system_automaton) {
    throw "System is not a symbolic automaton for a component!";
  }
  

  bool has_terminal_states=false;

  BDDFunction terminal_states=system_automaton->calcTerminalStates();
  if (terminal_states!=BDDFunction::ZERO()) {
    // there are terminal states, even if they may not be reachable
    has_terminal_states=true;

    system_automaton->addStoppedTransitions();
    TRACE_SYMB_COMPONENT_DOT("LTLModelChecker", "after_stopped", *system_automaton, true);
  }

  if (debugLevel>=3) {
    cout << "Translating AP to BDDFunctions\n";
  }
  BDD_NBA::ap_to_bddfunction_t ap_to_bdd;
  for (unsigned int i=0;i<ap_set->size();i++) {
    if (debugLevel>=3) {
      cout << "AP = " << ap_set->getAP(i) << "\n";
    }
    bool no_io=false;
    bool no_state=false;

    BDDFunction ap_sat;
    string guard=ap_set->getAP(i);

    if (guard=="stopped") {
      ap_sat=system_automaton->getStoppedTransitionLabel();
    } else if (guard.find("enabled ")==0) {
      guard=guard.substr(8);
      try {
	ap_sat=system_automaton->parse_enabled_io_guard(guard, false, false);
      } catch (...) {
	throw "No I/O guard for enabled!\n";
      }
    } else {
      try {
        if (debugLevel>=3) {
          cout << "Trying state guard\n";
        }
        ap_sat=system_automaton->parse_state_guard(guard);
      } catch (...) {
        if (debugLevel>=3) {
          cout << "No state guard\n";
        }
        no_state=true;
      }

      if (no_state) {      
        try {
          if (debugLevel>=3) {
            cout << "Trying I/O guard\n";
          }
          ap_sat=system_automaton->parse_io_guard(guard, true, true);
        } catch (...) {
          no_io=true;
          if (debugLevel>=3) {
            cout << "No I/O guard\n";
          }
        }
      }
      
      if (no_state && no_io) {
	throw "Atomic proposition "+guard+" is neither a valid state nor I/O guard";
      }
    }

    ap_to_bdd.push_back(ap_sat);
    
    if (debugLevel>=3) {
      cout << "Sat = \n";
      ap_sat.printPaths();
    }
  }
  
  cout << "Generate NBA for LTL formula\n";
  
  StopWatch nba_gen_watch;
  nba_gen_watch.start();
  std::string ltl2nba_internal_value=options.getOption("ltl2nba-internal").getValue();
  bool ltl2nba_internal;
  try {
    ltl2nba_internal=boost::lexical_cast<bool>(ltl2nba_internal_value);
  } catch (boost::bad_lexical_cast& e) {
    throw "Option --ltl2nba-internal has to be either 0 or 1";
  }

  BDD_NBA::ptr nba;
  if (ltl2nba_internal) {
    nba=BDD_NBA::ptr(new Symb_LTL_NBA(ltl_formula->negate(), 
                                      ltl_formula->getAPSet(),
                                      ap_to_bdd));
  } else {
    LTL2NBA_SPIN<BDD_NBA> ltl2nba(options.getOption("ltl2ba").getValue());
    nba=BDD_NBA::ptr(new BDD_NBA(ltl_formula->getAPSet(), ap_to_bdd));

    if (!ltl2nba.ltl2nba(*(ltl_formula->negate()), 
                         nba.get())) {
      throw "Could not generate NBA!";
    }
  }
  nba_gen_watch.stop();

  CHECK_SUCCESS(nba);

  if (debugLevel>=4) {
    nba->print();
  }


  StopWatch nba_product_watch;  
  nba_product_watch.start();

  cout << "Generating product automaton\n";
  SymbNBAProductAutomaton::ptr prod_automaton(new SymbNBAProductAutomaton(system_automaton,
                                                                          nba));

  nba_product_watch.stop();

  std::string nba_product_dot(options.getOption("nba-product-dot").getValue());  
  if (nba_product_dot!="") {
    ofstream dot_stream(nba_product_dot.c_str());
    if (dot_stream.good()) {
      prod_automaton->toDot(dot_stream);
    } else {
      throw "Can't open DOT file: "+nba_product_dot;
    }
  }

  Fair_Cycle_Detection fcd(prod_automaton->getDelta(),
			   prod_automaton->getOriginalStateVars(),
			   prod_automaton->getTransitionVars(),
			   prod_automaton->getAcceptingStates(),
			   prod_automaton->getInitialStates(),
			   true);

  bool precompute_reachable=options.getOption("precompute-reachable").getBoolValue();
  StopWatch precompute_reachable_watch;
  if (precompute_reachable) {
    cout << "Computing reachable states...\n";
    precompute_reachable_watch.start();
    fcd.setStateConstraint(system_automaton->calcReachableStates());
    precompute_reachable_watch.stop();
  }

  ModelCheckingResult::ptr mc_result(new ModelCheckingResult(system_automaton));

  StopWatch nba_check_watch;  
  nba_check_watch.start();

  cout << "LTL Modelchecking...\n";
  bool fair_cycle=fcd.find_cycle();
  mc_result->setCheckPassed(!fair_cycle);

  nba_check_watch.stop();

  cout << "Time for NBA generation = " << nba_gen_watch.getElapsedTime() << "s\n";
  cout << "Time for NBA product generation = " << nba_product_watch.getElapsedTime() << "s\n";
  if (precompute_reachable) {
    cout << "Time for computation of reachable state space = " << precompute_reachable_watch.getElapsedTime() << "s\n";
  }
  cout << "Time for NBA model checking = " << nba_check_watch.getElapsedTime() << "s\n";

  if (fair_cycle) {
    SymbTraceLasso::ptr lasso;
    lasso=fcd.find_witness();
  
    ModelCheckingResultTrace::ptr counter_example(new ModelCheckingResultTrace(ModelCheckingResultTrace::TRACE_COUNTEREXAMPLE,
                                                                               spec->getSpecification(),
                                                                               lasso));
    mc_result->addTrace(counter_example);

    std::string ltl_counterexample_dot(options.getOption("ltl-counterexample-dot").getValue());
    if (ltl_counterexample_dot!="") {
      ofstream dot_stream(ltl_counterexample_dot.c_str());
      if (dot_stream.good()) {
        prod_automaton->toDotWithTrace(dot_stream, 
                                       lasso->prefix->getAllStates() | lasso->cycle->getAllStates(),
                                       lasso->prefix->getAllTransitions(),
                                       lasso->cycle->getAllTransitions());
      } else {
        throw "Can't open DOT file: "+ltl_counterexample_dot;
      }
    }
  }
    
  return mc_result;
}


OptionContainer LTLModelChecker::getOptions() const {
  OptionContainer options;

  std::string ltl2ba_default=
#if (__WIN32__ || _WIN32)
    "ltl2ba.exe"
#else
    "ltl2ba"
#endif
    ;

  Option opt_ltl2ba("ltl2ba",
                    "The ltl2ba executable needed for LTL model checking if the internal translator is not used",
                    ltl2ba_default);
  options.addOption(opt_ltl2ba);

  Option opt_ltl2nba_internal("ltl2nba-internal",
                              "Use the internal symbolic LTL to NBA translation",
                              "0");
  options.addOption(opt_ltl2nba_internal);

  Option opt_product_dot("nba-product-dot",
                         "(LTL) Output the product automaton with the NBA as a DOT file",
                         "");
  options.addOption(opt_product_dot);

  Option opt_counterexample_dot("ltl-counterexample-dot",
                                "(LTL) Output the product automaton with the counterexample trace", 
                                "");
  options.addOption(opt_counterexample_dot);

  return options;
}



}
