/***************************************************************************
 *   Copyright (C) 2009 by Joachim Klein                                   *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Symb_LTL_NBA.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/common/bdd_helpers.hpp"

#include "vereofy/system/datadomain.hpp"
#include "vereofy/system/symb_automaton/symb_flat_datadomain.hpp"

using namespace vereofy;
using namespace vereofy::ltl_check;


Symb_LTL_NBA::Symb_LTL_NBA(LTLFormula_ptr ltl,
                           APSet_cp apset_,
                           const ap_to_bddfunction_t& ap_to_bddfunction_) :
  BDD_NBA(apset_, ap_to_bddfunction_) {

  std::cout << "Symbolic LTL to Buechi translation: EXPERIMENTAL" << std::endl;

  LTLFormula_ptr ltl_pnf=ltl->toPNF();

  // initially, all transitions are ok
  delta=BDDFunction::ONE();
  for (unsigned int i=0;i<apset->size();i++) {
    ap_vars.push_back(newStateVariable());
    delta&=EQUIV(ap_vars[i], ap_to_bddfunction.at(i));
  }

  initial=sat(ltl_pnf->getRootNode());

  TRACE_BDD_FUNCTION("Symb_LTL_NBA", initial, ltl_pnf->toStringPrefix());
  TRACE_BDD_FUNCTION("Symb_LTL_NBA", delta, "");
  degeneralize();
  TRACE_BDD_FUNCTION("Symb_LTL_NBA", delta, "");
  TRACE_BDD_FUNCTION("Symb_LTL_NBA", accepting, "");
}

BDDFunction Symb_LTL_NBA::sat(LTLNode_p ltl_node) {
  BDDFunction left_sat;
  BDDFunction right_sat;

  LTLNode_p left=ltl_node->getLeft();
  LTLNode_p right=ltl_node->getRight();

  if (left) {
    left_sat=sat(left);
    TRACE_BDD_FUNCTION("Symb_LTL_NBA", left_sat, LTLFormula(left, apset).toStringPrefix());
  }

  if (right) {
    right_sat=sat(right);
    TRACE_BDD_FUNCTION("Symb_LTL_NBA", right_sat, LTLFormula(right, apset).toStringPrefix());
  }

  switch (ltl_node->getType()) {
  case LTLNode::T_AP:
    return ap_vars.at(ltl_node->getAP());

  case LTLNode::T_TRUE:
    return BDDFunction::ONE();
  case LTLNode::T_FALSE:
    return BDDFunction::ZERO();
    
  case LTLNode::T_NOT:
    CHECK_SUCCESS(left->getType()==LTLNode::T_AP);
    return !left_sat;

  case LTLNode::T_AND:
    return left_sat & right_sat;
  case LTLNode::T_OR:
    return left_sat | right_sat;
  case LTLNode::T_XOR:
    return left_sat ^ right_sat;

  case LTLNode::T_IMPLICATE:
  case LTLNode::T_EQUIV:
    THROW_UNIMPLEMENTED();
    
  case LTLNode::T_NEXTSTEP: {
    BDDFunction next_var=newStateVariable();
    TRACE_BDD_FUNCTION("Symb_LTL_NBA", next_var, "");
    
    BDDFunction delta_next=EQUIV(next_var, left_sat.moveDown());
    TRACE_BDD_FUNCTION("Symb_LTL_NBA", delta_next, "");

    delta &= delta_next;
    return next_var;
  }

  case LTLNode::T_GLOBALLY:
  case LTLNode::T_FINALLY:
    THROW_UNIMPLEMENTED();

  case LTLNode::T_UNTIL: {
    BDDFunction until_var=newStateVariable();
    TRACE_BDD_FUNCTION("Symb_LTL_NBA", until_var, "");

    BDDFunction delta_until=EQUIV(until_var, 
                                    (left_sat & until_var.moveDown()) | right_sat);
    TRACE_BDD_FUNCTION("Symb_LTL_NBA", delta_until, "");    

    delta &= delta_until;


    gnba_accepting_states.push_back(!until_var | right_sat);
    return until_var;
  }
  case LTLNode::T_WEAKUNTIL:
    THROW_UNIMPLEMENTED();
  case LTLNode::T_RELEASE: {
    BDDFunction release_var=newStateVariable();
    TRACE_BDD_FUNCTION("Symb_LTL_NBA", release_var, "");
    
    BDDFunction delta_release=EQUIV(release_var, 
                                      (right_sat & release_var.moveDown()) | (left_sat & right_sat));
    TRACE_BDD_FUNCTION("Symb_LTL_NBA", delta_release, "");

    delta &= delta_release;
    return release_var;
  }

  case LTLNode::T_WEAKRELEASE:
    THROW_UNIMPLEMENTED();
  case LTLNode::T_BEFORE:
    THROW_UNIMPLEMENTED();
  }
  
  THROW_UNIMPLEMENTED();
}

void Symb_LTL_NBA::degeneralize() {
  if (gnba_accepting_states.size()==0) {
    // all states are accepting
    accepting=BDDFunction::ONE();
  } else if (gnba_accepting_states.size()==1) {
    accepting=gnba_accepting_states.at(0);
  } else {
    IntRangeDataDomain::ptr dd(new IntRangeDataDomain(0, gnba_accepting_states.size()-1));
    Symb_FlatDataDomain::ptr s_dd(new Symb_FlatDataDomain(dd));

    BDDFunctionContainer copy_vars;
    BDDFunctionContainer copy_vars_y;
    for (unsigned int i=0;i<s_dd->getLogSize();i++) {
      BDDFunction s=newStateVariable();
      copy_vars.push(s);
      copy_vars_y.push(s.moveDown());
    }

    for (unsigned int i=0;i<gnba_accepting_states.size();i++) {
      unsigned int next_copy=(i+1)%gnba_accepting_states.size();
      
      BDDFunction this_copy_x=pairwiseEQUIV(copy_vars, s_dd->getEncoding(i));
      BDDFunction this_copy_y=this_copy_x.moveDown();
      BDDFunction next_copy_y=pairwiseEQUIV(copy_vars_y, s_dd->getEncoding(next_copy));
      
      delta&=IMPLIES(this_copy_x & gnba_accepting_states.at(i), next_copy_y);
      delta&=IMPLIES(this_copy_x & !gnba_accepting_states.at(i), this_copy_y);

      if (i==0) {
        initial&=this_copy_x;
        accepting=this_copy_x & gnba_accepting_states.at(0);
      }
    }
  }  
}

