/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/



#include "symb_nba_product.hpp"
#include "vereofy/common/bdd_op_trace.hpp"

namespace vereofy {
namespace ltl_check {

SymbNBAProductAutomaton::SymbNBAProductAutomaton(SymbComponentInstance::ptr system_automaton_,
                                                 BDD_NBA::ptr nba_) :
  SymbAutomaton(*system_automaton_),   // clone from system_automaton_
  system_automaton(system_automaton_),
  nba(nba_) {

  BDDFunction system_delta=system_automaton->getDelta();
  BDDFunction nba_delta=nba->getDelta();
  
  delta = system_delta & nba_delta;
  BDD_OP_TRACE2("symb_nba_product_delta", delta, system_delta, nba_delta);
  
  csOriginalX   &= nba->getStateVars();
  csCopyY        = csOriginalX.moveDown();
  csCopyCopyZ    = csCopyY.moveDown();
  initialStates &= nba->getInitialStates();

  acceptingStates=nba->getAccepting();

  if (system_automaton->getStoppedTransitionLabel()!=BDDFunction::ZERO()) {
    stopped_transition_label=system_automaton->getStoppedTransitionLabel();
  }
}
  
BDDFunction SymbNBAProductAutomaton::getAcceptingStates() const {
  return acceptingStates;
}

void SymbNBAProductAutomaton::Product_State_Renderer::render(ostream& out,
                                                             const BDDFunction& assignment,
                                                             bool use_x) const {
  system_renderer->render(out, assignment, use_x);

  
  out << " | NBA=" << assignmentToHex(assignment, 
                                      use_x ? nba.getStateVars() : nba.getStateVars().moveDown());
                                
  if (bdd_sets_have_intersection(assignment, 
                                 use_x ? nba.getAccepting() : nba.getAccepting().moveDown())) {
    out << "!";
  }
}

bool SymbNBAProductAutomaton::Product_State_Renderer::render_dot_style(ostream& out, 
                                                                       const BDDFunction& assignment, 
                                                                       bool is_initial,
                                                                       bool is_terminal) const {
  bool is_fair_state=bdd_sets_have_intersection(assignment, nba.getAccepting());

  out << "shape="<< (is_fair_state ? "box" : "ellipse");
  
  out << ", color=";
  if (is_initial && is_terminal) {
    out << "violet";
  } else if (is_initial) {
          out << "blue";
  } else if (is_terminal) {
    out << "red";
  } else {
    out << "black";
  }

  if (bdd_sets_have_intersection(assignment, highlight_states)) {
    out << ", style=bold";
  } else {
    out << ", fontcolor=\"#808080\", color=\"#808080\"";
  }
  
  return true;
}

void SymbNBAProductAutomaton::Trace_IO_Renderer::render(ostream& out,
                                                        const BDDFunction& assignment, 
                                                        const BDDFunction& io_vars) const {
  system_renderer->render(out, assignment, io_vars);
}
  
bool SymbNBAProductAutomaton::Trace_IO_Renderer::render_dot_style(ostream& out, 
                                                                  const BDDFunction& assignment,
                                                                  const BDDFunction& io_vars,
                                                                  const BDDFunction& state_pair) const {
  BDDFunction t=assignment & state_pair;
  bool is_on_prefix=bdd_sets_have_intersection(t, transition_on_prefix);
  bool is_on_cycle=bdd_sets_have_intersection(t, transition_on_cycle);

  TRACE_BDD_FUNCTION("Trace_IO_Renderer_render_dot_style", t, "");
  TRACE_BDD_FUNCTION("Trace_IO_Renderer_render_dot_style", transition_on_prefix, is_on_prefix);
  TRACE_BDD_FUNCTION("Trace_IO_Renderer_render_dot_style", transition_on_cycle, is_on_cycle);


  enum {C_BLUE, C_RED, C_VIOLET, C_GRAY} color;
  if (is_on_prefix) {
    if (is_on_cycle) {
      color=C_VIOLET;
    } else {
      color=C_BLUE;
    }
  } else {
    if (is_on_cycle) {
      color=C_RED;
    } else {
      color=C_GRAY;
    }
  }
  
  string color_string;
  switch (color) {
  case C_BLUE:   color_string="#0000FF";break;
  case C_VIOLET: color_string="#FF00FF";break;
  case C_RED:    color_string="#FF0000";break;
  case C_GRAY:   color_string="#808080";break;
  }
  
  out << "color=\"" << color_string << "\", fontcolor=\"" << color_string << "\"";

  if (color!=C_GRAY) {
    out << ", style=bold";
  }

  return true;
}

SymbAutomaton::State_Renderer::ptr SymbNBAProductAutomaton::getDefaultStateRenderer() const {
  return SymbAutomaton::State_Renderer::ptr(new Product_State_Renderer(*system_automaton,
                                                                       *nba));
}
  
SymbAutomaton::IO_Renderer::ptr SymbNBAProductAutomaton::getDefaultIORenderer() const {
  return system_automaton->getDefaultIORenderer();
}


void SymbNBAProductAutomaton::toDotWithTrace(ostream& out, 
                                             const BDDFunction& states_on_trace,
                                             const BDDFunction& prefix_transitions,
                                             const BDDFunction& cycle_transitions) const {
  toDot(out, 
        Product_State_Renderer(*system_automaton, *nba, states_on_trace),
        Trace_IO_Renderer(*system_automaton, *nba, prefix_transitions, cycle_transitions));

}
  
}
}
