/***************************************************************************
 *   Copyright (C) 2008-2009 by Sascha Klueppelholz                        *
 *   klueppel@tcs.inf.tu-dresden.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include "ASL_Formula.hpp"
#include "vereofy/modelcheck/checkers/branching_time_check/mcreo.hpp"
#include "vereofy/modelcheck/checkers/branching_time_check/mcreo_functions.hpp"
#include "vereofy/modelcheck/specifications/asl/branching_time_logics_parser.hpp"
#include "vereofy/modelcheck/checkers/branching_time_check/hybridFiniteAutomaton.hpp"
#include "vereofy/reo_parser/guard_parser_helper.hpp"
#include "vereofy/common/functions.hpp"
#include "vereofy/common/weak_ptr_helper.hpp"

/**
 @author Sascha Klueppelholz
 */

using namespace std;
using namespace vereofy;

// initialize static container type
formula_map* formula::allFormulas=NULL;
min_type formula::minimazationMethod=MIN_UNDEFINED;

//********************************** public **********************************

const std::string formula::getParsableSpecification() const { 
  switch (formulaType) {
    case ATOM: return "\""+specAsString+"\"";
    case NEG:  return "!" + sub1->getParsableSpecification();
    case CONJ: return sub1->getParsableSpecification() + "&" + sub2->getParsableSpecification();
    case DISJ: return sub1->getParsableSpecification() + "|" + sub2->getParsableSpecification();
    case EXOR: return sub1->getParsableSpecification() + "^" + sub2->getParsableSpecification();
    default: return specAsString;
  }
}

/**
 * \brief A Method to clear computed satisfaction sets
 *
 * The satisfaction sets are of a given formula are stored within the formula object.
 * This method is used to restore the initial situation and therefore allows to compute the satisfaction set again.
 * Namely it clears the computed table.
 *
 * \sa mcreo.h
 */
void formula::clearComputation(){
  if (formulaType!=ONE && formulaType!=ZERO){    
    // model checking has to be redone. reset computed table of formula
    satSet.valid=false;
    satSet.value=BDDFunction::ZERO();
	  
    // and the flag for the stream listener
    done=false;
    
    // and subformulas if any
    if (sub1) sub1->clearComputation();
    if (sub2) sub2->clearComputation();
  } else{
    satSet.valid=true;
    satSet.value=(formulaType==ONE)?BDDFunction::ONE():BDDFunction::ZERO();
    done=true;
  }
}

const f_info& formula::getFixpointType(void) const{
  return wit_least->getFixpointType();
}

void formula::setFixpointType(const f_info& ft){
  wit_least->setFixpointType(ft);
}

/**
 * \brief This method prints a witness/counterexample
 *
 
 * \sa witness.h
 */
void formula::generateWitness(BDDFunction stateRestriction, unsigned int recursive, int parent_id, ModelCheckingResult& result) const{
  // witness not created if sat set was not computed, e.g., for 'ASL<< true | false >>' because true is already 'true'
  if (!wit_least) return;

  if (isASL){
    if (sub1) sub1->generateWitness(stateRestriction, recursive, parent_id, result);
    if (sub2) sub2->generateWitness(stateRestriction, recursive, parent_id, result);
    return;
  }
  
  switch (formulaType){    
    case EES: case EAS: case AES: case AAS:{
      wit_least->generateWitness(stateRestriction & regExpr->getInitialStates(), recursive, parent_id, result);
      break;
    }
      
    case ER: case AU:{
      if (!wit_greatest){
        wit_least->generateWitness(stateRestriction, recursive, parent_id, result);
        break;
      }
      
      if ((stateRestriction&!wit_greatest->result.top())==BDDFunction::ZERO()){
        wit_greatest->generateWitness(stateRestriction, recursive, parent_id, result);
      } else {
        wit_least->generateWitness(stateRestriction, recursive, parent_id, result);
      }
      
      break;
    }
		  
    default:{
      if (wit_least) wit_least->generateWitness(stateRestriction, recursive, parent_id, result);
      else if (wit_greatest) wit_greatest->generateWitness(stateRestriction, recursive, parent_id, result);
      break;
    }
  }
}

void formula::setDFAMinimizationMethod(const string& t) {
  if (t=="none") {
    formula::minimazationMethod=MIN_NONE;
  } else if (t=="brzozowski") {
    formula::minimazationMethod=MIN_BRZOZOWSKI;
  } else if (t=="hopcroft") {
    formula::minimazationMethod=MIN_HOPCROFT;
  } else {
    throw t+" is not a valid minimization method";
  }
}

//********************************** private **********************************

/**
 * \brief Private constructor of formula
 
 * Creates a new BTSL/ASL formula for the given expression.
 *
 * \sa mcreo.h
 */
formula::formula(const string& expr){	
	done=false;
	isASL=false;
  isBTSL=false;
	
	// try parsing the given expression
	specAsString=parse(expr);
	
	// reset satisfaction set stored in the function for this and all subformulas
	clearComputation();
}

string getExpression(const vparser::pt_node& node){
  string result = "";
  
  switch (node.value.id().to_long()){
    case branching_time_logics_parser::ID_IO_GUARD_ATOM: case 0:{
      if (node.children.size()==1) return getExpression(node.children.at(0));
      if (node.children.size()==2) result = getExpression(node.children.at(1));
      
      else {
        const vparser::pt_node& aNode = (node.children.size()==0)?(node):(node.children.at(1));
        result = "\"" + string(aNode.value.begin(),aNode.value.end()) + "\"";
      }
      
      return (node.children.size()==2)?("!("+result+")"):("("+result+")");
    }
      
    case branching_time_logics_parser::ID_IO_GUARD_STARED:{
      if (node.children.size()==1) return getExpression(node.children.at(0));
      result = getExpression(node.children.at(0)) + string(node.children.at(1).value.begin(),node.children.at(1).value.end());
      
      return "(" + result + ")";
    }
      
    case branching_time_logics_parser::ID_IO_SEQUENCE:{
      result = getExpression(node.children.at(0));
      for (unsigned int i=1;i<node.children.size();++i){
        result = result + ";" + getExpression(node.children.at(i));
      }
      
      return "(" + result + ")";
    }
      
    case branching_time_logics_parser::ID_IO_TERM_AND:{
      result = getExpression(node.children.at(0));
      for (unsigned int i=1;i<node.children.size();++i){
        result = result + "&" + getExpression(node.children.at(i));
      }
      
      return "(" + result + ")";
    }
      
    case branching_time_logics_parser::ID_IO_REGULAR_IO_EXPRESSION:{
      result = getExpression(node.children.at(0));
      for (unsigned int i=1;i<node.children.size();++i){
        result = result + "|" + getExpression(node.children.at(i));
      }
      
      return "(" + result + ")";
    }

    default:{
      THROW_UNIMPLEMENTED();
    }
  }
}

formula::ptr formula::toASLGraphStructure(const vparser::pt_node& node, const symbol_table_stack_t& symbol_table_stack){
  if (node.children.size()==1) return toASLGraphStructure(node.children.at(0), symbol_table_stack);
  
  // first formula instantiates a static map for all formulas
  if (formula::allFormulas==NULL) formula::allFormulas=new formula_map;
  
  formula::ptr fptr,child1,child2;
  vparser::pt_node alpha;
  set<string> portList;
  
  switch (node.value.id().to_long()){
    case branching_time_logics_parser::ID_CONST:{
      string value(node.value.begin(), node.value.end());
      if (value=="true") fptr=formula::ptr(new formula(value,ONE,alpha,child1,child2,false,false,portList));
      else if (value=="false") fptr=formula::ptr(new formula("false",ZERO,alpha,child1,child2,false,false,portList));
      break;
    }
      
    case branching_time_logics_parser::ID_ATOM:{
      fptr=formula::ptr(new formula(string(node.value.begin(), node.value.end()),ATOM,alpha,child1,child2,false,false,portList));
      break;
    }
      
    case branching_time_logics_parser::ID_ASL_STATE_FORMULA_AND:{
      // conjunction
      child1 = toASLGraphStructure(node.children.at(0), symbol_table_stack);
      child2 = toASLGraphStructure(node.children.at(1), symbol_table_stack);
      
      BTSLFT ftype = CONJ;
      
      string fkey=child1->getSpecification() + "&" + child2->getSpecification();
      fptr=formula::ptr(new formula(fkey,ftype,alpha,child1,child2,false,false,portList));
      break;
    }
      
    case branching_time_logics_parser::ID_ASL_STATE_TERM:{      
      if (node.children.size()==2){
        // negation
        child1=toASLGraphStructure(node.children.at(1), symbol_table_stack);
        string formula_key = "!" + child1->getSpecification();
        fptr=formula::ptr(new formula(formula_key,NEG,alpha,child1,child2,false,false,portList));
      }
      
      if (node.children.size()==3){
        if (node.children.at(1).value.id().to_long() == branching_time_logics_parser::ID_UNARY_CTL_OP){
          // CTL formula with unary path modality: EX, EF, EG, AX, AF, AG
          string value1(node.children.at(0).value.begin(), node.children.at(0).value.end());
          string value2(node.children.at(1).value.begin(), node.children.at(1).value.end());
          string fkey = value1 + value2;
          
          BTSLFT ftype=INVALID;
          if (fkey == "EX") ftype = EX;
          if (fkey == "EF") ftype = EF;
          if (fkey == "EG") ftype = EG;
          if (fkey == "AX") ftype = AX;
          if (fkey == "AF") ftype = AF;
          if (fkey == "AG") ftype = AG;
          
          child1 = toASLGraphStructure(node.children.at(2), symbol_table_stack);          
          fkey = fkey + "[" + child1->getSpecification()  + "]";
          fptr=formula::ptr(new formula(fkey,ftype,alpha,child1,child2,false,false,portList));          
        } else{
          if (node.children.at(1).value.id().to_long() == branching_time_logics_parser::ID_PREFIX_QUANTIFIER){
            // BTSL formula: EES, EAS, AES, AAS
            string value1(node.children.at(0).value.begin(), node.children.at(0).value.end());
            string fkey=(node.children.at(1).children.at(0).value.id().to_long() == branching_time_logics_parser::ID_EXISTS_PREFIX)?(value1+"ES"):(value1+"AS");
            alpha = node.children.at(1).children.at(0).children.at(0);
            
            BTSLFT ftype=INVALID;
            if (fkey == "EES") ftype = EES;
            if (fkey == "EAS") ftype = EAS;
            if (fkey == "AES") ftype = AES;
            if (fkey == "AAS") ftype = AAS;
            
            child1 = toASLGraphStructure(node.children.at(2), symbol_table_stack);
            fkey = fkey + "[" + getExpression(alpha) + "," + child1->getSpecification()  + "]";
            fptr=formula::ptr(new formula(fkey,ftype,alpha,child1,child2,true,false,portList));
          }
        }
      }
      
      if (node.children.size()==4){
        if (node.children.at(2).value.id().to_long() == branching_time_logics_parser::ID_ASL_STATE_FORMULA){
          // CTL formula with binary path modality: EU, ER, AU, AR
          string value1(node.children.at(0).value.begin(), node.children.at(0).value.end());
          string value2(node.children.at(1).value.begin(), node.children.at(1).value.end());
          string fkey = value1 + value2;
          
          BTSLFT ftype=INVALID;
          if (fkey == "EU") ftype = EU;
          if (fkey == "ER") ftype = ER;
          if (fkey == "AU") ftype = AU;
          if (fkey == "AR") ftype = AR;
          
          child1 = toASLGraphStructure(node.children.at(2), symbol_table_stack);
          child2 = toASLGraphStructure(node.children.at(3), symbol_table_stack);          
          fkey = fkey + "[" + child1->getSpecification()  + "," + child2->getSpecification() + "]";
          fptr=formula::ptr(new formula(fkey,ftype,alpha,child1,child2,false,false,portList));          
        }
        
        if (node.children.at(2).value.id().to_long() == branching_time_logics_parser::ID_UNARY_CTL_OP){
          // ASL formula with unary path modality: E{A,B,C}X, E{A,B,C}F, E{A,B,C}G, A{A,B,C}X, A{A,B,C}F, A{A,B,C}G
          string value1(node.children.at(0).value.begin(), node.children.at(0).value.end());
          string value2(node.children.at(2).value.begin(), node.children.at(2).value.end());
          string fkey = value1 + value2;
          
          BTSLFT ftype=INVALID;
          if (fkey == "EX") ftype = EX;
          if (fkey == "EF") ftype = EF;
          if (fkey == "EG") ftype = EG;
          if (fkey == "AX") ftype = AX;
          if (fkey == "AF") ftype = AF;
          if (fkey == "AG") ftype = AG;
          
          child1 = toASLGraphStructure(node.children.at(3), symbol_table_stack); 
          
          string list = "";
          if (node.children.at(1).children.size()){
            for (unsigned int i=0;i<node.children.at(1).children.at(0).children.size();++i){
              const vparser::pt_node& port_name_node=node.children.at(1).children.at(0).children.at(i);
              string port_name = vparser::guard_helper::t_port_name::getPortName(port_name_node, symbol_table_stack);
              
              if (i!=0) list = list + ",";
              list = (list + port_name);
              portList.insert(port_name);
            }
          }
          
          fkey = value1 + "{" + list + "}" + value2 + "[" + child1->getSpecification()  + "]";
          fptr=formula::ptr(new formula(fkey,ftype,alpha,child1,child2,false,true,portList));
        } else{
          if (node.children.at(1).value.id().to_long() == branching_time_logics_parser::ID_PREFIX_QUANTIFIER){
            // BTSL formula: EES, EAS, AES, AAS
            string value1(node.children.at(0).value.begin(), node.children.at(0).value.end());
            string fkey=(node.children.at(1).children.at(0).value.id().to_long() == branching_time_logics_parser::ID_EXISTS_PREFIX)?(value1+"ES"):(value1+"AS");
            alpha = node.children.at(1).children.at(0).children.at(0);
            
            BTSLFT ftype=INVALID;
            if (fkey == "EES") ftype = EES;
            if (fkey == "EAS") ftype = EAS;
            if (fkey == "AES") ftype = AES;
            if (fkey == "AAS") ftype = AAS;
            
            child1 = toASLGraphStructure(node.children.at(2), symbol_table_stack);
            fkey = fkey + "[" + getExpression(alpha) + "," + child1->getSpecification()  + "]";
            fptr=formula::ptr(new formula(fkey,ftype,alpha,child1,child2,true,false,portList));
          }else{
            if (node.children.at(2).value.id().to_long() == branching_time_logics_parser::ID_PREFIX_QUANTIFIER){
              // ASL formula with unary path modality: EE{A,B,C}S, EA{A,B,C}S, AE{A,B,C}S, AA{A,B,C}S
              string value1(node.children.at(0).value.begin(), node.children.at(0).value.end());
              string fkey=(node.children.at(2).children.at(0).value.id().to_long() == branching_time_logics_parser::ID_EXISTS_PREFIX)?(value1+"ES"):(value1+"AS");
              alpha = node.children.at(2).children.at(0).children.at(0);
              
              BTSLFT ftype=INVALID;
              if (fkey == "EES") ftype = EES;
              if (fkey == "EAS") ftype = EAS;
              if (fkey == "AES") ftype = AES;
              if (fkey == "AAS") ftype = AAS;
              
              child1 = toASLGraphStructure(node.children.at(3), symbol_table_stack);
              
              string list = "";
              if (node.children.at(1).children.size()){
                for (unsigned int i=0;i<node.children.at(1).children.at(0).children.size();++i){
                  const vparser::pt_node& port_name_node=node.children.at(1).children.at(0).children.at(i);
                  string port_name = vparser::guard_helper::t_port_name::getPortName(port_name_node, symbol_table_stack);
                  
                  if (i!=0) list = list + ",";
                  list = (list + port_name);
                  portList.insert(port_name);
                }
              }
              
              fkey=(ftype==EES || ftype==EAS)?(value1 + "{" + list + "}<"  + getExpression(alpha) + ">" + child1->getSpecification()):(value1 + "{" + list + "}["  + getExpression(alpha) + "]" + child1->getSpecification());
              fptr=formula::ptr(new formula(fkey,ftype,alpha,child1,child2,true,true,portList));
            }
          }
        }
      }
      
      if (node.children.size()==5){
        if (node.children.at(3).value.id().to_long() == branching_time_logics_parser::ID_ASL_STATE_FORMULA){
          // ASL formula with binary path modality: E{A,B,C}U, E{A,B,C}R, A{A,B,C}U, A{A,B,C}R
          string value1(node.children.at(0).value.begin(), node.children.at(0).value.end());
          string value2(node.children.at(2).value.begin(), node.children.at(2).value.end());
          string fkey = value1 + value2;
          
          BTSLFT ftype=INVALID;
          if (fkey == "EU") ftype = EU;
          if (fkey == "ER") ftype = ER;
          if (fkey == "AU") ftype = AU;
          if (fkey == "AR") ftype = AR;
          
          child1 = toASLGraphStructure(node.children.at(3), symbol_table_stack);
          child2 = toASLGraphStructure(node.children.at(4), symbol_table_stack);
          
          string list = "";
          
          if (node.children.at(1).children.size()){
            for (unsigned int i=0;i<node.children.at(1).children.at(0).children.size();++i){
              const vparser::pt_node& port_name_node=node.children.at(1).children.at(0).children.at(i);
              string port_name = vparser::guard_helper::t_port_name::getPortName(port_name_node, symbol_table_stack);
              if (i!=0) list = list + ",";
              list = (list + port_name);
              portList.insert(port_name);
            }
          }
          
          fkey=value1 + "{" + list + "}" + value2 + "[" + child1->getSpecification()  + "," + child2->getSpecification() + "]";
          fptr=formula::ptr(new formula(fkey,ftype,alpha,child1,child2,false,true,portList));          
        }
      }
      break;
    }
      
    case branching_time_logics_parser::ID_ASL_STATE_FORMULA:{
      // boolean operators ^ and |
      child1 = toASLGraphStructure(node.children.at(0), symbol_table_stack);
      child2 = toASLGraphStructure(node.children.at(1), symbol_table_stack);
      
      string op = string(node.value.begin(),node.value.end());
      BTSLFT ftype = INVALID;
      
      if (op == "^") ftype = EXOR;
      if (op == "|") ftype = DISJ;
      
      string fkey=child1->getSpecification() + op + child2->getSpecification();
      fptr=formula::ptr(new formula(fkey,ftype,alpha,child1,child2,false,false,portList));
      break;
    }
      
    default:{
      THROW_UNIMPLEMENTED();
    }
  }
  
  string spec = fptr->getSpecification();
  if (formula::allFormulas->count(spec)>0) {
    // formula already exists -> return pointer from map if it is still alive
    formula::ptr stored_fptr=get_shared_ptr_from_weak_ptr((*formula::allFormulas)[spec]);
    if (stored_fptr) {
      // still alive
      return stored_fptr;
    }
  }

  // otherwise insert new formula into map
  (*formula::allFormulas)[spec]=formula::ptr(fptr);  
  
  return fptr;
} 

formula::formula(const string& _spec, const BTSLFT _formulaType, const vparser::pt_node& alpha, const formula::ptr _sub1, const formula::ptr _sub2, const bool isBTSLFormula, const bool isASLFormula, const set<string> portList){
  specAsString    = _spec;
  formulaType     = _formulaType;
  if (_sub1) sub1 = _sub1;
  if (_sub2) sub2 = _sub2;
  isASL           = isASLFormula;
  isBTSL          = isBTSLFormula;

  // set minimazation method default (=brzozowski)
  if (minimazationMethod == MIN_UNDEFINED) minimazationMethod=MIN_BRZOZOWSKI;
  
  // Create finite automaton for alpha for BTSL formulas
  if ((formulaType == EES) | (formulaType == EAS) | (formulaType == AES) | (formulaType == AAS)){
    regExprRootNode = boost::shared_ptr<vparser::pt_node>(new vparser::pt_node(alpha));
    regExprAsString = getExpression(alpha);
    isBTSL = true;
  }
  
  // Copy controllable ports
  if (isASLFormula){
    controllablePorts = portList;
  }
  
	// reset satisfaction set stored in the function for this and all subformulas
	clearComputation();
}

string newFormat(const string& old){
  string result = "";
  if (old[0] != '(' && old[0] != '!') result = "\"";
  bool nextStep = true;
  bool insight = false;
  
  for (unsigned int i=0;i<old.length();++i){
    if (old[i] == ';' || old[i] == '|' || old[i] == '&'){
      if (nextStep) result = result + string("\"") + old.substr(i,1) + string("\"");
      else result += old[i] + string("\"");
      nextStep = true;
    } else if (old[i] == '*' || old[i] == '+') {
      if (nextStep) result = result + string("\"") + old.substr(i,1);
      else result = result + old.substr(i, 1);
      nextStep = false;
    }else if (old[i] == '('){
      result = result + "(\"";
    }else if (old[i] == ')'){
      if (nextStep) result = result + "\")";
      else result = result + ")";
      nextStep = false;
    }else if (old[i] == ',' && !insight){
      if (nextStep) result = result + string("\"&\"");
      else result += string("&\"");
      nextStep = true;
    } else if (old[i] == '{'){
      result += old[i];
      insight = true; 
    } else if (old[i] == '}'){
      result += old[i];
      insight = false;
    } else result += old[i];
  }
 
  if (nextStep) result+="\"";
  
  return result;
}


/**
 * \brief This method parses a given string
 *
 * This method parses a given string for a BTSL/ASL formula and generates subformulas and finite automata if needed.
 *
 * \sa mcreo.h consts.h
 */
string formula::parse(const string& expression){
  string trimmed_expression=trim(expression);
  
  // I know this expression already.... 
  if (formula::allFormulas->count(trimmed_expression)>0){
    formula::ptr f=get_shared_ptr_from_weak_ptr((*formula::allFormulas)[trimmed_expression]);
    if (f) {
      newSpec = f->newSpec;
      return f->specAsString;
    }
  }
  
  // initialize parsing	
  size_t pos=0;
  string phi1,phi2;
  bool error=false;
  
  // watch out for binary operators (BTSL/ASL: &|^) on level zero
  unsigned int OPPos=findOperatorPosition(trimmed_expression,BTSLBO);
  
  // found any binary operator?
  if (OPPos){
    // set formula type
    if (trimmed_expression[OPPos]=='|') formulaType=DISJ;
    else if (trimmed_expression[OPPos]=='&') formulaType=CONJ;
    else if (trimmed_expression[OPPos]=='^') formulaType=EXOR;
    else error=true;
    
    // find or add subformulas
    if (!error){
      sub1=findOrAdd(trimmed_expression.substr(0,OPPos));
      sub2=findOrAdd(trimmed_expression.substr(OPPos+1,trimmed_expression.length()-OPPos-1));
      
      // return the expression this formula represents
      newSpec = sub1->newSpec + trimmed_expression[OPPos] + sub2->newSpec;
      if (!error) return trim(expression);
    }
  }
  
  // looking for formula of the form E\psi
  else if (trimmed_expression.length()>2){
    if (trimmed_expression[0]=='E'){
      
      // preprocessing list of ports for ASL formulas E{P1,P2,..,Pn}\psi
      if (trimmed_expression[1]=='{'){
        isASL=true;
        newSpec = "{";
        pos=findMatchingClosing(trimmed_expression, '{', '}');
        if (pos!=string::npos){
          string work=trimmed_expression.substr(2,pos-2);
          while (work.find(',')!=string::npos){
            controllablePorts.insert(work.substr(0,work.find(',')));
            newSpec = newSpec + work.substr(0,work.find(',')) + ",";
            work.erase(0,work.find(',')+1);
          }
          if (work.length()){
            controllablePorts.insert(work);
            newSpec = newSpec + work;
          }
          newSpec = newSpec + "}";
          trimmed_expression.erase(1,pos);
        } else error=true;
      }
      
      // E<alpha>phi:	  
      if (trimmed_expression[1]=='<' && !error){
        pos=findMatchingClosing(trimmed_expression, '<', '>');
        if (pos!=string::npos){
          // set formula type EES
          formulaType=EES;
          
          // build automaton for alpha
          regExpr=SymbAutomatonLanguageAcceptor_ptr(new SymbAutomatonLanguageAcceptor(trimmed_expression.substr(2,pos-2)));
          
          // find or add subformula
          phi1=trimmed_expression.substr(pos+1,trimmed_expression.length()-pos-1);
          
          sub1=findOrAdd(phi1);
          
          // return the expression this formula represents
          newSpec = "E" + newSpec + "<" + newFormat(trimmed_expression.substr(2,pos-2)) + ">" + sub1->newSpec;
          if (!error) return trim(expression);
        } else error=true;
      }
      
      // E[alpha]phi:
      else if (trimmed_expression[1]=='[' && !error){
        pos=findMatchingClosing(trimmed_expression, '[', ']');
        if (pos!=string::npos){
          // set formula type EAS
          formulaType=EAS;
          
          // build automaton for alpha
          regExpr=SymbAutomatonLanguageAcceptor_ptr(new SymbAutomatonLanguageAcceptor(trimmed_expression.substr(2,pos-2)));
          
          // find or add subformula
          phi1=trimmed_expression.substr(pos+1,trimmed_expression.length()-pos-1);
          sub1=findOrAdd(phi1);
          
          // return the expression this formula represents
          newSpec = "E" + newSpec + "[" + newFormat(trimmed_expression.substr(2,pos-2)) + "]" + sub1->newSpec;
          if (!error) return trim(expression);
        } else error=true;
      }
      
      // looking for EX, EF, EG, EU, ER
      else if (!error&!((trimmed_expression[2]!='[')||(trimmed_expression[trimmed_expression.length()-1]!=']'))){
        // set formula type
        if (trimmed_expression[1]=='X') formulaType=EX;
        else if (trimmed_expression[1]=='F') formulaType=EF;
        else if (trimmed_expression[1]=='G') formulaType=EG;
        else if (trimmed_expression[1]=='U') formulaType=EU;
        else if (trimmed_expression[1]=='R') formulaType=ER;
        
        // find or add subformula(s)
        if (formulaType==EU || formulaType==ER){
          pos=findOperatorPosition(trimmed_expression.substr(3,trimmed_expression.length()-4),",");
          phi1=trimmed_expression.substr(3,pos);
          sub1=findOrAdd(phi1);  
          phi2=trimmed_expression.substr(pos+4,trimmed_expression.length()-5-pos);
          sub2=findOrAdd(phi2);
          newSpec = trimmed_expression.substr(0,1) + newSpec + trimmed_expression.substr(1,1) + "[" + sub1->newSpec + "," + sub2->newSpec + "]"; 
        } else{
          phi1=trimmed_expression.substr(3,trimmed_expression.length()-4);
          sub1=findOrAdd(phi1);
          newSpec = trimmed_expression[0] + newSpec + trimmed_expression[1] + "[" + sub1->newSpec + "]";
        }
        
        // return the expression this formula represents
        if (!error) return trim(expression);
      }
    }
    
    // and now the same for all formulas of the form A\psi
    // A<alpha>phi:
    else if (trimmed_expression[0]=='A'){
      
      // preprocessing list of ports for ASL-like formulas A{P1,P2,..,Pn}\psi
      if (trimmed_expression[1]=='{'){
        isASL=true;
        newSpec = "{";
        pos=findMatchingClosing(trimmed_expression, '{', '}');
        if (pos!=string::npos){
          string work=trimmed_expression.substr(2,pos-2);
          while (work.find(',')!=string::npos){
            controllablePorts.insert(work.substr(0,work.find(',')));
            newSpec = newSpec + work.substr(0,work.find(',')) + ",";
            work.erase(0,work.find(',')+1);
          }
          if (work.length()){
            controllablePorts.insert(work);
            newSpec = newSpec + work;
          }
          newSpec = newSpec + "}";
          trimmed_expression.erase(1,pos);
        } else error=true;
      }
      
      // A<alpha>phi:
      if (trimmed_expression[1]=='<'){
        pos=findMatchingClosing(trimmed_expression, '<', '>');
        if (pos!=string::npos){
          // set formula type AES
          formulaType=AES;
          
          // build automaton for alpha
          regExpr=SymbAutomatonLanguageAcceptor_ptr(new SymbAutomatonLanguageAcceptor(trimmed_expression.substr(2,pos-2)));
          
          // find or add subformula
          phi1=trimmed_expression.substr(pos+1,trimmed_expression.length()-pos-1);
          sub1=findOrAdd(phi1);
          
          // return the expression this formula represents
          newSpec = "A" + newSpec + "<" + newFormat(trimmed_expression.substr(2,pos-2)) + ">" + sub1->newSpec;
          if (!error) return trim(expression);
        } else error=true;
      }
      
      // A[alpha]phi:
      else if (trimmed_expression[1]=='['){
        pos=findMatchingClosing(trimmed_expression, '[', ']');
        if (pos!=string::npos){
          // set formula type AAS
          formulaType=AAS;
          
          // build automaton for alpha
          regExpr=SymbAutomatonLanguageAcceptor_ptr(new SymbAutomatonLanguageAcceptor(trimmed_expression.substr(2,pos-2)));
          
          // find or add subformula
          phi1=trimmed_expression.substr(pos+1,trimmed_expression.length()-pos-1);
          sub1=findOrAdd(phi1);
          
          // return the expression this formula represent
          newSpec = "A" + newSpec + "[" + newFormat(trimmed_expression.substr(2,pos-2)) + "]" + sub1->newSpec;
          if (!error) return trim(expression);
        } else error=true;
      }
      
      // looking for AX, AF, AG, AU, AR
      else if (!(error|=(trimmed_expression[2]!='[')||(trimmed_expression[trimmed_expression.length()-1]!=']'))){
        // set formula type
        if (trimmed_expression[1]=='X') formulaType=AX;
        else if (trimmed_expression[1]=='F') formulaType=AF;
        else if (trimmed_expression[1]=='G') formulaType=AG;
        else if (trimmed_expression[1]=='U') formulaType=AU;
        else if (trimmed_expression[1]=='R') formulaType=AR;
        
        // find or add subformula(s)
        if (formulaType==AU || formulaType==AR){
          pos=findOperatorPosition(trimmed_expression.substr(3,trimmed_expression.length()-4),",");
          phi1=trimmed_expression.substr(3,pos);
          sub1=findOrAdd(phi1);  
          phi2=trimmed_expression.substr(pos+4,trimmed_expression.length()-5-pos);
          sub2=findOrAdd(phi2);
          newSpec = trimmed_expression[0] + newSpec + trimmed_expression[1] + "[" + sub1->newSpec + "," + sub2->newSpec + "]"; 
        } else{
          phi1=trimmed_expression.substr(3,trimmed_expression.length()-4);
          sub1=findOrAdd(phi1);
          newSpec = trimmed_expression[0] + newSpec + trimmed_expression[1] + "[" + sub1->newSpec + "]"; 
        }
        
        // return the expression this formula represents
        if (!error) return trim(expression);
      }
    }
  }
  
  // remove brackets (phi)
  if ((trimmed_expression[0]=='(') && (trimmed_expression[trimmed_expression.length()-1]==')') && !error){ 
    return parse(trimmed_expression.substr(1,trimmed_expression.length()-2));
  }
  
  // negation !phi
  if (trimmed_expression[0]=='!'){
    // set formula type
    formulaType=NEG;
    
    // find or add subformula
    sub1=findOrAdd(trimmed_expression.substr(1,trimmed_expression.length()-1));
    
    // return the expression this formula represents
    newSpec = "!" + sub1->newSpec;
    if (!error) return trim(expression);
  }
  
  // the constant true formula
  else if (trimmed_expression=="true"){
    // set formula type
    formulaType=ONE;
    
    // the satisfaction set is constant
    satSet.value=BDDFunction::ONE();
    
    // return the expression this formula represents
    newSpec = expression;
    return expression;
  }
  
  // the constant false formula
  else if (trimmed_expression=="false"){
    // set formula type
    formulaType=ZERO;
    
    // the satisfaction set is constant
    satSet.value=BDDFunction::ZERO();
    
    // the satisfaction set is constant	
    newSpec = expression;
    return expression;
  }
  
  if (error) throw "Error parsing expression: " + expression;
  
  // and now the atomic case finally,
  // if the symbol is valid will be determined later according to the current CA
  formulaType=ATOM;
  satSet.value=BDDFunction::ZERO();
  satSet.valid=false;
  
  newSpec = "\"" + expression + "\"";
  return expression;
}

/**
 * \brief This method cares about system requests to add a formula
 *
 * Whenever a new (sub)formula should be created this methods starts a lookup to see if such formula already exists. (find operation)
 * If this is not the case a new formula will be created and added to the map of all formulas. (add operation) 
 * The parameter indicates the expression standing for the formula.
 * The method return a pointer to the (found or added) formula.
 * \sa mcreo.h consts.h
 */
formula::ptr formula::findOrAdd(const string& expr){
  // first formula instantiates a static map for all formulas
  if (formula::allFormulas==NULL) formula::allFormulas=new formula_map;
  
  string trimmed_expression=trim(expr);
  formula::ptr fptr;
  
  if (formula::allFormulas->count(trimmed_expression)>0) {
    // formula already exists -> return pointer from map
    formula::ptr stored_ptr=get_shared_ptr_from_weak_ptr((*formula::allFormulas)[trimmed_expression]);
    if (stored_ptr) {
      fptr=stored_ptr;
    }
  }
  
  if (!fptr) {
    // either formula is not in map or is expired
    fptr=formula::ptr(new formula(trimmed_expression));
    (*formula::allFormulas)[trimmed_expression]=fptr;
  }
  return fptr;
}

SymbAutomatonLanguageAcceptor_ptr formula::buildEA(const SymbAutomaton::ptr system, const bool det){
  if (regExpr){
    regExpr->parse(system); 
    if (det) regExpr->makeDeterministic();
    TRACE("EA","DOT", ofstream out(string("ea_test.dot").c_str()); regExpr->toDot(out));
  }
  
  SymbAutomatonLanguageAcceptor_ptr result = SymbAutomatonLanguageAcceptor_ptr(regExpr);
  return result;
}

SymbAutomatonLanguageAcceptor_ptr formula::composeEA(const SymbAutomaton::ptr system, const bool det){
  if (regExprRootNode){
    hybridFiniteAutomaton::ptr hfa = hybridFiniteAutomaton::toHybridFiniteAutomaton(*regExprRootNode,system);
        
    if(det){
      if (minimazationMethod == MIN_BRZOZOWSKI){
        hfa->minimizeBrzozowski();
      } else if (minimazationMethod == MIN_HOPCROFT){
        hfa->makeDeterministic();
        hfa->minimizeHopcroft();
      } else hfa->makeDeterministic();
    }
    
    hfa->removeUselessStates();
        
    regExpr = hfa->encode(system);
    TRACE("FA","DOT", ofstream out(string("fa_test.dot").c_str()); regExpr->toDot(out));
  }
  
  return regExpr;
}

void formula::createNewWitnessFor(formula::ptr f, SymbAutomaton::ptr a){
  wit_least=boost::shared_ptr<witness>(new witness(f,a));
}

void formula::appendToWitness(const BDDFunction& f){
  if (!wit_least) throw "No witness given!";
  wit_least->append(f);
}

void formula::createSecondWitnessFor(formula::ptr f, SymbAutomaton::ptr a, const f_info& ft, const BDDFunction& x){
  wit_greatest=boost::shared_ptr<witness>(new witness(f,a));
  wit_greatest->setFixpointType(ft);
  wit_greatest->append(x);
}
