/*
 * This file is part of the program vereofy.
 * Copyright (C) 2008-2009 Joachim Klein <klein.inf.tu-dresden.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 *  published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "LTLInfixParser.hpp"
#include "ltl_infix_parser.hpp"

#include "vereofy/common/errors.hpp"
#include "vereofy/common/tracing.hpp"

using namespace std;
using namespace vereofy::ltl_check;

template <class Node>
static
LTLNode_p toLTL_Node(const Node& node, 
                     APSet& apset,
                     APSet_cp predefined_apset) {
  LTLNode_p left, right;
  if (node.children.size()>=1) {
    left=toLTL_Node(node.children.at(0),
                    apset, predefined_apset);
  }
  if (node.children.size()==2) {
    right=toLTL_Node(node.children.at(1),
                    apset, predefined_apset);
  }

  switch (node.value.id().to_long()) {
  case ltl_infix_parser::ID_LTL_const_true:
    return LTLNode_p(new LTLNode(LTLNode::T_TRUE, 
                                 LTLNode_p(),
                                 LTLNode_p()));

  case ltl_infix_parser::ID_LTL_const_false:
    return LTLNode_p(new LTLNode(LTLNode::T_FALSE, 
                                 LTLNode_p(), 
                                 LTLNode_p()));

  case ltl_infix_parser::ID_LTL_term: {
    string value(node.value.begin(), node.value.end());
    if (value=="true" || value=="TRUE" || value=="t" || value=="1") {
      return LTLNode_p(new LTLNode(LTLNode::T_TRUE, 
                                   LTLNode_p(),
                                   LTLNode_p()));
    } else if (value=="false" || value=="FALSE" || value=="f" || value=="0") {
      return LTLNode_p(new LTLNode(LTLNode::T_FALSE, 
                                   LTLNode_p(),
                                   LTLNode_p()));
    } else {
      THROW_UNIMPLEMENTED();
    }
  }
  case ltl_infix_parser::ID_LTL_not_op:
    return LTLNode_p(new LTLNode(LTLNode::T_NOT, left));

  case ltl_infix_parser::ID_LTL_or_op: 
    return LTLNode_p(new LTLNode(LTLNode::T_OR, left, right));

  case ltl_infix_parser::ID_LTL_and_op:
    return LTLNode_p(new LTLNode(LTLNode::T_AND, left, right));
    
  case ltl_infix_parser::ID_LTL_xor_op:
    // a ^ b <=> !(a<->b)
    return LTLNode_p(new LTLNode(LTLNode::T_NOT, LTLNode_p(new LTLNode(LTLNode::T_EQUIV, left, right))));

  case ltl_infix_parser::ID_LTL_implies_op:
    return LTLNode_p(new LTLNode(LTLNode::T_IMPLICATE, left, right));

  case ltl_infix_parser::ID_LTL_equivalent_op:
    return LTLNode_p(new LTLNode(LTLNode::T_EQUIV, left, right));
    
  case ltl_infix_parser::ID_LTL_next_op: 
    return LTLNode_p(new LTLNode(LTLNode::T_NEXTSTEP, left));

  case ltl_infix_parser::ID_LTL_always_op:
    return LTLNode_p(new LTLNode(LTLNode::T_GLOBALLY, left));
    
  case ltl_infix_parser::ID_LTL_eventually_op:
    return LTLNode_p(new LTLNode(LTLNode::T_FINALLY, left));

  case ltl_infix_parser::ID_LTL_until_op:
    return LTLNode_p(new LTLNode(LTLNode::T_UNTIL, left, right));

  case ltl_infix_parser::ID_LTL_weak_until_op:
    return LTLNode_p(new LTLNode(LTLNode::T_WEAKUNTIL, left, right));

  case ltl_infix_parser::ID_LTL_release_op:
    return LTLNode_p(new LTLNode(LTLNode::T_RELEASE, left, right));

  case ltl_infix_parser::ID_LTL_io: {
    // <left>right = left & X right
    return LTLNode_p(new LTLNode(LTLNode::T_AND,
                                 left,
                                 LTLNode_p(new LTLNode(LTLNode::T_NEXTSTEP, right))));
  }

  case ltl_infix_parser::ID_LTL_ap: {
    int ap_i; // the AP index
    string ap(node.value.begin(),
              node.value.end());

    CHECK_SUCCESS(ap!="");

    if (predefined_apset!=0) {
      if ((ap_i=predefined_apset->find(ap))!=-1) {
	return LTLNode_p(new LTLNode(ap_i));
      } else {
	// not found in predefined APSet!
	throw std::string("Atomic proposition is not defined!");
      }
    } else {
      if ((ap_i=apset.find(ap))!=-1) {
	// AP exists already
	return LTLNode_p(new LTLNode(ap_i));
      } else {
	// create new AP
	ap_i=apset.addAP(ap);
	return LTLNode_p(new LTLNode(ap_i));
      }
    }
  }
  }
  
  THROW_UNIMPLEMENTED();
}

LTLFormula_ptr LTLInfixParser::parse(std::string const& formula,
                                     APSet_cp predefined_apset) {
    APSet_p apset(new APSet());
    
    ltl_infix_parser infix_p;
    tree_parse_info<> info = ast_parse(formula.c_str(),
                                       //                                       formula.end(),
                                       infix_p >> end_p, space_p);
    if (!info.full) {
      throw "Parse error, LTL formula";
    }

    // dump parse tree as XML
    TRACE("LTLInfixParser", "ParseTreeXML",     \
          tree_to_xml(cout, info.trees, "", infix_p.getRuleNames()); \
          );
    
    LTLNode_p ltl=toLTL_Node(info.trees[0], *apset, predefined_apset);
    
    APSet_cp apset_=(predefined_apset.get()!=0)
      ? predefined_apset : 
      boost::const_pointer_cast<APSet const>(apset);
    
    return LTLFormula_ptr(new LTLFormula(ltl, apset_));
}



