/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef EXECUTION_ENVIRONMENT_HPP
#define EXECUTION_ENVIRONMENT_HPP

#include "SymbolTable.hpp"
#include "SymbolTableStack.hpp"

#include <iostream>
#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include <vector>

namespace vereofy {
  class REO_Circuit;
  class SystemContext;
}

template <class Value_ptr, class symbol_table_t>
class ExecutionEnvironment {
public:
  typedef boost::shared_ptr<ExecutionEnvironment> ptr;
  
  typedef vector<Value_ptr> value_vector_t;
  typedef boost::shared_ptr<value_vector_t> value_vector_ptr;

  typedef SymbolTable<value_vector_ptr> value_table_t;
  typedef typename value_table_t::ptr value_table_ptr;

  typedef typename value_table_t::iterator iterator;

  ExecutionEnvironment(typename symbol_table_t::ptr global_symbols_, 
                       typename symbol_table_t::ptr local_symbols_, 
                       boost::shared_ptr<vereofy::SystemContext> system_context_,
                       value_table_ptr value_table_=value_table_ptr(new value_table_t()),
                       bool debug_=false) : 
    value_table(value_table_),
    global_symbols(global_symbols_), local_symbols(local_symbols_),
    system_context(system_context_), debug(debug_), debugStream(&cerr) {

    symbol_table_stack.push(global_symbols);
    symbol_table_stack.push(local_symbols);
  }

  bool flagDebug() const {
    return debug;
  }
  
  void setDebugStream(ostream& out) {
    debugStream=&out;
  }

  ostream& getDebugStream() {
    if (!debugStream) {
      throw "debugStream: NULL pointer!";
    }

    return *debugStream;
  }


  iterator begin() {return value_table.begin();}
  iterator end() {return value_table.end();}

  typename symbol_table_t::ptr getGlobalSymbols() {
    return global_symbols;
  }

  typename symbol_table_t::ptr getLocalSymbols() {
    return local_symbols;
  }

  SymbolTableStack<symbol_table_t> &getSymbolTableStack() {
    return symbol_table_stack;
  }

  boost::shared_ptr<vereofy::SystemContext> getSystemContext() {
    return system_context;
  }

  boost::shared_ptr<vereofy::REO_Circuit> getREOCircuit() {
    return reo_circuit;
  }

  void setREOCircuit(boost::shared_ptr<vereofy::REO_Circuit> circuit) {
    reo_circuit=circuit;
  }


  void setValue(const string& name, unsigned int index, Value_ptr value) {
    TRACE("ExecutionEnvironment", "setValue", \
          std::cout << "setValue(" << name << "[" << index << "], "; \
          value->print(std::cout);                                   \
          std::cout << ")\n";);

    value_vector_ptr v_vec=value_table->find(name);
    if (!v_vec) {
      v_vec=value_vector_ptr(new value_vector_t);
      value_table->add(name, v_vec);
    }

    if (index>=v_vec->size()) {
        v_vec->resize(index+1);
    }
    
    v_vec->at(index)=value;
  }

  unsigned int setValueNextFree(const string& name, Value_ptr value) {
    value_vector_ptr v_vec=getValueVector(name);
    unsigned int index;
    if (!v_vec) {
      index=0;
    } else {
      index=v_vec->size();
    }
    setValue(name, index, value);
    return index;
  }

  bool VarExists(const string& name) {
    value_vector_ptr v_vec=value_table->find(name);

    bool exists=v_vec;
    return exists;
  }

  bool VarIsNull(const string& name, unsigned int index) {
    value_vector_ptr v_vec=value_table->find(name);

    if (!v_vec) {
      throw string("Illegal access: Variable does not exist: ")+name;
    }

    if (index>=v_vec->size()) {
      return true;
    }

    bool is_null=!(v_vec->at(index));
    return is_null;
  }
  
  Value_ptr getValue(const string& name, unsigned int index) {
    value_vector_ptr v_vec=value_table->find(name);

    if (!v_vec) {
      throw string("Illegal access: Variable does not exist: ")+name;
    }

    if (index>=v_vec->size()) {
      throw string("Illegal access: Out of bounds! ")+name+"["+boost::lexical_cast<string>(index)+"]";
    }

    Value_ptr v=v_vec->at(index);

    TRACE("ExecutionEnvironment", "getValue", \
          std::cout << "getValue(" << name << "[" << index << "]) = "; \
          if (v) { v->print(std::cout); } else {std::cout << "NULL\n";} \
          cout << endl; \
          );

    return v;
  }

  value_vector_ptr getValueVector(const string& name) {
    return value_table->find(name);
  }
  
  value_table_ptr getValueTable() {
    return value_table;
  }
  
  void print(ostream& out) {
    print_valuetable(out, *value_table);
  }
  
  static 
  void print_valuetable(ostream& out, value_table_t& table, bool detailed=false) {
    for (typename value_table_t::symbol_map::iterator it=table.begin();
         it!=table.end();
         ++it) {
      string name=(*it).first;

      if (name=="<this>") {continue;}

      value_vector_ptr v_vec=(*it).second;
      if (v_vec) {
        for (size_t i=0;i<v_vec->size();i++) {
          out << name << "[" << i <<"] : ";
          Value_ptr value=v_vec->at(i);
          if (value) {
            value->print(out, true);
          } else {
            out << "NULL";
          }
          out << "\n";
        }
      }
    }
  }
  
private:
  value_table_ptr value_table;
  typename symbol_table_t::ptr global_symbols;
  typename symbol_table_t::ptr local_symbols;
  SymbolTableStack<symbol_table_t> symbol_table_stack;
  boost::shared_ptr<vereofy::SystemContext> system_context;

  boost::shared_ptr<vereofy::REO_Circuit> reo_circuit;

  bool debug;
  ostream* debugStream; // reference to debug stream
};


#endif  
