/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "spirit_init.hpp"

#include "REOParser.hpp"
#include "REOParseException.hpp"
#include "reo_parser.hpp"
#include "skip_parser.hpp"

#include <iostream>

namespace vereofy {
namespace vparser {

REOParser::REOParser(SystemContext::wptr system_context_) :
  system_context(system_context_) {
}

template <class Node>
void REOParser::add_parse_trees(const Node& top) {
  for (size_t i=0; i<top.children.size(); i++) {
    const Node& def=top.children.at(i);
    switch (def.value.id().to_long()) {
    case ID_C_circuit_def: {
      string name=circuit_helper::t_circuit_def::getName(def);

      if (debugLevel>=1) {
        cout << "Add " << name <<"\n";
      }
      typename CircuitInfo::ptr ci=boost::static_pointer_cast<CircuitInfo>(access_weak_ptr(system_context)->getGlobalSymbols()->find(name));
      ci->connector->behavior->parse_tree=def;
      break;
    }
#ifdef REO_PARSER_WITH_MODULES
    case ID_M_module_def: {
      string name=module_helper::t_module_def::getName(def);
      typename ModuleInfo::ptr mi=boost::static_pointer_cast<ModuleInfo>(access_weak_ptr(system_context)->getGlobalSymbols()->find(name));
      
      if (debugLevel>=1) {
        cout << "Add " << name <<"\n";
      }
      mi->module->behavior->parse_tree=def;
      break;
    }
#endif
    case ID_T_type_def:
    case ID_CO_const_def:
    case ID_C_include:
    case ID_C_alias_def:
    case ID_C_replace_def:
    case ID_C_global_ap_def:
    case ID_C_function_def:
      // ignore
      break;
    default:
      THROW_UNIMPLEMENTED();
    }
  }
}

bool REOParser::parse_file(const string& filename) {
  ifstream file_in(filename.c_str());
  if (!file_in) {
    std::cerr << "Unable to open file: " << filename << "\n";
    return false;
  }

  stringstream buf;
  // read contents into buffer
  while (file_in.good()) {
    const size_t BUF_SIZE=100;
    char buffer[BUF_SIZE];
    file_in.read(buffer, BUF_SIZE-1);
    streamsize read=file_in.gcount();
    if (read<0 || (unsigned long)read>=BUF_SIZE) {throw "implementation error";}

    // terminate buffer
    buffer[read]='\0';
    buf << buffer;
  }

  string text=buf.str();
  return parse(text, filename);
}

bool REOParser::parse(const string& text,
                      const string& filename) {

  symbol_table_stack_t symbol_table_stack;
  symbol_table_stack.push(access_weak_ptr(system_context)->getGlobalSymbols());

  reo_parser parser(symbol_table_stack, system_context);


  const char *c_text=text.c_str();
  iterator_t first(c_text, c_text + text.length(), filename);
  iterator_t last;

  skip_parser skip_grammar(access_weak_ptr(system_context)->getFlagMap());
  
  try {
    tree_parse_info<iterator_t, REO_PARSER_NODE_FACTORY > info
      = ast_parse<REO_PARSER_NODE_FACTORY> (first, last, parser >> end_p, 
                                            skip_grammar);
    if (info.full)
      {
#if defined(BOOST_SPIRIT_DUMP_PARSETREE_AS_XML)
        if (debugLevel>=4) {
          // dump parse tree as XML
          std::map<parser_id, std::string> rule_names;
          reo_parser::init_rule_names(rule_names);
          
          tree_to_xml(cout, info.trees, "", rule_names);
        }
#endif

        // print the result
        if (debugLevel>=2) {
          std::cout << "Parsing successful\n";
        }

        add_parse_trees(info.trees[0]);
      } else {
        cerr << "parsing failed\n";
        return false;
      }
  } catch (parser_error<string, iterator_t>& error) {
    REOParseException rpe(error.descriptor);
#ifdef REO_PARSER_POSITION_NODES
    file_position pos=error.where.get_position();
    rpe.setFile(pos.file);
    rpe.setLocation("("+
                    boost::lexical_cast<string>(pos.line)+
                    ","+
                    boost::lexical_cast<string>(pos.column)+
                    ")");
#endif
    rpe.raise();
  }
  return true;
}


pt_node REOParser::parse_io_guard(const string& guard, 
                                  const ComponentInterface::ptr c_interface,
                                  bool allow_state_references) const {
  return parse_guard(guard_parser::io_guard, guard, c_interface, allow_state_references);
}

pt_node REOParser::parse_state_guard(const string& guard, 
                                     const ComponentInterface::ptr c_interface) const {
  return parse_guard(guard_parser::state_guard, guard, c_interface);
}


pt_node REOParser::parse_guard(unsigned int type,
                               const string& guard, 
                               const ComponentInterface::ptr c_interface,
                               bool allow_state_references_for_io_guard) const {
  symbol_table_stack_t symbol_table_stack(c_interface->getSymbolTableStack());

  symbols<> reserved_keywords;
  guard_parser::init_reserved_keywords(reserved_keywords);
  guard_parser guard_grammar(symbol_table_stack, reserved_keywords, c_interface);

  guard_grammar.initialize(false, false, false);

  switch ((guard_parser::parser_type)type) {
  case guard_parser::io_guard: 
    guard_grammar.strong_guard=true;
    guard_grammar.allow_ports=true;
    guard_grammar.allow_labels=true;
    guard_grammar.allow_states=allow_state_references_for_io_guard;
    guard_grammar.allow_bool_expression_state_guard=false;
    guard_grammar.allow_bool_expression_io_guard=true;
    break;
  case guard_parser::state_guard:
    guard_grammar.strong_guard=true;
    guard_grammar.allow_ports=false;
    guard_grammar.allow_labels=false;
    guard_grammar.allow_states=true;
    guard_grammar.allow_bool_expression_state_guard=true;
    guard_grammar.allow_bool_expression_io_guard=false;
    break;
  default:
    THROW_UNIMPLEMENTED();
  }


  const char *c_guard=guard.c_str();

  iterator_t first(c_guard, c_guard + guard.length());
  iterator_t last;

  tree_parse_info<iterator_t, REO_PARSER_NODE_FACTORY > info;
  skip_parser skip_grammar(access_weak_ptr(system_context)->getFlagMap());

  try {
    switch ((guard_parser::parser_type)type) {
    case guard_parser::io_guard: 
      info = ast_parse<REO_PARSER_NODE_FACTORY> (first, last, 
                                                 guard_grammar.use_parser<guard_parser::bool_expression>() >> end_p, 
                                                 skip_grammar);
      break;
    case guard_parser::state_guard:
      info = ast_parse<REO_PARSER_NODE_FACTORY> (first, last, 
                                                 guard_grammar.use_parser<guard_parser::bool_expression>() >> end_p, 
                                                 skip_grammar);
      break;
    default:
      THROW_UNIMPLEMENTED();
    }
    
    if (info.full) {
#if defined(BOOST_SPIRIT_DUMP_PARSETREE_AS_XML)
      if (debugLevel>=4) {
        // dump parse tree as XML
        std::map<parser_id, std::string> rule_names;
        reo_parser::init_rule_names(rule_names);
        
        tree_to_xml(cout, info.trees, "", rule_names);
      }
#endif
      
      return info.trees[0];
    }
  } catch (parser_error<string, iterator_t>& error) {
    REOParseException rpe(error.descriptor);
#ifdef REO_PARSER_POSITION_NODES
    file_position pos=error.where.get_position();
    if (pos.line==1) {
      rpe.setLocation("column "+
                      boost::lexical_cast<string>(pos.column));
    } else {
      rpe.setLocation("("+
                      boost::lexical_cast<string>(pos.line)+
                      ","+
                      boost::lexical_cast<string>(pos.column)+
                      ")");
    }
#endif
    rpe.raise();
  }

  throw "parse not successful!";
}


void REOParser::parse_and_add_typedef(const string& type_def) {
  symbol_table_stack_t symbol_table_stack;
  symbol_table_stack.push(access_weak_ptr(system_context)->getGlobalSymbols());

  symbols<> reserved_keywords;
  type_parser type_grammar(symbol_table_stack, reserved_keywords);

  skip_parser skip_grammar(access_weak_ptr(system_context)->getFlagMap());

  const char *c_type_def=type_def.c_str();

  iterator_t first(c_type_def, c_type_def + type_def.length());
  iterator_t last;

  tree_parse_info<iterator_t, REO_PARSER_NODE_FACTORY > 
    info = ast_parse<REO_PARSER_NODE_FACTORY> (first, last, 
                                               type_grammar.use_parser<type_parser::type_def>() >> end_p, 
                                               skip_grammar);

  if (info.full) {
    // sucessful
    return;
  }

  throw "parse not sucessful";
}


void REOParser::parse_and_add_constdef(const string& const_def) {
  symbol_table_stack_t symbol_table_stack;
  symbol_table_stack.push(access_weak_ptr(system_context)->getGlobalSymbols());

  symbols<> reserved_keywords;
  const_def_parser const_def_grammar(symbol_table_stack, reserved_keywords, true);

  const char *c_const_def=const_def.c_str();

  iterator_t first(c_const_def, c_const_def + const_def.length());
  iterator_t last;

  skip_parser skip_grammar(access_weak_ptr(system_context)->getFlagMap());

  tree_parse_info<iterator_t, REO_PARSER_NODE_FACTORY > 
    info = ast_parse<REO_PARSER_NODE_FACTORY> (first, last, 
                                               const_def_grammar >> end_p, 
                                               skip_grammar);

  if (info.full) {
    // sucessful
    return;
  }

  throw "parse not sucessful";
}



// wrapper construct for subtree_to_xml
struct node_container_t {
    
  class const_iterator
    : public boost::iterator_facade<
    const_iterator
    , pt_node const
    , boost::forward_traversal_tag
    >
  {
  public:
    const_iterator() : m_node(0) {}
    explicit const_iterator(const pt_node& n) : m_node(&n) {}
    
  private:
    friend class boost::iterator_core_access;
    
    void increment() { m_node = 0; }    
    bool equal(const_iterator const& other) const {
      return this->m_node == other.m_node;
    }
    pt_node const& dereference() const { return *m_node; }

    pt_node const* m_node;
  };

  node_container_t(const pt_node& n_) : n(n_) {}

  const_iterator begin() const {return const_iterator(n);}
  const_iterator end() const {return const_iterator();}

  const pt_node& n;
};
  
  
void REOParser::subtree_to_xml(std::ostream& out, const pt_node& n) {
  std::map<parser_id, std::string> rule_names;
  reo_parser::init_rule_names(rule_names);

  tree_to_xml(out, node_container_t(n), "", rule_names);
}


}
}
