/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef CIRCUIT_PARSER_HPP
#define CIRCUIT_PARSER_HPP

#include "spirit_init.hpp"
using namespace vereofy::spirit;

#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>

#include <stack>
#include <string>

#include "reo_parser_types.hpp"
#include "circuit_types.hpp"

#include "position_node.hpp"
#include "position_node_factory.hpp"

#include "vereofy/common/singleton_set.hpp"

#include "parser_helper.hpp"
#include "circuit_parser_helper.hpp"
#include "common_parser.hpp"
#include "int_parser.hpp"

namespace vereofy {
namespace vparser {

struct circuit_parser : public grammar<circuit_parser>
{
 
  static 
  void init_rule_names(map<parser_id, std::string>& rule_names) {
    rule_names[ID_C_null_value] = "null_value";
    rule_names[ID_C_rsl_factor] = "rsl_factor";
    rule_names[ID_C_rsl_term] = "rsl_term";
    rule_names[ID_C_rsl_expression] = "rsl_expression";
    rule_names[ID_C_var_expression] = "var_expression";
    
    rule_names[ID_C_rsl_bool_simple] = "rsl_bool_simple";
    rule_names[ID_C_rsl_bool_factor] = "rsl_bool_factor";
    rule_names[ID_C_rsl_bool_and] = "rsl_bool_and";
    rule_names[ID_C_rsl_bool_or] = "rsl_bool_or";
    rule_names[ID_C_rsl_bool_expression] = "rsl_bool_expression";
    
    rule_names[ID_C_lhs_var] = "lhs_var";
    rule_names[ID_C_rhs_var] = "rhs_var";
    rule_names[ID_C_array_ref] = "array_ref";
    
    rule_names[ID_C_if_statement] = "if_statement";
    rule_names[ID_C_for_statement] = "for_statement";
    rule_names[ID_C_join_expression] = "join_expression";
    rule_names[ID_C_connector_spec] = "connector_spec";
    rule_names[ID_C_parameter_list] = "parameter_list";
    rule_names[ID_C_configuration_block] = "configuration_block";
    rule_names[ID_C_with_pair] = "with_pair";
    rule_names[ID_C_with_block] = "with_block";

    
    rule_names[ID_C_assignment] = "assignment";
    rule_names[ID_C_instantiation] = "instantiation";
    rule_names[ID_C_node_instantiation] = "node_instantiation";
    rule_names[ID_C_ap_statement] = "ap_statement";
    rule_names[ID_C_statement] = "statement";
    rule_names[ID_C_statements] = "statements";
    rule_names[ID_C_statement_block] = "statement_block";
    rule_names[ID_C_circuit_body] = "circuit_body";
    rule_names[ID_C_circuit_def] = "circuit_def";
  }

  static
  void init_reserved_keywords(symbols<>& keywords) {
    keywords="CIRCUIT", "new", "join", "for", "if", "else", "NULL", "CONFIG";
  }


  enum Errors {
    no_statement,
    no_assignment,
    no_left_curly,
    no_right_curly,
    no_left_paren,
    no_right_paren,
    no_semicolon,
    no_comma,
    no_equal
  };

  struct expect_parsers {
    expect_parsers() :
      statement(no_statement),
      assignment(no_assignment),
      left_curly(no_left_curly),
      right_curly(no_right_curly),
      left_paren(no_left_paren),
      right_paren(no_right_paren),
      semicolon(no_semicolon),
      comma(no_comma),
      equal(no_equal)
    {}

    vereofy::spirit::assertion<Errors> statement;
    vereofy::spirit::assertion<Errors> assignment;
    vereofy::spirit::assertion<Errors> left_curly;
    vereofy::spirit::assertion<Errors> right_curly;
    vereofy::spirit::assertion<Errors> left_paren;
    vereofy::spirit::assertion<Errors> right_paren;
    vereofy::spirit::assertion<Errors> semicolon;
    vereofy::spirit::assertion<Errors> comma;
    vereofy::spirit::assertion<Errors> equal;
  };


  template <class DefT>
  struct error_handler {
    const DefT& def;

    error_handler(const DefT& def_) : def(def_) {}

    template <typename ScannerT, typename ErrorT>
    error_status<>
    operator()(ScannerT const& scan, ErrorT const& error) const {
      switch (error.descriptor) {
      case no_statement:
        throw_(error.where, std::string("Expected a valid statement"));
      case no_assignment:
        throw_(error.where, std::string("Expected an assignment"));
      case no_left_curly:
        throw_(error.where, std::string("Expected opening '{'"));
      case no_left_paren:
        throw_(error.where, std::string("Expected opening '('"));
      case no_right_curly:
        throw_(error.where, std::string("Expected opening '}'"));
      case no_right_paren:
        throw_(error.where, std::string("Expected opening ')'"));
      case no_semicolon:
        throw_(error.where, std::string("Expected ';'"));
      case no_comma:
        throw_(error.where, std::string("Expected ','"));
      case no_equal:
        throw_(error.where, std::string("Expected '='"));
      default:
        return error_status<>(error_status<>::rethrow);
      }
    }
  };


  struct begin_circuit_def {
    symbol_table_stack_t &symbol_table_stack;
    const SystemContext::wptr& system_context;
    
    begin_circuit_def(symbol_table_stack_t &symbol_table_stack_,
                      const SystemContext::wptr& system_context_)
      : symbol_table_stack(symbol_table_stack_),
        system_context(system_context_) {
      if (debugLevel>=5) {
        cerr << "ST(begin_circuit_def): " << &symbol_table_stack_<< endl;
        cerr << "ST(begin_circuit_def): " << &symbol_table_stack<< endl;
      }
    }
    
    template<class Iter>
    const begin_circuit_def&
    operator()(Iter& i1, Iter& i2) const {
      std::string name(i1, i2);
      
      if (symbol_table_stack.find(name)) {
        throw_(i1, "There is already a definition for "+name);
      }

      CircuitInfo::ptr ci(new CircuitInfo(name));
      ci->connector=CircuitConnector::ptr(new CircuitConnector(system_context, name));
      symbol_table_stack.getTopTable()->add(name, ci);

      return *this;
    }
  };

  struct store_pos {

    template<class Node, class Iter>
    void operator()(Node& n, Iter& i1, Iter& i2) const {
#ifdef REO_PARSER_POSITION_NODES
      n.value.setPosition(i1.get_position());
      //      cerr << "store_pos" << n.value.pos->line << ":" << n.value.pos->column << " [" << string(i1, i2) << "]\n";
#endif
    }    
  };

  static
  std::map<parser_id, std::string> getRuleNames() {
    std::map<parser_id, std::string> rule_names;
    
    init_rule_names(rule_names);
    return rule_names;
  };


  circuit_parser(symbol_table_stack_t& symbol_table_stack_,
                 const symbols<>& reserved_keywords_,
                 SystemContext::wptr system_context_) : 
    symbol_table_stack(symbol_table_stack_),
    reserved_keywords(reserved_keywords_),
    system_context(system_context_) {
    if (debugLevel>=5) {
      cerr << "ST(circuit_parser): " << &symbol_table_stack<< endl;
    }
  }

  symbol_table_stack_t& symbol_table_stack;
  const symbols<>& reserved_keywords;
  SystemContext::wptr system_context;
  
  template <typename ScannerT>
    struct definition
    {
      definition(circuit_parser const& self) :
        integer_grammar(self.symbol_table_stack),
        identifier_grammar(last_identifier, self.reserved_keywords),
        common_grammar(self.reserved_keywords)
      {

        BOOST_SPIRIT_DEBUG_GRAMMAR(integer_grammar);
        BOOST_SPIRIT_DEBUG_GRAMMAR(identifier_grammar);
        BOOST_SPIRIT_DEBUG_GRAMMAR(common_grammar);
      
        ENTRY_GRAMMAR(identifier_p, identifier_parser, identifier_grammar, identifier_parser::identifier);
        ENTRY_GRAMMAR(const_expression_p, int_parser, integer_grammar, int_parser::const_expression);
        ENTRY_GRAMMAR(integer_p, int_parser, integer_grammar, int_parser::integer);

        ENTRY_GRAMMAR(quoted_string, common_parser, common_grammar, common_parser::quoted_string);
        ENTRY_GRAMMAR(parameter_def_p, common_parser, common_grammar, common_parser::parameter_def);


        //  Start grammar definition

        null_value =   root_node_d[lexeme_d[str_p("NULL")]];

        rsl_factor      =   integer_p
                        |   const_expression_p
                        |   var_expression
                        |   (no_node_d[ch_p('(')] >> rsl_expression >> no_node_d[ch_p(')')])
                        |   (root_node_d[ch_p('-')] >> rsl_factor);

        rsl_term        =   rsl_factor >>
                            *(  (root_node_d[ch_p('*')] >> rsl_factor)
                              | (root_node_d[ch_p('/')] >> rsl_factor)
                              | (root_node_d[ch_p('%')] >> rsl_factor)
                            );

        rsl_expression  =   rsl_term >>
                            *(  (root_node_d[ch_p('+')] >> rsl_term)
                              | (root_node_d[ch_p('-')] >> rsl_term)
                            );

        rsl_bool_simple =
          rsl_expression 
              >> ( (root_node_d[str_p("==")] >> rsl_expression)
                  |(root_node_d[str_p("!=")] >> rsl_expression)
                  |(root_node_d[str_p("<=")] >> rsl_expression)
                  |(root_node_d[str_p("<")] >>  rsl_expression)
                  |(root_node_d[str_p(">=")] >> rsl_expression)
                  |(root_node_d[str_p(">")] >>  rsl_expression)
                 );

        rsl_bool_factor = 
          rsl_bool_simple
              |  (no_node_d[ch_p('(')] >> rsl_bool_expression >> no_node_d[ch_p(')')])
              | (root_node_d[ch_p('!')] >> rsl_bool_factor);

        rsl_bool_and = 
              rsl_bool_factor
              >>  *( ( root_node_d[ch_p('&')] >> rsl_bool_and));
            
        rsl_bool_or   = 
          rsl_bool_and 
              >>  *( ( root_node_d[ch_p('|')] >> rsl_bool_and));

        rsl_bool_expression  = rsl_bool_or;

                    

            array_ref   = no_node_d[ch_p('[')] >> root_node_d[rsl_expression] >> no_node_d[ch_p(']')];

            lhs_var     = 
                           (identifier_p 
                            >> !array_ref
                            );

            rhs_var     = identifier_p >> !array_ref;
            
            var_expression = rhs_var >>
              !( root_node_d[ch_p('.')]
                 >> var_expression);

            connector_spec = no_node_d[ch_p('(')] 
              >> !infix_node_d[list_p((var_expression|null_value), ch_p(','))]
              >> !(ch_p(';') >> infix_node_d[list_p((var_expression|null_value), ch_p(','))])
              >> no_node_d[ch_p(')')];

            with_pair = identifier_p >> no_node_d[ch_p('=')] >> (identifier_p | integer_p);

            with_block =  no_node_d[keyword_d[as_lower_d["with"]]]
              >> +(with_pair);

            node_instantiation = root_node_d[str_p("NODE") | str_p("ROUTE_NODE")]; 
            
            instantiation  = root_node_d[keyword_d[as_lower_d["new"]]]
                 >> identifier_p  // !!! check for existence of circuit/module
                 >> !parameter_list
                 >> !connector_spec
                 >> !with_block;
                       
            join_expression = 
                 root_node_d[keyword_d[as_lower_d["join"]]]
              >> no_node_d[expect.left_paren(ch_p('('))] 
              >> var_expression
              >> no_node_d[expect.comma(ch_p(','))]
              >> var_expression
              >> no_node_d[expect.right_paren(ch_p(')'))];

            assignment = 
              lhs_var 
              >> root_node_d[expect.equal(ch_p('='))]
              >> (  null_value
                    | node_instantiation
                    | instantiation
                    | join_expression
                    | rsl_expression );

            

            
            if_statement  =
                  root_node_d[keyword_d[as_lower_d["if"]]]
              >>  no_node_d[expect.left_paren(ch_p('('))]
              >>  rsl_bool_expression
              >>  no_node_d[expect.right_paren(ch_p(')'))]
              >>  statement_block
              >> !(   keyword_d[as_lower_d["else"]]
                   >> statement_block
                  );

            for_statement = //access_node_d[
                 root_node_d[keyword_d[as_lower_d["for"]]]
              >> no_node_d[expect.left_paren(ch_p('('))]
              >> expect.assignment(assignment)
              >> no_node_d[expect.semicolon(ch_p(';'))]
              >> rsl_bool_expression
              >> no_node_d[expect.semicolon(ch_p(';'))]
              >> expect.assignment(assignment)
              >> no_node_d[expect.right_paren(ch_p(')'))]
              >> statement_block;
              //];// [do_for];
              

            ap_statement = 
              root_node_d[keyword_d[as_lower_d["ap"]]]
              >> no_node_d[expect.left_paren(ch_p('('))]
              >> quoted_string
              >> no_node_d[expect.comma(ch_p(','))]
              >> quoted_string
              >> no_node_d[expect.right_paren(ch_p(')'))];
              

            statement = 
              access_node_d[
                 (  
                  join_expression >> no_node_d[ch_p(';')]
                    | for_statement
                    | if_statement
                    | ap_statement >> no_node_d[ch_p(";")]
                    | instantiation >> no_node_d[ch_p(';')]
                    | assignment >> no_node_d[ch_p(';')] // assignment has to come last!
                 )][store_pos()];

            
            statement_block = expect.left_curly(no_node_d[ch_p('{')])
              >> statements
              >> expect.right_curly(no_node_d[ch_p('}')]);

            statements  =   *(~eps_p(ch_p('}') | keyword_d["TOPO"]) >> expect.statement(statement));

            parameter_list = no_node_d[ch_p('<')] >> infix_node_d[list_p(rsl_expression, ch_p(','))] >> no_node_d[ch_p('>')];

            configuration_block =
              no_node_d[keyword_d["TOPO"]]
              >> no_node_d[ch_p('(')]
              >> root_node_d[int_p]
              >> no_node_d[ch_p(')')]
              >> no_node_d[ch_p('=')]
              >> statement_block ;

            circuit_body =  no_node_d[ch_p('{')] 
              >> statements
              >> *configuration_block
              >> no_node_d[ch_p('}')];

            circuit_def     =   
              error_guard(
                          root_node_d[keyword_d["CIRCUIT"]]
                          >> identifier_p[begin_circuit_def(self.symbol_table_stack, self.system_context)]
                          >> !parameter_def_p
                          >> circuit_body
                          )[error_handler<circuit_parser>(self)];

            //  End grammar definition


            // turn on the debugging info.
            BOOST_SPIRIT_DEBUG_RULE(null_value);
            BOOST_SPIRIT_DEBUG_RULE(rsl_factor);
            BOOST_SPIRIT_DEBUG_RULE(rsl_term);
            BOOST_SPIRIT_DEBUG_RULE(rsl_expression);

            BOOST_SPIRIT_DEBUG_RULE(statements);
            BOOST_SPIRIT_DEBUG_RULE(statement);
            BOOST_SPIRIT_DEBUG_RULE(statement_block);
            BOOST_SPIRIT_DEBUG_RULE(assignment);
            BOOST_SPIRIT_DEBUG_RULE(connector_spec);
            BOOST_SPIRIT_DEBUG_RULE(instantiation);
            BOOST_SPIRIT_DEBUG_RULE(node_instantiation);
            BOOST_SPIRIT_DEBUG_RULE(join_expression);
            BOOST_SPIRIT_DEBUG_RULE(ap_statement);
            BOOST_SPIRIT_DEBUG_RULE(var_expression);
            BOOST_SPIRIT_DEBUG_RULE(array_ref);
            BOOST_SPIRIT_DEBUG_RULE(lhs_var);
            BOOST_SPIRIT_DEBUG_RULE(rhs_var);

            BOOST_SPIRIT_DEBUG_RULE(for_statement);
            BOOST_SPIRIT_DEBUG_RULE(if_statement);

            BOOST_SPIRIT_DEBUG_RULE(rsl_bool_expression);
            BOOST_SPIRIT_DEBUG_RULE(rsl_bool_and);
            BOOST_SPIRIT_DEBUG_RULE(rsl_bool_or);
            BOOST_SPIRIT_DEBUG_RULE(rsl_bool_factor);
            BOOST_SPIRIT_DEBUG_RULE(rsl_bool_simple);

            BOOST_SPIRIT_DEBUG_RULE(parameter_list);
            BOOST_SPIRIT_DEBUG_RULE(with_pair);
            BOOST_SPIRIT_DEBUG_RULE(with_block);
            BOOST_SPIRIT_DEBUG_RULE(configuration_block);
            BOOST_SPIRIT_DEBUG_RULE(circuit_body);
            BOOST_SPIRIT_DEBUG_RULE(circuit_def);
        }

      rule<ScannerT, parser_tag<ID_C_rsl_expression> >   rsl_expression;
      rule<ScannerT, parser_tag<ID_C_rsl_term> >         rsl_term;
      rule<ScannerT, parser_tag<ID_C_rsl_factor> >       rsl_factor;

      rule<ScannerT, parser_tag<ID_C_rsl_bool_expression> >   rsl_bool_expression;
      rule<ScannerT, parser_tag<ID_C_rsl_bool_and> >          rsl_bool_and;
      rule<ScannerT, parser_tag<ID_C_rsl_bool_or> >           rsl_bool_or;
      rule<ScannerT, parser_tag<ID_C_rsl_bool_factor> >       rsl_bool_factor;
      rule<ScannerT, parser_tag<ID_C_rsl_bool_simple> >       rsl_bool_simple;
 
      rule<ScannerT, parser_tag<ID_C_null_value> >   null_value;
      rule<ScannerT, parser_tag<ID_C_lhs_var> >   lhs_var;
      rule<ScannerT, parser_tag<ID_C_rhs_var> >   rhs_var;
      rule<ScannerT, parser_tag<ID_C_array_ref> >   array_ref;
      rule<ScannerT, parser_tag<ID_C_var_expression> >   var_expression;


      rule<ScannerT, parser_tag<ID_C_for_statement> >   for_statement;
      rule<ScannerT, parser_tag<ID_C_if_statement> >   if_statement;
      rule<ScannerT, parser_tag<ID_C_assignment> >   assignment;
      rule<ScannerT, parser_tag<ID_C_connector_spec> >   connector_spec;
      rule<ScannerT, parser_tag<ID_C_instantiation> >   instantiation;
      rule<ScannerT, parser_tag<ID_C_node_instantiation> >   node_instantiation;
      rule<ScannerT, parser_tag<ID_C_join_expression> >   join_expression;
      rule<ScannerT, parser_tag<ID_C_ap_statement> >   ap_statement;


      rule<ScannerT, parser_tag<ID_C_statement_block> >   statement_block;
      rule<ScannerT, parser_tag<ID_C_statement> >     statement;
      rule<ScannerT, parser_tag<ID_C_statements> >     statements;
      rule<ScannerT, parser_tag<ID_C_parameter_list> >     parameter_list;
      rule<ScannerT, parser_tag<ID_C_with_pair> >     with_pair;
      rule<ScannerT, parser_tag<ID_C_with_block> >     with_block;

      rule<ScannerT, parser_tag<ID_C_configuration_block> > configuration_block;
      rule<ScannerT, parser_tag<ID_C_circuit_body> >    circuit_body;
      rule<ScannerT, parser_tag<ID_C_circuit_def> >     circuit_def;

      rule<ScannerT, parser_tag<ID_C_circuit_def> > const&
      start() const { return circuit_def; }

      int_parser integer_grammar;
      identifier_parser identifier_grammar;
      common_parser common_grammar;
      
      guard<Errors> error_guard;
      expect_parsers expect;
      
      std::string last_identifier;
    };
};

} // namespace vereofy::parser
} // namespace vereofy

#endif
