/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef CIRCUIT_PARSER_HELPER_HPP
#define CIRCUIT_PARSER_HELPER_HPP

#include <boost/algorithm/string.hpp>

#include "vereofy/common/checks.hpp"
#include "vereofy/common/errors.hpp"
#include "vereofy/common/return_value.hpp"
#include "parser_helper.hpp"
#include "reo_parser_types.hpp"
#include "circuit_types.hpp"
#include "int_parser_helper.hpp"
#include "identifier_parser_helper.hpp"
#include "REOExecuteException.hpp"

namespace vereofy {
namespace vparser {

  struct circuit_helper {

    typedef identifier_helper ih;

    struct t_null_value {};
    
    
    struct t_rsl_factor {
      static
      bool hasOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_factor);
        
        if (string(n.value.begin(), n.value.end())=="-") {
          return true;
        }
        return false;
      }

      static
      bool hasExpression(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_factor);
        
        if (n.children.size()==1 && n.children.at(0).value.id()==parser_id(ID_C_rsl_expression)) {
          return true;
        }
        return false;
      }

      
      static
      bool hasVarExpression(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_factor);
        
        if (n.children.size()==1 && n.children.at(0).value.id()==parser_id(ID_C_var_expression)) {
          return true;
        }
        return false;
      }

      static
      bool hasConstExpression(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_factor);
        
        if (n.children.size()==1 && n.children.at(0).value.id()==parser_id(ID_INT_const_expression)) {
          return true;
        }
        return false;
      }
      
      static
      bool isInteger(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_factor);
        return n.children.size()==1 && n.children.at(0).value.id()==parser_id(ID_INT_integer);
      }
      
      static
      int getInteger(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_factor);
        return int_helper::t_integer::getInt(n.children.at(0));
      }
      
      static
      string getOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_factor);
        return string(n.value.begin(), n.value.end());
      }

      static
      const pt_node& getExpression(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_factor);
        ERROR_CHECK_ASSERT(hasExpression(n));
        return n.children.at(0);
      }
      
      static
      const pt_node& getVarExpression(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_factor);
        ERROR_CHECK_ASSERT(hasVarExpression(n));
        return n.children.at(0);
      }

      static
      const pt_node& getConstExpression(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_factor);
        ERROR_CHECK_ASSERT(hasConstExpression(n));
        return n.children.at(0);
      }
      
      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        if (isInteger(n)) {
          return Value::ptr(new IntValue(getInteger(n)));
        }
        
        if (hasOp(n)) {
          int i=boost::static_pointer_cast<IntValue>(t_generic::execute(env, n.children.at(0)))->value;
          if (getOp(n)=="-") {
            return Value::ptr(new IntValue(-i));
          }
        }
        
        if (hasVarExpression(n)) {
          return t_var_expression::execute(env, getVarExpression(n));
        }

        if (hasConstExpression(n)) {
          int_value_t v = int_helper::t_const_expression::evaluate(getConstExpression(n), env.getSymbolTableStack());
          
          CHECK_SUCCESS(v.isDefined());
          return Value::ptr(new IntValue(v.getValue()));
        }

        if (hasExpression(n)) {
          return t_generic::execute(env, getExpression(n));
        }

        throw "implementation error";
      }
    };
    
    struct t_rsl_term {
      static
      bool hasOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_term);
        return n.children.size()==2;
      }
      
      static
      string getOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_term);
        return string(n.value.begin(), n.value.end());
      }
      
      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        Value::ptr l=t_generic::execute(env, n.children.at(0));
        if (hasOp(n)) {
          int left=boost::static_pointer_cast<IntValue>(l)->value;
          int right=boost::static_pointer_cast<IntValue>(t_generic::execute(env, n.children.at(1)))->value;
          if (getOp(n)=="*") {
            return Value::ptr(new IntValue(left * right));
          } else if (getOp(n)=="/") {
            return Value::ptr(new IntValue(left / right));
          } else if (getOp(n)=="%") {
            return Value::ptr(new IntValue(left % right));
          } else {throw "implementation error";}
        } else {
          return l;
        }
      }
    };
    
    struct t_rsl_expression {
      static
      bool hasOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_expression);
        return n.children.size()==2;
      }

      static
      string getOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_expression);
        return string(n.value.begin(), n.value.end());
      }

      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        switch (n.value.id().to_long()) {
        case ID_C_rsl_term:
          return t_rsl_term::execute(env, n);
        case ID_C_rsl_factor:       
          return t_rsl_factor::execute(env, n);
        case ID_C_rsl_expression: {
          Value::ptr l=t_generic::execute(env, n.children.at(0));
          if (hasOp(n)) {
            int left=boost::static_pointer_cast<IntValue>(l)->value;
            int right=boost::static_pointer_cast<IntValue>(t_generic::execute(env, n.children.at(1)))->value;
            if (getOp(n)=="+") {
              return Value::ptr(new IntValue(left + right));
            } else if (getOp(n)=="-") {
              return Value::ptr(new IntValue(left - right));
            } else {throw "implementation error";}
          } else {
            return l;
          }
        }
        }
        throw "implementation error";
      }
    };

    struct t_array_ref {
      static
      const pt_node& getExpression(const pt_node& n) {
        ID_ASSERT_(n, ID_C_array_ref);
        return n.children.at(0);
      }
    };

    struct t_lhs_var {
      static
      const pt_node& getIdentifierNode(const pt_node& n) {
        ID_ASSERT_(n, ID_C_lhs_var);
        return n.children.at(0);
      }

      static
      string getIdentifier(const pt_node& n) {
        ID_ASSERT_(n, ID_C_lhs_var);
        return ih::t_identifier::getIdentifier(n.children.at(0));
      }

      static
      bool hasArray(const pt_node& n) {
        ID_ASSERT_(n, ID_C_lhs_var);
        return (n.children.size()==2);
      }

      static
      const pt_node& getArray(const pt_node& n) {
        ID_ASSERT_(n, ID_C_lhs_var);
        return n.children.at(1);
      }

      static
      const int getArrayIndex(execution_environment_t& env,
                              const pt_node& n) {
        ID_ASSERT_(n, ID_C_lhs_var);
        int index=0;
        if (hasArray(n)) {
          Value::ptr array_ref=t_rsl_expression::execute(env, t_array_ref::getExpression(getArray(n)));
          index=IntValue::getIntValue(array_ref);
        }
        return index;
      }
    
      static
      string toString(execution_environment_t& env, const pt_node& n) {
        int index=0;
        if (hasArray(n)) {
          Value::ptr array_ref=t_rsl_expression::execute(env, getArray(n));
          index=IntValue::getIntValue(array_ref);
        }
      
        return getIdentifier(n)+"["+boost::lexical_cast<string>(index)+"]";
      }
    };
  
    struct t_rhs_var {
      static
      const pt_node& getIdentifierNode(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rhs_var);
        return n.children.at(0);
      }

      static
      string getIdentifier(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rhs_var);
        return ih::t_identifier::getIdentifier(n.children.at(0));
      }

      static
      bool hasArray(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rhs_var);
        return (n.children.size()==2);
      }

      static
      const pt_node& getArray(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rhs_var);
        return n.children.at(1);
      }

      static
      const int getArrayIndex(execution_environment_t& env,
                              const pt_node& n) {
        ID_ASSERT_(n, ID_C_rhs_var);
        int index=0;
        if (hasArray(n)) {
          Value::ptr array_ref=t_rsl_expression::execute(env, t_array_ref::getExpression(getArray(n)));
          index=IntValue::getIntValue(array_ref);
        }
        return index;
      }

      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        unsigned int index=0;
        if (hasArray(n)) {
          const pt_node& array=getArray(n);

          Value::ptr i=t_rsl_expression::execute(env, t_array_ref::getExpression(array));

          index=boost::static_pointer_cast<IntValue>(i)->value;
        }

        return env.getValue(getIdentifier(n), index);
      }
    };

    struct t_var_expression {
      static
      const pt_node& getVar(const pt_node& n) {
        ID_ASSERT_(n, ID_C_var_expression);
        return n.children.at(0);
      }

      static
      bool hasDot(const pt_node& n) {
        ID_ASSERT_(n, ID_C_var_expression);
        if (string(n.value.begin(), n.value.end())==".") {
          return true;
        }
        return false;
      }

      static
      const pt_node& getAfterDot(const pt_node& n) {
        ID_ASSERT_(n, ID_C_var_expression);
        ERROR_CHECK_ASSERT(hasDot(n));
        return n.children.at(1);
      }

      static
      ChannelEndRef::ptr resolve_port_reference(ComponentRef::ptr cr,
                                                const string &name) {
        ComponentInterface::ptr ci=cr->reo_component->getComponentInstance()->getInterface();
        PortDescription::ptr pd=ci->getPortDescription(name);
        if (!pd) {
          return ChannelEndRef::ptr();
        }

        unsigned int index=ci->getIndexForPortDescription(pd);
        
        ChannelEndRef::ptr cer(new ChannelEndRef());
        if (pd->getType() == PortDescription::IN_PORT) {
          cer->reo_channel_end=cr->reo_component->getSourceCE(index);
        } else if (pd->getType() == PortDescription::OUT_PORT) {
          cer->reo_channel_end=cr->reo_component->getSinkCE(index);
        } else {
          throw "Trying to access exported port "+name+", is not exported";
        }

        return cer;
      }
                                                


      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        Value::ptr v=t_rhs_var::execute(env, getVar(n));
      
        if (hasDot(n)) {
          // Get value table t from v
          switch (v->type->type) {
          case VALUE_TYPE_INT:
          case VALUE_TYPE_BOOL:
          case VALUE_TYPE_NODE_REF:
          case VALUE_TYPE_CHANNEL_END_REF:
            throw "Can not access subscope of simple variable!";
          
          case VALUE_TYPE_COMPONENT_REF: {
            const pt_node& after_dot=getAfterDot(n);

            if (hasDot(after_dot)) {
              throw "Can only access exported nodes";
            }

            ComponentRef::ptr cr=boost::static_pointer_cast<ComponentRef>(v);

            const pt_node& var=t_var_expression::getVar(after_dot);
            string name=t_rhs_var::getIdentifier(var);
            string lowercase_name=name;
            boost::to_lower(lowercase_name);            
            if (lowercase_name=="sink" ||
                lowercase_name=="source" ||
                lowercase_name=="topo_port" ||
                lowercase_name=="reconf_port" ||
                lowercase_name=="recaonf") {
              name=lowercase_name;
            }

            bool has_index=t_rhs_var::getArrayIndex(env, var);
            unsigned int index=t_rhs_var::getArrayIndex(env, var); // 0 if !has_index

            if (name=="topo_port" ||
                name=="reconf_port" ||
                name=="reconf") {
              if (index!=0) {
                throw "Illegal index "+boost::lexical_cast<string>(index)+" for reconfiguration port!";
              }
              
              if (!cr->reo_component->getComponentInstance()->hasReconfigurationPort()) {
                throw "Trying to access RECONF for component that does not have it";
              }
              
              // rename
              name="source";
              index=cr->reo_component->getComponentInstance()->getReconfigurationPortSourceIndex();
              // continue...
            }
            

            ChannelEndRef::ptr cer(new ChannelEndRef());
            if (name=="source") {
              if (index>=cr->reo_component->getSourceCECount()) {
                throw string("Trying to access source[")+boost::lexical_cast<string>(index)+
                  "], but has only "+
                  boost::lexical_cast<string>(cr->reo_component->getSourceCECount())+
                  " source ports";
              }
              cer->reo_channel_end=cr->reo_component->getSourceCE(index);
            } else if (name=="sink") {
              if (index>=cr->reo_component->getSinkCECount()) {
                throw string("Trying to access sink[")+boost::lexical_cast<string>(index)+
                  "], but has only "+
                  boost::lexical_cast<string>(cr->reo_component->getSinkCECount())+
                  " sink ports";
              }
              cer->reo_channel_end=cr->reo_component->getSinkCE(index);
            } else {
              if (has_index) {
                name=name+"["+boost::lexical_cast<string>(index)+"]";
              }
              cer=resolve_port_reference(cr, name);
              if (!cer && !has_index) {
                cer=resolve_port_reference(cr, name+"[0]");
              }

              if (!cer) {
                throw "Trying to access exported port "+name+", does not exist";
              }
            }
            return cer;
          }
          case VALUE_TYPE_UNKNOWN:
            throw "implementation error";
          }
        }
        return v;
      }
    };
  
  
    struct t_rsl_bool_simple {
      static
      string getOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_bool_simple);
        return string(n.value.begin(), n.value.end());
      }
    
      static
      bool execute(execution_environment_t& env, const pt_node& n) {
        Value::ptr left=t_rsl_expression::execute(env, n.children.at(0));
        Value::ptr right=t_rsl_expression::execute(env, n.children.at(1));
      
        string op=getOp(n);
        if (op=="==") {
          // add equality check for refs...
        
          int i1=boost::static_pointer_cast<IntValue>(left)->value;
          int i2=boost::static_pointer_cast<IntValue>(right)->value;
          return i1==i2;
        
        } else if (op=="!=") {
          // add inequality check for refs...
        
          int i1=boost::static_pointer_cast<IntValue>(left)->value;
          int i2=boost::static_pointer_cast<IntValue>(right)->value;
          return i1!=i2;
        
        } else {
          int i1=boost::static_pointer_cast<IntValue>(left)->value;
          int i2=boost::static_pointer_cast<IntValue>(right)->value;
        
          if (op=="<=") {return i1<=i2;}
          if (op=="<") {return i1<i2;}
          if (op==">=") {return i1>=i2;}
          if (op==">") {return i1>i2;}
        }
        throw "implementation error";
      }
    };


    struct t_rsl_bool_factor {
      static
      bool hasOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_bool_factor);
        return getOp(n)=="!";
      }

      static
      string getOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_bool_factor);
        return string(n.value.begin(), n.value.end());
      }

      static
      bool execute(execution_environment_t& env, const pt_node& n) {
        bool left=t_rsl_bool_expression::execute(env, n.children.at(0));
        
        if (hasOp(n)) {
          if (getOp(n)=="!") {
            return !left;
          }
          throw "implementation error";
        }
        return left;
      }
    };


    struct t_rsl_bool_and {
      static
      bool hasOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_bool_and);
        return n.children.size()==2;
      }

      static
      string getOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_bool_and);
        return string(n.value.begin(), n.value.end());
      }

      static
      bool execute(execution_environment_t& env, const pt_node& n) {
        bool left=t_rsl_bool_expression::execute(env, n.children.at(0));
        
        if (hasOp(n)) {
          bool right=t_rsl_bool_expression::execute(env, n.children.at(1));
          if (getOp(n)=="&") {
            return (left & right);
          }
          throw "implementation error";
        }
        return left;
      }
    };

    struct t_rsl_bool_or {
      static
      bool hasOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_bool_or);
        return n.children.size()==2;
      }

      static
      string getOp(const pt_node& n) {
        ID_ASSERT_(n, ID_C_rsl_bool_or);
        return string(n.value.begin(), n.value.end());
      }

      static
      bool execute(execution_environment_t& env, const pt_node& n) {
        bool left=t_rsl_bool_expression::execute(env, n.children.at(0));
        
        if (hasOp(n)) {
          bool right=t_rsl_bool_expression::execute(env, n.children.at(1));
          if (getOp(n)=="|") {
            return (left | right);
          }
          throw "implementation error";
        }
        return left;
      }
    };


    struct t_rsl_bool_expression {
      static
      bool execute(execution_environment_t& env, const pt_node& n) {
        switch (n.value.id().to_long()) {
        case ID_C_rsl_bool_or:
          return t_rsl_bool_or::execute(env, n);
        case ID_C_rsl_bool_and:
          return t_rsl_bool_and::execute(env, n);
        case ID_C_rsl_bool_factor:
          return t_rsl_bool_factor::execute(env, n);
        case ID_C_rsl_bool_simple:
          return t_rsl_bool_simple::execute(env, n);
        case ID_C_rsl_bool_expression:
          return t_rsl_bool_or::execute(env, n.children.at(0));
        }
        throw "implementation error";
      }
    };


    struct t_for_statement {

      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        const pt_node& start=n.children.at(0);
        const pt_node& guard=n.children.at(1);
        const pt_node& step=n.children.at(2);
        const pt_node& block=n.children.at(3);

        t_assignment::execute(env, start);
        for (;
             t_rsl_bool_expression::execute(env, guard)==true;
             t_assignment::execute(env, step)) {
          t_statement_block::execute(env, block);
        }

        return Value::ptr();
      }
    };


    struct t_if_statement {
      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        const pt_node& if_guard=n.children.at(0);
        const pt_node& if_block=n.children.at(1);

        if (t_rsl_bool_expression::execute(env, if_guard)==true) {
          t_statement_block::execute(env, if_block);
        } else {
          if (n.children.size()==4) {
            const pt_node& else_block=n.children.at(3);
            t_statement_block::execute(env, else_block);
          }
        }
        return Value::ptr();
      }
    };

    struct t_connector_spec {

      static
      ReturnValue<unsigned int> getSemicolonPos(const pt_node& n) {
        ID_ASSERT_(n, ID_C_connector_spec);

        for (unsigned int i=0;i<n.children.size();i++) {
          const pt_node& cur=n.children.at(i);
          if (cur.value.id() != parser_id(ID_C_var_expression) &&
              cur.value.id() != parser_id(ID_C_null_value)) {
            if (string(cur.value.begin(), cur.value.end())!=";") {
              throw "implementation error";
            }
            return ReturnValue<unsigned int>(i);
          }
        }

        return ReturnValue<unsigned int>();
      }

      static
      unsigned int getSourceCount(const pt_node& n) {
        ID_ASSERT_(n, ID_C_connector_spec);

        ReturnValue<unsigned int> semicolon_pos=getSemicolonPos(n);
        if (semicolon_pos.valid) {
          return semicolon_pos.value;
        } else {
          // no semicolon -> only sources
          return n.children.size();
        }
      }

      static
      unsigned int getSinkCount(const pt_node& n) {
        ID_ASSERT_(n, ID_C_connector_spec);

        ReturnValue<unsigned int> semicolon_pos=getSemicolonPos(n);
        if (semicolon_pos.valid) {
          return n.children.size()-1-semicolon_pos.value;
        } else {
          // no semicolon -> no sinks
          return 0;
        }
      }

      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n,
                         ComponentRef::ptr cr) {
        REO_Component::ptr reo_component=cr->reo_component;

        unsigned int source_ces=reo_component->getSourceCECount();
        unsigned int sink_ces=reo_component->getSinkCECount();

        unsigned int source_params=getSourceCount(n);
        unsigned int sink_params=getSinkCount(n);

        if (source_ces != source_params) {
          throw string("ConnectorSpec: mismatch in number of source parameters, ")+
            boost::lexical_cast<std::string>(source_ces)+" expected, but "+
            boost::lexical_cast<std::string>(source_params)+" used";
        }

        if (sink_ces != sink_params) {
          throw string("ConnectorSpec: mismatch in number of sink parameters, ")+
            boost::lexical_cast<std::string>(sink_ces)+" expected, but "+
            boost::lexical_cast<std::string>(sink_params)+" used";
        }

        unsigned int cur_source=0;
        unsigned int cur_sink=0;

        bool before_semicolon=true;
        for (unsigned int i=0;i<n.children.size();i++) {
          if (i==source_params) {
            // at semicolon
            before_semicolon=false;
            continue;
          }

          const pt_node& cur=n.children.at(i);
          if (cur.value.id()==ID_C_null_value) {
            // NULL -> leave unconnected
            continue;
          }
          bool has_dot=t_var_expression::hasDot(cur);
          Value::ptr value;
          string target_name;
          unsigned int target_index=0;
          if (!has_dot) {
            target_name=t_rhs_var::getIdentifier(t_var_expression::getVar(cur));
            target_index=t_rhs_var::getArrayIndex(env, t_var_expression::getVar(cur));

            if ((!env.VarExists(target_name)) ||
                env.VarIsNull(target_name, target_index)) {
              value=Value::ptr();
            } else {
              value=env.getValue(target_name, target_index);
            }
          } else {
            // has dot
            value=t_var_expression::execute(env, cur);
          }

          TRACE("circuit_execute", "connector_spec",\
                cout << "Connecting " \
               << (before_semicolon ? "source " : "sink ")\
               << (before_semicolon ? cur_source : cur_sink) \
                << "\n"                                      \
                );

          REO_Channel_End::ptr exported_ce;
          if (before_semicolon) {
            exported_ce=reo_component->getSourceCE(cur_source);
          } else {
            exported_ce=reo_component->getSinkCE(cur_sink);
          }

          if (!value) {
            ERROR_CHECK_ASSERT(!has_dot);
            // just save the CE

            ChannelEndRef::ptr cer(new ChannelEndRef());
            cer->reo_channel_end=exported_ce;

            TRACE("circuit_execute", "connector_spec",  \
                  cout << "Just adding the exported CE\n");

            env.setValue(target_name, target_index, cer);
          } else {
            REO_Circuit::ptr reo_circuit=env.getREOCircuit();

            // value exists, have to join
            if (value->type==ValueType::NODE_REF) {
              TRACE("circuit_execute", "connector_spec", \
                    cout << "Joining with node ref");
              
              reo_circuit->join(boost::static_pointer_cast<NodeRef>(value)->reo_node, 
                                exported_ce);
            } else if (value->type==ValueType::CHANNEL_END_REF) {
              TRACE("circuit_execute", "connector_spec", \
                    cout << "Joining with channel end");
              REO_Node::ptr result_node;
              result_node=reo_circuit->join(boost::static_pointer_cast<ChannelEndRef>(value)->reo_channel_end, 
                                            exported_ce);

              if (!has_dot) {
                NodeRef::ptr nr(new NodeRef());
                nr->reo_node=result_node;
                env.setValue(target_name, target_index, nr);
              }
            } else {
              throw "Can't connect to variable that is not a node or channel end";
            }
          }

          
          // update cur_sink, cur_source
          if (before_semicolon) {
            cur_source++;
          } else {
            cur_sink++;
          }
        }

        return Value::ptr();
      }
    };

    struct t_node_instantiation {

      static
      REO_Node::node_semantics_t getSemantics(const pt_node& n) {
        string name(n.value.begin(), n.value.end());
        if (name=="NODE") {
          return REO_Node::NODE_REPLICATE;
        } else if (name=="ROUTE_NODE") {
          return REO_Node::NODE_ROUTE;
        } else {
          THROW_UNIMPLEMENTED();
        }
      }
      
      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        REO_Circuit::ptr circuit=env.getREOCircuit();

        REO_Node::ptr result_node=circuit->create_node(getSemantics(n));
        NodeRef::ptr nr(new NodeRef());
        nr->reo_node=result_node;
        
        return nr;
      }        
    };

    struct t_with_pair {
      static
      std::pair<std::string, std::string> 
      getPair(const pt_node& n) {
        ID_ASSERT_(n, ID_C_with_pair);

        string left=ih::t_identifier::getIdentifier(n.children.at(0));
        const pt_node& rhs=n.children.at(1);
        if (rhs.value.id()==parser_id(ID_I_identifier)) {
          return make_pair(left, ih::t_identifier::getIdentifier(rhs));
        } else if (rhs.value.id()==parser_id(ID_INT_integer)) {
          int v=int_helper::t_integer::getInt(rhs);
          return make_pair(left, boost::lexical_cast<string>(v));
        } else {
          THROW_UNIMPLEMENTED();
        }
      }
    };

    struct t_with_block {
      static
      unsigned int getCount(const pt_node& n) {
        ID_ASSERT_(n, ID_C_with_block);
        return n.children.size();
      }

      static
      std::pair<std::string, std::string> 
      getWithValue(const pt_node& n, unsigned int index) {
        ID_ASSERT_(n, ID_C_with_block);
      
        return t_with_pair::getPair(n.children.at(index));
      }

    };
    
    struct t_instantiation {
      static
      string getComponent(const pt_node& n) {
        ID_ASSERT_(n, ID_C_instantiation);
        return string(ih::t_identifier::getIdentifier(n.children.at(0)));
      }

      static
      unsigned int getParameterListPos(const pt_node& n) {
        ID_ASSERT_(n, ID_C_instantiation);
        for (unsigned int i=1;i<n.children.size();i++) {
          if (n.children.at(i).value.id()==parser_id(ID_C_parameter_list)) {
            return i;
          }
        }
        return 0;
      }


      static
      unsigned int getConnectorSpecPos(const pt_node& n) {
        ID_ASSERT_(n, ID_C_instantiation);
        for (unsigned int i=1;i<n.children.size();i++) {
          if (n.children.at(i).value.id()==parser_id(ID_C_connector_spec)) {
            return i;
          }
        }
        return 0;
      }

      static
      unsigned int getWithBlockPos(const pt_node& n) {
        ID_ASSERT_(n, ID_C_instantiation);
        for (unsigned int i=1;i<n.children.size();i++) {
          if (n.children.at(i).value.id()==parser_id(ID_C_with_block)) {
            return i;
          }
        }
        return 0;
      }

      static
      bool hasParameterList(const pt_node& n) {
        ID_ASSERT_(n, ID_C_instantiation);
        return getParameterListPos(n)!=0;
      }

      static
      bool hasConnectorSpec(const pt_node& n) {
        ID_ASSERT_(n, ID_C_instantiation);
        return getConnectorSpecPos(n)!=0;
      }

      static
      bool hasWithBlock(const pt_node& n) {
        ID_ASSERT_(n, ID_C_instantiation);
        return getWithBlockPos(n)!=0;
      }

      static
      const pt_node& getConnectorSpec(const pt_node& n) {
        ID_ASSERT_(n, ID_C_instantiation);
        if (hasConnectorSpec(n)) {
          return n.children.at(getConnectorSpecPos(n));
        } else {
          throw "implementation error";
        }
      }

      static
      const pt_node& getParameterList(const pt_node& n) {
        ID_ASSERT_(n, ID_C_instantiation);
        if (hasParameterList(n)) {
          return n.children.at(getParameterListPos(n));
        } else {
          throw "implementation error";
        }
      }

      static
      const pt_node& getWithBlock(const pt_node& n) {
        ID_ASSERT_(n, ID_C_instantiation);
        if (hasWithBlock(n)) {
          return n.children.at(getWithBlockPos(n));
        } else {
          throw "implementation error";
        }
      }


    
      static ComponentRef::ptr instantiate(execution_environment_t& env,
                                           const string& component, 
                                           const string& name,
                                           unsigned int name_index,
                                           const std::vector<Value::ptr>& parameters=std::vector<Value::ptr>()) {
        ComponentTypeIdentifier cti(component, parameters);
        if (env.getSystemContext()->getReplacementManager()->hasReplacement(cti.toString())) {
          ComponentTypeIdentifier cti_replacement(env.getSystemContext()->getReplacementManager()->getReplacement(cti.toString()));

          if (debugLevel>=2) {
            std::cout << "Replacing " << cti.toString() << " with " << cti_replacement.toString() << "\n";
          }
          return instantiate(env,
                             cti_replacement.type_name,
                             name,
                             name_index,
                             cti_replacement.parameters);
        }

        SymbolData::ptr si=env.getGlobalSymbols()->find(component);
      
        if (!si) {
          throw string("There is no component named '")+component+"'";
        }
      
        if (si->type!=SYMBOL_TYPE_COMPONENT) {
          throw string("The object named '")+component+"' is not a component!";
        }

        ComponentRef::ptr result;
      
        ComponentData::ptr cd=boost::static_pointer_cast<ComponentData>(si);
        switch (cd->comp_type) {
        case COMPONENT_CIRCUIT: {
          CircuitInfo::ptr ci=boost::static_pointer_cast<CircuitInfo>(si);
        
          ComponentInstanceCircuit::ptr cic(new ComponentInstanceCircuit(ci->connector, parameters, true));
          result=ComponentRef::ptr(new ComponentRef(cic));
          break;
        }
        case COMPONENT_REACTIVE_MODULE: {
          ModuleInfo::ptr mi=boost::static_pointer_cast<ModuleInfo>(si);
        
          ComponentInstanceReactiveModule::ptr cim(new ComponentInstanceReactiveModule(mi->module, parameters));
          result=ComponentRef::ptr(new ComponentRef(cim));
          break;
        }
        default:
          THROW_UNIMPLEMENTED();
        }

/*      string full_name;
        if (name=="") {
          full_name="$anonymous";
          unsigned int index=env.setValueNextFree(full_name, result);
          full_name+="["+boost::lexical_cast<string>(index)+"]";
        } else {
          env.setValue(name, name_index, result);
          full_name=name+"["+boost::lexical_cast<string>(name_index)+"]";
        }
        result->ci->setName(full_name);
*/

        env.getREOCircuit()->add_Component(result->reo_component);
        return result;
      }
    
      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n,
                         string name="", int index=0) {

        ComponentRef::ptr result;

        if (hasParameterList(n)) {
          std::vector<Value::ptr> parameters;
        
          for (unsigned int i=0;i<getParameterList(n).children.size();i++) {
            Value::ptr v=t_rsl_expression::execute(env, getParameterList(n).children.at(i));
            parameters.push_back(v);
          }
          result=instantiate(env, 
                             getComponent(n),
                             name,
                             index,
                             parameters);
        } else {
          result=instantiate(env, 
                             getComponent(n),
                             name,
                             index);
        }
      
        if (hasConnectorSpec(n)) {
          t_connector_spec::execute(env, getConnectorSpec(n), result);
        }

        if (hasWithBlock(n)) {
          ComponentInstance::ptr ci=result->ci;

          const pt_node& with_block=getWithBlock(n);
          for (unsigned int i=0;i<t_with_block::getCount(with_block);i++) {
            std::pair<std::string,std::string> v=t_with_block::getWithValue(with_block, i);
            ci->addWithParameter(v.first, v.second);
          }
        }

        return result;
      }
    };
  
    struct t_join_expression {
      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        vector<Value::ptr> to_join;
        for (size_t i=0;i<n.children.size();i++) {
          Value::ptr v=t_var_expression::execute(env, n.children.at(i));
          to_join.push_back(v);
        }

        if (to_join.size()==0) {
          throw "join without parameters";
        }

        if (to_join.size()==1) {
          throw "join with only 1 parameter";
        }

        REO_Circuit::ptr circuit=env.getREOCircuit();

        REO_Node::ptr result_node;
        result_node=join(circuit, to_join.at(0), to_join.at(1));
        for (unsigned int i=2;i<to_join.size();i++) {
          result_node=join(circuit, result_node, to_join.at(i));
        }

        NodeRef::ptr nr(new NodeRef());
        nr->reo_node=result_node;
        return nr;
      }

      static REO_Node::ptr join(REO_Circuit::ptr circuit, REO_Node::ptr node1, Value::ptr v2) {
        if (v2->type!=ValueType::NODE_REF &&
            v2->type!=ValueType::CHANNEL_END_REF)  {
          throw "Can only join nodes or channel ends";
        }

        bool v2_is_node=(v2->type==ValueType::NODE_REF);

        REO_Node::ptr v2_node;
        REO_Channel_End::ptr v2_ce;
        
        if (v2_is_node) {
          v2_node=boost::static_pointer_cast<NodeRef>(v2)->reo_node;
        } else {
          v2_ce=boost::static_pointer_cast<ChannelEndRef>(v2)->reo_channel_end;
        }
        
        if (v2_is_node) {
          return circuit->join(node1, v2_node);
        } else {
          return circuit->join(node1, v2_ce);
        }
      }

      static REO_Node::ptr join(REO_Circuit::ptr circuit, Value::ptr v1, Value::ptr v2) {
        if ((v1->type!=ValueType::NODE_REF &&
             v1->type!=ValueType::CHANNEL_END_REF) 
            ||
            (v2->type!=ValueType::NODE_REF &&
             v2->type!=ValueType::CHANNEL_END_REF))  {
          throw "Can only join nodes or channel ends";
        }

        bool v1_is_node=(v1->type==ValueType::NODE_REF);
        bool v2_is_node=(v2->type==ValueType::NODE_REF);

        REO_Node::ptr v1_node, v2_node;
        REO_Channel_End::ptr v1_ce, v2_ce;
        
        if (v1_is_node) {
          v1_node=boost::static_pointer_cast<NodeRef>(v1)->reo_node;
        } else {
          v1_ce=boost::static_pointer_cast<ChannelEndRef>(v1)->reo_channel_end;
        }

        if (v2_is_node) {
          v2_node=boost::static_pointer_cast<NodeRef>(v2)->reo_node;
        } else {
          v2_ce=boost::static_pointer_cast<ChannelEndRef>(v2)->reo_channel_end;
        }
        
        if (v1_is_node && v2_is_node) {
          return circuit->join(v1_node, v2_node);
        } else if (v1_is_node && !v2_is_node) { 
          return circuit->join(v1_node, v2_ce);
        } else if (!v1_is_node && v2_is_node) {
          return circuit->join(v2_node, v1_ce);
        } else if (!v1_is_node && !v2_is_node) {
          return circuit->join(v1_ce, v2_ce);
        }

        throw "implementation error";
      }
    };

    struct t_assignment {
      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        string name=t_lhs_var::getIdentifier(n.children.at(0));
        int index=t_lhs_var::getArrayIndex(env, n.children.at(0));
        
        Value::ptr old_value;
        if (env.VarExists(name) &&
            !env.VarIsNull(name, index)) {
          old_value=env.getValue(name, index);
        }

        // TODO: if necessary, free resources because old_value will be overwritten

        // right hand side
        Value::ptr new_value;
        switch (n.children.at(1).value.id().to_long()) {
        case ID_C_null_value: 
          new_value=Value::ptr();
          break;
        case ID_C_rsl_expression: 
          new_value=t_generic::execute(env, n.children.at(1));
          break;
        case ID_C_join_expression:
          new_value=t_join_expression::execute(env, n.children.at(1));
          break;
        case ID_C_node_instantiation: 
          new_value=t_node_instantiation::execute(env, n.children.at(1));
          break;
        case ID_C_instantiation: 
          new_value=t_instantiation::execute(env, n.children.at(1), name, index); 
          break;
        }
        env.setValue(name, index, new_value);


        // no return value
        return Value::ptr();
      }
    };


    struct t_ap_statement {

      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        REO_Circuit::ptr circuit=env.getREOCircuit();

        std::string ap_name=std::string(n.children.at(0).value.begin(), n.children.at(0).value.end());
        std::string ap_spec=std::string(n.children.at(1).value.begin(), n.children.at(1).value.end());

        circuit->attachAtomicProposition(ap_name, ap_spec);

        // no return value 
        return Value::ptr();
      }        
    };


    struct t_statement {
      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        if (env.flagDebug()) {
          ostream& dstream=env.getDebugStream();
          dstream << "Execute statement at ";
#ifdef REO_PARSER_POSITION_NODES
          if (!n.children.at(0).value.pos) {
            dstream << "<unknown>\n";
          } else {
            dstream << n.children.at(0).value.pos->line << ", " << n.children.at(0).value.pos->column << "\n";
          }
#else
          dstream << ":\n";
          dstream << string(n.children.at(0).value.begin(),
                            n.children.at(0).value.end());
          dstream << "\n";
#endif

        }

        try {
          return t_generic::execute(env, n.children.at(0));
        } catch (REOExecuteException& rer) {
          throw; // rethrow
        } catch (const std::string& err) {
          throw_exception(err, n);
        } catch (const char *err) {
          throw_exception(err, n);
        } catch (...) {
          throw;
        }

        // we should never get here
        THROW_UNIMPLEMENTED();
      }
      
      static
      void throw_exception(const std::string& message, const pt_node& n) {
        REOExecuteException rer(message);

#ifdef REO_PARSER_POSITION_NODES
        boost::shared_ptr<file_position> pos=n.children.at(0).value.pos;
        if (pos) {
          rer.setFile(pos->file);
          rer.setLocation("("+
                          boost::lexical_cast<string>(pos->line)+
                          ","+
                          boost::lexical_cast<string>(pos->column)+
                          ")");
        }
#endif
        rer.raise();
      }
        
    };

    struct t_statements {
      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        for (size_t i=0;i<n.children.size(); i++) {
          t_generic::execute(env, n.children.at(i));
        }
        return Value::ptr();
      }
    };


    struct t_statement_block {
      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        return t_generic::execute(env, n.children.at(0));
      }
    };

    struct t_parameter_list {};

    struct t_configuration_block {

      static
      const pt_node& getStatementBlock(const pt_node& n) {
        ID_ASSERT_(n, ID_C_configuration_block);
        return n.children.at(0);
      }

    };


    struct t_circuit_body {
      static
      const pt_node& getBaseStatementBlock(const pt_node& n) {
        ID_ASSERT_(n, ID_C_circuit_body);
        return n.children.at(0);
      }

      static
      unsigned int countConfigurations(const pt_node& n) {
        ID_ASSERT_(n, ID_C_circuit_body);
        return n.children.size()-1;
      }

      static
      const pt_node& getConfigurationBlock(const pt_node& n, unsigned int index) {
        ID_ASSERT_(n, ID_C_circuit_body);
        return n.children.at(index+1);
      }
    };
      
    struct t_circuit_def {
      static
      const pt_node& getStatementBlock(const pt_node& n) {
        ID_ASSERT_(n, ID_C_circuit_def);
        return t_circuit_body::getBaseStatementBlock(getCircuitBody(n));
      }

      static
      const pt_node& getCircuitBody(const pt_node& n) {
        ID_ASSERT_(n, ID_C_circuit_def);
        return n.children.at(n.children.size()-1);
      }

      static
      string getName(const pt_node& n) {
        ID_ASSERT_(n, ID_C_circuit_def);
        return ih::t_identifier::getIdentifier(n.children.at(0));
      }

      static
      bool hasParameterDef(const pt_node& n) {
        ID_ASSERT_(n, ID_C_circuit_def);
        return n.children.size()>2 && n.children.at(1).value.id()==parser_id(ID_COM_parameter_def);
      }

      static
      const pt_node& getParameterDef(const pt_node& n) {
        ID_ASSERT_(n, ID_C_circuit_def);
        if (hasParameterDef(n)) {
          return n.children.at(1);
        } else {
          throw "implementation error";
        }
      }
    };
   
    struct t_generic {
      template <class ValueType> static
      typename ValueType::ptr 
      execute(execution_environment_t& env, const pt_node& n) {
        return boost::static_pointer_cast<ValueType>(t_generic::execute(env, n));
      }

      static
      Value::ptr execute(execution_environment_t& env, const pt_node& n) {
        switch (n.value.id().to_long()) {
        case ID_INT_integer: //
        case ID_I_identifier: //
          break;
        
        case ID_C_rsl_factor:
          return t_rsl_factor::execute(env, n);
        case ID_C_rsl_term: 
          return t_rsl_term::execute(env, n);
        case ID_C_rsl_expression:
          return t_rsl_expression::execute(env, n);

        case ID_C_rsl_bool_simple:
        case ID_C_rsl_bool_factor:
        case ID_C_rsl_bool_and:
        case ID_C_rsl_bool_or:
        case ID_C_rsl_bool_expression:
          break;

        case ID_C_lhs_var:
          break;

        case ID_C_rhs_var:
          return t_rhs_var::execute(env, n);


        case ID_C_array_ref: //
          break;

        case ID_C_var_expression:
          return t_var_expression::execute(env, n);
        
        case ID_C_assignment:
          return t_assignment::execute(env, n);

        case ID_C_instantiation:
          // anonymous instantiation
          return t_instantiation::execute(env, n);

        case ID_C_join_expression:
          return t_join_expression::execute(env, n);

        case ID_C_ap_statement:
          return t_ap_statement::execute(env, n);

        case ID_C_if_statement:
          return t_if_statement::execute(env, n);
        
        case ID_C_for_statement:
          return t_for_statement::execute(env, n);
        
        case ID_C_statement:
          return t_statement::execute(env, n);

        case ID_C_statements:
          return t_statements::execute(env, n);

        case ID_C_statement_block:
          return t_statement_block::execute(env, n);
        
        case ID_C_circuit_def:
        case ID_C_global:
          break;
        }
        throw "implementation error";
      }
    };
  };

}
}

#endif 
