/***************************************************************************
 *   Copyright (C) 2008-2010 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef COMMON_PARSER_HPP
#define COMMON_PARSER_HPP

#include "spirit_init.hpp"
using namespace vereofy::spirit;

#include <boost/bind.hpp>

#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>

#include <string>

#include "reo_parser_types.hpp"
#include "position_node.hpp"
#include "position_node_factory.hpp"

#include "identifier_parser.hpp"


namespace vereofy {
namespace vparser {

struct common_parser : public grammar<common_parser>
{

  enum {
    quoted_string = 0,
    parameter_def = 1
  };

  enum Errors {
    no_parameter,
    no_greater
  };

  struct expect_parsers {
    expect_parsers() :
      parameter(no_parameter),
      greater(no_greater)
    {}
    
    vereofy::spirit::assertion<Errors> parameter;
    vereofy::spirit::assertion<Errors> greater;
  };



  static 
  void init_rule_names(map<parser_id, std::string>& rule_names) {
    rule_names[ID_COM_quoted_string] = "quoted_string";
    rule_names[ID_COM_parameter_def] = "parameter_def";
  }

  template <class DefT>
  struct error_handler {
    DefT& def;

    error_handler(DefT& def_) : def(def_) {}

    template <typename ScannerT, typename ErrorT>
    error_status<>
    operator()(ScannerT const& scan, ErrorT const& error) const {
      switch (error.descriptor) {
      case no_parameter:
        throw_(error.where, std::string("Expected a parameter name"));
      case no_greater:
        throw_(error.where, std::string("Expected '>'"));
      }
      
      // default action
      return error_status<>(error_status<>::rethrow);
    }
  };


  common_parser(const symbols<>& reserved_keywords_) :
    reserved_keywords(reserved_keywords_) 
  {}
  
  const symbols<>& reserved_keywords;

  template <typename ScannerT>
  struct definition :
    public grammar_def< rule<ScannerT, parser_tag<ID_COM_quoted_string> >,
                        rule<ScannerT, parser_tag<ID_COM_parameter_def> >   > 
  {
    definition(common_parser const& self_) 
      : identifier_grammar(self_.reserved_keywords)
    {

      BOOST_SPIRIT_DEBUG_GRAMMAR(identifier_grammar);

      ENTRY_GRAMMAR(identifier_p, identifier_parser, identifier_grammar, identifier_parser::identifier);

      //  Start grammar definition
      quoted_string =
        lexeme_d[ no_node_d[ch_p('"')] >> leaf_node_d[*(anychar_p - ch_p('"'))] >> no_node_d[ch_p('"')]];

      parameter_def  =  error_guard(
          no_node_d[ch_p('<')] 
          >> expect.parameter(infix_node_d[list_p(identifier_p, ch_p(','))]) 
          >> no_node_d[expect.greater(ch_p('>'))])
        [error_handler<definition>(*this)];
      
      //  End grammar definition

      this->start_parsers(quoted_string, parameter_def);

      BOOST_SPIRIT_DEBUG_RULE(quoted_string);
      BOOST_SPIRIT_DEBUG_RULE(parameter_def);
    }

    rule<ScannerT, parser_tag<ID_COM_quoted_string> >     quoted_string;
    rule<ScannerT, parser_tag<ID_COM_parameter_def> >     parameter_def;

    identifier_parser identifier_grammar;


    guard<Errors> error_guard;
    expect_parsers expect;
  };
};

}
}
#endif
