/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef GUARD_PARSER_HPP
#define GUARD_PARSER_HPP

#include "spirit_init.hpp"
using namespace vereofy::spirit;

#include <boost/bind.hpp>

#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>

#include <stack>
#include <string>

#include "reo_parser_types.hpp"
#include "position_node.hpp"
#include "position_node_factory.hpp"

#include "vereofy/common/singleton_set.hpp"

#include "vereofy/system/reactive_module/reactive_module.hpp"
#include "vereofy/system/datadomain.hpp"

#include "int_parser.hpp"
#include "int_parser_helper.hpp"


#include "type_parser.hpp"
#include "type_parser_helper.hpp"
#include "identifier_parser.hpp"
#include "guard_parser_helper.hpp"
#include "common_parser.hpp"

#include "parser_helper.hpp"
#include "module_parser_helper.hpp"


namespace vereofy {
namespace vparser {

struct guard_parser : public grammar<guard_parser>
{
  //  typedef guard_helper helper;

  enum parser_type {
    any_value = 0,
    guard_expression = 1,
    state_guard = 2,
    io_guard = 3,
    state_assignment = 4,
    port_name = 5,
    bool_expression = 6
  };

  enum Errors {
    state_var_expected,
    io_port_expected,
    no_compare_op,
    no_assignment_op,
    no_active_port_name,
    no_state_guard_expression,
    no_io_guard_expression,
    no_assignment_expression,
    is_not_a_subcomponent,
    no_left_parenthesis,
    no_right_parenthesis,
    bool_expression_expected,
    expression_expected,
    range_def_expected,
    no_right_square,
    no_right_curly,
    no_comma,
    no_semicolon
  };

  struct expect_parsers {
    expect_parsers() :
      state_var(state_var_expected),
      io_port(io_port_expected),
      compare_op(no_compare_op),
      assignment_op(no_assignment_op),
      active_port_name(no_active_port_name),
      state_guard_expression(no_state_guard_expression),
      io_guard_expression(no_io_guard_expression),
      assignment_expression(no_assignment_expression),
      not_a_subcomponent(is_not_a_subcomponent),
      left_parenthesis(no_left_parenthesis),
      right_parenthesis(no_right_parenthesis),
      right_square(no_right_square),
      right_curly(no_right_curly),
      bool_expression(bool_expression_expected),
      expression(expression_expected),
      range_def(range_def_expected),
      comma(no_comma),
      semicolon(no_semicolon)
    {}

    vereofy::spirit::assertion<Errors> state_var;
    vereofy::spirit::assertion<Errors> io_port;
    vereofy::spirit::assertion<Errors> compare_op;
    vereofy::spirit::assertion<Errors> assignment_op;
    vereofy::spirit::assertion<Errors> active_port_name;
    vereofy::spirit::assertion<Errors> state_guard_expression;
    vereofy::spirit::assertion<Errors> io_guard_expression;
    vereofy::spirit::assertion<Errors> assignment_expression;
    vereofy::spirit::assertion<Errors> not_a_subcomponent;
    vereofy::spirit::assertion<Errors> left_parenthesis;
    vereofy::spirit::assertion<Errors> right_parenthesis;
    vereofy::spirit::assertion<Errors> right_square;
    vereofy::spirit::assertion<Errors> right_curly;
    vereofy::spirit::assertion<Errors> bool_expression;
    vereofy::spirit::assertion<Errors> expression;
    vereofy::spirit::assertion<Errors> range_def;
    vereofy::spirit::assertion<Errors> comma;
    vereofy::spirit::assertion<Errors> semicolon;
  };

  static 
  void init_rule_names(map<parser_id, std::string>& rule_names) {
    rule_names[ID_G_any_value] = "any_value";
    rule_names[ID_G_function_call] = "function_call";
    rule_names[ID_G_bool_function_call] = "bool_function_call";
    rule_names[ID_G_guard_factor] = "guard_factor";
    rule_names[ID_G_guard_term] = "guard_term";
    rule_names[ID_G_guard_expression_full] = "guard_expression_full";
    rule_names[ID_G_guard_expression] = "guard_expression";

    rule_names[ID_G_int_range] = "int_range";
    
    rule_names[ID_G_state_var] = "state_var";
    rule_names[ID_G_simple_state_var] = "simple_state_var";
    rule_names[ID_G_subcomponent_access] = "subcomponent_access";
    rule_names[ID_G_port] = "port";
    rule_names[ID_G_port_data] = "port_data";
    rule_names[ID_G_active_ports] = "active_ports";
    rule_names[ID_G_compare_op] = "compare_op";
    rule_names[ID_G_port_activity_expression ] = "port_activity_expression ";
    rule_names[ID_G_port_data_expression] = "port_data_expression";
    rule_names[ID_G_has_label_expression] = "has_label_expression";
    rule_names[ID_G_label_id] = "label_id";

    rule_names[ID_G_range_def] = "range_def";
    rule_names[ID_G_compare_expression] = "compare_expression";
    rule_names[ID_G_bool_expression_term] = "bool_expression_term";
    rule_names[ID_G_bool_expression_ranged] = "bool_expression_ranged";
    rule_names[ID_G_bool_expression_implies_equiv] = "bool_expression_implies_equiv";
    rule_names[ID_G_bool_expression_and] = "bool_expression_and";
    rule_names[ID_G_bool_expression_xor] = "bool_expression_xor";
    rule_names[ID_G_bool_expression] = "bool_expression";
    rule_names[ID_G_if_expression] = "if_expression";
    rule_names[ID_G_ite_expression] = "ite_expression";
    rule_names[ID_G_ite_bool_expression] = "ite_bool_expression";
    rule_names[ID_G_ap_ref] = "ap_ref";
    rule_names[ID_G_simple_ap_ref] = "simple_ap_ref";
    rule_names[ID_G_struct_member_access] = "struct_member_access";
    rule_names[ID_G_port_name] = "port_name";

    rule_names[ID_G_state_guard] = "state_guard";
    rule_names[ID_G_io_guard] = "io_guard";
    rule_names[ID_G_state_assignment] = "state_assignment";
    rule_names[ID_G_add_label] = "add_label";
  }

  static
  void init_reserved_keywords(symbols<>& keywords) {
    keywords="MODULE", "has_label", "label";
  }


    // --- semantic actions ------

  typedef std::stack<vereofy::ComponentInterface::ptr> interface_stack_t;

  struct is_state_var {
    const interface_stack_t& stack_ref;
    std::string& id;
    bool check_type;
    DataDomain::data_domain_type type;

    is_state_var(const interface_stack_t& stack_ref_,
                 std::string& id_) :
      stack_ref(stack_ref_), 
      id(id_),
      check_type(false) {}

    is_state_var(const interface_stack_t& stack_ref_,
                 std::string& id_,
                 DataDomain::data_domain_type type_) :
      stack_ref(stack_ref_), 
      id(id_),
      check_type(true),
      type(type_) {}

    const bool
    operator()() const {
      if (debugLevel>=4) {
        cerr << "Checking for state variable " << id << " in interface " <<
          stack_ref.top() << endl;
      }
      bool rv=false;
      StateDescription::ptr sd=stack_ref.top()->getStateDescription(id);
      if (sd) {
        if (check_type) {
          DataDomain::ptr dd=sd->getType();
          if (dd->getType()==type) {
            rv=true;
          }
        } else {
          rv=true;
        }
      }

      if (debugLevel>=4) {
        cerr << "Result = " << rv <<endl;
      }
      
      return rv;
    }
  };


  struct is_io_port {
    const interface_stack_t& stack_ref;
    std::string& id;
    const int& index;
    bool has_index;

    is_io_port(const interface_stack_t& stack_ref_,
               std::string& id_) : stack_ref(stack_ref_), 
                              id(id_),
                              index(0),
                              has_index(false) {}

    is_io_port(const interface_stack_t& stack_ref_,
               std::string& id_,
               const int& index_) : stack_ref(stack_ref_), 
                                    id(id_),
                                    index(index_),
                                    has_index(true) {}


    const bool
    operator()() const {
      if (debugLevel>=4) {
        cerr << "Checking for I/O port ";
        if (has_index) {
          cerr << id << "[" << index << "]";
        } else {
          cerr << id;
        }
        cerr << " in interface " <<
          stack_ref.top() << endl;
      }

      bool rv=false;

      if (has_index) {
        std::string full_name=id+"["+boost::lexical_cast<std::string>(index)+"]";
        
        if (index < 0) {
          throw "index < 0: " + full_name;
        }

        rv=(bool)(stack_ref.top()->getPortDescription(full_name));
      } else {
        rv=(bool)(stack_ref.top()->getPortDescription(id));

        if (!rv) {
          // try with id[0]
          rv=(bool)(stack_ref.top()->getPortDescription(id+"[0]"));
        }
      }
      
      if (debugLevel>=4) {
        cerr << "Result = " << rv <<endl;
      }
      return rv;
    }
  };
  
  struct is_ap {
    const interface_stack_t& stack_ref;
    std::string& id;
    
    is_ap(const interface_stack_t& stack_ref_,
          std::string& id_) :
      stack_ref(stack_ref_), 
      id(id_) {}

    const bool
    operator()() const {
      if (debugLevel>=4) {
        cerr << "Checking for atomic proposition " << id << " in interface " <<
          stack_ref.top() << endl;
      }
      
      bool rv = stack_ref.top()->hasAtomicProposition(id);
      
      if (debugLevel>=4) {
        cerr << "Result = " << rv <<endl;
      }
      
      return rv;
    }
  };

  struct is_function {
    const guard_parser& self;
    std::string& id;
    
    is_function(const guard_parser& self_,
                std::string& id_) :
      self(self_),
      id(id_) {}

    const bool
    operator()() const {
      if (debugLevel>=4) {
        cerr << "Checking for function " << id << endl;
      }
      
      SymbolData::ptr s = self.symbol_table_stack.find(id);
      if (!s) {
        throw "Trying to call function "+id+"(...), is unknown!";
      }

      if (s->type != SYMBOL_TYPE_FUNCTION && s->type != SYMBOL_TYPE_TYPEDEF) {
        throw "Trying to call function "+id+"(...), is not a function!";
      }

      return true;
    }
  };


  struct is_virtual_value {
    const symbol_table_stack_t& stack_ref;
    std::string& id;
    bool check_type;
    DataDomain::data_domain_type type;

    is_virtual_value(const symbol_table_stack_t& stack_ref_,
                     std::string& id_) :
      stack_ref(stack_ref_), 
      id(id_),
      check_type(false) {}

    is_virtual_value(const symbol_table_stack_t& stack_ref_,
                     std::string& id_,
                     DataDomain::data_domain_type type_) :
      stack_ref(stack_ref_), 
      id(id_),
      check_type(true),
      type(type_) {}

    const bool
    operator()() const {
      bool rv=false;
      SymbolData::ptr sd=stack_ref.find(id);
      if (sd && sd->type==SYMBOL_TYPE_VIRTUAL_VALUE) {
        if (check_type) {
          DataDomain::ptr dd=static_cast<VirtualValueData&>(*sd).getDataDomain();
          if (dd->getType()==type) {
            rv=true;
          }
        } else {
          rv=true;
        }
      }
      
      return rv;
    }
  };

  
  struct reset_interface_stack {
    const guard_parser& self;
    interface_stack_t& stack_ref;

    reset_interface_stack(const guard_parser& self_,
                          interface_stack_t& stack_ref_) :
      self(self_), stack_ref(stack_ref_) {}

    template <class Iterator>
    const void
    operator()(Iterator& it1, Iterator& it2) const {
      while (!stack_ref.empty()) stack_ref.pop();

      stack_ref.push(self.component_interface);
    }
  };


  struct init {
    guard_parser& self;
    bool allow_states, allow_ports, allow_labels;
    bool change_strong_guard, strong_guard;

    init(guard_parser& self_,
         bool allow_states_,
         bool allow_ports_,
         bool allow_labels_) :
      self(self_), allow_states(allow_states_), allow_ports(allow_ports_), allow_labels(allow_labels_),
      change_strong_guard(false) {}

    init(guard_parser& self_,
         bool allow_states_,
         bool allow_ports_,
         bool allow_labels_,
         bool strong_guard_) :
      self(self_), allow_states(allow_states_), allow_ports(allow_ports_), allow_labels(allow_labels_),
      change_strong_guard(true), strong_guard(strong_guard_) {}

    template <class Iterator>
    const void
    operator()(Iterator& it1, Iterator& it2) const {
      if (change_strong_guard) {
        self.initialize(allow_states, allow_ports, allow_labels, strong_guard);
      } else {
        self.initialize(allow_states, allow_ports, allow_labels);
      }
    }
  };


  struct enter_subcomponent {
    const guard_parser& self;
    interface_stack_t& stack_ref;


    enter_subcomponent(const guard_parser& self_, interface_stack_t& stack_ref_) :
      self(self_),
      stack_ref(stack_ref_) {}

    template <class Node, class Iterator>
    const void
    operator()(const Node& node, Iterator& it1, Iterator& it2) const {
      std::string name=guard_helper::t_subcomponent_access::getSubcomponent(node);
      unsigned int index=0;
      bool used_index=false;
      if (guard_helper::t_subcomponent_access::hasSubcomponentIndex(node)) {
        index=guard_helper::t_subcomponent_access::getSubcomponentIndex(node, self.symbol_table_stack);
        used_index=true;
      }

      std::string full_name=name+"["+boost::lexical_cast<std::string>(index)+"]";
      ComponentInterface::ptr sub_interface=stack_ref.top()->getSubcomponentInterface(full_name);

      if (!sub_interface && index==0) {
        // try without index
        sub_interface=stack_ref.top()->getSubcomponentInterface(name);
      }
      
      if (!sub_interface) {
        throw_(it1, is_not_a_subcomponent);
      }

      stack_ref.push(sub_interface);
    }
  };


  struct exit_subcomponent {
    interface_stack_t& stack_ref;

    exit_subcomponent(interface_stack_t& stack_ref_) :
      stack_ref(stack_ref_) {}

    template <class Node, class Iterator>
    const void
    operator()(const Node& node, Iterator& it1, Iterator& it2) const {
      stack_ref.pop();
    }
  };


  struct determine_tmp_var_type {
    bool& bool_flag;

    determine_tmp_var_type(bool& bool_flag_) :
      bool_flag(bool_flag_) {}

    template <class Node, class Iterator>
    const void
    operator()(const Node& node, Iterator& it1, Iterator& it2) const {
      bool_flag=type_helper::t_type::isBoolDataType(node);
    }
  };

  struct enter_parametrized_expression {
    stack<symbol_table_ptr>& ranged_symbol_stack;
    symbol_table_stack_t& stack_ref;

    enter_parametrized_expression(stack<symbol_table_ptr>& ranged_symbol_stack_,
                                  symbol_table_stack_t& stack_ref_) :
      ranged_symbol_stack(ranged_symbol_stack_),
      stack_ref(stack_ref_) {}

    template <class Node, class Iterator>
    const void
    operator()(const Node& node, Iterator& it1, Iterator& it2) const {
      if (node.value.id() == parser_id(ID_G_range_def)) {
        handle_range_def(node, it1, it2);
      } else if (node.value.id() == parser_id(ID_G_tmp_var_def)) {
        handle_tmp_var_def(node, it1, it2);
      } else {
        THROW_UNIMPLEMENTED();
      }
    }

    template <class Node, class Iterator>
    const void
    handle_range_def(const Node& node, Iterator& it1, Iterator& it2) const {
      const string range_var =guard_helper::t_range_def::getRangeVar(node);
      const Node&  range_type_node=guard_helper::t_range_def::getRangeType(node);

      DataDomain::ptr range_type=type_helper::t_type::getDataDomain(range_type_node, stack_ref);

      if (range_type->getType()!=DataDomain::DD_FLAT ||
          static_cast<FlatDataDomain&>(*range_type).getFlatType()!=FlatDataDomain::DD_INT) {
        throw_(it1, "Not an integer range in range definition, '"+range_type->toString()+"' instead.");
      }

      IntRangeDataDomain::ptr range_int_dd = 
        boost::static_pointer_cast<IntRangeDataDomain>(range_type);

      symbol_table_t::ptr range_symbol_table(new symbol_table_t);
      
      // pick one of the values to put as constant in symbol table
      // for compilation reasons, will be replaced for each possible
      // value later during evaluation      
      int one_value = range_int_dd->getLowestInt();
      ConstantData::ptr cd(new ConstantData(range_var, one_value));

      range_symbol_table->add(range_var, cd);
      range_symbol_table->read_only=true;
      
      ranged_symbol_stack.push(range_symbol_table);
      stack_ref.push(range_symbol_table);
    }

    
    template <class Node, class Iterator>
    const void
    handle_tmp_var_def(const Node& node, Iterator& it1, Iterator& it2) const {
      const string tmp_var = guard_helper::t_tmp_var_def::getTmpVarName(node);
      const Node&  tmp_var_type_n = guard_helper::t_tmp_var_def::getTmpVarType(node);

      DataDomain::ptr tmp_var_type=type_helper::t_type::getDataDomain(tmp_var_type_n, stack_ref);

      symbol_table_t::ptr tmp_var_symbol_table(new symbol_table_t);
      
      tmp_var_symbol_table->add(tmp_var, 
                                VirtualValueData::ptr(new VirtualValueData(tmp_var, tmp_var_type)));
      tmp_var_symbol_table->read_only=true;
      
      ranged_symbol_stack.push(tmp_var_symbol_table);
      stack_ref.push(tmp_var_symbol_table);
    }

  };


  struct exit_parametrized_expression {
    stack<symbol_table_ptr>& ranged_symbol_stack;
    symbol_table_stack_t& stack_ref;
    unsigned int symbols;

    exit_parametrized_expression(stack<symbol_table_ptr>& ranged_symbol_stack_,
                                 symbol_table_stack_t& stack_ref_,
                                 unsigned int symbols_=1) :
      ranged_symbol_stack(ranged_symbol_stack_),
      stack_ref(stack_ref_),
      symbols(symbols_) {}

    template <class Iterator>
    const void
    operator()(Iterator& it1, Iterator& it2) const {
      for (unsigned int i=0;i<symbols;i++) {
        stack_ref.pop();
        ranged_symbol_stack.pop();
      }
    }
  };


  // this is the general error handler
  template <class DefT>
  struct error_handler {
    const DefT& def;

    error_handler(const DefT& def_) : def(def_) {}

    template <typename ScannerT, typename ErrorT>
    error_status<>
    operator()(ScannerT const& scan, ErrorT const& error) const {
      switch (error.descriptor) {
      case state_var_expected:
        throw_(error.where, std::string("Expected a state variable"));
      case io_port_expected:
        throw_(error.where, std::string("Expected a node name or I/O port"));

      case no_compare_op:
        throw_(error.where, std::string("Expected a comparison operator"));
      case no_assignment_op:
        throw_(error.where, std::string("Expected ':='"));

      case no_active_port_name:
        throw_(error.where, std::string("Expected a port name in active port expression"));

      case no_state_guard_expression:
        throw_(error.where, std::string("Expected a valid expression on right side of comparision of state guard"));
      case no_io_guard_expression:
        throw_(error.where, std::string("Expected a valid expression on right side of comparision of I/O guard"));
      case no_assignment_expression:
        throw_(error.where, std::string("Expected a valid expression on right side of assignment"));

      case is_not_a_subcomponent:
        return error_status<>(error_status<>::fail);

      case no_left_parenthesis:
        throw_(error.where, std::string("Expected opening '('"));
      case no_right_parenthesis:
        throw_(error.where, std::string("Expected closing ')'"));

      case no_right_square:
        throw_(error.where, std::string("Expected closing ']'"));
      case no_right_curly:
        throw_(error.where, std::string("Expected closing '}'"));
      case no_comma:
        throw_(error.where, std::string("Expected ','"));
      case no_semicolon:
        throw_(error.where, std::string("Expected ';'"));

      case bool_expression_expected:
        throw_(error.where, std::string("Boolean expression expected"));
      case expression_expected:
        throw_(error.where, std::string("Valid expression expected"));
      case range_def_expected:
        throw_(error.where, std::string("Valid range definition expected"));
      }
      return error_status<>(error_status<>::rethrow);
    }
  };


  void initialize(bool allow_states_, bool allow_ports_, bool allow_labels_, bool strong_guard_=false) {
    while (!interface_stack.empty()) interface_stack.pop();
    interface_stack.push(component_interface);

    allow_states=allow_states_;
    allow_ports=allow_ports_;
    allow_labels=allow_labels_;
    strong_guard=strong_guard_;
  }

  guard_parser(symbol_table_stack_t& symbol_table_stack_,
               const symbols<>& reserved_keywords_,
               const ComponentInterface::ptr& component_interface_) :
    symbol_table_stack(symbol_table_stack_),
    reserved_keywords(reserved_keywords_),
    component_interface(component_interface_),
    allow_ports(true),
    allow_states(true),
    allow_labels(true),
    allow_subcomponents(true),
    allow_bool_expression_state_guard(true),
    allow_bool_expression_io_guard(true),
    strong_guard(false)
  {}
  
  symbol_table_stack_t& symbol_table_stack;
  const symbols<>& reserved_keywords;
  const vereofy::ComponentInterface::ptr &component_interface;
  interface_stack_t interface_stack;
  
  bool allow_ports;
  bool allow_states;
  bool allow_labels;
  bool allow_subcomponents;

  bool allow_bool_expression_state_guard;
  bool allow_bool_expression_io_guard;

  bool strong_guard;  // set to true if we expect a state_guard / io_guard to be actually there


  bool tmp_var_bool;

  template <typename ScannerT>
  struct definition
    : public grammar_def< rule<ScannerT, parser_tag<ID_G_any_value> >,
                          rule<ScannerT, parser_tag<ID_G_guard_expression> >,
                          rule<ScannerT, parser_tag<ID_G_state_guard> >,
                          rule<ScannerT, parser_tag<ID_G_io_guard> >,
                          rule<ScannerT, parser_tag<ID_G_state_assignment> >,
                          rule<ScannerT, parser_tag<ID_G_port_name> >,
                          rule<ScannerT, parser_tag<ID_G_bool_expression> >
                          >
  {
    definition(const guard_parser& self_)
      : identifier_grammar(last_identifier, self_.reserved_keywords),
        integer_grammar(self_.symbol_table_stack),
        type_grammar(self_.symbol_table_stack, self_.reserved_keywords),
        common_grammar(self_.reserved_keywords),
        interface_stack(const_cast<guard_parser&>(self_).interface_stack)

    {
      guard_parser& self=const_cast<guard_parser&>(self_);

      BOOST_SPIRIT_DEBUG_GRAMMAR(identifier_grammar);
      BOOST_SPIRIT_DEBUG_GRAMMAR(integer_grammar);
      BOOST_SPIRIT_DEBUG_GRAMMAR(type_grammar);
      BOOST_SPIRIT_DEBUG_GRAMMAR(common_grammar);
      
      ENTRY_GRAMMAR(identifier_p, identifier_parser, identifier_grammar, identifier_parser::identifier);
      ENTRY_GRAMMAR(enum_value_p, identifier_parser, identifier_grammar, identifier_parser::enum_value);

      ENTRY_GRAMMAR(integer_p, int_parser, integer_grammar, int_parser::integer);
      ENTRY_GRAMMAR(const_expression_p, int_parser, integer_grammar, int_parser::const_expression);
      ENTRY_GRAMMAR(const_ref_int_p, int_parser, integer_grammar, int_parser::const_ref_int);

      ENTRY_GRAMMAR(type_p, type_parser, type_grammar, type_parser::type);
      ENTRY_GRAMMAR(type_def_p, type_parser, type_grammar, type_parser::type_def);

      ENTRY_GRAMMAR(quoted_string, common_parser, common_grammar, common_parser::quoted_string);

      //  Start grammar definition

      any_value = keyword_p("ANY");

      function_call = identifier_p 
        >> no_node_d[ch_p('(')]
        >> eps_p(is_function(self, last_identifier))
        >> !(guard_expression >> *(no_node_d[ch_p(',')] >> guard_expression))
        >> no_node_d[ch_p(')')];
      
      bool_function_call = identifier_p 
        >> no_node_d[ch_p('(')]
        >> eps_p(is_function(self, last_identifier))
        >> !(guard_expression >> *(no_node_d[ch_p(',')] >> guard_expression))
        >> no_node_d[ch_p(')')];

      virtual_value =
        identifier_p
        >> eps_p(is_virtual_value(self.symbol_table_stack, last_identifier))
        >> !(struct_member_access | array_access);

      guard_factor =  integer_p
        | const_ref_int_p
        | virtual_value
        | (eps_p(phoenix::var(self.allow_states)) >> state_var)
        | (eps_p(phoenix::var(self.allow_ports)) >> port_data)
        | ite_expression
        | iterate_expression
        | function_call
        | enum_value_p
        | (no_node_d[ch_p('(')] >> guard_expression_full >> no_node_d[ch_p(')')])
        | (root_node_d[ch_p('-')] >> guard_factor);
        
        
      guard_term   = guard_factor >>
        *(  (root_node_d[ch_p('*')] >> guard_factor)
            | (root_node_d[ch_p('/')] >> guard_factor)
            | (root_node_d[ch_p('%')] >> guard_factor)
              );
        
      guard_expression_full = 
         eps_p[reset_interface_stack(self, interface_stack)]
         >> guard_term >>
         *(  (root_node_d[ch_p('+')] >> guard_term)
             | (root_node_d[ch_p('-')] >> guard_term)
             );

      guard_expression = 
        const_expression_p | guard_expression_full;


      compare_expression =
        guard_expression
        >> compare_op 
        >> guard_expression;
      
      range_def = 
        identifier_p 
        >> root_node_d[keyword_p("in")]
        >> type_p;

      tmp_var_def = 
        access_node_d[type_p][determine_tmp_var_type(self.tmp_var_bool)]
        >> identifier_p
        >> no_node_d[str_p(":=")]
        >> ((eps_p(phoenix::var(self.tmp_var_bool)) 
             >> bool_expression) |
            (eps_p(!phoenix::var(self.tmp_var_bool)) 
             >> guard_expression));

      iterate_expression =
        root_node_d[str_p("ITERATE")]
        >> no_node_d[ch_p('(')]
        >> access_node_d[range_def][enter_parametrized_expression(ranged_symbol_stack, self.symbol_table_stack)]
        >> expect.semicolon(no_node_d[ch_p(';')])
        >> access_node_d[tmp_var_def][enter_parametrized_expression(ranged_symbol_stack, self.symbol_table_stack)]
        >> expect.semicolon(no_node_d[ch_p(';')])
        >> ((eps_p(phoenix::var(self.tmp_var_bool)) 
             >> bool_expression) |
            (eps_p(!phoenix::var(self.tmp_var_bool)) 
             >> guard_expression))
        >> no_node_d[ch_p(')')]
        >> eps_p[exit_parametrized_expression(ranged_symbol_stack, self.symbol_table_stack, 2)];
        
      bool_expression_ranged = 
        root_node_d[keyword_p("AND") | keyword_p("OR")]
        >> no_node_d[ch_p('(')]
        >> access_node_d[range_def][enter_parametrized_expression(ranged_symbol_stack, self.symbol_table_stack)]
        >> no_node_d[ch_p(';')]
        >> bool_expression
        >> no_node_d[ch_p(')')]
        >> eps_p[exit_parametrized_expression(ranged_symbol_stack, self.symbol_table_stack)];

      bool_expression_term = 
        root_node_d[keyword_p("true")]
        | root_node_d[keyword_p("false")]
        | (no_node_d[ch_p('(')] >> bool_expression >> no_node_d[ch_p(')')])
        | (root_node_d[ch_p('!')] >> bool_expression_term)
        | bool_expression_ranged
        | ite_bool_expression
        | compare_expression
        | (eps_p(phoenix::var(self.allow_bool_expression_state_guard)) >> eps_p[phoenix::var(self.strong_guard)=false] >> root_node_d[state_guard])
        | (eps_p(phoenix::var(self.allow_bool_expression_io_guard)) >> eps_p[phoenix::var(self.strong_guard)=false] >> root_node_d[io_guard]);


      bool_expression_implies_equiv =
        bool_expression_term >> *( ((root_node_d[str_p("->")]) | (root_node_d[str_p("<->")])) >> bool_expression_term );

      bool_expression_and = bool_expression_implies_equiv >> *( root_node_d[str_p("&")] >> bool_expression_implies_equiv );
      bool_expression_xor = bool_expression_and >> *( root_node_d[str_p("^")] >> bool_expression_and );

      // exported
      bool_expression  = bool_expression_xor >> *( root_node_d[str_p("|")] >> bool_expression_xor );

      if_expression = 
        root_node_d[keyword_p("IF")]
        >> no_node_d[ch_p('(')]
        >> bool_expression
        >> no_node_d[ch_p(')')];

      ite_expression =
        root_node_d[keyword_p("ITE")]
        >> no_node_d[ch_p('(')]
        >> bool_expression
        >> no_node_d[ch_p(',')]
        >> guard_expression
        >> no_node_d[ch_p(',')]
        >> guard_expression
        >> no_node_d[ch_p(')')];

      ite_bool_expression =
        root_node_d[keyword_p("ITE")]
        >> no_node_d[ch_p('(')]
        >> bool_expression
        >> no_node_d[ch_p(',')]
        >> bool_expression
        >> no_node_d[ch_p(',')]
        >> bool_expression
        >> no_node_d[ch_p(')')];


      simple_state_var = 
         identifier_p
         >> eps_p(is_state_var(interface_stack, last_identifier))
         >> (eps_p(is_state_var(interface_stack, last_identifier, 
                                DataDomain::DD_FLAT))
             | (eps_p(is_state_var(interface_stack, last_identifier, 
                                   DataDomain::DD_STRUCT))
                >> !struct_member_access) // optionally has struct member access
             | (eps_p(is_state_var(interface_stack, last_identifier, 
                                  DataDomain::DD_TAGGED_UNION))
                >> !struct_member_access) // optionally has struct member access
             | (eps_p(is_state_var(interface_stack, last_identifier, 
                                   DataDomain::DD_ARRAY))
                >> !array_access)
             );

      subcomponent_access =
        error_guard( // fail if this is no subcomponent
                    access_node_d[
                                  eps_p(phoenix::var(self.allow_subcomponents))
                                  >> identifier_p
                                  >> !(no_node_d[ch_p('[')] >> const_expression_p >> no_node_d[ch_p(']')])
                                  >> root_node_d[ch_p('.')]
                                  ][enter_subcomponent(self, interface_stack)])[error_handler<guard_parser>(self)];                    

      // todo: add expect for identifier
      struct_member_access = 
        no_node_d[str_p(".")]
        >> identifier_p
        >> !struct_member_access;

      array_access =
        no_node_d[ch_p('[')]
        >> expect.expression(guard_expression)
        >> expect.right_square(no_node_d[ch_p(']')]);

      state_var = 
        eps_p[reset_interface_stack(self, interface_stack)] >>
        (*subcomponent_access >> simple_state_var);

      ap_ref =
        eps_p[reset_interface_stack(self, interface_stack)] >>
        (*subcomponent_access >> simple_ap_ref);

      simple_ap_ref =
        identifier_p
        >> eps_p(is_ap(interface_stack, last_identifier));

      port = 
        (identifier_p 
         >> (no_node_d[ch_p('[')] 
             >> access_node_d[const_expression_p][int_helper::s_const_expression_evaluate(self.symbol_table_stack, last_index)] 
             >> no_node_d[ch_p(']')])
         >> eps_p(is_io_port(interface_stack, last_identifier, last_index))
         ) |
        (identifier_p
         >> eps_p(is_io_port(interface_stack, last_identifier))
         );

      // exported
      port_name = 
        eps_p[reset_interface_stack(self, interface_stack)]
        >> port;

      port_data = eps_p(ch_p('#')) >> expect.io_port(no_node_d[ch_p('#')] >> port) >> !struct_member_access;
      
      active_ports = no_node_d[ch_p('{')] 
        >> expect.active_port_name(infix_node_d[list_p(port, ch_p(','))])
        >> no_node_d[expect.right_curly(ch_p('}'))];
      
      compare_op = root_node_d[
                               str_p("==") 
                               | str_p("!=")
                               | str_p("<=")
                               | str_p("<")
                               | str_p(">=")
                               | str_p(">")
                               ];
      
      port_activity_expression = 
        active_ports 
        | (!ch_p('!') >> 
           expect.io_port(port));
      
      port_data_expression = port_data 
        >> expect.compare_op(compare_op)
        >> expect.io_guard_expression(guard_expression);

      label_id = 
        eps_p[reset_interface_stack(self, interface_stack)] >>
        (*subcomponent_access >> identifier_p);

      has_label_expression = 
        no_node_d[keyword_p("has_label")]
        >> no_node_d[ch_p('(')]
        >> label_id
        >> no_node_d[ch_p(')')];

      io_guard = error_guard(
                               keyword_p("epsilon") 
                             | keyword_p("dataflow")
                             | keyword_p("step")
                             | (eps_p(phoenix::var(self.allow_labels)) >> has_label_expression)
                             | if_expression
                             | bool_function_call
                             | port_data_expression 
                             | port_activity_expression  // has to come last!
                             )[error_handler<guard_parser>(self)];

      state_guard = error_guard(
                                (
                                 (if_expression)
                                |
                                 (bool_function_call)
                                |
                                 (ap_ref)
                                |
                                 (
                                  ((eps_p(phoenix::var(self.strong_guard))  >> expect.state_var(state_var)) |
                                   (~eps_p(phoenix::var(self.strong_guard)) >> state_var))
                                  >> expect.compare_op(compare_op)
                                  >> expect.state_guard_expression(guard_expression)
                                  )
                                 )
                                )[error_handler<guard_parser>(self)];
      
      add_label = 
        no_node_d[keyword_p("label")]
        >> no_node_d[ch_p('(')]
        >> identifier_p
        >> no_node_d[ch_p(')')]
        >> !(no_node_d[str_p(":=")]
             >> expect.assignment_expression(any_value | guard_expression)); 
        
      state_assignment = 
        add_label 
        | error_guard(
                      expect.state_var(state_var)
                      >> root_node_d[expect.assignment_op(str_p(":="))] 
                      >> expect.assignment_expression(any_value | guard_expression) 
                      )[error_handler<guard_parser>(self)];

        //  End grammar definition

      this->start_parsers(any_value,
                          guard_expression,
                          state_guard,
                          io_guard,
                          state_assignment,
                          port_name,
                          bool_expression);

      // turn on the debugging info.
        BOOST_SPIRIT_DEBUG_RULE(guard_factor);
        BOOST_SPIRIT_DEBUG_RULE(guard_term);
        BOOST_SPIRIT_DEBUG_RULE(guard_expression_full);
        BOOST_SPIRIT_DEBUG_RULE(guard_expression);
        BOOST_SPIRIT_DEBUG_RULE(any_value);
        BOOST_SPIRIT_DEBUG_RULE(function_call);
        BOOST_SPIRIT_DEBUG_RULE(bool_function_call);

        BOOST_SPIRIT_DEBUG_RULE(range_def);
        BOOST_SPIRIT_DEBUG_RULE(tmp_var_def);
        BOOST_SPIRIT_DEBUG_RULE(iterate_expression);
        BOOST_SPIRIT_DEBUG_RULE(compare_expression);

        BOOST_SPIRIT_DEBUG_RULE(bool_expression_term);
        BOOST_SPIRIT_DEBUG_RULE(bool_expression_ranged);
        BOOST_SPIRIT_DEBUG_RULE(bool_expression_implies_equiv);
        BOOST_SPIRIT_DEBUG_RULE(bool_expression_and);
        BOOST_SPIRIT_DEBUG_RULE(bool_expression_xor);
        BOOST_SPIRIT_DEBUG_RULE(bool_expression);
        BOOST_SPIRIT_DEBUG_RULE(if_expression);
        BOOST_SPIRIT_DEBUG_RULE(ite_expression);
        BOOST_SPIRIT_DEBUG_RULE(ite_bool_expression);

        BOOST_SPIRIT_DEBUG_RULE(ap_ref);
        BOOST_SPIRIT_DEBUG_RULE(simple_ap_ref);
        BOOST_SPIRIT_DEBUG_RULE(port_name);

        BOOST_SPIRIT_DEBUG_RULE(state_var);
        BOOST_SPIRIT_DEBUG_RULE(simple_state_var);
        BOOST_SPIRIT_DEBUG_RULE(struct_member_access);
        BOOST_SPIRIT_DEBUG_RULE(array_access);
        BOOST_SPIRIT_DEBUG_RULE(subcomponent_access);
        BOOST_SPIRIT_DEBUG_RULE(port);
        BOOST_SPIRIT_DEBUG_RULE(port_data);
        BOOST_SPIRIT_DEBUG_RULE(active_ports);
        BOOST_SPIRIT_DEBUG_RULE(compare_op);
        BOOST_SPIRIT_DEBUG_RULE(port_activity_expression );
        BOOST_SPIRIT_DEBUG_RULE(port_data_expression);
        BOOST_SPIRIT_DEBUG_RULE(has_label_expression);
        BOOST_SPIRIT_DEBUG_RULE(label_id);

        BOOST_SPIRIT_DEBUG_RULE(state_guard);
        BOOST_SPIRIT_DEBUG_RULE(io_guard);
        BOOST_SPIRIT_DEBUG_RULE(state_assignment);
        BOOST_SPIRIT_DEBUG_RULE(add_label);

    }
    
    rule<ScannerT, parser_tag<ID_G_any_value> >  any_value;
    rule<ScannerT, parser_tag<ID_G_function_call> >  function_call;
    rule<ScannerT, parser_tag<ID_G_bool_function_call> >  bool_function_call;
    rule<ScannerT, parser_tag<ID_G_guard_factor> >  guard_factor;
    rule<ScannerT, parser_tag<ID_G_guard_term> >  guard_term;
    rule<ScannerT, parser_tag<ID_G_guard_expression_full> >  guard_expression_full;
    rule<ScannerT, parser_tag<ID_G_guard_expression> >  guard_expression;

    rule<ScannerT, parser_tag<ID_G_virtual_value> >  virtual_value;
    rule<ScannerT, parser_tag<ID_G_range_def> > range_def;
    rule<ScannerT, parser_tag<ID_G_tmp_var_def> > tmp_var_def;
    rule<ScannerT, parser_tag<ID_G_iterate_expression> > iterate_expression;
    rule<ScannerT, parser_tag<ID_G_compare_expression> > compare_expression;

    rule<ScannerT, parser_tag<ID_G_bool_expression_term> > bool_expression_term;
    rule<ScannerT, parser_tag<ID_G_bool_expression_ranged> > bool_expression_ranged;
    rule<ScannerT, parser_tag<ID_G_bool_expression_implies_equiv> > bool_expression_implies_equiv;
    rule<ScannerT, parser_tag<ID_G_bool_expression_and> > bool_expression_and;
    rule<ScannerT, parser_tag<ID_G_bool_expression_xor> > bool_expression_xor;
    rule<ScannerT, parser_tag<ID_G_bool_expression> > bool_expression;
    rule<ScannerT, parser_tag<ID_G_if_expression> > if_expression;
    rule<ScannerT, parser_tag<ID_G_ite_expression> > ite_expression;
    rule<ScannerT, parser_tag<ID_G_ite_bool_expression> > ite_bool_expression;
    
    rule<ScannerT, parser_tag<ID_G_state_var> >  state_var;
    rule<ScannerT, parser_tag<ID_G_simple_state_var> >  simple_state_var;
    rule<ScannerT, parser_tag<ID_G_ap_ref> >  ap_ref;
    rule<ScannerT, parser_tag<ID_G_simple_ap_ref> >  simple_ap_ref;
    rule<ScannerT, parser_tag<ID_G_subcomponent_access> >  subcomponent_access;
    rule<ScannerT, parser_tag<ID_G_struct_member_access> >  struct_member_access;
    rule<ScannerT, parser_tag<ID_G_array_access> >  array_access;
    rule<ScannerT, parser_tag<ID_G_port> >  port;
    rule<ScannerT, parser_tag<ID_G_port_name> >  port_name;
    rule<ScannerT, parser_tag<ID_G_port_data> >  port_data;
    rule<ScannerT, parser_tag<ID_G_active_ports> >  active_ports;
    rule<ScannerT, parser_tag<ID_G_compare_op> >  compare_op;
    rule<ScannerT, parser_tag<ID_G_port_activity_expression > >  port_activity_expression ;
    rule<ScannerT, parser_tag<ID_G_port_data_expression> >  port_data_expression;
    rule<ScannerT, parser_tag<ID_G_has_label_expression> >  has_label_expression;
    rule<ScannerT, parser_tag<ID_G_label_id> >  label_id;
    rule<ScannerT, parser_tag<ID_G_state_guard> >  state_guard;
    rule<ScannerT, parser_tag<ID_G_io_guard> >  io_guard;
    rule<ScannerT, parser_tag<ID_G_state_assignment> >  state_assignment;
    rule<ScannerT, parser_tag<ID_G_add_label> >  add_label;

    guard<Errors> error_guard;

    identifier_parser identifier_grammar;
    int_parser integer_grammar;
    type_parser type_grammar;
    common_parser common_grammar;

    expect_parsers expect;
    
    bool factor_allow_port_data;
    bool factor_allow_state_var;

    interface_stack_t& interface_stack;
    stack<symbol_table_ptr> ranged_symbol_stack;

    std::string last_identifier;
    int last_index;
  };
};

}
}
#endif
