/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef GUARD_PARSER_HELPER_HPP
#define GUARD_PARSER_HELPER_HPP

#include "vereofy/common/checks.hpp"

#include "int_parser_helper.hpp"

namespace vereofy {
namespace vparser {

struct guard_helper {

  struct t_generic {
    static rule_id getType(const pt_node& n) {
      return (rule_id)n.value.id().to_long();
    }
  };

  struct t_generic_alternative {
    static const pt_node& get(const pt_node& n) {
      return n.children.at(0);
    }

    template <class IDType>
    static const bool is(const pt_node& n, IDType id) {
      return n.children.size()==1 &&
        n.children.at(0).value.id()==parser_id(id);
    }    
  };


  struct t_generic_identifier_wrapper {
    static string getIdentifier(const pt_node& n) {
      return string(n.children.at(0).value.begin(),
                    n.children.at(0).value.end());
    }
  };

  struct t_any_value {};

  struct t_compare_op {
    static comp_op_t getOp(const pt_node& n) {
      ID_ASSERT_(n, ID_G_compare_op);
      string op(n.value.begin(), n.value.end());

      if (op=="==") {
        return COMPARE_EQUAL;
      } else if (op=="!=") {
        return COMPARE_NOTEQUAL;
      } else if (op=="<") {
        return COMPARE_LESS;
      } else if (op=="<=") {
        return COMPARE_LESSEQUAL;
      } else if (op==">") {
        return COMPARE_GREATER;
      } else if (op==">=") {
        return COMPARE_GREATEREQUAL;
      } else {
        throw "implementation error";
      }
    }
  };

  struct t_function_call {
    // no ID_ASSERT_ checking for t_function_call, as it is aliased with t_bool_function_call
    static std::string getFunctionName(const pt_node& n) {
      return identifier_helper::t_identifier::getIdentifier(n.children.at(0));
    }

    static unsigned int getArity(const pt_node& n) {
      return n.children.size()-1;
    }


    static const pt_node& getArgumentExpression(const pt_node& n, unsigned int index) {
      return n.children.at(index+1);
    }
      
  };

  struct t_bool_function_call : public t_function_call {};

  struct t_if_expression {
    static const pt_node& getExpression(const pt_node& n) {
      ID_ASSERT_(n, ID_G_if_expression);
      return n.children.at(0);
    }
  };

  struct t_ite_expression {
    static const pt_node& getIf(const pt_node& n) {
      ID_ASSERT_(n, ID_G_ite_expression);
      return n.children.at(0);
    }

    static const pt_node& getThen(const pt_node& n) {
      ID_ASSERT_(n, ID_G_ite_expression);
      return n.children.at(1);
    }

    static const pt_node& getElse(const pt_node& n) {
      ID_ASSERT_(n, ID_G_ite_expression);
      return n.children.at(2);
    }
  };


  struct t_ite_bool_expression {
    static const pt_node& getIf(const pt_node& n) {
      ID_ASSERT_(n, ID_G_ite_bool_expression);
      return n.children.at(0);
    }

    static const pt_node& getThen(const pt_node& n) {
      ID_ASSERT_(n, ID_G_ite_bool_expression);
      return n.children.at(1);
    }

    static const pt_node& getElse(const pt_node& n) {
      ID_ASSERT_(n, ID_G_ite_bool_expression);
      return n.children.at(2);
    }
  };


  struct t_guard_factor {
    static bool isNegation(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_factor);
      return string(n.value.begin(), n.value.end())=="-";
    }

    static rule_id getChildType(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_factor);
      return (rule_id)n.children.at(0).value.id().to_long();
    }

    static const pt_node& getChild(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_factor);
      return n.children.at(0);
    }

    static bool noOps(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_factor);
      return !isNegation(n) && (getChildType(n)!=ID_G_guard_expression_full ||
                                // have a guard_expression_full
                                t_guard_expression_full::noOps(getChild(n))) ;
    }

    static const pt_node& getNoOpTerminal(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_factor);
      ERROR_CHECK_ASSERT(!isNegation(n));
      if (getChildType(n)==ID_G_guard_expression_full) {
        return t_guard_expression_full::getNoOpTerminal(getChild(n));
      } else {
        return getChild(n);
      }
    }
  };

  struct t_guard_term {
    static bool hasOp(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_term);
      return n.children.size()>1;
    }

    static string getOp(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_term);
      return string(n.value.begin(), n.value.end());
    }

    static const pt_node& getLeft(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_term);
      return n.children.at(0);
    }

    static const pt_node& getRight(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_term);
      return n.children.at(1);
    }

    static bool noOps(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_term);
      return !hasOp(n) && t_guard_factor::noOps(getLeft(n));
    }

    static const pt_node& getNoOpTerminal(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_term);
      ERROR_CHECK_ASSERT(!hasOp(n));
      return t_guard_factor::getNoOpTerminal(getLeft(n));
    }
  };

  struct t_guard_expression_full {
    static bool hasOp(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_expression_full);
      return n.children.size()>1;
    }

    static string getOp(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_expression_full);
      return string(n.value.begin(), n.value.end());
    }

    static const pt_node& getLeft(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_expression_full);
      return n.children.at(0);
    }

    static const pt_node& getRight(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_expression_full);
      return n.children.at(1);
    }

    static bool noOps(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_expression_full);
      return !hasOp(n) && t_guard_term::noOps(getLeft(n));
    }

    static const pt_node& getNoOpTerminal(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_expression_full);
      ERROR_CHECK_ASSERT(!hasOp(n));
      return t_guard_term::getNoOpTerminal(getLeft(n));
    }
  };

  struct t_guard_generic {

    static bool isIntExpressionType(const pt_node& n) {
      if (n.value.id()==parser_id(ID_G_guard_expression_full) ||
          n.value.id()==parser_id(ID_G_guard_term) ||
          n.value.id()==parser_id(ID_G_guard_factor)) {
        return true;
      }
      return false;
    }
      
    static bool noOps(const pt_node& n) {
      if (n.value.id()==parser_id(ID_G_guard_expression)) {
        return t_guard_expression_full::noOps(n);
      } else if (n.value.id()==parser_id(ID_G_guard_term)) {
        return t_guard_term::noOps(n);
      } else if (n.value.id()==parser_id(ID_G_guard_factor)) {
        return t_guard_factor::noOps(n);
      } else {
        THROW_UNIMPLEMENTED();
      }      
    }

    static const pt_node& getNoOpTerminal(const pt_node& n) {
      if (n.value.id()==parser_id(ID_G_guard_expression_full)) {
        return t_guard_expression_full::getNoOpTerminal(n);
      } else if (n.value.id()==parser_id(ID_G_guard_term)) {
        return t_guard_term::getNoOpTerminal(n);
      } else if (n.value.id()==parser_id(ID_G_guard_factor)) {
        return t_guard_factor::getNoOpTerminal(n);
      } else {
        THROW_UNIMPLEMENTED();
      }      
    }
  };

  struct t_guard_expression {
    static bool isConstantInt(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_expression);
      return n.children.at(0).value.id()==parser_id(ID_INT_const_expression);
    }

    static bool isFullExpression(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_expression);
      return n.children.at(0).value.id()==parser_id(ID_G_guard_expression_full);
    }

    static int_value_t getConstantInt(const pt_node& n,
                                      const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_G_guard_expression);
      ERROR_CHECK_ASSERT(isConstantInt(n));

      return int_helper::t_const_expression::evaluate(n.children.at(0), symbols);
    }

    static const pt_node& getFullExpression(const pt_node& n) {
      ID_ASSERT_(n, ID_G_guard_expression);
      return n.children.at(0);
    }
  };

  struct t_range_def {
    static string getRangeVar(const pt_node& n) {
      ID_ASSERT_(n, ID_G_range_def);
      return identifier_helper::t_identifier::getIdentifier(n.children.at(0));
    }

    static const pt_node& getRangeType(const pt_node& n) {
      ID_ASSERT_(n, ID_G_range_def);
      return n.children.at(1);
    }
  };

  struct t_tmp_var_def {
    static string getTmpVarName(const pt_node& n) {
      ID_ASSERT_(n, ID_G_tmp_var_def);
      return identifier_helper::t_identifier::getIdentifier(n.children.at(1));
    }

    static const pt_node& getTmpVarType(const pt_node& n) {
      ID_ASSERT_(n, ID_G_tmp_var_def);
      return n.children.at(0);
    }

    static const pt_node& getTmpVarInit(const pt_node& n) {
      ID_ASSERT_(n, ID_G_tmp_var_def);
      return n.children.at(2);
    }
  };


  struct t_bool_generic {
    static bool hasOp(const pt_node& n) {
      return n.children.size()>1;
    }

    static string getOp(const pt_node& n) {
      return string(n.value.begin(), n.value.end());
    }

    static const pt_node& getLeft(const pt_node& n) {
      return n.children.at(0);
    }

    static const pt_node& getRight(const pt_node& n) {
      return n.children.at(1);
    }
  };

  struct t_bool_expression_implies_equiv : public t_bool_generic {};
  struct t_bool_expression_and : public t_bool_generic {};
  struct t_bool_expression_xor : public t_bool_generic {};

  struct t_bool_expression_term {
    static bool isNegated(const pt_node& n) {
      ID_ASSERT_(n, ID_G_bool_expression_term);
      return string(n.value.begin(), n.value.end())=="!";
    }

    static bool isTrue(const pt_node& n) {
      ID_ASSERT_(n, ID_G_bool_expression_term);
      return string(n.value.begin(), n.value.end())=="true";
    }

    static bool isFalse(const pt_node& n) {
      ID_ASSERT_(n, ID_G_bool_expression_term);
      return string(n.value.begin(), n.value.end())=="false";
    }

    static rule_id getChildType(const pt_node& n) {
      ID_ASSERT_(n, ID_G_bool_expression_term);
      return (rule_id)n.children.at(0).value.id().to_long();
    }

    static const pt_node& getChild(const pt_node& n) {
      ID_ASSERT_(n, ID_G_bool_expression_term);
      return n.children.at(0);
    }
  };

  struct t_bool_expression_ranged {
    static string getOperator(const pt_node& n) {
      ID_ASSERT_(n, ID_G_bool_expression_ranged);
      return string(n.value.begin(), n.value.end());
    }
    
    static const pt_node& getRangeDef(const pt_node& n) {
      ID_ASSERT_(n, ID_G_bool_expression_ranged);
      return n.children.at(0);
    }

    static const pt_node& getExpression(const pt_node& n) {
      ID_ASSERT_(n, ID_G_bool_expression_ranged);
      return n.children.at(1);
    }
  };


  struct t_iterate_expression {

    template <class Node>
    static const Node& getRangeDef(const Node& n) {
      return n.children.at(0);
    }

    template <class Node>
    static const Node& getTmpVar(const Node& n) {
      return n.children.at(1);
    }

    template <class Node>
    static const Node& getExpression(const Node& n) {
      return n.children.at(2);
    }
  };


  struct port_name_t {
    port_name_t() : index(0), has_index(false) {}

    port_name_t(const string& name_, unsigned int index_) :
      name(name_), index(index_), has_index(true) {}

    port_name_t(const string& name_) :
      name(name_), index(0), has_index(false) {}

    std::string toString() const {
      return name + (has_index ? 
                     string("[")+boost::lexical_cast<std::string>(index)+"]" : 
                     string(""));
    }

    string name;
    unsigned int index;
    bool has_index;
  };


  struct t_port : public t_generic_identifier_wrapper {
    static bool hasIndex(const pt_node& n) {
      ID_ASSERT_(n, ID_G_port);
      return n.children.size()>1;
    }

    static unsigned int getIndex(const pt_node& n, const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_G_port);
      ERROR_CHECK_ASSERT(hasIndex(n));
      
      int_value_t port_index = 
        int_helper::t_const_expression::evaluate(n.children.at(1), symbols);

      if (!port_index.isDefined()) {
        throw "Port index is undefined for port "+getIdentifier(n);
      } else if (port_index.getValue()<0) {
        throw "Port index is negative: "+boost::lexical_cast<std::string>(port_index.getValue())+" for port "+getIdentifier(n);
      }

      return (unsigned int)port_index.getValue();
    }
    
    static port_name_t getPortName(const pt_node& n, const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_G_port);
      if (hasIndex(n)) {
        return port_name_t(getIdentifier(n), getIndex(n, symbols));
      } else {
        return port_name_t(getIdentifier(n));
      }
    }

    static std::string resolvePortName(const pt_node& n, const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_G_port);
      return getPortName(n, symbols).toString();
    }
  };

  struct t_port_name {
    static const pt_node& getPort(const pt_node& n) {
      ID_ASSERT_(n, ID_G_port_name);
      return n.children.at(0);
    }


    static std::string getPortName(const pt_node& n, const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_G_port_name);
      return t_port::resolvePortName(getPort(n), symbols);
    }
  };

  struct t_port_data {
    static port_name_t getPortName(const pt_node& n, const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_G_port_data);
      return t_port::getPortName(n.children.at(0), symbols);
    }

    static bool hasStructMemberAccess(const pt_node& n) {
      ID_ASSERT_(n, ID_G_port_data);
      return n.children.size()>1 && n.children.at(1).value.id()==parser_id(ID_G_struct_member_access);
    }

    static const pt_node& getStructMemberAccess(const pt_node& n) {
      ID_ASSERT_(n, ID_G_port_data);
      ERROR_CHECK_ASSERT(hasStructMemberAccess(n));
      return n.children.at(1);
    }

    
  };

  struct t_array_access {
    static const pt_node& getExpression(const pt_node& n) {
      ID_ASSERT_(n, ID_G_array_access);
      return n.children.at(0);
    }
  };


  struct t_simple_state_var : public t_generic_identifier_wrapper {

    static bool hasArrayAccess(const pt_node& n) {
      ID_ASSERT_2(n, ID_G_simple_state_var, ID_G_virtual_value);
      return n.children.size()==2 &&
        n.children.at(1).value.id()==parser_id(ID_G_array_access);
    }

    static const pt_node& getArrayAccess(const pt_node& n) {
      ID_ASSERT_2(n, ID_G_simple_state_var, ID_G_virtual_value);
      ERROR_CHECK_ASSERT(hasArrayAccess(n));
      return n.children.at(1);
    }

    static const pt_node& getArrayAccessExpression(const pt_node& n) {
      ID_ASSERT_2(n, ID_G_simple_state_var, ID_G_virtual_value);
      ERROR_CHECK_ASSERT(hasArrayAccess(n));
      return t_array_access::getExpression(n.children.at(1));
    }

    static bool hasStructMemberAccess(const pt_node& n) {
      ID_ASSERT_2(n, ID_G_simple_state_var, ID_G_virtual_value);
      return n.children.size()==2 &&
        n.children.at(1).value.id()==parser_id(ID_G_struct_member_access);
    }
    
    static
    const pt_node& getStructMemberAccess(const pt_node& n) {
      ID_ASSERT_2(n, ID_G_simple_state_var, ID_G_virtual_value);
      ERROR_CHECK_ASSERT(hasStructMemberAccess(n));
      return n.children.at(1);
    }

  };

  struct t_virtual_value : public t_simple_state_var {};

  struct t_subcomponent_access : public t_generic_identifier_wrapper {
    static bool hasSubcomponentIndex(const pt_node& n) {
      ID_ASSERT_(n, ID_G_subcomponent_access);
      return n.children.size()==2;
    }

    static string getSubcomponent(const pt_node& n) {
      ID_ASSERT_(n, ID_G_subcomponent_access);
      return identifier_helper::t_identifier::getIdentifier(n.children.at(0));
    }

    static bool isSubcomponentIndexDefined(const pt_node& n,
                                           const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_G_subcomponent_access);
      ERROR_CHECK_ASSERT(hasSubcomponentIndex(n));
      
      int_value_t index = int_helper::t_const_expression::evaluate(n.children.at(1), symbols);
      return index.isDefined();      
    }

    static unsigned int getSubcomponentIndex(const pt_node& n,
                                             const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_G_subcomponent_access);
      ERROR_CHECK_ASSERT(hasSubcomponentIndex(n));

      int_value_t index = int_helper::t_const_expression::evaluate(n.children.at(1), symbols);

      if (!index.isDefined()) {
        throw "Index in subcomponent access is undefined for "+getSubcomponent(n);
      } else if (index.getValue()<0) {
        throw "Index in subcomponent access is negative: "+boost::lexical_cast<std::string>(index.getValue())+" for "+getSubcomponent(n);
      }

      return index.getValue();
    }
  };

  struct t_struct_member_access : public t_generic_identifier_wrapper {
    static bool hasStructMemberAccess(const pt_node& n) {
      ID_ASSERT_(n, ID_G_struct_member_access);
      return n.children.size()==2 &&
        n.children.at(1).value.id()==parser_id(ID_G_struct_member_access);
    }
    
    static
    const pt_node& getStructMemberAccess(const pt_node& n) {
      ID_ASSERT_(n, ID_G_struct_member_access);
      ERROR_CHECK_ASSERT(hasStructMemberAccess(n));
      return n.children.at(1);
    }
  };

  struct t_generic_subcomponent_wrapper {
    static bool hasSubcomponentAccess(const pt_node& n) {
      return n.children.size()>1 &&
        n.children.at(0).value.id()==parser_id(ID_G_subcomponent_access);
    }

    static unsigned int countSubcomponentAccess(const pt_node& n) {
      return n.children.size()-1;
    }

    static const pt_node& getSubcomponent(const pt_node& n, unsigned int i) {
      ERROR_CHECK_ASSERT(hasSubcomponentAccess(n));
      ERROR_CHECK_ASSERT(i<countSubcomponentAccess(n));
      ERROR_CHECK_ASSERT(n.children.at(i).value.id()==parser_id(ID_G_subcomponent_access));

      return n.children.at(i);
    }
    
    static const pt_node& getSimple(const pt_node& n) {
      return n.children.at(n.children.size()-1);
    }
  };

  struct t_state_var : public t_generic_subcomponent_wrapper {};

  struct t_simple_ap_ref : public t_generic_identifier_wrapper {};
  struct t_ap_ref : public t_generic_subcomponent_wrapper {};

  struct t_active_ports {
    static port_name_t getPortName(const pt_node& n, unsigned int index, const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_G_active_ports);
      return t_port::getPortName(n.children.at(index), symbols);
    }

    static unsigned int getPortCount(const pt_node& n) {
      ID_ASSERT_(n, ID_G_active_ports);
      return n.children.size();
    }
  };

  struct t_port_data_expression {
    static comp_op_t getCompareOp(const pt_node& n) {
      ID_ASSERT_(n, ID_G_port_data_expression);
      return t_compare_op::getOp(n.children.at(1));
    }    
   
    static const pt_node& getLHS(const pt_node& n) {
      ID_ASSERT_(n, ID_G_port_data_expression);
      return n.children.at(0);
    }

    static const pt_node& getRHS(const pt_node& n) {
      ID_ASSERT_(n, ID_G_port_data_expression);
      return n.children.at(2);
    }
  };

  struct t_compare_expression {
    static comp_op_t getCompareOp(const pt_node& n) {
      ID_ASSERT_(n, ID_G_compare_expression);
      return t_compare_op::getOp(n.children.at(1));
    }    
   
    static const pt_node& getLHS(const pt_node& n) {
      ID_ASSERT_(n, ID_G_compare_expression);
      return n.children.at(0);
    }

    static const pt_node& getRHS(const pt_node& n) {
      ID_ASSERT_(n, ID_G_compare_expression);
      return n.children.at(2);
    }
  };

  struct t_state_guard {
    static bool isBoolFunctionCall(const pt_node& n) {
      ID_ASSERT_(n, ID_G_state_guard);
      if (n.children.size()==1 &&
          n.children.at(0).value.id() == parser_id(ID_G_bool_function_call)) {
        return true;
      }
      return false;
    }

    static bool isIfExpression(const pt_node& n) {
      ID_ASSERT_(n, ID_G_state_guard);
      if (n.children.size()==1 &&
          n.children.at(0).value.id() == parser_id(ID_G_if_expression)) {
        return true;
      }
      return false;
    }


    static bool isAPRef(const pt_node& n) {
      ID_ASSERT_(n, ID_G_state_guard);
      if (n.children.size()==1 &&
          n.children.at(0).value.id() == parser_id(ID_G_ap_ref)) {
        return true;
      }
      return false;
    }

    static comp_op_t getCompareOp(const pt_node& n) {
      ID_ASSERT_(n, ID_G_state_guard);
      return t_compare_op::getOp(n.children.at(1));
    }    
   
    static const pt_node& getLHS(const pt_node& n) {
      ID_ASSERT_(n, ID_G_state_guard);
      return n.children.at(0);
    }

    static const pt_node& getRHS(const pt_node& n) {
      ID_ASSERT_(n, ID_G_state_guard);
      return n.children.at(2);
    }
  };

  struct t_state_assignment {
    static bool is_add_label(const pt_node& n) {
      ID_ASSERT_(n, ID_G_state_assignment);
      return (n.children.at(0).value.id() == parser_id(ID_G_add_label));
    }

    static const pt_node& get_add_label(const pt_node& n) {
      ID_ASSERT_(n, ID_G_state_assignment);
      CHECK_SUCCESS(is_add_label(n));
      return n.children.at(0);
    }

    static const pt_node& getLHS(const pt_node& n) {
      ID_ASSERT_(n, ID_G_state_assignment);
      return n.children.at(0);
    }

    static const pt_node& getRHS(const pt_node& n) {
      ID_ASSERT_(n, ID_G_state_assignment);
      return n.children.at(1);
    }
  };


  struct t_add_label {
    static std::string getLabelName(const pt_node& n) {
      ID_ASSERT_(n, ID_G_add_label);
      std::string label_name=std::string(n.children.at(0).value.begin(), n.children.at(0).value.end());
      return label_name;
    }

    bool hasAssigment(const pt_node& n) {
      ID_ASSERT_(n, ID_G_add_label);
      return n.children.size()>1;
    }

    static const pt_node& getAssignment(const pt_node& n) {
      ID_ASSERT_(n, ID_G_add_label);
      return n.children.at(1);
    }
  };

  struct t_has_label_expression  {
    static const pt_node& getLabelID(const pt_node& n) {
      ID_ASSERT_(n, ID_G_has_label_expression);
      return n.children.at(0);
    }
  };

  struct t_label_id : public t_generic_subcomponent_wrapper {

    static std::string getLabelName(const pt_node& n) {
      ID_ASSERT_(n, ID_G_label_id);
      std::string label_name=
        string(getSimple(n).value.begin(),
               getSimple(n).value.end());

      return label_name;
    }
  };



  struct t_port_activity_expression {
    static const bool isActivePorts(const pt_node& n) {
      ID_ASSERT_(n, ID_G_port_activity_expression);
      return n.children.at(0).value.id()==parser_id(ID_G_active_ports);
    }    

    static const bool isSinglePort(const pt_node& n) {
      ID_ASSERT_(n, ID_G_port_activity_expression);
      return !isActivePorts(n);
    }    

    static const pt_node& getActivePorts(const pt_node& n) {
      ID_ASSERT_(n, ID_G_port_activity_expression);
      ERROR_CHECK_ASSERT(isActivePorts(n));
      return n.children.at(0);
    }    

    struct single_port_info {
      bool negated;
      port_name_t port_name;
    };

    static 
    single_port_info
    getSinglePort(const pt_node& n, const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_G_port_activity_expression);
      ERROR_CHECK_ASSERT(isSinglePort(n));
      single_port_info info;
      if (n.children.at(0).value.id()==parser_id(ID_G_port)) {
        info.negated=false;
        info.port_name=t_port::getPortName(n.children.at(0), symbols);
      } else {
        ERROR_CHECK_ASSERT(string(n.children.at(0).value.begin(), n.children.at(0).value.end())=="!");
        info.negated=true;
        info.port_name=info.port_name=t_port::getPortName(n.children.at(1), symbols);
      }
      return info;
    }    
  };
    
  struct t_io_guard : public t_generic_alternative {
    static const bool isEpsilon(const pt_node& n) {
      ID_ASSERT_(n, ID_G_io_guard);
      return string(n.value.begin(), n.value.end())=="epsilon";
    }    

    static const bool isDataflow(const pt_node& n) {
      ID_ASSERT_(n, ID_G_io_guard);
      return string(n.value.begin(), n.value.end())=="dataflow";
    }
    
    static const bool isStep(const pt_node& n) {
      ID_ASSERT_(n, ID_G_io_guard);
      return string(n.value.begin(), n.value.end())=="step";
    }
  };

};

}
}


#endif
