/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef INT_PARSER_HPP
#define INT_PARSER_HPP

#include "spirit_init.hpp"
using namespace vereofy::spirit;

#include <boost/bind.hpp>

#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>

#include <string>


#include "reo_parser_types.hpp"
#include "position_node.hpp"
#include "position_node_factory.hpp"

#include "parser_helper.hpp"
#include "module_parser_helper.hpp"

#include "identifier_parser.hpp"

namespace vereofy {
namespace vparser {

struct const_parser;

struct int_parser : public grammar<int_parser>
{
  enum {
    hex_string=0,
    integer=1,
    const_expression=2,
    const_ref_int=3
  };

  struct s_is_const_int_ref {
    const int_parser& self;
    const std::string& id_ref;

    s_is_const_int_ref(const int_parser& self_, const std::string& id_ref_) :
      self(self_), id_ref(id_ref_) {
    }

    bool operator()() const {
      SymbolData::ptr sd=self.symbol_table_stack.find(id_ref);

      if (sd && sd->type==SYMBOL_TYPE_CONSTANT) {
        return true;
      }
      return false;
    }
  };


  struct s_is_const_function {
    const int_parser& self;
    const std::string& id_ref;

    s_is_const_function(const int_parser& self_, const std::string& id_ref_) :
      self(self_), id_ref(id_ref_) {
    }

    bool operator()() const {
      SymbolData::ptr sd=self.symbol_table_stack.find(id_ref);

      if (sd && sd->type==SYMBOL_TYPE_FUNCTION) {
        FunctionData::ptr fd = boost::static_pointer_cast<FunctionData>(sd);
        
        if (fd->function->isConstEvaluatable()) {
          return true;
        }
      }

      return false;
    }
  };


  static 
  void init_rule_names(map<parser_id, std::string>& rule_names) {
    rule_names[ID_INT_hex_string] = "hex_string";
    rule_names[ID_INT_integer] = "integer";
    rule_names[ID_INT_const_ref_int] = "const_ref_int";
    rule_names[ID_INT_const_function_call] = "const_function_call";
    rule_names[ID_INT_const_factor] = "const_factor";
    rule_names[ID_INT_const_term] = "const_term";
    rule_names[ID_INT_const_expression] = "const_expression";
  }


  int_parser(symbol_table_stack_t& symbol_table_stack_) :
    symbol_table_stack(symbol_table_stack_) {}

  symbol_table_stack_t& symbol_table_stack;

  template <typename ScannerT>
  struct definition  : 
    public grammar_def < rule<ScannerT, parser_tag<ID_INT_hex_string> >, 
                         rule<ScannerT, parser_tag<ID_INT_integer> >, 
                         rule<ScannerT, parser_tag<ID_INT_const_expression> >,
                         rule<ScannerT, parser_tag<ID_INT_const_ref_int> > >
  {
    definition(int_parser const& self_)
      : identifier_grammar(last_identifier, reserved_keywords)
    {
      BOOST_SPIRIT_DEBUG_GRAMMAR(identifier_grammar);

      ENTRY_GRAMMAR(identifier_p, identifier_parser, identifier_grammar, identifier_parser::identifier);

      //  Start grammar definition
      hex_string = lexeme_d[leaf_node_d[
                            ch_p('0') 
                            >> ch_p('x')
                            >> +(range_p('0','9') | range_p('a','f') | range_p('A','F'))]];


      integer     =   access_node_d[int_p][s_int_value_assign()];

      const_ref_int = identifier_p >> eps_p(s_is_const_int_ref(self_, last_identifier));

      const_function_call = 
        identifier_p 
        >> eps_p(s_is_const_function(self_, last_identifier))
        >> no_node_d[ch_p('(')]
        >> !(const_expression >> *(no_node_d[ch_p(',')] >> const_expression))
        >> no_node_d[ch_p(')')];


      const_factor =  integer
                      | const_function_call
                      | const_ref_int
                      | (no_node_d[ch_p('(')] >> const_expression >> no_node_d[ch_p(')')])
                      | access_node_d[(root_node_d[ch_p('-')] >> const_factor)];
        
        
      const_term   =  
        const_factor
        >> *(    (root_node_d[ch_p('*')] >> const_factor)
               | (root_node_d[ch_p('/')] >> const_factor)
            );
      
      const_expression  = 
        const_term
        >> *(  (root_node_d[ch_p('+')] >> const_term)
             | (root_node_d[ch_p('-')] >> const_term)
            );

      //  End grammar definition
      this->start_parsers(hex_string, integer, const_expression, const_ref_int);

      BOOST_SPIRIT_DEBUG_RULE(hex_string);
      BOOST_SPIRIT_DEBUG_RULE(integer);
      BOOST_SPIRIT_DEBUG_RULE(const_ref_int);
      BOOST_SPIRIT_DEBUG_RULE(const_function_call);
      BOOST_SPIRIT_DEBUG_RULE(const_factor);
      BOOST_SPIRIT_DEBUG_RULE(const_term);
      BOOST_SPIRIT_DEBUG_RULE(const_expression);
    }

    rule<ScannerT, parser_tag<ID_INT_hex_string> >         hex_string;
    rule<ScannerT, parser_tag<ID_INT_integer> >            integer;
    rule<ScannerT, parser_tag<ID_INT_const_ref_int> >      const_ref_int;
    rule<ScannerT, parser_tag<ID_INT_const_function_call> > const_function_call;
    rule<ScannerT, parser_tag<ID_INT_const_term> >         const_term;
    rule<ScannerT, parser_tag<ID_INT_const_factor> >       const_factor;
    rule<ScannerT, parser_tag<ID_INT_const_expression> >   const_expression;

    identifier_parser identifier_grammar;
    std::string last_identifier;
    symbols<> reserved_keywords;
  };
};

}
}
#endif
