/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef INT_PARSER_HELPER_HPP
#define INT_PARSER_HELPER_HPP

#include "parser_helper.hpp"
#include "identifier_parser_helper.hpp"
#include "vereofy/common/errors.hpp"
#include "vereofy/common/value_or_undefined.hpp"
#include "vereofy/system/symb_automaton/symb_function.hpp"

namespace vereofy {
namespace vparser {

struct int_helper {

  struct t_integer {
    static int getInt(const pt_node& n) {
      return boost::lexical_cast<int>(string(n.value.begin(),
                                             n.value.end()));
    }
  };


  struct t_const_ref_int {
    static string getName(const pt_node& n) {
      return identifier_helper::t_identifier::getIdentifier(n.children.at(0));
    }

    static
    int_value_t evaluate(const pt_node& n, const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_INT_const_ref_int);

      string name=getName(n);
      
      SymbolData::ptr sd=symbols.find(name);
      
      if (!sd || sd->type!=SYMBOL_TYPE_CONSTANT) {
        throw string("No constant named ")+name;
      }
      
      ConstantData::ptr const_data=boost::static_pointer_cast<ConstantData>(sd);         return const_data->int_value;
    }
  };

  struct t_const_function_call {
    static std::string getFunctionName(const pt_node& n) {
      return identifier_helper::t_identifier::getIdentifier(n.children.at(0));
    }

    static unsigned int getArity(const pt_node& n) {
      return n.children.size()-1;
    }


    static const pt_node& getArgumentExpression(const pt_node& n, unsigned int index) {
      return n.children.at(index+1);
    }

    static int_value_t evaluate(const pt_node& n, const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_INT_const_function_call);

      std::string function_name=getFunctionName(n);
      SymbolData::ptr sd=symbols.find(function_name);
      if (!sd) {
        throw "Function call for function "+function_name+", does not exist";
      }
      if (sd->type!=SYMBOL_TYPE_FUNCTION) {
        throw "Function call for function "+function_name+", is not a function!";
      }
      FunctionData::ptr fd=boost::static_pointer_cast<FunctionData>(sd);
      SymbFunction::ptr symb_function=fd->function;
      
      if (getArity(n)!=symb_function->getArity()) {
        throw "Function call for function "+function_name+", mismatch in number of parameters (Expected "+
          boost::lexical_cast<std::string>(symb_function->getArity())+", got "+
          boost::lexical_cast<std::string>(getArity(n))+")";
      }

      SymbFunctionArgumentVector arguments;
      for (unsigned int i=0;i<getArity(n);i++) {
        const pt_node& argument_node=getArgumentExpression(n, i);
        int_value_t arg_value=t_const_expression_dispatch::evaluate(argument_node, symbols);

        if (!arg_value.isDefined()) {
          // one of the function argument values is undefined 
          // -> the return value is undefined
          return int_value_t::UNDEFINED();
        } else {
          int value = arg_value.getValue();
          arguments.push_back(SymbFunctionArgument(value));
        }
      }

      return symb_function->const_evaluate(arguments);
    }
  };

  struct t_const_factor {
    static
    int_value_t evaluate(const pt_node& n, const symbol_table_stack_t& symbols) {
      CHECK_SUCCESS(n.value.id() == parser_id(ID_INT_const_factor));

      string op(n.value.begin(), n.value.end());
      if (op=="-") {
        return 
          - t_const_expression_dispatch::evaluate(n.children.at(0), symbols);

      } else if (n.children.at(0).value.id()==parser_id(ID_INT_integer)) {
        return IntValue::getIntValue(n.children.at(0).value.value());

      } else if (n.children.at(0).value.id()==parser_id(ID_INT_const_ref_int)) {
        return t_const_ref_int::evaluate(n.children.at(0), symbols);
      } else if (n.children.at(0).value.id()==parser_id(ID_INT_const_function_call)) {
        return t_const_function_call::evaluate(n.children.at(0), symbols);
      } else {
        return t_const_expression_dispatch::evaluate(n.children.at(0), symbols);
      }
    }
  };


  struct t_const_term {
    static
    int_value_t evaluate(const pt_node& n, const symbol_table_stack_t& symbols) {
      CHECK_SUCCESS(n.value.id()==parser_id(ID_INT_const_term));

      string op(n.value.begin(), n.value.end());
      if (op=="*") {
        return 
          t_const_expression_dispatch::evaluate(n.children.at(0), symbols)
          *
          t_const_expression_dispatch::evaluate(n.children.at(1), symbols);

      } else if (op=="/") {
        int_value_t arg2=t_const_expression_dispatch::evaluate(n.children.at(1), symbols);
        if (!arg2.isDefined()) {return int_value_t::UNDEFINED();}
        if (arg2==0) {throw "div/0";}
        return 
          t_const_expression_dispatch::evaluate(n.children.at(0),symbols)
          /
          arg2;

      } else {
        return t_const_expression_dispatch::evaluate(n.children.at(0), symbols);
      }
    }
  };

  struct s_const_expression_evaluate {
    const symbol_table_stack_t& symbols;
    int& int_ref;

    s_const_expression_evaluate(const symbol_table_stack_t& symbols_, int& int_ref_) :
      symbols(symbols_),
      int_ref(int_ref_) {}

    template <class Iter>
    void operator()(const pt_node& n, Iter i1, Iter i2) const {
      int_ref=t_const_expression_dispatch::evaluate(n, symbols).getValue();
    }
  };


  struct t_const_expression {
    static
    int_value_t evaluate(const pt_node& n, const symbol_table_stack_t& symbols) {
      CHECK_SUCCESS(n.value.id() == parser_id(ID_INT_const_expression));

      string op(n.value.begin(), n.value.end());

      if (op=="+") {
        return 
          t_const_expression_dispatch::evaluate(n.children.at(0), symbols)
          +
          t_const_expression_dispatch::evaluate(n.children.at(1), symbols);
      } else if (op=="-") {
        return 
          t_const_expression_dispatch::evaluate(n.children.at(0), symbols)
          -
          t_const_expression_dispatch::evaluate(n.children.at(1), symbols);

      } else {
        return t_const_expression_dispatch::evaluate(n.children.at(0), symbols);
      }
    }
  };

  struct t_const_expression_dispatch {
    static
    int_value_t evaluate(const pt_node& n, const symbol_table_stack_t& symbols) {
      switch (n.value.id().to_long()) {
      case ID_INT_const_expression:
        return t_const_expression::evaluate(n,symbols);
      case ID_INT_const_term: 
        return t_const_term::evaluate(n,symbols);
      case ID_INT_const_factor:
        return t_const_factor::evaluate(n,symbols);
      default:
        THROW_UNIMPLEMENTED();
      }

    }
    

  };


};


}
}

#endif
