/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef MODULE_PARSER_HPP
#define MODULE_PARSER_HPP

#include "spirit_init.hpp"
using namespace vereofy::spirit;

#include <boost/bind.hpp>

#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>

#include <boost/typeof/typeof.hpp>

#include <string>


#include "reo_parser_types.hpp"
#include "position_node.hpp"
#include "position_node_factory.hpp"

#include "vereofy/common/singleton_set.hpp"

#include "vereofy/system/reactive_module/reactive_module.hpp"
#include "vereofy/system/datadomain.hpp"

#include "int_parser.hpp"
#include "int_parser_helper.hpp"

#include "type_parser.hpp"
//#include "type_parser_helper.hpp"

#include "guard_parser.hpp"

#include "identifier_parser.hpp"

#include "const_def_parser.hpp"

#include "parser_helper.hpp"
#include "module_parser_helper.hpp"

#include "common_parser.hpp"
#include "common_parser_helper.hpp"

namespace vereofy {
namespace vparser {

struct module_parser : public grammar<module_parser>
{
  typedef module_helper helper;


  enum Errors {
    no_module_name,
    no_interface_identifier,
    no_type,
    no_var_identifier,
    no_label_identifier,
    no_ap_identifier,
    no_init_expression,
    no_dc_open,
    no_dc_close,
    no_data_constraint,
    no_transition,
    no_left_curly,
    no_right_curly,
    no_left_paren,
    no_right_paren,
    no_left_square,
    no_right_square,
    no_less,
    no_greater,
    no_semicolon,
    no_colon,
    no_comma
  };

  struct expect_parsers {
    expect_parsers() :
      module_name(no_module_name),
      interface_identifier(no_interface_identifier),
      type(no_type),
      var_identifier(no_var_identifier),
      label_identifier(no_label_identifier),
      ap_identifier(no_ap_identifier),
      init_expression(no_init_expression),
      dc_open(no_dc_open),
      dc_close(no_dc_close),
      data_constraint(no_data_constraint),
      transition(no_transition),
      left_curly(no_left_curly),
      right_curly(no_right_curly),
      left_paren(no_left_paren),
      right_paren(no_right_paren),
      left_square(no_left_square),
      right_square(no_right_square),
      less(no_less),
      greater(no_greater),
      semicolon(no_semicolon),
      colon(no_colon),
      comma(no_comma)
    {}

    vereofy::spirit::assertion<Errors> module_name;
    vereofy::spirit::assertion<Errors> interface_identifier;
    vereofy::spirit::assertion<Errors> type;
    vereofy::spirit::assertion<Errors> var_identifier;
    vereofy::spirit::assertion<Errors> label_identifier;
    vereofy::spirit::assertion<Errors> ap_identifier;
    vereofy::spirit::assertion<Errors> init_expression;
    vereofy::spirit::assertion<Errors> dc_open;
    vereofy::spirit::assertion<Errors> dc_close;
    vereofy::spirit::assertion<Errors> data_constraint;
    vereofy::spirit::assertion<Errors> transition;
    vereofy::spirit::assertion<Errors> left_curly;
    vereofy::spirit::assertion<Errors> right_curly;
    vereofy::spirit::assertion<Errors> left_paren;
    vereofy::spirit::assertion<Errors> right_paren;
    vereofy::spirit::assertion<Errors> left_square;
    vereofy::spirit::assertion<Errors> right_square;
    vereofy::spirit::assertion<Errors> less;
    vereofy::spirit::assertion<Errors> greater;
    vereofy::spirit::assertion<Errors> semicolon;
    vereofy::spirit::assertion<Errors> colon;
    vereofy::spirit::assertion<Errors> comma;
  };


  static 
  void init_rule_names(std::map<parser_id, std::string>& rule_names) {
    rule_names[ID_M_l_transition] = "l_transition";
    rule_names[ID_M_m_transition] = "m_transition";
    rule_names[ID_M_r_transition] = "r_transition";

    rule_names[ID_M_transition] = "transition";
    rule_names[ID_M_compact_transition] = "compact_transition";
    rule_names[ID_M_transitions] = "transitions";

    rule_names[ID_M_var_init] = "var_init";
    rule_names[ID_M_var_def] = "var_def";
    rule_names[ID_M_label_def] = "label_def";
    rule_names[ID_M_defs] = "defs";
    rule_names[ID_M_ap_defs] = "ap_defs";
    rule_names[ID_M_ap_def] = "ap_def";

    rule_names[ID_M_interface_def] = "interface_def";
    rule_names[ID_M_interface_defs] = "interface_defs";

    rule_names[ID_M_body] = "body";

    rule_names[ID_M_module_def] = "module_def";
  }

  static
  void init_reserved_keywords(symbols<>& keywords) {
    keywords="MODULE", "label", "has_label";
  }


    // --- semantic actions ------

  struct s_var_def {
    const module_parser& self;

    s_var_def(const module_parser& self_) : self(self_) {}

    template <class Node, class It>
    void operator()(Node&n, It& i1, It& i2) const {
      std::string var_name=helper::t_var_def::getName(n);
      
      DataDomain::ptr dd=
        helper::t_var_def::getDataDomain(n, 
                                         self.symbol_table_stack);

      if (debugLevel>=3) {
        cerr << "Var Type for " << var_name << ": ";
        cerr << dd->toString();
      }

      try {
        dd->ensureFiniteness();
      } catch (std::string& e) {
        throw_(i1, std::string("In definition of state variable ")+var_name+": "+e);
      }
      
      n.value.value(module_var_type::ptr(new module_var_type(dd)));

      vereofy::StateDescription::ptr sd(new vereofy::StateDescription(var_name,
                                                                      dd));

      if (debugLevel>=4) {
        cerr << "Adding state description for " <<var_name<<endl;
        cerr << "  to interface " << self.cur_module->c_interface << endl;
      }

      self.cur_module->c_interface->addStateDescription(sd);
    }    
  };


  struct s_label_def {
    const module_parser& self;

    s_label_def(const module_parser& self_) : self(self_) {}

    template <class Node, class It>
    void operator()(Node&n, It& i1, It& i2) const {
      std::string label_name=helper::t_label_def::getName(n);
      
      vereofy::TransitionLabelDescription::ptr 
        ld(new vereofy::TransitionLabelDescription(TransitionLabelDescription::GENERIC,
                                                   label_name));

      if (debugLevel>=4) {
        cerr << "Adding label description for " <<label_name<<endl;
        cerr << "  to interface " << self.cur_module->c_interface << endl;
      }

      self.cur_module->c_interface->addTransitionLabelDescription(ld);
    }    
  };


  struct s_ap_def {
    const module_parser& self;

    s_ap_def(const module_parser& self_) : self(self_) {}

    template <class Node, class It>
    void operator()(Node&n, It& i1, It& i2) const {
      std::string ap_name=helper::t_ap_def::getName(n);
      
      if (debugLevel>=4) {
        cerr << "Adding placeholder for internal atomic proposition " <<ap_name<<endl;
        cerr << "  to interface " << self.cur_module->c_interface << endl;
      }

      if (self.cur_module->c_interface->hasAtomicProposition(ap_name)) {
        throw_(i1, std::string("Atomic proposition ")+ap_name+" is defined multiple times");
      }

      self.cur_module->c_interface->addAtomicProposition(ap_name, "%%internal_spec");  
      // TODO: would be nicer to actually store the text
    }    
  };



  struct s_interface_def {
    const vereofy::ReactiveModule::ptr& module;

    s_interface_def(const vereofy::ReactiveModule::ptr& module_) : module(module_) {}

    template <class Node, class It>
    void operator()(Node&n, It& i1, It& i2) const {
      std::string name=helper::t_interface_def::getName(n);

      if (debugLevel>=4) {
        cerr << "Adding port " <<name << 
          (helper::t_interface_def::isIn(n) ? " : in " : " : out ") << endl;
        cerr << "  to interface " << module->c_interface << endl;
      }
      if (helper::t_interface_def::isIn(n)) {
        if (!module->c_interface->addInPort(name,
                                            access_weak_ptr(module->context)->getDataDomain())) {
          throw_(i1, std::string("Couldn't add in port ")+name);
        }
      } else if (helper::t_interface_def::isOut(n)) {
        if (!module->c_interface->addOutPort(name,
                                             access_weak_ptr(module->context)->getDataDomain())) {
          throw_(i1, std::string("Couldn't add out port ")+name);
        }
      } else {throw "implementation error";}
    }    
  };


  struct s_module_def_begin {
    const module_parser& self;

    s_module_def_begin(const module_parser& self_) : self(self_) {}

    template<class Iter>
    const void
    operator()(Iter& i1, Iter& i2) const {
      std::string name(i1, i2);
 
      vereofy::ReactiveModule::ptr mp(new vereofy::ReactiveModule(self.system_context, name));

      access_weak_ptr(self.system_context)->requireDataDomain();

      module_parser& self_=const_cast<module_parser&>(self);
      self_.cur_module=mp;
      self_.cur_module_interface=mp->c_interface;
      self_.symbol_table_stack.push(mp->c_interface->getLocalSymbols());
    }
  };
  
  struct s_module_def_end {
    const module_parser& self;
    
    s_module_def_end(const module_parser& self_) : self(self_) {}
    
    template<class Node, class Iter>
    const void
    operator()(const Node&n, Iter& i1, Iter& i2) const {
      module_parser& self_=const_cast<module_parser&>(self);
      self_.symbol_table_stack.pop();
    }
  };

  struct s_parameter_def {
    const module_parser& self;

    s_parameter_def(const module_parser& self_) : self(self_) {}

    template<class Node, class Iter>
    const void
    operator()(const Node&n, Iter& i1, Iter& i2) const {
      for (unsigned int i=0;
           i<common_helper::t_parameter_def::countChildren(n);
           i++) {
        std::string name=
          identifier_helper::t_identifier::getIdentifier(common_helper::t_parameter_def::get(n, i));

        if (debugLevel>=5) {
          cerr << "Parameter: " << name << "\n";
        }

        // add place-holder constant into local symboltable
        ConstantData::ptr cd(new ConstantData(name, int_value_t::UNDEFINED()));

        self.symbol_table_stack.getTopTable()->add(name, cd);
      }
    }
  };

 
  template <class DefT>
  struct error_handler {
    DefT& def;

    error_handler(DefT& def_) : def(def_) {}

    template <typename ScannerT, typename ErrorT>
    error_status<>
    operator()(ScannerT const& scan, ErrorT const& error) const {
      switch (error.descriptor) {
      case no_module_name:
        throw_(error.where, std::string("Expected a module name"));
      case no_interface_identifier:
        throw_(error.where, std::string("Expected a name for this interface port"));
      case no_type:
        throw_(error.where, std::string("Expected a type for this variable"));
      case no_var_identifier:
        throw_(error.where, std::string("Expected a name for this variable"));
      case no_label_identifier:
        throw_(error.where, std::string("Expected a label name"));
      case no_ap_identifier:
        throw_(error.where, std::string("Expected a name for this atomic proposition"));
      case no_init_expression:
        throw_(error.where, std::string("Expected an initializing expression after :="));
      case no_dc_open:
        throw_(error.where, std::string("Expected '-[' or a state guard"));
      case no_dc_close:
        throw_(error.where, std::string("Expected ']->' or an I/O guard"));
      case no_data_constraint:
        throw_(error.where, std::string("Expect at least one I/O guard or 'epsilon'/'dataflow'/'step'"));
      case no_transition:
        throw_(error.where, std::string("Expected at least one transition"));

        
      case no_left_curly:
        throw_(error.where, std::string("Expected opening '{'"));
      case no_left_paren:
        throw_(error.where, std::string("Expected opening '('"));
      case no_left_square:
        throw_(error.where, std::string("Expected opening '['"));

      case no_right_curly:
        throw_(error.where, std::string("Expected closing '}'"));
      case no_right_paren:
        throw_(error.where, std::string("Expected closing ')'"));
      case no_right_square:
        throw_(error.where, std::string("Expected closing ']'"));

      case no_less:
        throw_(error.where, std::string("Expected '<'"));
      case no_greater:
        throw_(error.where, std::string("Expected '>'"));

      case no_semicolon:
        throw_(error.where, std::string("Expected ';'"));
      case no_colon:
        throw_(error.where, std::string("Expected ':'"));
      case no_comma:
        throw_(error.where, std::string("Expected ','"));
      }
      
      // default action
      return error_status<>(error_status<>::rethrow);
    }
  };


  struct store_pos {
    template<class Node, class Iter>
    void operator()(Node& n, Iter& i1, Iter& i2) const {
#ifdef REO_PARSER_POSITION_NODES
      n.value.setPosition(i1.get_position());
      //      cerr << "store_pos" << n.value.pos->line << ":" << n.value.pos->column << " [" << string(i1, i2) << "]\n";
#endif
    }    
  };



  module_parser(symbol_table_stack_t& symbol_table_stack_,
                const symbols<>& reserved_keywords_,
                vereofy::SystemContext::wptr system_context_) :
    symbol_table_stack(symbol_table_stack_),
    reserved_keywords(reserved_keywords_),
    system_context(system_context_) {}

  symbol_table_stack_t& symbol_table_stack;
  const symbols<>& reserved_keywords;
  vereofy::ReactiveModule::ptr cur_module;
  vereofy::ComponentInterface::ptr cur_module_interface;
  vereofy::SystemContext::wptr system_context;

  template <typename ScannerT>
  struct definition
  {
    definition(module_parser const& self)
      : identifier_grammar(last_identifier, self.reserved_keywords),
        integer_grammar(self.symbol_table_stack),
        type_grammar(self.symbol_table_stack, self.reserved_keywords),
        guard_grammar(self.symbol_table_stack, self.reserved_keywords, self.cur_module_interface),
        common_grammar(self.reserved_keywords)
    {
      
      BOOST_SPIRIT_DEBUG_GRAMMAR(identifier_grammar);
      BOOST_SPIRIT_DEBUG_GRAMMAR(integer_grammar);
      BOOST_SPIRIT_DEBUG_GRAMMAR(type_grammar);
      BOOST_SPIRIT_DEBUG_GRAMMAR(guard_grammar);
      BOOST_SPIRIT_DEBUG_GRAMMAR(common_grammar);

      ENTRY_GRAMMAR(identifier_p, identifier_parser, identifier_grammar, identifier_parser::identifier);
      ENTRY_GRAMMAR(enum_value_p, identifier_parser, identifier_grammar, identifier_parser::enum_value);

      ENTRY_GRAMMAR(integer_p, int_parser, integer_grammar, int_parser::integer);
      ENTRY_GRAMMAR(hex_string_p, int_parser, integer_grammar, int_parser::hex_string);
      ENTRY_GRAMMAR(const_expression_p, int_parser, integer_grammar, int_parser::const_expression);

      ENTRY_GRAMMAR(type_p, type_parser, type_grammar, type_parser::type);
      ENTRY_GRAMMAR(type_def_p, type_parser, type_grammar,type_parser::type_def);

      ENTRY_GRAMMAR(any_value_p, guard_parser, guard_grammar, guard_parser::any_value);
      ENTRY_GRAMMAR(guard_expression_p, guard_parser, guard_grammar, guard_parser::guard_expression);
      ENTRY_GRAMMAR(state_guard_p, guard_parser, guard_grammar, guard_parser::state_guard);
      ENTRY_GRAMMAR(bool_expression_p, guard_parser, guard_grammar, guard_parser::bool_expression);
      ENTRY_GRAMMAR(io_guard_p, guard_parser, guard_grammar, guard_parser::io_guard);
      ENTRY_GRAMMAR(state_assignment_p, guard_parser, guard_grammar, guard_parser::state_assignment);

      ENTRY_GRAMMAR(parameter_def_p, common_parser, common_grammar, common_parser::parameter_def);

      //  Start grammar definition
      
      l_transition = 
        ~eps_p(lexeme_d[str_p("-[")])  // skip if we have -[
        >> eps_p[guard_parser::init(guard_grammar, true, false, false, true)] >> state_guard_p
        >> *(no_node_d[ch_p('&')] 
             >> eps_p[guard_parser::init(guard_grammar, true, false, false, true)] 
             >> state_guard_p);

      m_transition =
        ~eps_p(lexeme_d[str_p("]->")])  // skip if we have ]->
        >> eps_p[guard_parser::init(guard_grammar, true, true, false, true)] 
        >> io_guard_p
        >> *(no_node_d[ch_p('&')] 
             >> eps_p[guard_parser::init(guard_grammar, true, true, false, true)] 
             >> io_guard_p);

      r_transition = 
        ~eps_p(ch_p(';'))   // skip if we have ;
        >> eps_p[guard_parser::init(guard_grammar, true, true, false)] 
        >> state_assignment_p 
        >> *(no_node_d[ch_p('&')] 
             >> eps_p[guard_parser::init(guard_grammar, true, true, false)] 
             >> state_assignment_p);
      
      transition = 
        access_node_d[
                      ~eps_p(ch_p('}')) // early abort if there is a }
                      >> !l_transition 
                      >> expect.dc_open(lexeme_d[str_p("-[")]) 
                      >> expect.data_constraint(m_transition)
                      >> expect.dc_close(lexeme_d[str_p("]->")])
                      >> !r_transition
                      >> no_node_d[expect.semicolon(ch_p(';'))] 
                      ][store_pos()];

      compact_transition =
        no_node_d[keyword_d["compact_v1"]]
        >> no_node_d[expect.left_curly(ch_p('{'))]
        >> lexeme_d[ leaf_node_d[str_p("->") >> *(anychar_p - ch_p('}'))] ]
        >> no_node_d[expect.right_curly(ch_p('}'))];

      transitions =  expect.transition(+(compact_transition | transition));
      
      var_init = no_node_d[str_p(":=")]  
        >> eps_p[guard_parser::init(guard_grammar, false, false, false)]
        >> expect.init_expression(any_value_p | guard_expression_p);
      
      var_def = access_node_d[
                                (root_node_d[str_p("var")]
                                 >> no_node_d[expect.colon(ch_p(':'))]
                                 >> expect.type(type_p)
                                 >> expect.var_identifier(identifier_p)
                                 >> !(no_node_d[ch_p('[')] 
                                      >> const_expression_p
                                      >> no_node_d[expect.right_square(ch_p(']'))]
                                      ) // array
                                 )
                              ][s_var_def(self)]
          >> !var_init
          >> no_node_d[expect.semicolon(ch_p(';'))]
          ;

      label_def = access_node_d[
                                (root_node_d[str_p("label")]
                                 >> no_node_d[expect.colon(ch_p(':'))]
                                 >> expect.label_identifier(identifier_p)
                                 //                                 >> expect.type(type_p)
                                 )
                              ][s_label_def(self)]
          >> no_node_d[expect.semicolon(ch_p(';'))]
          ;
        
        defs = *(var_def | type_def_p | label_def) ;
        
        ap_def = 
          access_node_d[
                        root_node_d[str_p("ap")]
                        >> no_node_d[expect.colon(ch_p(':'))]
                        >> expect.ap_identifier(identifier_p)
                        >> no_node_d[str_p("<=>")]
                        >> eps_p[guard_parser::init(guard_grammar, true, false, false)]
                        >> eps_p[phoenix::var(guard_grammar.allow_bool_expression_state_guard)=true]
                        >> eps_p[phoenix::var(guard_grammar.allow_bool_expression_io_guard)=false]
                        >> bool_expression_p
                        >> no_node_d[expect.semicolon(ch_p(';'))]
                        ][s_ap_def(self)];


        ap_defs = *(ap_def);

        interface_def = 
          access_node_d[ 
                        root_node_d[(str_p("in") | str_p("out"))]
                        >> no_node_d[expect.colon(ch_p(':'))] 
                        >> expect.interface_identifier(identifier_p)
                        >> no_node_d[expect.semicolon(ch_p(';'))]
                         ][s_interface_def(self.cur_module)];
        
        interface_defs = *interface_def;
          
        body = 
          interface_defs
          >> defs
          >> ap_defs
          >> eps_p[phoenix::var(guard_grammar.allow_ports)=true]
          >> eps_p[phoenix::var(guard_grammar.allow_states)=true]
          >> transitions;
        
        module_def = error_guard(
             access_node_d[
             root_node_d[keyword_d["MODULE"]]
             >> expect.module_name(identifier_p) [s_module_def_begin(self)]
             >> !access_node_d[parameter_def_p][s_parameter_def(self)]
             >> no_node_d[expect.left_curly(ch_p('{'))] 
             >> body 
             >> no_node_d[expect.right_curly(ch_p('}'))]
          ][s_module_def_end(self)])
          [error_handler<definition>(*this)];
        
        //  End grammar definition

#undef identifier_p     

      // turn on the debugging info.
        BOOST_SPIRIT_DEBUG_RULE(l_transition);
        BOOST_SPIRIT_DEBUG_RULE(m_transition);
        BOOST_SPIRIT_DEBUG_RULE(r_transition);
        
        BOOST_SPIRIT_DEBUG_RULE(transitions);
        BOOST_SPIRIT_DEBUG_RULE(transition);
        BOOST_SPIRIT_DEBUG_RULE(compact_transition);
        BOOST_SPIRIT_DEBUG_RULE(var_init);

        BOOST_SPIRIT_DEBUG_RULE(var_def);
        BOOST_SPIRIT_DEBUG_RULE(label_def);
        BOOST_SPIRIT_DEBUG_RULE(defs);
        BOOST_SPIRIT_DEBUG_RULE(ap_defs);
        BOOST_SPIRIT_DEBUG_RULE(ap_def);
        BOOST_SPIRIT_DEBUG_RULE(interface_defs);
        BOOST_SPIRIT_DEBUG_RULE(interface_def);
        BOOST_SPIRIT_DEBUG_RULE(body);

        BOOST_SPIRIT_DEBUG_RULE(module_def);
    }
    
    rule<ScannerT, parser_tag<ID_M_l_transition> >  l_transition;
    rule<ScannerT, parser_tag<ID_M_m_transition> >  m_transition;
    rule<ScannerT, parser_tag<ID_M_r_transition> >  r_transition;

    rule<ScannerT, parser_tag<ID_M_transitions> >     transitions;
    rule<ScannerT, parser_tag<ID_M_transition> >     transition;
    rule<ScannerT, parser_tag<ID_M_compact_transition> >  compact_transition;
    rule<ScannerT, parser_tag<ID_M_var_init> >     var_init;
    rule<ScannerT, parser_tag<ID_M_var_def> >     var_def;
    rule<ScannerT, parser_tag<ID_M_label_def> >     label_def;
    rule<ScannerT, parser_tag<ID_M_defs> >     defs;
    rule<ScannerT, parser_tag<ID_M_ap_def> >     ap_def;
    rule<ScannerT, parser_tag<ID_M_ap_defs> >     ap_defs;
    rule<ScannerT, parser_tag<ID_M_interface_def> >     interface_def;
    rule<ScannerT, parser_tag<ID_M_interface_defs> >     interface_defs;
    rule<ScannerT, parser_tag<ID_M_body> >     body;
    
    rule<ScannerT, parser_tag<ID_M_module_def> >     module_def;


    rule<ScannerT, parser_tag<ID_M_module_def> > const&
    start() const { return module_def; }


    guard<Errors> error_guard;

    identifier_parser identifier_grammar;
    int_parser integer_grammar;
    type_parser type_grammar;
    guard_parser guard_grammar;
    common_parser common_grammar;

    expect_parsers expect;
    
    std::string last_identifier;
  };
};

}
}
#endif
