/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef MODULE_PARSER_HELPER_HPP
#define MODULE_PARSER_HELPER_HPP

#include "vereofy/common/checks.hpp"
#include "identifier_parser_helper.hpp"
#include "type_parser_helper.hpp"

#include <boost/algorithm/string.hpp>

namespace vereofy {
namespace vparser {

struct module_helper {

  struct t_generic {
      static rule_id getType(const pt_node& n) {
      return (rule_id)n.value.id().to_long();
    }
  };

  struct t_generic_alternative {
    static const pt_node& get(const pt_node& n) {
      return n.children.at(0);
    }

    template <class IDType>
    static const bool is(const pt_node& n, IDType id) {
      return n.children.at(0).value.id()==parser_id(id);
    }    
  };

  struct t_generic_identifier_wrapper {
    static string getIdentifier(const pt_node& n) {
      return string(n.children.at(0).value.begin(),
                    n.children.at(0).value.end());
    }
  };


  struct t_var_init {
    static const pt_node& get(const pt_node& n) {
      ID_ASSERT_(n, ID_M_var_init);
      return n.children.at(0);
    }
  };

  struct t_var_def {
    static
    string getName(const pt_node& n) {
      ID_ASSERT_(n, ID_M_var_def);

      return string(n.children.at(1).value.begin(),
                    n.children.at(1).value.end());
    }

    static bool hasArray(const pt_node& n) {
      ID_ASSERT_(n, ID_M_var_def);

      return n.children.size()==4 ||
        (n.children.size()==3 && n.children.at(2).value.id()==parser_id(ID_INT_const_expression));
    }

    static bool hasInit(const pt_node& n) {
      ID_ASSERT_(n, ID_M_var_def);
      return hasArray(n) ? n.children.size()==4 : n.children.size()==3;
    }

    static const pt_node& getInit(const pt_node& n) {
      ID_ASSERT_(n, ID_M_var_def);

      ERROR_CHECK_ASSERT(hasInit(n));
      return hasArray(n) ? n.children.at(3) : n.children.at(2);
    }

    static DataDomain::ptr getDataDomain(const pt_node& n,
                                         const symbol_table_stack_t& symbols) {
      ID_ASSERT_(n, ID_M_var_def);

      DataDomain::ptr dd=
        type_helper::t_type::getDataDomain(n.children.at(0),
                                           symbols);
      
      if (dd->getType()==DataDomain::DD_FLAT &&
          static_cast<FlatDataDomain&>(*dd).getFlatType()==FlatDataDomain::DD_ENUM) {
        // name this enum type $var_name
        EnumDataDomain& enum_dd=static_cast<EnumDataDomain&>(*dd);
        enum_dd.setTypeName("$"+getName(n));
      }

      return dd;
    }

  };


  struct t_label_def {
    static
    string getName(const pt_node& n) {
      ID_ASSERT_(n, ID_M_label_def);

      return string(n.children.at(0).value.begin(),
                    n.children.at(0).value.end());
    }

  };

  struct t_defs : public t_generic_list {};

  struct t_ap_def {
    static
    string getName(const pt_node& n) {
      ID_ASSERT_(n, ID_M_ap_def);

      return string(n.children.at(0).value.begin(),
                    n.children.at(0).value.end());
    }

    static
    const pt_node& getBooleanStateGuard(const pt_node& n) {
      ID_ASSERT_(n, ID_M_ap_def);

      return n.children.at(1);
    }
  };

  struct t_ap_defs : public t_generic_list {};

  struct const_evaluate {
    rule_id rule;

    const_evaluate(rule_id rule_) : rule(rule_) {}
    
    template<class Iter>
    const const_evaluate&
    operator()(pt_node& n, Iter& i1, Iter& i2) const {
      //      return *this;
      
      switch (rule) {
      case ID_INT_integer:
        n.value.value(IntValue::ptr(new IntValue(boost::lexical_cast<int>(string(i1,i2)))));
        break;
      case ID_INT_const_factor: {
        if (n.value.id()==ID_INT_integer) {
          break;
        } else if (n.children.at(0).value.id()==parser_id(ID_INT_const_expression)) {
          ;
        } else if (string(n.value.begin(), 
                          n.value.end())=="-") {
          Value::ptr v=boost::static_pointer_cast<Value>(n.children.at(0).value.value());
          n.value.value(v);
        }
        break;
      } 
      default:
        throw "";
      }
      
      if (debugLevel>=5) {
        cerr << "Evaluated value: ";
        n.value.value()->print(cerr);
        cerr << "\n";
      }
      return *this;
    }
  };

  template <class V>
  struct print {
    print(V v, const string& name="") {
      cerr << "Printing closure value " << name << " : "<< v << "\n";
    }
  };

  struct t_interface_defs : t_generic_list {};

  struct t_interface_def {
    static
    string getName(const pt_node& n) {
      ID_ASSERT_(n, ID_M_interface_def);

      return string(n.children.at(0).value.begin(),
                    n.children.at(0).value.end());
    }

    static
    bool isOut(const pt_node& n) {
      ID_ASSERT_(n, ID_M_interface_def);

      return string(n.value.begin(),
                    n.value.end())=="out";
    }

    static
    bool isIn(const pt_node& n) {
      ID_ASSERT_(n, ID_M_interface_def);

      return string(n.value.begin(),
                    n.value.end())=="in";
    }
  };




  struct t_l_transition {
    static const pt_node& getStateGuard(const pt_node& n, 
                                        unsigned int index) {
      ID_ASSERT_(n, ID_M_l_transition);

      return n.children.at(index);
    }

    static unsigned int countStateGuards(const pt_node& n) {
      ID_ASSERT_(n, ID_M_l_transition);

      return n.children.size();
    }
  };


  struct t_m_transition {
    static const pt_node& getIOGuard(const pt_node& n, 
                                     unsigned int index) {
      ID_ASSERT_(n, ID_M_m_transition);

      return n.children.at(index);
    }

    static unsigned int countIOGuards(const pt_node& n) {
      ID_ASSERT_(n, ID_M_m_transition);

      return n.children.size();
    }
  };

  struct t_r_transition {
    static const pt_node& getAssignment(const pt_node& n, 
                                        unsigned int index) {
      ID_ASSERT_(n, ID_M_r_transition);

      return n.children.at(index);
    }

    static unsigned int countAssignments(const pt_node& n) {
      ID_ASSERT_(n, ID_M_r_transition);

      return n.children.size();
    }
  };

  struct t_compact_transition {
    static BDDFunction getAssignment(const pt_node& n, const BDDFunctionContainer& variables) {
      ID_ASSERT_(n, ID_M_compact_transition);
      
      std::string compact_transition_def(n.value.begin(), n.value.end());

      std::vector<std::string> transitions;
      boost::algorithm::split(transitions, 
                              compact_transition_def,
                              boost::algorithm::is_any_of(";\n"),
                              boost::algorithm::token_compress_on);
      
      BDDFunction bdd_transitions = BDDFunction::ZERO();
      for (std::size_t i = 0;
           i < transitions.size();
           ++i) {
        std::string t = transitions.at(i);
        // remove all whitespace
        boost::algorithm::erase_all(t, " ");
        if (t=="") {
          continue; // skip
        }
        if (boost::algorithm::starts_with(t, "->0x")) {
          t = t.substr(4);
          
          BDDFunction bdd_t = hexToAssignment(t, variables);
          bdd_transitions |= bdd_t;
        } else {
          throw "Illegal compact transition definition:\n"+t;
        }
      }
      
      return bdd_transitions;
    }
  };

  struct t_transition {
    struct info {
      bool has_l, has_m, has_r;
      unsigned int l_index, m_index, r_index;
    };

    static bool hasL(const pt_node& n) {
      ID_ASSERT_(n, ID_M_transition);
      return calc_info(n).has_l;
    }

    static bool hasM(const pt_node& n) {
      ID_ASSERT_(n, ID_M_transition);
      return calc_info(n).has_m;
    }

    static bool hasR(const pt_node& n) {
      ID_ASSERT_(n, ID_M_transition);
      return calc_info(n).has_r;
    }

    static const pt_node& getL(const pt_node& n) {
      ID_ASSERT_(n, ID_M_transition);
      info i=calc_info(n);
      ERROR_CHECK_ASSERT(i.has_l);
      return n.children.at(i.l_index);
    }

    static const pt_node& getM(const pt_node& n) {
      ID_ASSERT_(n, ID_M_transition);
      info i=calc_info(n);
      ERROR_CHECK_ASSERT(i.has_m);
      return n.children.at(i.m_index);
    }

    static const pt_node& getR(const pt_node& n) {
      ID_ASSERT_(n, ID_M_transition);
      info i=calc_info(n);
      ERROR_CHECK_ASSERT(i.has_r);
      return n.children.at(i.r_index);
    }

  private:
    static
    info calc_info(const pt_node& n) {
      info result;
      result.has_l=result.has_m=result.has_r=false;
      result.l_index=result.m_index=result.r_index=0;

      for (unsigned int i=0; i< n.children.size(); i++) {
        string v(n.children.at(i).value.begin(), n.children.at(i).value.end());
        if (v=="-[" || v=="]->") {
          continue;
        }

        switch (n.children.at(i).value.id().to_long()) {
        case ID_M_l_transition:
          result.l_index=i; 
          result.has_l=true; 
          break;
        case ID_M_m_transition:
          result.m_index=i; 
          result.has_m=true; 
          break;
        case ID_M_r_transition:
          result.r_index=i; 
          result.has_r=true; 
          break;
        default:
          throw "implementation error";
        }
      }
      return result;
    }
  };

  struct t_transitions {
    static const pt_node& getTransition(const pt_node& n, unsigned int index) {
      ID_ASSERT_(n, ID_M_transitions);

      return n.children.at(index);
    }

    static unsigned int countTransitions(const pt_node& n) {
      ID_ASSERT_(n, ID_M_transitions);

      return n.children.size();
    }
  };

  struct t_body {
  private:
    /** Return child with the given rule_id*/
    static const pt_node& getChild(const pt_node& n, rule_id id) {
      for (unsigned int i=0;i<n.children.size();i++) {
        if (n.children.at(i).value.id()==parser_id(id)) {
          return n.children.at(i);
        }
      }
      throw "Can not access child!";
    }

    static bool hasChild(const pt_node& n, rule_id id) {
      for (unsigned int i=0;i<n.children.size();i++) {
        if (n.children.at(i).value.id()==parser_id(id)) {
          return true;
        }
      }
      return false;
    }

  public:
    static bool hasInterfaceDefs(const pt_node& n) {
      ID_ASSERT_(n, ID_M_body);
      return hasChild(n, ID_M_interface_defs);
    }

    static bool hasDefs(const pt_node& n) {
      ID_ASSERT_(n, ID_M_body);
      return hasChild(n, ID_M_defs);
    }

    static bool hasAPDefs(const pt_node& n) {
      ID_ASSERT_(n, ID_M_body);
      return hasChild(n, ID_M_ap_defs);
    }

    static bool hasTransitions(const pt_node& n) {
      ID_ASSERT_(n, ID_M_body);
      return hasChild(n, ID_M_transitions);
    }

    static const pt_node& getInterfaceDefs(const pt_node& n) {
      ID_ASSERT_(n, ID_M_body);
      return getChild(n, ID_M_interface_defs);
    }

    static const pt_node& getDefs(const pt_node& n) {
      ID_ASSERT_(n, ID_M_body);
      return getChild(n, ID_M_defs);
    }

    static const pt_node& getAPDefs(const pt_node& n) {
      ID_ASSERT_(n, ID_M_body);
      return getChild(n, ID_M_ap_defs);
    }

    static const pt_node& getTransitions(const pt_node& n) {
      ID_ASSERT_(n, ID_M_body);
      return getChild(n, ID_M_transitions);
    }
  };

  
  struct t_module_def {
    static
    string getName(const pt_node& n) {
      ID_ASSERT_(n, ID_M_module_def);

      return string(n.children.at(0).value.begin(),
                    n.children.at(0).value.end());
    }    

    static
    const pt_node& getBody(const pt_node& n) {
      ID_ASSERT_(n, ID_M_module_def);

      return n.children.at(n.children.size()-1);
    }

    static
    bool hasParameterDef(const pt_node& n) {
      ID_ASSERT_(n, ID_M_module_def);

      return n.children.size()==3;
    }

    static
    const pt_node& getParameterDef(const pt_node& n) {
      ID_ASSERT_(n, ID_M_module_def);

      ERROR_CHECK_ASSERT(hasParameterDef(n));
      return n.children.at(1);
    }
  };

};

}
}


#endif
