/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef REO_PARSER_HPP
#define REO_PARSER_HPP

#include "spirit_init.hpp"
using namespace vereofy::spirit;

#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>

#include <stack>
#include <string>

#include "REOParser.hpp"

#include "reo_parser_types.hpp"
#include "position_node.hpp"
#include "position_node_factory.hpp"

#include "vereofy/common/singleton_set.hpp"
#include "vereofy/common/file_ops.hpp"

#include "parser_helper.hpp"
#include "circuit_parser_helper.hpp"
#include "reo_parser_helper.hpp"
#include "type_parser_helper.hpp"

#include "circuit_parser.hpp"
#include "guard_parser.hpp"
#include "module_parser.hpp"
#include "identifier_parser.hpp"
#include "identifier_parser_helper.hpp"
#include "int_parser.hpp"
#include "type_parser.hpp"
#include "const_def_parser.hpp"
#include "common_parser.hpp"
#include "guard_parser.hpp"

#include "vereofy/system/symb_automaton/symb_function_userdefined.hpp"

namespace vereofy {
namespace vparser {

struct reo_parser : public grammar<reo_parser>
{
 
  static 
  void init_rule_names(map<parser_id, std::string>& rule_names) {
    rule_names[ID_C_global]="global";
    rule_names[ID_C_include]="include";
    rule_names[ID_C_alias_def]="alias_def";
    rule_names[ID_C_replace_def] = "replace_def";
    rule_names[ID_C_global_ap_def] = "global_ap_def";
    rule_names[ID_C_function_arg] = "function_arg";
    rule_names[ID_C_function_arg_list] = "function_arg_list";
    rule_names[ID_C_function_def] = "function_def";
    
    circuit_parser::init_rule_names(rule_names);
    module_parser::init_rule_names(rule_names);
    type_parser::init_rule_names(rule_names);
    int_parser::init_rule_names(rule_names);
    identifier_parser::init_rule_names(rule_names);
    guard_parser::init_rule_names(rule_names);
    const_def_parser::init_rule_names(rule_names);
  }


  static
  void init_reserved_keywords(symbols<>& keywords) {
    circuit_parser::init_reserved_keywords(keywords);
    module_parser::init_reserved_keywords(keywords);
  }
  

  enum Errors {
    no_global_statement
  };

  struct expect_parsers {
    expect_parsers() :
      global_statement(no_global_statement)
    {}

    vereofy::spirit::assertion<Errors> global_statement;
  };


  template <class DefT>
  struct error_handler {
    const DefT& def;

    error_handler(const DefT& def_) : def(def_) {}

    template <typename ScannerT, typename ErrorT>
    error_status<>
    operator()(ScannerT const& scan, ErrorT const& error) const {
      switch (error.descriptor) {
      case no_global_statement:
          throw_(error.where, std::string("Expected a valid global level statement (e.g. CIRCUIT, MODULE, TYPE, CONST, AP or #include)"));
      default:
        return error_status<>(error_status<>::rethrow);
      }
    }
  };


#ifdef REO_PARSER_WITH_MODULES
  struct end_module_def {
    symbol_table_stack_t& symbol_table_stack;
    module_parser& mp;
    
    end_module_def(symbol_table_stack_t& symbol_table_stack_,
                   module_parser& mp_)
      : symbol_table_stack(symbol_table_stack_),
        mp(mp_) {}
    
    template<class Node, class Iter>
    const end_module_def&
    operator()(const Node& n, Iter& i1, Iter& i2) const {
      std::string name=module_helper::t_module_def::getName(n);
      
      if (symbol_table_stack.find(name)) {
        throw_(i1, "There is already a definition for "+name);
      }

      ModuleInfo::ptr mi(new ModuleInfo(name));            
      mi->module=mp.cur_module;
      mp.cur_module.reset();
      symbol_table_stack.getTopTable()->add(name, mi);

      return *this;
    }
  };
#endif

  struct perform_include {
    const reo_parser& self;

    perform_include(const reo_parser& self_) : self(self_) {}
    
    template <class Node, class Iterator>
    void operator()(Node& n, Iterator& it1, Iterator& it2) const {
      std::string filename(n.value.begin(), n.value.end());

      if (filename=="builtin") {
        access_weak_ptr(self.system_context)->loadBuiltins();
        return;
      }

      ReturnValue<std::string> file=access_weak_ptr(self.system_context)->findFile(filename);
      if (!file.valid) {
        throw std::string("Couldn't find file ") + filename;
      }
      
      ReturnValue<std::string> base_name = FileOps::get_basename(filename);
      if (base_name.valid && base_name.value!="" && access_weak_ptr(self.system_context)->ignoreMultipleIncludes()) {
        if (access_weak_ptr(self.system_context)->isFileMarkedAsIncluded(base_name.value)) {
          if (debugLevel>=3) {
            cout << "Skipping include file " << file.value << endl;
          }

          return;
        }
      }

      if (debugLevel>=3) {
        cout << "Including file " << file.value << endl;
      }
      REOParser rp(self.system_context);
      if (!rp.parse_file(file.value)) {
        throw std::string("Error including file ")+file.value;
      }

      if (base_name.valid && base_name.value!="") {
        access_weak_ptr(self.system_context)->markFileAsIncluded(base_name);
      }
    }
  };


  struct perform_alias_def {
    const reo_parser& self;

    perform_alias_def(const reo_parser& self_) : self(self_) {}
    
    template <class Node, class Iterator>
    void operator()(Node& n, Iterator& it1, Iterator& it2) const {
      std::string id_alias=identifier_helper::t_identifier::getIdentifier(n.children.at(0));
      std::string id_name=identifier_helper::t_identifier::getIdentifier(n.children.at(1));

      if (access_weak_ptr(self.system_context)->getGlobalSymbols()->find(id_alias)) {
        throw_(it1, "There is already a definition for "+id_alias);
      }

      SymbolData::ptr symbol=access_weak_ptr(self.system_context)->getGlobalSymbols()->find(id_name);
      if (!symbol) {
        throw_(it1, "There is no definition for "+id_name);
      }

      if (symbol->type!=SYMBOL_TYPE_COMPONENT) {
        throw_(it1, "Can only alias for components/connectors, which "+id_name+" is not");
      }

      if (debugLevel>=2) {
        cout << "Adding alias: " << id_alias <<  "=" << id_name << std::endl;
      }

      access_weak_ptr(self.system_context)->getGlobalSymbols()->add(id_alias, symbol);
    }
  };


  struct perform_replace_def {
    const reo_parser& self;

    perform_replace_def(const reo_parser& self_) : self(self_) {}
    
    template <class Node, class Iterator>
    void operator()(Node& n, Iterator& it1, Iterator& it2) const {
      std::string original(n.children.at(0).value.begin(), n.children.at(0).value.end());
      std::string replacement(n.children.at(1).value.begin(), n.children.at(1).value.end());

      access_weak_ptr(self.system_context)->getReplacementManager()->addReplacement(original, replacement);
    }
  };


  struct perform_global_ap_def {
    const reo_parser& self;

    perform_global_ap_def(const reo_parser& self_) : self(self_) {}
    
    template <class Node, class Iterator>
    void operator()(Node& n, Iterator& it1, Iterator& it2) const {
      std::string type_specifier(n.children.at(0).value.begin(), n.children.at(0).value.end());
      std::string ap_name(n.children.at(1).value.begin(), n.children.at(1).value.end());
      std::string ap_formula(n.children.at(2).value.begin(), n.children.at(2).value.end());

      access_weak_ptr(self.system_context)
        ->getTypeModificationRegistry()
        ->addAtomicProposition(type_specifier, ap_name, ap_formula);

      if (debugLevel>=2) {
        cout << "Adding AP, type: " << type_specifier <<  ", name:" << ap_name << ", formula: " << ap_formula << std::endl;
      }
    }
  };

  struct function_def_1 {
    reo_parser& self;

    function_def_1(const reo_parser& self_) : self(const_cast<reo_parser&>(self_)) {}
    
    template <class Node, class Iterator>
    void operator()(Node& n, Iterator& it1, Iterator& it2) const {
      self.function_ci=ComponentInterface::ptr(new ComponentInterface("function"));

      // std::cerr << reo_parser_helper::t_function_def::getName(n) << std::endl;
      std::string name = reo_parser_helper::t_function_def::getName(n);

      if (access_weak_ptr(self.system_context)->getGlobalSymbols()->find(name)) {
        throw_(it1, "Can not define function "+name+", symbol already in use!");
      }

      const Node& return_type_n=reo_parser_helper::t_function_def::getReturnType(n);
      DataDomain::ptr return_type = type_helper::t_type::getDataDomain(return_type_n, self.symbol_table_stack);
      //      std::cerr << "Return: " << return_type->toString() << "\n";

      if (reo_parser_helper::t_function_def::isReturnTypeBoolean(n)) {
        self.function_expression_bool=true;
      } else {
        self.function_expression_bool=false;
      }

      const Node& arg_list_n = reo_parser_helper::t_function_def::getArgumentList(n);
      for (unsigned int i=0;i<reo_parser_helper::t_function_arg_list::getSize(arg_list_n);i++) {
        const Node& arg_n = reo_parser_helper::t_function_arg_list::getArgument(arg_list_n, i);

        const std::string name = reo_parser_helper::t_function_arg::getName(arg_n);

        const Node& arg_type_n = reo_parser_helper::t_function_arg::getType(arg_n);
        DataDomain::ptr arg_type = type_helper::t_type::getDataDomain(arg_type_n, self.symbol_table_stack);
        StateDescription::ptr sd(new StateDescription(name, arg_type));
        self.function_ci->addStateDescription(sd);

        // std::cerr << "Argument: " << arg_type->toString() << " name\n";
      }
    }
  };

  struct function_def_2 {
    const reo_parser& self;

    function_def_2(const reo_parser& self_) : self(self_) {}
    
    template <class Node, class Iterator>
    void operator()(Node& n, Iterator& it1, Iterator& it2) const {
      std::string name = reo_parser_helper::t_function_def::getName(n);

      access_weak_ptr(self.system_context)->getGlobalSymbols()->
        add(name, SymbolData::ptr(new FunctionData(name, 
                                                   SymbFunction_Userdefined::fromDef(n, self.system_context))));
    }
  };

  
  static
  std::map<parser_id, std::string> getRuleNames() {
    std::map<parser_id, std::string> rule_names;
    
    init_rule_names(rule_names);
    return rule_names;
  };


  reo_parser(symbol_table_stack_t& symbol_table_stack_,
             vereofy::SystemContext::wptr system_context_) 
    : symbol_table_stack(symbol_table_stack_),
      system_context(system_context_) {
    if (debugLevel>=5) {
      cerr << "ST(reo_parser): " << &symbol_table_stack<< endl;
    }
    init_reserved_keywords(reserved_keywords);
  }

  symbol_table_stack_t& symbol_table_stack;
  vereofy::SystemContext::wptr system_context;
  symbols<> reserved_keywords;
  bool function_expression_bool;

  ComponentInterface::ptr function_ci;
  
  template <typename ScannerT>
    struct definition
    {
      definition(reo_parser const& self) :
        identifier_grammar(last_identifier, self.reserved_keywords),
        common_grammar(self.reserved_keywords),
        circuit_def(self.symbol_table_stack, self.reserved_keywords, self.system_context),
        module_def(self.symbol_table_stack, self.reserved_keywords, self.system_context),
        type_grammar(self.symbol_table_stack, self.reserved_keywords),
        const_def(self.symbol_table_stack, self.reserved_keywords, false),
        guard_grammar(self.symbol_table_stack, self.reserved_keywords, self.function_ci)
      {
        guard_grammar.allow_states=true;
        guard_grammar.allow_ports=false;

        BOOST_SPIRIT_DEBUG_GRAMMAR(identifier_grammar);
        BOOST_SPIRIT_DEBUG_GRAMMAR(common_grammar);
        BOOST_SPIRIT_DEBUG_GRAMMAR(type_grammar);

        ENTRY_GRAMMAR(type_def, type_parser, type_grammar, type_parser::type_def);
        ENTRY_GRAMMAR(type, type_parser, type_grammar, type_parser::type);
        ENTRY_GRAMMAR(identifier_p, identifier_parser, identifier_grammar, identifier_parser::identifier);
        ENTRY_GRAMMAR(quoted_string, common_parser, common_grammar, common_parser::quoted_string);
        ENTRY_GRAMMAR(guard_expression, guard_parser, guard_grammar, guard_parser::guard_expression);
        ENTRY_GRAMMAR(bool_expression, guard_parser, guard_grammar, guard_parser::bool_expression);

        //  Start grammar definition

        include = no_node_d[str_p("#include")]
          >> access_node_d[quoted_string][perform_include(self)];
        
        alias_def = no_node_d[str_p("ALIAS")]
          >> access_node_d[identifier_p >> root_node_d[ch_p('=')] >> identifier_p >> no_node_d[ch_p(';')]][perform_alias_def(self)];

        replace_def =
          access_node_d[
                        root_node_d[str_p("REPLACE")]
                        >> no_node_d[ch_p('(')]
                        >> quoted_string
                        >> no_node_d[ch_p(',')]
                        >> quoted_string
                        >> no_node_d[ch_p(')')]
                        >> no_node_d[ch_p(';')]
                        ][perform_replace_def(self)];

        global_ap_def =
          access_node_d[
                        root_node_d[str_p("AP")]
                        >> no_node_d[ch_p('(')]
                        >> quoted_string  // the type_specifier
                        >> no_node_d[ch_p(',')]
                        >> quoted_string  // the AP name
                        >> no_node_d[ch_p(',')]
                        >> quoted_string // the AP formula
                        >> no_node_d[ch_p(')')]
                        >> no_node_d[ch_p(';')]
                        ][perform_global_ap_def(self)];
        
        function_arg = type >> identifier_p;

        function_arg_list = 
          !(function_arg >> *(no_node_d[ch_p(',')] >> function_arg));

        function_def = 
          access_node_d[
                        access_node_d[
                                      root_node_d[str_p("FUNCTION")]
                                      >> type
                                      >> identifier_p
                                      >> no_node_d[ch_p('(')]
                                      >> function_arg_list
                                      >> no_node_d[ch_p(')')]][function_def_1(self)]
                        >> no_node_d[ch_p('=')]
                        >> ((eps_p(phoenix::var(self.function_expression_bool)) 
                             >> eps_p[guard_parser::init(guard_grammar, true, false,false)] >> bool_expression) |
                            (eps_p(!phoenix::var(self.function_expression_bool)) 
                             >> eps_p[guard_parser::init(guard_grammar, true, false,false)] >> guard_expression)) 
                        >> no_node_d[ch_p(';')]][function_def_2(self)];

        global = error_guard(
                 *(~eps_p(end_p) >> expect.global_statement(
                   circuit_def
                   | access_node_d[module_def][end_module_def(self.symbol_table_stack, module_def)]
                   | type_def
                   | const_def
                   | alias_def
                   | replace_def
                   | include
                   | global_ap_def
                   | function_def
                  ))
                )[error_handler<reo_parser>(self)]
          >> eps_p;  // eat whitespace at the end of file
        //  End grammar definition

        // turn on the debugging info.
        
        BOOST_SPIRIT_DEBUG_RULE(global);
        BOOST_SPIRIT_DEBUG_RULE(include);
        BOOST_SPIRIT_DEBUG_RULE(alias_def);
        BOOST_SPIRIT_DEBUG_RULE(replace_def);
        BOOST_SPIRIT_DEBUG_RULE(function_arg);
        BOOST_SPIRIT_DEBUG_RULE(function_arg_list);
        BOOST_SPIRIT_DEBUG_RULE(function_def);
        BOOST_SPIRIT_DEBUG_RULE(global_ap_def);
    
      }

      rule<ScannerT, parser_tag<ID_C_global> >     global;
      rule<ScannerT, parser_tag<ID_C_include> >     include;
      rule<ScannerT, parser_tag<ID_C_alias_def> >     alias_def;
      rule<ScannerT, parser_tag<ID_C_replace_def> >     replace_def;
      rule<ScannerT, parser_tag<ID_C_global_ap_def> >     global_ap_def;
      rule<ScannerT, parser_tag<ID_C_function_arg> >     function_arg;
      rule<ScannerT, parser_tag<ID_C_function_arg_list> >  function_arg_list;
      rule<ScannerT, parser_tag<ID_C_function_def> >     function_def;

      identifier_parser identifier_grammar;
      common_parser common_grammar;
      circuit_parser circuit_def;
#ifdef REO_PARSER_WITH_MODULES
      module_parser module_def;
#endif
      type_parser type_grammar;
      const_def_parser const_def;
      guard_parser guard_grammar;

      
      rule<ScannerT, parser_tag<ID_C_global> > const&
      start() const { return global; }

      guard<Errors> error_guard;
      expect_parsers expect;


      std::string last_identifier;
    };
};

} // namespace vereofy::parser
} // namespace vereofy

#endif
