#ifndef SKIP_PARSER_HPP
#define SKIP_PARSER_HPP

#include "spirit_init.hpp"

#include <map>
#include <string>

using namespace vereofy::spirit;

struct skip_parser : public parser<skip_parser>
{
  typedef skip_parser self_t;
  
  typedef std::map<std::string, bool> flag_map_t;
  
  const flag_map_t& flag_map;

  skip_parser(const flag_map_t& flag_map_) : flag_map(flag_map_) {}

  template <typename ScannerT>
  std::size_t skip_to_end_of_line(ScannerT const& scan) const {
    std::size_t len = 0;
    while (scan.first != scan.last && (*scan!='\n' && *scan!='\r')) {
      ++scan.first;
      ++len;
    }

    return len;
  }


  template <typename ScannerT>
  std::size_t skip_to_end_of_comment_block(ScannerT const& scan) const {
    bool have_star=false;
    bool was_closed=false;
    
    std::size_t len = 0;

    while (scan.first != scan.last) {
      if (have_star==true && *scan == '/') {
        ++scan.first;
        ++len;
        was_closed=true;
        break;
      } else if (*scan == '*') {
        have_star=true;
        ++scan.first;
        ++len;
      } else {
        have_star=false;
        ++scan.first;
        ++len;
      }
    }
    
    if (!was_closed && scan.first == scan.last) {
      throw_(scan.first, std::string("Missing closing */"));
    }

    return len;
  }

  template <typename ScannerT>
  std::size_t skip_whitespace(ScannerT const& scan) const {
    std::size_t len = 0;
    while (scan.first != scan.last && 
           (*scan ==' ' || *scan=='\t')) {
      ++scan.first;
      ++len;
    }

    return len;
  }


  struct scan_result_t {
    scan_result_t() : len(0) {}

    std::size_t len;
    std::string text;
  };

  template <typename ScannerT>
  scan_result_t
  scan_identifier(ScannerT const& scan) const {
    bool first=true;
    scan_result_t result;

    while (scan.first != scan.last) {
      if (isalpha(*scan) || (!first && isdigit(*scan)) || *scan=='_') {
        result.text+=*scan;
        ++scan.first;
        ++result.len;
      } else {
        break;
      }

      first=false;
    }

    return result;
  }


  template <typename ScannerT>
  std::size_t skip_to_end_of_if_block(ScannerT const& scan) const {
    std::size_t len = 0;

    const char* strings[2] = {"@if", "@endif"};
    const char* cur_pos[2] = {strings[0], strings[1]};
    bool have_slash=false;

    unsigned int open_if_blocks=0;
    bool finished=false;

    while (scan.first != scan.last && !finished) {
      if (*scan == '*') {
        if (have_slash) {
          ++scan.first;
          ++len;

          len+=skip_to_end_of_comment_block(scan);
          have_slash=false;
          continue;
        }
      }

      if (*scan == '/') {
        if (have_slash) {
          // C++ style line comment
          len+=skip_to_end_of_line(scan);
          have_slash=false;
          continue;

        } else {
          have_slash=true;
        }
      } else {
        have_slash=false;
      }

      if (*scan == *cur_pos[0]) {
        ++cur_pos[0];
      } else {
        cur_pos[0]=strings[0];
      }

      if (*scan == *cur_pos[1]) {
        ++cur_pos[1];
      } else {
        cur_pos[1]=strings[1];
      }

      if (*cur_pos[0]=='\0') {
        // matched @if
        open_if_blocks++;
      }

      if (*cur_pos[1]=='\0') {
        // matched @endif
        if (open_if_blocks==0) {
          finished=true;
        } else {
          --open_if_blocks;
        }
      }

      ++scan.first;
      ++len;
    }

    return len;
  }
  
  template <typename ScannerT>
  typename parser_result<self_t, ScannerT>::type
  parse(ScannerT const& scan) const
  {
    typename ScannerT::iterator_t save = scan.first;
    std::size_t len = 0;
    
    if (!scan.at_end()) {
      if (std::isspace(*scan)) {
        ++scan.first;
        ++len;
      } else if (*scan == '/') {
        ++scan.first;
        if (scan.first != scan.last && *scan == '/') {
          // C++ style comment till end of line
          ++scan.first;
          len=2;

          len+=skip_to_end_of_line(scan);

        } else if (scan.first != scan.last && *scan =='*') {
          // C style block comment
          ++scan.first;
          len=2;

          len+=skip_to_end_of_comment_block(scan);
        }
      } else if (*scan == '@') {
        // conditional compilation
        ++scan.first;
        ++len;
        
        if (scan.first==scan.last) {
          throw_(scan.first, std::string("Premature end-of-file after @"));
        }

        bool is_if_block=false;

        scan_result_t identifier=scan_identifier(scan);
        if (identifier.len>0) {
          len+=identifier.len;

          if (identifier.text=="if") {
            is_if_block=true;
            len+=skip_whitespace(scan);

          } else if (identifier.text=="endif") {
            // stand-alone @endif, return
            goto finished;
          } else {
            throw_(scan.first, std::string("Syntax error after @, has to be either @if, @endif, @+flag or @-flag"));
          }
        }


        std::string flag_name;
        bool positive=false;

        if (*scan == '+' || *scan == '-') {
          positive = (*scan=='+');
          ++len;
          ++scan.first;
          
          scan_result_t identifier=scan_identifier(scan);
          if (identifier.len==0) {
            throw_(scan.first, std::string("Expected flag name in conditional compilation construct"));
          }

          len+=identifier.len;
          flag_name=identifier.text;
        } else {
          throw_(scan.first, std::string("Expected + or - in conditional compilation construct"));
        }

        // determine whether to skip or to pass
        bool skip;
        if (flag_map.count(flag_name)>0) {
          skip = (flag_map.find(flag_name)->second != positive);
        } else {
          // not in map = flag is -
          skip = (false != positive);
        }

        if (skip) {
          if (is_if_block) {
            len+=skip_to_end_of_if_block(scan);
          } else {
            len+=skip_to_end_of_line(scan);
          }
        } else {
          // nothing to do, just skip over the already scanned part
        }
      }
    }

  finished:
    if (len)
      return scan.create_match(len, nil_t(), save, scan.first);

    scan.first=save;
    return scan.no_match();
  }
};


#endif
