/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef TYPE_PARSER_HPP
#define TYPE_PARSER_HPP

#include "spirit_init.hpp"
using namespace vereofy::spirit;

#include <boost/bind.hpp>

#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>

#include <map>
#include <string>

#include "reo_parser_types.hpp"
#include "position_node.hpp"
#include "position_node_factory.hpp"

#include "vereofy/common/errors.hpp"

#include "int_parser.hpp"
#include "identifier_parser.hpp"

#include "parser_helper.hpp"
#include "type_parser_helper.hpp"

namespace vereofy {
namespace vparser {

struct type_parser : public grammar<type_parser>
{
  enum {type = 0, type_def = 1};

  enum Errors {
    no_const_expression,
    no_left_parenthesis,
    no_right_parenthesis,
    no_left_curly,
    no_right_curly,
    no_comma,
    no_semicolon
  };

  struct expect_parsers {
    expect_parsers() :
      const_expression(no_const_expression),
      left_parenthesis(no_left_parenthesis),
      right_parenthesis(no_right_parenthesis),
      left_curly(no_left_curly),
      right_curly(no_right_curly),
      comma(no_comma),
      semicolon(no_semicolon)
    {}

    vereofy::spirit::assertion<Errors> const_expression;
    vereofy::spirit::assertion<Errors> left_parenthesis;
    vereofy::spirit::assertion<Errors> right_parenthesis;
    vereofy::spirit::assertion<Errors> left_curly;
    vereofy::spirit::assertion<Errors> right_curly;
    vereofy::spirit::assertion<Errors> comma;
    vereofy::spirit::assertion<Errors> semicolon;
  };

  template <class DefT>
  struct error_handler {
    const DefT& def;
    
    error_handler(const DefT& def_) : def(def_) {}
    
    template <typename ScannerT, typename ErrorT>
    error_status<>
    operator()(ScannerT const& scan, ErrorT const& error) const {
      switch (error.descriptor) {
      case no_const_expression:
        throw_(error.where, std::string("Expected a constant expression"));

      case no_left_parenthesis:
        throw_(error.where, std::string("Expected opening '('"));
      case no_right_parenthesis:
        throw_(error.where, std::string("Expected closing ')'"));

      case no_left_curly:
        throw_(error.where, std::string("Expected opening '{'"));
      case no_right_curly:
        throw_(error.where, std::string("Expected closing '}'"));
      case no_comma:
        throw_(error.where, std::string("Expected ','"));
      case no_semicolon:
        throw_(error.where, std::string("Expected ';'"));
      }
      return error_status<>(error_status<>::rethrow);
    }
  };

  static 
  void init_rule_names(std::map<parser_id, std::string>& rule_names) {
    rule_names[ID_T_int_range] = "int_range";
    rule_names[ID_T_enum_type] = "enum_type";
    rule_names[ID_T_int_type] = "int_type";
    rule_names[ID_T_named_type] = "named_type";
    rule_names[ID_T_flat_type] = "flat_type";
    rule_names[ID_T_array_type] = "array_type";
    rule_names[ID_T_struct_type] = "struct_type";
    rule_names[ID_T_tagged_union_type] = "tagged_union_type";
    rule_names[ID_T_struct_member_def] = "struct_member_def";
    rule_names[ID_T_type] = "type";
    rule_names[ID_T_type_def] = "type_def";
  }

  type_parser(symbol_table_stack_t& symbol_table_stack_,
              const symbols<>& reserved_keywords_) :
    symbol_table_stack(symbol_table_stack_),
    reserved_keywords(reserved_keywords_) {
  }

  symbol_table_stack_t& symbol_table_stack;
  const symbols<>& reserved_keywords;


  struct s_type_def {
    symbol_table_stack_t& symbol_table_stack;

    s_type_def(symbol_table_stack_t& symbol_table_stack_) :
      symbol_table_stack(symbol_table_stack_) {}

    template <class Node, class It>
    void operator()(Node&n, It& i1, It& i2) const {
      std::string name=type_helper::t_type_def::getName(n);
      DataDomain::ptr dd_type=
        type_helper::t_type_def::getDataDomain(n, symbol_table_stack);

      if (dd_type->getType()==DataDomain::DD_FLAT) {
        FlatDataDomain::ptr dd_flat_type=
          boost::static_pointer_cast<FlatDataDomain>(dd_type);
        
        if (dd_flat_type->getFlatType()==FlatDataDomain::DD_ENUM) {
          EnumDataDomain& enum_dd=static_cast<EnumDataDomain&>(*dd_flat_type);
          enum_dd.setTypeName(name);
        }
      }
        
      TypeDefData::ptr tdd(new TypeDefData(name, dd_type));
      
      if (symbol_table_stack.find(name)) {
        throw_(i1, "There is already a definition for "+name);
      }

      symbol_table_stack.getTopTable()->add(name, tdd);
    }
  };

  struct is_named_type {
    symbol_table_stack_t& symbol_table_stack;
    std::string& name_ref;
    
    is_named_type(symbol_table_stack_t& symbol_table_stack_,
                  std::string& name_ref_) :
      symbol_table_stack(symbol_table_stack_),
      name_ref(name_ref_) {}
    
    bool operator()() const {
      SymbolData::ptr sd=symbol_table_stack.find(name_ref);
      if (!sd) {
        return false;
      }

      if (sd->type!=SYMBOL_TYPE_TYPEDEF) {
        return false;
      }

      return true;
    }
  };


  template <typename ScannerT>
  struct definition :
    public grammar_def <rule<ScannerT, parser_tag<ID_T_type> >,
                        rule<ScannerT, parser_tag<ID_T_type_def> > >
  {
    definition(type_parser const& self)
      : integer_grammar(self.symbol_table_stack),
        identifier_grammar(last_identifier, self.reserved_keywords)
    {
      BOOST_SPIRIT_DEBUG_GRAMMAR(integer_grammar);
      BOOST_SPIRIT_DEBUG_GRAMMAR(identifier_grammar);

      ENTRY_GRAMMAR(identifier_p, identifier_parser, identifier_grammar, identifier_parser::identifier);
      ENTRY_GRAMMAR(integer_p, int_parser, integer_grammar, int_parser::integer);
      ENTRY_GRAMMAR(const_expression_p, int_parser, integer_grammar, int_parser::const_expression);

      //  Start grammar definition

      int_range = 
        (no_node_d[ch_p('(')]
         >> expect.const_expression(const_expression_p)
         >> expect.comma(no_node_d[ch_p(',')])
         >> expect.const_expression(const_expression_p)
         >> expect.right_parenthesis(no_node_d[ch_p(')')])
        )
        | 
        (
         const_expression_p
         >> no_node_d[str_p("..")] 
         >> expect.const_expression(const_expression_p)
         );
      
      int_type = root_node_d[keyword_p("int")] 
        >> !int_range;
      
      enum_type = no_node_d[keyword_p("enum")]
        >> expect.left_curly(no_node_d[ch_p('{')])
        >> infix_node_d[list_p(identifier_p, ch_p(','))]
        >> expect.right_curly(no_node_d[ch_p('}')]);

      named_type = 
           identifier_p 
        >> eps_p(is_named_type(self.symbol_table_stack, last_identifier));

      flat_type = 
        int_type
        | enum_type
        | named_type;
             
      array_type = 
        flat_type
        >> no_node_d[ch_p('[')]
        >> !const_expression_p
        >> no_node_d[ch_p(']')];

      struct_member_def = 
        flat_type >> identifier_p >> expect.semicolon(ch_p(';'));
      
      struct_type = 
        no_node_d[keyword_p("struct")]
        >> expect.left_curly(no_node_d[ch_p('{')])
        >> *struct_member_def
        >> expect.right_curly(no_node_d[ch_p('}')]);
 
      tagged_union_type = 
        no_node_d[keyword_p("tagged_union")]
        >> expect.left_curly(no_node_d[ch_p('{')])
        >> *struct_member_def
        >> expect.right_curly(no_node_d[ch_p('}')]);
 
      type = 
        error_guard(
                    struct_type 
                    | tagged_union_type
                    | array_type 
                    | flat_type
                    )[error_handler<type_parser>(self)];

      type_def = 
        error_guard(
                    access_node_d[
                               root_node_d[keyword_p("TYPE")]
                               >> identifier_p
                               >> no_node_d[ch_p('=')]
                               >> type
                               >> expect.semicolon(no_node_d[ch_p(';')])
                               ][s_type_def(self.symbol_table_stack)]
                    )[error_handler<type_parser>(self)];
    
        //  End grammar definition

      this->start_parsers(type, type_def);
        
      // turn on the debugging info.
      BOOST_SPIRIT_DEBUG_RULE(int_range);
      BOOST_SPIRIT_DEBUG_RULE(int_type);
      BOOST_SPIRIT_DEBUG_RULE(enum_type);
      BOOST_SPIRIT_DEBUG_RULE(named_type);
      BOOST_SPIRIT_DEBUG_RULE(flat_type);
      BOOST_SPIRIT_DEBUG_RULE(array_type);
      BOOST_SPIRIT_DEBUG_RULE(struct_type);
      BOOST_SPIRIT_DEBUG_RULE(tagged_union_type);
      BOOST_SPIRIT_DEBUG_RULE(struct_member_def);
      BOOST_SPIRIT_DEBUG_RULE(type);
      BOOST_SPIRIT_DEBUG_RULE(type_def);
    }
    
    rule<ScannerT, parser_tag<ID_T_int_range> >     int_range;
    rule<ScannerT, parser_tag<ID_T_int_type> >      int_type;
    rule<ScannerT, parser_tag<ID_T_enum_type> >     enum_type;
    rule<ScannerT, parser_tag<ID_T_named_type> >    named_type;
    rule<ScannerT, parser_tag<ID_T_flat_type> >     flat_type;
    rule<ScannerT, parser_tag<ID_T_array_type> >    array_type;
    rule<ScannerT, parser_tag<ID_T_struct_type> >   struct_type;
    rule<ScannerT, parser_tag<ID_T_tagged_union_type> >   tagged_union_type;
    rule<ScannerT, parser_tag<ID_T_struct_member_def> > struct_member_def;
    rule<ScannerT, parser_tag<ID_T_type> >          type;
    rule<ScannerT, parser_tag<ID_T_type_def> >      type_def;

    int_parser integer_grammar;
    identifier_parser identifier_grammar;

    guard<Errors> error_guard;
    expect_parsers expect;

    std::string last_identifier;
  };

};

}
}

#endif
