/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef TYPE_PARSER_HELPER_HPP
#define TYPE_PARSER_HELPER_HPP 


#include "int_parser_helper.hpp"
#include "identifier_parser_helper.hpp"

namespace vereofy {
namespace vparser {

struct type_helper {

  struct t_int_range {
    
    static int_value_t 
    getStart(const pt_node& n,
             const symbol_table_stack_t& symbols) {
      return int_helper::t_const_expression::evaluate(n.children.at(0), symbols);
    }

    static int_value_t 
    getEnd(const pt_node& n,
           const symbol_table_stack_t& symbols) {
      return int_helper::t_const_expression::evaluate(n.children.at(1), symbols);
    }

  };

  struct t_int_type {
    static
    const pt_node& getIntRange(const pt_node&n) {
      return n.children.at(0);
    }

    static
    bool hasIntRange(const pt_node&n) {
      return n.children.size()==1;
    }

    static
    FlatDataDomain::ptr getDataDomain(const pt_node& n,
                                      const symbol_table_stack_t& symbols) {
      if (hasIntRange(n)) {
        int_value_t range_start=t_int_range::getStart(getIntRange(n), symbols);
        int_value_t range_end=t_int_range::getEnd(getIntRange(n), symbols);
        return IntRangeDataDomain::ptr(new IntRangeDataDomain(range_start,
                                                              range_end));
      } else {
        return IntRangeDataDomain::ptr(new IntRangeDataDomain());
      }
    }
  };

  struct t_enum_type {
    static unsigned int getSize(const pt_node& n) {
      return n.children.size();
    }

    static string getIdentifier(const pt_node& n, unsigned int i) {
      return identifier_helper::t_identifier::getIdentifier(n.children.at(i));
    }

    static
    EnumDataDomain::ptr getDataDomain(const pt_node&n) {
      EnumDataDomain::ptr edd(new vereofy::EnumDataDomain());
      for (unsigned int i=0;i<getSize(n);i++) {
        edd->addElement(getIdentifier(n, i));
      }
      return edd;
    }
  };

  struct t_named_type {
    static string getIdentifier(const pt_node& n) {
      return identifier_helper::t_identifier::getIdentifier(n.children.at(0));
    }

    static
    DataDomain::ptr getDataDomain(const pt_node& n, 
                                  const symbol_table_stack_t& symbol_table_stack) {
      string name=getIdentifier(n);
      SymbolData::ptr sd=symbol_table_stack.find(name);
      if (!sd) {
        throw string("No data type named ")+name;
      } else if (sd->type!=SYMBOL_TYPE_TYPEDEF) {
        throw string("The symbol ")+name+" does not name a type!";
      }
        
      TypeDefData::ptr tdd=boost::static_pointer_cast<TypeDefData>(sd);
      
      return tdd->data_domain;
    }
  };


  struct t_flat_type {
    static
    DataDomain::ptr getDataDomain(const pt_node& n,
                                  const symbol_table_stack_t& symbols) {
      switch (n.children.at(0).value.id().to_long()) {
      case ID_T_int_type:
        return t_int_type::getDataDomain(n.children.at(0), symbols);
      case ID_T_enum_type:
        return t_enum_type::getDataDomain(n.children.at(0));
      case ID_T_named_type:
        return t_named_type::getDataDomain(n.children.at(0),
                                           symbols);
      default:
        THROW_UNIMPLEMENTED();
      }
    }

    static bool isNamedType(const pt_node& n) {
      return (n.value.id() == parser_id(ID_T_named_type));
    }
  };

  struct t_array_type {
    static
    DataDomain::ptr getDataDomain(const pt_node& n,
                                  const symbol_table_stack_t& symbols) {
      DataDomain::ptr base_type=
        t_flat_type::getDataDomain(n.children.at(0),
                                   symbols);
      
      switch (base_type->getType()) {
      case DataDomain::DD_FLAT: {
        FlatDataDomain::ptr flat_base=
          boost::static_pointer_cast<FlatDataDomain>(base_type);

        if (n.children.size()==1) {
          // no size
          return DataDomain::ptr(new ArrayDataDomain(flat_base));
        } else {
          int_value_t size=int_helper::t_const_expression::evaluate(n.children.at(1), symbols);

          if (size.isDefined() && size.getValue()<0) {
            throw "Array size "
              +boost::lexical_cast<std::string>(size.getValue())
              +" is negative!";
          }

          uint_value_t usize=(uint_value_t)size;
          
          return DataDomain::ptr(new ArrayDataDomain(flat_base, usize));
        }
      }
      case DataDomain::DD_ARRAY:
        throw "Multi-dimensional arrays are not supported";
      case DataDomain::DD_STRUCT:
      case DataDomain::DD_TAGGED_UNION:
        throw "Arrays of structured data types are not supported";
      }

      throw "implementation error";
    }
  };

  struct t_struct_member_def {
    static
    std::string getMemberName(const pt_node& n) {
      return identifier_helper::t_identifier::getIdentifier(n.children.at(1));
    }

    static
    DataDomain::ptr getDataDomain(const pt_node& n,
                                  const symbol_table_stack_t& symbols) {
      return t_flat_type::getDataDomain(n.children.at(0),
                                        symbols);
    }
  };

  struct t_struct_type {
    static
    DataDomain::ptr getDataDomain(const pt_node& n,
                                  const symbol_table_stack_t& symbols) {
      StructDataDomain::ptr struct_dd(new StructDataDomain());

      for (unsigned int i=0; i<n.children.size();i++) {
        const pt_node& n_struct_member_def=n.children.at(i);
        const std::string name=
          t_struct_member_def::getMemberName(n_struct_member_def);

        DataDomain::ptr member_type=
          t_struct_member_def::getDataDomain(n_struct_member_def,
                                             symbols);
        struct_dd->addMember(name, member_type);
      }
      
      return struct_dd;
    }
  };
  
  struct t_tagged_union_type {
    static
    DataDomain::ptr getDataDomain(const pt_node& n,
                                  const symbol_table_stack_t& symbols) {
      TaggedUnionDataDomain::ptr union_dd(new TaggedUnionDataDomain());

      for (unsigned int i=0; i<n.children.size();i++) {
        const pt_node& n_struct_member_def=n.children.at(i);
        const std::string name=
          t_struct_member_def::getMemberName(n_struct_member_def);

        DataDomain::ptr member_type=
          t_struct_member_def::getDataDomain(n_struct_member_def,
                                             symbols);
        union_dd->addMember(name, member_type);
      }
      
      return union_dd;
    }
  };


  struct t_type {
    static
    DataDomain::ptr getDataDomain(const pt_node& n,
                                  const symbol_table_stack_t& symbols) {
      switch (n.children.at(0).value.id().to_long()) {
      case ID_T_flat_type:
        return t_flat_type::getDataDomain(n.children.at(0), symbols);
      case ID_T_array_type:
        return t_array_type::getDataDomain(n.children.at(0), symbols);
      case ID_T_struct_type:
        return t_struct_type::getDataDomain(n.children.at(0), symbols);
      case ID_T_tagged_union_type:
        return t_tagged_union_type::getDataDomain(n.children.at(0), symbols);
      default:
        THROW_UNIMPLEMENTED();
      }
    }

    static bool isBoolDataType(const pt_node& n) {
      if (n.children.at(0).value.id() == parser_id(ID_T_flat_type) &&
          t_flat_type::isNamedType(n.children.at(0).children.at(0)) && 
          t_named_type::getIdentifier(n.children.at(0).children.at(0)) == "Bool") {
        return true;
      }
      return false;
    }
  };

  struct t_type_def {
    static string getName(const pt_node& n) {
      return identifier_helper::t_identifier::getIdentifier(n.children.at(0));
    }

    static 
    DataDomain::ptr getDataDomain(const pt_node& n, 
                                  const symbol_table_stack_t& symbols) {
      return t_type::getDataDomain(n.children.at(1), symbols);
    }
  };

};


}
}

#endif
