/**************************************************************************
*   Copyright (C) 2008-2009 by Johannes Osterholzer                       *
*   s5892660@inf.tu-dresden.de                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "dataconstraint.hpp"
#include "scheduleddataexpression.hpp"
#include <algorithm>
#include <iterator>
#include "helper.hpp"

void DataConstraint::printTree() const
{
  printTreeHelper(0);
}

DataConstraint::ptr DataConstraint::evaluateAST(iter_t const &i, const set<string> &allowedPorts)
{
//   cerr << "In DC::evaluateAST. i->value = " <<
//     string(i->value.begin(), i->value.end()) <<
//     " i->children.size() = " << i->children.size() << endl;

  // just abbreviations
  node_val_data<>::const_iterator_t begin = i->value.begin();
  node_val_data<>::const_iterator_t end = i->value.end();

  // test for all possible parse node types
  if(i->value.id() == sdexp::guard_litID) {
    if(*eatSpaces(begin, end) == '!') {
      assert(i->children.size() == 1);
      DataConstraint::ptr ret(new DCNegation(evaluateAST(i->children.begin(), allowedPorts)));
      return ret;
    } else if(*i->value.begin() == '=') {
      assert(i->children.size() == 2);
      iter_t left = i->children.begin();
      iter_t right = i->children.begin() + 1;
      if(left->value.id() == sdexp::portnameID && right->value.id() == sdexp::rhsvalueID) {
	string integer(right->value.begin(), right->value.end());

	string port(1, *eatSpaces(left->value.begin(), left->value.end()));
	if(find(allowedPorts.begin(), allowedPorts.end(), port) == allowedPorts.end())
	  throw "Invalid port access!";

	DataConstraint::ptr ret(new DCNumericalConstraint(port, atoi(integer.c_str())));
	return ret;
      } else if(left->value.id() == sdexp::portnameID && right->value.id() == sdexp::portnameID) {
	string lport(1, *eatSpaces(left->value.begin(), left->value.end()));
	string rport(1, *eatSpaces(right->value.begin(), right->value.end()));
	if(find(allowedPorts.begin(), allowedPorts.end(), lport) == allowedPorts.end() ||
	   find(allowedPorts.begin(), allowedPorts.end(), rport) == allowedPorts.end())
	  throw "Invalid port access!";

	DataConstraint::ptr ret(new DCPortEqualityConstraint(lport, rport));
	return ret;
      } else
	assert(0);
    } else if(*eatSpaces(begin, end) == '<') {
      iter_t left = i->children.begin();
      iter_t right = i->children.begin() + 1;
      assert(left->value.id() == sdexp::portnameID && right->value.id() == sdexp::valuesID);
      
      string port(1, *eatSpaces(left->value.begin(), left->value.end()));
      if(find(allowedPorts.begin(), allowedPorts.end(), port) == allowedPorts.end())
	throw "Invalid port access!";

      set<int> values;
      for(unsigned int j = 1; j < right->children.size(); j +=2) {
	iter_t val = right->children.begin() + j;
	assert(val->value.id() == sdexp::rhsvalueID);
	string integer(val->value.begin(), val->value.end());
	values.insert(atoi(integer.c_str()));
      }
      DataConstraint::ptr ret(new DCNumericalConstraint(port, values));
      return ret;
    } else {
      //cerr << "unknown symbol " << *i->value.begin() << endl;
      assert(0);
    }
  } else if(i->value.id() == sdexp::guard_conjID) {
    assert(*eatSpaces(begin, end) == '&');
    DataConstraint::ptr ret(new DCConjunction(evaluateAST(i->children.begin(), allowedPorts),
					      evaluateAST(i->children.begin() + 1, allowedPorts)));
    return ret;
  } else if(i->value.id() == sdexp::guardID) {
    assert(*eatSpaces(begin, end) == '|');
    DataConstraint::ptr ret(new DCDisjunction(evaluateAST(i->children.begin(), allowedPorts),
					      evaluateAST(i->children.begin() + 1, allowedPorts)));
    return ret;
  } else if(i->value.id() == sdexp::trueID) {
    return DataConstraint::ptr(new DCTrue);
  } else if(i->value.id() == sdexp::falseID) {
    return DataConstraint::ptr(new DCFalse);
  } else
    assert(0);
}

DataConstraint::ptr DataConstraint::negationFree() const
{
  return clone();
}

DataConstraint::ptr DataConstraint::canonicalForm() const
{
  throw "DataConstraint::canonicalForm: unexpected subformula";
  return clone();
}

void DataConstraint::reduceConstraints()
{
  throw "DataConstraint::reduceConstraints: unexpected subformula";
}

////////////
// DCTrue
////////////

DataConstraint::ptr DCTrue::clone() const
{
  DataConstraint::ptr ret(new DCTrue());
  return ret;
}

void DCTrue::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "DCTrue\n";
}

DataConstraint::ptr DCTrue::canonicalForm() const
{
  // [<>] = true, ie return a clause containing the empty dual clause
  shared_ptr<DCClause> ret(new DCClause);
  DCClause::set_t foo;
  DCDualClause dualClause;
  foo.insert(dualClause);
  ret->setElements(foo);
  return shared_static_cast<DataConstraint>(ret);
}


////////////
// DCFalse
////////////

DataConstraint::ptr DCFalse::clone() const
{
  DataConstraint::ptr ret(new DCFalse());
  return ret;
}

void DCFalse::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "DCFalse\n";
}

DataConstraint::ptr DCFalse::canonicalForm() const
{
  // [] = false, ie return the empty clause
  DataConstraint::ptr ret(new DCClause);
  return ret;
}

////////////
// DCNumericalConstraint
////////////

DCNumericalConstraint::DCNumericalConstraint()
  : DataConstraint()
{}

DCNumericalConstraint::DCNumericalConstraint(const string &_variable, int _value)
  : DataConstraint(), variable(_variable)
{
  values.insert(_value);
}

DCNumericalConstraint::DCNumericalConstraint(const string &_variable, const set<int> &_values)
  : DataConstraint(), variable(_variable), values(_values)
{}

DCNumericalConstraint::DCNumericalConstraint(const DCNumericalConstraint &rhs)
  : DataConstraint(), variable(rhs.variable), values(rhs.values)
{}

const string& DCNumericalConstraint::getVariable() const
{
  return variable;
}

const set<int>& DCNumericalConstraint::getValues() const
{
  return values;
}

void DCNumericalConstraint::setVariable(const string& _variable)
{
  variable = _variable;
}

void DCNumericalConstraint::setValue(int _value)
{
  values.clear();
  values.insert(_value);
}

void DCNumericalConstraint::setValues(const set<int> &_values)
{
  values = _values;
}

DataConstraint::ptr DCNumericalConstraint::clone() const
{
  DataConstraint::ptr ret(new DCNumericalConstraint(variable, values));
  return ret;
}

void DCNumericalConstraint::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "DCNumericalConstraint\n";
  cerr << string(level, ' ') << "variable = " << variable << endl;
  cerr << string(level, ' ') << "values = {";
  for(set<int>::iterator it = values.begin(); it != values.end();) {
    cerr << *it;
    if(++it != values.end())
      cerr << ", ";
  }
  cerr << "}\n";
}

DataConstraint::ptr DCNumericalConstraint::canonicalForm() const
{
  // just return a clause containing a dual clause containing the current constraint.
  shared_ptr<DCClause> clause(new DCClause);
  DCDualClause dualclause;
  
  DCDualClause::set_t conjuncts;
  conjuncts.insert(*this);
  dualclause.setElements(conjuncts);

  DCClause::set_t disjuncts;
  disjuncts.insert(dualclause);
  clause->setElements(disjuncts);
  
  return shared_static_cast<DataConstraint>(clause);
}

////////////
// DCPortEqualityConstraint
////////////

DCPortEqualityConstraint::DCPortEqualityConstraint()
  : DataConstraint()
{}

DCPortEqualityConstraint::DCPortEqualityConstraint(const string &_lhs, const string &_rhs)
  : DataConstraint(), lhs(_lhs), rhs(_rhs)
{}

DCPortEqualityConstraint::DCPortEqualityConstraint(const DCPortEqualityConstraint &other)
  : DataConstraint(), lhs(other.lhs), rhs(other.rhs)
{}

const string& DCPortEqualityConstraint::getLHS() const
{
  return lhs;
}

const string& DCPortEqualityConstraint::getRHS() const
{
  return rhs;
}

void DCPortEqualityConstraint::setLHS(const string& _lhs)
{
  lhs = _lhs;
}

void DCPortEqualityConstraint::setRHS(const string& _rhs)
{
  rhs = _rhs;
}

DataConstraint::ptr DCPortEqualityConstraint::clone() const
{
  DataConstraint::ptr ret(new DCPortEqualityConstraint(lhs, rhs));
  return ret;
}

void DCPortEqualityConstraint::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "DCPortEqualityConstraint\n";
  cerr << string(level, ' ') << "lhs = " << lhs << endl;
  cerr << string(level, ' ') << "rhs = " << rhs << endl;
}

DataConstraint::ptr DCPortEqualityConstraint::negationFree() const
{
  // in the course of making the equality constraint negation free, we transform it to a conjunction of numerical constraints
  DataConstraint::ptr ret(new DCFalse);
  for(int i = ScheduledDataExpression::getDataDomain().first; i <= ScheduledDataExpression::getDataDomain().second; ++i) {
    DataConstraint::ptr leftPort(new DCNumericalConstraint(lhs, i));
    DataConstraint::ptr rightPort(new DCNumericalConstraint(rhs, i));
    DataConstraint::ptr conj(new DCConjunction(leftPort, rightPort));

    DataConstraint::ptr newret(new DCDisjunction(conj, ret));
    ret = newret;
  }
  return ret;
}

///////////////
// DCNegation
///////////////

DCNegation::DCNegation()
  : DataConstraint()
{}

DCNegation::DCNegation(DataConstraint::ptr _child)
  : DataConstraint(), child(_child)
{}

DCNegation::DCNegation(const DCNegation& rhs)
  : DataConstraint(), child(rhs.child->clone())
{}

const DataConstraint& DCNegation::getChild() const
{
  return *child;
}

void DCNegation::setChild(DataConstraint::ptr _child)
{
  child = _child;
}

DataConstraint::ptr DCNegation::clone() const
{
  DataConstraint::ptr ret(new DCNegation(child->clone()));
  return ret;
}

void DCNegation::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "DCNegation\n";
  child->printTreeHelper(level + 2);
}

DataConstraint::ptr DCNegation::negationFree() const
{
  // we'll have to do some pattern matching here concerning the subformulae
  if(child->isNegation()) {
    // !!F -> F
    return shared_dynamic_cast<DCNegation>(child)->getChild().negationFree();
  } else if(child->isConjunction()) {
    // !(F & G) -> !F | !G
    DataConstraint::ptr leftNegation(
      new DCNegation(shared_dynamic_cast<DCConjunction>(child)->getLeftChild().clone()));
    DataConstraint::ptr rightNegation(
      new DCNegation(shared_dynamic_cast<DCConjunction>(child)->getRightChild().clone()));

    DataConstraint::ptr ret(new DCDisjunction(leftNegation->negationFree(), rightNegation->negationFree()));
    return ret;
  } else if(child->isDisjunction()) {
    // !(F | G) -> !F & !G
    DataConstraint::ptr leftNegation(
      new DCNegation(shared_dynamic_cast<DCDisjunction>(child)->getLeftChild().clone()));
    DataConstraint::ptr rightNegation(
      new DCNegation(shared_dynamic_cast<DCDisjunction>(child)->getRightChild().clone()));
    // TODO: refactor get*Child to shared_ptr so that i don't have to copy

    DataConstraint::ptr ret(new DCConjunction(leftNegation->negationFree(), rightNegation->negationFree()));
    return ret;
  } else if(child->isNumericalConstraint()) {
    // !<X, X <- Y>  ->  <X, X <- D \ Y>
    shared_ptr<DCNumericalConstraint> ch = shared_dynamic_cast<DCNumericalConstraint>(child);
    set<int> newValues;
    for(int i = ScheduledDataExpression::getDataDomain().first; i <= ScheduledDataExpression::getDataDomain().second; ++i)
      if(ch->getValues().find(i) == ch->getValues().end())
	newValues.insert(i);
    return DataConstraint::ptr(new DCNumericalConstraint(ch->getVariable(), newValues));
  } else if(child->isPortEqualityConstraint()) {
    // simplify port equality before going on
    DataConstraint::ptr portEqualitiesRemoved = child->negationFree();
    DataConstraint::ptr neg(new DCNegation(portEqualitiesRemoved));
    return neg->negationFree();
  } else if(child->isTrue()) {
    // triv
    DataConstraint::ptr ret(new DCFalse);
    return ret;
  } else if(child->isFalse()) {
    // triv
    DataConstraint::ptr ret(new DCTrue);
    return ret;
  } else return DataConstraint::negationFree();
}

///////////////////
// DCConjunction
///////////////////

DCConjunction::DCConjunction()
  : DataConstraint()
{}

DCConjunction::DCConjunction(DataConstraint::ptr _left, DataConstraint::ptr _right)
  : DataConstraint(), leftChild(_left), rightChild(_right)
{}

DCConjunction::DCConjunction(const DCConjunction &rhs)
  : DataConstraint(), leftChild(rhs.leftChild->clone()), rightChild(rhs.rightChild->clone())
{}

const DataConstraint& DCConjunction::getLeftChild() const
{
  return *leftChild;
}

const DataConstraint& DCConjunction::getRightChild() const
{
  return *rightChild;
}

void DCConjunction::setLeftChild(DataConstraint::ptr _left)
{
  leftChild = _left;
}

void DCConjunction::setRightChild(DataConstraint::ptr _right)
{
  rightChild = _right;
}

DataConstraint::ptr DCConjunction::clone() const
{
  DataConstraint::ptr ret(new DCConjunction(leftChild->clone(), rightChild->clone()));
  return ret;
}

void DCConjunction::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "DCConjunction\n";
  leftChild->printTreeHelper(level + 2);
  rightChild->printTreeHelper(level + 2);
}

DataConstraint::ptr DCConjunction::negationFree() const
{
  // nothing to do here
  DataConstraint::ptr ret(new DCConjunction(leftChild->negationFree(),
					    rightChild->negationFree()));
  return ret;
}

DataConstraint::ptr DCConjunction::canonicalForm() const
{
  DCClause::set_t newDisjuncts;
  // compute the canonical forms of subformulae
  shared_ptr<DCClause> left = shared_dynamic_cast<DCClause>(leftChild->canonicalForm());
  shared_ptr<DCClause> right = shared_dynamic_cast<DCClause>(rightChild->canonicalForm());

  // "multiply out" the elements of the clauses (apply distributivity)
  for(DCClause::set_t::iterator it = left->getElements().begin(); it != left->getElements().end(); ++it) {
    for(DCClause::set_t::iterator jt = right->getElements().begin(); jt != right->getElements().end(); ++jt) {
      DCDualClause::set_t newConjuncts;
      set_union(it->getElements().begin(), it->getElements().end(),
		jt->getElements().begin(), jt->getElements().end(),
		inserter(newConjuncts, newConjuncts.end()));
      
      newDisjuncts.insert(newConjuncts);
    }
  }

  DataConstraint::ptr ret(new DCClause(newDisjuncts));
  return ret;
}
  

///////////////////
// DCDisjunction
///////////////////

DCDisjunction::DCDisjunction()
  : DataConstraint()
{}

DCDisjunction::DCDisjunction(DataConstraint::ptr _left, DataConstraint::ptr _right)
  : DataConstraint(), leftChild(_left), rightChild(_right)
{}

DCDisjunction::DCDisjunction(const DCDisjunction &rhs)
  : DataConstraint(), leftChild(rhs.leftChild->clone()), rightChild(rhs.rightChild->clone())
{}

const DataConstraint& DCDisjunction::getLeftChild() const
{
  return *leftChild;
}

const DataConstraint& DCDisjunction::getRightChild() const
{
  return *rightChild;
}

void DCDisjunction::setLeftChild(DataConstraint::ptr _left)
{
  leftChild = _left;
}

void DCDisjunction::setRightChild(DataConstraint::ptr _right)
{
  rightChild = _right;
}

DataConstraint::ptr DCDisjunction::clone() const
{
  DataConstraint::ptr ret(new DCDisjunction(leftChild->clone(), rightChild->clone()));
  return ret;
}

void DCDisjunction::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "DCDisjunction\n";
  leftChild->printTreeHelper(level + 2);
  rightChild->printTreeHelper(level + 2);
}

DataConstraint::ptr DCDisjunction::negationFree() const
{
  // nothing to do here
  DataConstraint::ptr ret(new DCDisjunction(leftChild->negationFree(),
					    rightChild->negationFree()));
  return ret;
}

DataConstraint::ptr DCDisjunction::canonicalForm() const
{
  DCClause::set_t newDisjunct;
  // compute canonical forms of subformulae
  DCClause::set_t left = shared_dynamic_cast<DCClause>(leftChild->canonicalForm())->getElements();
  DCClause::set_t right = shared_dynamic_cast<DCClause>(rightChild->canonicalForm())->getElements();
  // new clause is the union of the two computed clauses
  set_union(left.begin(), left.end(),
	    right.begin(), right.end(),
	    inserter(newDisjunct, newDisjunct.end()));
  DataConstraint::ptr ret(new DCClause(newDisjunct));
  return ret;
}

/////////////
// DCClause
/////////////

DCClause::DCClause()
  : DataConstraint()
{}

DCClause::DCClause(const DCClause::set_t &_elements)
  : DataConstraint(), elements(_elements)
{}

DCClause::DCClause(const DCClause &rhs)
  : DataConstraint(), elements(rhs.elements)
{}

const DCClause::set_t& DCClause::getElements() const
{
  return elements;
}

void DCClause::setElements(const DCClause::set_t &_elements)
{
  elements = _elements;
}

DataConstraint::ptr DCClause::clone() const
{
  DataConstraint::ptr ret(new DCClause(elements));
  return ret;
}

void DCClause::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "DCClause\n";
  for(set_t::iterator it = elements.begin(); it != elements.end(); ++it)
    it->printTreeHelper(level + 2);
}

void DCClause::reduceConstraints()
{
  set_t reducedElements;
  // reduce all contained dual clauses
  for(set_t::iterator it = elements.begin(); it != elements.end(); ++it) {
    DCDualClause cpy(*it);
    cpy.reduceConstraints();

    // if there are empty value-sets in the dual clause, it is equivalent to false
    // and we can forget it!
    DCDualClause::set_t::iterator jt = cpy.getElements().begin();
    for(;jt != cpy.getElements().end(); ++jt)
      if(jt->getValues().empty())
	break;
    if(jt == cpy.getElements().end())
      reducedElements.insert(*it);
  }
  elements = reducedElements;
}

/////////////
// DCDualClause
/////////////

DCDualClause::DCDualClause()
  : DataConstraint()
{}

DCDualClause::DCDualClause(const DCDualClause::set_t &_elements)
  : DataConstraint(), elements(_elements)
{}

DCDualClause::DCDualClause(const DCDualClause &rhs)
  : DataConstraint(), elements(rhs.elements)
{}

const DCDualClause::set_t& DCDualClause::getElements() const
{
  return elements;
}

void DCDualClause::setElements(const DCDualClause::set_t &_elements)
{
  elements = _elements;
}

DataConstraint::ptr DCDualClause::clone() const
{
  DataConstraint::ptr ret(new DCDualClause(elements));
  return ret;
}

void DCDualClause::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "DCDualClause\n";
  for(set_t::iterator it = elements.begin(); it != elements.end(); ++it)
    it->printTreeHelper(level + 2);
}

void DCDualClause::reduceConstraints()
{
  set_t reducedElements;
  // intersect all contained numerical constraints for a certain port
  for(set_t::iterator it = elements.begin(); it != elements.end(); ++it) {
    DCNumericalConstraint constraint(*it);
      
    for(set_t::iterator jt = elements.begin(); jt != elements.end(); ++jt) {
      if(jt->getVariable() == constraint.getVariable()) {
	set<int> intersection;
	set_intersection(constraint.getValues().begin(), constraint.getValues().end(),
			 jt->getValues().begin(), jt->getValues().end(),
			 inserter(intersection, intersection.end()));
	constraint.setValues(intersection);
      }
    }
    reducedElements.insert(constraint);
  }
  elements = reducedElements;
}
