/**************************************************************************
*   Copyright (C) 2008-2009 by Johannes Osterholzer                       *
*   s5892660@inf.tu-dresden.de                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef DATACONSTRAINT_H
#define DATACONSTRAINT_H

#include <boost/spirit/home/classic.hpp>
#include <boost/spirit/home/classic/core.hpp>
#include <boost/spirit/home/classic/tree/ast.hpp>
#include <boost/spirit/home/classic/tree/common.hpp>
#include <boost/spirit/home/classic/tree/tree_to_xml.hpp>
#include <boost/spirit/home/classic/iterator.hpp>
#include <boost/smart_ptr.hpp>
#include <set>
#include "sde_grammar.hpp"

using namespace std;
using namespace boost;
using namespace boost::spirit::classic;

typedef char const* iterator_t;
typedef tree_match<char const*> parse_tree_match_t;
typedef parse_tree_match_t::const_tree_iterator iter_t;

// base class for data constraints
class DataConstraint {
  friend class ScheduledDataExpression;
public:
  typedef shared_ptr<DataConstraint> ptr;

  virtual ~DataConstraint() {}
  
  virtual DataConstraint::ptr clone() const/* = 0*/ //not abstract because STL::set complains about this. :|
    {
      return DataConstraint::ptr();
    } 

  // kind of useful for pattern-matching...
  virtual bool isTrue() const {return false;}
  virtual bool isFalse() const {return false;}
  virtual bool isNumericalConstraint() const {return false;}
  virtual bool isPortEqualityConstraint() const {return false;}
  virtual bool isNegation() const {return false;}
  virtual bool isConjunction() const {return false;}
  virtual bool isDisjunction() const {return false;}
  virtual bool isClause() const {return false;}
  virtual bool isDualClause() const {return false;}

  // prints current data constraint as tree
  void printTree() const;
  virtual void printTreeHelper(int) const /*= 0*/ {} //dito

  // creates negation normal form of the current DC
  // also replaces negated atoms by atoms with complemented value-set
  // and reduces port-equalities to normal atoms
  virtual DataConstraint::ptr negationFree() const;
  // creates canonical form of the DC, basically just the known algorithm
  // for creation of the DNF
  virtual DataConstraint::ptr canonicalForm() const;
  // simplifies the dual clauses of the current DC so that for every port-set
  // there is at most one atom
  virtual void reduceConstraints();

  // helper function to handle abstract syntax trees of data constraints
  static DataConstraint::ptr evaluateAST(iter_t const&, const set<string>&);
};

class DCTrue : public DataConstraint {
public:
  virtual DataConstraint::ptr clone () const;
  virtual bool isTrue() const {return true;}

  virtual void printTreeHelper(int) const;
  virtual DataConstraint::ptr canonicalForm() const;
};

class DCFalse : public DataConstraint {
public:
  virtual DataConstraint::ptr clone () const;
  virtual bool isFalse() const {return true;}

  virtual void printTreeHelper(int) const;
  virtual DataConstraint::ptr canonicalForm() const;
};

class DCNumericalConstraint : public DataConstraint {
public:
  DCNumericalConstraint();
  DCNumericalConstraint(const string&, int);
  DCNumericalConstraint(const string&, const set<int>&);
  DCNumericalConstraint(const DCNumericalConstraint&);

  virtual bool isNumericalConstraint() const {return true;}

  const string& getVariable() const;
  const set<int>& getValues() const;
  void setVariable(const string&);
  void setValue(int);
  void setValues(const set<int>&);

  virtual DataConstraint::ptr clone() const;
  virtual void printTreeHelper(int) const;

  bool operator<(const DCNumericalConstraint &rhs) const
    {
      if(variable < rhs.getVariable() || (variable == rhs.getVariable() && 
					  values < rhs.getValues()))
	return true;
      return false;
    }

  virtual DataConstraint::ptr canonicalForm() const;

protected:
  string variable;
  set<int> values;
};

class DCPortEqualityConstraint : public DataConstraint {
public:
  DCPortEqualityConstraint();
  DCPortEqualityConstraint(const string&, const string&);
  DCPortEqualityConstraint(const DCPortEqualityConstraint&);

  virtual bool isPortEqualityConstraint() const {return true;}

  const string& getLHS() const;
  const string& getRHS() const;
  void setLHS(const string&);
  void setRHS(const string&);

  virtual DataConstraint::ptr clone() const;
  virtual void printTreeHelper(int) const;

  virtual DataConstraint::ptr negationFree() const;

protected:
  string lhs;
  string rhs;
};

class DCNegation : public DataConstraint {
public:
  DCNegation();
  DCNegation(DataConstraint::ptr);
  DCNegation(const DCNegation&);

  virtual bool isNegation() const {return true;}

  const DataConstraint& getChild() const;
  void setChild(DataConstraint::ptr);

  virtual DataConstraint::ptr clone() const;
  virtual void printTreeHelper(int) const;

  virtual DataConstraint::ptr negationFree() const;

protected:
  DataConstraint::ptr child;
};

class DCConjunction : public DataConstraint {
public:
  DCConjunction();
  DCConjunction(DataConstraint::ptr, DataConstraint::ptr);
  DCConjunction(const DCConjunction&);

  virtual bool isConjunction() const {return true;}

  const DataConstraint& getLeftChild() const;
  const DataConstraint& getRightChild() const;
  void setLeftChild(DataConstraint::ptr);
  void setRightChild(DataConstraint::ptr);

  virtual DataConstraint::ptr clone() const;
  virtual void printTreeHelper(int) const;

  virtual DataConstraint::ptr negationFree() const;
  virtual DataConstraint::ptr canonicalForm() const;

protected:
  DataConstraint::ptr leftChild, rightChild;
};

class DCDisjunction : public DataConstraint {
public:
  DCDisjunction();
  DCDisjunction(DataConstraint::ptr, DataConstraint::ptr);
  DCDisjunction(const DCDisjunction&);

  virtual bool isDisjunction() const {return true;}

  const DataConstraint& getLeftChild() const;
  const DataConstraint& getRightChild() const;
  void setLeftChild(DataConstraint::ptr);
  void setRightChild(DataConstraint::ptr);

  virtual DataConstraint::ptr clone() const;
  virtual void printTreeHelper(int) const;

  virtual DataConstraint::ptr negationFree() const;
  virtual DataConstraint::ptr canonicalForm() const;

protected:
  DataConstraint::ptr leftChild, rightChild;

};

class DCDualClause : public DataConstraint {
public:
  typedef set<DCNumericalConstraint> set_t;

  DCDualClause();
  DCDualClause(const set_t&);
  DCDualClause(const DCDualClause&);

  virtual bool isDualClause() const {return true;}

  const set_t& getElements() const;
  void setElements(const set_t&);

  virtual DataConstraint::ptr clone() const;
  virtual void printTreeHelper(int) const;
  virtual void reduceConstraints();

  bool operator<(const DCDualClause &rhs) const
    {
      return elements < rhs.getElements();
    }

protected:
  set_t elements;
};

class DCClause : public DataConstraint {
public:
  typedef set<DCDualClause> set_t;

  DCClause();
  DCClause(const set_t&);
  DCClause(const DCClause&);

  virtual bool isClause() const {return true;}

  const set_t& getElements() const;
  void setElements(const set_t&);

  virtual DataConstraint::ptr clone() const;
  virtual void printTreeHelper(int) const;
  virtual void reduceConstraints();

protected:
  set_t elements;
};

#endif
