/**************************************************************************
*   Copyright (C) 2008-2009 by Johannes Osterholzer                       *
*   s5892660@inf.tu-dresden.de                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include <map>
#include <iostream>
#include <algorithm>
#include <boost/lambda/if.hpp>
#include <cctype>
#include <boost/lexical_cast.hpp>
#include "scheduleddataexpression.hpp"
#include "sde_grammar.hpp"
#include "helper.hpp"

//#define BOOST_SPIRIT_DUMP_PARSETREE_AS_XML

using boost::lexical_cast;

ScheduledDataExpression::ScheduledDataExpression()
{
  inCanonicalForm = false;
}

// Static members
pair<int,int> ScheduledDataExpression::dataDomain(0,1);
vector<string> ScheduledDataExpression::sources;
vector<string> ScheduledDataExpression::sinks;

vector<string> ScheduledDataExpression::getSources()
{
  return sources;
}

void ScheduledDataExpression::setSources(vector<string> _sources)
{
  // sanity check
  for(vector<string>::iterator it = _sources.begin(); it != _sources.end(); ++it)
    if(find(sinks.begin(), sinks.end(), *it) != sinks.end())
      throw "Sources and sinks must be disjoint sets";
  sources = _sources;
}

vector<string> ScheduledDataExpression::getSinks()
{
  return sinks;
}

void ScheduledDataExpression::setSinks(vector<string> _sinks)
{
  // sanity check
  for(vector<string>::iterator it = _sinks.begin(); it != _sinks.end(); ++it)
    if(find(sources.begin(), sources.end(), *it) != sources.end())
      throw "Sources and sinks must be disjoint sets";
  sinks = _sinks;
}

pair<int,int> ScheduledDataExpression::getDataDomain()
{
  return dataDomain;
}

void ScheduledDataExpression::setDataDomain(const pair<int,int>& _dataDomain)
{
  // sanity check
  if(_dataDomain.first >= _dataDomain.second)
    throw "Invalid data domain";
  dataDomain = _dataDomain;
}

// Static parsing functions

void ScheduledDataExpression::printTree() const
{
  printTreeHelper(0);
}

ScheduledDataExpression::ptr ScheduledDataExpression::readFromString(const string &str)
{
  sdexp sd;

  tree_parse_info<> info = ast_parse(str.c_str(), sd, space_p);
  if(info.full) {
#if defined(BOOST_SPIRIT_DUMP_PARSETREE_AS_XML)
    // dump parse tree as XML
    std::map<parser_id, std::string> rule_names;
    rule_names[sdexp::dataID] = "data";
    rule_names[sdexp::factorID] = "factor";
    rule_names[sdexp::termID] = "term";
    rule_names[sdexp::exprID] = "expr";
    rule_names[sdexp::guardID] = "guard";
    rule_names[sdexp::guard_conjID] = "guard_conj";
    rule_names[sdexp::guard_litID] = "guard_lit";
    rule_names[sdexp::portnameID] = "portname";
    rule_names[sdexp::portnamesID] = "portnames";
    rule_names[sdexp::starsID] = "stars";
    rule_names[sdexp::rhsvalueID] = "rhsvalue";
    rule_names[sdexp::valuesID] = "values";

    tree_to_xml(cout, info.trees, "blah", rule_names);
#endif
    cerr << "Parsed successfully!\n";
    return evaluateAST(info.trees.begin());
  } else {
    cerr << "Not parsed successfully. :(\n";
    return ScheduledDataExpression::ptr();
  }
}

ScheduledDataExpression::ptr ScheduledDataExpression::evaluateAST(const iter_t &i)
{
//   cerr << "In SDE::evaluateAST. i->value = " <<
//     string(i->value.begin(), i->value.end()) <<
//     " i->children.size() = " << i->children.size() << endl;
  node_val_data<>::const_iterator_t begin = i->value.begin();
  node_val_data<>::const_iterator_t end = i->value.end();
  
  // test for all possible parse node types
  if(i->value.id() == sdexp::dataID) {
    assert(i->children.size() == 5);
    set<string> ports = getPortnames(i->children.begin() + 1);
    DataConstraint::ptr tmp(DataConstraint::evaluateAST(i->children.begin() + 3, ports));
    ScheduledDataExpression::ptr ret(new SDEAtom(ports, tmp));
    return ret;
  } else if(i->value.id() == sdexp::factorID) {
    assert(*eatSpaces(begin, end) == 'e'); // can only be "epsilon"
    ScheduledDataExpression::ptr ret(new SDEEpsilon);
    return ret;
  } else if(i->value.id() == sdexp::starsID) {
    assert(i->children.size() == 2);
    if(*eatSpaces((i->children.end()-1)->value.begin(), (i->children.end()-1)->value.end()) == '^') {
      ScheduledDataExpression::ptr ret(new SDEManyOrInf(evaluateAST(i->children.begin())));
      return ret;
    } else if(*eatSpaces((i->children.end()-1)->value.begin(), (i->children.end()-1)->value.end()) == '~') {
      ScheduledDataExpression::ptr ret(new SDEInf(evaluateAST(i->children.begin())));
      return ret;
    } else
      assert(0);
  } else if(i->value.id() == sdexp::termID) {
    assert(*eatSpaces(begin, end) == ';');
    ScheduledDataExpression::ptr ret(new SDEConcatenation(evaluateAST(i->children.begin()),
							  evaluateAST(i->children.begin() + 1)));
    return ret;
  } else if(i->value.id() == sdexp::exprID) {
    assert(*eatSpaces(begin, end) == '|');
    ScheduledDataExpression::ptr ret(new SDEAlternative(evaluateAST(i->children.begin()),
							evaluateAST(i->children.begin() + 1)));
    return ret;
  } else
    assert(0);
}

set<string> ScheduledDataExpression::getPortnames(const iter_t &i)
{
//   cerr << "In getPortnames. i->value = " <<
//     string(i->value.begin(), i->value.end()) <<
//     " i->children.size() = " << i->children.size() << endl;
  
  set<string> ret;

  assert(i->value.id() == sdexp::portnamesID);
  // for(int j = 0; i->value.begin()+ j < i->value.end(); j++)
  //   cerr << *(i->value.begin() + j) << endl;
  //assert(*i->value.begin() == '{');
  assert(i->children.size() >= 2);
  if(i->children.size() == 2) {
    return ret;
  } else {
    for(unsigned int j = 1; j < i->children.size(); j += 2){
      iter_t ch = i->children.begin() + j;
      assert(ch->value.id() == sdexp::portnameID);
      assert(ch->children.size() == 0);
      
      char port = *eatSpaces(ch->value.begin(), ch->value.end());
      assert(port >= 'A' && port <= 'Z');
      ret.insert(string(1, port));
    }
    return ret;
  }
}

string ScheduledDataExpression::toREOCircuit() const
{
  // the implemented algorithm relies on the SD-expressions being in canonical form
  if(!inCanonicalForm)
    return canonicalForm()->toREOCircuit();

  string circuit; //the returned circuit
  unsigned int label = 0;

  vector<string> sourceOrder = sourceArgumentOrder();
  vector<string> sinkOrder = sinkArgumentOrder();

  // We need the following modules for the construction in the end.
  circuit += 
    "/* Reo circuit synthesized from scheduled data expression (insert name here)*/\n\n"
    "#include \"builtin\"\n\n"
    "MODULE Init {\n"
    "  out: a_init;\n"
    "  var: enum {FIRST, LAST} state := FIRST;\n\n"

    "  state==FIRST -[ {a_init} & #a_init == 0 ]-> state:=LAST;\n"
    "}\n\n";

  circuit +=
    "MODULE Stop {\n"
    "  in: b_stop;\n"
    "  var: enum {FIRST, LAST} state := FIRST;\n\n"

    "  state == FIRST -[ {b_stop} ]-> state:=LAST;\n"
    "}\n\n";

  circuit +=
    "MODULE Exrouter {\n"
    "  in: A;\n"
    "  out: B;\n"
    "  out: C;\n"
    "  -[ {A,B} & #A==#B ]->;\n"
    "  -[ {A,C} & #A==#C ]->;\n"
    "}\n\n";

  circuit +=
    "MODULE LossySync {\n"
    "  in: A;\n"
    "  out: B;\n"
    "  -[ {A,B} & #A==#B ]->;\n"
    "  -[ {A} ]->;\n"
    "}\n\n";

  circuit +=
    "MODULE PseudoChannel {\n"
    "  in: A;\n"
    "  out: B;\n"
    "  -[ {A,B} & #A==0 & #A!=0 ]->;\n"
    "}\n\n";
  
  // create the circuit for the expression (composition of the corr. circuits of its subexpressions)
  string name = toREOCircuitHelper(circuit, label);

  // now include the created circuit into the main circuit
  circuit += "CIRCUIT main {\n"
    "  component = new " + name + ";\n"
    "  init = new Init;\n"
    "  leftFifo = new FIFO1;\n"
    "  stop = new Stop;\n"
    "  rightFifo = new FIFO1;\n\n"
    "  leftSync = new SYNC;\n"
    "  rightSync = new SYNC;\n\n"

    "  join(init.sink, leftFifo.source);\n"
    "  join(leftFifo.sink, leftSync.source);\n"
    "  join(leftSync.sink, component.source[0]);\n"
    "  join(component.sink[0], rightSync.source);\n"
    "  join(rightSync.sink, rightFifo.source);\n"
    "  join(rightFifo.sink, stop.source);\n\n";

  // export the source and sink nodes of the expression
  for(unsigned int i = 0; i < sourceOrder.size(); ++i)
    circuit += "  source[" + lexical_cast<string>(i) + "] = component.source[" + 
      lexical_cast<string>(i+1) + "];\n";

  for(unsigned int i = 0; i < sinkOrder.size(); ++i)
    circuit += "  sink[" + lexical_cast<string>(i) +"] = component.sink[" + 
      lexical_cast<string>(i+1) + "];\n";

  //plug the exported port-copies together to a mixed node (port-copies must be shared between all subcircuits)
  for(unsigned int i; i < sourceOrder.size(); ++i) {
    circuit += "  join(component.source[" + lexical_cast<string>(sourceOrder.size() + i + 1) + "], component.sink["
      + lexical_cast<string>(sinkOrder.size() + i + 1) + "]);\n";
  }

  circuit += "}\n";

  return circuit;
}
  
  

//////////////
// SDEAtom
//////////////

SDEAtom::SDEAtom()
  : ScheduledDataExpression()
{}

SDEAtom::SDEAtom(set<string> _portset, DataConstraint::ptr _dataConstraint)
  : ScheduledDataExpression(), portset(_portset), dataConstraint(_dataConstraint)
{}

SDEAtom::SDEAtom(const SDEAtom &rhs)
  : ScheduledDataExpression(), portset(rhs.portset), dataConstraint(rhs.dataConstraint->clone())
{}

pair<string, string> SDEAtom::producerFromConstraint(const set<int> &constraint, unsigned int &label) const
{
  string name = string("Producer") + lexical_cast<string>(label);
  string module = string("MODULE ") + name + "{\n"
    "  in: A;\n"
    "  out: B;\n";
  // add a transition for each value to be produced
  for(set<int>::iterator it = constraint.begin(); it != constraint.end(); ++it)
    module +=
      "  -[{A,B} & #B==" + lexical_cast<string>(*it) + "]->;\n";

  module += "}\n\n";
  label++;
  return pair<string, string>(name, module);
}

pair<string, string> SDEAtom::filterFromConstraint(const set<int> &constraint, unsigned int &label) const
{
  string name = "Filter" + lexical_cast<string>(label);
  string module = "MODULE " + name + "{\n"
    "  in: A;\n"
    "  out: B;\n";
  // let accepted values pass, filter out the others
  for(int i = dataDomain.first; i <= dataDomain.second; ++i) {
    if(find(constraint.begin(), constraint.end(), i) != constraint.end())
      module += "  -[{A,B} & #A==#B & #A==" + lexical_cast<string>(i) + "]->;\n";
    else
      module += "  -[{A} & #A==" + lexical_cast<string>(i) + "]->;\n";
  }
  module += "}\n\n";

  label++;
  return pair<string, string>(name, module);
}    

string SDEAtom::toREOCircuitHelper(string &circuit, unsigned int &label) const
{
  string name = string("Atom") + lexical_cast<string>(label++);
  DCClause::set_t elems = shared_dynamic_cast<DCClause>(dataConstraint)->getElements();

  // at this point the atoms should be in canonical form
  if(elems.size() > 1) {
    throw "applying toREOCircuitHelper() to non-canonical atom";
    // if dc = false, we're nearly done, just allow no dataflow
  } else if(elems.size() == 0) {
    circuit += 
      "CIRCUIT " + name + "{\n"
      "  leftFifo = new FIFO1;\n"
      "  leftSync = new SYNC;\n"
      "  rightSync = new SYNC;\n"
      "  rightFifo = new FIFO1;\n"
      "  join(leftFifo.sink, leftSync.source);\n"
      "  join(leftSync.sink, rightSync.source);\n"
      "  join(rightSync.sink, rightFifo.source);\n\n"

      "  source[0] = leftFifo.source;\n"
      "  sink[0] = rightFifo.sink;\n\n"

      "  // Sources:\n";

    vector<string> sourceOrder = sourceArgumentOrder();
    for(unsigned int i = 0; i < sourceOrder.size(); i++) {
      string sync = "sync" + sourceOrder[i];
      string drain1 = "drain" + sourceOrder[i] + "1";
      string drain2 = "drain" + sourceOrder[i] + "2";
      string lossy = "lossy" + sourceOrder[i];
      
      circuit += 
	"  // " + sourceOrder[i] + ":\n"
	"  " + sync + " = new SYNC;\n"
	"  " + drain1 + " = new SYNC_DRAIN;\n"
	"  " + drain2 + " = new SYNC_DRAIN;\n"
	"  " + lossy + " = new LossySync;\n"
	"  join(" + sync + ".source, rightSync.source);\n"
	"  join(" + sync + ".source, " + drain2 + ".source[1]);\n"
	"  join(" + lossy + ".source, " + drain1 + ".source[1]);\n"
	"  source[" + lexical_cast<string>(i+1) + "] = " + drain1 + ".source[1];\n";
    }
    // export source-part of port-copy
    for(unsigned int i = 0; i < sourceOrder.size(); i++) {
      string drain = "drain" + sourceOrder[i] + "1";

      circuit +=
 	"  source[" + lexical_cast<string>(i+1+sourceOrder.size()) + "] = " + 
 	drain + ".source[0];\n";
    }

    circuit += "\n// Sinks:\n";
	
    vector<string> sinkOrder = sinkArgumentOrder();
    for(unsigned int i = 0; i < sinkOrder.size(); i++) {
      string sync = "sync" + sinkOrder[i];
      string pseudo = "pseudo" + sinkOrder[i];
      circuit +=
	"  // " + sinkOrder[i] + ":\n"
	"  " + sync + " = new SYNC;\n"
	"  " + pseudo + " = new PseudoChannel;\n"
	"  join(rightSync.source, " + sync + ".source);\n"
	"  join(" + sync + ".sink, " + pseudo + ".source);\n"
	"  sink[" + lexical_cast<string>(i+1) + "] = " + pseudo  + ".sink;\n";
    }
    // export sink-part of port-copy
    for(unsigned int i = 0; i < sourceOrder.size(); i++) {
      string sync = "sync" + sourceOrder[i];
      circuit +=
 	"  sink[" + lexical_cast<string>(i+1+sinkOrder.size()) + "] = " + 
 	sync  + ".sink;\n";
    }
    circuit += "}\n\n";
    // this isn't so simple, we'll have to create the corresponding producers/filters.
  } else {
    string tmpCircuit; // created since we will probably have to prepend producers/filters to the circuit
    DCDualClause::set_t numConstraints = elems.begin()->getElements();

    tmpCircuit += 
      "CIRCUIT " + name + "{\n"
      "  leftFifo = new FIFO1;\n"
      "  leftSync = new SYNC;\n"
      "  rightSync = new SYNC;\n"
      "  rightFifo = new FIFO1;\n"
      "  join(leftFifo.sink, leftSync.source);\n"
      "  join(leftSync.sink, rightSync.source);\n"
      "  join(rightSync.sink, rightFifo.source);\n\n"

      "  source[0] = leftFifo.source;\n"
      "  sink[0] = rightFifo.sink;\n\n"

      "  // Sources:\n";

    // handle each accessed source port
    vector<string> sourceOrder = sourceArgumentOrder();
    for(unsigned int i = 0; i < sourceOrder.size(); i++) {
      string sync = "sync" + sourceOrder[i];
      string drain1 = "drain" + sourceOrder[i] + "1";
      string drain2 = "drain" + sourceOrder[i] + "2";
      string lossy = "lossy" + sourceOrder[i];
      string filter = "filter" + sourceOrder[i];
      
      tmpCircuit += 
	"  //" + sourceOrder[i] + ":\n"
	"  " + sync + " = new SYNC;\n"
	"  " + drain1 + " = new SYNC_DRAIN;\n"
	"  " + drain2 + " = new SYNC_DRAIN;\n"
	"  " + lossy + " = new LossySync;\n";

      // if there is a constraint for the current source, we create a filter, else just a sync
      // (all values allowed)
      DCDualClause::set_t::iterator it = numConstraints.begin();
      for(;it != numConstraints.end(); ++it) {
	// if I don't copy the variable, the program crashes.
	// I just love C++.
	string tmp = it->getVariable();
	if(tmp == sourceOrder[i])
	  break;
      }

      if(it != numConstraints.end()) {
	pair<string, string> filterPair = filterFromConstraint(it->getValues(), label);
	circuit += filterPair.second;
	tmpCircuit +=
	  "  " + filter +" = new " + filterPair.first + ";\n";
      } else {
	tmpCircuit +=
	  "  " + filter +" = new SYNC;\n";
      }

      tmpCircuit += 
	"  join(" + filter + ".sink, " + lossy + ".source);\n"
	"  join(" + filter + ".source, " + drain2 + ".source[1]);\n"
	"  join(" + sync + ".source, rightSync.source);\n"
	"  join(" + drain1 + ".source[0], rightSync.source);\n"
	"  join(" + drain1 + ".source[1], " + lossy + ".sink);\n";
    }
    tmpCircuit += "\n";

    // export the sources
    for(unsigned int i = 0; i < sourceOrder.size(); i++)
      tmpCircuit +=
	"  source[" + lexical_cast<string>(i+1) + "] = filter" + sourceOrder[i]  + ".source;\n";
    // export source-part of port copy
    for(unsigned int i = 0; i < sourceOrder.size(); i++) {
      string drain = "drain" + sourceOrder[i] + "2";
      tmpCircuit +=
	"  source[" + lexical_cast<string>(i+1+sourceOrder.size()) + "] = " + 
	drain  + ".source[0];\n";
    }
    
    tmpCircuit += "\n  // Sinks:\n";
    
    // handle each accessed sink port
    vector<string> sinkOrder = sinkArgumentOrder();
    for(unsigned int i = 0; i < sinkOrder.size(); i++) {
      string sync = "sync" + sinkOrder[i];
      string producer = "producer" + sinkOrder[i];
      tmpCircuit +=
	"  // " + sinkOrder[i] + ":\n"
	"  " + sync + " = new SYNC;\n";
      
      // if there is a constraint for the current sink, we create the corr. producer, else 
      // an omnipotent producer.
      DCDualClause::set_t::iterator it = numConstraints.begin();
      for(; it != numConstraints.end(); ++it) {
	string tmp = it->getVariable();
	if(tmp == sinkOrder[i])
	  break;
      }
	
      if(it != numConstraints.end()) {
	pair<string, string> producerPair = producerFromConstraint(it->getValues(), label);
	circuit += producerPair.second;
	tmpCircuit +=
	  "  " + producer + " = new " + producerPair.first + ";\n";
      } else {
	// we need to construct a producer for all possible values.
	set<int> vals;
	for(int j = dataDomain.first; j <= dataDomain.second; ++j)
	  vals.insert(j);
	pair<string, string> producerPair = producerFromConstraint(vals, label);
	circuit += producerPair.second;
	tmpCircuit +=
	  "  " + producer + " = new " + producerPair.first + ";\n";
      }
      tmpCircuit += 
	"  join(rightSync.source, " + sync + ".source);\n"
	"  join(" + sync + ".sink, " + producer + ".source);\n";
    }
    tmpCircuit += "\n";

    // export the sinks
    for(unsigned int i = 0; i < sinkOrder.size(); i++)
      tmpCircuit += "  sink[" + lexical_cast<string>(i+1) + "] = producer" + sinkOrder[i] + ".sink;\n";
    // export sink-part of the port-copy
    for(unsigned int i = 0; i < sourceOrder.size(); i++) {
      string sync = "sync" + sourceOrder[i];
      tmpCircuit +=
	"  sink[" + lexical_cast<string>(i+1+sinkOrder.size()) + "] = " + 
	sync  + ".sink;\n";
    }
  
    tmpCircuit += "}\n\n";
    circuit += tmpCircuit;
  }
  return name;
}

ScheduledDataExpression::ptr SDEAtom::canonicalForm() const
{
  // create the negation normal form of the atom's data constraint and convert it to canonical form
  shared_ptr<DCClause> canonicalConstraint = shared_dynamic_cast<DCClause>(
    dataConstraint->negationFree()->canonicalForm());
  if(!canonicalConstraint)
    throw "unexpected error within canonicalForm()";

  // the constraint is reduced so that for every port in the clause there is at most one numerical constraint
  canonicalConstraint->reduceConstraints();

  // if the resulting constraint is a disjunction of >1 dual clauses, split it up to several new atoms
  ScheduledDataExpression::ptr replacement;
  if(canonicalConstraint->getElements().size() <= 1) {
    replacement = shared_ptr<SDEAtom>(new SDEAtom);
    replacement->inCanonicalForm = true;
    shared_dynamic_cast<SDEAtom>(replacement)->setDataConstraint(canonicalConstraint);
    shared_dynamic_cast<SDEAtom>(replacement)->setPortset(portset);
    return replacement;
  } else {
    DCClause::set_t::iterator it = canonicalConstraint->getElements().begin();
    DCClause::set_t leftElements;
    leftElements.insert(*(it++));
    DataConstraint::ptr ldc(new DCClause(leftElements));
    DCClause::set_t rightElements;
    rightElements.insert(*(it++));
    DataConstraint::ptr rdc(new DCClause(rightElements));
    replacement = ScheduledDataExpression::ptr(new SDEAlternative(
						 ScheduledDataExpression::ptr(new SDEAtom(portset, ldc)), 
						 ScheduledDataExpression::ptr(new SDEAtom(portset, rdc))));
    while(it != canonicalConstraint->getElements().end()) {
      DCClause::set_t elements;
      elements.insert(*(it++));
      DataConstraint::ptr dc(new DCClause(elements));
      replacement = ScheduledDataExpression::ptr(new SDEAlternative(
						   ScheduledDataExpression::ptr(new SDEAtom(portset, dc)),
						   replacement));
    }

    replacement->inCanonicalForm = true;
    return replacement;
  }
}

vector<string> SDEAtom::sinkArgumentOrder() const
{
  vector<string> ret;
  for(set<string>::iterator it = portset.begin(); it != portset.end(); ++it)
    if(find(sinks.begin(), sinks.end(), *it) != sinks.end())
      ret.push_back(*it);
  return ret;
}

vector<string> SDEAtom::sourceArgumentOrder() const
{
  vector<string> ret;
  for(set<string>::iterator it = portset.begin(); it != portset.end(); ++it)
    if(find(sources.begin(), sources.end(), *it) != sources.end())
      ret.push_back(*it);
  return ret;
}


const set<string>& SDEAtom::getPortset() const
{
  return portset;
}

void SDEAtom::setPortset(const set<string> &_portset)
{
  portset = _portset;
}

const DataConstraint& SDEAtom::getDataConstraint() const
{
  return *dataConstraint;
}

void SDEAtom::setDataConstraint(DataConstraint::ptr _dataConstraint)
{
  dataConstraint = _dataConstraint;
}

ScheduledDataExpression::ptr SDEAtom::clone() const
{
  ScheduledDataExpression::ptr ret(new SDEAtom(portset, dataConstraint->clone()));
  return ret;
}

void SDEAtom::printTreeHelper(int level) const
{
  string ports;
  for(set<string>::iterator it = portset.begin(); it != portset.end();) {
    ports += *it;
    if(++it != portset.end())
      ports += ", ";
  }
  cerr << string(level, ' ') << "SDEAtom\n";
  cerr << string(level, ' ') << "portnames = {" << ports << "}\n";
  dataConstraint->printTreeHelper(level + 2);
}

////////////////
// SDEEpsilon
////////////////

SDEEpsilon::SDEEpsilon()
  : ScheduledDataExpression()
{}

string SDEEpsilon::toREOCircuitHelper(string &circuit, unsigned int &label) const
{
  circuit +=
    "CIRCUIT epsilon" + lexical_cast<string>(label) + "{\n"
    "  fifo = new FIFO1;\n"
    "  source[0] = fifo.source;\n"
    "  sink[0] = fifo.sink;\n"
    "}\n\n";

  label++;
  return string("epsilon") + lexical_cast<string>(label - 1);
}

vector<string> SDEEpsilon::sinkArgumentOrder() const
{
  vector<string> ret;
  return ret;
}

vector<string> SDEEpsilon::sourceArgumentOrder() const
{
  vector<string> ret;
  return ret;
}

ScheduledDataExpression::ptr SDEEpsilon::clone() const
{
  ScheduledDataExpression::ptr ret(new SDEEpsilon);
  return ret;
}

void SDEEpsilon::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "SDEEpsilon\n";
}

/////////////////////
// SDEConcatenation
/////////////////////

SDEConcatenation::SDEConcatenation()
  : ScheduledDataExpression()
{}

SDEConcatenation::SDEConcatenation(ScheduledDataExpression::ptr _left, ScheduledDataExpression::ptr _right)
  : ScheduledDataExpression(), leftChild(_left), rightChild(_right)
{}

SDEConcatenation::SDEConcatenation(const SDEConcatenation &rhs)
  : ScheduledDataExpression(), leftChild(rhs.leftChild->clone()), rightChild(rhs.rightChild->clone())
{}

string SDEConcatenation::toREOCircuitHelper(string &circuit, unsigned int& label) const
{
  string name = "Concat" + lexical_cast<string>(label++);
  string leftName = leftChild->toREOCircuitHelper(circuit, label);
  string rightName = rightChild->toREOCircuitHelper(circuit, label);

  vector<string> lSourceOrder = leftChild->sourceArgumentOrder();
  vector<string> lSinkOrder = leftChild->sinkArgumentOrder();
  vector<string> rSourceOrder = rightChild->sourceArgumentOrder();
  vector<string> rSinkOrder = rightChild->sinkArgumentOrder();
  vector<string> sourceOrder = sourceArgumentOrder();
  vector<string> sinkOrder = sinkArgumentOrder();

  // create the circuits for the subexpressions
  circuit +=
    "CIRCUIT " + name + "{\n"
    "  left = new " + leftName + ";\n"
    "  right = new " + rightName + ";\n\n";

  // join corresponding exported nodes of the subcircuits
  for(unsigned int i = 0; i < lSourceOrder.size(); ++i)
    for(unsigned int j = 0; j < rSourceOrder.size(); ++j)
      if(lSourceOrder[i] == rSourceOrder[j])
	circuit += 
	  "  join(left.source[" + lexical_cast<string>(i+1)
	  + "], right.source[" + lexical_cast<string>(j+1) + "]);\n"
	  "  join(left.source[" + lexical_cast<string>(lSourceOrder.size()+i+1)
	  + "], right.source[" + lexical_cast<string>(rSourceOrder.size()+j+1) + "]);\n"
	  "  join(left.sink[" + lexical_cast<string>(lSinkOrder.size()+i+1)
	  + "], right.sink[" + lexical_cast<string>(rSinkOrder.size()+j+1) + "]);\n";

  for(unsigned int i = 0; i < lSinkOrder.size(); ++i)
    for(unsigned int j = 0; j < rSinkOrder.size(); ++j)
      if(lSinkOrder[i] == rSinkOrder[j])
	circuit += 
	  "  join(left.sink[" + lexical_cast<string>(i+1)
	  + "], right.sink[" + lexical_cast<string>(j+1) + "]);\n";

  circuit +=
    "\n  leftFifo = new FIFO1;\n"
    "  leftSync = new SYNC;\n"
    "  middleSync = new SYNC;\n"
    "  rightSync = new SYNC;\n"
    "  rightFifo = new FIFO1;\n"
    "  join(leftFifo.sink, leftSync.source);\n"
    "  join(leftSync.sink, left.source[0]);\n"
    "  join(left.sink[0], middleSync.source);\n"
    "  join(middleSync.sink, right.source[0]);\n"
    "  join(right.sink[0], rightSync.source);\n"
    "  join(rightSync.sink, rightFifo.source);\n\n"

    "  source[0] = leftFifo.source;\n"
    "  sink[0] = rightFifo.sink;\n\n";

  // in order to export the sources/sinks it has to be known from which subcircuit they come from
  for(unsigned int i = 0; i < sourceOrder.size(); ++i) {
    unsigned int j;
    for(j = 0; j < lSourceOrder.size(); ++j)
      if(sourceOrder[i] == lSourceOrder[j]) break;
    if(j != lSourceOrder.size()) {
      circuit +=
	"  source[" + lexical_cast<string>(i+1) + "] = left.source[" + lexical_cast<string>(j+1) + "];\n";
    } else {
      for(j = 0; j < rSourceOrder.size(); ++j)
	if(sourceOrder[i] == rSourceOrder[j]) break;
      if(j != rSourceOrder.size())
	circuit +=
	  "  source[" + lexical_cast<string>(i+1) + "] = right.source[" + lexical_cast<string>(j+1) + "];\n";
      else
	assert(0);
    }
  }
  for(unsigned int i = 0; i < sinkOrder.size(); ++i) {
    unsigned int j;
    for(j = 0; j < lSinkOrder.size(); ++j)
      if(sinkOrder[i] == lSinkOrder[j]) break;
    if(j != lSinkOrder.size()) {
      circuit +=
	"  sink[" + lexical_cast<string>(i+1) + "] = left.sink[" + lexical_cast<string>(j+1) + "];\n";
    } else {
      for(j = 0; j < rSinkOrder.size(); ++j)
	if(sinkOrder[i] == rSinkOrder[j]) break;
      if(j != rSinkOrder.size())
	circuit +=
	  "  sink[" + lexical_cast<string>(i+1) + "] = right.sink[" + lexical_cast<string>(j+1) + "];\n";
      else
	assert(0);
    }
  }
  for(unsigned int i = 0; i < sourceOrder.size(); ++i) {
    unsigned int j;
    for(j = 0; j < lSourceOrder.size(); ++j)
      if(sourceOrder[i] == lSourceOrder[j]) break;
    if(j != lSourceOrder.size()) {
      circuit +=
	"  source[" + lexical_cast<string>(sourceOrder.size()+i+1) + "] = left.source[" 
	+ lexical_cast<string>(lSourceOrder.size()+j+1) + "];\n"
      	"  sink[" + lexical_cast<string>(sinkOrder.size()+i+1) + "] = left.sink[" 
	+ lexical_cast<string>(lSinkOrder.size()+j+1) + "];\n";
    } else {
      for(j = 0; j < rSourceOrder.size(); ++j)
	if(sourceOrder[i] == rSourceOrder[j]) break;
      if(j != rSourceOrder.size())
	circuit +=
	  "  source[" + lexical_cast<string>(sourceOrder.size()+i+1) + "] = right.source[" 
	  + lexical_cast<string>(rSourceOrder.size()+j+1) + "];\n"
	  "  sink[" + lexical_cast<string>(sinkOrder.size()+i+1) + "] = right.sink[" 
	  + lexical_cast<string>(rSinkOrder.size()+j+1) + "];\n";
      else
	assert(0);
    }
  }

  circuit += "}\n\n";
  return name;
}

vector<string> SDEConcatenation::sinkArgumentOrder() const
{
  vector<string> ret;

  vector<string> left = leftChild->sinkArgumentOrder();
  vector<string> right = rightChild->sinkArgumentOrder();

  set_union(left.begin(), left.end(),
	    right.begin(), right.end(),
	    inserter(ret, ret.end()));

  return ret;
}

vector<string> SDEConcatenation::sourceArgumentOrder() const
{
  vector<string> ret;

  vector<string> left = leftChild->sourceArgumentOrder();
  vector<string> right = rightChild->sourceArgumentOrder();

  set_union(left.begin(), left.end(),
	    right.begin(), right.end(),
	    inserter(ret, ret.end()));

  return ret;
}

const ScheduledDataExpression& SDEConcatenation::getLeftChild() const
{
  return *leftChild;
}

const ScheduledDataExpression& SDEConcatenation::getRightChild() const
{
  return *rightChild;
}

void SDEConcatenation::setLeftChild(ScheduledDataExpression::ptr _left)
{
  leftChild = _left;
}

void SDEConcatenation::setRightChild(ScheduledDataExpression::ptr _right)
{
  rightChild = _right;
}

ScheduledDataExpression::ptr SDEConcatenation::clone() const
{
  ScheduledDataExpression::ptr ret(new SDEConcatenation(leftChild->clone(), rightChild->clone()));
  return ret;
}

void SDEConcatenation::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "SDEConcatenation\n";
  leftChild->printTreeHelper(level + 2);
  rightChild-> printTreeHelper(level + 2);
}

/////////////////////
// SDEAlternative
/////////////////////

SDEAlternative::SDEAlternative()
  : ScheduledDataExpression()
{}

SDEAlternative::SDEAlternative(ScheduledDataExpression::ptr _left, ScheduledDataExpression::ptr _right)
  : ScheduledDataExpression(), leftChild(_left), rightChild(_right)
{}

SDEAlternative::SDEAlternative(const SDEAlternative &rhs)
  : ScheduledDataExpression(), leftChild(rhs.leftChild->clone()), rightChild(rhs.rightChild->clone())
{}

string SDEAlternative::toREOCircuitHelper(string &circuit, unsigned int &label) const
{
  string name = "Alternative" + lexical_cast<string>(label++);
  string leftName = leftChild->toREOCircuitHelper(circuit, label);
  string rightName = rightChild->toREOCircuitHelper(circuit, label);

  vector<string> lSourceOrder = leftChild->sourceArgumentOrder();
  vector<string> lSinkOrder = leftChild->sinkArgumentOrder();
  vector<string> rSourceOrder = rightChild->sourceArgumentOrder();
  vector<string> rSinkOrder = rightChild->sinkArgumentOrder();
  vector<string> sourceOrder = sourceArgumentOrder();
  vector<string> sinkOrder = sinkArgumentOrder();

  // create circuits for subexpressions
  circuit +=
    "CIRCUIT " + name + "{\n"
    "  left = new " + leftName + ";\n"
    "  right = new " + rightName + ";\n\n";

  // join the corresponding sources/sinks of the subexpressions
  for(unsigned int i = 0; i < lSourceOrder.size(); ++i)
    for(unsigned int j = 0; j < rSourceOrder.size(); ++j)
      if(lSourceOrder[i] == rSourceOrder[j])
	circuit += 
	  "  join(left.source[" + lexical_cast<string>(i+1)
	  + "], right.source[" + lexical_cast<string>(j+1) + "]);\n"
	  "  join(left.source[" + lexical_cast<string>(lSourceOrder.size()+i+1)
	  + "], right.source[" + lexical_cast<string>(rSourceOrder.size()+j+1) + "]);\n"
	  "  join(left.sink[" + lexical_cast<string>(lSinkOrder.size()+i+1)
	  + "], right.sink[" + lexical_cast<string>(rSinkOrder.size()+j+1) + "]);\n";

  for(unsigned int i = 0; i < lSinkOrder.size(); ++i)
    for(unsigned int j = 0; j < rSinkOrder.size(); ++j)
      if(lSinkOrder[i] == rSinkOrder[j])
	circuit += 
	  "  join(left.sink[" + lexical_cast<string>(i+1)
	  + "], right.sink[" + lexical_cast<string>(j+1) + "]);\n";

  circuit +=
    "\n  leftFifo = new FIFO1;\n"
    "  leftSync = new SYNC;\n"
    "  router = new Exrouter;\n"
    "  nwSync = new SYNC;\n"
    "  neSync = new SYNC;\n"
    "  seSync = new SYNC;\n"
    "  swSync = new SYNC;\n"
    "  rightFifo = new FIFO1;\n"
    "  join(leftFifo.sink, leftSync.source);\n"
    "  join(leftSync.sink, router.source);\n"
    "  join(router.sink[0], nwSync.source);\n"
    "  join(router.sink[1], swSync.source);\n"
    "  join(nwSync.sink, left.source[0]);\n"
    "  join(swSync.sink, right.source[0]);\n"
    "  join(left.sink[0], neSync.source);\n"
    "  join(right.sink[0], seSync.source);\n"
    "  join(rightFifo.source, neSync.sink);\n"
    "  join(rightFifo.source, seSync.sink);\n\n"

    "  source[0] = leftFifo.source;\n"
    "  sink[0] = rightFifo.sink;\n\n";

  // in order to export the sources/sinks, it has to be determined from which subcircuit they come from
  for(unsigned int i = 0; i < sourceOrder.size(); ++i) {
    unsigned int j;
    for(j = 0; j < lSourceOrder.size(); ++j)
      if(sourceOrder[i] == lSourceOrder[j]) break;
    if(j != lSourceOrder.size()) {
      circuit +=
	"  source[" + lexical_cast<string>(i+1) + "] = left.source[" + lexical_cast<string>(j+1) + "];\n";
    } else {
      for(j = 0; j < rSourceOrder.size(); ++j)
	if(sourceOrder[i] == rSourceOrder[j]) break;
      if(j != rSourceOrder.size())
	circuit +=
	  "  source[" + lexical_cast<string>(i+1) + "] = right.source[" + lexical_cast<string>(j+1) + "];\n";
      else
	assert(0);
    }
  }
  for(unsigned int i = 0; i < sinkOrder.size(); ++i) {
    unsigned int j;
    for(j = 0; j < lSinkOrder.size(); ++j)
      if(sinkOrder[i] == lSinkOrder[j]) break;
    if(j != lSinkOrder.size()) {
      circuit +=
	"  sink[" + lexical_cast<string>(i+1) + "] = left.sink[" + lexical_cast<string>(j+1) + "];\n";
    } else {
      for(j = 0; j < rSinkOrder.size(); ++j)
	if(sinkOrder[i] == rSinkOrder[j]) break;
      if(j != rSinkOrder.size())
	circuit +=
	  "  sink[" + lexical_cast<string>(i+1) + "] = right.sink[" + lexical_cast<string>(j+1) + "];\n";
      else
	assert(0);
    }
  }
  for(unsigned int i = 0; i < sourceOrder.size(); ++i) {
    unsigned int j;
    for(j = 0; j < lSourceOrder.size(); ++j)
      if(sourceOrder[i] == lSourceOrder[j]) break;
    if(j != lSourceOrder.size()) {
      circuit +=
	"  source[" + lexical_cast<string>(sourceOrder.size()+i+1) + "] = left.source[" 
	+ lexical_cast<string>(lSourceOrder.size()+j+1) + "];\n"
      	"  sink[" + lexical_cast<string>(sinkOrder.size()+i+1) + "] = left.sink[" 
	+ lexical_cast<string>(lSinkOrder.size()+j+1) + "];\n";
    } else {
      for(j = 0; j < rSourceOrder.size(); ++j)
	if(sourceOrder[i] == rSourceOrder[j]) break;
      if(j != rSourceOrder.size())
	circuit +=
	  "  source[" + lexical_cast<string>(sourceOrder.size()+i+1) + "] = right.source[" 
	  + lexical_cast<string>(rSourceOrder.size()+j+1) + "];\n"
	  "  sink[" + lexical_cast<string>(sinkOrder.size()+i+1) + "] = right.sink[" 
	  + lexical_cast<string>(rSinkOrder.size()+j+1) + "];\n";
      else
	assert(0);
    }
  }
  circuit += "}\n\n";
  return name;
}

vector<string> SDEAlternative::sinkArgumentOrder() const
{
  vector<string> ret;

  vector<string> left = leftChild->sinkArgumentOrder();
  vector<string> right = rightChild->sinkArgumentOrder();

  set_union(left.begin(), left.end(),
	    right.begin(), right.end(),
	    inserter(ret, ret.end()));

  return ret;
}

vector<string> SDEAlternative::sourceArgumentOrder() const
{
  vector<string> ret;

  vector<string> left = leftChild->sourceArgumentOrder();
  vector<string> right = rightChild->sourceArgumentOrder();

  set_union(left.begin(), left.end(),
	    right.begin(), right.end(),
	    inserter(ret, ret.end()));

  return ret;
}

const ScheduledDataExpression& SDEAlternative::getLeftChild() const
{
  return *leftChild;
}

const ScheduledDataExpression& SDEAlternative::getRightChild() const
{
  return *rightChild;
}

void SDEAlternative::setLeftChild(ScheduledDataExpression::ptr _left)
{
  leftChild = _left;
}

void SDEAlternative::setRightChild(ScheduledDataExpression::ptr _right)
{
  rightChild = _right;
}

ScheduledDataExpression::ptr SDEAlternative::clone() const
{
  ScheduledDataExpression::ptr ret(new SDEAlternative(leftChild->clone(), rightChild->clone()));
  return ret;
}

void SDEAlternative::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "SDEAlternative\n";
  leftChild->printTreeHelper(level + 2);
  rightChild-> printTreeHelper(level + 2);
}

////////////
// SDEInf
////////////

SDEInf::SDEInf()
  : ScheduledDataExpression()
{}

SDEInf::SDEInf(ScheduledDataExpression::ptr _child)
  : ScheduledDataExpression(), child(_child)
{}

SDEInf::SDEInf(const SDEInf &rhs)
  : ScheduledDataExpression(), child(rhs.child->clone())
{}

string SDEInf::toREOCircuitHelper(string &circuit, unsigned int& label) const
{
  string name = "Inf" + lexical_cast<string>(label++);
  string componentName = child->toREOCircuitHelper(circuit, label);

  vector<string> cSourceOrder = child->sourceArgumentOrder();
  vector<string> cSinkOrder = child->sinkArgumentOrder();

  // create circuit for subexpression
  circuit +=
    "CIRCUIT " + name + "{\n"
    "  component = new " + componentName + ";\n";

  circuit +=
    "  leftFifo = new FIFO1;\n"
    "  leftSync = new SYNC;\n"
    "  middleSync = new SYNC;\n"
    "  router = new Exrouter;\n"
    "  rightSync = new SYNC;\n"
    "  backSync = new SYNC;\n"
    "  rightFifo = new FIFO1;\n"
    "  join(leftFifo.sink, leftSync.source);\n"
    "  join(leftSync.sink, component.source[0]);\n"
    "  join(component.sink[0], middleSync.source);\n"
    "  join(middleSync.sink, router.source);\n"
    "  join(router.sink[0], rightSync.source);\n"
    "  join(rightSync.sink, rightFifo.source);\n"
    "  join(router.sink[1], backSync.source);\n"
    "  join(backSync.sink, leftSync.source);\n\n"

    "  source[0] = leftFifo.source;\n"
    "  sink[0] = rightFifo.sink;\n\n";

  // export the sources/sinks of the subexpression
  for(unsigned int i = 0; i < cSourceOrder.size(); ++i)
    circuit +=
      "  source[" + lexical_cast<string>(i+1) + "] = component.source[" 
      + lexical_cast<string>(i+1) + "];\n";
  for(unsigned int i = 0; i < cSourceOrder.size(); ++i)
    circuit +=
      "  source[" + lexical_cast<string>(cSourceOrder.size()+i+1) + "] = component.source[" 
      + lexical_cast<string>(cSourceOrder.size()+i+1) + "];\n";
  for(unsigned int i = 0; i < cSinkOrder.size(); ++i)
    circuit +=
      "  sink[" + lexical_cast<string>(i+1) + "] = component.sink[" 
      + lexical_cast<string>(i+1) + "];\n";
  for(unsigned int i = 0; i < cSourceOrder.size(); ++i) {
    circuit +=
      "  sink[" + lexical_cast<string>(cSinkOrder.size()+i+1) + "] = component.sink["
      + lexical_cast<string>(cSinkOrder.size()+i+1) + "];\n";
  }

  circuit += "}\n\n";
  return name;
}

vector<string> SDEInf::sinkArgumentOrder() const
{
  return child->sinkArgumentOrder();
}

vector<string> SDEInf::sourceArgumentOrder() const
{
  return child->sourceArgumentOrder();
}

const ScheduledDataExpression& SDEInf::getChild() const
{
  return *child;
}

void SDEInf::setChild(ScheduledDataExpression::ptr _child)
{
  child = _child;
}

ScheduledDataExpression::ptr SDEInf::clone() const
{
  ScheduledDataExpression::ptr ret(new SDEInf(child->clone()));
  return ret;
}

void SDEInf::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "SDEInf\n";
  child->printTreeHelper(level + 2);
}

//////////////////
// SDEManyOrInf
//////////////////

SDEManyOrInf::SDEManyOrInf()
  : ScheduledDataExpression()
{}

SDEManyOrInf::SDEManyOrInf(ScheduledDataExpression::ptr _child)
  : ScheduledDataExpression(), child(_child)
{}

SDEManyOrInf::SDEManyOrInf(const SDEManyOrInf &rhs)
  : ScheduledDataExpression(), child(rhs.child->clone())
{}

string SDEManyOrInf::toREOCircuitHelper(string &circuit, unsigned int &label) const
{
  string name = "ManyInf" + lexical_cast<string>(label++);
  string componentName = child->toREOCircuitHelper(circuit, label);
  vector<string> cSourceOrder = child->sourceArgumentOrder();
  vector<string> cSinkOrder = child->sinkArgumentOrder();

  // create circuit for subexpression
  circuit +=
    "CIRCUIT " + name + "{\n"
    "  component = new " + componentName + ";\n"
    "  leftFifo = new FIFO1;\n"
    "  leftSync = new SYNC;\n"
    "  middleSync = new SYNC;\n"
    "  router = new Exrouter;\n"
    "  rightSync = new SYNC;\n"
    "  backSync = new SYNC;\n"
    "  pseudo = new PseudoChannel;\n"
    "  join(leftFifo.sink, leftSync.source);\n"
    "  join(leftSync.sink, component.source[0]);\n"
    "  join(component.sink[0], middleSync.source);\n"
    "  join(middleSync.sink, router.source);\n"
    "  join(router.sink[0], pseudo.source);\n"
    "  join(router.sink[1], backSync.source);\n"
    "  join(backSync.sink, leftSync.source);\n\n"

    "  source[0] = leftFifo.source;\n"
    "  sink[0] = pseudo.sink;\n\n";

  // export the sources/sinks of the subexpression
  for(unsigned int i = 0; i < cSourceOrder.size(); ++i)
    circuit +=
      "  source[" + lexical_cast<string>(i+1) + "] = component.source[" 
      + lexical_cast<string>(i+1) + "];\n";
  for(unsigned int i = 0; i < cSourceOrder.size(); ++i)
    circuit +=
      "  source[" + lexical_cast<string>(cSourceOrder.size()+i+1) + "] = component.source[" 
      + lexical_cast<string>(cSourceOrder.size()+i+1) + "];\n";
  for(unsigned int i = 0; i < cSinkOrder.size(); ++i)
    circuit +=
      "  sink[" + lexical_cast<string>(i+1) + "] = component.sink[" 
      + lexical_cast<string>(i+1) + "];\n";
  for(unsigned int i = 0; i < cSourceOrder.size(); ++i) {
    circuit +=
      "  sink[" + lexical_cast<string>(cSinkOrder.size()+i+1) + "] = component.sink["
      + lexical_cast<string>(cSinkOrder.size()+i+1) + "];\n";
  }
  


  circuit += "}\n\n";
  return name;
}

vector<string> SDEManyOrInf::sinkArgumentOrder() const
{
  return child->sinkArgumentOrder();
}

vector<string> SDEManyOrInf::sourceArgumentOrder() const
{
  return child->sourceArgumentOrder();
}


const ScheduledDataExpression& SDEManyOrInf::getChild() const
{
  return *child;
}

void SDEManyOrInf::setChild(ScheduledDataExpression::ptr _child)
{
  child = _child;
}

ScheduledDataExpression::ptr SDEManyOrInf::clone() const
{
  ScheduledDataExpression::ptr ret(new SDEManyOrInf(child->clone()));
  return ret;
}

void SDEManyOrInf::printTreeHelper(int level) const
{
  cerr << string(level, ' ') << "SDEManyOrInf\n";
  child->printTreeHelper(level + 2);
}
