/**************************************************************************
*   Copyright (C) 2008-2009 by Johannes Osterholzer                       *
*   s5892660@inf.tu-dresden.de                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef SCHEDULEDDATAEXPRESSION_HPP
#define SCHEDULEDDATAEXPRESSION_HPP

#include <boost/spirit/home/classic.hpp>
#include <boost/spirit/home/classic/tree/ast.hpp>
#include <boost/spirit/home/classic/tree/common.hpp>
#include <boost/spirit/home/classic/tree/tree_to_xml.hpp>
#include <boost/smart_ptr.hpp>
#include <boost/spirit/home/classic/iterator.hpp>
#include <string>
#include <set>
#include <vector>
#include <utility>
#include <iostream>
#include "dataconstraint.hpp"

using namespace std;
using namespace boost;
using namespace boost::spirit::classic;

// Abstract base class representing scheduled data expressions
class ScheduledDataExpression {
public:
  typedef shared_ptr<ScheduledDataExpression> ptr;

  ScheduledDataExpression();
  virtual ~ScheduledDataExpression() {}
  
  // Parses a string and constructs a SDE from it
  static ScheduledDataExpression::ptr readFromString(const string&);

  // creates a REO circuit (hopefully) equivalent to the input SDE
  string toREOCircuit() const;

  virtual string toREOCircuitHelper(string&, unsigned int&) const = 0;
  virtual ScheduledDataExpression::ptr clone() const = 0;
  
  // Prints the SDE as a tree.
  void printTree() const;
  virtual void printTreeHelper(int) const = 0;
  
  // returns equivalent SDE in canonical form
  virtual ScheduledDataExpression::ptr canonicalForm() const = 0;
  // order of sinks of the current expression, necessary for REO construction
  virtual vector<string> sinkArgumentOrder() const = 0;
  // dito
  virtual vector<string> sourceArgumentOrder() const = 0;

  static vector<string> getSources();
  static void setSources(vector<string>);

  static vector<string> getSinks();
  static void setSinks(vector<string>);

  static pair<int,int> getDataDomain();
  static void setDataDomain(const pair<int,int>&);

  // is this expression in canonical form?
  bool inCanonicalForm; //TODO: move to somewhere more sensible

protected:
  // helper function for handling the abstract syntax tree of a SDE
  static ScheduledDataExpression::ptr evaluateAST(iter_t const&);
  // helper function for getting port names out of a parsetree
  static set<string> getPortnames(const iter_t &i);

  // global source nodes
  static vector<string> sources;
  //global sink nodes
  static vector<string> sinks;
  // the data domain, minimum and maximum value
  static pair<int,int> dataDomain;
};

class SDEAtom :public ScheduledDataExpression {
public:
  SDEAtom();
  SDEAtom(set<string>, DataConstraint::ptr);
  SDEAtom(const SDEAtom&);

  virtual string toREOCircuitHelper(string&, unsigned int&) const;

  const set<string>& getPortset() const;
  void setPortset(const set<string>&);
  const DataConstraint& getDataConstraint() const;
  void setDataConstraint(DataConstraint::ptr);

  virtual ScheduledDataExpression::ptr clone() const;
  virtual void printTreeHelper(int) const;

  virtual ScheduledDataExpression::ptr canonicalForm() const;
  virtual vector<string> sinkArgumentOrder() const;
  virtual vector<string> sourceArgumentOrder() const;

protected:
  pair<string, string> producerFromConstraint(const set<int>&, unsigned int&) const;
  pair<string, string> filterFromConstraint(const set<int>&, unsigned int&) const;

  set<string> portset;
  DataConstraint::ptr dataConstraint;
};

class SDEEpsilon : public ScheduledDataExpression {
public:
  SDEEpsilon();
  
  virtual string toREOCircuitHelper(string&, unsigned int&) const;
  virtual ScheduledDataExpression::ptr clone() const;
  virtual void printTreeHelper(int) const;

  virtual ScheduledDataExpression::ptr canonicalForm() const 
    {
      ScheduledDataExpression::ptr ret(new SDEEpsilon);
      ret->inCanonicalForm = true;
      return ret;
    }
  virtual vector<string> sinkArgumentOrder() const;
  virtual vector<string> sourceArgumentOrder() const;

protected:
};

class SDEConcatenation : public ScheduledDataExpression {
public:
  SDEConcatenation();
  SDEConcatenation(ScheduledDataExpression::ptr, ScheduledDataExpression::ptr);
  SDEConcatenation(const SDEConcatenation&);

  virtual string toREOCircuitHelper(string&, unsigned int&) const;

  const ScheduledDataExpression& getLeftChild() const;
  const ScheduledDataExpression& getRightChild() const;
  void setLeftChild(ScheduledDataExpression::ptr);
  void setRightChild(ScheduledDataExpression::ptr);

  virtual ScheduledDataExpression::ptr clone() const;
  virtual void printTreeHelper(int) const;

  virtual ScheduledDataExpression::ptr canonicalForm() const
    {
      ScheduledDataExpression::ptr ret(new SDEConcatenation(leftChild->canonicalForm(),
								   rightChild->canonicalForm()));
      ret->inCanonicalForm = true;
      return ret;
    }
  virtual vector<string> sinkArgumentOrder() const;
  virtual vector<string> sourceArgumentOrder() const;

protected:
  ScheduledDataExpression::ptr leftChild, rightChild;
};

class SDEAlternative : public ScheduledDataExpression {
public:
  SDEAlternative();
  SDEAlternative(ScheduledDataExpression::ptr, ScheduledDataExpression::ptr);
  SDEAlternative(const SDEAlternative&);

  virtual string toREOCircuitHelper(string&, unsigned int&) const;

  const ScheduledDataExpression& getLeftChild() const;
  const ScheduledDataExpression& getRightChild() const;
  void setLeftChild(ScheduledDataExpression::ptr);
  void setRightChild(ScheduledDataExpression::ptr);

  virtual ScheduledDataExpression::ptr clone() const;
  virtual void printTreeHelper(int) const;

  virtual ScheduledDataExpression::ptr canonicalForm() const
    {
      ScheduledDataExpression::ptr ret(new SDEAlternative(leftChild->canonicalForm(),
							  rightChild->canonicalForm()));
      ret->inCanonicalForm = true;
      return ret;
    }
  virtual vector<string> sinkArgumentOrder() const;
  virtual vector<string> sourceArgumentOrder() const;


protected:
  ScheduledDataExpression::ptr leftChild, rightChild;
};

class SDEInf : public ScheduledDataExpression {
public:
  SDEInf();
  SDEInf(ScheduledDataExpression::ptr);
  SDEInf(const SDEInf&);

  virtual string toREOCircuitHelper(string&, unsigned int&) const;

  const ScheduledDataExpression& getChild() const;
  void setChild(ScheduledDataExpression::ptr);

  virtual ScheduledDataExpression::ptr clone() const;
  virtual void printTreeHelper(int) const;

  virtual ScheduledDataExpression::ptr canonicalForm() const
    {
      ScheduledDataExpression::ptr ret(new SDEInf(child->canonicalForm()));
      ret->inCanonicalForm = true;
      return ret;
    }
  virtual vector<string> sinkArgumentOrder() const;
  virtual vector<string> sourceArgumentOrder() const;


protected:
  ScheduledDataExpression::ptr child;
};

class SDEManyOrInf : public ScheduledDataExpression {
public:
  SDEManyOrInf();
  SDEManyOrInf(ScheduledDataExpression::ptr);
  SDEManyOrInf(const SDEManyOrInf&);

  virtual string toREOCircuitHelper(string&, unsigned int&) const;

  const ScheduledDataExpression& getChild() const;
  void setChild(ScheduledDataExpression::ptr);

  virtual ScheduledDataExpression::ptr clone() const;
  virtual void printTreeHelper(int) const;

  virtual ScheduledDataExpression::ptr canonicalForm() const
    {
      ScheduledDataExpression::ptr ret(new SDEManyOrInf(child->canonicalForm()));
      ret->inCanonicalForm = true;
      return ret;
    }
  virtual vector<string> sinkArgumentOrder() const;
  virtual vector<string> sourceArgumentOrder() const;

protected:
  ScheduledDataExpression::ptr child;
};


#endif
