/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "circuit_level.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/reo_parser/reo_parser_types.hpp"
#include "vereofy/reo_parser/circuit_types.hpp"
#include "vereofy/reo_parser/circuit_parser_helper.hpp"
#include "symb_circuit.hpp"

namespace vereofy {

CircuitLevel::CircuitLevel(CircuitConnector& connector_, value_table_ptr vt) :
  id(0), was_cloned(false), connector(connector_), values(vt), anonymous_counter(0), parent(0) {
}

CircuitLevel::ptr CircuitLevel::clone(unsigned int id_) const {
  std::pair<REO_Circuit::ptr, REO_Graph_Object::rgo_map> reo_clone=reo_circuit->clone();

  value_table_ptr cloned_vt=clone_and_map_value_table(values, reo_clone.second);
  CircuitLevel::ptr result(new CircuitLevel(connector, cloned_vt));
  result->reo_circuit=reo_clone.first;
  result->mapping=reo_clone.second;
  result->was_cloned=true;
  result->id=id_;

  return result;
}

REO_Circuit::ptr CircuitLevel::getREO_Circuit() {
  return reo_circuit;
}

CircuitLevel::value_table_ptr CircuitLevel::getValueTable() {
  return values;
}

void CircuitLevel::execute(const vparser::pt_node& statement_block,
                           symbol_table_ptr local_symbols) {
  if (!reo_circuit) {
    reo_circuit=REO_Circuit::ptr(new REO_Circuit(access_weak_ptr(connector.context)->getDataDomain()));
  }

  execution_environment_t env(access_weak_ptr(connector.context)->getGlobalSymbols(),
                              local_symbols,
                              access_weak_ptr(connector.context),
                              values);
  env.setREOCircuit(reo_circuit);

  vparser::circuit_helper::t_generic::execute(env, statement_block);
}

void CircuitLevel::build_name_map() {
  for (execution_environment_t::iterator it=values->begin();
       it!=values->end();
       ++it) {
    string name=it->first;
    bool is_sink_source=(name=="sink" || name=="source");
    execution_environment_t::value_vector_ptr values=it->second;

    for (unsigned int i=0;i<values->size();i++) {
      Value::ptr v=values->at(i);
      string full_name=name+"["+boost::lexical_cast<string>(i)+"]";

      if (!v) {
        if (is_sink_source) {
          throw full_name+" is undefined!";
        } else {
          // ignore
          continue;
        }
      }

      REO_Graph_Object::ptr rgo;
      if (v->type->type==VALUE_TYPE_COMPONENT_REF) {
        if (is_sink_source) {
          throw full_name+" is a component, "+name+" has to be a  node or channel end!";
        }

        REO_Component::ptr reo_component=static_cast<vparser::ComponentRef&>(*v).reo_component;
        reo_component->getComponentInstance()->setName(full_name);
        
        rgo=reo_component; 

      } else if (v->type==ValueType::NODE_REF) {
        REO_Node::ptr node=static_cast<vparser::NodeRef&>(*v).reo_node;
        node=reo_circuit->getRealNode(node);
        rgo=node;

        if (is_sink_source && reo_circuit->getNodeType(node)==REO_Node::NODE_MIXED) {
          throw "Trying to export node "+full_name+" as "+name+", but is a mixed node!";
        }
      } else if (v->type==ValueType::CHANNEL_END_REF) {
        REO_Channel_End::ptr reo_ce=static_cast<vparser::ChannelEndRef&>(*v).reo_channel_end;
        rgo=reo_ce;

        if (is_sink_source && reo_circuit->getConnectedNode(reo_ce)) {
          throw "Trying to export component port/channel end "+full_name+" as "+name+", but is connected to a node!";
        }
      } else {
        if (is_sink_source) {
          throw full_name+" is not a node or channel end!";
        }
        // ignore other data types (int, bool, etc...)
        continue;
      }

      reo_to_names.insert(make_pair(rgo, full_name));

      if (is_sink_source) {
        if (name=="sink") {
          exported_sinks.insert(rgo);
          sinks.push_back(rgo);
          CHECK(1, sinks.size()==i+1, "");
        } else if (name=="source") {
          exported_sources.insert(rgo);
          sources.push_back(rgo);
          CHECK(1, sources.size()==i+1, "");
        } else {
          throw "implementation error";
        }
      }
    }
  }
}


void CircuitLevel::plug(bool strict) {
  if (!strict) {
    return;
  }

  // deal with the nodes
  vector<REO_Node::ptr> nodes=reo_circuit->getAllNodes();
  
  for (unsigned int i=0;i<nodes.size();i++) {
    REO_Node::ptr reo_node=nodes.at(i);
    REO_Node::node_type_t type=reo_circuit->getNodeType(reo_node);

    if (is_exported_sink(reo_node) && is_exported_source(reo_node)) {
      throw "Node exported as sink and as source at the same time!";
    }

    switch (type) {
    case REO_Node::NODE_FREE:
      if (is_exported_sink(reo_node)) {
        // plug the source side
        reo_node->setSourcePlug(true);
      } else if (is_exported_source(reo_node)) {
        // plug the sink side
        reo_node->setSinkPlug(true);
      } else {
        // plug both sides
        reo_node->setSourcePlug(true);
        reo_node->setSinkPlug(true);
      }
      break;
    case REO_Node::NODE_SINK:
      if (!is_exported_sink(reo_node)) {
        // unexported sink node -> plug sink side
        reo_node->setSinkPlug(true);
      }
      break;
    case REO_Node::NODE_SOURCE:
      if (!is_exported_source(reo_node)) {
        // unexported sink node -> plug source side
        reo_node->setSourcePlug(true);
      }
      break;
    case REO_Node::NODE_MIXED:
      // no plugging to do here
      break;
    }
  }


  // now, deal with the channel ends
  vector<REO_Channel_End::ptr> ces=reo_circuit->getAllUnconnectedSinkCEs();
  for (unsigned int i=0;i<ces.size();i++) {
    if (!is_exported_sink(ces.at(i))) {
      ces.at(i)->setPlug(true);
    }
  }

  ces=reo_circuit->getAllUnconnectedSourceCEs();
  for (unsigned int i=0;i<ces.size();i++) {
    if (!is_exported_source(ces.at(i))) {
      ces.at(i)->setPlug(true);
    }
  }
}

void CircuitLevel::build_interface(ComponentInterface::ptr c_interface, bool strict) {
  anonymous_counter=0;

  // do sources/sinks first
  for (exported_rgos::iterator it=sources.begin();
       it!=sources.end();
       ++it) {
    build_interface_for(c_interface, *it, strict);
  }

  for (exported_rgos::iterator it=sinks.begin();
       it!=sinks.end();
       ++it) {
    build_interface_for(c_interface, *it, strict);
  }       

  // do the rest
  vector<REO_Graph_Object::ptr> rgos=reo_circuit->getAllGraphObjects();
  for (unsigned int i=0;i<rgos.size();i++) {
    if (is_exported_sink(rgos[i]) || is_exported_source(rgos[i])) {
      continue;  // already done above
    }

    build_interface_for(c_interface, rgos[i], strict);
  }
}

void CircuitLevel::build_interface_for(ComponentInterface::ptr c_interface,
                                                   REO_Graph_Object::ptr rgo,
                                                   bool strict) {
  if (rgo->wasCloned()) {
    REO_Graph_Object::ptr clonedFrom=rgo->getClonedFrom();

    CHECK(1, clonedFrom->get_graph_object_type() == rgo->get_graph_object_type(), "");


    // copy the port descriptions from the base objects
    switch (clonedFrom->get_graph_object_type()) {
    case REO_Graph_Object::REO_NODE: {
      PortDescription::ptr pd=static_cast<REO_Node&>(*clonedFrom).getPortDescription();
      CHECK_SUCCESS(pd);
      static_cast<REO_Node&>(*rgo).setPortDescription(pd);
      break;
    }
    case REO_Graph_Object::REO_CHANNEL_END: {
      PortDescription::ptr pd=static_cast<REO_Channel_End&>(*clonedFrom).get_PortDescription_Circuit();
      CHECK_SUCCESS(pd);
      static_cast<REO_Channel_End&>(*rgo).set_PortDescription_Circuit(pd);
      break;
    }      
    case REO_Graph_Object::REO_COMPONENT:
      // nothing to do
      break;

    default:
      THROW_UNIMPLEMENTED();
    }
    return;
  }


  REO_Graph_Object::Graph_Object_Type type=rgo->get_graph_object_type();

  TRACE("ComponentInstanceCircuit", "build_interface_for", cout << "RGO: " << rgo <<endl;);

  switch (type) {
  case REO_Graph_Object::REO_COMPONENT: {
    REO_Component::ptr component=boost::static_pointer_cast<REO_Component>(rgo);

    unsigned int name_count=reo_to_names.count(rgo);

    string name;
    if (name_count==0) {
      if (!was_cloned) {
        name="$comp"+boost::lexical_cast<std::string>(anonymous_counter++);
      } else {
        name="$comp_"+boost::lexical_cast<std::string>(id)+"_"+boost::lexical_cast<std::string>(anonymous_counter++);
      }

      reo_to_names.insert(make_pair(rgo, name));
      CHECK_SUCCESS( component->getComponentInstance()->getName()=="" );
      component->getComponentInstance()->setName(name);
    } else {
      name=reo_to_names.find(rgo)->second;    
    }
     
    if (name_count>=2) {
      std::cerr << "Warning: Component " << name << " has multiple names, but only this name will be in interface";
    }
    
    TRACE("ComponentInstanceCircuit", "build_interface_for", cout << "Add (sub-)component: " << name <<endl;);

    c_interface->addSubcomponent(name, component->getComponentInstance()->getInterface());
    if (parent) {
      parent->registerSubcomponent(component->getComponentInstance());
    }
    return;
  }
  case REO_Graph_Object::REO_NODE:
  case REO_Graph_Object::REO_CHANNEL_END: {
    REO_Circuit::reo_to_names_map::const_iterator it, it_end;
    tie(it, it_end)=reo_to_names.equal_range(rgo);

    string primary_name;
    bool is_anonymous=false;
    if (it==it_end) {
      // anonymous, no names
      primary_name=(type==REO_Graph_Object::REO_NODE ? "$node" : "$ce");
      if (was_cloned) {
        primary_name+="_"+boost::lexical_cast<std::string>(id)+"_";
      }
      primary_name+=boost::lexical_cast<std::string>(anonymous_counter++);
      is_anonymous=true;
    } else {
      primary_name=it->second;
      ++it;
    }

    TRACE("ComponentInstanceCircuit", "build_interface_for", cout << "Name: " << primary_name <<endl;);
    TRACE("ComponentInstanceCircuit", "build_interface_for", cout << "anonymous: " << is_anonymous <<endl;);

    PortDescription::ptr pd;
    if (type==REO_Graph_Object::REO_NODE) {
      pd=reo_circuit->getPortDescription(boost::static_pointer_cast<REO_Node>(rgo),
                                         primary_name);
    } else if (type==REO_Graph_Object::REO_CHANNEL_END) {
      pd=reo_circuit->getPortDescription(boost::static_pointer_cast<REO_Channel_End>(rgo),
                                         primary_name);
    }

    if (is_anonymous) {
      pd->setAnonymous(true);
    } 
    
    bool rv=c_interface->addPort(pd);
    if (!rv) {
      throw string("Couldn't add "+primary_name);
    }
    
    // handle extra names
    for (;it!=it_end;++it) {
      TRACE("ComponentInstanceCircuit", "build_interface_for", cout << "Alias: " << it->second <<endl;);
      c_interface->addPortAlias(pd, it->second);
    }

    return;
  }
  }

  THROW_UNIMPLEMENTED();
}

void CircuitLevel::toDot(ostream& out, bool is_subgraph) {
  reo_circuit->toDot(out, &reo_to_names, is_subgraph);
}

void CircuitLevel::toReoXML(ostream& out, const std::string& name, const std::vector<std::string>& input_files, bool is_subgraph) {
  reo_circuit->toReoXML(out, name, &reo_to_names, input_files, is_subgraph);
}

bool CircuitLevel::is_anonymous(REO_Graph_Object::ptr rgo) {
  return (reo_to_names.count(rgo) == 0);
}

bool CircuitLevel::is_exported_sink(REO_Graph_Object::ptr rgo) {
  return (exported_sinks.count(rgo) > 0);
}

bool CircuitLevel::is_exported_source(REO_Graph_Object::ptr rgo) {
  return (exported_sources.count(rgo) > 0);
}

void CircuitLevel::resetValueTable(const value_table_ptr& vt) {
  CHECK_SUCCESS(was_cloned);
  values=clone_and_map_value_table(vt,mapping);
}


CircuitLevel::value_table_ptr CircuitLevel::clone_and_map_value_table(const value_table_ptr& vt,
                                                                      const REO_Graph_Object::rgo_map& clone_mapping) const {
  value_table_ptr result(new execution_environment_t::value_table_t);

  for (execution_environment_t::value_table_t::symbol_map::iterator it=vt->begin();
       it!=vt->end();
       ++it) {
    string name=(*it).first;
    execution_environment_t::value_vector_ptr v_vec=(*it).second;

    execution_environment_t::value_vector_ptr target_vec(new execution_environment_t::value_vector_t(v_vec->size()));
    CHECK_SUCCESS(result->add(name, target_vec) == true);


    for (size_t i=0;i<v_vec->size();i++) {
      Value::ptr value=v_vec->at(i);
      if (value) {
        Value::ptr new_value=value->clone();

        switch (new_value->type->type) {
        case VALUE_TYPE_INT:
          // nothing to do
          break;
        case VALUE_TYPE_NODE_REF: {
          REO_Node::ptr old_rgo=static_cast<vparser::NodeRef&>(*value).reo_node;
          old_rgo=reo_circuit->getRealNode(old_rgo);
          CHECK_SUCCESS(clone_mapping.count(old_rgo)==1);
          REO_Node::ptr new_rgo=boost::dynamic_pointer_cast<REO_Node>(clone_mapping.find(old_rgo)->second);
          CHECK_SUCCESS(new_rgo);
          static_cast<vparser::NodeRef&>(*new_value).reo_node=new_rgo;
          break;
        }          
        case VALUE_TYPE_CHANNEL_END_REF: {
          REO_Channel_End::ptr old_rgo=static_cast<vparser::ChannelEndRef&>(*value).reo_channel_end;
          CHECK_SUCCESS(clone_mapping.count(old_rgo)==1);
          REO_Channel_End::ptr new_rgo=boost::dynamic_pointer_cast<REO_Channel_End>(clone_mapping.find(old_rgo)->second);
          CHECK_SUCCESS(new_rgo);
          static_cast<vparser::ChannelEndRef&>(*new_value).reo_channel_end=new_rgo;
          break;
        }
        case VALUE_TYPE_COMPONENT_REF: {
          REO_Component::ptr old_rgo=static_cast<vparser::ComponentRef&>(*value).reo_component;
          CHECK_SUCCESS(clone_mapping.count(old_rgo)==1);
          REO_Component::ptr new_rgo=boost::dynamic_pointer_cast<REO_Component>(clone_mapping.find(old_rgo)->second);
          CHECK_SUCCESS(new_rgo);
          static_cast<vparser::ComponentRef&>(*new_value).reo_component=new_rgo;
          break;
        }
        default:
          THROW_UNIMPLEMENTED();
        }

        target_vec->at(i)=new_value;
      }
    }

  }

  return result;
}
  
void CircuitLevel::set_parent(ComponentInstance *parent) {
  this->parent=parent;
  reo_circuit->setParent(parent);
}
  

void CircuitLevel::export_circuit(ostream& out) const {
  vector<REO_Graph_Object::ptr> components=reo_circuit->getAllComponentsInBFS();
  for (unsigned int i=0;i<components.size();i++) {
    if (components.at(i)->get_graph_object_type()!=REO_Graph_Object::REO_COMPONENT) {
      // Not a component
      continue;
    }
    REO_Component::ptr c=boost::static_pointer_cast<REO_Component>(components.at(i));
    ComponentInstance::ptr ci=c->getComponentInstance();
    string component_type=ci->getTypeName();

    out << "COMPONENT " << c << " \"" << component_type << "\" " << c->getSourceCECount() << " " << c->getSinkCECount() << endl;
  }

  vector<REO_Node::ptr> nodes=reo_circuit->getAllNodes();
  for (unsigned int i=0;i<nodes.size();i++) {
    REO_Node::ptr node=boost::static_pointer_cast<REO_Node>(nodes.at(i));
    out << "NODE " << node << endl;
  }

  vector<REO_Channel_End::ptr> ces=reo_circuit->getAllCEs();
  for (unsigned int i=0;i<ces.size();i++) {
    REO_Channel_End::ptr ce=boost::static_pointer_cast<REO_Channel_End>(ces.at(i));
    out << "PORT " << ce << " COMPONENT=" << reo_circuit->getComponentFor(ce) << " ";
    switch (ce->get_CE_Type()) {
    case REO_Channel_End::CE_SINK: out << "SINK"; break;
    case REO_Channel_End::CE_SOURCE: out << "SOURCE"; break;
    }

    PortDescription::ptr pd=ce->get_PortDescription_Component();
    out << " " << reo_circuit->getComponentFor(ce)->getComponentInstance()->getInterface()->getIndexForPortDescription(pd);
    REO_Node::ptr node=reo_circuit->getConnectedNode(ce);
    if (node) {
      out << " NODE=" << node;
    } else {
      out << " unconnected";
    }
    out << endl;
  }
  
  vector<REO_Graph_Object::ptr> rgos=reo_circuit->getAllGraphObjects();
  for (unsigned int i=0;i<rgos.size();i++) {
    REO_Graph_Object::ptr rgo=rgos.at(i);
    REO_Circuit::reo_to_names_map::const_iterator it, it_end;
    tie(it, it_end)=reo_to_names.equal_range(rgo);
    
    if (it!=it_end) {
      out << "NAMES FOR ";
      switch (rgo->get_graph_object_type()) {
      case REO_Graph_Object::REO_COMPONENT:
        out << "COMPONENT"; break;
      case REO_Graph_Object::REO_NODE:
        out << "NODE"; break;
      case REO_Graph_Object::REO_CHANNEL_END:
        out << "PORT"; break;
      }
      out << " " << rgo;
      for (;it!=it_end;++it) {
        out << " \"" << it->second << "\"";
      }

      out << endl;
    }
  }
}



void CircuitLevel::calculatePortPartitions(PortPartitions& port_partitions) {
  ComponentInstance* component=this->parent;

  vector<REO_Graph_Object::ptr> rgos=reo_circuit->getAllGraphObjects();
  for (unsigned int i=0;i<rgos.size();i++) {
    REO_Graph_Object::ptr rgo=rgos.at(i);

    switch (rgo->get_graph_object_type()) {
    case REO_Graph_Object::REO_COMPONENT: {
      REO_Component::ptr r_component=boost::static_pointer_cast<REO_Component>(rgo);
      
      r_component->getComponentInstance()->calculatePortPartitions(port_partitions);
      break;
    }      
    case REO_Graph_Object::REO_NODE: {
      REO_Node::ptr r_node=boost::static_pointer_cast<REO_Node>(rgo);
      
      vector<REO_Channel_End::ptr> out_ces=reo_circuit->getConnectedSourceChannelEnds(r_node);
      vector<REO_Channel_End::ptr> in_ces=reo_circuit->getConnectedSinkChannelEnds(r_node);
      
      // handle input side (non-deterministic choice)
      if (in_ces.size()==1) {
        // only one input -> identical with node
        port_partitions.identical_ports.make_equivalent(PortID(component, in_ces.at(0)->get_PortDescription_Circuit()),
                                                        PortID(component, r_node->getPortDescription()));
      } else {
        // multiple inputs -> non-det., mix with node
        for (unsigned int i=0;i<in_ces.size();i++) {
          port_partitions.mixed_data_ports.make_equivalent(PortID(component, in_ces.at(i)->get_PortDescription_Circuit()),
                                                           PortID(component, r_node->getPortDescription()));
        }
      }

      // handle output side
      switch (r_node->getNodeSemantic()) {
      case REO_Node::NODE_REPLICATE: {
        // all output ports of the node are identical to the node
        for (unsigned int i=0;i<out_ces.size();i++) {
          port_partitions.identical_ports.make_equivalent(PortID(component, out_ces.at(i)->get_PortDescription_Circuit()),
                                                          PortID(component, r_node->getPortDescription()));
        }
        break;
      }
      case REO_Node::NODE_ROUTE: {
        // all output ports of the node are mixed to the node
        for (unsigned int i=0;i<out_ces.size();i++) {
          port_partitions.mixed_data_ports.make_equivalent(PortID(component, out_ces.at(i)->get_PortDescription_Circuit()),
                                                           PortID(component, r_node->getPortDescription()));
        }
        break;
      }
      }


      // insert relations for all ports into if_active_same_data_ports
      set<PortID> port_ids;
      port_ids.insert(PortID(component, r_node->getPortDescription()));
      for (unsigned int i=0;i<in_ces.size();i++) {
        port_ids.insert(PortID(component, in_ces.at(i)->get_PortDescription_Circuit()));
      }
      for (unsigned int j=0;j<out_ces.size();j++) {
        port_ids.insert(PortID(component, out_ces.at(j)->get_PortDescription_Circuit()));
      }

      port_partitions.if_active_same_data_ports.make_all_related(port_ids);

      break;
    }
    case REO_Graph_Object::REO_CHANNEL_END: {
      REO_Channel_End::ptr r_ce=boost::static_pointer_cast<REO_Channel_End>(rgo);
      
      // The port description inside the component is identified with the port description
      // in the circuit
      port_partitions.identical_ports.make_equivalent(PortID(parent, r_ce->get_PortDescription_Circuit()),
                                                      PortID(reo_circuit->getComponentFor(r_ce)->getComponentInstance(), r_ce->get_PortDescription_Component()));
      break;
    }
    }
  }
}
  

}
