/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "component_instance_circuit.hpp"
#include "circuit_level.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/reo_parser/reo_parser_types.hpp"
#include "vereofy/reo_parser/circuit_types.hpp"
#include "vereofy/reo_parser/circuit_parser_helper.hpp"
#include "vereofy/reo_parser/common_parser_helper.hpp"
#include "symb_circuit.hpp"

namespace vereofy {

ComponentInstanceCircuit::ComponentInstanceCircuit(CircuitConnector::ptr circuit_connector_,
                                                   const std::vector<Value::ptr>& parameters_,
                                                   bool strict):
  ComponentInstance(COMPONENT_CIRCUIT,
                    circuit_connector_->context,
                    ComponentInterface::ptr(new ComponentInterface(""))),
  circuit_connector(circuit_connector_),
  parameters(parameters_) {

  execute();
  build_name_map();
  plug(strict);

  build_interface(strict);

  getInterface()->setGlobalSymbols(access_weak_ptr(circuit_connector->context)->getGlobalSymbols());
}

CircuitConnector::ptr ComponentInstanceCircuit::getCircuitConnector() {
  return circuit_connector;
}

REO_Circuit::ptr ComponentInstanceCircuit::getREO_Circuit() {
  return base->getREO_Circuit();
}

ComponentInstanceCircuit::CircuitLevel_ptr ComponentInstanceCircuit::getBaseLevel() {
  return base;
}

unsigned int ComponentInstanceCircuit::countConfigurations() const {
  return configurations.size();
}

ComponentInstanceCircuit::CircuitLevel_ptr ComponentInstanceCircuit::getConfiguration(unsigned int index) {
  return configurations.at(index);
}

StateDescription::ptr ComponentInstanceCircuit::getConfigState() const {
  return config_state;
}

PortDescription::ptr ComponentInstanceCircuit::getConfigPort() const {
  return config_port;
}

bool ComponentInstanceCircuit::hasInitialConfiguration() const {
  return (countConfigurations()>0) && (getWithParam("initial_topo")!="");
}
 
unsigned int ComponentInstanceCircuit::getInitialConfiguration() const {
  CHECK_SUCCESS(hasInitialConfiguration());
  try {
    unsigned int id=boost::lexical_cast<unsigned int>(getWithParam("initial_topo"));

    if (id>=countConfigurations()) {
      throw "Invalid topology ID: "+boost::lexical_cast<string>(id);
    }
    return id;
  } catch (boost::bad_lexical_cast&) {
    throw "Invalid topology ID: "+getWithParam("initial_topo");
  }
}

SymbComponentInstance::ptr ComponentInstanceCircuit::getSymbAutomaton(ComponentInstance::ptr self,
                                                                      Symb_DataDomain::ptr s_dd) {
  ERROR_CHECK_ASSERT(self.get()==this);

  ComponentInstanceCircuit::ptr cic=boost::static_pointer_cast<ComponentInstanceCircuit>(self);
  
  SymbCircuit::ptr ca_c(new SymbCircuit(cic, s_dd));
  return ca_c;
}

void ComponentInstanceCircuit::execute() {
  const vparser::pt_node& parse_tree=getCircuitConnector()->behavior->parse_tree;
  
  const string& component=getCircuitConnector()->name;

  execution_environment_t env(access_weak_ptr(getCircuitConnector()->context)->getGlobalSymbols(), 
                              getInterface()->getLocalSymbols(), 
                              access_weak_ptr(getCircuitConnector()->context));
  execution_environment_t::value_table_ptr values=env.getValueTable();

  typedef vparser::circuit_helper ch;

  // parameters
  if (ch::t_circuit_def::hasParameterDef(parse_tree)) {
    vector<string> parameter_def=vparser::common_helper::t_parameter_def::getIdentifiers(ch::t_circuit_def::getParameterDef(parse_tree));
    
    if (parameter_def.size()!=parameters.size()) {
      throw string("Parameter mismatch in instantiation: ")+component+ " expects "+
        boost::lexical_cast<string>(parameter_def.size())+" parameters, but "+
        boost::lexical_cast<string>(parameters.size())+" were supplied!";
    }
    
    for (unsigned int i=0;i<parameter_def.size();i++) {
      string p_name=parameter_def.at(i);
      Value::ptr p=parameters.at(i);
      
      if (!p) {
        throw "NULL value in parameters!";
      }
      
      // check that value has not been assigned previously
      if (env.VarExists(p_name) && !env.VarIsNull(p_name, 0)) {
        throw "Multiple parameters with the same name '"+p_name+"'";
      }
      
      switch (p->type->type) {
      case VALUE_TYPE_INT:
        // ok
        env.setValue(p_name, 0, p);
        getInterface()->getLocalSymbols()->replace(p_name, 
                                                   ConstantData::ptr(new ConstantData(p_name,
                                                                                      IntValue::getIntValue(p))));
        break;
      case VALUE_TYPE_BOOL:
        THROW_UNIMPLEMENTED();
      default:
        throw "Can not pass parameter of type "+p->type->typeToString()+" during instantiation";
      }
    }
  } else {
    // no parameter def
    if (parameters.size()!=0) {
      throw string("Parameter mismatch in instantiation: ")+component+ " does not expect parameters!";
    }
  }

  

  setTypeName(component+getParametersAsString(parameters));
  getInterface()->setTypeName(getTypeName());

  base=CircuitLevel::ptr(new CircuitLevel(*circuit_connector, values));

  const vparser::pt_node& circuit_body=ch::t_circuit_def::getCircuitBody(parse_tree);
  base->execute(ch::t_circuit_body::getBaseStatementBlock(circuit_body),
                getInterface()->getLocalSymbols());
  base->set_parent(this);

  for (unsigned int c=0; c < ch::t_circuit_body::countConfigurations(circuit_body); c++) {
    CircuitLevel::ptr level(base->clone(c));
    level->execute(ch::t_configuration_block::getStatementBlock(ch::t_circuit_body::getConfigurationBlock(circuit_body, c)),
                   getInterface()->getLocalSymbols());
    level->set_parent(this);

    level->resetValueTable(base->getValueTable());

    configurations.push_back(level);
  }

  // attach the atomic propositions attached to the REO_Circuit
  attachAtomicPropositions(base->getREO_Circuit()->getAtomicPropositions());

  // attach the atomic propositions specified for this type
  attachAtomicPropositionsForType();
}

void ComponentInstanceCircuit::build_name_map() {
  base->build_name_map();
  for (unsigned int c=0; c < configurations.size(); c++) {
    configurations.at(c)->build_name_map();
  }
}


void ComponentInstanceCircuit::plug(bool strict) {
  if (configurations.size()>0) {
    strict=true;
  }

  base->plug(strict);
  for (unsigned int c=0; c < configurations.size(); c++) {
    configurations.at(c)->plug(strict);
  }
}

void ComponentInstanceCircuit::build_interface(bool strict) {
  if (configurations.size()>0) {
    strict=true;

    IntRangeDataDomain::ptr dd(new IntRangeDataDomain(0, configurations.size()-1));
    config_state=StateDescription::ptr(new StateDescription("TOPO", dd));
    getInterface()->addStateDescription(config_state);
  }

  base->build_interface(getInterface(), strict);
  for (unsigned int c=0; c < configurations.size(); c++) {
    configurations.at(c)->build_interface(getInterface(), strict);
  }

  // TODO: change to real data domain when there is support for multiple data domains
  DataDomain::ptr data_domain=access_weak_ptr(circuit_connector->context)->getDataDomain();
  config_port=PortDescription::ptr(new PortDescription("RECONF_PORT",
                                                       PortDescription::IN_PORT,
                                                       data_domain));
  if (configurations.size()>0) {
    getInterface()->addPort(config_port);
  }
}

void ComponentInstanceCircuit::toReoXML(ostream& out, const std::vector<std::string>& input_files) {
  if (configurations.size()>0) {
    throw "Can not convert with multiple configurations to REO diagram";
  } else {
    base->toReoXML(out, getTypeName(), input_files);
  }
}


void ComponentInstanceCircuit::toDot(ostream& out) {
  if (configurations.size()>0) {
    out << "digraph {\n";
    for (unsigned int i=0;i<configurations.size();i++) {
      out << "subgraph cluster_" << i << "{\n";
      configurations.at(i)->toDot(out, true);
      out << "\ngraph [label=\"TOPO = " << i << "\"];\n";
      out << "}\n";
    }
    out << "}\n";
  } else {
    base->toDot(out);
  }
}



void ComponentInstanceCircuit::export_circuit(ostream& stream) const {
  base->export_circuit(stream);
}


bool ComponentInstanceCircuit::hasReconfigurationPort() const {
  return (configurations.size()>0);
}

unsigned int ComponentInstanceCircuit::getReconfigurationPortSourceIndex() const {
  if (hasReconfigurationPort()) {
    return getInterface()->getInPortCount()-1;
  }
  throw "error";  
}

void ComponentInstanceCircuit::calculatePortPartitions(PortPartitions& port_partitions) {
  if (configurations.size()>0) {
    // TODO: calculate port partitions for reconfigurable circuits...
    return;
  }

  base->calculatePortPartitions(port_partitions);
}

}
