/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "symb_circuit.hpp"
#include "symb_configurator.hpp"
#include "vereofy/system/symb_automaton/symb_node.hpp"
#include "vereofy/common/errors.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/system/symb_automaton/var_positions.hpp"
#include "vereofy/system/circuit_connector/circuit_level.hpp"

namespace vereofy {

SymbCircuit::SymbCircuit(ComponentInstanceCircuit::ptr ci_circuit_,
                         Symb_DataDomain::ptr dataDomain) :
  SymbComponentInstance(ci_circuit_->getInterface(), dataDomain),
  first(true) {

  setComponentInstance(ci_circuit_);
}

struct config_functions {
  BDDFunction delta;
  BDDFunction csPorts;
  BDDFunction csDCVariables;
  BDDFunction csTransitionLabelVariables;

  SymbAutomaton::symb_ce_set symb_ces;
  SymbAutomaton::symb_tl_set symb_tls;
};
  
void SymbCircuit::build_automaton(BuildContext::ptr build_context) {
  if (build_automaton_with_serialized_varorder(build_context)) {
    return;
  }  

  TRACE("SymbCircuit", "build", cout << "Start building the circuit\n");

  // store build context
  this->build_context=build_context;

  ComponentInstanceCircuit::ptr cic=boost::static_pointer_cast<ComponentInstanceCircuit>(info->component_instance);

  TRACE("building", "circuit",                                        \
        std::cout << "Build automaton [CIRCUIT], ";                   \
        if (info->component_instance) {                               \
          std::cout << info->component_instance->getFQName() <<       \
            " (" <<info->component_instance->getTypeName() << ")";    \
        }                                                             \
        std::cout << endl;                                            \
        );

  vector<CircuitLevel::ptr> levels;
  levels.push_back(cic->getBaseLevel());
  for (unsigned int c=0; c < cic->countConfigurations(); c++) {
    levels.push_back(cic->getConfiguration(c));
  }

  bool no_configurations=(levels.size()==1);

  config_functions base_functions;
  std::vector<config_functions> level_functions;

  //TODO find a way to generate a better variable ordering, this variable is never used
  VarPositions::ptr var_positions = VarPositions::ptr(new VarPositions());

  for (unsigned int l=0; l<levels.size(); l++) {
    REO_Circuit::ptr reo_circuit=levels.at(l)->getREO_Circuit();
    vector<REO_Graph_Object::ptr> allObjects = reo_circuit->getAllComponentsInDFS();
    
    TRACE_REO_CIRCUIT_DOT("SymbCircuit.build", "reo_circuit_DOT", *reo_circuit);

    TRACE("building", "circuit",                                \
          std::cout << " Building level = " << l << endl;       \
          );

    if (!no_configurations) {
      if (l>0) {
        // reset for configuration...
        delta=base_functions.delta;
        csPorts=base_functions.csPorts;
        csDCVariables=base_functions.csDCVariables;
        csTransitionLabelVariables=base_functions.csTransitionLabelVariables;
        symb_ces=base_functions.symb_ces;
        symb_tls=base_functions.symb_tls;
      }
    }

    for (unsigned int i=0;i<allObjects.size();i++) {
      REO_Graph_Object::ptr rgo=allObjects.at(i);
      TRACE("SymbCircuit", "build", cout << i << " : " << rgo->get_Description() << endl);
      
      if (no_configurations) {
        switch (rgo->get_graph_object_type()) {
        case REO_Graph_Object::REO_COMPONENT: {
          REO_Component::ptr component=boost::static_pointer_cast<REO_Component>(rgo);
          
          build_component(reo_circuit, component, build_context);
          plug_component(reo_circuit, component);
         
          VarManager::getInstance()->selectiveReorderingHook(2);
    
          break;      
        }
        case REO_Graph_Object::REO_NODE: {
        REO_Node::ptr node=boost::static_pointer_cast<REO_Node>(rgo);
        
        build_node(reo_circuit, node, build_context);

        VarManager::getInstance()->selectiveReorderingHook(2);

        break;      
        } 
        default:
          THROW_UNIMPLEMENTED();
        }
      } else {
        // multiple configurations

        if (l==0) {
          // base level
          if (rgo->get_graph_object_type()==REO_Graph_Object::REO_COMPONENT) {
            REO_Component::ptr component=boost::static_pointer_cast<REO_Component>(rgo);
            build_component(reo_circuit, component, build_context);

            VarManager::getInstance()->selectiveReorderingHook(2);
            
            // note: don't plug components here!
          } else {
            // don't build nodes in the base level if we have multiple configurations
          }
        } else {
          if (rgo->wasCloned()) {
            // this object also exists in the base level
            if (rgo->get_graph_object_type()==REO_Graph_Object::REO_NODE) {
              REO_Node::ptr node=boost::static_pointer_cast<REO_Node>(rgo);
              build_node(reo_circuit, node, build_context);

              VarManager::getInstance()->selectiveReorderingHook(2);
            } else {
              // component, should be already built in the base level
              REO_Component::ptr component=boost::static_pointer_cast<REO_Component>(rgo);
              plug_component(reo_circuit, component);

              VarManager::getInstance()->selectiveReorderingHook(2);
            }
          } else {
            // this is an object only in this configuration

            switch (rgo->get_graph_object_type()) {
            case REO_Graph_Object::REO_COMPONENT: {
              REO_Component::ptr component=boost::static_pointer_cast<REO_Component>(rgo);
              build_component(reo_circuit, component, build_context);

              VarManager::getInstance()->selectiveReorderingHook(2);
              break;      
            }
            case REO_Graph_Object::REO_NODE: {
              REO_Node::ptr node=boost::static_pointer_cast<REO_Node>(rgo);
              build_node(reo_circuit, node, build_context);

              VarManager::getInstance()->selectiveReorderingHook(2);
              break;      
            } 
            default:
              THROW_UNIMPLEMENTED();
            }
          }
        }
      }  


      CHECK(5, check_symb_ce_sanity(), "");
      TRACE_SYMB_COMPONENT_DOT("SymbCircuit.build", "product_dot", *this, false);
      TRACE_SYMB_COMPONENT_PATHS("SymbCircuit.build", "product_path", *this);
    }

    if (!no_configurations) {
      if (l==0) {
        // store base 
        base_functions.delta=delta;
        base_functions.csPorts=csPorts;
        base_functions.csDCVariables=csDCVariables;
        base_functions.csTransitionLabelVariables=csTransitionLabelVariables;
        base_functions.symb_ces=symb_ces;
        base_functions.symb_tls=symb_tls;
      } else {
        // store level
        level_functions.push_back(config_functions());
        config_functions& functions=level_functions.back();
        
        functions.delta=delta;
        functions.csPorts=csPorts;
        functions.csDCVariables=csDCVariables;
        functions.csTransitionLabelVariables=csTransitionLabelVariables;
        functions.symb_ces=symb_ces;
        functions.symb_tls=symb_tls;
      }
    }
  }

  if (!no_configurations) {
    // merge everything
    delta=BDDFunction::ZERO();
    csPorts=BDDFunction::ONE();
    csDCVariables=BDDFunction::ONE();
    csTransitionLabelVariables=BDDFunction::ONE();
    symb_ces.clear();
    symb_tls.clear();

    Symb_Channel_End::ptr config_port_ce_init;
    // check whether the config port has already a SymbChannelEnd, use that
    if (info->port_to_bdd.count(cic->getConfigPort())>0) {
      config_port_ce_init=info->port_to_bdd[cic->getConfigPort()];
    }

    SymbConfigurator configurator(dataDomain,
                                  cic->getConfigPort(),
                                  cic->getConfigState(),
                                  config_port_ce_init);

    SymbolicState::ptr config_state=configurator.getConfigState();
    Symb_FlatDataDomain::ptr config_state_dd=boost::dynamic_pointer_cast<Symb_FlatDataDomain>(config_state->data_domain);

    for (unsigned int l=0;l<level_functions.size();l++) {
      
      BDDFunction config_state_restriction=pairwiseEQUIV(*config_state->vars.X, 
                                                         config_state_dd->getEncoding(l));
      BDDFunction delta_with_restriction=level_functions.at(l).delta & config_state_restriction;
      delta|=delta_with_restriction;

      TRACE_BDD_FUNCTION("SymbCircuit.merge_levels", level_functions.at(l).delta, l);
      TRACE_BDD_FUNCTION("SymbCircuit.merge_levels", config_state_restriction, l);
      TRACE_BDD_FUNCTION("SymbCircuit.merge_levels", delta_with_restriction, l);
      TRACE_BDD_FUNCTION("SymbCircuit.merge_levels", delta, l);

      csPorts&=level_functions.at(l).csPorts;
      TRACE_BDD_FUNCTION("SymbCircuit.merge_levels", level_functions.at(l).csPorts, l);
      TRACE_BDD_FUNCTION("SymbCircuit.merge_levels", csPorts, l);

      csDCVariables&=level_functions.at(l).csDCVariables;
      csTransitionLabelVariables&=level_functions.at(l).csTransitionLabelVariables;
      
      symb_ces.insert(level_functions.at(l).symb_ces.begin(), 
                      level_functions.at(l).symb_ces.begin());

      symb_tls.insert(level_functions.at(l).symb_tls.begin(), 
                      level_functions.at(l).symb_tls.begin());
    }

    TRACE_BDD_FUNCTION("SymbCircuit.merge_levels", delta, "before configurator product");
    TRACE_BDD_FUNCTION("SymbCircuit.merge_levels", csPorts, "before configurator product");
    TRACE_BDD_FUNCTION("SymbCircuit.merge_levels", csOriginalX, "before configurator product");

    product(configurator);

    addPort(cic->getConfigPort(), configurator.getConfigPort());
    addState(cic->getConfigState(), configurator.getConfigState());

    if (cic->hasInitialConfiguration()) {
      initialStates&=pairwiseEQUIV(*config_state->vars.X, 
                                   config_state_dd->getEncoding(cic->getInitialConfiguration()));
    }

    TRACE_BDD_FUNCTION("SymbCircuit.merge_levels", delta, "after configurator product");
    TRACE_BDD_FUNCTION("SymbCircuit.merge_levels", csPorts, "after configurator product");
    TRACE_BDD_FUNCTION("SymbCircuit.merge_levels", csOriginalX, "after configurator product");
  }

  CHECK(5, check_symb_ce_sanity(), "");
  
#define HIDE_ANONYMOUS
#ifdef HIDE_ANONYMOUS
  TRACE("SymbCircuit", "build", cout << "Hiding anonymous channel ends\n");
  
  TRACE_SYMB_COMPONENT_DOT("SymbCircuit.build", "circuit_before_hide_dot", *this, false);
  TRACE_SYMB_COMPONENT_PATHS("SymbCircuit.build", "circuit_before__hide_path", *this);
  
  hideAnonymous(build_context);

  CHECK(5, check_symb_ce_sanity(), "");
#endif


  TRACE("building", "circuit",                                          \
        std::cout << "Calculate atomic propositions [CIRCUIT]" << endl;);
  // calculate the generic atomic propositions after building the automaton
  calculateAtomicPropositions();  
  
  TRACE_SYMB_COMPONENT_DOT("SymbCircuit.build", "circuit_dot", *this, false);
  TRACE_SYMB_COMPONENT_PATHS("SymbCircuit.build", "circuit_dot_path", *this);

  // @TODO: cache!
  if (build_context->param_constrain_to_reachable_circuits>=1){
    if (build_context->param_constrain_to_reachable_circuits>=2){
      TRACE("building", "RSL", std::cout << " Compute reachable part of the system"<< endl;);
      reachabilityConstrain = calcReachableStates();
    }
    TRACE("building", "RSL", std::cout << " Constrain to reachable part"<< endl;);
    delta = delta.constrain(reachabilityConstrain);
  }

  if (build_context->param_abstract_data_domain) {
    info->abstract_port_data_domain=true;
  }
	
  TRACE("building", "circuit",                                          \
        std::cout << "Finished building automaton [CIRCUIT]" << endl;);
}

void SymbCircuit::build_component(REO_Circuit::ptr reo_circuit, REO_Component::ptr component,
                                  BuildContext::ptr build_context) {
  port_to_bdd_map bdd_ports_init;
  
  for (unsigned int i=0;i<component->getSourceCECount();i++) {
    REO_Channel_End::ptr ce=component->getSourceCE(i);
    
    PortDescription::ptr pd_in_circuit=ce->get_PortDescription_Circuit();
    if (info->port_to_bdd.count(pd_in_circuit)>0) {
      bdd_ports_init[ce->get_PortDescription_Component()]=
        info->port_to_bdd[pd_in_circuit];
    }
  }
  
  for (unsigned int i=0;i<component->getSinkCECount();i++) {
    REO_Channel_End::ptr ce=component->getSinkCE(i);
    
    PortDescription::ptr pd_in_circuit=ce->get_PortDescription_Circuit();
    if (info->port_to_bdd.count(pd_in_circuit)>0) {
      bdd_ports_init[ce->get_PortDescription_Component()]=
        info->port_to_bdd[pd_in_circuit];
    }
  }
        
  ComponentInstance::ptr ci=component->getComponentInstance();
  
  //TODO find a way to generate a better variable ordering: use var_positions as parameter
  SymbComponentInstance::ptr sci=ci->getSymbAutomaton(ci, dataDomain);
  
  sci->set_var_info(bdd_ports_init,
                    state_to_bdd_map(),
                    VarPositions::ptr(new VarPositions()));

  sci->build_automaton(build_context);
  VarManager::getInstance()->selectiveReorderingHook(3);

  sci->hideInternalPorts(build_context);
  VarManager::getInstance()->selectiveReorderingHook(2);

  TRACE("SymbCircuit", "build", cout << "Subcomponent " << ci->getName() << endl);
  TRACE_SYMB_COMPONENT_DOT("SymbCircuit.build", "component_dot", *sci, false);
  TRACE_SYMB_COMPONENT_PATHS("SymbCircuit.build", "component_path", *sci);
  
  addSubcomponentAutomaton(sci);
  addSubcomponentInfo(sci->info);
  addSubcomponentRestrictions(sci);
  
  port_to_bdd_map& bdd_ports_in_component=sci->info->port_to_bdd;
  
  TRACE("SymbCircuit", "build", cout << "VARORDER: size Delta new Component: " << sci->getDelta().size() << endl
        << "VARORDER: " << BDDVarOrder::getVarOrder() << endl << "VARORDER: ";
        sci->getDelta().printEssentialVariables() );
  if (first) {
    TRACE("SymbCircuit", "build", cout << "Copy (component)\n");
    delta=sci->getDelta();
    initialStates=sci->getInitialStates();
    csPorts&=sci->getPortVars();
    csDCVariables&=sci->getDataConstraintVars();
    csTransitionLabelVariables&=sci->getTransitionLabelVars();
    csOriginalX&=sci->getOriginalStateVars();
    csCopyY&=sci->getCopyStateVars();
    csCopyCopyZ&=sci->getCopyCopyStateVars();
    haltSystem&=sci->getHaltSystem();

    copy_symb_tls(*sci);

    first=false;
  } else {
    TRACE("SymbCircuit", "build", cout << "Building product (component) " << ci->getFQName() << "\n");          
    TRACE("building", "circuit", std::cout << " PRODUCT (component) " << ci->getFQName() << "\n");

    product(*sci);
    TRACE("SymbCircuit", "build", cout << "size Delta circuit after product: " << getDelta().size() << endl);
  }
  
  for (unsigned int i=0;i<component->getSourceCECount();i++) {
    REO_Channel_End::ptr ce=component->getSourceCE(i);
    
    PortDescription::ptr pd_in_circuit=ce->get_PortDescription_Circuit();
    addPort(pd_in_circuit, bdd_ports_in_component[ce->get_PortDescription_Component()]);
  }
  
  for (unsigned int i=0;i<component->getSinkCECount();i++) {
    REO_Channel_End::ptr ce=component->getSinkCE(i);
    
    PortDescription::ptr pd_in_circuit=ce->get_PortDescription_Circuit();
    addPort(pd_in_circuit, bdd_ports_in_component[ce->get_PortDescription_Component()]);
  }
  
  CHECK(5, check_symb_ce_sanity(), "");
  
  if (build_context->param_constrain_to_reachable_circuits==1){
    TRACE("building", "RSL", std::cout << " Combine reachablility constraints of sub-subcomponents"<< endl;);
    reachabilityConstrain &= sci->getReachabilityConstrain();
  }
  
}

void SymbCircuit::plug_component(REO_Circuit::ptr reo_circuit, REO_Component::ptr component) {
  BDDFunction plug_restriction=BDDFunction::ONE();


  for (unsigned int i=0;i<component->getSourceCECount();i++) {
    REO_Channel_End::ptr ce=component->getSourceCE(i);
    
    PortDescription::ptr pd_in_circuit=ce->get_PortDescription_Circuit();
    
    if (ce->isPlugged()) {
      TRACE("SymbCircuit", "build", cout << "Plugging source channel end " << i << " of " << component << "\n");          
      // inhibit data flow on plugged channel end
      plug_restriction&=!get_symb_ce_for_port(pd_in_circuit)->getPortActive();
    }
  }
  
  for (unsigned int i=0;i<component->getSinkCECount();i++) {
    REO_Channel_End::ptr ce=component->getSinkCE(i);
    
    PortDescription::ptr pd_in_circuit=ce->get_PortDescription_Circuit();
    
    if (ce->isPlugged()) {
      TRACE("SymbCircuit", "build", cout << "Plugging sink channel end " << i << " of " << component << "\n");          
      // inhibit data flow on plugged channel end
      plug_restriction&=!get_symb_ce_for_port(pd_in_circuit)->getPortActive();
    }
  }


  if (plug_restriction!=BDDFunction::ONE()) {
    delta&=plug_restriction;
  }
}

  void SymbCircuit::build_node(REO_Circuit::ptr reo_circuit, 
                               REO_Node::ptr node, 
                               BuildContext::ptr build_context,
                               BDDFunction augmentation) {
  vector<REO_Channel_End::ptr> in_ces=reo_circuit->getConnectedSinkChannelEnds(node);
  vector<REO_Channel_End::ptr> out_ces=reo_circuit->getConnectedSourceChannelEnds(node);
  
  vector<PortDescription::ptr> in_pd;
  vector<PortDescription::ptr> out_pd;
  
  // used to check whether there are connected components with more than one connected port
  bool multiple_connections=false;
  typedef shared_ptr_set<REO_Component::ptr>::type connected_component_map;
  connected_component_map connected_components;

  TRACE("building", "node",                                             \
        std::cout << "Build node (" << in_ces.size() << "," << out_ces.size() << ") " \
        << (info->component_instance->getFQName()!=""                   \
            ? (info->component_instance->getFQName()+"."+node->getPortDescription()->getName()) \
            : (node->getPortDescription()->getName())) << "\n");
  
  for (unsigned int i=0;i<in_ces.size();i++) {
    PortDescription::ptr pd=in_ces.at(i)->get_PortDescription_Circuit();
    in_pd.push_back(pd);

    REO_Component::ptr c=reo_circuit->getComponentFor(in_ces.at(i));
    if (connected_components.count(c)>0) {
      // this component was already connected via another port
      multiple_connections=true;
    }
    connected_components.insert(c);
  }

  for (unsigned int i=0;i<out_ces.size();i++) {
    PortDescription::ptr pd=out_ces.at(i)->get_PortDescription_Circuit();
    out_pd.push_back(pd);

    REO_Component::ptr c=reo_circuit->getComponentFor(out_ces.at(i));
    if (connected_components.count(c)>0) {
      // this component was already connected via another port
      multiple_connections=true;
    }
    connected_components.insert(c);
  }

  bool share_channel_ends=true;
#ifdef SYMB_BUILD_CONSERVATIVE
  // share channel ends only if there are no multiple connections
  share_channel_ends=!multiple_connections;
#endif

  PortDescription::ptr node_port=node->getPortDescription();
  
  //TODO find a way to generate a better variable ordering: use var_positions as parameter
  SymbNode::ptr s_node(new SymbNode(boost::static_pointer_cast<SymbCircuit>(this->shared_from_this()),
                                    dataDomain,
                                    node->getNodeSemantic(),
                                    in_pd,
                                    out_pd,
                                    node_port,
                                    share_channel_ends,
                                    VarPositions::ptr(new VarPositions()),
                                    build_context));

  s_node->addAugmentation(augmentation);
  
  TRACE_SYMB_COMPONENT_DOT("SymbCircuit.build", "node_dot", *s_node, false);
  TRACE_SYMB_COMPONENT_PATHS("SymbCircuit.build", "node_path", *s_node);
  
  TRACE("SymbCircuit", "build", cout << "VARORDER: size Delta new Node: " << s_node->getDelta().size() << endl
        << "VARORDER: " << BDDVarOrder::getVarOrder() << endl << "VARORDER: ";
        s_node->getDelta().printEssentialVariables() );

  
  if (first) {
    TRACE("SymbCircuit", "build", cout << "Copy (node)\n");
    
    delta=s_node->getDelta();
    initialStates=s_node->getInitialStates();
    csPorts&=s_node->getPortVars();
    csDCVariables&=s_node->getDataConstraintVars();
    csTransitionLabelVariables&=s_node->getTransitionLabelVars();
    csOriginalX&=s_node->getOriginalStateVars();
    csCopyY&=s_node->getCopyStateVars();
    csCopyCopyZ&=s_node->getCopyCopyStateVars();
    haltSystem&=s_node->getHaltSystem();
    
    copy_symb_tls(*s_node);

    first=false;
  } else {
    TRACE("SymbCircuit", "build", cout << "Building product (node) " << node->getPortDescription()->getName() <<"\n");
    TRACE("building", "node", std::cout << " PRODUCT (node) " <<        \
          (info->component_instance->getFQName()!=""                    \
           ? (info->component_instance->getFQName()+"."+node->getPortDescription()->getName()) \
           : (node->getPortDescription()->getName())) << "\n");

    product(*s_node);
    TRACE("SymbCircuit", "build", cout << "size Delta circuit after product: " << getDelta().size() << endl);
  }
  
  port_to_bdd_map& bdd_ports_in_node=s_node->info->port_to_bdd;
  
  for (unsigned int i=0;i<in_pd.size();i++) {
    PortDescription::ptr pd_in_node=s_node->info->c_interface->getInPort(i);
    addPort(in_pd.at(i), bdd_ports_in_node[pd_in_node]);
  }
  
  for (unsigned int i=0;i<out_pd.size();i++) {
    PortDescription::ptr pd_in_node=s_node->info->c_interface->getOutPort(i);
    addPort(out_pd.at(i), bdd_ports_in_node[pd_in_node]);
  }
  
  // add port_to_bdd for node port description
  addPort(node->getPortDescription(), s_node->getCEforPort());
  
  
  if (node->isSourcePlugged() || node->isSinkPlugged()) {
    TRACE("SymbCircuit", "build", cout << "Plugging node " << node->getPortDescription()->getName() << "\n"); 
    // inhibit data flow on plugged channel end
    delta&=!get_symb_ce_for_port(node->getPortDescription())->getPortActive();
  }
  
  CHECK(5, check_symb_ce_sanity(), "");
}


void SymbCircuit::hideAnonymous(BuildContext::ptr build_context) {
  typedef shared_ptr_id_set<Symb_Channel_End::ptr>::type ce_set;
  ce_set ces;
  ce_set non_anonymous_ces;
  
  for (ce_to_port_map::const_iterator it=info->ce_to_port.begin();
       it!=info->ce_to_port.end();
       ++it) {
    Symb_Channel_End::ptr ce=it->first;
    PortDescription::ptr pd=it->second;
    
    if (debugLevel>=6) {
      cout << &info->ce_to_port << " : CE " << ce <<endl;
      cout << &info->ce_to_port << " : PD " << pd->getName() <<endl;
    }
    
    if (!ce->isVisible()) {
      // already hidden
      if (debugLevel>=6) {
        cout << "already hidden\n";
      }
      continue;
    }
    
    ces.insert(ce);
    
    if (!pd->isAnonymous()) {
      // ok, found one non-anonymous reference
      if (debugLevel>=6) {
        cout << "non anonymous\n";
      }
      non_anonymous_ces.insert(ce);
    }
  }
  
  std::vector<Automaton_Channel_End::ptr> to_hide;

  for (ce_set::iterator it=ces.begin();
       it!=ces.end();
       ++it) {
    if (non_anonymous_ces.count(*it)>0) {
      // skip
      if (debugLevel>=6) {
        cout << "skip " << *it << endl;
      }
    } else {
      Symb_Channel_End::ptr ce=*it;
      if (debugLevel>=6) {
        cout << "hide " << ce << endl;
      }
      
      to_hide.push_back(ce);
    }
  }

  TRACE_SYMB_COMPONENT_DOT("SymbCircuit.hideAnonymous", "hide", *this, false);
  if (build_context->param_retain_hidden_ports ||
      (build_context->param_retain_top_level_ports && (!info->component_instance->getParent()))) {
    // retain all ports, or just the toplevel ports and we are the top level
    hide_semi(to_hide);
  } else {
    hide(to_hide);

    for (unsigned int i=0;i<to_hide.size();i++) {
      Symb_Channel_End::ptr ce=boost::static_pointer_cast<Symb_Channel_End>(to_hide.at(i));
      ce->hide();
      info->ce_to_port.erase(ce);
    }
  }
}
  

void SymbCircuit::hideInternalPorts(BuildContext::ptr build_context) {
  ComponentInstanceCircuit::ptr cic=boost::static_pointer_cast<ComponentInstanceCircuit>(info->component_instance);

  vector<CircuitLevel::ptr> levels;
  CircuitLevel::ptr base=cic->getBaseLevel();
  levels.push_back(base);
  for (unsigned int i=0;i<cic->countConfigurations();i++) {
    levels.push_back(cic->getConfiguration(i));
  }
  
  for (unsigned int i=0;i<levels.size();i++) {
    CircuitLevel::ptr cur_level=levels.at(i);

    REO_Circuit::ptr reo_circuit=cur_level->getREO_Circuit();
    vector<REO_Node::ptr> nodes=reo_circuit->getAllNodes();
  
    for (unsigned int j=0;j<nodes.size();j++) {
      REO_Node::ptr node=nodes.at(j);

      if (node->wasCloned() && i!=0) {
        // was cloned and we are not in base
        continue;
      }

      if (base->is_exported_sink(node) ||
          base->is_exported_source(node)) {
        // don't hide
      } else {
        node->setHidden(true);
        node->getPortDescription()->setAnonymous(true);
      }
    }

    vector<REO_Channel_End::ptr> ces=reo_circuit->getAllCEs();
    for (unsigned int j=0;j<ces.size();j++) {
      REO_Channel_End::ptr ce=ces.at(j);
      
      if (ce->wasCloned() && i!=0) {
        // was cloned and we are not in base
        continue;
      }

      if (base->is_exported_sink(ce) ||
          base->is_exported_source(ce)) {
        // don't hide
      } else {
        ce->get_PortDescription_Circuit()->setAnonymous(true);
      }
    }
  }
    
  hideAnonymous(build_context);
}


}
