/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/



#include <boost/algorithm/string.hpp>
#include <boost/tuple/tuple.hpp>
#include "vereofy/common/errors.hpp"
#include "vereofy/common/weak_ptr_helper.hpp"
#include "vereofy/common/xml_helpers.hpp"
#include "vereofy/common/weak_ptr_helper.hpp"
#include "component_instance.hpp"
#include "vereofy/system/symb_automaton/symb_component_instance.hpp"

namespace vereofy {

ComponentInstance::ComponentInstance(component_type_t component_type_,
				     SystemContext::wptr system_context_,
				     ComponentInterface::ptr c_interface_) :
  component_type(component_type_),
  system_context(system_context_),
  c_interface(c_interface_),
  parent(0) {
}

component_type_t ComponentInstance::getType() const {
  return component_type;
}

ComponentInterface::ptr ComponentInstance::getInterface() const {
  return c_interface;
}

SystemContext::wptr ComponentInstance::getSystemContext() const {
  return system_context;
}

void ComponentInstance::setInterface(ComponentInterface::ptr c_interface_) {
  c_interface=c_interface_;
}


void ComponentInstance::setName(const string& name_) {
  name=name_;
}

const string& ComponentInstance::getName() const {
  return name;
}


void ComponentInstance::setTypeName(const string& name) {
  type_name=name;
}

const string& ComponentInstance::getTypeName() const {
  return type_name;
}


string ComponentInstance::getParametersAsString(const std::vector<Value::ptr>& parameters) const {
  if (parameters.size()==0) {
    return "";
  }

  stringstream out;
  out << "<";

  for (unsigned int i=0;i<parameters.size();i++) {
    if (i>0) {
      out << ",";
    }
    Value::ptr v=parameters.at(i);
      
    if (!v) {
      throw "NULL value in parameters!";
    }
    
    switch (v->type->type) {
    case VALUE_TYPE_INT:
      out << IntValue::getIntValue(v);
      break;
    case VALUE_TYPE_BOOL: 
      THROW_UNIMPLEMENTED();
    default:
      THROW_UNIMPLEMENTED();
    }
  }
  out << ">";
  return out.str();
}

void ComponentInstance::addWithParameter(const std::string& key_, const std::string& value) {
  string key=key_;
  boost::to_lower(key);
  with_params.insert(make_pair(key, value));
}

std::string ComponentInstance::getWithParam(const std::string& key) const {
  with_params_map::const_iterator it=with_params.find(key);
  if (it!=with_params.end()) {
    return it->second;
  } else {
    return "";
  }
}

void ComponentInstance::print_interface_xml(std::ostream& out) const {
  getInterface()->print_xml(out, "");
}


bool ComponentInstance::hasReconfigurationPort() const {
  return false;
}

unsigned int ComponentInstance::getReconfigurationPortSourceIndex() const {
  THROW_UNIMPLEMENTED();
}


void ComponentInstance::setParent(ComponentInstance* parent_) {
  parent=parent_;
}

ComponentInstance* ComponentInstance::getParent() const {
  return parent;
}

ComponentInstance::ptr ComponentInstance::getSubcomponent(const std::string& name) const {
  subcomponent_map_t::const_iterator it=subcomponents.find(name);
  if (it==subcomponents.end()) {
    return ComponentInstance::ptr();
  }
  return it->second;
}

void ComponentInstance::registerSubcomponent(ComponentInstance::ptr ci) {
  CHECK_SUCCESS(ci->getName()!="");
  CHECK_SUCCESS((!getSubcomponent(ci->getName())) ||
                getSubcomponent(ci->getName())==ci);
  subcomponents.insert(make_pair(ci->getName(), ci));
}


std::string ComponentInstance::getFQName() const {
  if (!getParent()) {
    return getName();
  } else {
    std::string parent_name=getParent()->getFQName();
    if (parent_name!="") {
      return parent_name+ "." + getName();
    } else {
      return getName();
    }
  }
}

void ComponentInstance::attachAtomicPropositions(const AtomicPropositionSpec::vector& atomic_propositions) {
  for (unsigned int i=0;i<atomic_propositions.size();i++) {
    if (c_interface->hasAtomicProposition(atomic_propositions.at(i).name)) {
      throw std::string("Atomic proposition "+
                        atomic_propositions.at(i).name+
                        " defined multiple times for the given component");
    }

    c_interface->addAtomicProposition(atomic_propositions.at(i).name,
                                      atomic_propositions.at(i).specification);
  }
}


void ComponentInstance::attachAtomicPropositionsForType() {
  TypeModificationRegistry::ptr registry=access_weak_ptr(system_context)->getTypeModificationRegistry();
  attachAtomicPropositions(registry->getAtomicPropositionsForType(getTypeName()));
}


PortID ComponentInstance::lookupPort(const std::string& fq_port_name,
                                     bool exact) const {
  if (fq_port_name=="") {
    throw lookup_failure_exception("Lookup for port failed, empty name!");
  }


  if (fq_port_name.find(".")!=string::npos) {
    // we have subcomponent.suffix...
    
    std::string subcomponent, suffix;
    boost::tie(subcomponent, suffix) = splitAt(fq_port_name, '.');
    if (subcomponent=="") {
      std::string ci_name=getFQName();
      throw lookup_failure_exception("Lookup for port '"+fq_port_name+"' failed"
                                     +(ci_name!="" ? (" in "+ci_name) : "")+
                                     ", starts with a dot!");
    }
  
    ComponentInstance::ptr ci=getSubcomponent(subcomponent);
    if (!ci) {
      std::string ci_name=getFQName();
      throw lookup_failure_exception("Lookup for subcomponent "+subcomponent
                                     +(ci_name!="" ? (" in "+ci_name) : "")
                                     +" failed!");
    }
    return ci->lookupPort(suffix);
  } else {
    // we have PORT or PORT[..]
    std::string port_name = fq_port_name;
    std::string ci_name=getFQName();
    
    PortDescription::ptr port=getInterface()->getPortDescription(port_name);

    if (!port && !exact) {
      // exact name does not match, we may try variants
      
      std::string ci_name=getFQName();

      if (port_name.find('[')!=string::npos) {
        // has index
        
        string name_without_index, index_str;
        boost::tie(name_without_index, index_str)=splitAt(port_name, '[');
        
        if (name_without_index=="") {
          throw lookup_failure_exception("Lookup for port '"+fq_port_name
                                         +(ci_name!="" ? (" in "+ci_name) : "")
                                         +" failed, starts with a square bracket [ !");
        }
        
        if (!boost::algorithm::ends_with(index_str, "]")) {
          throw lookup_failure_exception("Lookup for port "+fq_port_name
                                         +(ci_name!="" ? (" in "+ci_name) : "")
                                         +" failed, missing ] at end!");
        }
        
        // remove ]
        index_str=index_str.substr(0, index_str.length()-1);

        unsigned int index=0;
        try {
          index=boost::lexical_cast<unsigned int>(index_str);
        } catch (boost::bad_lexical_cast& err) {
          throw lookup_failure_exception("Lookup for port "+fq_port_name
                                         +(ci_name!="" ? (" in "+ci_name) : "")
                                         +" failed, invalid index '"+index_str+"'!");
        }

        // has index 'index'
        if (index==0) {
          // index==0 -> try without
          port=getInterface()->getPortDescription(name_without_index);          
        } else {
          // nop, nothing to try 
        }
      } else {  // has no index
        // no index, try name[0]
        port=getInterface()->getPortDescription(port_name+"[0]");        
      } 
    }
    
    if (!port) {
        throw lookup_failure_exception("Lookup for port "+fq_port_name
                                       +(ci_name!="" ? (" in "+ci_name) : "")
                                       +" failed, does not exist!");
    }
    return PortID(const_cast<ComponentInstance*>(this), port);
  }
}
  
StateID ComponentInstance::lookupState(const std::string& fq_state_name) const {
  std::pair<std::string, std::string> split_string=splitAt(fq_state_name, '.');
  CHECK_SUCCESS(split_string.first!="");
  
  if (split_string.second!="") {
    // we have A.X
    ComponentInstance::ptr ci=getSubcomponent(split_string.first);
    if (!ci) {
      std::string ci_name=getFQName();
      throw lookup_failure_exception("Lookup for subcomponent "+split_string.first
                                     +(ci_name!="" ? (" in "+ci_name) : "")
                                     +" failed!");
    }
    return ci->lookupState(split_string.second);
  } else {
    // we have X
    StateDescription::ptr state=getInterface()->getStateDescription(split_string.first);
    if (!state) {
      std::string ci_name=getFQName();
      throw lookup_failure_exception("Lookup for state "+split_string.first
                                     +(ci_name!="" ? (" in "+ci_name) : "")
                                     +" failed!");
    }
    return StateID(const_cast<ComponentInstance*>(this), state);
  }
}
  
ComponentInstance::ptr ComponentInstance::lookupComponent(const std::string& fq_component_name) const {
  std::pair<std::string, std::string> split_string=splitAt(fq_component_name, '.');
  CHECK_SUCCESS(split_string.first!="");
  
  ComponentInstance::ptr ci=getSubcomponent(split_string.first);
  if (!ci) {
    std::string ci_name=getFQName();
    throw lookup_failure_exception("Lookup for subcomponent "+split_string.first
                                   +(ci_name!="" ? (" in "+ci_name) : "")
                                   +" failed!");
  }
  if (split_string.second!="") {
    return ci->lookupComponent(split_string.second);
  }

  return ci;
}
  
}
