/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "vereofy/common/errors.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/common/weak_ptr_helper.hpp"
#include "vereofy/common/xml_helpers.hpp"
#include "component_interface.hpp"

using namespace std;

namespace vereofy {

  ComponentInterface::ComponentInterface(const std::string& type_name_) :
    local_symbols(new symbol_table_t()), type_name(type_name_) {}

  // returns the number of "in" ports 
  unsigned int ComponentInterface::getInPortCount() const {
    return in_ports.size();
  }
  
  // returns the number of "out" ports 
  unsigned int ComponentInterface::getOutPortCount() const {
    return out_ports.size();
  }

  // returns the number of "internal" ports 
  unsigned int ComponentInterface::getInternalPortCount() const {
    return internal_ports.size();
  }

  // returns the number of "anonymous" ports 
  unsigned int ComponentInterface::getAnonymousPortCount() const {
    return anonymous_ports.size();
  }
  
  // returns the number of state descriptions
  unsigned int ComponentInterface::getStateDescriptionCount() const {
    return state_descriptions.size();
  }

  // returns the number of transition label descriptions
  unsigned int ComponentInterface::getTransitionLabelDescriptionCount() const {
    return transition_label_descriptions.size();
  }

  // returns the number of state descriptions
  unsigned int ComponentInterface::getSubcomponentCount() const {
    return subcomponent_index.size();
  }
  
  // returns the description of the i-th "in" port
  PortDescription::ptr ComponentInterface::getInPort(unsigned int i) const {
    return in_ports.at(i);
  }
  
  // returns the description of the i-th "out" port
  PortDescription::ptr ComponentInterface::getOutPort(unsigned int i) const {
    return out_ports.at(i);
  }

  // returns the description of the i-th "internal" port
  PortDescription::ptr ComponentInterface::getInternalPort(unsigned int i) const {
    return internal_ports.at(i);
  }

  // returns the description of the i-th "anonymous" port
  PortDescription::ptr ComponentInterface::getAnonymousPort(unsigned int i) const {
    return anonymous_ports.at(i);
  }
  
  StateDescription::ptr ComponentInterface::getStateDescription(unsigned int i) const {
    return state_descriptions.at(i);
  }

  TransitionLabelDescription::ptr ComponentInterface::getTransitionLabelDescription(unsigned int i) const {
    return transition_label_descriptions.at(i);
  }
  
  unsigned int ComponentInterface::getIndexForPortDescription(PortDescription::ptr pd) const {
    const index_map* i_map = NULL;
    switch (pd->getType()) {
    case PortDescription::IN_PORT:
      i_map=&in_port_index;
      break;
    case PortDescription::OUT_PORT:
      i_map=&out_port_index;
      break;
    case PortDescription::INTERNAL_PORT:
      i_map=&internal_port_index;
      break;
    }
    
    index_map::const_iterator it=i_map->find(pd->getName());
    CHECK_SUCCESS(it!=i_map->end());

    return it->second;
  }

  PortDescription::ptr ComponentInterface::getPortDescription(const string& name) const {
    index_map::const_iterator it=
      in_port_index.find(name);
    
    if (it!=in_port_index.end()) {
      // success
      return getInPort(it->second);
    }
    
    it=out_port_index.find(name);
    if (it!=out_port_index.end()) {
      // success
      return getOutPort(it->second);
    }

    it=internal_port_index.find(name);
    if (it!=internal_port_index.end()) {
      // success
      return getInternalPort(it->second);
    }

    it=anonymous_port_index.find(name);
    if (it!=anonymous_port_index.end()) {
      // success
      return getAnonymousPort(it->second);
    }
    
    return PortDescription::ptr();
  }

  StateDescription::ptr ComponentInterface::getStateDescription(const string& name) const {
    index_map::const_iterator it=
      state_index.find(name);
    
    if (it!=state_index.end()) {
      // success
      return getStateDescription(it->second);
    }
    return StateDescription::ptr();
  }

  TransitionLabelDescription::ptr ComponentInterface::getTransitionLabelDescription(const string& name) const {
    index_map::const_iterator it=
      transition_label_index.find(name);
    
    if (it!=transition_label_index.end()) {
      // success
      return getTransitionLabelDescription(it->second);
    }
    return TransitionLabelDescription::ptr();
  }


  ComponentInterface::ptr ComponentInterface::getSubcomponentInterface(const string& name) const {
    index_map::const_iterator it=
      subcomponent_index.find(name);
    
    if (it!=subcomponent_index.end()) {
      // success
      return getSubcomponentInterface(it->second);
    }
    return ComponentInterface::ptr();
  }

  ComponentInterface::ptr ComponentInterface::getSubcomponentInterface(unsigned int i) const {
    return subcomponents.at(i).first;
  }

  const std::string& ComponentInterface::getSubcomponentName(unsigned int i) const {
    return subcomponents.at(i).second;
  }
  
  const vector<StateDescription::ptr>& ComponentInterface::getStateDescriptions() const {
    return state_descriptions;
  }
  
  bool ComponentInterface::addPort(PortDescription::ptr pd) {
    if (pd->isAnonymous()) {
      if (addToIndex(anonymous_port_index, pd->getName(), anonymous_ports.size())) {
	anonymous_ports.push_back(pd);
      } else {
	return false;
      }
      return true;
    }

    switch (pd->getType()) {
    case PortDescription::IN_PORT:
      if (addToIndex(in_port_index, pd->getName(), in_ports.size())) {
	in_ports.push_back(pd);
      } else {
	return false;
      }
      return true;
    case PortDescription::OUT_PORT:
      if (addToIndex(out_port_index, pd->getName(), out_ports.size())) {
	out_ports.push_back(pd);
      } else {
	return false;
      }
      return true;
    case PortDescription::INTERNAL_PORT:
      if (addToIndex(internal_port_index, pd->getName(), internal_ports.size())) {
	internal_ports.push_back(pd);
      } else {
	return false;
      }
      return true;
    }

    throw "implementation error";
  }
  
  bool ComponentInterface::addPortAlias(PortDescription::ptr pd, const string& alias) {
    vector<PortDescription::ptr>* pds=NULL;
    index_map* index=NULL;
    switch (pd->getType()) {
    case PortDescription::IN_PORT:
      pds=&in_ports;
      index=&in_port_index;
      break;
    case PortDescription::OUT_PORT:
      pds=&out_ports;
      index=&out_port_index;
      break;
    case PortDescription::INTERNAL_PORT:
      pds=&internal_ports;
      index=&internal_port_index;
      break;
    }
    ERROR_CHECK_ASSERT(pds);

    for (unsigned int i=0;i<pds->size();i++) {
      if (pds->at(i)==pd) {
        pd->addAlias(alias);
        index->insert(make_pair(alias, i));
        return true;
      }      
    }
    throw "Could'nt find port description!";
  }

  bool ComponentInterface::addInPort(const string& name, DataDomain::ptr data_domain) {
    PortDescription::ptr pd(new PortDescription(name, PortDescription::IN_PORT, data_domain));
    return addPort(pd);
  }
  
  bool ComponentInterface::addOutPort(const string& name, DataDomain::ptr data_domain) {
    PortDescription::ptr pd(new PortDescription(name, PortDescription::OUT_PORT, data_domain));
    return addPort(pd);
  }
  
  bool ComponentInterface::addStateDescription(StateDescription::ptr description) {
    if (addToIndex(state_index, description->getName(), state_descriptions.size())) {
      state_descriptions.push_back(description);
    } else {
      return false;
    }
    return true;
  }

  void ComponentInterface::replaceStateDescription(StateDescription::ptr new_sd) {
    const std::string& name=new_sd->getName();

    index_map::iterator it_old_sd=state_index.find(name);
    
    CHECK_SUCCESS(it_old_sd!=state_index.end());
    unsigned int index=it_old_sd->second;

    state_descriptions.at(index)=new_sd;
  }

  bool ComponentInterface::addTransitionLabelDescription(TransitionLabelDescription::ptr description) {
    if (addToIndex(transition_label_index, description->getName(), transition_label_index.size())) {
      transition_label_descriptions.push_back(description);
    } else {
      return false;
    }
    return true;
  }

  bool ComponentInterface::addSubcomponent(const string& name, ComponentInterface::ptr c_interface) {
    bool rv=addToIndex(subcomponent_index, name, subcomponent_index.size());
    if (rv) {
      subcomponents.push_back(make_pair(c_interface, name));
    }

    return rv;
  }

  const std::string& ComponentInterface::getTypeName() const {
    return type_name;
  }

  void ComponentInterface::setTypeName(const std::string& type_name_) {
    type_name=type_name_;
  }

  void ComponentInterface::print(ostream& out) {
    out << "\nIn ports: " << getInPortCount() << "\n";
    for (unsigned int i=0;i<getInPortCount();i++) {
      out << "  " << i << " : ";
      getInPort(i)->print(out);
      out << "\n";
    }

    out << "\nOut ports: " << getOutPortCount() << "\n";
    for (unsigned int i=0;i<getOutPortCount();i++) {
      out << "  " << i << " : ";
      getOutPort(i)->print(out);
      out << "\n";
    }

    out << "\nInternal ports: " << getInternalPortCount() << "\n";
    for (unsigned int i=0;i<getInternalPortCount();i++) {
      out << "  " << i << " : ";
      getInternalPort(i)->print(out);
      out << "\n";
    }

    if (getTransitionLabelDescriptionCount()>0) {
      out << "\nLabels: " << getTransitionLabelDescriptionCount() << "\n";
      for (unsigned int i=0;i<getTransitionLabelDescriptionCount();i++) {
        out << "  ";
        getTransitionLabelDescription(i)->print(out);
        out << "\n";
      }
    }
  }

  void ComponentInterface::print_xml(ostream& out) {
    xml_tag_open(out, "interface");
    print_xml(out, "");
    xml_tag_close(out, "interface");
  }

  void ComponentInterface::print_xml(ostream& out, const std::string &name) {
    xml_tag_open(out, "componentdescription", 
                 xml_parameter("name", name)+
                 xml_parameter("type", getTypeName()));
    
    for (unsigned int i=0;i<getStateDescriptionCount();i++) {
      StateDescription::ptr sd=getStateDescription(i);
      sd->print_xml(out);
    }

    for (unsigned int i=0;i<getInPortCount();i++) {
      getInPort(i)->print_xml(out, i);
    }

    for (unsigned int i=0;i<getOutPortCount();i++) {
      getOutPort(i)->print_xml(out, i);
    }

    for (unsigned int i=0;i<getInternalPortCount();i++) {
      getInternalPort(i)->print_xml(out, i);
    }

    for (unsigned int i=0;i<getTransitionLabelDescriptionCount();i++) {
      getTransitionLabelDescription(i)->print_xml(out, i);
    }


    for (unsigned int i=0;i<getSubcomponentCount();i++) {
      ComponentInterface::ptr ci=getSubcomponentInterface(i);
      const std::string& subname=getSubcomponentName(i);
      ci->print_xml(out, subname);
    }
    xml_tag_close(out, "componentdescription");
  }

  symbol_table_ptr ComponentInterface::getLocalSymbols() {
    return local_symbols;
  }

  void ComponentInterface::setLocalSymbols(symbol_table_ptr local_symbols_) {
    symbol_table_stack.pop_all();
    local_symbols=local_symbols_;

    // regenerate symbol table stack
    if (!global_symbols.expired()) {
      symbol_table_stack.push(global_symbols);
    }
    if (local_symbols) {
      symbol_table_stack.push(local_symbols);
    }
  }


  symbol_table_ptr ComponentInterface::getGlobalSymbols() {
    // return shared_ptr for the global symbol table, don't check whether
    // the pointer is empty or not
    return get_shared_ptr_from_weak_ptr(global_symbols);
  }

  void ComponentInterface::setGlobalSymbols(symbol_table_ptr global_symbols_) {
    symbol_table_stack.pop_all();

    global_symbols=global_symbols_;

    // regenerate symbol table stack
    if (!global_symbols.expired()) {
      symbol_table_stack.push(global_symbols);
    }
    if (local_symbols) {
      symbol_table_stack.push(local_symbols);
    }
  }

  const symbol_table_stack_t& ComponentInterface::getSymbolTableStack() {
    return symbol_table_stack;
  }




  bool ComponentInterface::addAtomicProposition(const std::string& name, const std::string& formula) {
    CHECK_SUCCESS(!hasAtomicProposition(name));

    if (addToIndex(atomic_proposition_index, name, atomic_propositions.size())) {
      atomic_propositions.push_back(make_pair(name,formula));
    } else {
      return false;
    }
    return true;
  }

  unsigned int ComponentInterface::getAtomicPropositionCount() const {
    return atomic_propositions.size();
  }

  const std::string& ComponentInterface::getAtomicProposition(unsigned int index) const {
    return atomic_propositions.at(index).second;
  }

  const std::string& ComponentInterface::getAtomicPropositionName(unsigned int index) const {
    return atomic_propositions.at(index).first;
  }

  unsigned int ComponentInterface::getAtomicPropositionIndex(const std::string& name) const {
    CHECK_SUCCESS(hasAtomicProposition(name));
    
    return atomic_proposition_index.find(name)->second;
  }
  
  const std::string& ComponentInterface::getAtomicProposition(const std::string& name) const {
    CHECK_SUCCESS(hasAtomicProposition(name));
    
    return atomic_propositions.at(getAtomicPropositionIndex(name)).second;
  }

  bool ComponentInterface::hasAtomicProposition(const std::string& name) const {
    return (atomic_proposition_index.count(name) >= 1);
  }


  
  bool ComponentInterface::addToIndex(index_map& i_map, string name, unsigned int index) {
    bool success=i_map.insert(index_map::value_type(name,index)).second;
    return success;
  }
  


}
