/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef COMPONENT_INTERFACE_HPP
#define COMPONENT_INTERFACE_HPP

#include <string>
#include <vector>
#include <map>
#include <boost/shared_ptr.hpp>

#include "vereofy/reo_parser/reo_parser_types.hpp"
#include "vereofy/system/state_description.hpp"
#include "vereofy/system/port_description.hpp"
#include "vereofy/system/transition_label_description.hpp"

namespace vereofy {
  class ComponentInterface {
  public:
    typedef boost::shared_ptr<ComponentInterface> ptr;

    ComponentInterface(const std::string& type_name);

    // returns the number of "in" ports 
    unsigned int getInPortCount() const;

    // returns the number of "out" ports 
    unsigned int getOutPortCount() const;

    // returns the number of "internal" ports 
    unsigned int getInternalPortCount() const;

    // returns the number of "anonymous" ports 
    unsigned int getAnonymousPortCount() const;

    // returns the number of state descriptions
    unsigned int getStateDescriptionCount() const;

    // returns the number of transition label descriptions
    unsigned int getTransitionLabelDescriptionCount() const;

    // returns the number of subcomponent interfaces
    unsigned int getSubcomponentCount() const;

    // returns the description of the i-th "in" port
    PortDescription::ptr getInPort(unsigned int i) const; 

    // returns the description of the i-th "out" port
    PortDescription::ptr getOutPort(unsigned int i) const;

    // returns the description of the i-th "internal" port
    PortDescription::ptr getInternalPort(unsigned int i) const;

    // returns the description of the i-th "anonymous" port
    PortDescription::ptr getAnonymousPort(unsigned int i) const;


    StateDescription::ptr getStateDescription(unsigned int i) const;
    PortDescription::ptr getPortDescription(const std::string& name) const;
    StateDescription::ptr getStateDescription(const std::string& name) const;
    TransitionLabelDescription::ptr getTransitionLabelDescription(unsigned int i) const;
    TransitionLabelDescription::ptr getTransitionLabelDescription(const std::string& name) const;

    ComponentInterface::ptr getSubcomponentInterface(const std::string& name) const;
    ComponentInterface::ptr getSubcomponentInterface(unsigned int i) const;
    const std::string& getSubcomponentName(unsigned int i) const;

    unsigned int getIndexForPortDescription(PortDescription::ptr pd) const;
 
    const std::vector<StateDescription::ptr>& getStateDescriptions() const;

    bool addPort(PortDescription::ptr port_description);
    bool addInPort(const std::string& name, DataDomain::ptr data_domain);
    bool addOutPort(const std::string& name, DataDomain::ptr data_domain);
    bool addPortAlias(PortDescription::ptr pd, const std::string& alias);

    bool addStateDescription(StateDescription::ptr description);
    void replaceStateDescription(StateDescription::ptr new_sd);

    bool addTransitionLabelDescription(TransitionLabelDescription::ptr description);

    bool addSubcomponent(const std::string& name, ComponentInterface::ptr c_interface);

    bool addAtomicProposition(const std::string& name, const std::string& formula);
    unsigned int getAtomicPropositionCount() const;
    const std::string& getAtomicPropositionName(unsigned int index) const;
    unsigned int getAtomicPropositionIndex(const std::string& name) const;
    const std::string& getAtomicProposition(unsigned int index) const;
    const std::string& getAtomicProposition(const std::string& name) const;
    bool hasAtomicProposition(const std::string& name) const;

    const std::string& getTypeName() const;
    void setTypeName(const std::string& type_name);

    void print(ostream& out);
    void print_xml(ostream& out);
    void print_xml(ostream& out, const std::string& name);

    symbol_table_ptr getLocalSymbols();
    void setLocalSymbols(symbol_table_ptr local_symbols);

    symbol_table_ptr getGlobalSymbols();
    void setGlobalSymbols(symbol_table_ptr global_symbols);

    const symbol_table_stack_t& getSymbolTableStack();
  private:
    typedef std::map<std::string, unsigned int> index_map;

    // shared_ptr to local_symbols
    symbol_table_ptr local_symbols;
    // weak_ptr to global_symbols (owned by SystemContext)
    symbol_table_wptr global_symbols;

    symbol_table_stack_t symbol_table_stack;
    
    std::vector<StateDescription::ptr> state_descriptions;
    std::vector<TransitionLabelDescription::ptr> transition_label_descriptions;
    std::vector<PortDescription::ptr> in_ports;
    std::vector<PortDescription::ptr> out_ports;
    std::vector<PortDescription::ptr> internal_ports;
    std::vector<PortDescription::ptr> anonymous_ports;
    std::vector<std::pair<ComponentInterface::ptr, std::string> > subcomponents;
    std::vector<std::pair<std::string, std::string> > atomic_propositions;

    index_map in_port_index;
    index_map out_port_index;
    index_map internal_port_index;
    index_map anonymous_port_index;
    index_map state_index;
    index_map transition_label_index;
    index_map subcomponent_index;
    index_map atomic_proposition_index;

    std::string type_name;

    // --- private functions ---
    bool addToIndex(index_map& i_map, std::string name, unsigned int index);
  };
} 

#endif
