/***************************************************************************
 *   Copyright (C) 2008-2009 by Tobias Blechmann                           *
 *   vereofy@blechmann.de                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "vereofy/common/errors.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/common/xml_helpers.hpp"

#include "datadomain.hpp"
#include <boost/lexical_cast.hpp>
#include <iostream>
#include <set>
#include <string>
#include <cmath>

using namespace std;

namespace vereofy {

  DataDomain::DataDomain(data_domain_type type_) 
    : type(type_) {
  }

  DataDomain::data_domain_type DataDomain::getType() const {
    return type;
  }

  bool DataDomain::isSubsetOf(const DataDomain& other) const {
    throw "implementation error";
  }
  
  bool DataDomain::isSameAs(const DataDomain& other) const {
    return isSubsetOf(other) && other.isSubsetOf(*this);
  }

  bool DataDomain::canMergeWith(const DataDomain& other) const {
    CHECK_SUCCESS(isFullyDefined());

    // default implementation
    return isSameAs(other);
  }

  DataDomain::ptr DataDomain::mergedWith(const DataDomain& other) const {
    // default
    THROW_UNIMPLEMENTED();
  }

  bool DataDomain::isConvertibleTo(const DataDomain& other) const {
    CHECK_SUCCESS(isFullyDefined());
    return isSubsetOf(other) || (other.isFullyDefined() && other.isSubsetOf(*this));
  }

  DataDomain::ptr DataDomain::getConversionTarget(DataDomain::ptr target) const {
    if (isConvertibleTo(*target)) {
      return target;
    } else {
      throw "Can not generate conversion target from "+toString()+" to "+target->toString();
    }
  }

  string DataDomain::toString() const {
    throw "implementation error";
  }

  void DataDomain::to_xml(std::ostream& out) const {
    throw "implementation error";
  }

  bool DataDomain::isFinite() const {
    // default implementation
    return true;
  }

  bool DataDomain::isFullyDefined() const {
    // default implementation
    return true;
  }
  
  void DataDomain::ensureFiniteness() const {
    if (!isFinite()) {
      throw std::string("Data type "+toString()+" is not finite!");
    }
  }

  void DataDomain::ensureFullyDefinedness() const {
    if (!isFullyDefined()) {
      throw std::string("Data type "+toString()+" is not fully defined!");
    }
  }

  FlatDataDomain::FlatDataDomain(flat_data_domain_type type_) 
    : DataDomain(DD_FLAT), type(type_) {
  }


  bool FlatDataDomain::isValidValue(int i) const {
    return false;
  }

  bool FlatDataDomain::isValidValue(const string& v) const {
    return false;
  }

  bool FlatDataDomain::hasIntegers() const {
    return false;
  }

  bool FlatDataDomain::hasEnums() const {
    return false;
  }
  
  int FlatDataDomain::getLowestInt() const {
    throw "implementation error";
  }
  
  int FlatDataDomain::getHighestInt() const {
    throw "implementation error";
  }

  FlatDataDomain::flat_data_domain_type FlatDataDomain::getFlatType() const {
    return type;
  }

  unsigned int FlatDataDomain::getConstructorArgumentsCount() const {
    return 1;
  }
  
  string FlatDataDomain::toString() const {
    throw "implementation error";
  }


  IntRangeDataDomain::IntRangeDataDomain() :
    FlatDataDomain(DD_INT),
    has_range(false),
    lower_bound(int_value_t::UNDEFINED()), 
    upper_bound(int_value_t::UNDEFINED()) {
  }

  IntRangeDataDomain::IntRangeDataDomain(int_value_t lower_bound_, int_value_t upper_bound_) :
    FlatDataDomain(DD_INT),
    has_range(true),
    lower_bound(lower_bound_), upper_bound(upper_bound_) {

    if (isFullyDefined() &&
        lower_bound>upper_bound) {
      throw "In integer range "+toString()+
        " the lower bound is greater than the upper bound!";
    }
  }

  IntRangeDataDomain::IntRangeDataDomain(const IntRangeDataDomain& other) :
    FlatDataDomain(DD_INT),
    has_range(other.has_range),
    lower_bound(other.lower_bound),
    upper_bound(other.upper_bound) {
  }

  bool IntRangeDataDomain::hasRange() const {
    return has_range;
  }

  bool IntRangeDataDomain::hasDefinedRange() const {
    return has_range && 
      lower_bound.isDefined() &&
      upper_bound.isDefined();
  }

  bool IntRangeDataDomain::isValidValue(int i) const {
    CHECK_SUCCESS(hasDefinedRange());
    return (i>=getLowestInt() && i<=getHighestInt());
  }
  
  bool IntRangeDataDomain::hasIntegers() const {
    return true;
  }
  
  int IntRangeDataDomain::getLowestInt() const {
    CHECK_SUCCESS(hasDefinedRange());
    return lower_bound.getValue();
  }

  int IntRangeDataDomain::getHighestInt() const {
    CHECK_SUCCESS(hasDefinedRange());
    return upper_bound.getValue();
  }

  void IntRangeDataDomain::expandBounds(int new_lower_bound, int new_upper_bound) {
    CHECK_SUCCESS(hasDefinedRange());
    if (lower_bound < new_lower_bound || 
        upper_bound > new_upper_bound) {
      throw "IntRangeDataDomain::expandBounds: Can not shrink range!";
    }

    lower_bound=new_lower_bound;
    upper_bound=new_upper_bound;
  }

  void IntRangeDataDomain::mergeBounds(const IntRangeDataDomain& other) {
    CHECK_SUCCESS(hasDefinedRange());
    lower_bound=std::min<int>(getLowestInt(), other.getLowestInt());
    upper_bound=std::max<int>(getHighestInt(), other.getHighestInt());
  }

  string IntRangeDataDomain::toString() const {
    if (!hasRange()) {
      return "int";
    } else {
      return 
        string("int("+
               boost::lexical_cast<string>(lower_bound)+
               ","+
               boost::lexical_cast<string>(upper_bound)+
               ")");
    }
  }

  void IntRangeDataDomain::to_xml(std::ostream& out) const {
    CHECK_SUCCESS(hasDefinedRange());
    xml_tag_simple(out, "type", 
                   xml_parameter("class", "int")+
                   xml_parameter("lowerbound", getLowestInt())+
                   xml_parameter("upperbound", getHighestInt()));
  }

  bool IntRangeDataDomain::isSubsetOf(const DataDomain& other) const {
    CHECK_SUCCESS(isFinite());
    CHECK_SUCCESS(isFullyDefined());
    CHECK_SUCCESS(other.isFinite());
    CHECK_SUCCESS(other.isFullyDefined());

    if (other.getType()!=DataDomain::DD_FLAT) {
      return false;
    }

    const FlatDataDomain& flat_other=static_cast<const FlatDataDomain&>(other);
    switch (flat_other.getFlatType()) {
    case DD_INT:
    case DD_BOOL:
    case DD_UNION:
      return getLowestInt() >= flat_other.getLowestInt() 
	&& getHighestInt() <= flat_other.getHighestInt();
    case DD_ENUM:
      return false;
    default:
      throw "implementation error";
    }
  }

  bool IntRangeDataDomain::isConvertibleTo(const DataDomain& other) const {
    CHECK_SUCCESS(isFullyDefined());
    CHECK_SUCCESS(other.isFullyDefined());
    
    if (other.isFinite()) {
      return this->isSubsetOf(other) || other.isSubsetOf(*this);
    } else {
      if (other.getType()!=DataDomain::DD_FLAT) {
        return false;
      }
      
      const FlatDataDomain& flat_other=static_cast<const FlatDataDomain&>(other);
      switch (flat_other.getFlatType()) {
      case DD_INT:
      case DD_BOOL:
      case DD_UNION:
        return true;
      case DD_ENUM:
        return false;
      default:
        throw "implementation error";
      }
    }
  }


  DataDomain::ptr IntRangeDataDomain::getConversionTarget(DataDomain::ptr target) const {
    CHECK_SUCCESS(isFullyDefined());
    CHECK_SUCCESS(target->isFullyDefined());

    if (isConvertibleTo(*target)) {
      if (!target->isFinite()) {
        // target has no range, just use the current datadomain
        return DataDomain::ptr(new IntRangeDataDomain(*this));
      } else {
        return target;
      }
    } else {
      throw "Can not generate conversion target from "+toString()+" to "+target->toString();
    }
  }


  bool IntRangeDataDomain::canMergeWith(const DataDomain& other) const {
    CHECK_SUCCESS(hasDefinedRange());
    CHECK_SUCCESS(other.isFinite() && other.isFullyDefined());

    if (other.getType() == DD_FLAT &&
        static_cast<const FlatDataDomain&>(other).getFlatType() == DD_INT) {
      return true;
    }

    return false;
  }

  DataDomain::ptr IntRangeDataDomain::mergedWith(const DataDomain& other) const {
    CHECK_SUCCESS(canMergeWith(other));

    const IntRangeDataDomain& other_int = static_cast<const IntRangeDataDomain&>(other);

    IntRangeDataDomain::ptr result(new IntRangeDataDomain(getLowestInt(),
                                                          getHighestInt()));
    result->mergeBounds(other_int);

    return result;
  }

  bool IntRangeDataDomain::isFinite() const {
    return hasRange();
  }

  bool IntRangeDataDomain::isFullyDefined() const {
    return (!hasRange()) || 
      hasDefinedRange();
  }

  void IntRangeDataDomain::ensureFiniteness() const {
    if (!isFinite()) {
      throw std::string("Integer data type "+toString()+" is not finite!");
    }
  }

  void IntRangeDataDomain::ensureFullyDefinedness() const {
    if (!isFullyDefined()) {
      throw std::string("Integer data type "+toString()+" has undefined range!");
    }
  }


  BoolDataDomain::BoolDataDomain() : IntRangeDataDomain(0,1) {
    type=DD_BOOL;
  }

  string BoolDataDomain::toString() const {
    return string("bool");
  }

  BoolDataDomain::ptr BoolDataDomain::getInstance() {
    if (!instance) 
      instance=BoolDataDomain::ptr(new BoolDataDomain());
      
    return instance;
  }


  // static instance
  BoolDataDomain::ptr BoolDataDomain::instance;
 
  EnumDataDomain::EnumDataDomain(const string& type_name_) :
    FlatDataDomain(DD_ENUM), 
    type_name(type_name_) {}
  
  bool EnumDataDomain::hasEnums() const {
    return true;
  }

  bool EnumDataDomain::addElement(const string& element) {
    if (elements.count(element)>0) {
      return false; // already there
    }
    
    unsigned int index=elements.size();
    elements[element]=index;
    CHECK(1, indizes.size()==index, "");
    indizes.push_back(element);
    return true;
  }
  
  bool EnumDataDomain::isValidValue(const string& element) const {
    return elements.count(element)>0;
  }
  
  bool EnumDataDomain::isValidIndex(unsigned int index) const {
    return index<countElements();
  }
    
  void EnumDataDomain::setTypeName(const string& type_name_) {
    type_name=type_name_;
  }   
     
  const string& EnumDataDomain::getTypeName() const {
    return type_name;
  }
  
  unsigned int EnumDataDomain::countElements() const {
    return elements.size();
  }

  ReturnValue<unsigned int> EnumDataDomain::getIndexForElement(const string& element) const {
    element_to_index_map::const_iterator it=elements.find(element);
    if (it==elements.end()) {
      return ReturnValue<unsigned int>(); // invalid
    }
    if (debugLevel>=5) {
      cerr << "Index for " << getTypeName() << "::" << element << " = " << it->second;
    }
    return it->second;
  }

  const string& EnumDataDomain::getElementForIndex(unsigned int index) const {
    // valid index check via .at()
    return indizes.at(index);
  }
  
  EnumDataDomain::const_iterator EnumDataDomain::begin() const {
    return elements.begin();
  }
  
  EnumDataDomain::const_iterator EnumDataDomain::end() const {
    return elements.end();
  }

  string EnumDataDomain::toString() const {
    std::stringstream out;
    out << "enum {";
    bool first=true;
    for (EnumDataDomain::const_iterator it=begin();it!=end();++it) {
      if (!first) out << ",";
      first=false;
      
      out << it->first;
    }
    out << "}";
    return out.str();
  }

  void EnumDataDomain::to_xml(std::ostream& out) const {
    xml_tag_open(out, "type", 
                 xml_parameter("class", "enum"));

    for (EnumDataDomain::const_iterator it=begin();
         it!=end();
         ++it) {
      xml_tag_simple(out, "element", xml_parameter("name", it->first));
    }

    xml_tag_close(out, "type");
  }


  bool EnumDataDomain::isSubsetOf(const DataDomain& other) const {
    if (other.getType()!=DataDomain::DD_FLAT) {
      return false;
    }

    const FlatDataDomain& flat_other=static_cast<const FlatDataDomain&>(other);
    switch (flat_other.getFlatType()) {
    case DD_INT:
    case DD_BOOL:
      return false;

    case DD_ENUM: {
      const EnumDataDomain& enum_other=static_cast<const EnumDataDomain&>(flat_other);
      return (type_name==enum_other.type_name)
	&& (elements==enum_other.elements);
    }
    case DD_UNION:
      throw "not implemented";
    default:
      throw "implementation error";
    }
  }
  
  UnionDataDomain::UnionDataDomain() : 
    FlatDataDomain(DD_UNION) {}

  bool UnionDataDomain::addDataDomain(FlatDataDomain::ptr dd) {
    switch (dd->getFlatType()) {
    case DD_INT:
    case DD_BOOL: {
      IntRangeDataDomain& dd_int = static_cast<IntRangeDataDomain&>(*dd);
      if (!int_range) {
	int_range=IntRangeDataDomain::ptr(new IntRangeDataDomain(dd_int));
      } else {
	int_range->mergeBounds(dd_int);
      }
      break;
    }
    case DD_ENUM: {
      EnumDataDomain::ptr enum_domain=boost::static_pointer_cast<EnumDataDomain>(dd);
      if (enum_domains.count(enum_domain->getTypeName())>0) {
	throw string("Can not add another enum domain with name ")+enum_domain->getTypeName();
      }
      enum_domains[enum_domain->getTypeName()]=enum_domain;
      break;
    }
    case DD_UNION: {
      throw "not yet implemented";
    }
    }
    return true;
  }

  bool UnionDataDomain::isValidValue(int i) const {
    return int_range && int_range->isValidValue(i);
  }

  bool UnionDataDomain::isValidValue(const string& v) const {
    size_t colon_pos=v.find("::");

    if (colon_pos==string::npos) {
      // no ::

      bool valid_element_found=false;
      for (enum_domain_map::const_iterator it=enum_domains.begin();
	   it!=enum_domains.end();
	   ++it) {
	EnumDataDomain::ptr ed=it->second;
	bool r=ed->isValidValue(v);
	
	if (valid_element_found && r) {
	  // ambiguous!
	  cerr << "Warning: Ambiguous enum element " << v << endl;
	  return false;
	} else {
	  valid_element_found|=r;
	}
      }
      return valid_element_found;

    } else {
      string type(v, 0, colon_pos); // get stuff before ::
      if (colon_pos+2>v.length()) {
	throw "Illegal argument, enum element ends with ::";
      }
      string element(v, colon_pos+2); // get stuff after ::
      if (element.find("::")!=string::npos) {
	throw "Illegal argument, enum element can not contain another ::";
      }
      
      enum_domain_map::const_iterator it=enum_domains.find(type);
      if (it==enum_domains.end()) {
	// don't know about this enum domain
	return false;
      }

      EnumDataDomain::ptr ed=it->second;
      return ed->isValidValue(v);
    }
  }

  bool UnionDataDomain::hasIntegers() const {
    return (bool)int_range;
  }


  bool UnionDataDomain::hasEnums() const {
    return enum_domains.size()>0;
  }

  int UnionDataDomain::getLowestInt() const {
    if (!int_range) {throw "implementation error";}
    return int_range->getLowestInt();
  }

  int UnionDataDomain::getHighestInt() const {
    if (!int_range) {throw "implementation error";}
    return int_range->getHighestInt();
  }

  string UnionDataDomain::toString() const {
    return string("union");
  }

  bool UnionDataDomain::isSubsetOf(const DataDomain& other) const {
    if (other.getType()!=DataDomain::DD_FLAT) {
      return false;
    }

    const FlatDataDomain& flat_other=static_cast<const FlatDataDomain&>(other);
    switch (flat_other.getFlatType()) {
    case DD_INT:
    case DD_BOOL:
      if (hasEnums()) {
	return false;
      } else {
	if (hasIntegers()) {
	  return int_range->isSubsetOf(flat_other);
	}
      }
      return false;
      
    case DD_ENUM:
      THROW_UNIMPLEMENTED();
    case DD_UNION:
      THROW_UNIMPLEMENTED();
    default:
      throw "implementation error";
    }
  }

  ArrayDataDomain::ArrayDataDomain(FlatDataDomain::ptr base_type_, 
				   uint_value_t array_size_) 
    : DataDomain(DD_ARRAY), 
      base_type(base_type_), 
      no_size(false),
      array_size(array_size_) {

    uint_value_t max_value;
    if (array_size.isDefined()) {
      if (array_size.getValue()==0) {
        max_value = 0;
      } else {
        max_value = array_size-1;
      }
    } else {
      max_value = uint_value_t::UNDEFINED();
    }

    index_domain = IntRangeDataDomain::ptr(new IntRangeDataDomain(0, (int_value_t)max_value));
  }

  ArrayDataDomain::ArrayDataDomain(FlatDataDomain::ptr base_type_)
    : DataDomain(DD_ARRAY), base_type(base_type_), 
      no_size(true),
      array_size(uint_value_t::UNDEFINED()), 
      index_domain(new IntRangeDataDomain()) {
  }

  bool ArrayDataDomain::hasNoSize() const {
    return no_size;
  }
  
  bool ArrayDataDomain::hasUndefinedSize() const {
    return !array_size.isDefined();
  }

  FlatDataDomain::ptr ArrayDataDomain::getBaseType() const {
    return base_type;
  }

  IntRangeDataDomain::ptr ArrayDataDomain::getIndexDomain() const {
    return index_domain;
  }

  unsigned int ArrayDataDomain::getArraySize() const {
    CHECK_SUCCESS(array_size.isDefined());
    return array_size.getValue();
  }

  string ArrayDataDomain::toString() const {
    if (hasNoSize()) {
      return base_type->toString() + "[]";
    } else {
      return base_type->toString() + "["+boost::lexical_cast<std::string>(array_size)+"]";
    }
  }

  void ArrayDataDomain::to_xml(std::ostream& out) const {
    CHECK_SUCCESS(isFullyDefined() && isFinite());

    xml_tag_open(out, "type", 
                 xml_parameter("class", "array")
                 +xml_parameter("size", array_size));
    base_type->to_xml(out);
    xml_tag_close(out, "type");
  }

  unsigned int ArrayDataDomain::getConstructorArgumentsCount() const {
    THROW_UNIMPLEMENTED();
  }


  bool ArrayDataDomain::isSameAs(const DataDomain& other) const {
    CHECK_SUCCESS(isFullyDefined());  // see below

    if (other.getType()!=DataDomain::DD_ARRAY) {
      return false;
    }

    const ArrayDataDomain& array_other=static_cast<const ArrayDataDomain&>(other);
    if (! getBaseType()->isSameAs(*array_other.getBaseType())) {
      return false;
    }

    if (hasNoSize()!=array_other.hasNoSize()) {
      return false;
    }

    if (hasNoSize() && array_other.hasNoSize()) {
      return true;
    }

    // from here, both have sizes, and as we check for isFullyDefined() above,
    // both actually have defined sizes
    if (getArraySize()!=array_other.getArraySize()) {
      return false;
    }
    return true;
  }

  bool ArrayDataDomain::isSubsetOf(const DataDomain& other) const {
    return isSameAs(other);
  }

  bool ArrayDataDomain::isConvertibleTo(const DataDomain& other) const {
    CHECK_SUCCESS(isFinite());
    CHECK_SUCCESS(isFullyDefined());
    CHECK_SUCCESS(other.isFullyDefined());
    
    if (other.getType()!=DataDomain::DD_ARRAY) {
      return false;
    }

    const ArrayDataDomain& array_other=static_cast<const ArrayDataDomain&>(other);
    if (array_other.hasNoSize()) {
      // ok, no problem with the size
    } else if (array_other.getArraySize()!=getArraySize()) {
      // mismatch in size
      return false;
    }

    if (! getBaseType()->isConvertibleTo(*array_other.getBaseType())) {
      return false;
    }
  
    return true;
  }

  DataDomain::ptr ArrayDataDomain::getConversionTarget(DataDomain::ptr target) const {
    if (!isConvertibleTo(*target)) {
      throw "Can not generate conversion target from "+toString()+" to "+target->toString();
    }

    if (target->isFinite()) {
      return target;
    }

    // target is not finite, construct correct target
    ArrayDataDomain::ptr array_target=boost::static_pointer_cast<ArrayDataDomain>(target);

    FlatDataDomain::ptr fq_target_base=
      boost::dynamic_pointer_cast<FlatDataDomain>(getBaseType()->getConversionTarget(array_target->getBaseType()));

    CHECK_SUCCESS(fq_target_base);

    unsigned int fq_target_size;
    if (array_target->hasNoSize()) {
      fq_target_size=getArraySize();
    } else {
      fq_target_size=array_target->getArraySize();
    }

    return DataDomain::ptr(new ArrayDataDomain(fq_target_base, fq_target_size));
  }
  
  bool ArrayDataDomain::isFinite() const {
    return !hasNoSize() && getBaseType()->isFinite();
  }
  
  void ArrayDataDomain::ensureFiniteness() const {
    if (hasNoSize()) {
      throw std::string("Array data type ")+toString()+" is not finite, as it has no array size!";
    }
    
    getBaseType()->ensureFiniteness();
  }

  bool ArrayDataDomain::isFullyDefined() const {
    if (hasNoSize()) {
      return getBaseType()->isFullyDefined();
    } else if (hasUndefinedSize()) {
      return false;
    } else {
      return getBaseType()->isFullyDefined();
    }
  }
  
  void ArrayDataDomain::ensureFullyDefinedness() const {
    if (!hasNoSize() && hasUndefinedSize()) {
      throw std::string("Array data type ")+toString()+" is not fully defined!";
    }
    
    getBaseType()->ensureFiniteness();
  }

  StructDataDomain::StructDataDomain() : DataDomain(DD_STRUCT) {
  }

  StructDataDomain::StructDataDomain(data_domain_type type) : DataDomain(type) {
  }


  unsigned int StructDataDomain::getNumberOfMembers() const {
    return member_types.size();
  }

  unsigned int StructDataDomain::getConstructorArgumentsCount() const {
    return getNumberOfMembers();
  }

  bool StructDataDomain::isSameAs(const DataDomain& other) const {
    if (other.getType()!=DataDomain::DD_STRUCT) {
      return false;
    }

    const StructDataDomain& struct_other=static_cast<const StructDataDomain&>(other);
    if (getNumberOfMembers()!=struct_other.getNumberOfMembers()) {
      return false;
    }

    for (unsigned int i=0;i<getNumberOfMembers();i++) {
      if (getMemberName(i)!=struct_other.getMemberName(i)) {
        return false;
      }
      if (! getMemberType(i)->isSameAs(*struct_other.getMemberType(i))) {
        return false;
      }
    }
    
    return true;
  }

  bool StructDataDomain::isSubsetOf(const DataDomain& other) const {
    return isSameAs(other);
  }


  bool StructDataDomain::isMember(const std::string& name) const {
    return (member_name_to_index.count(name)>0);
  }
  
  unsigned int StructDataDomain::getIndexOfMember(const std::string& name) const {
    member_name_to_index_map_t::const_iterator it=member_name_to_index.find(name);
    if (it==member_name_to_index.end()) {
      throw name+" is not a member of this struct data type";
    }
    return it->second;
  }
  
  DataDomain::ptr StructDataDomain::getMemberType(unsigned int index) const {
    return member_types.at(index).second;
  }

  const std::string& StructDataDomain::getMemberName(unsigned int index) const {
    return member_types.at(index).first;
  }

  unsigned int StructDataDomain::addMember(const std::string& name, DataDomain::ptr member_type) {
    if (isMember(name)) {
      throw name+" is already member of this struct data type!";
    }

    unsigned int index=member_types.size();
    member_types.push_back(make_pair(name, member_type));
    member_name_to_index[name]=index;

    return index;
  }

  string StructDataDomain::toString() const {
    string result="{";
    for (unsigned int i=0;i<getNumberOfMembers();i++) {
      if (i>0) {result+=',';}

      result+=getMemberName(i)+":"+getMemberType(i)->toString();
    }
    result+="}";
    return result;
  }

  void StructDataDomain::to_xml(std::ostream& out) const {
    xml_tag_open(out, "type", 
                 xml_parameter("class", "struct"));

    for (unsigned int i=0;i<getNumberOfMembers();i++) {
      xml_tag_open(out, "element", xml_parameter("name", getMemberName(i)));
      getMemberType(i)->to_xml(out);
      xml_tag_close(out, "element");
    }

    xml_tag_close(out, "type");
  }

  bool StructDataDomain::isFinite() const {
    bool is_finite = true;
    for (unsigned int i=0;i<getNumberOfMembers() && is_finite;i++) {
      is_finite &= getMemberType(i)->isFinite();
    }
    return is_finite;
  }

  void StructDataDomain::ensureFiniteness() const {
    for (unsigned int i=0;i<getNumberOfMembers();i++) {
      getMemberType(i)->ensureFiniteness();
    }
  }

  bool StructDataDomain::isFullyDefined() const {
    bool is_fully_defined = true;
    for (unsigned int i=0;i<getNumberOfMembers() && is_fully_defined;i++) {
      is_fully_defined &= getMemberType(i)->isFullyDefined();
    }
    return is_fully_defined;
  }

  void StructDataDomain::ensureFullyDefinedness() const {
    for (unsigned int i=0;i<getNumberOfMembers();i++) {
      getMemberType(i)->ensureFullyDefinedness();
    }
  }



  TaggedUnionDataDomain::TaggedUnionDataDomain() : 
    StructDataDomain(DD_TAGGED_UNION), tag_domain(new EnumDataDomain()) {
  }

  EnumDataDomain::ptr TaggedUnionDataDomain::getTag() {
    return tag_domain;
  }

  unsigned int TaggedUnionDataDomain::getConstructorArgumentsCount() const {
    return 2;
  }

  bool TaggedUnionDataDomain::isSameAs(const DataDomain& other) const {
    if (other.getType()!=DataDomain::DD_TAGGED_UNION) {
      return false;
    }

    const TaggedUnionDataDomain& tagged_union_other=static_cast<const TaggedUnionDataDomain&>(other);
    if (getNumberOfMembers()!=tagged_union_other.getNumberOfMembers()) {
      return false;
    }

    for (unsigned int i=0;i<getNumberOfMembers();i++) {
      if (getMemberName(i)!=tagged_union_other.getMemberName(i)) {
        return false;
      }
      if (! getMemberType(i)->isSameAs(*tagged_union_other.getMemberType(i))) {
        return false;
      }
    }
    
    return true;
  }

  bool TaggedUnionDataDomain::isSubsetOf(const DataDomain& other) const {
    return isSameAs(other);
  }

  unsigned int TaggedUnionDataDomain::addMember(const std::string& name, DataDomain::ptr member_type) {
    if (isMember(name)) {
      throw name+" is already member of this tagged_union data type!";
    }

    if (name==getTagName()) {
      throw "No member named 'tag' is allowed in tagged_union";
    }

    unsigned int index=StructDataDomain::addMember(name, member_type);
    getTag()->addElement(name);

    CHECK_SUCCESS(getTag()->getIndexForElement(name) == index);

    return index;
  }

  string TaggedUnionDataDomain::toString() const {
    string result="{";
    for (unsigned int i=0;i<getNumberOfMembers();i++) {
      if (i>0) {result+='+';}

      result+=getMemberName(i)+":"+getMemberType(i)->toString();
    }
    result+="}";
    return result;
  }

  void TaggedUnionDataDomain::to_xml(std::ostream& out) const {
    xml_tag_open(out, "type", 
                 xml_parameter("class", "tagged_union"));

    for (unsigned int i=0;i<getNumberOfMembers();i++) {
      xml_tag_open(out, "element", xml_parameter("name", getMemberName(i)));
      getMemberType(i)->to_xml(out);
      xml_tag_close(out, "element");
    }

    xml_tag_close(out, "type");
  }

}
