/***************************************************************************
*   Copyright (C) 2008-2009 by Tobias Blechmann                           *
*   vereofy@blechmann.de                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef DATADOMAIN_H
#define DATADOMAIN_H

#include "vereofy/setup.hpp"
#include "vereofy/common/return_value.hpp"
#include "vereofy/common/value_or_undefined.hpp"
#include <boost/utility.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <vector>
#include <string>
#include <map>


namespace vereofy{

  /**
   * This is the base class for all Data Domains
   */
  class DataDomain : public boost::enable_shared_from_this<DataDomain> {
  public:
    typedef boost::shared_ptr<DataDomain> ptr;

    enum data_domain_type {DD_FLAT, DD_ARRAY, DD_STRUCT, DD_TAGGED_UNION};

    DataDomain(data_domain_type type);
    virtual ~DataDomain() {};

    data_domain_type getType() const;
    virtual std::string toString() const;
    virtual void to_xml(std::ostream& out) const;

    virtual bool isSubsetOf(const DataDomain& other) const;
    virtual bool isSameAs(const DataDomain& other) const;
    virtual bool isConvertibleTo(const DataDomain& other) const;
    virtual DataDomain::ptr getConversionTarget(DataDomain::ptr target) const;

    virtual bool canMergeWith(const DataDomain& other) const;
    // mergedWith may only be called if 1) this->isSameAs(other)==false and 2) this->canMergeWith(other)==true
    virtual DataDomain::ptr mergedWith(const DataDomain& other) const;

    virtual bool isFinite() const;
    virtual bool isFullyDefined() const;

    virtual void ensureFiniteness() const;  // throw std::string() exception if not finite
    virtual void ensureFullyDefinedness() const;  // throw std::string() exception if data domain is not fully defined

    virtual unsigned int getConstructorArgumentsCount() const = 0;
  private:
    data_domain_type type;
  };

  /**
   * This is the base class for all the 'flat' Data Domains, i.e. data domains
   * without internal structure. These data domains can consist of integers
   * and of enumeration elements.
   */
  class FlatDataDomain : public DataDomain {
  public:
    typedef boost::shared_ptr<FlatDataDomain> ptr;

    enum flat_data_domain_type {DD_INT, DD_BOOL, DD_ENUM, DD_UNION};

    FlatDataDomain(flat_data_domain_type type);

    virtual bool isValidValue(int i) const;
    virtual bool isValidValue(const std::string& v) const;

    virtual bool hasIntegers() const;
    virtual bool hasEnums() const;
    virtual int getLowestInt() const;
    virtual int getHighestInt() const;

    virtual std::string toString() const;

    flat_data_domain_type getFlatType() const;

    virtual unsigned int getConstructorArgumentsCount() const;

  protected:
    flat_data_domain_type type;
  };

  class IntRangeDataDomain : public FlatDataDomain {
  public:
    typedef boost::shared_ptr<IntRangeDataDomain> ptr;

    /** Constructor for an unranged int data domain */
    IntRangeDataDomain();

    /** Constructor for a ranged int data domain, lower and upper bound may be
     *  int_value_t::UNDEFINED()
     */
    IntRangeDataDomain(int_value_t lower_bound, int_value_t upper_bound);

    /** Copy constructor */
    IntRangeDataDomain(const IntRangeDataDomain& other);

    /** Returns true if this data domain has a range, i.e., not just 'int' */
    bool hasRange() const;

    /** Returns true if this data domain has a properly defined range,
     *  i.e., known values for the upper and lower bound
     */
    bool hasDefinedRange() const;
  

    virtual bool isValidValue(int i) const;
    virtual bool hasIntegers() const;
    virtual int getLowestInt() const;
    virtual int getHighestInt() const;

    void expandBounds(int new_lower_bound, int new_upper_bound);
    void mergeBounds(const IntRangeDataDomain& other);

    virtual bool isSubsetOf(const DataDomain& other) const;
    virtual bool isConvertibleTo(const DataDomain& other) const;
    virtual DataDomain::ptr getConversionTarget(DataDomain::ptr target) const;

    virtual bool canMergeWith(const DataDomain& other) const;
    virtual DataDomain::ptr mergedWith(const DataDomain& other) const;

    virtual std::string toString() const;
    virtual void to_xml(std::ostream& out) const;

    virtual bool isFinite() const;
    virtual bool isFullyDefined() const;

    virtual void ensureFiniteness() const;  // throw std::string() exception if not finite
    virtual void ensureFullyDefinedness() const;  // throw std::string() exception if data domain is not fully defined


  private:
    bool has_range;
    int_value_t lower_bound, upper_bound;
  };

  class BoolDataDomain : public IntRangeDataDomain {
  public:
    typedef boost::shared_ptr<BoolDataDomain> ptr;

    BoolDataDomain();

    static BoolDataDomain::ptr getInstance();
    virtual std::string toString() const;
  private:
    static BoolDataDomain::ptr instance;
  };

  class EnumDataDomain : public FlatDataDomain {
  public:
    typedef boost::shared_ptr<EnumDataDomain> ptr;
    typedef std::map<std::string, unsigned int> element_to_index_map;
    typedef element_to_index_map::const_iterator const_iterator;
    typedef std::vector<std::string> index_to_element_map;
    
    EnumDataDomain(const std::string& type_name_ = "<anonymous>");
    
    virtual bool hasEnums() const;
    bool addElement(const std::string& element);
    virtual bool isValidValue(const std::string& element) const;

    void setTypeName(const std::string& type_name);
    const std::string& getTypeName() const;

    ReturnValue<unsigned int> getIndexForElement(const std::string& element) const;
    const std::string& getElementForIndex(unsigned int index) const;
    bool isValidIndex(unsigned int index) const;
    unsigned int countElements() const;
    const_iterator begin() const;
    const_iterator end() const;

    virtual std::string toString() const;
    virtual void to_xml(std::ostream& out) const;

    virtual bool isSubsetOf(const DataDomain& other) const;
  private:
    std::string type_name;
    element_to_index_map elements;
    index_to_element_map indizes;
  };


  class UnionDataDomain : public FlatDataDomain {
    typedef boost::shared_ptr<UnionDataDomain> ptr;

    UnionDataDomain();
    bool addDataDomain(FlatDataDomain::ptr dd);

    virtual bool isValidValue(int i) const;
    virtual bool isValidValue(const std::string& v) const;

    virtual bool hasEnums() const;
    virtual bool hasIntegers() const;
    virtual int getLowestInt() const;
    virtual int getHighestInt() const;

    virtual std::string toString() const;
    virtual bool isSubsetOf(const DataDomain& other) const;
  private:
    typedef std::map<std::string, EnumDataDomain::ptr> enum_domain_map;
    enum_domain_map enum_domains;
    IntRangeDataDomain::ptr int_range;
  };


  /**
   * This is a data domain representing base_type^array_size
   */
  class ArrayDataDomain : public DataDomain {
  public:
    typedef boost::shared_ptr<ArrayDataDomain> ptr;

    ArrayDataDomain(FlatDataDomain::ptr base_type, uint_value_t array_size);
    ArrayDataDomain(FlatDataDomain::ptr base_type); // no size

    bool hasNoSize() const;
    bool hasUndefinedSize() const;

    virtual bool isSubsetOf(const DataDomain& other) const;
    virtual bool isSameAs(const DataDomain& other) const;
    virtual bool isConvertibleTo(const DataDomain& other) const;
    virtual DataDomain::ptr getConversionTarget(DataDomain::ptr target) const;

    unsigned int getArraySize() const;
    FlatDataDomain::ptr getBaseType() const;
    IntRangeDataDomain::ptr getIndexDomain() const;

    virtual std::string toString() const;
    virtual void to_xml(std::ostream& out) const;

    virtual unsigned int getConstructorArgumentsCount() const;

    virtual bool isFinite() const;
    virtual void ensureFiniteness() const;  // throw std::string() exception if not finite

    virtual bool isFullyDefined() const;
    virtual void ensureFullyDefinedness() const;  // throw std::string() exception if not fully defined

  private:
    FlatDataDomain::ptr base_type;
    bool no_size;
    uint_value_t array_size;
    IntRangeDataDomain::ptr index_domain;
  };


  /**
   * This is a data domain representing a structured data type 
   * consisting of multiple other data types
   */
  class StructDataDomain : public DataDomain {
  public:
    typedef boost::shared_ptr<StructDataDomain> ptr;

    StructDataDomain();

    virtual bool isSubsetOf(const DataDomain& other) const;
    virtual bool isSameAs(const DataDomain& other) const;

    unsigned int getNumberOfMembers() const;
    bool isMember(const std::string& name) const;
    unsigned int getIndexOfMember(const std::string& name) const;
    const std::string& getMemberName(unsigned int index) const;
    DataDomain::ptr getMemberType(unsigned int index) const;

    unsigned int addMember(const std::string& name, DataDomain::ptr member_type);

    virtual std::string toString() const;
    virtual void to_xml(std::ostream& out) const;

    virtual unsigned int getConstructorArgumentsCount() const;

    virtual bool isFinite() const;
    virtual bool isFullyDefined() const;

    virtual void ensureFiniteness() const;  // throw std::string() exception if not finite
    virtual void ensureFullyDefinedness() const;  // throw std::string() exception if data domain is not fully defined

  protected:
    StructDataDomain(data_domain_type type);

  private:
    std::vector< std::pair<std::string, DataDomain::ptr> > member_types;
    typedef std::map<std::string, unsigned int> member_name_to_index_map_t;
    member_name_to_index_map_t member_name_to_index;
  };


  class TaggedUnionDataDomain : public StructDataDomain {
  public:
    typedef boost::shared_ptr<TaggedUnionDataDomain> ptr;
    
    TaggedUnionDataDomain();

    virtual bool isSubsetOf(const DataDomain& other) const;
    virtual bool isSameAs(const DataDomain& other) const;

    unsigned int addMember(const std::string& name, DataDomain::ptr member_type);

    EnumDataDomain::ptr getTag();
    std::string getTagName() const {return "tag";}

    virtual std::string toString() const;
    virtual void to_xml(std::ostream& out) const;

    virtual unsigned int getConstructorArgumentsCount() const;
  private:
    EnumDataDomain::ptr tag_domain;
  };

} //end namespace

#endif

