/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef PORT_DESCRIPTION_HPP
#define PORT_DESCRIPTION_HPP

#include <string>
#include <set>
#include <boost/shared_ptr.hpp>
#include <boost/enable_shared_from_this.hpp>

#include "vereofy/common/vereofy_id.hpp"
#include "vereofy/common/xml_helpers.hpp"
#include "vereofy/common/errors.hpp"
#include "vereofy/common/shared_ptr_map.hpp"
#include "datadomain.hpp"

namespace vereofy {
  class PortDescription : public VereofyID, public boost::enable_shared_from_this<PortDescription> {
  public:
    typedef boost::shared_ptr<PortDescription> ptr;
    typedef shared_ptr_id_set<PortDescription::ptr>::type ptr_set;

    typedef std::set<std::string> alias_set_t;

    enum port_type {IN_PORT, OUT_PORT, INTERNAL_PORT};

    PortDescription(const std::string& name_,
		    port_type type_,
                    DataDomain::ptr data_domain_,
		    bool is_anonymous_=false) : 
      name(name_), type(type_), data_domain(data_domain_), is_anonymous(is_anonymous_) {}

    // get a name for this port
    const std::string& getName() const {return name;}
    port_type getType() const {return type;}
    DataDomain::ptr getDataDomain() const {return data_domain;}

    void addAlias(const std::string& alias) {aliases.insert(alias);}
    alias_set_t::const_iterator aliases_begin() const {return aliases.begin();}
    alias_set_t::const_iterator aliases_end() const {return aliases.end();}
    std::string getAliasString() const {
      std::string alias_string;
      for (alias_set_t::const_iterator it=aliases_begin();
           it!=aliases_end();
           ++it) {
        if (it!=aliases_begin()) alias_string+=" ";
        alias_string+=*it;
      }
      return alias_string;
    }


    bool isAnonymous() const {return is_anonymous;}
    void setAnonymous(bool flag=true) {is_anonymous=flag;}

    std::string getTypeString() const {
      switch (getType()) {
      case IN_PORT: return "IN"; break;
      case OUT_PORT: return "OUT"; break;
      case INTERNAL_PORT: return "INTERNAL"; break;
      }
      THROW_UNIMPLEMENTED();
    }

    void print(std::ostream& out) {
      out << name << " (" << getTypeString() << ")";

      if (aliases.size()>0) {
        out << " Aliases:";

        for (alias_set_t::const_iterator it=aliases_begin();
             it!=aliases_end();
             ++it) {
          out << " " << *it;
        }
      }
    }

    void print_xml(std::ostream& out, unsigned int index) {
      xml_tag_open(out, "portdescription", 
                   xml_parameter("name", getName())+xml_parameter("type", getTypeString())+
                   xml_parameter("index", boost::lexical_cast<std::string>(index)));
      for (PortDescription::alias_set_t::const_iterator it_alias=aliases_begin();
           it_alias!=aliases_end();
           ++it_alias) {
        xml_tag_simple(out, "alias", xml_parameter("name", *it_alias));
      }
      
      getDataDomain()->to_xml(out);
      xml_tag_close(out, "portdescription");
    }
      
  private:
    std::string name;
    port_type type;
    DataDomain::ptr data_domain;
    bool is_anonymous;
    alias_set_t aliases;

  };
} 


#endif
