/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "component_instance_rm.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/system/types.hpp"
#include "vereofy/system/reactive_module/symb_ca_reactive_module.hpp"
#include "vereofy/reo_parser/identifier_parser_helper.hpp"
#include "vereofy/reo_parser/module_parser_helper.hpp"
#include "vereofy/reo_parser/common_parser_helper.hpp"
#include "vereofy/common/errors.hpp"

#include <boost/algorithm/string.hpp>

namespace vereofy {

using namespace vparser;

ComponentInstanceReactiveModule::ComponentInstanceReactiveModule(ReactiveModule::ptr module_,
                                                                 const std::vector<Value::ptr>& parameters)
  : ComponentInstance(COMPONENT_REACTIVE_MODULE, module_->context),
    module(module_) {
  
  // make a copy of the interface
  ComponentInterface::ptr new_interface(new ComponentInterface(*module->c_interface));
  // make a copy of the local symbol table
  new_interface->setLocalSymbols(symbol_table_ptr(new symbol_table_t(*new_interface->getLocalSymbols())));
  

  if (module_helper::t_module_def::hasParameterDef(module->behavior->parse_tree)) {
    vector<string> parameter_def=
      common_helper::t_parameter_def::getIdentifiers(module_helper::t_module_def::getParameterDef(module->behavior->parse_tree));
    
    if (parameter_def.size()!=parameters.size()) {
      throw string("Parameter mismatch in instantiation: ")+module->name+ " expects "+
        boost::lexical_cast<string>(parameter_def.size())+" parameters, but "+
        boost::lexical_cast<string>(parameters.size())+" were supplied!";
    }
    
    for (unsigned int i=0;i<parameter_def.size();i++) {
      string p_name=parameter_def.at(i);
      Value::ptr p=parameters.at(i);
      
      if (!p) {
        throw "NULL value in parameters!";
      }
      
      switch (p->type->type) {
      case VALUE_TYPE_INT:
        new_interface->getLocalSymbols()->replace(p_name, 
                                                  ConstantData::ptr(new ConstantData(p_name,
                                                                                     IntValue::getIntValue(p))));
        break;
      case VALUE_TYPE_BOOL: 
        THROW_UNIMPLEMENTED();
      default:
        throw "Can not pass parameter of type "+p->type->typeToString()+" during instantiation";
      }
    }

    // update the interface to potentially take parameters into account
    update_interface(new_interface);
  } else {
    // no parameter def
    if (parameters.size()!=0) {
      throw string("Parameter mismatch in instantiation: ")+module->name+ " does not expect parameters!";
    }
  }

  setInterface(new_interface);
  setTypeName(module->name+getParametersAsString(parameters));
  attachAtomicPropositionsForType();
  new_interface->setTypeName(getTypeName());
}

ReactiveModule::ptr ComponentInstanceReactiveModule::getModule() {
  return module;
}


SymbComponentInstance::ptr ComponentInstanceReactiveModule::getSymbAutomaton(ComponentInstance::ptr self,
                                                                     vereofy::Symb_DataDomain::ptr s_dd) {
  ERROR_CHECK_ASSERT(self.get()==this);

  ComponentInstanceReactiveModule::ptr cirm=boost::static_pointer_cast<ComponentInstanceReactiveModule>(self);
  
  Symb_CA_ReactiveModule::ptr ca_rm(new Symb_CA_ReactiveModule(cirm, s_dd));
  return ca_rm;
}


void ComponentInstanceReactiveModule::update_interface(ComponentInterface::ptr ci) {
  const vparser::pt_node& parse_tree=module->behavior->parse_tree;

  const vparser::pt_node& body=module_helper::t_module_def::getBody(parse_tree);
  
  if (!module_helper::t_body::hasDefs(body)) {
    return;
  }

  const vparser::pt_node& defs=module_helper::t_body::getDefs(body);
 
  for (unsigned int i=0;i<module_helper::t_defs::countChildren(defs);i++) {
    const vparser::pt_node& def=module_helper::t_defs::get(defs, i);
    if (def.value.id()!=parser_id(ID_M_var_def))
      continue;

    string var_name=module_helper::t_var_def::getName(def);
    DataDomain::ptr dd=module_helper::t_var_def::getDataDomain(def, 
                                                               ci->getSymbolTableStack());

    try {
      dd->ensureFiniteness();
      dd->ensureFullyDefinedness();
    } catch (std::string& e) {
      throw std::string("In definition of state variable ")+var_name+": "+e;
    }

    vereofy::StateDescription::ptr sd(new vereofy::StateDescription(var_name,
                                                                    dd));
    
    ci->replaceStateDescription(sd);
  }
}


void ComponentInstanceReactiveModule::calculatePortPartitions(PortPartitions& port_partitions) {
  // add PortIDs to the partitions
  for (unsigned int i=0;i<getInterface()->getInPortCount();i++) {
    port_partitions.identical_ports.add(PortID(this, getInterface()->getInPort(i)));
    port_partitions.mixed_data_ports.add(PortID(this, getInterface()->getInPort(i)));
  }

  for (unsigned int i=0;i<getInterface()->getOutPortCount();i++) {
    port_partitions.identical_ports.add(PortID(this, getInterface()->getOutPort(i)));
    port_partitions.mixed_data_ports.add(PortID(this, getInterface()->getOutPort(i)));
  }


  if (getTypeName() == "SYNC") {
    port_partitions.identical_ports.make_equivalent(PortID(this, getInterface()->getInPort(0)),
                                                    PortID(this, getInterface()->getOutPort(0)));
    return;
  }

  if (getTypeName() == "SYNC_DRAIN_EQUAL_DATA") {
    port_partitions.identical_ports.make_equivalent(PortID(this, getInterface()->getInPort(0)),
                                                    PortID(this, getInterface()->getInPort(1)));
    return;
  }

  // TODO SYNC_DRAIN and SYNC_SPOUT are not mixed, but are not identical either

  if (getTypeName() == "LOSSY_SYNC_ND"  ||
      boost::algorithm::starts_with(getTypeName(),"FILTER_EQUAL<") ||
      boost::algorithm::starts_with(getTypeName(),"FILTER_NOT_EQUAL<") ||
      boost::algorithm::starts_with(getTypeName(),"FILTER_LESS<") ||
      boost::algorithm::starts_with(getTypeName(),"FILTER_LESS_EQUAL<") ||
      boost::algorithm::starts_with(getTypeName(),"FILTER_GREATER<") ||
      boost::algorithm::starts_with(getTypeName(),"FILTER_GREATER_EQUAL<")) {
    // data bits of both ports should be mixed

    port_partitions.mixed_data_ports.make_equivalent(PortID(this, getInterface()->getInPort(0)),
                                                     PortID(this, getInterface()->getOutPort(0)));
    return;
  }
}

}
