/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/




#include "symb_ca_reactive_module.hpp"
#include "vereofy/reo_parser/REOParser.hpp"
#include "vereofy/reo_parser/module_parser.hpp"
#include "vereofy/reo_parser/guard_parser.hpp"
#include "vereofy/reo_parser/guard_parser_helper.hpp"
#include "vereofy/reo_parser/int_parser_helper.hpp"
#include "vereofy/reo_parser/parser_helper.hpp"
#include "vereofy/common/errors.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/common/bdd_op_trace.hpp"
#include "vereofy/system/symb_automaton/symb_array_datadomain.hpp"

#include <sstream>

using namespace std;
using namespace vereofy;
using namespace vereofy::vparser;

Symb_CA_ReactiveModule::Symb_CA_ReactiveModule(ComponentInstanceReactiveModule::ptr ci_module_,
                                               Symb_DataDomain::ptr dataDomain) :
  SymbComponentInstance(ci_module_->getInterface(), dataDomain),
  ci_module(ci_module_),
  have_rm_variables(false) {

  setComponentInstance(ci_module_);
}

void Symb_CA_ReactiveModule::perform_assignments(const vector<bdd_assignment_vector>& assignments, 
                                                 BDDFunction assignment,
                                                 BDDFunction csTargets,
                                                 unsigned int index,
                                                 vector<BDDFunction>& result) const {
  const bdd_assignment_vector& cur=assignments.at(index);

  if (index+1==assignments.size()) {
    // last 
    for (unsigned int j=0;j<cur.size();j++) {
      const bdd_assignment& cur_assignment=cur.at(j);

      BDDFunction a=assignment & cur_assignment.assignment;
      BDDFunction t=csTargets & container2cs(cur_assignment.targets);

      BDDFunction other_states_unchanged=no_state_changes.exists(t);
      BDDFunction a_=a & other_states_unchanged;
      
      if (debugLevel>=4) {
        cout << "perform_assignments:\n";
        cout << "Assignment:\n";
        a.printPaths();
        cout << "Targets:\n";
        t.printPaths();
        cout << "Other states unchanged:\n";
        other_states_unchanged.printPaths();
        cout << "Real assignment:\n";
        a_.printPaths();
      }

      result.push_back(a_);
    }
  } else {
    for (unsigned int j=0;j<cur.size();j++) {
      const bdd_assignment& cur_assignment=cur.at(j);
      perform_assignments(assignments,
                          assignment & cur_assignment.assignment,
                          csTargets & container2cs(cur_assignment.targets),
                          index+1,
                          result);
    }
  }
}

void Symb_CA_ReactiveModule::build_automaton(BuildContext::ptr build_context) {
  if (build_automaton_with_serialized_varorder(build_context)) {
    return;
  }  

  TRACE("building", "CARML",                                            \
        std::cout << "Build automaton [CARML], ";                       \
        if (info->component_instance) {                                 \
          std::cout << info->component_instance->getFQName() <<         \
            " (" <<info->component_instance->getTypeName() << ")";      \
        }
        std::cout << endl;);

  // store build context
  this->build_context=build_context;

  const vparser::pt_node& parse_tree=ci_module->getModule()->behavior->parse_tree;


  unsigned int i;


  // add transition labels
  for (i=0;i<ci_module->getInterface()->getTransitionLabelDescriptionCount();i++) {
    addTransitionLabel(ci_module->getInterface()->getTransitionLabelDescription(i));
  }

  // add out ports
  for (i=0;i<ci_module->getInterface()->getOutPortCount();i++) {
    addPort(ci_module->getInterface()->getOutPort(i));
  }

  // add states
  if (ci_module->getInterface()->getStateDescriptionCount()==0) {
#ifdef WITH_DUMMY_STATE
    addDummyState();
#endif
  } else {
    for (i=0;i<ci_module->getInterface()->getStateDescriptionCount();i++) {
      addState(ci_module->getInterface()->getStateDescription(i));
    }
  }

  // add in ports
  for (i=0;i<ci_module->getInterface()->getInPortCount();i++) {
    addPort(ci_module->getInterface()->getInPort(i));
  }


  const vparser::pt_node& body=module_helper::t_module_def::getBody(parse_tree);

  handle_initialization(body, build_context);
  handle_atomic_proposition_defs(body, build_context);
  handle_transitions(body, build_context);

  // calculate the generic atomic propositions at the end
  TRACE("building", "CARML", std::cout << " Calculating atomic propositions" << std::endl);          
  calculateAtomicPropositions();
}

void Symb_CA_ReactiveModule::handle_initialization(const vparser::pt_node& body, BuildContext::ptr build_context) {
  BDDFunction initial_states=BDDFunction::ONE();

#ifdef WITH_DUMMY_STATE
  if (dummy_state_var.X) {
    // dummy state
    initial_states&=!dummy_state_var.X->at(0);
  }
#endif

  if (!module_helper::t_body::hasDefs(body)) {
      return;
  }

  const vparser::pt_node& defs=module_helper::t_body::getDefs(body);

  for (unsigned int i=0;i<module_helper::t_defs::countChildren(defs);i++) {
    const vparser::pt_node& def=module_helper::t_defs::get(defs, i);
    if (def.value.id()==parser_id(ID_M_var_def) &&
        module_helper::t_var_def::hasInit(def)) {
      string name=module_helper::t_var_def::getName(def);
      const vparser::pt_node& init=module_helper::t_var_def::getInit(def);

      bool any_value_init=false;
      if (module_helper::t_var_init::get(init).value.id() == parser_id(ID_G_any_value)) {
        any_value_init=true;
        // TODO: modify for validity
        continue;
      }

      SymbolicState::ptr ss=getState(name);

      BDDFunction init_var;
      switch (ss->data_domain->getType()) {
      case DataDomain::DD_FLAT:
      case DataDomain::DD_STRUCT: 
      case DataDomain::DD_TAGGED_UNION: {
        bdd_assignment_target target=getAssignmentTargetFlat(name, false);
        SymbExpression init_expression=calc_expression(module_helper::t_var_init::get(init),
                                                       target.data_domain, true);
        
        init_var=perform_assignment(target, init_expression).merge();
        break;
      }
      case DataDomain::DD_ARRAY: {
        Symb_ArrayDataDomain::ptr s_array_dd=
          boost::static_pointer_cast<Symb_ArrayDataDomain>(ss->data_domain);
        ArrayDataDomain::ptr array_dd=s_array_dd->getArrayDataDomain();

        init_var=BDDFunction::ONE();
        for (unsigned int i=0;i<array_dd->getArraySize();i++) {
          bdd_assignment_target target=getAssignmentTargetArray(name, i, false);
          SymbExpression init_expression=calc_expression(module_helper::t_var_init::get(init),
                                                                target.data_domain, true);
        
          init_var&=perform_assignment(target, init_expression).merge();
        }
        break;
      }
      }
      
      if (debugLevel>=3) {
        cerr << "Init " << name << endl;
        init_var.printPaths();
      }
      initial_states&=init_var;
    }
  }

  initialStates=initial_states;

  if (build_context->restriction_manager &&
      build_context->restriction_manager->restrict_delta_x) {
    initialStates&=build_context->restriction_manager->getStateRestrictions(*this, true);
  }

  if (debugLevel>=3) {
    cerr << "Initial States:\n";
    initialStates.printPaths();
  }
}

void Symb_CA_ReactiveModule::handle_atomic_proposition_defs(const vparser::pt_node& body, BuildContext::ptr build_context) {
  if (!module_helper::t_body::hasAPDefs(body)) {
    return;
  }

  const vparser::pt_node& ap_defs=module_helper::t_body::getAPDefs(body);

  for (unsigned int i=0;i<module_helper::t_ap_defs::countChildren(ap_defs);i++) {
    const vparser::pt_node& ap_def=module_helper::t_ap_defs::get(ap_defs, i);

    const std::string& ap_name=module_helper::t_ap_def::getName(ap_def);
    const vparser::pt_node& ap_state_guard=module_helper::t_ap_def::getBooleanStateGuard(ap_def);

    BDDFunction guard_function=bool_expression_to_bdd(ap_state_guard);

    info->addAtomicProposition(ap_name, guard_function);
  }
}

BDDFunction Symb_CA_ReactiveModule::get_no_generic_labels() const {
  BDDFunction csGenericLabelVars = BDDFunction::ONE();
  
  for (transition_label_to_bdd_map::const_iterator it=info->transition_label_to_bdd.begin();
       it!=info->transition_label_to_bdd.end();
       ++it) {
    TransitionLabelDescription::ptr tl=it->first;
    Symb_Transition_Label_Var::ptr symb_tl=it->second;
    
    if (tl->getType()==TransitionLabelDescription::GENERIC) {
      csGenericLabelVars &= symb_tl->getHasLabelVar();
    }
  }

  return csGenericLabelVars.change();
}

void Symb_CA_ReactiveModule::handle_transitions(const vparser::pt_node& body, BuildContext::ptr build_context) {
  if (!module_helper::t_body::hasTransitions(body)) {
    delta=BDDFunction::ZERO();
    return;
  }

  const vparser::pt_node& transitions=module_helper::t_body::getTransitions(body);

  delta=BDDFunction::ZERO();  
  for (unsigned int t=0;t<module_helper::t_transitions::countTransitions(transitions);t++) {
    BDDFunction state_guard=BDDFunction::ONE();
    BDDFunction io_guard=BDDFunction::ONE();
    BDDFunction state_assign=BDDFunction::ONE();
    BDDFunction label_assign=get_no_generic_labels();

    const vparser::pt_node& transition=module_helper::t_transitions::getTransition(transitions, t);

#ifdef REO_PARSER_POSITION_NODES
    TRACE("building", "CARML",                                          \
          std::cout << " Adding transition "<< t;                       \
          std::cout << " at " << getFilePositionToString(*transition.value.pos); \
          std::cout << endl;);
#else
    TRACE("building", "CARML", std::cout << " Adding transition "<< t << endl;);

#endif


    BDDFunction transition_bdd=BDDFunction::ZERO();

    if (transition.value.id() == parser_id(ID_M_compact_transition)) {
      transition_bdd = module_helper::t_compact_transition::getAssignment(transition, getRMVariables());
      TRACE_BDD_FUNCTION("compact_transition", transition_bdd, "");
    } else if (transition.value.id() == parser_id(ID_M_transition)) {
      // standard transition
      if (module_helper::t_transition::hasL(transition)) {
        const vparser::pt_node& l_trans=module_helper::t_transition::getL(transition);
        for (unsigned int g=0;g<module_helper::t_l_transition::countStateGuards(l_trans);g++) {
          BDDFunction cur_guard=state_guard_to_bdd(module_helper::t_l_transition::getStateGuard(l_trans, g));
          state_guard&=cur_guard;
        }
      }
      
      if (module_helper::t_transition::hasM(transition)) {
        const vparser::pt_node& m_trans=module_helper::t_transition::getM(transition);
        for (unsigned int g=0;g<module_helper::t_m_transition::countIOGuards(m_trans);g++) {
          BDDFunction cur_guard=io_guard_to_bdd(module_helper::t_m_transition::getIOGuard(m_trans, g));
          io_guard&=cur_guard;
        }
      }
      
      if (module_helper::t_transition::hasR(transition)) {
        const vparser::pt_node& r_trans=module_helper::t_transition::getR(transition);
        vector<bdd_assignment_vector> assignments;
        
        bool has_state_assignment = false;
        
        for (unsigned int a=0;a<module_helper::t_r_transition::countAssignments(r_trans);a++) {
          const vparser::pt_node& assignment=module_helper::t_r_transition::getAssignment(r_trans, a);
          
          if (guard_helper::t_state_assignment::is_add_label(assignment)) {
            continue;  // skip labels
          }
          
          has_state_assignment = true;
          const vparser::pt_node& lhs=guard_helper::t_state_assignment::getLHS(assignment);
          const vparser::pt_node& rhs=guard_helper::t_state_assignment::getRHS(assignment);
          
          bdd_assignment_target_vector targets=getAssignmentTargets(lhs);
          
          if (rhs.value.id() == parser_id(ID_G_any_value)) {
            assignments.push_back(perform_assignment_any_value(targets));
          } else {
            SymbExpression rhs_expression=calc_expression(rhs, targets.data_domain, true);
            assignments.push_back(perform_assignment(targets, rhs_expression));
          }
        }
        
        if (has_state_assignment) {
          vector<BDDFunction> a_vector;
          perform_assignments(assignments, BDDFunction::ONE(), BDDFunction::ONE(), 0, a_vector);
          state_assign=BDDFunction::ZERO();
          for (unsigned int i=0;i<a_vector.size();i++) {
            state_assign|=a_vector.at(i);
          }
        } else {
          state_assign=no_state_changes;
        }
        
        
        
        // do labels
        for (unsigned int a=0;a<module_helper::t_r_transition::countAssignments(r_trans);a++) {
          const vparser::pt_node& assignment=module_helper::t_r_transition::getAssignment(r_trans, a);
          
          if (!guard_helper::t_state_assignment::is_add_label(assignment)) {
            continue;  // skip state_assignments
          }
          
          const vparser::pt_node& add_label=guard_helper::t_state_assignment::get_add_label(assignment);
          
          const std::string label_name=guard_helper::t_add_label::getLabelName(add_label);
          
          Symb_Transition_Label_Var::ptr symb_tl = info->getTransitionLabel(label_name);
          if (!symb_tl) {
            throw "Label "+label_name+" does not exist!";
          }
          
          BDDFunction has_label = symb_tl->getHasLabelVar();
          if (bdd_is_bit_set(label_assign, has_label)) {
            throw "Trying to add label "+label_name+", already there!";
          }
          label_assign = label_assign.change(has_label);  // switch to "bit is set"
          
          // TODO: do value assignments
        }
      } else {
        state_assign=no_state_changes;
      }
      
      if (debugLevel>=3) {
        std::cerr << "State guard:\n";
        state_guard.printPaths();
        
        std::cerr << "IO guard:\n";
        io_guard.printPaths();
        
        std::cerr << "State assignments:\n";
        state_assign.printPaths();
        
        std::cerr << "Label assignments:\n";
        label_assign.printPaths();
      }
     
      transition_bdd=state_guard & io_guard & state_assign & label_assign;
    } else {
      // neither transition nor compact_transition
      THROW_UNIMPLEMENTED();
    }

    if (build_context->param_abstract_data_domain) {
      // restrict the ports to their valid values
      restrictToValidValuesPorts();

      // abstract away from concrete data values
      transition_bdd=transition_bdd.exists(getDataConstraintVars());
    }

    bool no_warning=false;

    if (build_context->restriction_manager &&
        build_context->restriction_manager->restrict_delta_x_with_initial) {
      transition_bdd&=getInitialStates();
      no_warning=true;
    }

    if (build_context->restriction_manager &&
        build_context->restriction_manager->restrict_delta_x) {
      transition_bdd&=build_context->restriction_manager->getStateRestrictions(*this, true);
      no_warning=true;
    }

    if (build_context->restriction_manager &&
        build_context->restriction_manager->restrict_delta_y) {
      transition_bdd&=build_context->restriction_manager->getStateRestrictions(*this, false);
      no_warning=true;
    }
    
    if (debugLevel>=3) {
      std::cerr << "Transition:\n";
      transition_bdd.printPaths();
    }

    if (!no_warning && transition_bdd==BDDFunction::ZERO()) {
      std::cerr << "Warning: transition equals false";
#ifdef REO_PARSER_POSITION_NODES
      std::cerr << " at " << getFilePositionToString(*transition.value.pos);
#endif

      if (info->component_instance) {
        std::cerr << " " << info->component_instance->getFQName() 
                  << " (" <<info->component_instance->getTypeName() << ")";
      }
      std::cerr << std::endl;
    }

    // Perform disjunction with delta
    VarManager::getInstance()->selectiveReorderingHook(4);
    BDDFunction delta_before=delta;
    delta|=transition_bdd;
    if (t>0) {
      BDD_OP_TRACE2("carml_delta", delta, delta_before, transition_bdd);
    }
    VarManager::getInstance()->selectiveReorderingHook(3);
  }

  // @TODO: cache!
  if (build_context->param_constrain_to_reachable_components){
    TRACE("building", "CARML", std::cout << " Compute reachable part of the component" << endl;);
    BDDFunction rs = calcReachableStates();
    reachabilityConstrain = rs;
    TRACE("building", "CARML", std::cout << " Constrain to reachable part of the component" << endl;);
    delta = delta.constrain(rs);
  }

  if (build_context->param_abstract_data_domain) {
    info->abstract_port_data_domain=true;
  }
}

const BDDFunctionContainer& Symb_CA_ReactiveModule::getRMVariables() {
  unsigned int i;

  if (have_rm_variables) {
    return rm_variables;
  }
  
  // add from states
  for (i=0;i<ci_module->getInterface()->getStateDescriptionCount();i++) {
    const SymbolicState::ptr ss_state = getState(ci_module->getInterface()->getStateDescription(i));
    rm_variables.push(*ss_state->vars.X);
  }

  // add to states
  for (i=0;i<ci_module->getInterface()->getStateDescriptionCount();i++) {
    const SymbolicState::ptr ss_state = getState(ci_module->getInterface()->getStateDescription(i));
    rm_variables.push(*ss_state->vars.Y);
  }
  
  // add in ports
  for (i=0;i<ci_module->getInterface()->getInPortCount();i++) {
    Symb_Channel_End::ptr symb_ce = info->port_to_bdd.find(ci_module->getInterface()->getInPort(i))->second;
    rm_variables.push(symb_ce->getPortActive());
    rm_variables.push(symb_ce->getDCVars());
  }
  
  // add out ports
  for (i=0;i<ci_module->getInterface()->getOutPortCount();i++) {
    Symb_Channel_End::ptr symb_ce = info->port_to_bdd.find(ci_module->getInterface()->getOutPort(i))->second;
    rm_variables.push(symb_ce->getPortActive());
    rm_variables.push(symb_ce->getDCVars());
  }
  
  // add transition labels
  for (i=0;i<ci_module->getInterface()->getTransitionLabelDescriptionCount();i++) {
    const Symb_Transition_Label_Var::ptr s_tl = info->getTransitionLabel(ci_module->getInterface()->getTransitionLabelDescription(i));
    rm_variables.push(s_tl->getHasLabelVar());
    if (s_tl->hasValueDomain()) {
      rm_variables.push(s_tl->getValueVars());
    }
  }

  have_rm_variables = true;
  return rm_variables;
}


