/***************************************************************************
 *   Copyright (C) 2008-2009 by Tobias Blechmann                           *
 *   blechman@tcs.inf.tu-dresden.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 @author Tobias Blechmann
 */

#include "reo_circuit.hpp"
#include <iostream>
#include <stack>
#include <vector>

#include <boost/pending/queue.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/graph/breadth_first_search.hpp>
#include <boost/graph/topological_sort.hpp>
#include <boost/graph/iteration_macros.hpp>
#include <boost/property_map/property_map.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/graph/graphviz.hpp>
#include <boost/graph/undirected_dfs.hpp>

#include "vereofy/common/xml_helpers.hpp"
#include "vereofy/system/builtin_components/builtin_mapper.hpp"

#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH

using namespace std;

namespace vereofy {    
  
  REO_Circuit::REO_Circuit(DataDomain::ptr port_data_domain_) :
    graph(new graph_t()), port_data_domain(port_data_domain_) {
  }
  
  
  REO_Circuit::~REO_Circuit() {
  }

  std::pair<REO_Circuit::ptr, REO_Graph_Object::rgo_map> REO_Circuit::clone() const {
    REO_Graph_Object::rgo_map mapping;
    REO_Circuit::ptr result(new REO_Circuit(port_data_domain));
    
    BGL_FORALL_VERTICES(v, *graph, graph_t) {
      REO_Graph_Object::ptr rgo=(*graph)[v].ptr;
      
      if (rgo->get_graph_object_type()==REO_Graph_Object::REO_CHANNEL_END) {
        continue; // Channel Ends are automatically cloned by their components, skip here
      }

      REO_Graph_Object::ptr cloned_rgo=rgo->clone(rgo, mapping);
      switch (cloned_rgo->get_graph_object_type()) {
      case REO_Graph_Object::REO_COMPONENT:
        result->add_Component(boost::static_pointer_cast<REO_Component>(cloned_rgo));
        break;
      case REO_Graph_Object::REO_NODE: {
        vertex_t cloned_vertex=add_vertex(*result->graph);
        (*result->graph)[cloned_vertex].ptr=cloned_rgo;
        cloned_rgo->setBOOST_Vertex(cloned_vertex);
        break;
      }
      default:
        throw "implementation error";
      }
    }

    BGL_FORALL_EDGES(e, *graph, graph_t) {
      REO_Graph_Object::ptr s=(*graph)[source(e, *graph)].ptr;
      REO_Graph_Object::ptr t=(*graph)[target(e, *graph)].ptr;

      CHECK(1, mapping.count(s)==1, "");
      CHECK(1, mapping.count(t)==1, "");
      REO_Graph_Object::ptr cloned_s=mapping[s];
      REO_Graph_Object::ptr cloned_t=mapping[t];

      add_edge(cloned_s->getBOOST_Vertex(), cloned_t->getBOOST_Vertex(), *result->graph);
    }

    return make_pair(result, mapping);
  }
  
  void REO_Circuit::add_Component(REO_Component::ptr comp) {
    TRACE("REO_Circuit.add_Component", "add", cout << "Add " << comp << " to " << this << endl;);
    vertex_t comp_vertex = add_vertex(*graph);
    (*graph)[comp_vertex].ptr = comp;
    comp->setBOOST_Vertex(comp_vertex);
    
    //add exported channel ends to actual REO_Circuit
    unsigned int source_ce_count = comp->getSourceCECount();
    for (unsigned int i=0; i < source_ce_count; ++i) {
      REO_Channel_End::ptr ce_ptr = comp->getSourceCE(i);
      vertex_t new_vertex = add_vertex(*graph);
      (*graph)[new_vertex].ptr = ce_ptr;
      ce_ptr->setBOOST_Vertex(new_vertex);
      add_edge(comp_vertex, new_vertex, *graph);
    }
    
    unsigned int sink_ce_count = comp->getSinkCECount();
    for (unsigned int i=0; i < sink_ce_count; ++i) {
      REO_Channel_End::ptr ce_ptr = comp->getSinkCE(i);
      vertex_t new_vertex = add_vertex(*graph);
      (*graph)[new_vertex].ptr = ce_ptr;
      ce_ptr->setBOOST_Vertex(new_vertex);
      add_edge(comp_vertex, new_vertex, *graph);
    }
    
  }
  
  REO_Node::ptr REO_Circuit::create_node(REO_Node::node_semantics_t node_semantic) {
    vertex_t new_vertex = add_vertex(*graph);
    REO_Node::ptr new_node = REO_Node::ptr(new REO_Node(node_semantic));
    new_node->setBOOST_Vertex(new_vertex);
    (*graph)[ new_vertex ].ptr = new_node;
    return new_node;
  }
  
  REO_Node::ptr REO_Circuit::join(REO_Channel_End::ptr ce1, REO_Channel_End::ptr ce2) {
    TRACE("REO_Circuit", "join", cout << "join(CE, CE) with " << ce1 << ", " << ce2 << endl);
    
    REO_Node::ptr n1=getConnectedNode(ce1);
    REO_Node::ptr n2=getConnectedNode(ce2);
    
    if (n1||n2) {
      if (n1 && n2) {
        return join(n1, n2);
      } else if (n1) {
        return join(n1, ce2);
      } else {
        return join(n2, ce1);
      }
    }
  
    REO_Node::ptr new_node = create_node();
    do_join(new_node, ce1);
    
    if (ce1==ce2) {
      // trying to join the same channel-end, don't add second edge
      // TODO: emit warning?
    } else {
      do_join(new_node, ce2);
    }

    return new_node;
  }
  
  
  REO_Node::ptr REO_Circuit::join(REO_Node::ptr node, REO_Channel_End::ptr ce) {
    TRACE("REO_Circuit", "join", cout << "join(Node, CE) with " << node << ", " << ce << endl);
    
    REO_Node::ptr n_ce=getConnectedNode(ce);
    if (n_ce) {
      return join(node, n_ce);
    }
    
    REO_Node::ptr real_node = getRealNode(node);
    
    TRACE("REO_Circuit", "join", cout << "join(Realnode, CE) with " << real_node << ", " << ce << endl);
    return do_join(real_node, ce);
  }
  
  
  REO_Node::ptr REO_Circuit::join(REO_Node::ptr _node1, REO_Node::ptr _node2) {
    TRACE("REO_Circuit", "join", cout << "join(Node, Node) with " << _node1 << ", " << _node2 << endl);
    REO_Node::ptr node1 = getRealNode(_node1);
    REO_Node::ptr node2 = getRealNode(_node2);
    TRACE("REO_Circuit", "join", cout << "join(Realnode, Realnode) with " << node1 << ", " << node2 << endl);
    
    if (node1==node2) {
      return node1;
    }

    return do_join(node1, node2);
  }
  

  REO_Node::ptr REO_Circuit::do_join(REO_Node::ptr node, REO_Channel_End::ptr ce) {
    REO_Node::ptr n_ce=getConnectedNode(ce);
    if (n_ce) {
      // if channel end was already connected, disconnect first
      TRACE("REO_Circuit", "do_join", cout << "outdegree(n_ce): " << \
            out_degree(n_ce->getBOOST_Vertex(), *graph)<<endl);

      TRACE_REO_OBJECT("REO_Circuit", "do_join", n_ce);
      TRACE_REO_OBJECT("REO_Circuit", "do_join", ce);

      remove_edge(n_ce->getBOOST_Vertex(), ce->getBOOST_Vertex(), *graph);
      TRACE("REO_Circuit", "do_join", cout << "outdegree(n_ce): " << \
            out_degree(n_ce->getBOOST_Vertex(), *graph)<<endl);
    }

    // ensure that node + ce are not plugged
    CHECK_SUCCESS(!node->isSourcePlugged());
    CHECK_SUCCESS(!node->isSinkPlugged());
    CHECK_SUCCESS(!ce->isPlugged());

    TRACE_REO_OBJECT("REO_Circuit", "do_join", node);
    TRACE_REO_OBJECT("REO_Circuit", "do_join", ce);
    add_edge(node->getBOOST_Vertex(), ce->getBOOST_Vertex(), *graph);
    return node;
  }

  REO_Node::ptr REO_Circuit::do_join(REO_Node::ptr node1, REO_Node::ptr node2) {
    if (node1->getNodeSemantic()!=node2->getNodeSemantic()) {
      throw "Can not join nodes with different semantic!";
    }

    REO_Node::ptr new_node = create_node(node1->getNodeSemantic());

    TRACE_REO_OBJECT("REO_Circuit", "do_join", node1);
    TRACE_REO_OBJECT("REO_Circuit", "do_join", node2);
    TRACE_REO_OBJECT("REO_Circuit", "do_join", new_node);

    // ensure that node + ce are not plugged
    CHECK_SUCCESS(!node1->isSourcePlugged());
    CHECK_SUCCESS(!node1->isSinkPlugged());
    CHECK_SUCCESS(!node2->isSourcePlugged());
    CHECK_SUCCESS(!node2->isSinkPlugged());

    // actually join the channel ends from node1 and node2 onto the new node
    vector<REO_Channel_End::ptr> ce_vec = getConnectedChannelEnds(node1);
    for (unsigned int i = 0; i<ce_vec.size(); ++i) {
      do_join(new_node, ce_vec[i]);
    }
    ce_vec = getConnectedChannelEnds(node2);
    for (unsigned int i = 0; i<ce_vec.size(); ++i) {
      do_join(new_node, ce_vec[i]);
    }

    CHECK(2, out_degree(node1->getBOOST_Vertex(), *graph)==0, \
          boost::lexical_cast<string>(out_degree(node1->getBOOST_Vertex(), *graph)));
    remove_vertex(node1->getBOOST_Vertex(), *graph);
    node1->setParent(new_node);

    CHECK(2, out_degree(node2->getBOOST_Vertex(), *graph)==0, \
          boost::lexical_cast<string>(node2->getBOOST_Vertex())+" = "+boost::lexical_cast<string>(out_degree(node2->getBOOST_Vertex(), *graph)));
    remove_vertex(node2->getBOOST_Vertex(), *graph);
    node2->setParent(new_node);

    return new_node;
  }
  
  PortDescription::ptr REO_Circuit::getPortDescription(REO_Channel_End::ptr ce, const string& name) {
    if (ce->pd_in_circuit) {
      return ce->pd_in_circuit;
    }
    
    PortDescription::port_type type;
    if (ce->isPlugged() || getConnectedNode(ce)) {
      type=PortDescription::INTERNAL_PORT;
    } else {
      if (ce->get_CE_Type()==REO_Channel_End::CE_SINK)
        type=PortDescription::OUT_PORT;
      else 
        type=PortDescription::IN_PORT;
    }
    
    
    
    PortDescription::ptr pd(new PortDescription(name, type, port_data_domain));

    ce->pd_in_circuit=pd;
    return pd;
  }
  
  PortDescription::ptr REO_Circuit::getPortDescription(REO_Node::ptr node, const string& name) {
    if (node->hasParent()) {
      throw "Node is not valid anymore";
    }
    
    if (node->getPortDescription()) {
      return node->getPortDescription();
    }

    PortDescription::ptr pd;

    switch (getNodeType(node)) {
    case REO_Node::NODE_FREE:
    case REO_Node::NODE_MIXED:
      pd=PortDescription::ptr(new PortDescription(name, PortDescription::INTERNAL_PORT, port_data_domain));
      break;
    case REO_Node::NODE_SINK:
      pd=PortDescription::ptr(new PortDescription(name, PortDescription::OUT_PORT, port_data_domain));
      break;
    case REO_Node::NODE_SOURCE:
      pd=PortDescription::ptr(new PortDescription(name, PortDescription::IN_PORT, port_data_domain));
      break;
    }
    
    node->setPortDescription(pd);
    return pd;
  }
  
  vector<REO_Channel_End::ptr> REO_Circuit::getConnectedChannelEnds(REO_Node::ptr node) const {
    if (node->hasParent()) {
      throw "Node is not valid anymore";
    }
    
    vector<REO_Channel_End::ptr> result;
    out_edge_iter_t out_it, out_end;
    edge_descriptor_t edge;
    for (tie(out_it, out_end) = out_edges(node->getBOOST_Vertex(), *graph);  out_it != out_end; ++out_it) {
      edge = *out_it;
      vertex_t src = source(edge, *graph);
      vertex_t targ = target(edge, *graph);
      vertex_t channelEnd_vertex = (node->getBOOST_Vertex()==src) ? targ:src;
      REO_Channel_End::ptr ce_ptr = dynamic_pointer_cast<REO_Channel_End>( (*graph)[ channelEnd_vertex ].ptr );
      result.push_back( ce_ptr );
    }
    return result;
  }
  
  
  vector<REO_Channel_End::ptr> REO_Circuit::getConnectedSourceChannelEnds(REO_Node::ptr node) {
    if (node->hasParent()) {
      throw "Node is not valid anymore";
    }
    
    vector<REO_Channel_End::ptr> result;
    out_edge_iter_t out_it, out_end;
    edge_descriptor_t edge;
    for (tie(out_it, out_end) = out_edges(node->getBOOST_Vertex(), *graph);  out_it != out_end; ++out_it) {
      edge = *out_it;
      vertex_t src = source(edge, *graph);
      vertex_t targ = target(edge, *graph);
      vertex_t channelEnd_vertex = (node->getBOOST_Vertex()==src) ? targ:src;
      REO_Channel_End::ptr ce_ptr = dynamic_pointer_cast<REO_Channel_End>( (*graph)[ channelEnd_vertex ].ptr );
      if (ce_ptr->get_CE_Type() == REO_Channel_End::CE_SOURCE) {
        result.push_back( ce_ptr );
      }
    }
    return result;
  }
  
  
  vector<REO_Channel_End::ptr> REO_Circuit::getConnectedSinkChannelEnds(REO_Node::ptr node) {
    if (node->hasParent()) {
      throw "Node is not valid anymore";
    }
    
    vector<REO_Channel_End::ptr> result;
    out_edge_iter_t out_it, out_end;
    edge_descriptor_t edge;
    for (tie(out_it, out_end) = out_edges(node->getBOOST_Vertex(), *graph);  out_it != out_end; ++out_it) {
      edge = *out_it;
      vertex_t src = source(edge, *graph);
      vertex_t targ = target(edge, *graph);
      vertex_t channelEnd_vertex = (node->getBOOST_Vertex()==src) ? targ:src;
      REO_Channel_End::ptr ce_ptr = dynamic_pointer_cast<REO_Channel_End>( (*graph)[ channelEnd_vertex ].ptr );
      if (ce_ptr->get_CE_Type() == REO_Channel_End::CE_SINK) {
        result.push_back( ce_ptr );
      }
    }
    return result;
  }
  
  REO_Node::node_type_t REO_Circuit::getNodeType(REO_Node::ptr node) {
    // source ces = outgoing 
    // sink ces = incoming
    unsigned int source_ce_count=getConnectedSourceChannelEnds(node).size();
    unsigned int sink_ce_count=getConnectedSinkChannelEnds(node).size();

    if (source_ce_count>0 && sink_ce_count>0) {
      CHECK(1, !node->isSinkPlugged(), "");
      CHECK(1, !node->isSourcePlugged(), "");
      return REO_Node::NODE_MIXED;
    }

    if (source_ce_count>0) {
      // check that the node is not plugged, although there are CEs at that side
      CHECK(1, !node->isSinkPlugged(), "");

      if (node->isSourcePlugged()) 
        return REO_Node::NODE_MIXED;
      else
        // there are source channel ends, but no sink channel ends -> NODE_SOURCE
        return REO_Node::NODE_SOURCE;
    }

    if (sink_ce_count>0) {
      // check that the node is not plugged, although there are CEs at that side
      CHECK(1, !node->isSourcePlugged(), "");
      
      if (node->isSinkPlugged()) 
        return REO_Node::NODE_MIXED;
      else
        // there are sink channel ends, but no source channel ends -> NODE_SINK
        return REO_Node::NODE_SINK;
    }
    
    // else: sink_ce_count==0, source_ce_count==0
    CHECK(1, sink_ce_count==0 && source_ce_count==0, "");
    
    if (node->isSourcePlugged() && node->isSinkPlugged()) {
      return REO_Node::NODE_MIXED;
    } else if (node->isSourcePlugged()) {
      return REO_Node::NODE_SINK;
    } else if (node->isSinkPlugged()) {
      return REO_Node::NODE_SOURCE;
    } else {
      return REO_Node::NODE_FREE;      
    }
  }
  
  REO_Node::ptr REO_Circuit::getConnectedNode(REO_Channel_End::ptr ce) const {
    out_edge_iter_t out_it, out_end;
    edge_descriptor_t edge;
    for (tie(out_it, out_end) = out_edges(ce->getBOOST_Vertex(), *graph);  out_it != out_end; ++out_it) {
      edge = *out_it;
      vertex_t src = source(edge, *graph);
      vertex_t targ = target(edge, *graph);
      vertex_t other_vertex = (ce->getBOOST_Vertex()==src) ? targ:src;
      
      REO_Graph_Object::ptr go=(*graph)[other_vertex].ptr;
      if (go->get_graph_object_type()==REO_Graph_Object::REO_NODE) {
        return boost::dynamic_pointer_cast<REO_Node>(go);
      }
    }
    
    return REO_Node::ptr();
  }
  
  REO_Component::ptr REO_Circuit::getComponentFor(REO_Channel_End::ptr ce) const {
    out_edge_iter_t out_it, out_end;
    edge_descriptor_t edge;
    for (tie(out_it, out_end) = out_edges(ce->getBOOST_Vertex(), *graph);  out_it != out_end; ++out_it) {
      edge = *out_it;
      vertex_t src = source(edge, *graph);
      vertex_t targ = target(edge, *graph);
      vertex_t other_vertex = (ce->getBOOST_Vertex()==src) ? targ:src;
      
      REO_Graph_Object::ptr go=(*graph)[other_vertex].ptr;
      if (go->get_graph_object_type()==REO_Graph_Object::REO_COMPONENT) {
        return boost::dynamic_pointer_cast<REO_Component>(go);
      }
    }
    
    return REO_Component::ptr();
  }


  vector<REO_Graph_Object::ptr> REO_Circuit::getAllGraphObjects() {
    vector<REO_Graph_Object::ptr> result;
    std::pair<vertex_iter_t, vertex_iter_t> vp;
    for (vp = vertices(*graph); vp.first != vp.second; ++vp.first) {
      result.push_back( (*graph)[*vp.first].ptr );
    }
    return result;
  }
  
  vector<REO_Node::ptr> REO_Circuit::getAllNodes() {
    vector<REO_Node::ptr> result;
    std::pair<vertex_iter_t, vertex_iter_t> vp;
    for (vp = vertices(*graph); vp.first != vp.second; ++vp.first) {
      if ((*graph)[*vp.first].ptr->get_graph_object_type() == REO_Graph_Object::REO_NODE) {
        result.push_back( dynamic_pointer_cast<REO_Node>((*graph)[*vp.first].ptr) );
      }
    }
    return result;
  }

  vector<REO_Node::ptr> REO_Circuit::getAllNodes(REO_Node::node_type_t filter) {
    vector<REO_Node::ptr> result;
    std::pair<vertex_iter_t, vertex_iter_t> vp;
    for (vp = vertices(*graph); vp.first != vp.second; ++vp.first) {
      if ((*graph)[*vp.first].ptr->get_graph_object_type() == REO_Graph_Object::REO_NODE) {
        REO_Node::ptr node=dynamic_pointer_cast<REO_Node>((*graph)[*vp.first].ptr);
        if (getNodeType(node)==filter) {
          result.push_back(node);
        }
      }
    }
    return result;
  }
  
  vector<REO_Channel_End::ptr> REO_Circuit::getAllCEs() {
    vector<REO_Channel_End::ptr> result;
    std::pair<vertex_iter_t, vertex_iter_t> vp;
    for (vp = vertices(*graph); vp.first != vp.second; ++vp.first) {
      if ((*graph)[*vp.first].ptr->get_graph_object_type() == REO_Graph_Object::REO_CHANNEL_END) {
        result.push_back( dynamic_pointer_cast<REO_Channel_End>((*graph)[*vp.first].ptr) );
      }
    }
    return result;
  }
  
  
  vector<REO_Channel_End::ptr> REO_Circuit::getAllUnconnectedSinkCEs() {
    vector<REO_Channel_End::ptr> tmp = getAllCEs();
    vector<REO_Channel_End::ptr> result;    
    for (unsigned int i=0; i<tmp.size(); ++i) {
      if (tmp[i]->get_CE_Type() == REO_Channel_End::CE_SINK &&
          !getConnectedNode(tmp[i])) {
        result.push_back(tmp[i]);
      }
    }
    return result;
  }
  
  
  vector<REO_Channel_End::ptr> REO_Circuit::getAllUnconnectedSourceCEs() {
    vector<REO_Channel_End::ptr> tmp = getAllCEs();
    vector<REO_Channel_End::ptr> result;    
    for (unsigned int i=0; i<tmp.size(); ++i) {
      if (tmp[i]->get_CE_Type() == REO_Channel_End::CE_SOURCE &&
          !getConnectedNode(tmp[i])) {
        result.push_back(tmp[i]);
      }
    }
    return result;
  }  
  
  
  void REO_Circuit::hide(REO_Node::ptr node) {
    node->setHidden(true);
  }
  
  void REO_Circuit::createMissingPortDescriptions() {
    vector<REO_Node::ptr> nodes=getAllNodes();
    
    for (unsigned int i=0;i<nodes.size();i++) {
      if (!nodes.at(i)->getPortDescription()) {
        std::string name="$node"+boost::lexical_cast<string>(i);
        // TODO: check that this node is not sink/source
        PortDescription::ptr pd(new PortDescription(name, PortDescription::INTERNAL_PORT, port_data_domain, true));
        nodes.at(i)->setPortDescription(pd);
        anonymous_port_descriptions.push_back(pd);
      }
    }
    
    vector<REO_Channel_End::ptr> ces=getAllCEs();
    for (unsigned int i=0;i<ces.size();i++) {
      if (!ces.at(i)->get_PortDescription_Circuit()) {
        std::string name="$ce"+boost::lexical_cast<string>(i);
        PortDescription::ptr pd(new PortDescription(name, PortDescription::INTERNAL_PORT, port_data_domain, true));
        ces.at(i)->pd_in_circuit=pd;
        anonymous_port_descriptions.push_back(pd);
      }
    }
  }
  
  const vector<PortDescription::ptr>& 
  REO_Circuit::getAnonymousPortDescriptions() const {
    return anonymous_port_descriptions;
  }
  
  
  
  
  struct BFS_visitor_t : public default_bfs_visitor {
    BFS_visitor_t(vector<vertex_t>& _vertex_vec) : vertex_vec(_vertex_vec) {};
    template < typename Vertex, typename Graph >
    void discover_vertex(Vertex u, const Graph & g) const {
      TRACE("REOCircuit.bfs", "discover_vertex", cout << u << endl);
      vertex_vec.push_back(u);
    }
  private:
    vector<vertex_t>& vertex_vec;
  };
  
  
  vector<REO_Graph_Object::ptr> REO_Circuit::getAllComponentsInBFS() {    
    vector<REO_Graph_Object::ptr> result;
    
    vector<vertex_t> vertex_vec;
    BFS_visitor_t vis(vertex_vec);
    std::map<vertex_t, default_color_type> color_map;

    std::pair<vertex_iter_t, vertex_iter_t> vp;
    for (vp = vertices(*graph); vp.first != vp.second; ++vp.first) {
      color_map[*vp.first]=white_color;
    }
    
    for (vp = vertices(*graph); vp.first != vp.second; ++vp.first) {
      boost::queue<vertex_t> Q;
      if (color_map[*vp.first]==white_color) {
        TRACE("REOCircuit.getAllComponentsInBFS", "visit", cout << *vp.first << endl);
        breadth_first_visit(*graph, *vp.first, Q, vis, 
                            make_assoc_property_map(color_map));
      }
    }
    
    for (unsigned int i=0;i<vertex_vec.size();i++) {
      vertex_t v=vertex_vec[i];
      REO_Graph_Object::ptr rgo=(*graph)[v].ptr;
      
      TRACE("REOCircuit.getAllComponentsInBFS", "vertex", cout << "BFS " << i << ", vertex " << vertex_vec[i] << " : " << rgo->get_Description() << endl);
      
      switch (rgo->get_graph_object_type()) {
        case REO_Graph_Object::REO_NODE:
        case REO_Graph_Object::REO_COMPONENT:
          result.push_back(rgo);
          break;
          
        case REO_Graph_Object::REO_CHANNEL_END:
          break;
          
        default:
          THROW_UNIMPLEMENTED();
      }
    }    
    
    return result;
  }

  
  struct DFS_visitor_t : public default_dfs_visitor {
    DFS_visitor_t(vector<vertex_t>& _vertex_vec) : vertex_vec(_vertex_vec) {};
    template < typename Vertex, typename Graph >
    void discover_vertex(Vertex u, const Graph & g) const {
      TRACE("REOCircuit.dfs", "discover_vertex", cout << u << endl);
      vertex_vec.push_back(u);
    }
  private:
    vector<vertex_t>& vertex_vec;
  };

  
  vector<REO_Graph_Object::ptr> REO_Circuit::getAllComponentsInDFS() {
    vector<REO_Graph_Object::ptr> result;
    
    vector<vertex_t> vertex_vec;
    std::map<vertex_t, default_color_type> color_map;

    //    std::pair<vertex_iter_t, vertex_iter_t> vp;
    //    for (vp = vertices(*graph); vp.first != vp.second; ++vp.first) {
    //      color_map[*vp.first]=white_color;
    //    }

    DFS_visitor_t vis(vertex_vec);
    
    depth_first_search(*graph, vis, make_assoc_property_map(color_map));

    for (unsigned int i=0;i<vertex_vec.size();i++) {
      vertex_t v=vertex_vec[i];
      REO_Graph_Object::ptr rgo=(*graph)[v].ptr;
      
      TRACE("REOCircuit.getAllComponentsInDFS", "vertex", cout << "DFS " << i << ", vertex " << vertex_vec[i] << " : " << rgo->get_Description() << endl);
      
      switch (rgo->get_graph_object_type()) {
        case REO_Graph_Object::REO_NODE:
        case REO_Graph_Object::REO_COMPONENT:
          result.push_back(rgo);
          break;
          
        case REO_Graph_Object::REO_CHANNEL_END:
          break;
          
        default:
          THROW_UNIMPLEMENTED();
      }
    }    
    
    return result;
  }

  
  vector<REO_Graph_Object::ptr> REO_Circuit::getAllComponentsInDFS(bool sort_nodes, bool nodes_first, bool sort_sink_source, bool sink_first) {
    throw "Currently broken";

    vector<REO_Graph_Object::ptr> result;
    vector<vertex_t> vertex_vec;
    stack<vertex_t> vertex_stack;

    std::map<vertex_t, default_color_type> color_map;
    std::pair<vertex_iter_t, vertex_iter_t> vp;
    for (vp = vertices(*graph); vp.first != vp.second; ++vp.first) {
      color_map[*vp.first]=white_color;
    }
    
    for (vp = vertices(*graph); vp.first != vp.second; ++vp.first) {
      if (color_map[*vp.first]==white_color) {
        TRACE("REOCircuit.getAllComponentsInDFS", "visit", cout << *vp.first << endl);
        vertex_stack.push(*vp.first);
        while (!vertex_stack.empty()) {
          vertex_t actual_vertex = vertex_stack.top();
          if (color_map[actual_vertex] == white_color) {
            //this is a new vertex
            color_map[actual_vertex] = gray_color;
            if (sort_nodes) {
              REO_Graph_Object::ptr rgo=(*graph)[actual_vertex].ptr;
              if (nodes_first) {
                if (rgo->get_graph_object_type()==REO_Graph_Object::REO_NODE) vertex_vec.push_back(actual_vertex);
              } else {
                if (rgo->get_graph_object_type()!=REO_Graph_Object::REO_NODE) vertex_vec.push_back(actual_vertex);
              }
            }
            //now find all nodes with direct connection
            vector<vertex_t> connected_vertices;            
            out_edge_iter_t out_it, out_end;
            edge_descriptor_t edge;
            for (tie(out_it, out_end) = out_edges(actual_vertex, *graph);  out_it != out_end; ++out_it) {
              edge = *out_it;
              vertex_t src = source(edge, *graph);
              vertex_t targ = target(edge, *graph);
              vertex_t channelEnd_vertex = (actual_vertex==src)?targ:src;
              if (color_map[channelEnd_vertex]==white_color) connected_vertices.push_back(channelEnd_vertex);
            }
            //the vertices in connected_vertices are all un-visited
            //now sort them and add them in the vertex_stack
            if (sort_sink_source) {
              for (unsigned int i=0; i<connected_vertices.size(); i++) {
                REO_Graph_Object::ptr rgo=(*graph)[connected_vertices.at(i)].ptr;
                if (rgo->get_graph_object_type()==REO_Graph_Object::REO_CHANNEL_END) {
                  REO_Channel_End::ptr ce_ptr = dynamic_pointer_cast<REO_Channel_End>( (*graph)[ connected_vertices.at(i) ].ptr );
                  if (sink_first) {
                    //we should visit the sink channel ends first
                    //since vertex_stack is a stack we have to put the SOURCEs in first
                    if (ce_ptr->get_CE_Type() != REO_Channel_End::CE_SINK) 
                      vertex_stack.push(connected_vertices.at(i)); 
                  } else {
                    if (ce_ptr->get_CE_Type() == REO_Channel_End::CE_SINK) 
                      vertex_stack.push(connected_vertices.at(i));
                  }
                } else vertex_stack.push(connected_vertices.at(i));
              }
              for (unsigned int i=0; i<connected_vertices.size(); i++) {
                REO_Graph_Object::ptr rgo=(*graph)[connected_vertices.at(i)].ptr;
                if (rgo->get_graph_object_type()==REO_Graph_Object::REO_CHANNEL_END) {
                  REO_Channel_End::ptr ce_ptr = dynamic_pointer_cast<REO_Channel_End>( (*graph)[ connected_vertices.at(i) ].ptr );
                  if (sink_first) {
                    //this is basically the other way around we have to add everything we did not add in the for loop above
                    if (ce_ptr->get_CE_Type() == REO_Channel_End::CE_SINK) 
                      vertex_stack.push(connected_vertices.at(i)); 
                  } else {
                    if (ce_ptr->get_CE_Type() != REO_Channel_End::CE_SINK) 
                      vertex_stack.push(connected_vertices.at(i));
                  }
                }
              }
            } else {
              for (unsigned int i=0; i<connected_vertices.size(); i++) vertex_stack.push(connected_vertices.at(i));
            }
          } else { 
            //color should be gray == this vertex was already visited
            vertex_stack.pop();
            if (sort_nodes) {
              REO_Graph_Object::ptr rgo=(*graph)[actual_vertex].ptr;
              if (nodes_first) {
                if (rgo->get_graph_object_type()!=REO_Graph_Object::REO_NODE) vertex_vec.push_back(actual_vertex);
              } else {
                if (rgo->get_graph_object_type()==REO_Graph_Object::REO_NODE) vertex_vec.push_back(actual_vertex);
              }
            } else vertex_vec.push_back(actual_vertex);
          }
        }
      }
    }
    
    for (unsigned int i=0;i<vertex_vec.size();i++) {
      vertex_t v=vertex_vec[i];
      REO_Graph_Object::ptr rgo=(*graph)[v].ptr;
      
      TRACE("REOCircuit.getAllComponentsInDFS", "vertex", cout << "DFS " << i << ", vertex " << vertex_vec[i] << " : " << rgo->get_Description() << endl);
      
      switch (rgo->get_graph_object_type()) {
        case REO_Graph_Object::REO_NODE:
        case REO_Graph_Object::REO_COMPONENT:
          result.push_back(rgo);
          break;
          
        case REO_Graph_Object::REO_CHANNEL_END:
          break;
          
        default:
          THROW_UNIMPLEMENTED();
      }
    }    
    
    return result;
  }


  bool REO_Circuit::isInFlowDirection(REO_Graph_Object::ptr from, REO_Graph_Object::ptr to) {
    bool swapped=false;
    if (to->get_graph_object_type()==REO_Graph_Object::REO_COMPONENT) {
      swapped=true;
      from.swap(to);
    } else if (to->get_graph_object_type()==REO_Graph_Object::REO_NODE) {
      swapped=true;
      from.swap(to);
    }

    bool in_flow;

    // "to" should be channel end now
    CHECK_SUCCESS(to->get_graph_object_type()==REO_Graph_Object::REO_CHANNEL_END);
    REO_Channel_End::ptr ce=boost::static_pointer_cast<REO_Channel_End>(to);
    
    if (from->get_graph_object_type()==REO_Graph_Object::REO_COMPONENT) {
      if (ce->get_CE_Type()==REO_Channel_End::CE_SINK) {
        in_flow=true;
      } else {
        in_flow=false;
      }
    } else if (from->get_graph_object_type()==REO_Graph_Object::REO_NODE) {
      if (ce->get_CE_Type()==REO_Channel_End::CE_SINK) {
        in_flow=false;
      } else {
        in_flow=true;
      }
    } else {
      THROW_UNIMPLEMENTED();
    }

    return swapped ? !in_flow : in_flow;
  }
 
  std::vector<REO_Graph_Object::ptr> REO_Circuit::getAllComponentsInTopologicalSort() {
    typedef boost::adjacency_list<boost::setS, boost::listS, 
      boost::bidirectionalS, 
      vertex_property,
      boost::no_property> dir_graph_t;

    typedef graph_traits< dir_graph_t >::vertex_descriptor dir_vertex_t;
    typedef graph_traits< dir_graph_t >::vertex_iterator dir_vertex_iter_t;
    typedef graph_traits< dir_graph_t >::edge_descriptor dir_edge_descriptor_t;

    dir_graph_t d_graph;

    std::map<vertex_t, dir_vertex_t> map_to_dir;

    // copy vertices
    BGL_FORALL_VERTICES(v, *graph, graph_t) {
      dir_vertex_t dir_v=add_vertex(d_graph);
      d_graph[dir_v].vertex=v;
      map_to_dir.insert(make_pair(v, dir_v));
    }

    BGL_FORALL_EDGES(e, *graph, graph_t) {
      vertex_t s=source(e, *graph);
      vertex_t t=target(e, *graph);

      CHECK(1, map_to_dir.count(s)==1, "");
      CHECK(1, map_to_dir.count(t)==1, "");

      if (isInFlowDirection((*graph)[s].ptr, (*graph)[t].ptr)) {
        add_edge(map_to_dir[s], map_to_dir[t], d_graph);
      } else {
        add_edge(map_to_dir[t], map_to_dir[s], d_graph);
      }
    }

    dfs_remove_back_edges(d_graph);
    
    typedef std::vector<dir_vertex_t> top_ordering_t;
    top_ordering_t top_ordering;
    std::map<dir_vertex_t, default_color_type> color_map;
    topological_sort(d_graph, 
                     std::back_inserter(top_ordering),
                     boost::color_map(make_assoc_property_map(color_map)) );

    std::vector<REO_Graph_Object::ptr> result;
    for (top_ordering_t::reverse_iterator it=top_ordering.rbegin();
         it!=top_ordering.rend();
         ++it) {

      dir_vertex_t d_v=*it;
      vertex_t v=d_graph[d_v].vertex;
      REO_Graph_Object::ptr rgo=(*graph)[v].ptr;
      
      switch (rgo->get_graph_object_type()) {
        case REO_Graph_Object::REO_NODE:
        case REO_Graph_Object::REO_COMPONENT:
          result.push_back(rgo);
          break;
          
        case REO_Graph_Object::REO_CHANNEL_END:
          break;
          
        default:
          THROW_UNIMPLEMENTED();
      }
    }

    return result;
  }
  
 
  void REO_Circuit::print() const {
    toDotBoost(cout);
  }
  
  void REO_Circuit::toDot(ostream& out,
                          const reo_to_names_map* names,
                          bool is_subgraph) const {
    if (!is_subgraph) {
      out << "digraph {\n";
    }

    std::pair<vertex_iter_t, vertex_iter_t> vp;
    for (vp = vertices(*graph); vp.first != vp.second; ++vp.first) {
      REO_Graph_Object::ptr rgo=(*graph)[*vp.first].ptr;
      switch (rgo->get_graph_object_type()) {
      case REO_Graph_Object::REO_COMPONENT: {
        REO_Component::ptr r_component=boost::static_pointer_cast<REO_Component>(rgo);
        out << "_" << rgo << " [shape=plaintext,label=<";

        out << "<table><tr>";
        
        for (unsigned int i=0;i<r_component->getSourceCECount();i++) {
          REO_Channel_End::ptr ce=r_component->getSourceCE(i);
          PortDescription::ptr pd_in_circuit=ce->get_PortDescription_Circuit();
          string port_name=(pd_in_circuit ? 
                            boost::lexical_cast<string>(pd_in_circuit) :
                            r_component->getNameForCE(ce));
          
          string nr_cell="<td bgcolor=\"blue\" port=\"_"+
            port_name+"\">"
            +boost::lexical_cast<string>(i)+
            (ce->isPlugged() ? "*" :"")+
            "</td>";

          if (names && names->count(ce)>0) {
            out << "<td><table border=\"0\">";
            out << "<tr>" << nr_cell << "</tr>";
            
            for (reo_to_names_const_iterator it=names->equal_range(ce);
                 it.first!=it.second;
                 ++it.first) {
              string name=it.first->second;
              out << "<tr><td>";
              if (boost::starts_with(name, "source")) {
                out << "<font color=\"blue\">" << name << "</font>";
              } else {
                out << name;
              }
              out << "</td></tr>";              
            }
            out << "</table></td>";
          } else {
            out << nr_cell;
          }
        }

        out << "<td>";
        out << r_component->get_Label();
        out << "</td>";

        for (unsigned int i=0;i<r_component->getSinkCECount();i++) {
          REO_Channel_End::ptr ce=r_component->getSinkCE(i);
          PortDescription::ptr pd_in_circuit=ce->get_PortDescription_Circuit();
          string port_name=(pd_in_circuit ? 
                            boost::lexical_cast<string>(pd_in_circuit) :
                            r_component->getNameForCE(ce));
         
          string nr_cell="<td bgcolor=\"lightblue\" port=\"_"+
            port_name+"\">"
            +boost::lexical_cast<string>(i)+
            (ce->isPlugged() ? "*" :"")+
            "</td>";

          if (names && names->count(ce)>0) {
            out << "<td><table border=\"0\">";
            out << "<tr>" << nr_cell << "</tr>";
            
            for (reo_to_names_const_iterator it=names->equal_range(ce);
                 it.first!=it.second;
                 ++it.first) {
              string name=it.first->second;
              out << "<tr><td>";
              if (boost::starts_with(name, "sink")) {
                out << "<font color=\"blue\">" << name << "</font>";
              } else {
                out << name;
              }
              out << "</td></tr>";              
            }
            out << "</table></td>";
          } else {
            out << nr_cell;
          }
        }

        out << "</tr></table>";
        out <<">];\n";
        break;
      }
      case REO_Graph_Object::REO_NODE: {
        REO_Node::ptr r_node=boost::static_pointer_cast<REO_Node>(rgo);
        out << "_" << r_node;
        if (names && names->count(r_node)>0) {
          out << "[shape=ellipse, label=<";
          out << "<table border=\"0\">";
        
          if (r_node->isSourcePlugged()) {
            out << "<tr><td>*</td></tr>";
          }

          switch (r_node->getNodeSemantic()) {
          case REO_Node::NODE_REPLICATE:
            // nothing to do
            break;
          case REO_Node::NODE_ROUTE:
            out << "<tr><td><font color=\"red\">X</font></td></tr>";
            break;
          }
  
          for (reo_to_names_const_iterator it=names->equal_range(r_node);
               it.first!=it.second;
               ++it.first) {
            string name=it.first->second;
            out << "<tr><td>";
            if (boost::starts_with(name, "sink") ||
                boost::starts_with(name, "source")) {
              out << "<font color=\"blue\">" << name << "</font>";
            } else {
              out << name;
            }
            out << "</td></tr>";              
          }


          if (r_node->isSinkPlugged()) {
            out << "<tr><td>*</td></tr>";
          }

          out << "</table>";
          out << ">]";
        } else {
          switch (r_node->getNodeSemantic()) {
          case REO_Node::NODE_REPLICATE:
            out << "[shape=point]";
            break;
          case REO_Node::NODE_ROUTE:
            out << "[shape=point,color=red]";
            break;
          }
        }
        out << ";";
        std::vector<REO_Channel_End::ptr> ces=getConnectedChannelEnds(r_node);

        for (unsigned int i=0;i<ces.size();i++) {
          REO_Channel_End::ptr ce=ces.at(i);
          REO_Component::ptr component=getComponentFor(ce);

          PortDescription::ptr pd_in_circuit=ce->get_PortDescription_Circuit();
          string port_name=(pd_in_circuit ? 
                            boost::lexical_cast<string>(pd_in_circuit) :
                            component->getNameForCE(ce));

          string name="_"+
            boost::lexical_cast<string>(component)+
            ":_"+port_name;

          switch (ce->get_CE_Type()) {
          case REO_Channel_End::CE_SOURCE:
            out << "_" << r_node << " -> " << name << ";\n" ;
            break;
          case REO_Channel_End::CE_SINK:
            out << name << " -> _" << r_node << ";\n" ;
            break;
          }
        }
        break;
      }
      case REO_Graph_Object::REO_CHANNEL_END:
        break;
      }
    }

    if (!is_subgraph) {
      out << "}\n";
    }
  }



  static void xml_createNode(ostream& out,
                             const std::string& id,
                             const std::string &name,
                             const std::string& semantics,
                             const std::string& source_ends,
                             const std::string& sink_ends,
                             const std::string& aliases) {

    xml_tag_open(out, "nodes",
                 xml_parameter("xmi:type", "cwi.reo:Node")+
                 xml_parameter("xmi:id", id)+
                 xml_parameter("type", semantics)+
                 xml_parameter("sourceEnds", source_ends)+
                 xml_parameter("sinkEnds", sink_ends)+
                 xml_parameter("name", name));            
    xml_tag_simple(out, "properties",
                   xml_parameter("xmi:type", "cwi.reo:Property")+
                   xml_parameter("key", "vereofy_node_aliases")+
                   xml_parameter("value", aliases));
    xml_tag_close(out, "nodes");
  }

  static std::string xml_createNodeForUnconnectedEnd(ostream& out,
                                                     REO_Channel_End::ptr ce) {
    std::string connected_node_id=boost::lexical_cast<string>(ce.get())+"_node";
    
    bool is_sink=ce->get_CE_Type()==REO_Channel_End::CE_SINK;

    std::string node_name=ce->get_PortDescription_Circuit()->getName();
    if (boost::starts_with(node_name, "$")) {
      node_name="";
    }
    
    std::string ce_id=boost::lexical_cast<string>(ce.get());

    xml_createNode(out, 
                   connected_node_id,
                   node_name,
                   "replicate",
                   !is_sink ? ce_id : std::string(""),
                   is_sink ? ce_id : std::string(""),
                   ce->get_PortDescription_Circuit()->getAliasString());

    return connected_node_id;
  }





  void REO_Circuit::toReoXML(ostream& out,
                             const std::string& name, 
                             const reo_to_names_map* names,
                             const std::vector<std::string>& input_files,
                             bool is_subgraph) const {

    BuiltinMapper mapper;

    stringstream later_nodes;

    bool only_builtin=true;

    if (!is_subgraph) {
      out << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
      out << "<xmi:XMI xmi:version=\"2.0\" xmlns:xmi=\"http://www.omg.org/XMI\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:channels=\"http://www.cwi.nl/reo/channels\" xmlns:cwi.reo=\"http://www.cwi.nl/reo\" xmlns:notation=\"http://www.eclipse.org/gmf/runtime/1.0.2/notation\" xsi:schemaLocation=\"http://www.cwi.nl/reo/channels http://www.cwi.nl/reo#//channels\">\n";
      
      xml_tag_open(out, "cwi.reo:Module",
                   xml_parameter("xmi:id", "module"));
    }

    std::string connector_name=name;

    boost::algorithm::replace_all(connector_name, "<", "_");
    boost::algorithm::replace_all(connector_name, ">", "");
    boost::algorithm::replace_all(connector_name, ",", "_");

    xml_tag_open(out, "connectors",
                 xml_parameter("xmi:type", "cwi.reo:Connector")+
                 xml_parameter("name", connector_name)+
                 xml_parameter("xmi:id", this));
    out << "\n";

    std::pair<vertex_iter_t, vertex_iter_t> vp;
    for (vp = vertices(*graph); vp.first != vp.second; ++vp.first) {
      REO_Graph_Object::ptr rgo=(*graph)[*vp.first].ptr;
      switch (rgo->get_graph_object_type()) {
      case REO_Graph_Object::REO_COMPONENT: {
        REO_Component::ptr r_component=boost::static_pointer_cast<REO_Component>(rgo);

        stringstream component_xml;

        BuiltinMapper::mapped::ptr mapped=mapper.mapToReoXML(component_xml, r_component);

        if (!mapped) {
          std::string component_name=r_component->getComponentInstance()->getName();
          if (boost::starts_with(component_name, "$")) {
            // anonymous name
            component_name="";
          }
          
          xml_tag_open(component_xml, "primitives",
                       xml_parameter("xmi:type", "cwi.reo:Component")+
                       xml_parameter("xmi:id", r_component.get())+
                       xml_parameter("name", component_name).only_if(component_name!=""));

          xml_tag_simple(component_xml, "properties",
                         xml_parameter("xmi:type", "cwi.reo:Property")+
                         xml_parameter("key", "type")+
                         xml_parameter("value", "vereofy:"+r_component->getComponentInstance()->getTypeName()));

          only_builtin=false;

          mapped.reset(new BuiltinMapper::mapped("primitives", "component"));
        }

        component_xml << "\n";

        for (unsigned int i=0;i<r_component->getSourceCECount();i++) {
          REO_Channel_End::ptr ce=r_component->getSourceCE(i);

          REO_Node::ptr connected_node=getConnectedNode(ce);
          string connected_node_id;
          if (connected_node) {
            connected_node_id=boost::lexical_cast<string>(connected_node.get());
          } else {
            connected_node_id=xml_createNodeForUnconnectedEnd(later_nodes, ce);
          }

          xml_tag_simple(component_xml, "sourceEnds",
                         xml_parameter("xmi:type", "cwi.reo:SourceEnd")+
                         xml_parameter("xmi:id", ce.get())+
                         xml_parameter("name", "source["+boost::lexical_cast<string>(i)+"]")+
                         xml_parameter("node", connected_node_id));
          component_xml << "\n";
        }

        for (unsigned int i=0;i<r_component->getSinkCECount();i++) {
          REO_Channel_End::ptr ce=r_component->getSinkCE(i);

          REO_Node::ptr connected_node=getConnectedNode(ce);
          string connected_node_id;
          if (connected_node) {
            connected_node_id=boost::lexical_cast<string>(connected_node.get());
          } else {
            connected_node_id=xml_createNodeForUnconnectedEnd(later_nodes, ce);
          }
          
          xml_tag_simple(component_xml, "sinkEnds",
                         xml_parameter("xmi:type", "cwi.reo:SinkEnd")+
                         xml_parameter("xmi:id", ce.get())+
                         xml_parameter("name", "sink["+boost::lexical_cast<string>(i)+"]")+
                         xml_parameter("node", connected_node_id));
          component_xml << "\n";
        }

        xml_tag_close(component_xml, mapped->xml_tag);
        component_xml << "\n";

        out << component_xml.str();

        break;
      }
      case REO_Graph_Object::REO_NODE: {
        REO_Node::ptr r_node=boost::static_pointer_cast<REO_Node>(rgo);

        string semantics;
        switch (r_node->getNodeSemantic()) {
          case REO_Node::NODE_REPLICATE:
            semantics="replicate";
            break;
          case REO_Node::NODE_ROUTE:
            semantics="route";
            break;
        }
        
        string name="";
        reo_to_names_const_iterator it=names->equal_range(r_node);
        if (it.first!=it.second) {
          name=it.first->second;
        }

        if (boost::starts_with(name, "$")) {
          name="";
        }        

        bool first_source=true, first_sink=true;
        stringstream source_ends;
        stringstream sink_ends;


        std::vector<REO_Channel_End::ptr> ces=getConnectedChannelEnds(r_node);

        for (unsigned int i=0;i<ces.size();i++) {
          REO_Channel_End::ptr ce=ces.at(i);
          REO_Component::ptr component=getComponentFor(ce);

          switch (ce->get_CE_Type()) {
          case REO_Channel_End::CE_SOURCE:
            if (!first_source) source_ends << " ";
            first_source=false;
            source_ends << ce.get();
            break;
          case REO_Channel_End::CE_SINK:
            if (!first_sink) sink_ends << " ";
            first_sink=false;
            sink_ends << ce.get();
            break;
          }
        }

        xml_createNode(out, 
                       boost::lexical_cast<std::string>(r_node.get()),
                       name,
                       semantics,
                       source_ends.str(),
                       sink_ends.str(),
                       r_node->getPortDescription()->getAliasString());
        break;
      }
      case REO_Graph_Object::REO_CHANNEL_END:
        break;
      }
    }

    out << later_nodes.str() << "\n";

    xml_tag_close(out, "connectors");

    if (!is_subgraph) {
      if (!only_builtin && input_files.size()>0) {
        std::string vereofy_includes="";

        for (unsigned int i=0;i<input_files.size();i++) {
          vereofy_includes+="#include \""+input_files.at(i)+"\" ";
        }

        xml_tag_simple(out, "properties",
                       xml_parameter("xmi:type", "cwi.reo:Property")+
                       xml_parameter("key", "vereofy_preamble")+
                       xml_parameter("value", vereofy_includes));
      }

      xml_tag_close(out, "cwi.reo:Module");

      xml_tag_simple(out, "notation:Diagram",
                     xml_parameter("type", "Reo")+
                     xml_parameter("element", "module")+
                     xml_parameter("measurementUnit", "Pixel"));
      xml_tag_close(out, "xmi:XMI");
    }
  }



  void REO_Circuit::toDotBoost(ostream& out) const {
    map<vertex_t, unsigned int> id_map;

    unsigned int i=0;
    BGL_FORALL_VERTICES(v, *graph, graph_t)
      id_map[v]=i++;

    boost::write_graphviz(out, *graph, 
                          Vertex_Property_Writer(*graph),
                          default_writer(),
                          default_writer(),
                          make_assoc_property_map(id_map));
  }
  
  
  REO_Node::ptr REO_Circuit::getRealNode(REO_Node::ptr node) const {
    while (node->hasParent()) {
      node=node->getParent();
    }
    return node;
  }


  void REO_Circuit::setParent(ComponentInstance* parent) {
    vector<REO_Graph_Object::ptr> components=getAllComponentsInDFS();
    for (vector<REO_Graph_Object::ptr>::iterator it=components.begin();
         it!=components.end();
         ++it) {
      REO_Graph_Object::ptr rgo=*it;
      if (rgo->get_graph_object_type() == REO_Graph_Object::REO_COMPONENT) {
        (boost::static_pointer_cast<REO_Component>(rgo))->getComponentInstance()->setParent(parent);
      }
    }    
  }

  const AtomicPropositionSpec::vector &REO_Circuit::getAtomicPropositions() const {
    return atomic_propositions;
  }

  
  void REO_Circuit::attachAtomicProposition(const std::string& name, const std::string& spec) {
    atomic_propositions.push_back(AtomicPropositionSpec(name, spec));
  }
  

  // ===================== node sorter =============================

  REO_Circuit::node_sorter::node_sorter(bool nodes_first_) : 
    nodes_first(nodes_first_) {}

  bool REO_Circuit::node_sorter::operator()(const REO_Graph_Object::ptr a,
                                            const REO_Graph_Object::ptr b) const {
    if (a->get_graph_object_type()==b->get_graph_object_type()) {
      return false;
    }
    
    if (a->get_graph_object_type()==REO_Graph_Object::REO_NODE) {
      if (nodes_first) {return true;} else {return false;}
    }
    
    if (nodes_first) {return false;} else {return true;}
  }

}

