/***************************************************************************
 *   Copyright (C) 2008-2009 by Tobias Blechmann                           *
 *   blechman@tcs.inf.tu-dresden.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 @author Tobias Blechmann
 */

#ifndef REO_GRAPH_HPP
#define REO_GRAPH_HPP

#include <iostream>
#include <boost/smart_ptr.hpp>

#include "vereofy/common/errors.hpp"
#include "vereofy/common/shared_ptr_map.hpp"
#include "vereofy/common/graph_helpers.hpp"
#include "component_instance.hpp"
#include "component_interface.hpp"

#include "reo_channel_end.hpp"
#include "reo_node.hpp"
#include "reo_component.hpp"

#include "atomic_proposition_spec.hpp"

namespace vereofy {

#define TRACE_REO_CIRCUIT_DOT(scope, name, circuit)			\
  TRACE(scope, name,							\
	{								\
    std::string filename=std::string(scope)+"_"+name+".";				\
    filename+=boost::lexical_cast<std::string>(Tracing::getTraceCount(TRACE_GET_FULL_NAME(scope, name))); \
    filename+=".dot";							\
    cout << "Trace: Writing DOT file " << filename <<endl;		\
    ofstream out(filename.c_str());					\
    if (out) {								\
      (circuit).toDot(out);						\
    } else {								\
      cout << "Couldn't write " << filename << endl;			\
    }									\
  }									\
	)

  
  class REO_Circuit {
  public:
    typedef boost::shared_ptr<REO_Circuit> ptr;
    typedef shared_ptr_multimap<REO_Graph_Object::ptr, std::string>::type reo_to_names_map;
    typedef pair<reo_to_names_map::const_iterator, reo_to_names_map::const_iterator> reo_to_names_const_iterator;

    REO_Circuit(DataDomain::ptr port_data_domain);
    ~REO_Circuit();

    std::pair<REO_Circuit::ptr, REO_Graph_Object::rgo_map> clone() const;

    void print() const;
    void toDot(ostream& out,
               const reo_to_names_map* names=0,
               bool is_subgraph=false) const;
    void toDotBoost(ostream& out) const;
    void toReoXML(ostream& out,
                  const std::string& name, 
                  const reo_to_names_map* names=0,
                  const std::vector<std::string>& input_files=std::vector<std::string>(),
                  bool is_subgraph=false) const;
    
    void add_Component(REO_Component::ptr);
    REO_Node::ptr create_node(REO_Node::node_semantics_t node_semantic=REO_Node::NODE_REPLICATE);
    REO_Node::ptr join(REO_Channel_End::ptr, REO_Channel_End::ptr);
    REO_Node::ptr join(REO_Node::ptr, REO_Channel_End::ptr);
    REO_Node::ptr join(REO_Node::ptr, REO_Node::ptr);
    void hide(REO_Node::ptr);

    PortDescription::ptr getPortDescription(REO_Channel_End::ptr, const std::string& name);
    PortDescription::ptr getPortDescription(REO_Node::ptr, const std::string& name);

    vector<REO_Channel_End::ptr> getConnectedChannelEnds(REO_Node::ptr) const;
    vector<REO_Channel_End::ptr> getConnectedSourceChannelEnds(REO_Node::ptr);
    vector<REO_Channel_End::ptr> getConnectedSinkChannelEnds(REO_Node::ptr);

    REO_Node::node_type_t getNodeType(REO_Node::ptr);
    
    REO_Node::ptr getRealNode(REO_Node::ptr) const;

    REO_Node::ptr getConnectedNode(REO_Channel_End::ptr ce) const;
    REO_Component::ptr getComponentFor(REO_Channel_End::ptr ce) const;

    vector<REO_Graph_Object::ptr> getAllGraphObjects();
    vector<REO_Node::ptr> getAllNodes();
    vector<REO_Node::ptr> getAllNodes(REO_Node::node_type_t filter);

    vector<REO_Channel_End::ptr> getAllCEs();
    vector<REO_Channel_End::ptr> getAllUnconnectedSinkCEs();
    vector<REO_Channel_End::ptr> getAllUnconnectedSourceCEs();
    vector<REO_Graph_Object::ptr> getAllComponentsInBFS();
    vector<REO_Graph_Object::ptr> getAllComponentsInDFS();
    vector<REO_Graph_Object::ptr> getAllComponentsInDFS(bool sort_nodes, bool nodes_first, bool sort_inout, bool in_first);
    vector<REO_Graph_Object::ptr> getAllComponentsInTopologicalSort();

    const vector<PortDescription::ptr>& getAnonymousPortDescriptions() const;

    bool isInFlowDirection(REO_Graph_Object::ptr from, REO_Graph_Object::ptr to);

    void createMissingPortDescriptions();

    // set the parent component instance for all the subcomponents
    void setParent(ComponentInstance* parent);

    const AtomicPropositionSpec::vector &getAtomicPropositions() const;
    void attachAtomicProposition(const std::string& name, const std::string& spec);

    struct node_sorter {
    public:
      node_sorter(bool nodes_first=true);

      bool operator()(const REO_Graph_Object::ptr a,
                      const REO_Graph_Object::ptr b) const;
    private:
      bool nodes_first;
    };

  private:
    // perform the actual join in the graph
    REO_Node::ptr do_join(REO_Node::ptr, REO_Channel_End::ptr);
    REO_Node::ptr do_join(REO_Node::ptr, REO_Node::ptr);

    vector<PortDescription::ptr> anonymous_port_descriptions;

    AtomicPropositionSpec::vector atomic_propositions;

    graph_ptr_t graph;
    DataDomain::ptr port_data_domain;
  };
  
  
  struct Vertex_Property_Writer {
    Vertex_Property_Writer(graph_t &_g) : g(_g) {}
    template <class Vertex>
    void operator() (std::ostream &out, Vertex u) {
      out << " [label=" << g[u].ptr->get_Description() << "]";
    }
    graph_t &g;
  };  

} //end namespace

#endif
