/***************************************************************************
 *   Copyright (C) 2008-2009 by Tobias Blechmann                           *
 *   blechman@tcs.inf.tu-dresden.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 @author Tobias Blechmann
 */

#include "reo_component.hpp"
#include "vereofy/common/errors.hpp"
#include <iostream>
#include <boost/lexical_cast.hpp>

#include <boost/foreach.hpp>
#include <boost/algorithm/string/replace.hpp>
#define foreach BOOST_FOREACH

using namespace std;

namespace vereofy {

  REO_Component::ptr REO_Component::getFor(ComponentInstance::ptr _my_component) {
    REO_Component::ptr result = REO_Component::ptr(new REO_Component(_my_component));

    ComponentInterface::ptr comp_interface = _my_component->getInterface();
    unsigned int in_count = comp_interface->getInPortCount();
    for (unsigned int i=0; i<in_count; ++i) {
      result->source_channel_ends.push_back(REO_Channel_End::ptr(new REO_Channel_End(result, comp_interface->getInPort(i)))); 
    }
    
    unsigned int out_count = comp_interface->getOutPortCount();
    for (unsigned int i=0; i<out_count; ++i) {
      result->sink_channel_ends.push_back(REO_Channel_End::ptr(new REO_Channel_End(result, comp_interface->getOutPort(i)))); 
    }
    return result;
  }

  REO_Graph_Object::ptr REO_Component::clone(REO_Graph_Object::ptr self, REO_Graph_Object::rgo_map& mapping) const {
    CHECK_SUCCESS(this==self.get());

    REO_Component::ptr result = REO_Component::ptr(new REO_Component(my_component));
    result->setClonedFrom(self);
    
    // insert result into mapping, needed for CE cloning
    CHECK_SUCCESS(mapping.insert(make_pair(self, result)).second == true);
    
    for (unsigned int i=0; i<source_channel_ends.size(); ++i) {
      REO_Channel_End::ptr ce=source_channel_ends.at(i);
      result->source_channel_ends.push_back(boost::static_pointer_cast<REO_Channel_End>(ce->clone(ce, mapping)));
    }

    for (unsigned int i=0; i<sink_channel_ends.size(); ++i) {
      REO_Channel_End::ptr ce=sink_channel_ends.at(i);
      result->sink_channel_ends.push_back(boost::static_pointer_cast<REO_Channel_End>(ce->clone(ce, mapping)));
    }

    return result;
  }

  ComponentInstance::ptr REO_Component::getComponentInstance() {
    return my_component;
  }
  
  REO_Component::REO_Component(ComponentInstance::ptr _my_component) : REO_Graph_Object(REO_COMPONENT), my_component(_my_component) {
  }
  
  std::string REO_Component::get_Description() {
    return "\"REO_Component "+getComponentInstance()->getTypeName()+" "+getComponentInstance()->getName()+"\"";
  }
  
  std::string REO_Component::get_Label() {
    string s=getComponentInstance()->getTypeName()+" "+getComponentInstance()->getName();
    boost::replace_all(s, "<", "&lt;");
    boost::replace_all(s, ">", "&gt;");
    return s;
  }

  unsigned int REO_Component::getSourceCECount() const {
    return source_channel_ends.size();
  }
  

  REO_Channel_End::ptr REO_Component::getSourceCE(unsigned int number) const {
    return source_channel_ends.at(number);
  }  
  

  unsigned int REO_Component::getSinkCECount() const {
    return sink_channel_ends.size();
  }
  

 REO_Channel_End::ptr REO_Component::getSinkCE(unsigned int number) const {
   return sink_channel_ends.at(number);
 }


  // return a name for the channel end (not efficient)
  string REO_Component::getNameForCE(REO_Channel_End::ptr ce) const {
    const vector<REO_Channel_End::ptr>& v=(ce->get_CE_Type()==REO_Channel_End::CE_SOURCE ? source_channel_ends : sink_channel_ends);

    for (unsigned int i=0;i<v.size();i++) {
      if (v.at(i)==ce) {
        return (ce->get_CE_Type()==REO_Channel_End::CE_SOURCE ? "source" : "sink")+boost::lexical_cast<string>(i);
      }
    }
    
    // Can't find the CE!
    throw "implementation error";
  }
  
}
