/***************************************************************************
 *   Copyright (C) 2009 by Joachim Klein                                   *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "restriction_manager.hpp"
#include "vereofy/common/errors.hpp"

#include "vereofy/system/symb_automaton/symb_component_instance.hpp"
#include <boost/tuple/tuple.hpp>

using namespace std;

namespace vereofy {
  RestrictionManager::RestrictionManager() :
    restrict_delta_x_with_initial(false),
    restrict_delta_x(false),
    restrict_delta_y(false) {
  }

  void RestrictionManager::addStateRestriction(const std::string& component_instance_name, 
                                               const std::string& restriction) {
    if (hasStateRestrictions(component_instance_name)) {
      state_restrictions[component_instance_name].push_back(restriction);
    } else {
      std::vector<std::string> v;
      v.push_back(restriction);

      state_restrictions.insert(make_pair(component_instance_name, v));
    }
  }
  
  bool RestrictionManager::hasStateRestrictions(const std::string& component_instance_name) const {
    return (state_restrictions.count(component_instance_name)>0);
  }

  std::pair<RestrictionManager::restriction_iterator_t,
            RestrictionManager::restriction_iterator_t> 
  RestrictionManager::getStateRestrictions(const std::string& component_instance_name) const {

    restriction_map_t::const_iterator it=state_restrictions.find(component_instance_name);
    CHECK_SUCCESS(it!=state_restrictions.end());

    return make_pair(it->second.begin(), it->second.end());
  }


  void RestrictionManager::print(std::ostream& out) const {
    for (restriction_map_t::const_iterator it=state_restrictions.begin();
         it!=state_restrictions.end();
         ++it) {
      for (RestrictionManager::restriction_iterator_t r_it=it->second.begin();
           r_it!=it->second.end();
           ++r_it) {
        out << it->first << " " << *r_it << "\n";
      }
    }
  }


  BDDFunction RestrictionManager::getStateRestrictions(SymbComponentInstance& sci, bool state_x) const {

    // TODO rework to work also when there are no subcomponent automata

    BDDFunction result=BDDFunction::ONE();

    if (!sci.info->component_instance) {
      // we don't have information about the component instance name -> ignore
      return result;
    }

    std::string name=sci.info->component_instance->getFQName();
    //    std::cerr << "Check for name = " << name << endl;

    if (hasStateRestrictions(name)) {
      restriction_iterator_t begin, end;
      boost::tie(begin, end) = getStateRestrictions(name);
      for (restriction_iterator_t it=begin;
           it!=end;
           ++it) {
        result &= sci.parse_state_guard(*it);
        //        std::cerr << "Adding restriction = " << *it << endl;
      }
    }

    if (!state_x) {
      result=result.moveDown();
    }

    for (subcomponent_info_map::const_iterator it=sci.info->subcomponent_infos_begin();
         it!=sci.info->subcomponent_infos_end();
         ++it) {
      result&=getStateRestrictions(*sci.info->getSubcomponentAutomaton(it->first), state_x);
    }

    return state_x ? result : result.moveDown();
  }
}



