/***************************************************************************
 *   Copyright (C) 2008-2009 by Johannes Osterholzer                       *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


/**
   @author Johannes Osterholzer
*/

#include "bddsubsetiterator.hpp"
#include "vereofy/common/functions.hpp"
#include <iostream>
#include <string>
#include <cmath>
#include <climits>

using namespace std;

namespace vereofy {
  unsigned int ulongBits = sizeof(unsigned long) * 8;

  //bddset - the set whose subsets are to be iterated
  //csVars - cubeset of the variables the set is built over.
  BDDSubsetIterator::BDDSubsetIterator(BDDFunction bddset, BDDFunction csVars)
  {
    end = false;

    int i = 0;

    while(bddset != 0.0) {
      long size = BDDVarOrder::size();
      char *buffer = new char[size];
      bddset.getPath(buffer);

      // Accumulate the assignments of the essential variables in this element
      BDDFunction element = 1.0;
      for(int j = 0; j < size; j++) {
	if(buffer[j] == '0')
	  element &= !BDDFunction::projection(j);
	else if(buffer[j] == '1')
	  element &= BDDFunction::projection(j);
	else if(buffer[j] == '-')
	  element &= ((csVars & !BDDFunction::projection(j)) == 0.0) ? !BDDFunction::projection(j) : 1.0;
	else
	  throw "Unexpected char in JINC-Path!";
      }

      elements.push_back(element);
      bddset &= !element;

      i++;
      delete[] buffer;
    }

    //cout << i << " elements.\n";
  
    // Initially, the bitmask is all zeroed out.
    bitmask.insert(bitmask.begin(), (int)ceil((double)i / ulongBits), 0); 
  }

  BDDSubsetIterator::~BDDSubsetIterator()
  {
    debugOut(7, "BDDSubsetIterator::~BDDSubsetIterator()", this);
    debugOut(7, string("Forbidden Subsets: ") + forbiddenSets.size(), this);
  }

  BDDFunction BDDSubsetIterator::operator*()
  {
    BDDFunction result = 0.0;
    if(end)	// we've already iterated all subsets
      return result;
    for(unsigned int i = 0; i < bitmask.size(); i++)
      for(unsigned int j = 0; j < ulongBits; j++) {
	if(i * ulongBits + j >= elements.size())
	  goto dropout;
	if(bitmask[i] & (1 << j))
	  result |= elements[i * ulongBits + j];
      }
  dropout:
    return result;
  }

  BDDSubsetIterator BDDSubsetIterator::operator++()
  {
    do {
      nextSet();
    } while(isForbidden() & !end);

    return *this;
  }

  void BDDSubsetIterator::forbidSet()
  {
    forbiddenSets.push_back(bitmask);
    //cout << forbiddenSets.size() << " sets are forbidden." << endl;
  }

  void BDDSubsetIterator::nextSet()
  {
    if(end == true)
      return;

    unsigned int i = 0;
    for(; i < bitmask.size(); i++) {
      if(bitmask[i] == ULONG_MAX) {
	bitmask[i] = 0;
      } else {
	bitmask[i]++;
	break;
      }
    }

    if(i == bitmask.size() || ((i == bitmask.size() - 1) && ((bitmask[i] & (ULONG_MAX << (elements.size() % ulongBits))) != 0)))
      end = true;
  }

  bool BDDSubsetIterator::isForbidden()
  {
    bool ret = false;
    // iterate over all forbidden sets
    for(unsigned int i = 0; i < forbiddenSets.size(); i++) {
      unsigned int j = 0;
      // if for all elements of forbiddenSets[i] a set bit implies the bit is also set in the bitmask, then the bitmask is forbidden
      for(j = 0; j < forbiddenSets[i].size(); j++) {
	if((bitmask[j] & (forbiddenSets[i])[j]) != (forbiddenSets[i])[j])
	  break;
      }

      if(j == forbiddenSets[i].size()) {
	ret = true;
	break;
      }
    }
  
    return ret;
  }

}
