

#include "serialized_varorder.hpp"
#include "vereofy/common/bdd_container_helpers.hpp"
#include <boost/algorithm/string.hpp>

using namespace vereofy;

unsigned int SerializedVarorder::counter=0;


SerializedVarorder::malformed_varorder_exception::malformed_varorder_exception(const string &message) :
  std::runtime_error("Malformed variable order: "+message) {
}

SerializedVarorder::varorder_mismatch_exception::varorder_mismatch_exception(const string &message) :
  std::runtime_error("Variable order mismatch with system: "+message) {
}


void SerializedVarorder::writeVarorder(std::ostream& out,
                                       SymbComponentInstance::ptr system) {
  VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

  out << "Vereofy Serialized Varorder v1\n";
  
  out << "Variables:\n";
  for (unsigned int i=0;i<BDDVarOrder::size();i++) {
    string name=bdd_var_name(i);
    char prefix=name.at(0);
    out << prefix;
  }
  out << "\n\n";

  writeComponents(out, system->info);
  writeStates(out, system->info);
  writePorts(out, system->info);
}


void SerializedVarorder::writeComponents(std::ostream& out, SymbComponentInfo::ptr info) {
  ComponentInstance::ptr  ci=info->component_instance;
  CHECK_SUCCESS(ci);

  out << "Component-Name:\n" << ci->getFQName() << "\n";
  out << "Type:\n" << ci->getTypeName() << "\n";

  for (subcomponent_info_map::const_iterator it=info->subcomponent_infos_begin();
       it!=info->subcomponent_infos_end();
       ++it) {
    writeComponents(out, it->second);
  }
}

void SerializedVarorder::writeStates(std::ostream& out, SymbComponentInfo::ptr info) {
  ComponentInstance::ptr  ci=info->component_instance;
  CHECK_SUCCESS(ci);

  for (state_to_bdd_map::const_iterator it=info->state_to_bdd.begin();
       it!=info->state_to_bdd.end();
       ++it) {
    StateID state_id(ci, it->first);
    out << "State-Name:\n" << state_id << "\n";
    out << "Type:\n" << it->first->getType()->toString() << "\n";
    out << "Variables:\n";
    SymbolicState::ptr ss=it->second;
    for (unsigned int i=0;i<ss->vars.X->size();i++) {
      if (i>0) out << " ";
      out << ss->vars.X->at(i).getFirstVariableIndex();
    }
    out << "\n";
  }

  for (subcomponent_info_map::const_iterator it=info->subcomponent_infos_begin();
       it!=info->subcomponent_infos_end();
       ++it) {
    writeStates(out, it->second);
  }
}


void SerializedVarorder::writePorts(std::ostream& out, SymbComponentInfo::ptr info) {
  ComponentInstance::ptr  ci=info->component_instance;
  CHECK_SUCCESS(ci);

  for (port_to_bdd_map::const_iterator it=info->port_to_bdd.begin();
       it!=info->port_to_bdd.end();
       ++it) {
    PortID port_id(ci, it->first);
    out << "Port-Name:\n" << port_id << "\n";
    out << "Type:\n" << it->first->getDataDomain()->toString() << "\n";
    out << "Variables:\n";
    Symb_Channel_End::ptr symb_ce=it->second;

    symb_ce->serialize_vars(out);
  }

  for (subcomponent_info_map::const_iterator it=info->subcomponent_infos_begin();
       it!=info->subcomponent_infos_end();
       ++it) {
    writePorts(out, it->second);
  }
}


void SerializedVarorder::readVarorder(std::istream& in,
                                      ComponentInstance::ptr system,
                                      BuildContext::ptr build_context,
                                      bool verify_only) {
  VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

  SerializedVarorder sv;

  std::string input;
  std::getline(in, input);
  if (input!="Vereofy Serialized Varorder v1") {
    throw malformed_varorder_exception("Missing or invalid identification string");
  }

  std::getline(in, input);
  if (input!="Variables:") {
    throw malformed_varorder_exception("Missing or invalid Variables specification");
  }
  std::getline(in, input);
  if (!verify_only) {
    sv.generateVariables(input);
  }
  
  std::getline(in, input);
  if (input!="") {
    throw malformed_varorder_exception("Missing empty line after Variables");
  }
  
  try {
    while (sv.readItem(in, system, build_context, verify_only)) {
      ;
    }
  } catch (ComponentInstance::lookup_failure_exception& e) {
    throw varorder_mismatch_exception(e.what());
  }
}
  
void SerializedVarorder::generateVariables(const std::string& v, bool prepend) {
  char last=' ';
  
  VarManager* var_manager=VarManager::getInstance();

  CHECK_SUCCESS(vars.size()==0);  // generateVariables can only be called once per SerializedVarorder object
  std::vector<BDDFunction> tmp_vars;

  std::string var_codes=v;
  
  if (prepend) {
    // build backwards because prepending inserts in reverse order
    unsigned int i=0;
    for (std::string::const_reverse_iterator it=v.rbegin();
         it!=v.rend();
         ++it) {
      var_codes[i++]=*it;
    }
  }

  for (std::string::iterator it=var_codes.begin();
       it!=var_codes.end();
       ++it) {
    char ch = *it;

    switch (ch) {
    case '*': {
      BDDFunctionContainer f=var_manager->createVars("*__"+boost::lexical_cast<std::string>(counter++), 1, prepend);
      tmp_vars.push_back(f.at(0));
      break;
    }
    case '#': {
      BDDFunctionContainer f=var_manager->createVars("#__"+boost::lexical_cast<std::string>(counter++), 1, prepend);
      tmp_vars.push_back(f.at(0));
      break;
    }
    case 'X': {
      if (prepend) CHECK_SUCCESS(last=='Y');
      StateVars s=var_manager->createStateVars(1, prepend);
      if (prepend) {
        tmp_vars.push_back(s.Z->at(0));
        tmp_vars.push_back(s.Y->at(0));
        tmp_vars.push_back(s.X->at(0));
      } else {
        tmp_vars.push_back(s.X->at(0));
        tmp_vars.push_back(s.Y->at(0));
        tmp_vars.push_back(s.Z->at(0));
      }
      break;
    }
    case 'Y':
      if (prepend) {
        CHECK_SUCCESS(last=='Z');
      } else {
        CHECK_SUCCESS(last=='X');
      }
      break;
    case 'Z':
      if (!prepend) CHECK_SUCCESS(last=='Y');
      break;
    default:
      throw malformed_varorder_exception(std::string("Unknown variables prefix ")+ch);
    }

    last=ch;

    TRACE("SerializedVarorder", "generateVariables",            \
          for (unsigned int i=0;i<BDDVarOrder::size();i++) {    \
            std::cout << bdd_var_name(i) << " ";                \
          }                                                     \
          std::cout << "\n";                                    \
          );
  }

  if (prepend) {
    // reverse variables for subsequent use...
    vars.insert(vars.end(),
                tmp_vars.rbegin(), tmp_vars.rend());
  } else {
    vars=tmp_vars;
  }
}


BDDFunctionContainer SerializedVarorder::getVars(const std::string& variables) {
  std::vector<std::string> v;
  boost::split(v, variables, boost::is_any_of(" "));

  BDDFunctionContainer result;
  if (variables=="") {
    return result; // no variables
  }

  for (unsigned int i=0;i<v.size();i++) {
    unsigned int index=boost::lexical_cast<unsigned int>(v.at(i));
    if (index>=vars.size()) {
      // out of range
      throw malformed_varorder_exception("Illegal variable index "+v.at(i));
    }
    result.push(vars.at(index));
  }
  return result;
}  

bool SerializedVarorder::readItem(std::istream& in,
                                  ComponentInstance::ptr system,
                                  BuildContext::ptr build_context,
                                  bool verify_only) {
  std::string input;
  std::getline(in, input);
  if (input=="") return false; 

  enum {ITEM_COMPONENT, ITEM_PORT, ITEM_STATE} item_t;
  
  if (input=="State-Name:") {
    item_t=ITEM_STATE;
  } else if (input=="Port-Name:") {
    item_t=ITEM_PORT;
  } else if (input=="Component-Name:") {
    item_t=ITEM_COMPONENT;
  } else {
    throw malformed_varorder_exception("Unknown line: "+input);
  }

  std::string name, item_type, variables;
  std::getline(in, name);
  
  std::getline(in, input);
  if (input!="Type:") {
    throw malformed_varorder_exception("Missing Type: for "+name);
  }
  std::getline(in, item_type);

  if (item_t!=ITEM_COMPONENT) {
    std::getline(in, input);
    if (input!="Variables:") {
      throw malformed_varorder_exception("Missing Variables: for "+name);
    }
    std::getline(in, variables);
  }
  
  switch (item_t) {
  case (ITEM_PORT): {
    // port
    // lookup ID
    PortID port_id=system->lookupPort(name);
    std::string portActiveVar, dataVars;
    boost::tie(portActiveVar, dataVars)=splitAt(variables, ' ');
    
    // verify type equality
    if (port_id.getPortDescription().getDataDomain()->toString() != item_type) {
      throw varorder_mismatch_exception("Mismatch of data types for "+
                                        boost::lexical_cast<std::string>(port_id)+
                                        "\n In system: "+port_id.getPortDescription().getDataDomain()->toString()+
                                        "\n In varorder: "+item_type);
    }

    if (verify_only) {
      return true;
    }

    Symb_Channel_End::ptr symb_ce;
    bool was_shared=false;
    if (vars_to_symb_ce.count(variables) > 0) {
      // there is already a channel end with these variables, use this
      symb_ce=vars_to_symb_ce.find(variables)->second;
    } else {    
      // split off portActive bit

      if (dataVars!="") {
        // try to find shared data symb_ce
        if (data_vars_to_symb_ce.count(dataVars) > 0) {
          BDDFunction portActive=getVars(portActiveVar).at(0);
          symb_ce=data_vars_to_symb_ce.find(dataVars)->second->getSharedDataChannelEnd(portActive);
          was_shared=true;
        }
      }

      if (!symb_ce) {
        Symb_DataDomain::ptr dd=Symb_DataDomain::getFor(port_id.getPortDescription().getDataDomain());
        
        symb_ce=Symb_Channel_End::ptr(new Symb_Channel_End(dd, getVars(variables)));
      }

      // insert the resulting symb_ce for these variables, set insert does not overwrite existing entry
      vars_to_symb_ce.insert(std::make_pair(variables, symb_ce));
      if (dataVars!="") {
        data_vars_to_symb_ce.insert(std::make_pair(dataVars, symb_ce));
      }
    }

    TRACE("SerializedVarorder", "port", \
          std::cout << "Deserialized Port: " << port_id << " = " << symb_ce << (was_shared ? " (shared)": "") << "\n";\
          symb_ce->serialize_vars(std::cout););
    build_context->port_id_to_deserialized_symb_ce->add(port_id, symb_ce);
    break;
  }
  case ITEM_STATE: {
    // state
    // lookup ID
    StateID state_id=system->lookupState(name);

    // verify data type
    if (state_id.getStateDescription().getType()->toString() != item_type) {
      throw varorder_mismatch_exception("Mismatch of data types for "+
                                        boost::lexical_cast<std::string>(state_id)+
                                        "\n In system: "+state_id.getStateDescription().getType()->toString()+
                                        "\n In varorder: "+item_type);
    }

    if (verify_only) {
      return true;
    }

    SymbolicState::ptr ss(new SymbolicState());
    ss->data_domain=Symb_DataDomain::getFor(state_id.getStateDescription().getType());

    ss->vars.X.reset(new BDDFunctionContainer(getVars(variables)));
    ss->vars.Y.reset(new BDDFunctionContainer(container_moveDown(*ss->vars.X)));
    ss->vars.Z.reset(new BDDFunctionContainer(container_moveDown(*ss->vars.Y)));
    
    CHECK_SUCCESS(ss->data_domain->getLogSize() == ss->vars.X->size());

    TRACE("SerializedVarorder", "statet", 
          std::cout << "Deserialized State: " << state_id << "\n");
    build_context->state_id_to_deserialized_symbolic_state->add(state_id, ss);
    break;
  }
  case (ITEM_COMPONENT): {
    ComponentInstance::ptr ci;
    if (name=="") {
      ci=system;
    } else {
      ci=system->lookupComponent(name);
    }

    if (ci->getTypeName() != item_type) {
      throw varorder_mismatch_exception("Mismatch of component types for "+
                                        (name=="" ? "system" : name)+
                                        "\n In system: "+ci->getTypeName()+
                                        "\n In varorder: "+item_type);
    }
  }
  }

  return true;
}


bool SerializedVarorder::findAndReadVarorder(ComponentInstance::ptr system,
                                             BuildContext::ptr build_context) {
  if (build_context->varorder_dir=="") {
    return false;
  }

  std::string filename=getFilename(system->getTypeName());

#ifdef _WIN32
  std::string complete_filename=build_context->varorder_dir+"\\"+filename;
#else
  std::string complete_filename=build_context->varorder_dir+"/"+filename;
#endif
  
  ifstream in(complete_filename.c_str());
  if (in.fail()) {
    return false;
  }

  // check if varorder in file matches this component
  try {
    readVarorder(in, system, build_context, 
                 true  // verify_only
                 );
  } catch (malformed_varorder_exception& e) {
    std::cout << "Trying to read varorder for " << system->getTypeName() << " from file " << complete_filename << ": Malformed varorder\n";
    return false;
  } catch (varorder_mismatch_exception& e) {
    std::cout << "Trying to read varorder for " << system->getTypeName() << " from file " << complete_filename << ": Does not match component\n";
    return false;
  }

  in.clear(); // remove the eof flag 
  in.seekg(0, ios::beg);   // go back to beginning of varorder file

  // actually read varorder
  std::cout << "Deserialize varorder for " << system->getTypeName() << " from file " << complete_filename << "\n";
  readVarorder(in, system, build_context, false);

  return true;
}

std::string SerializedVarorder::getFilename(const std::string& type_name) {
  std::string t=type_name;

  boost::algorithm::replace_all(t, "<", "-");
  boost::algorithm::replace_all(t, ">", "");
  
  return t+".varorder";
}


