/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include <sstream>
#include "symb_array_datadomain.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/common/xml_helpers.hpp"
#include "symb_expression.hpp"

namespace vereofy {

  Symb_ArrayDataDomain::Symb_ArrayDataDomain(ArrayDataDomain::ptr data_domain) :
    Symb_DataDomain(data_domain),
    symb_base(Symb_DataDomain::getFor(data_domain->getBaseType())),
    symb_index(boost::static_pointer_cast<Symb_FlatDataDomain>(Symb_DataDomain::getFor(data_domain->getIndexDomain()))) {
  }
    
  unsigned int Symb_ArrayDataDomain::getLogSize() const {
    return symb_base->getLogSize() * getArrayDataDomain()->getArraySize();
  }

  double Symb_ArrayDataDomain::countValidValues() const {
    double base_count=symb_base->countValidValues();
    double result=1;

    for (unsigned int i=0;i<getArrayDataDomain()->getArraySize();i++) {
      result*=base_count;
    }

    return result;
  }

  BDDFunction Symb_ArrayDataDomain::restrictToValidValues(const BDDFunctionContainer& encoding) const {
    BDDFunction result=BDDFunction::ONE();
    for (unsigned int i=0;i<getArrayDataDomain()->getArraySize();i++) {
      result&=getSymbBase()->restrictToValidValues(getEncoding(encoding, i));
    }
    return result;
  }

  ArrayDataDomain::ptr Symb_ArrayDataDomain::getArrayDataDomain() const {
    return boost::static_pointer_cast<ArrayDataDomain>(getDataDomain());
  }
  
  Symb_DataDomain::ptr Symb_ArrayDataDomain::getSymbBase() const {
    return symb_base;
  }

  Symb_FlatDataDomain::ptr Symb_ArrayDataDomain::getIndexDomain() const {
    return symb_index;
  }
  
  BDDFunctionContainer Symb_ArrayDataDomain::getEncoding(const BDDFunctionContainer& encoding, 
							 unsigned int index) const {
    CHECK(1, encoding.size()==getLogSize(), "");
    if (index>=getArrayDataDomain()->getArraySize()) {
      throw "index out of range";
    }
    
    unsigned int start_index=index * symb_base->getLogSize();
    BDDFunctionContainer result;
    for (unsigned int i=0;i<symb_base->getLogSize();i++) {
      CHECK(1, start_index+i<encoding.size(), "");
      result.push(encoding.at(start_index+i));
    }
    return result;
  }

  
  string Symb_ArrayDataDomain::reverseEncoding(const BDDFunction& assignment, 
					       const BDDFunctionContainer& vars) const {
    stringstream out;

    out << "{";
    for (unsigned int i=0;i<getArrayDataDomain()->getArraySize();i++) {
      if (i!=0) {
	out << ",";
      }

      BDDFunctionContainer i_encoding=getEncoding(vars, i);
      out << symb_base->reverseEncoding(assignment, i_encoding);
    }
    out << "}";
    return out.str();
  }

  void Symb_ArrayDataDomain::reverseEncodingToXML(std::ostream& out, 
                                                   const BDDFunction& assignment,
                                                   const BDDFunctionContainer& vars) const {
    xml_tag_open(out, "value");

    for (unsigned int i=0;i<getArrayDataDomain()->getArraySize();i++) {
      xml_tag_open(out, "element", xml_parameter("name", i));

      BDDFunctionContainer i_encoding=getEncoding(vars, i);
      symb_base->reverseEncodingToXML(out, assignment, i_encoding);

      xml_tag_close(out, "element");
    }

    xml_tag_close(out, "value");
  }

  void Symb_ArrayDataDomain::reverseEncodingToGuards(const BDDFunction& assignment, 
                                                     const BDDFunctionContainer& vars,
                                                     const std::string& prefix,
                                                     std::vector<std::string> &result) const {
    for (unsigned int i=0;i<getArrayDataDomain()->getArraySize();i++) {
      BDDFunctionContainer i_encoding=getEncoding(vars, i);
      std::string new_prefix=prefix+"["+boost::lexical_cast<string>(i)+"]";
      symb_base->reverseEncodingToGuards(assignment, i_encoding,
                                         new_prefix,
                                         result);
    }
  }

  BDDFunction Symb_ArrayDataDomain::getEquivalent(const BDDFunction& assignments, 
                                                  const BDDFunctionContainer& encoding) const {
    BDDFunction result=assignments;

    for (unsigned int i=0;i<getArrayDataDomain()->getArraySize();i++) {
      result = symb_base->getEquivalent(result, getEncoding(encoding, i));
    }
    
    return result;
  }

  boost::shared_ptr<SymbFunction_Constructor> Symb_ArrayDataDomain::getConstructorFunction() {
    THROW_UNIMPLEMENTED();
  }


  BDDFunction Symb_ArrayDataDomain::performComparison(comp_op_t compare_op,
                                                      const SymbExpression& left, 
                                                      const SymbExpression& right) const {

    switch (compare_op) {
    case COMPARE_EQUAL:
    case COMPARE_NOTEQUAL:
      return Symb_DataDomain::performComparison(compare_op, left, right);
    default:
      throw "Can only perform == or != comparison for whole arrays!";
    }
  }

  SymbExpression Symb_ArrayDataDomain::castExpression(const SymbExpression& expression) const {
    Symb_DataDomain::ptr sdd_other=expression.getDataDomain();
    CHECK_SUCCESS(sdd_other->getDataDomain()->isConvertibleTo(*getDataDomain()));
    
    if (getDataDomain()->isSameAs(*sdd_other->getDataDomain())) {
      // nothing to do
      return expression;
    }
    
    Symb_ArrayDataDomain::ptr sdd_other_array=boost::dynamic_pointer_cast<Symb_ArrayDataDomain>(sdd_other);
    CHECK_SUCCESS(sdd_other_array);
    CHECK_SUCCESS(getArrayDataDomain()->getArraySize() == sdd_other_array->getArrayDataDomain()->getArraySize());
    
    BDDFunctionContainer values;
    BDDFunction cast_conditions = BDDFunction::ONE();
    for (unsigned int i=0;i<getArrayDataDomain()->getArraySize();i++) {
      // cast each element
      BDDFunctionContainer i_values_from=sdd_other_array->getEncoding(expression.getValues(), i);
      SymbExpression i_from(sdd_other_array->getSymbBase(),
                            i_values_from);

      SymbExpression i_to = getSymbBase()->castExpression(i_from);
      for (unsigned int j=0;j<i_to.getValues().size();j++) {
        values.push(i_to.getValues().at(j));
      }
      cast_conditions&=i_to.getCondition();
      
    }
    CHECK_SUCCESS(values.size()==getLogSize());
    
    return SymbExpression(Symb_DataDomain::ptr(new Symb_ArrayDataDomain(*this)),
                          values,
                          expression.getCondition() & cast_conditions);
  }
}

