/***************************************************************************
 *   Copyright (C) 2008-2009 by Tobias Blechmann                           *
 *   blechman@tcs.inf.tu-dresden.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
@author Tobias Blechmann
*/

/*! \file symautomaton.cpp
  This file provides the SymbAutomaton class
*/

#include "vereofy/common/functions.hpp"
#include "vereofy/common/stopwatch.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/common/bdd_op_trace.hpp"
#include "vereofy/common/errors.hpp"

#include "symb_automaton.hpp"
#include "symb_state.hpp"
#include "symb_transition.hpp"
#include "symb_transition_label.hpp"
#include "symb_trace_explorer_automaton.hpp"
#include <stdexcept>

#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#include <boost/lexical_cast.hpp>
#include <boost/scoped_ptr.hpp>

#include <iostream>
#include <sstream>

using namespace std;

namespace vereofy {
/*!
  The object containing all information about a symbolic constraint automata
*/


  // static variable:
  SymbAutomaton::hide_type SymbAutomaton::default_hide_type=SymbAutomaton::HIDE_EPSILON;

  SymbAutomaton::SymbAutomaton(Symb_DataDomain::ptr dataDomain_) :
    dataDomain(dataDomain_)
  {
    debugOut(3, "SymbAutomaton::SymbAutomaton()", this);
    varManager = VarManager::getInstance();
    clear();
  }

  // clone from other symb automaton
  SymbAutomaton::SymbAutomaton(const SymbAutomaton& other, bool share_symb_ces) :
    dataDomain(other.dataDomain),
    delta(other.delta), 
    initialStates(other.initialStates), 
    csPorts(other.csPorts),
    csDCVariables(other.csDCVariables),
    csTransitionLabelVariables(other.csTransitionLabelVariables), 
    csOriginalX(other.csOriginalX),
    csCopyY(other.csCopyY),
    csCopyCopyZ(other.csCopyCopyZ),
    haltSystem(other.haltSystem),
    stopped_transition_label(other.stopped_transition_label) {
    
    if (share_symb_ces) {
      copy_symb_ces(other);
    } else {
      for (symb_ce_set::const_iterator it=other.symb_ces.begin();
           it!=other.symb_ces.end();
           ++it) {
        add_symb_ce_to_set((*it)->clone());
      }
    }

    copy_symb_tls(other);
  }
  
  SymbAutomaton::~SymbAutomaton() 
  {
  }

  
  void SymbAutomaton::clear() 
  {
    delta=BDDFunction::ZERO();
    initialStates=BDDFunction::ONE();
    csPorts=BDDFunction::ONE();
    csDCVariables=BDDFunction::ONE();
    csTransitionLabelVariables=BDDFunction::ONE();
    csOriginalX=BDDFunction::ONE();
    csCopyY=BDDFunction::ONE();
    csCopyCopyZ=BDDFunction::ONE();
    haltSystem=BDDFunction::ONE();

    symb_ces.clear();

    cachedReachableStates=BDDCachedFunction();
    cachedTerminalStates=BDDCachedFunction();

    stopped_transition_label=BDDFunction::ZERO();
  }


  void SymbAutomaton::product(const Automaton& _rhs) 
  {
    debugOut(3, "SymbAutomaton::join (product automaton of two automata) ", this);
    // TODO: verify that the data domain is the same as for rhs
    const SymbAutomaton& rhs = dynamic_cast<const SymbAutomaton&>(_rhs);

    TRACE_SYMB_AUTOMATON_PATHS("SymbAutomaton.product", "lhs_paths", *this);
    TRACE_SYMB_AUTOMATON_PATHS("SymbAutomaton.product", "rhs_paths", rhs);

    BDDFunction sync = this->delta & rhs.delta;
    TRACE_BDD_FUNCTION("SymbAutomaton.product", sync, "");

    BDDFunction only_left = (this->delta 
                             & rhs.getPortVars().change() 
                             & rhs.getTransitionLabelVars().change()
                             & rhs.haltSystem);
    TRACE_BDD_FUNCTION("SymbAutomaton.product", only_left, "");

    BDDFunction only_right = (rhs.delta 
                              & this->getPortVars().change() 
                              & this->getTransitionLabelVars().change()
                              & this->haltSystem);
    TRACE_BDD_FUNCTION("SymbAutomaton.product", only_right, "");

    BDDFunction async = only_left | only_right;
    TRACE_BDD_FUNCTION("SymbAutomaton.product", sync, "");

    BDDFunction new_delta = async | sync;
    BDD_OP_TRACE2("symb_automaton_product_delta", new_delta, this->delta, rhs.delta);

    delta = new_delta;
    initialStates = this->initialStates & rhs.initialStates;


    // combine csVars, etc...
    do_product_for_meta_information(rhs);

    TRACE_SYMB_AUTOMATON_PATHS("SymbAutomaton.product", "product_paths", *this);
  }

  void SymbAutomaton::sync_product(const SymbAutomaton& rhs) 
  {
    debugOut(3, "SymbAutomaton::sync_product (synchronous product automaton of two automata) ", this);
    // TODO: verify that the data domain is the same as for rhs

    TRACE_SYMB_AUTOMATON_PATHS("SymbAutomaton.sync_product", "lhs_paths", *this);
    TRACE_SYMB_AUTOMATON_PATHS("SymbAutomaton.sync_product", "rhs_paths", rhs);

    BDDFunction sync = this->delta & rhs.delta;
    TRACE_BDD_FUNCTION("SymbAutomaton.product", sync, "");

    BDDFunction new_delta = sync;
    BDD_OP_TRACE2("symb_automaton_product_delta", new_delta, this->delta, rhs.delta);

    delta = new_delta;
    initialStates = this->initialStates & rhs.initialStates;

    // combine csVars, etc...
    do_product_for_meta_information(rhs);

    TRACE_SYMB_AUTOMATON_PATHS("SymbAutomaton.product", "product_paths", *this);
  }


  Automaton::ptr SymbAutomaton::copyProduct(const Automaton& _rhs) const 
  {
    debugOut(3, "SymbAutomaton::copyProduct", this);
    // TODO: verify that the data domain is the same as for rhs
    const SymbAutomaton& rhs = dynamic_cast<const SymbAutomaton&>(_rhs);
    SymbAutomaton::ptr newAutomaton(new SymbAutomaton(rhs));
    newAutomaton->product(rhs);

    return newAutomaton;
  }

  void SymbAutomaton::do_product_for_meta_information(const SymbAutomaton& rhs) {
    csPorts &= rhs.csPorts;
    csDCVariables &= rhs.csDCVariables;
    csTransitionLabelVariables &= rhs.csTransitionLabelVariables;
    csOriginalX &= rhs.csOriginalX;
    csCopyY &= rhs.csCopyY;
    csCopyCopyZ &= rhs.csCopyCopyZ;
    haltSystem &= rhs.haltSystem;

    copy_symb_ces(rhs);
    copy_symb_tls(rhs);
  }  
 
  Automaton::ptr SymbAutomaton::copyUnite(const Automaton& _rhs) const
  {
    debugOut(3, "SymbAutomaton::copyUnite (disjoint union of two automata) ", this);
    // TODO: verify that the data domain is the same as for rhs
    const SymbAutomaton& rhs = dynamic_cast<const SymbAutomaton&>(_rhs);
    SymbAutomaton::ptr newAutomaton(new SymbAutomaton(dataDomain));

    StateVars new_state_var= newAutomaton->createStateVars(1);
    BDDFunction var_first = ! new_state_var.X->at(0);
    BDDFunction var_second = new_state_var.X->at(0);
    newAutomaton->delta = (var_first & var_first.moveDown() & delta & rhs.haltSystem) | 
                          (var_second & var_second.moveDown() & rhs.delta & haltSystem);
    
    newAutomaton->initialStates = (var_first & initialStates & rhs.csOriginalX.change()) 
                                | (var_second & rhs.initialStates & csOriginalX.change());
    
    newAutomaton->csPorts = csPorts & rhs.csPorts;
    newAutomaton->csDCVariables = csDCVariables & rhs.csDCVariables;
    newAutomaton->csTransitionLabelVariables = 
      csTransitionLabelVariables & rhs.csTransitionLabelVariables;
    newAutomaton->csOriginalX &=  csOriginalX & rhs.csOriginalX;
    newAutomaton->csCopyY &= csCopyY & rhs.csCopyY;
    newAutomaton->csCopyCopyZ &= csCopyCopyZ & rhs.csCopyCopyZ;
    newAutomaton->haltSystem &= haltSystem & rhs.haltSystem;

    newAutomaton->copy_symb_ces(*this);
    newAutomaton->copy_symb_ces(rhs);
    newAutomaton->copy_symb_tls(*this);
    newAutomaton->copy_symb_tls(rhs);
    return newAutomaton;
  }



  SymbAutomaton::hide_type SymbAutomaton::getDefaultHideType() {
    return default_hide_type;
  }

  void SymbAutomaton::setDefaultHideType(SymbAutomaton::hide_type t) {
    default_hide_type=t;
  }

  void SymbAutomaton::setDefaultHideType(const std::string& t) {
    if (t=="transitive") {
      default_hide_type=HIDE_TRANSITIVE;
    } else if (t=="transitive-full") {
      default_hide_type=HIDE_TRANSITIVE_FULL;
    } else if (t=="epsilon") {
      default_hide_type=HIDE_EPSILON;
    } else {
      throw t+" is not a valid hide variant";
    }
  }

  void SymbAutomaton::hide(const Automaton_Channel_End::ptr& ce) {
    hide(ce, default_hide_type);
  }

  void SymbAutomaton::hide(const std::vector<Automaton_Channel_End::ptr>& ces) {
    hide(ces, default_hide_type);
  }

  void SymbAutomaton::hide(const std::vector<Automaton_Channel_End::ptr>& ces, hide_type t) {
    if (ces.size()==0) {
      if (t == HIDE_TRANSITIVE_FULL) {
        // call hide transitive full to remove epsilon transitions
        hide_transitive_full();
      }
      return;
    }

    if (t==HIDE_TRANSITIVE) {
      for (unsigned int i=0;i<ces.size();i++) {
        // successively hide all the CEs
        hide(ces.at(i), t);
      }
      return;
    }

    BDDFunction port_active_vars=BDDFunction::ONE();
    BDDFunction data_vars=BDDFunction::ONE();

    symb_ce_set remaining_symb_ces_in_automaton=getSymbChannelEnds();
    
    for (unsigned int i=0;i<ces.size();i++) {
      Symb_Channel_End::ptr ce=boost::static_pointer_cast<Symb_Channel_End>(ces.at(i)); 
      port_active_vars&=ce->getPortActive();

      if (ce->areDataVariablesShared()) {
        if (ce->getSharedDataCEsFromSet(remaining_symb_ces_in_automaton).size()>0) {
          // there is still some symb ce with these data vars
          // don't remove variables, but this ce will be removed below
        } else {
          // no further ces sharing these data variables in this automaton
          // -> these data variables should be hidden
          data_vars&=container2cs(ce->getDCVars());
        }
      } else {
        data_vars&=container2cs(ce->getDCVars());
      }
      
      remaining_symb_ces_in_automaton.erase(ce);
    }

    switch (t) {
    case HIDE_EPSILON:
      hide_epsilon(port_active_vars, data_vars);
      break;
    case HIDE_TRANSITIVE_FULL:
      // transitive-full:
      // hide ports to epsilon transitions
      // then transitive hiding of the epsilon transitions
      hide_epsilon(port_active_vars, data_vars);
      hide_transitive_full();
      break;
    default:
      THROW_UNIMPLEMENTED();
    }

    for (unsigned int i=0;i<ces.size();i++) {
      Symb_Channel_End::ptr ce=boost::static_pointer_cast<Symb_Channel_End>(ces.at(i)); 
      remove_symb_ce_from_set(ce);
    }
  }

  void SymbAutomaton::hide(const Automaton_Channel_End::ptr& ce_, hide_type t) {
    Symb_Channel_End::ptr ce=boost::static_pointer_cast<Symb_Channel_End>(ce_);

    BDDFunction port_active_var=ce->getPortActive();
    BDDFunction data_vars=BDDFunction::ONE();
    if (ce->areDataVariablesShared() && ce->getSharedDataCEsFromSet(getSymbChannelEnds()).size()>0) {
      // there are other symb ces sharing these data variables, don't remove
    } else {
      data_vars=container2cs(ce->getDCVars());
    }

    switch (t) {
    case HIDE_TRANSITIVE:
      hide_transitive(port_active_var, data_vars);
      break;
    case HIDE_TRANSITIVE_FULL:
      // transitive-full:
      // hide ports to epsilon transitions
      // then transitive hiding of the epsilon transitions
      hide_epsilon(port_active_var, data_vars);
      hide_transitive_full();
      break;
    case HIDE_EPSILON:
      hide_epsilon(port_active_var, data_vars);
      break;
    default:
      THROW_UNIMPLEMENTED();
    }

    remove_symb_ce_from_set(ce);
  }


  // currently supports only one port_active_var
  void SymbAutomaton::hide_transitive(const BDDFunction& port_active_var, const BDDFunction& data_vars) {
    TRACE_SYMB_AUTOMATON_PATHS("SymbAutomaton.hide", "before_hide_paths", *this);
    TRACE_SYMB_AUTOMATON_DOT("SymbAutomaton.hide", "before_dot", *this);

    // transitive hide only works with unannotated transitions
    if (!cubeSetIsEmpty(getTransitionLabelVars())) {
        THROW_UNIMPLEMENTED();
    }

    BDDFunction portToHideActive = port_active_var;
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", portToHideActive, "");

    BDDFunction portToHideData = data_vars;
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", portToHideData, "");


    BDDFunction transitionsHidePortActive = delta & portToHideActive & csPorts.exists(portToHideActive).change();
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", transitionsHidePortActive, "");
    
    //deltaIndependentOfHidePort = all transitions with some other port than hidePort active
    //and all transitions with no port active (epsilon labelled transitions)
    BDDFunction deltaIndependentOfHidePort = delta & ( (!(csPorts.exists(portToHideActive).change())) | (csPorts.change()) ) ;

    TRACE_BDD_FUNCTION("SymbAutomaton.hide", deltaIndependentOfHidePort, "");

    // deltaIndependentOfHidePort still contains labels 
    deltaIndependentOfHidePort = deltaIndependentOfHidePort.exists(portToHideActive & portToHideData);
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", deltaIndependentOfHidePort, "");

    // transitionsHidePortActive only has unlabeled transitions, but for all transitions in this set
    // only the hidePort was active in the original system
    transitionsHidePortActive = transitionsHidePortActive.exists(getTransitionVars());
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", transitionsHidePortActive, "");
    
    // simple reachability analysis, starting in the set of starting states
    // and using only transitions for which only the hidingPort is active 
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", initialStates, "before hide steps");
    BDDFunction oldInitialStates;
    do {
      oldInitialStates = initialStates;
      BDDFunction newInitialStates = (initialStates & transitionsHidePortActive).exists(csOriginalX).moveUp();
            
      initialStates |= newInitialStates;
      TRACE_BDD_FUNCTION("SymbAutomaton.hide", initialStates, "after hide step");
    } while (initialStates != oldInitialStates);    
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", initialStates, "after hide steps");
    
    
    //deltaIndependentOfHidePort = labeled transitions from states represented with X variables to states rep. with Z variables
    BDDFunction workingSet=deltaIndependentOfHidePort.moveDown(csCopyY);
    deltaIndependentOfHidePort=workingSet;
    //repeat until empty
    while (workingSet != BDDFunction::ZERO()) {
      TRACE_BDD_FUNCTION("SymbAutomaton.hide", workingSet, "");

      //workingSet = all transitions X -> Y-> Z, X->Y is a transition for the hidingPort and Y->Z is a transition with silent hidingPort
      workingSet = (transitionsHidePortActive & workingSet.moveDown(csOriginalX));
      workingSet = workingSet.exists(csCopyY);
      TRACE_BDD_FUNCTION("SymbAutomaton.hide", workingSet, "after exists(csCopyY)");

      //remove transitions with hidePortActive from workingSet 
      workingSet &= !deltaIndependentOfHidePort;
      TRACE_BDD_FUNCTION("SymbAutomaton.hide", workingSet, "after &=!deltaIndependentOfHidePort");
      
      //add the new transitions to deltaIndependentOfHidePort (the name is bad chosen afterwards)
      deltaIndependentOfHidePort |= workingSet;
      TRACE_BDD_FUNCTION("SymbAutomaton.hide", deltaIndependentOfHidePort, "");
    }
    
    //delta is from X -> Y (variables)
    delta = deltaIndependentOfHidePort.moveUp(csCopyCopyZ);
    csPorts = csPorts.exists(portToHideActive);
    csDCVariables = csDCVariables.exists(portToHideData);

    TRACE_BDD_FUNCTION("SymbAutomaton.hide", delta, "");
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", csPorts, "");      
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", csDCVariables, "");

    TRACE_SYMB_AUTOMATON_PATHS("SymbAutomaton.hide", "after_hide_paths", *this);
    TRACE_SYMB_AUTOMATON_DOT("SymbAutomaton.hide", "after_hide_dot", *this);
  }


  void SymbAutomaton::hide_transitive_full() {
    TRACE_SYMB_AUTOMATON_PATHS("SymbAutomaton.hide", "before_hide_paths", *this);
    TRACE_SYMB_AUTOMATON_DOT("SymbAutomaton.hide", "before_dot", *this);

    // transitive hide only works with unannotated transitions
    CHECK_SUCCESS(getTransitionVars() == (getPortVars() & getDataConstraintVars()));

    BDDFunction epsilonTransitionsXaY = delta & csPorts.change();  // all transitions with no ports actice
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", epsilonTransitionsXaY, "");
    
    // normalTransitions: those where at least one port is active.
    BDDFunction normalTransitionsXaY = delta & !epsilonTransitionsXaY;
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", normalTransitionsXaY, "");

    // All (x,y) with x-epsilon->y
    BDDFunction epsilonReachXY = 
      epsilonTransitionsXaY.exists(getTransitionVars());

    // simple reachability analysis, starting in the set of starting states
    // and using only transitions for which only the hidingPort is active 
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", initialStates, "before hide steps");
    BDDFunction oldInitialStates;
    do {
      oldInitialStates = initialStates;
      BDDFunction newInitialStates = (initialStates & epsilonReachXY).exists(csOriginalX).moveUp();
            
      initialStates |= newInitialStates;
      TRACE_BDD_FUNCTION("SymbAutomaton.hide", initialStates, "after hide step");
    } while (initialStates != oldInitialStates);    
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", initialStates, "after hide steps");

    BDDFunction old_transitions;
    do {
      old_transitions = normalTransitionsXaY;
      
      BDDFunction new_transitions = epsilonReachXY & normalTransitionsXaY.moveDown(csOriginalX & csCopyY);  // XYaZ
      new_transitions=new_transitions.exists(csCopyY); // XaZ
      new_transitions=new_transitions.moveUp(csCopyCopyZ); // XaY

      normalTransitionsXaY|=new_transitions;
    } while (old_transitions!=normalTransitionsXaY);

    delta = normalTransitionsXaY;

    TRACE_BDD_FUNCTION("SymbAutomaton.hide", delta, "");
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", csPorts, "");      
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", csDCVariables, "");

    TRACE_SYMB_AUTOMATON_PATHS("SymbAutomaton.hide", "after_hide_paths", *this);
    TRACE_SYMB_AUTOMATON_DOT("SymbAutomaton.hide", "after_hide_dot", *this);
  }


  void SymbAutomaton::hide_epsilon(const BDDFunction& port_active_vars, const BDDFunction& data_vars) {
    TRACE_SYMB_AUTOMATON_PATHS("SymbAutomaton.hide", "before_hide_paths", *this);
    TRACE_SYMB_AUTOMATON_DOT("SymbAutomaton.hide", "before_dot", *this);

    BDDFunction portToHideActive = port_active_vars;
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", portToHideActive, "");

    BDDFunction portToHideData = data_vars;

    BDDFunction portToHideVars = portToHideActive & portToHideData;
    
    delta=delta.exists(portToHideVars);
    csPorts = csPorts.exists(portToHideActive);
    csDCVariables = csDCVariables.exists(portToHideData);

    TRACE_BDD_FUNCTION("SymbAutomaton.hide", delta, "");
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", csPorts, "");      
    TRACE_BDD_FUNCTION("SymbAutomaton.hide", csDCVariables, "");

    TRACE_SYMB_AUTOMATON_PATHS("SymbAutomaton.hide", "after_hide_paths", *this);
    TRACE_SYMB_AUTOMATON_DOT("SymbAutomaton.hide", "after_hide_dot", *this);
  }

  void SymbAutomaton::hide_semi(const std::vector<Automaton_Channel_End::ptr>& ces) {
    for (unsigned int i=0;i<ces.size();i++) {
      Symb_Channel_End& ce=static_cast<Symb_Channel_End&>(*ces.at(i)); 
      ce.semi_hide();
    }
  }



  // given a transition assignment, return all transitions that are equivalent 
  // (have different data assignment for the inactive ports)
  BDDFunction SymbAutomaton::getEquivalentTransitions(const BDDFunction& transition) const {
    CHECK(3, BDDChecks::isAssignment(transition, getTransitionVars()), "");
    CHECK(5, check_symb_ce_sanity(), "");
    
    if (getStoppedTransitionLabel()!=BDDFunction::ZERO() &&
        bdd_is_bit_set(transition, getStoppedTransitionLabel())) {
      return getStoppedTransitionLabel();
    }

    BDDFunction result=transition;
    for (symb_ce_set::const_iterator it=getSymbChannelEnds().begin();
         it!=getSymbChannelEnds().end();
         ++it) {
      const Symb_Channel_End& ce=**it;
      if (!bdd_is_bit_set(transition, ce.getPortActive())) {
        // the port for this CE is not active
        // -> remove assignment for the corresponding data constraints
        if (ce.areDataVariablesShared()) {
          // only remove the data variables iff all shared symb_channel_ends 
          // are not active
          if (ces_are_all_inactive(ce.getSharedDataCEsFromSet(getSymbChannelEnds()), transition)) {
            result=result.exists(container2cs(ce.getDCVars()));
          }
        } else {
          result=result.exists(container2cs(ce.getDCVars()));
        }
      } else {
        // get equivalent data values
        result=ce.getDataDomain()->getEquivalent(result, ce.getDCVars());
      }
    }
    
    return result;
  }

  // canonicalizeTransitions takes a transition_set and removes all but one for each set
  // of equivalent transitions. E.g., for a transition with an inactive port, the (irrelevant) 
  // data bits are all set to exactly one value (NULL)
  // The resulting transition set is returned
  BDDFunction SymbAutomaton::canonicalizeTransitions(const BDDFunction& transition_set) const {
    CHECK(5, check_symb_ce_sanity(), "");

    BDDFunction result=transition_set;
    
    // handle stopped transitions
    if (getStoppedTransitionLabel()!=BDDFunction::ZERO()) {
      // there may be stopped transitions
      BDDFunction all_ports_inactive=BDDFunction::ONE();
      for (symb_ce_set::const_iterator it=getSymbChannelEnds().begin();
           it!=getSymbChannelEnds().end();
           ++it) {
        const Symb_Channel_End& ce=**it;
        all_ports_inactive&=(!ce.getPortActive() & ce.data_equal(ce.getDataDomain()->getNULL()));
      }
      
      result&=
        (!getStoppedTransitionLabel()) |                               //    not stopped
        (getStoppedTransitionLabel() & all_ports_inactive);            // or all ports inactive
    }

    for (symb_ce_set::const_iterator it=getSymbChannelEnds().begin();
         it!=getSymbChannelEnds().end();
         ++it) {
      const Symb_Channel_End& ce=**it;

      if (ce.areDataVariablesShared()) {
        // if data variables are shared, the data variables should be set to NULL only
        // iff all shared ces are inactive
        result&=ce.getPortActive() | (!ce.getPortActive() &
                                      IMPLIES(ces_are_all_inactive(ce.getSharedDataCEsFromSet(getSymbChannelEnds())),
                                              ce.data_equal(ce.getDataDomain()->getNULL())));
      } else {
        result&=ce.getPortActive() | (!ce.getPortActive() & ce.data_equal(ce.getDataDomain()->getNULL()));
      }
    }
    
    return result;
  }


  SymbAutomaton::operator std::string() const 
  {
    using boost::lexical_cast;
    debugOut(3, "SymbAutomaton::operator string", this);
    string result; // = "**************\n*SymbAutomaton* \n**************\n";
    result += "delta essential variables: " + setOfStrToStr(delta.getEssentialVariables()) + "\n";
    result += "delta BDDNodes: " + lexical_cast<string>(delta.size()) + "\n";
    result += "number of starting states: " + lexical_cast<string>((initialStates & csAllOtherVariables(csOriginalX)).countAssignments()) + "\n";
    BDDFunction reachableStates = calcReachableStates();
    double nrOfReachable = (reachableStates & csAllOtherVariables(csOriginalX)).countAssignments();
    result += "number of reachable states: " + lexical_cast<string>(nrOfReachable) + "\n";
    if (nrOfReachable < 100) result += "\n" + toDotString() + "\n";
    else result += "more than 100 reachable states, refusing to print dot-representation\n";
    return result;  
  }

  string SymbAutomaton::stateToBin(const BDDFunction& state_assignment, bool use_x) const {  
    return stateToBin(state_assignment, (use_x ? csOriginalX : csCopyY));
  }
  
  string SymbAutomaton::stateToBin(const BDDFunction& state_assignment, 
				   const BDDFunction& csVars) const {
    return assignmentToBin(state_assignment, csVars);
  }


  void SymbAutomaton::State_Renderer::render(ostream& out, const BDDFunction& assignment, bool use_x) const {
    BDDFunction state_vars=(use_x ? automaton.csOriginalX : automaton.csCopyY);

    bool first=true;
    while (state_vars!=BDDFunction::ONE()) {
      BDDFunction state=popCubeSet(state_vars, state_vars);

      if (!first) {
	out << ",";
      } 
      first=false;
	
      out << bdd_var_name(state);
      if (bdd_is_bit_set(assignment, state)) {
	out << "+";
      } else {
	out << " ";
      }
    }
  }

  bool SymbAutomaton::State_Renderer::render_dot_style(ostream& out, const BDDFunction& assignment, bool is_initial, bool is_terminal) const {
    return false;
  }

  void SymbAutomaton::IO_Renderer::render(ostream& out, const BDDFunction& assignment, const BDDFunction& csVars) const {
    BDDFunction io_vars=csVars;

    bool first=true;
    while (io_vars!=BDDFunction::ONE()) {
      BDDFunction io=popCubeSet(io_vars, io_vars);

      if (!first) {
	out << " ";
      }
      first=false;
	
      out << bdd_var_name(io);
      if (bdd_is_bit_set(assignment, io)) {
	out << "+";
      } else {
	out << " ";
      }
    }
  }

  bool SymbAutomaton::IO_Renderer::render_dot_style(ostream& out, const BDDFunction& assignment, const BDDFunction& csVars,
                                                    const BDDFunction& state_pair) const {
    return false;
  }

  void SymbAutomaton::printPaths() const {
    cout << BDDVarOrder::getVarOrder() << "\n";

    cout << "Delta:\n";
    delta.printPaths();

    cout << "Reachable:\n";
    (delta & calcReachableStates()).printPaths();

    cout << "initialStates:\n";
    initialStates.printPaths();

    cout << "csPorts:\n";
    csPorts.printPaths();

    cout << "csDCVariables:\n";
    csDCVariables.printPaths();

    cout << "csTransitionLabelVariables:\n";
    csTransitionLabelVariables.printPaths();

    cout << "csOriginalX:\n";
    csOriginalX.printPaths();

    cout << "csCopyY:\n";
    csCopyY.printPaths();

    cout << "csCopyCopyZ:\n";
    csCopyCopyZ.printPaths();

    if (debugLevel>=10) {
      cout << "haltSystem:\n";
      haltSystem.printPaths();
    }
  }

  string SymbAutomaton::toDotString() const 
  {
    stringstream sstream;
    toDot(sstream);
    return sstream.str();
  }

  SymbAutomaton::State_Renderer::ptr SymbAutomaton::getDefaultStateRenderer() const {
    return State_Renderer::ptr(new State_Renderer(*this));
  }
  
  SymbAutomaton::IO_Renderer::ptr SymbAutomaton::getDefaultIORenderer() const {
    return IO_Renderer::ptr(new IO_Renderer(*this));
  }

  void SymbAutomaton::toDot(ostream& out) const {
    toDot(out, 
          *getDefaultStateRenderer(),
          *getDefaultIORenderer());
  }

  void SymbAutomaton::toDot(ostream& out, const State_Renderer& state_renderer, const IO_Renderer& io_renderer) const {
    VarManager::ReorderLock reorder_lock; // inhibit reordering for the rest of this scope
    
    out << "digraph G {\n";
    BDDFunction reachable = calcReachableStates();
    BDDFunction reachableTerminalStates = calcReachableTerminalStates();

    BDDFunction reachableCopy = reachable;
    while (reachableCopy != BDDFunction::ZERO()) {
      SymbState state=SymbState::popState(reachableCopy, csOriginalX);
      out << "  \"" << stateToBin(state.getX(), true) << "\" [";

      bool is_initial=bdd_sets_have_intersection(state.getX(), initialStates);
      bool is_terminal=bdd_sets_have_intersection(state.getX(), reachableTerminalStates);

      if (state_renderer.render_dot_style(out, state.getX(), is_initial, is_terminal)) {
        // ok, has rendered the style itself
      } else {
        out << "shape=box, color=";
        if (is_initial && is_terminal) {
          out << "violet";
        } else if (is_initial) {
          out << "blue";
        } else if (is_terminal) {
          out << "red";
        } else {
          out << "black";
        }
      }

      // add label
      out << ", label=\"";
      state_renderer.render(out, state.getX(), true);
      out << "\"]\n";
    }
    
    out << endl;
    
    if (debugLevel>=10) {
      cout << "Reachable:\n";
      (delta & reachable).printPaths();
    }

    BDDFunction deltaCopy = (delta & reachable).exists(getTransitionVars());
    while (deltaCopy != BDDFunction::ZERO()) {
      BDDFunction state_pair = popAssignment(deltaCopy, csOriginalX & csCopyY);
      TRACE_BDD_FUNCTION("SymbAutomaton.toDot", state_pair, "");
      string from=stateToBin(state_pair, true);
      string to=stateToBin(state_pair, false);
      if (debugLevel>=8) {
	cerr << "From "<< from << " to " << to << endl;
      }
      BDDFunction transitions = delta & state_pair;
      while (transitions != BDDFunction::ZERO()) {
	out << "  \"" << from << "\" -> \"" << to << "\" [";

	BDDFunction transition_instance = popAssignment(transitions, getTransitionVars());
	TRACE_BDD_FUNCTION("SymbAutomaton.toDot", transition_instance, "");

        // also remove all equivalent transitions (with different data values for inactive ports)
        BDDFunction equivalent_transitions=getEquivalentTransitions(transition_instance);
	TRACE_BDD_FUNCTION("SymbAutomaton.toDot", equivalent_transitions, "");
        transitions&=!equivalent_transitions;

        if (io_renderer.render_dot_style(out, transition_instance, getTransitionVars(),
                                         state_pair)) {
          out << ",";
        }
        
        out << "label=\"";
	io_renderer.render(out, transition_instance, getTransitionVars());
	out << "\"];\n";
      }
    }

    if (debugLevel>=4) {
      out << " ports [shape=box, label=\"csPorts=" << cubeSetToString(csPorts) << "\"];\n";
      out << " dc_vars [shape=box, label=\"csDCVariables=" << cubeSetToString(csDCVariables) << "\"];\n";
      out << " labels [shape=box, label=\"csTransitionLabelVariables=" << cubeSetToString(csTransitionLabelVariables) << "\"];\n";
      out << " original_x [shape=box, label=\"csOriginalX=" << cubeSetToString(csOriginalX) << "\"];\n";
      out << " copy_y [shape=box, label=\"csCopyY=" << cubeSetToString(csCopyY) << "\"];\n";
      out << " copy_z [shape=box, label=\"csCopyCopyZ=" << cubeSetToString(csCopyCopyZ) << "\"];\n";
    }
    out << "}\n";
  }


  void SymbAutomaton::extractExplicitAutomaton(ExtractExplicitAutomatonFunctor& functor) const {
    // we don't need to inhibit reordering, as we are mapping BDDFunctions
    //  to IDs without relying on a stable varorder
    
    BDDFunction reachable = calcReachableStates();
    BDDFunction reachableTerminalStates = calcReachableTerminalStates();

    functor.start();

    std::map<BDDFunction, unsigned long, BDDFunction::less> state_to_id;
    unsigned long next_state_id = 0;

    BDDFunction reachableCopy = reachable;
    while (reachableCopy != BDDFunction::ZERO()) {
      SymbState state=SymbState::popState(reachableCopy, csOriginalX);

      bool is_initial=bdd_sets_have_intersection(state.getX(), initialStates);
      bool is_terminal=bdd_sets_have_intersection(state.getX(), reachableTerminalStates);

      if (next_state_id >= ULONG_MAX) {
        throw "Too many states to make explicit!";
      }

      unsigned long id = next_state_id++;
      state_to_id[state.getX()]=id;
      
      functor.explicit_state(id, state.getX(), is_initial, is_terminal);
    }

    BDDFunction from_to_states = (delta & reachable).exists(getTransitionVars());
    BDDFunction from_states = from_to_states.exists(csCopyY);

    while (from_states != BDDFunction::ZERO()) {
      BDDFunction state_from = popAssignment(from_states, csOriginalX);
      unsigned long from = state_to_id[state_from];

      BDDFunction to_states = 
        (from_to_states & state_from).exists(csOriginalX);

      while (to_states != BDDFunction::ZERO()) {
        BDDFunction state_to = popAssignment(to_states, csCopyY);
        unsigned long to = state_to_id[state_to.moveUp()];

        BDDFunction transitions = delta & (state_from & state_to);
        while (transitions != BDDFunction::ZERO()) {
          BDDFunction transition_instance = popAssignment(transitions, getTransitionVars());
          
          // also remove all equivalent transitions (with different data values for inactive ports)
          BDDFunction equivalent_transitions=getEquivalentTransitions(transition_instance);
          transitions&=!equivalent_transitions;
          
          functor.explicit_transition(from, to, transition_instance);
        }
      }
    }

    functor.finished();
  }


  void SymbAutomaton::print(std::ostream& out, const SymbTrace& trace) const {
    print(out, trace, State_Renderer(*this), IO_Renderer(*this));
  }
  
  void SymbAutomaton::print(std::ostream& out, 
                            const SymbTrace& trace, 
                            const SymbAutomaton::State_Renderer& state_renderer,
                            const SymbAutomaton::IO_Renderer& io_renderer) const {
    
    for (unsigned int i=0;i<trace.countStates();i++) {
      state_renderer.render(out, trace.getState(i).getX(), true);
      out << "\n";
      
      if (i<trace.countLabels()) {
        out << "  ";
        io_renderer.render(out, trace.getLabel(i).getLabel(), trace.getLabel(i).getCsLabelVars());
        out << "\n";
      }
    }
  }

  void SymbAutomaton::print(std::ostream& out, 
                            const SymbState& state, 
                            const SymbAutomaton::State_Renderer& state_renderer) const {
    state_renderer.render(out, state.getX(), true);
  }
  
  void SymbAutomaton::print(std::ostream& out, 
                            const SymbTransitionLabel& label, 
                            const SymbAutomaton::IO_Renderer& io_renderer) const {
    io_renderer.render(out, label.getLabel(), label.getCsLabelVars());
  }
  
  double SymbAutomaton::getNumberOfReachableStates() const 
  {
    BDDFunction reachableStates = calcReachableStates();
    double nrOfReachable = (reachableStates & csAllOtherVariables(csOriginalX)).countAssignments();
    return nrOfReachable;
  }

  void SymbAutomaton::makeDeterministic()
  {
    BDDFunction n = BDDFunction::ZERO(); // necessary for the function call to pass the type system.
    makeDeterministic(n);
  }

  SymbAutomaton::ptr SymbAutomaton::makeDeterministicCopy() const
  {
    BDDFunction n = BDDFunction::ZERO();
    return makeDeterministicCopy(n);
  }

/**
 * This function converts a (nondeterministic) constraint automaton to an equivalent deterministic one.
 */
  void SymbAutomaton::makeDeterministic(BDDFunction& accepting)
  {
    *this = *(makeDeterministicCopy(accepting));
  }

/**
 * This function converts a (nondeterministic) constraint automaton to an equivalent deterministic one, doesn't touch the original.
 * accepting is a pointer to a set of accepting states. This is used for the derivative class ea.
 * @return Returns a deterministic constraint automaton equivalent to the current
 */
  SymbAutomaton::ptr SymbAutomaton::makeDeterministicCopy(BDDFunction& accepting) const
  {
    // determinisation currently only works with unannotated transitions
    CHECK_SUCCESS(getTransitionVars() == (getPortVars() & getDataConstraintVars()));


    BDDFunction containedStates = 0.0;	// associates to each state of the deterministic automaton the set of states in the original automaton 
       					// which it "contains". Disjunction of the conjunction of the corresponding states and statesets.

    unsigned long long nrStates; 				// counts the number of states in the generated deterministic automaton
    vector<BDDFunction> stateVars;				// holds the variables used for encoding these states
    BDDFunction workset = 0.0;				// contains states yet to expand

    BDDFunction oldAcceptingStates = 0.0;
    BDDFunction newAcceptingStates = 0.0;

    if(accepting != 0.0)
      oldAcceptingStates = accepting;

    // The generated automaton has the same ports as the original.
    SymbAutomaton::ptr result(new SymbAutomaton(dataDomain));
    result->csPorts = csPorts;
    result->csDCVariables = csDCVariables;
    result->copy_symb_ces(*this);

    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    // The initial state is the first to expand.
    StateVars newvar = result->createStateVars(1);
    BDDFunction x0 = container2cs(*(newvar.X));
    result->initialStates = !x0;

    // Is one of the initial states accepting?
    if((initialStates & oldAcceptingStates) != 0.0)
      newAcceptingStates |= result->initialStates;

    containedStates |= result->initialStates & initialStates;
    nrStates = 1;
    stateVars.push_back(x0);
    workset = !x0;

    while(workset != 0.0) {
      debugOut(7, "\n\n===NEW LOOP=======\n\n", this);

      // Select a not yet expanded state
      BDDFunction currentState = popAssignment(workset, result->csOriginalX);

      // The corresponding stateset:
      BDDFunction currentStateSet = containedStates.cofactor(currentState);
    
      if(debugLevel >= 7)  // debugBDD doesn't seem to regard the debug level
	debugBDD(currentStateSet, "current state set");

      // Find all possible port activations for the current stateset
      BDDFunction portDisj = (delta & currentStateSet).exists(csDCVariables & csOriginalX & csCopyY); // disjunction of port labelings
    
      if(debugLevel >= 7)
	debugBDD(portDisj, "possible port activations");

      // For each of the possible port activations, determine successor state sets
      while(portDisj != 0.0) {
	BDDFunction activation = popAssignment(portDisj, csPorts);

	if(debugLevel >= 7)
	  debugBDD(activation, "current activation");

	// Determine the set of (activation-)successors.
	BDDFunction successorSet = (delta & currentStateSet & activation).exists(csOriginalX & csDCVariables & csPorts).moveUp(); // lol

	if(debugLevel >= 7)
	  debugBDD(successorSet, "corresponding successor set");

	// No successor states, we're done here.
	if(successorSet == 0.0)
	  continue;

	// Now iterate through all the subsets of this successor set
	BDDSubsetIterator it(successorSet, csOriginalX);
	while(!it.iterated()) {
	  BDDFunction subset = *it;
	  ++it;

	  if(subset == 0.0)
	    continue;

	  if(debugLevel >= 7)
	    debugBDD(subset, "current subset of successor set");

	  // 1. compute g_det = g_1 & !g_2
	  // 2. ???
	  // 3. profit!!!

	  // We have to watch every transition for itself
	  BDDFunction subsetCopy = subset;
	  BDDFunction complementCopy = successorSet & !subset;
	  BDDFunction g_1 = 1.0;
	  BDDFunction g_2 = 0.0;

	  while(subsetCopy != 0.0) {
	    BDDFunction successorState = popAssignment(subsetCopy, csOriginalX);

	    BDDFunction g = (delta 
			     & currentStateSet 
			     & successorState.moveDown() 
			     & activation).exists(csOriginalX & csCopyY & csPorts);
	    if(g != 0.0)
	      g_1 &= g;
	  }

	  while(complementCopy != 0.0) {
	    BDDFunction successorState = popAssignment(complementCopy, csOriginalX);

	    BDDFunction g = (delta
			     & currentStateSet
			     & successorState.moveDown()
			     & activation).exists(csOriginalX & csCopyY & csPorts);
	    if(g != 0.0)
	      g_2 |= g;
	  }	  

	  if(g_1 == 0.0) {
	    // we can skip all supersets of the current set.
	    it.forbidSet();
	    continue;
	  }

	  BDDFunction gdet = g_1 & !g_2;
	
	  if(gdet == 0.0)
	    continue;

	  if(debugLevel >= 7)
	    debugBDD(gdet, "g_det");

	  // OK, now we've found a transition in the deterministic CA.
	  // We need to make sure that there's an associated state for this stateset
	  BDDFunction associatedState = containedStates.equivalenceClasses(subset, csOriginalX);

	  if(associatedState == 0.0) {
	    // Insert new state... this means perhaps we need to add a new state variable. 
	    if((1ULL << stateVars.size()) == nrStates) {
	      // Enrich everything with the new variable
	      newvar = result->createStateVars(1);
	      BDDFunction var = container2cs(*(newvar.X));
	      result->delta &= !var & !(var.moveDown());
	      result->initialStates &= !var;
	    
	      containedStates &= !var;

	      workset &= !var;

	      newAcceptingStates &= !var;

	      currentState &= !var;
	      stateVars.push_back(var);
	    }
	  
	    // All's clear for adding the state
	    nrStates++;
	    containedStates |= stateNumberToState(nrStates - 1, stateVars) & subset;

	    if((subset & oldAcceptingStates) != 0.0)
	      newAcceptingStates |= stateNumberToState(nrStates - 1, stateVars);
	  
	    debugOut(7, string("adding new state number ") + boost::lexical_cast<string>(nrStates), this);

	    // This state hasn't been expanded yet!
	    workset |= stateNumberToState(nrStates - 1, stateVars);
	  }

	  // Finally, add the transition!
	  associatedState = containedStates.equivalenceClasses(subset, csOriginalX);
	  result->delta |= currentState & associatedState.moveDown() & activation & gdet;
	}
      }
    }
    if(accepting != 0.0)
      accepting = newAcceptingStates;
    return result;
  }



  BDDFunction SymbAutomaton::getDelta() const 
  {
    debugOut(42, "SymbAutomaton::getDelta", this);
    return delta;
  }


  void SymbAutomaton::setInitialStates(const BDDFunction& initial_states) {
    debugOut(3, "SymbAutomaton::setInitialStates", this);
    // TODO: sanity checks?
    initialStates=initial_states;
  }
  
  void SymbAutomaton::setDelta(const BDDFunction& new_delta) {
    debugOut(3, "SymbAutomaton::setDelta", this);
    // TODO: sanity checks?
    delta=new_delta;
  }


  BDDFunction SymbAutomaton::getInitialStates() const
  {
    debugOut(42, "SymbAutomaton::getInitialStates", this);
    return initialStates;
  }

  BDDFunction SymbAutomaton::getTransitionVars() const
  {
    debugOut(42, "SymbAutomaton::getTransitionVars", this);
    return csPorts & csDCVariables & csTransitionLabelVariables;
  }

  BDDFunction SymbAutomaton::getPortVars() const
  {
    debugOut(42, "SymbAutomaton::getPortVars", this);
    return csPorts;
  }

  BDDFunction SymbAutomaton::getDataConstraintVars() const
  {
    debugOut(42, "SymbAutomaton::getDataConstraintVars", this);
    return csDCVariables;
  }

  BDDFunction SymbAutomaton::getTransitionLabelVars() const
  {
    debugOut(42, "SymbAutomaton::getTransitionLabelVars", this);
    return csTransitionLabelVariables;
  }

  BDDFunction SymbAutomaton::getOriginalStateVars() const
  {
    debugOut(42, "SymbAutomaton::getOriginalX", this);
    return csOriginalX;
  }


  BDDFunction SymbAutomaton::getCopyStateVars() const
  {
    debugOut(42, "SymbAutomaton::getCopyStateVars", this);
    return csCopyY;
  }

  BDDFunction SymbAutomaton::getCopyCopyStateVars() const
  {
    debugOut(42, "SymbAutomaton::getCopyStateVars", this);
    return csCopyCopyZ;
  }


  BDDFunction SymbAutomaton::getHaltSystem() const
  {
    debugOut(42, "SymbAutomaton::getHaltSystem");
    return haltSystem;
  }

  const SymbAutomaton::symb_ce_set& SymbAutomaton::getSymbChannelEnds() const {
    return symb_ces;
  }

  BDDFunction SymbAutomaton::calcReachableStates() const
  {
    BDDFunctionContainer dependencies;
    dependencies.push(delta);
    dependencies.push(initialStates);
    dependencies.push(getTransitionVars() & csOriginalX);

    if (cachedReachableStates.isValid() &&
        !cachedReachableStates.haveDependenciesChanged(dependencies)) {
      // valid result was cached
      return cachedReachableStates.getValue();
    }

    debugOut(2, "SymbAutomaton::calcReachableStates", this);
    TRACE("SymbAutomaton", "calcReachableStates", cout << "SymbAutomaton::calcReachableStates started" << endl);
    StopWatch stopwatch;
    stopwatch.start();
    BDDFunction sysCopy = delta.exists(getTransitionVars());
    BDDFunction oldReachable=BDDFunction::ZERO();
    BDDFunction reachable=initialStates;  
    int steps = 0;
    debugOut(3, "SymbAutomaton::calcReachableStates starting analysis");
    while (reachable!=oldReachable) {
      oldReachable=reachable;
      ++steps;
      reachable |= ((sysCopy & reachable).exists(csOriginalX)).moveUp();
      TRACE("SymbAutomaton", "calcReachableStates", cout << (steps>1?", ":"") << steps << flush);
    }
    stopwatch.stop();
    TRACE("SymbAutomaton", "calcReachableStates", cout << endl << "SymbAutomaton::calcReachableStates finished" << endl);
    TRACE("SymbAutomaton", "calcReachableStates", cout << "number of reachable states: " << (reachable & csAllOtherVariables(csOriginalX)).countAssignments() << endl);
    debugOut(2, "Reachable analysis: " + boost::lexical_cast<string>(steps) +" steps, "+ stopwatch.toString());  

    // cache result
    const_cast<BDDCachedFunction&>(cachedReachableStates)=BDDCachedFunction(reachable, dependencies);
    return reachable;
  }

  BDDFunction SymbAutomaton::calcTerminalStates() const {
    BDDFunctionContainer dependencies;
    dependencies.push(delta);
    dependencies.push(getTransitionVars() & csOriginalX);

    if (cachedTerminalStates.isValid() &&
        !cachedTerminalStates.haveDependenciesChanged(dependencies)) {
      // valid result was cached
      return cachedTerminalStates.getValue();
    }

    BDDFunction non_terminal_states=delta.exists( getTransitionVars() & csCopyY);
    BDDFunction terminal_states=!non_terminal_states;
    TRACE_BDD_FUNCTION("SymbAutomaton.calcTerminalStates", terminal_states, "");

    BDD_CHECK2(3, BDDChecks::essentialVarsInCubeSet(terminal_states, csOriginalX), "", terminal_states, csOriginalX);

    // cache result
    const_cast<BDDCachedFunction&>(cachedTerminalStates)=BDDCachedFunction(terminal_states, dependencies);
    return terminal_states;
  }

  BDDFunction SymbAutomaton::calcReachableTerminalStates() const {
    return calcReachableStates() & calcTerminalStates();
  }

  
  void SymbAutomaton::addChannelEndToAutomaton(Symb_Channel_End::ptr _symb_ce) {
    csPorts &= _symb_ce->getPortActive();
    csDCVariables &= container2cs(_symb_ce->getDCVars());
    add_symb_ce_to_set(_symb_ce);
  }

  void SymbAutomaton::addTransitionLabelToAutomaton(Symb_Transition_Label_Var::ptr _symb_tl) {
    csTransitionLabelVariables &= _symb_tl->getVars();
    add_symb_tl_to_set(_symb_tl);
  }
  
  bool SymbAutomaton::hasTransitionLabel(Symb_Transition_Label_Var::ptr _symb_tl) const {
    return symb_tls.count(_symb_tl)>0;
  }


  BDDFunction SymbAutomaton::csAllOtherVariables(const BDDFunction& csVariables) const 
  {
    unsigned long size = BDDVarOrder::size();
    boost::scoped_array<char> bufCsVariables(new char[size]);                                    
    csVariables.getPath(bufCsVariables.get());
    boost::scoped_array<char> bufResult(new char[size]);
    for (unsigned long i=0; i<size;i++) {
      bufResult[i] = (bufCsVariables[i]=='1')?'-':'1';
    }
    BDDFunction result=BDDFunction(bufResult.get());
    return result;
  }

  void SymbAutomaton::registerStateVars(const BDDFunctionContainer& x_vars) {
    for (unsigned int i=0; i < x_vars.size(); ++i) {
      BDDFunction X = x_vars.at(i); 
      csOriginalX &= X;
      BDDFunction Y = X.moveDown();
      csCopyY &= Y;
      haltSystem &= EQUIV(X, Y);      
      BDDFunction Z = Y.moveDown();
      csCopyCopyZ &= Z;
    }
  }
  
  StateVars SymbAutomaton::createStateVars(unsigned int number)
  {
    StateVars result = varManager->createStateVars(number);

    registerStateVars(*result.X);
    return result;
  }



  // create nr_of_states new state vars, mixed with var_positions_in_out
  // if var_positions_in_out is too short, append new positions
  StateVars SymbAutomaton::createStateVars(unsigned int number, 
					   BDDFunctionContainer& var_positions_in_out) {
    StateVars result = varManager->createStateVars(number, var_positions_in_out);

    registerStateVars(*result.X);
    return result;
  }


  StateVars SymbAutomaton::findOrAddGlobalState(const string& name, unsigned int bits) {
    StateVars result = varManager->findOrAddGlobalState(name, bits);

    registerStateVars(*result.X);
    return result;
  }

  Symb_DataDomain::ptr SymbAutomaton::getDataDomain() const {
    return dataDomain;
  }

  void SymbAutomaton::add_symb_ce_to_set(const Symb_Channel_End::ptr& symb_ce) {
    symb_ces.insert(symb_ce);
  }

  void SymbAutomaton::add_symb_tl_to_set(const Symb_Transition_Label_Var::ptr& symb_tl) {
    symb_tls.insert(symb_tl);
  }

  /**
   * Removes all Symb_Channel_Ends from the list that are for the same port
   */
  void SymbAutomaton::remove_symb_ce_from_set(const Symb_Channel_End::ptr& symb_ce) {
    symb_ces.erase(symb_ce);
  }
   
  void SymbAutomaton::copy_symb_ces(const SymbAutomaton& other) {
    symb_ces.insert(other.symb_ces.begin(),
                    other.symb_ces.end());
  }

  void SymbAutomaton::copy_symb_tls(const SymbAutomaton& other) {
    symb_tls.insert(other.symb_tls.begin(),
                    other.symb_tls.end());
  }


  bool SymbAutomaton::check_symb_ce_sanity() const {
    BDDFunction port_vars=BDDFunction::ONE();
    BDDFunction dc_vars=BDDFunction::ONE();
    
    for (symb_ce_set::const_iterator it=symb_ces.begin();
         it!=symb_ces.end();
         ++it) {
      port_vars&=(*it)->getPortActive();
      dc_vars&=container2cs((*it)->getDCVars());
    }

    if (getStoppedTransitionLabel()!=BDDFunction::ZERO()) {
      port_vars&=getStoppedTransitionLabel();
    }

    if (!(port_vars==getPortVars())) {
      cout << "\nPort active mismatch\n";
      port_vars.printEssentialVariables();
      port_vars.printPaths();

      getPortVars().printEssentialVariables();
      getPortVars().printPaths();
      return false;
    }
    if (!(dc_vars==getDataConstraintVars()))
      return false;
    return true;
  }


  bool SymbAutomaton::ces_are_all_inactive(const Symb_Channel_End::ptr_set& ces, const BDDFunction& assignment) const {
    for (Symb_Channel_End::ptr_set::const_iterator it=ces.begin();
         it!=ces.end();
         ++it) {
      const Symb_Channel_End& ce=**it;
      if (bdd_is_bit_set(assignment, ce.getPortActive())) {
        return false;
      }
    }

    return true;
  }

  BDDFunction SymbAutomaton::ces_are_all_inactive(const Symb_Channel_End::ptr_set& ces) const {
    BDDFunction result=BDDFunction::ONE();
    for (Symb_Channel_End::ptr_set::const_iterator it=ces.begin();
         it!=ces.end();
         ++it) {
      const Symb_Channel_End& ce=**it;
      result&=!ce.getPortActive();
    }
    return result;
  }



/**
 * Helper function to extract elements from a BDDFunction representing a set.
 * Discards non-essential variables.
 * @param fun is the BDDFunction representing the set.
 * @param csVars is supposed to be a conjunction of the variables considered to be essential for the set.
 */
  BDDFunction SymbAutomaton::getElem(BDDFunction fun, BDDFunction csVars) const
  {
    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    BDDFunction nonEssentialVars = 1.0;
    set<string> essentialVars;
    for(unsigned long k = 0; k < BDDVarOrder::size(); k++)
      if((csVars.cofactor(!BDDFunction::projection(k))) != 0.0){
	nonEssentialVars &= BDDFunction::projection(k);
      } else {
	#ifdef JINC_NEW
	essentialVars.insert(BDDVarOrder::getVariableName(k));
        #else
	essentialVars.insert(BDDVarOrder::getVariable(k));
	#endif
      }
    BDDFunction result = BDDFunction(fun.getAssignment());
    result = result.exists(nonEssentialVars);
    
    // We need to enrich the element with essential variables which are neglected by the assignment.
    set<string> someEssentialVars = result.getEssentialVariables();
    for(set<string>::iterator it = essentialVars.begin(); it != essentialVars.end(); it++)
      if(someEssentialVars.find(*it) == someEssentialVars.end())
	result &= !BDDFunction(*it);

    return result;
  }

/**
 * Function which considers a state number as a bitmask and builds the corresponding state over a set of variables.
 * @param number is the number of the state to convert.
 * @param stateVars is the list of variable names over which the state is to be represented.
 */
  BDDFunction SymbAutomaton::stateNumberToState(long long number, vector<BDDFunction> stateVars) const
  {
    BDDFunction result = 1.0;
    for(unsigned int i = 0; i < stateVars.size(); i++) {
      if(number & (1 << i))
	result &= stateVars[i];
      else
	result &= !stateVars[i];
    }
    return result;
  }
  
  /* Rename ce_a to ce_b, i.e. in the automaton ce_a and ce_b behave identical.
   * If ce_a or ce_b have shared data variables, this is achieved by synching the 
   * ce_a and ce_b, ce_a will still be in the automaton.
   * Otherwise, ce_a will be composed with ce_b and subsequently removed from
   * the automaton. It should not be touched afterwards.
   */ 
  void SymbAutomaton::rename_ce(Symb_Channel_End::ptr ce_a, Symb_Channel_End::ptr ce_b) {

    CHECK(5, check_symb_ce_sanity(), "");
    
    if (ce_a->areDataVariablesShared() || ce_b->areDataVariablesShared()) {
      // if they are shared, we can not necessarily compose, synch the ports instead
      // and keep both symb_ces
      // this is not as efficient, TODO: better 
      delta&=ce_a->synched(*ce_b);

      return;
    }

    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    delta=delta.compose(ce_a->getPortActive().getFirstVariableIndex(), ce_b->getPortActive());
    csPorts=csPorts.exists(ce_a->getPortActive());
    BDDFunctionContainer DC_a = ce_a->getDCVars();
    BDDFunctionContainer DC_b = ce_b->getDCVars();
    for (unsigned int i=0; i<DC_a.size(); ++i) {
      delta=delta.compose(DC_a.at(i).getFirstVariableIndex(), DC_b.at(i));
    }
    
    csDCVariables=csDCVariables.exists(container2cs(DC_a));
    remove_symb_ce_from_set(ce_a);

    CHECK(5, check_symb_ce_sanity(), "");
  }



  void SymbAutomaton::addStoppedTransitions() {
    CHECK_SUCCESS(getStoppedTransitionLabel()==BDDFunction::ZERO());
    
    BDDFunction terminal_states=calcTerminalStates();
    
    stopped_transition_label=VarManager::getInstance()->findOrAddGlobalBits("STOPPED", 1).at(0);
    csPorts&=stopped_transition_label;
    delta&=!stopped_transition_label;
    
    delta|=terminal_states & stopped_transition_label & getHaltSystem();
  }


  BDDFunction SymbAutomaton::getStoppedTransitionLabel() const {
    return stopped_transition_label;
  }  

  SymbTraceExplorer::ptr SymbAutomaton::getTraceExplorer() {
    return SymbTraceExplorer::ptr(new SymbTraceExplorerAutomaton(shared_from_this()));
  }


  bool SymbAutomaton::isValidTrace(SymbTrace& trace) const {
    if (trace.countStates()>0) {
      TRACE_BDD_FUNCTION("SymbAutomaton.isValidTrace", trace.getState(0).getX(), "");
      TRACE_BDD_FUNCTION("SymbAutomaton.isValidTrace", getInitialStates(), "");
      
      bool rv=bdd_is_element_of_set(trace.getState(0).getX(),
                                    getInitialStates());
      if (!rv) return false;
    }

    for (unsigned int i=0;i<trace.countTransitions();i++) {
      SymbTransition transition=trace.getTransition(i);
      TRACE_BDD_FUNCTION("SymbAutomaton.isValidTrace", transition, "");

      bool rv=bdd_is_element_of_set(transition.getTransitionFunction(),
                                    getDelta());

      if (!rv) return false;
    }

    return true;
  }  
}
