/***************************************************************************
 *   Copyright (C) 2008-2009 by Tobias Blechmann                           *
 *   blechman@tcs.inf.tu-dresden.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
@author Tobias Blechmann
*/

#ifndef SYMB_AUTOMATON_H
#define SYMB_AUTOMATON_H

#include <string>
#include <vector>
#include <set>
#include <iostream>
#include <boost/smart_ptr.hpp>

#include "vereofy/setup.hpp"
#include "vereofy/system/automaton.hpp"
#include "vereofy/system/reo_channel_end.hpp"
#include "symb_channel_end.hpp"
#include "symb_transition_label_var.hpp"
#include "symb_datadomain.hpp"
#include "symb_trace.hpp"
#include "symb_trace_explorer.hpp"
#include "varmanager.hpp"
#include "bddsubsetiterator.hpp"
#include "vereofy/common/tracing.hpp"
#include "vereofy/common/bdd_cached_function.hpp"


#define TRACE_SYMB_AUTOMATON_PATHS(scope, name, automaton)		\
  TRACE(scope, name,							\
  {									\
    cout << "Trace: Paths for " << scope << "::" << name << " ";	\
    cout << Tracing::getTraceCount(TRACE_GET_FULL_NAME(scope, name));	\
    cout << endl;							\
    (automaton).printPaths();						\
  }									\
	)


#define TRACE_SYMB_AUTOMATON_DOT(scope, name, automaton)		\
  TRACE(scope, name,							\
  {									\
    string filename=string(scope)+"_"+name+".";				\
    filename+=boost::lexical_cast<string>(Tracing::getTraceCount(TRACE_GET_FULL_NAME(scope, name))); \
    filename+=".dot";							\
    cout << "Trace: Writing DOT file " << filename <<endl;		\
    ofstream out(filename.c_str());					\
    if (out) {								\
      (automaton).toDot(out);						\
    } else {								\
      cout << "Couldn't write " << filename << endl;			\
    }									\
  }									\
	)

namespace vereofy {
  class SymbAutomaton;
  
    /// The SymbAutomaton class provides everything needed to represent symbolic constraint automaton
  class SymbAutomaton: public Automaton, public boost::enable_shared_from_this<SymbAutomaton> {
  public:
    typedef boost::shared_ptr<SymbAutomaton> ptr;
    typedef shared_ptr_id_set<Symb_Channel_End::ptr>::type symb_ce_set;
    typedef shared_ptr_id_set<Symb_Transition_Label_Var::ptr>::type symb_tl_set;

    enum hide_type {HIDE_TRANSITIVE, HIDE_TRANSITIVE_FULL, HIDE_EPSILON};

    SymbAutomaton(Symb_DataDomain::ptr dataDomain);
    SymbAutomaton(const SymbAutomaton& other, bool share_symb_ces=true);  // clone
    virtual ~SymbAutomaton();
    virtual void clear();
    
    virtual void product(const Automaton&);  // CA product
    virtual void sync_product(const SymbAutomaton&);  // synchronous product
    virtual Automaton::ptr copyProduct(const Automaton&) const;
    virtual Automaton::ptr copyUnite(const Automaton&) const;
    
    virtual void hide(const Automaton_Channel_End::ptr&);
    virtual void hide(const std::vector<Automaton_Channel_End::ptr>& ces);
    virtual void hide(const Automaton_Channel_End::ptr&, hide_type t);
    virtual void hide(const std::vector<Automaton_Channel_End::ptr>& ces, hide_type t);

    virtual void hide_semi(const std::vector<Automaton_Channel_End::ptr>& ces);
    
    virtual double getNumberOfReachableStates() const;

    virtual SymbAutomaton::ptr makeDeterministicCopy() const;
    virtual void makeDeterministic();
    virtual SymbAutomaton::ptr makeDeterministicCopy(BDDFunction& accepting) const;
    virtual void makeDeterministic(BDDFunction& accepting);
    
    BDDFunction getDelta() const;
    BDDFunction getInitialStates() const;
    BDDFunction getTransitionVars() const; // ports+dcs+labels
    BDDFunction getPortVars() const;
    BDDFunction getDataConstraintVars() const;
    BDDFunction getTransitionLabelVars() const;    
    BDDFunction getOriginalStateVars() const;
    BDDFunction getCopyStateVars() const;
    BDDFunction getCopyCopyStateVars() const;
    BDDFunction getHaltSystem() const;

    void setInitialStates(const BDDFunction& initial_states);
    void setDelta(const BDDFunction& delta);

    BDDFunction calcReachableStates() const;
    BDDFunction calcTerminalStates() const;
    BDDFunction calcReachableTerminalStates() const;

    string stateToBin(const BDDFunction& state_assignment, bool use_x) const;
    string stateToBin(const BDDFunction& state_assignment, const BDDFunction& csVars) const;

    const symb_ce_set& getSymbChannelEnds() const;
    void rename_ce(Symb_Channel_End::ptr, Symb_Channel_End::ptr);      

    const symb_tl_set& getSymbTransitionLabelVars() const;

    bool isValidTrace(SymbTrace& trace) const;
    
    // given a transition assignment, return all transitions that are equivalent 
    // (have different data assignment for the inactive ports)
    virtual BDDFunction getEquivalentTransitions(const BDDFunction& transition) const;

    // canonicalizeTransitions takes a transition_set and removes all but one for each set
    // of equivalent transitions. E.g., for a transition with an inactive port, the (irrelevant) 
    // data bits are all set to exactly one value (NULL).
    // The resulting transition set is returned
    virtual BDDFunction canonicalizeTransitions(const BDDFunction& transition_set) const;

    // ----- output functions -----

    struct State_Renderer {
      typedef boost::shared_ptr<State_Renderer> ptr;

      State_Renderer(const SymbAutomaton& automaton_) : automaton(automaton_) {};
      virtual ~State_Renderer() {}

      virtual void render(ostream& out, const BDDFunction& assignment, bool use_x) const;
      virtual bool render_dot_style(ostream& out, const BDDFunction& assignment, bool is_initial, bool is_terminal) const;

    protected:
      const SymbAutomaton& automaton;
    };

    struct IO_Renderer {
      typedef boost::shared_ptr<IO_Renderer> ptr;

      IO_Renderer(const SymbAutomaton& automaton_) : automaton(automaton_) {};
      virtual ~IO_Renderer() {}

      virtual void render(ostream& out, const BDDFunction& assignment, const BDDFunction& io_vars) const;
      virtual bool render_dot_style(ostream& out, 
                                    const BDDFunction& assignment, 
                                    const BDDFunction& io_vars,
                                    const BDDFunction& state_pair) const;

    protected:
      const SymbAutomaton& automaton;
    };

    virtual State_Renderer::ptr getDefaultStateRenderer() const;
    virtual IO_Renderer::ptr getDefaultIORenderer() const;

    virtual operator std::string() const;

    virtual std::string toDotString() const;
    virtual void toDot(std::ostream& out) const;
    virtual void toDot(std::ostream& out, const State_Renderer&, const IO_Renderer&) const;

    virtual void print(std::ostream& out, const SymbTrace& trace) const;
    virtual void print(std::ostream& out, const SymbTrace& trace, 
                       const State_Renderer&, const IO_Renderer&) const;

    virtual void print(std::ostream& out, const SymbState& state, 
                       const State_Renderer&) const;
    virtual void print(std::ostream& out, const SymbTransitionLabel& label, 
                       const IO_Renderer&) const;

    void printPaths() const;

    struct ExtractExplicitAutomatonFunctor {
      ExtractExplicitAutomatonFunctor() {}
      virtual ~ExtractExplicitAutomatonFunctor() {}

      virtual void start() {}
      virtual void finished() {}

      virtual void explicit_state(unsigned long state_id,
                                  BDDFunction state_assignment,
                                  bool is_initial,
                                  bool is_terminal) = 0; 

      virtual void explicit_transition(unsigned long state_id_from,
                                       unsigned long state_id_to,
                                       BDDFunction transition_assignment) = 0; 
    };

    void extractExplicitAutomaton(ExtractExplicitAutomatonFunctor& functor) const;

    SymbTraceExplorer::ptr getTraceExplorer();

    Symb_DataDomain::ptr getDataDomain() const;

    static hide_type getDefaultHideType();
    static void setDefaultHideType(hide_type t);
    static void setDefaultHideType(const std::string& t);

    void addStoppedTransitions();
    BDDFunction getStoppedTransitionLabel() const;

    bool hasTransitionLabel(Symb_Transition_Label_Var::ptr) const;

  protected:
    VarManager* varManager;
    Symb_DataDomain::ptr dataDomain;
    BDDFunction delta;
    BDDFunction initialStates;
    BDDFunction csPorts;
    BDDFunction csDCVariables;
    BDDFunction csTransitionLabelVariables;
    BDDFunction csOriginalX;
    BDDFunction csCopyY;
    BDDFunction csCopyCopyZ;
    BDDFunction haltSystem;

    BDDCachedFunction cachedReachableStates;
    BDDCachedFunction cachedTerminalStates;

    BDDFunction stopped_transition_label;

    // store Symb_Channel_Ends to have correlation between active bit and data bits
    symb_ce_set symb_ces;
    
    symb_tl_set symb_tls; // transition labels

    //TODO: SymbAutomaton has to keep a shared_ptr on the Channel_Ends belonging to this automaton
    void addChannelEndToAutomaton(Symb_Channel_End::ptr);
    void registerStateVars(const BDDFunctionContainer& x_vars);
    
    void addTransitionLabelToAutomaton(Symb_Transition_Label_Var::ptr);

    BDDFunction csAllOtherVariables(const BDDFunction&) const;      
    
    StateVars createStateVars(unsigned int);

    // create nr_of_states new state vars, mixed with var_positions_in_out
    // if var_positions_in_out is too short, append new positions
    StateVars createStateVars(unsigned int number, BDDFunctionContainer& var_positions_in_out);

    StateVars findOrAddGlobalState(const string& name, unsigned int bits);

    bool check_symb_ce_sanity() const;
    void copy_symb_ces(const SymbAutomaton& other);
    void copy_symb_tls(const SymbAutomaton& other);

    bool ces_are_all_inactive(const Symb_Channel_End::ptr_set& ces, const BDDFunction& assignment) const;
    BDDFunction ces_are_all_inactive(const Symb_Channel_End::ptr_set& ces) const;

    // Helper functions for makeDeterministic
    BDDFunction stateNumberToState(long long number, vector<BDDFunction> stateVars) const;
    BDDFunction getElem(BDDFunction fun, BDDFunction csVars) const;

    // do the product of the meta_information (csVariables, ...)
    void do_product_for_meta_information(const SymbAutomaton& other);

    // default hide type
    static hide_type default_hide_type; 

  private:
    void add_symb_ce_to_set(const Symb_Channel_End::ptr& symb_ce);
    void remove_symb_ce_from_set(const Symb_Channel_End::ptr& symb_ce);

    void add_symb_tl_to_set(const Symb_Transition_Label_Var::ptr& symb_tl);


  protected:
    void hide_epsilon(const BDDFunction& port_active_vars, const BDDFunction& data_vars);
    void hide_transitive(const BDDFunction& port_active_vars, const BDDFunction& data_vars);
    void hide_transitive_full();
  };
  
} //end namespace

#endif
