/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/




#include "symb_bfs.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include <vector>

using namespace std;


namespace vereofy {

  SymbBFS::SymbBFS(BDDFunction delta_,
                   BDDFunction csStateVars_,
                   BDDFunction csTransitionVars_,
                   BDDFunction stateConstraint_) :
    delta(delta_),
    csStateVars(csStateVars_),
    csTransitionVars(csTransitionVars_),
    stateConstraint(stateConstraint_),
    state_normalizer(BDDFunction::ONE()),
    transition_normalizer(BDDFunction::ONE()) {
    
    
  }
  
  void SymbBFS::addStateNormalizer(BDDFunction state_normalizer_) {
    state_normalizer&=state_normalizer_;
  }
   
  void SymbBFS::addTransitionNormalizer(BDDFunction transition_normalizer_) {
    transition_normalizer&=transition_normalizer_;
  }

  BDDFunction SymbBFS::postImage(const BDDFunction& state_set) {
    const BDDFunction& Q=state_set;

    BDDFunction delta_Q=delta & Q;
    BDDFunction Q_post_y=delta_Q.exists(csStateVars & csTransitionVars);
    BDDFunction Q_post=Q_post_y.moveUp();

    return Q_post.constrain(stateConstraint);
  }
  
  BDDFunction SymbBFS::preImage(const BDDFunction& state_set) {
    const BDDFunction& Q=state_set;

    BDDFunction csStateVarsY=csStateVars.moveDown();

    BDDFunction Q_y=Q.moveDown();
    BDDFunction delta_Q_y=delta & Q_y;

    BDDFunction Q_pre=delta_Q_y.exists(csStateVarsY & csTransitionVars);

    return Q_pre.constrain(stateConstraint);
  }

  ReturnValue<SymbTrace> SymbBFS::findTraceBetweenStateSets(BDDFunction from_set, 
                                                            BDDFunction to_set) {
    typedef vector<BDDFunction> set_vector_t;

    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    set_vector_t bfs_sets;
    
    BDDFunction reached_from_states=BDDFunction::ZERO();

    bfs_sets.push_back(to_set);

    TRACE_BDD_FUNCTION("SymbBFS_findTraceBetweenStateSets", from_set, "");
    TRACE_BDD_FUNCTION("SymbBFS_findTraceBetweenStateSets", to_set, "");

    BDDFunction border=preImage(to_set);

    BDDFunction visited=border;
    BDDFunction visited_old=BDDFunction::ZERO();
    while (visited != visited_old) {
      TRACE_BDD_FUNCTION("SymbBFS_findTraceBetweenStateSets", border, "");
      bfs_sets.push_back(border);
      
      if (bdd_sets_have_intersection(from_set, border)) {
        // we have reached at least one state from the from_set
        reached_from_states=from_set & border;
        break;
      }

      border=preImage(border);
      visited_old=visited;
      visited|=border;
    }

    if (reached_from_states == BDDFunction::ZERO()) {
      // from_set was not reached
      return ReturnValue<SymbTrace>();
    }

    TRACE("SymbBFS_findTraceBetweenStateSets", "bfs_result", \
          for (unsigned int i=0;i<bfs_sets.size();i++) {     \
            std::cout << "bfs_sets["<<i<<"]:\n";             \
            bfs_sets[i].printPaths();
          });

    TRACE_BDD_FUNCTION("SymbBFS_findTraceBetweenStateSets", reached_from_states, "");

    SymbTrace trace;
    SymbState s_from=SymbState::pickState(reached_from_states,
                                          csStateVars);
    
    unsigned int i=0;
    TRACE_BDD_FUNCTION("SymbBFS_findTraceBetweenStateSets", (BDDFunction)s_from, i++);
    set_vector_t::reverse_iterator it=bfs_sets.rbegin();
    // skip the last bfs_set, as it contains s_from
    ++it;

    for (;it!=bfs_sets.rend();++it) {
      const BDDFunction& Q_post=*it;
      SymbTransition t=SymbTransition::pickTransitionBetweenSets(delta,
                                                                 s_from,
                                                                 Q_post,
                                                                 csStateVars,
                                                                 csTransitionVars);
      trace.appendTransition(t);
      s_from=t.getToState();
      TRACE_BDD_FUNCTION("SymbBFS_findTraceBetweenStateSets", s_from, i++);
    }

    return trace;
  }

}
