/***************************************************************************
 *   Copyright (C) 2008-2009 by Tobias Blechmann                           *
 *   blechman@tcs.inf.tu-dresden.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 @author Tobias Blechmann
*/

#include <boost/lexical_cast.hpp>
#include <iostream>

#include "symb_channel_end.hpp"
#include "varmanager.hpp"

#include "vereofy/common/bdd_helpers.hpp"

using namespace std;

namespace vereofy {

  int Symb_Channel_End::nr_of_channel_ends = 0;

  Symb_Channel_End::Symb_Channel_End(Symb_DataDomain::ptr dataDomain_) :
    dataDomain(dataDomain_), state(CE_NORMAL) {
#ifdef JINC_NEW  
    debugOut(3, "no JINC_NEW Symb_Channel_End::Symb_Channel_End()", this );
    //TODO add version without names for variables when new JINC is used
#endif
    string ce_nr = boost::lexical_cast<string>(nr_of_channel_ends++);
    BDDFunctionContainer dc_vars=VarManager::getInstance()->createVars("#"+ce_nr, dataDomain->getLogSize());
    vecDCVars=dc_vars_vector(new BDDFunctionContainer(dc_vars));
    portActive = VarManager::getInstance()->createVars("*"+ce_nr, 1).at(0);
  }

  
  Symb_Channel_End::Symb_Channel_End(Symb_DataDomain::ptr dataDomain_, VarPositions::ptr var_positions_in_out) :
    dataDomain(dataDomain_), state(CE_NORMAL) {
#ifdef JINC_NEW  
    debugOut(3, "no JINC_NEW Symb_Channel_End::Symb_Channel_End()", this );
    //TODO add version without names for variables when new JINC is used
#endif
    string ce_nr = boost::lexical_cast<string>(nr_of_channel_ends++);
    BDDFunctionContainer dc_vars=VarManager::getInstance()->createVars("#"+ce_nr, dataDomain->getLogSize(), var_positions_in_out->positions);
    vecDCVars=dc_vars_vector(new BDDFunctionContainer(dc_vars));
    portActive = VarManager::getInstance()->createVars("*"+ce_nr, 1, var_positions_in_out->positions).at(0);
  }

  Symb_Channel_End::Symb_Channel_End(Symb_DataDomain::ptr dataDomain_, const BDDFunctionContainer& variables) :
    dataDomain(dataDomain_),  state(CE_NORMAL) {
    CHECK_SUCCESS(dataDomain->getLogSize()+1 == variables.size());
    
    portActive=variables.at(0);
    vecDCVars=dc_vars_vector(new BDDFunctionContainer());
    for (unsigned int i=1;i<variables.size();i++) {
      vecDCVars->push(variables.at(i));
    }
  }

  Symb_Channel_End::Symb_Channel_End(Symb_DataDomain::ptr dataDomain_,
                                     BDDFunction portActive_,
                                     dc_vars_vector vecDCVars_) :
    dataDomain(dataDomain_),  state(CE_NORMAL), portActive(portActive_), vecDCVars(vecDCVars_) {
  }


  Symb_Channel_End::Symb_Channel_End(const Symb_Channel_End& other) : 
    dataDomain(other.dataDomain), state(other.state), portActive(other.portActive), vecDCVars(other.vecDCVars) {
  }

  Symb_Channel_End::ptr Symb_Channel_End::clone() const {
    return Symb_Channel_End::ptr(new Symb_Channel_End(*this));
  }

  Symb_Channel_End::ptr Symb_Channel_End::getSharedDataChannelEnd() const {
    CHECK_SUCCESS(mayShareDataVariables());
    string ce_nr = boost::lexical_cast<string>(nr_of_channel_ends++);

    // generate new portActive variable near the original
    BDDFunctionContainer var_positions;
    var_positions.push(this->portActive);

    BDDFunction new_portActive = VarManager::getInstance()->createVars("*"+ce_nr, 1, var_positions).at(0);

    return Symb_Channel_End::ptr(new Symb_Channel_End(this->dataDomain,
                                                      new_portActive,
                                                      this->vecDCVars));
  }

  Symb_Channel_End::ptr Symb_Channel_End::getSharedDataChannelEnd(const BDDFunction& portActive) const {
    CHECK_SUCCESS(mayShareDataVariables());

    return Symb_Channel_End::ptr(new Symb_Channel_End(this->dataDomain,
                                                      portActive,
                                                      this->vecDCVars));
  }


  BDDFunction Symb_Channel_End::equals(const Symb_Channel_End& ce) const {
    debugOut(3, "Symb_Channel_End equals(" + boost::lexical_cast<string>(&ce) + ")", this );
    // TODO: verify that the data domains agree
    BDDFunction result = portActive & ce.portActive;
    for (unsigned long i=0; i < dataDomain->getLogSize(); ++i) {
      result &= !(vecDCVars->at(i) ^ ce.vecDCVars->at(i));
    }
    return result;    
  }

  
  BDDFunction Symb_Channel_End::synched(const Symb_Channel_End& ce) const {
    debugOut(3, "Symb_Channel_End synched(" + boost::lexical_cast<string>(&ce) + ")", this );
    // TODO: verify that the data domains agree
    BDDFunction result = 
      (portActive & ce.portActive & pairwiseEQUIV(*vecDCVars, *ce.vecDCVars) )
       | (!portActive & !ce.portActive);

    return result;    
  }

  
  BDDFunction Symb_Channel_End::data_equal(const BDDFunctionContainer& _vecDCVars) const {
    debugOut(3, "Symb_Channel_End equals(BDDFunctionContainer)", this);
    BDDFunction result = 1;
    for (unsigned long i=0; i < dataDomain->getLogSize(); ++i) {
      result &= !(vecDCVars->at(i) ^ _vecDCVars.at(i));
    }
    return result;    
  }

  
  Symb_Channel_End::channel_end_state_t Symb_Channel_End::getState() const {
    return state;
  }

  bool Symb_Channel_End::isHidden() const {
    return state == CE_HIDDEN;
  }

  bool Symb_Channel_End::isSemiHidden() const {
    return state == CE_SEMI_HIDDEN;
  }

  bool Symb_Channel_End::isVisible() const {
    return state == CE_NORMAL;
  }

  bool Symb_Channel_End::mayShareDataVariables() const {
    return vecDCVars && !isHidden();
  }

  bool Symb_Channel_End::areDataVariablesShared() const {
    return (vecDCVars && !vecDCVars.unique());
  }  

  bool Symb_Channel_End::areDataVariablesSharedWith(const Symb_Channel_End::ptr& other) const {
    return (vecDCVars && vecDCVars==other->vecDCVars);
  }  

  void Symb_Channel_End::hide() {
    state = CE_HIDDEN;
  }

  void Symb_Channel_End::semi_hide() {
    CHECK_SUCCESS(state == CE_NORMAL);
    state=CE_SEMI_HIDDEN;
  }
  
  string Symb_Channel_End::getVariableName() const {
    return bdd_var_name(portActive);
  }


  bool Symb_Channel_End::isSamePort(const Symb_Channel_End& other) const {
    return getPortActive()==other.getPortActive();
  }

  void Symb_Channel_End::print_vars(std::ostream& out, std::string name) const {
    out << portActive.getFirstVariableIndex() << " " << bdd_var_name(portActive) << " " << name << (isHidden() ? " (h)" : "") << " *\n";
    for (unsigned int i=0;i<vecDCVars->size();i++) {
      BDDFunction p=vecDCVars->at(i);
      out << p.getFirstVariableIndex() << " " << bdd_var_name(p) << " " << name << (isHidden() ? " (h)" : "") << " #" << i << "\n";
    }
  }

  void Symb_Channel_End::serialize_vars(std::ostream& out) const {
    VarManager::ReorderLock lock;

    out << portActive.getFirstVariableIndex();
    for (unsigned int i=0;i<vecDCVars->size();i++) {
      out << " " << vecDCVars->at(i).getFirstVariableIndex();
    }    
    out << "\n";
  }


  Symb_Channel_End::ptr_set Symb_Channel_End::getSharedDataCEsFromSet(const Symb_Channel_End::ptr_set& ces) const {
    ptr_set result;

    for (ptr_set::const_iterator it=ces.begin();
         it!=ces.end();
         ++it) {
      const Symb_Channel_End::ptr& cur=*it;

      if (cur.get()!=this &&    // not ourself
          areDataVariablesSharedWith(cur)) {
        result.insert(cur);
      }
    }
    return result;
  }

}
