/***************************************************************************
 *   Copyright (C) 2009 by Joachim Klein                                   *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "symb_component_info.hpp"
#include "symb_component_instance.hpp"

#include "vereofy/system/port_id.hpp"
#include "vereofy/system/state_id.hpp"

using namespace vereofy;

const SymbolicState::ptr 
SymbComponentInfo::getState(const string& name) const {
  StateDescription::ptr sd=c_interface->getStateDescription(name);

  if (!sd)
    return SymbolicState::ptr();

  return getState(sd);
}

const SymbolicState::ptr 
SymbComponentInfo::getState(StateDescription::ptr state_d) const {
  state_to_bdd_map::const_iterator it=state_to_bdd.find(state_d);
  if (it==state_to_bdd.end()) {
    return SymbolicState::ptr();
  }
  return it->second;
}

const Symb_Transition_Label_Var::ptr 
SymbComponentInfo::getTransitionLabel(const string& name) const {
  TransitionLabelDescription::ptr ld=c_interface->getTransitionLabelDescription(name);

  if (!ld)
    return Symb_Transition_Label_Var::ptr();

  return getTransitionLabel(ld);
}

const Symb_Transition_Label_Var::ptr 
SymbComponentInfo::getTransitionLabel(TransitionLabelDescription::ptr transition_label_d) const {
  transition_label_to_bdd_map::const_iterator it=transition_label_to_bdd.find(transition_label_d);
  if (it==transition_label_to_bdd.end()) {
    return Symb_Transition_Label_Var::ptr();
  }
  return it->second;
}


void SymbComponentInfo::addSubcomponentAutomaton(const string& name, SymbComponentInstance::ptr sci) {
  bool result=subcomponent_automata.insert(make_pair(name, sci)).second;
  CHECK_SUCCESS(result);
}


void SymbComponentInfo::addSubcomponentInfo(const string& name, SymbComponentInfo::ptr info) {
  bool result=subcomponent_infos.insert(make_pair(name, info)).second;
  CHECK_SUCCESS(result);
}


SymbComponentInstance::ptr SymbComponentInfo::getSubcomponentAutomaton(const string& name) const {
  subcomponent_automaton_map::const_iterator it=subcomponent_automata.find(name);
  if (it==subcomponent_automata.end()) {
    return SymbComponentInstance::ptr();
  }
  return it->second;
}

SymbComponentInfo::ptr SymbComponentInfo::getSubcomponentInfo(const string& name) const {
  subcomponent_info_map::const_iterator it=subcomponent_infos.find(name);
  if (it==subcomponent_infos.end()) {
    return SymbComponentInfo::ptr();
  }
  return it->second;
}

subcomponent_info_map::const_iterator SymbComponentInfo::subcomponent_infos_begin() const {
  return subcomponent_infos.begin();
}

subcomponent_info_map::const_iterator SymbComponentInfo::subcomponent_infos_end() const {
  return subcomponent_infos.end();
}


const BDDFunction& SymbComponentInfo::getAtomicProposition(const std::string& ap_name) const {
  CHECK_SUCCESS(hasAtomicProposition(ap_name));

  return atomic_propositions.find(ap_name)->second;
}


bool SymbComponentInfo::hasAtomicProposition(const std::string& ap_name) const {
  return (atomic_propositions.count(ap_name) >= 1);
}

void SymbComponentInfo::addAtomicProposition(const std::string& name, const BDDFunction& ap_function) {
  CHECK_SUCCESS(!hasAtomicProposition(name));

  atomic_propositions.insert(make_pair(name, ap_function));
}
  
SymbComponentInfo::atomic_proposition_map::const_iterator SymbComponentInfo::atomic_propositions_begin() const {
  return atomic_propositions.begin();
}

SymbComponentInfo::atomic_proposition_map::const_iterator SymbComponentInfo::atomic_propositions_end() const {
  return atomic_propositions.end();
}


void SymbComponentInfo::print_var_info(std::ostream& out) const {
  if (!component_instance) return;

  for (state_to_bdd_map::const_iterator it=state_to_bdd.begin();
       it!=state_to_bdd.end();
       ++it) {
    for (unsigned int i=0;i<it->second->vars.X->size();i++) {
      BDDFunction p=it->second->vars.X->at(i);
      out << p.getFirstVariableIndex() << " " << bdd_var_name(p) << " " << StateID(component_instance, it->first) << " #" << i << "\n";
    }
  }

  for (port_to_bdd_map::const_iterator it=port_to_bdd.begin();
       it!=port_to_bdd.end();
       ++it) {
    it->second->print_vars(out, boost::lexical_cast<string>(PortID(component_instance, it->first)));
  }

  for (subcomponent_info_map::const_iterator it=subcomponent_infos_begin();
       it!=subcomponent_infos_end();
       ++it) {
    it->second->print_var_info(out);
  }
}
