/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "symb_component_instance.hpp"
#include "vereofy/reo_parser/REOParser.hpp"
#include "vereofy/reo_parser/module_parser.hpp"
#include "vereofy/reo_parser/guard_parser.hpp"
#include "vereofy/reo_parser/guard_parser_helper.hpp"
#include "vereofy/reo_parser/int_parser_helper.hpp"
#include "vereofy/common/setup.hpp"
#include "vereofy/common/errors.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/common/xml_helpers.hpp"
#include "vereofy/system/symb_automaton/symb_array_datadomain.hpp"
#include "vereofy/system/symb_automaton/symb_struct_datadomain.hpp"
#include "vereofy/system/symb_automaton/symb_tagged_union_datadomain.hpp"
#include "vereofy/system/symb_automaton/symb_function_arithmetic.hpp"
#include "vereofy/system/symb_automaton/symb_function_constructor.hpp"

#include "vereofy/system/symb_automaton/serialized_varorder.hpp"

#include <boost/algorithm/string.hpp>

#include <sstream>

using namespace std;
using namespace vereofy;
using namespace vereofy::vparser;

SymbComponentInstance::SymbComponentInstance(ComponentInterface::ptr c_interface_,
					     Symb_DataDomain::ptr dataDomain) :
  SymbAutomaton(dataDomain),
  info(new SymbComponentInfo(c_interface_)),
  show_only_toplevel_ports(true),
  reachabilityConstrain(BDDFunction::ONE()) {
  init();
}

void SymbComponentInstance::set_var_info(port_to_bdd_map port_map_init,
					 state_to_bdd_map state_map_init,
                                         VarPositions::ptr var_positions_) {
  info->port_to_bdd=port_map_init;
  info->state_to_bdd=state_map_init;
  var_positions=var_positions_;
}


void SymbComponentInstance::setShowOnlyToplevelPorts(bool enabled) {
  show_only_toplevel_ports=enabled;
}

void SymbComponentInstance::init() {
  if (!var_positions) {
    var_positions = VarPositions::ptr(new VarPositions());
  }
  no_state_changes=BDDFunction::ONE();
}


const SymbolicState::ptr 
SymbComponentInstance::getState(const string& name) const {
  return info->getState(name);
}

const SymbolicState::ptr 
SymbComponentInstance::getState(StateDescription::ptr state_d) const {
  return info->getState(state_d);
}

SymbComponentInstance::ptr SymbComponentInstance::getSubcomponentAutomaton(const string& name) const {
  return info->getSubcomponentAutomaton(name);
}

SymbComponentInfo::ptr SymbComponentInstance::getSubcomponentInfo(const string& name) const {
  return info->getSubcomponentInfo(name);
}


Symb_Channel_End::ptr SymbComponentInstance::addPort(const PortDescription::ptr port) {
  if (info->port_to_bdd.count(port)==0) {
    // no symb_ce yet, try to find equivalent port
    Symb_Channel_End::ptr symb_ce=find_symb_ce_for_port(port);
    if (symb_ce) {
      return addPort(port, symb_ce);
    }

#ifdef MIX_POSITIONS
    Symb_Channel_End::ptr ce(new Symb_Channel_End(dataDomain, var_positions));
#else
    Symb_Channel_End::ptr ce(new Symb_Channel_End(dataDomain));
#endif
    return addPort(port, ce);
  } else {
    return addPort(port, info->port_to_bdd[port]);
  }
}

Symb_Channel_End::ptr SymbComponentInstance::addPort(const PortDescription::ptr port, 
                                                     const Symb_Channel_End::ptr symb_ce) {
  if (info->port_to_bdd.count(port)>0) {
    if (!(info->port_to_bdd[port]==symb_ce)) {
      throw "addPort: SymbolicChannelEnds don't agree";
    }
  } else {
    info->port_to_bdd[port]=symb_ce;
  }

  register_symb_ce_for_port(port, symb_ce);
  addChannelEndToAutomaton(symb_ce);

  TRACE("SymbComponentInstance.addPort", "insert",                      \
        cout << "addPort ";                                             \
        if (info->component_instance)                                   \
          cout << info->component_instance->getFQName();                \
        else                                                            \
          cout << &info->ce_to_port;                                    \
        cout << " " << port->getName() << ": " << symb_ce << endl;);

  info->ce_to_port.insert(make_pair(symb_ce, port));

  TRACE("SymbComponentInstance.addPort", "var_name", cout << "SymbCE for " << port->getName() << " : " << symb_ce->getVariableName()<<"\n"<< endl);

  return symb_ce;
}


Symb_Transition_Label_Var::ptr SymbComponentInstance::addTransitionLabel(const TransitionLabelDescription::ptr transition_label_description) {
  if (info->transition_label_to_bdd.count(transition_label_description)==0) {

#ifdef MIX_POSITIONS
    Symb_Transition_Label_Var::ptr symb_tl(new Symb_Transition_Label_Var(transition_label_description, var_positions));
#else
    Symb_Transition_Label_Var::ptr symb_tl(new Symb_Transition_Label_Var(transition_label_description));
#endif
    return addTransitionLabel(transition_label_description, symb_tl);
  } else {
    return addTransitionLabel(transition_label_description,
                              info->transition_label_to_bdd[transition_label_description]);
  }
}


Symb_Transition_Label_Var::ptr SymbComponentInstance::addTransitionLabel(const TransitionLabelDescription::ptr label_d, 
                                                                         const Symb_Transition_Label_Var::ptr symb_tl) {
  if (info->transition_label_to_bdd.count(label_d)>0) {
    if (!(info->transition_label_to_bdd[label_d]==symb_tl)) {
      throw "addTransitionLabel: Symb_Transition_Label_Var don't agree";
    }
  } else {
    info->transition_label_to_bdd[label_d]=symb_tl;
  }

  //  register_symb_ce_for_port(port, symb_ce);
  addTransitionLabelToAutomaton(symb_tl);

  //  info->ce_to_port.insert(make_pair(symb_ce, port));

  TRACE("SymbComponentInstance.addTransitonLabel", "var_name", cout << "Symb_Transition_Label_Var for " << label_d->getName() << " : " << symb_tl->getVariableName()<<"\n"<< endl);

  return symb_tl;
}



void SymbComponentInstance::addState(const StateDescription::ptr sd) {
  if (info->state_to_bdd.count(sd)==0) {
    if (build_context &&
        build_context->state_id_to_deserialized_symbolic_state &&
        info->component_instance) {
      StateID state_id(info->component_instance, sd);
      SymbolicState::ptr ss=build_context->state_id_to_deserialized_symbolic_state->get(state_id);
      if (ss) {
        TRACE("SymbComponentInstance", "deserialized", 
              std::cout << "Found deserialized SymbolicState for " << state_id << "\n");
        addState(sd, ss);
        return;
      }

      // else, continue & generate new state
    }


    SymbolicState::ptr ss(new SymbolicState);
    ss->data_domain=Symb_DataDomain::getFor(sd->getType());
    
    unsigned int needed_bits=ss->data_domain->getLogSize();

    if (debugLevel>=3) {
      std::cout << "Creating " << needed_bits << " state variables for variable " << sd->getName() << std::endl;
    }
    
#ifdef MIX_POSITIONS
    ss->vars=createStateVars(needed_bits, var_positions->positions);
#else
    ss->vars=createStateVars(needed_bits);
#endif

    addState(sd, ss);
  } else {
    throw "Trying to add state that is already there!";
  }
}

void SymbComponentInstance::addState(const StateDescription::ptr sd, 
                                     const SymbolicState::ptr ss) {
  if (info->state_to_bdd.count(sd)!=0) {
    throw "Trying to add state that is already there!";
  }
  
  info->state_to_bdd[sd]=ss;
  
  // register the state variables
  registerStateVars(*ss->vars.X);

  // update no_state_changes
  no_state_changes=no_state_change_bdd();
  
  // restriction to ensure that the state values are in range
  BDDFunction state_values_are_in_range=ss->data_domain->restrictToValidValues(*ss->vars.X);
  add_state_value_restriction(state_values_are_in_range);
}


#ifdef WITH_DUMMY_STATE
void SymbComponentInstance::addDummyState() {
#define GLOBAL_DUMMY_STATE
#ifdef GLOBAL_DUMMY_STATE
  dummy_state_var=findOrAddGlobalState("DUMMY", 1);
#else
  dummy_state_var=createStateVars(1);
#endif
  BDDFunction state_values_are_in_range=!dummy_state_var.X->at(0);
  add_state_value_restriction(state_values_are_in_range);
  no_state_changes=no_state_change_bdd();
}
#endif // WITH_DUMMY_STATE

void SymbComponentInstance::addSubcomponentAutomaton(SymbComponentInstance::ptr sub_component) {
  string name;
  if (sub_component->info->component_instance && 
      sub_component->info->component_instance->getName()!="") {
    name=sub_component->info->component_instance->getName();
    if (info->getSubcomponentAutomaton(name)) {
      cerr << "Warning: more than one subcomponent in SymbAutomaton with name " << name << endl;
      name=name+"$"+boost::lexical_cast<string>(info->getSubcomponentAutomataCount());
    }
  } else {
    name="$"+boost::lexical_cast<string>(info->getSubcomponentAutomataCount());
  }

  CHECK(2, !info->getSubcomponentAutomaton(name), "");
  info->addSubcomponentAutomaton(name, sub_component);
}

void SymbComponentInstance::addSubcomponentInfo(SymbComponentInfo::ptr sub_component_info) {
  string name;
  if (sub_component_info->component_instance && 
      sub_component_info->component_instance->getName()!="") {
    name=sub_component_info->component_instance->getName();
    if (info->getSubcomponentInfo(name)) {
      cerr << "Warning: more than one subcomponent in SymbAutomaton with name " << name << endl;
      name=name+"$"+boost::lexical_cast<string>(info->getSubcomponentInfoCount());
    }
  } else {
    name="$"+boost::lexical_cast<string>(info->getSubcomponentInfoCount());
  }

  CHECK(2, !info->getSubcomponentInfo(name), "");
  info->addSubcomponentInfo(name, sub_component_info);
}

void SymbComponentInstance::addSubcomponentRestrictions(SymbComponentInstance::ptr sub_component) {
  copy_state_value_restrictions(*sub_component);

  // TODO: insert non hidden internal ports
  //  ce_to_port.insert(sub_component->ce_to_port.begin(),
  //		    sub_component->ce_to_port.end());
}

void SymbComponentInstance::setComponentInstance(ComponentInstance::ptr ci) {
  info->component_instance=ci;
}

void SymbComponentInstance::hide(const Automaton_Channel_End::ptr& ce_) {
  Symb_Channel_End::ptr ce=boost::static_pointer_cast<Symb_Channel_End>(ce_);

  CHECK(5, check_symb_ce_sanity(), "");
  SymbAutomaton::hide(ce);
  CHECK(5, check_symb_ce_sanity(), "");
}

void SymbComponentInstance::hide(const std::vector<Automaton_Channel_End::ptr>& ces) {
  CHECK(5, check_symb_ce_sanity(), "");
  SymbAutomaton::hide(ces);
  CHECK(5, check_symb_ce_sanity(), "");
}


BDDFunctionContainer SymbComponentInstance::getPortRestrictions() {
  BDDFunctionContainer result;

  for (symb_ce_set::const_iterator it=getSymbChannelEnds().begin();
       it!=getSymbChannelEnds().end();
       ++it) {
    const Symb_Channel_End& ce=**it;
    
    // either port is not active or the data values are in range
    BDDFunction port_data_is_in_range=
      !ce.getPortActive() |
      dataDomain->restrictToValidValues(ce.getDCVars());

    result.push(port_data_is_in_range);
  }
  return result;
}

// result.first = restrictions for initialStates
// result.second = restrictions for delta
std::pair<BDDFunctionContainer, BDDFunctionContainer> 
SymbComponentInstance::getStateRestrictions() {
  std::pair<BDDFunctionContainer, BDDFunctionContainer> result;
  
  for (state_value_restriction_list::const_iterator it=state_value_restrictions.begin();
       it!=state_value_restrictions.end();
       ++it) {
    const BDDFunction& restriction=*it;
    result.first.push(restriction);
    result.second.push(restriction & restriction.moveDown());
  }
  return result;
}


void SymbComponentInstance::restrictToValidValuesPorts() {
  BDDFunctionContainer delta_restrictions=getPortRestrictions();

  // delta
  delta = containerApplyAND(delta, delta_restrictions);
}


void SymbComponentInstance::restrictToValidValuesStates() {
  BDDFunctionContainer init_restrictions;
  BDDFunctionContainer delta_restrictions;

  boost::tie(init_restrictions, delta_restrictions) =
    getStateRestrictions();

  // initial states
  initialStates = containerApplyAND(initialStates, init_restrictions);

  // delta
  delta = containerApplyAND(delta, delta_restrictions);
}

void SymbComponentInstance::restrictToValidValues() {
  BDDFunctionContainer init_restrictions;
  BDDFunctionContainer delta_restrictions;

  boost::tie(init_restrictions, delta_restrictions) =
    getStateRestrictions();

  if (!build_context->param_abstract_data_domain) {
    delta_restrictions.push(getPortRestrictions());
  }

  // initial states
  initialStates = containerApplyAND(initialStates, init_restrictions);

  // delta
  delta = containerApplyAND(delta, delta_restrictions);
}

BDDFunction SymbComponentInstance::no_state_change_bdd() const {
  BDDFunction result=BDDFunction::ONE();

#ifdef WITH_DUMMY_STATE
  if (dummy_state_var.X) {
    // dummy state
    return pairwiseEQUIV(*dummy_state_var.X, *dummy_state_var.Y);
  }
#endif // WITH_DUMMY_STATE

  for (state_to_bdd_map::const_iterator it=info->state_to_bdd.begin();
       it!=info->state_to_bdd.end();
       ++it) {
    StateDescription::ptr sd=it->first;
    SymbolicState::ptr ss=it->second;
    
    result&=pairwiseEQUIV(*(ss->vars.X), *(ss->vars.Y));
  }

  return result;
}

SymbExpression SymbComponentInstance::bdd_expression_vector::flatten() const {
  if (expression_vector.size()==1) {
    return expression_vector.at(0);
  }

  BDDFunctionContainer flat_values;
  for (unsigned int i=0;i<data_domain->getLogSize();i++) {
    flat_values.push(BDDFunction::ZERO());
  }
  BDDFunction or_of_conditions=BDDFunction::ZERO();

  for (unsigned int i=0;i<expression_vector.size();i++) {  
    BDDFunction condition=expression_vector.at(i).getCondition();
    or_of_conditions|=condition;
    
    for (unsigned int bit=0; bit < data_domain->getLogSize(); bit++) {
      BDDFunction value=expression_vector.at(i).getValues().at(bit);
      flat_values[bit]|=condition & value;
    }
  }
  
  return SymbExpression(data_domain,
                        flat_values,
                        or_of_conditions);
}

BDDFunction SymbComponentInstance::perform_compare(comp_op_t comp_op,
                                                   const SymbExpression& left,
                                                   const SymbExpression& right) const {
  return left.getDataDomain()->performComparison(comp_op, left, right) &
    left.getCondition() & right.getCondition();
};


SymbComponentInstance::bdd_assignment_vector
SymbComponentInstance::perform_assignment_any_value(const bdd_assignment_target_vector& left) const {
  bdd_assignment_vector result;

  for (unsigned int i=0; i<left.target_vector.size(); i++) {
    const bdd_assignment_target& target=left.target_vector.at(i);

    bdd_assignment a;
    a.assignment=target.condition;
    a.targets=target.targets;

    // perform co-assignments
    for (list<bdd_assignment>::const_iterator it=target.co_assignments.begin();
         it!=target.co_assignments.end();
         ++it) {
      a.assignment &= it->assignment;
      a.targets.push(it->targets);
    }

    result.push_back(a);
  }
  
  return result;
}

SymbComponentInstance::bdd_assignment_vector 
SymbComponentInstance::perform_assignment(const bdd_assignment_target_vector& left,
					  const SymbExpression& value) const {
  bdd_assignment_vector result;

  for (unsigned int i=0; i<left.target_vector.size(); i++) {
    const bdd_assignment_target& target=left.target_vector.at(i);

    const BDDFunction& target_condition=target.condition;
    const BDDFunction& value_condition=value.getCondition();

    BDDFunction condition=target_condition & value_condition;
    if (condition==BDDFunction::ZERO()) {
      continue;
    }
    
    bdd_assignment a;
    a.assignment=target.data_domain->performAssignment(target.targets,
                                                       value);
    a.assignment&=condition;  // add condition
    a.targets=target.targets;
    
    // perform co-assignments
    for (list<bdd_assignment>::const_iterator it=target.co_assignments.begin();
         it!=target.co_assignments.end();
         ++it) {
      a.assignment &= it->assignment;
      a.targets.push(it->targets);
    }
      
    result.push_back(a);
  }
  return result;
}

BDDFunction SymbComponentInstance::state_guard_to_bdd(const vparser::pt_node& node) {
  if (guard_helper::t_state_guard::isBoolFunctionCall(node)) {
    return handle_bool_function_call(node.children.at(0));
  }

  if (guard_helper::t_state_guard::isIfExpression(node)) {
    return bool_expression_to_bdd(node.children.at(0).children.at(0));
  }

  if (guard_helper::t_state_guard::isAPRef(node)) {
    SymbComponentInfo::ptr comp_info=resolveSubcomponentAccess(node.children.at(0));
    const vparser::pt_node& simple_ap_ref=guard_helper::t_generic_subcomponent_wrapper::getSimple(node.children.at(0));

    string ap_name=guard_helper::t_simple_ap_ref::getIdentifier(simple_ap_ref);

    if (!comp_info->hasAtomicProposition(ap_name)) {
      throw "Trying to access atomic proposition "+ap_name+", does not exist in the automaton";
    }

    return comp_info->getAtomicProposition(ap_name);
  }

  const vparser::pt_node& lhs=guard_helper::t_state_guard::getLHS(node);
  SymbExpression lhs_e=accessStateVar(lhs);

  comp_op_t compare_op=guard_helper::t_state_guard::getCompareOp(node);

  const vparser::pt_node& rhs=
    guard_helper::t_state_guard::getRHS(node);
  
  SymbExpression rhs_e = calc_expression(rhs, lhs_e.getDataDomain(), false);

  return perform_compare(compare_op, lhs_e, rhs_e);
}

BDDFunction SymbComponentInstance::bool_expression_to_bdd(const vparser::pt_node& node) {
  switch (node.value.id().to_long()) {
  case ID_G_bool_expression:
  case ID_G_bool_expression_xor:
  case ID_G_bool_expression_and:
  case ID_G_bool_expression_implies_equiv: {
    if (!guard_helper::t_bool_generic::hasOp(node)) {
      return bool_expression_to_bdd(guard_helper::t_bool_generic::getLeft(node));
    }
    std::string op=guard_helper::t_bool_generic::getOp(node);
    BDDFunction left=bool_expression_to_bdd(guard_helper::t_bool_generic::getLeft(node));
    BDDFunction right=bool_expression_to_bdd(guard_helper::t_bool_generic::getRight(node));
    
    if (op=="|") return left | right;
    if (op=="&") return left & right;
    if (op=="->") return IMPLIES(left,right);
    if (op=="<->") return EQUIV(left,right);
    if (op=="^") return left ^ right;
    
    throw "Unimplemented boolean operator :"+op;
  }
  case ID_G_bool_expression_term: {
    if (guard_helper::t_bool_expression_term::isTrue(node)) return BDDFunction::ONE();
    if (guard_helper::t_bool_expression_term::isFalse(node)) return BDDFunction::ZERO();
    if (guard_helper::t_bool_expression_term::isNegated(node)) 
      return !bool_expression_to_bdd(guard_helper::t_bool_expression_term::getChild(node));

    return bool_expression_to_bdd(guard_helper::t_bool_expression_term::getChild(node));
  }
  case ID_G_bool_expression_ranged: {
    const vparser::pt_node& range_def = guard_helper::t_bool_expression_ranged::getRangeDef(node);
    
    const string range_var = guard_helper::t_range_def::getRangeVar(range_def);
    DataDomain::ptr range_type=type_helper::t_type::getDataDomain(guard_helper::t_range_def::getRangeType(range_def),
                                                                  info->getComponentInterface()->getSymbolTableStack());
    
    if (range_type->getType()!=DataDomain::DD_FLAT) {
      throw "Parametrized boolean expressions with complex parameters not currently supported";
    }
    
    FlatDataDomain::ptr range_flat_dd = 
      boost::static_pointer_cast<FlatDataDomain>(range_type);
    
    if (!range_flat_dd->getFlatType() == FlatDataDomain::DD_INT) {
      throw "Parametrized boolean expressions with non-integer parameters not currently supported";
    }
    
    IntRangeDataDomain::ptr int_dd = boost::static_pointer_cast<IntRangeDataDomain>(range_flat_dd);
    
    
    BDDFunction result;
    if (guard_helper::t_bool_expression_ranged::getOperator(node) == "AND") {
      result = BDDFunction::ONE();
    } else if (guard_helper::t_bool_expression_ranged::getOperator(node) == "OR") {
      result = BDDFunction::ZERO();
    } else {
      throw "Unknown parametrized boolean operator "+guard_helper::t_bool_expression_ranged::getOperator(node);
    }
    
    symbol_table_stack_t& symbol_tables=const_cast<symbol_table_stack_t&>(info->getComponentInterface()->getSymbolTableStack());
    
    for (int i = int_dd->getLowestInt(); i <= int_dd->getHighestInt(); i++) {
      // prepare new symbol table for the constant
      symbol_table_t::ptr range_symbol_table(new symbol_table_t);
      
      range_symbol_table->add(range_var, 
                              ConstantData::ptr(new ConstantData(range_var, i)));
      range_symbol_table->read_only=true;
      
      symbol_tables.push(range_symbol_table);
      BDDFunction result_i = bool_expression_to_bdd(guard_helper::t_bool_expression_ranged::getExpression(node));
      symbol_tables.pop();

      if (guard_helper::t_bool_expression_ranged::getOperator(node) == "AND") {
        result &= result_i;
      } else if (guard_helper::t_bool_expression_ranged::getOperator(node) == "OR") {
        result |= result_i;
      } else {
        THROW_UNIMPLEMENTED();
      }
    }
    return result;
  }   


  case ID_G_compare_expression: {
    const vparser::pt_node& lhs=guard_helper::t_compare_expression::getLHS(node);
    comp_op_t compare_op=guard_helper::t_compare_expression::getCompareOp(node);
    const vparser::pt_node& rhs=guard_helper::t_compare_expression::getRHS(node);

    SymbExpression lhs_e=calc_expression(lhs, Symb_DataDomain::ptr(), false);    
    SymbExpression rhs_e=calc_expression(rhs, lhs_e.getDataDomain(), false);
    
    return perform_compare(compare_op, lhs_e, rhs_e);
  }    
  case ID_G_ite_bool_expression: {
    const vparser::pt_node& if_node = guard_helper::t_ite_bool_expression::getIf(node);
    const vparser::pt_node& then_node = guard_helper::t_ite_bool_expression::getThen(node);
    const vparser::pt_node& else_node = guard_helper::t_ite_bool_expression::getElse(node);
    
    BDDFunction if_function = bool_expression_to_bdd(if_node);
    BDDFunction then_function = bool_expression_to_bdd(then_node);
    BDDFunction else_function = bool_expression_to_bdd(else_node);

    BDDFunction ite_function = ITE(if_function, then_function, else_function);

    TRACE_BDD_FUNCTION("ite_bool_expression", if_function, "");
    TRACE_BDD_FUNCTION("ite_bool_expression", then_function, "");
    TRACE_BDD_FUNCTION("ite_bool_expression", else_function, "");
    TRACE_BDD_FUNCTION("ite_bool_expression", ite_function, "");

    return ite_function;
  }
  case ID_G_state_guard:
    return state_guard_to_bdd(node);
  case ID_G_io_guard:
    return io_guard_to_bdd(node);
  default:
    THROW_UNIMPLEMENTED();
  }
}

SymbExpression
SymbComponentInstance::accessMember(const vparser::pt_node& node_struct_member_access,
                                    const BDDFunctionContainer& struct_encoding,
                                    Symb_DataDomain::ptr sdd_) const {
  string member_name=guard_helper::t_struct_member_access::getIdentifier(node_struct_member_access);

  switch (sdd_->getType()) {
  case DataDomain::DD_STRUCT: {
    Symb_StructDataDomain::ptr sdd=boost::static_pointer_cast<Symb_StructDataDomain>(sdd_);
    StructDataDomain::ptr dd=sdd->getStructDataDomain();
  
    if (!dd->isMember(member_name)) {
      throw member_name+" is not a valid member of this struct!";
    }
    unsigned int member_index=dd->getIndexOfMember(member_name);
    
    Symb_DataDomain::ptr member_dd=sdd->getSymbDataDomainForMember(member_index);
    
    if (guard_helper::t_struct_member_access::hasStructMemberAccess(node_struct_member_access)) {
      return accessMember(guard_helper::t_struct_member_access::getStructMemberAccess(node_struct_member_access),
                          sdd->getEncodingForMember(struct_encoding, member_index),
                          member_dd);
    } else {
      return SymbExpression(member_dd, 
                            sdd->getEncodingForMember(struct_encoding, member_index));
    }
  }
  case DataDomain::DD_TAGGED_UNION: {
    Symb_TaggedUnionDataDomain::ptr sdd=boost::static_pointer_cast<Symb_TaggedUnionDataDomain>(sdd_);
    TaggedUnionDataDomain::ptr dd=sdd->getTaggedUnionDataDomain();
    
    if (!dd->isMember(member_name) && !(member_name == dd->getTagName())) {
      throw member_name+" is not a valid member of this tagged union!";
    }

    if (member_name==dd->getTagName()) {
      // accessing tag
      if (guard_helper::t_struct_member_access::hasStructMemberAccess(node_struct_member_access)) {
        throw "Trying to access member of tag";
      }

      return SymbExpression(sdd->getSymbDataDomainForTag(), 
                            sdd->getEncodingForTag(struct_encoding));
    }

    // normal member

    unsigned int member_index=dd->getIndexOfMember(member_name);
    Symb_DataDomain::ptr member_dd=sdd->getSymbDataDomainForMember(member_index);

    SymbExpression result;
    if (guard_helper::t_struct_member_access::hasStructMemberAccess(node_struct_member_access)) {
      result=accessMember(guard_helper::t_struct_member_access::getStructMemberAccess(node_struct_member_access),
                          sdd->getEncodingForMember(struct_encoding, member_index),
                          member_dd);
    } else {
      result=SymbExpression(member_dd, 
                            sdd->getEncodingForMember(struct_encoding, member_index));
    }

    BDDFunction type_condition = sdd->isMember(member_index, struct_encoding);
    result.setCondition(result.getCondition() & type_condition);

    return result;
  }
  default:
    throw "Trying to access member "+member_name+" of non-structured data domain!";
  }

  THROW_UNIMPLEMENTED();
}


SymbExpression
SymbComponentInstance::accessArray(const vparser::pt_node& node_array_access,
                                   const BDDFunctionContainer& array_encoding,
                                   Symb_DataDomain::ptr sdd_) const {
  CHECK_SUCCESS(node_array_access.value.id() == ID_G_array_access);

  if (sdd_->getType() != DataDomain::DD_ARRAY) {
    throw "Trying to access member of array, but data type is "+sdd_->getDataDomain()->toString();
  }
  
  Symb_ArrayDataDomain::ptr s_array_dd=
    boost::static_pointer_cast<Symb_ArrayDataDomain>(sdd_);
      
  ArrayDataDomain::ptr array_dd=s_array_dd->getArrayDataDomain();
      
  Symb_DataDomain::ptr base_dd=s_array_dd->getSymbBase();
  if (base_dd->getType()!=DataDomain::DD_FLAT) {
    THROW_UNIMPLEMENTED();
  }
  
  Symb_FlatDataDomain::ptr fdd=boost::static_pointer_cast<Symb_FlatDataDomain>(base_dd);
  
  const vparser::pt_node& array_ref_expression=
    guard_helper::t_array_access::getExpression(node_array_access);
  
  if (guard_helper::t_guard_expression::isConstantInt(array_ref_expression)) {
    int index=
      guard_helper::t_guard_expression::getConstantInt(array_ref_expression, 
                                                       info->c_interface->getSymbolTableStack()).getValue();
    
    if (index<0 ||
        (unsigned int)index>=array_dd->getArraySize()) {
      throw "Index out of range!";
    }
    
    return SymbExpression(fdd, s_array_dd->getEncoding(array_encoding, index));
  } else if (guard_helper::t_guard_expression::isFullExpression(array_ref_expression)) {
    Symb_FlatDataDomain::ptr index_domain=s_array_dd->getIndexDomain();
    
    SymbExpression ref_expression=calc_expression(guard_helper::t_guard_expression::getFullExpression(array_ref_expression),
                                                   index_domain, true);
        
    bdd_expression_vector result(fdd);
    
    for (unsigned int i=0;i<array_dd->getArraySize();i++) {
      BDDFunctionContainer index_encoding=index_domain->getEncoding(i);
      
      BDDFunction condition=perform_compare(COMPARE_EQUAL,
                                            SymbExpression(index_domain, index_encoding),
                                            ref_expression);
        
      result.expression_vector.push_back(SymbExpression(fdd, 
                                                        s_array_dd->getEncoding(array_encoding, i),
                                                        condition));
    }
    
    return result.flatten();
  } else {
    THROW_UNIMPLEMENTED();
  }
}

SymbExpression 
SymbComponentInstance::handle_tmp_var_def(const vparser::pt_node& node) const {
  const vparser::pt_node& tmp_var_type_n = guard_helper::t_tmp_var_def::getTmpVarType(node);
  DataDomain::ptr dd = type_helper::t_type::getDataDomain(tmp_var_type_n,
                                                          info->getComponentInterface()->getSymbolTableStack());

  const vparser::pt_node& tmp_var_init_n = guard_helper::t_tmp_var_def::getTmpVarInit(node);
  
  if (type_helper::t_type::isBoolDataType(tmp_var_type_n)) {
    BDDFunction init = const_cast<SymbComponentInstance&>(*this).bool_expression_to_bdd(tmp_var_init_n);
    
    BDDFunctionContainer init_values;
    init_values.push(init);

    return SymbExpression(Symb_DataDomain::getFor(dd), init_values);
  } else {
    // TODO: refactor dd
    SymbExpression init = calc_expression(tmp_var_init_n,
                                          Symb_DataDomain::ptr(),
                                          false);

    DataDomain::ptr target_dd=init.getDataDomain()->getDataDomain()->getConversionTarget(dd);
    Symb_DataDomain::ptr target_sdd=Symb_DataDomain::getFor(target_dd);
    SymbExpression converted=target_sdd->castExpression(init);

    converted.restrictToValidValues();
    
    return converted;
  }
}


SymbExpression 
SymbComponentInstance::handle_iterate_expression(const vparser::pt_node& node) const {
  const vparser::pt_node& range_def = guard_helper::t_iterate_expression::getRangeDef(node);
  const string range_var = guard_helper::t_range_def::getRangeVar(range_def);
  DataDomain::ptr range_type=type_helper::t_type::getDataDomain(guard_helper::t_range_def::getRangeType(range_def),
                                                                info->getComponentInterface()->getSymbolTableStack());
  if (range_type->getType()!=DataDomain::DD_FLAT) {
    throw "Range with complex data types currently not supported";
  }

  FlatDataDomain::ptr range_flat_dd = 
    boost::static_pointer_cast<FlatDataDomain>(range_type);
  
  if (!range_flat_dd->getFlatType() == FlatDataDomain::DD_INT) {
    throw "Range with non-integer parameters not currently supported";
  }
  
  IntRangeDataDomain::ptr int_dd = boost::static_pointer_cast<IntRangeDataDomain>(range_flat_dd);

  const vparser::pt_node& tmp_var_def = guard_helper::t_iterate_expression::getTmpVar(node);
  std::string tmp_var_name = guard_helper::t_tmp_var_def::getTmpVarName(tmp_var_def);

  SymbExpression tmp_var_init = handle_tmp_var_def(tmp_var_def);
  const vparser::pt_node& tmp_var_type_n = guard_helper::t_tmp_var_def::getTmpVarType(tmp_var_def);
  DataDomain::ptr result_dd = type_helper::t_type::getDataDomain(tmp_var_type_n,
                                                                 info->getComponentInterface()->getSymbolTableStack());  
  SymbExpression::ptr tmp_var(new SymbExpression(tmp_var_init));

  symbol_table_stack_t& symbol_tables=const_cast<symbol_table_stack_t&>(info->getComponentInterface()->getSymbolTableStack());
  
  for (int i = int_dd->getLowestInt(); i <= int_dd->getHighestInt(); i++) {
    // prepare new symbol table for this iteration
    symbol_table_t::ptr range_symbol_table(new symbol_table_t);
    
    range_symbol_table->add(range_var, 
                            ConstantData::ptr(new ConstantData(range_var, i)));
    range_symbol_table->add(tmp_var_name,
                            VirtualValueData::ptr(new VirtualValueData(tmp_var_name, tmp_var)));
    range_symbol_table->read_only=true;
    
    symbol_tables.push(range_symbol_table);
    
    SymbExpression iteration_result = calc_expression(guard_helper::t_iterate_expression::getExpression(node),
                                                      tmp_var->getDataDomain(),
                                                      false);

    DataDomain::ptr target_dd=iteration_result.getDataDomain()->getDataDomain()->getConversionTarget(result_dd);
    Symb_DataDomain::ptr target_sdd=Symb_DataDomain::getFor(target_dd);
    SymbExpression converted=target_sdd->castExpression(iteration_result);
    converted.restrictToValidValues();

    symbol_tables.pop();

    // store converted result, accumulate conditions
    tmp_var = SymbExpression::ptr(new SymbExpression(converted.getDataDomain(),
                                                     converted.getValues(),
                                                     converted.getCondition() & tmp_var->getCondition()));
  }
  return *tmp_var;
}

SymbComponentInfo::ptr
SymbComponentInstance::resolveSubcomponentAccess(const vparser::pt_node& node) const {
  SymbComponentInfo::ptr cur_info=this->info;
  
  for (unsigned int i=0; 
       i<guard_helper::t_generic_subcomponent_wrapper::countSubcomponentAccess(node); 
       i++) {
    const vparser::pt_node& subcomponent_node=guard_helper::t_generic_subcomponent_wrapper::getSubcomponent(node, i);
    string subcomponent=guard_helper::t_subcomponent_access::getSubcomponent(subcomponent_node);
    bool has_index=false;
    unsigned int index=0;
    
    string fullname=subcomponent;
    if (guard_helper::t_subcomponent_access::hasSubcomponentIndex(subcomponent_node)) {
      index=guard_helper::t_subcomponent_access::getSubcomponentIndex(subcomponent_node, 
                                                                      cur_info->c_interface->getSymbolTableStack());
      has_index=true;
      fullname+="["+boost::lexical_cast<string>(index)+"]";
    }

    SymbComponentInfo::ptr sub_info=cur_info->getSubcomponentInfo(fullname);
    if (!sub_info) {
      if (has_index && index==0) {
	sub_info=cur_info->getSubcomponentInfo(subcomponent);
      } else if (!has_index) {
	sub_info=cur_info->getSubcomponentInfo(subcomponent+"[0]");
      }
    }
    
    if (!sub_info) {
      throw "Couldn't find subcomponent "+fullname;
    }

    cur_info=sub_info;
  }

  return cur_info;
}


SymbExpression
SymbComponentInstance::accessStateVar(const vparser::pt_node& node_state_var) const {

  // if necessary, resolve subcomponent access, get the correct SymbComponentInfo scope
  SymbComponentInfo::ptr comp_info=resolveSubcomponentAccess(node_state_var);
  const vparser::pt_node& simple_state_var=guard_helper::t_generic_subcomponent_wrapper::getSimple(node_state_var);

  string var_name=
    guard_helper::t_simple_state_var::getIdentifier(simple_state_var);
    
  bool has_array=
    guard_helper::t_simple_state_var::hasArrayAccess(simple_state_var);

  SymbolicState::ptr ss=comp_info->getState(var_name);
  if (!ss) {throw string("Can't find symbolic state for variable "+var_name);}

  SymbExpression state_expression(ss->data_domain,
                                  *ss->vars.X);

  switch (state_expression.getDataDomain()->getType()) {
  case DataDomain::DD_FLAT: {
    CHECK(1, !has_array, ""); 
    return state_expression;
  }
  case DataDomain::DD_STRUCT:
  case DataDomain::DD_TAGGED_UNION: {
    if (guard_helper::t_simple_state_var::hasStructMemberAccess(simple_state_var)) {
      return accessMember(guard_helper::t_simple_state_var::getStructMemberAccess(simple_state_var),
                          state_expression.getValues(),
                          state_expression.getDataDomain());
    } else {
      // just get the whole struct variable
      CHECK(1, !has_array, ""); 
      return state_expression;
    }
  }
  case DataDomain::DD_ARRAY: {
    if (has_array) {    
      const vparser::pt_node& array_access=
        guard_helper::t_simple_state_var::getArrayAccess(simple_state_var);

      return accessArray(array_access,
                         state_expression.getValues(),
                         state_expression.getDataDomain());
      
    } else {
      // just provide access to whole array
      return state_expression;
    }
  }
  }
  THROW_UNIMPLEMENTED();
}



SymbComponentInstance::bdd_assignment_target_vector
SymbComponentInstance::getAssignmentTargets(const vparser::pt_node& node_state_var,
					    bool assign_to_y) const {
  if (guard_helper::t_state_var::hasSubcomponentAccess(node_state_var)) {
    throw "Can't assign something to a subcomponent state!";
  }
  const vparser::pt_node& simple_state_var=guard_helper::t_state_var::getSimple(node_state_var);
  
  string lhs_var_name=guard_helper::t_simple_state_var::getIdentifier(simple_state_var);
 
  SymbolicState::ptr lhs_ss=getState(lhs_var_name);
 
  switch (lhs_ss->data_domain->getType()) {
  case DataDomain::DD_FLAT: {
    return getAssignmentTargetFlat(lhs_var_name, assign_to_y);
  }
  case DataDomain::DD_STRUCT: 
  case DataDomain::DD_TAGGED_UNION: {
    if (guard_helper::t_simple_state_var::hasStructMemberAccess(simple_state_var)) {
      return getAssignmentTargetMemberAccess(assign_to_y ? *(lhs_ss->vars.Y) : *(lhs_ss->vars.X),
                                             guard_helper::t_simple_state_var::getStructMemberAccess(simple_state_var),
                                             lhs_ss->data_domain);
    } else {
      // assignment target for whole struct/tagged_union
      return getAssignmentTargetFlat(lhs_var_name, assign_to_y);
    }
  }
  case DataDomain::DD_ARRAY: {
    if (!guard_helper::t_simple_state_var::hasArrayAccess(simple_state_var)) {
      // assignment target for whole array
      return getAssignmentTargetFlat(lhs_var_name, assign_to_y);
    }
    
    // with array access:
    Symb_ArrayDataDomain::ptr s_array_dd=
      boost::static_pointer_cast<Symb_ArrayDataDomain>(lhs_ss->data_domain);
    ArrayDataDomain::ptr array_dd=s_array_dd->getArrayDataDomain();

    Symb_DataDomain::ptr base_dd=s_array_dd->getSymbBase();
    if (base_dd->getType()!=DataDomain::DD_FLAT) {
      THROW_UNIMPLEMENTED();
    }
    Symb_FlatDataDomain::ptr fdd=boost::static_pointer_cast<Symb_FlatDataDomain>(base_dd);

    const vparser::pt_node& array_ref_expression=
      guard_helper::t_simple_state_var::getArrayAccessExpression(simple_state_var);
            
    if (guard_helper::t_guard_expression::isConstantInt(array_ref_expression)) {
      int index=
	guard_helper::t_guard_expression::getConstantInt(array_ref_expression,
                                                         info->c_interface->getSymbolTableStack()).getValue();
      
      if (index<0 ||
	  (unsigned int)index>=array_dd->getArraySize()) {
	throw "Index out of range!";
      }
      
      return getAssignmentTargetArray(lhs_var_name, index, assign_to_y);
    } else if (guard_helper::t_guard_expression::isFullExpression(array_ref_expression)) {
      Symb_FlatDataDomain::ptr index_domain=s_array_dd->getIndexDomain();

      SymbExpression ref_expression=calc_expression(guard_helper::t_guard_expression::getFullExpression(array_ref_expression),
                                                     index_domain, true);
      
      bdd_assignment_target_vector result(fdd);

      for (unsigned int i=0;i<array_dd->getArraySize();i++) {
	BDDFunctionContainer index_encoding=index_domain->getEncoding(i);
	
        BDDFunction condition=perform_compare(COMPARE_EQUAL,
                                              SymbExpression(index_domain, index_encoding),
                                              ref_expression);
        
        result.target_vector.push_back(bdd_assignment_target(fdd, 
                                                             s_array_dd->getEncoding(*(assign_to_y ? lhs_ss->vars.Y : lhs_ss->vars.X), i),
                                                             condition));
      }
      
      return result;
    } else {
      THROW_UNIMPLEMENTED();
    }
  }
  }
  
  throw "implementation error";
}


/**
   Returns an assignment target for variable 'name' as a 
   flat target (i.e. the whole encoding bits)
 */
SymbComponentInstance::bdd_assignment_target
SymbComponentInstance::getAssignmentTargetFlat(const string& name,
					       bool assign_to_y) const {
  SymbolicState::ptr lhs_ss=getState(name);
  if (!lhs_ss) {throw string("Can't find symbolic state for variable "+name);}
  
  const BDDFunctionContainer& lhs_encoding=
    assign_to_y 
    ? *(lhs_ss->vars.Y) 
    : *(lhs_ss->vars.X);
  
  return bdd_assignment_target(lhs_ss->data_domain, lhs_encoding);
}

SymbComponentInstance::bdd_assignment_target
SymbComponentInstance::getAssignmentTargetMemberAccess(const BDDFunctionContainer& struct_encoding,
                                                       const vparser::pt_node& node_struct_member_access,
                                                       Symb_DataDomain::ptr sdd_) const {
  string member_name=guard_helper::t_struct_member_access::getIdentifier(node_struct_member_access);
  
  switch (sdd_->getType()) {
  case DataDomain::DD_STRUCT: {
    Symb_StructDataDomain::ptr sdd=boost::static_pointer_cast<Symb_StructDataDomain>(sdd_);
    StructDataDomain::ptr dd=sdd->getStructDataDomain();
    
    if (!dd->isMember(member_name)) {
      throw member_name+" is not a valid member of this struct!";
    }
    unsigned int member_index=dd->getIndexOfMember(member_name);
    
    Symb_DataDomain::ptr member_dd=sdd->getSymbDataDomainForMember(member_index);
    
    if (guard_helper::t_struct_member_access::hasStructMemberAccess(node_struct_member_access)) {
      return getAssignmentTargetMemberAccess(sdd->getEncodingForMember(struct_encoding, member_index),
                                             guard_helper::t_struct_member_access::getStructMemberAccess(node_struct_member_access),
                                             member_dd);
   } else {
      return bdd_assignment_target(member_dd,
                                   sdd->getEncodingForMember(struct_encoding, member_index));
    }
  }
  case DataDomain::DD_TAGGED_UNION: {
    Symb_TaggedUnionDataDomain::ptr sdd=boost::static_pointer_cast<Symb_TaggedUnionDataDomain>(sdd_);
    TaggedUnionDataDomain::ptr dd=sdd->getTaggedUnionDataDomain();
    
    string member_name=guard_helper::t_struct_member_access::getIdentifier(node_struct_member_access);
    
    if (!dd->isMember(member_name) && !(member_name==dd->getTagName())) {
      throw member_name+" is not a valid member of this taggedUnion!";
    }
    
    if (member_name == dd->getTagName()) {
      throw "Trying to assign to the tag of a tagged_union, which is illegal!";
    } else {
      // normal member
      unsigned int member_index=dd->getIndexOfMember(member_name);
      Symb_DataDomain::ptr member_dd=sdd->getSymbDataDomainForMember(member_index);

      SymbComponentInstance::bdd_assignment_target result;
      if (guard_helper::t_struct_member_access::hasStructMemberAccess(node_struct_member_access)) {
        result=getAssignmentTargetMemberAccess(sdd->getEncodingForMember(struct_encoding, member_index),
                                               guard_helper::t_struct_member_access::getStructMemberAccess(node_struct_member_access),
                                               member_dd);
      } else {
        result=bdd_assignment_target(member_dd,
                                     sdd->getEncodingForMember(struct_encoding, member_index));
      }


      bdd_assignment co_assignment;
      co_assignment.targets = sdd->getEncodingForTag(struct_encoding);
      co_assignment.assignment = sdd->getSymbDataDomainForTag()->performAssignment(co_assignment.targets,
                                                                                   SymbExpression(sdd->getSymbDataDomainForTag(),
                                                                                                  sdd->getSymbDataDomainForTag()->getEncoding(member_index)));
      result.co_assignments.push_back(co_assignment);
      return result;
    }
  }
  default:
    throw "Trying to assign to member "+member_name+", but data type is non-structured";
  }
  
  // should never be reached
  THROW_UNIMPLEMENTED();
}


SymbComponentInstance::bdd_assignment_target
SymbComponentInstance::getAssignmentTargetArray(const string& array_name,
						unsigned int index,
						bool assign_to_y) const {

  SymbolicState::ptr ss=getState(array_name);
  if (!ss) {throw string("Can't find symbolic state for variable "+array_name);}
  
  if (ss->data_domain->getType()!=DataDomain::DD_ARRAY) {
    THROW_UNIMPLEMENTED();
  }
  
  Symb_ArrayDataDomain::ptr s_array_dd=
    boost::static_pointer_cast<Symb_ArrayDataDomain>(ss->data_domain);
  
  ArrayDataDomain::ptr array_dd=s_array_dd->getArrayDataDomain();
  
  if (index>=array_dd->getArraySize()) {
    throw "Index out of range!";
  }
  
  Symb_DataDomain::ptr base_dd=s_array_dd->getSymbBase();
  if (base_dd->getType()!=DataDomain::DD_FLAT) {
    THROW_UNIMPLEMENTED();
  }
  
  Symb_FlatDataDomain::ptr lhs_data_domain=
    boost::static_pointer_cast<Symb_FlatDataDomain>(base_dd);
  BDDFunctionContainer lhs_encoding=
    s_array_dd->getEncoding(*(assign_to_y ? ss->vars.Y : ss->vars.X),
			      index);
  
  return bdd_assignment_target(lhs_data_domain, lhs_encoding);
}


BDDFunction SymbComponentInstance::io_guard_to_bdd(const vparser::pt_node& node) {
  if (guard_helper::t_io_guard::isEpsilon(node)) {
    PortDescription::ptr_set no_active_ports;
    return only_ports_active(no_active_ports);
  } else if (guard_helper::t_io_guard::isDataflow(node)) {
    PortDescription::ptr_set no_active_ports;
    return !only_ports_active(no_active_ports);
  } else if (guard_helper::t_io_guard::isStep(node)) {
    return BDDFunction::ONE();
  } else if (guard_helper::t_io_guard::is(node, ID_G_bool_function_call)) {
    return handle_bool_function_call(guard_helper::t_io_guard::get(node));
  } else if (guard_helper::t_io_guard::is(node, ID_G_if_expression)) {
    return bool_expression_to_bdd(guard_helper::t_if_expression::getExpression(guard_helper::t_io_guard::get(node)));
  } else if (guard_helper::t_io_guard::is(node, ID_G_port_activity_expression)) {
    const vparser::pt_node& pae=guard_helper::t_io_guard::get(node);
    
    if (guard_helper::t_port_activity_expression::isSinglePort(pae)) {
      guard_helper::t_port_activity_expression::single_port_info sp_info;
      sp_info=guard_helper::t_port_activity_expression::getSinglePort(pae, info->c_interface->getSymbolTableStack());
      Symb_Channel_End::ptr port_ce=get_symb_ce_for_port(get_port_description(sp_info.port_name));
      return port_active(port_ce, !sp_info.negated);
    } else if (guard_helper::t_port_activity_expression::isActivePorts(pae)) {
      const vparser::pt_node& active_ports=
	guard_helper::t_port_activity_expression::getActivePorts(pae);

      PortDescription::ptr_set active_port_pds;
      for (unsigned int i=0;
	   i<guard_helper::t_active_ports::getPortCount(active_ports);
	   ++i) {
	PortDescription::ptr pd=get_port_description(guard_helper::t_active_ports::getPortName(active_ports, 
											       i,
											       info->c_interface->getSymbolTableStack()));
	active_port_pds.insert(pd);
      }
      return only_ports_active(active_port_pds);
    } else {
      throw "implementation error";
    }
  } else if (guard_helper::t_io_guard::is(node, ID_G_port_data_expression)) {
    const vparser::pt_node& pde=guard_helper::t_io_guard::get(node);

    const vparser::pt_node& port_data_node=guard_helper::t_port_data_expression::getLHS(pde);
    
    PortDescription::ptr lhs_port=
      get_port_description(guard_helper::t_port_data::getPortName(port_data_node,
								  info->c_interface->getSymbolTableStack()));

    Symb_DataDomain::ptr lhs_data_domain;
    BDDFunctionContainer lhs_encoding;
    BDDFunction lhs_condition=BDDFunction::ONE();
    
    if (guard_helper::t_port_data::hasStructMemberAccess(port_data_node)) {
      
      if (dataDomain->getType()!=DataDomain::DD_STRUCT &&
          dataDomain->getType()!=DataDomain::DD_TAGGED_UNION) {
        throw "Trying to access member of port data, but data domain is not structured";
      }

      SymbExpression e=accessMember(guard_helper::t_port_data::getStructMemberAccess(port_data_node),
                                    get_data_vars_for_port(lhs_port),
                                    dataDomain);
      
      lhs_data_domain=e.getDataDomain();
      lhs_encoding=e.getValues();
      lhs_condition=e.getCondition();
    } else {
      lhs_data_domain=dataDomain;
      lhs_encoding=get_data_vars_for_port(lhs_port);
    }
    
    comp_op_t compare_op=guard_helper::t_port_data_expression::getCompareOp(pde);
    
    const vparser::pt_node& rhs=
      guard_helper::t_port_data_expression::getRHS(pde);
    
    BDDFunction result=port_active(get_symb_ce_for_port(lhs_port));
    
    SymbExpression lhs_expression(lhs_data_domain, lhs_encoding, lhs_condition);
    SymbExpression rhs_expression=calc_expression(rhs, lhs_data_domain, false);
    
    result&=perform_compare(compare_op, lhs_expression, rhs_expression);
    
    return result;
  } else if (guard_helper::t_io_guard::is(node, ID_G_has_label_expression)) {
    const vparser::pt_node& has_label_expression=guard_helper::t_io_guard::get(node);
    const vparser::pt_node& label_id=
      guard_helper::t_has_label_expression::getLabelID(has_label_expression);
    
    // if necessary, resolve subcomponent access, get the correct SymbComponentInfo scope
    SymbComponentInfo::ptr comp_info=resolveSubcomponentAccess(label_id);
    string label_name=guard_helper::t_label_id::getLabelName(label_id);
    
    Symb_Transition_Label_Var::ptr symb_tl = comp_info->getTransitionLabel(label_name);
    if (!symb_tl) {
      throw "Trying to access label "+label_name+", does not exist";
    }

    if (symb_tls.count(symb_tl)==0) {
      throw "Trying to access label "+label_name+", does not exist anymore";
    }

    return symb_tl->getHasLabelVar();    
  } else {
    throw "implementation error";
  }
}

SymbExpression
SymbComponentInstance::get_int_expression(int i_,
                                          Symb_DataDomain::ptr target_data_domain,
                                          bool strict_target) const {
    if (i_<0) {
      throw "negative ints not yet supported";
    }
    unsigned int i=(unsigned int)i_;
    if (strict_target) {
      CHECK_SUCCESS(target_data_domain);

      if (   (target_data_domain->getType()==DataDomain::DD_FLAT &&
              !(static_cast<Symb_FlatDataDomain&>(*target_data_domain).getFlatDataDomain()->isValidValue(i)))
           || target_data_domain->getType()!=DataDomain::DD_FLAT) {
        cerr << "Value out of range of data domain!\n";
        cerr << " Value: " << i << "\n";
        cerr << " Data domain: " << target_data_domain->getDataDomain()->toString() << "\n";
      }
    }

    return SymbExpression(i);
}

SymbExpression
SymbComponentInstance::handle_function_call(const vparser::pt_node& expression_node,
                                            Symb_DataDomain::ptr target_data_domain,
                                            bool strict_target) const {
  std::string function_name=guard_helper::t_function_call::getFunctionName(expression_node);
  SymbolData::ptr sd=info->c_interface->getGlobalSymbols()->find(function_name);
  if (!sd) {
    throw "Function call for function "+function_name+", does not exist";
  }

  SymbFunction::ptr symb_function;
  
  switch (sd->type) {
  case SYMBOL_TYPE_FUNCTION: {
    FunctionData::ptr fd=boost::static_pointer_cast<FunctionData>(sd);
    symb_function=fd->function;
    break;
  }
  case SYMBOL_TYPE_TYPEDEF: {
    TypeDefData::ptr tdd=boost::static_pointer_cast<TypeDefData>(sd);
    DataDomain::ptr dd=tdd->data_domain;
    Symb_DataDomain::ptr sdd=Symb_DataDomain::getFor(dd);

    symb_function=sdd->getConstructorFunction();
    break;
  }
  default:
    throw "Function call for function "+function_name+", is not a function!";
  }

  unsigned int call_arity=guard_helper::t_function_call::getArity(expression_node);
  if (call_arity!=symb_function->getArity()) {
    throw "Function call for function "+function_name+", mismatch in number of parameters (Expected "+
      boost::lexical_cast<std::string>(symb_function->getArity())+", got "+
      boost::lexical_cast<std::string>(call_arity)+")";
  }


  BDDFunction argument_conditions=BDDFunction::ONE();
  SymbFunctionArgumentVector arguments;
  for (unsigned int i=0;i<call_arity;i++) {
    const vparser::pt_node& argument_node=guard_helper::t_function_call::getArgumentExpression(expression_node, i);
    if (argument_node.value.id() == parser_id(ID_G_guard_expression)
        && guard_helper::t_guard_expression::isConstantInt(argument_node)) {
      int i=guard_helper::t_guard_expression::getConstantInt(argument_node, 
                                                             info->c_interface->getSymbolTableStack()).getValue();
      if (i<0) {throw "Function call: Negative values not yet supported!";}
      
      arguments.push_back(SymbFunctionArgument((unsigned int) i));
      
    } else {
      // non-const expression
      SymbExpression argument=calc_expression(argument_node,
                                              symb_function->getExpectedArgumentType(i, arguments),
                                              false);
      SymbExpression::ptr argument_expression(new SymbExpression(argument));
      arguments.push_back(SymbFunctionArgument(argument_expression));
      argument_conditions&=argument_expression->getCondition();
    }
  }
  
  SymbExpression::ptr result=symb_function->calculate(arguments);
  result->setCondition(result->getCondition() & argument_conditions);
  return *result;
}    

BDDFunction
SymbComponentInstance::handle_bool_function_call(const vparser::pt_node& node) const {
  CHECK_SUCCESS(node.value.id() == parser_id(ID_G_bool_function_call));
  SymbExpression result=handle_function_call(node, Symb_DataDomain::ptr(), false);

  DataDomain::ptr result_dd=result.getDataDomain()->getDataDomain();
  if (result_dd->getType()!=DataDomain::DD_FLAT ||
      boost::static_pointer_cast<FlatDataDomain>(result_dd)->getFlatType()!=FlatDataDomain::DD_INT ||
      boost::static_pointer_cast<IntRangeDataDomain>(result_dd)->getHighestInt()>1) {

    std::string msg = std::string("Call to function ")+
      guard_helper::t_function_call::getFunctionName(node)+
      " has no boolean return value, return value is of type "+
      result_dd->toString();

    throw msg;
  }

  return perform_compare(COMPARE_NOTEQUAL, result, SymbExpression(0));
}

SymbExpression
SymbComponentInstance::handle_ite_expression(const vparser::pt_node& expression_node,
                                             Symb_DataDomain::ptr target_data_domain,
                                             bool strict_target) const {
  CHECK_SUCCESS(expression_node.value.id() == parser_id(ID_G_ite_expression));

  const vparser::pt_node& if_node = guard_helper::t_ite_expression::getIf(expression_node);
  const vparser::pt_node& then_node = guard_helper::t_ite_expression::getThen(expression_node);
  const vparser::pt_node& else_node = guard_helper::t_ite_expression::getElse(expression_node);

  // const_cast necessary as bool_expression_to_bdd is not syntactially const (in reality, it is)
  BDDFunction if_function = const_cast<SymbComponentInstance&>(*this).bool_expression_to_bdd(if_node);
  
  SymbExpression then_expression = calc_expression(then_node, target_data_domain, false);
  SymbExpression else_expression = calc_expression(else_node, target_data_domain, false);

  Symb_DataDomain::ptr sdd_then = then_expression.getDataDomain();
  Symb_DataDomain::ptr sdd_else = else_expression.getDataDomain();
  DataDomain::ptr dd_then = sdd_then->getDataDomain();
  DataDomain::ptr dd_else = sdd_else->getDataDomain();

  if (dd_then->isSameAs(*dd_else)) {
    // cast else_expression to data domain of then_expression, just to be sure
    else_expression=sdd_then->castExpression(else_expression);
  } else {
    if (dd_then->canMergeWith(*dd_else)) {
      DataDomain::ptr dd_merged = dd_then->mergedWith(*dd_else);
      Symb_DataDomain::ptr sdd_merged = Symb_DataDomain::getFor(dd_merged);
      
      then_expression = sdd_merged->castExpression(then_expression);
      else_expression = sdd_merged->castExpression(else_expression);
    } else {
      throw std::string("In ITE(...) expression, data types of 'then' and 'else' expression differ: \n")+
        dd_then->toString()+" and "+dd_else->toString();
    }
  }

  // Construct result expression
  BDDFunctionContainer result;
  CHECK_SUCCESS(then_expression.getValues().size() == else_expression.getValues().size());
  
  for (unsigned int i=0;i<then_expression.getValues().size();i++) {
    result.push(ITE(if_function, 
                    then_expression.getValues().at(i), 
                    else_expression.getValues().at(i))); 
  }

  BDDFunction condition = ITE(if_function, then_expression.getCondition(), else_expression.getCondition());

  return SymbExpression(then_expression.getDataDomain(),
                        result,
                        condition);
}


SymbExpression
SymbComponentInstance::calc_expression(const vparser::pt_node& expression_node,
				       Symb_DataDomain::ptr target_data_domain,
                                       bool strict_target) const {
  if (expression_node.value.id() == parser_id(ID_G_guard_expression)
      && guard_helper::t_guard_expression::isConstantInt(expression_node)) {
    int i=guard_helper::t_guard_expression::getConstantInt(expression_node, 
                                                           info->c_interface->getSymbolTableStack()).getValue();

    return get_int_expression(i, target_data_domain, strict_target);
  }

  switch (guard_helper::t_generic::getType(expression_node)) {
  case ID_G_guard_expression:
    if (!guard_helper::t_guard_expression::isFullExpression(expression_node)) {
      THROW_UNIMPLEMENTED();
    }
    return calc_expression(guard_helper::t_guard_expression::getFullExpression(expression_node),
                           target_data_domain, strict_target);
  case ID_G_guard_expression_full:
    if (guard_helper::t_guard_expression_full::hasOp(expression_node)) {
      const vparser::pt_node& left=guard_helper::t_guard_expression_full::getLeft(expression_node);
      const vparser::pt_node& right=guard_helper::t_guard_expression_full::getRight(expression_node);

      SymbExpression left_e=calc_expression(left,
                                            target_data_domain, false);
      SymbExpression right_e=calc_expression(right,
                                             target_data_domain, false);

      if (guard_helper::t_guard_expression_full::getOp(expression_node)=="+") {
        SymbFunction_uint_add uint_add;
	SymbFunctionArgumentVector arguments(2);
	arguments[0]=SymbFunctionArgument(SymbExpression::ptr(new SymbExpression(left_e)));
	arguments[1]=SymbFunctionArgument(SymbExpression::ptr(new SymbExpression(right_e)));
        
        SymbExpression result=*uint_add.calculate(arguments);
        result.setCondition( result.getCondition() & left_e.getCondition() & right_e.getCondition() );
        return result;
      } else if (guard_helper::t_guard_expression_full::getOp(expression_node)=="-") {
        SymbFunction_uint_sub uint_sub;
	SymbFunctionArgumentVector arguments(2);
	arguments[0]=SymbFunctionArgument(SymbExpression::ptr(new SymbExpression(left_e)));
	arguments[1]=SymbFunctionArgument(SymbExpression::ptr(new SymbExpression(right_e)));
        
        SymbExpression result=*uint_sub.calculate(arguments);
        result.setCondition( result.getCondition() & left_e.getCondition() & right_e.getCondition() );
        return result;
      } else {
        THROW_UNIMPLEMENTED();
      }
    } else {
      // no operator
      return calc_expression(guard_helper::t_guard_expression_full::getLeft(expression_node),
                             target_data_domain, strict_target);
    }
  case ID_G_guard_term:{
    if (guard_helper::t_guard_term::hasOp(expression_node)) {
      const vparser::pt_node& left=guard_helper::t_guard_term::getLeft(expression_node);
      const vparser::pt_node& right=guard_helper::t_guard_term::getRight(expression_node);

      
      SymbExpression left_e=calc_expression(left,
                                            target_data_domain, false);
      
      SymbExpression right_e=calc_expression(right,
                                             target_data_domain, false);
      
      if (guard_helper::t_guard_term::getOp(expression_node)=="*") {
        SymbFunction_uint_mult uint_mult;
	SymbFunctionArgumentVector arguments(2);
	arguments[0]=SymbFunctionArgument(SymbExpression::ptr(new SymbExpression(left_e)));
	arguments[1]=SymbFunctionArgument(SymbExpression::ptr(new SymbExpression(right_e)));
        
        SymbExpression result=*uint_mult.calculate(arguments);
        result.setCondition( result.getCondition() & left_e.getCondition() & right_e.getCondition() );
        return result;
      } else if (guard_helper::t_guard_term::getOp(expression_node)=="/") {
        SymbFunction_uint_div uint_div;
	SymbFunctionArgumentVector arguments(2);
	arguments[0]=SymbFunctionArgument(SymbExpression::ptr(new SymbExpression(left_e)));
	arguments[1]=SymbFunctionArgument(SymbExpression::ptr(new SymbExpression(right_e)));
        
        SymbExpression result=*uint_div.calculate(arguments);
        result.setCondition( result.getCondition() & left_e.getCondition() & right_e.getCondition() );
        return result;
      } else if (guard_helper::t_guard_term::getOp(expression_node)=="%") {
        SymbFunction_uint_mod uint_mod;
	SymbFunctionArgumentVector arguments(2);
	arguments[0]=SymbFunctionArgument(SymbExpression::ptr(new SymbExpression(left_e)));
	arguments[1]=SymbFunctionArgument(SymbExpression::ptr(new SymbExpression(right_e)));
        
        SymbExpression result=*uint_mod.calculate(arguments);
        result.setCondition( result.getCondition() & left_e.getCondition() & right_e.getCondition() );
        return result;
      }

      THROW_UNIMPLEMENTED();
    } else {
      return calc_expression(guard_helper::t_guard_term::getLeft(expression_node),
                             target_data_domain, strict_target);
    }
  }
  case ID_G_guard_factor:
    if (guard_helper::t_guard_factor::isNegation(expression_node)) {
      // TODO: handle negation
      THROW_UNIMPLEMENTED();
    }
    // handle other cases
    return calc_expression(guard_helper::t_guard_factor::getChild(expression_node),
                           target_data_domain, strict_target);
  case ID_INT_integer: {
    int i=int_helper::t_integer::getInt(expression_node);
    return get_int_expression(i, target_data_domain, strict_target);
  }
  case ID_INT_const_ref_int: {
    int i=int_helper::t_const_ref_int::evaluate(expression_node, info->getComponentInterface()->getSymbolTableStack()).getValue();
    return get_int_expression(i, target_data_domain, strict_target);
  }
  case ID_G_state_var: {
    return accessStateVar(expression_node);
  }
  case ID_G_virtual_value: {
    SymbolData::ptr sd = info->getComponentInterface()->getSymbolTableStack().find(guard_helper::t_virtual_value::getIdentifier(expression_node));
    CHECK_SUCCESS(sd);
    CHECK_SUCCESS(sd->type == SYMBOL_TYPE_VIRTUAL_VALUE);
    
    VirtualValueData& vvd = static_cast<VirtualValueData&>(*sd);
    CHECK_SUCCESS(vvd.getExpression());
    
    if (guard_helper::t_virtual_value::hasStructMemberAccess(expression_node)) {
      if (vvd.getDataDomain()->getType()!=DataDomain::DD_STRUCT &&
          vvd.getDataDomain()->getType()!=DataDomain::DD_TAGGED_UNION) {
        throw "Trying to access member, but data domain is not structured ("+vvd.getDataDomain()->toString()+")";
      }

      return accessMember(guard_helper::t_virtual_value::getStructMemberAccess(expression_node),
                          vvd.getExpression()->getValues(),
                          vvd.getExpression()->getDataDomain());
    } else if (guard_helper::t_virtual_value::hasArrayAccess(expression_node)) {
      if (vvd.getDataDomain()->getType()!=DataDomain::DD_ARRAY) {
        throw "Trying to access member, but data domain is not array ("+vvd.getDataDomain()->toString()+")";
      }

      return accessArray(guard_helper::t_virtual_value::getArrayAccess(expression_node),
                         vvd.getExpression()->getValues(),
                         vvd.getExpression()->getDataDomain());
    } else {
        return *vvd.getExpression();
    }
  }    
  case ID_G_port_data: {
    Symb_DataDomain::ptr data_domain;
    BDDFunctionContainer encoding;
    BDDFunction condition=BDDFunction::ONE();
    
    PortDescription::ptr port=
      get_port_description(guard_helper::t_port_data::getPortName(expression_node, 
                                                                  info->c_interface->getSymbolTableStack()));
    
    if (guard_helper::t_port_data::hasStructMemberAccess(expression_node)) {
      if (dataDomain->getType()!=DataDomain::DD_STRUCT &&
          dataDomain->getType()!=DataDomain::DD_TAGGED_UNION) {
        throw "Trying to access member of port data, but data domain is not structured";
      }

      SymbExpression e=accessMember(guard_helper::t_port_data::getStructMemberAccess(expression_node),
                                    get_data_vars_for_port(port),
                                    dataDomain);

      data_domain=e.getDataDomain();
      encoding=e.getValues();
      condition=e.getCondition();
    } else {
      data_domain=dataDomain;
      encoding=get_data_vars_for_port(port);
    }
    
    return SymbExpression(data_domain, 
                          encoding,
                          condition & port_active(get_symb_ce_for_port(port)));
  }
  case ID_I_enum_value: {
    if (identifier_helper::t_enum_value::isFullyQualifiedName(expression_node)) {
      string enum_type=identifier_helper::t_enum_value::getTypePart(expression_node);
      string element=identifier_helper::t_enum_value::getNamePart(expression_node);
      
      DataDomain::ptr enum_dd=findType(enum_type);
      
      if (!enum_dd ||
          enum_dd->getType()!=DataDomain::DD_FLAT) {
        throw string("For ")+enum_type+"::"+element+", "+enum_type+" is not a valid type";
      }
      
      if (!static_cast<FlatDataDomain&>(*enum_dd).isValidValue(element)) {
        throw string("For type ")+enum_type+", "+element+" is not a valid element";
      }

      Symb_FlatDataDomain::ptr s_enum_dd=
        boost::static_pointer_cast<Symb_FlatDataDomain>(Symb_DataDomain::getFor(enum_dd));
      
      BDDFunctionContainer e_encoding=s_enum_dd->getEncoding(element);
      return SymbExpression(s_enum_dd, e_encoding);
    } else {
      string element=identifier_helper::t_enum_value::getNamePart(expression_node);
      if (!target_data_domain ||
          target_data_domain->getType()!=DataDomain::DD_FLAT ||
          !boost::static_pointer_cast<Symb_FlatDataDomain>(target_data_domain)->getFlatDataDomain()->isValidValue(element)) {
        throw "Enum element "+element+" is not in lhs data domain!";
        // return false
        return SymbExpression(target_data_domain, target_data_domain->getNULL(), BDDFunction::ZERO());
      }
      Symb_FlatDataDomain::ptr s_enum_dd=boost::static_pointer_cast<Symb_FlatDataDomain>(target_data_domain);
      
      BDDFunctionContainer e_encoding=s_enum_dd->getEncoding(element);
      return SymbExpression(s_enum_dd, e_encoding);
    }
  }
  case ID_G_function_call: 
    return handle_function_call(expression_node, target_data_domain, strict_target);
  case ID_G_ite_expression:
    return handle_ite_expression(expression_node, target_data_domain, strict_target);
  case ID_G_iterate_expression:
    return handle_iterate_expression(expression_node);
  default:
    throw "implementation error";
  }
}

PortDescription::ptr SymbComponentInstance::get_port_description(const guard_helper::port_name_t& port_name) const {
  CHECK_SUCCESS(info->component_instance);
  PortID port_id = info->component_instance->lookupPort(port_name.toString(), false);

  CHECK_SUCCESS(info->component_instance.get() == &port_id.getComponentInstance());
  return port_id.getPortDescription().shared_from_this();
}

Symb_Channel_End::ptr SymbComponentInstance::get_symb_ce_for_port(PortDescription::ptr pd) const {
 port_to_bdd_map::const_iterator it=info->port_to_bdd.find(pd);
  if (it==info->port_to_bdd.end()) {
    throw string("port_to_bdd map: No port "+pd->getName());
  }

  return it->second;
}

Symb_Channel_End::ptr SymbComponentInstance::find_shared_symb_ce_for_port(const PortID& port_id) const {
  if (build_context &&
      (!build_context->param_no_shared_data) &&
      build_context->port_partitions) {
    
    PortID port_id_r=build_context->port_partitions->identical_ports.getRepresentative(port_id);
    TRACE("SymbComponentInstance.find_shared_symb_ce_for_port", "search", \
          std::cout << "Search shared for " << port_id << " [" << port_id_r << "]\n");

    const set<PortID>& related_ports=build_context->port_partitions->if_active_same_data_ports.get_related(port_id_r);
    for (set<PortID>::const_iterator it=related_ports.begin();
         it!=related_ports.end();
         ++it) {
      const PortID& candidate_portid=*it;
      PortID candidate_port_id_r=build_context->port_partitions->identical_ports.getRepresentative(candidate_portid);
      TRACE("SymbComponentInstance.find_shared_symb_ce_for_port", "search", \
            std::cout << "Candidate: " << candidate_portid << " [" << candidate_port_id_r << "]\n");
      
      Symb_Channel_End::ptr candidate_symb_ce=find_identical_symb_ce_for_port(candidate_port_id_r);

      if (!candidate_symb_ce) {
        TRACE("SymbComponentInstance.find_shared_symb_ce_for_port", "search", \
              std::cout << "  -> No Symb_ce\n");
        continue;
      }

      if (!candidate_symb_ce->mayShareDataVariables()) {
        TRACE("SymbComponentInstance.find_shared_symb_ce_for_port", "search", \
              std::cout << "  -> Symb_ce, but may not share variables\n");
        continue;
      }

      if (candidate_symb_ce->areDataVariablesShared()) {
        // we have to check that all shared are also ok for us
        const set<PortID>& related_ports_of_candidate=
          build_context->port_partitions->if_active_same_data_ports.get_related(candidate_port_id_r);

        TRACE("SymbComponentInstance.find_shared_symb_ce_for_port", "check_shared", \
              std::cout << " -> Symb_ce has shared data variables, check\n");

        bool ok=true;
        // if every related port of the candidate either has no shared symb ce with the candidate
        // or if it has a shared symb_ce we are also related to that port, we are fine...
        for (set<PortID>::const_iterator it=related_ports_of_candidate.begin();
             it!=related_ports_of_candidate.end();
             ++it) {
          TRACE("SymbComponentInstance.find_shared_symb_ce_for_port", "check_shared", \
                std::cout << "  Check " << *it << "\n");

          Symb_Channel_End::ptr symb_ce=find_identical_symb_ce_for_port(*it);
          if (!symb_ce ||                                              // no symb_ce yet
              !symb_ce->areDataVariablesShared() ||                    // symb_ce not shared
              !symb_ce->areDataVariablesSharedWith(candidate_symb_ce)  // not shared with our candidate symb_ce
              ) {
            // ok, this port does not pose a problem

            TRACE("SymbComponentInstance.find_shared_symb_ce_for_port", "check_shared", \
                  std::cout << "  " << *it << " is no problem\n");
          
            continue;
          }

          // this port is a problem iff we are not related to it
          if (build_context->port_partitions->if_active_same_data_ports.are_related(port_id_r, *it)) {
            // ok, we are related, everything ok
            TRACE("SymbComponentInstance.find_shared_symb_ce_for_port", "check_shared", \
                  std::cout << "  " << *it << " is no problem, we are also related\n");

            continue;
          } else {
            TRACE("SymbComponentInstance.find_shared_symb_ce_for_port", "check_shared", \
                  std::cout << "  FAIL: " << *it << " is not related to us\n");

            ok=false;
            break;
          }
        }        

        if (!ok) continue;
      } else {
        TRACE("SymbComponentInstance.find_shared_symb_ce_for_port", "search", \
              std::cout << "  -> Symb_ce, not shared yet\n");
      }      

      // ok, we can share
      TRACE("SymbComponentInstance.find_shared_symb_ce_for_port", "found", \
            std::cout << "  ==> Share " << port_id << " and " << candidate_port_id_r << "\n");

      Symb_Channel_End::ptr result=candidate_symb_ce->getSharedDataChannelEnd();

      TRACE_BDD_FUNCTION("SymbComponentInstance.find_shared_symb_ce_for_port", candidate_symb_ce->getPortActive(), "");
      TRACE_BDD_FUNCTION("SymbComponentInstance.find_shared_symb_ce_for_port", result->getPortActive(), "");

      CHECK(1, result->areDataVariablesSharedWith(candidate_symb_ce), "");
      CHECK(1, candidate_symb_ce->areDataVariablesSharedWith(result), "");

      return result;
    }
  }

  return Symb_Channel_End::ptr();
}

Symb_Channel_End::ptr SymbComponentInstance::find_identical_symb_ce_for_port(const PortID& port_id) const {
  if (build_context &&
      build_context->port_partitions) {
    TRACE("SymbComponentInstance.find_identical_symb_ce_for_port", "search", \
          std::cout << "Search identical for " << port_id << "\n");
    PortID identical_class_id=build_context->port_partitions->identical_ports.getRepresentative(port_id);
    if (build_context->port_id_to_symb_ce->has_symb_ce(identical_class_id)) {
      return build_context->port_id_to_symb_ce->get(identical_class_id);
    }
  }  
    
  return Symb_Channel_End::ptr();
}

Symb_Channel_End::ptr SymbComponentInstance::find_symb_ce_for_port(PortDescription::ptr pd) const {
  port_to_bdd_map::const_iterator it=info->port_to_bdd.find(pd);
  if (it!=info->port_to_bdd.end()) {
    return it->second;
  } else {
      // no symb_ce yet, try to find equivalent port

    if (info->component_instance) {
      PortID port_id(info->component_instance, pd);
      
      Symb_Channel_End::ptr result;

      if (build_context->port_id_to_deserialized_symb_ce) {
        result=build_context->port_id_to_deserialized_symb_ce->get(port_id);
        
        if (result) {
          TRACE("SymbComponentInstance", "deserialized", 
                std::cout << "Found deserialized SymbCE for port " << port_id << "\n");
        }
      }

      if (!result) {
        result=find_identical_symb_ce_for_port(port_id);
      }

      if (!result) {
        result=find_shared_symb_ce_for_port(port_id);
      }
      return result;
    }
  }

  return Symb_Channel_End::ptr();
}



// register a mapping between PortID(component_instance, port) and ce
// internally used for collaborative pd->ce mapping (with Symb_Node)
// the symb_ce does not automatically become an active ce in this automaton,
// for that, use addPort()
void SymbComponentInstance::register_symb_ce_for_port(PortDescription::ptr port,
                                                      Symb_Channel_End::ptr symb_ce) const {
  if (info->component_instance) {
    // if we are a valid component, add information to global PortID2SymbCEMap
    PortID port_id(info->component_instance, port);
    build_context->port_id_to_symb_ce->add(port_id, symb_ce);

    if (build_context &&
        build_context->port_partitions) {
      PortID identical_class_id=build_context->port_partitions->identical_ports.getRepresentative(port_id);
      // store the symb_ce also for the whole equivalence class (i.e. the representative)
      // don't overwrite earlier assignment
      if (!build_context->port_id_to_symb_ce->has_symb_ce(identical_class_id)) {
        build_context->port_id_to_symb_ce->add(identical_class_id, symb_ce);
      }
    }
  }
}

BDDFunction SymbComponentInstance::port_active(Symb_Channel_End::ptr ce, bool active) const {
  if (active) {
    return ce->getPortActive();
  } else {
    return !ce->getPortActive();
  }
}

const BDDFunctionContainer& SymbComponentInstance::get_data_vars_for_port(PortDescription::ptr pd) const {
  Symb_Channel_End::ptr ce=get_symb_ce_for_port(pd);
  return ce->getDCVars();
}


/**
 * Calculates port activity functions for the top-level ports.
 * The first part of the tuple corresponds to the activity where all the ports from port_set are active.
 * The second part of the tuple corresponds to the activity where all the top-level ports *not* in port_set are active.
 * Both parts may overlap due to identification of ports.
 */
std::pair<BDDFunction,BDDFunction>
 SymbComponentInstance::toplevel_port_activity(const PortDescription::ptr_set& port_set) const {
  BDDFunction activity_ports=BDDFunction::ONE();
  BDDFunction activity_complement=BDDFunction::ONE();
  
  for (port_to_bdd_map::const_iterator it_all_ports=info->port_to_bdd.begin();
       it_all_ports!=info->port_to_bdd.end();
       ++it_all_ports) {
    PortDescription::ptr pd=it_all_ports->first;
    bool in_port_set= (port_set.count(pd)!=0);

    Symb_Channel_End::ptr ce=it_all_ports->second;

    TRACE("SymbComponentInstance", "toplevel_port_activity", \
	  cout << it_all_ports->first->getName(););
    if (!ce->isVisible() || pd->isAnonymous()) {
      // skip
      if (in_port_set) {
        throw "Port "+pd->getName()+" used in port set, is not visible!";
      }
      
      TRACE("SymbComponentInstance", "toplevel_port_activity", \
	    cout << " [skipped]\n";);
      continue;
    }

    if (in_port_set) {
      TRACE("SymbComponentInstance", "toplevel_port_activity", \
	    cout << " [in_port_set]\n";);
      activity_ports&=ce->getPortActive();      
    } else {
      TRACE("SymbComponentInstance", "toplevel_port_activity", \
	    cout << " [!in_port_set]\n";);
      activity_complement&=ce->getPortActive();
    }
  }
  
  return make_pair(activity_ports, activity_complement);
}


BDDFunction SymbComponentInstance::only_ports_active(const PortDescription::ptr_set& active_ports) const {
  pair<BDDFunction, BDDFunction> activity = toplevel_port_activity(active_ports);

  return activity.first & cubeSetNone(activity.second);
}

void SymbComponentInstance::State_Render_Personality_XML::start(ostream& out) {
  xml_tag_open(out, "statevector");
}

void SymbComponentInstance::State_Render_Personality_XML::stop(ostream& out) {
  xml_tag_close(out, "statevector");
}
 
void SymbComponentInstance::State_Render_Personality_XML::start_subcomponent(ostream& out, const std::string& name) {
  xml_tag_open(out, "component", xml_parameter("name", name));
}

void SymbComponentInstance::State_Render_Personality_XML::stop_subcomponent(ostream& out) {
  xml_tag_close(out, "component");
}

void SymbComponentInstance::State_Render_Personality_XML::state_value(ostream& out, 
                                                                      const BDDFunction& assignment, 
                                                                      const string& name, 
                                                                      SymbolicState::ptr ss, 
                                                                      bool use_x) {
  xml_tag_open(out, "state", xml_parameter("name", name));
  ss->data_domain->reverseEncodingToXML(out,
                                        assignment,
                                        (use_x ? 
                                         *(ss->vars.X) : 
                                         *(ss->vars.Y)));
  xml_tag_close(out, "state");
}



void SymbComponentInstance::State_Render_Personality_Text::start(ostream& out) {
  first=true;
  update_current_prefix();
}
 
void SymbComponentInstance::State_Render_Personality_Text::start_subcomponent(ostream& out, const std::string& name) {
  prefixes.push_back(name+".");
  update_current_prefix();

  if (!first) 
    out << " ";
  first=true;
}

void SymbComponentInstance::State_Render_Personality_Text::stop_subcomponent(ostream& out) {
  CHECK_SUCCESS(!prefixes.empty());
  prefixes.pop_back();
  update_current_prefix();
}

void SymbComponentInstance::State_Render_Personality_Text::state_value(ostream& out, 
                                                                       const BDDFunction& assignment, 
                                                                       const string& name, 
                                                                       SymbolicState::ptr ss, 
                                                                       bool use_x) {
  if (!first) {
    out << ",";
  }
  first=false;

  out << current_prefix << name << "=";
  out << ss->data_domain->reverseEncoding(assignment, 
					  (use_x ? 
					   *(ss->vars.X) : 
					   *(ss->vars.Y)));
}

void SymbComponentInstance::State_Render_Personality_Text::update_current_prefix() {
  stringstream ss;
  for (unsigned int i=0;i<prefixes.size();i++) {
    ss << prefixes[i];
  }
  current_prefix=ss.str();
}



void SymbComponentInstance::State_Render_Personality_Guards::start(ostream& out) {
  update_current_prefix();
  first=true;
}

void SymbComponentInstance::State_Render_Personality_Guards::stop(ostream& out) {
  if (!store_in_restriction_manager) {
    out << "\n";
  }
}
 
void SymbComponentInstance::State_Render_Personality_Guards::start_subcomponent(ostream& out, const std::string& name) {
  if (prefixes.size()==0) {
    prefixes.push_back(name);
  } else {
    prefixes.push_back("."+name);
  }
  update_current_prefix();
}

void SymbComponentInstance::State_Render_Personality_Guards::stop_subcomponent(ostream& out) {
  CHECK_SUCCESS(!prefixes.empty());
  prefixes.pop_back();
  update_current_prefix();
}

void SymbComponentInstance::State_Render_Personality_Guards::state_value(ostream& out, 
                                                                         const BDDFunction& assignment, 
                                                                         const string& name, 
                                                                         SymbolicState::ptr ss, 
                                                                         bool use_x) {
  vector<string> results;

  ss->data_domain->reverseEncodingToGuards(assignment, 
                                           (use_x ? 
                                            *(ss->vars.X) : 
                                            *(ss->vars.Y)),
                                           name,
                                           results);

  for (unsigned int i=0;i<results.size();i++) {
    if (!store_in_restriction_manager) {
      if (!first) {
        out << " & ";
      } else {
        first=false;
      }
      out << "\"" << current_prefix << (current_prefix=="" ? "" : ".") << results.at(i) << "\"";
    } else {
      restriction_manager->addStateRestriction(current_prefix, results.at(i));
    }
  }
}

void SymbComponentInstance::State_Render_Personality_Guards::update_current_prefix() {
  stringstream ss;
  for (unsigned int i=0;i<prefixes.size();i++) {
    ss << prefixes[i];
  }
  current_prefix=ss.str();
}


void SymbComponentInstance::Component_State_Renderer::render_states(ostream& out,
                                                                    const SymbComponentInfo& info,
								    const BDDFunction& assignment,
								    bool use_x) const {
  
  for (state_to_bdd_map::const_iterator it=info.state_to_bdd.begin();
       it!=info.state_to_bdd.end();
       ++it) {
    const StateDescription::ptr sd=it->first;
    const SymbolicState::ptr ss=it->second;
    
    personality->state_value(out, assignment, sd->getName(), ss, use_x);
  }
  
  for (subcomponent_info_map::const_iterator it_sub=info.subcomponent_infos_begin();
       it_sub!=info.subcomponent_infos_end();
       ++it_sub) {
    const string& subcomponent_name=it_sub->first;

    SymbComponentInfo& sub_info=*it_sub->second;
    personality->start_subcomponent(out, subcomponent_name);
    render_states(out, sub_info, assignment, use_x);
    personality->stop_subcomponent(out);
  }
}

void SymbComponentInstance::Component_State_Renderer::render(ostream& out,
							     const BDDFunction& assignment,
							     bool use_x) const {
  const SymbComponentInstance& sci=static_cast<const SymbComponentInstance&>(automaton);

  personality->start(out);
  render_states(out, *sci.info, assignment, use_x);
  personality->stop(out);
}


void SymbComponentInstance::IO_Render_Personality_Text::is_stopped(ostream& out) {
  out << "stopped";
}

void SymbComponentInstance::IO_Render_Personality_Text::start(ostream& out) {
  update_current_prefix();

  first=true;
  out << "{";
}

void SymbComponentInstance::IO_Render_Personality_Text::stop(ostream& out) {
  out << "}";
}

void SymbComponentInstance::IO_Render_Personality_Text::start_subcomponent(ostream& out, const std::string& name) {
  prefixes.push_back(name+".");
  update_current_prefix();
}

void SymbComponentInstance::IO_Render_Personality_Text::stop_subcomponent(ostream& out) {
  CHECK_SUCCESS(!prefixes.empty());
  prefixes.pop_back();
  update_current_prefix();
}


void SymbComponentInstance::IO_Render_Personality_Text::update_current_prefix() {
  stringstream ss;
  for (unsigned int i=0;i<prefixes.size();i++) {
    ss << prefixes[i];
  }
  current_prefix=ss.str();
}

void SymbComponentInstance::IO_Render_Personality_Text::port_active(ostream& out,
                                                                    const BDDFunction& assignment,
                                                                    PortDescription::ptr pd,
                                                                    Symb_Channel_End::ptr ce,
                                                                    bool output_data) {
  if (!first) {
    out << ",";
  }
  first=false;

  out << current_prefix << pd->getName();
  if (output_data) {
    out << "=";
    out << ce->getDataDomain()->reverseEncoding(assignment, ce->getDCVars());
  }  
}


void SymbComponentInstance::IO_Render_Personality_Text::label_active(ostream& out,
                                                                     const BDDFunction& assignment,
                                                                     Symb_Transition_Label_Var::ptr symb_tl) {
  if (!first) {
    out << ",";
  }
  first=false;

  out << "label(" << current_prefix << access_weak_ptr(symb_tl->getTransitionLabelDescription())->getName() << ")";
  if (symb_tl->hasValueDomain()) {
    out << "=";
    out << symb_tl->getDataDomain()->reverseEncoding(assignment, symb_tl->getValueVars());
  }  
}

void SymbComponentInstance::IO_Render_Personality_Text::label_inactive(ostream& out,
                                                                       Symb_Transition_Label_Var::ptr symb_tl) {
  // nothing to do
}

void SymbComponentInstance::IO_Render_Personality_XML::is_stopped(ostream& out) {
  xml_tag_simple(out, "stopped");
}

void SymbComponentInstance::IO_Render_Personality_XML::start(ostream& out) {
  xml_tag_open(out, "portvector");
}

void SymbComponentInstance::IO_Render_Personality_XML::stop(ostream& out) {
  xml_tag_close(out, "portvector");
}

void SymbComponentInstance::IO_Render_Personality_XML::start_subcomponent(ostream& out, const std::string& name) {
  xml_tag_open(out, "component", xml_parameter("name", name));
}

void SymbComponentInstance::IO_Render_Personality_XML::stop_subcomponent(ostream& out) {
  xml_tag_close(out, "component");
}

void SymbComponentInstance::IO_Render_Personality_XML::port_active(ostream& out,
                                                                   const BDDFunction& assignment,
                                                                   PortDescription::ptr pd,
                                                                   Symb_Channel_End::ptr ce,
                                                                   bool output_data) {
  xml_tag_open(out, "port", xml_parameter("name", pd->getName())+xml_parameter("active","1"));
  if (output_data) {
    ce->getDataDomain()->reverseEncodingToXML(out, assignment, ce->getDCVars());
  }
  xml_tag_close(out, "port");
}

void SymbComponentInstance::IO_Render_Personality_XML::port_inactive(ostream& out,
                                                                     PortDescription::ptr pd) {
  xml_tag_simple(out, "port", xml_parameter("name", pd->getName())+xml_parameter("active","0"));
}                                                                     


void SymbComponentInstance::IO_Render_Personality_XML::label_active(ostream& out,
                                                                    const BDDFunction& assignment,
                                                                    Symb_Transition_Label_Var::ptr symb_tl) {                                                                 xml_tag_open(out, "label", xml_parameter("name", access_weak_ptr(symb_tl->getTransitionLabelDescription())->getName())+xml_parameter("active","1"));
  if (symb_tl->hasValueDomain()) {
    symb_tl->getDataDomain()->reverseEncodingToXML(out, assignment, symb_tl->getValueVars());
  }
  xml_tag_close(out, "label");
}

void SymbComponentInstance::IO_Render_Personality_XML::label_inactive(ostream& out,
                                                                      Symb_Transition_Label_Var::ptr symb_tl) {
  xml_tag_open(out, "label", xml_parameter("name", access_weak_ptr(symb_tl->getTransitionLabelDescription())->getName())+xml_parameter("active","0"));
  xml_tag_close(out, "label");
}


void SymbComponentInstance::Component_IO_Renderer::render_io(ostream& out,
                                                             const SymbComponentInfo& info,
                                                             const BDDFunction& assignment,
                                                             unsigned int level) const {

  if (!only_toplevel_ports || level==0) {
    for (port_to_bdd_map::const_iterator it=info.port_to_bdd.begin();
         it!=info.port_to_bdd.end();
         ++it) {
      const PortDescription::ptr pd=it->first;
      const Symb_Channel_End::ptr ce=it->second;
      
      if (ce->isVisible() || ce->isSemiHidden()) {
        // ok
      } else {
        personality->port_hidden(out, pd);
        continue;
      }
      
      TRACE_BDD_FUNCTION("IO_render", ce->getPortActive(), pd->getName());
      bool port_is_active=bdd_is_bit_set(assignment, ce->getPortActive());
      
      if (pd->getName()=="") {
        throw "Unnamed channel end (io_string)";
      }
      
      if (pd->getName().at(0)=='$') {
        // anonymous
        personality->port_hidden(out, pd);
        continue;
      }
      
      bool output_data = !info.abstract_port_data_domain;
      if (port_is_active) {
        personality->port_active(out, assignment, pd, ce, output_data);
      } else {
        personality->port_inactive(out, pd);
      }
    }
  }
    
    
  if (show_labels) {
    for (transition_label_to_bdd_map::const_iterator it=info.transition_label_to_bdd.begin();
         it!=info.transition_label_to_bdd.end();
         ++it) {
      const TransitionLabelDescription::ptr tl=it->first;
      const Symb_Transition_Label_Var::ptr symb_tl=it->second;
        
      if (!automaton.hasTransitionLabel(symb_tl)) {
        // skip label if it is not contained in the Symb_Component_Instance
        continue;
      }

      TRACE_BDD_FUNCTION("IO_render", symb_tl->getHasLabelVar(), tl->getName());
      bool has_label=bdd_is_bit_set(assignment, symb_tl->getHasLabelVar());
        
      if (tl->getName()=="") {
        throw "Unnamed label (io_string)";
      }
        
      if (has_label) {
        personality->label_active(out, assignment, symb_tl);
      } else {
        personality->label_inactive(out, symb_tl);
      }
    }
  }

  if (only_toplevel_ports && !show_labels) 
    return;

  for (subcomponent_info_map::const_iterator it_sub=info.subcomponent_infos_begin();
       it_sub!=info.subcomponent_infos_end();
       ++it_sub) {
    const string& subcomponent_name=it_sub->first;

    SymbComponentInfo& sub_info=*it_sub->second;
    personality->start_subcomponent(out, subcomponent_name);
    render_io(out, sub_info, assignment, level+1);
    personality->stop_subcomponent(out);
  }
}

void SymbComponentInstance::Component_IO_Renderer::render(ostream& out, const BDDFunction& assignment, const BDDFunction& csVars) const {
  const SymbComponentInstance& sci=static_cast<const SymbComponentInstance&>(automaton);

  if (sci.getStoppedTransitionLabel()!=BDDFunction::ZERO() &&
      bdd_is_bit_set(assignment, sci.getStoppedTransitionLabel())) {
    personality->is_stopped(out);
    return;
  }

  personality->start(out);

  TRACE_BDD_FUNCTION("IO_render", assignment, "");

  render_io(out, *sci.info, assignment, 0);
  personality->stop(out);
}


SymbAutomaton::State_Renderer::ptr SymbComponentInstance::getDefaultStateRenderer() const {
  return SymbAutomaton::State_Renderer::ptr(new Component_State_Renderer(*this));
}
  
SymbAutomaton::IO_Renderer::ptr SymbComponentInstance::getDefaultIORenderer() const {
  return SymbAutomaton::IO_Renderer::ptr(new Component_IO_Renderer(*this, show_only_toplevel_ports));
}

void SymbComponentInstance::toDot(ostream& out) const {
  toDot(out, false);
}

void SymbComponentInstance::toDot(ostream& out, bool using_interface) const {
  if (using_interface) {
    SymbAutomaton::toDot(out, Component_State_Renderer(*this), Component_IO_Renderer(*this, show_only_toplevel_ports));
  } else {
    SymbAutomaton::toDot(out, *SymbAutomaton::getDefaultStateRenderer(), *SymbAutomaton::getDefaultIORenderer());
  }
}

void SymbComponentInstance::print(std::ostream& out, const SymbTrace& trace) const {
  SymbAutomaton::print(out, trace, Component_State_Renderer(*this), Component_IO_Renderer(*this, show_only_toplevel_ports));
}

void SymbComponentInstance::print_xml(std::ostream& out, const SymbTrace& trace) const {
    for (unsigned int i=0;i<trace.countStates();i++) {
      print_xml(out, trace.getState(i));
      out << "\n";
      
      if (i<trace.countLabels()) {
        print_xml(out, trace.getLabel(i));
        out << "\n";
      }
    }
}



void SymbComponentInstance::print(std::ostream& out, const SymbState& state) const {
  SymbAutomaton::print(out, state, Component_State_Renderer(*this));
}

void SymbComponentInstance::print_xml(std::ostream& out, const SymbState& state) const {
  SymbAutomaton::print(out, state, Component_State_Renderer(*this,
                                                            State_Render_Personality::ptr(new State_Render_Personality_XML())));
  
}


void SymbComponentInstance::print(std::ostream& out, const SymbTransitionLabel& label) const {
  SymbAutomaton::print(out, 
                       label, 
                       Component_IO_Renderer(*this, 
                                             show_only_toplevel_ports,
                                             true // show labels
                                             ));
}

void SymbComponentInstance::print_xml(std::ostream& out, const SymbTransitionLabel& label) const {
  SymbAutomaton::print(out, 
                       label, 
                       Component_IO_Renderer(*this, 
                                             false,
                                             true, // show labels
                                             IO_Render_Personality::ptr(new IO_Render_Personality_XML())));
}


void SymbComponentInstance::hideInternalPorts(BuildContext::ptr build_context) {
  if (getDefaultHideType() == SymbAutomaton::HIDE_TRANSITIVE_FULL) {
    // call hide_transitive_full to remove the epsilon transitions
    hide_transitive_full();
  }
  return;
}


BDDFunction SymbComponentInstance::parse_state_guard(const string& guard) {
  REOParser rp(info->component_instance->getSystemContext());
  vparser::pt_node tree=rp.parse_state_guard(guard, info->c_interface);
  BDDFunction result=bool_expression_to_bdd(tree);

  TRACE_BDD_FUNCTION("parse_state_guard", result, guard);
  return result;
}

BDDFunction SymbComponentInstance::parse_io_guard(const string& guard, 
                                                  bool allow_state_references,
                                                  bool stop_check) {
  REOParser rp(info->component_instance->getSystemContext());
  vparser::pt_node tree=rp.parse_io_guard(guard, info->c_interface, allow_state_references);
  BDDFunction result=bool_expression_to_bdd(tree);

  if (stop_check && getStoppedTransitionLabel()!=BDDFunction::ZERO()) {
    result&=!getStoppedTransitionLabel();
  }

  TRACE_BDD_FUNCTION("parse_io_guard", result, guard);
  return result;
}

BDDFunction SymbComponentInstance::parse_enabled_io_guard(const string& guard, bool allow_state_references, bool stop_check) {
  BDDFunction bdd_guard=parse_io_guard(guard, allow_state_references, stop_check);

  TRACE_BDD_FUNCTION("parse_enabled_io_guard", bdd_guard, guard);

  BDDFunction matching_transitions=getDelta() & bdd_guard;
  BDDFunction matching_states=
    matching_transitions.exists(getCopyStateVars() & getTransitionVars());

  TRACE_BDD_FUNCTION("parse_enabled_io_guard", matching_states, guard);

  return matching_states;
}


DataDomain::ptr SymbComponentInstance::findType(const string& name) const {
  if (info->c_interface->getLocalSymbols()) {
    SymbolData::ptr sd=info->c_interface->getLocalSymbols()->find(name);
    if (sd && sd->type==SYMBOL_TYPE_TYPEDEF) {
      return boost::static_pointer_cast<TypeDefData>(sd)->data_domain;
    }
  }

  if (info->c_interface->getGlobalSymbols()) {
    SymbolData::ptr sd=info->c_interface->getGlobalSymbols()->find(name);
    if (sd && sd->type==SYMBOL_TYPE_TYPEDEF) {
      return boost::static_pointer_cast<TypeDefData>(sd)->data_domain;
    }
  }
  
  return DataDomain::ptr();
}


void SymbComponentInstance::add_state_value_restriction(const BDDFunction& restriction) {
  state_value_restrictions.push_back(restriction);
}

void SymbComponentInstance::copy_state_value_restrictions(const SymbComponentInstance& other) {
  state_value_restrictions.insert(state_value_restrictions.end(),
                                  other.state_value_restrictions.begin(),
                                  other.state_value_restrictions.end());
}


void SymbComponentInstance::generateRestrictionsFromState(const SymbState& state, RestrictionManager& restriction_manager) const {
  std::stringstream out;
  SymbAutomaton::print(out, 
                       state, 
                       Component_State_Renderer(*this,
                                                State_Render_Personality::ptr(new State_Render_Personality_Guards(restriction_manager))));
}

// calculate not yet calculated atomic propositions specified in the component interface
void SymbComponentInstance::calculateAtomicPropositions() {
  ComponentInterface::ptr c_interface=info->getComponentInterface();

  for (unsigned int i=0;i<c_interface->getAtomicPropositionCount();i++) {
    const std::string& ap_name=c_interface->getAtomicPropositionName(i);

    if (info->hasAtomicProposition(ap_name)) {
      // skip, probably filled in an earlier step
      // TODO: distinguish between already handled earlier and multiple definitions
      continue;
    }
    
    // calculate the characteristic function
    BDDFunction ap_function;
    try {
      ap_function=parse_state_guard(c_interface->getAtomicProposition(i));
    } catch (const char *parse_error) {
      std::string ci_name;
      if (info->component_instance) {
        ci_name=info->component_instance->getFQName();
      }
      
      throw "During calculation of atomic proposition "+
        (ci_name!="" ? (ci_name+".") : "")+
        ap_name+": "+parse_error;
    } catch (REOParseException& err) {
      std::string ci_name;
      if (info->component_instance) {
        ci_name=info->component_instance->getFQName();
      }
      
      std::string e = "During calculation of atomic proposition "+
        (ci_name!="" ? (ci_name+".") : "")+
        ap_name+": "+err.what();
      throw e;
    }

    info->addAtomicProposition(ap_name, ap_function);
  }
}

bool SymbComponentInstance::build_automaton_with_serialized_varorder(BuildContext::ptr build_context) {
  if (!build_context->try_loading_serialized_varorder) {
    return false;
  }

  VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope
    
  if (SerializedVarorder::findAndReadVarorder(info->component_instance,
                                              build_context)) {
    // there was actually a varorder for this component instance
    BuildContext::ptr new_build_context(new BuildContext(*build_context)); // copy
    new_build_context->try_loading_serialized_varorder=false;  // inhibit further loading

    build_automaton(new_build_context);  // recursively call build_automaton
    return true;
  }

  return false;
}


// given a transition assignment, return all transitions that are equivalent 
// (have different data assignment for the inactive ports)
BDDFunction SymbComponentInstance::getEquivalentTransitions(const BDDFunction& transition) const {
  CHECK(3, BDDChecks::isAssignment(transition, getTransitionVars()), "");
  CHECK(5, check_symb_ce_sanity(), "");
  
  if (info->abstract_port_data_domain) {
    if (getStoppedTransitionLabel()!=BDDFunction::ZERO() &&
        bdd_is_bit_set(transition, getStoppedTransitionLabel())) {
      return getStoppedTransitionLabel();
    }
  
    BDDFunction result=transition;
    // as the data domain is abstracted away, just existential quantify all DC vars
    result=result.exists(getDataConstraintVars());

    return result;
  } else {
    return SymbAutomaton::getEquivalentTransitions(transition);
  }
}


// canonicalizeTransitions takes a transition_set and removes all but one for each set
// of equivalent transitions. E.g., for a transition with an inactive port, the (irrelevant) 
// data bits are all set to exactly one value (NULL)
// The resulting transition set is returned
BDDFunction SymbComponentInstance::canonicalizeTransitions(const BDDFunction& transition_set) const {
  CHECK(5, check_symb_ce_sanity(), "");
  
  // normal handling done by SymbAutomaton
  BDDFunction result=SymbAutomaton::canonicalizeTransitions(transition_set);
  
  // handle data abstract 
  if (info->abstract_port_data_domain) {
    for (symb_ce_set::const_iterator it=getSymbChannelEnds().begin();
         it!=getSymbChannelEnds().end();
         ++it) {
      const Symb_Channel_End& ce=**it;
      
      // all data values are NULL
      result&=ce.data_equal(ce.getDataDomain()->getNULL());
    }
  }
  
  return result;
}





// callback from ExtractExplicitAutomatonFunctor
void SymbComponentInstance::
     ExtractExplicitComponentInstanceFunctor::explicit_state(unsigned long state_id,
                                                             BDDFunction state_assignment,
                                                             bool is_initial,
                                                             bool is_terminal) {
  SymbComponentInstance::Component_State_Renderer csr(*sci);
  
  stringstream ss;
  csr.render(ss, state_assignment, true);
  
  std::vector<std::string> true_aps;
  extract_aps(*sci->info, state_assignment, "", true_aps);

  explicit_state(state_id, ss.str(), true_aps, is_initial, is_terminal);
}


// callback from ExtractExplicitAutomatonFunctor
void SymbComponentInstance::
     ExtractExplicitComponentInstanceFunctor::explicit_transition(unsigned long state_id_from,
                                                                  unsigned long state_id_to,
                                                                  BDDFunction transition_assignment) {
  SymbComponentInstance::Component_IO_Renderer ior(*sci, true, false);
  
  stringstream ss;
  ior.render(ss, transition_assignment, sci->getTransitionVars());
  
  std::vector<std::string> transition_labels;
  extract_labels(*sci->info, 
                 transition_assignment, 
                 "", 
                 transition_labels);
  
  explicit_transition(state_id_from,
                      state_id_to, 
                      ss.str(), 
                      transition_labels);
}



void SymbComponentInstance::
     ExtractExplicitComponentInstanceFunctor::extract_aps(SymbComponentInfo& info, 
                                                          BDDFunction state_assignment,
                                                          const std::string& prefix,
                                                          std::vector<std::string>& true_aps) {

  for (SymbComponentInfo::atomic_proposition_map::const_iterator it = info.atomic_propositions_begin();
       it != info.atomic_propositions_end();
       ++it) {
    const std::string& name = it->first;
    BDDFunction ap = it->second;

    if ( (ap & state_assignment) != BDDFunction::ZERO()) {
      std::string fq_name;
      if (prefix!="") {
        fq_name=prefix+"."+name;
      } else {
        fq_name=name;
      }

      true_aps.push_back(fq_name);
    }
  }
 
  for (subcomponent_info_map::const_iterator it_sub=info.subcomponent_infos_begin();
       it_sub!=info.subcomponent_infos_end();
       ++it_sub) {
    const string& subcomponent_name=it_sub->first;

    SymbComponentInfo& sub_info=*it_sub->second;
    std::string new_prefix;
    if (prefix!="") {
      new_prefix = prefix+"."+subcomponent_name;
    } else {
      new_prefix = subcomponent_name;
    }
    extract_aps(sub_info, state_assignment, new_prefix, true_aps);
  }
}


void SymbComponentInstance::
ExtractExplicitComponentInstanceFunctor::extract_labels(SymbComponentInfo& info, 
                                                        BDDFunction transition_assignment,
                                                        const std::string& prefix,
                                                        std::vector<std::string>& transition_labels) {


  for (transition_label_to_bdd_map::const_iterator it=info.transition_label_to_bdd.begin();
         it!=info.transition_label_to_bdd.end();
         ++it) {
    const TransitionLabelDescription::ptr tl=it->first;
    const Symb_Transition_Label_Var::ptr symb_tl=it->second;
    
    if (!sci->hasTransitionLabel(symb_tl)) {
      // skip label if it is not contained in the Symb_Component_Instance
      continue;
    }
    
    bool has_label=bdd_is_bit_set(transition_assignment, symb_tl->getHasLabelVar());
    
    if (tl->getName()=="") {
      throw "Unnamed label (extract_labels)";
    }
    
    if (has_label) {
      std::string fq_name;
      if (prefix!="") {
        fq_name=prefix+"."+tl->getName();
      } else {
        fq_name=tl->getName();
      }
      
      transition_labels.push_back(fq_name);
    }
  }
  
  for (subcomponent_info_map::const_iterator it_sub=info.subcomponent_infos_begin();
       it_sub!=info.subcomponent_infos_end();
       ++it_sub) {
    const string& subcomponent_name=it_sub->first;

    SymbComponentInfo& sub_info=*it_sub->second;
    std::string new_prefix;
    if (prefix!="") {
      new_prefix = prefix+"."+subcomponent_name;
    } else {
      new_prefix = subcomponent_name;
    }
    extract_labels(sub_info, transition_assignment, new_prefix, transition_labels);
  }
}
