/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef SYMB_COMPONENT_INSTANCE_HPP
#define SYMB_COMPONENT_INSTANCE_HPP

#include "vereofy/common/shared_ptr_map.hpp"
#include "vereofy/setup.hpp"
#include "vereofy/system/symb_automaton/symb_automaton.hpp"
#include "vereofy/system/symb_automaton/symb_datadomain.hpp"
#include "vereofy/system/symb_automaton/symb_expression.hpp"
#include "vereofy/system/component_interface.hpp"
#include "vereofy/system/component_instance.hpp"
#include "vereofy/reo_parser/reo_parser_types.hpp"
#include "vereofy/reo_parser/guard_parser_helper.hpp"
#include "vereofy/system/symb_automaton/var_positions.hpp"
#include "vereofy/system/symb_automaton/build_context.hpp"

#include "vereofy/system/symb_automaton/symb_component_info.hpp"
#include "vereofy/system/restriction_manager.hpp"

#include "vereofy/common/tracing.hpp"

#define TRACE_SYMB_COMPONENT_DOT(scope, name, automaton, using_interface) \
  TRACE(scope, name,							\
  {									\
    string filename=string(scope)+"_"+name+".";				\
    filename+=boost::lexical_cast<string>(Tracing::getTraceCount(TRACE_GET_FULL_NAME(scope, name))); \
    filename+=".dot";							\
    cout << "Trace: Writing DOT file " << filename <<endl;		\
    ofstream out(filename.c_str());					\
    if (out) {								\
      (automaton).toDot(out, using_interface);				\
    } else {								\
      cout << "Couldn't write " << filename << endl;			\
    }									\
  }									\
	)

#define TRACE_SYMB_COMPONENT_PATHS(scope, name, automaton)		\
  TRACE_SYMB_AUTOMATON_PATHS(scope, name, automaton)

namespace vereofy {
  // forward
  class ComponentInstance;

  
  class SymbComponentInstance : public SymbAutomaton {
  public:
    
    typedef boost::shared_ptr<SymbComponentInstance> ptr;
    
    SymbComponentInfo::ptr info;
    bool show_only_toplevel_ports;
    
    SymbComponentInstance(ComponentInterface::ptr c_interface_,
			  Symb_DataDomain::ptr dataDomain);

    virtual void set_var_info(port_to_bdd_map port_map_init,
                              state_to_bdd_map state_map_init,
                              VarPositions::ptr var_positions_);

    void setShowOnlyToplevelPorts(bool enabled);

    virtual void build_automaton(BuildContext::ptr build_context) = 0;

    const SymbolicState::ptr getState(const string& name) const;
    const SymbolicState::ptr getState(StateDescription::ptr state_d) const;

    SymbComponentInstance::ptr getSubcomponentAutomaton(const string& name) const;
    SymbComponentInfo::ptr getSubcomponentInfo(const string& name) const;

    virtual void hide(const Automaton_Channel_End::ptr&);
    virtual void hide(const std::vector<Automaton_Channel_End::ptr>&);


    void restrictToValidValues();

    BDDFunctionContainer getPortRestrictions();
    // result.first = restrictions for initialStates
    // result.second = restrictions for delta
    std::pair<BDDFunctionContainer, BDDFunctionContainer> getStateRestrictions();

    void restrictToValidValuesPorts();
    void restrictToValidValuesStates();

    virtual BDDFunction parse_state_guard(const string& guard);
    virtual BDDFunction parse_io_guard(const string& guard, bool allow_state_references, bool stop_check);
    virtual BDDFunction parse_enabled_io_guard(const string& guard, bool allow_state_references, bool stop_check);


    struct State_Render_Personality {
      typedef boost::shared_ptr<State_Render_Personality> ptr;
      
      virtual ~State_Render_Personality() {}

      virtual void start(ostream& out) {}
      virtual void stop(ostream& out) {}
      virtual void start_subcomponent(ostream& out, const std::string& name) {};
      virtual void stop_subcomponent(ostream& out) {};
      virtual void state_value(ostream& out, 
                               const BDDFunction& assignment, 
                               const string& name, 
                               SymbolicState::ptr ss, 
                               bool use_x) {}
    };

    struct State_Render_Personality_Text : public State_Render_Personality {
      typedef boost::shared_ptr<State_Render_Personality_Text> ptr;
      
      std::vector<std::string> prefixes;
      std::string current_prefix;
      bool first;

      virtual void start(ostream& out);
      virtual void start_subcomponent(ostream& out, const std::string& name);
      virtual void stop_subcomponent(ostream& out);
      virtual void state_value(ostream& out, 
                               const BDDFunction& assignment, 
                               const string& name, 
                               SymbolicState::ptr ss, 
                               bool use_x);

    private:
      void update_current_prefix();
    };

    
    struct State_Render_Personality_XML : public State_Render_Personality {
      typedef boost::shared_ptr<State_Render_Personality_XML> ptr;
      
      virtual void start(ostream& out);
      virtual void stop(ostream& out);
      virtual void start_subcomponent(ostream& out, const std::string& name);
      virtual void stop_subcomponent(ostream& out);
      virtual void state_value(ostream& out, 
                               const BDDFunction& assignment, 
                               const string& name, 
                               SymbolicState::ptr ss, 
                               bool use_x);
    };


    struct State_Render_Personality_Guards : public State_Render_Personality {
      typedef boost::shared_ptr<State_Render_Personality_Guards> ptr;
      
      std::vector<std::string> prefixes;
      std::string current_prefix;
      bool first;

      bool store_in_restriction_manager;
      RestrictionManager* restriction_manager;

      State_Render_Personality_Guards() :
        store_in_restriction_manager(false), restriction_manager(0) {}

      State_Render_Personality_Guards(RestrictionManager& restriction_manager_) :
        store_in_restriction_manager(true), restriction_manager(&restriction_manager_) {}

      virtual void start(ostream& out);
      virtual void stop(ostream& out);
      virtual void start_subcomponent(ostream& out, const std::string& name);
      virtual void stop_subcomponent(ostream& out);
      virtual void state_value(ostream& out, 
                               const BDDFunction& assignment, 
                               const string& name, 
                               SymbolicState::ptr ss, 
                               bool use_x);

    private:
      void update_current_prefix();
    };


    struct Component_State_Renderer : public SymbAutomaton::State_Renderer {
      State_Render_Personality::ptr personality;

      Component_State_Renderer(const SymbComponentInstance& sci,
                               State_Render_Personality::ptr personality_=
                                 State_Render_Personality::ptr(new State_Render_Personality_Text)) 
        : SymbAutomaton::State_Renderer(sci), personality(personality_) {}
      
      virtual void render_states(ostream& out, const SymbComponentInfo& info, const BDDFunction& assignment, bool use_x) const;
      virtual void render(ostream& out, const BDDFunction& assignment, bool use_x) const;
    };



    struct IO_Render_Personality {
      typedef boost::shared_ptr<IO_Render_Personality> ptr;
      
      virtual ~IO_Render_Personality() {}

      virtual void is_stopped(ostream& out) {};   // called instead of start()...stop()
      virtual void start(ostream& out) {};
      virtual void stop(ostream& out) {};
      virtual void start_subcomponent(ostream& out, const std::string& name) {};
      virtual void stop_subcomponent(ostream& out) {};
      virtual void port_active(ostream& out,
                               const BDDFunction& assignment,
                               PortDescription::ptr pd,
                               Symb_Channel_End::ptr ce,
                               bool output_data) {};
      virtual void port_inactive(ostream& out,
                                 PortDescription::ptr pd) {};
      virtual void port_hidden(ostream& out,
                               PortDescription::ptr pd) {};
      virtual void label_active(ostream& out,
                                const BDDFunction& assignment,
                                Symb_Transition_Label_Var::ptr symb_tl) {};
      virtual void label_inactive(ostream& out,
                                  Symb_Transition_Label_Var::ptr symb_tl) {};
    };
    
    struct IO_Render_Personality_Text : public IO_Render_Personality {
      typedef boost::shared_ptr<IO_Render_Personality_Text> ptr;

      std::vector<std::string> prefixes;
      std::string current_prefix;

      bool first;

      virtual void is_stopped(ostream& out); 
      virtual void start(ostream& out);
      virtual void stop(ostream& out);
      virtual void start_subcomponent(ostream& out, const std::string& name);
      virtual void stop_subcomponent(ostream& out);
      virtual void port_active(ostream& out,
                               const BDDFunction& assignment,
                               PortDescription::ptr pd,
                               Symb_Channel_End::ptr ce,
                               bool output_data);
      virtual void label_active(ostream& out,
                                const BDDFunction& assignment,
                                Symb_Transition_Label_Var::ptr symb_tl);
      virtual void label_inactive(ostream& out,
                                Symb_Transition_Label_Var::ptr symb_tl);



    private:
      void update_current_prefix();
    };

    struct IO_Render_Personality_XML : public IO_Render_Personality {
      typedef boost::shared_ptr<IO_Render_Personality_XML> ptr;

      virtual void is_stopped(ostream& out); 
      virtual void start(ostream& out);
      virtual void stop(ostream& out);
      virtual void start_subcomponent(ostream& out, const std::string& name);
      virtual void stop_subcomponent(ostream& out);
      virtual void port_active(ostream& out,
                               const BDDFunction& assignment,
                               PortDescription::ptr pd,
                               Symb_Channel_End::ptr ce,
                               bool output_data);
      virtual void port_inactive(ostream& out,
                                 PortDescription::ptr pd);
      virtual void label_active(ostream& out,
                                const BDDFunction& assignment,
                                Symb_Transition_Label_Var::ptr symb_tl);
      virtual void label_inactive(ostream& out,
                                  Symb_Transition_Label_Var::ptr symb_tl);
    };

    
    struct Component_IO_Renderer : public SymbAutomaton::IO_Renderer {
      bool only_toplevel_ports;
      bool show_labels;
      IO_Render_Personality::ptr personality;

      Component_IO_Renderer(const SymbComponentInstance& sci,
                            bool only_toplevel_ports_=true,
                            bool show_labels_=true,
                            IO_Render_Personality::ptr personality_=IO_Render_Personality::ptr(new IO_Render_Personality_Text()))
        : SymbAutomaton::IO_Renderer(sci), 
          only_toplevel_ports(only_toplevel_ports_), 
          show_labels(show_labels_),
          personality(personality_) {}

      virtual void render_io(ostream& out, const SymbComponentInfo& info, const BDDFunction& assignment, unsigned int level) const;
      virtual void render(ostream& out, const BDDFunction& assignment, const BDDFunction& csVars) const;
    };

    virtual State_Renderer::ptr getDefaultStateRenderer() const;
    virtual IO_Renderer::ptr getDefaultIORenderer() const;

    virtual void toDot(ostream& out) const;
    virtual void toDot(ostream& out, bool using_interface) const;

    virtual void print(std::ostream& out, const SymbTrace& trace) const;
    virtual void print(std::ostream& out, const SymbState& state) const;
    virtual void print(std::ostream& out, const SymbTransitionLabel& label) const;

    virtual void print_xml(std::ostream& out, const SymbTrace& trace) const;
    virtual void print_xml(std::ostream& out, const SymbState& state) const;
    virtual void print_xml(std::ostream& out, const SymbTransitionLabel& label) const;


    struct ExtractExplicitComponentInstanceFunctor : public ExtractExplicitAutomatonFunctor {
      ExtractExplicitComponentInstanceFunctor(SymbComponentInstance::ptr sci_) :
        sci(sci_) {}

      virtual void start() {}
      virtual void finished() {}

      // to be implemented by derived classes
      // state_id: the ID of the state
      // state: a string representation of the state valuation
      // true_aps: a vector of the atomic propositions that are true in this state
      // is_initial: true if state is initial
      // is_terminal: true if state is terminal
      virtual void explicit_state(unsigned long state_id,
                                  const std::string& state,
                                  const std::vector<std::string>& true_aps,
                                  bool is_initial,
                                  bool is_terminal) = 0;

      // to be implemented by derived classes
      // state_id_from: the ID of the from state of this transition
      // state_id_to:   the ID of the to state of this transition
      // transition: a string representation of the transition (active ports and port data values)
      // transition_labels: a vector of the labels attached to the transition
      virtual void explicit_transition(unsigned long state_id_from,
                                       unsigned long state_id_to,
                                       const std::string& transition,
                                       const std::vector<std::string>& transition_labels) = 0;

      // callback from ExtractExplicitAutomatonFunctor
      virtual void explicit_state(unsigned long state_id,
                                  BDDFunction state_assignment,
                                  bool is_initial,
                                  bool is_terminal);

      // callback from ExtractExplicitAutomatonFunctor
      virtual void explicit_transition(unsigned long state_id_from,
                                       unsigned long state_id_to,
                                       BDDFunction transition_assignment);

      SymbComponentInstance::ptr sci;

    private:
      void extract_aps(SymbComponentInfo& info, 
                       BDDFunction state_assignment,
                       const std::string& prefix,
                       std::vector<std::string>& true_aps);
      void extract_labels(SymbComponentInfo& info, 
                          BDDFunction transition_assignment,
                          const std::string& prefix,
                          std::vector<std::string>& transition_labels);        
    };


    virtual void hideInternalPorts(BuildContext::ptr build_context);
	  
    virtual BDDFunction getReachabilityConstrain(void) const{ return reachabilityConstrain; };

    // get the registered Symb_Channel_End for the given port, throws exception if none is registered
    Symb_Channel_End::ptr get_symb_ce_for_port(PortDescription::ptr pd) const;

    // find an appropriate Symb_Channel_End for the given port, return NULL if none found
    Symb_Channel_End::ptr find_symb_ce_for_port(PortDescription::ptr pd) const;

    // register a symb_ce for the given port. only use internally (Symb_Node)
    void register_symb_ce_for_port(PortDescription::ptr pd,
                                   Symb_Channel_End::ptr ce) const;

    void generateRestrictionsFromState(const SymbState& state, RestrictionManager& restriction_manager) const;

    /**
     * Calculates port activity functions for the top-level ports.
     * The first part of the tuple corresponds to the activity where all the ports from port_set are active.
     * The second part of the tuple corresponds to the activity where all the top-level ports *not* in port_set are active.
     * Both parts may overlap due to identification of ports.
     */
    std::pair<BDDFunction,BDDFunction> toplevel_port_activity(const PortDescription::ptr_set& port_set) const;

  protected:
    Symb_Channel_End::ptr addPort(const PortDescription::ptr);
    Symb_Channel_End::ptr addPort(const PortDescription::ptr, const Symb_Channel_End::ptr);
    void addState(StateDescription::ptr);
    void addState(StateDescription::ptr, const SymbolicState::ptr);
    void addDummyState();

    Symb_Transition_Label_Var::ptr addTransitionLabel(TransitionLabelDescription::ptr);
    Symb_Transition_Label_Var::ptr addTransitionLabel(TransitionLabelDescription::ptr,
                                                      const Symb_Transition_Label_Var::ptr);

    void addSubcomponentAutomaton(SymbComponentInstance::ptr);
    void addSubcomponentInfo(SymbComponentInfo::ptr);
    void addSubcomponentRestrictions(SymbComponentInstance::ptr sub_component);

    void setComponentInstance(ComponentInstance::ptr);

    struct bdd_expression_vector {
      bdd_expression_vector(Symb_DataDomain::ptr data_domain_) :
	data_domain(data_domain_) {}

      bdd_expression_vector(const SymbExpression& expression) :
	data_domain(expression.getDataDomain()) {
	expression_vector.push_back(expression);
      }

      SymbExpression flatten() const;

      Symb_DataDomain::ptr data_domain;
      vector<SymbExpression> expression_vector;
    };

    struct bdd_assignment {
      BDDFunction assignment;
      BDDFunctionContainer targets;
    };

    struct bdd_assignment_target {
      explicit bdd_assignment_target() {}

      bdd_assignment_target(Symb_DataDomain::ptr data_domain_,
			    BDDFunctionContainer targets_,
			    BDDFunction condition_=BDDFunction::ONE()) :
	data_domain(data_domain_), targets(targets_), condition(condition_) {}

      Symb_DataDomain::ptr data_domain;
      BDDFunctionContainer targets;
      BDDFunction condition;

      std::list<bdd_assignment> co_assignments;
    };

    struct bdd_assignment_target_vector {
      bdd_assignment_target_vector(Symb_DataDomain::ptr data_domain_) :
	data_domain(data_domain_) {}

      bdd_assignment_target_vector(const bdd_assignment_target& target) :
	data_domain(target.data_domain) {
	target_vector.push_back(target);
      }

      Symb_DataDomain::ptr data_domain;
      vector<bdd_assignment_target> target_vector;
    };

    struct bdd_assignment_vector : public vector<bdd_assignment> {
      bdd_assignment_vector() {}

      BDDFunction merge() const {
	BDDFunction result=BDDFunction::ZERO();
	for (unsigned int i=0;i<size();i++) {
	  result|=at(i).assignment;
	}
	return result;
      }
    };
    
    SymbExpression accessStateVar(const vparser::pt_node& node_state_var) const;
    SymbExpression accessMember(const vparser::pt_node& node_struct_member_access,
                                const BDDFunctionContainer& struct_encoding,
                                Symb_DataDomain::ptr sdd) const;
    SymbExpression accessArray(const vparser::pt_node& node_array_access,
                               const BDDFunctionContainer& array_encoding,
                               Symb_DataDomain::ptr sdd) const;
    
    bdd_assignment_target_vector getAssignmentTargets(const vparser::pt_node& node_state_var,
						      bool assign_to_y=true) const;

    bdd_assignment_target getAssignmentTargetFlat(const string& name,
						  bool assign_to_y=true) const;
    bdd_assignment_target getAssignmentTargetMemberAccess(const BDDFunctionContainer& struct_encoding,
                                                          const vparser::pt_node& node_struct_member_access,
                                                          Symb_DataDomain::ptr sdd_) const;

    bdd_assignment_target getAssignmentTargetArray(const string& name,
						   unsigned int index,
						   bool assign_to_y=true) const;
	 
    BDDFunction state_guard_to_bdd(const vparser::pt_node& node);

    BDDFunction io_guard_to_bdd(const vparser::pt_node& node);

    BDDFunction bool_expression_to_bdd(const vparser::pt_node& node);


    BDDFunction no_state_change_bdd() const;


    SymbExpression get_int_expression(int i,
                                      Symb_DataDomain::ptr target_data_domain,
                                      bool strict_target) const;

    SymbExpression handle_function_call(const vparser::pt_node& expression_node,
                                               Symb_DataDomain::ptr target_data_domain,
                                               bool strict_target) const;
    
    BDDFunction handle_bool_function_call(const vparser::pt_node& expression_node) const;
    
    SymbExpression handle_ite_expression(const vparser::pt_node& expression_node,
                                                Symb_DataDomain::ptr target_data_domain,
                                                bool strict_target) const;

    SymbExpression handle_tmp_var_def(const vparser::pt_node& node) const;
    SymbExpression handle_iterate_expression(const vparser::pt_node& node) const;

    SymbExpression calc_expression(const vparser::pt_node& expression_node,
					  Symb_DataDomain::ptr target_data_domain,
                                          bool strict_target) const;

    BDDFunction perform_compare(comp_op_t comp_op, const SymbExpression& left, const SymbExpression& right) const;

    bdd_assignment_vector perform_assignment_any_value(const bdd_assignment_target_vector& left) const;

    bdd_assignment_vector perform_assignment(const bdd_assignment_target_vector& left,
					     const SymbExpression& value) const;


    PortDescription::ptr get_port_description(const vparser::guard_helper::port_name_t& port_name) const;
    
    BDDFunction port_active(Symb_Channel_End::ptr ce, bool active=true) const;
    const BDDFunctionContainer& get_data_vars_for_port(PortDescription::ptr pd) const;



    BDDFunction only_ports_active(const PortDescription::ptr_set& active_ports) const;

    // calculate not yet calculated atomic propositions specified in the component interface
    void calculateAtomicPropositions();

    DataDomain::ptr findType(const string& name) const;

    bool build_automaton_with_serialized_varorder(BuildContext::ptr build_context);

    virtual BDDFunction getEquivalentTransitions(const BDDFunction& transition) const;
    virtual BDDFunction canonicalizeTransitions(const BDDFunction& transition_set) const;

    VarPositions::ptr var_positions;

    // reference to the build context
    BuildContext::ptr build_context;

    // store restrictions for state values (only on the X variables)
    typedef std::list<BDDFunction> state_value_restriction_list;
    state_value_restriction_list state_value_restrictions;

    BDDFunction no_state_changes;
    BDDFunction reachabilityConstrain;
#ifdef WITH_DUMMY_STATE
    StateVars dummy_state_var;
#endif

  private:
    void init();

    void add_state_value_restriction(const BDDFunction& restriction);
    void copy_state_value_restrictions(const SymbComponentInstance& other);

    Symb_Channel_End::ptr find_identical_symb_ce_for_port(const PortID& port_id) const;
    Symb_Channel_End::ptr find_shared_symb_ce_for_port(const PortID& port_id) const;

    SymbComponentInfo::ptr resolveSubcomponentAccess(const vparser::pt_node& node) const;

  };
}
#endif
