/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "symb_datadomain.hpp"
#include "symb_flat_datadomain.hpp"
#include "symb_array_datadomain.hpp"
#include "symb_struct_datadomain.hpp"
#include "symb_tagged_union_datadomain.hpp"
#include "symb_expression.hpp"

#include "vereofy/common/errors.hpp"
#include "vereofy/common/bdd_helpers.hpp"



namespace vereofy {

  Symb_DataDomain::Symb_DataDomain(DataDomain::ptr data_domain_) : 
    data_domain(data_domain_) {}

  Symb_DataDomain::ptr Symb_DataDomain::getFor(DataDomain::ptr data_domain) {
    data_domain->ensureFiniteness();
    data_domain->ensureFullyDefinedness();

    switch (data_domain->getType()) {
    case DataDomain::DD_FLAT: {
      FlatDataDomain::ptr flat_dd=
        boost::static_pointer_cast<FlatDataDomain>(data_domain);
      return Symb_DataDomain::ptr(new Symb_FlatDataDomain(flat_dd));
    }
    case DataDomain::DD_ARRAY: {
      ArrayDataDomain::ptr array_dd=
        boost::static_pointer_cast<ArrayDataDomain>(data_domain);
      return Symb_DataDomain::ptr(new Symb_ArrayDataDomain(array_dd));
    }
    case DataDomain::DD_STRUCT: {
      StructDataDomain::ptr struct_dd=
        boost::static_pointer_cast<StructDataDomain>(data_domain);
      return Symb_DataDomain::ptr(new Symb_StructDataDomain(struct_dd));
    }
    case DataDomain::DD_TAGGED_UNION: {
      TaggedUnionDataDomain::ptr tagged_union_dd=
        boost::static_pointer_cast<TaggedUnionDataDomain>(data_domain);
      return Symb_DataDomain::ptr(new Symb_TaggedUnionDataDomain(tagged_union_dd));
    }

    default:
      THROW_UNIMPLEMENTED();
    }
  }

  DataDomain::data_domain_type Symb_DataDomain::getType() const {
    return data_domain->getType();
  }

  const DataDomain::ptr Symb_DataDomain::getDataDomain() const {
    return data_domain;
  }

  BDDFunctionContainer Symb_DataDomain::getNULL() const {
    BDDFunctionContainer result;
    for (unsigned long i=0;i<getLogSize();++i){
      result.push(BDDFunction::ZERO());
    }
    return result;
  }
  
  SymbExpression Symb_DataDomain::castExpression(const SymbExpression& expression) const {
    if (getDataDomain()->isSameAs(*expression.getDataDomain()->getDataDomain())) {
      return expression;
    }

    throw "Can not cast expression!";
  }


  BDDFunction Symb_DataDomain::performComparison(comp_op_t compare_op,
                                                    const SymbExpression& left, 
                                                    const SymbExpression& right) const {
    // default implemntation (only == and !=):
    // if the data domains agree, perform bitwise comparison

    bool negate=false;
    switch (compare_op) {
    case COMPARE_EQUAL:
      negate=false;
      break;
    case COMPARE_NOTEQUAL:
      negate=true;
      break;
    case COMPARE_LESS:
    case COMPARE_LESSEQUAL:
    case COMPARE_GREATER:
    case COMPARE_GREATEREQUAL:
      throw "Unsupported comparison operator for these types";
    }
    
    // verify that the data domains are equal
    if (left.getDataDomain()->getDataDomain()->isSameAs(*right.getDataDomain()->getDataDomain())) {
      // generate equality test
      BDDFunction equal=pairwiseEQUIV(left.getValues(), right.getValues());

      if (negate) {
        return !equal;
      } else {
        return equal;
      }      
    } else {
      throw "Can not perform comparison for these different data types";
    }
  }

  
  BDDFunction Symb_DataDomain::performAssignment(const BDDFunctionContainer& targets,
                                                 const SymbExpression& right) const {
    // default implemntation, perform bitwise assignment if data domains agree

    if (getDataDomain()->isSameAs(*right.getDataDomain()->getDataDomain())) {
      return pairwiseEQUIV(targets, right.getValues());
    } else {
      throw "Can not assign values from differing data types, "+getDataDomain()->toString()+" and "+right.getDataDomain()->getDataDomain()->toString()+"!";
    }
  }  

  BDDFunction Symb_DataDomain::getEquivalent(const BDDFunction& assignments, 
                                             const BDDFunctionContainer& encoding) const {
    return assignments;
  }

}
