/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "vereofy/common/math_helpers.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/common/xml_helpers.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/common/errors.hpp"
#include "symb_flat_datadomain.hpp"
#include "symb_function_constructor.hpp"
#include "symb_expression.hpp"
#include <iostream>
#include <boost/lexical_cast.hpp>
using namespace vereofy;

Symb_FlatDataDomain::Symb_FlatDataDomain(FlatDataDomain::ptr data_domain) : 
  Symb_DataDomain(data_domain) {
  if (getFlatDataDomain()->getFlatType() == FlatDataDomain::DD_INT) {
    CHECK_SUCCESS(boost::static_pointer_cast<IntRangeDataDomain>(data_domain)->hasRange());
  }
}

const FlatDataDomain::ptr Symb_FlatDataDomain::getFlatDataDomain() const {
  return boost::static_pointer_cast<FlatDataDomain>(getDataDomain());
}


unsigned int Symb_FlatDataDomain::getLogSize() const {
  if (debugLevel>=6) {
    std::cerr << "Log Size for " << getDataDomain()->toString() << " : ";
  }
  unsigned int size;
  
  switch (getFlatDataDomain()->getFlatType()) {
  case FlatDataDomain::DD_INT: {
    IntRangeDataDomain::ptr dd=boost::static_pointer_cast<IntRangeDataDomain>(getFlatDataDomain());
    CHECK_SUCCESS(dd->hasRange());

    if (dd->getLowestInt()<0) {
      throw "Data domains with negative numbers currently not supported!";
    }

    if (dd->getLowestInt()==0 &&
	dd->getHighestInt()==0) {
      return 0;
    }

    size=logSize(dd->getHighestInt()+1);
    break;
  }
  case FlatDataDomain::DD_ENUM: {
    EnumDataDomain::ptr dd=boost::static_pointer_cast<EnumDataDomain>(getFlatDataDomain());
    size=logSize(dd->countElements());  
    break;
  }
  default:
    throw "not yet implemented";
  }

  if (debugLevel>=6) {
    std::cerr << size << std::endl;
  }
  return size;
}


double Symb_FlatDataDomain::countValidValues() const {
  switch (getFlatDataDomain()->getFlatType()) {
  case FlatDataDomain::DD_INT: {
    IntRangeDataDomain::ptr dd=boost::static_pointer_cast<IntRangeDataDomain>(getFlatDataDomain());
    CHECK_SUCCESS(dd->hasRange());
    return dd->getHighestInt()+1;
  }
  case FlatDataDomain::DD_ENUM: {
    EnumDataDomain::ptr dd=boost::static_pointer_cast<EnumDataDomain>(getFlatDataDomain());
    return dd->countElements();
  }
  default:
    throw "not yet implemented";
  }
}


BDDFunction Symb_FlatDataDomain::restrictToValidValues(const BDDFunctionContainer& encoding) const {
  switch (getFlatDataDomain()->getFlatType()) {
  case FlatDataDomain::DD_INT: {
    unsigned int biggest_value=getFlatDataDomain()->getHighestInt();
    BDDFunction restriction=containerLEQ(encoding, getEncoding(biggest_value));

    TRACE_BDD_FUNCTION("Symb_FlatDataDomain.restrictToValidValues", restriction, biggest_value);
    return restriction;
  }
  case FlatDataDomain::DD_ENUM: {
    const EnumDataDomain& enum_dd=static_cast<const EnumDataDomain&>(*getFlatDataDomain());
    unsigned int biggest_value=enum_dd.countElements();

    BDDFunction restriction;

    if (biggest_value==0) {
      restriction=containerLEQ(encoding, getEncoding(0));
    } else {
      restriction=containerLEQ(encoding, getEncoding(biggest_value-1));
    }

    TRACE_BDD_FUNCTION("Symb_FlatDataDomain.restrictToValidValues", restriction, "enum");
    return restriction;
  }
  default:
    THROW_UNIMPLEMENTED();
  }
}

BDDFunctionContainer Symb_FlatDataDomain::getEncoding(unsigned long value) const {
  BDDFunctionContainer result;

  for (unsigned long i=0;i<getLogSize();++i){
    result.push((value%2)?BDDFunction::ONE():BDDFunction::ZERO());
    value=value>>1;
  }

  return result;
}


SymbExpression Symb_FlatDataDomain::cast_up(const SymbExpression& expression) const {
  CHECK(1, expression.getDataDomain()->getDataDomain()->isSubsetOf(*getDataDomain()), "");
  CHECK(1, expression.getDataDomain()->getLogSize()==expression.getValues().size(), "");

  if (expression.getValues().size() == getLogSize()) {
    return SymbExpression(boost::const_pointer_cast<Symb_DataDomain>(shared_from_this()),
                          expression.getValues(),
                          expression.getCondition());
  } else if (expression.getValues().size() < getLogSize()) {
    // copy encoding
    BDDFunctionContainer result(expression.getValues());
    unsigned int remaining_bits=getLogSize()-expression.getValues().size();
    // fill the remaining bits with ZERO
    for (unsigned int i=0;i<remaining_bits;i++) {
      result.push(BDDFunction::ZERO());
    }

    return SymbExpression(boost::const_pointer_cast<Symb_DataDomain>(shared_from_this()),
                          result,
                          expression.getCondition());
  } else {
    throw "Can not cast up!";
  }
}

SymbExpression Symb_FlatDataDomain::cast_down(const SymbExpression& expression) const {
  CHECK_SUCCESS(expression.getDataDomain()->getDataDomain()->getType()==DataDomain::DD_FLAT);

  CHECK(1, getFlatDataDomain()->isSubsetOf(*expression.getDataDomain()->getDataDomain()), "");
  CHECK(1, expression.getDataDomain()->getLogSize()==expression.getValues().size(), "");

  if (expression.getValues().size() == getLogSize()) {
    return SymbExpression(boost::const_pointer_cast<Symb_DataDomain>(shared_from_this()),
                          expression.getValues(),
                          expression.getCondition());
  } else if (getLogSize() < expression.getValues().size()) {
    BDDFunctionContainer result;
    
    BDDFunction condition_rest_zero=BDDFunction::ONE();
    for (unsigned int i=0;i<expression.getValues().size();i++) {
      if (i<getLogSize()) {
        // copy lower bits
        result.push(expression.getValues().at(i));
      } else {
        // ensure that bits are zero
        condition_rest_zero &= EQUIV(expression.getValues().at(i), BDDFunction::ZERO());
      }
    }
    
    return SymbExpression(boost::const_pointer_cast<Symb_DataDomain>(shared_from_this()),
                          result,
                          expression.getCondition() & condition_rest_zero);
  } else {
    throw "Can not cast down!";
  }
}


BDDFunctionContainer Symb_FlatDataDomain::getEncoding(const string& element) const {
  return getEncoding(getIntValueForEnumElement(element));
}


unsigned int Symb_FlatDataDomain::reverseConstExpression(const SymbExpression& expression) const {
  CHECK(1, expression.getValues().size()==getLogSize(), "");
  CHECK_SUCCESS(expression.getDataDomain()->getDataDomain()->isSameAs(*getDataDomain()));

  unsigned long int_value=0;
  for (unsigned int i=0;i<getLogSize();i++) {
    if (expression.getValues().at(i) == BDDFunction::ONE()) {
      int_value+= 1L << i;
    } else if (expression.getValues().at(i) == BDDFunction::ZERO()) {
      // nothing to do
    } else {
      throw "reverseConstExpression: expression is not a constant!";
    }
  }

  return int_value;
}


string Symb_FlatDataDomain::reverseEncoding(const BDDFunction& assignment, const BDDFunctionContainer& vars) const {
  CHECK(1, vars.size()==getLogSize(), "");

  BDD_CHECK2(2, BDDChecks::isAssignment(assignment, container2cs(vars)), "", assignment, container2cs(vars));

  unsigned long int_value=0;
  for (unsigned int i=0;i<getLogSize();i++) {
    if (bdd_is_bit_set(assignment, vars.at(i))) {
      int_value+= 1L << i;
    }
  }

  switch (getFlatDataDomain()->getFlatType()) {
  case FlatDataDomain::DD_INT: 
    return boost::lexical_cast<string>(int_value);
  case FlatDataDomain::DD_ENUM: {
    const EnumDataDomain& enum_dd=static_cast<const EnumDataDomain&>(*getFlatDataDomain());
    
    if (enum_dd.isValidIndex(int_value)) {
      //      return enum_dd.getTypeName()+"::"+enum_dd.getElementForIndex(int_value);
      return enum_dd.getElementForIndex(int_value);
    } else {
      cerr << "Warning: index " << int_value << " out of range for enum type " << enum_dd.getTypeName() << endl;
      return boost::lexical_cast<string>(int_value);
    }
  }
  default:
    THROW_UNIMPLEMENTED();
  }

  throw "implementation error";
}

void Symb_FlatDataDomain::reverseEncodingToXML(std::ostream& out, 
                                               const BDDFunction& assignment,
                                               const BDDFunctionContainer& vars) const {
  xml_tag_open(out, "value", true);
  xml_text(out, reverseEncoding(assignment, vars));
  xml_tag_close(out, "value");
}


void Symb_FlatDataDomain::reverseEncodingToGuards(const BDDFunction& assignment, 
                                                  const BDDFunctionContainer& vars,
                                                  const std::string& prefix,
                                                  std::vector<std::string> &result) const {
  std::string s=prefix;
  s+="==";
  s+=reverseEncoding(assignment, vars);
  result.push_back(s);
}


int Symb_FlatDataDomain::getIntValueForEnumElement(const string& element) const {
  switch (getFlatDataDomain()->getFlatType()) {
  case FlatDataDomain::DD_INT:
  case FlatDataDomain::DD_BOOL:
    throw "Can't have enum elements for integer data domain!";

  case FlatDataDomain::DD_ENUM: {
    EnumDataDomain::ptr dd=boost::static_pointer_cast<EnumDataDomain>(getFlatDataDomain());
    ReturnValue<unsigned int> i=dd->getIndexForElement(element);
    if (!i.valid) {
      throw string("Element ")+element+" is not valid!";
    }
    return i.value;	
  }
  default:
    throw "not yet implemented!";
  }
}


SymbExpression Symb_FlatDataDomain::castExpression(const SymbExpression& expression) const {
  Symb_DataDomain::ptr sdd_other=expression.getDataDomain();

  if (getDataDomain()->isSameAs(*sdd_other->getDataDomain())) {
    // nothing to do
    return expression;
    }

  if (getDataDomain()->isSubsetOf(*sdd_other->getDataDomain())) {
    // cast down
    CHECK_SUCCESS(sdd_other->getDataDomain()->getType() == DataDomain::DD_FLAT);
    return cast_down(expression);
  }

  if (sdd_other->getDataDomain()->isSubsetOf(*getDataDomain())) {
    // cast up
    CHECK_SUCCESS(sdd_other->getDataDomain()->getType() == DataDomain::DD_FLAT);
    return cast_up(expression);
  }

  throw "Cast error!";
}


BDDFunction Symb_FlatDataDomain::performComparison(comp_op_t compare_op,
                                                   const SymbExpression& left, 
                                                   const SymbExpression& right) const {
  // verify that we are the Symb_DataDomain of left
  CHECK_SUCCESS(left.getDataDomain().get()==this);

  switch (getFlatDataDomain()->getFlatType()) {
  case FlatDataDomain::DD_ENUM:
    // call the default implementation in Symb_DataDomain
    return this->Symb_DataDomain::performComparison(compare_op, left, right);
  case FlatDataDomain::DD_INT:
  case FlatDataDomain::DD_BOOL:
    // ok, handle below
    break;
  case FlatDataDomain::DD_UNION:
    THROW_UNIMPLEMENTED();
  }

  // check compatibility of right data domain
  if (right.getDataDomain()->getDataDomain()->getType()!=DataDomain::DD_FLAT ||
      ((const FlatDataDomain&)*right.getDataDomain()->getDataDomain()).getFlatType()!=FlatDataDomain::DD_INT) {
    throw "Can not compare incompatible types!";
  }

  // both data domains are integer type

  BDDFunctionContainer left_values=left.getValues();
  BDDFunctionContainer right_values=right.getValues();

  // extend values to have equal size
  containerFillToEqualSize(left_values, right_values);
  
  switch (compare_op) {
  case COMPARE_EQUAL: {
    // equal
    return pairwiseEQUIV(left_values, right_values);
  }
  case COMPARE_NOTEQUAL: {
    // not equal
    return !pairwiseEQUIV(left_values, right_values);
  }
  case COMPARE_LESS: {
    // less
    return containerLE(left_values, right_values);
  }
  case COMPARE_LESSEQUAL: {
    // less
    return containerLEQ(left_values, right_values);
  }
  case COMPARE_GREATER: {
    // greater 
    return !containerLEQ(left_values, right_values);
  }
  case COMPARE_GREATEREQUAL: {
    // greater equal
    return !containerLE(left_values, right_values);
  }
  default:
    THROW_UNIMPLEMENTED();
  }
}



BDDFunction Symb_FlatDataDomain::performAssignment(const BDDFunctionContainer& targets,
                                                   const SymbExpression& right) const {

  DataDomain::ptr dd_right=right.getDataDomain()->getDataDomain();
  if (dd_right->getType()!=DataDomain::DD_FLAT) {
    throw "Can not assign different data types!";
  }

  const FlatDataDomain& fdd_right=static_cast<const FlatDataDomain&>(*dd_right);
  if (getFlatDataDomain()->getFlatType() != fdd_right.getFlatType()) {
    throw "Can not assign different data types!";
  }

  switch (getFlatDataDomain()->getFlatType()) {
  case FlatDataDomain::DD_ENUM:
    // call the default implementation in Symb_DataDomain
    return this->Symb_DataDomain::performAssignment(targets, right);
  case FlatDataDomain::DD_INT:
  case FlatDataDomain::DD_BOOL:
    // ok, handle below
    break;
  case FlatDataDomain::DD_UNION:
    THROW_UNIMPLEMENTED();
  }

  // ok, both data types are integers

  BDDFunctionContainer values;
  BDDFunction constraints=BDDFunction::ONE();

  unsigned int target_size=getLogSize();
  unsigned int value_size=right.getDataDomain()->getLogSize();

  // handle different bit sizes:
  
  // loop over all bits (both target and value bits)
  for (unsigned int i=0; i<target_size || i<value_size; i++) {
    if (i < target_size) {
      if (i < value_size) {
        // copy value to target
        values.push(right.getValues().at(i));
      } else {
        // target has more bits than value -> fill with 0
        values.push(BDDFunction::ZERO());
      }
    } else {
      // target has less bits than value, require that those bits in the 
      // value are 0
      constraints&=EQUIV(right.getValues().at(i), BDDFunction::ZERO());
    }    
  }

  return pairwiseEQUIV(targets, values) & constraints;
}


namespace vereofy {
  class SymbFunction_Constructor_Flat : public SymbFunction_Constructor {
  public:
    typedef boost::shared_ptr<SymbFunction_Constructor_Flat> ptr;
    
    SymbFunction_Constructor_Flat(Symb_FlatDataDomain::ptr dd) : 
      SymbFunction_Constructor(dd) {}
      
    virtual SymbExpression::ptr calculate(SymbFunctionArgumentVector& arguments) const {
      checkArguments(arguments);
      
      SymbExpression::ptr expression=arguments.at(0).getSymbExpression();

      return SymbExpression::ptr(new SymbExpression(symb_dd->castExpression(*expression)));
    }

    virtual Symb_DataDomain::ptr getExpectedArgumentType(unsigned int index,
                                                         const SymbFunctionArgumentVector& previous_arguments) const {
      CHECK_SUCCESS(index==0);

      Symb_FlatDataDomain::ptr flat_dd=
        boost::static_pointer_cast<Symb_FlatDataDomain>(symb_dd);
      
      // in case of enum, return enum domain so that there can be 
      // correct mapping
      switch (flat_dd->getFlatDataDomain()->getFlatType()) {
      case FlatDataDomain::DD_INT:
        return Symb_DataDomain::ptr();
      case FlatDataDomain::DD_ENUM:
        return symb_dd;
      default:
        THROW_UNIMPLEMENTED();
      }
    }
  };
}

boost::shared_ptr<SymbFunction_Constructor> Symb_FlatDataDomain::getConstructorFunction() {
  boost::shared_ptr<Symb_FlatDataDomain> sdd=
    boost::static_pointer_cast<Symb_FlatDataDomain>(this->shared_from_this());

  return SymbFunction_Constructor::ptr(new SymbFunction_Constructor_Flat(sdd));
}

