/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "symb_function_arithmetic.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/common/math_helpers.hpp"
#include <boost/lexical_cast.hpp>

using namespace vereofy;
using namespace std;

SymbExpression::ptr SymbFunction_uint_add::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  SymbExpression::ptr a=arguments.at(0).getSymbExpression();
  SymbExpression::ptr b=arguments.at(1).getSymbExpression();

  Symb_DataDomain::ptr s_dd_a=a->getDataDomain();
  Symb_DataDomain::ptr s_dd_b=b->getDataDomain();

  IntRangeDataDomain::ptr dd_a=arguments.at(0).getIntRangeDataDomain();
  IntRangeDataDomain::ptr dd_b=arguments.at(1).getIntRangeDataDomain();

  // calculate new boundaries from the data domains
  unsigned int new_min=0;
  unsigned int new_max=dd_a->getHighestInt() + dd_b->getHighestInt();

  IntRangeDataDomain::ptr dd_result(new IntRangeDataDomain(new_min, new_max));
  Symb_DataDomain::ptr sdd_result=Symb_DataDomain::getFor(dd_result);

  BDDFunctionContainer values_a=a->getValues();
  BDDFunctionContainer values_b=b->getValues();

  containerFillToEqualSize(values_a, values_b);
  BDDFunctionContainer values_result=containerSUM(values_a, values_b);

  values_result=containerTruncate(values_result, sdd_result->getLogSize());

  SymbExpression::ptr result(new SymbExpression(sdd_result,
                                                values_result));
  return result;
}

unsigned int SymbFunction_uint_add::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);
  return arguments.at(0).getConstValue() + arguments.at(1).getConstValue();
}


SymbExpression::ptr SymbFunction_uint_sub::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  SymbExpression::ptr a=arguments.at(0).getSymbExpression();
  SymbExpression::ptr b=arguments.at(1).getSymbExpression();

  Symb_DataDomain::ptr s_dd_a=a->getDataDomain();
  Symb_DataDomain::ptr s_dd_b=b->getDataDomain();

  IntRangeDataDomain::ptr dd_a=arguments.at(0).getIntRangeDataDomain();
  IntRangeDataDomain::ptr dd_b=arguments.at(1).getIntRangeDataDomain();

  Symb_DataDomain::ptr sdd_result=s_dd_a;

  BDDFunctionContainer values_a=a->getValues();
  BDDFunctionContainer values_b=b->getValues();

  containerFillToEqualSize(values_a, values_b);
  pair<BDDFunctionContainer,BDDFunction> values_result=containerSUB(values_a, values_b);

  values_result.first=containerTruncate(values_result.first, sdd_result->getLogSize());

  const BDDFunction& underflow=values_result.second;

  SymbExpression::ptr result(new SymbExpression(sdd_result,
                                                values_result.first,
                                                !underflow));
  return result;
}

unsigned int SymbFunction_uint_sub::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  int r=arguments.at(0).getConstValue() - arguments.at(1).getConstValue();
  if (r<0) {
    throw "Result of subtraction is negative, currently not supported";
  }
  return (unsigned int)r;
}



SymbExpression::ptr SymbFunction_uint_mult::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  SymbExpression::ptr a=arguments.at(0).getSymbExpression();
  SymbExpression::ptr b=arguments.at(1).getSymbExpression();

  Symb_DataDomain::ptr s_dd_a=a->getDataDomain();
  Symb_DataDomain::ptr s_dd_b=b->getDataDomain();

  IntRangeDataDomain::ptr dd_a=arguments.at(0).getIntRangeDataDomain();
  IntRangeDataDomain::ptr dd_b=arguments.at(1).getIntRangeDataDomain();

  // calculate new boundaries from the data domains
  unsigned int new_min=0;
  unsigned int new_max=dd_a->getHighestInt() * dd_b->getHighestInt();

  IntRangeDataDomain::ptr dd_result(new IntRangeDataDomain(new_min, new_max));
  Symb_DataDomain::ptr sdd_result=Symb_DataDomain::getFor(dd_result);

  BDDFunctionContainer values_a=a->getValues();
  BDDFunctionContainer values_b=b->getValues();

  containerFillToEqualSize(values_a, values_b);
  BDDFunctionContainer values_result=containerMULT(values_a, values_b);

  values_result=containerTruncate(values_result, sdd_result->getLogSize());

  SymbExpression::ptr result(new SymbExpression(sdd_result,
                                                values_result));
  return result;
}

unsigned int SymbFunction_uint_mult::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);
  return arguments.at(0).getConstValue() * arguments.at(1).getConstValue();
}



SymbExpression::ptr SymbFunction_uint_div::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  SymbExpression::ptr a=arguments.at(0).getSymbExpression();
  SymbExpression::ptr b=arguments.at(1).getSymbExpression();

  Symb_DataDomain::ptr s_dd_a=a->getDataDomain();
  Symb_DataDomain::ptr s_dd_b=b->getDataDomain();

  IntRangeDataDomain::ptr dd_a=arguments.at(0).getIntRangeDataDomain();
  IntRangeDataDomain::ptr dd_b=arguments.at(1).getIntRangeDataDomain();

  // calculate new boundaries from the data domains
  unsigned int new_min=0;
  unsigned int new_max=dd_a->getHighestInt();

  IntRangeDataDomain::ptr dd_result(new IntRangeDataDomain(new_min, new_max));
  Symb_DataDomain::ptr sdd_result=Symb_DataDomain::getFor(dd_result);

  BDDFunctionContainer values_a=a->getValues();
  BDDFunctionContainer values_b=b->getValues();

  BDDFunction b_is_zero=pairwiseEQUIV_with_fill(values_b, BDDFunctionContainer());

  BDDFunctionContainer values_result=containerDIV(values_a, values_b);

  SymbExpression::ptr result(new SymbExpression(sdd_result,
                                                values_result,
                                                !b_is_zero));
  return result;
}

unsigned int SymbFunction_uint_div::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);
  if (arguments.at(1).getConstValue()) {
    throw "DIV/ZERO!";
  }

  return arguments.at(0).getConstValue() * arguments.at(1).getConstValue();
}


SymbExpression::ptr SymbFunction_uint_mod::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  SymbExpression::ptr a=arguments.at(0).getSymbExpression();
  SymbExpression::ptr b=arguments.at(1).getSymbExpression();

  Symb_DataDomain::ptr s_dd_a=a->getDataDomain();
  Symb_DataDomain::ptr s_dd_b=b->getDataDomain();

  IntRangeDataDomain::ptr dd_a=arguments.at(0).getIntRangeDataDomain();
  IntRangeDataDomain::ptr dd_b=arguments.at(1).getIntRangeDataDomain();

  // calculate new boundaries from the data domains
  unsigned int new_min=0;
  unsigned int new_max=dd_a->getHighestInt();

  IntRangeDataDomain::ptr dd_result(new IntRangeDataDomain(new_min, new_max));
  Symb_DataDomain::ptr sdd_result=Symb_DataDomain::getFor(dd_result);

  BDDFunctionContainer values_a=a->getValues();
  BDDFunctionContainer values_b=b->getValues();

  BDDFunction b_is_zero=pairwiseEQUIV_with_fill(values_b, BDDFunctionContainer());

  BDDFunctionContainer values_result=containerDIV(values_a, values_b, true);

  SymbExpression::ptr result(new SymbExpression(sdd_result,
                                                values_result,
                                                !b_is_zero));
  return result;
}

unsigned int SymbFunction_uint_mod::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);
  if (arguments.at(1).getConstValue()) {
    throw "DIV/ZERO!";
  }

  return arguments.at(0).getConstValue() % arguments.at(1).getConstValue();
}




SymbExpression::ptr SymbFunction_uint_mod_bits::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  SymbExpression::ptr v=arguments.at(0).getSymbExpression();
  unsigned int k=arguments.at(1).getConstValue();

  Symb_DataDomain::ptr s_dd_v=v->getDataDomain();
  IntRangeDataDomain::ptr dd_v=arguments.at(0).getIntRangeDataDomain();

  // calculate new boundaries from the data domains
  unsigned int new_min=0;
  unsigned int new_max=(1 << k)-1;

  IntRangeDataDomain::ptr dd_result(new IntRangeDataDomain(new_min, new_max));
  Symb_DataDomain::ptr sdd_result=Symb_DataDomain::getFor(dd_result);

  BDDFunctionContainer values_v=v->getValues();
  BDDFunctionContainer values_result;
  for (unsigned int i=0;i<sdd_result->getLogSize();i++) {
    if (i>=values_v.size()) {
      values_result.push(BDDFunction::ZERO());
    } else {
      values_result.push(values_v.at(i));
    }
  }

  SymbExpression::ptr result(new SymbExpression(sdd_result,
                                                values_result));
  return result;
}

unsigned int SymbFunction_uint_mod_bits::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);
  return arguments.at(0).getConstValue() % (1 << arguments.at(1).getConstValue());
}




SymbExpression::ptr SymbFunction_uint_compare::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);


  SymbExpression::ptr a=arguments.at(0).getSymbExpression();
  SymbExpression::ptr b=arguments.at(1).getSymbExpression();

  Symb_DataDomain::ptr s_dd_a=a->getDataDomain();

  BDDFunction result_value=s_dd_a->performComparison(compare_op,
                                                     *a, *b);
  
  IntRangeDataDomain::ptr bool_dd(new IntRangeDataDomain(0,1));
  Symb_DataDomain::ptr symb_bool=Symb_DataDomain::getFor(bool_dd);
  
  SymbExpression::ptr result(new SymbExpression(symb_bool,
                                                result_value));
  return result;
}

unsigned int SymbFunction_uint_compare::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  unsigned int a=arguments.at(0).getConstValue();
  unsigned int b=arguments.at(1).getConstValue();

  switch (compare_op) {
  case COMPARE_EQUAL: return (a==b ? 1 : 0);
  case COMPARE_NOTEQUAL: return (a!=b ? 1 : 0);
  case COMPARE_LESS: return (a<b ? 1 : 0);
  case COMPARE_LESSEQUAL: return (a<=b ? 1 : 0);
  case COMPARE_GREATER: return (a>b ? 1 : 0);
  case COMPARE_GREATEREQUAL: return (a>=b ? 1 : 0);
  }
  THROW_UNIMPLEMENTED();
}



SymbExpression::ptr SymbFunction_uint_max::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  SymbExpression::ptr a=arguments.at(0).getSymbExpression();
  SymbExpression::ptr b=arguments.at(1).getSymbExpression();

  Symb_DataDomain::ptr s_dd_a=a->getDataDomain();
  Symb_DataDomain::ptr s_dd_b=b->getDataDomain();

  IntRangeDataDomain::ptr dd_a=arguments.at(0).getIntRangeDataDomain();
  IntRangeDataDomain::ptr dd_b=arguments.at(1).getIntRangeDataDomain();

  // calculate new boundaries from the data domains
  unsigned int new_min=0;
  unsigned int new_max=dd_a->getHighestInt();
  if ((unsigned int)dd_b->getHighestInt() > new_max) {
    new_max=dd_b->getHighestInt();
  }

  IntRangeDataDomain::ptr dd_result(new IntRangeDataDomain(new_min, new_max));
  Symb_DataDomain::ptr sdd_result=Symb_DataDomain::getFor(dd_result);

  BDDFunctionContainer values_a=a->getValues();
  BDDFunctionContainer values_b=b->getValues();

  containerFillToEqualSize(values_a, values_b);

  BDDFunction a_less_than_b=containerLEQ(values_a, values_b);
  BDDFunctionContainer values_result;
  for (unsigned int i=0;i<values_a.size();i++) {
    values_result.push((a_less_than_b & values_b.at(i)) |
                       (!a_less_than_b & values_a.at(i)));
  }

  SymbExpression::ptr result(new SymbExpression(sdd_result,
                                                values_result));
  return result;
}

unsigned int SymbFunction_uint_max::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);
  
  unsigned int a=arguments.at(0).getConstValue();
  unsigned int b=arguments.at(1).getConstValue();

  return (a<b) ? b : a;
}


SymbExpression::ptr SymbFunction_uint_min::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  SymbExpression::ptr a=arguments.at(0).getSymbExpression();
  SymbExpression::ptr b=arguments.at(1).getSymbExpression();

  Symb_DataDomain::ptr s_dd_a=a->getDataDomain();
  Symb_DataDomain::ptr s_dd_b=b->getDataDomain();

  IntRangeDataDomain::ptr dd_a=arguments.at(0).getIntRangeDataDomain();
  IntRangeDataDomain::ptr dd_b=arguments.at(1).getIntRangeDataDomain();

  // calculate new boundaries from the data domains
  unsigned int new_min=0;
  unsigned int new_max=dd_a->getHighestInt();
  if ((unsigned int)dd_b->getHighestInt() < new_max) {
    new_max=dd_b->getHighestInt();
  }

  IntRangeDataDomain::ptr dd_result(new IntRangeDataDomain(new_min, new_max));
  Symb_DataDomain::ptr sdd_result=Symb_DataDomain::getFor(dd_result);

  BDDFunctionContainer values_a=a->getValues();
  BDDFunctionContainer values_b=b->getValues();

  containerFillToEqualSize(values_a, values_b);

  BDDFunction a_less_than_b=containerLEQ(values_a, values_b);
  BDDFunctionContainer values_result;
  for (unsigned int i=0;i<sdd_result->getLogSize();i++) {
    values_result.push((a_less_than_b & values_a.at(i)) |
                       (!a_less_than_b & values_b.at(i)));
  }

  SymbExpression::ptr result(new SymbExpression(sdd_result,
                                                values_result));
  return result;
}

unsigned int SymbFunction_uint_min::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);
  
  unsigned int a=arguments.at(0).getConstValue();
  unsigned int b=arguments.at(1).getConstValue();

  return (a<b) ? a : b;
}




SymbExpression::ptr SymbFunction_uint_pow2::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  return SymbExpression::ptr(new SymbExpression(const_evaluate(arguments)));
}

unsigned int SymbFunction_uint_pow2::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);
  unsigned int a=arguments.at(0).getConstValue();
  
  return (1 << a);
}


SymbExpression::ptr SymbFunction_uint_bits_needed::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  return SymbExpression::ptr(new SymbExpression(const_evaluate(arguments)));
}

unsigned int SymbFunction_uint_bits_needed::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);
  unsigned int a=arguments.at(0).getConstValue();
  
  return logSize(a+1);
}


SymbExpression::ptr SymbFunction_uint_ones_complement::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  unsigned int bits=arguments.at(0).getConstValue();
  BDDFunctionContainer arg=arguments.at(1).getSymbExpression()->getValues();
  
  BDDFunctionContainer values;
  for (unsigned int i=0;i<bits;i++) {
    if (i<arg.size()) {
      values.push(arg.at(i));
    } else {
      values.push(BDDFunction::ZERO());
    }
  }

  BDDFunctionContainer result=containerOnesComplement(values);

  IntRangeDataDomain::ptr dd(new IntRangeDataDomain(0, (1<<bits)-1));
  Symb_DataDomain::ptr s_dd=Symb_DataDomain::getFor(dd);
  CHECK_SUCCESS(s_dd->getLogSize() == result.size());

  return SymbExpression::ptr(new SymbExpression(s_dd,
                                                result));
}

unsigned int SymbFunction_uint_ones_complement::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  unsigned int bits=arguments.at(0).getConstValue();
  unsigned int value=arguments.at(1).getConstValue();

  unsigned int result = (~value) % (1 << bits);
  return result;
}


SymbExpression::ptr SymbFunction_uint_twos_complement::calculate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  unsigned int bits=arguments.at(0).getConstValue();
  BDDFunctionContainer arg=arguments.at(1).getSymbExpression()->getValues();
  
  BDDFunctionContainer values;
  for (unsigned int i=0;i<bits;i++) {
    if (i<arg.size()) {
      values.push(arg.at(i));
    } else {
      values.push(BDDFunction::ZERO());
    }
  }

  BDDFunctionContainer result=containerTwosComplement(values);

  IntRangeDataDomain::ptr dd(new IntRangeDataDomain(0, (1<<bits)-1));
  Symb_DataDomain::ptr s_dd=Symb_DataDomain::getFor(dd);
  CHECK_SUCCESS(s_dd->getLogSize() == result.size());

  return SymbExpression::ptr(new SymbExpression(s_dd,
                                                result));
}

unsigned int SymbFunction_uint_twos_complement::const_evaluate(SymbFunctionArgumentVector& arguments) const {
  checkArguments(arguments);

  unsigned int bits=arguments.at(0).getConstValue();
  unsigned int value=arguments.at(1).getConstValue();

  unsigned int result = ((~value)+1) % (1 << bits);
  return result;
}
