/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "symb_function_seq_number_arithmetic.hpp"
#include "vereofy/common/bdd_helpers.hpp"

namespace vereofy {

  SymbExpression::ptr SymbFunction_seq_number_add::calculate(SymbFunctionArgumentVector& arguments) const {
    checkArguments(arguments);
    
    unsigned int bits=arguments.at(0).getConstValue();
    
    SymbFunction_uint_add uint_add;
    SymbFunction_uint_mod_bits uint_mod;
    
    SymbFunctionArgumentVector add_arguments;
    add_arguments.push_back(arguments.at(1));
    add_arguments.push_back(arguments.at(2));
    
    SymbExpression::ptr sum=uint_add.calculate(add_arguments);
    
    SymbFunctionArgumentVector mod_arguments;
    mod_arguments.push_back(SymbFunctionArgument(sum));
    mod_arguments.push_back(SymbFunctionArgument(bits));
    
    return uint_mod.calculate(mod_arguments);
  }
  
  unsigned int SymbFunction_seq_number_add::const_evaluate(SymbFunctionArgumentVector& arguments) const {
    checkArguments(arguments);
    unsigned int bits=arguments.at(0).getConstValue();
    unsigned int a=arguments.at(1).getConstValue();
    unsigned int b=arguments.at(2).getConstValue();
   
    return (a+b) % (1 << bits);
  }


  SymbExpression::ptr SymbFunction_seq_number_less::calculate(SymbFunctionArgumentVector& arguments) const {
    checkArguments(arguments);
    
    unsigned int bits=arguments.at(0).getConstValue();
    
    SymbExpression::ptr a=arguments.at(1).getSymbExpression();
    SymbExpression::ptr b=arguments.at(2).getSymbExpression();

    // sequence number arithmetic:
    // a < b iff a-b < 0
    BDDFunctionContainer a_values=a->getValues();
    a_values=containerToSize(a_values, bits);
    
    BDDFunctionContainer b_values=b->getValues();
    b_values=containerToSize(b_values, bits);

    BDDFunctionContainer b_complemented=containerTwosComplement(b_values);
    
    // a-b = a+(-b)
    BDDFunctionContainer sum=containerSUM(a_values, b_complemented);

    // < 0 -> bit (bits-1) is one
    BDDFunction less_than_zero=sum.at(bits-1);
    BDDFunction result_value=less_than_zero;

    IntRangeDataDomain::ptr bool_dd(new IntRangeDataDomain(0,1));
    Symb_DataDomain::ptr symb_bool=Symb_DataDomain::getFor(bool_dd);
    
    SymbExpression::ptr result(new SymbExpression(symb_bool,
                                                  result_value));

    return result;
  }

  unsigned int SymbFunction_seq_number_less::const_evaluate(SymbFunctionArgumentVector& arguments) const {
    THROW_UNIMPLEMENTED();
  }


  SymbExpression::ptr SymbFunction_seq_number_less_equal::calculate(SymbFunctionArgumentVector& arguments) const {
    checkArguments(arguments);
    
    SymbExpression::ptr a_less_b=SymbFunction_seq_number_less().calculate(arguments);
    
    SymbExpression::ptr a=arguments.at(1).getSymbExpression();
    SymbExpression::ptr b=arguments.at(2).getSymbExpression();

    BDDFunction result_value=a->getDataDomain()->performComparison(COMPARE_EQUAL,
                                                                   *a, *b);
    result_value|=a_less_b->getValues().at(0);

    IntRangeDataDomain::ptr bool_dd(new IntRangeDataDomain(0,1));
    Symb_DataDomain::ptr symb_bool=Symb_DataDomain::getFor(bool_dd);
    
    SymbExpression::ptr result(new SymbExpression(symb_bool,
                                                  result_value));

    return result;
  }

  unsigned int SymbFunction_seq_number_less_equal::const_evaluate(SymbFunctionArgumentVector& arguments) const {
    THROW_UNIMPLEMENTED();
  }


  SymbExpression::ptr SymbFunction_seq_number_greater::calculate(SymbFunctionArgumentVector& arguments) const {
    checkArguments(arguments);
    
    SymbExpression::ptr a_leq_b=SymbFunction_seq_number_less_equal().calculate(arguments);
    BDDFunction result_value=!a_leq_b->getValues().at(0);

    IntRangeDataDomain::ptr bool_dd(new IntRangeDataDomain(0,1));
    Symb_DataDomain::ptr symb_bool=Symb_DataDomain::getFor(bool_dd);
    
    SymbExpression::ptr result(new SymbExpression(symb_bool,
                                                  result_value));

    return result;
  }

  unsigned int SymbFunction_seq_number_greater::const_evaluate(SymbFunctionArgumentVector& arguments) const {
    THROW_UNIMPLEMENTED();
  }


  SymbExpression::ptr SymbFunction_seq_number_greater_equal::calculate(SymbFunctionArgumentVector& arguments) const {
    checkArguments(arguments);
    
    SymbExpression::ptr a_less_b=SymbFunction_seq_number_less().calculate(arguments);
    BDDFunction result_value=!a_less_b->getValues().at(0);

    IntRangeDataDomain::ptr bool_dd(new IntRangeDataDomain(0,1));
    Symb_DataDomain::ptr symb_bool=Symb_DataDomain::getFor(bool_dd);
    
    SymbExpression::ptr result(new SymbExpression(symb_bool,
                                                  result_value));

    return result;
  }

  unsigned int SymbFunction_seq_number_greater_equal::const_evaluate(SymbFunctionArgumentVector& arguments) const {
    THROW_UNIMPLEMENTED();
  }


 
  SymbExpression::ptr SymbFunction_seq_number_max::calculate(SymbFunctionArgumentVector& arguments) const {
    checkArguments(arguments);

    unsigned int bits=arguments.at(0).getConstValue();
    
    SymbExpression::ptr a=arguments.at(1).getSymbExpression();
    SymbExpression::ptr b=arguments.at(2).getSymbExpression();

    BDDFunctionContainer a_values=a->getValues();
    a_values=containerToSize(a_values, bits);
    
    BDDFunctionContainer b_values=b->getValues();
    b_values=containerToSize(b_values, bits);

    SymbExpression::ptr seq_number_less=SymbFunction_seq_number_less().calculate(arguments);
    CHECK_SUCCESS(seq_number_less->getValues().size() == 1);
    BDDFunctionContainer result_values=containerChoose(seq_number_less->getValues().at(0),
                                                       b_values,
                                                       a_values);
    
    IntRangeDataDomain::ptr dd(new IntRangeDataDomain(0,(1<<bits)-1));
    Symb_DataDomain::ptr sdd=Symb_DataDomain::getFor(dd);
    
    SymbExpression::ptr result(new SymbExpression(sdd,
                                                  result_values));
    return result;
  }

  unsigned int SymbFunction_seq_number_max::const_evaluate(SymbFunctionArgumentVector& arguments) const {
    unsigned int a=arguments.at(1).getConstValue();
    unsigned int b=arguments.at(2).getConstValue();

    return (SymbFunction_seq_number_less().const_evaluate(arguments) ? b : a);
  }

}
