/***************************************************************************
 *   Copyright (C) 2008-2010 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "symb_function_userdefined.hpp"
#include "vereofy/common/errors.hpp"
#include "vereofy/reo_parser/reo_parser_helper.hpp"
#include "vereofy/reo_parser/type_parser_helper.hpp"
#include "vereofy/system/symb_automaton/symb_component_info.hpp"
#include "vereofy/system/symb_automaton/symb_component_instance.hpp"

using namespace vereofy;
using namespace vereofy::vparser;
using namespace std;


namespace vereofy {
  namespace SymbFunctionUserdefined_private {
    class SCI_UD : public SymbComponentInstance {
    public:
      typedef boost::shared_ptr<SCI_UD> ptr;
      
      SCI_UD(ComponentInterface::ptr c_interface_,
             SymbComponentInfo::ptr c_info_) : 
        SymbComponentInstance(c_interface_, Symb_DataDomain::ptr()) {
        info=c_info_;
      }
      
      virtual void build_automaton(BuildContext::ptr build_context) {
        THROW_UNIMPLEMENTED();
      }
      
      SymbExpression calc_expression(const vparser::pt_node& n) {
        return SymbComponentInstance::calc_expression(n,
                                                      Symb_DataDomain::ptr(), 
                                                      false);
      }

      BDDFunction calc_bool_expression(const vparser::pt_node& n) {
        return SymbComponentInstance::bool_expression_to_bdd(n);
      }

    };
  }
}



SymbFunction_Userdefined::SymbFunction_Userdefined(unsigned int arity_, 
                                                   SymbFunction::return_value_t rv_,
                                                   const vparser::pt_node& function_def_,
                                                   SystemContext::wptr context_) :
  SymbFunction(arity_, rv_, false), function_def(function_def_), context(context_) {
}

SymbFunction_Userdefined::~SymbFunction_Userdefined() {}

SymbExpression::ptr SymbFunction_Userdefined::calculate(SymbFunctionArgumentVector& arguments) const {
  ComponentInterface::ptr ci(new ComponentInterface("function"));
  ci->setGlobalSymbols(access_weak_ptr(context)->getGlobalSymbols());
  SymbComponentInfo::ptr sci(new SymbComponentInfo(ci));

  std::string function_name= reo_parser_helper::t_function_def::getName(function_def);

  const vparser::pt_node& arg_list_n = reo_parser_helper::t_function_def::getArgumentList(function_def);
  unsigned int argument_count = reo_parser_helper::t_function_arg_list::getSize(arg_list_n);
  CHECK_SUCCESS(argument_count == arguments.size());

  const vparser::pt_node& return_type_n = reo_parser_helper::t_function_def::getReturnType(function_def);
  DataDomain::ptr return_type = type_helper::t_type::getDataDomain(return_type_n, ci->getSymbolTableStack());

  BDDFunction arg_conditions = BDDFunction::ONE();

  for (unsigned int i=0;i<argument_count;i++) {
    const vparser::pt_node& arg_n = reo_parser_helper::t_function_arg_list::getArgument(arg_list_n, i);
    const std::string name = reo_parser_helper::t_function_arg::getName(arg_n);
    const vparser::pt_node& arg_type_n = reo_parser_helper::t_function_arg::getType(arg_n);
    DataDomain::ptr arg_type = type_helper::t_type::getDataDomain(arg_type_n, ci->getSymbolTableStack());

    SymbFunctionArgument arg = arguments.at(i);
    DataDomain::ptr arg_dd = arg.getSymbExpression()->getDataDomain()->getDataDomain();
    
    if (!arg_dd->isConvertibleTo(*arg_type)) {
      std::string error_msg=
        "In call of function "
        +function_name
        +", argument "
        +boost::lexical_cast<std::string>(i+1)
        +": Argument type mismatch! (Expected: "+arg_type->toString()+", found: "+arg_dd->toString()+")";
      
      throw error_msg;
    }

    DataDomain::ptr target_dd=arg_dd->getConversionTarget(arg_type);
    Symb_DataDomain::ptr target_sdd=Symb_DataDomain::getFor(target_dd);
    SymbExpression converted_arg=target_sdd->castExpression(*arg.getSymbExpression());

    arg_conditions &= converted_arg.getCondition();
    arg_conditions &= target_sdd->restrictToValidValues(converted_arg.getValues());

    StateDescription::ptr sd(new StateDescription(name, arg_dd));
    ci->addStateDescription(sd);
    
    StateVars v;
    v.X=boost::shared_ptr<BDDFunctionContainer>(new BDDFunctionContainer(converted_arg.getValues()));
    SymbolicState::ptr ss(new SymbolicState());
    ss->vars=v;
    ss->data_domain=target_sdd;

    sci->state_to_bdd.insert(make_pair(sd, ss));
  }


  vereofy::SymbFunctionUserdefined_private::SCI_UD::ptr sc_instance(new vereofy::SymbFunctionUserdefined_private::SCI_UD(ci, sci));


  const vparser::pt_node& function_expression_n = reo_parser_helper::t_function_def::getFunctionExpression(function_def);
  
  SymbExpression result;
  if (reo_parser_helper::t_function_def::isFunctionExpressionBoolean(function_def)) {
    IntRangeDataDomain::ptr bool_dd(new IntRangeDataDomain(0,1));
    result = SymbExpression(Symb_DataDomain::getFor(bool_dd),
                            sc_instance->calc_bool_expression(function_expression_n));
  } else {
    result = sc_instance->calc_expression(function_expression_n);
  }

  result.setCondition(result.getCondition() & arg_conditions);

  DataDomain::ptr return_dd = result.getDataDomain()->getDataDomain();
  
  if (!return_dd->isConvertibleTo(*return_type)) {
    std::string error_msg=
      "In call of function "
      +function_name
      +", the function return value of type "
      +return_dd->toString()+" is not convertible to "+return_type->toString();
    
    throw error_msg;
  }

  DataDomain::ptr return_target_dd=return_dd->getConversionTarget(return_type);
  Symb_DataDomain::ptr return_target_sdd=Symb_DataDomain::getFor(return_target_dd);
  SymbExpression converted_return=return_target_sdd->castExpression(result);

  // ensure that all return values are valid
  converted_return.setCondition(converted_return.getCondition() & 
                                return_target_sdd->restrictToValidValues(converted_return.getValues()));

  return SymbExpression::ptr(new SymbExpression(converted_return));
}




SymbFunction_Userdefined::ptr SymbFunction_Userdefined::fromDef(const vparser::pt_node& function_def_, SystemContext::wptr context_) {
  const vparser::pt_node& argument_list_n=reo_parser_helper::t_function_def::getArgumentList(function_def_);

  unsigned int arity=reo_parser_helper::t_function_arg_list::getSize(argument_list_n);

  ComponentInterface::ptr ci(new ComponentInterface("function"));
  ci->setGlobalSymbols(access_weak_ptr(context_)->getGlobalSymbols());

  const vparser::pt_node& return_type_n = reo_parser_helper::t_function_def::getReturnType(function_def_);
  DataDomain::ptr return_type = type_helper::t_type::getDataDomain(return_type_n, ci->getSymbolTableStack());

  return_value_t rv = RETURN_VALUE_COMPLEX;
  if (reo_parser_helper::t_function_def::isReturnTypeBoolean(function_def_)) {
    rv = RETURN_VALUE_BOOLEAN;
  } else if (return_type->getType() == DataDomain::DD_FLAT &&
             boost::static_pointer_cast<FlatDataDomain>(return_type)->getFlatType()==FlatDataDomain::DD_INT) {
    rv = RETURN_VALUE_INTEGER;
  }

  return SymbFunction_Userdefined::ptr(new SymbFunction_Userdefined(arity, rv, function_def_, context_));
}

