/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "symb_node.hpp"
#include <boost/lexical_cast.hpp>
#include "vereofy/common/bdd_helpers.hpp"

namespace vereofy {
  
  SymbNode::SymbNode(SymbComponentInstance::ptr parent,
                     Symb_DataDomain::ptr dataDomain,
                     REO_Node::node_semantics_t node_semantic,
                     const vector<PortDescription::ptr>& in,
                     const vector<PortDescription::ptr>& out,
                     PortDescription::ptr node_port,
                     bool share_channel_ends,
                     VarPositions::ptr var_positions_,
                     BuildContext::ptr build_context) :
  SymbComponentInstance(ComponentInterface::ptr(new ComponentInterface("$node")), dataDomain) {
    this->build_context=build_context;
    
    switch (node_semantic) {
    case REO_Node::NODE_REPLICATE:
      build_merge_copy_node(parent,
                            dataDomain,
                            in, out,
                            node_port,
                            share_channel_ends,
                            var_positions_,
                            build_context);
      return;
    case REO_Node::NODE_ROUTE:
      build_route_node(parent,
                       dataDomain,
                       in, out,
                       node_port,
                       share_channel_ends,
                       var_positions_,
                       build_context);
      return;
    }

    THROW_UNIMPLEMENTED();
  }
  
  void SymbNode::build_merge_copy_node(SymbComponentInstance::ptr parent,
                                       Symb_DataDomain::ptr dataDomain,
                                       const vector<PortDescription::ptr>& in,
                                       const vector<PortDescription::ptr>& out,
                                       PortDescription::ptr node_port,
                                       bool share_channel_ends,
                                       VarPositions::ptr var_positions_,
                                       BuildContext::ptr build_context) {
    //TODO: get rid of this nasty write on protected variable
    var_positions = var_positions_;
    
#ifdef WITH_DUMMY_STATE
    addDummyState();
    initialStates==!dummy_state_var.X->at(0);
#endif

    delta=BDDFunction::ZERO();

    CHECK_SUCCESS(node_port);
    Symb_Channel_End::ptr ce_copy;

    ce_copy=parent->find_symb_ce_for_port(node_port);

    if (!ce_copy && share_channel_ends) {
#ifndef SYMB_NODE_DONT_SHARE_CES

      // try to find a suitable symb channel end for the copy port
      if (in.size()==1) {
        // only one in port, if we have a symb ce -> can be used as copy port
        ce_copy=parent->find_symb_ce_for_port(in.at(0));
      } else if (out.size()>0) {
        // if there is an out port with symb ce, we can take that
        for (unsigned int i=0;i<out.size();i++) {
          ce_copy=parent->find_symb_ce_for_port(out.at(i));
          if (ce_copy) {
            break;
          }
        }
      }
#endif
    }

    if (!ce_copy) {
      // generate a copy port
      ce_copy=addPort(node_port);
    } else {
      addPort(node_port, ce_copy);
    }
    parent->register_symb_ce_for_port(node_port, ce_copy);
    
    ce_port=ce_copy;
    
    BDDFunction copy_semantics=BDDFunction::ONE();
    TRACE_BDD_FUNCTION("SymbNode.build", copy_semantics, "");
    
    for (unsigned int i=0;i<out.size();i++) {
      PortDescription::ptr circuit_port=out.at(i);
      
      PortDescription::ptr out_port(new PortDescription("out"+boost::lexical_cast<string>(i),
                                                         PortDescription::OUT_PORT,
                                                         dataDomain->getDataDomain()));
      info->c_interface->addPort(out_port);

      Symb_Channel_End::ptr symb_ce = parent->find_symb_ce_for_port(circuit_port);
      if (!symb_ce && share_channel_ends) {
        symb_ce=ce_copy;
      }
      
      if (!symb_ce) {
        // generate new symb_ce
        symb_ce=addPort(out_port);
      } else {
        // add port to node
        addPort(out_port, symb_ce);
      }
      parent->register_symb_ce_for_port(circuit_port, symb_ce);

      if (ce_copy != symb_ce) {
        // we have to ensure copy semantics between ce_copy and symb_ce

        if (build_context->param_abstract_data_domain) {
          // abstract data domain: only activity bits active
          copy_semantics&=ce_copy->getPortActive() & symb_ce->getPortActive();
        } else {
          // standard
          copy_semantics&=ce_copy->equals(*symb_ce);
        }
      }
      TRACE_BDD_FUNCTION("SymbNode.build", copy_semantics, circuit_port->getName()+", "+out_port->getName());
    }
    
    TRACE_BDD_FUNCTION("SymbNode.build", copy_semantics, "[result]");
    
    BDDFunction no_in_port_active=BDDFunction::ONE();
    TRACE_BDD_FUNCTION("SymbNode.build", no_in_port_active, "");
    
    // add in ports
    for (unsigned int i=0;i<in.size();i++) {
      PortDescription::ptr circuit_port=in.at(i);
      
      PortDescription::ptr in_port(new PortDescription("in"+boost::lexical_cast<string>(i),
                                                       PortDescription::IN_PORT,
                                                       dataDomain->getDataDomain()));
      info->c_interface->addPort(in_port);

      Symb_Channel_End::ptr symb_ce=parent->find_symb_ce_for_port(circuit_port);
      if (!symb_ce && in.size()==1 && share_channel_ends) {
        // only one in port and does not have a symb channel end -> put copy port
        symb_ce=ce_copy;
      }

      if (!symb_ce) {
        // generate new symb_ce
        symb_ce=addPort(in_port);        
      } else {
        // add symb_ce to node port
        addPort(in_port, symb_ce);
      }
      parent->register_symb_ce_for_port(circuit_port, symb_ce);

      
      Symb_Channel_End::ptr ce_in=info->port_to_bdd[in_port];
      no_in_port_active&=!ce_in->getPortActive();
      TRACE_BDD_FUNCTION("SymbNode.build", no_in_port_active, in_port->getName());
    }
    
    for (unsigned int i=0;i<in.size();i++) {
      PortDescription::ptr in_port=info->c_interface->getInPort(i);
      Symb_Channel_End::ptr ce_in=info->port_to_bdd[in_port];
      
      BDDFunction only_this_port_active=no_in_port_active.exists(ce_in->getPortActive());
      TRACE_BDD_FUNCTION("SymbNode.build", only_this_port_active, in_port->getName());
      
      if (build_context->param_abstract_data_domain) {
        // abstract data domain
        delta |= ce_copy->getPortActive() & ce_in->getPortActive() & no_state_changes & only_this_port_active;
      } else {
        // standard
        delta |= ce_copy->equals(*ce_in) & no_state_changes & only_this_port_active;
      }
      TRACE_BDD_FUNCTION("SymbNode.build", delta, in_port->getName());
    }

    if (in.size()==0) {
      // no in CEs -> no input restrictions, but copy port has to be active
      delta=ce_copy->getPortActive();
    }
    
    delta&=copy_semantics;
    TRACE_BDD_FUNCTION("SymbNode.build", delta, "");

    CHECK(5, check_symb_ce_sanity(), "");
  }




  void SymbNode::build_route_node(SymbComponentInstance::ptr parent,
                                  Symb_DataDomain::ptr dataDomain,
                                  const vector<PortDescription::ptr>& in,
                                  const vector<PortDescription::ptr>& out,
                                  PortDescription::ptr node_port,
                                  bool share_channel_ends,
                                  VarPositions::ptr var_positions_,
                                  BuildContext::ptr build_context) {
    //TODO: get rid of this nasty write on protected variable
    var_positions = var_positions_;
    
#ifdef WITH_DUMMY_STATE
    addDummyState();
    initialStates==!dummy_state_var.X->at(0);
#endif

    delta=BDDFunction::ZERO();

    CHECK_SUCCESS(node_port);


    Symb_Channel_End::ptr ce_node=parent->find_symb_ce_for_port(node_port);

    if (!ce_node) {
      // generate a port for the node
      ce_node=addPort(node_port);
    }

    addPort(node_port, ce_node);
    parent->register_symb_ce_for_port(node_port, ce_node);

    ce_port=ce_node;
    
    // -------------- the in ports ------------------------------

    BDDFunction in_choice=BDDFunction::ZERO();

    BDDFunction no_in_port_active=BDDFunction::ONE();
    TRACE_BDD_FUNCTION("SymbNode.build", no_in_port_active, "");
    
    // add in ports
    for (unsigned int i=0;i<in.size();i++) {
      PortDescription::ptr circuit_port=in.at(i);
      
      PortDescription::ptr in_port(new PortDescription("in"+boost::lexical_cast<string>(i),
                                                       PortDescription::IN_PORT,
                                                       dataDomain->getDataDomain()));
      info->c_interface->addPort(in_port);
      
      Symb_Channel_End::ptr symb_ce=parent->find_symb_ce_for_port(circuit_port);
      if (!symb_ce && in.size()==1 && share_channel_ends) {
          // only one in port and does not have a symb channel end -> put node port
        symb_ce=ce_node;
      }

      if (!symb_ce) {
        // generate new symb_ce
        symb_ce=addPort(in_port);        
      } else {
        // add symb_ce to node port
        addPort(in_port, symb_ce);
      }
      parent->register_symb_ce_for_port(circuit_port, symb_ce);
      
      Symb_Channel_End::ptr ce_in=info->port_to_bdd[in_port];
      no_in_port_active&=!ce_in->getPortActive();
      TRACE_BDD_FUNCTION("SymbNode.build", no_in_port_active, in_port->getName());
    }
    
    for (unsigned int i=0;i<in.size();i++) {
      PortDescription::ptr in_port=info->c_interface->getInPort(i);
      Symb_Channel_End::ptr ce_in=info->port_to_bdd[in_port];
      
      BDDFunction only_this_port_active=no_in_port_active.exists(ce_in->getPortActive());
      TRACE_BDD_FUNCTION("SymbNode.build", only_this_port_active, in_port->getName());
      
      if (build_context->param_abstract_data_domain) {
        // abstract, ignore data domain, both ports active
        in_choice |= ce_node->getPortActive() & ce_in->getPortActive() & no_state_changes & only_this_port_active;
      } else {
        // standard        
        in_choice |= ce_node->equals(*ce_in) & no_state_changes & only_this_port_active;
      }
      TRACE_BDD_FUNCTION("SymbNode.build", delta, in_port->getName());
    }

    

    // -------------- the out ports ------------------------------

    BDDFunction out_choice=BDDFunction::ZERO();

    BDDFunction no_out_port_active=BDDFunction::ONE();
    TRACE_BDD_FUNCTION("SymbNode.build", no_out_port_active, "");
    
    // add out ports
    for (unsigned int i=0;i<out.size();i++) {
      PortDescription::ptr circuit_port=out.at(i);
      
      PortDescription::ptr out_port(new PortDescription("out"+boost::lexical_cast<string>(i),
                                                        PortDescription::OUT_PORT,
                                                        dataDomain->getDataDomain()));
      info->c_interface->addPort(out_port);

      Symb_Channel_End::ptr symb_ce=parent->find_symb_ce_for_port(circuit_port);
      if (!symb_ce && out.size()==1 && share_channel_ends) {
        // only one out port and does not have a symb channel end -> put node port
        symb_ce=ce_node;
      }

      if (!symb_ce) {
        symb_ce=addPort(out_port);        
      } else {
        addPort(out_port, symb_ce);
      }
      parent->register_symb_ce_for_port(circuit_port, symb_ce);      

      Symb_Channel_End::ptr ce_out=info->port_to_bdd[out_port];
      no_out_port_active&=!ce_out->getPortActive();
      TRACE_BDD_FUNCTION("SymbNode.build", no_out_port_active, out_port->getName());
    }
    
    for (unsigned int i=0;i<out.size();i++) {
      PortDescription::ptr out_port=info->c_interface->getOutPort(i);
      Symb_Channel_End::ptr ce_out=info->port_to_bdd[out_port];
      
      BDDFunction only_this_port_active=no_out_port_active.exists(ce_out->getPortActive());
      TRACE_BDD_FUNCTION("SymbNode.build", only_this_port_active, out_port->getName());

      if (build_context->param_abstract_data_domain) {
        out_choice |= ce_node->getPortActive() & ce_out->getPortActive() & no_state_changes & only_this_port_active;
      } else {
        // standard
        out_choice |= ce_node->equals(*ce_out) & no_state_changes & only_this_port_active;
      }
      TRACE_BDD_FUNCTION("SymbNode.build", delta, out_port->getName());
    }
    
    if (in.size()==0) {
      in_choice=BDDFunction::ONE();
    }

    if (out.size()==0) {
      out_choice=BDDFunction::ONE();
    }

    
    if (in.size()==0 && out.size()==0) {
      // no in and out CEs -> no input and output restrictions, but node port has to be active
      delta=ce_node->getPortActive();
    } else {
      // normal construction:
      delta=in_choice & out_choice;
    }
    
    TRACE_BDD_FUNCTION("SymbNode.build", delta, "");
    CHECK(5, check_symb_ce_sanity(), "");
  }


  
  void SymbNode::addAugmentation(BDDFunction augmentation) {
    delta&=augmentation;
  }
  
  Symb_Channel_End::ptr SymbNode::getCEforPort() const {
    return ce_port;
  }
}
