/***************************************************************************
 *   Copyright (C) 2009 by Joachim Klein                                   *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include <sstream>
#include "symb_struct_datadomain.hpp"
#include "symb_function_constructor.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/common/xml_helpers.hpp"

namespace vereofy {

  Symb_StructDataDomain::Symb_StructDataDomain(StructDataDomain::ptr data_domain) :
    Symb_DataDomain(data_domain),
    logsize(0) {

    for (unsigned int i=0;i<data_domain->getNumberOfMembers();i++) {
      DataDomain::ptr member_domain=data_domain->getMemberType(i);
      Symb_DataDomain::ptr symb_member_domain=Symb_DataDomain::getFor(member_domain);
      encoding_info info;
      info.start_bit=logsize;
      info.length=symb_member_domain->getLogSize();
      info.symb_datadomain=symb_member_domain;
      encodings.push_back(info);
      
      logsize+=info.length;
    }
  }

  unsigned int Symb_StructDataDomain::getLogSize() const {
    return logsize;
  }

  double Symb_StructDataDomain::countValidValues() const {
    double result=1;

    for (unsigned int i=0;i<getStructDataDomain()->getNumberOfMembers();i++) {
      DataDomain::ptr member_domain=getStructDataDomain()->getMemberType(i);
      Symb_DataDomain::ptr symb_member_domain=Symb_DataDomain::getFor(member_domain);

      result*=symb_member_domain->countValidValues();
    }

    return result;
  }

  BDDFunction Symb_StructDataDomain::restrictToValidValues(const BDDFunctionContainer& encoding) const {
    BDDFunction result=BDDFunction::ONE();
    for (unsigned int i=0;i<encodings.size();i++) {
      result&=encodings.at(i).symb_datadomain->restrictToValidValues(getEncodingForMember(encoding, i));
    }
    return result;
  }

  StructDataDomain::ptr Symb_StructDataDomain::getStructDataDomain() const {
    return boost::static_pointer_cast<StructDataDomain>(getDataDomain());
  }

  Symb_DataDomain::ptr Symb_StructDataDomain::getSymbDataDomainForMember(unsigned int index) const {
    return encodings.at(index).symb_datadomain;
  }

  BDDFunctionContainer Symb_StructDataDomain::getEncodingForMember(const BDDFunctionContainer& encoding,
                                                                   unsigned int index) const {
    CHECK(1, encoding.size()==getLogSize(), "");
    if (index>=getStructDataDomain()->getNumberOfMembers()) {
      throw "member index out of range";
    }

    unsigned int start_bit=encodings.at(index).start_bit;
    unsigned int length=encodings.at(index).length;

    BDDFunctionContainer result;
    for (unsigned int i=0;i<length;i++) {
      CHECK(1, i+start_bit<encoding.size(), "");
      result.push(encoding[i+start_bit]);
    }

    return result;
  }
  
  string Symb_StructDataDomain::reverseEncoding(const BDDFunction& assignment, const BDDFunctionContainer& vars) const {
    stringstream out;
    
    out << "{";
    for (unsigned int i=0;i<getStructDataDomain()->getNumberOfMembers();i++) {
      if (i!=0) {
	out << ",";
      }

      BDDFunctionContainer i_encoding=getEncodingForMember(vars, i);
      out << encodings.at(i).symb_datadomain->reverseEncoding(assignment, i_encoding);
    }
    out << "}";
    return out.str();
  }

  void Symb_StructDataDomain::reverseEncodingToXML(std::ostream& out, 
                                                   const BDDFunction& assignment,
                                                   const BDDFunctionContainer& vars) const {
    xml_tag_open(out, "value");

    for (unsigned int i=0;i<getStructDataDomain()->getNumberOfMembers();i++) {
      xml_tag_open(out, "element", xml_parameter("name", getStructDataDomain()->getMemberName(i)));

      BDDFunctionContainer i_encoding=getEncodingForMember(vars, i);
      getSymbDataDomainForMember(i)->reverseEncodingToXML(out, assignment, i_encoding);

      xml_tag_close(out, "element");
    }

    xml_tag_close(out, "value");
  }

  void Symb_StructDataDomain::reverseEncodingToGuards(const BDDFunction& assignment, 
                                                      const BDDFunctionContainer& vars,
                                                      const std::string& prefix,
                                                      std::vector<std::string> &result) const {
    for (unsigned int i=0;i<getStructDataDomain()->getNumberOfMembers();i++) {
      std::string new_prefix=prefix+"."+getStructDataDomain()->getMemberName(i);

      BDDFunctionContainer i_encoding=getEncodingForMember(vars, i);
      getSymbDataDomainForMember(i)->reverseEncodingToGuards(assignment, i_encoding,
                                                             new_prefix,
                                                             result);
    }
  }

  BDDFunction Symb_StructDataDomain::getEquivalent(const BDDFunction& assignments, 
                                                   const BDDFunctionContainer& encoding) const {
    BDDFunction result=assignments;
    for (unsigned int i=0;i<getStructDataDomain()->getNumberOfMembers();i++) {
      result = getSymbDataDomainForMember(i)->getEquivalent(result, getEncodingForMember(encoding, i));
    }
    return result;
  }  

  

  class SymbFunction_Constructor_Struct : public SymbFunction_Constructor {
  public:
    typedef boost::shared_ptr<SymbFunction_Constructor_Struct> ptr;
    
    SymbFunction_Constructor_Struct(Symb_StructDataDomain::ptr dd) : 
      SymbFunction_Constructor(dd) {}
      
    virtual SymbExpression::ptr calculate(SymbFunctionArgumentVector& arguments) const {
      checkArguments(arguments);

      BDDFunctionContainer values;
      BDDFunction condition = BDDFunction::ONE();

      Symb_StructDataDomain::ptr struct_dd=
        boost::static_pointer_cast<Symb_StructDataDomain>(symb_dd);

      for (unsigned int i=0;i<arguments.size();i++) {
        SymbExpression::ptr arg=arguments.at(i).getSymbExpression();
        
        SymbExpression cast_arg=struct_dd->getSymbDataDomainForMember(i)->castExpression(*arg);
        values.push(cast_arg.getValues());
        condition&=cast_arg.getCondition();
      }
      
      CHECK(1, values.size()==struct_dd->getLogSize(), "");

      return SymbExpression::ptr(new SymbExpression(symb_dd, values, condition));
    }

    virtual Symb_DataDomain::ptr getExpectedArgumentType(unsigned int index,
                                                         const SymbFunctionArgumentVector& previous_arguments) const {
      Symb_StructDataDomain::ptr struct_dd=
        boost::static_pointer_cast<Symb_StructDataDomain>(symb_dd);

      return struct_dd->getSymbDataDomainForMember(index);
    }

  };


  boost::shared_ptr<SymbFunction_Constructor> Symb_StructDataDomain::getConstructorFunction() {
    boost::shared_ptr<Symb_StructDataDomain> sdd=
      boost::static_pointer_cast<Symb_StructDataDomain>(this->shared_from_this());
    
    return SymbFunction_Constructor::ptr(new SymbFunction_Constructor_Struct(sdd));
  }
  
}
