/***************************************************************************
 *   Copyright (C) 2009 by Joachim Klein                                   *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include <sstream>
#include "symb_tagged_union_datadomain.hpp"
#include "symb_expression.hpp"
#include "symb_function_constructor.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/common/xml_helpers.hpp"

namespace vereofy {

  Symb_TaggedUnionDataDomain::Symb_TaggedUnionDataDomain(TaggedUnionDataDomain::ptr data_domain) :
    Symb_DataDomain(data_domain),
    logsize(0) {

    tag_domain=boost::static_pointer_cast<Symb_FlatDataDomain>(Symb_DataDomain::getFor(getTaggedUnionDataDomain()->getTag()));

    for (unsigned int i=0;i<data_domain->getNumberOfMembers();i++) {
      DataDomain::ptr member_domain=data_domain->getMemberType(i);
      Symb_DataDomain::ptr symb_member_domain=Symb_DataDomain::getFor(member_domain);
      unsigned int length=symb_member_domain->getLogSize();

      encodings.push_back(symb_member_domain);

      // calculate max of member lengths
      if (length>logsize) {
        logsize=length;
      }
    }

    logsize+=tag_domain->getLogSize();
  }

  unsigned int Symb_TaggedUnionDataDomain::getLogSize() const {
    return logsize;
  }

  double Symb_TaggedUnionDataDomain::countValidValues() const {
    double result=0;

    for (unsigned int i=0;i<getTaggedUnionDataDomain()->getNumberOfMembers();i++) {
      DataDomain::ptr member_domain=getTaggedUnionDataDomain()->getMemberType(i);
      Symb_DataDomain::ptr symb_member_domain=Symb_DataDomain::getFor(member_domain);

      result+=symb_member_domain->countValidValues();
    }

    return result;
  }

  BDDFunction Symb_TaggedUnionDataDomain::restrictToValidValues(const BDDFunctionContainer& encoding) const {
    BDDFunction result=BDDFunction::ZERO();
    for (unsigned int i=0;i<encodings.size();i++) {
      BDDFunction restrict_for_member=encodings.at(i)->restrictToValidValues(getEncodingForMember(encoding, i));

      result|= isMember(i, encoding) & restrict_for_member;
    }
    return result;
  }

  TaggedUnionDataDomain::ptr Symb_TaggedUnionDataDomain::getTaggedUnionDataDomain() const {
    return boost::static_pointer_cast<TaggedUnionDataDomain>(getDataDomain());
  }

  Symb_FlatDataDomain::ptr Symb_TaggedUnionDataDomain::getSymbDataDomainForTag() const {
    return tag_domain;
  }

  Symb_DataDomain::ptr Symb_TaggedUnionDataDomain::getSymbDataDomainForMember(unsigned int index) const {
    return encodings.at(index);
  }

  BDDFunctionContainer Symb_TaggedUnionDataDomain::getEncodingForTag(const BDDFunctionContainer& encoding) const {
    CHECK(1, encoding.size()==getLogSize(), "");

    unsigned int start_bit=0;
    unsigned int length=tag_domain->getLogSize();

    BDDFunctionContainer result;
    for (unsigned int i=0;i<length;i++) {
      CHECK(1, i+start_bit<encoding.size(), "");
      result.push(encoding[i+start_bit]);
    }

    return result;
  }

  BDDFunctionContainer Symb_TaggedUnionDataDomain::getEncodingForMember(const BDDFunctionContainer& encoding,
                                                                        unsigned int index) const {
    CHECK(1, encoding.size()==getLogSize(), "");
    if (index>=getTaggedUnionDataDomain()->getNumberOfMembers()) {
      throw "member index out of range";
    }

    unsigned int start_bit=tag_domain->getLogSize();
    unsigned int length=encodings.at(index)->getLogSize();

    BDDFunctionContainer result;
    for (unsigned int i=0;i<length;i++) {
      CHECK(1, i+start_bit<encoding.size(), "");
      result.push(encoding[i+start_bit]);
    }

    return result;
  }
  
  BDDFunction Symb_TaggedUnionDataDomain::isMember(unsigned int index, const BDDFunctionContainer& encoding) const {
    return tag_domain->performComparison(COMPARE_EQUAL,
                                         SymbExpression(tag_domain, getEncodingForTag(encoding)),
                                         SymbExpression(tag_domain, tag_domain->getEncoding(index)));
  }
  
  SymbExpression Symb_TaggedUnionDataDomain::accessMember(const std::string& name, const BDDFunctionContainer& encoding) const {
    TaggedUnionDataDomain::ptr dd=getTaggedUnionDataDomain();

    if (name == dd->getTagName()) {
      return SymbExpression(getSymbDataDomainForTag(),
                            getEncodingForTag(encoding));
    }

    if (!dd->isMember(name)) {
      throw name+" is not a valid member of this tagged union!";
    }
    
    unsigned int member_index=dd->getIndexOfMember(name);
      
    Symb_DataDomain::ptr member_dd=getSymbDataDomainForMember(member_index);
    CHECK_SUCCESS(member_dd);
    
    return SymbExpression(member_dd, 
                          getEncodingForMember(encoding, member_index),
                          isMember(member_index, encoding));
  }
  
  string Symb_TaggedUnionDataDomain::reverseEncoding(const BDDFunction& assignment, const BDDFunctionContainer& vars) const {
    stringstream out;
    
    out << "{";

    string member_name=tag_domain->reverseEncoding(assignment, getEncodingForTag(vars));
    unsigned int tag_index = getTaggedUnionDataDomain()->getTag()->getIndexForElement(member_name);

    out << member_name << " : ";

    BDDFunctionContainer i_encoding=getEncodingForMember(vars, tag_index);
    out << encodings.at(tag_index)->reverseEncoding(assignment, i_encoding);

    out << "}";
    return out.str();
  }

  void Symb_TaggedUnionDataDomain::reverseEncodingToXML(std::ostream& out, 
                                                   const BDDFunction& assignment,
                                                   const BDDFunctionContainer& vars) const {

    string member_name=tag_domain->reverseEncoding(assignment, getEncodingForTag(vars));
    unsigned int tag_index = getTaggedUnionDataDomain()->getTag()->getIndexForElement(member_name);

    xml_tag_open(out, "value", xml_parameter("tagged_union_tag", member_name));

    xml_tag_open(out, "element", xml_parameter("name", member_name));

    BDDFunctionContainer i_encoding=getEncodingForMember(vars, tag_index);
    getSymbDataDomainForMember(tag_index)->reverseEncodingToXML(out, assignment, i_encoding);

    xml_tag_close(out, "element");
    xml_tag_close(out, "value");
  }


  void Symb_TaggedUnionDataDomain::reverseEncodingToGuards(const BDDFunction& assignment, 
                                                           const BDDFunctionContainer& vars,
                                                           const std::string& prefix,
                                                           std::vector<std::string>& result) const {
    string member_name=tag_domain->reverseEncoding(assignment, getEncodingForTag(vars));
    unsigned int tag_index = getTaggedUnionDataDomain()->getTag()->getIndexForElement(member_name);
    
    std::string new_prefix=prefix+"."+member_name;
    BDDFunctionContainer i_encoding=getEncodingForMember(vars, tag_index);
    getSymbDataDomainForMember(tag_index)->reverseEncodingToGuards(assignment, i_encoding,
                                                                   new_prefix,
                                                                   result);
  }

  // returns cubeset with the variables that are not used for a given member
  BDDFunction Symb_TaggedUnionDataDomain::getUnusedEncodingForMember(const BDDFunctionContainer& encoding, unsigned int index) const {
    CHECK(1, encoding.size()==getLogSize(), "");
    if (index>=getTaggedUnionDataDomain()->getNumberOfMembers()) {
      throw "member index out of range";
    }

    unsigned int start_bit=tag_domain->getLogSize();
    unsigned int used_length=encodings.at(index)->getLogSize();

    BDDFunction unused=BDDFunction::ONE();
    for (unsigned int i=start_bit+used_length;i<getLogSize();i++) {
      CHECK(1, i<encoding.size(), "");
      unused &= encoding.at(i);
    }

    return unused;    
  }


  BDDFunction Symb_TaggedUnionDataDomain::getEquivalent(const BDDFunction& assignments, 
                                                        const BDDFunctionContainer& encoding) const {
    if (getTaggedUnionDataDomain()->getNumberOfMembers()==0) {
      return assignments;
    }

    BDDFunction result=BDDFunction::ZERO();
    for (unsigned int i=0;i<getTaggedUnionDataDomain()->getNumberOfMembers();i++) {
        BDDFunction assignments_for_this_member=isMember(i, encoding) & assignments;
        if (assignments_for_this_member==BDDFunction::ZERO()) {
          continue; 
        }
        result |= assignments_for_this_member.exists(getUnusedEncodingForMember(encoding, i));
    }

    return result;
  }  


  class SymbFunction_Constructor_TaggedUnion : public SymbFunction_Constructor {
  public:
    typedef boost::shared_ptr<SymbFunction_Constructor_TaggedUnion> ptr;
    
    SymbFunction_Constructor_TaggedUnion(Symb_TaggedUnionDataDomain::ptr dd) : 
      SymbFunction_Constructor(dd) {}

    virtual SymbExpression::ptr calculate(SymbFunctionArgumentVector& arguments) const {
      checkArguments(arguments);

      Symb_TaggedUnionDataDomain::ptr tagged_union_dd=
        boost::static_pointer_cast<Symb_TaggedUnionDataDomain>(symb_dd);

      BDDFunctionContainer values;
      BDDFunction condition = BDDFunction::ONE();

      unsigned int member_index=getMemberIndex(arguments);
      values.push(tagged_union_dd->getSymbDataDomainForTag()->getEncoding(member_index));

      SymbExpression::ptr value=arguments.at(1).getSymbExpression();
      SymbExpression value_cast=tagged_union_dd->getSymbDataDomainForMember(member_index)->castExpression(*value);
      values.push(value_cast.getValues());
      condition&=value_cast.getCondition();

      // fill container to logSize (unused bits)
      for (unsigned int i=values.size(); i<symb_dd->getLogSize(); i++) {
        values.push(BDDFunction::ONE());
      }
      
      return SymbExpression::ptr(new SymbExpression(symb_dd, values, condition));
    }

    virtual Symb_DataDomain::ptr getExpectedArgumentType(unsigned int index,
                                                         const SymbFunctionArgumentVector& previous_arguments) const {
      Symb_TaggedUnionDataDomain::ptr tagged_union_dd=
        boost::static_pointer_cast<Symb_TaggedUnionDataDomain>(symb_dd);

      if (index==0) {
        return tagged_union_dd->getSymbDataDomainForTag();
      } else if (index==1) {
        unsigned int member_index=getMemberIndex(previous_arguments);
        return tagged_union_dd->getSymbDataDomainForMember(member_index);
      } 
      
      throw "implementation error";
    }

  private:
    unsigned int getMemberIndex(const SymbFunctionArgumentVector& arguments) const {
      SymbExpression::ptr tag=arguments.at(0).getSymbExpression();
      CHECK_SUCCESS(tag->isConstant());

      Symb_TaggedUnionDataDomain::ptr tagged_union_dd=
        boost::static_pointer_cast<Symb_TaggedUnionDataDomain>(symb_dd);

      unsigned int member_index=tagged_union_dd->getSymbDataDomainForTag()->reverseConstExpression(*tag);
      return member_index;
    }
  };


  boost::shared_ptr<SymbFunction_Constructor> Symb_TaggedUnionDataDomain::getConstructorFunction() {
    boost::shared_ptr<Symb_TaggedUnionDataDomain> sdd=
      boost::static_pointer_cast<Symb_TaggedUnionDataDomain>(this->shared_from_this());
    
    return SymbFunction_Constructor::ptr(new SymbFunction_Constructor_TaggedUnion(sdd));
  }


};
