/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/



#include "symb_trace.hpp"
#include "symb_component_instance.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/common/xml_helpers.hpp"

namespace vereofy {

  SymbTrace::SymbTrace() {
  }


  void SymbTrace::appendTrace(const SymbTrace& trace) {
    if (trace.countStates()==0) {
      // nothing to do
      return;
    }
    
    if (this->countStates()>0) {
      // check that the first state of trace agrees with our last state 
      SymbState first=trace.getState(0);
      ERROR_CHECK_ASSERT(first == this->getLastState());
    }
    
    states.insert(states.end(), ++trace.states.begin(), trace.states.end());
    labels.insert(labels.end(), trace.labels.begin(), trace.labels.end());
  }

  void SymbTrace::appendTransition(const SymbTransition& transition) {
    SymbState from=transition.getFromState();
    SymbState to=transition.getToState();
    SymbTransitionLabel label=transition.getLabel();
    
    if (states.empty()) {
      states.push_back(from);
    } else {
      // check that the from state of transition agrees with our last state 
      ERROR_CHECK_ASSERT(from == this->getLastState());
    }
    
    labels.push_back(label);
    states.push_back(to);
  }

  void SymbTrace::appendState(const SymbState& state) {
    ERROR_CHECK_ASSERT(states.size()==0);
    states.push_back(state);
  }

  unsigned int SymbTrace::countStates() const {
    return states.size();
  }

  unsigned int SymbTrace::countLabels() const {
    return labels.size();
  }

  unsigned int SymbTrace::countTransitions() const {
    unsigned int states=countStates();
    return (states==0) ? 0 : countStates()-1;
  }

  const SymbState& SymbTrace::getState(unsigned int index) const {
    return states.at(index);
  }

  const SymbState& SymbTrace::getLastState() const {
    return states.at(countStates()-1);
  }
  
  const SymbTransitionLabel& SymbTrace::getLabel(unsigned int index) const {
    return labels.at(index);
  }

  SymbTransition SymbTrace::getTransition(unsigned int index) const {
    return SymbTransition(getState(index), getState(index+1), getLabel(index));
  }

  BDDFunction SymbTrace::getAllStates() const {
    BDDFunction result=BDDFunction::ZERO();
    for (unsigned int i=0;i<countStates();i++) {
      result|=getState(i).getX();
    }
    return result;
  }
    
  BDDFunction SymbTrace::getAllTransitions() const {
    BDDFunction result=BDDFunction::ZERO();
    for (unsigned int i=0;i<countTransitions();i++) {
      result|=getTransition(i).getTransitionFunction();
    }
    return result;
  }
  
  void SymbTrace::print_xml(SymbComponentInstance& system, std::ostream& out) const {
    xml_tag_open(out, "tracefragment", xml_parameter("type", ""));
    out << "\n";
    system.print_xml(out, *this);
    xml_tag_close(out, "tracefragment");
    out << "\n";
  }

  
}
