/***************************************************************************
 *   Copyright (C) 2009 by Joachim Klein                                   *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "symb_trace_explorer.hpp"

using namespace vereofy;

/** Sets the progress flag to a new value.
 *  The presence of this flag indicates that the user of SymbTraceExplorer
 *  would prefer SuccessorStates to be chosen in a more directed way. The 
 *  meaning of this depends on the actual implementation, but might be 
 *  successor states that are on a shortesst trace to a demonstration
 *  that the formula is violated.
 */
void SymbTraceExplorer::setProgressFlag(bool progress_) {
  progress=progress_;
}

/** Returns the value of the progress flag */
bool SymbTraceExplorer::isProgressFlag() const {
  return progress;
}
    
/** Returns a reference to the currently active state.
 *  Throws exception if no initial state has been chosen */
const SymbState& SymbTraceExplorer::getCurrentState() const {
  return trace.getLastState();
}
    
/** Returns a reference to the current trace */
const SymbTrace& SymbTraceExplorer::getCurrentTrace() const {
  return trace;
}

/** Chose initial as the initial state of the trace.
 *  initial becomes the current state.
 */
void SymbTraceExplorer::chooseInitialState(const SymbState& initial) {
  ERROR_CHECK_ASSERT(trace.countStates()==0);

  // check that this is a valid initial state
  CHECK(3, getInitialStates().contains(initial), "");

  trace.appendState(initial);
}


/** Choose the transition as the next step.
 *  Precondition: transition.getFromState()==getCurrentState()
 *  Postcondition: getCurrentState()==transition.getToState()
 */
void SymbTraceExplorer::chooseNextStep(const SymbTransition& transition) {
  ERROR_CHECK_ASSERT(transition.getFromState() == getCurrentState());
  CHECK(3, getSuccessorStates().contains(transition.getToState()), "");
  CHECK(3, getTransitionSetToState(transition.getToState(), false).contains(transition), "");

  trace.appendTransition(transition);
}


// -------------- Helper functions --------------------------------------

/** Returns the projection on the underlying main automaton state 
 *  (i.e. removes the state variables from a product construction that don't belong
 *   to the main automaton from the cubeset). */
SymbState SymbTraceExplorer::canonicalProjection(const SymbState& state) const {
  // default implementation: return identity
  return state;
}

void SymbTraceExplorer::randomTrace(unsigned int steps) {
  for (unsigned int i=0;i<steps;i++) {
    if (trace.countStates()==0) {
      // choose an initial state
      SymbStateSet initialStates=getInitialStates();
      TRACE_BDD_FUNCTION("randomTrace", initialStates.getX(), "");

      if (initialStates.isEmpty()) {return;}

      SymbState state=initialStates.randomState();
      TRACE_BDD_FUNCTION("randomTrace", state.getX(), "");

      chooseInitialState(state);
    } else {
      SymbStateSet states=getSuccessorStates();
      TRACE_BDD_FUNCTION("randomTrace", states.getX(), "");

      if (states.isEmpty()) {return;}

      SymbState state=states.randomState();
      TRACE_BDD_FUNCTION("randomTrace", state.getX(), "");

      SymbTransitionSet transitions=getTransitionSetToState(state, true);
      TRACE_BDD_FUNCTION("randomTrace", transitions.getDelta(), "");

      if (transitions.isEmpty()) {return;}

      SymbTransition transition=transitions.randomTransition();
      TRACE_BDD_FUNCTION("randomTrace", transition.getTransitionFunction(), "");

      chooseNextStep(transition);
    }
  }
}
