/***************************************************************************
 *   Copyright (C) 2009 by Joachim Klein                                   *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef SYMB_TRACE_EXPLORER_HPP
#define SYMB_TRACE_EXPLORER_HPP

#include "symb_state_set.hpp"
#include "symb_transition_set.hpp"
#include "symb_trace_lasso.hpp"
#include "vereofy/common/return_value.hpp"
#include <iostream>
#include <boost/shared_ptr.hpp>

namespace vereofy {

  /**
   * SymbTraceExplorer is a virtual base class providing a standardized interface 
   * to allow state space/trace exploration in symbolic automata or subsets of 
   * traces.
   * Generation of a trace works by first choosing an initial state and 
   * then choosing from the available successors of the current state (and
   * the transition labels).
   */
  class SymbTraceExplorer {
  public:
    typedef boost::shared_ptr<SymbTraceExplorer> ptr;

    struct not_supported_exception {};

    SymbTraceExplorer() : progress(false) {}
    virtual ~SymbTraceExplorer() {}
    
    /** Sets the progress flag to a new value.
     *  The presence of this flag indicates that the user of SymbTraceExplorer
     *  would prefer SuccessorStates to be chosen in a more directed way. The 
     *  meaning of this depends on the actual implementation, but might be 
     *  successor states that are on a shortesst trace to a demonstration
     *  that the formula is violated.
     */
    virtual void setProgressFlag(bool progress);

    /** Returns the value of the progress flag */
    virtual bool isProgressFlag() const;
    
    /** Returns a reference to the currently active state.
     *  Throws exception if no initial state has been chosen */
    virtual const SymbState& getCurrentState() const;
    
    /** Returns a reference to the current trace */
    virtual const SymbTrace& getCurrentTrace() const;




    /** Get a state set of possible starting states */
    virtual SymbStateSet getInitialStates() const = 0;

    /** Chose initial as the initial state of the trace.
     *  initial becomes the current state.
     */
    virtual void chooseInitialState(const SymbState& initial);


    /** Get the set of possible successor states for the current state */
    virtual SymbStateSet getSuccessorStates() const = 0;

    /** Get the set of possible successor states for the current state at depth steps 
     *  (all states reachable via a path of length steps) */
    virtual SymbStateSet getSuccessorStates(unsigned int steps) const = 0;
    
    /** Get the set of transitions between the current state and the to state
     *  If canonical_transitions is true, the transition set will contain
     *  exactly one assignment per transition
     */
    virtual SymbTransitionSet getTransitionSetToState(const SymbState& to, 
                                                      bool canonical_transitions) const = 0;

    /** Choose the transition as the next step.
     *  Precondition: transition.getFromState()==getCurrentState()
     *  Postcondition: getCurrentState()==transition.getToState()
     */
    virtual void chooseNextStep(const SymbTransition& transition) ;
    
    /** Returns true if the transition set may change depending on the 
     *  history of the current trace. */
    virtual bool isHistoryDependant() const = 0;

    // -------------- Helper functions --------------------------------------

    /** Get the set of all valid transitions
     *  This function may throw not_supported_exception if it is not supported. */
    virtual SymbTransitionSet getTransitionSet() const = 0;

    /** Returns the projection on the underlying main automaton state 
     *  (i.e. removes the state variables from a product construction that don't belong
     *   to the main automaton from the cubeset). */
    virtual SymbState canonicalProjection(const SymbState& state) const;

    /** Get an infinite trace starting from state from.
     *  This function may throw not_supported_exception if it is not supported.
     *  If there is no infinite trace, the ReturnValue is marked as invalid.
     */
    virtual ReturnValue<SymbTraceLasso::ptr> getInfiniteTrace(const SymbState& from) const = 0;
    
    /** Get an infinite trace extending the current trace.
     *  This function may throw not_supported_exception if it is not supported.
     *  If there is no infinite trace, the ReturnValue is marked as invalid.
     */
    virtual ReturnValue<SymbTraceLasso::ptr> extendToInfiniteTrace() const = 0;

    void randomTrace(unsigned int steps);
   
  private:
    bool progress;
    SymbTrace trace;
  };

}

#endif
