/***************************************************************************
 *   Copyright (C) 2009 by Joachim Klein                                   *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "symb_bfs.hpp"
#include "symb_trace_explorer_automaton.hpp"

using namespace vereofy;

SymbTraceExplorerAutomaton::SymbTraceExplorerAutomaton(SymbAutomaton::ptr automaton_) :
  automaton(automaton_) { }
    
/** Get a state set of possible starting states */
SymbStateSet SymbTraceExplorerAutomaton::getInitialStates() const {
  return SymbStateSet(automaton->getInitialStates(), 
                      automaton->getOriginalStateVars());
}

/** Get the set of possible successor states from the current state */
SymbStateSet SymbTraceExplorerAutomaton::getSuccessorStates() const {
  return getSuccessorStates(1);
}

/** Get the set of possible successor states from the current state at depth steps 
 *  (all states reachable via a path of length steps) */
SymbStateSet SymbTraceExplorerAutomaton::getSuccessorStates(unsigned int steps) const {
  CHECK(1, getCurrentState().getCsStateVars() == automaton->getOriginalStateVars(), "");

  SymbBFS symb_bfs(automaton->getDelta(), 
                   automaton->getOriginalStateVars(),
                   automaton->getTransitionVars());

  BDDFunction states=getCurrentState().getX();
  for (unsigned int i=0;i<steps;i++) {
    states=symb_bfs.postImage(states);
  }

  return SymbStateSet(states, automaton->getOriginalStateVars());
}

/** Get the set of transitions from the current state to to
 *  If canonical_transitions is true, the transition set will contain
 *  exactly one assignment per transition
 */
SymbTransitionSet SymbTraceExplorerAutomaton::getTransitionSetToState(const SymbState& to,
                                                                      bool canonical_transitions) const {
  CHECK(1, getCurrentState().getCsStateVars() == automaton->getOriginalStateVars(), "");
  CHECK(1, to.getCsStateVars() == automaton->getOriginalStateVars(), "");

  BDDFunction from_to_delta=automaton->getDelta() & getCurrentState().getX() & to.getY();
  if (canonical_transitions) {
    from_to_delta=automaton->canonicalizeTransitions(from_to_delta);
  }

  return SymbTransitionSet(from_to_delta,
                           automaton->getOriginalStateVars(),
                           automaton->getTransitionVars());
}

/** Returns true if the transition set may change depending on the 
 *  history of the current trace. */
bool SymbTraceExplorerAutomaton::isHistoryDependant() const {
  return false;
}



/** Get the set of all valid transitions */
SymbTransitionSet SymbTraceExplorerAutomaton::getTransitionSet() const {
  return SymbTransitionSet(automaton->getDelta(),
                           automaton->getOriginalStateVars(),
                           automaton->getTransitionVars());
}



/** Get an infinite trace starting from state from.
 *  This function currently throws not_supported_exception for SymbTraceExplorerAutomaton */
ReturnValue<SymbTraceLasso::ptr> SymbTraceExplorerAutomaton::getInfiniteTrace(const SymbState& from) const {
  throw SymbTraceExplorer::not_supported_exception();
}

/** Get an infinite trace extending the current trace.
 *  This function currently throws not_supported_exception for SymbTraceExplorerAutomaton */
ReturnValue<SymbTraceLasso::ptr> SymbTraceExplorerAutomaton::extendToInfiniteTrace() const {
  throw SymbTraceExplorer::not_supported_exception();
}

