/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef SYMB_TRANSITION_HPP
#define SYMB_TRANSITION_HPP 

#include "symb_state.hpp"
#include "symb_transition_label.hpp"

namespace vereofy {
  
  class SymbTransition {
  public:
    typedef boost::shared_ptr<SymbTransition> ptr;
    
    SymbTransition(const BDDFunction& transition_, 
                   BDDFunction csStateVars_,
                   BDDFunction csLabelVars_) :
      transition(transition_), 
      csStateVars(csStateVars_),
      csLabelVars(csLabelVars_) {
      BDD_CHECK2(3, BDDChecks::isAssignment(transition, getCsTransitionVars()),  "", transition, getCsTransitionVars());
    }

    SymbTransition(const SymbState& from,
                   const SymbState& to,
                   const SymbTransitionLabel& label) :
      transition(from.getX() & to.getY() & label.getLabel()),
      csStateVars(from.getCsStateVars()),
      csLabelVars(label.getCsLabelVars()) {
    }
    
    SymbState getFromState() const {
      BDDFunction from_state=
        transition.exists(csLabelVars & csStateVars.moveDown());
      
      return SymbState(from_state, csStateVars);
    }
    
    SymbState getToState() const {
      BDDFunction to_state=
        transition.exists(csLabelVars & csStateVars).moveUp();

      return SymbState(to_state, csStateVars);
    }
  
    SymbTransitionLabel getLabel() const {
      BDDFunction label=
        transition.exists(csStateVars & csStateVars.moveDown());
      
      return SymbTransitionLabel(label, csLabelVars);
    }

    const BDDFunction& getTransitionFunction() const {
        return transition;
    }

    const BDDFunction& getCsStateVars() const {
      return csStateVars;
    }

    const BDDFunction& getCsLabelVars() const {
      return csLabelVars;
    }
    
    const BDDFunction getCsTransitionVars() const {
      return csStateVars & csLabelVars & csStateVars.moveDown();
    }

    static SymbTransition pickTransition(const BDDFunction& delta,
                                         const SymbState& from,
                                         const SymbState& to,
                                         const BDDFunction& csLabelVars) {
      CHECK(1, from.getCsStateVars()==to.getCsStateVars(), "");
      
      BDDFunction csTransitionVars=from.getCsStateVars()
        & to.getCsStateVars().moveDown()
        & csLabelVars;

      BDDFunction delta_restricted=delta & from.getX() & to.getY();
      BDDFunction transition=popAssignment(delta_restricted, csTransitionVars);

      return SymbTransition(transition, 
                            from.getCsStateVars(),
                            csLabelVars);
    }

    static SymbTransition pickTransitionBetweenSets(const BDDFunction& delta,
                                                    const BDDFunction& from_state_set,
                                                    const BDDFunction& to_state_set,
                                                    const BDDFunction& csStateVars,
                                                    const BDDFunction& csLabelVars) {

      TRACE_BDD_FUNCTION("SymbTransition_pickTransitionBetweenSets", delta, "");
      TRACE_BDD_FUNCTION("SymbTransition_pickTransitionBetweenSets", from_state_set, "");
      TRACE_BDD_FUNCTION("SymbTransition_pickTransitionBetweenSets", to_state_set, "");

      BDDFunction csTransitionVars=csStateVars 
        & csStateVars.moveDown()
        & csLabelVars;

      TRACE_BDD_FUNCTION("SymbTransition_pickTransitionBetweenSets", csTransitionVars, "");

      BDDFunction delta_restricted=delta & from_state_set & to_state_set.moveDown();
      TRACE_BDD_FUNCTION("SymbTransition_pickTransitionBetweenSets", delta_restricted, "");

      BDDFunction transition=popAssignment(delta_restricted, csTransitionVars);
      TRACE_BDD_FUNCTION("SymbTransition_pickTransitionBetweenSets", transition, "");

      return SymbTransition(transition, 
                            csStateVars,
                            csLabelVars);
    }


    static SymbTransition popTransition(BDDFunction& delta,
                                        const SymbState& from,
                                        const SymbState& to,
                                        const BDDFunction& csLabelVars) {
      SymbTransition transition=pickTransition(delta, from, to, csLabelVars);
      delta&=!transition.getTransitionFunction();
      return transition;
    }                                      


  private:
    BDDFunction transition;
    BDDFunction csStateVars;
    BDDFunction csLabelVars;
  };
}

#endif
