/***************************************************************************
 *   Copyright (C) 2011 by Joachim Klein                                   *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 @author Joachim Klein
*/

#include <boost/lexical_cast.hpp>
#include <iostream>

#include "symb_transition_label_var.hpp"
#include "varmanager.hpp"

#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/common/weak_ptr_helper.hpp"

using namespace std;

namespace vereofy {

  int Symb_Transition_Label_Var::nr_of_labels = 0;

  Symb_Transition_Label_Var::Symb_Transition_Label_Var(TransitionLabelDescription::wptr label_description_)
    : label_description(label_description_) {

    if (access_weak_ptr(label_description)->hasValueDomain()) {
      dataDomain=Symb_DataDomain::getFor(access_weak_ptr(label_description)->getValueDomain());
    }

    string label_nr = boost::lexical_cast<string>(nr_of_labels++);
    if (hasValueDomain()) {
      BDDFunctionContainer value_vars=VarManager::getInstance()->createVars("#l"+label_nr, dataDomain->getLogSize());
      vecValueVars=value_vars_vector(new BDDFunctionContainer(value_vars));
    }
    label_var = VarManager::getInstance()->createVars("*l"+label_nr, 1).at(0);
  }

  
  Symb_Transition_Label_Var::Symb_Transition_Label_Var(TransitionLabelDescription::wptr label_description_, VarPositions::ptr var_positions_in_out) 
    : label_description(label_description_) {

    if (access_weak_ptr(label_description)->hasValueDomain()) {
      dataDomain=Symb_DataDomain::getFor(access_weak_ptr(label_description)->getValueDomain());
    }

    string label_nr = boost::lexical_cast<string>(nr_of_labels++);
    if (hasValueDomain()) {
      BDDFunctionContainer value_vars=VarManager::getInstance()->createVars("#l"+label_nr, dataDomain->getLogSize(), var_positions_in_out->positions);
      vecValueVars=value_vars_vector(new BDDFunctionContainer(value_vars));
    }
    label_var = VarManager::getInstance()->createVars("*l"+label_nr, 1, var_positions_in_out->positions).at(0);
  }

  Symb_Transition_Label_Var::Symb_Transition_Label_Var(TransitionLabelDescription::wptr label_description_, const BDDFunctionContainer& variables) 
    : label_description(label_description_) {

    if (access_weak_ptr(label_description)->hasValueDomain()) {
      dataDomain=Symb_DataDomain::getFor(access_weak_ptr(label_description)->getValueDomain());
    }
    
    CHECK_SUCCESS( 1+(hasValueDomain() ? dataDomain->getLogSize() : 0) == variables.size());
    
    label_var=variables.at(0);
    vecValueVars=value_vars_vector(new BDDFunctionContainer());
    for (unsigned int i=1;i<variables.size();i++) {
      vecValueVars->push(variables.at(i));
    }
  }

  Symb_Transition_Label_Var::Symb_Transition_Label_Var(TransitionLabelDescription::wptr label_description_,
                                                       BDDFunction label_var_,
                                                       value_vars_vector vecValueVars_) 
    : label_description(label_description_),  label_var(label_var_), vecValueVars(vecValueVars_) {

    if (access_weak_ptr(label_description)->hasValueDomain()) {
      dataDomain=Symb_DataDomain::getFor(access_weak_ptr(label_description)->getValueDomain());
    }
  }


  const BDDFunction Symb_Transition_Label_Var::getVars() const {
    BDDFunction cs=getHasLabelVar();
    if (hasValueDomain()) {
      cs&=container2cs(getValueVars());
    }

    return cs;
  }

  BDDFunction Symb_Transition_Label_Var::value_equal(const BDDFunctionContainer& _vecValueVars) const {
    debugOut(3, "Symb_Transition_Label_Var equals(BDDFunctionContainer)", this);
    BDDFunction result = 1;
    for (unsigned long i=0; i < dataDomain->getLogSize(); ++i) {
      result &= !(vecValueVars->at(i) ^ _vecValueVars.at(i));
    }
    return result;    
  }
  
  string Symb_Transition_Label_Var::getVariableName() const {
    return bdd_var_name(label_var);
  }


  void Symb_Transition_Label_Var::print_vars(std::ostream& out, std::string name) const {
    out << label_var.getFirstVariableIndex() << " " << bdd_var_name(label_var) << " " << name << " *l\n";
    for (unsigned int i=0;i<vecValueVars->size();i++) {
      BDDFunction p=vecValueVars->at(i);
      out << p.getFirstVariableIndex() << " " << bdd_var_name(p) << " " << name << " #l" << i << "\n";
    }
  }

  void Symb_Transition_Label_Var::serialize_vars(std::ostream& out) const {
    VarManager::ReorderLock lock;

    out << label_var.getFirstVariableIndex();
    for (unsigned int i=0;i<vecValueVars->size();i++) {
      out << " " << vecValueVars->at(i).getFirstVariableIndex();
    }    
    out << "\n";
  }

}
