/***************************************************************************
*   Copyright (C) 2008-2009 by Tobias Blechmann                           *
*   vereofy@blechmann.de                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "varmanager.hpp"
#include "vereofy/common/bdd_helpers.hpp"
#include "vereofy/common/functions.hpp"
#include "vereofy/common/errors.hpp"
#include "vereofy/common/checks.hpp"

#include <iostream>

#include <boost/lexical_cast.hpp>
#include <boost/foreach.hpp>
#include <boost/algorithm/string.hpp>

using namespace std;

#define foreach BOOST_FOREACH
using boost::lexical_cast;

namespace vereofy {

  VarManager::VarManager() {
    init();
  }

  void VarManager::init() {
    dynamic_reorder_enabled=false;
    reorder_lock=0;
    selective_reorder_level=0;
    numberOfStateVars = 0;
  }


  VarManager* VarManager::getInstance() {
    static VarManager instance;
    return &instance;
  }


  StateVars VarManager::findOrAddGlobalState(const string& name, unsigned int bits) {
    if (globalStateVars.count(name)>0) {
      const StateVars& vars=globalStateVars[name];

      if (vars.X->size()!=bits) {
        throw "Can't have global state var "+name+" with differing number of bits!";
      }
      return vars;
    }

    StateVars vars=createStateVars(bits);
    globalStateVars.insert(make_pair(name, vars));
    return vars;
  }

  BDDFunctionContainer VarManager::findOrAddGlobalBits(const string& name, unsigned int bits) {
    if (globalBits.count(name)>0) {
      const BDDFunctionContainer& vars=globalBits[name];
      
      if (vars.size()!=bits) {
        throw "Can't have global "+name+" with differing number of bits!";
      }
      return vars;
    }

    BDDFunctionContainer vars=createVars(name, bits);
    globalBits.insert(make_pair(name, vars));
    return vars;
  }

  StateVars VarManager::createStateVars(unsigned int number, bool prepend) {
    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    unsigned int position = prepend ? 0 : BDDVarOrder::size();

    StateVars result;
    result.X = boost::shared_ptr<BDDFunctionContainer> (new BDDFunctionContainer);
    result.Y = boost::shared_ptr<BDDFunctionContainer> (new BDDFunctionContainer);
    result.Z = boost::shared_ptr<BDDFunctionContainer> (new BDDFunctionContainer);
    for (unsigned int i=0; i<number; ++i) {    
      string curNumber = lexical_cast<string>(numberOfStateVars);
      std::vector<std::string> names;
      names.push_back("X"+curNumber);
      names.push_back("Y"+curNumber);
      names.push_back("Z"+curNumber);

      BDDFunctionContainer projections=createVarsInGroup(names, position);
      result.X->push(projections.at(0));
      result.Y->push(projections.at(1));
      result.Z->push(projections.at(2));

      ++numberOfStateVars;
    }
    
    return result;  
  }

  // if var_positions_in_out is too short, append new positions
  StateVars VarManager::createStateVars(unsigned int number, BDDFunctionContainer& var_positions_in_out) {
    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    StateVars result;
    result.X = boost::shared_ptr<BDDFunctionContainer> (new BDDFunctionContainer);
    result.Y = boost::shared_ptr<BDDFunctionContainer> (new BDDFunctionContainer);
    result.Z = boost::shared_ptr<BDDFunctionContainer> (new BDDFunctionContainer);
    for (unsigned int i=0; i<number; ++i) {
      string curNumber = lexical_cast<string>(numberOfStateVars);
      unsigned long position=0;
      bool has_position=false;
      if (i<var_positions_in_out.size()) {
        position=var_positions_in_out.at(i).getFirstVariableIndex();
        has_position=true;
      } else {
        has_position=false;
      }

      if (!has_position) {
        // choose position to append variable
        position=BDDVarOrder::size();
      }
      
      std::vector<std::string> names;
      names.push_back("X"+curNumber);
      names.push_back("Y"+curNumber);
      names.push_back("Z"+curNumber);

      BDDFunctionContainer projections=createVarsInGroup(names, position);
      result.X->push(projections.at(0));
      result.Y->push(projections.at(1));
      result.Z->push(projections.at(2));

      if (!has_position) {
        // add new position
        var_positions_in_out.push(projections.at(0));
      }

      ++numberOfStateVars;
    }

    return result;  
  }




  BDDFunctionContainer VarManager::createVars(const string& prefix, unsigned int number, bool prepend) {
    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    unsigned int position = prepend ? 0 : BDDVarOrder::size();

    BDDFunctionContainer result;
    for (unsigned int i=0; i<number; ++i) {    
      string curNumber = lexical_cast<string>(i);
      string var_name =prefix + (number>1 ? string("_")+curNumber : "");

      std::vector<std::string> names(1, var_name); // initialize with var_name
      BDDFunctionContainer projections=createVarsInGroup(names, position);
      result.push(projections.at(0));
    }
    return result;  
  }

  // if var_positions_in_out is too short, append new positions
  BDDFunctionContainer VarManager::createVars(const string& prefix, unsigned int number, BDDFunctionContainer& var_positions_in_out) {
    VarManager::ReorderLock reorder_lock; // inhibit reordering within this scope

    BDDFunctionContainer result;
    for (unsigned int i=0; i<number; ++i) {
      string curNumber = lexical_cast<string>(i);
      string var_name =prefix + (number>1 ? string("_")+curNumber : "");
      unsigned long position;
      bool has_position=false;
      if (i<var_positions_in_out.size()) {
        position=var_positions_in_out.at(i).getFirstVariableIndex();
        has_position=true;
      } else {
        has_position=false;
      }

      if (!has_position) {
        // choose position to append variable
        position=BDDVarOrder::size();
      }

      std::vector<std::string> names(1, var_name); // initialize with var_name
      
      BDDFunctionContainer projections=createVarsInGroup(names, position);
      result.push(projections.at(0));

      // add new position
      var_positions_in_out.push(projections.at(0));
    }

    return result;  
  }

  BDDFunction VarManager::getEncoding(unsigned long elementNr, const BDDFunctionContainer& varProjections) {
    BDDFunction result = BDDFunction::ONE();
    for(unsigned long i=0; i < varProjections.size(); ++i) {
      result &= (elementNr & (1<<i))==0 ? !varProjections.at(i):varProjections.at(i);
    }
    return result;
  }

  void VarManager::reset() {
    globalStateVars.clear();
    globalBits.clear();
    group_list.clear();
    init();

    BDD::reset();
    CHECK(3, BDDVarOrder::size()==0, "There are BDDFunctions still in use! "+BDDVarOrder::getVarOrder());
  }

  void VarManager::enableDynamicReordering() {
    dynamic_reorder_enabled=true;
    setReorderingState();
  }

  void VarManager::disableDynamicReordering() {
    dynamic_reorder_enabled=false;
    setReorderingState();
  }

  void VarManager::setDynamicReorderingMaxGrowth(float maxGrowth) {
    BDD::setDynamicReorderingMaxGrowth(maxGrowth);
  }

  void VarManager::setDynamicReorderingTimeout(unsigned int timeout_ms) {
    BDD::setDynamicReorderingTimeout(timeout_ms);
  }
  
  void VarManager::acquireReorderingLock() {
    ++reorder_lock;
    if (reorder_lock==1) {
      // outmost lock acquired, change in reordering state
      setReorderingState();
    }
  }

  void VarManager::releaseReorderingLock() {
    CHECK_SUCCESS(reorder_lock>0);
    if (--reorder_lock == 0) {
      // outmost lock released, change in reordering state
      setReorderingState();
    }
  }

  void VarManager::setReorderingState() {
    if (dynamic_reorder_enabled && reorder_lock==0) {
      // enable reordering
      BDD::enableDynamicReordering();
    } else {
      // disable reordering
      BDD::disableDynamicReordering();
    }
  }


  void VarManager::enableSelectiveReordering(unsigned int level) {
    selective_reorder_level=level;
  }
  
  void VarManager::disableSelectiveReordering() {
    selective_reorder_level=0;
  }

  
  void VarManager::selectiveReorderingHook(unsigned int level, bool force_reordering) {
    if (reorder_lock==0 && level>0 && level<=selective_reorder_level) {    
      BDD::reorderingHook(force_reordering);
    }
  }


  VarManager::ReorderLock::ReorderLock() {
    VarManager::getInstance()->acquireReorderingLock();
  }
  
  VarManager::ReorderLock::~ReorderLock() {
    VarManager::getInstance()->releaseReorderingLock();
  }

  BDDFunctionContainer VarManager::createVarsInGroup(const std::vector<std::string>& names,
                                                     unsigned int position) {
    if (names.size()>0 &&
        bdd_var_exists(names.at(0))) {
      // the first variable exists, we assume all variables exist
      BDDFunctionContainer result;
      for (unsigned int i=0;i<names.size();i++) {
        ERROR_CHECK_ASSERT(bdd_var_exists(names.at(i)));
        BDDFunction p=BDDFunction::projection(names.at(i));
        result.push(p);
        // TODO: Check that all variables belong to the same group
      }
      return result;
    }
    
    BDDGroup_ptr varGroup(new BDDGroup);
    std::list<BDDFunction> projection_functions;
    
    for (std::vector<std::string>::const_reverse_iterator it=names.rbegin();
         it!=names.rend();
         it++) {
      
      BDDVarOrder::insertVariable(*it, position);
      projection_functions.push_front(BDDFunction::projection(position));
#ifdef JINC_NEW 
      varGroup->groupVariable(position);
#else  // JINC_OLD
      varGroup->addVariable(*it);
#endif
    }

    group_list.push_back(varGroup);
    
    BDDFunctionContainer result;
    for (std::list<BDDFunction>::iterator it=projection_functions.begin();
         it!=projection_functions.end();
         ++it) {
      result.push(*it);
    }
    return result;
  }

  void VarManager::writeVarOrder(std::ostream& out) {
    unsigned int written=0, skipped=0;
    for (unsigned int i=0;i<BDDVarOrder::size();i++) {
#ifdef JINC_NEW
      string name=BDDVarOrder::getVariableName(i);
#else
      string name=BDDVarOrder::getVariable(i);
#endif
      
      if (boost::starts_with(name, "Y") || boost::starts_with(name, "Z")) {
        skipped++;
        continue; // skip
      }
      
      written++;
      out << name << "\n";
    }
    if (debugLevel>=1) {
      std::cout << "VarManager: Wrote " << written << " variables to file, skipped " << skipped << std::endl;
    }
  }

  void VarManager::readVarOrder(std::istream& in) {
    std::vector<std::string> vars;
    while (!in.eof()) {
      std::string name;
      getline(in, name);
      if (in.fail()) { break; }
      vars.push_back(name);
    }

    if (debugLevel>=1) {
      std::cout << "VarManager: Read " << vars.size() << " variables from file." << std::endl;
    }

    // reverse iterate over variable names
    for (std::vector<std::string>::reverse_iterator it=vars.rbegin();
         it!=vars.rend();
         ++it) {
      const std::string& name=*it;

      std::vector<std::string> names;
      if (boost::starts_with(name, "X")) {
        std::string suffix=name.substr(1);
        names.push_back(name);
        names.push_back("Y"+suffix);
        names.push_back("Z"+suffix);
      } else {
        names.push_back(name);
      }

      createVarsInGroup(names, 0);  // prepend variable
    }
  }  


}
