/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "system_context.hpp"
#include "vereofy/reo_parser/circuit_parser_helper.hpp"
#include "vereofy/common/errors.hpp"
#include "vereofy/common/file_ops.hpp"

#include "vereofy/reo_parser/REOParser.hpp"
#include "vereofy/system/builtin_components/builtin_lib.hpp"
#include "vereofy/system/builtin_components/builtin_functions.hpp"


#include <iostream>

namespace vereofy {

SystemContext::ptr SystemContext::newSystemContext() {
  SystemContext::ptr sc(new SystemContext());
  sc->initStandardDefinitions();
  return sc;
}

SystemContext::SystemContext() :
  global_symbols(new symbol_table_t()),
  builtins_were_loaded(false),
  ignore_multiple_includes(false),
  replacement_manager(new ReplacementManager()) {
  global_symbols->name="global";

  type_modification_registry.reset(new TypeModificationRegistry());
}

void SystemContext::addBasicChannel(const string& name) {
  if (global_symbols->find(name)) {
    throw string("There is already a definition for ")+name;
  }
  
  vparser::BasicChannelInfo::ptr bci(new vparser::BasicChannelInfo(name));            
  global_symbols->add(name, bci);
}


void SystemContext::loadBuiltins() {
  if (!builtins_were_loaded) {

    vparser::REOParser rp(shared_from_this());
    rp.parse(getBuiltinLib(), "builtin");

    builtins_were_loaded=true;
  }
}

symbol_table_t::ptr SystemContext::getGlobalSymbols() {
  return global_symbols;
}

ReplacementManager::ptr SystemContext::getReplacementManager() {
  return replacement_manager;
}

TypeModificationRegistry::ptr SystemContext::getTypeModificationRegistry() {
  return type_modification_registry;
}

void SystemContext::requireDataDomain() {
  if (!getDataDomain()) {
    vereofy::FlatDataDomain::ptr standard_dd(new vereofy::IntRangeDataDomain(0,1));
    cerr << "No data domain specified, using int(0,1) as standard data domain\n";

    TypeDefData::ptr tdd(new TypeDefData("Data", standard_dd));
    getGlobalSymbols()->add("Data", tdd);    
  }
}

ComponentInstance::ptr SystemContext::instantiateSystem(const ComponentTypeIdentifier& component_type,
                                                        bool strict) {
  if (replacement_manager->hasReplacement(component_type.toString())) {
    ComponentTypeIdentifier cti_replacement(replacement_manager->getReplacement(component_type.toString()));
    
    std::cout << "Replacing " << component_type.toString() << " with " << cti_replacement.toString() << "\n";

    return instantiateSystem(cti_replacement, strict);
  }

  requireDataDomain();

  SymbolData::ptr si=global_symbols->find(component_type.type_name);
  
  if (!si) {
    throw string("There is no component named '")+component_type.type_name+"'";
  }
  
  if (si->type!=SYMBOL_TYPE_COMPONENT) {
    throw string("The object named '")+component_type.type_name+"' is not a component!";
  }
  
  ComponentData::ptr cd=boost::static_pointer_cast<ComponentData>(si);
  
  switch (cd->comp_type) {
  case COMPONENT_CIRCUIT: {
    vparser::CircuitInfo::ptr ci=boost::static_pointer_cast<vparser::CircuitInfo>(si);
    return ComponentInstance::ptr(new ComponentInstanceCircuit(ci->connector, component_type.parameters, strict));
  }
  case COMPONENT_REACTIVE_MODULE: {
    vparser::ModuleInfo::ptr mi=boost::static_pointer_cast<vparser::ModuleInfo>(si);
    return ComponentInstance::ptr(new ComponentInstanceReactiveModule(mi->module, component_type.parameters));
  }
  default:
    THROW_UNIMPLEMENTED();
  }
}


DataDomain::ptr SystemContext::getDataDomain() {
  SymbolData::ptr sd=global_symbols->find("Data");
  if (!sd) {
    return DataDomain::ptr();
  }
    
  if (sd->type!=SYMBOL_TYPE_TYPEDEF) {
    return DataDomain::ptr();
  }

  DataDomain::ptr data_dd=
    boost::static_pointer_cast<TypeDefData>(sd)->data_domain;

  try {
    data_dd->ensureFiniteness();
    data_dd->ensureFullyDefinedness();
  } catch (std::string& e) {
    throw std::string("Global data domain: ")+e;
  }

  return data_dd;
}

void SystemContext::addIncludeDirs(const std::vector<string>& include_dirs_) {
  include_dirs.insert(include_dirs.end(), include_dirs_.begin(), include_dirs_.end());
}

ReturnValue<string> SystemContext::findFile(const std::string& filename) const {
  if (!FileOps::file_exists(filename)) {
    for (unsigned int i=0;i<include_dirs.size();i++) {
      string name=include_dirs.at(i)+"/"+filename;
      if (FileOps::file_exists(name)) {
	return name;
      }
    }
  } else {
    return filename;
  }
  return ReturnValue<string>();
}

void SystemContext::initStandardDefinitions() {
  vparser::REOParser rp(shared_from_this());
  rp.parse_and_add_typedef("TYPE Bool = int(0,1);");
  rp.parse_and_add_constdef("CONST TRUE = 1;");
  rp.parse_and_add_constdef("CONST FALSE = 0;");

  // load builtin functions
  loadBuiltinFunctions(*getGlobalSymbols());
}

  
void SystemContext::setFlag(const std::string& flag, bool enabled) {
  flags[flag]=enabled;
}


bool SystemContext::getFlag(const std::string& flag) const {
  flag_map::const_iterator it=flags.find(flag);
  if (it==flags.end()) {return false;}
  
  return it->second;
}

const SystemContext::flag_map& SystemContext::getFlagMap() const {
  return flags;
}

void SystemContext::markFileAsIncluded(const std::string& filename) {
  if (filename!="") {
    included_files.insert(filename);
  }
}

bool SystemContext::isFileMarkedAsIncluded(const std::string& filename) {
  return included_files.count(filename);
}

bool SystemContext::ignoreMultipleIncludes() const {
  return ignore_multiple_includes;
}

void SystemContext::setIgnoreMultipleIncludes(bool value) {
  ignore_multiple_includes=value;
}



}
