/**************************************************************************
*   Copyright (C) 2008-2009 by Johannes Osterholzer                       *
*   s5892660@inf.tu-dresden.de                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include <iostream>
#include "vereofy/setup.hpp"

#include "vereofy/system/automatonfactory.hpp"
#include "vereofy/system/reo_circuit.hpp"
#include "vereofy/system/symb_automaton/symb_channel_end.hpp"
#include "vereofy/system/symb_automaton/symb_automaton.hpp"
#include "vereofy/system/symb_automaton/symb_expression.hpp"
#include "vereofy/common/functions.hpp"

#include <vector>
#include <stdlib.h>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#include <boost/lexical_cast.hpp>
using boost::lexical_cast;

using namespace vereofy;

void testSimpleSystem();
void testNonBinarySystem();
void testSemaphores(int procs, int sems);
void testYetAnotherSystem();
void testNonInjectiveDelta();

void usage(char *name);


int main(int argc, char **argv) {
  if(argc == 1) {
    usage(argv[0]);
    return -1;
  }
  if(argv[1][0] == 't')
    testSimpleSystem();
  else if(argv[1][0] == 'n')
    testNonBinarySystem();
  else if(argv[1][0] == 's')
    (argc >= 4) ? testSemaphores(atoi(argv[2]), atoi(argv[3])) : usage(argv[0]);
  else if(argv[1][0] == 'y')
    testYetAnotherSystem();
  else if(argv[1][0] == 'i')
    testNonInjectiveDelta();
  else
    usage(argv[0]);
  return EXIT_SUCCESS;
}


class SimpleSystem : public SymbAutomaton {
public:
  SimpleSystem(Symb_FlatDataDomain::ptr dd) :
    SymbAutomaton(dd) {

    Symb_Channel_End::ptr channelA(new Symb_Channel_End(dd));
    Symb_Channel_End::ptr channelB(new Symb_Channel_End(dd));

    StateVars vars = createStateVars(2);
    BDDFunction x0 = vars.X->at(0);
    BDDFunction x0_ = x0.moveDown();
    BDDFunction x1 = vars.X->at(1);
    BDDFunction x1_ = x1.moveDown();

    addChannelEndToAutomaton(channelA);
    addChannelEndToAutomaton(channelB);

    BDDFunction A = channelA->getPortActive();
    BDDFunction B = channelB->getPortActive();
    
    BDDFunction d_A = channelA->getDCVars().at(0); //binary data domain
    BDDFunction d_B = channelB->getDCVars().at(0);

    initialStates=!x0 & !x1;
    setDelta(!x1 & !x0 & A & B & !x1_ & x0_ |
	     !x1 & !x0 & A & B & channelA->equals(*channelB) & x1_ & !x0_ |
	     x1 & !x0 & !A & B & x1_ & x0_ |
	     x1 & x0 & A & !B & x1_ & !x0_ |
	     !x1 & x0 & A & B & !x1_ & x0_);
  }
};

void testSimpleSystem()
{
  // The following is probably the wrong way (R) to create a SymbAutomaton.
  // However, it should be ok for testing purposes.

  // Test: correct :)

  // create a test data domain
  IntRangeDataDomain::ptr dataDomain(new IntRangeDataDomain(0,1));
  Symb_FlatDataDomain::ptr symb_dataDomain(new Symb_FlatDataDomain(dataDomain));

  SymbAutomaton::ptr a(new SimpleSystem(symb_dataDomain));

  cout << "--- nondeterministic automaton a ---\n";
  a->getDelta().printEssentialVariables();
  a->getDelta().printPaths();

  cout << "--- determistic automaton det(a)\n";
  a->makeDeterministic();
  a->getDelta().printEssentialVariables();
  a->getDelta().printPaths();
}



class NonBinarySystem : public SymbAutomaton {
public:
  NonBinarySystem(Symb_FlatDataDomain::ptr dd) :
    SymbAutomaton(dd) {

    Symb_Channel_End::ptr channelA(new Symb_Channel_End(dd));
    Symb_Channel_End::ptr channelB(new Symb_Channel_End(dd));
    
    StateVars vars = createStateVars(3);
    BDDFunction x0 = vars.X->at(0);
    BDDFunction x0_ = x0.moveDown();
    BDDFunction x1 = vars.X->at(1);
    BDDFunction x1_ = x1.moveDown();
    BDDFunction x2 = vars.X->at(2);
    BDDFunction x2_ = x2.moveDown();
    
    addChannelEndToAutomaton(channelA);
    addChannelEndToAutomaton(channelB);
    
    BDDFunction A = channelA->getPortActive();
    BDDFunction B = channelB->getPortActive();
  
    SymbExpression dataA(dd, channelA->getDCVars());
    SymbExpression dataB(dd, channelB->getDCVars());
  
    initialStates=!x0 & !x1 & !x2;
    setDelta(!x2 & !x1 & !x0 & A & B & !x2_ & !x1_ & x0_ |
             !x2 & !x1 & !x0 & A & B & dd->performComparison(COMPARE_EQUAL, dataA, SymbExpression(0)) & !x2_ & x1_ & !x0_ |
		!x2 & !x1 & !x0 & A & B & dd->performComparison(COMPARE_EQUAL, dataB, SymbExpression(2)) & !x2_ & x1_ & x0_ |
             !x2 & !x1 & x0 & A & B & !dd->performComparison(COMPARE_EQUAL, dataA, dataB) & x2_ & !x1_ & !x0_ |
		!x2 & x1 & !x0 & !A & B & x2_ & !x1_ & !x0_ |
		!x2 & x1 & x0 & !A & B & x2_ & !x1_ & !x0_ |
		x2 & !x1 & !x0 & A & !B & !x2_ & x1_ & !x0_ |
		x2 & !x1 & !x0 & A & !B & !x2_ & x1_ & x0_);
  }
};

void testNonBinarySystem()
{
  // Test: correct. :)

  // create a test data domain
  IntRangeDataDomain::ptr dataDomain(new IntRangeDataDomain(0,3));
  Symb_FlatDataDomain::ptr symb_dataDomain(new Symb_FlatDataDomain(dataDomain));

  SymbAutomaton::ptr a(new NonBinarySystem(symb_dataDomain));

  cout << "--- nondeterministic automaton a ---\n";
  a->getDelta().printEssentialVariables();
  a->getDelta().printPaths();

  cout << "--- determistic automaton det(a)\n";
  SymbAutomaton::ptr det(a->makeDeterministicCopy());
  det->getDelta().printEssentialVariables();
  det->getDelta().printPaths();
}

class YetAnotherSystem : public SymbAutomaton {
public:
  YetAnotherSystem(Symb_FlatDataDomain::ptr dd) :
    SymbAutomaton(dd) {

    Symb_Channel_End::ptr channelA(new Symb_Channel_End(dd));
    Symb_Channel_End::ptr channelB(new Symb_Channel_End(dd));

    StateVars vars = createStateVars(2);
    BDDFunction x0 = vars.X->at(0);
    BDDFunction x1 = vars.X->at(1);
    BDDFunction x0_ = x0.moveDown(), x1_ = x1.moveDown();

    initialStates = !x1 & !x0 | x1 & !x0;
    setDelta(!x1 & !x0 & channelA->getPortActive() & !x1_ & x0_ |
             x1 & !x0 & channelB->getPortActive() & x1_ & x0_);
  }
};

void testYetAnotherSystem()
{
  IntRangeDataDomain::ptr dataDomain(new IntRangeDataDomain(0,1));
  Symb_FlatDataDomain::ptr symb_dataDomain(new Symb_FlatDataDomain(dataDomain));

  SymbAutomaton::ptr a(new YetAnotherSystem(symb_dataDomain));

  cout << "--- nondeterministic automaton a ---\n";
  a->getDelta().printEssentialVariables();
  a->getDelta().printPaths();

  cout << "--- determistic automaton det(a)\n";
  SymbAutomaton::ptr det(a->makeDeterministicCopy());
  det->getDelta().printEssentialVariables();
  det->getDelta().printPaths();
}

class NonInjectiveDelta : public SymbAutomaton {
public:
  NonInjectiveDelta(Symb_FlatDataDomain:: ptr dd) 
   : SymbAutomaton(dd) {
    Symb_Channel_End::ptr channelA(new Symb_Channel_End(dd));

    StateVars vars = createStateVars(2);
    BDDFunction x0 = vars.X->at(0);
    BDDFunction x1 = vars.X->at(1);
    BDDFunction x0_ = x0.moveDown(), x1_ = x1.moveDown();

    SymbExpression dataA(dd, channelA->getDCVars());

    initialStates = !x1 & !x0 | !x1 & x0;
    setDelta(
      !x1 & !x0 & dd->performComparison(COMPARE_EQUAL, dataA, SymbExpression(0)) & x1_ & !x0_ |
      !x1 & !x0 & dd->performComparison(COMPARE_EQUAL, dataA, SymbExpression(0)) & x1_ &  x0_ |
      !x1 &  x0 & dd->performComparison(COMPARE_EQUAL, dataA, SymbExpression(1)) & x1_ & !x0_);
  }
};

void testNonInjectiveDelta()
{
  IntRangeDataDomain::ptr dataDomain(new IntRangeDataDomain(0,1));
  Symb_FlatDataDomain::ptr symb_dataDomain(new Symb_FlatDataDomain(dataDomain));

  SymbAutomaton::ptr a(new NonInjectiveDelta(symb_dataDomain));

  cout << "--- nondeterministic automaton a ---\n";
  a->getDelta().printEssentialVariables();
  a->getDelta().printPaths();

  cout << "--- determistic automaton det(a)\n";
  SymbAutomaton::ptr det(a->makeDeterministicCopy());
  det->getDelta().printEssentialVariables();
  det->getDelta().printPaths();
}


void testSemaphores(int procs, int sems)
{
  cout << "Not implemented (yet).\n";
}

void usage(char *name)
{
  cout << "Usage:\t" << name << " t - for simple test system\n";
  cout << "\t" << name << " n - for nonbinary test system\n";
  cout << "\t" << name << " s [nrOfProcs] [nrOfSemaphores] - for semaphore test system\n";
  cout << "\t" << name << " y - for yet another system (with 2 initial states)\n";
  cout << "\t" << name << " i - for non-injective delta\n";
}
