/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "vereofy_context.hpp"

#include "vereofy/setup.hpp"
#include "vereofy/common/tracing.hpp"
#include "vereofy/common/checks.hpp"
#include "vereofy/common/version.hpp"
#include "vereofy/common/CSVFile.hpp"

#include <boost/shared_ptr.hpp>
#include <boost/program_options.hpp>
#include <iostream>

#include "vereofy/common/stopwatch.hpp"
#include "vereofy/common/file_ops.hpp"
#include "vereofy/system/circuit_connector/component_instance_circuit.hpp"
#include "vereofy/system/symb_automaton/symb_automaton.hpp"
#include "vereofy/system/symb_automaton/symb_bfs.hpp"
#include "vereofy/system/component_type_identifier.hpp"
#include "vereofy/system/system_statistics.hpp"

#include "vereofy/system/symb_automaton/serialized_varorder.hpp"
#include "vereofy/system/symb_automaton/extract_explicit.hpp"

#include "vereofy/modelcheck/checkers/ModelChecker.hpp"
#include "vereofy/modelcheck/checkers/ModelCheckerInit.hpp"
#include "vereofy/modelcheck/checkers/branching_time_check/ASLModelCheckingResult.hpp"

#include "vereofy/modelcheck/checkers/bisim_equiv/bisimulation.hpp"

#include "vereofy/modelcheck/specifications/Specification.hpp"
#include "vereofy/modelcheck/specifications/SpecParser.hpp"
#include "vereofy/modelcheck/specifications/SpecParserInit.hpp"

#include "vereofy/modelcheck/specifications/asl/ASL_Formula.hpp"

#include "vereofy/reo_parser/REOParser.hpp"
#include "vereofy/reo_parser/REOParseException.hpp"
#include "vereofy/reo_parser/REOExecuteException.hpp"
#include "vereofy/system/reactive_module/symb_ca_reactive_module.hpp"

#include <boost/algorithm/string.hpp>
#include <cstdlib>

using namespace std;
using namespace vereofy;

namespace po=boost::program_options;

// static var:
bool VereofyContext::initialized=false;


// helper function

static po::options_description getModelCheckerOptions(OptionContainer& checker_options) {
  po::options_description opt_checking("Modelchecker options");
  
  for (OptionContainer::iterator it=checker_options.begin();
       it!=checker_options.end();
       ++it) {
    Option& option=it->second;

    opt_checking.add_options()
      (option.getName().c_str(),
       po::value<string>(&option.getValueRef())->default_value(option.getDefaultValue().c_str()),
       option.getDescription().c_str());
  }

  return opt_checking;
}




VereofyContext::VereofyContext() :
  bisim_check_mode(false) {

  initialize(); // static initializations

  debugLevel=0;
  
  // set JINC garbage collection delay
  BDD::setGarbageCollectionDelay(1500);
  
  dynamic_reordering_during_build=false;
  dynamic_reordering_during_modelchecking=false;
  
#ifndef JINC_NEW
  // JINC_OLD and not 32bit pointers are problematic
  if (sizeof(int*)!=4 || sizeof(int*)!=sizeof(unsigned int)) {
    std::cout << "Size of pointers: " << sizeof(int*)*8 << " bits\n";
    throw string("JINC_OLD library is only compatible with 32bit compilation. Please switch to JINC_NEW");
  }
#endif
  
  system_context=SystemContext::newSystemContext();
}

VereofyContext::~VereofyContext() {
  system_context.reset();
  automaton.reset();
  system.reset();
  
  VarManager::getInstance()->reset();
}

void VereofyContext::parse(const std::string& content) {
  vparser::REOParser reo_parser(system_context);
  reo_parser.parse(content);
}

void VereofyContext::parse_file(const std::string& file_name) {
  ReturnValue<string> real_circuit_file=system_context->findFile(file_name);

  if (!real_circuit_file.valid) {
    throw "Can not find file '"+file_name+"'";
  }

  vparser::REOParser reo_parser(system_context);
  
  if (!reo_parser.parse_file(real_circuit_file.value)) {
    throw "Unknown parse error parsing file "+ real_circuit_file.value;
  }    
}

void VereofyContext::instantiate_system(const std::string& name) {
  boost::scoped_ptr<ComponentTypeIdentifier> component_type_identifier;

  if (parameters.size()>0) {
    // there are already parameters (probably via -P command line parameters)
    component_type_identifier.reset(new ComponentTypeIdentifier(name, parameters));
  } else {
    component_type_identifier.reset(new ComponentTypeIdentifier(name));
  }

  system=system_context->instantiateSystem(*component_type_identifier, strict);

  if (dot_circuit!="" && system->getType()==COMPONENT_CIRCUIT) {
    ofstream dot_circuit_stream(dot_circuit.c_str());
    if (dot_circuit_stream.good()) {
      print_circuit_dot(dot_circuit_stream);
      dot_circuit_stream.close();
    } else {
      throw "Can't open DOT file: "+dot_circuit;
    }
  }

  if (export_circuit_reo_xml!="" && system->getType()==COMPONENT_CIRCUIT) {
    ofstream reo_circuit_stream(export_circuit_reo_xml.c_str());
    if (reo_circuit_stream.good()) {
      export_as_reo_xml(reo_circuit_stream, param_input_files);
      reo_circuit_stream.close();
    } else {
      throw "Can't open REO file: "+export_circuit_reo_xml;
    }
  }

  if (param_show_system_stats) {
    SystemStatistics sys_stat(system, system_context->getDataDomain());
    sys_stat.printReport(std::cout);
  }
}

void VereofyContext::print_interface(ostream& out) {
  system->getInterface()->print(out);
}

void VereofyContext::print_interface_xml(ostream& out) {
  system->print_interface_xml(out);
}

void VereofyContext::print_circuit_dot(ostream& out) {
  if (system->getType()==COMPONENT_CIRCUIT) {
    ComponentInstanceCircuit::ptr cic=boost::static_pointer_cast<ComponentInstanceCircuit>(system);
    cic->toDot(out);
  } else {
    out << "digraph A {}\n";
  }
}

void VereofyContext::export_as_reo_xml(ostream& out, const std::vector<std::string>& input_files) {
  CHECK_SUCCESS(system);

  if (system->getType()==COMPONENT_CIRCUIT) {
    ComponentInstanceCircuit::ptr cic=boost::static_pointer_cast<ComponentInstanceCircuit>(system);
    cic->toReoXML(out, input_files);
  } else {
    throw "Can not export MODULE to REO XML";
  }
}


void VereofyContext::build_system(RestrictionManager::ptr restriction_manager) {
  BuildContext::ptr build_context(new BuildContext());
  build_context->param_constrain_to_reachable_components = (param_constrain_reachable_components==1);
  build_context->param_constrain_to_reachable_circuits = param_constrain_reachable_circuits;
  build_context->param_retain_hidden_ports = param_retain_hidden_ports;
  build_context->param_retain_top_level_ports = param_retain_top_level_ports;
  build_context->param_abstract_data_domain = param_abstract_data_domain;
  build_context->param_no_shared_data = param_no_shared_data;

  build_context->restriction_manager=restriction_manager;
  
  if (varorder_dir!="") {
    build_context->try_loading_serialized_varorder=true;
    build_context->varorder_dir=varorder_dir;
  } else {
    build_context->try_loading_serialized_varorder=false;
  }

  if (param_optimize_ports) {
    build_context->port_partitions=PortPartitions::ptr(new PortPartitions());
    system->calculatePortPartitions(*build_context->port_partitions);
    build_context->port_partitions->if_active_same_data_ports=
      build_context->port_partitions->if_active_same_data_ports.lifted(build_context->port_partitions->identical_ports);
    
    
    TRACE("PortEquivalence", "identical", std::cout << "Identical ports:\n"; build_context->port_partitions->identical_ports.print(std::cout););
    TRACE("PortEquivalence", "mixed", std::cout << "Mixed ports:\n"; build_context->port_partitions->mixed_data_ports.print(std::cout););
    TRACE("PortRelation", "if_active_same_data", std::cout << "Ports, that, if active together, have the same data value:\n"; build_context->port_partitions->if_active_same_data_ports.print(std::cout););
  }
  
  if (dynamic_reordering_during_build) {
    enableDynamicReordering();
  } else {
    disableDynamicReordering();
  }

  if (read_varorder_file!="") {
    ReturnValue<string> varorder_file=system_context->findFile(read_varorder_file);
    if (!varorder_file.valid) {
      throw string("Can't find file ")+read_varorder_file;
    }
    ifstream i_varorder(varorder_file.value.c_str());
    if (!i_varorder.good()) {
      throw string("Can't open/read ")+varorder_file.value;
    }
    SerializedVarorder::readVarorder(i_varorder,
                                     system,
                                     build_context,
                                     false);
    
    // disable dynamic reordering during build if we have varorder
    disableDynamicReordering();
  }

  // symbolic automaton
  DataDomain::ptr dd=system_context->getDataDomain();
  
  Symb_DataDomain::ptr s_dd(Symb_DataDomain::getFor(dd));
  
  StopWatch building_stopwatch;
  
  BDD::setGarbageCollectionDelay(param_gcd_building);
  cout << "Generating CA for system...\n";
  building_stopwatch.start();
  automaton=system->getSymbAutomaton(system, s_dd);
  automaton->setShowOnlyToplevelPorts(!param_show_all_ports);

  automaton->build_automaton(build_context);

  // for selective reordering, force one last reordering after the whole CA is built
  VarManager::getInstance()->selectiveReorderingHook(1, true);

  if (bisim_check_mode || param_hide_internals) {
    automaton->hideInternalPorts(build_context);
  }

  if (SymbAutomaton::getDefaultHideType()==SymbAutomaton::HIDE_TRANSITIVE_FULL) {
    // call hide to remove all epsilons with empty list of channel ends
    automaton->hide(std::vector<Automaton_Channel_End::ptr>());
  }

  building_stopwatch.stop();
  

  TRACE("print_vars", "system",                         \
        ofstream of("vars.txt");                        \
        automaton->info->print_var_info(of));
  
  cout << "Generation of CA finished, time = " << building_stopwatch.getElapsedTimeString() << "s\n";
  if (debugLevel>=2) {
    cout << "CA initialStates() = " << automaton->getInitialStates().size() << " BDD Nodes\n";
    cout << "CA delta() = " << automaton->getDelta().size() << " BDD Nodes\n";
    if (debugLevel>=4)
      cout << "Reachable States = " << automaton->getNumberOfReachableStates() << endl;
  }

  TRACE("stats", "ca",                          \
        cout << "CA initialStates() = " << automaton->getInitialStates().size() << " BDD Nodes\n"; \
        cout << "CA delta() = " << automaton->getDelta().size() << " BDD Nodes\n";);
        

  TRACE("stats", "reachable",                                           \
        cout << "Reachable States = " << automaton->getNumberOfReachableStates() << endl;);
  
  if (!without_restrict) {
    StopWatch restrict_stopwatch;
    cout << "Restricting to valid values...\n";
    restrict_stopwatch.start();
    automaton->restrictToValidValues();
    restrict_stopwatch.stop();

    // for selective reordering, force one last reordering after the whole CA is built
    VarManager::getInstance()->selectiveReorderingHook(1, true);
    
    cout << "Restricting finished , time = " << restrict_stopwatch.getElapsedTimeString() << "s\n";
    if (debugLevel>=2) {
      cout << "CA initialStates() = " << automaton->getInitialStates().size() << " BDD Nodes\n";
      cout << "CA delta() = " << automaton->getDelta().size() << " BDD Nodes\n";
      if (debugLevel>=4) 
        cout << "Reachable States = " << automaton->getNumberOfReachableStates() << endl;
    }
  }

  if (print_varorder) {
    std::cout << BDDVarOrder::getVarOrder() << std::endl;
  }


  if (dot_file!="") {
    ofstream dot_stream(dot_file.c_str());
    if (dot_stream.good()) {
      print_automaton_dot(dot_stream);
    } else {
      throw "Can't open DOT file: "+dot_file;
    }
  }


  if (extract_explicit_file!="") {
    ofstream extract_explicit_stream(extract_explicit_file.c_str());
    if (extract_explicit_stream.good()) {
      extract_explicit(automaton, extract_explicit_stream);
    } else {
      throw "Can't open file: "+extract_explicit_file;
    }
  }
}


void VereofyContext::print_automaton_dot(ostream& out) {
  CHECK_SUCCESS(automaton);
  automaton->toDot(out, true);
}

ModelCheckingResult::ptr VereofyContext::model_check(const std::string& formula) {
  check_phase_stopwatch.start();

  if (dynamic_reordering_during_modelchecking) {
    enableDynamicReordering();
  } else {
    disableDynamicReordering();
  }

  Specification::ptr spec=SpecParser::parse_global(formula,
                                                   system_context,
                                                   system->getInterface());

  ModelChecker::ptr checker=ModelChecker::getChecker(spec->getType());
  
  BDD::setGarbageCollectionDelay(param_gcd_checking);
  
  if (!checker) throw "No ModelChecker for this specification defined!";
  ModelCheckingResult::ptr result=checker->analyze(automaton, spec, modelchecker_options);
  
  if (!result) {
    throw "No valid ModelCheckingResult!";
  }

  if (param_check_result) {
    if (!result->canCheckResult()) {
      std::cout << "\nChecking result not supported for this property." << std::endl;
    } else {
      std::cout << "\nChecking result..." << std::endl;
      
      // MAY modify automaton!!!
      bool result_ok = result->checkResult(system_context, automaton, modelchecker_options);
      
      if (!result_ok) {
        throw std::string("Result check FAILED!");
      } else {
        std::cout << "Result OK\n";
      }
    }
  }
  
  check_phase_stopwatch.stop();
  if (csv_file) {
    csv_file->addColumn("check_time", check_phase_stopwatch.getElapsedTimeString());
    csv_file->write();
  }
  
  return result;
}

void VereofyContext::initialize() {
  if (!initialized) {
    // static initialization
    init_spec_parsers();
    init_modelcheckers();

    initialized=true;
  }
}

void VereofyContext::add_parameters(unsigned int argc, char **argv) {
  if (cmd_line_parameters.size()==0) new_parameter_section();

  for (unsigned int i=1;i<argc;i++) {
    cmd_line_parameters.back().push_back(argv[i]);
  }
}

void VereofyContext::add_parameter(const string& parameter) {
  if (cmd_line_parameters.size()==0) new_parameter_section();

  cmd_line_parameters.back().push_back(parameter);
}

void VereofyContext::new_parameter_section() {
  cmd_line_parameters.push_back(vector<string>());
}

bool VereofyContext::apply_parameters() {
  modelchecker_options=ModelChecker::getAllOptions();
  po::options_description opt_checkers=getModelCheckerOptions(modelchecker_options);
  
  po::options_description opt_general("General options");
  opt_general.add_options()
    ("help", "produce a help message")
    ;
  
  po::options_description opt_system("System options");
  opt_system.add_options()
    ("system", 
     po::value<string>(&system_name)->default_value("main"),
     "the name of the circuit/module to check")
    ("check-equivalence-with", 
     po::value<string>(&check_equivalence_with),
     "the name of the second circuit/module for bisimulation")
    ("system-parameter,P", 
     po::value<vector<string> >(),
     "parameter for the system")
    ("formula",
     po::value<string>(&formula),
     "the formula to be checked")
    ("print_trace_to",
     po::value<string>(&print_trace_to),
     "Print a trace from an initial state to a state satisfying the given state formula")
    ("random-trace",
     po::value<signed int>(&random_trace_steps),
     "Print a random trace with n steps")
    ("random-trace-iterative",
     po::value<signed int>(&random_trace_steps_iterative),
     "Don't build whole system, do iterative random tracing")     
    ("input",
     po::value<vector<string> >(),
     "read the specified input file")
    ("to-dot",
     po::value<string>(&dot_file),
     "Output DOT file of system")
    ("circuit-dot",
     po::value<string>(&dot_circuit),
     "Output DOT file of the system circuit")
    ("export-circuit",
     po::value<string>(&export_circuit),
     "Export the system circuit to a file")
    ("export-to-reo-xml",
     po::value<string>(&export_circuit_reo_xml),
     "Export the system circuit to REO XML file")
    ("data-domain",
     po::value<string>(&data_domain),
     "Data domain")
    ("include-dir,I", 
     po::value<vector<string> >(),
     "include directory")
    ("ignore-multiple-includes",
     po::value<bool>(&param_ignore_multiple_includes)->default_value(false),
     "Ignore multiple includes of the same file")
    ("automatic-includes",
     po::value<bool>(&param_automatic_includes)->default_value("false"),
     "Automatically add directories of --input files to the include directories")
    ("constant,C",
     po::value<vector<string> >(),
     "constant definition")
    ("flag,F",
     po::value<vector<string> >(),
     "Compile-time flags, either +flag or -flag")
    ("replace,R",
     po::value<vector<string> >(),
     "replace one component type with another. Syntax: type_1 -> type_2")
    ("hide",
     po::value<string>(),
     "The variant of the hide operator to use [transitive, transitive-full, epsilon]"
     )
    ("minimize-dfa",
     po::value<string>(),
     "Indicate what method should be used for minimization of DFAs [none, hopcroft, brzozowski]"
     )
    ("hide-internals",
     po::value<bool>(&param_hide_internals)->default_value(false),
     "Hide internals (e.g. non-exported ports/nodes) of the final system")
    ("retain-hidden-ports",
     po::value<bool>(&param_retain_hidden_ports)->default_value(false),
     "Retain the I/O activity at hidden ports")
    ("retain-top-level-ports",
     po::value<bool>(&param_retain_top_level_ports)->default_value(false),
     "Retain the I/O activity at all ports in the top level of the system")
    ("show-all-ports",
     po::value<bool>(&param_show_all_ports)->default_value(false),
     "0 = show only top-level visible ports, 1 = show all available port information")
    ("system-stats",
     po::value<bool>(&param_show_system_stats)->default_value(false),
     "show statistics about the system")
    ("stats",
     po::value<string>(&stat_file),
     "Print statistics to file")
    ("print-interface",
     po::value<bool>(&param_print_interface)->default_value(false),
     "Print the interface of the system")
    ("without-restrict",
     po::value<bool>(&without_restrict)->default_value(false),
     "Without restricting to valid values")
    ("abstract-data-domain",
     po::value<bool>(&param_abstract_data_domain)->default_value(false),
     "Abstract away the actual values at ports")
    ("strict",
     po::value<bool>(&strict)->default_value(false),
     "Generate system as if it were a subcomponent")
    ("load-builtins", "Load the builtin library of connectors")
    ("parse-only", 
     po::value<bool>(&param_parse_only)->default_value(false),
     "Only parse the input, don't build the system")
    ("trace-xml", 
     po::value<std::string>(&trace_xml_file),
     "Write trace information in XML format to file")
    ;
  
  po::options_description opt_debug("Debug options");
  opt_debug.add_options()
    ("debug",
     po::value<int>(&debugLevel),
     "Debug Level (0=no info, 5=max info)")
    ("trace,T", 
     po::value<vector<string> >(),
     "trace points for debugging")
    ("check-level",
     po::value<unsigned int>(&check_level),
     "level of internal sanity checks (0=off)")
    ("print-varorder",
     po::value<bool>(&print_varorder)->default_value(false),
     "Print the variable ordering")
    ("random-seed",
     po::value<unsigned int>(),
     "Seed the PRNG with the given value");
  
  po::options_description opt_bdd("BDD options");
  opt_bdd.add_options()
    ("dynamic-reorder", 
     po::value<bool>(&dynamic_reordering)->default_value(0),
     "Dynamic reordering of the variables")
    ("dynamic-reorder-max-growth",
     po::value<float>(&dynamic_reordering_max_growth)->default_value(MAX_GROWTH_DEFAULT),
     "The maximum intermediate growth of the BDD size during dynamic reordering (1.0 = no growth, 2.0 = may double)")
    ("dynamic-reorder-timeout",
     po::value<unsigned int>(&dynamic_reordering_timeout)->default_value(0),
     "A timeout (in milliseconds) for unsuccessful sifting for each variable during reordering. 0 = no timeout")
    ("selective-reorder",
     po::value<unsigned int>(&selective_reorder_level)->default_value(0),
     "The level of selective reordering (0 = none, 1-3 increasingly aggressive)")
    ("unique-table-load-factor",
     po::value<double>(&unique_table_load_factor)->default_value(1.0),
     "The load factor for the unique tables where the hash table size is doubled")
    ("gcd-building",
     po::value<unsigned int>(&param_gcd_building)->default_value(1500),
     "BDD garbage collection delay (CA building phase)")
    ("gcd-checking",
     po::value<unsigned int>(&param_gcd_checking)->default_value(200000),
     "BDD garbage collection delay (checking phase)")
    ("read-varorder",
     po::value<string>(&read_varorder_file),
     "Read the variable order from a file")
    ("varorder-dir",
     po::value<string>(&varorder_dir)->default_value(""),
     "Specify a directory where variable orders for component types are stored and loaded on-demand")
    ("write-varorder",
     po::value<string>(&write_varorder_file),
     "Write the variable order after creation of the automaton to a file")
    ("constrain-to-reachable-components",
     po::value<unsigned int>(&param_constrain_reachable_components)->default_value(0),
     "Try to optimize transition relation by using BDD constrain on reachable fragment after components have been build. (0 = off, 1 = on)")
    ("constrain-to-reachable-circuits",
     po::value<unsigned int>(&param_constrain_reachable_circuits)->default_value(0),
     "Try to optimize transition relation by using BDD constrain on circuit level. (0 = off, 1 = use over-approximation of reachable, 2 = exact calculation of reachable)")
    ("optimize-ports",
     po::value<bool>(&param_optimize_ports)->default_value(1),
     "Optimize ports by using equivalence relations of identical or similar ports")
    ("no-shared-data",
     po::value<bool>(&param_no_shared_data)->default_value(0),
     "Disallow the sharing of data variables in ports");


  po::options_description opt_special("Special options");
  opt_special.add_options()
    ("read-varorder-old",
     po::value<string>(&read_varorder_file_old),
     "Read the variable order from a file (old format)")
    ("write-varorder-old",
     po::value<string>(&write_varorder_file_old),
     "Write the variable order after creation of the automaton to a file (old format)")
    ("dynamic-reorder-reverse",
     po::value<bool>(&dynamic_reordering_reverse)->default_value(false),
     "Start the dynamic reordering from the other side")
    ("check-result",
     po::value<bool>(&param_check_result)->default_value(false),
     "Check Model Checking result (if possible)")
    ("extract-explicit",
     po::value<string>(&extract_explicit_file),
     "Extract an explicit automaton to file");

  po::options_description opt_visible("Options");
  opt_visible.add(opt_general).add(opt_system).add(opt_bdd).add(opt_debug);
  
#ifdef HAVE_CHECKERS
  opt_visible.add(opt_checkers);
#endif
  
  po::options_description opt_all("All Options");
  opt_all.add(opt_visible).add(opt_special);

  po::variables_map vm;
  for (unsigned int i=0;i<cmd_line_parameters.size();i++) {
    po::store(po::command_line_parser(cmd_line_parameters[i]).options(opt_all).run(), vm);
  }
  po::notify(vm);    
  
  if (vm.count("help")) {
    cout << opt_visible;
    return false;
  }

  if (vm.count("load-builtins")) {
    param_load_builtins=true;
  } else {
    param_load_builtins=false;
  }

  if (vm.count("system-parameter")) {
    parameters=vm["system-parameter"].as< vector<string> >();
  }

  if (vm.count("input")) {
    param_input_files=vm["input"].as< vector<string> >();
  }
  
  if (vm.count("include-dir")) {
    include_dirs=vm["include-dir"].as< vector<string> >();
  }

  if (vm.count("constant")) {
    param_constant_defs=vm["constant"].as< vector<string> >();
  }

  if (vm.count("replace")) {
    vector<string> replacements=vm["replace"].as< vector<string> >();

    for (unsigned int i=0;i<replacements.size();i++) {
      string replacement = replacements.at(i);
      size_t arrow_pos = replacement.find("->");
      if (arrow_pos == string::npos) {
        throw "Illegal replacement, missing ->:\n"+replacement;
      }
      
      string original=replacement.substr(0, arrow_pos);
      string replace_with=replacement.substr(arrow_pos+2);

      boost::trim(original);
      boost::trim(replace_with);

      system_context->getReplacementManager()->addReplacement(original, replace_with);
    }    
  }

  if (vm.count("flag")) {
    flags=vm["flag"].as< vector<string> >();

    for (unsigned int i=0;i<flags.size();i++) {
      if (flags[i].size()==0) {
        throw std::string("Empty flag used as parameter");
      }

      unsigned char enabled_char=flags[i].at(0);
      bool enabled=false;
      if (enabled_char == '+') {enabled=true;}
      else if (enabled_char == '-') {enabled=false;}
      else {
        throw std::string("Flag parameter either has to start with + or -");
      }

      std::string flag_name=flags[i].substr(1);
      if (flag_name=="") {
        throw std::string("Flag name is empty!");
      }

      // TODO: check that flag_name is a valid identifier

      system_context->setFlag(flag_name, enabled);
    }    
  }

  if (!vm.count("random-trace")) {
    random_trace_steps=-1;
  }

  if (!vm.count("random-trace-iterative")) {
    random_trace_steps_iterative=-1;
  }
    
  if (vm.count("trace")) {
    trace_points=vm["trace"].as< vector<string> >();
    
    tracingEnabled=true;
    for (unsigned int i=0;i<trace_points.size();i++) {
      Tracing::activateTracePoint(trace_points.at(i));
      cerr << "Activate trace point : " << trace_points.at(i) << endl;
    }
  }

  if (vm.count("random-seed")) {
    cerr << "Setting random seed to " << vm["random-seed"].as<unsigned int>() << std::endl;
    srand(vm["random-seed"].as<unsigned int>());
  }

  if (dynamic_reordering) {
    dynamic_reordering_during_build=true;
    dynamic_reordering_during_modelchecking=true;
  }

  BDD::setUniqueTableMaxLoadFactor(unique_table_load_factor);

  if (read_varorder_file_old!="") {
    ifstream i_varorder(read_varorder_file_old.c_str());
    if (!i_varorder.good()) {
      throw string("Can't read ")+read_varorder_file_old;
    }
    VarManager::getInstance()->readVarOrder(i_varorder);
    
    // disable dynamic reordering during build if we have varorder
    dynamic_reordering_during_build=false;
  }

  if (selective_reorder_level>0) {
    VarManager::getInstance()->enableSelectiveReordering(selective_reorder_level);
  }

  if (vm.count("hide")) {
    SymbAutomaton::setDefaultHideType(vm["hide"].as<string>());
  }
  
  if (vm.count("minimize-dfa")) {
    formula::setDFAMinimizationMethod(vm["minimize-dfa"].as<string>());
  }

  if (param_automatic_includes) {
    for (unsigned int i=0;i<param_input_files.size();i++) {
      if (FileOps::file_exists(param_input_files.at(i))) {
        // this file is specified directly, add its directory to the include dirs
        ReturnValue<std::string> dirname=FileOps::get_dirname(param_input_files.at(i));
        if (dirname.valid) {
          include_dirs.push_back(dirname);
          if (debugLevel>=2) {
            std::cout << "Adding automatic include directory: "<< dirname.value << std::endl;
          }
        }
      }
    }
  }

  
  if (include_dirs.size()>0) {
    system_context->addIncludeDirs(include_dirs);
  }


  // process parameters for the system
  vparser::REOParser reo_parser(system_context);

  if (param_constant_defs.size()>0) {
    for (unsigned int i=0;i<param_constant_defs.size();i++) {
      reo_parser.parse_and_add_constdef("CONST "+param_constant_defs.at(i)+";");
    }
  }
 
  if (data_domain!="") {
    reo_parser.parse_and_add_typedef("TYPE Data = "+data_domain+";");
  }
  

  if (check_equivalence_with!="") {
    bisim_check_mode=true;
  } else if (starts_with(formula, "BISIM<<") && ends_with(formula, ">>")) {
    bisim_check_mode=true;

    check_equivalence_with=std::string(formula.begin() + 7,
                                       formula.end() - 2);
  }

  if (bisim_check_mode) {
    if (!strict) {
      std::cout << "Note: Bisimulation checking requires --strict=1, automatically enabled\n";
      strict=true;
    }

    if (param_retain_hidden_ports) {
      std::cout << "Note: Bisimulation checking requires --retain-hidden-ports=0, automatically disabled.\n";
      param_retain_hidden_ports=false;
    }

    if (param_retain_top_level_ports) {
      std::cout << "Note: Bisimulation checking requires --retain-top-level-ports=0, automatically disabled.\n";
      param_retain_top_level_ports=false;
    }
  }

  if ((SymbAutomaton::getDefaultHideType()==SymbAutomaton::HIDE_TRANSITIVE ||
       SymbAutomaton::getDefaultHideType()==SymbAutomaton::HIDE_TRANSITIVE_FULL) &&
      (param_retain_hidden_ports || param_retain_top_level_ports)) {
    if (param_retain_hidden_ports) {
      throw "Parameters --retain-hidden-ports=1 and transitive hiding not supported at the same time, use --hide=epsilon.";
    }
    if (param_retain_top_level_ports) {
      throw "Parameters --retain-top-level-ports=1 and transitive hiding not supported at the same time, use --hide=epsilon.";
    }
  }

  if (SymbAutomaton::getDefaultHideType()!=SymbAutomaton::HIDE_EPSILON &&
      random_trace_steps_iterative>0) {
    std::cout << "Note: --random-trace-iterative requires epsilon hiding, automatically enabled.\n";
    SymbAutomaton::setDefaultHideType(SymbAutomaton::HIDE_EPSILON);
  }

  system_context->setIgnoreMultipleIncludes(param_ignore_multiple_includes);
  
  return true;
}

void VereofyContext::enableDynamicReordering() {
    VarManager::getInstance()->enableDynamicReordering();
    VarManager::getInstance()->setDynamicReorderingMaxGrowth(dynamic_reordering_max_growth);
    VarManager::getInstance()->setDynamicReorderingTimeout(dynamic_reordering_timeout);
    
#ifdef JINC_NEW
    BDD::dynamicReorderingReverse=dynamic_reordering_reverse;
#endif
}

void VereofyContext::disableDynamicReordering() {
    VarManager::getInstance()->disableDynamicReordering();
}

ModelCheckingResult::ptr VereofyContext::bisimulation_check(const std::string& other_system) {
  SymbComponentInstance::ptr first_automaton;
  ComponentInstance::ptr first_system;
  first_automaton = automaton;
  automaton.reset();
  first_system = system;
  system.reset();
  parameters.clear();
  
  instantiate_system(other_system);
  build_system();

  build_phase_stopwatch.stop();
  if (csv_file) {
    csv_file->addColumn("build_time", build_phase_stopwatch.getElapsedTimeString());
    csv_file->addColumn("delta_size", boost::lexical_cast<string>(automaton->getDelta().size() + first_automaton->getDelta().size()));
  }

  BDD::setGarbageCollectionDelay(param_gcd_checking);

  if (dynamic_reordering_during_modelchecking) {
    enableDynamicReordering();
  } else {
    disableDynamicReordering();
  }

  check_phase_stopwatch.start();
  Bisimulation bisim = Bisimulation(first_automaton, automaton);
  ModelCheckingResult::ptr result=bisim.compute();
  check_phase_stopwatch.stop();

  if (csv_file) {
    csv_file->addColumn("check_time", check_phase_stopwatch.getElapsedTimeString());
    csv_file->write();
  }

  return result;
}

vereofy::SymbComponentInstance::ptr VereofyContext::getSystemAutomaton() {
  return automaton;
}


unsigned int VereofyContext::random_trace(unsigned int steps, bool xml_format, std::ostream &out) {
  SymbTraceExplorer::ptr trace_explorer=automaton->getTraceExplorer();
  trace_explorer->randomTrace(steps);
  
  if (xml_format) {
    print_xml_vereofy_trace_open(out, true);
    xml_tag_open(out,"traces");
    xml_tag_open(out,"trace", xml_parameter("type", "random"));
    trace_explorer->getCurrentTrace().print_xml(*automaton, out);
    xml_tag_close(out,"trace");
    xml_tag_close(out,"traces");
    print_xml_vereofy_trace_close(out);
  } else {
    automaton->print(out, trace_explorer->getCurrentTrace());
  }

  return trace_explorer->getCurrentTrace().countStates();
}

int VereofyContext::random_trace_iterative(unsigned int steps, bool xml_format, std::ostream& out) {
  RestrictionManager::ptr restrictions(new RestrictionManager());
  restrictions->restrict_delta_x_with_initial=true;
  restrictions->restrict_delta_x=false;
  restrictions->restrict_delta_y=false;

  instantiate_system(system_name);

  if (xml_format) {
    print_xml_vereofy_trace_open(out, true);
    xml_tag_open(out,"traces");
    xml_tag_open(out,"trace", xml_parameter("type", "random"));
    xml_tag_open(out, "tracefragment", xml_parameter("type", ""));
  }

  for (unsigned int i=0;i<steps;i++) {
    automaton.reset();
    VarManager::getInstance()->reset();

    std::cout << "\n=======================\nCalculating step " << i << "\n=======================\n"<< endl;

    build_system(restrictions);
    
    if (i>0) {
      // Relax initial states
      automaton->setInitialStates(restrictions->getStateRestrictions(*automaton, true));
    }

    SymbTraceExplorer::ptr trace_explorer=automaton->getTraceExplorer();
    trace_explorer->randomTrace(2);
    const SymbTrace& step=trace_explorer->getCurrentTrace();

    if (i==0 && step.countStates()>=1) {
      // output initial state
      const SymbState& first_state=step.getState(0);
      if (xml_format) {
        automaton->print_xml(out, first_state);
      } else {
        automaton->print(out, first_state);
        out << "\n";
      }
    }
    
    if (step.countStates()<=1) {
      // terminal state
      break;
    }

    const SymbState& next_state=step.getState(1);
    const SymbTransitionLabel& label=step.getLabel(0);

    if (xml_format) {
      automaton->print_xml(out, label);
      automaton->print_xml(out, next_state);
    } else {
      automaton->print(out, label);
      out << "\n";
      automaton->print(out, next_state);
      out << "\n";
    }

    out.flush();

    RestrictionManager::ptr new_restrictions(new RestrictionManager());
    automaton->generateRestrictionsFromState(next_state, *new_restrictions);
    new_restrictions->restrict_delta_x_with_initial=false;
    new_restrictions->restrict_delta_x=true;
    new_restrictions->restrict_delta_y=false;

    restrictions=new_restrictions;
  }

  if (xml_format) {
    xml_tag_close(out, "tracefragment");
    xml_tag_close(out,"trace");
    xml_tag_close(out,"traces");
    print_xml_vereofy_trace_close(out);
  }

  return 0;
}


void VereofyContext::print_xml_vereofy_trace_open(std::ostream& out, bool output_interface) {
  xml_header(out);
  xml_tag_open(out, "vereofy_trace");
  out << "\n";

  if (output_interface) {
    // print interface
    system->getInterface()->print_xml(out);
    out << "\n";
  }
}

void VereofyContext::print_xml_vereofy_trace_close(std::ostream& out) {
  xml_tag_close(out, "vereofy_trace");
  out << "\n";
}

// handle varorder writing if --write-varorder parameters are set
void VereofyContext::varorder_write() {
  if (write_varorder_file_old!="") {
    ofstream o_varorder(write_varorder_file_old.c_str());
    if (!o_varorder.good()) {
      throw string("Can't write to ")+write_varorder_file_old;
    }
    VarManager::getInstance()->writeVarOrder(o_varorder);
  }
  
  if (write_varorder_file!="") {
    ofstream o_varorder(write_varorder_file.c_str());
    if (!o_varorder.good()) {
      throw string("Can't write to ")+write_varorder_file;
    }
    SerializedVarorder::writeVarorder(o_varorder, automaton);
  }
}

  
int VereofyContext::run(unsigned int argc, char**argv) {
  std::cout << "Vereofy Model Checker (" << VEREOFY_VERSION << "),  http://www.vereofy.de\n" << std::endl;
  std::cout << "Copyright (C) 2008-2012, Vereofy Group, TU Dresden" << std::endl;
  try {
    add_parameters(argc, argv);
    if (argc==1) {
      add_parameter("--help");
    }

    bool rv=apply_parameters();
    if (!rv) {
      // --help
      return 0;
    }

    build_phase_stopwatch.start();
    
    if (param_load_builtins) {
      system_context->loadBuiltins();
    }

    for (unsigned int i=0;i<param_input_files.size();i++) {
      parse_file(param_input_files.at(i));
    }


    if (stat_file!="") {
      csv_file.reset(new CSVFile(stat_file));
    }
    
    if (random_trace_steps_iterative>0) {
      int rv;
      if (trace_xml_file!="") {
        ofstream s(trace_xml_file.c_str());
        if (!s.good()) {
          throw "Couldn't open "+trace_xml_file;
        }
        rv=random_trace_iterative((unsigned int)random_trace_steps_iterative, true, s);
      } else {
        std::stringstream s;
        rv=random_trace_iterative((unsigned int)random_trace_steps_iterative, false, s);

        std::cout << "\n\nTrace:\n";
        std::cout << s.str() << "\n";
      }
      return rv;
    }

    instantiate_system(system_name);
    if (export_circuit!="" && system->getType()==COMPONENT_CIRCUIT) {
      ofstream export_circuit_stream(export_circuit.c_str());
      if (export_circuit_stream.good()) {
        ComponentInstanceCircuit::ptr cic=boost::static_pointer_cast<ComponentInstanceCircuit>(system);
        cic->export_circuit(export_circuit_stream);
        export_circuit_stream.close();
      } else {
        throw "Can't open file: "+export_circuit;
      }
      
      return 0;
    }
    
    if (param_print_interface) {
      cout << "\n=====================================================\n";
      cout << "Interface: \n";

      print_interface(cout);

      cout << "\n=====================================================\n";
    }


    if (param_parse_only) {
      cout << "Parse successful" << std::endl;
      return 0;
    }

    build_system();
    if (!bisim_check_mode) {
      build_phase_stopwatch.stop();
      if (csv_file) {
        csv_file->addColumn("build_time", build_phase_stopwatch.getElapsedTimeString());
        csv_file->addColumn("delta_size", boost::lexical_cast<string>(automaton->getDelta().size()));
        csv_file->write();
      }
    }

    varorder_write();

    if (random_trace_steps>0) {
      if (trace_xml_file!="") {
        ofstream s(trace_xml_file.c_str());
        if (!s.good()) {
          throw "Couldn't open "+trace_xml_file;
        }
        unsigned int states=random_trace(random_trace_steps, true, s);

        std::cout << "Trace written to " << trace_xml_file << ", length: " << 
          states << " states\n";
      } else {
        cout << "Trace:\n";
        random_trace(random_trace_steps, false, std::cout);
      }

      return 0;
    }

    if (print_trace_to!="") {
      BDDFunction to_guard_set=automaton->parse_state_guard(print_trace_to);
    
      SymbBFS bfs(automaton->getDelta(),
                  automaton->getOriginalStateVars(),
                  automaton->getTransitionVars());
      ReturnValue<SymbTrace> trace=bfs.findTraceBetweenStateSets(automaton->getInitialStates(),
                                                                 to_guard_set);
    
      if (trace.valid) {
        cout << "Found trace to '"<< print_trace_to << "', length " << trace.value.countStates() << "\n\n";

        if (trace_xml_file!="") {
          ofstream s(trace_xml_file.c_str());
          if (!s.good()) {
            throw "Couldn't open "+trace_xml_file;
          }
          print_xml_vereofy_trace_open(s, true);
          xml_tag_open(s,"traces");
          xml_tag_open(s,"trace", xml_parameter("type", "example"));
          trace.value.print_xml(*automaton, s);
          xml_tag_close(s,"trace");
          xml_tag_close(s,"traces");
          print_xml_vereofy_trace_close(s);

          std::cout << "Trace written to " << trace_xml_file << ".\n";
        } else {
          cout << "Trace:\n";
          automaton->print(cout, trace.value);
        }
        return 0;
      } else {
        cout << "Couldn't find trace to '"<< print_trace_to << "'\n\n";
        return 1;
      }
    }

    vereofy::ModelCheckingResult::ptr result;

    if (bisim_check_mode) {
      result=bisimulation_check(check_equivalence_with);

    } else {
      // formula check mode

      if (formula=="") {
        return 0;
      }
      
      cout << "Checking formula " << formula << endl;
      result=model_check(formula);
    }


    if (!result) {
      throw "No valid ModelCheckingResult!";
    }

    cout << "\n\n";
    if (bisim_check_mode) {
      cout << "Checking bisimulation equivalence:" << endl;
    } else {
      cout << "Formula: " << formula << endl;
    }
    if (trace_xml_file!="") {
      ofstream s(trace_xml_file.c_str());
      if (!s.good()) {
        throw "Couldn't open "+trace_xml_file;
      }
      result->print_pass_fail(std::cout);
      print_xml_vereofy_trace_open(s, true);
      result->print_xml(s);
      print_xml_vereofy_trace_close(s);

    } else {
      result->print(cout);
    }
    
    if (!bisim_check_mode && result->countTraces()>0 && result->getTrace(0)->getTrace()) {
      CHECK(5, automaton->isValidTrace(*result->getTrace(0)->getTrace()->merge()), "The top-level trace is not valid in the system automaton!");
    }

    if (csv_file) {
      csv_file->addColumn("result", boost::lexical_cast<std::string>(result->checkPassed()));
    }
  
    if (result->checkPassed()) {
      return 0;
    } else {
      return 1;
    }


  } catch (const char* err) {
    cerr << "Exception! " << err << endl;
  } catch (string& err) {
    cerr << "Exception! " << err << endl;
  } catch (REOParseException& rpe) {
    cerr << rpe.what() << endl;
  } catch (REOExecuteException& ree) {
    cerr << ree.what() << endl;
  } catch (std::exception& err) {
    cerr << "Exception! " << err.what() << endl;
  }
  return 2;
  
}


