/***************************************************************************
 *   Copyright (C) 2008-2009 by Joachim Klein                              *
 *   vereofy@tcs.inf.tu-dresden.de                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "vereofy_context_for_java.hpp"
#include "vereofy_context.hpp"
#include "vereofy/modelcheck/checkers/ModelCheckingResult.hpp"
#include "vereofy/system/symb_automaton/symb_component_instance.hpp"
#include "vereofy/system/symb_automaton/symb_trace.hpp"
#include "vereofy/system/symb_automaton/symb_trace_explorer.hpp"
#include "vereofy/common/version.hpp"
#include "vereofy/common/interruptible.hpp"

#include <sstream>

using namespace vereofy;

VereofyContextForJava::VereofyContextForJava() {
  context=VereofyContext::ptr(new VereofyContext());
  resetInterruptFlag();
}

VereofyContextForJava::~VereofyContextForJava() {
  modelchecking_results.clear();
  context.reset();
  resetInterruptFlag();
}

void VereofyContextForJava::new_parameter_section() {
  context->new_parameter_section();
}

void VereofyContextForJava::add_parameter(const std::string& parameter) {
  context->add_parameter(parameter);
}

void VereofyContextForJava::apply_parameters() {
  context->apply_parameters();
}

void VereofyContextForJava::parse(const std::string& content) {
  context->parse(content);
}

void VereofyContextForJava::parse_file(const std::string& file_name) {
  context->parse_file(file_name);
}

void VereofyContextForJava::instantiate_system(const std::string& name) {
  context->instantiate_system(name);
}

// has to have instantiated system
void VereofyContextForJava::export_to_reo_xml(const std::string& filename, const std::string& system_file) {
  std::ofstream out(filename.c_str());
  if (!out.good()) {
    throw "Can't open "+filename+" for writing!";
  }

  std::vector<std::string> input_files;
  input_files.push_back(system_file);
  context->export_as_reo_xml(out, input_files);
  out.close();
}


std::string VereofyContextForJava::print_interface() {
  std::stringstream ss;
  context->print_interface(ss);
  return ss.str();
}

std::string VereofyContextForJava::print_interface_xml() {
  std::stringstream ss;
  context->print_interface_xml(ss);
  return ss.str();
}

void VereofyContextForJava::print_circuit_dot(const std::string& filename) {
  ofstream dot_circuit_stream(filename.c_str());
  if (dot_circuit_stream.good()) {
    context->print_circuit_dot(dot_circuit_stream);
  } else {
    throw std::string("Can not open output file: ")+filename;
  }
}

void VereofyContextForJava::build_system() {
  context->build_system();
  context->varorder_write();
}

void VereofyContextForJava::print_automaton_dot(const std::string& filename) {
  ofstream dot_stream(filename.c_str());
  if (dot_stream.good()) {
    context->print_automaton_dot(dot_stream);
  } else {
    throw std::string("Can not open output file: ")+filename;
  }
}

std::string VereofyContextForJava::get_random_trace_xml(unsigned int steps) {
  stringstream s;
  context->random_trace(steps, true, s);

  return s.str();
}

unsigned int VereofyContextForJava::model_check(const std::string& formula) {
  modelchecking_results.push_back(context->model_check(formula));
  return modelchecking_results.size()-1;
}

unsigned int VereofyContextForJava::bisimulation_check(const std::string& system_1, const std::string& system_2) {
  instantiate_system(system_1);
  context->build_system();
  modelchecking_results.push_back(context->bisimulation_check(system_2));
  return modelchecking_results.size()-1;
}

bool VereofyContextForJava::model_checking_passed(unsigned int index) {
  return modelchecking_results.at(index)->checkPassed();
}

std::string VereofyContextForJava::model_checking_info(unsigned int index) {
  std::stringstream ss; 
  modelchecking_results.at(index)->print(ss);

  return ss.str();
}

ModelCheckingResult_for_java VereofyContextForJava::model_checking_result(unsigned int index) {
  ModelCheckingResult_for_java result;
  result.context=context;
  result.result=modelchecking_results.at(index);
  return result;
}

SymbTrace_for_java VereofyContextForJava::get_random_trace(unsigned int length) {
  SymbComponentInstance::ptr automaton=context->getSystemAutomaton();
  CHECK_SUCCESS(automaton);

  SymbTraceExplorer::ptr trace_explorer=automaton->getTraceExplorer();
  trace_explorer->randomTrace(length);
  
  // store trace
  random_trace=SymbTrace::ptr(new SymbTrace(trace_explorer->getCurrentTrace()));
  
  SymbTrace_for_java trace;
  trace.context=context;
  trace.trace=random_trace;
  return trace;
}


std::string VereofyContextForJava::get_statistics() {
  std::stringstream result;

  result << "BDD size: " << boost::lexical_cast<string>(BDD::size()) << "\n";

  if (context->getSystemAutomaton()) {
    result << "BDD size of Delta = " << context->getSystemAutomaton()->getDelta().size() << "\n";
    result << "BDD size of Initial States = "<< context->getSystemAutomaton()->getInitialStates().size() << "\n";

    result << "\n\nReachable states = " << context->getSystemAutomaton()->getNumberOfReachableStates() << "\n";
  }

  return result.str();
}

std::string VereofyContextForJava::get_version() {
  return VEREOFY_VERSION;
}


void VereofyContextForJava::asyncRaiseInterruptFlag() {
  raiseInterruptFlag();
}
