#! /usr/bin/perl -w

use strict;
use Cwd 'abs_path';

my $clean=0;
my $run_tests=0;
my $processors;
my $cmake_args='';
my $has_cxx_flags=0;
my $with_coverage;

my $build_dir = "build";

while (@ARGV && $ARGV[0]=~/^--(.+)$/) {
  if ($1 eq 'clean') {
    $clean=1;
    shift;
    next;
  }

  if ($1=~/^build-dir=(.+)$/) {
    $build_dir=$1;
    shift;
    next;
  }

  if ($1 eq 'run-tests') {
    $run_tests=1;
    shift;
    next;
  }

  if ($1 =~ /^with-coverage=(.+)$/) {
    if ($1 eq 'lcov' ||
        $1 eq 'gcovr') {
      $with_coverage = $1;
    } else {
      die "Unknown coverage option '$1'\n Supported: 'lcov' and 'gcovr'";
    }
    shift;
    next;
  }

  if ($1=~/^procs=(\d+)$/) {
    $processors=$1;
    shift;
    next;
  }

  if ($1=~/^cmake-arg=(.+)$/) {
    my $arg=$1;
    if ($arg=~/-DCMAKE_CXX_FLAGS/) {
      $has_cxx_flags=1;
    }
    $cmake_args.=' '."'".$arg."'";
    shift;
    next;
  }

  # unknown
  print STDERR "Unknown parameter $1 \n";
  exit(1);
}

my $work_dir=shift || `pwd`;
chomp($work_dir);

chdir($work_dir) or exit(1);
if (-e "$build_dir") {
  if (-d "$build_dir") {
    # build directory exists
    if ($clean) {
      execute("rm -rf $build_dir");
      execute("mkdir $build_dir");
    }
  } else {
    print STDERR "$build_dir is not a directory!";
    exit(1);
  }
} else {
  # build does not yet exist
  execute("mkdir $build_dir");
}

chdir("$build_dir") or exit(1);

if (!$has_cxx_flags) {
  $cmake_args.=" -DCMAKE_CXX_FLAGS:STRING='-Wall'";
}

if ($with_coverage) {
  $cmake_args.=" -DCMAKE_BUILD_TYPE:STRING=Debug";
  $cmake_args.=" -DCMAKE_CXX_FLAGS_DEBUG:STRING='-Wall -O0 -g -fprofile-arcs -ftest-coverage'";
  $cmake_args.=" -DCMAKE_EXE_LINKER_FLAGS_DEBUG:STRING='-g -fprofile-arcs -ftest-coverage'";
}

execute("cmake $cmake_args $work_dir/src");
execute('make'.(defined $processors ? " -j$processors" : ''));

if ($run_tests || $with_coverage) {
  if ($with_coverage eq 'lcov' && !$clean) {
    execute('lcov -z --directory .');
  }
  
  print "Running tests\n";
  execute('../tools/run-tests.sh'.(defined $processors ? " --procs=$processors" : ''));

  if ($with_coverage eq 'gcovr') {
    my $path = abs_path('..');
    execute("gcovr -r '$path' --xml > vereofy.gcovr.xml");
  } elsif ($with_coverage eq 'lcov') {
    execute('lcov --directory . --capture --output-file vereofy.lcov.info');
    my $path = abs_path('..');
    my $src_dir=$path.'/src/vereofy/*';
    execute("lcov -e vereofy.lcov.info '$src_dir' --output-file vereofy.lcov-filtered.info");
    execute('genhtml --legend --show-details -o lcov_html vereofy.lcov-filtered.info');
  }
}

sub execute {
  my $command_line=shift;
  
  print $command_line, "\n";
  system($command_line);
  my $rv=$?;
  if ($rv & 127) {
    print "Died: ".($rv&127);
    exit(1);
  }

  $rv=$rv>>8;
  if ($rv!=0) {
    print "Failed, rv = $rv\n";
    exit($rv);
  }

  return;
}
