#! /usr/bin/perl -w

use strict;

use Text::Balanced qw (extract_delimited extract_bracketed extract_quotelike);
use Dumpvalue;

my $html = 0;
my $latex = 1;

my $dumper = new Dumpvalue;

sub eat_ws($) {
  my $text = shift;
  $text=~s/^\s*//;
  return $text;
}

sub make_tree {
  my $original_text = shift;
  my $text = $original_text;

  my @result;

  while (1) {
    $text = eat_ws($text);
    if ($text eq "") {
      return \@result;
    }

    if ($text=~/^([a-zA-Z_:.~]+)(.*)$/) {
      my $identifier = $1;
      my $remaining = eat_ws($2);

      my %node = (identifier => $identifier);
      while ($remaining =~ /^[(\[]/) {
        (my $extracted, $remaining)=extract_bracketed($remaining, "(['\"])");

        my $in_brackets=substr($extracted, 1, -1);
        $in_brackets =~ s/^\s+//;
        $in_brackets =~ s/\s+$//;
        
        $node{children}=[] if (!defined $node{children});
        push @{ $node{children} },  make_tree($in_brackets);
        $remaining=eat_ws($remaining);
      }

      push @result, \%node;

      $text = $remaining;
      next;
    } elsif ($text =~ /^\(/) {
      (my $extracted, my $remaining)=extract_bracketed($text, "(['\"])");

      die "$remaining\n$original_text" if ($extracted eq '');

      my $in_brackets=substr($extracted, 1, -1);
      $in_brackets =~ s/^\s+//;
      $in_brackets =~ s/\s+$//;
      my %node = ( subnode => make_tree($in_brackets) );
      push @result, \%node;

      $text=$remaining;
      next;
    } elsif ($text =~ /^['"]/) {
      (my $extracted, my $remaining)=extract_quotelike($text);
      
      my %node = ( quoted => $extracted );
      push @result, \%node;

      $text=$remaining;
      next;
    } else {
      my $ch = substr($text,0,1);
      $text=substr($text, 1);

      if ($ch eq '>' &&
          defined $result[-1] &&
          defined $result[-1]->{'ch'} &&
          $result[-1]->{'ch'} eq '>') {
        $result[-1]->{'ch'}='>>';
      } else {
        my %node = ( ch => $ch );
        
        push @result, \%node;
      }
      next;
    }
  }
}

sub split_rules {
  my $a = shift;

  my @rules;
  my @rule;
  foreach (@$a) {
    if (defined $_->{ch} &&
        $_->{ch} eq ';') {
      # end of rule
      push @rules, [@rule];
      @rule = ();
    } else {
      push @rule, $_;
    }
  }

  if (@rule) {
    push @rules, [@rule];
  }

  return \@rules;
}

sub delete_marked {
  my $a = shift;

  my $deleted;
  do {
    $deleted = 0;
    for (my $i=0;$i<@$a;$i++) {
      if (defined $a->[$i]->{delete}) {
        splice(@$a, $i, 1);
        $deleted++;
        last;
      }
    }
  } until ($deleted == 0);
}

sub process_rule {
  my $node = shift;
  my $mods = shift || {};

  foreach (@$node) {
    die if !(ref ($_) eq 'HASH');

    if (defined $_->{identifier} &&
        $_->{identifier} eq 'access_node_d') {
      # access_node_d[child1][child2]
      # replace by child1
      $_ = {subnode => process_rule($_->{children}->[0], $mods)}; 
    } elsif (defined $_->{identifier} &&
             ($_->{identifier} eq 'lexeme_d' ||
             $_->{identifier} eq 'keyword_d')) {
      # replace by child1
      $_ = {subnode => process_rule($_->{children}->[0], $mods)}; 
    } elsif (defined $_->{identifier} &&
        $_->{identifier} =~ /^expect\./ ) {
      # expect.xxx(...)
      # replace by child1
      $_ = {subnode => process_rule($_->{children}->[0], $mods)}; 
    } elsif (defined $_->{identifier} &&
        $_->{identifier} eq 'error_guard') {
      # error_guard[child1][child2]
      # replace by child1
      $_ = {subnode => process_rule($_->{children}->[0], $mods)}; 
    } elsif (defined $_->{identifier} &&
        $_->{identifier} eq 'list_p') {
      # list_p(rules, sep)
      $_ = {list => process_rule([$_->{children}->[0]->[0]], $mods),
            separator => process_rule([$_->{children}->[0]->[2]], $mods)}; 
    } elsif (defined $_->{identifier} &&
             ($_->{identifier} eq 'str_p' ||
              $_->{identifier} eq 'ch_p')) {
      die if ! (scalar @{ $_->{children} } == 1);
      die if ! (scalar @{ $_->{children}->[0] } == 1);
      $_->{str} = $_->{children}->[0]->[0]->{quoted};
      delete $_->{children};
      delete $_->{identifier};

      die if !defined $_->{str};
    } elsif (defined $_->{quoted}) {
      $_->{str} = $_->{quoted};
      delete $_->{quoted};
    } elsif (defined $_->{identifier} &&
             $_->{identifier} =~ /^~?eps_p/) {
      $_->{eps_p}=$_->{identifier};
      delete $_->{identifier};
    } elsif (defined $_->{identifier} &&
             ($_->{identifier} eq 'no_node_d' ||
              $_->{identifier} eq 'root_node_d' ||
              $_->{identifier} eq 'infix_node_d' ||
              $_->{identifier} eq 'leaf_node_d' ||
              $_->{identifier} eq 'as_lower_d' ||
              $_->{identifier} eq 'keyword_p' 
             )) {
      my %sub_mods = %{$mods};
      $sub_mods{$_->{identifier}}=1;

      # replace by processed subrule
      $_ = {subnode => process_rule($_->{children}->[0], \%sub_mods) };
    } elsif (defined $_->{subnode}) {
      $_->{subnode} = process_rule($_->{subnode}, $mods)
    } elsif (defined $_->{identifier} &&
             (!defined $_->{children} ||
              (scalar @{$_->{children}} == 1))) {
      # this is actually a rule
      $_->{rule} = $_->{identifier};
      %$_ = (%$_, %$mods);
      delete $_->{identifier};
    }
  }

  
  # remove eps_p
  for (my $i=0;$i<@$node;$i++) {
    if (defined $node->[$i]->{eps_p}) {
      $node->[$i]->{delete}=1;

      if (defined $node->[$i+1]->{ch} &&
          ($node->[$i+1]->{ch} eq '>>' ||
           $node->[$i+1]->{ch} eq '|')) {
        $node->[$i+1]->{delete}=1;
      }
    }
  }
  delete_marked($node);

#goto ende;
  # handle !, * and +
  my @result;
  my $cur;
  foreach (reverse @$node) {
    if (!defined $cur) {
      $cur = $_;
      next;
    }

    if (defined $_->{ch} &&
        $_->{ch} =~ /^[!*+]$/) {
      $cur = { 
              op => $_->{ch},
              subnode => [$cur]
             };
      next;
    }
    
    push @result, $cur;
    $cur = $_;
  }
  push @result, $cur;
  $node = [reverse @result];

#goto ende;
  # handle |
  @result=();
  my @tmp=();
  
  if (defined $node->[1]->{ch} &&
      $node->[1]->{ch} eq '=') {
    # rule head
    
    push @result, shift @$node;
    push @result, shift @$node;
  }

  my $choices;
  foreach (@$node) {
    if (defined $_->{ch} &&
        $_->{ch} eq '|') {

      if (!defined $choices) {
        $choices = {
                choices => [],
               };
      }

      if (@tmp == 1) {
        push @{ $choices->{choices} }, $tmp[0];
      } else {
        push @{ $choices->{choices} }, { subnode => [@tmp] };
      }
      @tmp = ();
    } else {
      push @tmp, $_;
    }

  }
  if (defined $choices) {
    if (@tmp == 1) {
      push @{ $choices->{choices} }, $tmp[0];
    } else {
      push @{ $choices->{choices} }, { subnode => [@tmp] };
    }
    push @result, $choices;
  } else {
    push @result, @tmp;
  }

  $node=[@result];

 ende:
  return $node;
}

sub flatten_subnodes {
  my $node = shift;

  foreach (@$node) {
    if (defined $_->{choices}) {
      $_->{choices} = flatten_subnodes($_->{choices});
    } elsif (defined $_->{subnode}) {
      $_->{subnode}=flatten_subnodes($_->{subnode});
      if (scalar @{$_->{subnode}} == 1 &&
          defined $_->{subnode}->[0]->{subnode}) {
        my $sub_node = $_->{subnode}->[0];
        
        if ((!defined $sub_node->{op}) ||
            (!defined $_->{op})) {
          # we can merge
          my $op = $sub_node->{op} || $_->{op};

          $_ = $sub_node;
          $_->{op} = $op if defined $op;
        }
      }

      if (scalar @{$_->{subnode}} == 1 &&
          !defined $_->{op}) {
        $_ = $_->{subnode}->[0];
      }
    }
  }

  return $node;
}


sub output_rule_html {
  my $rule = shift;

  print "<table border=1>\n <tr>\n";

  if (defined $rule->[1]->{ch} &&
      $rule->[1]->{ch} eq '=') {
    # rule head
    my $rule_name = $rule->[0]->{rule};
    
    print '<td><a name="', $rule_name, '">', $rule_name, '</a></td>', "\n";
    print '<td>:=</td>', "\n";
    
    shift @$rule;
    shift @$rule;
  }
  
  foreach (@$rule) {
    if (defined $_->{rule}) {
      print "  <td>\n";
      print '<span class="rule">',
        '<a href="#', $_->{rule}, '">',
        $_->{rule},
        '</a></span>', "\n";

      # TODO modifiers
      print "  </td>\n";
      next;
    } elsif (defined $_->{op}) {
      print "  <td>\n";
      print $_->{op};
      print " </td><td>";
      output_rule_html($_->{subnode});
      print "  </td>\n";
      next;
    } elsif (defined $_->{ch}) {
      my $ch = $_->{ch};
      if ($ch eq '>>') {
        next;
      }

      print "  <td>\n";
      print $ch;
      print "  </td>\n";
      next;
    } elsif (defined $_->{str}) {
      print '  <td><span class="literal">';
      print $_->{str};
      print "  </span></td>\n";
      next;
    } elsif (defined $_->{choices}) {
      print "  <td>\n";
      print "  <table>\n";
      my $first=1;
      foreach (@{$_->{choices}}) {
        print "<tr><td>\n";
        print " | " if (!$first);
        $first=0;
        print "</td>";
        print "<td>";
        output_rule_html([$_]);
        print "</td>";
        print "</tr>\n";
      }
      print "  </table>\n";
      print "  </td>\n";
      next;
    } elsif (defined $_->{subnode}) {
      print "  <td>\n";
        output_rule_html($_->{subnode});      
      print "  </td>\n";
      next;
    } elsif (defined $_->{list}) {
      print "  <td>\n";
      print "list";
      print "  </td>\n";
      next;
    }

    next if !(scalar keys %$_);
    $dumper->dumpValue($_);
    die;
  }
  print " </tr></table>\n";
}

sub html_header {
  print '<html><head>';
  print ' <style>';
#  print '  .literal {font: monospace bold; }';
  print ' </style>';
  print '</head><body>', "\n";
}

sub html_begin_file {
  my $file = shift;
  print '<h2>', $file, '</h2>', "\n";
}

sub html_footer {
  print '</body></html>', "\n";
}


sub latexify {
  my $text = shift;

  $text=~s/_/\\_/g;

  return $text;
}

sub latex_rule_label {
  my $rule_name = shift;
  
  if ($rule_name =~ /^(.+)_p$/) {
    $rule_name = $1;  # remove _p
  }

  $rule_name =~ s/_/-/g;
  return 'rule:'.$rule_name;
}

sub latex_header {
  print '\\documentclass[a4paper]{article}', "\n";
  print '\\usepackage[landscape]{geometry}', "\n";
  print '\\usepackage{syntax}', "\n";
  print '\\usepackage[colorlinks=true,linkcolor=blue]{hyperref}', "\n";
  print '\\begin{document}', "\n";
  print '\\pagestyle{headings}', "\n";
}

sub latex_footer {

  print '\newpage', "\n";
  print '\section{Builtin}', "\n";
  print '\begin{description}', "\n";
  print '\item[anychar] \hypertarget{rule:anychar}{Any character}', "\n";
  print '\item[alnum] \hypertarget{rule:alnum}{An alphanumeric character}', "\n";
  print '\item[alpha] \hypertarget{rule:alpha}{An alphabetic character}', "\n";
  print '\item[int] \hypertarget{rule:int}{An integer}', "\n";
  print '\end{description}', "\n";
  print '\\end{document}', "\n";
}

sub latex_begin_file {
  my $file = shift;

  print "\\newpage \\section{", latexify($file), "}\n";
}

sub output_rule_latex {
  my $rule = shift;

  if (defined $rule->[1]->{ch} &&
      $rule->[1]->{ch} eq '=') {
    # rule head
    my $rule_name = $rule->[0]->{rule};
    
    print '\\noindent ';
    print '\begin{minipage}{\textwidth}', "\n";
    print 
      '\\hypertarget{', latex_rule_label($rule_name), '}{',
      '\\synt{', latexify($rule_name), '} := ', "}\n"; 
#    print "Rule ", latexify($rule_name), ": \\\\";
    
    print "\\begin{syntdiag}\n";

    shift @$rule;
    shift @$rule;

    foreach (@$rule) {
      output_node_latex($_);
    }

    print "\\end{syntdiag}";
    print '\end{minipage}';
    print "\n\n\\vspace{1cm}";
    
  } else {
    die;
  }
}

sub output_nodes_latex {
  my $v = shift;

  if (ref($v) eq 'ARRAY') {
    foreach (@$v) {
      output_node_latex($_);
    }
  } else {
    output_node_latex($v);
  }
}

sub output_node_latex {
  my $n = shift;
  
  die "ref($n)" if ! (ref($n) eq 'HASH');
  
  if (defined $n->{rule}) {
    print '\synt{';
    print '\\hyperlink{', latex_rule_label($n->{rule}), '}{';
    print latexify($n->{rule});
    print '}} ';
    # TODO modifiers
  } elsif (defined $n->{op}) {
    if ($n->{op} eq '*') {
      print '\\begin{rep}', "\n";
      print '\\begin{stack} \\\\ ', "\n";
      output_nodes_latex($n->{subnode});
      print '\\end{stack}', "\n";
      print '\\end{rep}', "\n"; 
    } elsif ($n->{op} eq '+') {
      print '\\begin{rep}', "\n";
      output_nodes_latex($n->{subnode});
      print '\\end{rep}', "\n";
    } elsif ($n->{op} eq '!') {
      print '\\begin{stack} \\\\ ', "\n";
      output_nodes_latex($n->{subnode});
      print '\\end{stack}', "\n";
    } else {
      die "$n->{op}";
    }
  } elsif (defined $n->{ch}) {
    my $ch = $n->{ch};
    if ($ch eq '>>') {
      return;
    } elsif ($ch eq '-') {
      print ' \tok{-} ';
      return;
    }

    die "ch = $ch";
  } elsif (defined $n->{str}) {
    # strip quotes
    my $t=substr($n->{str}, 1, -1);
#    print STDERR $t, "\n";
    print "`\\verb|", $t, "|' ";
  } elsif (defined $n->{choices}) {
    print '\\begin{stack} ', "\n";
    
    my $first = 1;
    foreach (@{$n->{choices}}) {
      print '\\\\', "\n"  if !$first;
      $first = 0;
      output_node_latex($_);
    }
    print '\\end{stack} ', "\n";
  } elsif (defined $n->{subnode}) {
    # TODO: check
    output_nodes_latex($n->{subnode});      
  } elsif (defined $n->{list}) {
    print '\\begin{rep}', "\n";
    output_nodes_latex($n->{list});
    print '\\\\ ';
    output_nodes_latex($n->{separator});
    print '\\end{rep}';
  } else {
    return if !(scalar keys %$n);
    $dumper->dumpValue($_);
    die;
  }
}


sub output_rule {
  my $rule = shift;

  $dumper->dumpValue($rule);
}

html_header if $html;
latex_header if $latex;

foreach my $file (@ARGV) {
  my $text='';
  my $in_grammar_def = 0;
  
  open(IN, "< $file") or die;
  while (<IN>) {
    chomp;
    
    if ($in_grammar_def) {
      if (/End grammar definition/) {
        $in_grammar_def = 0;
        next;
      } else {
        s/\/\/(.+)$//;
        $text.=$_;
      }
    } else {
      if (/Start grammar definition/) {
        $in_grammar_def = 1;
      }
    }
  }

  next if ($text eq ''); # skip this file
  
  my $a = make_tree($text);
  my $rules = split_rules($a);

  if ($html) {
    html_begin_file($file);
  } elsif ($latex) {
    latex_begin_file($file);
  } else {
    print "\n\n========== ", $file, " ===================== \n\n";
  }
  
  my $i;
  foreach (@$rules) {
    $_=process_rule($_);
    $_=flatten_subnodes($_);
    
    if ($html) {
      print "\nRule ", $i++, "\n";
      output_rule_html($_);
    } elsif ($latex) {
      output_rule_latex($_);
    } else {
      print "\nRule ", $i++, "\n";
      output_rule($_);
    }
  }
}
  
html_footer if $html;
latex_footer if $latex;

