#! /usr/bin/perl -w

use strict;
use File::Spec;

my $path_to_script=(File::Spec->splitpath($0))[1];

my $all_types=1;

my $tap_output;
my $xml_output=0;
my $xml_output_prefix;

my $extra_args='';

my $negate_properties=0;
my $show_commandline=0;
my $continue_on_error=0;


my $export_vchecks=0;
my $vcheck_prefix='';

my $stat_file;
my $log_failures_file;

my %formula_types;

while (@ARGV) {
  if ($ARGV[0]=~/^--formulas=(.+)$/) {
    $formula_types{$1}=1;
    $all_types=0;
    shift;
  } elsif ($ARGV[0] eq '--tap') {
    $tap_output=1;
    shift;
  } elsif ($ARGV[0]=~/^--xml=(.+)$/) {
    $xml_output=1;
    $xml_output_prefix=$1;
    shift;
  } elsif ($ARGV[0] eq '--negate') {
    $negate_properties=1;
    shift;
  } elsif ($ARGV[0]=~/--extra-arg=(.+)$/) {
    $extra_args.=' '.$1;
    shift;
  } elsif ($ARGV[0]=~/--stats=(.+)$/) {
    $stat_file=$1;
    if (-f $stat_file) {
      unlink($stat_file);
    }
    shift;
  } elsif ($ARGV[0]=~/--log-failures$/) {
    $log_failures_file="failures.log";
    shift;
  } elsif ($ARGV[0]=~/^--show-commandline$/) {
    $show_commandline=1;
    shift;
  } elsif ($ARGV[0]=~/^--continue-on-error$/) {
    $continue_on_error=1;
    shift;
  } elsif ($ARGV[0] eq '--export-vchecks') {
    $export_vchecks=1;
    shift;
  } elsif ($ARGV[0]=~/^--vcheck-prefix=(.+)$/) {
    $vcheck_prefix=$1;
    shift;
  } else {
    last;
  }
}

my $vereofy_bin=shift or die;
my $directory=shift or die;

$|=1;



my @tasks;
my $property_count=0;

foreach my $filename (@ARGV) {
  if (!open(IN, "< $filename")) {
    print "Could not open $filename, quit...\n";
    exit(1);
  }

  my $text;
  while ($text=<IN>) {
    my $line=$text;
    chomp($text);

    next if ($text=~/^\s*#/); # comment

    while ($text=~/\\$/) {
      my $continuation=<IN>;
      if (!defined $continuation) {
        print STDERR "Continuation without next line!";
        exit 1;
      }

      chomp($continuation);
      $continuation=~s/^\s+/ /;
      $text=~s/\\$//;
      $text.=$continuation;
    }

    $text=~s/^(.*?)\/\/.+$/$1/;
    $text=~s/^\s*//g;
    $text=~s/\s*$//g;
    next if ($text=~/^$/);


    
    if ($text=~/^System\s*=\s*([^ ]+) ([^ ]+)/) {
      push @tasks, {
                    type => 'system',
                    filename => $1,
                    systemname => $2
                   };
      next;
    }
    
    if ($text=~/^Data\s*=\s*(.+)$/) {
      push @tasks, {
                    type => 'datadomain',
                    datadomain => $1
                   };
      next;
    }

    if ($text=~/^ID\s*=\s*([a-zA-Z0-9_\-]+)$/) {
      push @tasks, {
                    type => 'id',
                    id => $1
                    };
      next;
    }
    
    if ($text=~/^Property\(([01])(!?)\)\s*=\s*([A-Z]+)\s*(<<.+>>)$/) {
      my $property_true=$1;
      my $property_negatable=$2;
      my $property_type=$3;
      my $property=$property_type.$4;
      
      push @tasks, {
                    type => 'property',
                    property => $property,
                    property_type => $property_type,
                    property_true => $property_true
                   };
      
      if ($negate_properties) {
        if ($property_negatable ||
            is_always_negatable($property_type)) {

          my $negated_property=negate_property($property);
          if (defined $negated_property) {
            push @tasks, {
                          type=>'property',
                          property => $negated_property,
                          property_type => $property_type,
                          property_true => !$property_true
                         };
            $property_count++;
          }
        }
      }

      $property_count++;
      next;
    }
    
    if ($text=~/^Commandline\s*=\s*(.*)$/) {
      my $args=$1;

      push @tasks, {
                    type => 'commandline',
                    commandline => $args
                    };
      next;
    }
    

    if ($tap_output) {
      print "not ok\n";
    }
    print "# Don't know what to do with the following line:\n";
    print "# ", $line, "\n";
  }
}

if ($tap_output) {
  print "1..", $property_count, "\n";
}

my $failed=execute(\@tasks);
if ($tap_output && $failed>0) {
  print "# FAILED:  ", $failed, "\n";
}
exit($failed);



sub tap_ok {
  my $property = shift;
  my $count = shift;
  my $passed = shift;

  print "ok ", $count, " ", $property, " ", $passed, "\n";
}

sub tap_skip {
  my $property = shift;
  my $count = shift;
  
  print "ok ", $count, " # skip ", $property, "\n";
}


sub tap_error {
  my $property = shift;
  my $count = shift;
  my $reason = shift;
  my $output = shift;
  
  print "not ok ", $count, " ", $property, " ", $reason, "\n";
  if (defined $output) {
    print "# Output was:\n";
    foreach (@$output) {
      print "# ", $_;
    }
  }
}

sub execute {
  my $tasks=shift;

  my $have_system=0;
  my $system_name;
  my $data_domain;
  my $system_file;

  my $cmdline_args='';

  my $current_id='property';
  my $id_counter=0;

  my $i=0;
  my $failed=0;

  foreach (@$tasks) {
    ## check for pause
    if (-f "pause.$$") {
      print STDERR "==== PAUSE at ", scalar localtime(), " =====\n";
      print STDERR "Waiting for file pause.$$ to disappear...\n";
      while (-f "pause.$$") {
        sleep 2;
      }
      print STDERR "==== CONTINUE at ", scalar localtime(), " ====\n";
    }

    my $type=$_->{type};
    if ($type eq 'system') {
      finish_xml_output() if ($xml_output && $have_system);

      $have_system=1;
      $system_file=$_->{filename};
      $system_name=$_->{systemname};
      
      print '# ' if ($tap_output);
      print "System = ", $_->{filename}, ", $system_name\n";

      start_xml_output($system_name) if ($xml_output);
    } elsif ($type eq 'datadomain') {
      $data_domain=$_->{datadomain};

      print '# ' if ($tap_output);
      print "Data = $data_domain\n";

    } elsif ($type eq 'commandline') {
      $cmdline_args=$_->{commandline};
      print '# ' if ($tap_output);
      print "Commandline arguments = $cmdline_args\n";

    } elsif ($type eq 'id') {
      $current_id=$_->{id};
      $id_counter=0;
    } elsif ($type eq 'property') {
      $i++;
      die "Need system!" if !$have_system;

      my $property_true=$_->{property_true};
      my $property_type=$_->{property_type};
      my $property=$_->{property};

      
      if ($all_types ||
          $formula_types{$property_type}) {
        # ok
      } else {
        xml_test_case($property, "skipped") if ($xml_output);	
        if ($tap_output) {
          tap_skip($property, $i);
        } else {
          print $property, "\t[SKIPPED]\n";
        }
        next;
      }

      my $data_domain_arg=(defined $data_domain ? "--data-domain='$data_domain'" : '');
      
      my $stat_arg='';
      if (defined $stat_file) {
        $stat_arg='--stats=stat.csv';
      }
      
      my $input_arg='';
      if ($system_file eq 'builtin') {
        $input_arg='--load-builtins';
      } else {
        $input_arg="'--input=$system_file'";
      }

      $id_counter++;

      if ($export_vchecks) {
        export_vcheck($current_id, $id_counter, $system_file, $system_name, $data_domain, $property, $extra_args." ".$cmdline_args);

        next;
      }

      my $cmd_line="$vereofy_bin -I $directory $input_arg --system='$system_name' $data_domain_arg --formula='$property' $extra_args $cmdline_args $stat_arg 2>&1";

      if ($show_commandline) {
        print "# Command: ", $cmd_line, "\n";
      }      

      print $property, "\t" if (!$tap_output);

      my @result=`$cmd_line`;
      unshift(@result, "$cmd_line\n");
      #print @result;
      my $rv=$?;
      if ($rv & 127) {
	$failed++;
        my $error = "vereofy died: ".($rv&127);
        xml_test_case($property, "fail") if ($xml_output);
        if (defined $log_failures_file) {
          open(FAILURE, ">> $log_failures_file");
          print FAILURE $cmd_line, "\n";
          close(FAILURE);
        }
        if ($tap_output) {
          tap_error($property, $i, $error, \@result);
          next;
        } else {
          print STDERR "$error\n";
          print STDERR @result;
          if (!$continue_on_error) {
            exit(1);
          }
        }
      }
      $rv=$rv>>8;
      
      if (defined $stat_file) {
        system("perl $path_to_script/csv-stat-add.pl $stat_file stat.csv ".
               "'Key' '".$current_id.'-'.$id_counter."' ".
               "'System' '".csv_quote($system_name)."' ".
               "'Cmdline' '".csv_quote($cmdline_args)."'");
      }

      my $pass_or_fail;
      if ($rv==0) {
	$pass_or_fail="PASSED";
      } elsif ($rv==1) {
	$pass_or_fail="FAILED";
      } else {
	$failed++;
        my $error = "got error code ".$rv;

        xml_test_case($property, "fail") if ($xml_output);
        if (defined $log_failures_file) {
          open(FAILURE, ">> $log_failures_file");
          print FAILURE $cmd_line, "\n";
          close(FAILURE);
        }
        if ($tap_output) {
          tap_error($property, $i, $error, \@result);
          next;
        } else {
          print "[ERROR]\n";
          print STDERR $error;
          print STDERR @result;
          if (!$continue_on_error) {
            exit(1);
          }
        }
      }
      
      my $result_ok=($rv!=$property_true);
      $failed++ if !$result_ok;

      if ($xml_output) {
        xml_test_case($property, ($result_ok ? "pass" : "fail"));
      }

      if ($tap_output) {
        if ($result_ok) {
          tap_ok($property, $i, $pass_or_fail);
        } else {
          tap_error($property, $i, " Result = ".$pass_or_fail, \@result);
        }
      } else {
        if (!$result_ok) {
          print $pass_or_fail, "\t[ERROR]\n";
          print STDERR @result;
        } else {
          print $pass_or_fail, "\t[OK]\n";
        }
      }
    } else {
      die "Unknown task type".$type;
    }
  }

  finish_xml_output() if ($xml_output);

  return $failed;
}


sub clean_name {
  my $safe = shift;
  $safe =~ s/[^-:_A-Za-z0-9]+/_/gs;
  return $safe;
}

my $xml_text;
my $xml_test_count;
my $xml_test_fail_count;
my $xml_testsuite_name;

my $xml_file;

sub start_xml_output {
  my $system=shift;
  $system=clean_name($system);

  $xml_text="";
  $xml_test_count=0;
  $xml_test_fail_count=0;

  my $xml_file_name=$xml_output_prefix.'.'.$system.".xml";
  if (!open($xml_file, "> $xml_file_name")) {
    print "Can't open XML output file $xml_file_name\n";
    exit(1);
  }

  $xml_testsuite_name=(File::Spec->splitpath($xml_output_prefix))[2];
  $xml_testsuite_name=clean_name($xml_testsuite_name);
}

sub finish_xml_output {
  print $xml_file '<?xml version=\'1.0\' encoding=\'UTF-8\' ?>', "\n";
  print $xml_file '<testsuite name="',
    $xml_testsuite_name, '" tests="', $xml_test_count, '" failures="', $xml_test_fail_count, '">', "\n";
  print $xml_file $xml_text, "\n";
  print $xml_file '</testsuite>', "\n";

  close($xml_file);
}

sub xml_test_case {
  my $name=shift;
  my $result=shift;

  $xml_text.='<testcase name="'. clean_name($name). '">'. "\n";
  if ($result eq 'pass') {
    # ok
  } elsif ($result eq 'fail') {
    $xml_text.='<failure/>'."\n";
  } elsif ($result eq 'skipped') {
    $xml_text.='<skipped/>'."\n";
  } else {
    $xml_text.='<error/>'."\n";
  }
  $xml_text.='</testcase>'."\n";
}


sub is_always_negatable {
  my $property_type=shift;

  return ($property_type eq 'ASL');
}

sub negate_property {
  my $property=shift;

  if ($property=~/^([A-Z]+)\s*<<(.+)>>$/) {
    my $type=$1;
    my $body=$2;

    if ($type eq 'LTL') {
      return "LTL<< ! ( $body ) >>";
    } elsif ($type eq 'ASL') {
      return "ASL<< !( $body )>>";
    } else {
      print STDERR "Can't negate formula type $type, skipping\n";
      return undef;
    }
  } else {
    print STDERR "Unknown property type, skipping negation\n";
    return undef;
  }
}

sub xml_quote {
  my $text=shift;

  $text=~s/&/&amp;/g;
  $text=~s/"/&quot;/g;
  $text=~s/'/&apos;/g;
  $text=~s/</&lt;/g;
  $text=~s/>/&gt;/g;

  return $text;
}

sub csv_quote {
  my $text=shift;

  $text=~s/"//g;
  $text=~s/'//g;
  $text=~s/;//g;

  return $text;
}


sub export_vcheck {
  my $current_id=shift;
  my $id_counter=shift;
  my $system_file=shift;
  my $system_name=shift;
  my $data_domain=shift;
  my $property=shift;
  my $args=shift;
  
  my $filename=$current_id."-".$id_counter.".vcheck";
  
  my $formula_type;
  my $formula;

  if ($property=~/^(.+)<<(.+)>>$/) {
    $formula_type=$1;
    $formula=$2;
  } else {
    print STDERR "Malformed property: ", $property, "\n";
  }


  my $input_file=$system_file;

  $data_domain="" if !defined $data_domain;

  open(VCHECK, "> $filename") or die "Can not open ".$filename." for writing!";
  
  print "Writing $filename.\n";

  print VCHECK '<?xml version="1.0" encoding=\'ISO-8859-1\'?>', "\n";
  print VCHECK '<vereofy_check>', "\n";
  print VCHECK '<formula_type value="', xml_quote($formula_type), '" />', "\n";
  print VCHECK '<formula value="', xml_quote($formula), '" />', "\n";
  print VCHECK '<target_file value="', xml_quote($vcheck_prefix.$input_file), '" />', "\n";
  print VCHECK '<system_name value="', xml_quote($system_name), '" />', "\n";
  print VCHECK '<data_domain value="', xml_quote($data_domain), '" />', "\n";
  print VCHECK '<modelchecker_options>', "\n";
  
  $args=' '.$args;
  while ($args=~/^\s*(--?.+?)($|\s--?.*$)/) {
    my $arg=$1;
    $args=$2;
    print VCHECK '<option active="1" value="', xml_quote($arg), '" />', "\n";
  }

  if (!$args=~/^\s*$/) {
    print "Don't know what to do with args: ", $args, "\n";
  }

  print VCHECK '</modelchecker_options>', "\n";
  print VCHECK '</vereofy_check>', "\n";
}

